/*

 srs_init.c -- populating the SPATIAL_REF_SYS table

 version 2.4, 2010 April 5

 Author: Sandro Furieri a.furieri@lqt.it

 -----------------------------------------------------------------------------
 
 Version: MPL 1.1/GPL 2.0/LGPL 2.1
 
 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at
 http://www.mozilla.org/MPL/
 
Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the SpatiaLite library

The Initial Developer of the Original Code is Alessandro Furieri
 
Portions created by the Initial Developer are Copyright (C) 2008
the Initial Developer. All Rights Reserved.

Contributor(s):

Alternatively, the contents of this file may be used under the terms of
either the GNU General Public License Version 2 or later (the "GPL"), or
the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
in which case the provisions of the GPL or the LGPL are applicable instead
of those above. If you wish to allow use of your version of this file only
under the terms of either the GPL or the LGPL, and not to allow others to
use your version of this file under the terms of the MPL, indicate your
decision by deleting the provisions above and replace them with the notice
and other provisions required by the GPL or the LGPL. If you do not delete
the provisions above, a recipient may use your version of this file under
the terms of any one of the MPL, the GPL or the LGPL.
 
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#ifdef SPL_AMALGAMATION	/* spatialite-amalgamation */
#include <spatialite/sqlite3.h>
#else
#include <sqlite3.h>
#endif

#ifdef _WIN32
#define strcasecmp	_stricmp
#endif /* not WIN32 */

#include <spatialite.h>

struct epsg_defs
{
    int srid;
    char *auth_name;
    int auth_srid;
    char *ref_sys_name;
    char *proj4text;
    char *srs_wkt;
    struct epsg_defs *next;
};

static void
free_epsg_def (struct epsg_defs *ptr)
{
/* memory cleanup - destroying an EPSG def item */
    if (ptr->auth_name)
	free (ptr->auth_name);
    if (ptr->ref_sys_name)
	free (ptr->ref_sys_name);
    if (ptr->proj4text)
	free (ptr->proj4text);
    if (ptr->srs_wkt)
	free (ptr->srs_wkt);
    free (ptr);
}

static struct epsg_defs *
add_epsg_def (struct epsg_defs **first, struct epsg_defs **last, int srid,
	      const char *auth_name, int auth_srid, const char *ref_sys_name)
{
/* appending an EPSG def to the list */
    int len;
    struct epsg_defs *p = malloc (sizeof (struct epsg_defs));
    if (!p)
	return NULL;
    p->srid = srid;
    p->auth_name = NULL;
    p->auth_srid = auth_srid;
    p->ref_sys_name = NULL;
    p->proj4text = NULL;
    p->srs_wkt = NULL;
    p->next = NULL;
    if (auth_name)
      {
	  len = strlen (auth_name);
	  if (len > 0)
	    {
		p->auth_name = malloc (len + 1);
		if (p->auth_name == NULL)
		    goto error;
		strcpy (p->auth_name, auth_name);
	    }
      }
    if (ref_sys_name)
      {
	  len = strlen (ref_sys_name);
	  if (len > 0)
	    {
		p->ref_sys_name = malloc (len + 1);
		if (p->ref_sys_name == NULL)
		    goto error;
		strcpy (p->ref_sys_name, ref_sys_name);
	    }
      }
    if (*first == NULL)
	*first = p;
    if (*last != NULL)
	(*last)->next = p;
    *last = p;
    return p;
  error:
    free_epsg_def (p);
    return NULL;
}

static void
add_proj4text (struct epsg_defs *p, int count, const char *text)
{
/* creating the PROJ4TEXT string */
    int len;
    int olen;
    char *string;
    if (text == NULL)
	return;
    len = strlen (text);
    if (!count)
      {
	  p->proj4text = malloc (len + 1);
	  if (p->proj4text == NULL)
	      return;
	  strcpy (p->proj4text, text);
	  return;
      }
    if (p->proj4text == NULL)
	return;
    olen = strlen (p->proj4text);
    string = malloc (len + olen + 1);
    if (string == NULL)
	return;
    strcpy (string, p->proj4text);
    free (p->proj4text);
    p->proj4text = string;
    strcat (p->proj4text, text);
}

static void
add_srs_wkt (struct epsg_defs *p, int count, const char *text)
{
/* creating the SRS_WKT string */
    int len;
    int olen;
    char *string;
    if (text == NULL)
	return;
    len = strlen (text);
    if (!count)
      {
	  p->srs_wkt = malloc (len + 1);
	  if (p->srs_wkt == NULL)
	      return;
	  strcpy (p->srs_wkt, text);
	  return;
      }
    if (p->srs_wkt == NULL)
	return;
    olen = strlen (p->srs_wkt);
    string = malloc (len + olen + 1);
    if (string == NULL)
	return;
    strcpy (string, p->srs_wkt);
    free (p->srs_wkt);
    p->srs_wkt = string;
    strcat (p->srs_wkt, text);
}

static void
initialize_epsg_00 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 2000, "epsg", 2000,
		      "Anguilla 1957 / British West Indies Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-62 +k=0.9995000000000001 +x");
    add_proj4text (p, 1, "_0=400000 +y_0=0 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Anguilla 1957 / British West Indies Grid\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"Anguilla 1957\",DATUM[\"Anguilla_1957\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7012\"]],AUTHORITY[\"EPSG\",\"6600\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4600\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-62],PARAMETER[\"scale_factor\",0.9995],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",400000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"2000\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2001, "epsg", 2001,
		      "Antigua 1943 / British West Indies Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-62 +k=0.9995000000000001 +x");
    add_proj4text (p, 1, "_0=400000 +y_0=0 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Antigua 1943 / British West Indies Grid\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"Antigua 1943\",DATUM[\"Antigua_1943\",SPHEROID[\"Cla");
    add_srs_wkt (p, 2,
		 "rke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"7012\"]],AUTHORITY[\"EPSG\",\"6601\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"4601\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETE");
    add_srs_wkt (p, 8,
		 "R[\"latitude_of_origin\",0],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 9,
		 "\",-62],PARAMETER[\"scale_factor\",0.9995],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",400000],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 11,
		 "UTHORITY[\"EPSG\",\"2001\"],AXIS[\"Easting\",EAST],AXIS[");
    add_srs_wkt (p, 12, "\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2002, "epsg", 2002,
		      "Dominica 1945 / British West Indies Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-62 +k=0.9995000000000001 +x");
    add_proj4text (p, 1,
		   "_0=400000 +y_0=0 +ellps=clrk80 +towgs84=725,685,536,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Dominica 1945 / British West Indies Grid\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"Dominica 1945\",DATUM[\"Dominica_1945\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7012\"]],TOWGS84[725,685,536,0,0,0,0],AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"6602\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 5,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4602\"]],UN");
    add_srs_wkt (p, 7,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 9,
		 ",0],PARAMETER[\"central_meridian\",-62],PARAMETER[\"scal");
    add_srs_wkt (p, 10,
		 "e_factor\",0.9995],PARAMETER[\"false_easting\",400000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2002");
    add_srs_wkt (p, 12, "\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2003, "epsg", 2003,
		      "Grenada 1953 / British West Indies Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-62 +k=0.9995000000000001 +x");
    add_proj4text (p, 1,
		   "_0=400000 +y_0=0 +ellps=clrk80 +towgs84=72,213.7,93,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Grenada 1953 / British West Indies Grid\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"Grenada 1953\",DATUM[\"Grenada_1953\",SPHEROID[\"Cla");
    add_srs_wkt (p, 2,
		 "rke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"7012\"]],TOWGS84[72,213.7,93,0,0,0,0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6603\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4603\"]],UNIT[");
    add_srs_wkt (p, 7,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-62],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.9995],PARAMETER[\"false_easting\",400000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2003\"");
    add_srs_wkt (p, 12, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2004, "epsg", 2004,
		      "Montserrat 1958 / British West Indies Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-62 +k=0.9995000000000001 +x");
    add_proj4text (p, 1,
		   "_0=400000 +y_0=0 +ellps=clrk80 +towgs84=174,359,365,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Montserrat 1958 / British West Indies Grid\",GE");
    add_srs_wkt (p, 1,
		 "OGCS[\"Montserrat 1958\",DATUM[\"Montserrat_1958\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7012\"]],TOWGS84[174,359,365,0,0,0,0],AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"6604\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 5,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4604\"");
    add_srs_wkt (p, 7,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 9,
		 "igin\",0],PARAMETER[\"central_meridian\",-62],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"scale_factor\",0.9995],PARAMETER[\"false_easting\",400");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12,
		 "\"2004\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 2005, "epsg", 2005,
		      "St. Kitts 1955 / British West Indies Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-62 +k=0.9995000000000001 +x");
    add_proj4text (p, 1, "_0=400000 +y_0=0 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"St. Kitts 1955 / British West Indies Grid\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"St. Kitts 1955\",DATUM[\"St_Kitts_1955\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6605\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4605\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-62],PARAMETER[\"scale_factor\",0.9995],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",400000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"2005\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 12, "AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2006, "epsg", 2006,
		      "St. Lucia 1955 / British West Indies Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-62 +k=0.9995000000000001 +x");
    add_proj4text (p, 1,
		   "_0=400000 +y_0=0 +ellps=clrk80 +towgs84=-149,128,296,0,0");
    add_proj4text (p, 2, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"St. Lucia 1955 / British West Indies Grid\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"St. Lucia 1955\",DATUM[\"St_Lucia_1955\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7012\"]],TOWGS84[-149,128,296,0,0,0,0],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6606\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4606\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",-62],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",0.9995],PARAMETER[\"false_easting\",40000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12,
		 "2006\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 2007, "epsg", 2007,
		      "St. Vincent 45 / British West Indies Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-62 +k=0.9995000000000001 +x");
    add_proj4text (p, 1,
		   "_0=400000 +y_0=0 +ellps=clrk80 +towgs84=195.671,332.517,");
    add_proj4text (p, 2, "274.607,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"St. Vincent 45 / British West Indies Grid\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"St. Vincent 1945\",DATUM[\"St_Vincent_1945\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7012\"]],TOWGS84[195.671,332.517,274.607,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6607\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4607\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 8,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 9,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",-6");
    add_srs_wkt (p, 10,
		 "2],PARAMETER[\"scale_factor\",0.9995],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",400000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 12,
		 "ITY[\"EPSG\",\"2007\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 2008, "epsg", 2008,
		      "NAD27(CGQ77) / SCoPQ zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-55.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(CGQ77) / SCoPQ zone 2\",GEOGCS[\"NAD27(CG");
    add_srs_wkt (p, 1,
		 "Q77)\",DATUM[\"North_American_Datum_1927_CGQ77\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4609\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-55.5],PARAMETER[\"scale_factor\",0.9999],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",304800],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",0],AUTHORITY[\"EPSG\",\"2008\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2009, "epsg", 2009,
		      "NAD27(CGQ77) / SCoPQ zone 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-58.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(CGQ77) / SCoPQ zone 3\",GEOGCS[\"NAD27(CG");
    add_srs_wkt (p, 1,
		 "Q77)\",DATUM[\"North_American_Datum_1927_CGQ77\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4609\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-58.5],PARAMETER[\"scale_factor\",0.9999],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",304800],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",0],AUTHORITY[\"EPSG\",\"2009\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2010, "epsg", 2010,
		      "NAD27(CGQ77) / SCoPQ zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-61.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(CGQ77) / SCoPQ zone 4\",GEOGCS[\"NAD27(CG");
    add_srs_wkt (p, 1,
		 "Q77)\",DATUM[\"North_American_Datum_1927_CGQ77\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4609\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-61.5],PARAMETER[\"scale_factor\",0.9999],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",304800],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",0],AUTHORITY[\"EPSG\",\"2010\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2011, "epsg", 2011,
		      "NAD27(CGQ77) / SCoPQ zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-64.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(CGQ77) / SCoPQ zone 5\",GEOGCS[\"NAD27(CG");
    add_srs_wkt (p, 1,
		 "Q77)\",DATUM[\"North_American_Datum_1927_CGQ77\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4609\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-64.5],PARAMETER[\"scale_factor\",0.9999],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",304800],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",0],AUTHORITY[\"EPSG\",\"2011\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2012, "epsg", 2012,
		      "NAD27(CGQ77) / SCoPQ zone 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-67.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(CGQ77) / SCoPQ zone 6\",GEOGCS[\"NAD27(CG");
    add_srs_wkt (p, 1,
		 "Q77)\",DATUM[\"North_American_Datum_1927_CGQ77\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4609\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-67.5],PARAMETER[\"scale_factor\",0.9999],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",304800],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",0],AUTHORITY[\"EPSG\",\"2012\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2013, "epsg", 2013,
		      "NAD27(CGQ77) / SCoPQ zone 7");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-70.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(CGQ77) / SCoPQ zone 7\",GEOGCS[\"NAD27(CG");
    add_srs_wkt (p, 1,
		 "Q77)\",DATUM[\"North_American_Datum_1927_CGQ77\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4609\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-70.5],PARAMETER[\"scale_factor\",0.9999],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",304800],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",0],AUTHORITY[\"EPSG\",\"2013\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2014, "epsg", 2014,
		      "NAD27(CGQ77) / SCoPQ zone 8");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-73.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(CGQ77) / SCoPQ zone 8\",GEOGCS[\"NAD27(CG");
    add_srs_wkt (p, 1,
		 "Q77)\",DATUM[\"North_American_Datum_1927_CGQ77\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4609\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-73.5],PARAMETER[\"scale_factor\",0.9999],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",304800],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",0],AUTHORITY[\"EPSG\",\"2014\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2015, "epsg", 2015,
		      "NAD27(CGQ77) / SCoPQ zone 9");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-76.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(CGQ77) / SCoPQ zone 9\",GEOGCS[\"NAD27(CG");
    add_srs_wkt (p, 1,
		 "Q77)\",DATUM[\"North_American_Datum_1927_CGQ77\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4609\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-76.5],PARAMETER[\"scale_factor\",0.9999],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",304800],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",0],AUTHORITY[\"EPSG\",\"2015\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2016, "epsg", 2016,
		      "NAD27(CGQ77) / SCoPQ zone 10");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-79.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(CGQ77) / SCoPQ zone 10\",GEOGCS[\"NAD27(C");
    add_srs_wkt (p, 1,
		 "GQ77)\",DATUM[\"North_American_Datum_1927_CGQ77\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4609\"]],UNIT[\"metre\",1,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-79.5],PARAMETER[\"scale_factor\",0.9999],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",304800],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",0],AUTHORITY[\"EPSG\",\"2016\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2017, "epsg", 2017,
		      "NAD27(76) / MTM zone 8");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-73.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(76) / MTM zone 8\",GEOGCS[\"NAD27(76)\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1927_1976\",SPHEROID[\"Clark");
    add_srs_wkt (p, 2,
		 "e 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"4608\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETE");
    add_srs_wkt (p, 8,
		 "R[\"latitude_of_origin\",0],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 9,
		 "\",-73.5],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",304800],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"2017\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 2018, "epsg", 2018,
		      "NAD27(76) / MTM zone 9");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-76.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(76) / MTM zone 9\",GEOGCS[\"NAD27(76)\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1927_1976\",SPHEROID[\"Clark");
    add_srs_wkt (p, 2,
		 "e 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"4608\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETE");
    add_srs_wkt (p, 8,
		 "R[\"latitude_of_origin\",0],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 9,
		 "\",-76.5],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",304800],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"2018\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 2019, "epsg", 2019,
		      "NAD27(76) / MTM zone 10");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-79.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(76) / MTM zone 10\",GEOGCS[\"NAD27(76)\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1927_1976\",SPHEROID[\"Clar");
    add_srs_wkt (p, 2,
		 "ke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4608\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-79.5],PARAMETER[\"scale_factor\",0.9999],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",304800],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"2019\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 2020, "epsg", 2020,
		      "NAD27(76) / MTM zone 11");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-82.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(76) / MTM zone 11\",GEOGCS[\"NAD27(76)\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1927_1976\",SPHEROID[\"Clar");
    add_srs_wkt (p, 2,
		 "ke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4608\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-82.5],PARAMETER[\"scale_factor\",0.9999],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",304800],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"2020\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2021, "epsg", 2021,
		      "NAD27(76) / MTM zone 12");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-81 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(76) / MTM zone 12\",GEOGCS[\"NAD27(76)\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1927_1976\",SPHEROID[\"Clar");
    add_srs_wkt (p, 2,
		 "ke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4608\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-81],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",304800],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"2021\"],AXIS[\"Easting\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2022, "epsg", 2022,
		      "NAD27(76) / MTM zone 13");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-84 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(76) / MTM zone 13\",GEOGCS[\"NAD27(76)\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1927_1976\",SPHEROID[\"Clar");
    add_srs_wkt (p, 2,
		 "ke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4608\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-84],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",304800],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"2022\"],AXIS[\"Easting\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2023, "epsg", 2023,
		      "NAD27(76) / MTM zone 14");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-87 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(76) / MTM zone 14\",GEOGCS[\"NAD27(76)\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1927_1976\",SPHEROID[\"Clar");
    add_srs_wkt (p, 2,
		 "ke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4608\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-87],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",304800],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"2023\"],AXIS[\"Easting\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2024, "epsg", 2024,
		      "NAD27(76) / MTM zone 15");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-90 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(76) / MTM zone 15\",GEOGCS[\"NAD27(76)\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1927_1976\",SPHEROID[\"Clar");
    add_srs_wkt (p, 2,
		 "ke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4608\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-90],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",304800],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"2024\"],AXIS[\"Easting\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2025, "epsg", 2025,
		      "NAD27(76) / MTM zone 16");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-93 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(76) / MTM zone 16\",GEOGCS[\"NAD27(76)\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1927_1976\",SPHEROID[\"Clar");
    add_srs_wkt (p, 2,
		 "ke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4608\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-93],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",304800],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"2025\"],AXIS[\"Easting\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2026, "epsg", 2026,
		      "NAD27(76) / MTM zone 17");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-96 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(76) / MTM zone 17\",GEOGCS[\"NAD27(76)\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1927_1976\",SPHEROID[\"Clar");
    add_srs_wkt (p, 2,
		 "ke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4608\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-96],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",304800],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"2026\"],AXIS[\"Easting\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2027, "epsg", 2027,
		      "NAD27(76) / UTM zone 15N");
    add_proj4text (p, 0, "+proj=utm +zone=15 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(76) / UTM zone 15N\",GEOGCS[\"NAD27(76)\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1927_1976\",SPHEROID[\"Cla");
    add_srs_wkt (p, 2,
		 "rke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4608\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-93],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"2027\"],AXIS[\"Easting\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2028, "epsg", 2028,
		      "NAD27(76) / UTM zone 16N");
    add_proj4text (p, 0, "+proj=utm +zone=16 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(76) / UTM zone 16N\",GEOGCS[\"NAD27(76)\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1927_1976\",SPHEROID[\"Cla");
    add_srs_wkt (p, 2,
		 "rke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4608\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-87],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"2028\"],AXIS[\"Easting\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2029, "epsg", 2029,
		      "NAD27(76) / UTM zone 17N");
    add_proj4text (p, 0, "+proj=utm +zone=17 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(76) / UTM zone 17N\",GEOGCS[\"NAD27(76)\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1927_1976\",SPHEROID[\"Cla");
    add_srs_wkt (p, 2,
		 "rke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4608\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-81],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"2029\"],AXIS[\"Easting\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2030, "epsg", 2030,
		      "NAD27(76) / UTM zone 18N");
    add_proj4text (p, 0, "+proj=utm +zone=18 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(76) / UTM zone 18N\",GEOGCS[\"NAD27(76)\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1927_1976\",SPHEROID[\"Cla");
    add_srs_wkt (p, 2,
		 "rke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4608\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-75],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"2030\"],AXIS[\"Easting\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2031, "epsg", 2031,
		      "NAD27(CGQ77) / UTM zone 17N");
    add_proj4text (p, 0, "+proj=utm +zone=17 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(CGQ77) / UTM zone 17N\",GEOGCS[\"NAD27(CG");
    add_srs_wkt (p, 1,
		 "Q77)\",DATUM[\"North_American_Datum_1927_CGQ77\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4609\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-81],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 11,
		 "ing\",0],AUTHORITY[\"EPSG\",\"2031\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2032, "epsg", 2032,
		      "NAD27(CGQ77) / UTM zone 18N");
    add_proj4text (p, 0, "+proj=utm +zone=18 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(CGQ77) / UTM zone 18N\",GEOGCS[\"NAD27(CG");
    add_srs_wkt (p, 1,
		 "Q77)\",DATUM[\"North_American_Datum_1927_CGQ77\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4609\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-75],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 11,
		 "ing\",0],AUTHORITY[\"EPSG\",\"2032\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2033, "epsg", 2033,
		      "NAD27(CGQ77) / UTM zone 19N");
    add_proj4text (p, 0, "+proj=utm +zone=19 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(CGQ77) / UTM zone 19N\",GEOGCS[\"NAD27(CG");
    add_srs_wkt (p, 1,
		 "Q77)\",DATUM[\"North_American_Datum_1927_CGQ77\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4609\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-69],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 11,
		 "ing\",0],AUTHORITY[\"EPSG\",\"2033\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2034, "epsg", 2034,
		      "NAD27(CGQ77) / UTM zone 20N");
    add_proj4text (p, 0, "+proj=utm +zone=20 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(CGQ77) / UTM zone 20N\",GEOGCS[\"NAD27(CG");
    add_srs_wkt (p, 1,
		 "Q77)\",DATUM[\"North_American_Datum_1927_CGQ77\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4609\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-63],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 11,
		 "ing\",0],AUTHORITY[\"EPSG\",\"2034\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2035, "epsg", 2035,
		      "NAD27(CGQ77) / UTM zone 21N");
    add_proj4text (p, 0, "+proj=utm +zone=21 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(CGQ77) / UTM zone 21N\",GEOGCS[\"NAD27(CG");
    add_srs_wkt (p, 1,
		 "Q77)\",DATUM[\"North_American_Datum_1927_CGQ77\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4609\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-57],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 11,
		 "ing\",0],AUTHORITY[\"EPSG\",\"2035\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2036, "epsg", 2036,
		      "NAD83(CSRS98) / New Brunswick Stereo (deprecated)");
    add_proj4text (p, 0,
		   "+proj=sterea +lat_0=46.5 +lon_0=-66.5 +k=0.999912 +x_0=2");
    add_proj4text (p, 1,
		   "500000 +y_0=7500000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / New Brunswick Stereo (deprecate");
    add_srs_wkt (p, 1,
		 "d)\",GEOGCS[\"NAD83(CSRS98)\",DATUM[\"NAD83_Canadian_Spa");
    add_srs_wkt (p, 2,
		 "tial_Reference_System\",SPHEROID[\"GRS 1980\",6378137,29");
    add_srs_wkt (p, 3,
		 "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 5,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 6,
		 "74532925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"4140\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 8,
		 "9001\"]],PROJECTION[\"Oblique_Stereographic\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"latitude_of_origin\",46.5],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 10,
		 "an\",-66.5],PARAMETER[\"scale_factor\",0.999912],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_easting\",2500000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 12,
		 "\",7500000],AUTHORITY[\"EPSG\",\"2036\"],AXIS[\"Northing");
    add_srs_wkt (p, 13, "\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 2037, "epsg", 2037,
		      "NAD83(CSRS98) / UTM zone 19N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / UTM zone 19N (deprecated)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(CSRS98)\",DATUM[\"NAD83_Canadian_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4140\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 8,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 9,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",-69],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"2037\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 2038, "epsg", 2038,
		      "NAD83(CSRS98) / UTM zone 20N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / UTM zone 20N (deprecated)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(CSRS98)\",DATUM[\"NAD83_Canadian_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4140\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 8,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 9,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",-63],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"2038\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 2039, "epsg", 2039,
		      "Israel / Israeli TM Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31.73439361111111 +lon_0=35.204516944");
    add_proj4text (p, 1,
		   "44445 +k=1.0000067 +x_0=219529.584 +y_0=626907.39 +ellps");
    add_proj4text (p, 2, "=GRS80 +towgs84=-48,55,52,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Israel / Israeli TM Grid\",GEOGCS[\"Israel\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"Israel\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 2,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-48,55,52,0,0,0");
    add_srs_wkt (p, 3,
		 ",0],AUTHORITY[\"EPSG\",\"6141\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 5,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 6,
		 "G\",\"4141\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 7,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 8,
		 "titude_of_origin\",31.73439361111111],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",35.20451694444445],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",1.0000067],PARAMETER[\"false_easting\",219529.584],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",626907.39],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12,
		 ",\"2039\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORT");
    add_srs_wkt (p, 13, "H]]");
    p = add_epsg_def (first, last, 2040, "epsg", 2040,
		      "Locodjo 1965 / UTM zone 30N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=30 +ellps=clrk80 +towgs84=-125,53,467,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Locodjo 1965 / UTM zone 30N\",GEOGCS[\"Locodjo ");
    add_srs_wkt (p, 1,
		 "1965\",DATUM[\"Locodjo_1965\",SPHEROID[\"Clarke 1880 (RG");
    add_srs_wkt (p, 2,
		 "S)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[-125,53,467,0,0,0,0],AUTHORITY[\"EPSG\",\"6142\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4142\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-3],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 10,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"2040\"],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2041, "epsg", 2041,
		      "Abidjan 1987 / UTM zone 30N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=30 +ellps=clrk80 +towgs84=-124.76,53,466");
    add_proj4text (p, 1, ".79,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Abidjan 1987 / UTM zone 30N\",GEOGCS[\"Abidjan ");
    add_srs_wkt (p, 1,
		 "1987\",DATUM[\"Abidjan_1987\",SPHEROID[\"Clarke 1880 (RG");
    add_srs_wkt (p, 2,
		 "S)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[-124.76,53,466.79,0,0,0,0],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 4,
		 "143\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 5,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 6,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4143\"]],UNIT[\"metr");
    add_srs_wkt (p, 7,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transve");
    add_srs_wkt (p, 8,
		 "rse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"central_meridian\",-3],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"2041\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2042, "epsg", 2042,
		      "Locodjo 1965 / UTM zone 29N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=29 +ellps=clrk80 +towgs84=-125,53,467,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Locodjo 1965 / UTM zone 29N\",GEOGCS[\"Locodjo ");
    add_srs_wkt (p, 1,
		 "1965\",DATUM[\"Locodjo_1965\",SPHEROID[\"Clarke 1880 (RG");
    add_srs_wkt (p, 2,
		 "S)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[-125,53,467,0,0,0,0],AUTHORITY[\"EPSG\",\"6142\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4142\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-9],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 10,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"2042\"],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2043, "epsg", 2043,
		      "Abidjan 1987 / UTM zone 29N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=29 +ellps=clrk80 +towgs84=-124.76,53,466");
    add_proj4text (p, 1, ".79,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Abidjan 1987 / UTM zone 29N\",GEOGCS[\"Abidjan ");
    add_srs_wkt (p, 1,
		 "1987\",DATUM[\"Abidjan_1987\",SPHEROID[\"Clarke 1880 (RG");
    add_srs_wkt (p, 2,
		 "S)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[-124.76,53,466.79,0,0,0,0],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 4,
		 "143\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 5,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 6,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4143\"]],UNIT[\"metr");
    add_srs_wkt (p, 7,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transve");
    add_srs_wkt (p, 8,
		 "rse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"central_meridian\",-9],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"2043\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2044, "epsg", 2044,
		      "Hanoi 1972 / Gauss-Kruger zone 18");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=18500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=-17.51,-108.32,-62.39,0,0,0,0 +u");
    add_proj4text (p, 2, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Hanoi 1972 / Gauss-Kruger zone 18\",GEOGCS[\"Ha");
    add_srs_wkt (p, 1,
		 "noi 1972\",DATUM[\"Hanoi_1972\",SPHEROID[\"Krassowsky 19");
    add_srs_wkt (p, 2,
		 "40\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84");
    add_srs_wkt (p, 3,
		 "[-17.51,-108.32,-62.39,0,0,0,0],AUTHORITY[\"EPSG\",\"614");
    add_srs_wkt (p, 4,
		 "7\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4147\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",105],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "1],PARAMETER[\"false_easting\",18500000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"2044\"],AXIS[\"X\"");
    add_srs_wkt (p, 12, ",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2045, "epsg", 2045,
		      "Hanoi 1972 / Gauss-Kruger zone 19");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=19500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=-17.51,-108.32,-62.39,0,0,0,0 +u");
    add_proj4text (p, 2, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Hanoi 1972 / Gauss-Kruger zone 19\",GEOGCS[\"Ha");
    add_srs_wkt (p, 1,
		 "noi 1972\",DATUM[\"Hanoi_1972\",SPHEROID[\"Krassowsky 19");
    add_srs_wkt (p, 2,
		 "40\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84");
    add_srs_wkt (p, 3,
		 "[-17.51,-108.32,-62.39,0,0,0,0],AUTHORITY[\"EPSG\",\"614");
    add_srs_wkt (p, 4,
		 "7\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4147\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",111],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "1],PARAMETER[\"false_easting\",19500000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"2045\"],AXIS[\"X\"");
    add_srs_wkt (p, 12, ",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2056, "epsg", 2056, "CH1903+ / LV95");
    add_proj4text (p, 0,
		   "+proj=somerc +lat_0=46.95240555555556 +lon_0=7.439583333");
    add_proj4text (p, 1,
		   "333333 +k_0=1 +x_0=2600000 +y_0=1200000 +ellps=bessel +t");
    add_proj4text (p, 2,
		   "owgs84=674.374,15.056,405.346,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"CH1903+ / LV95\",GEOGCS[\"CH1903+\",DATUM[\"CH1");
    add_srs_wkt (p, 1,
		 "903\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7004\"]],TOWGS84[674.374,15.056,405.");
    add_srs_wkt (p, 3,
		 "346,0,0,0,0],AUTHORITY[\"EPSG\",\"6150\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 4,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 5,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"4150\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"9001\"]],PROJECTION[\"Hotine_Oblique_Mercator\"],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"latitude_of_center\",46.95240555555556],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"longitude_of_center\",7.439583333333333],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"azimuth\",90],PARAMETER[\"rectified_grid_angle\",90]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting");
    add_srs_wkt (p, 12,
		 "\",2600000],PARAMETER[\"false_northing\",1200000],AUTHOR");
    add_srs_wkt (p, 13,
		 "ITY[\"EPSG\",\"2056\"],AXIS[\"Y\",EAST],AXIS[\"X\",NORTH");
    add_srs_wkt (p, 14, "]]");
    p = add_epsg_def (first, last, 2057, "epsg", 2057,
		      "Rassadiran / Nakhl e Taqi");
    add_proj4text (p, 0,
		   "+proj=omerc +lat_0=27.51882880555555 +lonc=52.6035391666");
    add_proj4text (p, 1,
		   "6667 +alpha=0.5716611944444444 +k=0.999895934 +x_0=65837");
    add_proj4text (p, 2,
		   "7.437 +y_0=3044969.194 +ellps=intl +towgs84=-133.63,-157");
    add_proj4text (p, 3, ".5,-158.62,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Rassadiran / Nakhl e Taqi\",GEOGCS[\"Rassadiran");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"Rassadiran\",SPHEROID[\"International 1924\",");
    add_srs_wkt (p, 2,
		 "6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-133.6");
    add_srs_wkt (p, 3,
		 "3,-157.5,-158.62,0,0,0,0],AUTHORITY[\"EPSG\",\"6153\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4153\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Hotine_Oblique_");
    add_srs_wkt (p, 8,
		 "Mercator\"],PARAMETER[\"latitude_of_center\",27.51882880");
    add_srs_wkt (p, 9,
		 "555555],PARAMETER[\"longitude_of_center\",52.60353916666");
    add_srs_wkt (p, 10,
		 "667],PARAMETER[\"azimuth\",0.5716611944444444],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"rectified_grid_angle\",0.5716611944444444],PARAMETER[");
    add_srs_wkt (p, 12,
		 "\"scale_factor\",0.999895934],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 13,
		 ",658377.437],PARAMETER[\"false_northing\",3044969.194],A");
    add_srs_wkt (p, 14,
		 "UTHORITY[\"EPSG\",\"2057\"],AXIS[\"Easting\",EAST],AXIS[");
    add_srs_wkt (p, 15, "\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2058, "epsg", 2058,
		      "ED50(ED77) / UTM zone 38N");
    add_proj4text (p, 0, "+proj=utm +zone=38 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50(ED77) / UTM zone 38N\",GEOGCS[\"ED50(ED77)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"European_Datum_1950_1977\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6154\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"4154\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",45],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"2058\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 2059, "epsg", 2059,
		      "ED50(ED77) / UTM zone 39N");
    add_proj4text (p, 0, "+proj=utm +zone=39 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50(ED77) / UTM zone 39N\",GEOGCS[\"ED50(ED77)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"European_Datum_1950_1977\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6154\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"4154\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",51],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"2059\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 2060, "epsg", 2060,
		      "ED50(ED77) / UTM zone 40N");
    add_proj4text (p, 0, "+proj=utm +zone=40 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50(ED77) / UTM zone 40N\",GEOGCS[\"ED50(ED77)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"European_Datum_1950_1977\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6154\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"4154\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",57],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"2060\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 2061, "epsg", 2061,
		      "ED50(ED77) / UTM zone 41N");
    add_proj4text (p, 0, "+proj=utm +zone=41 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50(ED77) / UTM zone 41N\",GEOGCS[\"ED50(ED77)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"European_Datum_1950_1977\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6154\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"4154\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",63],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"2061\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 2062, "epsg", 2062,
		      "Madrid 1870 (Madrid) / Spain");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40 +lat_0=40 +lon_0=0 +k_0=0.9988085293");
    add_proj4text (p, 1,
		   " +x_0=600000 +y_0=600000 +a=6378298.3 +b=6356657.1426695");
    add_proj4text (p, 2, "61 +pm=madrid +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Madrid 1870 (Madrid) / Spain\",GEOGCS[\"Madrid ");
    add_srs_wkt (p, 1,
		 "1870 (Madrid)\",DATUM[\"Madrid_1870_Madrid\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Struve 1860\",6378298.3,294.73,AUTHORITY[\"EPSG\",\"7028");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6903\"]],PRIMEM[\"Madrid\",-3.");
    add_srs_wkt (p, 4,
		 "687938888888889,AUTHORITY[\"EPSG\",\"8905\"]],UNIT[\"deg");
    add_srs_wkt (p, 5,
		 "ree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"4903\"]],UNIT[\"metre\",1,AUTHORITY");
    add_srs_wkt (p, 7,
		 "[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Coni");
    add_srs_wkt (p, 8,
		 "c_1SP\"],PARAMETER[\"latitude_of_origin\",40],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",0],PARAMETER[\"scale_factor\",0.998");
    add_srs_wkt (p, 10,
		 "8085293],PARAMETER[\"false_easting\",600000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",600000],AUTHORITY[\"EPSG\",\"2062\"],AX");
    add_srs_wkt (p, 12, "IS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2063, "epsg", 2063,
		      "Dabola 1981 / UTM zone 28N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=28 +a=6378249.2 +b=6356515 +towgs84=-23,");
    add_proj4text (p, 1, "259,-9,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Dabola 1981 / UTM zone 28N (deprecated)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"Conakry 1905\",DATUM[\"Conakry_1905\",SPHEROID[\"Cla");
    add_srs_wkt (p, 2,
		 "rke 1880 (IGN)\",6378249.2,293.4660212936265,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7011\"]],TOWGS84[-23,259,-9,0,0,0,0],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6315\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4315\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",-15],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12,
		 "2063\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 2064, "epsg", 2064,
		      "Dabola 1981 / UTM zone 29N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=29 +a=6378249.2 +b=6356515 +towgs84=-23,");
    add_proj4text (p, 1, "259,-9,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Dabola 1981 / UTM zone 29N (deprecated)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"Conakry 1905\",DATUM[\"Conakry_1905\",SPHEROID[\"Cla");
    add_srs_wkt (p, 2,
		 "rke 1880 (IGN)\",6378249.2,293.4660212936265,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7011\"]],TOWGS84[-23,259,-9,0,0,0,0],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6315\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4315\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",-9],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",0.9996],PARAMETER[\"false_easting\",500000");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 12,
		 "064\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2065, "epsg", 2065,
		      "S-JTSK (Ferro) / Krovak");
    add_proj4text (p, 0,
		   "+proj=krovak +lat_0=49.5 +lon_0=42.5 +alpha=30.288139722");
    add_proj4text (p, 1,
		   "22222 +k=0.9999 +x_0=0 +y_0=0 +ellps=bessel +pm=ferro +u");
    add_proj4text (p, 2, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"S-JTSK (Ferro) / Krovak\",GEOGCS[\"S-JTSK (Ferr");
    add_srs_wkt (p, 1,
		 "o)\",DATUM[\"S_JTSK_Ferro\",SPHEROID[\"Bessel 1841\",637");
    add_srs_wkt (p, 2,
		 "7397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6818\"]],PRIMEM[\"Ferro\",-17.6666666666");
    add_srs_wkt (p, 4,
		 "6667,AUTHORITY[\"EPSG\",\"8909\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4818\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Krovak\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "center\",49.5],PARAMETER[\"longitude_of_center\",42.5],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"azimuth\",30.28813972222222],PARAMETER[\"pseu");
    add_srs_wkt (p, 10,
		 "do_standard_parallel_1\",78.5],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 11,
		 ",0.9999],PARAMETER[\"false_easting\",0],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"2065\"],AXIS[\"X\",");
    add_srs_wkt (p, 13, "SOUTH],AXIS[\"Y\",WEST]]");
    p = add_epsg_def (first, last, 2066, "epsg", 2066,
		      "Mount Dillon / Tobago Grid");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=11.25217861111111 +lon_0=-60.686008888");
    add_proj4text (p, 1,
		   "88889 +x_0=37718.66159325 +y_0=36209.91512952 +a=6378293");
    add_proj4text (p, 2,
		   ".645208759 +b=6356617.987679838 +to_meter=0.201166195164");
    add_proj4text (p, 3, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Mount Dillon / Tobago Grid\",GEOGCS[\"Mount Dil");
    add_srs_wkt (p, 1,
		 "lon\",DATUM[\"Mount_Dillon\",SPHEROID[\"Clarke 1858\",63");
    add_srs_wkt (p, 2,
		 "78293.645208759,294.2606763692569,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "007\"]],AUTHORITY[\"EPSG\",\"6157\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4157\"]],UNIT[\"Clarke's link\",0.201166195164,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9039\"]],PROJECTION[\"Cassini_Soldn");
    add_srs_wkt (p, 8,
		 "er\"],PARAMETER[\"latitude_of_origin\",11.25217861111111");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-60.68600888888889],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_easting\",187500],PARAMETER[\"false_north");
    add_srs_wkt (p, 11,
		 "ing\",180000],AUTHORITY[\"EPSG\",\"2066\"],AXIS[\"Eastin");
    add_srs_wkt (p, 12, "g\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2067, "epsg", 2067,
		      "Naparima 1955 / UTM zone 20N");
    add_proj4text (p, 0, "+proj=utm +zone=20 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Naparima 1955 / UTM zone 20N\",GEOGCS[\"Naparim");
    add_srs_wkt (p, 1,
		 "a 1955\",DATUM[\"Naparima_1955\",SPHEROID[\"Internationa");
    add_srs_wkt (p, 2,
		 "l 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6158\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"415");
    add_srs_wkt (p, 6,
		 "8\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",-63],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 10,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 11,
		 "\",\"2067\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NO");
    add_srs_wkt (p, 12, "RTH]]");
    p = add_epsg_def (first, last, 2068, "epsg", 2068, "ELD79 / Libya zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=9 +k=0.9999 +x_0=200000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ELD79 / Libya zone 5\",GEOGCS[\"ELD79\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Libyan_Datum_1979\",SPHEROID[\"International 19");
    add_srs_wkt (p, 2,
		 "24\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4159\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",9],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "cale_factor\",0.9999],PARAMETER[\"false_easting\",200000");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 11, "068\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2069, "epsg", 2069, "ELD79 / Libya zone 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=11 +k=0.9999 +x_0=200000 +y_");
    add_proj4text (p, 1, "0=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ELD79 / Libya zone 6\",GEOGCS[\"ELD79\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Libyan_Datum_1979\",SPHEROID[\"International 19");
    add_srs_wkt (p, 2,
		 "24\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4159\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",11],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",20000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2069\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2070, "epsg", 2070, "ELD79 / Libya zone 7");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=13 +k=0.9999 +x_0=200000 +y_");
    add_proj4text (p, 1, "0=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ELD79 / Libya zone 7\",GEOGCS[\"ELD79\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Libyan_Datum_1979\",SPHEROID[\"International 19");
    add_srs_wkt (p, 2,
		 "24\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4159\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",13],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",20000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2070\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2071, "epsg", 2071, "ELD79 / Libya zone 8");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=0.9999 +x_0=200000 +y_");
    add_proj4text (p, 1, "0=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ELD79 / Libya zone 8\",GEOGCS[\"ELD79\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Libyan_Datum_1979\",SPHEROID[\"International 19");
    add_srs_wkt (p, 2,
		 "24\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4159\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",15],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",20000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2071\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2072, "epsg", 2072, "ELD79 / Libya zone 9");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=17 +k=0.9999 +x_0=200000 +y_");
    add_proj4text (p, 1, "0=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ELD79 / Libya zone 9\",GEOGCS[\"ELD79\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Libyan_Datum_1979\",SPHEROID[\"International 19");
    add_srs_wkt (p, 2,
		 "24\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4159\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",17],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",20000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2072\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2073, "epsg", 2073, "ELD79 / Libya zone 10");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=19 +k=0.9999 +x_0=200000 +y_");
    add_proj4text (p, 1, "0=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ELD79 / Libya zone 10\",GEOGCS[\"ELD79\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"European_Libyan_Datum_1979\",SPHEROID[\"International ");
    add_srs_wkt (p, 2,
		 "1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4159\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 8,
		 "igin\",0],PARAMETER[\"central_meridian\",19],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",20000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2073\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2074, "epsg", 2074, "ELD79 / Libya zone 11");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=0.9999 +x_0=200000 +y_");
    add_proj4text (p, 1, "0=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ELD79 / Libya zone 11\",GEOGCS[\"ELD79\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"European_Libyan_Datum_1979\",SPHEROID[\"International ");
    add_srs_wkt (p, 2,
		 "1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4159\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 8,
		 "igin\",0],PARAMETER[\"central_meridian\",21],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",20000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2074\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2075, "epsg", 2075, "ELD79 / Libya zone 12");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=23 +k=0.9999 +x_0=200000 +y_");
    add_proj4text (p, 1, "0=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ELD79 / Libya zone 12\",GEOGCS[\"ELD79\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"European_Libyan_Datum_1979\",SPHEROID[\"International ");
    add_srs_wkt (p, 2,
		 "1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4159\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 8,
		 "igin\",0],PARAMETER[\"central_meridian\",23],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",20000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2075\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2076, "epsg", 2076, "ELD79 / Libya zone 13");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=25 +k=0.9999 +x_0=200000 +y_");
    add_proj4text (p, 1, "0=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ELD79 / Libya zone 13\",GEOGCS[\"ELD79\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"European_Libyan_Datum_1979\",SPHEROID[\"International ");
    add_srs_wkt (p, 2,
		 "1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4159\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 8,
		 "igin\",0],PARAMETER[\"central_meridian\",25],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",20000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2076\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2077, "epsg", 2077, "ELD79 / UTM zone 32N");
    add_proj4text (p, 0, "+proj=utm +zone=32 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ELD79 / UTM zone 32N\",GEOGCS[\"ELD79\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Libyan_Datum_1979\",SPHEROID[\"International 19");
    add_srs_wkt (p, 2,
		 "24\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4159\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",9],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "cale_factor\",0.9996],PARAMETER[\"false_easting\",500000");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 11,
		 "077\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2078, "epsg", 2078, "ELD79 / UTM zone 33N");
    add_proj4text (p, 0, "+proj=utm +zone=33 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ELD79 / UTM zone 33N\",GEOGCS[\"ELD79\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Libyan_Datum_1979\",SPHEROID[\"International 19");
    add_srs_wkt (p, 2,
		 "24\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4159\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",15],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "2078\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 2079, "epsg", 2079, "ELD79 / UTM zone 34N");
    add_proj4text (p, 0, "+proj=utm +zone=34 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ELD79 / UTM zone 34N\",GEOGCS[\"ELD79\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Libyan_Datum_1979\",SPHEROID[\"International 19");
    add_srs_wkt (p, 2,
		 "24\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4159\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",21],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "2079\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 2080, "epsg", 2080, "ELD79 / UTM zone 35N");
    add_proj4text (p, 0, "+proj=utm +zone=35 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ELD79 / UTM zone 35N\",GEOGCS[\"ELD79\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Libyan_Datum_1979\",SPHEROID[\"International 19");
    add_srs_wkt (p, 2,
		 "24\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4159\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",27],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "2080\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 2081, "epsg", 2081,
		      "Chos Malal 1914 / Argentina zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-69 +k=1 +x_0=2500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Chos Malal 1914 / Argentina zone 2\",GEOGCS[\"C");
    add_srs_wkt (p, 1,
		 "hos Malal 1914\",DATUM[\"Chos_Malal_1914\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6160\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4160\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",-90],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 9,
		 "-69],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",2500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"2081\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 2082, "epsg", 2082,
		      "Pampa del Castillo / Argentina zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-69 +k=1 +x_0=2500000 +y_0");
    add_proj4text (p, 1,
		   "=0 +ellps=intl +towgs84=27.5,14,186.4,0,0,0,0 +units=m +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pampa del Castillo / Argentina zone 2\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"Pampa del Castillo\",DATUM[\"Pampa_del_Castillo\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7022\"]],TOWGS84[27.5,14,186.4,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6161\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4161\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",-90],PARAMETER[\"central_meridian\",-69],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",2500000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2082");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2083, "epsg", 2083,
		      "Hito XVIII 1963 / Argentina zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-69 +k=1 +x_0=2500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Hito XVIII 1963 / Argentina zone 2\",GEOGCS[\"H");
    add_srs_wkt (p, 1,
		 "ito XVIII 1963\",DATUM[\"Hito_XVIII_1963\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6254\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4254\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",-90],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 9,
		 "-69],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",2500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"2083\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 2084, "epsg", 2084,
		      "Hito XVIII 1963 / UTM zone 19S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Hito XVIII 1963 / UTM zone 19S\",GEOGCS[\"Hito ");
    add_srs_wkt (p, 1,
		 "XVIII 1963\",DATUM[\"Hito_XVIII_1963\",SPHEROID[\"Intern");
    add_srs_wkt (p, 2,
		 "ational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6254\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4254\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-69],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",10000000],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"2084\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2085, "epsg", 2085,
		      "NAD27 / Cuba Norte (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=22.35 +lat_0=22.35 +lon_0=-81 +k_0=0.99");
    add_proj4text (p, 1,
		   "993602 +x_0=500000 +y_0=280296.016 +ellps=clrk66 +datum=");
    add_proj4text (p, 2, "NAD27 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Cuba Norte (deprecated)\",GEOGCS[\"NAD2");
    add_srs_wkt (p, 1,
		 "7\",DATUM[\"North_American_Datum_1927\",SPHEROID[\"Clark");
    add_srs_wkt (p, 2,
		 "e 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"7008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1SP\"],");
    add_srs_wkt (p, 8,
		 "PARAMETER[\"latitude_of_origin\",22.35],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",-81],PARAMETER[\"scale_factor\",0.9999360");
    add_srs_wkt (p, 10,
		 "2],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_northing\",280296.016],AUTHORITY[\"EPSG\",\"2085\"],AXI");
    add_srs_wkt (p, 12, "S[\"Y\",NORTH],AXIS[\"X\",EAST]]");
    p = add_epsg_def (first, last, 2086, "epsg", 2086,
		      "NAD27 / Cuba Sur (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=20.71666666666667 +lat_0=20.71666666666");
    add_proj4text (p, 1,
		   "667 +lon_0=-76.83333333333333 +k_0=0.99994848 +x_0=50000");
    add_proj4text (p, 2,
		   "0 +y_0=229126.939 +ellps=clrk66 +datum=NAD27 +units=m +n");
    add_proj4text (p, 3, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Cuba Sur (deprecated)\",GEOGCS[\"NAD27\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1");
    add_srs_wkt (p, 2,
		 "866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1SP\"],PAR");
    add_srs_wkt (p, 8,
		 "AMETER[\"latitude_of_origin\",20.71666666666667],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-76.83333333333333],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",0.99994848],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",229126.939],AUTHORIT");
    add_srs_wkt (p, 12,
		 "Y[\"EPSG\",\"2086\"],AXIS[\"Y\",NORTH],AXIS[\"X\",EAST]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 2087, "epsg", 2087, "ELD79 / TM 12 NE");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=12 +k=0.9996 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ELD79 / TM 12 NE\",GEOGCS[\"ELD79\",DATUM[\"Eur");
    add_srs_wkt (p, 1,
		 "opean_Libyan_Datum_1979\",SPHEROID[\"International 1924\"");
    add_srs_wkt (p, 2,
		 ",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4159\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",0],PARAMETER[\"central_meridian\",12],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",0.9996],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2087\"");
    add_srs_wkt (p, 11, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2088, "epsg", 2088, "Carthage / TM 11 NE");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=11 +k=0.9996 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=0 +a=6378249.2 +b=6356515 +datum=carthage +units=m +no");
    add_proj4text (p, 2, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Carthage / TM 11 NE\",GEOGCS[\"Carthage\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Carthage\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,2");
    add_srs_wkt (p, 2,
		 "93.4660212936265,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6223\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4223\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",11],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "2088\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 2089, "epsg", 2089,
		      "Yemen NGN96 / UTM zone 38N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=38 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Yemen NGN96 / UTM zone 38N\",GEOGCS[\"Yemen NGN");
    add_srs_wkt (p, 1,
		 "96\",DATUM[\"Yemen_National_Geodetic_Network_1996\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "6163\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4163\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 8,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"central_meridian\",45],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2089\"],AXIS");
    add_srs_wkt (p, 12, "[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2090, "epsg", 2090,
		      "Yemen NGN96 / UTM zone 39N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=39 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Yemen NGN96 / UTM zone 39N\",GEOGCS[\"Yemen NGN");
    add_srs_wkt (p, 1,
		 "96\",DATUM[\"Yemen_National_Geodetic_Network_1996\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "6163\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4163\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 8,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"central_meridian\",51],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2090\"],AXIS");
    add_srs_wkt (p, 12, "[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2091, "epsg", 2091,
		      "South Yemen / Gauss Kruger zone 8 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=45 +k=1 +x_0=8500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=-76,-138,67,0,0,0,0 +units=m +no_d");
    add_proj4text (p, 2, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"South Yemen / Gauss Kruger zone 8 (deprecated)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"South Yemen\",DATUM[\"South_Yemen\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[-76,-138,67,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "6164\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4164\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 8,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"central_meridian\",45],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",1],PARAMETER[\"false_easting\",8500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"2091\"],AXIS[\"X");
    add_srs_wkt (p, 12, "\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2092, "epsg", 2092,
		      "South Yemen / Gauss Kruger zone 9 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=51 +k=1 +x_0=9500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=-76,-138,67,0,0,0,0 +units=m +no_d");
    add_proj4text (p, 2, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"South Yemen / Gauss Kruger zone 9 (deprecated)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"South Yemen\",DATUM[\"South_Yemen\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[-76,-138,67,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "6164\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4164\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 8,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"central_meridian\",51],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",1],PARAMETER[\"false_easting\",9500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"2092\"],AXIS[\"X");
    add_srs_wkt (p, 12, "\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2093, "epsg", 2093,
		      "Hanoi 1972 / GK 106 NE");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=106 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=-17.51,-108.32,-62.39,0,0,0,0 +uni");
    add_proj4text (p, 2, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Hanoi 1972 / GK 106 NE\",GEOGCS[\"Hanoi 1972\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Hanoi_1972\",SPHEROID[\"Krassowsky 1940\",637824");
    add_srs_wkt (p, 2,
		 "5,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[-17.51,-10");
    add_srs_wkt (p, 3,
		 "8.32,-62.39,0,0,0,0],AUTHORITY[\"EPSG\",\"6147\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4147\"]],UNIT[\"metre\",1,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",106],PARAMETER[\"scale_factor\",1],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"2093\"],AXIS[\"X\",NORTH],AXIS[");
    add_srs_wkt (p, 12, "\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2094, "epsg", 2094, "WGS 72BE / TM 106 NE");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=106 +k=0.9996 +x_0=500000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.814,-0.38 +unit");
    add_proj4text (p, 2, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / TM 106 NE\",GEOGCS[\"WGS 72BE\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"WG");
    add_srs_wkt (p, 2,
		 "S 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWG");
    add_srs_wkt (p, 3,
		 "S84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",106],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"2094\"],AXIS[\"Eas");
    add_srs_wkt (p, 12, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2095, "epsg", 2095, "Bissau / UTM zone 28N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=28 +ellps=intl +towgs84=-173,253,27,0,0,");
    add_proj4text (p, 1, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Bissau / UTM zone 28N\",GEOGCS[\"Bissau\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Bissau\",SPHEROID[\"International 1924\",6378388,297,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-173,253,27,0,0,0,");
    add_srs_wkt (p, 3,
		 "0],AUTHORITY[\"EPSG\",\"6165\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4165\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"lat");
    add_srs_wkt (p, 8,
		 "itude_of_origin\",0],PARAMETER[\"central_meridian\",-15]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_ea");
    add_srs_wkt (p, 10,
		 "sting\",500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"2095\"],AXIS[\"Easting\",EAST],AXIS[\"North");
    add_srs_wkt (p, 12, "ing\",NORTH]]");
    p = add_epsg_def (first, last, 2096, "epsg", 2096,
		      "Korean 1985 / Korea East Belt");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38 +lon_0=129 +k=1 +x_0=200000 +y_0=5");
    add_proj4text (p, 1, "00000 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Korean 1985 / Korea East Belt\",GEOGCS[\"Korean");
    add_srs_wkt (p, 1,
		 " 1985\",DATUM[\"Korean_Datum_1985\",SPHEROID[\"Bessel 18");
    add_srs_wkt (p, 2,
		 "41\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6162\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4162\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"lat");
    add_srs_wkt (p, 8,
		 "itude_of_origin\",38],PARAMETER[\"central_meridian\",129");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",200000],PARAMETER[\"false_northing\",500000],AUTHORI");
    add_srs_wkt (p, 11,
		 "TY[\"EPSG\",\"2096\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 2097, "epsg", 2097,
		      "Korean 1985 / Korea Central Belt");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38 +lon_0=127 +k=1 +x_0=200000 +y_0=5");
    add_proj4text (p, 1, "00000 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Korean 1985 / Korea Central Belt\",GEOGCS[\"Kor");
    add_srs_wkt (p, 1,
		 "ean 1985\",DATUM[\"Korean_Datum_1985\",SPHEROID[\"Bessel");
    add_srs_wkt (p, 2,
		 " 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 3,
		 "4\"]],AUTHORITY[\"EPSG\",\"6162\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4162\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",38],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 9,
		 "127],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",200000],PARAMETER[\"false_northing\",500000],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"2097\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EA");
    add_srs_wkt (p, 12, "ST]]");
    p = add_epsg_def (first, last, 2098, "epsg", 2098,
		      "Korean 1985 / Korea West Belt");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38 +lon_0=125 +k=1 +x_0=200000 +y_0=5");
    add_proj4text (p, 1, "00000 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Korean 1985 / Korea West Belt\",GEOGCS[\"Korean");
    add_srs_wkt (p, 1,
		 " 1985\",DATUM[\"Korean_Datum_1985\",SPHEROID[\"Bessel 18");
    add_srs_wkt (p, 2,
		 "41\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6162\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4162\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"lat");
    add_srs_wkt (p, 8,
		 "itude_of_origin\",38],PARAMETER[\"central_meridian\",125");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",200000],PARAMETER[\"false_northing\",500000],AUTHORI");
    add_srs_wkt (p, 11,
		 "TY[\"EPSG\",\"2098\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 2099, "epsg", 2099,
		      "Qatar 1948 / Qatar Grid");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=25.38236111111111 +lon_0=50.7613888888");
    add_proj4text (p, 1,
		   "8889 +x_0=100000 +y_0=100000 +ellps=helmert +units=m +no");
    add_proj4text (p, 2, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Qatar 1948 / Qatar Grid\",GEOGCS[\"Qatar 1948\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Qatar_1948\",SPHEROID[\"Helmert 1906\",6378200,");
    add_srs_wkt (p, 2,
		 "298.3,AUTHORITY[\"EPSG\",\"7020\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "6286\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 4,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 5,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4286\"]],UNIT[\"met");
    add_srs_wkt (p, 6,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Cassin");
    add_srs_wkt (p, 7,
		 "i_Soldner\"],PARAMETER[\"latitude_of_origin\",25.3823611");
    add_srs_wkt (p, 8,
		 "1111111],PARAMETER[\"central_meridian\",50.7613888888888");
    add_srs_wkt (p, 9,
		 "9],PARAMETER[\"false_easting\",100000],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_northing\",100000],AUTHORITY[\"EPSG\",\"2099\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2100, "epsg", 2100, "GGRS87 / Greek Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=24 +k=0.9996 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=0 +ellps=GRS80 +datum=GGRS87 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GGRS87 / Greek Grid\",GEOGCS[\"GGRS87\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Greek_Geodetic_Reference_System_1987\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 2,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 3,
		 ",TOWGS84[-199.87,74.79,246.62,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6121\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4121\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"central_meridian\",24],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 10,
		 "or\",0.9996],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2100\"],AX");
    add_srs_wkt (p, 12, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2101, "epsg", 2101,
		      "Lake / Maracaibo Grid M1");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=10.16666666666667 +lat_0=10.16666666666");
    add_proj4text (p, 1,
		   "667 +lon_0=-71.60561777777777 +k_0=1 +x_0=0 +y_0=-52684.");
    add_proj4text (p, 2, "972 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Lake / Maracaibo Grid M1\",GEOGCS[\"Lake\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"Lake\",SPHEROID[\"International 1924\",6378388,297,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6249\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4249\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Confo");
    add_srs_wkt (p, 7,
		 "rmal_Conic_1SP\"],PARAMETER[\"latitude_of_origin\",10.16");
    add_srs_wkt (p, 8,
		 "666666666667],PARAMETER[\"central_meridian\",-71.6056177");
    add_srs_wkt (p, 9,
		 "7777777],PARAMETER[\"scale_factor\",1],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",0],PARAMETER[\"false_northing\",-52684.972],A");
    add_srs_wkt (p, 11,
		 "UTHORITY[\"EPSG\",\"2101\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 12, "NORTH]]");
    p = add_epsg_def (first, last, 2102, "epsg", 2102, "Lake / Maracaibo Grid");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=10.16666666666667 +lat_0=10.16666666666");
    add_proj4text (p, 1,
		   "667 +lon_0=-71.60561777777777 +k_0=1 +x_0=200000 +y_0=14");
    add_proj4text (p, 2, "7315.028 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Lake / Maracaibo Grid\",GEOGCS[\"Lake\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Lake\",SPHEROID[\"International 1924\",6378388,297,AUTHO");
    add_srs_wkt (p, 2,
		 "RITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6249\"]],P");
    add_srs_wkt (p, 3,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 4,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 5,
		 "2\"]],AUTHORITY[\"EPSG\",\"4249\"]],UNIT[\"metre\",1,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conforma");
    add_srs_wkt (p, 7,
		 "l_Conic_1SP\"],PARAMETER[\"latitude_of_origin\",10.16666");
    add_srs_wkt (p, 8,
		 "666666667],PARAMETER[\"central_meridian\",-71.6056177777");
    add_srs_wkt (p, 9,
		 "7777],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_ea");
    add_srs_wkt (p, 10,
		 "sting\",200000],PARAMETER[\"false_northing\",147315.028]");
    add_srs_wkt (p, 11,
		 ",AUTHORITY[\"EPSG\",\"2102\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 2103, "epsg", 2103,
		      "Lake / Maracaibo Grid M3");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=10.16666666666667 +lat_0=10.16666666666");
    add_proj4text (p, 1,
		   "667 +lon_0=-71.60561777777777 +k_0=1 +x_0=500000 +y_0=44");
    add_proj4text (p, 2, "7315.028 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Lake / Maracaibo Grid M3\",GEOGCS[\"Lake\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"Lake\",SPHEROID[\"International 1924\",6378388,297,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6249\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4249\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Confo");
    add_srs_wkt (p, 7,
		 "rmal_Conic_1SP\"],PARAMETER[\"latitude_of_origin\",10.16");
    add_srs_wkt (p, 8,
		 "666666666667],PARAMETER[\"central_meridian\",-71.6056177");
    add_srs_wkt (p, 9,
		 "7777777],PARAMETER[\"scale_factor\",1],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",447315.0");
    add_srs_wkt (p, 11,
		 "28],AUTHORITY[\"EPSG\",\"2103\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 2104, "epsg", 2104,
		      "Lake / Maracaibo La Rosa Grid");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=10.16666666666667 +lat_0=10.16666666666");
    add_proj4text (p, 1,
		   "667 +lon_0=-71.60561777777777 +k_0=1 +x_0=-17044 +y_0=-2");
    add_proj4text (p, 2, "3139.97 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Lake / Maracaibo La Rosa Grid\",GEOGCS[\"Lake\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Lake\",SPHEROID[\"International 1924\",6378388,");
    add_srs_wkt (p, 2,
		 "297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "249\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4249\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_1SP\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",10.16666666666667],PARAMETER[\"central_meridian\",-71.6");
    add_srs_wkt (p, 9,
		 "0561777777777],PARAMETER[\"scale_factor\",1],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",-17044],PARAMETER[\"false_northing\",-23");
    add_srs_wkt (p, 11,
		 "139.97],AUTHORITY[\"EPSG\",\"2104\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 12, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2105, "epsg", 2105,
		      "NZGD2000 / Mount Eden 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-36.87972222222222 +lon_0=174.7641666");
    add_proj4text (p, 1,
		   "666667 +k=0.9999 +x_0=400000 +y_0=800000 +ellps=GRS80 +t");
    add_proj4text (p, 2, "owgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Mount Eden 2000\",GEOGCS[\"NZGD2000\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GR");
    add_srs_wkt (p, 2,
		 "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019");
    add_srs_wkt (p, 3,
		 "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",-36.879722222");
    add_srs_wkt (p, 9,
		 "22222],PARAMETER[\"central_meridian\",174.7641666666667]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",400000],PARAMETER[\"false_northing\",800000],AUT");
    add_srs_wkt (p, 12,
		 "HORITY[\"EPSG\",\"2105\"],AXIS[\"Northing\",NORTH],AXIS[");
    add_srs_wkt (p, 13, "\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 2106, "epsg", 2106,
		      "NZGD2000 / Bay of Plenty 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-37.76111111111111 +lon_0=176.4661111");
    add_proj4text (p, 1,
		   "111111 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Bay of Plenty 2000\",GEOGCS[\"NZGD20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"616");
    add_srs_wkt (p, 4,
		 "7\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",-37.761111");
    add_srs_wkt (p, 9,
		 "11111111],PARAMETER[\"central_meridian\",176.46611111111");
    add_srs_wkt (p, 10,
		 "11],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 11,
		 "ing\",400000],PARAMETER[\"false_northing\",800000],AUTHO");
    add_srs_wkt (p, 12,
		 "RITY[\"EPSG\",\"2106\"],AXIS[\"Northing\",NORTH],AXIS[\"");
    add_srs_wkt (p, 13, "Easting\",EAST]]");
    p = add_epsg_def (first, last, 2107, "epsg", 2107,
		      "NZGD2000 / Poverty Bay 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-38.62444444444444 +lon_0=177.8855555");
    add_proj4text (p, 1,
		   "555556 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Poverty Bay 2000\",GEOGCS[\"NZGD2000");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",-38.624444444");
    add_srs_wkt (p, 9,
		 "44444],PARAMETER[\"central_meridian\",177.8855555555556]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting");
    add_srs_wkt (p, 11,
		 "\",400000],PARAMETER[\"false_northing\",800000],AUTHORIT");
    add_srs_wkt (p, 12,
		 "Y[\"EPSG\",\"2107\"],AXIS[\"Northing\",NORTH],AXIS[\"Eas");
    add_srs_wkt (p, 13, "ting\",EAST]]");
    p = add_epsg_def (first, last, 2108, "epsg", 2108,
		      "NZGD2000 / Hawkes Bay 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-39.65083333333333 +lon_0=176.6736111");
    add_proj4text (p, 1,
		   "111111 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Hawkes Bay 2000\",GEOGCS[\"NZGD2000\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GR");
    add_srs_wkt (p, 2,
		 "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019");
    add_srs_wkt (p, 3,
		 "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",-39.650833333");
    add_srs_wkt (p, 9,
		 "33333],PARAMETER[\"central_meridian\",176.6736111111111]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting");
    add_srs_wkt (p, 11,
		 "\",400000],PARAMETER[\"false_northing\",800000],AUTHORIT");
    add_srs_wkt (p, 12,
		 "Y[\"EPSG\",\"2108\"],AXIS[\"Northing\",NORTH],AXIS[\"Eas");
    add_srs_wkt (p, 13, "ting\",EAST]]");
    p = add_epsg_def (first, last, 2109, "epsg", 2109,
		      "NZGD2000 / Taranaki 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-39.13555555555556 +lon_0=174.2277777");
    add_proj4text (p, 1,
		   "777778 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Taranaki 2000\",GEOGCS[\"NZGD2000\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",-39.13555555555");
    add_srs_wkt (p, 9,
		 "556],PARAMETER[\"central_meridian\",174.2277777777778],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",400000],PARAMETER[\"false_northing\",800000],AUTHORITY[");
    add_srs_wkt (p, 12,
		 "\"EPSG\",\"2109\"],AXIS[\"Northing\",NORTH],AXIS[\"Easti");
    add_srs_wkt (p, 13, "ng\",EAST]]");
    p = add_epsg_def (first, last, 2110, "epsg", 2110,
		      "NZGD2000 / Tuhirangi 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-39.51222222222222 +lon_0=175.64 +k=1");
    add_proj4text (p, 1,
		   " +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs84=0,0,0,0,0");
    add_proj4text (p, 2, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Tuhirangi 2000\",GEOGCS[\"NZGD2000\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GR");
    add_srs_wkt (p, 2,
		 "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019");
    add_srs_wkt (p, 3,
		 "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",-39.512222222");
    add_srs_wkt (p, 9,
		 "22222],PARAMETER[\"central_meridian\",175.64],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"scale_factor\",1],PARAMETER[\"false_easting\",400000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",800000],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12,
		 "\"2110\"],AXIS[\"Northing\",NORTH],AXIS[\"Easting\",EAST");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 2111, "epsg", 2111,
		      "NZGD2000 / Wanganui 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-40.24194444444444 +lon_0=175.4880555");
    add_proj4text (p, 1,
		   "555555 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Wanganui 2000\",GEOGCS[\"NZGD2000\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",-40.24194444444");
    add_srs_wkt (p, 9,
		 "444],PARAMETER[\"central_meridian\",175.4880555555555],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",400000],PARAMETER[\"false_northing\",800000],AUTHORITY[");
    add_srs_wkt (p, 12,
		 "\"EPSG\",\"2111\"],AXIS[\"Northing\",NORTH],AXIS[\"Easti");
    add_srs_wkt (p, 13, "ng\",EAST]]");
    p = add_epsg_def (first, last, 2112, "epsg", 2112,
		      "NZGD2000 / Wairarapa 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-40.92527777777777 +lon_0=175.6472222");
    add_proj4text (p, 1,
		   "222222 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Wairarapa 2000\",GEOGCS[\"NZGD2000\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GR");
    add_srs_wkt (p, 2,
		 "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019");
    add_srs_wkt (p, 3,
		 "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",-40.925277777");
    add_srs_wkt (p, 9,
		 "77777],PARAMETER[\"central_meridian\",175.6472222222222]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting");
    add_srs_wkt (p, 11,
		 "\",400000],PARAMETER[\"false_northing\",800000],AUTHORIT");
    add_srs_wkt (p, 12,
		 "Y[\"EPSG\",\"2112\"],AXIS[\"Northing\",NORTH],AXIS[\"Eas");
    add_srs_wkt (p, 13, "ting\",EAST]]");
    p = add_epsg_def (first, last, 2113, "epsg", 2113,
		      "NZGD2000 / Wellington 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-41.3011111111111 +lon_0=174.77638888");
    add_proj4text (p, 1,
		   "88889 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs84");
    add_proj4text (p, 2, "=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Wellington 2000\",GEOGCS[\"NZGD2000\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GR");
    add_srs_wkt (p, 2,
		 "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019");
    add_srs_wkt (p, 3,
		 "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",-41.301111111");
    add_srs_wkt (p, 9,
		 "1111],PARAMETER[\"central_meridian\",174.7763888888889],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",400000],PARAMETER[\"false_northing\",800000],AUTHORITY[");
    add_srs_wkt (p, 12,
		 "\"EPSG\",\"2113\"],AXIS[\"Northing\",NORTH],AXIS[\"Easti");
    add_srs_wkt (p, 13, "ng\",EAST]]");
    p = add_epsg_def (first, last, 2114, "epsg", 2114,
		      "NZGD2000 / Collingwood 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-40.71472222222223 +lon_0=172.6719444");
    add_proj4text (p, 1,
		   "444444 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Collingwood 2000\",GEOGCS[\"NZGD2000");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",-40.714722222");
    add_srs_wkt (p, 9,
		 "22223],PARAMETER[\"central_meridian\",172.6719444444444]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting");
    add_srs_wkt (p, 11,
		 "\",400000],PARAMETER[\"false_northing\",800000],AUTHORIT");
    add_srs_wkt (p, 12,
		 "Y[\"EPSG\",\"2114\"],AXIS[\"Northing\",NORTH],AXIS[\"Eas");
    add_srs_wkt (p, 13, "ting\",EAST]]");
    p = add_epsg_def (first, last, 2115, "epsg", 2115,
		      "NZGD2000 / Nelson 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-41.27444444444444 +lon_0=173.2991666");
    add_proj4text (p, 1,
		   "666667 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Nelson 2000\",GEOGCS[\"NZGD2000\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 2,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 3,
		 "],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 6,
		 "2\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,AUT");
    add_srs_wkt (p, 7,
		 "HORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merca");
    add_srs_wkt (p, 8,
		 "tor\"],PARAMETER[\"latitude_of_origin\",-41.274444444444");
    add_srs_wkt (p, 9,
		 "44],PARAMETER[\"central_meridian\",173.2991666666667],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 11,
		 "400000],PARAMETER[\"false_northing\",800000],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"2115\"],AXIS[\"Northing\",NORTH],AXIS[\"Easting");
    add_srs_wkt (p, 13, "\",EAST]]");
    p = add_epsg_def (first, last, 2116, "epsg", 2116,
		      "NZGD2000 / Karamea 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-41.28972222222222 +lon_0=172.1088888");
    add_proj4text (p, 1,
		   "888889 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Karamea 2000\",GEOGCS[\"NZGD2000\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GRS ");
    add_srs_wkt (p, 2,
		 "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",-41.28972222222");
    add_srs_wkt (p, 9,
		 "222],PARAMETER[\"central_meridian\",172.1088888888889],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",400000],PARAMETER[\"false_northing\",800000],AUTHORITY[");
    add_srs_wkt (p, 12,
		 "\"EPSG\",\"2116\"],AXIS[\"Northing\",NORTH],AXIS[\"Easti");
    add_srs_wkt (p, 13, "ng\",EAST]]");
    p = add_epsg_def (first, last, 2117, "epsg", 2117,
		      "NZGD2000 / Buller 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-41.81055555555555 +lon_0=171.5811111");
    add_proj4text (p, 1,
		   "111111 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Buller 2000\",GEOGCS[\"NZGD2000\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 2,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 3,
		 "],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 6,
		 "2\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,AUT");
    add_srs_wkt (p, 7,
		 "HORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merca");
    add_srs_wkt (p, 8,
		 "tor\"],PARAMETER[\"latitude_of_origin\",-41.810555555555");
    add_srs_wkt (p, 9,
		 "55],PARAMETER[\"central_meridian\",171.5811111111111],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 11,
		 "400000],PARAMETER[\"false_northing\",800000],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"2117\"],AXIS[\"Northing\",NORTH],AXIS[\"Easting");
    add_srs_wkt (p, 13, "\",EAST]]");
    p = add_epsg_def (first, last, 2118, "epsg", 2118, "NZGD2000 / Grey 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-42.33361111111111 +lon_0=171.5497222");
    add_proj4text (p, 1,
		   "222222 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Grey 2000\",GEOGCS[\"NZGD2000\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GRS 198");
    add_srs_wkt (p, 2,
		 "0\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],");
    add_srs_wkt (p, 3,
		 "TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",-42.33361111111111]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"central_meridian\",171.5497222222222],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",400");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",800000],AUTHORITY[\"EP");
    add_srs_wkt (p, 12,
		 "SG\",\"2118\"],AXIS[\"Northing\",NORTH],AXIS[\"Easting\"");
    add_srs_wkt (p, 13, ",EAST]]");
    p = add_epsg_def (first, last, 2119, "epsg", 2119, "NZGD2000 / Amuri 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-42.68888888888888 +lon_0=173.01 +k=1");
    add_proj4text (p, 1,
		   " +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs84=0,0,0,0,0");
    add_proj4text (p, 2, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Amuri 2000\",GEOGCS[\"NZGD2000\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 2,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 3,
		 ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6,
		 "\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",-42.6888888888888");
    add_srs_wkt (p, 9,
		 "8],PARAMETER[\"central_meridian\",173.01],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",400000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",800000],AUTHORITY[\"EPSG\",\"21");
    add_srs_wkt (p, 12,
		 "19\"],AXIS[\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 2120, "epsg", 2120,
		      "NZGD2000 / Marlborough 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-41.54444444444444 +lon_0=173.8019444");
    add_proj4text (p, 1,
		   "444444 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Marlborough 2000\",GEOGCS[\"NZGD2000");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",-41.544444444");
    add_srs_wkt (p, 9,
		 "44444],PARAMETER[\"central_meridian\",173.8019444444444]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting");
    add_srs_wkt (p, 11,
		 "\",400000],PARAMETER[\"false_northing\",800000],AUTHORIT");
    add_srs_wkt (p, 12,
		 "Y[\"EPSG\",\"2120\"],AXIS[\"Northing\",NORTH],AXIS[\"Eas");
    add_srs_wkt (p, 13, "ting\",EAST]]");
    p = add_epsg_def (first, last, 2121, "epsg", 2121,
		      "NZGD2000 / Hokitika 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-42.88611111111111 +lon_0=170.9797222");
    add_proj4text (p, 1,
		   "222222 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Hokitika 2000\",GEOGCS[\"NZGD2000\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",-42.88611111111");
    add_srs_wkt (p, 9,
		 "111],PARAMETER[\"central_meridian\",170.9797222222222],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",400000],PARAMETER[\"false_northing\",800000],AUTHORITY[");
    add_srs_wkt (p, 12,
		 "\"EPSG\",\"2121\"],AXIS[\"Northing\",NORTH],AXIS[\"Easti");
    add_srs_wkt (p, 13, "ng\",EAST]]");
    p = add_epsg_def (first, last, 2122, "epsg", 2122,
		      "NZGD2000 / Okarito 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-43.11 +lon_0=170.2608333333333 +k=1 ");
    add_proj4text (p, 1,
		   "+x_0=400000 +y_0=800000 +ellps=GRS80 +towgs84=0,0,0,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Okarito 2000\",GEOGCS[\"NZGD2000\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GRS ");
    add_srs_wkt (p, 2,
		 "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",-43.11],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",170.2608333333333],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",1],PARAMETER[\"false_easting\",400000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",800000],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 12,
		 "122\"],AXIS[\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 2123, "epsg", 2123,
		      "NZGD2000 / Jacksons Bay 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-43.97777777777778 +lon_0=168.6061111");
    add_proj4text (p, 1,
		   "111111 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Jacksons Bay 2000\",GEOGCS[\"NZGD200");
    add_srs_wkt (p, 1,
		 "0\",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",-43.977777777");
    add_srs_wkt (p, 9,
		 "77778],PARAMETER[\"central_meridian\",168.6061111111111]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting");
    add_srs_wkt (p, 11,
		 "\",400000],PARAMETER[\"false_northing\",800000],AUTHORIT");
    add_srs_wkt (p, 12,
		 "Y[\"EPSG\",\"2123\"],AXIS[\"Northing\",NORTH],AXIS[\"Eas");
    add_srs_wkt (p, 13, "ting\",EAST]]");
    p = add_epsg_def (first, last, 2124, "epsg", 2124,
		      "NZGD2000 / Mount Pleasant 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-43.59055555555556 +lon_0=172.7269444");
    add_proj4text (p, 1,
		   "444445 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Mount Pleasant 2000\",GEOGCS[\"NZGD2");
    add_srs_wkt (p, 1,
		 "000\",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"616");
    add_srs_wkt (p, 4,
		 "7\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",-43.590555");
    add_srs_wkt (p, 9,
		 "55555556],PARAMETER[\"central_meridian\",172.72694444444");
    add_srs_wkt (p, 10,
		 "45],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 11,
		 "ing\",400000],PARAMETER[\"false_northing\",800000],AUTHO");
    add_srs_wkt (p, 12,
		 "RITY[\"EPSG\",\"2124\"],AXIS[\"Northing\",NORTH],AXIS[\"");
    add_srs_wkt (p, 13, "Easting\",EAST]]");
    p = add_epsg_def (first, last, 2125, "epsg", 2125,
		      "NZGD2000 / Gawler 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-43.74861111111111 +lon_0=171.3605555");
    add_proj4text (p, 1,
		   "555555 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Gawler 2000\",GEOGCS[\"NZGD2000\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 2,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 3,
		 "],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 6,
		 "2\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,AUT");
    add_srs_wkt (p, 7,
		 "HORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merca");
    add_srs_wkt (p, 8,
		 "tor\"],PARAMETER[\"latitude_of_origin\",-43.748611111111");
    add_srs_wkt (p, 9,
		 "11],PARAMETER[\"central_meridian\",171.3605555555555],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 11,
		 "400000],PARAMETER[\"false_northing\",800000],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"2125\"],AXIS[\"Northing\",NORTH],AXIS[\"Easting");
    add_srs_wkt (p, 13, "\",EAST]]");
    p = add_epsg_def (first, last, 2126, "epsg", 2126,
		      "NZGD2000 / Timaru 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-44.40194444444445 +lon_0=171.0572222");
    add_proj4text (p, 1,
		   "222222 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Timaru 2000\",GEOGCS[\"NZGD2000\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 2,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 3,
		 "],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 6,
		 "2\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,AUT");
    add_srs_wkt (p, 7,
		 "HORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merca");
    add_srs_wkt (p, 8,
		 "tor\"],PARAMETER[\"latitude_of_origin\",-44.401944444444");
    add_srs_wkt (p, 9,
		 "45],PARAMETER[\"central_meridian\",171.0572222222222],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 11,
		 "400000],PARAMETER[\"false_northing\",800000],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"2126\"],AXIS[\"Northing\",NORTH],AXIS[\"Easting");
    add_srs_wkt (p, 13, "\",EAST]]");
    p = add_epsg_def (first, last, 2127, "epsg", 2127,
		      "NZGD2000 / Lindis Peak 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-44.735 +lon_0=169.4675 +k=1 +x_0=400");
    add_proj4text (p, 1,
		   "000 +y_0=800000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +uni");
    add_proj4text (p, 2, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Lindis Peak 2000\",GEOGCS[\"NZGD2000");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",-44.735],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"central_meridian\",169.4675],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",1],PARAMETER[\"false_easting\",400000],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_northing\",800000],AUTHORITY[\"EPSG\",\"2127\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 2128, "epsg", 2128,
		      "NZGD2000 / Mount Nicholas 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-45.13277777777778 +lon_0=168.3986111");
    add_proj4text (p, 1,
		   "111111 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Mount Nicholas 2000\",GEOGCS[\"NZGD2");
    add_srs_wkt (p, 1,
		 "000\",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"616");
    add_srs_wkt (p, 4,
		 "7\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",-45.132777");
    add_srs_wkt (p, 9,
		 "77777778],PARAMETER[\"central_meridian\",168.39861111111");
    add_srs_wkt (p, 10,
		 "11],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 11,
		 "ing\",400000],PARAMETER[\"false_northing\",800000],AUTHO");
    add_srs_wkt (p, 12,
		 "RITY[\"EPSG\",\"2128\"],AXIS[\"Northing\",NORTH],AXIS[\"");
    add_srs_wkt (p, 13, "Easting\",EAST]]");
    p = add_epsg_def (first, last, 2129, "epsg", 2129,
		      "NZGD2000 / Mount York 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-45.56361111111111 +lon_0=167.7386111");
    add_proj4text (p, 1,
		   "111111 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Mount York 2000\",GEOGCS[\"NZGD2000\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GR");
    add_srs_wkt (p, 2,
		 "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019");
    add_srs_wkt (p, 3,
		 "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",-45.563611111");
    add_srs_wkt (p, 9,
		 "11111],PARAMETER[\"central_meridian\",167.7386111111111]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting");
    add_srs_wkt (p, 11,
		 "\",400000],PARAMETER[\"false_northing\",800000],AUTHORIT");
    add_srs_wkt (p, 12,
		 "Y[\"EPSG\",\"2129\"],AXIS[\"Northing\",NORTH],AXIS[\"Eas");
    add_srs_wkt (p, 13, "ting\",EAST]]");
    p = add_epsg_def (first, last, 2130, "epsg", 2130,
		      "NZGD2000 / Observation Point 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-45.81611111111111 +lon_0=170.6283333");
    add_proj4text (p, 1,
		   "333333 +k=1 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Observation Point 2000\",GEOGCS[\"NZ");
    add_srs_wkt (p, 1,
		 "GD2000\",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 8,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",-45.81");
    add_srs_wkt (p, 9,
		 "611111111111],PARAMETER[\"central_meridian\",170.6283333");
    add_srs_wkt (p, 10,
		 "333333],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",400000],PARAMETER[\"false_northing\",800000],A");
    add_srs_wkt (p, 12,
		 "UTHORITY[\"EPSG\",\"2130\"],AXIS[\"Northing\",NORTH],AXI");
    add_srs_wkt (p, 13, "S[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 2131, "epsg", 2131,
		      "NZGD2000 / North Taieri 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-45.86138888888889 +lon_0=170.2825 +k");
    add_proj4text (p, 1,
		   "=0.99996 +x_0=400000 +y_0=800000 +ellps=GRS80 +towgs84=0");
    add_proj4text (p, 2, ",0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / North Taieri 2000\",GEOGCS[\"NZGD200");
    add_srs_wkt (p, 1,
		 "0\",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",-45.861388888");
    add_srs_wkt (p, 9,
		 "88889],PARAMETER[\"central_meridian\",170.2825],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"scale_factor\",0.99996],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 11,
		 "400000],PARAMETER[\"false_northing\",800000],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"2131\"],AXIS[\"Northing\",NORTH],AXIS[\"Easting");
    add_srs_wkt (p, 13, "\",EAST]]");
    p = add_epsg_def (first, last, 2132, "epsg", 2132, "NZGD2000 / Bluff 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-46.6 +lon_0=168.3427777777778 +k=1 +");
    add_proj4text (p, 1,
		   "x_0=400000 +y_0=800000 +ellps=GRS80 +towgs84=0,0,0,0,0,0");
    add_proj4text (p, 2, ",0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Bluff 2000\",GEOGCS[\"NZGD2000\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 2,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 3,
		 ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6,
		 "\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",-46.6],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",168.3427777777778],PARAMETER[\"scal");
    add_srs_wkt (p, 10,
		 "e_factor\",1],PARAMETER[\"false_easting\",400000],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",800000],AUTHORITY[\"EPSG\",\"2132");
    add_srs_wkt (p, 12, "\"],AXIS[\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 2133, "epsg", 2133,
		      "NZGD2000 / UTM zone 58S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / UTM zone 58S\",GEOGCS[\"NZGD2000\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GRS ");
    add_srs_wkt (p, 2,
		 "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",165],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 10,
		 "6],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_northing\",10000000],AUTHORITY[\"EPSG\",\"2133\"],AXIS[");
    add_srs_wkt (p, 12, "\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2134, "epsg", 2134,
		      "NZGD2000 / UTM zone 59S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=59 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / UTM zone 59S\",GEOGCS[\"NZGD2000\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GRS ");
    add_srs_wkt (p, 2,
		 "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",171],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 10,
		 "6],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_northing\",10000000],AUTHORITY[\"EPSG\",\"2134\"],AXIS[");
    add_srs_wkt (p, 12, "\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2135, "epsg", 2135,
		      "NZGD2000 / UTM zone 60S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=60 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / UTM zone 60S\",GEOGCS[\"NZGD2000\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GRS ");
    add_srs_wkt (p, 2,
		 "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",177],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 10,
		 "6],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_northing\",10000000],AUTHORITY[\"EPSG\",\"2135\"],AXIS[");
    add_srs_wkt (p, 12, "\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2136, "epsg", 2136,
		      "Accra / Ghana National Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=4.666666666666667 +lon_0=-1 +k=0.9997");
    add_proj4text (p, 1,
		   "5 +x_0=274319.7391633579 +y_0=0 +a=6378300 +b=6356751.68");
    add_proj4text (p, 2, "9189189 +to_meter=0.3047997101815088 +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Accra / Ghana National Grid\",GEOGCS[\"Accra\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Accra\",SPHEROID[\"War Office\",6378300,296,AUTH");
    add_srs_wkt (p, 2,
		 "ORITY[\"EPSG\",\"7029\"]],AUTHORITY[\"EPSG\",\"6168\"]],");
    add_srs_wkt (p, 3,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 4,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 5,
		 "22\"]],AUTHORITY[\"EPSG\",\"4168\"]],UNIT[\"Gold Coast f");
    add_srs_wkt (p, 6,
		 "oot\",0.3047997101815088,AUTHORITY[\"EPSG\",\"9094\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 8,
		 "of_origin\",4.666666666666667],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",-1],PARAMETER[\"scale_factor\",0.99975],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",900000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"2136\"],AXIS[\"Easting\",EAST],AX");
    add_srs_wkt (p, 12, "IS[\"Northing\",NORTH]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_01 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 2137, "epsg", 2137, "Accra / TM 1 NW");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-1 +k=0.9996 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=0 +a=6378300 +b=6356751.689189189 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Accra / TM 1 NW\",GEOGCS[\"Accra\",DATUM[\"Accr");
    add_srs_wkt (p, 1,
		 "a\",SPHEROID[\"War Office\",6378300,296,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 2,
		 "\",\"7029\"]],AUTHORITY[\"EPSG\",\"6168\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 3,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 4,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"4168\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 7,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 8,
		 "dian\",-1],PARAMETER[\"scale_factor\",0.9996],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 10,
		 "],AUTHORITY[\"EPSG\",\"2137\"],AXIS[\"Easting\",EAST],AX");
    add_srs_wkt (p, 11, "IS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2138, "epsg", 2138,
		      "NAD27(CGQ77) / Quebec Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=60 +lat_2=46 +lat_0=44 +lon_0=-68.5 +x_");
    add_proj4text (p, 1, "0=0 +y_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27(CGQ77) / Quebec Lambert\",GEOGCS[\"NAD27(");
    add_srs_wkt (p, 1,
		 "CGQ77)\",DATUM[\"North_American_Datum_1927_CGQ77\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4609\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_C");
    add_srs_wkt (p, 8,
		 "onic_2SP\"],PARAMETER[\"standard_parallel_1\",60],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"standard_parallel_2\",46],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 10,
		 "origin\",44],PARAMETER[\"central_meridian\",-68.5],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_easting\",0],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"2138\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 13, "\",NORTH]]");
    p = add_epsg_def (first, last, 2139, "epsg", 2139,
		      "NAD83(CSRS98) / SCoPQ zone 2 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-55.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1,
		   "+y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / SCoPQ zone 2 (deprecated)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(CSRS98)\",DATUM[\"NAD83_Canadian_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4140\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 8,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 9,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",-55.5],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_eas");
    add_srs_wkt (p, 11,
		 "ting\",304800],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 12,
		 "[\"EPSG\",\"2139\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2140, "epsg", 2140,
		      "NAD83(CSRS98) / MTM zone 3 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-58.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1,
		   "+y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / MTM zone 3 (deprecated)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(CSRS98)\",DATUM[\"NAD83_Canadian_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "9433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4140\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-58.5],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",304800],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"2140\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 2141, "epsg", 2141,
		      "NAD83(CSRS98) / MTM zone 4 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-61.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1,
		   "+y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / MTM zone 4 (deprecated)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(CSRS98)\",DATUM[\"NAD83_Canadian_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "9433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4140\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-61.5],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",304800],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"2141\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 2142, "epsg", 2142,
		      "NAD83(CSRS98) / MTM zone 5 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-64.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1,
		   "+y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / MTM zone 5 (deprecated)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(CSRS98)\",DATUM[\"NAD83_Canadian_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "9433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4140\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-64.5],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",304800],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"2142\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 2143, "epsg", 2143,
		      "NAD83(CSRS98) / MTM zone 6 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-67.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1,
		   "+y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / MTM zone 6 (deprecated)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(CSRS98)\",DATUM[\"NAD83_Canadian_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "9433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4140\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-67.5],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",304800],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"2143\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 2144, "epsg", 2144,
		      "NAD83(CSRS98) / MTM zone 7 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-70.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1,
		   "+y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / MTM zone 7 (deprecated)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(CSRS98)\",DATUM[\"NAD83_Canadian_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "9433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4140\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-70.5],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",304800],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"2144\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 2145, "epsg", 2145,
		      "NAD83(CSRS98) / MTM zone 8 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-73.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1,
		   "+y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / MTM zone 8 (deprecated)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(CSRS98)\",DATUM[\"NAD83_Canadian_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "9433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4140\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-73.5],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",304800],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"2145\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 2146, "epsg", 2146,
		      "NAD83(CSRS98) / MTM zone 9 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-76.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1,
		   "+y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / MTM zone 9 (deprecated)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(CSRS98)\",DATUM[\"NAD83_Canadian_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "9433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4140\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-76.5],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",304800],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"2146\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 2147, "epsg", 2147,
		      "NAD83(CSRS98) / MTM zone 10 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-79.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1,
		   "+y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / MTM zone 10 (deprecated)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NAD83(CSRS98)\",DATUM[\"NAD83_Canadian_Spatial_Refe");
    add_srs_wkt (p, 2,
		 "rence_System\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 3,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 6,
		 "99433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4140\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-79.5],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",304800],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"2147\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 2148, "epsg", 2148,
		      "NAD83(CSRS98) / UTM zone 21N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / UTM zone 21N (deprecated)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(CSRS98)\",DATUM[\"NAD83_Canadian_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4140\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 8,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 9,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",-57],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"2148\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 2149, "epsg", 2149,
		      "NAD83(CSRS98) / UTM zone 18N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / UTM zone 18N (deprecated)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(CSRS98)\",DATUM[\"NAD83_Canadian_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4140\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 8,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 9,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",-75],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"2149\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 2150, "epsg", 2150,
		      "NAD83(CSRS98) / UTM zone 17N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=17 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / UTM zone 17N (deprecated)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(CSRS98)\",DATUM[\"NAD83_Canadian_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4140\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 8,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 9,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",-81],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"2150\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 2151, "epsg", 2151,
		      "NAD83(CSRS98) / UTM zone 13N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=13 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / UTM zone 13N (deprecated)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(CSRS98)\",DATUM[\"NAD83_Canadian_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4140\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 8,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 9,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",-105],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 11,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 12,
		 "\"EPSG\",\"2151\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 13, "g\",NORTH]]");
    p = add_epsg_def (first, last, 2152, "epsg", 2152,
		      "NAD83(CSRS98) / UTM zone 12N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=12 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / UTM zone 12N (deprecated)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(CSRS98)\",DATUM[\"NAD83_Canadian_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4140\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 8,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 9,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",-111],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 11,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 12,
		 "\"EPSG\",\"2152\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 13, "g\",NORTH]]");
    p = add_epsg_def (first, last, 2153, "epsg", 2153,
		      "NAD83(CSRS98) / UTM zone 11N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=11 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / UTM zone 11N (deprecated)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(CSRS98)\",DATUM[\"NAD83_Canadian_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4140\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 8,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 9,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",-117],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 11,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 12,
		 "\"EPSG\",\"2153\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 13, "g\",NORTH]]");
    p = add_epsg_def (first, last, 2154, "epsg", 2154, "RGF93 / Lambert-93");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49 +lat_2=44 +lat_0=46.5 +lon_0=3 +x_0=");
    add_proj4text (p, 1,
		   "700000 +y_0=6600000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGF93 / Lambert-93\",GEOGCS[\"RGF93\",DATUM[\"R");
    add_srs_wkt (p, 1,
		 "eseau_Geodesique_Francais_1993\",SPHEROID[\"GRS 1980\",6");
    add_srs_wkt (p, 2,
		 "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS");
    add_srs_wkt (p, 3,
		 "84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4171\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",49],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_2\",44],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 10,
		 "\",46.5],PARAMETER[\"central_meridian\",3],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",700000],PARAMETER[\"false_northing\",6600");
    add_srs_wkt (p, 12,
		 "000],AUTHORITY[\"EPSG\",\"2154\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 13, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2155, "epsg", 2155,
		      "American Samoa 1962 / American Samoa Lambert (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-14.26666666666667 +lat_0=-14.266666666");
    add_proj4text (p, 1,
		   "66667 +lon_0=170 +k_0=1 +x_0=152400.3048006096 +y_0=0 +e");
    add_proj4text (p, 2,
		   "llps=clrk66 +towgs84=-115,118,426,0,0,0,0 +units=us-ft +");
    add_proj4text (p, 3, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"American Samoa 1962 / American Samoa Lambert (d");
    add_srs_wkt (p, 1,
		 "eprecated)\",GEOGCS[\"American Samoa 1962\",DATUM[\"Amer");
    add_srs_wkt (p, 2,
		 "ican_Samoa_1962\",SPHEROID[\"Clarke 1866\",6378206.4,294");
    add_srs_wkt (p, 3,
		 ".9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-11");
    add_srs_wkt (p, 4,
		 "5,118,426,0,0,0,0],AUTHORITY[\"EPSG\",\"6169\"]],PRIMEM[");
    add_srs_wkt (p, 5,
		 "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg");
    add_srs_wkt (p, 6,
		 "ree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"4169\"]],UNIT[\"US survey foot\",0.");
    add_srs_wkt (p, 8,
		 "3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTIO");
    add_srs_wkt (p, 9,
		 "N[\"Lambert_Conformal_Conic_1SP\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 10,
		 "of_origin\",-14.26666666666667],PARAMETER[\"central_meri");
    add_srs_wkt (p, 11,
		 "dian\",170],PARAMETER[\"scale_factor\",1],PARAMETER[\"fa");
    add_srs_wkt (p, 12,
		 "lse_easting\",500000],PARAMETER[\"false_northing\",0],AU");
    add_srs_wkt (p, 13,
		 "THORITY[\"EPSG\",\"2155\"],AXIS[\"X\",EAST],AXIS[\"Y\",N");
    add_srs_wkt (p, 14, "ORTH]]");
    p = add_epsg_def (first, last, 2156, "epsg", 2156,
		      "NAD83(HARN) / UTM zone 59S (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=59 +south +ellps=GRS80 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / UTM zone 59S (deprecated)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_N");
    add_srs_wkt (p, 2,
		 "etwork\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]]");
    add_srs_wkt (p, 4,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 5,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 6,
		 "122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",171],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 10,
		 "6],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_northing\",10000000],AUTHORITY[\"EPSG\",\"2156\"],AXIS[");
    add_srs_wkt (p, 12, "\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2157, "epsg", 2157,
		      "IRENET95 / Irish Transverse Mercator");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=53.5 +lon_0=-8 +k=0.99982 +x_0=600000");
    add_proj4text (p, 1,
		   " +y_0=750000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IRENET95 / Irish Transverse Mercator\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "IRENET95\",DATUM[\"IRENET95\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6173\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4173\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",53.5],PARAMETER[\"central_m");
    add_srs_wkt (p, 9,
		 "eridian\",-8],PARAMETER[\"scale_factor\",0.99982],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",600000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 11,
		 "\",750000],AUTHORITY[\"EPSG\",\"2157\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 12, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2158, "epsg", 2158,
		      "IRENET95 / UTM zone 29N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=29 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IRENET95 / UTM zone 29N\",GEOGCS[\"IRENET95\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"IRENET95\",SPHEROID[\"GRS 1980\",6378137,298.2572");
    add_srs_wkt (p, 2,
		 "22101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,");
    add_srs_wkt (p, 3,
		 "0],AUTHORITY[\"EPSG\",\"6173\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4173\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"lat");
    add_srs_wkt (p, 8,
		 "itude_of_origin\",0],PARAMETER[\"central_meridian\",-9],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"2158\"],AXIS[\"Easting\",EAST],AXIS[\"Northi");
    add_srs_wkt (p, 12, "ng\",NORTH]]");
    p = add_epsg_def (first, last, 2159, "epsg", 2159,
		      "Sierra Leone 1924 / New Colony Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=6.666666666666667 +lon_0=-12 +k=1 +x_");
    add_proj4text (p, 1,
		   "0=152399.8550907544 +y_0=0 +a=6378300 +b=6356751.6891891");
    add_proj4text (p, 2, "89 +to_meter=0.3047997101815088 +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Sierra Leone 1924 / New Colony Grid\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Sierra Leone 1924\",DATUM[\"Sierra_Leone_Colony_1924\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"War Office\",6378300,296,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7029\"]],AUTHORITY[\"EPSG\",\"6174\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4174\"]],UNIT[\"Gold Coast foot\",0.304799710");
    add_srs_wkt (p, 7,
		 "1815088,AUTHORITY[\"EPSG\",\"9094\"]],PROJECTION[\"Trans");
    add_srs_wkt (p, 8,
		 "verse_Mercator\"],PARAMETER[\"latitude_of_origin\",6.666");
    add_srs_wkt (p, 9,
		 "666666666667],PARAMETER[\"central_meridian\",-12],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12,
		 "2159\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 2160, "epsg", 2160,
		      "Sierra Leone 1924 / New War Office Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=6.666666666666667 +lon_0=-12 +k=1 +x_");
    add_proj4text (p, 1,
		   "0=243839.7681452071 +y_0=182879.8261089053 +a=6378300 +b");
    add_proj4text (p, 2,
		   "=6356751.689189189 +to_meter=0.3047997101815088 +no_defs");
    add_proj4text (p, 3, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Sierra Leone 1924 / New War Office Grid\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"Sierra Leone 1924\",DATUM[\"Sierra_Leone_Colony_1924");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"War Office\",6378300,296,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7029\"]],AUTHORITY[\"EPSG\",\"6174\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4174\"]],UNIT[\"Gold Coast foot\",0.304799");
    add_srs_wkt (p, 7,
		 "7101815088,AUTHORITY[\"EPSG\",\"9094\"]],PROJECTION[\"Tr");
    add_srs_wkt (p, 8,
		 "ansverse_Mercator\"],PARAMETER[\"latitude_of_origin\",6.");
    add_srs_wkt (p, 9,
		 "666666666666667],PARAMETER[\"central_meridian\",-12],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",8");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",600000],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"2160\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 2161, "epsg", 2161,
		      "Sierra Leone 1968 / UTM zone 28N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=28 +ellps=clrk80 +towgs84=-88,4,101,0,0,");
    add_proj4text (p, 1, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Sierra Leone 1968 / UTM zone 28N\",GEOGCS[\"Sie");
    add_srs_wkt (p, 1,
		 "rra Leone 1968\",DATUM[\"Sierra_Leone_1968\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7012\"]],TOWGS84[-88,4,101,0,0,0,0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"6175\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4175\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 8,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 9,
		 "0],PARAMETER[\"central_meridian\",-15],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.9996],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2161\"");
    add_srs_wkt (p, 12, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2162, "epsg", 2162,
		      "Sierra Leone 1968 / UTM zone 29N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=29 +ellps=clrk80 +towgs84=-88,4,101,0,0,");
    add_proj4text (p, 1, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Sierra Leone 1968 / UTM zone 29N\",GEOGCS[\"Sie");
    add_srs_wkt (p, 1,
		 "rra Leone 1968\",DATUM[\"Sierra_Leone_1968\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7012\"]],TOWGS84[-88,4,101,0,0,0,0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"6175\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4175\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 8,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 9,
		 "0],PARAMETER[\"central_meridian\",-9],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.9996],PARAMETER[\"false_easting\",500000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2162\"");
    add_srs_wkt (p, 12, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2163, "epsg", 2163, "unnamed");
    add_proj4text (p, 0,
		   "+proj=laea +lat_0=45 +lon_0=-100 +x_0=0 +y_0=0 +a=637099");
    add_proj4text (p, 1, "7 +b=6370997 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"unnamed\",GEOGCS[\"unnamed ellipse\",DATUM[\"un");
    add_srs_wkt (p, 1,
		 "known\",SPHEROID[\"unnamed\",6370997,0]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 2,
		 "ich\",0],UNIT[\"degree\",0.0174532925199433]],PROJECTION");
    add_srs_wkt (p, 3,
		 "[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 4,
		 "of_center\",45],PARAMETER[\"longitude_of_center\",-100],");
    add_srs_wkt (p, 5,
		 "PARAMETER[\"false_easting\",0],PARAMETER[\"false_northin");
    add_srs_wkt (p, 6,
		 "g\",0],UNIT[\"Meter\",1],AUTHORITY[\"EPSG\",\"2163\"]]");
    p = add_epsg_def (first, last, 2164, "epsg", 2164,
		      "Locodjo 1965 / TM 5 NW");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-5 +k=0.9996 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=0 +ellps=clrk80 +towgs84=-125,53,467,0,0,0,0 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Locodjo 1965 / TM 5 NW\",GEOGCS[\"Locodjo 1965\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Locodjo_1965\",SPHEROID[\"Clarke 1880 (RGS)\",6");
    add_srs_wkt (p, 2,
		 "378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84");
    add_srs_wkt (p, 3,
		 "[-125,53,467,0,0,0,0],AUTHORITY[\"EPSG\",\"6142\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4142\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",-5],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 11,
		 "hing\",0],AUTHORITY[\"EPSG\",\"2164\"],AXIS[\"Easting\",");
    add_srs_wkt (p, 12, "EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2165, "epsg", 2165,
		      "Abidjan 1987 / TM 5 NW");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-5 +k=0.9996 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=0 +ellps=clrk80 +towgs84=-124.76,53,466.79,0,0,0,0 +un");
    add_proj4text (p, 2, "its=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Abidjan 1987 / TM 5 NW\",GEOGCS[\"Abidjan 1987\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Abidjan_1987\",SPHEROID[\"Clarke 1880 (RGS)\",6");
    add_srs_wkt (p, 2,
		 "378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84");
    add_srs_wkt (p, 3,
		 "[-124.76,53,466.79,0,0,0,0],AUTHORITY[\"EPSG\",\"6143\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4143\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-5],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 10,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"2165\"],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2166, "epsg", 2166,
		      "Pulkovo 1942(83) / Gauss Kruger zone 3 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=9 +k=1 +x_0=3500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(83) / Gauss Kruger zone 3 (depreca");
    add_srs_wkt (p, 1,
		 "ted)\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo_1942_");
    add_srs_wkt (p, 2,
		 "83\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6178\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4178\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",9],PARAMETER[\"scale_factor\",1],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_easting\",3500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"2166\"],AXIS[\"X\",NORTH],AXIS[");
    add_srs_wkt (p, 12, "\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2167, "epsg", 2167,
		      "Pulkovo 1942(83) / Gauss Kruger zone 4 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=12 +k=1 +x_0=4500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(83) / Gauss Kruger zone 4 (depreca");
    add_srs_wkt (p, 1,
		 "ted)\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo_1942_");
    add_srs_wkt (p, 2,
		 "83\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6178\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4178\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",12],PARAMETER[\"scale_factor\",1],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",4500000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 11,
		 "\",0],AUTHORITY[\"EPSG\",\"2167\"],AXIS[\"X\",NORTH],AXI");
    add_srs_wkt (p, 12, "S[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2168, "epsg", 2168,
		      "Pulkovo 1942(83) / Gauss Kruger zone 5 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=1 +x_0=5500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(83) / Gauss Kruger zone 5 (depreca");
    add_srs_wkt (p, 1,
		 "ted)\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo_1942_");
    add_srs_wkt (p, 2,
		 "83\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6178\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4178\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",15],PARAMETER[\"scale_factor\",1],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",5500000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 11,
		 "\",0],AUTHORITY[\"EPSG\",\"2168\"],AXIS[\"X\",NORTH],AXI");
    add_srs_wkt (p, 12, "S[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2169, "epsg", 2169,
		      "Luxembourg 1930 / Gauss");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=49.83333333333334 +lon_0=6.1666666666");
    add_proj4text (p, 1,
		   "66667 +k=1 +x_0=80000 +y_0=100000 +ellps=intl +towgs84=-");
    add_proj4text (p, 2,
		   "193,13.7,-39.3,-0.41,-2.933,2.688,0.43 +units=m +no_defs");
    add_proj4text (p, 3, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Luxembourg 1930 / Gauss\",GEOGCS[\"Luxembourg 1");
    add_srs_wkt (p, 1,
		 "930\",DATUM[\"Luxembourg_1930\",SPHEROID[\"International");
    add_srs_wkt (p, 2,
		 " 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS8");
    add_srs_wkt (p, 3,
		 "4[-193,13.7,-39.3,-0.41,-2.933,2.688,0.43],AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"6181\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 5,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4181\"]],UN");
    add_srs_wkt (p, 7,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 9,
		 ",49.83333333333334],PARAMETER[\"central_meridian\",6.166");
    add_srs_wkt (p, 10,
		 "666666666667],PARAMETER[\"scale_factor\",1],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",80000],PARAMETER[\"false_northing\",1000");
    add_srs_wkt (p, 12,
		 "00],AUTHORITY[\"EPSG\",\"2169\"],AXIS[\"X\",NORTH],AXIS[");
    add_srs_wkt (p, 13, "\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2170, "epsg", 2170,
		      "MGI / Slovenia Grid (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=0 +ellps=bessel +datum=hermannskogel +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / Slovenia Grid (deprecated)\",GEOGCS[\"MGI");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"Militar_Geographische_Institute\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7004\"]],TOWGS84[577.326,90.129,463.919,5.137,1.474,5");
    add_srs_wkt (p, 4,
		 ".297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 5,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 6,
		 "0.0174532925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"4312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 8,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 10,
		 "an\",15],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"2170\"],AXIS[\"X\",NORTH],AXIS[\"Y\"");
    add_srs_wkt (p, 13, ",EAST]]");
    p = add_epsg_def (first, last, 2171, "epsg", 2171,
		      "Pulkovo 1942(58) / Poland zone I (deprecated)");
    add_proj4text (p, 0,
		   "+proj=sterea +lat_0=50.625 +lon_0=21.08333333333333 +k=0");
    add_proj4text (p, 1,
		   ".9998 +x_0=4637000 +y_0=5647000 +ellps=krass +units=m +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / Poland zone I (deprecated)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo_1942_58\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4179\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Oblique_Stereographic\"],");
    add_srs_wkt (p, 8,
		 "PARAMETER[\"latitude_of_origin\",50.625],PARAMETER[\"cen");
    add_srs_wkt (p, 9,
		 "tral_meridian\",21.08333333333333],PARAMETER[\"scale_fac");
    add_srs_wkt (p, 10,
		 "tor\",0.9998],PARAMETER[\"false_easting\",4637000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",5647000],AUTHORITY[\"EPSG\",\"21");
    add_srs_wkt (p, 12, "71\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2172, "epsg", 2172,
		      "Pulkovo 1942(58) / Poland zone II");
    add_proj4text (p, 0,
		   "+proj=sterea +lat_0=53.00194444444445 +lon_0=21.50277777");
    add_proj4text (p, 1,
		   "777778 +k=0.9998 +x_0=4603000 +y_0=5806000 +ellps=krass ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / Poland zone II\",GEOGCS[\"Pu");
    add_srs_wkt (p, 1,
		 "lkovo 1942(58)\",DATUM[\"Pulkovo_1942_58\",SPHEROID[\"Kr");
    add_srs_wkt (p, 2,
		 "assowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4179\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Oblique_Stereographic\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",53.00194444444445],PARAMETER[\"centr");
    add_srs_wkt (p, 9,
		 "al_meridian\",21.50277777777778],PARAMETER[\"scale_facto");
    add_srs_wkt (p, 10,
		 "r\",0.9998],PARAMETER[\"false_easting\",4603000],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",5806000],AUTHORITY[\"EPSG\",\"2172");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2173, "epsg", 2173,
		      "Pulkovo 1942(58) / Poland zone III");
    add_proj4text (p, 0,
		   "+proj=sterea +lat_0=53.58333333333334 +lon_0=17.00833333");
    add_proj4text (p, 1,
		   "333333 +k=0.9998 +x_0=3501000 +y_0=5999000 +ellps=krass ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / Poland zone III\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942(58)\",DATUM[\"Pulkovo_1942_58\",SPHEROID[\"K");
    add_srs_wkt (p, 2,
		 "rassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4179\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Oblique_Stereographic\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",53.58333333333334],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",17.00833333333333],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 10,
		 "or\",0.9998],PARAMETER[\"false_easting\",3501000],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",5999000],AUTHORITY[\"EPSG\",\"217");
    add_srs_wkt (p, 12, "3\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2174, "epsg", 2174,
		      "Pulkovo 1942(58) / Poland zone IV");
    add_proj4text (p, 0,
		   "+proj=sterea +lat_0=51.67083333333333 +lon_0=16.67222222");
    add_proj4text (p, 1,
		   "222222 +k=0.9998 +x_0=3703000 +y_0=5627000 +ellps=krass ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / Poland zone IV\",GEOGCS[\"Pu");
    add_srs_wkt (p, 1,
		 "lkovo 1942(58)\",DATUM[\"Pulkovo_1942_58\",SPHEROID[\"Kr");
    add_srs_wkt (p, 2,
		 "assowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4179\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Oblique_Stereographic\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",51.67083333333333],PARAMETER[\"centr");
    add_srs_wkt (p, 9,
		 "al_meridian\",16.67222222222222],PARAMETER[\"scale_facto");
    add_srs_wkt (p, 10,
		 "r\",0.9998],PARAMETER[\"false_easting\",3703000],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",5627000],AUTHORITY[\"EPSG\",\"2174");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2175, "epsg", 2175,
		      "Pulkovo 1942(58) / Poland zone V");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18.95833333333333 +k=0.99998");
    add_proj4text (p, 1,
		   "3 +x_0=237000 +y_0=-4700000 +ellps=krass +units=m +no_de");
    add_proj4text (p, 2, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / Poland zone V\",GEOGCS[\"Pul");
    add_srs_wkt (p, 1,
		 "kovo 1942(58)\",DATUM[\"Pulkovo_1942_58\",SPHEROID[\"Kra");
    add_srs_wkt (p, 2,
		 "ssowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4179\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"lat");
    add_srs_wkt (p, 8,
		 "itude_of_origin\",0],PARAMETER[\"central_meridian\",18.9");
    add_srs_wkt (p, 9,
		 "5833333333333],PARAMETER[\"scale_factor\",0.999983],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_easting\",237000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 11,
		 "ng\",-4700000],AUTHORITY[\"EPSG\",\"2175\"],AXIS[\"X\",N");
    add_srs_wkt (p, 12, "ORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2176, "epsg", 2176,
		      "ETRS89 / Poland CS2000 zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=0.999923 +x_0=5500000 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / Poland CS2000 zone 5\",GEOGCS[\"ETRS89");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"European_Terrestrial_Reference_System_1989\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",15],PARAMETER[\"scale_factor\",0.999923],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",5500000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 11,
		 "g\",0],AUTHORITY[\"EPSG\",\"2176\"],AXIS[\"x\",NORTH],AX");
    add_srs_wkt (p, 12, "IS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 2177, "epsg", 2177,
		      "ETRS89 / Poland CS2000 zone 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18 +k=0.999923 +x_0=6500000 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / Poland CS2000 zone 6\",GEOGCS[\"ETRS89");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"European_Terrestrial_Reference_System_1989\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",18],PARAMETER[\"scale_factor\",0.999923],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",6500000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 11,
		 "g\",0],AUTHORITY[\"EPSG\",\"2177\"],AXIS[\"x\",NORTH],AX");
    add_srs_wkt (p, 12, "IS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 2178, "epsg", 2178,
		      "ETRS89 / Poland CS2000 zone 7");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=0.999923 +x_0=7500000 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / Poland CS2000 zone 7\",GEOGCS[\"ETRS89");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"European_Terrestrial_Reference_System_1989\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",21],PARAMETER[\"scale_factor\",0.999923],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",7500000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 11,
		 "g\",0],AUTHORITY[\"EPSG\",\"2178\"],AXIS[\"x\",NORTH],AX");
    add_srs_wkt (p, 12, "IS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 2179, "epsg", 2179,
		      "ETRS89 / Poland CS2000 zone 8");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=24 +k=0.999923 +x_0=8500000 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / Poland CS2000 zone 8\",GEOGCS[\"ETRS89");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"European_Terrestrial_Reference_System_1989\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",24],PARAMETER[\"scale_factor\",0.999923],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",8500000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 11,
		 "g\",0],AUTHORITY[\"EPSG\",\"2179\"],AXIS[\"x\",NORTH],AX");
    add_srs_wkt (p, 12, "IS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 2180, "epsg", 2180, "ETRS89 / Poland CS92");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=19 +k=0.9993 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=-5300000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / Poland CS92\",GEOGCS[\"ETRS89\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"European_Terrestrial_Reference_System_1989\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",19],PARAMETER[\"scale_factor\",0.9993],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",500000],PARAMETER[\"false_northing\",-530000");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"2180\"],AXIS[\"x\",NORTH],AXIS[\"");
    add_srs_wkt (p, 12, "y\",EAST]]");
    p = add_epsg_def (first, last, 2188, "epsg", 2188,
		      "Azores Occidental 1939 / UTM zone 25N");
    add_proj4text (p, 0, "+proj=utm +zone=25 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Azores Occidental 1939 / UTM zone 25N\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"Azores Occidental 1939\",DATUM[\"Azores_Occidental_Isl");
    add_srs_wkt (p, 2,
		 "ands_1939\",SPHEROID[\"International 1924\",6378388,297,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6182\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4182\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-33],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"2188\"],AXIS[\"Eas");
    add_srs_wkt (p, 12, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2189, "epsg", 2189,
		      "Azores Central 1948 / UTM zone 26N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=26 +ellps=intl +towgs84=-104,167,-38,0,0");
    add_proj4text (p, 1, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Azores Central 1948 / UTM zone 26N\",GEOGCS[\"A");
    add_srs_wkt (p, 1,
		 "zores Central 1948\",DATUM[\"Azores_Central_Islands_1948");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7022\"]],TOWGS84[-104,167,-38,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6183\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"418");
    add_srs_wkt (p, 7,
		 "3\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",0],PARAMETER[\"central_meridian\",-27],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12,
		 "\",\"2189\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NO");
    add_srs_wkt (p, 13, "RTH]]");
    p = add_epsg_def (first, last, 2190, "epsg", 2190,
		      "Azores Oriental 1940 / UTM zone 26N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=26 +ellps=intl +towgs84=-203,141,53,0,0,");
    add_proj4text (p, 1, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Azores Oriental 1940 / UTM zone 26N\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Azores Oriental 1940\",DATUM[\"Azores_Oriental_Islands_1");
    add_srs_wkt (p, 2,
		 "940\",SPHEROID[\"International 1924\",6378388,297,AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"7022\"]],TOWGS84[-203,141,53,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6184\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "184\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",-27],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12,
		 "G\",\"2190\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 2191, "epsg", 2191,
		      "Madeira 1936 / UTM zone 28N (deprecated)");
    add_proj4text (p, 0, "+proj=utm +zone=28 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Madeira 1936 / UTM zone 28N (deprecated)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"Madeira 1936\",DATUM[\"Madeira_1936\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6185\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "2925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 6,
		 "G\",\"4185\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 7,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 8,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-15");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 10,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 11,
		 "TY[\"EPSG\",\"2191\"],AXIS[\"Easting\",EAST],AXIS[\"Nort");
    add_srs_wkt (p, 12, "hing\",NORTH]]");
    p = add_epsg_def (first, last, 2192, "epsg", 2192,
		      "ED50 / France EuroLambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46.8 +lat_0=46.8 +lon_0=2.3372291666666");
    add_proj4text (p, 1,
		   "67 +k_0=0.99987742 +x_0=600000 +y_0=2200000 +ellps=intl ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / France EuroLambert\",GEOGCS[\"ED50\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"European_Datum_1950\",SPHEROID[\"International 1924");
    add_srs_wkt (p, 2,
		 "\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Lambert_Conformal_Conic_1SP\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 8,
		 "f_origin\",46.8],PARAMETER[\"central_meridian\",2.337229");
    add_srs_wkt (p, 9,
		 "166666667],PARAMETER[\"scale_factor\",0.99987742],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",600000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 11,
		 "\",2200000],AUTHORITY[\"EPSG\",\"2192\"],AXIS[\"X\",EAST");
    add_srs_wkt (p, 12, "],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2193, "epsg", 2193,
		      "NZGD2000 / New Zealand Transverse Mercator 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=173 +k=0.9996 +x_0=1600000 +");
    add_proj4text (p, 1,
		   "y_0=10000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / New Zealand Transverse Mercator 2000");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"NZGD2000\",DATUM[\"New_Zealand_Geodetic_Datu");
    add_srs_wkt (p, 2,
		 "m_2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4167");
    add_srs_wkt (p, 7,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 8,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 9,
		 "origin\",0],PARAMETER[\"central_meridian\",173],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"scale_factor\",0.9996],PARAMETER[\"false_easting\",1");
    add_srs_wkt (p, 11,
		 "600000],PARAMETER[\"false_northing\",10000000],AUTHORITY");
    add_srs_wkt (p, 12,
		 "[\"EPSG\",\"2193\"],AXIS[\"Northing\",NORTH],AXIS[\"East");
    add_srs_wkt (p, 13, "ing\",EAST]]");
    p = add_epsg_def (first, last, 2194, "epsg", 2194,
		      "American Samoa 1962 / American Samoa Lambert (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-14.26666666666667 +lat_0=-14.266666666");
    add_proj4text (p, 1,
		   "66667 +lon_0=-170 +k_0=1 +x_0=152400.3048006096 +y_0=0 +");
    add_proj4text (p, 2,
		   "ellps=clrk66 +towgs84=-115,118,426,0,0,0,0 +units=us-ft ");
    add_proj4text (p, 3, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"American Samoa 1962 / American Samoa Lambert (d");
    add_srs_wkt (p, 1,
		 "eprecated)\",GEOGCS[\"American Samoa 1962\",DATUM[\"Amer");
    add_srs_wkt (p, 2,
		 "ican_Samoa_1962\",SPHEROID[\"Clarke 1866\",6378206.4,294");
    add_srs_wkt (p, 3,
		 ".9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-11");
    add_srs_wkt (p, 4,
		 "5,118,426,0,0,0,0],AUTHORITY[\"EPSG\",\"6169\"]],PRIMEM[");
    add_srs_wkt (p, 5,
		 "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg");
    add_srs_wkt (p, 6,
		 "ree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"4169\"]],UNIT[\"US survey foot\",0.");
    add_srs_wkt (p, 8,
		 "3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTIO");
    add_srs_wkt (p, 9,
		 "N[\"Lambert_Conformal_Conic_1SP\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 10,
		 "of_origin\",-14.26666666666667],PARAMETER[\"central_meri");
    add_srs_wkt (p, 11,
		 "dian\",-170],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 12,
		 "alse_easting\",500000],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 13,
		 "UTHORITY[\"EPSG\",\"2194\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 14, "NORTH]]");
    p = add_epsg_def (first, last, 2195, "epsg", 2195,
		      "NAD83(HARN) / UTM zone 2S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=2 +south +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / UTM zone 2S\",GEOGCS[\"NAD83(HARN");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-171],PARAMETER[\"scale_factor\",0.9996],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",1");
    add_srs_wkt (p, 11,
		 "0000000],AUTHORITY[\"EPSG\",\"2195\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2196, "epsg", 2196,
		      "ETRS89 / Kp2000 Jutland");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=9.5 +k=0.99995 +x_0=200000 +");
    add_proj4text (p, 1, "y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / Kp2000 Jutland\",GEOGCS[\"ETRS89\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"European_Terrestrial_Reference_System_1989\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",9.5],PARAMETER[\"scale_factor\",0.99995],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",200000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"2196\"],AXIS[\"Easting\",EAST],AX");
    add_srs_wkt (p, 12, "IS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2197, "epsg", 2197,
		      "ETRS89 / Kp2000 Zealand");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=12 +k=0.99995 +x_0=500000 +y");
    add_proj4text (p, 1, "_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / Kp2000 Zealand\",GEOGCS[\"ETRS89\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"European_Terrestrial_Reference_System_1989\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",12],PARAMETER[\"scale_factor\",0.99995],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"2197\"],AXIS[\"Easting\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2198, "epsg", 2198,
		      "ETRS89 / Kp2000 Bornholm");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=1 +x_0=900000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / Kp2000 Bornholm\",GEOGCS[\"ETRS89\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"European_Terrestrial_Reference_System_1989\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",15],PARAMETER[\"scale_factor\",1],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",900000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 11,
		 "RITY[\"EPSG\",\"2198\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 12, "rthing\",NORTH]]");
    p = add_epsg_def (first, last, 2199, "epsg", 2199,
		      "Albanian 1987 / Gauss Kruger zone 4 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=1 +x_0=4500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Albanian 1987 / Gauss Kruger zone 4 (deprecated");
    add_srs_wkt (p, 1,
		 ")\",GEOGCS[\"Albanian 1987\",DATUM[\"Albanian_1987\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7024\"]],AUTHORITY[\"EPSG\",\"6191\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4191\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",21],PARAMETER[\"scale_factor\",1],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_easting\",4500000],PARAMETER[\"false_northing\",0],AU");
    add_srs_wkt (p, 11,
		 "THORITY[\"EPSG\",\"2199\"],AXIS[\"X\",NORTH],AXIS[\"Y\",");
    add_srs_wkt (p, 12, "EAST]]");
    p = add_epsg_def (first, last, 2200, "epsg", 2200,
		      "ATS77 / New Brunswick Stereographic (ATS77)");
    add_proj4text (p, 0,
		   "+proj=sterea +lat_0=46.5 +lon_0=-66.5 +k=0.999912 +x_0=3");
    add_proj4text (p, 1,
		   "00000 +y_0=800000 +a=6378135 +b=6356750.304921594 +units");
    add_proj4text (p, 2, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ATS77 / New Brunswick Stereographic (ATS77)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"ATS77\",DATUM[\"Average_Terrestrial_System_1977\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Average Terrestrial System 1977\",6378135,29");
    add_srs_wkt (p, 3,
		 "8.257,AUTHORITY[\"EPSG\",\"7041\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "6122\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4122\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Obliqu");
    add_srs_wkt (p, 8,
		 "e_Stereographic\"],PARAMETER[\"latitude_of_origin\",46.5");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-66.5],PARAMETER[\"scal");
    add_srs_wkt (p, 10,
		 "e_factor\",0.999912],PARAMETER[\"false_easting\",300000]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_northing\",800000],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12,
		 ",\"2200\"],AXIS[\"Northing\",NORTH],AXIS[\"Easting\",EAS");
    add_srs_wkt (p, 13, "T]]");
    p = add_epsg_def (first, last, 2201, "epsg", 2201, "REGVEN / UTM zone 18N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"REGVEN / UTM zone 18N\",GEOGCS[\"REGVEN\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Red_Geodesica_Venezolana\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6189\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4189\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-75],PARAMETER[\"scale_factor\",0.9996],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",500000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"2201\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2202, "epsg", 2202, "REGVEN / UTM zone 19N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"REGVEN / UTM zone 19N\",GEOGCS[\"REGVEN\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Red_Geodesica_Venezolana\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6189\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4189\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-69],PARAMETER[\"scale_factor\",0.9996],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",500000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"2202\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2203, "epsg", 2203, "REGVEN / UTM zone 20N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"REGVEN / UTM zone 20N\",GEOGCS[\"REGVEN\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Red_Geodesica_Venezolana\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6189\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4189\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-63],PARAMETER[\"scale_factor\",0.9996],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",500000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"2203\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2204, "epsg", 2204, "NAD27 / Tennessee");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=35.25 +lat_2=36.41666666666666 +lat_0=3");
    add_proj4text (p, 1,
		   "4.66666666666666 +lon_0=-86 +x_0=609601.2192024384 +y_0=");
    add_proj4text (p, 2,
		   "30480.06096012192 +ellps=clrk66 +datum=NAD27 +units=us-f");
    add_proj4text (p, 3, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Tennessee\",GEOGCS[\"NAD27\",DATUM[\"No");
    add_srs_wkt (p, 1,
		 "rth_American_Datum_1927\",SPHEROID[\"Clarke 1866\",63782");
    add_srs_wkt (p, 2,
		 "06.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal_");
    add_srs_wkt (p, 8,
		 "Conic_2SP\"],PARAMETER[\"standard_parallel_1\",35.25],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"standard_parallel_2\",36.41666666666666],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"latitude_of_origin\",34.66666666666666],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"central_meridian\",-86],PARAMETER[\"false_easting\",2");
    add_srs_wkt (p, 12,
		 "000000],PARAMETER[\"false_northing\",100000],AUTHORITY[\"");
    add_srs_wkt (p, 13, "EPSG\",\"2204\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2205, "epsg", 2205,
		      "NAD83 / Kentucky North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.96666666666667 +lat_2=38.96666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=37.5 +lon_0=-84.25 +x_0=500000 +y_0=0 +ellps=");
    add_proj4text (p, 2, "GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Kentucky North\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 8,
		 "rd_parallel_1\",37.96666666666667],PARAMETER[\"standard_");
    add_srs_wkt (p, 9,
		 "parallel_2\",38.96666666666667],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 10,
		 "origin\",37.5],PARAMETER[\"central_meridian\",-84.25],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 12,
		 "hing\",0],AUTHORITY[\"EPSG\",\"2205\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 13, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2206, "epsg", 2206,
		      "ED50 / 3-degree Gauss-Kruger zone 9");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=27 +k=1 +x_0=9500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / 3-degree Gauss-Kruger zone 9\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "ED50\",DATUM[\"European_Datum_1950\",SPHEROID[\"Internat");
    add_srs_wkt (p, 2,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4230\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",27],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",950");
    add_srs_wkt (p, 10,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"2206\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2207, "epsg", 2207,
		      "ED50 / 3-degree Gauss-Kruger zone 10");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=30 +k=1 +x_0=10500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / 3-degree Gauss-Kruger zone 10\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "ED50\",DATUM[\"European_Datum_1950\",SPHEROID[\"Internat");
    add_srs_wkt (p, 2,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4230\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",30],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",105");
    add_srs_wkt (p, 10,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"2207\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2208, "epsg", 2208,
		      "ED50 / 3-degree Gauss-Kruger zone 11");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=33 +k=1 +x_0=11500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / 3-degree Gauss-Kruger zone 11\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "ED50\",DATUM[\"European_Datum_1950\",SPHEROID[\"Internat");
    add_srs_wkt (p, 2,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4230\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",33],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",115");
    add_srs_wkt (p, 10,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"2208\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2209, "epsg", 2209,
		      "ED50 / 3-degree Gauss-Kruger zone 12");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=36 +k=1 +x_0=12500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / 3-degree Gauss-Kruger zone 12\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "ED50\",DATUM[\"European_Datum_1950\",SPHEROID[\"Internat");
    add_srs_wkt (p, 2,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4230\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",36],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",125");
    add_srs_wkt (p, 10,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"2209\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2210, "epsg", 2210,
		      "ED50 / 3-degree Gauss-Kruger zone 13");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=39 +k=1 +x_0=13500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / 3-degree Gauss-Kruger zone 13\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "ED50\",DATUM[\"European_Datum_1950\",SPHEROID[\"Internat");
    add_srs_wkt (p, 2,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4230\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",39],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",135");
    add_srs_wkt (p, 10,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"2210\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2211, "epsg", 2211,
		      "ED50 / 3-degree Gauss-Kruger zone 14");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=42 +k=1 +x_0=14500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / 3-degree Gauss-Kruger zone 14\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "ED50\",DATUM[\"European_Datum_1950\",SPHEROID[\"Internat");
    add_srs_wkt (p, 2,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4230\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",42],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",145");
    add_srs_wkt (p, 10,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"2211\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2212, "epsg", 2212,
		      "ED50 / 3-degree Gauss-Kruger zone 15");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=45 +k=1 +x_0=15500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / 3-degree Gauss-Kruger zone 15\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "ED50\",DATUM[\"European_Datum_1950\",SPHEROID[\"Internat");
    add_srs_wkt (p, 2,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4230\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",45],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",155");
    add_srs_wkt (p, 10,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"2212\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2213, "epsg", 2213, "ETRS89 / TM 30 NE");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=30 +k=0.9996 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / TM 30 NE\",GEOGCS[\"ETRS89\",DATUM[\"E");
    add_srs_wkt (p, 1,
		 "uropean_Terrestrial_Reference_System_1989\",SPHEROID[\"G");
    add_srs_wkt (p, 2,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "9\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",3");
    add_srs_wkt (p, 9,
		 "0],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"2213\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 12, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 2214, "epsg", 2214,
		      "Douala 1948 / AOF west (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=10.5 +k=0.999 +x_0=1000000 +");
    add_proj4text (p, 1,
		   "y_0=1000000 +ellps=intl +towgs84=-206.1,-174.7,-87.7,0,0");
    add_proj4text (p, 2, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Douala 1948 / AOF west (deprecated)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Douala 1948\",DATUM[\"Douala_1948\",SPHEROID[\"Internati");
    add_srs_wkt (p, 2,
		 "onal 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[-206.1,-174.7,-87.7,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "6192\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4192\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 8,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"central_meridian\",10.5],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 10,
		 "or\",0.999],PARAMETER[\"false_easting\",1000000],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",1000000],AUTHORITY[\"EPSG\",\"2214");
    add_srs_wkt (p, 12, "\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2215, "epsg", 2215,
		      "Manoca 1962 / UTM zone 32N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=32 +a=6378249.2 +b=6356515 +towgs84=-70.");
    add_proj4text (p, 1, "9,-151.8,-41.4,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Manoca 1962 / UTM zone 32N\",GEOGCS[\"Manoca 19");
    add_srs_wkt (p, 1,
		 "62\",DATUM[\"Manoca_1962\",SPHEROID[\"Clarke 1880 (IGN)\"");
    add_srs_wkt (p, 2,
		 ",6378249.2,293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[-70.9,-151.8,-41.4,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6193\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4193\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"central_meridian\",9],PARAMETER[\"scale_facto");
    add_srs_wkt (p, 10,
		 "r\",0.9996],PARAMETER[\"false_easting\",500000],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2215\"],AXI");
    add_srs_wkt (p, 12, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2216, "epsg", 2216,
		      "Qornoq 1927 / UTM zone 22N");
    add_proj4text (p, 0, "+proj=utm +zone=22 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Qornoq 1927 / UTM zone 22N\",GEOGCS[\"Qornoq 19");
    add_srs_wkt (p, 1,
		 "27\",DATUM[\"Qornoq_1927\",SPHEROID[\"International 1924");
    add_srs_wkt (p, 2,
		 "\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6194\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4194\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",0],PARAMETER[\"central_meridian\",-51],PARAMETER[\"sc");
    add_srs_wkt (p, 9,
		 "ale_factor\",0.9996],PARAMETER[\"false_easting\",500000]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"22");
    add_srs_wkt (p, 11,
		 "16\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2217, "epsg", 2217,
		      "Qornoq 1927 / UTM zone 23N");
    add_proj4text (p, 0, "+proj=utm +zone=23 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Qornoq 1927 / UTM zone 23N\",GEOGCS[\"Qornoq 19");
    add_srs_wkt (p, 1,
		 "27\",DATUM[\"Qornoq_1927\",SPHEROID[\"International 1924");
    add_srs_wkt (p, 2,
		 "\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6194\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4194\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",0],PARAMETER[\"central_meridian\",-45],PARAMETER[\"sc");
    add_srs_wkt (p, 9,
		 "ale_factor\",0.9996],PARAMETER[\"false_easting\",500000]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"22");
    add_srs_wkt (p, 11,
		 "17\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2219, "epsg", 2219, "ATS77 / UTM zone 19N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +a=6378135 +b=6356750.304921594 +unit");
    add_proj4text (p, 1, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ATS77 / UTM zone 19N\",GEOGCS[\"ATS77\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Average_Terrestrial_System_1977\",SPHEROID[\"Average Ter");
    add_srs_wkt (p, 2,
		 "restrial System 1977\",6378135,298.257,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7041\"]],AUTHORITY[\"EPSG\",\"6122\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4122\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-69],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"2219\"],AXIS[\"Easting\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2220, "epsg", 2220, "ATS77 / UTM zone 20N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +a=6378135 +b=6356750.304921594 +unit");
    add_proj4text (p, 1, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ATS77 / UTM zone 20N\",GEOGCS[\"ATS77\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Average_Terrestrial_System_1977\",SPHEROID[\"Average Ter");
    add_srs_wkt (p, 2,
		 "restrial System 1977\",6378135,298.257,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7041\"]],AUTHORITY[\"EPSG\",\"6122\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4122\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-63],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"2220\"],AXIS[\"Easting\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2222, "epsg", 2222,
		      "NAD83 / Arizona East (ft)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-110.1666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=213360 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=ft");
    add_proj4text (p, 2, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Arizona East (ft)\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",31],PARAMETER[\"central_meridian\",-110.1");
    add_srs_wkt (p, 9,
		 "666666666667],PARAMETER[\"scale_factor\",0.9999],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",700000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"2222\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 2223, "epsg", 2223,
		      "NAD83 / Arizona Central (ft)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-111.9166666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=213360 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=ft");
    add_proj4text (p, 2, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Arizona Central (ft)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",31],PARAMETER[\"central_meridian\",-111.");
    add_srs_wkt (p, 9,
		 "9166666666667],PARAMETER[\"scale_factor\",0.9999],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",700000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 11,
		 "\",0],AUTHORITY[\"EPSG\",\"2223\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2224, "epsg", 2224,
		      "NAD83 / Arizona West (ft)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-113.75 +k=0.999933333 +x_0");
    add_proj4text (p, 1,
		   "=213360 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=ft +no_d");
    add_proj4text (p, 2, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Arizona West (ft)\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",31],PARAMETER[\"central_meridian\",-113.7");
    add_srs_wkt (p, 9,
		 "5],PARAMETER[\"scale_factor\",0.999933333],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",700000],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 11,
		 "UTHORITY[\"EPSG\",\"2224\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 12, "NORTH]]");
    p = add_epsg_def (first, last, 2225, "epsg", 2225,
		      "NAD83 / California zone 1 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.66666666666666 +lat_2=40 +lat_0=39.3");
    add_proj4text (p, 1,
		   "3333333333334 +lon_0=-122 +x_0=2000000.0001016 +y_0=5000");
    add_proj4text (p, 2,
		   "00.0001016001 +ellps=GRS80 +datum=NAD83 +units=us-ft +no");
    add_proj4text (p, 3, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / California zone 1 (ftUS)\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS ");
    add_srs_wkt (p, 2,
		 "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",41.");
    add_srs_wkt (p, 9,
		 "66666666666666],PARAMETER[\"standard_parallel_2\",40],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"latitude_of_origin\",39.33333333333334],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"central_meridian\",-122],PARAMETER[\"false_easting");
    add_srs_wkt (p, 12,
		 "\",6561666.667],PARAMETER[\"false_northing\",1640416.667");
    add_srs_wkt (p, 13,
		 "],AUTHORITY[\"EPSG\",\"2225\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 14, "\",NORTH]]");
    p = add_epsg_def (first, last, 2226, "epsg", 2226,
		      "NAD83 / California zone 2 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.83333333333334 +lat_2=38.33333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=37.66666666666666 +lon_0=-122 +x_0=2000000.00");
    add_proj4text (p, 2,
		   "01016 +y_0=500000.0001016001 +ellps=GRS80 +datum=NAD83 +");
    add_proj4text (p, 3, "units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / California zone 2 (ftUS)\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS ");
    add_srs_wkt (p, 2,
		 "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",39.");
    add_srs_wkt (p, 9,
		 "83333333333334],PARAMETER[\"standard_parallel_2\",38.333");
    add_srs_wkt (p, 10,
		 "33333333334],PARAMETER[\"latitude_of_origin\",37.6666666");
    add_srs_wkt (p, 11,
		 "6666666],PARAMETER[\"central_meridian\",-122],PARAMETER[");
    add_srs_wkt (p, 12,
		 "\"false_easting\",6561666.667],PARAMETER[\"false_northin");
    add_srs_wkt (p, 13,
		 "g\",1640416.667],AUTHORITY[\"EPSG\",\"2226\"],AXIS[\"X\"");
    add_srs_wkt (p, 14, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2227, "epsg", 2227,
		      "NAD83 / California zone 3 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.43333333333333 +lat_2=37.06666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.5 +lon_0=-120.5 +x_0=2000000.0001016 +y_0=");
    add_proj4text (p, 2,
		   "500000.0001016001 +ellps=GRS80 +datum=NAD83 +units=us-ft");
    add_proj4text (p, 3, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / California zone 3 (ftUS)\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS ");
    add_srs_wkt (p, 2,
		 "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",38.");
    add_srs_wkt (p, 9,
		 "43333333333333],PARAMETER[\"standard_parallel_2\",37.066");
    add_srs_wkt (p, 10,
		 "66666666667],PARAMETER[\"latitude_of_origin\",36.5],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-120.5],PARAMETER[\"false_eas");
    add_srs_wkt (p, 12,
		 "ting\",6561666.667],PARAMETER[\"false_northing\",1640416");
    add_srs_wkt (p, 13,
		 ".667],AUTHORITY[\"EPSG\",\"2227\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 14, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2228, "epsg", 2228,
		      "NAD83 / California zone 4 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.25 +lat_2=36 +lat_0=35.3333333333333");
    add_proj4text (p, 1,
		   "4 +lon_0=-119 +x_0=2000000.0001016 +y_0=500000.000101600");
    add_proj4text (p, 2, "1 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / California zone 4 (ftUS)\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS ");
    add_srs_wkt (p, 2,
		 "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",37.");
    add_srs_wkt (p, 9,
		 "25],PARAMETER[\"standard_parallel_2\",36],PARAMETER[\"la");
    add_srs_wkt (p, 10,
		 "titude_of_origin\",35.33333333333334],PARAMETER[\"centra");
    add_srs_wkt (p, 11,
		 "l_meridian\",-119],PARAMETER[\"false_easting\",6561666.6");
    add_srs_wkt (p, 12,
		 "67],PARAMETER[\"false_northing\",1640416.667],AUTHORITY[");
    add_srs_wkt (p, 13,
		 "\"EPSG\",\"2228\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2229, "epsg", 2229,
		      "NAD83 / California zone 5 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=35.46666666666667 +lat_2=34.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=33.5 +lon_0=-118 +x_0=2000000.0001016 +y_0=50");
    add_proj4text (p, 2,
		   "0000.0001016001 +ellps=GRS80 +datum=NAD83 +units=us-ft +");
    add_proj4text (p, 3, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / California zone 5 (ftUS)\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS ");
    add_srs_wkt (p, 2,
		 "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",35.");
    add_srs_wkt (p, 9,
		 "46666666666667],PARAMETER[\"standard_parallel_2\",34.033");
    add_srs_wkt (p, 10,
		 "33333333333],PARAMETER[\"latitude_of_origin\",33.5],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-118],PARAMETER[\"false_easti");
    add_srs_wkt (p, 12,
		 "ng\",6561666.667],PARAMETER[\"false_northing\",1640416.6");
    add_srs_wkt (p, 13,
		 "67],AUTHORITY[\"EPSG\",\"2229\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 14, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 2230, "epsg", 2230,
		      "NAD83 / California zone 6 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.88333333333333 +lat_2=32.78333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=32.16666666666666 +lon_0=-116.25 +x_0=2000000");
    add_proj4text (p, 2,
		   ".0001016 +y_0=500000.0001016001 +ellps=GRS80 +datum=NAD8");
    add_proj4text (p, 3, "3 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / California zone 6 (ftUS)\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS ");
    add_srs_wkt (p, 2,
		 "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",33.");
    add_srs_wkt (p, 9,
		 "88333333333333],PARAMETER[\"standard_parallel_2\",32.783");
    add_srs_wkt (p, 10,
		 "33333333333],PARAMETER[\"latitude_of_origin\",32.1666666");
    add_srs_wkt (p, 11,
		 "6666666],PARAMETER[\"central_meridian\",-116.25],PARAMET");
    add_srs_wkt (p, 12,
		 "ER[\"false_easting\",6561666.667],PARAMETER[\"false_nort");
    add_srs_wkt (p, 13,
		 "hing\",1640416.667],AUTHORITY[\"EPSG\",\"2230\"],AXIS[\"");
    add_srs_wkt (p, 14, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2231, "epsg", 2231,
		      "NAD83 / Colorado North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.78333333333333 +lat_2=39.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=39.33333333333334 +lon_0=-105.5 +x_0=914401.8");
    add_proj4text (p, 2,
		   "288036576 +y_0=304800.6096012192 +ellps=GRS80 +datum=NAD");
    add_proj4text (p, 3, "83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Colorado North (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",40.78333");
    add_srs_wkt (p, 9,
		 "333333333],PARAMETER[\"standard_parallel_2\",39.71666666");
    add_srs_wkt (p, 10,
		 "666667],PARAMETER[\"latitude_of_origin\",39.333333333333");
    add_srs_wkt (p, 11,
		 "34],PARAMETER[\"central_meridian\",-105.5],PARAMETER[\"f");
    add_srs_wkt (p, 12,
		 "alse_easting\",3000000],PARAMETER[\"false_northing\",100");
    add_srs_wkt (p, 13,
		 "0000],AUTHORITY[\"EPSG\",\"2231\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 14, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2232, "epsg", 2232,
		      "NAD83 / Colorado Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.75 +lat_2=38.45 +lat_0=37.8333333333");
    add_proj4text (p, 1,
		   "3334 +lon_0=-105.5 +x_0=914401.8288036576 +y_0=304800.60");
    add_proj4text (p, 2,
		   "96012192 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_proj4text (p, 3, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Colorado Central (ftUS)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 2,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",39.75");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"standard_parallel_2\",38.45],PARAMETER[\"l");
    add_srs_wkt (p, 10,
		 "atitude_of_origin\",37.83333333333334],PARAMETER[\"centr");
    add_srs_wkt (p, 11,
		 "al_meridian\",-105.5],PARAMETER[\"false_easting\",300000");
    add_srs_wkt (p, 12,
		 "0],PARAMETER[\"false_northing\",1000000],AUTHORITY[\"EPS");
    add_srs_wkt (p, 13, "G\",\"2232\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2233, "epsg", 2233,
		      "NAD83 / Colorado South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.43333333333333 +lat_2=37.23333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=36.66666666666666 +lon_0=-105.5 +x_0=914401.8");
    add_proj4text (p, 2,
		   "288036576 +y_0=304800.6096012192 +ellps=GRS80 +datum=NAD");
    add_proj4text (p, 3, "83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Colorado South (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",38.43333");
    add_srs_wkt (p, 9,
		 "333333333],PARAMETER[\"standard_parallel_2\",37.23333333");
    add_srs_wkt (p, 10,
		 "333333],PARAMETER[\"latitude_of_origin\",36.666666666666");
    add_srs_wkt (p, 11,
		 "66],PARAMETER[\"central_meridian\",-105.5],PARAMETER[\"f");
    add_srs_wkt (p, 12,
		 "alse_easting\",3000000],PARAMETER[\"false_northing\",100");
    add_srs_wkt (p, 13,
		 "0000],AUTHORITY[\"EPSG\",\"2233\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 14, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2234, "epsg", 2234,
		      "NAD83 / Connecticut (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.86666666666667 +lat_2=41.2 +lat_0=40");
    add_proj4text (p, 1,
		   ".83333333333334 +lon_0=-72.75 +x_0=304800.6096012192 +y_");
    add_proj4text (p, 2,
		   "0=152400.3048006096 +ellps=GRS80 +datum=NAD83 +units=us-");
    add_proj4text (p, 3, "ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Connecticut (ftUS)\",GEOGCS[\"NAD83\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal_");
    add_srs_wkt (p, 8,
		 "Conic_2SP\"],PARAMETER[\"standard_parallel_1\",41.866666");
    add_srs_wkt (p, 9,
		 "66666667],PARAMETER[\"standard_parallel_2\",41.2],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"latitude_of_origin\",40.83333333333334],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"central_meridian\",-72.75],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",1000000],PARAMETER[\"false_northing\",500000],AUTHORITY");
    add_srs_wkt (p, 13,
		 "[\"EPSG\",\"2234\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2235, "epsg", 2235,
		      "NAD83 / Delaware (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38 +lon_0=-75.41666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "995 +x_0=200000.0001016002 +y_0=0 +ellps=GRS80 +datum=NA");
    add_proj4text (p, 2, "D83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Delaware (ftUS)\",GEOGCS[\"NAD83\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",63");
    add_srs_wkt (p, 2,
		 "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",38],PARAMETER[\"centr");
    add_srs_wkt (p, 9,
		 "al_meridian\",-75.41666666666667],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 10,
		 "or\",0.999995],PARAMETER[\"false_easting\",656166.667],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2235");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2236, "epsg", 2236,
		      "NAD83 / Florida East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=24.33333333333333 +lon_0=-81 +k=0.999");
    add_proj4text (p, 1,
		   "941177 +x_0=200000.0001016002 +y_0=0 +ellps=GRS80 +datum");
    add_proj4text (p, 2, "=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Florida East (ftUS)\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercato");
    add_srs_wkt (p, 8,
		 "r\"],PARAMETER[\"latitude_of_origin\",24.33333333333333]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"central_meridian\",-81],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",0.999941177],PARAMETER[\"false_easting\",656166.");
    add_srs_wkt (p, 11,
		 "667],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"2236\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2237, "epsg", 2237,
		      "NAD83 / Florida West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=24.33333333333333 +lon_0=-82 +k=0.999");
    add_proj4text (p, 1,
		   "941177 +x_0=200000.0001016002 +y_0=0 +ellps=GRS80 +datum");
    add_proj4text (p, 2, "=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Florida West (ftUS)\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercato");
    add_srs_wkt (p, 8,
		 "r\"],PARAMETER[\"latitude_of_origin\",24.33333333333333]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"central_meridian\",-82],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",0.999941177],PARAMETER[\"false_easting\",656166.");
    add_srs_wkt (p, 11,
		 "667],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"2237\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2238, "epsg", 2238,
		      "NAD83 / Florida North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.75 +lat_2=29.58333333333333 +lat_0=2");
    add_proj4text (p, 1,
		   "9 +lon_0=-84.5 +x_0=600000 +y_0=0 +ellps=GRS80 +datum=NA");
    add_proj4text (p, 2, "D83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Florida North (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",30.75],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"standard_parallel_2\",29.58333333333333],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"latitude_of_origin\",29],PARAMETER[\"central_mer");
    add_srs_wkt (p, 11,
		 "idian\",-84.5],PARAMETER[\"false_easting\",1968500],PARA");
    add_srs_wkt (p, 12,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2238\"]");
    add_srs_wkt (p, 13, ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2239, "epsg", 2239,
		      "NAD83 / Georgia East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=-82.16666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=200000.0001016002 +y_0=0 +ellps=GRS80 +datum=NAD8");
    add_proj4text (p, 2, "3 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Georgia East (ftUS)\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercato");
    add_srs_wkt (p, 8,
		 "r\"],PARAMETER[\"latitude_of_origin\",30],PARAMETER[\"ce");
    add_srs_wkt (p, 9,
		 "ntral_meridian\",-82.16666666666667],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",0.9999],PARAMETER[\"false_easting\",656166.667],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"223");
    add_srs_wkt (p, 12, "9\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2240, "epsg", 2240,
		      "NAD83 / Georgia West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=-84.16666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=699999.9998983998 +y_0=0 +ellps=GRS80 +datum=NAD8");
    add_proj4text (p, 2, "3 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Georgia West (ftUS)\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercato");
    add_srs_wkt (p, 8,
		 "r\"],PARAMETER[\"latitude_of_origin\",30],PARAMETER[\"ce");
    add_srs_wkt (p, 9,
		 "ntral_meridian\",-84.16666666666667],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",0.9999],PARAMETER[\"false_easting\",2296583.333]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"22");
    add_srs_wkt (p, 12, "40\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2241, "epsg", 2241,
		      "NAD83 / Idaho East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-112.1666666");
    add_proj4text (p, 1,
		   "666667 +k=0.9999473679999999 +x_0=200000.0001016002 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Idaho East (ftUS)\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",41.66666666666666],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"central_meridian\",-112.1666666666667],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.999947368],PARAMETER[\"false_eas");
    add_srs_wkt (p, 11,
		 "ting\",656166.667],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 12,
		 "RITY[\"EPSG\",\"2241\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORT");
    add_srs_wkt (p, 13, "H]]");
    p = add_epsg_def (first, last, 2242, "epsg", 2242,
		      "NAD83 / Idaho Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-114 +k=0.99");
    add_proj4text (p, 1,
		   "99473679999999 +x_0=500000.0001016001 +y_0=0 +ellps=GRS8");
    add_proj4text (p, 2, "0 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Idaho Central (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",41.66666666666666");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-114],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.999947368],PARAMETER[\"false_easting\",16404");
    add_srs_wkt (p, 11,
		 "16.667],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2242\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2243, "epsg", 2243,
		      "NAD83 / Idaho West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-115.75 +k=0");
    add_proj4text (p, 1,
		   ".999933333 +x_0=800000.0001016001 +y_0=0 +ellps=GRS80 +d");
    add_proj4text (p, 2, "atum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Idaho West (ftUS)\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",41.66666666666666],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"central_meridian\",-115.75],PARAMETER[\"scal");
    add_srs_wkt (p, 10,
		 "e_factor\",0.999933333],PARAMETER[\"false_easting\",2624");
    add_srs_wkt (p, 11,
		 "666.667],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12, "G\",\"2243\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2244, "epsg", 2244,
		      "NAD83 / Indiana East (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=37.5 +lon_0=-85.66666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=99999.99989839978 +y_0=249364.9987299975 +");
    add_proj4text (p, 2, "ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Indiana East (ftUS) (deprecated)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"7019\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"4269\"]],UNIT[\"US survey foot\",0.30480060");
    add_srs_wkt (p, 7,
		 "96012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",37.5");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-85.66666666666667],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",0.999966667],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",328083.333],PARAMETER[\"false_northing\",818125");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"2244\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 13, "\",NORTH]]");
    p = add_epsg_def (first, last, 2245, "epsg", 2245,
		      "NAD83 / Indiana West (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=37.5 +lon_0=-87.08333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=900000 +y_0=249364.9987299975 +ellps=GRS80");
    add_proj4text (p, 2, " +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Indiana West (ftUS) (deprecated)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"7019\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"4269\"]],UNIT[\"US survey foot\",0.30480060");
    add_srs_wkt (p, 7,
		 "96012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",37.5");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-87.08333333333333],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",0.999966667],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",2952750],PARAMETER[\"false_northing\",818125],A");
    add_srs_wkt (p, 12,
		 "UTHORITY[\"EPSG\",\"2245\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 13, "NORTH]]");
    p = add_epsg_def (first, last, 2246, "epsg", 2246,
		      "NAD83 / Kentucky North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.96666666666667 +lat_2=38.96666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=37.5 +lon_0=-84.25 +x_0=500000.0001016001 +y_");
    add_proj4text (p, 2, "0=0 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Kentucky North (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",37.96666");
    add_srs_wkt (p, 9,
		 "666666667],PARAMETER[\"standard_parallel_2\",38.96666666");
    add_srs_wkt (p, 10,
		 "666667],PARAMETER[\"latitude_of_origin\",37.5],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"central_meridian\",-84.25],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",1640416.667],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 13,
		 "\"EPSG\",\"2246\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2247, "epsg", 2247,
		      "NAD83 / Kentucky South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.93333333333333 +lat_2=36.73333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=36.33333333333334 +lon_0=-85.75 +x_0=500000.0");
    add_proj4text (p, 2,
		   "001016001 +y_0=500000.0001016001 +ellps=GRS80 +datum=NAD");
    add_proj4text (p, 3, "83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Kentucky South (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",37.93333");
    add_srs_wkt (p, 9,
		 "333333333],PARAMETER[\"standard_parallel_2\",36.73333333");
    add_srs_wkt (p, 10,
		 "333333],PARAMETER[\"latitude_of_origin\",36.333333333333");
    add_srs_wkt (p, 11,
		 "34],PARAMETER[\"central_meridian\",-85.75],PARAMETER[\"f");
    add_srs_wkt (p, 12,
		 "alse_easting\",1640416.667],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 13,
		 ",1640416.667],AUTHORITY[\"EPSG\",\"2247\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 14, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2248, "epsg", 2248,
		      "NAD83 / Maryland (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.45 +lat_2=38.3 +lat_0=37.66666666666");
    add_proj4text (p, 1,
		   "666 +lon_0=-77 +x_0=399999.9998983998 +y_0=0 +ellps=GRS8");
    add_proj4text (p, 2, "0 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Maryland (ftUS)\",GEOGCS[\"NAD83\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",63");
    add_srs_wkt (p, 2,
		 "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal_Con");
    add_srs_wkt (p, 8,
		 "ic_2SP\"],PARAMETER[\"standard_parallel_1\",39.45],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"standard_parallel_2\",38.3],PARAMETER[\"latitude_");
    add_srs_wkt (p, 10,
		 "of_origin\",37.66666666666666],PARAMETER[\"central_merid");
    add_srs_wkt (p, 11,
		 "ian\",-77],PARAMETER[\"false_easting\",1312333.333],PARA");
    add_srs_wkt (p, 12,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2248\"]");
    add_srs_wkt (p, 13, ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2249, "epsg", 2249,
		      "NAD83 / Massachusetts Mainland (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=42.68333333333333 +lat_2=41.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=41 +lon_0=-71.5 +x_0=200000.0001016002 +y_0=7");
    add_proj4text (p, 2,
		   "50000 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Massachusetts Mainland (ftUS)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "42.68333333333333],PARAMETER[\"standard_parallel_2\",41.");
    add_srs_wkt (p, 10,
		 "71666666666667],PARAMETER[\"latitude_of_origin\",41],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"central_meridian\",-71.5],PARAMETER[\"false_eas");
    add_srs_wkt (p, 12,
		 "ting\",656166.667],PARAMETER[\"false_northing\",2460625]");
    add_srs_wkt (p, 13,
		 ",AUTHORITY[\"EPSG\",\"2249\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 14, ",NORTH]]");
    p = add_epsg_def (first, last, 2250, "epsg", 2250,
		      "NAD83 / Massachusetts Island (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.48333333333333 +lat_2=41.28333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=41 +lon_0=-70.5 +x_0=500000.0001016001 +y_0=0");
    add_proj4text (p, 2, " +ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Massachusetts Island (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"G");
    add_srs_wkt (p, 2,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "9\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "41.48333333333333],PARAMETER[\"standard_parallel_2\",41.");
    add_srs_wkt (p, 10,
		 "28333333333333],PARAMETER[\"latitude_of_origin\",41],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"central_meridian\",-70.5],PARAMETER[\"false_eas");
    add_srs_wkt (p, 12,
		 "ting\",1640416.667],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 13,
		 "ORITY[\"EPSG\",\"2250\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 14, "TH]]");
    p = add_epsg_def (first, last, 2251, "epsg", 2251,
		      "NAD83 / Michigan North (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.08333333333334 +lat_2=45.48333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=44.78333333333333 +lon_0=-87 +x_0=7999999.999");
    add_proj4text (p, 2,
		   "968001 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=ft +no_de");
    add_proj4text (p, 3, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Michigan North (ft)\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"standard_parallel_1\",47.08333333333334],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_2\",45.48333333333333],PARAMETER[\"lat");
    add_srs_wkt (p, 10,
		 "itude_of_origin\",44.78333333333333],PARAMETER[\"central");
    add_srs_wkt (p, 11,
		 "_meridian\",-87],PARAMETER[\"false_easting\",26246719.16");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 13, "251\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2252, "epsg", 2252,
		      "NAD83 / Michigan Central (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.7 +lat_2=44.18333333333333 +lat_0=43");
    add_proj4text (p, 1,
		   ".31666666666667 +lon_0=-84.36666666666666 +x_0=5999999.9");
    add_proj4text (p, 2,
		   "99976001 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=ft +no_");
    add_proj4text (p, 3, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Michigan Central (ft)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"standard_parallel_1\",45.7],PARAMETER[\"standard_para");
    add_srs_wkt (p, 9,
		 "llel_2\",44.18333333333333],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",43.31666666666667],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 11,
		 "84.36666666666666],PARAMETER[\"false_easting\",19685039.");
    add_srs_wkt (p, 12,
		 "37],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 13, "2252\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2253, "epsg", 2253,
		      "NAD83 / Michigan South (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43.66666666666666 +lat_2=42.1 +lat_0=41");
    add_proj4text (p, 1,
		   ".5 +lon_0=-84.36666666666666 +x_0=3999999.999984 +y_0=0 ");
    add_proj4text (p, 2, "+ellps=GRS80 +datum=NAD83 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Michigan South (ft)\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"standard_parallel_1\",43.66666666666666],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_2\",42.1],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",41.5],PARAMETER[\"central_meridian\",-84.3666666666");
    add_srs_wkt (p, 11,
		 "6666],PARAMETER[\"false_easting\",13123359.58],PARAMETER");
    add_srs_wkt (p, 12,
		 "[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2253\"],AXIS");
    add_srs_wkt (p, 13, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2254, "epsg", 2254,
		      "NAD83 / Mississippi East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=29.5 +lon_0=-88.83333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "9995 +x_0=300000.0000000001 +y_0=0 +ellps=GRS80 +datum=N");
    add_proj4text (p, 2, "AD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Mississippi East (ftUS)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 2,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",29.5],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-88.83333333333333],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",0.99995],PARAMETER[\"false_easting\",98425");
    add_srs_wkt (p, 11,
		 "0.0000000002],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 12,
		 "\"EPSG\",\"2254\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2255, "epsg", 2255,
		      "NAD83 / Mississippi West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=29.5 +lon_0=-90.33333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "9995 +x_0=699999.9998983998 +y_0=0 +ellps=GRS80 +datum=N");
    add_proj4text (p, 2, "AD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Mississippi West (ftUS)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 2,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",29.5],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-90.33333333333333],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",0.99995],PARAMETER[\"false_easting\",22965");
    add_srs_wkt (p, 11,
		 "83.333],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2255\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2256, "epsg", 2256, "NAD83 / Montana (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49 +lat_2=45 +lat_0=44.25 +lon_0=-109.5");
    add_proj4text (p, 1,
		   " +x_0=599999.9999976 +y_0=0 +ellps=GRS80 +datum=NAD83 +u");
    add_proj4text (p, 2, "nits=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Montana (ft)\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"stan");
    add_srs_wkt (p, 8,
		 "dard_parallel_1\",49],PARAMETER[\"standard_parallel_2\",");
    add_srs_wkt (p, 9,
		 "45],PARAMETER[\"latitude_of_origin\",44.25],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "central_meridian\",-109.5],PARAMETER[\"false_easting\",1");
    add_srs_wkt (p, 11,
		 "968503.937],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12, "EPSG\",\"2256\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2257, "epsg", 2257,
		      "NAD83 / New Mexico East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-104.3333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "909091 +x_0=165000 +y_0=0 +ellps=GRS80 +datum=NAD83 +uni");
    add_proj4text (p, 2, "ts=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / New Mexico East (ftUS)\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 2,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",31],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-104.3333333333333],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",0.999909091],PARAMETER[\"false_easting\",541");
    add_srs_wkt (p, 11,
		 "337.5],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2257\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2258, "epsg", 2258,
		      "NAD83 / New Mexico Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-106.25 +k=0.9999 +x_0=5000");
    add_proj4text (p, 1,
		   "00.0001016001 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=us");
    add_proj4text (p, 2, "-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / New Mexico Central (ftUS)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_M");
    add_srs_wkt (p, 8,
		 "ercator\"],PARAMETER[\"latitude_of_origin\",31],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-106.25],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",0.9999],PARAMETER[\"false_easting\",1640416.667],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2258\"]");
    add_srs_wkt (p, 12, ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2259, "epsg", 2259,
		      "NAD83 / New Mexico West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-107.8333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "916667 +x_0=830000.0001016001 +y_0=0 +ellps=GRS80 +datum");
    add_proj4text (p, 2, "=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / New Mexico West (ftUS)\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 2,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",31],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-107.8333333333333],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",0.999916667],PARAMETER[\"false_easting\",272");
    add_srs_wkt (p, 11,
		 "3091.667],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 12, "SG\",\"2259\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2260, "epsg", 2260,
		      "NAD83 / New York East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38.83333333333334 +lon_0=-74.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=150000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=");
    add_proj4text (p, 2, "us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / New York East (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",38.83333333333334");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-74.5],PARAMETER[\"scal");
    add_srs_wkt (p, 10,
		 "e_factor\",0.9999],PARAMETER[\"false_easting\",492125],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2260");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2261, "epsg", 2261,
		      "NAD83 / New York Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40 +lon_0=-76.58333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "9375 +x_0=249999.9998983998 +y_0=0 +ellps=GRS80 +datum=N");
    add_proj4text (p, 2, "AD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / New York Central (ftUS)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 2,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",40],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-76.58333333333333],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",0.9999375],PARAMETER[\"false_easting\",82020");
    add_srs_wkt (p, 11,
		 "8.3330000002],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 12,
		 "\"EPSG\",\"2261\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2262, "epsg", 2262,
		      "NAD83 / New York West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40 +lon_0=-78.58333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "9375 +x_0=350000.0001016001 +y_0=0 +ellps=GRS80 +datum=N");
    add_proj4text (p, 2, "AD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / New York West (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",40],PARAMETER[\"c");
    add_srs_wkt (p, 9,
		 "entral_meridian\",-78.58333333333333],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.9999375],PARAMETER[\"false_easting\",1148291.");
    add_srs_wkt (p, 11,
		 "667],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"2262\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2263, "epsg", 2263,
		      "NAD83 / New York Long Island (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.03333333333333 +lat_2=40.66666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=40.16666666666666 +lon_0=-74 +x_0=300000.0000");
    add_proj4text (p, 2,
		   "000001 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=us-ft +no");
    add_proj4text (p, 3, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / New York Long Island (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"G");
    add_srs_wkt (p, 2,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "9\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "41.03333333333333],PARAMETER[\"standard_parallel_2\",40.");
    add_srs_wkt (p, 10,
		 "66666666666666],PARAMETER[\"latitude_of_origin\",40.1666");
    add_srs_wkt (p, 11,
		 "6666666666],PARAMETER[\"central_meridian\",-74],PARAMETE");
    add_srs_wkt (p, 12,
		 "R[\"false_easting\",984250.0000000002],PARAMETER[\"false");
    add_srs_wkt (p, 13,
		 "_northing\",0],AUTHORITY[\"EPSG\",\"2263\"],AXIS[\"X\",E");
    add_srs_wkt (p, 14, "AST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2264, "epsg", 2264,
		      "NAD83 / North Carolina (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.16666666666666 +lat_2=34.33333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=33.75 +lon_0=-79 +x_0=609601.2192024384 +y_0=");
    add_proj4text (p, 2, "0 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / North Carolina (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",36.16666");
    add_srs_wkt (p, 9,
		 "666666666],PARAMETER[\"standard_parallel_2\",34.33333333");
    add_srs_wkt (p, 10,
		 "333334],PARAMETER[\"latitude_of_origin\",33.75],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"central_meridian\",-79],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 12,
		 "2000000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 13, "G\",\"2264\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2265, "epsg", 2265,
		      "NAD83 / North Dakota North (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.73333333333333 +lat_2=47.43333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=47 +lon_0=-100.5 +x_0=599999.9999976 +y_0=0 +");
    add_proj4text (p, 2, "ellps=GRS80 +datum=NAD83 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / North Dakota North (ft)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 2,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4269\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "02\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"standard_parallel_1\",48.73333333333333],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"standard_parallel_2\",47.43333333333333],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "latitude_of_origin\",47],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 11,
		 "-100.5],PARAMETER[\"false_easting\",1968503.937],PARAMET");
    add_srs_wkt (p, 12,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2265\"],AX");
    add_srs_wkt (p, 13, "IS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2266, "epsg", 2266,
		      "NAD83 / North Dakota South (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.48333333333333 +lat_2=46.18333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=45.66666666666666 +lon_0=-100.5 +x_0=599999.9");
    add_proj4text (p, 2,
		   "999976 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=ft +no_de");
    add_proj4text (p, 3, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / North Dakota South (ft)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 2,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4269\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "02\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"standard_parallel_1\",47.48333333333333],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"standard_parallel_2\",46.18333333333333],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "latitude_of_origin\",45.66666666666666],PARAMETER[\"cent");
    add_srs_wkt (p, 11,
		 "ral_meridian\",-100.5],PARAMETER[\"false_easting\",19685");
    add_srs_wkt (p, 12,
		 "03.937],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 13, "\",\"2266\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2267, "epsg", 2267,
		      "NAD83 / Oklahoma North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.76666666666667 +lat_2=35.56666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=35 +lon_0=-98 +x_0=600000 +y_0=0 +ellps=GRS80");
    add_proj4text (p, 2, " +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Oklahoma North (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",36.76666");
    add_srs_wkt (p, 9,
		 "666666667],PARAMETER[\"standard_parallel_2\",35.56666666");
    add_srs_wkt (p, 10,
		 "666667],PARAMETER[\"latitude_of_origin\",35],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "central_meridian\",-98],PARAMETER[\"false_easting\",1968");
    add_srs_wkt (p, 12,
		 "500],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 13, "\"2267\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2268, "epsg", 2268,
		      "NAD83 / Oklahoma South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=35.23333333333333 +lat_2=33.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=33.33333333333334 +lon_0=-98 +x_0=600000 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Oklahoma South (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",35.23333");
    add_srs_wkt (p, 9,
		 "333333333],PARAMETER[\"standard_parallel_2\",33.93333333");
    add_srs_wkt (p, 10,
		 "333333],PARAMETER[\"latitude_of_origin\",33.333333333333");
    add_srs_wkt (p, 11,
		 "34],PARAMETER[\"central_meridian\",-98],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_easting\",1968500],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 13,
		 "HORITY[\"EPSG\",\"2268\"],AXIS[\"X\",EAST],AXIS[\"Y\",NO");
    add_srs_wkt (p, 14, "RTH]]");
    p = add_epsg_def (first, last, 2269, "epsg", 2269,
		      "NAD83 / Oregon North (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46 +lat_2=44.33333333333334 +lat_0=43.6");
    add_proj4text (p, 1,
		   "6666666666666 +lon_0=-120.5 +x_0=2500000.0001424 +y_0=0 ");
    add_proj4text (p, 2, "+ellps=GRS80 +datum=NAD83 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Oregon North (ft)\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"standard_parallel_1\",46],PARAMETER[\"standard_paralle");
    add_srs_wkt (p, 9,
		 "l_2\",44.33333333333334],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 10,
		 ",43.66666666666666],PARAMETER[\"central_meridian\",-120.");
    add_srs_wkt (p, 11,
		 "5],PARAMETER[\"false_easting\",8202099.738],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"2269\"],AXIS[\"");
    add_srs_wkt (p, 13, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2270, "epsg", 2270,
		      "NAD83 / Oregon South (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44 +lat_2=42.33333333333334 +lat_0=41.6");
    add_proj4text (p, 1,
		   "6666666666666 +lon_0=-120.5 +x_0=1500000.0001464 +y_0=0 ");
    add_proj4text (p, 2, "+ellps=GRS80 +datum=NAD83 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Oregon South (ft)\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"standard_parallel_1\",44],PARAMETER[\"standard_paralle");
    add_srs_wkt (p, 9,
		 "l_2\",42.33333333333334],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 10,
		 ",41.66666666666666],PARAMETER[\"central_meridian\",-120.");
    add_srs_wkt (p, 11,
		 "5],PARAMETER[\"false_easting\",4921259.843],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"2270\"],AXIS[\"");
    add_srs_wkt (p, 13, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2271, "epsg", 2271,
		      "NAD83 / Pennsylvania North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.95 +lat_2=40.88333333333333 +lat_0=4");
    add_proj4text (p, 1,
		   "0.16666666666666 +lon_0=-77.75 +x_0=600000 +y_0=0 +ellps");
    add_proj4text (p, 2, "=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Pennsylvania North (ftUS)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",41.");
    add_srs_wkt (p, 9,
		 "95],PARAMETER[\"standard_parallel_2\",40.88333333333333]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"latitude_of_origin\",40.16666666666666],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"central_meridian\",-77.75],PARAMETER[\"false_ea");
    add_srs_wkt (p, 12,
		 "sting\",1968500],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 13,
		 "TY[\"EPSG\",\"2271\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]");
    add_srs_wkt (p, 14, "]");
    p = add_epsg_def (first, last, 2272, "epsg", 2272,
		      "NAD83 / Pennsylvania South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.96666666666667 +lat_2=39.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=39.33333333333334 +lon_0=-77.75 +x_0=600000 +");
    add_proj4text (p, 2,
		   "y_0=0 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Pennsylvania South (ftUS)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",40.");
    add_srs_wkt (p, 9,
		 "96666666666667],PARAMETER[\"standard_parallel_2\",39.933");
    add_srs_wkt (p, 10,
		 "33333333333],PARAMETER[\"latitude_of_origin\",39.3333333");
    add_srs_wkt (p, 11,
		 "3333334],PARAMETER[\"central_meridian\",-77.75],PARAMETE");
    add_srs_wkt (p, 12,
		 "R[\"false_easting\",1968500],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 13,
		 ",0],AUTHORITY[\"EPSG\",\"2272\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 14, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 2273, "epsg", 2273,
		      "NAD83 / South Carolina (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=34.83333333333334 +lat_2=32.5 +lat_0=31");
    add_proj4text (p, 1,
		   ".83333333333333 +lon_0=-81 +x_0=609600 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +datum=NAD83 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / South Carolina (ft)\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"standard_parallel_1\",34.83333333333334],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_2\",32.5],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",31.83333333333333],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 11,
		 "81],PARAMETER[\"false_easting\",2000000],PARAMETER[\"fal");
    add_srs_wkt (p, 12,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"2273\"],AXIS[\"X\"");
    add_srs_wkt (p, 13, ",EAST],AXIS[\"Y\",NORTH]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_02 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 2274, "epsg", 2274,
		      "NAD83 / Tennessee (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.41666666666666 +lat_2=35.25 +lat_0=3");
    add_proj4text (p, 1,
		   "4.33333333333334 +lon_0=-86 +x_0=600000 +y_0=0 +ellps=GR");
    add_proj4text (p, 2, "S80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Tennessee (ftUS)\",GEOGCS[\"NAD83\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6");
    add_srs_wkt (p, 2,
		 "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"426");
    add_srs_wkt (p, 6,
		 "9\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal_Co");
    add_srs_wkt (p, 8,
		 "nic_2SP\"],PARAMETER[\"standard_parallel_1\",36.41666666");
    add_srs_wkt (p, 9,
		 "666666],PARAMETER[\"standard_parallel_2\",35.25],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"latitude_of_origin\",34.33333333333334],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "central_meridian\",-86],PARAMETER[\"false_easting\",1968");
    add_srs_wkt (p, 12,
		 "500],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 13, "\"2274\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2275, "epsg", 2275,
		      "NAD83 / Texas North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.18333333333333 +lat_2=34.65 +lat_0=3");
    add_proj4text (p, 1,
		   "4 +lon_0=-101.5 +x_0=200000.0001016002 +y_0=999999.99989");
    add_proj4text (p, 2,
		   "83998 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Texas North (ftUS)\",GEOGCS[\"NAD83\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal_");
    add_srs_wkt (p, 8,
		 "Conic_2SP\"],PARAMETER[\"standard_parallel_1\",36.183333");
    add_srs_wkt (p, 9,
		 "33333333],PARAMETER[\"standard_parallel_2\",34.65],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"latitude_of_origin\",34],PARAMETER[\"central_meri");
    add_srs_wkt (p, 11,
		 "dian\",-101.5],PARAMETER[\"false_easting\",656166.667],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_northing\",3280833.333],AUTHORITY[\"EPS");
    add_srs_wkt (p, 13, "G\",\"2275\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2276, "epsg", 2276,
		      "NAD83 / Texas North Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.96666666666667 +lat_2=32.13333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=31.66666666666667 +lon_0=-98.5 +x_0=600000 +y");
    add_proj4text (p, 2,
		   "_0=2000000.0001016 +ellps=GRS80 +datum=NAD83 +units=us-f");
    add_proj4text (p, 3, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Texas North Central (ftUS)\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GR");
    add_srs_wkt (p, 2,
		 "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4269\"]],UNIT[\"US survey foot\",0.30480060960121");
    add_srs_wkt (p, 7,
		 "92,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Co");
    add_srs_wkt (p, 8,
		 "nformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",3");
    add_srs_wkt (p, 9,
		 "3.96666666666667],PARAMETER[\"standard_parallel_2\",32.1");
    add_srs_wkt (p, 10,
		 "3333333333333],PARAMETER[\"latitude_of_origin\",31.66666");
    add_srs_wkt (p, 11,
		 "666666667],PARAMETER[\"central_meridian\",-98.5],PARAMET");
    add_srs_wkt (p, 12,
		 "ER[\"false_easting\",1968500],PARAMETER[\"false_northing");
    add_srs_wkt (p, 13,
		 "\",6561666.667],AUTHORITY[\"EPSG\",\"2276\"],AXIS[\"X\",");
    add_srs_wkt (p, 14, "EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2277, "epsg", 2277,
		      "NAD83 / Texas Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=31.88333333333333 +lat_2=30.11666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=29.66666666666667 +lon_0=-100.3333333333333 +");
    add_proj4text (p, 2,
		   "x_0=699999.9998983998 +y_0=3000000 +ellps=GRS80 +datum=N");
    add_proj4text (p, 3, "AD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Texas Central (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",31.88333");
    add_srs_wkt (p, 9,
		 "333333333],PARAMETER[\"standard_parallel_2\",30.11666666");
    add_srs_wkt (p, 10,
		 "666667],PARAMETER[\"latitude_of_origin\",29.666666666666");
    add_srs_wkt (p, 11,
		 "67],PARAMETER[\"central_meridian\",-100.3333333333333],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_easting\",2296583.333],PARAMETER[\"fals");
    add_srs_wkt (p, 13,
		 "e_northing\",9842500.000000002],AUTHORITY[\"EPSG\",\"227");
    add_srs_wkt (p, 14, "7\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2278, "epsg", 2278,
		      "NAD83 / Texas South Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.28333333333333 +lat_2=28.38333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=27.83333333333333 +lon_0=-99 +x_0=600000 +y_0");
    add_proj4text (p, 2,
		   "=3999999.9998984 +ellps=GRS80 +datum=NAD83 +units=us-ft ");
    add_proj4text (p, 3, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Texas South Central (ftUS)\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GR");
    add_srs_wkt (p, 2,
		 "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4269\"]],UNIT[\"US survey foot\",0.30480060960121");
    add_srs_wkt (p, 7,
		 "92,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Co");
    add_srs_wkt (p, 8,
		 "nformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",3");
    add_srs_wkt (p, 9,
		 "0.28333333333333],PARAMETER[\"standard_parallel_2\",28.3");
    add_srs_wkt (p, 10,
		 "8333333333333],PARAMETER[\"latitude_of_origin\",27.83333");
    add_srs_wkt (p, 11,
		 "333333333],PARAMETER[\"central_meridian\",-99],PARAMETER");
    add_srs_wkt (p, 12,
		 "[\"false_easting\",1968500],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 13,
		 ",13123333.333],AUTHORITY[\"EPSG\",\"2278\"],AXIS[\"X\",E");
    add_srs_wkt (p, 14, "AST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2279, "epsg", 2279,
		      "NAD83 / Texas South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=27.83333333333333 +lat_2=26.16666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=25.66666666666667 +lon_0=-98.5 +x_0=300000.00");
    add_proj4text (p, 2,
		   "00000001 +y_0=5000000.0001016 +ellps=GRS80 +datum=NAD83 ");
    add_proj4text (p, 3, "+units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Texas South (ftUS)\",GEOGCS[\"NAD83\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal_");
    add_srs_wkt (p, 8,
		 "Conic_2SP\"],PARAMETER[\"standard_parallel_1\",27.833333");
    add_srs_wkt (p, 9,
		 "33333333],PARAMETER[\"standard_parallel_2\",26.166666666");
    add_srs_wkt (p, 10,
		 "66667],PARAMETER[\"latitude_of_origin\",25.6666666666666");
    add_srs_wkt (p, 11,
		 "7],PARAMETER[\"central_meridian\",-98.5],PARAMETER[\"fal");
    add_srs_wkt (p, 12,
		 "se_easting\",984250.0000000002],PARAMETER[\"false_northi");
    add_srs_wkt (p, 13,
		 "ng\",16404166.667],AUTHORITY[\"EPSG\",\"2279\"],AXIS[\"X");
    add_srs_wkt (p, 14, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2280, "epsg", 2280,
		      "NAD83 / Utah North (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.78333333333333 +lat_2=40.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=40.33333333333334 +lon_0=-111.5 +x_0=500000.0");
    add_proj4text (p, 2,
		   "001504 +y_0=999999.9999960001 +ellps=GRS80 +datum=NAD83 ");
    add_proj4text (p, 3, "+units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Utah North (ft)\",GEOGCS[\"NAD83\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",63");
    add_srs_wkt (p, 2,
		 "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "standard_parallel_1\",41.78333333333333],PARAMETER[\"sta");
    add_srs_wkt (p, 9,
		 "ndard_parallel_2\",40.71666666666667],PARAMETER[\"latitu");
    add_srs_wkt (p, 10,
		 "de_of_origin\",40.33333333333334],PARAMETER[\"central_me");
    add_srs_wkt (p, 11,
		 "ridian\",-111.5],PARAMETER[\"false_easting\",1640419.948");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_northing\",3280839.895],AUTHORITY[\"");
    add_srs_wkt (p, 13, "EPSG\",\"2280\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2281, "epsg", 2281,
		      "NAD83 / Utah Central (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.65 +lat_2=39.01666666666667 +lat_0=3");
    add_proj4text (p, 1,
		   "8.33333333333334 +lon_0=-111.5 +x_0=500000.0001504 +y_0=");
    add_proj4text (p, 2,
		   "1999999.999992 +ellps=GRS80 +datum=NAD83 +units=ft +no_d");
    add_proj4text (p, 3, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Utah Central (ft)\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"standard_parallel_1\",40.65],PARAMETER[\"standard_para");
    add_srs_wkt (p, 9,
		 "llel_2\",39.01666666666667],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",38.33333333333334],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 11,
		 "111.5],PARAMETER[\"false_easting\",1640419.948],PARAMETE");
    add_srs_wkt (p, 12,
		 "R[\"false_northing\",6561679.79],AUTHORITY[\"EPSG\",\"22");
    add_srs_wkt (p, 13, "81\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2282, "epsg", 2282,
		      "NAD83 / Utah South (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.35 +lat_2=37.21666666666667 +lat_0=3");
    add_proj4text (p, 1,
		   "6.66666666666666 +lon_0=-111.5 +x_0=500000.0001504 +y_0=");
    add_proj4text (p, 2,
		   "2999999.999988 +ellps=GRS80 +datum=NAD83 +units=ft +no_d");
    add_proj4text (p, 3, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Utah South (ft)\",GEOGCS[\"NAD83\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",63");
    add_srs_wkt (p, 2,
		 "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "standard_parallel_1\",38.35],PARAMETER[\"standard_parall");
    add_srs_wkt (p, 9,
		 "el_2\",37.21666666666667],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 10,
		 "\",36.66666666666666],PARAMETER[\"central_meridian\",-11");
    add_srs_wkt (p, 11,
		 "1.5],PARAMETER[\"false_easting\",1640419.948],PARAMETER[");
    add_srs_wkt (p, 12,
		 "\"false_northing\",9842519.685],AUTHORITY[\"EPSG\",\"228");
    add_srs_wkt (p, 13, "2\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2283, "epsg", 2283,
		      "NAD83 / Virginia North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.2 +lat_2=38.03333333333333 +lat_0=37");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-78.5 +x_0=3500000.0001016 +y_0=2");
    add_proj4text (p, 2,
		   "000000.0001016 +ellps=GRS80 +datum=NAD83 +units=us-ft +n");
    add_proj4text (p, 3, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Virginia North (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",39.2],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"standard_parallel_2\",38.03333333333333],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"latitude_of_origin\",37.66666666666666],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"central_meridian\",-78.5],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",11482916.667],PARAMETER[\"false_northing\",6561666.667]");
    add_srs_wkt (p, 13,
		 ",AUTHORITY[\"EPSG\",\"2283\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 14, ",NORTH]]");
    p = add_epsg_def (first, last, 2284, "epsg", 2284,
		      "NAD83 / Virginia South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.96666666666667 +lat_2=36.76666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.33333333333334 +lon_0=-78.5 +x_0=3500000.0");
    add_proj4text (p, 2,
		   "001016 +y_0=999999.9998983998 +ellps=GRS80 +datum=NAD83 ");
    add_proj4text (p, 3, "+units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Virginia South (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",37.96666");
    add_srs_wkt (p, 9,
		 "666666667],PARAMETER[\"standard_parallel_2\",36.76666666");
    add_srs_wkt (p, 10,
		 "666667],PARAMETER[\"latitude_of_origin\",36.333333333333");
    add_srs_wkt (p, 11,
		 "34],PARAMETER[\"central_meridian\",-78.5],PARAMETER[\"fa");
    add_srs_wkt (p, 12,
		 "lse_easting\",11482916.667],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 13,
		 ",3280833.333],AUTHORITY[\"EPSG\",\"2284\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 14, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2285, "epsg", 2285,
		      "NAD83 / Washington North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.73333333333333 +lat_2=47.5 +lat_0=47");
    add_proj4text (p, 1,
		   " +lon_0=-120.8333333333333 +x_0=500000.0001016001 +y_0=0");
    add_proj4text (p, 2, " +ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Washington North (ftUS)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 2,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",48.73");
    add_srs_wkt (p, 9,
		 "333333333333],PARAMETER[\"standard_parallel_2\",47.5],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"latitude_of_origin\",47],PARAMETER[\"central_m");
    add_srs_wkt (p, 11,
		 "eridian\",-120.8333333333333],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",1640416.667],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 13,
		 "\"EPSG\",\"2285\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2286, "epsg", 2286,
		      "NAD83 / Washington South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.33333333333334 +lat_2=45.83333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=45.33333333333334 +lon_0=-120.5 +x_0=500000.0");
    add_proj4text (p, 2,
		   "001016001 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=us-ft ");
    add_proj4text (p, 3, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Washington South (ftUS)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 2,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",47.33");
    add_srs_wkt (p, 9,
		 "333333333334],PARAMETER[\"standard_parallel_2\",45.83333");
    add_srs_wkt (p, 10,
		 "333333334],PARAMETER[\"latitude_of_origin\",45.333333333");
    add_srs_wkt (p, 11,
		 "33334],PARAMETER[\"central_meridian\",-120.5],PARAMETER[");
    add_srs_wkt (p, 12,
		 "\"false_easting\",1640416.667],PARAMETER[\"false_northin");
    add_srs_wkt (p, 13,
		 "g\",0],AUTHORITY[\"EPSG\",\"2286\"],AXIS[\"X\",EAST],AXI");
    add_srs_wkt (p, 14, "S[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2287, "epsg", 2287,
		      "NAD83 / Wisconsin North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46.76666666666667 +lat_2=45.56666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=45.16666666666666 +lon_0=-90 +x_0=600000 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Wisconsin North (ftUS)\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 2,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",46.76");
    add_srs_wkt (p, 9,
		 "666666666667],PARAMETER[\"standard_parallel_2\",45.56666");
    add_srs_wkt (p, 10,
		 "666666667],PARAMETER[\"latitude_of_origin\",45.166666666");
    add_srs_wkt (p, 11,
		 "66666],PARAMETER[\"central_meridian\",-90],PARAMETER[\"f");
    add_srs_wkt (p, 12,
		 "alse_easting\",1968500],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 13,
		 "AUTHORITY[\"EPSG\",\"2287\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 14, ",NORTH]]");
    p = add_epsg_def (first, last, 2288, "epsg", 2288,
		      "NAD83 / Wisconsin Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.5 +lat_2=44.25 +lat_0=43.83333333333");
    add_proj4text (p, 1,
		   "334 +lon_0=-90 +x_0=600000 +y_0=0 +ellps=GRS80 +datum=NA");
    add_proj4text (p, 2, "D83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Wisconsin Central (ftUS)\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS ");
    add_srs_wkt (p, 2,
		 "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",45.");
    add_srs_wkt (p, 9,
		 "5],PARAMETER[\"standard_parallel_2\",44.25],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "latitude_of_origin\",43.83333333333334],PARAMETER[\"cent");
    add_srs_wkt (p, 11,
		 "ral_meridian\",-90],PARAMETER[\"false_easting\",1968500]");
    add_srs_wkt (p, 12,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"22");
    add_srs_wkt (p, 13, "88\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2289, "epsg", 2289,
		      "NAD83 / Wisconsin South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.06666666666667 +lat_2=42.73333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=42 +lon_0=-90 +x_0=600000 +y_0=0 +ellps=GRS80");
    add_proj4text (p, 2, " +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Wisconsin South (ftUS)\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 2,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",44.06");
    add_srs_wkt (p, 9,
		 "666666666667],PARAMETER[\"standard_parallel_2\",42.73333");
    add_srs_wkt (p, 10,
		 "333333333],PARAMETER[\"latitude_of_origin\",42],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"central_meridian\",-90],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 12,
		 "1968500],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 13, "G\",\"2289\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2290, "epsg", 2290,
		      "ATS77 / Prince Edward Isl. Stereographic (ATS77)");
    add_proj4text (p, 0,
		   "+proj=sterea +lat_0=47.25 +lon_0=-63 +k=0.999912 +x_0=70");
    add_proj4text (p, 1,
		   "0000 +y_0=400000 +a=6378135 +b=6356750.304921594 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ATS77 / Prince Edward Isl. Stereographic (ATS77");
    add_srs_wkt (p, 1,
		 ")\",GEOGCS[\"ATS77\",DATUM[\"Average_Terrestrial_System_");
    add_srs_wkt (p, 2,
		 "1977\",SPHEROID[\"Average Terrestrial System 1977\",6378");
    add_srs_wkt (p, 3,
		 "135,298.257,AUTHORITY[\"EPSG\",\"7041\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 4,
		 "SG\",\"6122\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4122\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 8,
		 "Oblique_Stereographic\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 9,
		 ",47.25],PARAMETER[\"central_meridian\",-63],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",0.999912],PARAMETER[\"false_easting\",700");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",400000],AUTHORITY[\"EP");
    add_srs_wkt (p, 12,
		 "SG\",\"2290\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 2291, "epsg", 2291,
		      "NAD83(CSRS98) / Prince Edward Isl. Stereographic (NAD83) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=sterea +lat_0=47.25 +lon_0=-63 +k=0.999912 +x_0=40");
    add_proj4text (p, 1,
		   "0000 +y_0=800000 +a=6378135 +b=6356750.304921594 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / Prince Edward Isl. Stereographi");
    add_srs_wkt (p, 1,
		 "c (NAD83) (deprecated)\",GEOGCS[\"ATS77\",DATUM[\"Averag");
    add_srs_wkt (p, 2,
		 "e_Terrestrial_System_1977\",SPHEROID[\"Average Terrestri");
    add_srs_wkt (p, 3,
		 "al System 1977\",6378135,298.257,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 4,
		 "41\"]],AUTHORITY[\"EPSG\",\"6122\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4122\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 8,
		 "001\"]],PROJECTION[\"Oblique_Stereographic\"],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"latitude_of_origin\",47.25],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 10,
		 "an\",-63],PARAMETER[\"scale_factor\",0.999912],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_easting\",400000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 12,
		 "800000],AUTHORITY[\"EPSG\",\"2291\"],AXIS[\"E(X)\",EAST]");
    add_srs_wkt (p, 13, ",AXIS[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 2292, "epsg", 2292,
		      "NAD83(CSRS98) / Prince Edward Isl. Stereographic (NAD83) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=sterea +lat_0=47.25 +lon_0=-63 +k=0.999912 +x_0=40");
    add_proj4text (p, 1,
		   "0000 +y_0=800000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +un");
    add_proj4text (p, 2, "its=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS98) / Prince Edward Isl. Stereographi");
    add_srs_wkt (p, 1,
		 "c (NAD83) (deprecated)\",GEOGCS[\"NAD83(CSRS98)\",DATUM[");
    add_srs_wkt (p, 2,
		 "\"NAD83_Canadian_Spatial_Reference_System\",SPHEROID[\"G");
    add_srs_wkt (p, 3,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 4,
		 "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6140\"");
    add_srs_wkt (p, 5,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 6,
		 "UNIT[\"degree\",0.0174532925199433,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9108\"]],AUTHORITY[\"EPSG\",\"4140\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 8,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Oblique_Stere");
    add_srs_wkt (p, 9,
		 "ographic\"],PARAMETER[\"latitude_of_origin\",47.25],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"central_meridian\",-63],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 11,
		 "\",0.999912],PARAMETER[\"false_easting\",400000],PARAMET");
    add_srs_wkt (p, 12,
		 "ER[\"false_northing\",800000],AUTHORITY[\"EPSG\",\"2292\"");
    add_srs_wkt (p, 13, "],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 2294, "epsg", 2294,
		      "ATS77 / MTM Nova Scotia zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-61.5 +k=0.9999 +x_0=4500000");
    add_proj4text (p, 1,
		   " +y_0=0 +a=6378135 +b=6356750.304921594 +units=m +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ATS77 / MTM Nova Scotia zone 4\",GEOGCS[\"ATS77");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"Average_Terrestrial_System_1977\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Average Terrestrial System 1977\",6378135,298.257,AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"7041\"]],AUTHORITY[\"EPSG\",\"6122\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6,
		 "\"]],AUTHORITY[\"EPSG\",\"4122\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"ce");
    add_srs_wkt (p, 9,
		 "ntral_meridian\",-61.5],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 10,
		 "9],PARAMETER[\"false_easting\",4500000],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"2294\"],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2295, "epsg", 2295,
		      "ATS77 / MTM Nova Scotia zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-64.5 +k=0.9999 +x_0=5500000");
    add_proj4text (p, 1,
		   " +y_0=0 +a=6378135 +b=6356750.304921594 +units=m +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ATS77 / MTM Nova Scotia zone 5\",GEOGCS[\"ATS77");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"Average_Terrestrial_System_1977\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Average Terrestrial System 1977\",6378135,298.257,AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"7041\"]],AUTHORITY[\"EPSG\",\"6122\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6,
		 "\"]],AUTHORITY[\"EPSG\",\"4122\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"ce");
    add_srs_wkt (p, 9,
		 "ntral_meridian\",-64.5],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 10,
		 "9],PARAMETER[\"false_easting\",5500000],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"2295\"],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2308, "epsg", 2308, "Batavia / TM 109 SE");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=109 +k=0.9996 +x_0=500000 +y");
    add_proj4text (p, 1, "_0=10000000 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Batavia / TM 109 SE\",GEOGCS[\"Batavia\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"Batavia\",SPHEROID[\"Bessel 1841\",6377397.155,299.152");
    add_srs_wkt (p, 2,
		 "8128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "6211\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 4,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 5,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4211\"]],UNIT[\"met");
    add_srs_wkt (p, 6,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 7,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 8,
		 "AMETER[\"central_meridian\",109],PARAMETER[\"scale_facto");
    add_srs_wkt (p, 9,
		 "r\",0.9996],PARAMETER[\"false_easting\",500000],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"2308");
    add_srs_wkt (p, 11, "\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2309, "epsg", 2309, "WGS 84 / TM 116 SE");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=116 +k=0.9996 +x_0=500000 +y");
    add_proj4text (p, 1,
		   "_0=10000000 +ellps=WGS84 +datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / TM 116 SE\",GEOGCS[\"WGS 84\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]]");
    add_srs_wkt (p, 3,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 4,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 5,
		 "122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,A");
    add_srs_wkt (p, 6,
		 "UTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 7,
		 "cator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "central_meridian\",116],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 9,
		 "6],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_northing\",10000000],AUTHORITY[\"EPSG\",\"2309\"],AXIS[");
    add_srs_wkt (p, 11, "\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2310, "epsg", 2310, "WGS 84 / TM 132 SE");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=132 +k=0.9996 +x_0=500000 +y");
    add_proj4text (p, 1,
		   "_0=10000000 +ellps=WGS84 +datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / TM 132 SE\",GEOGCS[\"WGS 84\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]]");
    add_srs_wkt (p, 3,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 4,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 5,
		 "122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,A");
    add_srs_wkt (p, 6,
		 "UTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 7,
		 "cator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "central_meridian\",132],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 9,
		 "6],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_northing\",10000000],AUTHORITY[\"EPSG\",\"2310\"],AXIS[");
    add_srs_wkt (p, 11, "\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2311, "epsg", 2311, "WGS 84 / TM 6 NE");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=6 +k=0.9996 +x_0=500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=WGS84 +datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / TM 6 NE\",GEOGCS[\"WGS 84\",DATUM[\"WG");
    add_srs_wkt (p, 1,
		 "S_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHO");
    add_srs_wkt (p, 2,
		 "RITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],P");
    add_srs_wkt (p, 3,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 4,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 5,
		 "2\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merca");
    add_srs_wkt (p, 7,
		 "tor\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"c");
    add_srs_wkt (p, 8,
		 "entral_meridian\",6],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"2311\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2312, "epsg", 2312, "Garoua / UTM zone 33N");
    add_proj4text (p, 0, "+proj=utm +zone=33 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Garoua / UTM zone 33N\",GEOGCS[\"Garoua\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Garoua\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,2");
    add_srs_wkt (p, 2,
		 "93.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"6197\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 4,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4197\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",15],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 9,
		 "or\",0.9996],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2312\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2313, "epsg", 2313,
		      "Kousseri / UTM zone 33N");
    add_proj4text (p, 0, "+proj=utm +zone=33 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kousseri / UTM zone 33N\",GEOGCS[\"Kousseri\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Kousseri\",SPHEROID[\"Clarke 1880 (RGS)\",6378249");
    add_srs_wkt (p, 2,
		 ".145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6198\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4198\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",0],PARAMETER[\"central_meridian\",15],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",0.9996],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2313\"");
    add_srs_wkt (p, 11, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2314, "epsg", 2314,
		      "Trinidad 1903 / Trinidad Grid (ftCla)");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=10.44166666666667 +lon_0=-61.333333333");
    add_proj4text (p, 1,
		   "33334 +x_0=86501.46392052001 +y_0=65379.0134283 +a=63782");
    add_proj4text (p, 2,
		   "93.645208759 +b=6356617.987679838 +to_meter=0.3047972654");
    add_proj4text (p, 3, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Trinidad 1903 / Trinidad Grid (ftCla)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"Trinidad 1903\",DATUM[\"Trinidad_1903\",SPHEROID[\"Cla");
    add_srs_wkt (p, 2,
		 "rke 1858\",6378293.645208759,294.2606763692569,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7007\"]],AUTHORITY[\"EPSG\",\"6302\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4302\"]],UNIT[\"Clarke's foot\",0");
    add_srs_wkt (p, 7,
		 ".3047972654,AUTHORITY[\"EPSG\",\"9005\"]],PROJECTION[\"C");
    add_srs_wkt (p, 8,
		 "assini_Soldner\"],PARAMETER[\"latitude_of_origin\",10.44");
    add_srs_wkt (p, 9,
		 "166666666667],PARAMETER[\"central_meridian\",-61.3333333");
    add_srs_wkt (p, 10,
		 "3333334],PARAMETER[\"false_easting\",283800],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",214500],AUTHORITY[\"EPSG\",\"2314\"],AX");
    add_srs_wkt (p, 12, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2315, "epsg", 2315,
		      "Campo Inchauspe / UTM zone 19S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Campo Inchauspe / UTM zone 19S\",GEOGCS[\"Campo");
    add_srs_wkt (p, 1,
		 " Inchauspe\",DATUM[\"Campo_Inchauspe\",SPHEROID[\"Intern");
    add_srs_wkt (p, 2,
		 "ational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6221\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4221\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-69],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",10000000],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"2315\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2316, "epsg", 2316,
		      "Campo Inchauspe / UTM zone 20S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Campo Inchauspe / UTM zone 20S\",GEOGCS[\"Campo");
    add_srs_wkt (p, 1,
		 " Inchauspe\",DATUM[\"Campo_Inchauspe\",SPHEROID[\"Intern");
    add_srs_wkt (p, 2,
		 "ational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6221\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4221\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-63],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",10000000],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"2316\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2317, "epsg", 2317, "PSAD56 / ICN Regional");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=9 +lat_2=3 +lat_0=6 +lon_0=-66 +x_0=100");
    add_proj4text (p, 1, "0000 +y_0=1000000 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PSAD56 / ICN Regional\",GEOGCS[\"PSAD56\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Provisional_South_American_Datum_1956\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4248\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"standard_parallel_1\",9],PARAMETER[\"standard_par");
    add_srs_wkt (p, 9,
		 "allel_2\",3],PARAMETER[\"latitude_of_origin\",6],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"central_meridian\",-66],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",1000000],PARAMETER[\"false_northing\",1000000],AUTHORIT");
    add_srs_wkt (p, 12,
		 "Y[\"EPSG\",\"2317\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 2318, "epsg", 2318,
		      "Ain el Abd / Aramco Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=17 +lat_2=33 +lat_0=25.08951 +lon_0=48 ");
    add_proj4text (p, 1, "+x_0=0 +y_0=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Ain el Abd / Aramco Lambert\",GEOGCS[\"Ain el A");
    add_srs_wkt (p, 1,
		 "bd\",DATUM[\"Ain_el_Abd_1970\",SPHEROID[\"International ");
    add_srs_wkt (p, 2,
		 "1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6204\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4204\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 8,
		 "rd_parallel_1\",17],PARAMETER[\"standard_parallel_2\",33");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"latitude_of_origin\",25.08951],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "central_meridian\",48],PARAMETER[\"false_easting\",0],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2318\"");
    add_srs_wkt (p, 12, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2319, "epsg", 2319, "ED50 / TM27");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=27 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / TM27\",GEOGCS[\"ED50\",DATUM[\"European_");
    add_srs_wkt (p, 1,
		 "Datum_1950\",SPHEROID[\"International 1924\",6378388,297");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6230");
    add_srs_wkt (p, 3,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",27],PARAMETER[\"scale_factor\",1");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"false_easting\",500000],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "northing\",0],AUTHORITY[\"EPSG\",\"2319\"],AXIS[\"X\",NO");
    add_srs_wkt (p, 11, "RTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2320, "epsg", 2320, "ED50 / TM30");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=30 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / TM30\",GEOGCS[\"ED50\",DATUM[\"European_");
    add_srs_wkt (p, 1,
		 "Datum_1950\",SPHEROID[\"International 1924\",6378388,297");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6230");
    add_srs_wkt (p, 3,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",30],PARAMETER[\"scale_factor\",1");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"false_easting\",500000],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "northing\",0],AUTHORITY[\"EPSG\",\"2320\"],AXIS[\"X\",NO");
    add_srs_wkt (p, 11, "RTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2321, "epsg", 2321, "ED50 / TM33");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=33 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / TM33\",GEOGCS[\"ED50\",DATUM[\"European_");
    add_srs_wkt (p, 1,
		 "Datum_1950\",SPHEROID[\"International 1924\",6378388,297");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6230");
    add_srs_wkt (p, 3,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",33],PARAMETER[\"scale_factor\",1");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"false_easting\",500000],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "northing\",0],AUTHORITY[\"EPSG\",\"2321\"],AXIS[\"X\",NO");
    add_srs_wkt (p, 11, "RTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2322, "epsg", 2322, "ED50 / TM36");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=36 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / TM36\",GEOGCS[\"ED50\",DATUM[\"European_");
    add_srs_wkt (p, 1,
		 "Datum_1950\",SPHEROID[\"International 1924\",6378388,297");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6230");
    add_srs_wkt (p, 3,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",36],PARAMETER[\"scale_factor\",1");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"false_easting\",500000],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "northing\",0],AUTHORITY[\"EPSG\",\"2322\"],AXIS[\"X\",NO");
    add_srs_wkt (p, 11, "RTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2323, "epsg", 2323, "ED50 / TM39");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=39 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / TM39\",GEOGCS[\"ED50\",DATUM[\"European_");
    add_srs_wkt (p, 1,
		 "Datum_1950\",SPHEROID[\"International 1924\",6378388,297");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6230");
    add_srs_wkt (p, 3,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",39],PARAMETER[\"scale_factor\",1");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"false_easting\",500000],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "northing\",0],AUTHORITY[\"EPSG\",\"2323\"],AXIS[\"X\",NO");
    add_srs_wkt (p, 11, "RTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2324, "epsg", 2324, "ED50 / TM42");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=42 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / TM42\",GEOGCS[\"ED50\",DATUM[\"European_");
    add_srs_wkt (p, 1,
		 "Datum_1950\",SPHEROID[\"International 1924\",6378388,297");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6230");
    add_srs_wkt (p, 3,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",42],PARAMETER[\"scale_factor\",1");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"false_easting\",500000],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "northing\",0],AUTHORITY[\"EPSG\",\"2324\"],AXIS[\"X\",NO");
    add_srs_wkt (p, 11, "RTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2325, "epsg", 2325, "ED50 / TM45");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=45 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / TM45\",GEOGCS[\"ED50\",DATUM[\"European_");
    add_srs_wkt (p, 1,
		 "Datum_1950\",SPHEROID[\"International 1924\",6378388,297");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6230");
    add_srs_wkt (p, 3,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",45],PARAMETER[\"scale_factor\",1");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"false_easting\",500000],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "northing\",0],AUTHORITY[\"EPSG\",\"2325\"],AXIS[\"X\",NO");
    add_srs_wkt (p, 11, "RTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2326, "epsg", 2326,
		      "Hong Kong 1980 Grid System");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=22.31213333333334 +lon_0=114.17855555");
    add_proj4text (p, 1,
		   "55556 +k=1 +x_0=836694.05 +y_0=819069.8 +ellps=intl +tow");
    add_proj4text (p, 2,
		   "gs84=-162.619,-276.959,-161.764,0.067753,-2.24365,-1.158");
    add_proj4text (p, 3, "83,-1.09425 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Hong Kong 1980 Grid System\",GEOGCS[\"Hong Kong");
    add_srs_wkt (p, 1,
		 " 1980\",DATUM[\"Hong_Kong_1980\",SPHEROID[\"Internationa");
    add_srs_wkt (p, 2,
		 "l 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS");
    add_srs_wkt (p, 3,
		 "84[-162.619,-276.959,-161.764,0.067753,-2.24365,-1.15883");
    add_srs_wkt (p, 4,
		 ",-1.09425],AUTHORITY[\"EPSG\",\"6611\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 5,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 6,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"4611\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 8,
		 ",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"latitude_of_origin\",22.31213333333334],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "central_meridian\",114.1785555555556],PARAMETER[\"scale_");
    add_srs_wkt (p, 11,
		 "factor\",1],PARAMETER[\"false_easting\",836694.05],PARAM");
    add_srs_wkt (p, 12,
		 "ETER[\"false_northing\",819069.8],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 13,
		 "326\"],AXIS[\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 2327, "epsg", 2327,
		      "Xian 1980 / Gauss-Kruger zone 13");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=75 +k=1 +x_0=13500000 +y_0=0");
    add_proj4text (p, 1, " +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger zone 13\",GEOGCS[\"Xia");
    add_srs_wkt (p, 1,
		 "n 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",6378");
    add_srs_wkt (p, 2,
		 "140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",75],PARAMETER[\"scale_");
    add_srs_wkt (p, 9,
		 "factor\",1],PARAMETER[\"false_easting\",13500000],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2327\"],A");
    add_srs_wkt (p, 11, "XIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2328, "epsg", 2328,
		      "Xian 1980 / Gauss-Kruger zone 14");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=81 +k=1 +x_0=14500000 +y_0=0");
    add_proj4text (p, 1, " +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger zone 14\",GEOGCS[\"Xia");
    add_srs_wkt (p, 1,
		 "n 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",6378");
    add_srs_wkt (p, 2,
		 "140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",81],PARAMETER[\"scale_");
    add_srs_wkt (p, 9,
		 "factor\",1],PARAMETER[\"false_easting\",14500000],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2328\"],A");
    add_srs_wkt (p, 11, "XIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2329, "epsg", 2329,
		      "Xian 1980 / Gauss-Kruger zone 15");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=87 +k=1 +x_0=15500000 +y_0=0");
    add_proj4text (p, 1, " +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger zone 15\",GEOGCS[\"Xia");
    add_srs_wkt (p, 1,
		 "n 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",6378");
    add_srs_wkt (p, 2,
		 "140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",87],PARAMETER[\"scale_");
    add_srs_wkt (p, 9,
		 "factor\",1],PARAMETER[\"false_easting\",15500000],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2329\"],A");
    add_srs_wkt (p, 11, "XIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2330, "epsg", 2330,
		      "Xian 1980 / Gauss-Kruger zone 16");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=93 +k=1 +x_0=16500000 +y_0=0");
    add_proj4text (p, 1, " +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger zone 16\",GEOGCS[\"Xia");
    add_srs_wkt (p, 1,
		 "n 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",6378");
    add_srs_wkt (p, 2,
		 "140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",93],PARAMETER[\"scale_");
    add_srs_wkt (p, 9,
		 "factor\",1],PARAMETER[\"false_easting\",16500000],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2330\"],A");
    add_srs_wkt (p, 11, "XIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2331, "epsg", 2331,
		      "Xian 1980 / Gauss-Kruger zone 17");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=99 +k=1 +x_0=17500000 +y_0=0");
    add_proj4text (p, 1, " +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger zone 17\",GEOGCS[\"Xia");
    add_srs_wkt (p, 1,
		 "n 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",6378");
    add_srs_wkt (p, 2,
		 "140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",99],PARAMETER[\"scale_");
    add_srs_wkt (p, 9,
		 "factor\",1],PARAMETER[\"false_easting\",17500000],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2331\"],A");
    add_srs_wkt (p, 11, "XIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2332, "epsg", 2332,
		      "Xian 1980 / Gauss-Kruger zone 18");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=18500000 +y_0=");
    add_proj4text (p, 1, "0 +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger zone 18\",GEOGCS[\"Xia");
    add_srs_wkt (p, 1,
		 "n 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",6378");
    add_srs_wkt (p, 2,
		 "140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",105],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",1],PARAMETER[\"false_easting\",18500000],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2332\"],");
    add_srs_wkt (p, 11, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2333, "epsg", 2333,
		      "Xian 1980 / Gauss-Kruger zone 19");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=19500000 +y_0=");
    add_proj4text (p, 1, "0 +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger zone 19\",GEOGCS[\"Xia");
    add_srs_wkt (p, 1,
		 "n 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",6378");
    add_srs_wkt (p, 2,
		 "140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",111],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",1],PARAMETER[\"false_easting\",19500000],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2333\"],");
    add_srs_wkt (p, 11, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2334, "epsg", 2334,
		      "Xian 1980 / Gauss-Kruger zone 20");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=1 +x_0=20500000 +y_0=");
    add_proj4text (p, 1, "0 +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger zone 20\",GEOGCS[\"Xia");
    add_srs_wkt (p, 1,
		 "n 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",6378");
    add_srs_wkt (p, 2,
		 "140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",117],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",1],PARAMETER[\"false_easting\",20500000],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2334\"],");
    add_srs_wkt (p, 11, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2335, "epsg", 2335,
		      "Xian 1980 / Gauss-Kruger zone 21");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=1 +x_0=21500000 +y_0=");
    add_proj4text (p, 1, "0 +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger zone 21\",GEOGCS[\"Xia");
    add_srs_wkt (p, 1,
		 "n 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",6378");
    add_srs_wkt (p, 2,
		 "140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",123],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",1],PARAMETER[\"false_easting\",21500000],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2335\"],");
    add_srs_wkt (p, 11, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2336, "epsg", 2336,
		      "Xian 1980 / Gauss-Kruger zone 22");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=129 +k=1 +x_0=22500000 +y_0=");
    add_proj4text (p, 1, "0 +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger zone 22\",GEOGCS[\"Xia");
    add_srs_wkt (p, 1,
		 "n 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",6378");
    add_srs_wkt (p, 2,
		 "140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",129],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",1],PARAMETER[\"false_easting\",22500000],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2336\"],");
    add_srs_wkt (p, 11, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2337, "epsg", 2337,
		      "Xian 1980 / Gauss-Kruger zone 23");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=135 +k=1 +x_0=23500000 +y_0=");
    add_proj4text (p, 1, "0 +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger zone 23\",GEOGCS[\"Xia");
    add_srs_wkt (p, 1,
		 "n 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",6378");
    add_srs_wkt (p, 2,
		 "140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",135],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",1],PARAMETER[\"false_easting\",23500000],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2337\"],");
    add_srs_wkt (p, 11, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2338, "epsg", 2338,
		      "Xian 1980 / Gauss-Kruger CM 75E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=75 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger CM 75E\",GEOGCS[\"Xian");
    add_srs_wkt (p, 1,
		 " 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",63781");
    add_srs_wkt (p, 2,
		 "40,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",75],PARAMETER[\"scale_");
    add_srs_wkt (p, 9,
		 "factor\",1],PARAMETER[\"false_easting\",500000],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2338\"],AXI");
    add_srs_wkt (p, 11, "S[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2339, "epsg", 2339,
		      "Xian 1980 / Gauss-Kruger CM 81E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=81 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger CM 81E\",GEOGCS[\"Xian");
    add_srs_wkt (p, 1,
		 " 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",63781");
    add_srs_wkt (p, 2,
		 "40,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",81],PARAMETER[\"scale_");
    add_srs_wkt (p, 9,
		 "factor\",1],PARAMETER[\"false_easting\",500000],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2339\"],AXI");
    add_srs_wkt (p, 11, "S[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2340, "epsg", 2340,
		      "Xian 1980 / Gauss-Kruger CM 87E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=87 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger CM 87E\",GEOGCS[\"Xian");
    add_srs_wkt (p, 1,
		 " 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",63781");
    add_srs_wkt (p, 2,
		 "40,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",87],PARAMETER[\"scale_");
    add_srs_wkt (p, 9,
		 "factor\",1],PARAMETER[\"false_easting\",500000],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2340\"],AXI");
    add_srs_wkt (p, 11, "S[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2341, "epsg", 2341,
		      "Xian 1980 / Gauss-Kruger CM 93E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=93 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger CM 93E\",GEOGCS[\"Xian");
    add_srs_wkt (p, 1,
		 " 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",63781");
    add_srs_wkt (p, 2,
		 "40,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",93],PARAMETER[\"scale_");
    add_srs_wkt (p, 9,
		 "factor\",1],PARAMETER[\"false_easting\",500000],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2341\"],AXI");
    add_srs_wkt (p, 11, "S[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2342, "epsg", 2342,
		      "Xian 1980 / Gauss-Kruger CM 99E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=99 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger CM 99E\",GEOGCS[\"Xian");
    add_srs_wkt (p, 1,
		 " 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",63781");
    add_srs_wkt (p, 2,
		 "40,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",99],PARAMETER[\"scale_");
    add_srs_wkt (p, 9,
		 "factor\",1],PARAMETER[\"false_easting\",500000],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2342\"],AXI");
    add_srs_wkt (p, 11, "S[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2343, "epsg", 2343,
		      "Xian 1980 / Gauss-Kruger CM 105E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger CM 105E\",GEOGCS[\"Xia");
    add_srs_wkt (p, 1,
		 "n 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",6378");
    add_srs_wkt (p, 2,
		 "140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",105],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",1],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2343\"],AX");
    add_srs_wkt (p, 11, "IS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2344, "epsg", 2344,
		      "Xian 1980 / Gauss-Kruger CM 111E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger CM 111E\",GEOGCS[\"Xia");
    add_srs_wkt (p, 1,
		 "n 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",6378");
    add_srs_wkt (p, 2,
		 "140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",111],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",1],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2344\"],AX");
    add_srs_wkt (p, 11, "IS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2345, "epsg", 2345,
		      "Xian 1980 / Gauss-Kruger CM 117E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger CM 117E\",GEOGCS[\"Xia");
    add_srs_wkt (p, 1,
		 "n 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",6378");
    add_srs_wkt (p, 2,
		 "140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",117],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",1],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2345\"],AX");
    add_srs_wkt (p, 11, "IS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2346, "epsg", 2346,
		      "Xian 1980 / Gauss-Kruger CM 123E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger CM 123E\",GEOGCS[\"Xia");
    add_srs_wkt (p, 1,
		 "n 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",6378");
    add_srs_wkt (p, 2,
		 "140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",123],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",1],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2346\"],AX");
    add_srs_wkt (p, 11, "IS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2347, "epsg", 2347,
		      "Xian 1980 / Gauss-Kruger CM 129E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=129 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger CM 129E\",GEOGCS[\"Xia");
    add_srs_wkt (p, 1,
		 "n 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",6378");
    add_srs_wkt (p, 2,
		 "140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",129],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",1],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2347\"],AX");
    add_srs_wkt (p, 11, "IS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2348, "epsg", 2348,
		      "Xian 1980 / Gauss-Kruger CM 135E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=135 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / Gauss-Kruger CM 135E\",GEOGCS[\"Xia");
    add_srs_wkt (p, 1,
		 "n 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 1980\",6378");
    add_srs_wkt (p, 2,
		 "140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",135],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",1],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2348\"],AX");
    add_srs_wkt (p, 11, "IS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2349, "epsg", 2349,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 25");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=75 +k=1 +x_0=25500000 +y_0=0");
    add_proj4text (p, 1, " +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 25\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",75],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",255000");
    add_srs_wkt (p, 10,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2349\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2350, "epsg", 2350,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 26");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=78 +k=1 +x_0=26500000 +y_0=0");
    add_proj4text (p, 1, " +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 26\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",78],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",265000");
    add_srs_wkt (p, 10,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2350\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2351, "epsg", 2351,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 27");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=81 +k=1 +x_0=27500000 +y_0=0");
    add_proj4text (p, 1, " +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 27\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",81],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",275000");
    add_srs_wkt (p, 10,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2351\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2352, "epsg", 2352,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 28");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=84 +k=1 +x_0=28500000 +y_0=0");
    add_proj4text (p, 1, " +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 28\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",84],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",285000");
    add_srs_wkt (p, 10,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2352\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2353, "epsg", 2353,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 29");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=87 +k=1 +x_0=29500000 +y_0=0");
    add_proj4text (p, 1, " +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 29\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",87],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",295000");
    add_srs_wkt (p, 10,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2353\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2354, "epsg", 2354,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 30");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=90 +k=1 +x_0=30500000 +y_0=0");
    add_proj4text (p, 1, " +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 30\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",90],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",305000");
    add_srs_wkt (p, 10,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2354\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2355, "epsg", 2355,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 31");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=93 +k=1 +x_0=31500000 +y_0=0");
    add_proj4text (p, 1, " +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 31\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",93],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",315000");
    add_srs_wkt (p, 10,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2355\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2356, "epsg", 2356,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 32");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=96 +k=1 +x_0=32500000 +y_0=0");
    add_proj4text (p, 1, " +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 32\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",96],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",325000");
    add_srs_wkt (p, 10,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2356\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2357, "epsg", 2357,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 33");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=99 +k=1 +x_0=33500000 +y_0=0");
    add_proj4text (p, 1, " +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 33\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",99],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",335000");
    add_srs_wkt (p, 10,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2357\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2358, "epsg", 2358,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 34");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=102 +k=1 +x_0=34500000 +y_0=");
    add_proj4text (p, 1, "0 +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 34\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",102],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",34500");
    add_srs_wkt (p, 10,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"2358\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2359, "epsg", 2359,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 35");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=35500000 +y_0=");
    add_proj4text (p, 1, "0 +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 35\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",105],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",35500");
    add_srs_wkt (p, 10,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"2359\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2360, "epsg", 2360,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 36");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=108 +k=1 +x_0=36500000 +y_0=");
    add_proj4text (p, 1, "0 +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 36\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",108],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",36500");
    add_srs_wkt (p, 10,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"2360\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2361, "epsg", 2361,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 37");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=37500000 +y_0=");
    add_proj4text (p, 1, "0 +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 37\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",111],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",37500");
    add_srs_wkt (p, 10,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"2361\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2362, "epsg", 2362,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 38");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=114 +k=1 +x_0=38500000 +y_0=");
    add_proj4text (p, 1, "0 +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 38\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",114],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",38500");
    add_srs_wkt (p, 10,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"2362\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2363, "epsg", 2363,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 39");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=1 +x_0=39500000 +y_0=");
    add_proj4text (p, 1, "0 +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 39\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",117],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",39500");
    add_srs_wkt (p, 10,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"2363\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2364, "epsg", 2364,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 40");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=120 +k=1 +x_0=40500000 +y_0=");
    add_proj4text (p, 1, "0 +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 40\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",120],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",40500");
    add_srs_wkt (p, 10,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"2364\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2365, "epsg", 2365,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 41");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=1 +x_0=41500000 +y_0=");
    add_proj4text (p, 1, "0 +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 41\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",123],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",41500");
    add_srs_wkt (p, 10,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"2365\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2366, "epsg", 2366,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 42");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=126 +k=1 +x_0=42500000 +y_0=");
    add_proj4text (p, 1, "0 +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 42\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",126],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",42500");
    add_srs_wkt (p, 10,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"2366\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2367, "epsg", 2367,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 43");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=129 +k=1 +x_0=43500000 +y_0=");
    add_proj4text (p, 1, "0 +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 43\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",129],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",43500");
    add_srs_wkt (p, 10,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"2367\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2368, "epsg", 2368,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 44");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=132 +k=1 +x_0=44500000 +y_0=");
    add_proj4text (p, 1, "0 +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 44\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",132],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",44500");
    add_srs_wkt (p, 10,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"2368\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2369, "epsg", 2369,
		      "Xian 1980 / 3-degree Gauss-Kruger zone 45");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=135 +k=1 +x_0=45500000 +y_0=");
    add_proj4text (p, 1, "0 +a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger zone 45\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",135],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",45500");
    add_srs_wkt (p, 10,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"2369\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2370, "epsg", 2370,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 75E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=75 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 75E\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 198");
    add_srs_wkt (p, 2,
		 "0\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",75],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",1],PARAMETER[\"false_easting\",500000]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23");
    add_srs_wkt (p, 11, "70\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2371, "epsg", 2371,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 78E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=78 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 78E\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 198");
    add_srs_wkt (p, 2,
		 "0\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",78],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",1],PARAMETER[\"false_easting\",500000]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23");
    add_srs_wkt (p, 11, "71\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2372, "epsg", 2372,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 81E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=81 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 81E\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 198");
    add_srs_wkt (p, 2,
		 "0\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",81],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",1],PARAMETER[\"false_easting\",500000]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23");
    add_srs_wkt (p, 11, "72\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2373, "epsg", 2373,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 84E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=84 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 84E\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 198");
    add_srs_wkt (p, 2,
		 "0\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",84],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",1],PARAMETER[\"false_easting\",500000]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23");
    add_srs_wkt (p, 11, "73\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2374, "epsg", 2374,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 87E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=87 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 87E\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 198");
    add_srs_wkt (p, 2,
		 "0\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",87],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",1],PARAMETER[\"false_easting\",500000]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23");
    add_srs_wkt (p, 11, "74\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2375, "epsg", 2375,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 90E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=90 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 90E\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 198");
    add_srs_wkt (p, 2,
		 "0\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",90],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",1],PARAMETER[\"false_easting\",500000]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23");
    add_srs_wkt (p, 11, "75\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2376, "epsg", 2376,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 93E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=93 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 93E\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 198");
    add_srs_wkt (p, 2,
		 "0\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",93],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",1],PARAMETER[\"false_easting\",500000]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23");
    add_srs_wkt (p, 11, "76\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2377, "epsg", 2377,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 96E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=96 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 96E\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 198");
    add_srs_wkt (p, 2,
		 "0\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",96],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",1],PARAMETER[\"false_easting\",500000]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23");
    add_srs_wkt (p, 11, "77\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2378, "epsg", 2378,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 99E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=99 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 99E\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 198");
    add_srs_wkt (p, 2,
		 "0\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4610");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",99],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",1],PARAMETER[\"false_easting\",500000]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23");
    add_srs_wkt (p, 11, "78\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2379, "epsg", 2379,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 102E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=102 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 102E\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",102],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2379\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2380, "epsg", 2380,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 105E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 105E\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",105],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2380\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2381, "epsg", 2381,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 108E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=108 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 108E\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",108],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2381\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2382, "epsg", 2382,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 111E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 111E\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",111],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2382\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2383, "epsg", 2383,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 114E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=114 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 114E\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",114],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2383\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2384, "epsg", 2384,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 117E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 117E\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",117],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2384\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2385, "epsg", 2385,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 120E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=120 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 120E\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",120],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2385\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2386, "epsg", 2386,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 123E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 123E\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",123],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2386\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2387, "epsg", 2387,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 126E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=126 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 126E\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",126],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2387\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2388, "epsg", 2388,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 129E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=129 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 129E\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",129],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2388\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2389, "epsg", 2389,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 132E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=132 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 132E\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",132],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2389\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2390, "epsg", 2390,
		      "Xian 1980 / 3-degree Gauss-Kruger CM 135E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=135 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+a=6378140 +b=6356755.288157528 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Xian 1980 / 3-degree Gauss-Kruger CM 135E\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian 19");
    add_srs_wkt (p, 2,
		 "80\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"461");
    add_srs_wkt (p, 6,
		 "0\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",135],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "2390\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2391, "epsg", 2391, "KKJ / Finland zone 1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=1 +x_0=1500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"KKJ / Finland zone 1\",GEOGCS[\"KKJ\",DATUM[\"K");
    add_srs_wkt (p, 1,
		 "artastokoordinaattijarjestelma_1966\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6123\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"4123\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",21],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",1");
    add_srs_wkt (p, 10,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 11, "\",\"2391\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2392, "epsg", 2392, "KKJ / Finland zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=24 +k=1 +x_0=2500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"KKJ / Finland zone 2\",GEOGCS[\"KKJ\",DATUM[\"K");
    add_srs_wkt (p, 1,
		 "artastokoordinaattijarjestelma_1966\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6123\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"4123\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",24],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",2");
    add_srs_wkt (p, 10,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 11, "\",\"2392\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2393, "epsg", 2393,
		      "KKJ / Finland Uniform Coordinate System");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=27 +k=1 +x_0=3500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"KKJ / Finland Uniform Coordinate System\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"KKJ\",DATUM[\"Kartastokoordinaattijarjestelma_1966\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6123\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4123\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",27],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",3500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"2393\"],AXIS[\"X\",NORTH],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",EAST]]");
    p = add_epsg_def (first, last, 2394, "epsg", 2394, "KKJ / Finland zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=30 +k=1 +x_0=4500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"KKJ / Finland zone 4\",GEOGCS[\"KKJ\",DATUM[\"K");
    add_srs_wkt (p, 1,
		 "artastokoordinaattijarjestelma_1966\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6123\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"4123\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",30],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",4");
    add_srs_wkt (p, 10,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 11, "\",\"2394\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2395, "epsg", 2395,
		      "South Yemen / Gauss-Kruger zone 8");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=45 +k=1 +x_0=8500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=-76,-138,67,0,0,0,0 +units=m +no_d");
    add_proj4text (p, 2, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"South Yemen / Gauss-Kruger zone 8\",GEOGCS[\"So");
    add_srs_wkt (p, 1,
		 "uth Yemen\",DATUM[\"South_Yemen\",SPHEROID[\"Krassowsky ");
    add_srs_wkt (p, 2,
		 "1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS");
    add_srs_wkt (p, 3,
		 "84[-76,-138,67,0,0,0,0],AUTHORITY[\"EPSG\",\"6164\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6,
		 "\"]],AUTHORITY[\"EPSG\",\"4164\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"ce");
    add_srs_wkt (p, 9,
		 "ntral_meridian\",45],PARAMETER[\"scale_factor\",1],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_easting\",8500000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 11,
		 "ng\",0],AUTHORITY[\"EPSG\",\"2395\"],AXIS[\"X\",NORTH],A");
    add_srs_wkt (p, 12, "XIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2396, "epsg", 2396,
		      "South Yemen / Gauss-Kruger zone 9");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=51 +k=1 +x_0=9500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=-76,-138,67,0,0,0,0 +units=m +no_d");
    add_proj4text (p, 2, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"South Yemen / Gauss-Kruger zone 9\",GEOGCS[\"So");
    add_srs_wkt (p, 1,
		 "uth Yemen\",DATUM[\"South_Yemen\",SPHEROID[\"Krassowsky ");
    add_srs_wkt (p, 2,
		 "1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS");
    add_srs_wkt (p, 3,
		 "84[-76,-138,67,0,0,0,0],AUTHORITY[\"EPSG\",\"6164\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6,
		 "\"]],AUTHORITY[\"EPSG\",\"4164\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"ce");
    add_srs_wkt (p, 9,
		 "ntral_meridian\",51],PARAMETER[\"scale_factor\",1],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_easting\",9500000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 11,
		 "ng\",0],AUTHORITY[\"EPSG\",\"2396\"],AXIS[\"X\",NORTH],A");
    add_srs_wkt (p, 12, "XIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2397, "epsg", 2397,
		      "Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=9 +k=1 +x_0=3500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 3");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo_1942_83\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6178\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4178\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",9],PARAMETER[\"scale_factor\",1],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_easting\",3500000],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 11,
		 "UTHORITY[\"EPSG\",\"2397\"],AXIS[\"X\",NORTH],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",EAST]]");
    p = add_epsg_def (first, last, 2398, "epsg", 2398,
		      "Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=12 +k=1 +x_0=4500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 4");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo_1942_83\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6178\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4178\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",12],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",4500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"2398\"],AXIS[\"X\",NORTH],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",EAST]]");
    p = add_epsg_def (first, last, 2399, "epsg", 2399,
		      "Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=1 +x_0=5500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 5");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo_1942_83\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6178\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4178\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",15],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",5500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"2399\"],AXIS[\"X\",NORTH],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",EAST]]");
    p = add_epsg_def (first, last, 2400, "epsg", 2400,
		      "RT90 2.5 gon W (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15.80827777777778 +k=1 +x_0=");
    add_proj4text (p, 1, "1500000 +y_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RT90 2.5 gon W (deprecated)\",GEOGCS[\"RT90\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Rikets_koordinatsystem_1990\",SPHEROID[\"Bessel 1");
    add_srs_wkt (p, 2,
		 "841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6124\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4124\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"lat");
    add_srs_wkt (p, 8,
		 "itude_of_origin\",0],PARAMETER[\"central_meridian\",15.8");
    add_srs_wkt (p, 9,
		 "0827777777778],PARAMETER[\"scale_factor\",1],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",1500000],PARAMETER[\"false_northing\",0]");
    add_srs_wkt (p, 11,
		 ",AUTHORITY[\"EPSG\",\"2400\"],AXIS[\"X\",NORTH],AXIS[\"Y");
    add_srs_wkt (p, 12, "\",EAST]]");
    p = add_epsg_def (first, last, 2401, "epsg", 2401,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 25");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=75 +k=1 +x_0=25500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 25\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",7");
    add_srs_wkt (p, 9,
		 "5],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",25500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"2401\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2402, "epsg", 2402,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 26");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=78 +k=1 +x_0=26500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 26\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",7");
    add_srs_wkt (p, 9,
		 "8],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",26500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"2402\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2403, "epsg", 2403,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 27");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=81 +k=1 +x_0=27500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 27\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",8");
    add_srs_wkt (p, 9,
		 "1],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",27500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"2403\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2404, "epsg", 2404,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 28");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=84 +k=1 +x_0=28500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 28\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",8");
    add_srs_wkt (p, 9,
		 "4],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",28500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"2404\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2405, "epsg", 2405,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 29");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=87 +k=1 +x_0=29500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 29\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",8");
    add_srs_wkt (p, 9,
		 "7],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",29500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"2405\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2406, "epsg", 2406,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 30");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=90 +k=1 +x_0=30500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 30\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",9");
    add_srs_wkt (p, 9,
		 "0],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",30500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"2406\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2407, "epsg", 2407,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 31");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=93 +k=1 +x_0=31500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 31\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",9");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",31500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"2407\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2408, "epsg", 2408,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 32");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=96 +k=1 +x_0=32500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 32\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",9");
    add_srs_wkt (p, 9,
		 "6],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",32500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"2408\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2409, "epsg", 2409,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 33");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=99 +k=1 +x_0=33500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 33\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",9");
    add_srs_wkt (p, 9,
		 "9],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",33500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"2409\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2410, "epsg", 2410,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 34");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=102 +k=1 +x_0=34500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 34\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "02],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",34500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"2410\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 2411, "epsg", 2411,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 35");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=35500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 35\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "05],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",35500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"2411\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 2412, "epsg", 2412,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 36");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=108 +k=1 +x_0=36500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 36\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "08],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",36500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"2412\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 2413, "epsg", 2413,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 37");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=37500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 37\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "11],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",37500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"2413\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 2414, "epsg", 2414,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 38");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=114 +k=1 +x_0=38500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 38\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "14],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",38500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"2414\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_03 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 2415, "epsg", 2415,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 39");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=1 +x_0=39500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 39\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "17],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",39500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"2415\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 2416, "epsg", 2416,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 40");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=120 +k=1 +x_0=40500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 40\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "20],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",40500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"2416\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 2417, "epsg", 2417,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 41");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=1 +x_0=41500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 41\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "23],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",41500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"2417\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 2418, "epsg", 2418,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 42");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=126 +k=1 +x_0=42500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 42\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "26],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",42500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"2418\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 2419, "epsg", 2419,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 43");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=129 +k=1 +x_0=43500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 43\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "29],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",43500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"2419\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 2420, "epsg", 2420,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 44");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=132 +k=1 +x_0=44500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 44\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "32],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",44500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"2420\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 2421, "epsg", 2421,
		      "Beijing 1954 / 3-degree Gauss-Kruger zone 45");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=135 +k=1 +x_0=45500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger zone 45\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "35],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",45500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"2421\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 2422, "epsg", 2422,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 75E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=75 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 75E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",7");
    add_srs_wkt (p, 9,
		 "5],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11, "EPSG\",\"2422\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2423, "epsg", 2423,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 78E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=78 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 78E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",7");
    add_srs_wkt (p, 9,
		 "8],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11, "EPSG\",\"2423\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2424, "epsg", 2424,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 81E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=81 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 81E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",8");
    add_srs_wkt (p, 9,
		 "1],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11, "EPSG\",\"2424\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2425, "epsg", 2425,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 84E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=84 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 84E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",8");
    add_srs_wkt (p, 9,
		 "4],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11, "EPSG\",\"2425\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2426, "epsg", 2426,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 87E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=87 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 87E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",8");
    add_srs_wkt (p, 9,
		 "7],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11, "EPSG\",\"2426\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2427, "epsg", 2427,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 90E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=90 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 90E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",9");
    add_srs_wkt (p, 9,
		 "0],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11, "EPSG\",\"2427\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2428, "epsg", 2428,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 93E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=93 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 93E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",9");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11, "EPSG\",\"2428\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2429, "epsg", 2429,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 96E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=96 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 96E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",9");
    add_srs_wkt (p, 9,
		 "6],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11, "EPSG\",\"2429\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2430, "epsg", 2430,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 99E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=99 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 99E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",9");
    add_srs_wkt (p, 9,
		 "9],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11, "EPSG\",\"2430\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2431, "epsg", 2431,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 102E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=102 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 102E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "02],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"2431\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2432, "epsg", 2432,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 105E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 105E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "05],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"2432\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2433, "epsg", 2433,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 108E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=108 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 108E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "08],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"2433\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2434, "epsg", 2434,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 111E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 111E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "11],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"2434\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2435, "epsg", 2435,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 114E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=114 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 114E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "14],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"2435\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2436, "epsg", 2436,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 117E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 117E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "17],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"2436\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2437, "epsg", 2437,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 120E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=120 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 120E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "20],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"2437\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2438, "epsg", 2438,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 123E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 123E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "23],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"2438\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2439, "epsg", 2439,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 126E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=126 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 126E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "26],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"2439\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2440, "epsg", 2440,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 129E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=129 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 129E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "29],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"2440\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2441, "epsg", 2441,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 132E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=132 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 132E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "32],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"2441\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2442, "epsg", 2442,
		      "Beijing 1954 / 3-degree Gauss-Kruger CM 135E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=135 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / 3-degree Gauss-Kruger CM 135E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "35],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"2442\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2443, "epsg", 2443,
		      "JGD2000 / Japan Plane Rectangular CS I");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=33 +lon_0=129.5 +k=0.9999 +x_0=0 +y_0");
    add_proj4text (p, 1,
		   "=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / Japan Plane Rectangular CS I\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",33],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"central_meridian\",129.5],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.9999],PARAMETER[\"false_easting\",0],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2443\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2444, "epsg", 2444,
		      "JGD2000 / Japan Plane Rectangular CS II");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=33 +lon_0=131 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / Japan Plane Rectangular CS II\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",33],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"central_meridian\",131],PARAMETER[\"scale_fa");
    add_srs_wkt (p, 10,
		 "ctor\",0.9999],PARAMETER[\"false_easting\",0],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_northing\",0],AUTHORITY[\"EPSG\",\"2444\"],AXIS[");
    add_srs_wkt (p, 12, "\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2445, "epsg", 2445,
		      "JGD2000 / Japan Plane Rectangular CS III");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36 +lon_0=132.1666666666667 +k=0.9999");
    add_proj4text (p, 1,
		   " +x_0=0 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +unit");
    add_proj4text (p, 2, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / Japan Plane Rectangular CS III\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",36],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"central_meridian\",132.1666666666667],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",0.9999],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2445\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2446, "epsg", 2446,
		      "JGD2000 / Japan Plane Rectangular CS IV");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=33 +lon_0=133.5 +k=0.9999 +x_0=0 +y_0");
    add_proj4text (p, 1,
		   "=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / Japan Plane Rectangular CS IV\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",33],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"central_meridian\",133.5],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.9999],PARAMETER[\"false_easting\",0],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2446\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2447, "epsg", 2447,
		      "JGD2000 / Japan Plane Rectangular CS V");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36 +lon_0=134.3333333333333 +k=0.9999");
    add_proj4text (p, 1,
		   " +x_0=0 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +unit");
    add_proj4text (p, 2, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / Japan Plane Rectangular CS V\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",36],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"central_meridian\",134.3333333333333],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",0.9999],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2447\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2448, "epsg", 2448,
		      "JGD2000 / Japan Plane Rectangular CS VI");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36 +lon_0=136 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / Japan Plane Rectangular CS VI\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",36],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"central_meridian\",136],PARAMETER[\"scale_fa");
    add_srs_wkt (p, 10,
		 "ctor\",0.9999],PARAMETER[\"false_easting\",0],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_northing\",0],AUTHORITY[\"EPSG\",\"2448\"],AXIS[");
    add_srs_wkt (p, 12, "\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2449, "epsg", 2449,
		      "JGD2000 / Japan Plane Rectangular CS VII");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36 +lon_0=137.1666666666667 +k=0.9999");
    add_proj4text (p, 1,
		   " +x_0=0 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +unit");
    add_proj4text (p, 2, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / Japan Plane Rectangular CS VII\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",36],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"central_meridian\",137.1666666666667],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",0.9999],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2449\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2450, "epsg", 2450,
		      "JGD2000 / Japan Plane Rectangular CS VIII");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36 +lon_0=138.5 +k=0.9999 +x_0=0 +y_0");
    add_proj4text (p, 1,
		   "=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / Japan Plane Rectangular CS VIII\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]],UNIT[");
    add_srs_wkt (p, 7,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",3");
    add_srs_wkt (p, 9,
		 "6],PARAMETER[\"central_meridian\",138.5],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",0.9999],PARAMETER[\"false_easting\",0],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2450\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2451, "epsg", 2451,
		      "JGD2000 / Japan Plane Rectangular CS IX");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36 +lon_0=139.8333333333333 +k=0.9999");
    add_proj4text (p, 1,
		   " +x_0=0 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +unit");
    add_proj4text (p, 2, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / Japan Plane Rectangular CS IX\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",36],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"central_meridian\",139.8333333333333],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",0.9999],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2451\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2452, "epsg", 2452,
		      "JGD2000 / Japan Plane Rectangular CS X");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40 +lon_0=140.8333333333333 +k=0.9999");
    add_proj4text (p, 1,
		   " +x_0=0 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +unit");
    add_proj4text (p, 2, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / Japan Plane Rectangular CS X\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",40],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"central_meridian\",140.8333333333333],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",0.9999],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2452\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2453, "epsg", 2453,
		      "JGD2000 / Japan Plane Rectangular CS XI");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=44 +lon_0=140.25 +k=0.9999 +x_0=0 +y_");
    add_proj4text (p, 1,
		   "0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / Japan Plane Rectangular CS XI\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",44],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"central_meridian\",140.25],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.9999],PARAMETER[\"false_easting\",0],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2453\"],AX");
    add_srs_wkt (p, 12, "IS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2454, "epsg", 2454,
		      "JGD2000 / Japan Plane Rectangular CS XII");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=44 +lon_0=142.25 +k=0.9999 +x_0=0 +y_");
    add_proj4text (p, 1,
		   "0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / Japan Plane Rectangular CS XII\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",44],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"central_meridian\",142.25],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.9999],PARAMETER[\"false_easting\",0],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2454\"],AX");
    add_srs_wkt (p, 12, "IS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2455, "epsg", 2455,
		      "JGD2000 / Japan Plane Rectangular CS XIII");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=44 +lon_0=144.25 +k=0.9999 +x_0=0 +y_");
    add_proj4text (p, 1,
		   "0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / Japan Plane Rectangular CS XIII\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]],UNIT[");
    add_srs_wkt (p, 7,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",4");
    add_srs_wkt (p, 9,
		 "4],PARAMETER[\"central_meridian\",144.25],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",0.9999],PARAMETER[\"false_easting\",0],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2455\"]");
    add_srs_wkt (p, 12, ",AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2456, "epsg", 2456,
		      "JGD2000 / Japan Plane Rectangular CS XIV");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=26 +lon_0=142 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / Japan Plane Rectangular CS XIV\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",26],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"central_meridian\",142],PARAMETER[\"scale_fa");
    add_srs_wkt (p, 10,
		 "ctor\",0.9999],PARAMETER[\"false_easting\",0],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_northing\",0],AUTHORITY[\"EPSG\",\"2456\"],AXIS[");
    add_srs_wkt (p, 12, "\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2457, "epsg", 2457,
		      "JGD2000 / Japan Plane Rectangular CS XV");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=26 +lon_0=127.5 +k=0.9999 +x_0=0 +y_0");
    add_proj4text (p, 1,
		   "=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / Japan Plane Rectangular CS XV\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",26],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"central_meridian\",127.5],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.9999],PARAMETER[\"false_easting\",0],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2457\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2458, "epsg", 2458,
		      "JGD2000 / Japan Plane Rectangular CS XVI");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=26 +lon_0=124 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / Japan Plane Rectangular CS XVI\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",26],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"central_meridian\",124],PARAMETER[\"scale_fa");
    add_srs_wkt (p, 10,
		 "ctor\",0.9999],PARAMETER[\"false_easting\",0],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_northing\",0],AUTHORITY[\"EPSG\",\"2458\"],AXIS[");
    add_srs_wkt (p, 12, "\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2459, "epsg", 2459,
		      "JGD2000 / Japan Plane Rectangular CS XVII");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=26 +lon_0=131 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / Japan Plane Rectangular CS XVII\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]],UNIT[");
    add_srs_wkt (p, 7,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",2");
    add_srs_wkt (p, 9,
		 "6],PARAMETER[\"central_meridian\",131],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.9999],PARAMETER[\"false_easting\",0],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2459\"],AX");
    add_srs_wkt (p, 12, "IS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2460, "epsg", 2460,
		      "JGD2000 / Japan Plane Rectangular CS XVIII");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=20 +lon_0=136 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / Japan Plane Rectangular CS XVIII\",GE");
    add_srs_wkt (p, 1,
		 "OGCS[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 8,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 9,
		 "20],PARAMETER[\"central_meridian\",136],PARAMETER[\"scal");
    add_srs_wkt (p, 10,
		 "e_factor\",0.9999],PARAMETER[\"false_easting\",0],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2460\"],A");
    add_srs_wkt (p, 12, "XIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2461, "epsg", 2461,
		      "JGD2000 / Japan Plane Rectangular CS XIX");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=26 +lon_0=154 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / Japan Plane Rectangular CS XIX\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",26],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"central_meridian\",154],PARAMETER[\"scale_fa");
    add_srs_wkt (p, 10,
		 "ctor\",0.9999],PARAMETER[\"false_easting\",0],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_northing\",0],AUTHORITY[\"EPSG\",\"2461\"],AXIS[");
    add_srs_wkt (p, 12, "\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2462, "epsg", 2462,
		      "Albanian 1987 / Gauss-Kruger zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=1 +x_0=4500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Albanian 1987 / Gauss-Kruger zone 4\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Albanian 1987\",DATUM[\"Albanian_1987\",SPHEROID[\"Krass");
    add_srs_wkt (p, 2,
		 "owsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6191\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4191\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",21],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 10,
		 "4500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 11, "G\",\"2462\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2463, "epsg", 2463,
		      "Pulkovo 1995 / Gauss-Kruger CM 21E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 21E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",21],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2463\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2464, "epsg", 2464,
		      "Pulkovo 1995 / Gauss-Kruger CM 27E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=27 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 27E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",27],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2464\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2465, "epsg", 2465,
		      "Pulkovo 1995 / Gauss-Kruger CM 33E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=33 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 33E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",33],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2465\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2466, "epsg", 2466,
		      "Pulkovo 1995 / Gauss-Kruger CM 39E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=39 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 39E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",39],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2466\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2467, "epsg", 2467,
		      "Pulkovo 1995 / Gauss-Kruger CM 45E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=45 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 45E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",45],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2467\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2468, "epsg", 2468,
		      "Pulkovo 1995 / Gauss-Kruger CM 51E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=51 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 51E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",51],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2468\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2469, "epsg", 2469,
		      "Pulkovo 1995 / Gauss-Kruger CM 57E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=57 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 57E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",57],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2469\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2470, "epsg", 2470,
		      "Pulkovo 1995 / Gauss-Kruger CM 63E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=63 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 63E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",63],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2470\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2471, "epsg", 2471,
		      "Pulkovo 1995 / Gauss-Kruger CM 69E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=69 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 69E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",69],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2471\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2472, "epsg", 2472,
		      "Pulkovo 1995 / Gauss-Kruger CM 75E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=75 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 75E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",75],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2472\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2473, "epsg", 2473,
		      "Pulkovo 1995 / Gauss-Kruger CM 81E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=81 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 81E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",81],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2473\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2474, "epsg", 2474,
		      "Pulkovo 1995 / Gauss-Kruger CM 87E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=87 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 87E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",87],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2474\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2475, "epsg", 2475,
		      "Pulkovo 1995 / Gauss-Kruger CM 93E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=93 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 93E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",93],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2475\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2476, "epsg", 2476,
		      "Pulkovo 1995 / Gauss-Kruger CM 99E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=99 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 99E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",99],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2476\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2477, "epsg", 2477,
		      "Pulkovo 1995 / Gauss-Kruger CM 105E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 105E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",105],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2477\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2478, "epsg", 2478,
		      "Pulkovo 1995 / Gauss-Kruger CM 111E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 111E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",111],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2478\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2479, "epsg", 2479,
		      "Pulkovo 1995 / Gauss-Kruger CM 117E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 117E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",117],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2479\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2480, "epsg", 2480,
		      "Pulkovo 1995 / Gauss-Kruger CM 123E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 123E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",123],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2480\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2481, "epsg", 2481,
		      "Pulkovo 1995 / Gauss-Kruger CM 129E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=129 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 129E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",129],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2481\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2482, "epsg", 2482,
		      "Pulkovo 1995 / Gauss-Kruger CM 135E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=135 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 135E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",135],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2482\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2483, "epsg", 2483,
		      "Pulkovo 1995 / Gauss-Kruger CM 141E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=141 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 141E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",141],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2483\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2484, "epsg", 2484,
		      "Pulkovo 1995 / Gauss-Kruger CM 147E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=147 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 147E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",147],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2484\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2485, "epsg", 2485,
		      "Pulkovo 1995 / Gauss-Kruger CM 153E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=153 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 153E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",153],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2485\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2486, "epsg", 2486,
		      "Pulkovo 1995 / Gauss-Kruger CM 159E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=159 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 159E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",159],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2486\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2487, "epsg", 2487,
		      "Pulkovo 1995 / Gauss-Kruger CM 165E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=165 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 165E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",165],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2487\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2488, "epsg", 2488,
		      "Pulkovo 1995 / Gauss-Kruger CM 171E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=171 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 171E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",171],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2488\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2489, "epsg", 2489,
		      "Pulkovo 1995 / Gauss-Kruger CM 177E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=177 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 177E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",177],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2489\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2490, "epsg", 2490,
		      "Pulkovo 1995 / Gauss-Kruger CM 177W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-177 +k=1 +x_0=500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 177W\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",-177],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2490\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2491, "epsg", 2491,
		      "Pulkovo 1995 / Gauss-Kruger CM 171W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-171 +k=1 +x_0=500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger CM 171W\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",-171],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2491\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2492, "epsg", 2492,
		      "Pulkovo 1942 / Gauss-Kruger CM 9E (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=9 +k=1 +x_0=500000 +y_0=0 +e");
    add_proj4text (p, 1, "llps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 9E (deprecated)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",9],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2492\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2493, "epsg", 2493,
		      "Pulkovo 1942 / Gauss-Kruger CM 15E (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 15E (deprecated)");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"7024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",15],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2493\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2494, "epsg", 2494,
		      "Pulkovo 1942 / Gauss-Kruger CM 21E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 21E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",21],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2494\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2495, "epsg", 2495,
		      "Pulkovo 1942 / Gauss-Kruger CM 27E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=27 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 27E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",27],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2495\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2496, "epsg", 2496,
		      "Pulkovo 1942 / Gauss-Kruger CM 33E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=33 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 33E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",33],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2496\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2497, "epsg", 2497,
		      "Pulkovo 1942 / Gauss-Kruger CM 39E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=39 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 39E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",39],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2497\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2498, "epsg", 2498,
		      "Pulkovo 1942 / Gauss-Kruger CM 45E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=45 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 45E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",45],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2498\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2499, "epsg", 2499,
		      "Pulkovo 1942 / Gauss-Kruger CM 51E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=51 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 51E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",51],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2499\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2500, "epsg", 2500,
		      "Pulkovo 1942 / Gauss-Kruger CM 57E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=57 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 57E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",57],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2500\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2501, "epsg", 2501,
		      "Pulkovo 1942 / Gauss-Kruger CM 63E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=63 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 63E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",63],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2501\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2502, "epsg", 2502,
		      "Pulkovo 1942 / Gauss-Kruger CM 69E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=69 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 69E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",69],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2502\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2503, "epsg", 2503,
		      "Pulkovo 1942 / Gauss-Kruger CM 75E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=75 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 75E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",75],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2503\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2504, "epsg", 2504,
		      "Pulkovo 1942 / Gauss-Kruger CM 81E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=81 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 81E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",81],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2504\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2505, "epsg", 2505,
		      "Pulkovo 1942 / Gauss-Kruger CM 87E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=87 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 87E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",87],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2505\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2506, "epsg", 2506,
		      "Pulkovo 1942 / Gauss-Kruger CM 93E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=93 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 93E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",93],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2506\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2507, "epsg", 2507,
		      "Pulkovo 1942 / Gauss-Kruger CM 99E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=99 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 99E\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",99],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2507\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2508, "epsg", 2508,
		      "Pulkovo 1942 / Gauss-Kruger CM 105E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 105E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",105],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2508\"]");
    add_srs_wkt (p, 12, ",AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2509, "epsg", 2509,
		      "Pulkovo 1942 / Gauss-Kruger CM 111E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 111E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",111],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2509\"]");
    add_srs_wkt (p, 12, ",AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2510, "epsg", 2510,
		      "Pulkovo 1942 / Gauss-Kruger CM 117E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 117E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",117],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2510\"]");
    add_srs_wkt (p, 12, ",AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2511, "epsg", 2511,
		      "Pulkovo 1942 / Gauss-Kruger CM 123E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 123E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",123],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2511\"]");
    add_srs_wkt (p, 12, ",AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2512, "epsg", 2512,
		      "Pulkovo 1942 / Gauss-Kruger CM 129E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=129 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 129E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",129],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2512\"]");
    add_srs_wkt (p, 12, ",AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2513, "epsg", 2513,
		      "Pulkovo 1942 / Gauss-Kruger CM 135E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=135 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 135E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",135],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2513\"]");
    add_srs_wkt (p, 12, ",AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2514, "epsg", 2514,
		      "Pulkovo 1942 / Gauss-Kruger CM 141E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=141 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 141E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",141],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2514\"]");
    add_srs_wkt (p, 12, ",AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2515, "epsg", 2515,
		      "Pulkovo 1942 / Gauss-Kruger CM 147E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=147 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 147E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",147],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2515\"]");
    add_srs_wkt (p, 12, ",AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2516, "epsg", 2516,
		      "Pulkovo 1942 / Gauss-Kruger CM 153E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=153 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 153E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",153],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2516\"]");
    add_srs_wkt (p, 12, ",AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2517, "epsg", 2517,
		      "Pulkovo 1942 / Gauss-Kruger CM 159E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=159 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 159E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",159],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2517\"]");
    add_srs_wkt (p, 12, ",AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2518, "epsg", 2518,
		      "Pulkovo 1942 / Gauss-Kruger CM 165E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=165 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 165E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",165],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2518\"]");
    add_srs_wkt (p, 12, ",AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2519, "epsg", 2519,
		      "Pulkovo 1942 / Gauss-Kruger CM 171E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=171 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 171E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",171],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2519\"]");
    add_srs_wkt (p, 12, ",AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2520, "epsg", 2520,
		      "Pulkovo 1942 / Gauss-Kruger CM 177E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=177 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 177E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",177],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2520\"]");
    add_srs_wkt (p, 12, ",AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2521, "epsg", 2521,
		      "Pulkovo 1942 / Gauss-Kruger CM 177W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-177 +k=1 +x_0=500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 177W\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",-177],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",1],PARAMETER[\"false_easting\",500000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2521\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2522, "epsg", 2522,
		      "Pulkovo 1942 / Gauss-Kruger CM 171W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-171 +k=1 +x_0=500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger CM 171W\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",-171],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",1],PARAMETER[\"false_easting\",500000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2522\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2523, "epsg", 2523,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 7");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=1 +x_0=7500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 7\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",21],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",75000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2523\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2524, "epsg", 2524,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 8");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=24 +k=1 +x_0=8500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 8\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",24],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",85000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2524\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2525, "epsg", 2525,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 9");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=27 +k=1 +x_0=9500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 9\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",27],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",95000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2525\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2526, "epsg", 2526,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 10");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=30 +k=1 +x_0=10500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 10\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",30],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",105");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2526\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2527, "epsg", 2527,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 11");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=33 +k=1 +x_0=11500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 11\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",33],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",115");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2527\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2528, "epsg", 2528,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 12");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=36 +k=1 +x_0=12500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 12\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",36],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",125");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2528\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2529, "epsg", 2529,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 13");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=39 +k=1 +x_0=13500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 13\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",39],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",135");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2529\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2530, "epsg", 2530,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 14");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=42 +k=1 +x_0=14500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 14\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",42],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",145");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2530\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2531, "epsg", 2531,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 15");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=45 +k=1 +x_0=15500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 15\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",45],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",155");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2531\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2532, "epsg", 2532,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 16");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=48 +k=1 +x_0=16500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 16\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",48],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",165");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2532\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2533, "epsg", 2533,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 17");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=51 +k=1 +x_0=17500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 17\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",51],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",175");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2533\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2534, "epsg", 2534,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 18");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=54 +k=1 +x_0=18500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 18\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",54],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",185");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2534\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2535, "epsg", 2535,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 19");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=57 +k=1 +x_0=19500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 19\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",57],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",195");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2535\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2536, "epsg", 2536,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 20");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=60 +k=1 +x_0=20500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 20\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",60],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",205");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2536\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2537, "epsg", 2537,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 21");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=63 +k=1 +x_0=21500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 21\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",63],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",215");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2537\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2538, "epsg", 2538,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 22");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=66 +k=1 +x_0=22500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 22\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",66],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",225");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2538\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2539, "epsg", 2539,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 23");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=69 +k=1 +x_0=23500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 23\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",69],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",235");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2539\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2540, "epsg", 2540,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 24");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=72 +k=1 +x_0=24500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 24\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",72],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",245");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2540\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2541, "epsg", 2541,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 25");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=75 +k=1 +x_0=25500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 25\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",75],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",255");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2541\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2542, "epsg", 2542,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 26");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=78 +k=1 +x_0=26500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 26\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",78],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",265");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2542\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_04 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 2543, "epsg", 2543,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 27");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=81 +k=1 +x_0=27500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 27\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",81],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",275");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2543\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2544, "epsg", 2544,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 28");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=84 +k=1 +x_0=28500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 28\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",84],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",285");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2544\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2545, "epsg", 2545,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 29");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=87 +k=1 +x_0=29500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 29\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",87],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",295");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2545\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2546, "epsg", 2546,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 30");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=90 +k=1 +x_0=30500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 30\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",90],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",305");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2546\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2547, "epsg", 2547,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 31");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=93 +k=1 +x_0=31500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 31\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",93],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",315");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2547\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2548, "epsg", 2548,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 32");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=96 +k=1 +x_0=32500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 32\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",96],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",325");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2548\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2549, "epsg", 2549,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 33");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=99 +k=1 +x_0=33500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 33\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",99],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",335");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2549\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2550, "epsg", 2550,
		      "Samboja / UTM zone 50S (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=50 +south +ellps=bessel +towgs84=-404.78");
    add_proj4text (p, 1, ",685.68,45.47,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Samboja / UTM zone 50S (deprecated)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Samboja\",DATUM[\"Samboja\",SPHEROID[\"Bessel 1841\",637");
    add_srs_wkt (p, 2,
		 "7397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS");
    add_srs_wkt (p, 3,
		 "84[-404.78,685.68,45.47,0,0,0,0],AUTHORITY[\"EPSG\",\"61");
    add_srs_wkt (p, 4,
		 "25\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.0174532925199433,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9108\"]],AUTHORITY[\"EPSG\",\"4125\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",117],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"2550\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2551, "epsg", 2551,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 34");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=102 +k=1 +x_0=34500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 34\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",102],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",34");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2551\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2552, "epsg", 2552,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 35");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=35500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 35\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",105],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",35");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2552\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2553, "epsg", 2553,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 36");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=108 +k=1 +x_0=36500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 36\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",108],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",36");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2553\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2554, "epsg", 2554,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 37");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=37500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 37\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",111],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",37");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2554\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2555, "epsg", 2555,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 38");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=114 +k=1 +x_0=38500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 38\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",114],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",38");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2555\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2556, "epsg", 2556,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 39");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=1 +x_0=39500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 39\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",117],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",39");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2556\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2557, "epsg", 2557,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 40");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=120 +k=1 +x_0=40500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 40\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",120],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",40");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2557\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2558, "epsg", 2558,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 41");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=1 +x_0=41500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 41\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",123],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",41");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2558\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2559, "epsg", 2559,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 42");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=126 +k=1 +x_0=42500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 42\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",126],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",42");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2559\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2560, "epsg", 2560,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 43");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=129 +k=1 +x_0=43500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 43\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",129],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",43");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2560\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2561, "epsg", 2561,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 44");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=132 +k=1 +x_0=44500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 44\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",132],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",44");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2561\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2562, "epsg", 2562,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 45");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=135 +k=1 +x_0=45500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 45\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",135],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",45");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2562\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2563, "epsg", 2563,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 46");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=138 +k=1 +x_0=46500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 46\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",138],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",46");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2563\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2564, "epsg", 2564,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 47");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=141 +k=1 +x_0=47500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 47\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",141],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",47");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2564\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2565, "epsg", 2565,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 48");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=144 +k=1 +x_0=48500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 48\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",144],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",48");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2565\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2566, "epsg", 2566,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 49");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=147 +k=1 +x_0=49500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 49\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",147],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",49");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2566\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2567, "epsg", 2567,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 50");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=150 +k=1 +x_0=50500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 50\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",150],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2567\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2568, "epsg", 2568,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 51");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=153 +k=1 +x_0=51500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 51\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",153],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",51");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2568\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2569, "epsg", 2569,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 52");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=156 +k=1 +x_0=52500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 52\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",156],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",52");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2569\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2570, "epsg", 2570,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 53");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=159 +k=1 +x_0=53500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 53\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",159],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",53");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2570\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2571, "epsg", 2571,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 54");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=162 +k=1 +x_0=54500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 54\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",162],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",54");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2571\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2572, "epsg", 2572,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 55");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=165 +k=1 +x_0=55500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 55\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",165],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",55");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2572\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2573, "epsg", 2573,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 56");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=168 +k=1 +x_0=56500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 56\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",168],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",56");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2573\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2574, "epsg", 2574,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 57");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=171 +k=1 +x_0=57500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 57\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",171],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",57");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2574\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2575, "epsg", 2575,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 58");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=174 +k=1 +x_0=58500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 58\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",174],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",58");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2575\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2576, "epsg", 2576,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 59");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=177 +k=1 +x_0=59500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 59\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",177],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",59");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2576\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2577, "epsg", 2577,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 60 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=180 +k=1 +x_0=60000000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 60 (d");
    add_srs_wkt (p, 1,
		 "eprecated)\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_194");
    add_srs_wkt (p, 2,
		 "2\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0");
    add_srs_wkt (p, 4,
		 ".85,-0.12],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 5,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 6,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 8,
		 ",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"latitude_of_origin\",0],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 10,
		 "\",180],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",60000000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 12,
		 "ORITY[\"EPSG\",\"2577\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EA");
    add_srs_wkt (p, 13, "ST]]");
    p = add_epsg_def (first, last, 2578, "epsg", 2578,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 61");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-177 +k=1 +x_0=61500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 61\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-177],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",6");
    add_srs_wkt (p, 11,
		 "1500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12, "G\",\"2578\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2579, "epsg", 2579,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 62");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-174 +k=1 +x_0=62500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 62\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-174],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",6");
    add_srs_wkt (p, 11,
		 "2500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12, "G\",\"2579\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2580, "epsg", 2580,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 63");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-171 +k=1 +x_0=63500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 63\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-171],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",6");
    add_srs_wkt (p, 11,
		 "3500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12, "G\",\"2580\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2581, "epsg", 2581,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 64");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-168 +k=1 +x_0=64500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 64\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-168],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",6");
    add_srs_wkt (p, 11,
		 "4500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12, "G\",\"2581\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2582, "epsg", 2582,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 21E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 21E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",21],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2582\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2583, "epsg", 2583,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 24E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=24 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 24E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",24],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2583\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2584, "epsg", 2584,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 27E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=27 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 27E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",27],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2584\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2585, "epsg", 2585,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 30E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=30 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 30E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",30],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2585\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2586, "epsg", 2586,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 33E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=33 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 33E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",33],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2586\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2587, "epsg", 2587,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 36E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=36 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 36E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",36],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2587\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2588, "epsg", 2588,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 39E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=39 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 39E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",39],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2588\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2589, "epsg", 2589,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 42E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=42 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 42E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",42],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2589\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2590, "epsg", 2590,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 45E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=45 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 45E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",45],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2590\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2591, "epsg", 2591,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 48E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=48 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 48E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",48],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2591\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2592, "epsg", 2592,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 51E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=51 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 51E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",51],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2592\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2593, "epsg", 2593,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 54E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=54 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 54E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",54],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2593\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2594, "epsg", 2594,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 57E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=57 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 57E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",57],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2594\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2595, "epsg", 2595,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 60E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=60 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 60E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",60],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2595\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2596, "epsg", 2596,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 63E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=63 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 63E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",63],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2596\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2597, "epsg", 2597,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 66E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=66 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 66E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",66],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2597\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2598, "epsg", 2598,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 69E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=69 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 69E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",69],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2598\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2599, "epsg", 2599,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 72E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=72 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 72E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",72],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2599\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2600, "epsg", 2600,
		      "Lietuvos Koordinoei Sistema 1994 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=24 +k=0.9998 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Lietuvos Koordinoei Sistema 1994 (deprecated)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"LKS94\",DATUM[\"Lithuania_1994_ETRS89\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "6126\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4669\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 8,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"central_meridian\",24],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",0.9998],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2600\"],AXIS");
    add_srs_wkt (p, 12, "[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2601, "epsg", 2601,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 75E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=75 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 75E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",75],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2601\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2602, "epsg", 2602,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 78E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=78 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 78E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",78],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2602\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2603, "epsg", 2603,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 81E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=81 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 81E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",81],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2603\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2604, "epsg", 2604,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 84E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=84 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 84E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",84],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2604\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2605, "epsg", 2605,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 87E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=87 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 87E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",87],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2605\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2606, "epsg", 2606,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 90E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=90 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 90E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",90],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2606\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2607, "epsg", 2607,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 93E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=93 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 93E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",93],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2607\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2608, "epsg", 2608,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 96E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=96 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 96E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",96],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2608\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2609, "epsg", 2609,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 99E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=99 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 99E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",99],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2609\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2610, "epsg", 2610,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 102E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=102 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 102E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",102],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2610\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2611, "epsg", 2611,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 105E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 105E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",105],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2611\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2612, "epsg", 2612,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 108E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=108 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 108E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",108],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2612\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2613, "epsg", 2613,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 111E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 111E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",111],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2613\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2614, "epsg", 2614,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 114E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=114 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 114E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",114],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2614\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2615, "epsg", 2615,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 117E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 117E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",117],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2615\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2616, "epsg", 2616,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 120E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=120 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 120E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",120],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2616\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2617, "epsg", 2617,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 123E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 123E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",123],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2617\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2618, "epsg", 2618,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 126E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=126 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 126E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",126],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2618\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2619, "epsg", 2619,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 129E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=129 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 129E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",129],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2619\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2620, "epsg", 2620,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 132E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=132 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 132E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",132],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2620\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2621, "epsg", 2621,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 135E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=135 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 135E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",135],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2621\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2622, "epsg", 2622,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 138E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=138 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 138E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",138],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2622\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2623, "epsg", 2623,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 141E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=141 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 141E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",141],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2623\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2624, "epsg", 2624,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 144E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=144 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 144E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",144],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2624\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2625, "epsg", 2625,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 147E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=147 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 147E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",147],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2625\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2626, "epsg", 2626,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 150E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=150 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 150E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",150],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2626\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2627, "epsg", 2627,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 153E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=153 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 153E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",153],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2627\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2628, "epsg", 2628,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 156E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=156 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 156E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",156],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2628\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2629, "epsg", 2629,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 159E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=159 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 159E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",159],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2629\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2630, "epsg", 2630,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 162E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=162 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 162E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",162],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2630\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2631, "epsg", 2631,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 165E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=165 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 165E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",165],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2631\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2632, "epsg", 2632,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 168E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=168 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 168E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",168],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2632\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2633, "epsg", 2633,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 171E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=171 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 171E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",171],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2633\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2634, "epsg", 2634,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 174E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=174 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 174E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",174],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2634\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2635, "epsg", 2635,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 177E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=177 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 177E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",177],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2635\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2636, "epsg", 2636,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 180E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=180 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 180E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",180],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2636\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2637, "epsg", 2637,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 177W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-177 +k=1 +x_0=500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 177W\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-177],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2637\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2638, "epsg", 2638,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 174W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-174 +k=1 +x_0=500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 174W\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-174],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2638\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2639, "epsg", 2639,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 171W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-171 +k=1 +x_0=500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 171W\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-171],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2639\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2640, "epsg", 2640,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 168W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-168 +k=1 +x_0=500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 168W\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-168],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2640\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2641, "epsg", 2641,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 7");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=1 +x_0=7500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 7\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",21],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",7500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"2641\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2642, "epsg", 2642,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 8");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=24 +k=1 +x_0=8500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 8\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",24],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",8500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"2642\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2643, "epsg", 2643,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 9");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=27 +k=1 +x_0=9500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 9\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",27],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",9500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"2643\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2644, "epsg", 2644,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 10");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=30 +k=1 +x_0=10500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 10\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",30],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",105");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2644\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2645, "epsg", 2645,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 11");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=33 +k=1 +x_0=11500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 11\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",33],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",115");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2645\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2646, "epsg", 2646,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 12");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=36 +k=1 +x_0=12500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 12\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",36],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",125");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2646\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2647, "epsg", 2647,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 13");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=39 +k=1 +x_0=13500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 13\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",39],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",135");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2647\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2648, "epsg", 2648,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 14");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=42 +k=1 +x_0=14500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 14\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",42],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",145");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2648\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2649, "epsg", 2649,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 15");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=45 +k=1 +x_0=15500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 15\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",45],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",155");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2649\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2650, "epsg", 2650,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 16");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=48 +k=1 +x_0=16500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 16\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",48],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",165");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2650\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2651, "epsg", 2651,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 17");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=51 +k=1 +x_0=17500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 17\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",51],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",175");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2651\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2652, "epsg", 2652,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 18");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=54 +k=1 +x_0=18500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 18\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",54],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",185");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2652\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2653, "epsg", 2653,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 19");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=57 +k=1 +x_0=19500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 19\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",57],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",195");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2653\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2654, "epsg", 2654,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 20");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=60 +k=1 +x_0=20500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 20\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",60],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",205");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2654\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2655, "epsg", 2655,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 21");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=63 +k=1 +x_0=21500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 21\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",63],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",215");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2655\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2656, "epsg", 2656,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 22");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=66 +k=1 +x_0=22500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 22\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",66],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",225");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2656\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2657, "epsg", 2657,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 23");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=69 +k=1 +x_0=23500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 23\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",69],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",235");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2657\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2658, "epsg", 2658,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 24");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=72 +k=1 +x_0=24500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 24\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",72],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",245");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2658\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2659, "epsg", 2659,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 25");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=75 +k=1 +x_0=25500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 25\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",75],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",255");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2659\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2660, "epsg", 2660,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 26");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=78 +k=1 +x_0=26500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 26\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",78],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",265");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2660\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2661, "epsg", 2661,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 27");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=81 +k=1 +x_0=27500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 27\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",81],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",275");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2661\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2662, "epsg", 2662,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 28");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=84 +k=1 +x_0=28500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 28\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",84],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",285");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2662\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2663, "epsg", 2663,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 29");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=87 +k=1 +x_0=29500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 29\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",87],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",295");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2663\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2664, "epsg", 2664,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 30");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=90 +k=1 +x_0=30500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 30\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",90],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",305");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2664\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2665, "epsg", 2665,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 31");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=93 +k=1 +x_0=31500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 31\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",93],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",315");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2665\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2666, "epsg", 2666,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 32");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=96 +k=1 +x_0=32500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 32\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",96],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",325");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2666\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2667, "epsg", 2667,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 33");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=99 +k=1 +x_0=33500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 33\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",99],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",335");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2667\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2668, "epsg", 2668,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 34");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=102 +k=1 +x_0=34500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 34\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",102],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",34");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2668\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2669, "epsg", 2669,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 35");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=35500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 35\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",105],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",35");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2669\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2670, "epsg", 2670,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 36");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=108 +k=1 +x_0=36500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 36\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",108],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",36");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2670\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_05 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 2671, "epsg", 2671,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 37");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=37500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 37\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",111],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",37");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2671\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2672, "epsg", 2672,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 38");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=114 +k=1 +x_0=38500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 38\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",114],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",38");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2672\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2673, "epsg", 2673,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 39");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=1 +x_0=39500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 39\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",117],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",39");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2673\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2674, "epsg", 2674,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 40");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=120 +k=1 +x_0=40500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 40\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",120],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",40");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2674\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2675, "epsg", 2675,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 41");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=1 +x_0=41500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 41\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",123],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",41");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2675\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2676, "epsg", 2676,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 42");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=126 +k=1 +x_0=42500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 42\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",126],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",42");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2676\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2677, "epsg", 2677,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 43");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=129 +k=1 +x_0=43500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 43\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",129],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",43");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2677\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2678, "epsg", 2678,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 44");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=132 +k=1 +x_0=44500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 44\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",132],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",44");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2678\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2679, "epsg", 2679,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 45");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=135 +k=1 +x_0=45500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 45\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",135],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",45");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2679\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2680, "epsg", 2680,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 46");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=138 +k=1 +x_0=46500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 46\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",138],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",46");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2680\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2681, "epsg", 2681,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 47");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=141 +k=1 +x_0=47500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 47\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",141],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",47");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2681\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2682, "epsg", 2682,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 48");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=144 +k=1 +x_0=48500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 48\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",144],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",48");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2682\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2683, "epsg", 2683,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 49");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=147 +k=1 +x_0=49500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 49\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",147],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",49");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2683\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2684, "epsg", 2684,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 50");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=150 +k=1 +x_0=50500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 50\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",150],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2684\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2685, "epsg", 2685,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 51");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=153 +k=1 +x_0=51500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 51\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",153],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",51");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2685\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2686, "epsg", 2686,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 52");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=156 +k=1 +x_0=52500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 52\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",156],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",52");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2686\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2687, "epsg", 2687,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 53");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=159 +k=1 +x_0=53500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 53\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",159],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",53");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2687\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2688, "epsg", 2688,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 54");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=162 +k=1 +x_0=54500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 54\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",162],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",54");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2688\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2689, "epsg", 2689,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 55");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=165 +k=1 +x_0=55500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 55\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",165],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",55");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2689\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2690, "epsg", 2690,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 56");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=168 +k=1 +x_0=56500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 56\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",168],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",56");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2690\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2691, "epsg", 2691,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 57");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=171 +k=1 +x_0=57500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 57\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",171],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",57");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2691\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2692, "epsg", 2692,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 58");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=174 +k=1 +x_0=58500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 58\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",174],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",58");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2692\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2693, "epsg", 2693,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 59");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=177 +k=1 +x_0=59500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 59\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",177],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",59");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"2693\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2694, "epsg", 2694,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 60 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=180 +k=1 +x_0=60000000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 60 (d");
    add_srs_wkt (p, 1,
		 "eprecated)\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_199");
    add_srs_wkt (p, 2,
		 "5\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,");
    add_srs_wkt (p, 4,
		 "0.16,-0.12],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 5,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 6,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 8,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 10,
		 "an\",180],PARAMETER[\"scale_factor\",1],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",60000000],PARAMETER[\"false_northing\",0],AU");
    add_srs_wkt (p, 12,
		 "THORITY[\"EPSG\",\"2694\"],AXIS[\"X\",NORTH],AXIS[\"Y\",");
    add_srs_wkt (p, 13, "EAST]]");
    p = add_epsg_def (first, last, 2695, "epsg", 2695,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 61");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-177 +k=1 +x_0=61500000 +y_0");
    add_proj4text (p, 1,
		   "=0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16");
    add_proj4text (p, 2, ",-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 61\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-177],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",6");
    add_srs_wkt (p, 11,
		 "1500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12, "G\",\"2695\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2696, "epsg", 2696,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 62");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-174 +k=1 +x_0=62500000 +y_0");
    add_proj4text (p, 1,
		   "=0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16");
    add_proj4text (p, 2, ",-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 62\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-174],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",6");
    add_srs_wkt (p, 11,
		 "2500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12, "G\",\"2696\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2697, "epsg", 2697,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 63");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-171 +k=1 +x_0=63500000 +y_0");
    add_proj4text (p, 1,
		   "=0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16");
    add_proj4text (p, 2, ",-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 63\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-171],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",6");
    add_srs_wkt (p, 11,
		 "3500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12, "G\",\"2697\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2698, "epsg", 2698,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 64");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-168 +k=1 +x_0=64500000 +y_0");
    add_proj4text (p, 1,
		   "=0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16");
    add_proj4text (p, 2, ",-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 64\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-168],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",6");
    add_srs_wkt (p, 11,
		 "4500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12, "G\",\"2698\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2699, "epsg", 2699,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 21E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 21E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",21],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2699\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2700, "epsg", 2700,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 24E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=24 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 24E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",24],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2700\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2701, "epsg", 2701,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 27E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=27 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 27E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",27],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2701\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2702, "epsg", 2702,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 30E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=30 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 30E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",30],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2702\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2703, "epsg", 2703,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 33E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=33 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 33E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",33],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2703\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2704, "epsg", 2704,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 36E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=36 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 36E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",36],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2704\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2705, "epsg", 2705,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 39E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=39 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 39E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",39],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2705\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2706, "epsg", 2706,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 42E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=42 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 42E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",42],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2706\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2707, "epsg", 2707,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 45E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=45 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 45E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",45],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2707\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2708, "epsg", 2708,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 48E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=48 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 48E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",48],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2708\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2709, "epsg", 2709,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 51E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=51 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 51E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",51],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2709\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2710, "epsg", 2710,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 54E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=54 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 54E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",54],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2710\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2711, "epsg", 2711,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 57E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=57 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 57E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",57],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2711\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2712, "epsg", 2712,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 60E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=60 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 60E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",60],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2712\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2713, "epsg", 2713,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 63E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=63 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 63E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",63],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2713\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2714, "epsg", 2714,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 66E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=66 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 66E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",66],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2714\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2715, "epsg", 2715,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 69E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=69 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 69E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",69],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2715\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2716, "epsg", 2716,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 72E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=72 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 72E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",72],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2716\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2717, "epsg", 2717,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 75E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=75 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 75E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",75],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2717\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2718, "epsg", 2718,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 78E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=78 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 78E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",78],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2718\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2719, "epsg", 2719,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 81E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=81 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 81E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",81],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2719\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2720, "epsg", 2720,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 84E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=84 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 84E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",84],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2720\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2721, "epsg", 2721,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 87E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=87 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 87E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",87],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2721\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2722, "epsg", 2722,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 90E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=90 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 90E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",90],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2722\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2723, "epsg", 2723,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 93E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=93 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 93E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",93],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2723\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2724, "epsg", 2724,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 96E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=96 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 96E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",96],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2724\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2725, "epsg", 2725,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 99E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=99 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 99E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",99],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2725\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2726, "epsg", 2726,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 102E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=102 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 102E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",102],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2726\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2727, "epsg", 2727,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 105E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 105E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",105],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2727\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2728, "epsg", 2728,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 108E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=108 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 108E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",108],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2728\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2729, "epsg", 2729,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 111E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 111E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",111],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2729\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2730, "epsg", 2730,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 114E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=114 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 114E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",114],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2730\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2731, "epsg", 2731,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 117E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 117E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",117],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2731\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2732, "epsg", 2732,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 120E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=120 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 120E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",120],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2732\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2733, "epsg", 2733,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 123E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 123E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",123],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2733\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2734, "epsg", 2734,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 126E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=126 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 126E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",126],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2734\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2735, "epsg", 2735,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 129E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=129 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 129E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",129],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2735\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2736, "epsg", 2736, "Tete / UTM zone 36S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=36 +south +ellps=clrk66 +units=m +no_def");
    add_proj4text (p, 1, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tete / UTM zone 36S\",GEOGCS[\"Tete\",DATUM[\"T");
    add_srs_wkt (p, 1,
		 "ete\",SPHEROID[\"Clarke 1866\",6378206.4,294.97869821390");
    add_srs_wkt (p, 2,
		 "06,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"61");
    add_srs_wkt (p, 3,
		 "27\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4127\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",33],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 9,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_northing\",10000000],AUTHORITY[\"EPSG\",\"2736\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2737, "epsg", 2737, "Tete / UTM zone 37S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=37 +south +ellps=clrk66 +units=m +no_def");
    add_proj4text (p, 1, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tete / UTM zone 37S\",GEOGCS[\"Tete\",DATUM[\"T");
    add_srs_wkt (p, 1,
		 "ete\",SPHEROID[\"Clarke 1866\",6378206.4,294.97869821390");
    add_srs_wkt (p, 2,
		 "06,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"61");
    add_srs_wkt (p, 3,
		 "27\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4127\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",39],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 9,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_northing\",10000000],AUTHORITY[\"EPSG\",\"2737\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2738, "epsg", 2738,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 132E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=132 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 132E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",132],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2738\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2739, "epsg", 2739,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 135E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=135 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 135E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",135],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2739\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2740, "epsg", 2740,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 138E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=138 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 138E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",138],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2740\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2741, "epsg", 2741,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 141E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=141 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 141E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",141],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2741\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2742, "epsg", 2742,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 144E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=144 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 144E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",144],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2742\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2743, "epsg", 2743,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 147E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=147 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 147E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",147],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2743\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2744, "epsg", 2744,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 150E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=150 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 150E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",150],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2744\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2745, "epsg", 2745,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 153E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=153 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 153E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",153],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2745\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2746, "epsg", 2746,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 156E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=156 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 156E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",156],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2746\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2747, "epsg", 2747,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 159E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=159 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 159E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",159],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2747\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2748, "epsg", 2748,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 162E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=162 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 162E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",162],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2748\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2749, "epsg", 2749,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 165E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=165 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 165E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",165],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2749\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2750, "epsg", 2750,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 168E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=168 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 168E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",168],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2750\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2751, "epsg", 2751,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 171E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=171 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 171E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",171],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2751\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2752, "epsg", 2752,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 174E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=174 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 174E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",174],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2752\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2753, "epsg", 2753,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 177E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=177 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 177E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",177],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2753\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2754, "epsg", 2754,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 180E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=180 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 180E\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",180],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2754\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2755, "epsg", 2755,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 177W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-177 +k=1 +x_0=500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 177W\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-177],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2755\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2756, "epsg", 2756,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 174W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-174 +k=1 +x_0=500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 174W\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-174],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2756\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2757, "epsg", 2757,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 171W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-171 +k=1 +x_0=500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 171W\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-171],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2757\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2758, "epsg", 2758,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 168W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-168 +k=1 +x_0=500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 168W\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-168],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2758\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 2759, "epsg", 2759,
		      "NAD83(HARN) / Alabama East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30.5 +lon_0=-85.83333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "9996 +x_0=200000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Alabama East\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",30.5],PARAMETER[\"central_m");
    add_srs_wkt (p, 9,
		 "eridian\",-85.83333333333333],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.99996],PARAMETER[\"false_easting\",200000],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_northing\",0],AUTHORITY[\"EPSG\",\"2759\"],AXIS[");
    add_srs_wkt (p, 12, "\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2760, "epsg", 2760,
		      "NAD83(HARN) / Alabama West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=-87.5 +k=0.999933333 +x_0=6");
    add_proj4text (p, 1, "00000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Alabama West\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",30],PARAMETER[\"central_mer");
    add_srs_wkt (p, 9,
		 "idian\",-87.5],PARAMETER[\"scale_factor\",0.999933333],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",600000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",0],AUTHORITY[\"EPSG\",\"2760\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2761, "epsg", 2761,
		      "NAD83(HARN) / Arizona East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-110.1666666666667 +k=0.999");
    add_proj4text (p, 1, "9 +x_0=213360 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Arizona East\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",31],PARAMETER[\"central_mer");
    add_srs_wkt (p, 9,
		 "idian\",-110.1666666666667],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9999],PARAMETER[\"false_easting\",213360],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"2761\"],AXIS[\"X");
    add_srs_wkt (p, 12, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2762, "epsg", 2762,
		      "NAD83(HARN) / Arizona Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-111.9166666666667 +k=0.999");
    add_proj4text (p, 1, "9 +x_0=213360 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Arizona Central\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",31],PARAMETER[\"central_m");
    add_srs_wkt (p, 9,
		 "eridian\",-111.9166666666667],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9999],PARAMETER[\"false_easting\",213360],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"2762\"],AXIS[\"");
    add_srs_wkt (p, 12, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2763, "epsg", 2763,
		      "NAD83(HARN) / Arizona West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-113.75 +k=0.999933333 +x_0");
    add_proj4text (p, 1, "=213360 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Arizona West\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",31],PARAMETER[\"central_mer");
    add_srs_wkt (p, 9,
		 "idian\",-113.75],PARAMETER[\"scale_factor\",0.999933333]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_easting\",213360],PARAMETER[\"false_n");
    add_srs_wkt (p, 11,
		 "orthing\",0],AUTHORITY[\"EPSG\",\"2763\"],AXIS[\"X\",EAS");
    add_srs_wkt (p, 12, "T],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2764, "epsg", 2764,
		      "NAD83(HARN) / Arkansas North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.23333333333333 +lat_2=34.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=34.33333333333334 +lon_0=-92 +x_0=400000 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Arkansas North\",GEOGCS[\"NAD83(H");
    add_srs_wkt (p, 1,
		 "ARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",36.2333333333333");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"standard_parallel_2\",34.93333333333333],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",34.33333333333334],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-92],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 12,
		 "g\",400000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13, "EPSG\",\"2764\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2765, "epsg", 2765,
		      "NAD83(HARN) / Arkansas South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=34.76666666666667 +lat_2=33.3 +lat_0=32");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-92 +x_0=400000 +y_0=400000 +ellp");
    add_proj4text (p, 2, "s=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Arkansas South\",GEOGCS[\"NAD83(H");
    add_srs_wkt (p, 1,
		 "ARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",34.7666666666666");
    add_srs_wkt (p, 9,
		 "7],PARAMETER[\"standard_parallel_2\",33.3],PARAMETER[\"l");
    add_srs_wkt (p, 10,
		 "atitude_of_origin\",32.66666666666666],PARAMETER[\"centr");
    add_srs_wkt (p, 11,
		 "al_meridian\",-92],PARAMETER[\"false_easting\",400000],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_northing\",400000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 13, "2765\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2766, "epsg", 2766,
		      "NAD83(HARN) / California zone 1");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.66666666666666 +lat_2=40 +lat_0=39.3");
    add_proj4text (p, 1,
		   "3333333333334 +lon_0=-122 +x_0=2000000 +y_0=500000 +ellp");
    add_proj4text (p, 2, "s=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / California zone 1\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",41.6666666666666");
    add_srs_wkt (p, 9,
		 "6],PARAMETER[\"standard_parallel_2\",40],PARAMETER[\"lat");
    add_srs_wkt (p, 10,
		 "itude_of_origin\",39.33333333333334],PARAMETER[\"central");
    add_srs_wkt (p, 11,
		 "_meridian\",-122],PARAMETER[\"false_easting\",2000000],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_northing\",500000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 13, "2766\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2767, "epsg", 2767,
		      "NAD83(HARN) / California zone 2");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.83333333333334 +lat_2=38.33333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=37.66666666666666 +lon_0=-122 +x_0=2000000 +y");
    add_proj4text (p, 2, "_0=500000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / California zone 2\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",39.8333333333333");
    add_srs_wkt (p, 9,
		 "4],PARAMETER[\"standard_parallel_2\",38.33333333333334],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",37.66666666666666],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-122],PARAMETER[\"false_easti");
    add_srs_wkt (p, 12,
		 "ng\",2000000],PARAMETER[\"false_northing\",500000],AUTHO");
    add_srs_wkt (p, 13,
		 "RITY[\"EPSG\",\"2767\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORT");
    add_srs_wkt (p, 14, "H]]");
    p = add_epsg_def (first, last, 2768, "epsg", 2768,
		      "NAD83(HARN) / California zone 3");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.43333333333333 +lat_2=37.06666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.5 +lon_0=-120.5 +x_0=2000000 +y_0=500000 +");
    add_proj4text (p, 2, "ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / California zone 3\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",38.4333333333333");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"standard_parallel_2\",37.06666666666667],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",36.5],PARAMETER[\"centr");
    add_srs_wkt (p, 11,
		 "al_meridian\",-120.5],PARAMETER[\"false_easting\",200000");
    add_srs_wkt (p, 12,
		 "0],PARAMETER[\"false_northing\",500000],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 13, "\",\"2768\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2769, "epsg", 2769,
		      "NAD83(HARN) / California zone 4");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.25 +lat_2=36 +lat_0=35.3333333333333");
    add_proj4text (p, 1,
		   "4 +lon_0=-119 +x_0=2000000 +y_0=500000 +ellps=GRS80 +uni");
    add_proj4text (p, 2, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / California zone 4\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",37.25],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_2\",36],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 10,
		 "gin\",35.33333333333334],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 11,
		 "-119],PARAMETER[\"false_easting\",2000000],PARAMETER[\"f");
    add_srs_wkt (p, 12,
		 "alse_northing\",500000],AUTHORITY[\"EPSG\",\"2769\"],AXI");
    add_srs_wkt (p, 13, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2770, "epsg", 2770,
		      "NAD83(HARN) / California zone 5");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=35.46666666666667 +lat_2=34.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=33.5 +lon_0=-118 +x_0=2000000 +y_0=500000 +el");
    add_proj4text (p, 2, "lps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / California zone 5\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",35.4666666666666");
    add_srs_wkt (p, 9,
		 "7],PARAMETER[\"standard_parallel_2\",34.03333333333333],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",33.5],PARAMETER[\"centr");
    add_srs_wkt (p, 11,
		 "al_meridian\",-118],PARAMETER[\"false_easting\",2000000]");
    add_srs_wkt (p, 12,
		 ",PARAMETER[\"false_northing\",500000],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 13, ",\"2770\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2771, "epsg", 2771,
		      "NAD83(HARN) / California zone 6");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.88333333333333 +lat_2=32.78333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=32.16666666666666 +lon_0=-116.25 +x_0=2000000");
    add_proj4text (p, 2, " +y_0=500000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / California zone 6\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",33.8833333333333");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"standard_parallel_2\",32.78333333333333],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",32.16666666666666],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-116.25],PARAMETER[\"false_ea");
    add_srs_wkt (p, 12,
		 "sting\",2000000],PARAMETER[\"false_northing\",500000],AU");
    add_srs_wkt (p, 13,
		 "THORITY[\"EPSG\",\"2771\"],AXIS[\"X\",EAST],AXIS[\"Y\",N");
    add_srs_wkt (p, 14, "ORTH]]");
    p = add_epsg_def (first, last, 2772, "epsg", 2772,
		      "NAD83(HARN) / Colorado North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.78333333333333 +lat_2=39.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=39.33333333333334 +lon_0=-105.5 +x_0=914401.8");
    add_proj4text (p, 2, "289 +y_0=304800.6096 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Colorado North\",GEOGCS[\"NAD83(H");
    add_srs_wkt (p, 1,
		 "ARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",40.7833333333333");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"standard_parallel_2\",39.71666666666667],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",39.33333333333334],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-105.5],PARAMETER[\"false_eas");
    add_srs_wkt (p, 12,
		 "ting\",914401.8289],PARAMETER[\"false_northing\",304800.");
    add_srs_wkt (p, 13,
		 "6096],AUTHORITY[\"EPSG\",\"2772\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 14, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2773, "epsg", 2773,
		      "NAD83(HARN) / Colorado Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.75 +lat_2=38.45 +lat_0=37.8333333333");
    add_proj4text (p, 1,
		   "3334 +lon_0=-105.5 +x_0=914401.8289 +y_0=304800.6096 +el");
    add_proj4text (p, 2, "lps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Colorado Central\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",39.75],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_2\",38.45],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 10,
		 "origin\",37.83333333333334],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 11,
		 "\",-105.5],PARAMETER[\"false_easting\",914401.8289],PARA");
    add_srs_wkt (p, 12,
		 "METER[\"false_northing\",304800.6096],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 13, ",\"2773\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2774, "epsg", 2774,
		      "NAD83(HARN) / Colorado South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.43333333333333 +lat_2=37.23333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=36.66666666666666 +lon_0=-105.5 +x_0=914401.8");
    add_proj4text (p, 2, "289 +y_0=304800.6096 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Colorado South\",GEOGCS[\"NAD83(H");
    add_srs_wkt (p, 1,
		 "ARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",38.4333333333333");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"standard_parallel_2\",37.23333333333333],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",36.66666666666666],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-105.5],PARAMETER[\"false_eas");
    add_srs_wkt (p, 12,
		 "ting\",914401.8289],PARAMETER[\"false_northing\",304800.");
    add_srs_wkt (p, 13,
		 "6096],AUTHORITY[\"EPSG\",\"2774\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 14, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2775, "epsg", 2775,
		      "NAD83(HARN) / Connecticut");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.86666666666667 +lat_2=41.2 +lat_0=40");
    add_proj4text (p, 1,
		   ".83333333333334 +lon_0=-72.75 +x_0=304800.6096 +y_0=1524");
    add_proj4text (p, 2, "00.3048 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Connecticut\",GEOGCS[\"NAD83(HARN");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",41.86666666666667],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"standard_parallel_2\",41.2],PARAMETER[\"latit");
    add_srs_wkt (p, 10,
		 "ude_of_origin\",40.83333333333334],PARAMETER[\"central_m");
    add_srs_wkt (p, 11,
		 "eridian\",-72.75],PARAMETER[\"false_easting\",304800.609");
    add_srs_wkt (p, 12,
		 "6],PARAMETER[\"false_northing\",152400.3048],AUTHORITY[\"");
    add_srs_wkt (p, 13, "EPSG\",\"2775\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2776, "epsg", 2776,
		      "NAD83(HARN) / Delaware");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38 +lon_0=-75.41666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "995 +x_0=200000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Delaware\",GEOGCS[\"NAD83(HARN)\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",38],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 9,
		 "\",-75.41666666666667],PARAMETER[\"scale_factor\",0.9999");
    add_srs_wkt (p, 10,
		 "95],PARAMETER[\"false_easting\",200000],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"2776\"],AXIS[\"X\",");
    add_srs_wkt (p, 12, "EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2777, "epsg", 2777,
		      "NAD83(HARN) / Florida East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=24.33333333333333 +lon_0=-81 +k=0.999");
    add_proj4text (p, 1,
		   "941177 +x_0=200000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Florida East\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",24.33333333333333],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-81],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".999941177],PARAMETER[\"false_easting\",200000],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2777\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2778, "epsg", 2778,
		      "NAD83(HARN) / Florida West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=24.33333333333333 +lon_0=-82 +k=0.999");
    add_proj4text (p, 1,
		   "941177 +x_0=200000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Florida West\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",24.33333333333333],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-82],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".999941177],PARAMETER[\"false_easting\",200000],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2778\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2779, "epsg", 2779,
		      "NAD83(HARN) / Florida North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.75 +lat_2=29.58333333333333 +lat_0=2");
    add_proj4text (p, 1,
		   "9 +lon_0=-84.5 +x_0=600000 +y_0=0 +ellps=GRS80 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Florida North\",GEOGCS[\"NAD83(HA");
    add_srs_wkt (p, 1,
		 "RN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"standard_parallel_1\",30.75],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_2\",29.58333333333333],PARAMETER[\"lat");
    add_srs_wkt (p, 10,
		 "itude_of_origin\",29],PARAMETER[\"central_meridian\",-84");
    add_srs_wkt (p, 11,
		 ".5],PARAMETER[\"false_easting\",600000],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"2779\"],AXIS[\"X\",");
    add_srs_wkt (p, 13, "EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2780, "epsg", 2780,
		      "NAD83(HARN) / Georgia East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=-82.16666666666667 +k=0.999");
    add_proj4text (p, 1, "9 +x_0=200000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Georgia East\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",30],PARAMETER[\"central_mer");
    add_srs_wkt (p, 9,
		 "idian\",-82.16666666666667],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9999],PARAMETER[\"false_easting\",200000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"2780\"],AXIS[\"X");
    add_srs_wkt (p, 12, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2781, "epsg", 2781,
		      "NAD83(HARN) / Georgia West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=-84.16666666666667 +k=0.999");
    add_proj4text (p, 1, "9 +x_0=700000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Georgia West\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",30],PARAMETER[\"central_mer");
    add_srs_wkt (p, 9,
		 "idian\",-84.16666666666667],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9999],PARAMETER[\"false_easting\",700000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"2781\"],AXIS[\"X");
    add_srs_wkt (p, 12, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2782, "epsg", 2782,
		      "NAD83(HARN) / Hawaii zone 1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=18.83333333333333 +lon_0=-155.5 +k=0.");
    add_proj4text (p, 1,
		   "999966667 +x_0=500000 +y_0=0 +ellps=GRS80 +units=m +no_d");
    add_proj4text (p, 2, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Hawaii zone 1\",GEOGCS[\"NAD83(HA");
    add_srs_wkt (p, 1,
		 "RN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",18.83333333333333],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-155.5],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.999966667],PARAMETER[\"false_easting\",500000],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2782\"],A");
    add_srs_wkt (p, 12, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2783, "epsg", 2783,
		      "NAD83(HARN) / Hawaii zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=20.33333333333333 +lon_0=-156.6666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999966667 +x_0=500000 +y_0=0 +ellps=GRS80 +u");
    add_proj4text (p, 2, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Hawaii zone 2\",GEOGCS[\"NAD83(HA");
    add_srs_wkt (p, 1,
		 "RN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",20.33333333333333],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-156.6666666666667],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",0.999966667],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2783\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2784, "epsg", 2784,
		      "NAD83(HARN) / Hawaii zone 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=21.16666666666667 +lon_0=-158 +k=0.99");
    add_proj4text (p, 1,
		   "999 +x_0=500000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Hawaii zone 3\",GEOGCS[\"NAD83(HA");
    add_srs_wkt (p, 1,
		 "RN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",21.16666666666667],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-158],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.99999],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"2784\"],AXIS[\"");
    add_srs_wkt (p, 12, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2785, "epsg", 2785,
		      "NAD83(HARN) / Hawaii zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=21.83333333333333 +lon_0=-159.5 +k=0.");
    add_proj4text (p, 1,
		   "99999 +x_0=500000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Hawaii zone 4\",GEOGCS[\"NAD83(HA");
    add_srs_wkt (p, 1,
		 "RN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",21.83333333333333],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-159.5],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.99999],PARAMETER[\"false_easting\",500000],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_northing\",0],AUTHORITY[\"EPSG\",\"2785\"],AXIS[");
    add_srs_wkt (p, 12, "\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2786, "epsg", 2786,
		      "NAD83(HARN) / Hawaii zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=21.66666666666667 +lon_0=-160.1666666");
    add_proj4text (p, 1,
		   "666667 +k=1 +x_0=500000 +y_0=0 +ellps=GRS80 +units=m +no");
    add_proj4text (p, 2, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Hawaii zone 5\",GEOGCS[\"NAD83(HA");
    add_srs_wkt (p, 1,
		 "RN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",21.66666666666667],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-160.1666666666667],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",1],PARAMETER[\"false_easting\",500000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2786\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2787, "epsg", 2787,
		      "NAD83(HARN) / Idaho East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-112.1666666");
    add_proj4text (p, 1,
		   "666667 +k=0.9999473679999999 +x_0=200000 +y_0=0 +ellps=G");
    add_proj4text (p, 2, "RS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Idaho East\",GEOGCS[\"NAD83(HARN)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",41.66666666666666],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-112.1666666666667],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",0.999947368],PARAMETER[\"false_easting\",200");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"2787\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2788, "epsg", 2788,
		      "NAD83(HARN) / Idaho Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-114 +k=0.99");
    add_proj4text (p, 1,
		   "99473679999999 +x_0=500000 +y_0=0 +ellps=GRS80 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Idaho Central\",GEOGCS[\"NAD83(HA");
    add_srs_wkt (p, 1,
		 "RN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",41.66666666666666],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-114],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.999947368],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2788\"],AX");
    add_srs_wkt (p, 12, "IS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2789, "epsg", 2789,
		      "NAD83(HARN) / Idaho West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-115.75 +k=0");
    add_proj4text (p, 1,
		   ".999933333 +x_0=800000 +y_0=0 +ellps=GRS80 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Idaho West\",GEOGCS[\"NAD83(HARN)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",41.66666666666666],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-115.75],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.999933333],PARAMETER[\"false_easting\",800000],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2789\"],A");
    add_srs_wkt (p, 12, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2790, "epsg", 2790,
		      "NAD83(HARN) / Illinois East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36.66666666666666 +lon_0=-88.33333333");
    add_proj4text (p, 1,
		   "333333 +k=0.9999749999999999 +x_0=300000 +y_0=0 +ellps=G");
    add_proj4text (p, 2, "RS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Illinois East\",GEOGCS[\"NAD83(HA");
    add_srs_wkt (p, 1,
		 "RN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",36.66666666666666],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-88.33333333333333],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",0.999975],PARAMETER[\"false_easting\",3000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "2790\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2791, "epsg", 2791,
		      "NAD83(HARN) / Illinois West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36.66666666666666 +lon_0=-90.16666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999941177 +x_0=700000 +y_0=0 +ellps=GRS80 +u");
    add_proj4text (p, 2, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Illinois West\",GEOGCS[\"NAD83(HA");
    add_srs_wkt (p, 1,
		 "RN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",36.66666666666666],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-90.16666666666667],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",0.999941177],PARAMETER[\"false_easting\",7");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2791\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2792, "epsg", 2792,
		      "NAD83(HARN) / Indiana East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=37.5 +lon_0=-85.66666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=100000 +y_0=250000 +ellps=GRS80 +units=m +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Indiana East\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",37.5],PARAMETER[\"central_m");
    add_srs_wkt (p, 9,
		 "eridian\",-85.66666666666667],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.999966667],PARAMETER[\"false_easting\",100000],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",250000],AUTHORITY[\"EPSG\",\"2792");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2793, "epsg", 2793,
		      "NAD83(HARN) / Indiana West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=37.5 +lon_0=-87.08333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=900000 +y_0=250000 +ellps=GRS80 +units=m +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Indiana West\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",37.5],PARAMETER[\"central_m");
    add_srs_wkt (p, 9,
		 "eridian\",-87.08333333333333],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.999966667],PARAMETER[\"false_easting\",900000],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",250000],AUTHORITY[\"EPSG\",\"2793");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2794, "epsg", 2794,
		      "NAD83(HARN) / Iowa North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43.26666666666667 +lat_2=42.06666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=41.5 +lon_0=-93.5 +x_0=1500000 +y_0=1000000 +");
    add_proj4text (p, 2, "ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Iowa North\",GEOGCS[\"NAD83(HARN)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",43.26666666666667],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"standard_parallel_2\",42.06666666666667],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"latitude_of_origin\",41.5],PARAMETER[\"central_m");
    add_srs_wkt (p, 11,
		 "eridian\",-93.5],PARAMETER[\"false_easting\",1500000],PA");
    add_srs_wkt (p, 12,
		 "RAMETER[\"false_northing\",1000000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 13, "2794\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2795, "epsg", 2795,
		      "NAD83(HARN) / Iowa South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.78333333333333 +lat_2=40.61666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=40 +lon_0=-93.5 +x_0=500000 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Iowa South\",GEOGCS[\"NAD83(HARN)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",41.78333333333333],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"standard_parallel_2\",40.61666666666667],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"latitude_of_origin\",40],PARAMETER[\"central_mer");
    add_srs_wkt (p, 11,
		 "idian\",-93.5],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 12,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2795\"],");
    add_srs_wkt (p, 13, "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2796, "epsg", 2796,
		      "NAD83(HARN) / Kansas North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.78333333333333 +lat_2=38.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=38.33333333333334 +lon_0=-98 +x_0=400000 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Kansas North\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"standard_parallel_1\",39.78333333333333]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"standard_parallel_2\",38.71666666666667],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"latitude_of_origin\",38.33333333333334],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"central_meridian\",-98],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",400000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 13, "G\",\"2796\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2797, "epsg", 2797,
		      "NAD83(HARN) / Kansas South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.56666666666667 +lat_2=37.26666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.66666666666666 +lon_0=-98.5 +x_0=400000 +y");
    add_proj4text (p, 2, "_0=400000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Kansas South\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"standard_parallel_1\",38.56666666666667]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"standard_parallel_2\",37.26666666666667],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"latitude_of_origin\",36.66666666666666],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"central_meridian\",-98.5],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 12,
		 "g\",400000],PARAMETER[\"false_northing\",400000],AUTHORI");
    add_srs_wkt (p, 13,
		 "TY[\"EPSG\",\"2797\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]");
    add_srs_wkt (p, 14, "]");
    p = add_epsg_def (first, last, 2798, "epsg", 2798,
		      "NAD83(HARN) / Kentucky North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.96666666666667 +lat_2=38.96666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=37.5 +lon_0=-84.25 +x_0=500000 +y_0=0 +ellps=");
    add_proj4text (p, 2, "GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Kentucky North\",GEOGCS[\"NAD83(H");
    add_srs_wkt (p, 1,
		 "ARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",37.9666666666666");
    add_srs_wkt (p, 9,
		 "7],PARAMETER[\"standard_parallel_2\",38.96666666666667],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",37.5],PARAMETER[\"centr");
    add_srs_wkt (p, 11,
		 "al_meridian\",-84.25],PARAMETER[\"false_easting\",500000");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 13, "798\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_06 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 2799, "epsg", 2799,
		      "NAD83(HARN) / Kentucky South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.93333333333333 +lat_2=36.73333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=36.33333333333334 +lon_0=-85.75 +x_0=500000 +");
    add_proj4text (p, 2, "y_0=500000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Kentucky South\",GEOGCS[\"NAD83(H");
    add_srs_wkt (p, 1,
		 "ARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",37.9333333333333");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"standard_parallel_2\",36.73333333333333],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",36.33333333333334],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-85.75],PARAMETER[\"false_eas");
    add_srs_wkt (p, 12,
		 "ting\",500000],PARAMETER[\"false_northing\",500000],AUTH");
    add_srs_wkt (p, 13,
		 "ORITY[\"EPSG\",\"2799\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 14, "TH]]");
    p = add_epsg_def (first, last, 2800, "epsg", 2800,
		      "NAD83(HARN) / Louisiana North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=32.66666666666666 +lat_2=31.16666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=30.5 +lon_0=-92.5 +x_0=1000000 +y_0=0 +ellps=");
    add_proj4text (p, 2, "GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Louisiana North\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",32.6666666666666");
    add_srs_wkt (p, 9,
		 "6],PARAMETER[\"standard_parallel_2\",31.16666666666667],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",30.5],PARAMETER[\"centr");
    add_srs_wkt (p, 11,
		 "al_meridian\",-92.5],PARAMETER[\"false_easting\",1000000");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 13, "800\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2801, "epsg", 2801,
		      "NAD83(HARN) / Louisiana South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.7 +lat_2=29.3 +lat_0=28.5 +lon_0=-91");
    add_proj4text (p, 1,
		   ".33333333333333 +x_0=1000000 +y_0=0 +ellps=GRS80 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Louisiana South\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",30.7],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"standard_parallel_2\",29.3],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 10,
		 "igin\",28.5],PARAMETER[\"central_meridian\",-91.33333333");
    add_srs_wkt (p, 11,
		 "333333],PARAMETER[\"false_easting\",1000000],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"2801\"],AXIS[\"");
    add_srs_wkt (p, 13, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2802, "epsg", 2802,
		      "NAD83(HARN) / Maine East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=43.66666666666666 +lon_0=-68.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=300000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Maine East\",GEOGCS[\"NAD83(HARN)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",43.66666666666666],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-68.5],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9999],PARAMETER[\"false_easting\",300000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"2802\"],AXIS[\"X");
    add_srs_wkt (p, 12, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2803, "epsg", 2803,
		      "NAD83(HARN) / Maine West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.83333333333334 +lon_0=-70.16666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999966667 +x_0=900000 +y_0=0 +ellps=GRS80 +u");
    add_proj4text (p, 2, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Maine West\",GEOGCS[\"NAD83(HARN)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",42.83333333333334],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-70.16666666666667],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",0.999966667],PARAMETER[\"false_easting\",900");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"2803\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2804, "epsg", 2804,
		      "NAD83(HARN) / Maryland");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.45 +lat_2=38.3 +lat_0=37.66666666666");
    add_proj4text (p, 1,
		   "666 +lon_0=-77 +x_0=400000 +y_0=0 +ellps=GRS80 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Maryland\",GEOGCS[\"NAD83(HARN)\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"standard_parallel_1\",39.45],PARAMETER[\"stan");
    add_srs_wkt (p, 9,
		 "dard_parallel_2\",38.3],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 10,
		 ",37.66666666666666],PARAMETER[\"central_meridian\",-77],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_easting\",400000],PARAMETER[\"false_no");
    add_srs_wkt (p, 12,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"2804\"],AXIS[\"X\",EAST");
    add_srs_wkt (p, 13, "],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2805, "epsg", 2805,
		      "NAD83(HARN) / Massachusetts Mainland");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=42.68333333333333 +lat_2=41.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=41 +lon_0=-71.5 +x_0=200000 +y_0=750000 +ellp");
    add_proj4text (p, 2, "s=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Massachusetts Mainland\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_C");
    add_srs_wkt (p, 8,
		 "onic_2SP\"],PARAMETER[\"standard_parallel_1\",42.6833333");
    add_srs_wkt (p, 9,
		 "3333333],PARAMETER[\"standard_parallel_2\",41.7166666666");
    add_srs_wkt (p, 10,
		 "6667],PARAMETER[\"latitude_of_origin\",41],PARAMETER[\"c");
    add_srs_wkt (p, 11,
		 "entral_meridian\",-71.5],PARAMETER[\"false_easting\",200");
    add_srs_wkt (p, 12,
		 "000],PARAMETER[\"false_northing\",750000],AUTHORITY[\"EP");
    add_srs_wkt (p, 13, "SG\",\"2805\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2806, "epsg", 2806,
		      "NAD83(HARN) / Massachusetts Island");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.48333333333333 +lat_2=41.28333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=41 +lon_0=-70.5 +x_0=500000 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Massachusetts Island\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Networ");
    add_srs_wkt (p, 2,
		 "k\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_C");
    add_srs_wkt (p, 8,
		 "onic_2SP\"],PARAMETER[\"standard_parallel_1\",41.4833333");
    add_srs_wkt (p, 9,
		 "3333333],PARAMETER[\"standard_parallel_2\",41.2833333333");
    add_srs_wkt (p, 10,
		 "3333],PARAMETER[\"latitude_of_origin\",41],PARAMETER[\"c");
    add_srs_wkt (p, 11,
		 "entral_meridian\",-70.5],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 12,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 13, "\"2806\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2807, "epsg", 2807,
		      "NAD83(HARN) / Michigan North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.08333333333334 +lat_2=45.48333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=44.78333333333333 +lon_0=-87 +x_0=8000000 +y_");
    add_proj4text (p, 2, "0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Michigan North\",GEOGCS[\"NAD83(H");
    add_srs_wkt (p, 1,
		 "ARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",47.0833333333333");
    add_srs_wkt (p, 9,
		 "4],PARAMETER[\"standard_parallel_2\",45.48333333333333],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",44.78333333333333],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-87],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 12,
		 "g\",8000000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13, "EPSG\",\"2807\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2808, "epsg", 2808,
		      "NAD83(HARN) / Michigan Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.7 +lat_2=44.18333333333333 +lat_0=43");
    add_proj4text (p, 1,
		   ".31666666666667 +lon_0=-84.36666666666666 +x_0=6000000 +");
    add_proj4text (p, 2, "y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Michigan Central\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",45.7],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"standard_parallel_2\",44.18333333333333],PARAMETER[\"l");
    add_srs_wkt (p, 10,
		 "atitude_of_origin\",43.31666666666667],PARAMETER[\"centr");
    add_srs_wkt (p, 11,
		 "al_meridian\",-84.36666666666666],PARAMETER[\"false_east");
    add_srs_wkt (p, 12,
		 "ing\",6000000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 13,
		 "[\"EPSG\",\"2808\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2809, "epsg", 2809,
		      "NAD83(HARN) / Michigan South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43.66666666666666 +lat_2=42.1 +lat_0=41");
    add_proj4text (p, 1,
		   ".5 +lon_0=-84.36666666666666 +x_0=4000000 +y_0=0 +ellps=");
    add_proj4text (p, 2, "GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Michigan South\",GEOGCS[\"NAD83(H");
    add_srs_wkt (p, 1,
		 "ARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",43.6666666666666");
    add_srs_wkt (p, 9,
		 "6],PARAMETER[\"standard_parallel_2\",42.1],PARAMETER[\"l");
    add_srs_wkt (p, 10,
		 "atitude_of_origin\",41.5],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 11,
		 ",-84.36666666666666],PARAMETER[\"false_easting\",4000000");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 13, "809\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2810, "epsg", 2810,
		      "NAD83(HARN) / Minnesota North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.63333333333333 +lat_2=47.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=46.5 +lon_0=-93.09999999999999 +x_0=800000 +y");
    add_proj4text (p, 2, "_0=100000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Minnesota North\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",48.6333333333333");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"standard_parallel_2\",47.03333333333333],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",46.5],PARAMETER[\"centr");
    add_srs_wkt (p, 11,
		 "al_meridian\",-93.1],PARAMETER[\"false_easting\",800000]");
    add_srs_wkt (p, 12,
		 ",PARAMETER[\"false_northing\",100000],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 13, ",\"2810\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2811, "epsg", 2811,
		      "NAD83(HARN) / Minnesota Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.05 +lat_2=45.61666666666667 +lat_0=4");
    add_proj4text (p, 1,
		   "5 +lon_0=-94.25 +x_0=800000 +y_0=100000 +ellps=GRS80 +un");
    add_proj4text (p, 2, "its=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Minnesota Central\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",47.05],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_2\",45.61666666666667],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "latitude_of_origin\",45],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 11,
		 "-94.25],PARAMETER[\"false_easting\",800000],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",100000],AUTHORITY[\"EPSG\",\"2811\"],AX");
    add_srs_wkt (p, 13, "IS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2812, "epsg", 2812,
		      "NAD83(HARN) / Minnesota South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.21666666666667 +lat_2=43.78333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=43 +lon_0=-94 +x_0=800000 +y_0=100000 +ellps=");
    add_proj4text (p, 2, "GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Minnesota South\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",45.2166666666666");
    add_srs_wkt (p, 9,
		 "7],PARAMETER[\"standard_parallel_2\",43.78333333333333],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",43],PARAMETER[\"central");
    add_srs_wkt (p, 11,
		 "_meridian\",-94],PARAMETER[\"false_easting\",800000],PAR");
    add_srs_wkt (p, 12,
		 "AMETER[\"false_northing\",100000],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 13, "812\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2813, "epsg", 2813,
		      "NAD83(HARN) / Mississippi East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=29.5 +lon_0=-88.83333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "9995 +x_0=300000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Mississippi East\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",29.5],PARAMETER[\"central");
    add_srs_wkt (p, 9,
		 "_meridian\",-88.83333333333333],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",0.99995],PARAMETER[\"false_easting\",300000],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2813\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2814, "epsg", 2814,
		      "NAD83(HARN) / Mississippi West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=29.5 +lon_0=-90.33333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "9995 +x_0=700000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Mississippi West\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",29.5],PARAMETER[\"central");
    add_srs_wkt (p, 9,
		 "_meridian\",-90.33333333333333],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",0.99995],PARAMETER[\"false_easting\",700000],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2814\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2815, "epsg", 2815,
		      "NAD83(HARN) / Missouri East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=35.83333333333334 +lon_0=-90.5 +k=0.9");
    add_proj4text (p, 1,
		   "99933333 +x_0=250000 +y_0=0 +ellps=GRS80 +units=m +no_de");
    add_proj4text (p, 2, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Missouri East\",GEOGCS[\"NAD83(HA");
    add_srs_wkt (p, 1,
		 "RN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",35.83333333333334],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-90.5],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.999933333],PARAMETER[\"false_easting\",250000],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2815\"],A");
    add_srs_wkt (p, 12, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2816, "epsg", 2816,
		      "NAD83(HARN) / Missouri Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=35.83333333333334 +lon_0=-92.5 +k=0.9");
    add_proj4text (p, 1,
		   "99933333 +x_0=500000 +y_0=0 +ellps=GRS80 +units=m +no_de");
    add_proj4text (p, 2, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Missouri Central\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",35.83333333333334],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"central_meridian\",-92.5],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",0.999933333],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2816\"]");
    add_srs_wkt (p, 12, ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2817, "epsg", 2817,
		      "NAD83(HARN) / Missouri West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36.16666666666666 +lon_0=-94.5 +k=0.9");
    add_proj4text (p, 1,
		   "99941177 +x_0=850000 +y_0=0 +ellps=GRS80 +units=m +no_de");
    add_proj4text (p, 2, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Missouri West\",GEOGCS[\"NAD83(HA");
    add_srs_wkt (p, 1,
		 "RN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",36.16666666666666],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-94.5],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.999941177],PARAMETER[\"false_easting\",850000],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2817\"],A");
    add_srs_wkt (p, 12, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2818, "epsg", 2818, "NAD83(HARN) / Montana");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49 +lat_2=45 +lat_0=44.25 +lon_0=-109.5");
    add_proj4text (p, 1, " +x_0=600000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Montana\",GEOGCS[\"NAD83(HARN)\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"standard_parallel_1\",49],PARAMETER[\"standar");
    add_srs_wkt (p, 9,
		 "d_parallel_2\",45],PARAMETER[\"latitude_of_origin\",44.2");
    add_srs_wkt (p, 10,
		 "5],PARAMETER[\"central_meridian\",-109.5],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_easting\",600000],PARAMETER[\"false_northing\",0],AU");
    add_srs_wkt (p, 12,
		 "THORITY[\"EPSG\",\"2818\"],AXIS[\"X\",EAST],AXIS[\"Y\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 2819, "epsg", 2819,
		      "NAD83(HARN) / Nebraska");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43 +lat_2=40 +lat_0=39.83333333333334 +");
    add_proj4text (p, 1,
		   "lon_0=-100 +x_0=500000 +y_0=0 +ellps=GRS80 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Nebraska\",GEOGCS[\"NAD83(HARN)\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"standard_parallel_1\",43],PARAMETER[\"standar");
    add_srs_wkt (p, 9,
		 "d_parallel_2\",40],PARAMETER[\"latitude_of_origin\",39.8");
    add_srs_wkt (p, 10,
		 "3333333333334],PARAMETER[\"central_meridian\",-100],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_easting\",500000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 12,
		 "ng\",0],AUTHORITY[\"EPSG\",\"2819\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 13, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2820, "epsg", 2820,
		      "NAD83(HARN) / Nevada East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-115.5833333333333 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=200000 +y_0=8000000 +ellps=GRS80 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Nevada East\",GEOGCS[\"NAD83(HARN");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",34.75],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-115.5833333333333],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9999],PARAMETER[\"false_easting\",200000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",8000000],AUTHORITY[\"EPSG\",\"2820\"],A");
    add_srs_wkt (p, 12, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2821, "epsg", 2821,
		      "NAD83(HARN) / Nevada Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-116.6666666666667 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=500000 +y_0=6000000 +ellps=GRS80 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Nevada Central\",GEOGCS[\"NAD83(H");
    add_srs_wkt (p, 1,
		 "ARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",34.75],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-116.6666666666667],PARAMETER[\"scale_facto");
    add_srs_wkt (p, 10,
		 "r\",0.9999],PARAMETER[\"false_easting\",500000],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_northing\",6000000],AUTHORITY[\"EPSG\",\"2821\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2822, "epsg", 2822,
		      "NAD83(HARN) / Nevada West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-118.5833333333333 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=800000 +y_0=4000000 +ellps=GRS80 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Nevada West\",GEOGCS[\"NAD83(HARN");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",34.75],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-118.5833333333333],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9999],PARAMETER[\"false_easting\",800000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",4000000],AUTHORITY[\"EPSG\",\"2822\"],A");
    add_srs_wkt (p, 12, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2823, "epsg", 2823,
		      "NAD83(HARN) / New Hampshire");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.5 +lon_0=-71.66666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=300000 +y_0=0 +ellps=GRS80 +units=m +no_de");
    add_proj4text (p, 2, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / New Hampshire\",GEOGCS[\"NAD83(HA");
    add_srs_wkt (p, 1,
		 "RN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",42.5],PARAMETER[\"central_m");
    add_srs_wkt (p, 9,
		 "eridian\",-71.66666666666667],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.999966667],PARAMETER[\"false_easting\",300000],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2823\"],A");
    add_srs_wkt (p, 12, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2824, "epsg", 2824,
		      "NAD83(HARN) / New Jersey");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38.83333333333334 +lon_0=-74.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=150000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / New Jersey\",GEOGCS[\"NAD83(HARN)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",38.83333333333334],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-74.5],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9999],PARAMETER[\"false_easting\",150000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"2824\"],AXIS[\"X");
    add_srs_wkt (p, 12, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2825, "epsg", 2825,
		      "NAD83(HARN) / New Mexico East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-104.3333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "909091 +x_0=165000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / New Mexico East\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",31],PARAMETER[\"central_m");
    add_srs_wkt (p, 9,
		 "eridian\",-104.3333333333333],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.999909091],PARAMETER[\"false_easting\",165000],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2825\"],A");
    add_srs_wkt (p, 12, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2826, "epsg", 2826,
		      "NAD83(HARN) / New Mexico Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-106.25 +k=0.9999 +x_0=5000");
    add_proj4text (p, 1, "00 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / New Mexico Central\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",31],PARAMETER[\"central_m");
    add_srs_wkt (p, 9,
		 "eridian\",-106.25],PARAMETER[\"scale_factor\",0.9999],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 11,
		 "hing\",0],AUTHORITY[\"EPSG\",\"2826\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 12, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2827, "epsg", 2827,
		      "NAD83(HARN) / New Mexico West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-107.8333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "916667 +x_0=830000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / New Mexico West\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",31],PARAMETER[\"central_m");
    add_srs_wkt (p, 9,
		 "eridian\",-107.8333333333333],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.999916667],PARAMETER[\"false_easting\",830000],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2827\"],A");
    add_srs_wkt (p, 12, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2828, "epsg", 2828,
		      "NAD83(HARN) / New York East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38.83333333333334 +lon_0=-74.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=150000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / New York East\",GEOGCS[\"NAD83(HA");
    add_srs_wkt (p, 1,
		 "RN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",38.83333333333334],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-74.5],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9999],PARAMETER[\"false_easting\",150000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"2828\"],AXIS[\"");
    add_srs_wkt (p, 12, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2829, "epsg", 2829,
		      "NAD83(HARN) / New York Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40 +lon_0=-76.58333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "9375 +x_0=250000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / New York Central\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",40],PARAMETER[\"central_m");
    add_srs_wkt (p, 9,
		 "eridian\",-76.58333333333333],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9999375],PARAMETER[\"false_easting\",250000],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2829\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2830, "epsg", 2830,
		      "NAD83(HARN) / New York West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40 +lon_0=-78.58333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "9375 +x_0=350000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / New York West\",GEOGCS[\"NAD83(HA");
    add_srs_wkt (p, 1,
		 "RN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",40],PARAMETER[\"central_mer");
    add_srs_wkt (p, 9,
		 "idian\",-78.58333333333333],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9999375],PARAMETER[\"false_easting\",350000],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_northing\",0],AUTHORITY[\"EPSG\",\"2830\"],AXIS[");
    add_srs_wkt (p, 12, "\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2831, "epsg", 2831,
		      "NAD83(HARN) / New York Long Island");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.03333333333333 +lat_2=40.66666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=40.16666666666666 +lon_0=-74 +x_0=300000 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / New York Long Island\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Networ");
    add_srs_wkt (p, 2,
		 "k\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_C");
    add_srs_wkt (p, 8,
		 "onic_2SP\"],PARAMETER[\"standard_parallel_1\",41.0333333");
    add_srs_wkt (p, 9,
		 "3333333],PARAMETER[\"standard_parallel_2\",40.6666666666");
    add_srs_wkt (p, 10,
		 "6666],PARAMETER[\"latitude_of_origin\",40.16666666666666");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"central_meridian\",-74],PARAMETER[\"false_");
    add_srs_wkt (p, 12,
		 "easting\",300000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 13,
		 "ITY[\"EPSG\",\"2831\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH");
    add_srs_wkt (p, 14, "]]");
    p = add_epsg_def (first, last, 2832, "epsg", 2832,
		      "NAD83(HARN) / North Dakota North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.73333333333333 +lat_2=47.43333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=47 +lon_0=-100.5 +x_0=600000 +y_0=0 +ellps=GR");
    add_proj4text (p, 2, "S80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / North Dakota North\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",48.7333333333333");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"standard_parallel_2\",47.43333333333333],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",47],PARAMETER[\"central");
    add_srs_wkt (p, 11,
		 "_meridian\",-100.5],PARAMETER[\"false_easting\",600000],");
    add_srs_wkt (p, 12,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"283");
    add_srs_wkt (p, 13, "2\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2833, "epsg", 2833,
		      "NAD83(HARN) / North Dakota South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.48333333333333 +lat_2=46.18333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=45.66666666666666 +lon_0=-100.5 +x_0=600000 +");
    add_proj4text (p, 2, "y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / North Dakota South\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",47.4833333333333");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"standard_parallel_2\",46.18333333333333],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",45.66666666666666],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-100.5],PARAMETER[\"false_eas");
    add_srs_wkt (p, 12,
		 "ting\",600000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 13,
		 "[\"EPSG\",\"2833\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2834, "epsg", 2834,
		      "NAD83(HARN) / Ohio North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.7 +lat_2=40.43333333333333 +lat_0=39");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-82.5 +x_0=600000 +y_0=0 +ellps=G");
    add_proj4text (p, 2, "RS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Ohio North\",GEOGCS[\"NAD83(HARN)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",41.7],PARAMETER[\"st");
    add_srs_wkt (p, 9,
		 "andard_parallel_2\",40.43333333333333],PARAMETER[\"latit");
    add_srs_wkt (p, 10,
		 "ude_of_origin\",39.66666666666666],PARAMETER[\"central_m");
    add_srs_wkt (p, 11,
		 "eridian\",-82.5],PARAMETER[\"false_easting\",600000],PAR");
    add_srs_wkt (p, 12,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2834\"");
    add_srs_wkt (p, 13, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2835, "epsg", 2835,
		      "NAD83(HARN) / Ohio South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.03333333333333 +lat_2=38.73333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=38 +lon_0=-82.5 +x_0=600000 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Ohio South\",GEOGCS[\"NAD83(HARN)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",40.03333333333333],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"standard_parallel_2\",38.73333333333333],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"latitude_of_origin\",38],PARAMETER[\"central_mer");
    add_srs_wkt (p, 11,
		 "idian\",-82.5],PARAMETER[\"false_easting\",600000],PARAM");
    add_srs_wkt (p, 12,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2835\"],");
    add_srs_wkt (p, 13, "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2836, "epsg", 2836,
		      "NAD83(HARN) / Oklahoma North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.76666666666667 +lat_2=35.56666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=35 +lon_0=-98 +x_0=600000 +y_0=0 +ellps=GRS80");
    add_proj4text (p, 2, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Oklahoma North\",GEOGCS[\"NAD83(H");
    add_srs_wkt (p, 1,
		 "ARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",36.7666666666666");
    add_srs_wkt (p, 9,
		 "7],PARAMETER[\"standard_parallel_2\",35.56666666666667],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",35],PARAMETER[\"central");
    add_srs_wkt (p, 11,
		 "_meridian\",-98],PARAMETER[\"false_easting\",600000],PAR");
    add_srs_wkt (p, 12,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2836\"");
    add_srs_wkt (p, 13, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2837, "epsg", 2837,
		      "NAD83(HARN) / Oklahoma South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=35.23333333333333 +lat_2=33.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=33.33333333333334 +lon_0=-98 +x_0=600000 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Oklahoma South\",GEOGCS[\"NAD83(H");
    add_srs_wkt (p, 1,
		 "ARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",35.2333333333333");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"standard_parallel_2\",33.93333333333333],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",33.33333333333334],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-98],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 12,
		 "g\",600000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13, "EPSG\",\"2837\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2838, "epsg", 2838,
		      "NAD83(HARN) / Oregon North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46 +lat_2=44.33333333333334 +lat_0=43.6");
    add_proj4text (p, 1,
		   "6666666666666 +lon_0=-120.5 +x_0=2500000 +y_0=0 +ellps=G");
    add_proj4text (p, 2, "RS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Oregon North\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"standard_parallel_1\",46],PARAMETER[\"st");
    add_srs_wkt (p, 9,
		 "andard_parallel_2\",44.33333333333334],PARAMETER[\"latit");
    add_srs_wkt (p, 10,
		 "ude_of_origin\",43.66666666666666],PARAMETER[\"central_m");
    add_srs_wkt (p, 11,
		 "eridian\",-120.5],PARAMETER[\"false_easting\",2500000],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2838");
    add_srs_wkt (p, 13, "\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2839, "epsg", 2839,
		      "NAD83(HARN) / Oregon South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44 +lat_2=42.33333333333334 +lat_0=41.6");
    add_proj4text (p, 1,
		   "6666666666666 +lon_0=-120.5 +x_0=1500000 +y_0=0 +ellps=G");
    add_proj4text (p, 2, "RS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Oregon South\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"standard_parallel_1\",44],PARAMETER[\"st");
    add_srs_wkt (p, 9,
		 "andard_parallel_2\",42.33333333333334],PARAMETER[\"latit");
    add_srs_wkt (p, 10,
		 "ude_of_origin\",41.66666666666666],PARAMETER[\"central_m");
    add_srs_wkt (p, 11,
		 "eridian\",-120.5],PARAMETER[\"false_easting\",1500000],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2839");
    add_srs_wkt (p, 13, "\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2840, "epsg", 2840,
		      "NAD83(HARN) / Rhode Island");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.08333333333334 +lon_0=-71.5 +k=0.9");
    add_proj4text (p, 1,
		   "9999375 +x_0=100000 +y_0=0 +ellps=GRS80 +units=m +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Rhode Island\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",41.08333333333334],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-71.5],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.99999375],PARAMETER[\"false_easting\",100000],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2840\"],AX");
    add_srs_wkt (p, 12, "IS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2841, "epsg", 2841,
		      "NAD83(HARN) / South Dakota North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.68333333333333 +lat_2=44.41666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=43.83333333333334 +lon_0=-100 +x_0=600000 +y_");
    add_proj4text (p, 2, "0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / South Dakota North\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",45.6833333333333");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"standard_parallel_2\",44.41666666666666],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",43.83333333333334],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-100],PARAMETER[\"false_easti");
    add_srs_wkt (p, 12,
		 "ng\",600000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13, "EPSG\",\"2841\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2842, "epsg", 2842,
		      "NAD83(HARN) / South Dakota South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.4 +lat_2=42.83333333333334 +lat_0=42");
    add_proj4text (p, 1,
		   ".33333333333334 +lon_0=-100.3333333333333 +x_0=600000 +y");
    add_proj4text (p, 2, "_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / South Dakota South\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",44.4],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"standard_parallel_2\",42.83333333333334],PARAMETER[\"l");
    add_srs_wkt (p, 10,
		 "atitude_of_origin\",42.33333333333334],PARAMETER[\"centr");
    add_srs_wkt (p, 11,
		 "al_meridian\",-100.3333333333333],PARAMETER[\"false_east");
    add_srs_wkt (p, 12,
		 "ing\",600000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 13,
		 "\"EPSG\",\"2842\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2843, "epsg", 2843,
		      "NAD83(HARN) / Tennessee");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.41666666666666 +lat_2=35.25 +lat_0=3");
    add_proj4text (p, 1,
		   "4.33333333333334 +lon_0=-86 +x_0=600000 +y_0=0 +ellps=GR");
    add_proj4text (p, 2, "S80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Tennessee\",GEOGCS[\"NAD83(HARN)\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"standard_parallel_1\",36.41666666666666],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"standard_parallel_2\",35.25],PARAMETER[\"latitud");
    add_srs_wkt (p, 10,
		 "e_of_origin\",34.33333333333334],PARAMETER[\"central_mer");
    add_srs_wkt (p, 11,
		 "idian\",-86],PARAMETER[\"false_easting\",600000],PARAMET");
    add_srs_wkt (p, 12,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2843\"],AX");
    add_srs_wkt (p, 13, "IS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2844, "epsg", 2844,
		      "NAD83(HARN) / Texas North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.18333333333333 +lat_2=34.65 +lat_0=3");
    add_proj4text (p, 1,
		   "4 +lon_0=-101.5 +x_0=200000 +y_0=1000000 +ellps=GRS80 +u");
    add_proj4text (p, 2, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Texas North\",GEOGCS[\"NAD83(HARN");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",36.18333333333333],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"standard_parallel_2\",34.65],PARAMETER[\"lati");
    add_srs_wkt (p, 10,
		 "tude_of_origin\",34],PARAMETER[\"central_meridian\",-101");
    add_srs_wkt (p, 11,
		 ".5],PARAMETER[\"false_easting\",200000],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_northing\",1000000],AUTHORITY[\"EPSG\",\"2844\"],AXIS[");
    add_srs_wkt (p, 13, "\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2845, "epsg", 2845,
		      "NAD83(HARN) / Texas North Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.96666666666667 +lat_2=32.13333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=31.66666666666667 +lon_0=-98.5 +x_0=600000 +y");
    add_proj4text (p, 2, "_0=2000000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Texas North Central\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Co");
    add_srs_wkt (p, 8,
		 "nic_2SP\"],PARAMETER[\"standard_parallel_1\",33.96666666");
    add_srs_wkt (p, 9,
		 "666667],PARAMETER[\"standard_parallel_2\",32.13333333333");
    add_srs_wkt (p, 10,
		 "333],PARAMETER[\"latitude_of_origin\",31.66666666666667]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"central_meridian\",-98.5],PARAMETER[\"false");
    add_srs_wkt (p, 12,
		 "_easting\",600000],PARAMETER[\"false_northing\",2000000]");
    add_srs_wkt (p, 13,
		 ",AUTHORITY[\"EPSG\",\"2845\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 14, ",NORTH]]");
    p = add_epsg_def (first, last, 2846, "epsg", 2846,
		      "NAD83(HARN) / Texas Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=31.88333333333333 +lat_2=30.11666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=29.66666666666667 +lon_0=-100.3333333333333 +");
    add_proj4text (p, 2,
		   "x_0=700000 +y_0=3000000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Texas Central\",GEOGCS[\"NAD83(HA");
    add_srs_wkt (p, 1,
		 "RN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"standard_parallel_1\",31.88333333333333]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"standard_parallel_2\",30.11666666666667],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"latitude_of_origin\",29.66666666666667],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"central_meridian\",-100.3333333333333],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_easting\",700000],PARAMETER[\"false_northing\",300");
    add_srs_wkt (p, 13,
		 "0000],AUTHORITY[\"EPSG\",\"2846\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 14, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2847, "epsg", 2847,
		      "NAD83(HARN) / Texas South Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.28333333333333 +lat_2=28.38333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=27.83333333333333 +lon_0=-99 +x_0=600000 +y_0");
    add_proj4text (p, 2, "=4000000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Texas South Central\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Co");
    add_srs_wkt (p, 8,
		 "nic_2SP\"],PARAMETER[\"standard_parallel_1\",30.28333333");
    add_srs_wkt (p, 9,
		 "333333],PARAMETER[\"standard_parallel_2\",28.38333333333");
    add_srs_wkt (p, 10,
		 "333],PARAMETER[\"latitude_of_origin\",27.83333333333333]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"central_meridian\",-99],PARAMETER[\"false_e");
    add_srs_wkt (p, 12,
		 "asting\",600000],PARAMETER[\"false_northing\",4000000],A");
    add_srs_wkt (p, 13,
		 "UTHORITY[\"EPSG\",\"2847\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 14, "NORTH]]");
    p = add_epsg_def (first, last, 2848, "epsg", 2848,
		      "NAD83(HARN) / Texas South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=27.83333333333333 +lat_2=26.16666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=25.66666666666667 +lon_0=-98.5 +x_0=300000 +y");
    add_proj4text (p, 2, "_0=5000000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Texas South\",GEOGCS[\"NAD83(HARN");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",27.83333333333333],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"standard_parallel_2\",26.16666666666667],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"latitude_of_origin\",25.66666666666667],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"central_meridian\",-98.5],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",300000],PARAMETER[\"false_northing\",5000000],AUTHORITY");
    add_srs_wkt (p, 13,
		 "[\"EPSG\",\"2848\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2849, "epsg", 2849,
		      "NAD83(HARN) / Utah North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.78333333333333 +lat_2=40.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=40.33333333333334 +lon_0=-111.5 +x_0=500000 +");
    add_proj4text (p, 2, "y_0=1000000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Utah North\",GEOGCS[\"NAD83(HARN)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",41.78333333333333],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"standard_parallel_2\",40.71666666666667],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"latitude_of_origin\",40.33333333333334],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"central_meridian\",-111.5],PARAMETER[\"false_easting");
    add_srs_wkt (p, 12,
		 "\",500000],PARAMETER[\"false_northing\",1000000],AUTHORI");
    add_srs_wkt (p, 13,
		 "TY[\"EPSG\",\"2849\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]");
    add_srs_wkt (p, 14, "]");
    p = add_epsg_def (first, last, 2850, "epsg", 2850,
		      "NAD83(HARN) / Utah Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.65 +lat_2=39.01666666666667 +lat_0=3");
    add_proj4text (p, 1,
		   "8.33333333333334 +lon_0=-111.5 +x_0=500000 +y_0=2000000 ");
    add_proj4text (p, 2, "+ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Utah Central\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"standard_parallel_1\",40.65],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_2\",39.01666666666667],PARAMETER[\"lat");
    add_srs_wkt (p, 10,
		 "itude_of_origin\",38.33333333333334],PARAMETER[\"central");
    add_srs_wkt (p, 11,
		 "_meridian\",-111.5],PARAMETER[\"false_easting\",500000],");
    add_srs_wkt (p, 12,
		 "PARAMETER[\"false_northing\",2000000],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 13, ",\"2850\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2851, "epsg", 2851,
		      "NAD83(HARN) / Utah South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.35 +lat_2=37.21666666666667 +lat_0=3");
    add_proj4text (p, 1,
		   "6.66666666666666 +lon_0=-111.5 +x_0=500000 +y_0=3000000 ");
    add_proj4text (p, 2, "+ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Utah South\",GEOGCS[\"NAD83(HARN)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",38.35],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_2\",37.21666666666667],PARAMETER[\"lati");
    add_srs_wkt (p, 10,
		 "tude_of_origin\",36.66666666666666],PARAMETER[\"central_");
    add_srs_wkt (p, 11,
		 "meridian\",-111.5],PARAMETER[\"false_easting\",500000],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_northing\",3000000],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 13, "\"2851\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2852, "epsg", 2852, "NAD83(HARN) / Vermont");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.5 +lon_0=-72.5 +k=0.999964286 +x_0");
    add_proj4text (p, 1, "=500000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Vermont\",GEOGCS[\"NAD83(HARN)\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",42.5],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-72.5],PARAMETER[\"scale_factor\",0.999964286],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_easting\",500000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 11,
		 "ng\",0],AUTHORITY[\"EPSG\",\"2852\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 12, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2853, "epsg", 2853,
		      "NAD83(HARN) / Virginia North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.2 +lat_2=38.03333333333333 +lat_0=37");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-78.5 +x_0=3500000 +y_0=2000000 +");
    add_proj4text (p, 2, "ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Virginia North\",GEOGCS[\"NAD83(H");
    add_srs_wkt (p, 1,
		 "ARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",39.2],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"standard_parallel_2\",38.03333333333333],PARAMETER[\"l");
    add_srs_wkt (p, 10,
		 "atitude_of_origin\",37.66666666666666],PARAMETER[\"centr");
    add_srs_wkt (p, 11,
		 "al_meridian\",-78.5],PARAMETER[\"false_easting\",3500000");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_northing\",2000000],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 13, "\",\"2853\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2854, "epsg", 2854,
		      "NAD83(HARN) / Virginia South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.96666666666667 +lat_2=36.76666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.33333333333334 +lon_0=-78.5 +x_0=3500000 +");
    add_proj4text (p, 2, "y_0=1000000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Virginia South\",GEOGCS[\"NAD83(H");
    add_srs_wkt (p, 1,
		 "ARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",37.9666666666666");
    add_srs_wkt (p, 9,
		 "7],PARAMETER[\"standard_parallel_2\",36.76666666666667],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",36.33333333333334],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-78.5],PARAMETER[\"false_east");
    add_srs_wkt (p, 12,
		 "ing\",3500000],PARAMETER[\"false_northing\",1000000],AUT");
    add_srs_wkt (p, 13,
		 "HORITY[\"EPSG\",\"2854\"],AXIS[\"X\",EAST],AXIS[\"Y\",NO");
    add_srs_wkt (p, 14, "RTH]]");
    p = add_epsg_def (first, last, 2855, "epsg", 2855,
		      "NAD83(HARN) / Washington North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.73333333333333 +lat_2=47.5 +lat_0=47");
    add_proj4text (p, 1,
		   " +lon_0=-120.8333333333333 +x_0=500000 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Washington North\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",48.7333333333333");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"standard_parallel_2\",47.5],PARAMETER[\"l");
    add_srs_wkt (p, 10,
		 "atitude_of_origin\",47],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 11,
		 "120.8333333333333],PARAMETER[\"false_easting\",500000],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2855");
    add_srs_wkt (p, 13, "\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2856, "epsg", 2856,
		      "NAD83(HARN) / Washington South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.33333333333334 +lat_2=45.83333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=45.33333333333334 +lon_0=-120.5 +x_0=500000 +");
    add_proj4text (p, 2, "y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Washington South\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",47.3333333333333");
    add_srs_wkt (p, 9,
		 "4],PARAMETER[\"standard_parallel_2\",45.83333333333334],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",45.33333333333334],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-120.5],PARAMETER[\"false_eas");
    add_srs_wkt (p, 12,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 13,
		 "[\"EPSG\",\"2856\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2857, "epsg", 2857,
		      "NAD83(HARN) / West Virginia North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.25 +lat_2=39 +lat_0=38.5 +lon_0=-79.");
    add_proj4text (p, 1, "5 +x_0=600000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / West Virginia North\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Co");
    add_srs_wkt (p, 8,
		 "nic_2SP\"],PARAMETER[\"standard_parallel_1\",40.25],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"standard_parallel_2\",39],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 10,
		 "f_origin\",38.5],PARAMETER[\"central_meridian\",-79.5],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_easting\",600000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 12,
		 "thing\",0],AUTHORITY[\"EPSG\",\"2857\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 13, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2858, "epsg", 2858,
		      "NAD83(HARN) / West Virginia South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.88333333333333 +lat_2=37.48333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=37 +lon_0=-81 +x_0=600000 +y_0=0 +ellps=GRS80");
    add_proj4text (p, 2, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / West Virginia South\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Co");
    add_srs_wkt (p, 8,
		 "nic_2SP\"],PARAMETER[\"standard_parallel_1\",38.88333333");
    add_srs_wkt (p, 9,
		 "333333],PARAMETER[\"standard_parallel_2\",37.48333333333");
    add_srs_wkt (p, 10,
		 "333],PARAMETER[\"latitude_of_origin\",37],PARAMETER[\"ce");
    add_srs_wkt (p, 11,
		 "ntral_meridian\",-81],PARAMETER[\"false_easting\",600000");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 13, "858\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2859, "epsg", 2859,
		      "NAD83(HARN) / Wisconsin North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46.76666666666667 +lat_2=45.56666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=45.16666666666666 +lon_0=-90 +x_0=600000 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Wisconsin North\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",46.7666666666666");
    add_srs_wkt (p, 9,
		 "7],PARAMETER[\"standard_parallel_2\",45.56666666666667],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",45.16666666666666],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-90],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 12,
		 "g\",600000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13, "EPSG\",\"2859\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2860, "epsg", 2860,
		      "NAD83(HARN) / Wisconsin Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.5 +lat_2=44.25 +lat_0=43.83333333333");
    add_proj4text (p, 1,
		   "334 +lon_0=-90 +x_0=600000 +y_0=0 +ellps=GRS80 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Wisconsin Central\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",45.5],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"standard_parallel_2\",44.25],PARAMETER[\"latitude_of_o");
    add_srs_wkt (p, 10,
		 "rigin\",43.83333333333334],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 11,
		 ",-90],PARAMETER[\"false_easting\",600000],PARAMETER[\"fa");
    add_srs_wkt (p, 12,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"2860\"],AXIS[\"X\"");
    add_srs_wkt (p, 13, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2861, "epsg", 2861,
		      "NAD83(HARN) / Wisconsin South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.06666666666667 +lat_2=42.73333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=42 +lon_0=-90 +x_0=600000 +y_0=0 +ellps=GRS80");
    add_proj4text (p, 2, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Wisconsin South\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",44.0666666666666");
    add_srs_wkt (p, 9,
		 "7],PARAMETER[\"standard_parallel_2\",42.73333333333333],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",42],PARAMETER[\"central");
    add_srs_wkt (p, 11,
		 "_meridian\",-90],PARAMETER[\"false_easting\",600000],PAR");
    add_srs_wkt (p, 12,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2861\"");
    add_srs_wkt (p, 13, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2862, "epsg", 2862,
		      "NAD83(HARN) / Wyoming East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-105.1666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "999375 +x_0=200000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Wyoming East\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",40.5],PARAMETER[\"central_m");
    add_srs_wkt (p, 9,
		 "eridian\",-105.1666666666667],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9999375],PARAMETER[\"false_easting\",200000],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2862\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2863, "epsg", 2863,
		      "NAD83(HARN) / Wyoming East Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-107.3333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "999375 +x_0=400000 +y_0=100000 +ellps=GRS80 +units=m +no");
    add_proj4text (p, 2, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Wyoming East Central\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Networ");
    add_srs_wkt (p, 2,
		 "k\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",40.5],PARAMETER[\"c");
    add_srs_wkt (p, 9,
		 "entral_meridian\",-107.3333333333333],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.9999375],PARAMETER[\"false_easting\",400000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",100000],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"2863\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2864, "epsg", 2864,
		      "NAD83(HARN) / Wyoming West Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-108.75 +k=0.9999375 +x_0");
    add_proj4text (p, 1, "=600000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Wyoming West Central\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Networ");
    add_srs_wkt (p, 2,
		 "k\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",40.5],PARAMETER[\"c");
    add_srs_wkt (p, 9,
		 "entral_meridian\",-108.75],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 10,
		 "9999375],PARAMETER[\"false_easting\",600000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"2864\"],AXIS[\"");
    add_srs_wkt (p, 12, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2865, "epsg", 2865,
		      "NAD83(HARN) / Wyoming West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-110.0833333333333 +k=0.9");
    add_proj4text (p, 1,
		   "999375 +x_0=800000 +y_0=100000 +ellps=GRS80 +units=m +no");
    add_proj4text (p, 2, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Wyoming West\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",40.5],PARAMETER[\"central_m");
    add_srs_wkt (p, 9,
		 "eridian\",-110.0833333333333],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9999375],PARAMETER[\"false_easting\",800000],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_northing\",100000],AUTHORITY[\"EPSG\",\"2865\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2866, "epsg", 2866,
		      "NAD83(HARN) / Puerto Rico and Virgin Is.");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=18.43333333333333 +lat_2=18.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=17.83333333333333 +lon_0=-66.43333333333334 +");
    add_proj4text (p, 2,
		   "x_0=200000 +y_0=200000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Puerto Rico and Virgin Is.\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_");
    add_srs_wkt (p, 2,
		 "Network\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Confo");
    add_srs_wkt (p, 8,
		 "rmal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",18.4");
    add_srs_wkt (p, 9,
		 "3333333333333],PARAMETER[\"standard_parallel_2\",18.0333");
    add_srs_wkt (p, 10,
		 "3333333333],PARAMETER[\"latitude_of_origin\",17.83333333");
    add_srs_wkt (p, 11,
		 "333333],PARAMETER[\"central_meridian\",-66.4333333333333");
    add_srs_wkt (p, 12,
		 "4],PARAMETER[\"false_easting\",200000],PARAMETER[\"false");
    add_srs_wkt (p, 13,
		 "_northing\",200000],AUTHORITY[\"EPSG\",\"2866\"],AXIS[\"");
    add_srs_wkt (p, 14, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2867, "epsg", 2867,
		      "NAD83(HARN) / Arizona East (ft)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-110.1666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=213360 +y_0=0 +ellps=GRS80 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Arizona East (ft)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"foot\",0.3048,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9002\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",31],PARAMETER[\"centr");
    add_srs_wkt (p, 9,
		 "al_meridian\",-110.1666666666667],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 10,
		 "or\",0.9999],PARAMETER[\"false_easting\",700000],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2867\"],AX");
    add_srs_wkt (p, 12, "IS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2868, "epsg", 2868,
		      "NAD83(HARN) / Arizona Central (ft)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-111.9166666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=213360 +y_0=0 +ellps=GRS80 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Arizona Central (ft)\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Networ");
    add_srs_wkt (p, 2,
		 "k\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"foot\",0.3048,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9002\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",31],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-111.9166666666667],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.9999],PARAMETER[\"false_easting\",700000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2868\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2869, "epsg", 2869,
		      "NAD83(HARN) / Arizona West (ft)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-113.75 +k=0.999933333 +x_0");
    add_proj4text (p, 1, "=213360 +y_0=0 +ellps=GRS80 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Arizona West (ft)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"foot\",0.3048,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9002\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",31],PARAMETER[\"centr");
    add_srs_wkt (p, 9,
		 "al_meridian\",-113.75],PARAMETER[\"scale_factor\",0.9999");
    add_srs_wkt (p, 10,
		 "33333],PARAMETER[\"false_easting\",700000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"2869\"],AXIS[\"X");
    add_srs_wkt (p, 12, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2870, "epsg", 2870,
		      "NAD83(HARN) / California zone 1 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.66666666666666 +lat_2=40 +lat_0=39.3");
    add_proj4text (p, 1,
		   "3333333333334 +lon_0=-122 +x_0=2000000.0001016 +y_0=5000");
    add_proj4text (p, 2, "00.0001016001 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / California zone 1 (ftUS)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Ne");
    add_srs_wkt (p, 2,
		 "twork\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey fo");
    add_srs_wkt (p, 7,
		 "ot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 9,
		 "andard_parallel_1\",41.66666666666666],PARAMETER[\"stand");
    add_srs_wkt (p, 10,
		 "ard_parallel_2\",40],PARAMETER[\"latitude_of_origin\",39");
    add_srs_wkt (p, 11,
		 ".33333333333334],PARAMETER[\"central_meridian\",-122],PA");
    add_srs_wkt (p, 12,
		 "RAMETER[\"false_easting\",6561666.667],PARAMETER[\"false");
    add_srs_wkt (p, 13,
		 "_northing\",1640416.667],AUTHORITY[\"EPSG\",\"2870\"],AX");
    add_srs_wkt (p, 14, "IS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2871, "epsg", 2871,
		      "NAD83(HARN) / California zone 2 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.83333333333334 +lat_2=38.33333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=37.66666666666666 +lon_0=-122 +x_0=2000000.00");
    add_proj4text (p, 2,
		   "01016 +y_0=500000.0001016001 +ellps=GRS80 +units=us-ft +");
    add_proj4text (p, 3, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / California zone 2 (ftUS)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Ne");
    add_srs_wkt (p, 2,
		 "twork\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey fo");
    add_srs_wkt (p, 7,
		 "ot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 9,
		 "andard_parallel_1\",39.83333333333334],PARAMETER[\"stand");
    add_srs_wkt (p, 10,
		 "ard_parallel_2\",38.33333333333334],PARAMETER[\"latitude");
    add_srs_wkt (p, 11,
		 "_of_origin\",37.66666666666666],PARAMETER[\"central_meri");
    add_srs_wkt (p, 12,
		 "dian\",-122],PARAMETER[\"false_easting\",6561666.667],PA");
    add_srs_wkt (p, 13,
		 "RAMETER[\"false_northing\",1640416.667],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 14, "\",\"2871\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2872, "epsg", 2872,
		      "NAD83(HARN) / California zone 3 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.43333333333333 +lat_2=37.06666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.5 +lon_0=-120.5 +x_0=2000000.0001016 +y_0=");
    add_proj4text (p, 2,
		   "500000.0001016001 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / California zone 3 (ftUS)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Ne");
    add_srs_wkt (p, 2,
		 "twork\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey fo");
    add_srs_wkt (p, 7,
		 "ot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 9,
		 "andard_parallel_1\",38.43333333333333],PARAMETER[\"stand");
    add_srs_wkt (p, 10,
		 "ard_parallel_2\",37.06666666666667],PARAMETER[\"latitude");
    add_srs_wkt (p, 11,
		 "_of_origin\",36.5],PARAMETER[\"central_meridian\",-120.5");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_easting\",6561666.667],PARAMETER[\"f");
    add_srs_wkt (p, 13,
		 "alse_northing\",1640416.667],AUTHORITY[\"EPSG\",\"2872\"");
    add_srs_wkt (p, 14, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2873, "epsg", 2873,
		      "NAD83(HARN) / California zone 4 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.25 +lat_2=36 +lat_0=35.3333333333333");
    add_proj4text (p, 1,
		   "4 +lon_0=-119 +x_0=2000000.0001016 +y_0=500000.000101600");
    add_proj4text (p, 2, "1 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / California zone 4 (ftUS)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Ne");
    add_srs_wkt (p, 2,
		 "twork\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey fo");
    add_srs_wkt (p, 7,
		 "ot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 9,
		 "andard_parallel_1\",37.25],PARAMETER[\"standard_parallel");
    add_srs_wkt (p, 10,
		 "_2\",36],PARAMETER[\"latitude_of_origin\",35.33333333333");
    add_srs_wkt (p, 11,
		 "334],PARAMETER[\"central_meridian\",-119],PARAMETER[\"fa");
    add_srs_wkt (p, 12,
		 "lse_easting\",6561666.667],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 13,
		 "1640416.667],AUTHORITY[\"EPSG\",\"2873\"],AXIS[\"X\",EAS");
    add_srs_wkt (p, 14, "T],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2874, "epsg", 2874,
		      "NAD83(HARN) / California zone 5 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=35.46666666666667 +lat_2=34.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=33.5 +lon_0=-118 +x_0=2000000.0001016 +y_0=50");
    add_proj4text (p, 2, "0000.0001016001 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / California zone 5 (ftUS)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Ne");
    add_srs_wkt (p, 2,
		 "twork\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey fo");
    add_srs_wkt (p, 7,
		 "ot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 9,
		 "andard_parallel_1\",35.46666666666667],PARAMETER[\"stand");
    add_srs_wkt (p, 10,
		 "ard_parallel_2\",34.03333333333333],PARAMETER[\"latitude");
    add_srs_wkt (p, 11,
		 "_of_origin\",33.5],PARAMETER[\"central_meridian\",-118],");
    add_srs_wkt (p, 12,
		 "PARAMETER[\"false_easting\",6561666.667],PARAMETER[\"fal");
    add_srs_wkt (p, 13,
		 "se_northing\",1640416.667],AUTHORITY[\"EPSG\",\"2874\"],");
    add_srs_wkt (p, 14, "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2875, "epsg", 2875,
		      "NAD83(HARN) / California zone 6 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.88333333333333 +lat_2=32.78333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=32.16666666666666 +lon_0=-116.25 +x_0=2000000");
    add_proj4text (p, 2,
		   ".0001016 +y_0=500000.0001016001 +ellps=GRS80 +units=us-f");
    add_proj4text (p, 3, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / California zone 6 (ftUS)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Ne");
    add_srs_wkt (p, 2,
		 "twork\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey fo");
    add_srs_wkt (p, 7,
		 "ot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 9,
		 "andard_parallel_1\",33.88333333333333],PARAMETER[\"stand");
    add_srs_wkt (p, 10,
		 "ard_parallel_2\",32.78333333333333],PARAMETER[\"latitude");
    add_srs_wkt (p, 11,
		 "_of_origin\",32.16666666666666],PARAMETER[\"central_meri");
    add_srs_wkt (p, 12,
		 "dian\",-116.25],PARAMETER[\"false_easting\",6561666.667]");
    add_srs_wkt (p, 13,
		 ",PARAMETER[\"false_northing\",1640416.667],AUTHORITY[\"E");
    add_srs_wkt (p, 14, "PSG\",\"2875\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2876, "epsg", 2876,
		      "NAD83(HARN) / Colorado North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.78333333333333 +lat_2=39.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=39.33333333333334 +lon_0=-105.5 +x_0=914401.8");
    add_proj4text (p, 2,
		   "288036576 +y_0=304800.6096012192 +ellps=GRS80 +units=us-");
    add_proj4text (p, 3, "ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Colorado North (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",40.78333333333333],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",39.71666666666667],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",39.33333333333334],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 12,
		 "\",-105.5],PARAMETER[\"false_easting\",3000000],PARAMETE");
    add_srs_wkt (p, 13,
		 "R[\"false_northing\",1000000],AUTHORITY[\"EPSG\",\"2876\"");
    add_srs_wkt (p, 14, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2877, "epsg", 2877,
		      "NAD83(HARN) / Colorado Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.75 +lat_2=38.45 +lat_0=37.8333333333");
    add_proj4text (p, 1,
		   "3334 +lon_0=-105.5 +x_0=914401.8288036576 +y_0=304800.60");
    add_proj4text (p, 2, "96012192 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Colorado Central (ftUS)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Net");
    add_srs_wkt (p, 2,
		 "work\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 6,
		 "2\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foo");
    add_srs_wkt (p, 7,
		 "t\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"sta");
    add_srs_wkt (p, 9,
		 "ndard_parallel_1\",39.75],PARAMETER[\"standard_parallel_");
    add_srs_wkt (p, 10,
		 "2\",38.45],PARAMETER[\"latitude_of_origin\",37.833333333");
    add_srs_wkt (p, 11,
		 "33334],PARAMETER[\"central_meridian\",-105.5],PARAMETER[");
    add_srs_wkt (p, 12,
		 "\"false_easting\",3000000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 13,
		 "1000000],AUTHORITY[\"EPSG\",\"2877\"],AXIS[\"X\",EAST],A");
    add_srs_wkt (p, 14, "XIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2878, "epsg", 2878,
		      "NAD83(HARN) / Colorado South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.43333333333333 +lat_2=37.23333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=36.66666666666666 +lon_0=-105.5 +x_0=914401.8");
    add_proj4text (p, 2,
		   "288036576 +y_0=304800.6096012192 +ellps=GRS80 +units=us-");
    add_proj4text (p, 3, "ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Colorado South (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",38.43333333333333],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",37.23333333333333],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",36.66666666666666],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 12,
		 "\",-105.5],PARAMETER[\"false_easting\",3000000],PARAMETE");
    add_srs_wkt (p, 13,
		 "R[\"false_northing\",1000000],AUTHORITY[\"EPSG\",\"2878\"");
    add_srs_wkt (p, 14, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2879, "epsg", 2879,
		      "NAD83(HARN) / Connecticut (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.86666666666667 +lat_2=41.2 +lat_0=40");
    add_proj4text (p, 1,
		   ".83333333333334 +lon_0=-72.75 +x_0=304800.6096012192 +y_");
    add_proj4text (p, 2,
		   "0=152400.3048006096 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Connecticut (ftUS)\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",0.30");
    add_srs_wkt (p, 7,
		 "48006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 9,
		 "rallel_1\",41.86666666666667],PARAMETER[\"standard_paral");
    add_srs_wkt (p, 10,
		 "lel_2\",41.2],PARAMETER[\"latitude_of_origin\",40.833333");
    add_srs_wkt (p, 11,
		 "33333334],PARAMETER[\"central_meridian\",-72.75],PARAMET");
    add_srs_wkt (p, 12,
		 "ER[\"false_easting\",1000000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 13,
		 "\",500000],AUTHORITY[\"EPSG\",\"2879\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 14, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2880, "epsg", 2880,
		      "NAD83(HARN) / Delaware (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38 +lon_0=-75.41666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "995 +x_0=200000.0001016002 +y_0=0 +ellps=GRS80 +units=us");
    add_proj4text (p, 2, "-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Delaware (ftUS)\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",0.30");
    add_srs_wkt (p, 7,
		 "48006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 9,
		 ",38],PARAMETER[\"central_meridian\",-75.41666666666667],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"scale_factor\",0.999995],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",656166.667],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 12,
		 "HORITY[\"EPSG\",\"2880\"],AXIS[\"X\",EAST],AXIS[\"Y\",NO");
    add_srs_wkt (p, 13, "RTH]]");
    p = add_epsg_def (first, last, 2881, "epsg", 2881,
		      "NAD83(HARN) / Florida East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=24.33333333333333 +lon_0=-81 +k=0.999");
    add_proj4text (p, 1,
		   "941177 +x_0=200000.0001016002 +y_0=0 +ellps=GRS80 +units");
    add_proj4text (p, 2, "=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Florida East (ftUS)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",");
    add_srs_wkt (p, 7,
		 "0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECT");
    add_srs_wkt (p, 8,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 9,
		 "gin\",24.33333333333333],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 10,
		 "-81],PARAMETER[\"scale_factor\",0.999941177],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",656166.667],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"2881\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 2882, "epsg", 2882,
		      "NAD83(HARN) / Florida West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=24.33333333333333 +lon_0=-82 +k=0.999");
    add_proj4text (p, 1,
		   "941177 +x_0=200000.0001016002 +y_0=0 +ellps=GRS80 +units");
    add_proj4text (p, 2, "=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Florida West (ftUS)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",");
    add_srs_wkt (p, 7,
		 "0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECT");
    add_srs_wkt (p, 8,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 9,
		 "gin\",24.33333333333333],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 10,
		 "-82],PARAMETER[\"scale_factor\",0.999941177],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",656166.667],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"2882\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 2883, "epsg", 2883,
		      "NAD83(HARN) / Florida North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.75 +lat_2=29.58333333333333 +lat_0=2");
    add_proj4text (p, 1,
		   "9 +lon_0=-84.5 +x_0=600000 +y_0=0 +ellps=GRS80 +units=us");
    add_proj4text (p, 2, "-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Florida North (ftUS)\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Networ");
    add_srs_wkt (p, 2,
		 "k\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",30.75],PARAMETER[\"standard_parallel_2\"");
    add_srs_wkt (p, 10,
		 ",29.58333333333333],PARAMETER[\"latitude_of_origin\",29]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"central_meridian\",-84.5],PARAMETER[\"false");
    add_srs_wkt (p, 12,
		 "_easting\",1968500],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 13,
		 "ORITY[\"EPSG\",\"2883\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 14, "TH]]");
    p = add_epsg_def (first, last, 2884, "epsg", 2884,
		      "NAD83(HARN) / Georgia East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=-82.16666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=200000.0001016002 +y_0=0 +ellps=GRS80 +units=us-f");
    add_proj4text (p, 2, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Georgia East (ftUS)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",");
    add_srs_wkt (p, 7,
		 "0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECT");
    add_srs_wkt (p, 8,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 9,
		 "gin\",30],PARAMETER[\"central_meridian\",-82.16666666666");
    add_srs_wkt (p, 10,
		 "667],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",656166.667],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"2884\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 2885, "epsg", 2885,
		      "NAD83(HARN) / Georgia West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=-84.16666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=699999.9998983998 +y_0=0 +ellps=GRS80 +units=us-f");
    add_proj4text (p, 2, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Georgia West (ftUS)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",");
    add_srs_wkt (p, 7,
		 "0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECT");
    add_srs_wkt (p, 8,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 9,
		 "gin\",30],PARAMETER[\"central_meridian\",-84.16666666666");
    add_srs_wkt (p, 10,
		 "667],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",2296583.333],PARAMETER[\"false_northing\",0]");
    add_srs_wkt (p, 12,
		 ",AUTHORITY[\"EPSG\",\"2885\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 2886, "epsg", 2886,
		      "NAD83(HARN) / Idaho East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-112.1666666");
    add_proj4text (p, 1,
		   "666667 +k=0.9999473679999999 +x_0=200000.0001016002 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Idaho East (ftUS)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",0.30");
    add_srs_wkt (p, 7,
		 "48006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 9,
		 ",41.66666666666666],PARAMETER[\"central_meridian\",-112.");
    add_srs_wkt (p, 10,
		 "1666666666667],PARAMETER[\"scale_factor\",0.999947368],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_easting\",656166.667],PARAMETER[\"false");
    add_srs_wkt (p, 12,
		 "_northing\",0],AUTHORITY[\"EPSG\",\"2886\"],AXIS[\"X\",E");
    add_srs_wkt (p, 13, "AST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2887, "epsg", 2887,
		      "NAD83(HARN) / Idaho Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-114 +k=0.99");
    add_proj4text (p, 1,
		   "99473679999999 +x_0=500000.0001016001 +y_0=0 +ellps=GRS8");
    add_proj4text (p, 2, "0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Idaho Central (ftUS)\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Networ");
    add_srs_wkt (p, 2,
		 "k\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 9,
		 "igin\",41.66666666666666],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 10,
		 ",-114],PARAMETER[\"scale_factor\",0.999947368],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_easting\",1640416.667],PARAMETER[\"false_northi");
    add_srs_wkt (p, 12,
		 "ng\",0],AUTHORITY[\"EPSG\",\"2887\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 13, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2888, "epsg", 2888,
		      "NAD83(HARN) / Idaho West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-115.75 +k=0");
    add_proj4text (p, 1,
		   ".999933333 +x_0=800000.0001016001 +y_0=0 +ellps=GRS80 +u");
    add_proj4text (p, 2, "nits=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Idaho West (ftUS)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",0.30");
    add_srs_wkt (p, 7,
		 "48006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 9,
		 ",41.66666666666666],PARAMETER[\"central_meridian\",-115.");
    add_srs_wkt (p, 10,
		 "75],PARAMETER[\"scale_factor\",0.999933333],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",2624666.667],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"2888\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 2889, "epsg", 2889,
		      "NAD83(HARN) / Indiana East (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=37.5 +lon_0=-85.66666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=99999.99989839978 +y_0=249364.9987299975 +");
    add_proj4text (p, 2, "ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Indiana East (ftUS) (deprecated)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regi");
    add_srs_wkt (p, 2,
		 "onal_Network\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 3,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"61");
    add_srs_wkt (p, 4,
		 "52\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US surv");
    add_srs_wkt (p, 7,
		 "ey foot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",37.5],PARAMETER[\"central_meridian\",-85");
    add_srs_wkt (p, 10,
		 ".66666666666667],PARAMETER[\"scale_factor\",0.999966667]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_easting\",328083.333],PARAMETER[\"fal");
    add_srs_wkt (p, 12,
		 "se_northing\",818125],AUTHORITY[\"EPSG\",\"2889\"],AXIS[");
    add_srs_wkt (p, 13, "\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2890, "epsg", 2890,
		      "NAD83(HARN) / Indiana West (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=37.5 +lon_0=-87.08333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=900000 +y_0=249364.9987299975 +ellps=GRS80");
    add_proj4text (p, 2, " +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Indiana West (ftUS) (deprecated)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regi");
    add_srs_wkt (p, 2,
		 "onal_Network\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 3,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"61");
    add_srs_wkt (p, 4,
		 "52\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US surv");
    add_srs_wkt (p, 7,
		 "ey foot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",37.5],PARAMETER[\"central_meridian\",-87");
    add_srs_wkt (p, 10,
		 ".08333333333333],PARAMETER[\"scale_factor\",0.999966667]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_easting\",2952750],PARAMETER[\"false_");
    add_srs_wkt (p, 12,
		 "northing\",818125],AUTHORITY[\"EPSG\",\"2890\"],AXIS[\"X");
    add_srs_wkt (p, 13, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2891, "epsg", 2891,
		      "NAD83(HARN) / Kentucky North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.96666666666667 +lat_2=38.96666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=37.5 +lon_0=-84.25 +x_0=500000.0001016001 +y_");
    add_proj4text (p, 2, "0=0 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Kentucky North (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",37.96666666666667],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",38.96666666666667],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",37.5],PARAMETER[\"central_meridian\",-84.25],PA");
    add_srs_wkt (p, 12,
		 "RAMETER[\"false_easting\",1640416.667],PARAMETER[\"false");
    add_srs_wkt (p, 13,
		 "_northing\",0],AUTHORITY[\"EPSG\",\"2891\"],AXIS[\"X\",E");
    add_srs_wkt (p, 14, "AST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2892, "epsg", 2892,
		      "NAD83(HARN) / Kentucky South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.93333333333333 +lat_2=36.73333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=36.33333333333334 +lon_0=-85.75 +x_0=500000.0");
    add_proj4text (p, 2,
		   "001016001 +y_0=500000.0001016001 +ellps=GRS80 +units=us-");
    add_proj4text (p, 3, "ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Kentucky South (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",37.93333333333333],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",36.73333333333333],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",36.33333333333334],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 12,
		 "\",-85.75],PARAMETER[\"false_easting\",1640416.667],PARA");
    add_srs_wkt (p, 13,
		 "METER[\"false_northing\",1640416.667],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 14, ",\"2892\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2893, "epsg", 2893,
		      "NAD83(HARN) / Maryland (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.45 +lat_2=38.3 +lat_0=37.66666666666");
    add_proj4text (p, 1,
		   "666 +lon_0=-77 +x_0=399999.9998983998 +y_0=0 +ellps=GRS8");
    add_proj4text (p, 2, "0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Maryland (ftUS)\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",0.30");
    add_srs_wkt (p, 7,
		 "48006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 9,
		 "rallel_1\",39.45],PARAMETER[\"standard_parallel_2\",38.3");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"latitude_of_origin\",37.66666666666666],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"central_meridian\",-77],PARAMETER[\"false_east");
    add_srs_wkt (p, 12,
		 "ing\",1312333.333],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 13,
		 "RITY[\"EPSG\",\"2893\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORT");
    add_srs_wkt (p, 14, "H]]");
    p = add_epsg_def (first, last, 2894, "epsg", 2894,
		      "NAD83(HARN) / Massachusetts Mainland (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=42.68333333333333 +lat_2=41.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=41 +lon_0=-71.5 +x_0=200000.0001016002 +y_0=7");
    add_proj4text (p, 2, "50000 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Massachusetts Mainland (ftUS)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Region");
    add_srs_wkt (p, 2,
		 "al_Network\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US surv");
    add_srs_wkt (p, 7,
		 "ey foot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",42.68333333333333],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "standard_parallel_2\",41.71666666666667],PARAMETER[\"lat");
    add_srs_wkt (p, 11,
		 "itude_of_origin\",41],PARAMETER[\"central_meridian\",-71");
    add_srs_wkt (p, 12,
		 ".5],PARAMETER[\"false_easting\",656166.667],PARAMETER[\"");
    add_srs_wkt (p, 13,
		 "false_northing\",2460625],AUTHORITY[\"EPSG\",\"2894\"],A");
    add_srs_wkt (p, 14, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2895, "epsg", 2895,
		      "NAD83(HARN) / Massachusetts Island (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.48333333333333 +lat_2=41.28333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=41 +lon_0=-70.5 +x_0=500000.0001016001 +y_0=0");
    add_proj4text (p, 2, " +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Massachusetts Island (ftUS)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional");
    add_srs_wkt (p, 2,
		 "_Network\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey ");
    add_srs_wkt (p, 7,
		 "foot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",41.48333333333333],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",41.28333333333333],PARAMETER[\"latitu");
    add_srs_wkt (p, 11,
		 "de_of_origin\",41],PARAMETER[\"central_meridian\",-70.5]");
    add_srs_wkt (p, 12,
		 ",PARAMETER[\"false_easting\",1640416.667],PARAMETER[\"fa");
    add_srs_wkt (p, 13,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"2895\"],AXIS[\"X\"");
    add_srs_wkt (p, 14, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2896, "epsg", 2896,
		      "NAD83(HARN) / Michigan North (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.08333333333334 +lat_2=45.48333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=44.78333333333333 +lon_0=-87 +x_0=7999999.999");
    add_proj4text (p, 2, "968001 +y_0=0 +ellps=GRS80 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Michigan North (ft)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"foot\",0.3048,AUT");
    add_srs_wkt (p, 7,
		 "HORITY[\"EPSG\",\"9002\"]],PROJECTION[\"Lambert_Conforma");
    add_srs_wkt (p, 8,
		 "l_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",47.0833");
    add_srs_wkt (p, 9,
		 "3333333334],PARAMETER[\"standard_parallel_2\",45.4833333");
    add_srs_wkt (p, 10,
		 "3333333],PARAMETER[\"latitude_of_origin\",44.78333333333");
    add_srs_wkt (p, 11,
		 "333],PARAMETER[\"central_meridian\",-87],PARAMETER[\"fal");
    add_srs_wkt (p, 12,
		 "se_easting\",26246719.16],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 13,
		 "],AUTHORITY[\"EPSG\",\"2896\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 14, "\",NORTH]]");
    p = add_epsg_def (first, last, 2897, "epsg", 2897,
		      "NAD83(HARN) / Michigan Central (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.7 +lat_2=44.18333333333333 +lat_0=43");
    add_proj4text (p, 1,
		   ".31666666666667 +lon_0=-84.36666666666666 +x_0=5999999.9");
    add_proj4text (p, 2, "99976001 +y_0=0 +ellps=GRS80 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Michigan Central (ft)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"foot\",0.3048,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9002\"]],PROJECTION[\"Lambert_Conform");
    add_srs_wkt (p, 8,
		 "al_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",45.7],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"standard_parallel_2\",44.18333333333333],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"latitude_of_origin\",43.31666666666667],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"central_meridian\",-84.36666666666666],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_easting\",19685039.37],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 13,
		 ",0],AUTHORITY[\"EPSG\",\"2897\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 14, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 2898, "epsg", 2898,
		      "NAD83(HARN) / Michigan South (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43.66666666666666 +lat_2=42.1 +lat_0=41");
    add_proj4text (p, 1,
		   ".5 +lon_0=-84.36666666666666 +x_0=3999999.999984 +y_0=0 ");
    add_proj4text (p, 2, "+ellps=GRS80 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Michigan South (ft)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"foot\",0.3048,AUT");
    add_srs_wkt (p, 7,
		 "HORITY[\"EPSG\",\"9002\"]],PROJECTION[\"Lambert_Conforma");
    add_srs_wkt (p, 8,
		 "l_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",43.6666");
    add_srs_wkt (p, 9,
		 "6666666666],PARAMETER[\"standard_parallel_2\",42.1],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"latitude_of_origin\",41.5],PARAMETER[\"central_m");
    add_srs_wkt (p, 11,
		 "eridian\",-84.36666666666666],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",13123359.58],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 13,
		 "\"EPSG\",\"2898\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2899, "epsg", 2899,
		      "NAD83(HARN) / Mississippi East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=29.5 +lon_0=-88.83333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "9995 +x_0=300000.0000000001 +y_0=0 +ellps=GRS80 +units=u");
    add_proj4text (p, 2, "s-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Mississippi East (ftUS)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Net");
    add_srs_wkt (p, 2,
		 "work\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 6,
		 "2\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foo");
    add_srs_wkt (p, 7,
		 "t\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",29.5],PARAMETER[\"central_meridian\",-88.83333");
    add_srs_wkt (p, 10,
		 "333333333],PARAMETER[\"scale_factor\",0.99995],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_easting\",984250.0000000002],PARAMETER[\"false_");
    add_srs_wkt (p, 12,
		 "northing\",0],AUTHORITY[\"EPSG\",\"2899\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 13, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2900, "epsg", 2900,
		      "NAD83(HARN) / Mississippi West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=29.5 +lon_0=-90.33333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "9995 +x_0=699999.9998983998 +y_0=0 +ellps=GRS80 +units=u");
    add_proj4text (p, 2, "s-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Mississippi West (ftUS)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Net");
    add_srs_wkt (p, 2,
		 "work\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 6,
		 "2\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foo");
    add_srs_wkt (p, 7,
		 "t\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",29.5],PARAMETER[\"central_meridian\",-90.33333");
    add_srs_wkt (p, 10,
		 "333333333],PARAMETER[\"scale_factor\",0.99995],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_easting\",2296583.333],PARAMETER[\"false_northi");
    add_srs_wkt (p, 12,
		 "ng\",0],AUTHORITY[\"EPSG\",\"2900\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 13, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2901, "epsg", 2901,
		      "NAD83(HARN) / Montana (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49 +lat_2=45 +lat_0=44.25 +lon_0=-109.5");
    add_proj4text (p, 1,
		   " +x_0=599999.9999976 +y_0=0 +ellps=GRS80 +units=ft +no_d");
    add_proj4text (p, 2, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Montana (ft)\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"foot\",0.3048,AUTHORITY[");
    add_srs_wkt (p, 7,
		 "\"EPSG\",\"9002\"]],PROJECTION[\"Lambert_Conformal_Conic");
    add_srs_wkt (p, 8,
		 "_2SP\"],PARAMETER[\"standard_parallel_1\",49],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"standard_parallel_2\",45],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",44.25],PARAMETER[\"central_meridian\",-109.5],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_easting\",1968503.937],PARAMETER[\"false_no");
    add_srs_wkt (p, 12,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"2901\"],AXIS[\"X\",EAST");
    add_srs_wkt (p, 13, "],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2902, "epsg", 2902,
		      "NAD83(HARN) / New Mexico East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-104.3333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "909091 +x_0=165000 +y_0=0 +ellps=GRS80 +units=us-ft +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / New Mexico East (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 9,
		 "igin\",31],PARAMETER[\"central_meridian\",-104.333333333");
    add_srs_wkt (p, 10,
		 "3333],PARAMETER[\"scale_factor\",0.999909091],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_easting\",541337.5],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"2902\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 2903, "epsg", 2903,
		      "NAD83(HARN) / New Mexico Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-106.25 +k=0.9999 +x_0=5000");
    add_proj4text (p, 1,
		   "00.0001016001 +y_0=0 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / New Mexico Central (ftUS)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_N");
    add_srs_wkt (p, 2,
		 "etwork\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]]");
    add_srs_wkt (p, 4,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 5,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 6,
		 "122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey f");
    add_srs_wkt (p, 7,
		 "oot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",31],PARAMETER[\"central_meridian\",-106.25],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_eas");
    add_srs_wkt (p, 11,
		 "ting\",1640416.667],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 12,
		 "ORITY[\"EPSG\",\"2903\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 13, "TH]]");
    p = add_epsg_def (first, last, 2904, "epsg", 2904,
		      "NAD83(HARN) / New Mexico West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-107.8333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "916667 +x_0=830000.0001016001 +y_0=0 +ellps=GRS80 +units");
    add_proj4text (p, 2, "=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / New Mexico West (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 9,
		 "igin\",31],PARAMETER[\"central_meridian\",-107.833333333");
    add_srs_wkt (p, 10,
		 "3333],PARAMETER[\"scale_factor\",0.999916667],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_easting\",2723091.667],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",0],AUTHORITY[\"EPSG\",\"2904\"],AXIS[\"X\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2905, "epsg", 2905,
		      "NAD83(HARN) / New York East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38.83333333333334 +lon_0=-74.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=150000 +y_0=0 +ellps=GRS80 +units=us-ft +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / New York East (ftUS)\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Networ");
    add_srs_wkt (p, 2,
		 "k\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 9,
		 "igin\",38.83333333333334],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 10,
		 ",-74.5],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",492125],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 12,
		 "UTHORITY[\"EPSG\",\"2905\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 13, "NORTH]]");
    p = add_epsg_def (first, last, 2906, "epsg", 2906,
		      "NAD83(HARN) / New York Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40 +lon_0=-76.58333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "9375 +x_0=249999.9998983998 +y_0=0 +ellps=GRS80 +units=u");
    add_proj4text (p, 2, "s-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / New York Central (ftUS)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Net");
    add_srs_wkt (p, 2,
		 "work\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 6,
		 "2\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foo");
    add_srs_wkt (p, 7,
		 "t\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",40],PARAMETER[\"central_meridian\",-76.5833333");
    add_srs_wkt (p, 10,
		 "3333333],PARAMETER[\"scale_factor\",0.9999375],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_easting\",820208.3330000002],PARAMETER[\"false_");
    add_srs_wkt (p, 12,
		 "northing\",0],AUTHORITY[\"EPSG\",\"2906\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 13, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2907, "epsg", 2907,
		      "NAD83(HARN) / New York West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40 +lon_0=-78.58333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "9375 +x_0=350000.0001016001 +y_0=0 +ellps=GRS80 +units=u");
    add_proj4text (p, 2, "s-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / New York West (ftUS)\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Networ");
    add_srs_wkt (p, 2,
		 "k\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 9,
		 "igin\",40],PARAMETER[\"central_meridian\",-78.5833333333");
    add_srs_wkt (p, 10,
		 "3333],PARAMETER[\"scale_factor\",0.9999375],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",1148291.667],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"2907\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 2908, "epsg", 2908,
		      "NAD83(HARN) / New York Long Island (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.03333333333333 +lat_2=40.66666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=40.16666666666666 +lon_0=-74 +x_0=300000.0000");
    add_proj4text (p, 2, "000001 +y_0=0 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / New York Long Island (ftUS)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional");
    add_srs_wkt (p, 2,
		 "_Network\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey ");
    add_srs_wkt (p, 7,
		 "foot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",41.03333333333333],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",40.66666666666666],PARAMETER[\"latitu");
    add_srs_wkt (p, 11,
		 "de_of_origin\",40.16666666666666],PARAMETER[\"central_me");
    add_srs_wkt (p, 12,
		 "ridian\",-74],PARAMETER[\"false_easting\",984250.0000000");
    add_srs_wkt (p, 13,
		 "002],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 14, "\"2908\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2909, "epsg", 2909,
		      "NAD83(HARN) / North Dakota North (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.73333333333333 +lat_2=47.43333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=47 +lon_0=-100.5 +x_0=599999.9999976 +y_0=0 +");
    add_proj4text (p, 2, "ellps=GRS80 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / North Dakota North (ft)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Net");
    add_srs_wkt (p, 2,
		 "work\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 6,
		 "2\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"foot\",0.3048");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9002\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",48.");
    add_srs_wkt (p, 9,
		 "73333333333333],PARAMETER[\"standard_parallel_2\",47.433");
    add_srs_wkt (p, 10,
		 "33333333333],PARAMETER[\"latitude_of_origin\",47],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"central_meridian\",-100.5],PARAMETER[\"false_easti");
    add_srs_wkt (p, 12,
		 "ng\",1968503.937],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 13,
		 "ITY[\"EPSG\",\"2909\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH");
    add_srs_wkt (p, 14, "]]");
    p = add_epsg_def (first, last, 2910, "epsg", 2910,
		      "NAD83(HARN) / North Dakota South (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.48333333333333 +lat_2=46.18333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=45.66666666666666 +lon_0=-100.5 +x_0=599999.9");
    add_proj4text (p, 2, "999976 +y_0=0 +ellps=GRS80 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / North Dakota South (ft)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Net");
    add_srs_wkt (p, 2,
		 "work\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 6,
		 "2\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"foot\",0.3048");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9002\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",47.");
    add_srs_wkt (p, 9,
		 "48333333333333],PARAMETER[\"standard_parallel_2\",46.183");
    add_srs_wkt (p, 10,
		 "33333333333],PARAMETER[\"latitude_of_origin\",45.6666666");
    add_srs_wkt (p, 11,
		 "6666666],PARAMETER[\"central_meridian\",-100.5],PARAMETE");
    add_srs_wkt (p, 12,
		 "R[\"false_easting\",1968503.937],PARAMETER[\"false_north");
    add_srs_wkt (p, 13,
		 "ing\",0],AUTHORITY[\"EPSG\",\"2910\"],AXIS[\"X\",EAST],A");
    add_srs_wkt (p, 14, "XIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2911, "epsg", 2911,
		      "NAD83(HARN) / Oklahoma North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.76666666666667 +lat_2=35.56666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=35 +lon_0=-98 +x_0=600000 +y_0=0 +ellps=GRS80");
    add_proj4text (p, 2, " +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Oklahoma North (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",36.76666666666667],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",35.56666666666667],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",35],PARAMETER[\"central_meridian\",-98],PARAMET");
    add_srs_wkt (p, 12,
		 "ER[\"false_easting\",1968500],PARAMETER[\"false_northing");
    add_srs_wkt (p, 13,
		 "\",0],AUTHORITY[\"EPSG\",\"2911\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 14, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2912, "epsg", 2912,
		      "NAD83(HARN) / Oklahoma South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=35.23333333333333 +lat_2=33.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=33.33333333333334 +lon_0=-98 +x_0=600000 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Oklahoma South (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",35.23333333333333],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",33.93333333333333],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",33.33333333333334],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 12,
		 "\",-98],PARAMETER[\"false_easting\",1968500],PARAMETER[\"");
    add_srs_wkt (p, 13,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"2912\"],AXIS[\"");
    add_srs_wkt (p, 14, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2913, "epsg", 2913,
		      "NAD83(HARN) / Oregon North (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46 +lat_2=44.33333333333334 +lat_0=43.6");
    add_proj4text (p, 1,
		   "6666666666666 +lon_0=-120.5 +x_0=2500000.0001424 +y_0=0 ");
    add_proj4text (p, 2, "+ellps=GRS80 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Oregon North (ft)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"foot\",0.3048,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9002\"]],PROJECTION[\"Lambert_Conformal_Co");
    add_srs_wkt (p, 8,
		 "nic_2SP\"],PARAMETER[\"standard_parallel_1\",46],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"standard_parallel_2\",44.33333333333334],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"latitude_of_origin\",43.66666666666666],PARAMETER[\"ce");
    add_srs_wkt (p, 11,
		 "ntral_meridian\",-120.5],PARAMETER[\"false_easting\",820");
    add_srs_wkt (p, 12,
		 "2099.738],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 13, "SG\",\"2913\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2914, "epsg", 2914,
		      "NAD83(HARN) / Oregon South (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44 +lat_2=42.33333333333334 +lat_0=41.6");
    add_proj4text (p, 1,
		   "6666666666666 +lon_0=-120.5 +x_0=1500000.0001464 +y_0=0 ");
    add_proj4text (p, 2, "+ellps=GRS80 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Oregon South (ft)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"foot\",0.3048,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9002\"]],PROJECTION[\"Lambert_Conformal_Co");
    add_srs_wkt (p, 8,
		 "nic_2SP\"],PARAMETER[\"standard_parallel_1\",44],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"standard_parallel_2\",42.33333333333334],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"latitude_of_origin\",41.66666666666666],PARAMETER[\"ce");
    add_srs_wkt (p, 11,
		 "ntral_meridian\",-120.5],PARAMETER[\"false_easting\",492");
    add_srs_wkt (p, 12,
		 "1259.843],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 13, "SG\",\"2914\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2915, "epsg", 2915,
		      "NAD83(HARN) / Tennessee (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.41666666666666 +lat_2=35.25 +lat_0=3");
    add_proj4text (p, 1,
		   "4.33333333333334 +lon_0=-86 +x_0=600000 +y_0=0 +ellps=GR");
    add_proj4text (p, 2, "S80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Tennessee (ftUS)\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",0.30");
    add_srs_wkt (p, 7,
		 "48006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 9,
		 "rallel_1\",36.41666666666666],PARAMETER[\"standard_paral");
    add_srs_wkt (p, 10,
		 "lel_2\",35.25],PARAMETER[\"latitude_of_origin\",34.33333");
    add_srs_wkt (p, 11,
		 "333333334],PARAMETER[\"central_meridian\",-86],PARAMETER");
    add_srs_wkt (p, 12,
		 "[\"false_easting\",1968500],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 13,
		 ",0],AUTHORITY[\"EPSG\",\"2915\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 14, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 2916, "epsg", 2916,
		      "NAD83(HARN) / Texas North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.18333333333333 +lat_2=34.65 +lat_0=3");
    add_proj4text (p, 1,
		   "4 +lon_0=-101.5 +x_0=200000.0001016002 +y_0=999999.99989");
    add_proj4text (p, 2, "83998 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Texas North (ftUS)\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",0.30");
    add_srs_wkt (p, 7,
		 "48006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 9,
		 "rallel_1\",36.18333333333333],PARAMETER[\"standard_paral");
    add_srs_wkt (p, 10,
		 "lel_2\",34.65],PARAMETER[\"latitude_of_origin\",34],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-101.5],PARAMETER[\"false_eas");
    add_srs_wkt (p, 12,
		 "ting\",656166.667],PARAMETER[\"false_northing\",3280833.");
    add_srs_wkt (p, 13,
		 "333],AUTHORITY[\"EPSG\",\"2916\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 14, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2917, "epsg", 2917,
		      "NAD83(HARN) / Texas North Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.96666666666667 +lat_2=32.13333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=31.66666666666667 +lon_0=-98.5 +x_0=600000 +y");
    add_proj4text (p, 2,
		   "_0=2000000.0001016 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Texas North Central (ftUS)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_");
    add_srs_wkt (p, 2,
		 "Network\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey ");
    add_srs_wkt (p, 7,
		 "foot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",33.96666666666667],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",32.13333333333333],PARAMETER[\"latitu");
    add_srs_wkt (p, 11,
		 "de_of_origin\",31.66666666666667],PARAMETER[\"central_me");
    add_srs_wkt (p, 12,
		 "ridian\",-98.5],PARAMETER[\"false_easting\",1968500],PAR");
    add_srs_wkt (p, 13,
		 "AMETER[\"false_northing\",6561666.667],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 14, ",\"2917\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2918, "epsg", 2918,
		      "NAD83(HARN) / Texas Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=31.88333333333333 +lat_2=30.11666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=29.66666666666667 +lon_0=-100.3333333333333 +");
    add_proj4text (p, 2,
		   "x_0=699999.9998983998 +y_0=3000000 +ellps=GRS80 +units=u");
    add_proj4text (p, 3, "s-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Texas Central (ftUS)\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Networ");
    add_srs_wkt (p, 2,
		 "k\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",31.88333333333333],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",30.11666666666667],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",29.66666666666667],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 12,
		 "\",-100.3333333333333],PARAMETER[\"false_easting\",22965");
    add_srs_wkt (p, 13,
		 "83.333],PARAMETER[\"false_northing\",9842500.000000002],");
    add_srs_wkt (p, 14,
		 "AUTHORITY[\"EPSG\",\"2918\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 15, ",NORTH]]");
    p = add_epsg_def (first, last, 2919, "epsg", 2919,
		      "NAD83(HARN) / Texas South Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.28333333333333 +lat_2=28.38333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=27.83333333333333 +lon_0=-99 +x_0=600000 +y_0");
    add_proj4text (p, 2, "=3999999.9998984 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Texas South Central (ftUS)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_");
    add_srs_wkt (p, 2,
		 "Network\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey ");
    add_srs_wkt (p, 7,
		 "foot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",30.28333333333333],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",28.38333333333333],PARAMETER[\"latitu");
    add_srs_wkt (p, 11,
		 "de_of_origin\",27.83333333333333],PARAMETER[\"central_me");
    add_srs_wkt (p, 12,
		 "ridian\",-99],PARAMETER[\"false_easting\",1968500],PARAM");
    add_srs_wkt (p, 13,
		 "ETER[\"false_northing\",13123333.333],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 14, ",\"2919\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2920, "epsg", 2920,
		      "NAD83(HARN) / Texas South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=27.83333333333333 +lat_2=26.16666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=25.66666666666667 +lon_0=-98.5 +x_0=300000.00");
    add_proj4text (p, 2,
		   "00000001 +y_0=5000000.0001016 +ellps=GRS80 +units=us-ft ");
    add_proj4text (p, 3, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Texas South (ftUS)\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",0.30");
    add_srs_wkt (p, 7,
		 "48006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 9,
		 "rallel_1\",27.83333333333333],PARAMETER[\"standard_paral");
    add_srs_wkt (p, 10,
		 "lel_2\",26.16666666666667],PARAMETER[\"latitude_of_origi");
    add_srs_wkt (p, 11,
		 "n\",25.66666666666667],PARAMETER[\"central_meridian\",-9");
    add_srs_wkt (p, 12,
		 "8.5],PARAMETER[\"false_easting\",984250.0000000002],PARA");
    add_srs_wkt (p, 13,
		 "METER[\"false_northing\",16404166.667],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 14, ",\"2920\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2921, "epsg", 2921,
		      "NAD83(HARN) / Utah North (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.78333333333333 +lat_2=40.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=40.33333333333334 +lon_0=-111.5 +x_0=500000.0");
    add_proj4text (p, 2,
		   "001504 +y_0=999999.9999960001 +ellps=GRS80 +units=ft +no");
    add_proj4text (p, 3, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Utah North (ft)\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"foot\",0.3048,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9002\"]],PROJECTION[\"Lambert_Conformal_Co");
    add_srs_wkt (p, 8,
		 "nic_2SP\"],PARAMETER[\"standard_parallel_1\",41.78333333");
    add_srs_wkt (p, 9,
		 "333333],PARAMETER[\"standard_parallel_2\",40.71666666666");
    add_srs_wkt (p, 10,
		 "667],PARAMETER[\"latitude_of_origin\",40.33333333333334]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"central_meridian\",-111.5],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_easting\",1640419.948],PARAMETER[\"false_northing\",32");
    add_srs_wkt (p, 13,
		 "80839.895],AUTHORITY[\"EPSG\",\"2921\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 14, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2922, "epsg", 2922,
		      "NAD83(HARN) / Utah Central (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.65 +lat_2=39.01666666666667 +lat_0=3");
    add_proj4text (p, 1,
		   "8.33333333333334 +lon_0=-111.5 +x_0=500000.0001504 +y_0=");
    add_proj4text (p, 2, "1999999.999992 +ellps=GRS80 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Utah Central (ft)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"foot\",0.3048,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9002\"]],PROJECTION[\"Lambert_Conformal_Co");
    add_srs_wkt (p, 8,
		 "nic_2SP\"],PARAMETER[\"standard_parallel_1\",40.65],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"standard_parallel_2\",39.01666666666667],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"latitude_of_origin\",38.33333333333334],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "central_meridian\",-111.5],PARAMETER[\"false_easting\",1");
    add_srs_wkt (p, 12,
		 "640419.948],PARAMETER[\"false_northing\",6561679.79],AUT");
    add_srs_wkt (p, 13,
		 "HORITY[\"EPSG\",\"2922\"],AXIS[\"X\",EAST],AXIS[\"Y\",NO");
    add_srs_wkt (p, 14, "RTH]]");
    p = add_epsg_def (first, last, 2923, "epsg", 2923,
		      "NAD83(HARN) / Utah South (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.35 +lat_2=37.21666666666667 +lat_0=3");
    add_proj4text (p, 1,
		   "6.66666666666666 +lon_0=-111.5 +x_0=500000.0001504 +y_0=");
    add_proj4text (p, 2, "2999999.999988 +ellps=GRS80 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Utah South (ft)\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"foot\",0.3048,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9002\"]],PROJECTION[\"Lambert_Conformal_Co");
    add_srs_wkt (p, 8,
		 "nic_2SP\"],PARAMETER[\"standard_parallel_1\",38.35],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"standard_parallel_2\",37.21666666666667],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"latitude_of_origin\",36.66666666666666],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "central_meridian\",-111.5],PARAMETER[\"false_easting\",1");
    add_srs_wkt (p, 12,
		 "640419.948],PARAMETER[\"false_northing\",9842519.685],AU");
    add_srs_wkt (p, 13,
		 "THORITY[\"EPSG\",\"2923\"],AXIS[\"X\",EAST],AXIS[\"Y\",N");
    add_srs_wkt (p, 14, "ORTH]]");
    p = add_epsg_def (first, last, 2924, "epsg", 2924,
		      "NAD83(HARN) / Virginia North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.2 +lat_2=38.03333333333333 +lat_0=37");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-78.5 +x_0=3500000.0001016 +y_0=2");
    add_proj4text (p, 2, "000000.0001016 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Virginia North (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",39.2],PARAMETER[\"standard_parallel_2\",");
    add_srs_wkt (p, 10,
		 "38.03333333333333],PARAMETER[\"latitude_of_origin\",37.6");
    add_srs_wkt (p, 11,
		 "6666666666666],PARAMETER[\"central_meridian\",-78.5],PAR");
    add_srs_wkt (p, 12,
		 "AMETER[\"false_easting\",11482916.667],PARAMETER[\"false");
    add_srs_wkt (p, 13,
		 "_northing\",6561666.667],AUTHORITY[\"EPSG\",\"2924\"],AX");
    add_srs_wkt (p, 14, "IS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2925, "epsg", 2925,
		      "NAD83(HARN) / Virginia South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.96666666666667 +lat_2=36.76666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.33333333333334 +lon_0=-78.5 +x_0=3500000.0");
    add_proj4text (p, 2,
		   "001016 +y_0=999999.9998983998 +ellps=GRS80 +units=us-ft ");
    add_proj4text (p, 3, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Virginia South (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",37.96666666666667],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",36.76666666666667],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",36.33333333333334],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 12,
		 "\",-78.5],PARAMETER[\"false_easting\",11482916.667],PARA");
    add_srs_wkt (p, 13,
		 "METER[\"false_northing\",3280833.333],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 14, ",\"2925\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2926, "epsg", 2926,
		      "NAD83(HARN) / Washington North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.73333333333333 +lat_2=47.5 +lat_0=47");
    add_proj4text (p, 1,
		   " +lon_0=-120.8333333333333 +x_0=500000.0001016001 +y_0=0");
    add_proj4text (p, 2, " +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Washington North (ftUS)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Net");
    add_srs_wkt (p, 2,
		 "work\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 6,
		 "2\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foo");
    add_srs_wkt (p, 7,
		 "t\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"sta");
    add_srs_wkt (p, 9,
		 "ndard_parallel_1\",48.73333333333333],PARAMETER[\"standa");
    add_srs_wkt (p, 10,
		 "rd_parallel_2\",47.5],PARAMETER[\"latitude_of_origin\",4");
    add_srs_wkt (p, 11,
		 "7],PARAMETER[\"central_meridian\",-120.8333333333333],PA");
    add_srs_wkt (p, 12,
		 "RAMETER[\"false_easting\",1640416.667],PARAMETER[\"false");
    add_srs_wkt (p, 13,
		 "_northing\",0],AUTHORITY[\"EPSG\",\"2926\"],AXIS[\"X\",E");
    add_srs_wkt (p, 14, "AST],AXIS[\"Y\",NORTH]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_07 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 2927, "epsg", 2927,
		      "NAD83(HARN) / Washington South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.33333333333334 +lat_2=45.83333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=45.33333333333334 +lon_0=-120.5 +x_0=500000.0");
    add_proj4text (p, 2, "001016001 +y_0=0 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Washington South (ftUS)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Net");
    add_srs_wkt (p, 2,
		 "work\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 6,
		 "2\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foo");
    add_srs_wkt (p, 7,
		 "t\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"sta");
    add_srs_wkt (p, 9,
		 "ndard_parallel_1\",47.33333333333334],PARAMETER[\"standa");
    add_srs_wkt (p, 10,
		 "rd_parallel_2\",45.83333333333334],PARAMETER[\"latitude_");
    add_srs_wkt (p, 11,
		 "of_origin\",45.33333333333334],PARAMETER[\"central_merid");
    add_srs_wkt (p, 12,
		 "ian\",-120.5],PARAMETER[\"false_easting\",1640416.667],P");
    add_srs_wkt (p, 13,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2927");
    add_srs_wkt (p, 14, "\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2928, "epsg", 2928,
		      "NAD83(HARN) / Wisconsin North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46.76666666666667 +lat_2=45.56666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=45.16666666666666 +lon_0=-90 +x_0=600000 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Wisconsin North (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",46.76666666666667],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",45.56666666666667],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",45.16666666666666],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 12,
		 "\",-90],PARAMETER[\"false_easting\",1968500],PARAMETER[\"");
    add_srs_wkt (p, 13,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"2928\"],AXIS[\"");
    add_srs_wkt (p, 14, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2929, "epsg", 2929,
		      "NAD83(HARN) / Wisconsin Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.5 +lat_2=44.25 +lat_0=43.83333333333");
    add_proj4text (p, 1,
		   "334 +lon_0=-90 +x_0=600000 +y_0=0 +ellps=GRS80 +units=us");
    add_proj4text (p, 2, "-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Wisconsin Central (ftUS)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Ne");
    add_srs_wkt (p, 2,
		 "twork\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey fo");
    add_srs_wkt (p, 7,
		 "ot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 9,
		 "andard_parallel_1\",45.5],PARAMETER[\"standard_parallel_");
    add_srs_wkt (p, 10,
		 "2\",44.25],PARAMETER[\"latitude_of_origin\",43.833333333");
    add_srs_wkt (p, 11,
		 "33334],PARAMETER[\"central_meridian\",-90],PARAMETER[\"f");
    add_srs_wkt (p, 12,
		 "alse_easting\",1968500],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 13,
		 "AUTHORITY[\"EPSG\",\"2929\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 14, ",NORTH]]");
    p = add_epsg_def (first, last, 2930, "epsg", 2930,
		      "NAD83(HARN) / Wisconsin South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.06666666666667 +lat_2=42.73333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=42 +lon_0=-90 +x_0=600000 +y_0=0 +ellps=GRS80");
    add_proj4text (p, 2, " +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Wisconsin South (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",44.06666666666667],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",42.73333333333333],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",42],PARAMETER[\"central_meridian\",-90],PARAMET");
    add_srs_wkt (p, 12,
		 "ER[\"false_easting\",1968500],PARAMETER[\"false_northing");
    add_srs_wkt (p, 13,
		 "\",0],AUTHORITY[\"EPSG\",\"2930\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 14, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2931, "epsg", 2931, "Beduaram / TM 13 NE");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=13 +k=0.9996 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=0 +a=6378249.2 +b=6356515 +towgs84=-106,-87,188,0,0,0,");
    add_proj4text (p, 2, "0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beduaram / TM 13 NE\",GEOGCS[\"Beduaram\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Beduaram\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,2");
    add_srs_wkt (p, 2,
		 "93.4660212936265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-");
    add_srs_wkt (p, 3,
		 "106,-87,188,0,0,0,0],AUTHORITY[\"EPSG\",\"6213\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4213\"]],UNIT[\"metre\",1,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",13],PARAMETER[\"scale_factor\",0.9996],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_easting\",500000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 11,
		 "ng\",0],AUTHORITY[\"EPSG\",\"2931\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 12, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2932, "epsg", 2932,
		      "QND95 / Qatar National Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=24.45 +lon_0=51.21666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "9999 +x_0=200000 +y_0=300000 +ellps=intl +towgs84=-119.4");
    add_proj4text (p, 2,
		   "25,-303.659,-11.0006,1.1643,0.174458,1.09626,3.65706 +un");
    add_proj4text (p, 3, "its=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"QND95 / Qatar National Grid\",GEOGCS[\"QND95\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Qatar_National_Datum_1995\",SPHEROID[\"Internati");
    add_srs_wkt (p, 2,
		 "onal 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[-119.425,-303.659,-11.0006,1.1643,0.174458,1.09626");
    add_srs_wkt (p, 4,
		 ",3.65706],AUTHORITY[\"EPSG\",\"6614\"]],PRIMEM[\"Greenwi");
    add_srs_wkt (p, 5,
		 "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0");
    add_srs_wkt (p, 6,
		 "1745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY");
    add_srs_wkt (p, 7,
		 "[\"EPSG\",\"4614\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 8,
		 ",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"latitude_of_origin\",24.45],PARAMETER[\"central_meri");
    add_srs_wkt (p, 10,
		 "dian\",51.21666666666667],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 11,
		 "9999],PARAMETER[\"false_easting\",200000],PARAMETER[\"fa");
    add_srs_wkt (p, 12,
		 "lse_northing\",300000],AUTHORITY[\"EPSG\",\"2932\"],AXIS");
    add_srs_wkt (p, 13, "[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2933, "epsg", 2933, "Segara / UTM zone 50S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=50 +south +ellps=bessel +units=m +no_def");
    add_proj4text (p, 1, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Segara / UTM zone 50S\",GEOGCS[\"Segara\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Gunung_Segara\",SPHEROID[\"Bessel 1841\",6377397.155,");
    add_srs_wkt (p, 2,
		 "299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6613\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4613\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",117],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",0.9996],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11,
		 "\"2933\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH");
    add_srs_wkt (p, 12, "]]");
    p = add_epsg_def (first, last, 2934, "epsg", 2934,
		      "Segara (Jakarta) / NEIEZ (deprecated)");
    add_proj4text (p, 0,
		   "+proj=merc +lon_0=110 +k=0.997 +x_0=3900000 +y_0=900000 ");
    add_proj4text (p, 1, "+ellps=bessel +pm=jakarta +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Segara (Jakarta) / NEIEZ (deprecated)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"Segara (Jakarta)\",DATUM[\"Gunung_Segara_Jakarta\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6820\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Jakarta\",106.8077194444444,AUTHORITY[\"EPSG\",\"8908");
    add_srs_wkt (p, 5,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 6,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4820\"]],UNIT[\"metr");
    add_srs_wkt (p, 7,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Mercato");
    add_srs_wkt (p, 8,
		 "r_1SP\"],PARAMETER[\"central_meridian\",110],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.997],PARAMETER[\"false_easting\",390000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",900000],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 11, "\",\"2934\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2935, "epsg", 2935,
		      "Pulkovo 1942 / CS63 zone A1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0.1166666666666667 +lon_0=41.53333333");
    add_proj4text (p, 1,
		   "333333 +k=1 +x_0=1300000 +y_0=0 +ellps=krass +units=m +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / CS63 zone A1\",GEOGCS[\"Pulkovo ");
    add_srs_wkt (p, 1,
		 "1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassowsky 1940");
    add_srs_wkt (p, 2,
		 "\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[2");
    add_srs_wkt (p, 3,
		 "3.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0.11");
    add_srs_wkt (p, 9,
		 "66666666666667],PARAMETER[\"central_meridian\",41.533333");
    add_srs_wkt (p, 10,
		 "33333333],PARAMETER[\"scale_factor\",1],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",1300000],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 12,
		 "HORITY[\"EPSG\",\"2935\"],AXIS[\"X\",NORTH],AXIS[\"Y\",E");
    add_srs_wkt (p, 13, "AST]]");
    p = add_epsg_def (first, last, 2936, "epsg", 2936,
		      "Pulkovo 1942 / CS63 zone A2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0.1166666666666667 +lon_0=44.53333333");
    add_proj4text (p, 1,
		   "333333 +k=1 +x_0=2300000 +y_0=0 +ellps=krass +units=m +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / CS63 zone A2\",GEOGCS[\"Pulkovo ");
    add_srs_wkt (p, 1,
		 "1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassowsky 1940");
    add_srs_wkt (p, 2,
		 "\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[2");
    add_srs_wkt (p, 3,
		 "3.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0.11");
    add_srs_wkt (p, 9,
		 "66666666666667],PARAMETER[\"central_meridian\",44.533333");
    add_srs_wkt (p, 10,
		 "33333333],PARAMETER[\"scale_factor\",1],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",2300000],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 12,
		 "HORITY[\"EPSG\",\"2936\"],AXIS[\"X\",NORTH],AXIS[\"Y\",E");
    add_srs_wkt (p, 13, "AST]]");
    p = add_epsg_def (first, last, 2937, "epsg", 2937,
		      "Pulkovo 1942 / CS63 zone A3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0.1166666666666667 +lon_0=47.53333333");
    add_proj4text (p, 1,
		   "333333 +k=1 +x_0=3300000 +y_0=0 +ellps=krass +units=m +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / CS63 zone A3\",GEOGCS[\"Pulkovo ");
    add_srs_wkt (p, 1,
		 "1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassowsky 1940");
    add_srs_wkt (p, 2,
		 "\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[2");
    add_srs_wkt (p, 3,
		 "3.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0.11");
    add_srs_wkt (p, 9,
		 "66666666666667],PARAMETER[\"central_meridian\",47.533333");
    add_srs_wkt (p, 10,
		 "33333333],PARAMETER[\"scale_factor\",1],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",3300000],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 12,
		 "HORITY[\"EPSG\",\"2937\"],AXIS[\"X\",NORTH],AXIS[\"Y\",E");
    add_srs_wkt (p, 13, "AST]]");
    p = add_epsg_def (first, last, 2938, "epsg", 2938,
		      "Pulkovo 1942 / CS63 zone A4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0.1166666666666667 +lon_0=50.53333333");
    add_proj4text (p, 1,
		   "333333 +k=1 +x_0=4300000 +y_0=0 +ellps=krass +units=m +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / CS63 zone A4\",GEOGCS[\"Pulkovo ");
    add_srs_wkt (p, 1,
		 "1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassowsky 1940");
    add_srs_wkt (p, 2,
		 "\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[2");
    add_srs_wkt (p, 3,
		 "3.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0.11");
    add_srs_wkt (p, 9,
		 "66666666666667],PARAMETER[\"central_meridian\",50.533333");
    add_srs_wkt (p, 10,
		 "33333333],PARAMETER[\"scale_factor\",1],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",4300000],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 12,
		 "HORITY[\"EPSG\",\"2938\"],AXIS[\"X\",NORTH],AXIS[\"Y\",E");
    add_srs_wkt (p, 13, "AST]]");
    p = add_epsg_def (first, last, 2939, "epsg", 2939,
		      "Pulkovo 1942 / CS63 zone K2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0.1333333333333333 +lon_0=50.76666666");
    add_proj4text (p, 1,
		   "666667 +k=1 +x_0=2300000 +y_0=0 +ellps=krass +units=m +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / CS63 zone K2\",GEOGCS[\"Pulkovo ");
    add_srs_wkt (p, 1,
		 "1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassowsky 1940");
    add_srs_wkt (p, 2,
		 "\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[2");
    add_srs_wkt (p, 3,
		 "3.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0.13");
    add_srs_wkt (p, 9,
		 "33333333333333],PARAMETER[\"central_meridian\",50.766666");
    add_srs_wkt (p, 10,
		 "66666667],PARAMETER[\"scale_factor\",1],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",2300000],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 12,
		 "HORITY[\"EPSG\",\"2939\"],AXIS[\"X\",NORTH],AXIS[\"Y\",E");
    add_srs_wkt (p, 13, "AST]]");
    p = add_epsg_def (first, last, 2940, "epsg", 2940,
		      "Pulkovo 1942 / CS63 zone K3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0.1333333333333333 +lon_0=53.76666666");
    add_proj4text (p, 1,
		   "666667 +k=1 +x_0=3300000 +y_0=0 +ellps=krass +units=m +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / CS63 zone K3\",GEOGCS[\"Pulkovo ");
    add_srs_wkt (p, 1,
		 "1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassowsky 1940");
    add_srs_wkt (p, 2,
		 "\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[2");
    add_srs_wkt (p, 3,
		 "3.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0.13");
    add_srs_wkt (p, 9,
		 "33333333333333],PARAMETER[\"central_meridian\",53.766666");
    add_srs_wkt (p, 10,
		 "66666667],PARAMETER[\"scale_factor\",1],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",3300000],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 12,
		 "HORITY[\"EPSG\",\"2940\"],AXIS[\"X\",NORTH],AXIS[\"Y\",E");
    add_srs_wkt (p, 13, "AST]]");
    p = add_epsg_def (first, last, 2941, "epsg", 2941,
		      "Pulkovo 1942 / CS63 zone K4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0.1333333333333333 +lon_0=56.76666666");
    add_proj4text (p, 1,
		   "666667 +k=1 +x_0=4300000 +y_0=0 +ellps=krass +units=m +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / CS63 zone K4\",GEOGCS[\"Pulkovo ");
    add_srs_wkt (p, 1,
		 "1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassowsky 1940");
    add_srs_wkt (p, 2,
		 "\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[2");
    add_srs_wkt (p, 3,
		 "3.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0.13");
    add_srs_wkt (p, 9,
		 "33333333333333],PARAMETER[\"central_meridian\",56.766666");
    add_srs_wkt (p, 10,
		 "66666667],PARAMETER[\"scale_factor\",1],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",4300000],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 12,
		 "HORITY[\"EPSG\",\"2941\"],AXIS[\"X\",NORTH],AXIS[\"Y\",E");
    add_srs_wkt (p, 13, "AST]]");
    p = add_epsg_def (first, last, 2942, "epsg", 2942,
		      "Porto Santo / UTM zone 28N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=28 +ellps=intl +towgs84=-499,-249,314,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Porto Santo / UTM zone 28N\",GEOGCS[\"Porto San");
    add_srs_wkt (p, 1,
		 "to\",DATUM[\"Porto_Santo_1936\",SPHEROID[\"International");
    add_srs_wkt (p, 2,
		 " 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS8");
    add_srs_wkt (p, 3,
		 "4[-499,-249,314,0,0,0,0],AUTHORITY[\"EPSG\",\"6615\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 6,
		 "2\"]],AUTHORITY[\"EPSG\",\"4615\"]],UNIT[\"metre\",1,AUT");
    add_srs_wkt (p, 7,
		 "HORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merca");
    add_srs_wkt (p, 8,
		 "tor\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"c");
    add_srs_wkt (p, 9,
		 "entral_meridian\",-15],PARAMETER[\"scale_factor\",0.9996");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"false_easting\",500000],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "northing\",0],AUTHORITY[\"EPSG\",\"2942\"],AXIS[\"Eastin");
    add_srs_wkt (p, 12, "g\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2943, "epsg", 2943,
		      "Selvagem Grande / UTM zone 28N");
    add_proj4text (p, 0, "+proj=utm +zone=28 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Selvagem Grande / UTM zone 28N\",GEOGCS[\"Selva");
    add_srs_wkt (p, 1,
		 "gem Grande\",DATUM[\"Selvagem_Grande\",SPHEROID[\"Intern");
    add_srs_wkt (p, 2,
		 "ational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6616\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4616\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-15],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"2943\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 12, "g\",NORTH]]");
    p = add_epsg_def (first, last, 2944, "epsg", 2944,
		      "NAD83(CSRS) / SCoPQ zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-55.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / SCoPQ zone 2\",GEOGCS[\"NAD83(CSR");
    add_srs_wkt (p, 1,
		 "S)\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-55.5],PARAMETER[\"scale_factor\",0.9999],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_easting\",304800],PARAMETER[\"false_northin");
    add_srs_wkt (p, 11,
		 "g\",0],AUTHORITY[\"EPSG\",\"2944\"],AXIS[\"X\",EAST],AXI");
    add_srs_wkt (p, 12, "S[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2945, "epsg", 2945,
		      "NAD83(CSRS) / MTM zone 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-58.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / MTM zone 3\",GEOGCS[\"NAD83(CSRS)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-58.5],PARAMETER[\"scale_factor\",0.9999],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",304800],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"2945\"],AXIS[\"E(X)\",EAST],AXI");
    add_srs_wkt (p, 12, "S[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 2946, "epsg", 2946,
		      "NAD83(CSRS) / MTM zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-61.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / MTM zone 4\",GEOGCS[\"NAD83(CSRS)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-61.5],PARAMETER[\"scale_factor\",0.9999],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",304800],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"2946\"],AXIS[\"E(X)\",EAST],AXI");
    add_srs_wkt (p, 12, "S[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 2947, "epsg", 2947,
		      "NAD83(CSRS) / MTM zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-64.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / MTM zone 5\",GEOGCS[\"NAD83(CSRS)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-64.5],PARAMETER[\"scale_factor\",0.9999],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",304800],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"2947\"],AXIS[\"E(X)\",EAST],AXI");
    add_srs_wkt (p, 12, "S[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 2948, "epsg", 2948,
		      "NAD83(CSRS) / MTM zone 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-67.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / MTM zone 6\",GEOGCS[\"NAD83(CSRS)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-67.5],PARAMETER[\"scale_factor\",0.9999],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",304800],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"2948\"],AXIS[\"E(X)\",EAST],AXI");
    add_srs_wkt (p, 12, "S[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 2949, "epsg", 2949,
		      "NAD83(CSRS) / MTM zone 7");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-70.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / MTM zone 7\",GEOGCS[\"NAD83(CSRS)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-70.5],PARAMETER[\"scale_factor\",0.9999],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",304800],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"2949\"],AXIS[\"E(X)\",EAST],AXI");
    add_srs_wkt (p, 12, "S[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 2950, "epsg", 2950,
		      "NAD83(CSRS) / MTM zone 8");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-73.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / MTM zone 8\",GEOGCS[\"NAD83(CSRS)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-73.5],PARAMETER[\"scale_factor\",0.9999],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",304800],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"2950\"],AXIS[\"E(X)\",EAST],AXI");
    add_srs_wkt (p, 12, "S[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 2951, "epsg", 2951,
		      "NAD83(CSRS) / MTM zone 9");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-76.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / MTM zone 9\",GEOGCS[\"NAD83(CSRS)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-76.5],PARAMETER[\"scale_factor\",0.9999],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",304800],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"2951\"],AXIS[\"E(X)\",EAST],AXI");
    add_srs_wkt (p, 12, "S[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 2952, "epsg", 2952,
		      "NAD83(CSRS) / MTM zone 10");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-79.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / MTM zone 10\",GEOGCS[\"NAD83(CSRS");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-79.5],PARAMETER[\"scale_factor\",0.9999],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_easting\",304800],PARAMETER[\"false_northin");
    add_srs_wkt (p, 11,
		 "g\",0],AUTHORITY[\"EPSG\",\"2952\"],AXIS[\"E(X)\",EAST],");
    add_srs_wkt (p, 12, "AXIS[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 2953, "epsg", 2953,
		      "NAD83(CSRS) / New Brunswick Stereographic");
    add_proj4text (p, 0,
		   "+proj=sterea +lat_0=46.5 +lon_0=-66.5 +k=0.999912 +x_0=2");
    add_proj4text (p, 1, "500000 +y_0=7500000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / New Brunswick Stereographic\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(CSRS)\",DATUM[\"NAD83_Canadian_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"614");
    add_srs_wkt (p, 4,
		 "0\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Oblique_St");
    add_srs_wkt (p, 8,
		 "ereographic\"],PARAMETER[\"latitude_of_origin\",46.5],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"central_meridian\",-66.5],PARAMETER[\"scale_fa");
    add_srs_wkt (p, 10,
		 "ctor\",0.999912],PARAMETER[\"false_easting\",2500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",7500000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12,
		 "2953\"],AXIS[\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 2954, "epsg", 2954,
		      "NAD83(CSRS) / Prince Edward Isl. Stereographic (NAD83)");
    add_proj4text (p, 0,
		   "+proj=sterea +lat_0=47.25 +lon_0=-63 +k=0.999912 +x_0=40");
    add_proj4text (p, 1, "0000 +y_0=800000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / Prince Edward Isl. Stereographic ");
    add_srs_wkt (p, 1,
		 "(NAD83)\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83_Canadian_");
    add_srs_wkt (p, 2,
		 "Spatial_Reference_System\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 3,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4617\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Oblique_Stereographic\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",47.25],PARAMETER[\"central_meridian\",-63],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"scale_factor\",0.999912],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",400000],PARAMETER[\"false_northing\",800000],AUTHORITY[");
    add_srs_wkt (p, 12,
		 "\"EPSG\",\"2954\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NOR");
    add_srs_wkt (p, 13, "TH]]");
    p = add_epsg_def (first, last, 2955, "epsg", 2955,
		      "NAD83(CSRS) / UTM zone 11N");
    add_proj4text (p, 0, "+proj=utm +zone=11 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / UTM zone 11N\",GEOGCS[\"NAD83(CSR");
    add_srs_wkt (p, 1,
		 "S)\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-117],PARAMETER[\"scale_factor\",0.9996],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",500000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 11,
		 "\",0],AUTHORITY[\"EPSG\",\"2955\"],AXIS[\"Easting\",EAST");
    add_srs_wkt (p, 12, "],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2956, "epsg", 2956,
		      "NAD83(CSRS) / UTM zone 12N");
    add_proj4text (p, 0, "+proj=utm +zone=12 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / UTM zone 12N\",GEOGCS[\"NAD83(CSR");
    add_srs_wkt (p, 1,
		 "S)\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-111],PARAMETER[\"scale_factor\",0.9996],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",500000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 11,
		 "\",0],AUTHORITY[\"EPSG\",\"2956\"],AXIS[\"Easting\",EAST");
    add_srs_wkt (p, 12, "],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2957, "epsg", 2957,
		      "NAD83(CSRS) / UTM zone 13N");
    add_proj4text (p, 0, "+proj=utm +zone=13 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / UTM zone 13N\",GEOGCS[\"NAD83(CSR");
    add_srs_wkt (p, 1,
		 "S)\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-105],PARAMETER[\"scale_factor\",0.9996],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",500000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 11,
		 "\",0],AUTHORITY[\"EPSG\",\"2957\"],AXIS[\"Easting\",EAST");
    add_srs_wkt (p, 12, "],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2958, "epsg", 2958,
		      "NAD83(CSRS) / UTM zone 17N");
    add_proj4text (p, 0, "+proj=utm +zone=17 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / UTM zone 17N\",GEOGCS[\"NAD83(CSR");
    add_srs_wkt (p, 1,
		 "S)\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-81],PARAMETER[\"scale_factor\",0.9996],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"2958\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 12, "AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2959, "epsg", 2959,
		      "NAD83(CSRS) / UTM zone 18N");
    add_proj4text (p, 0, "+proj=utm +zone=18 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / UTM zone 18N\",GEOGCS[\"NAD83(CSR");
    add_srs_wkt (p, 1,
		 "S)\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-75],PARAMETER[\"scale_factor\",0.9996],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"2959\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 12, "AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2960, "epsg", 2960,
		      "NAD83(CSRS) / UTM zone 19N");
    add_proj4text (p, 0, "+proj=utm +zone=19 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / UTM zone 19N\",GEOGCS[\"NAD83(CSR");
    add_srs_wkt (p, 1,
		 "S)\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-69],PARAMETER[\"scale_factor\",0.9996],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"2960\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 12, "AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2961, "epsg", 2961,
		      "NAD83(CSRS) / UTM zone 20N");
    add_proj4text (p, 0, "+proj=utm +zone=20 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / UTM zone 20N\",GEOGCS[\"NAD83(CSR");
    add_srs_wkt (p, 1,
		 "S)\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-63],PARAMETER[\"scale_factor\",0.9996],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"2961\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 12, "AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2962, "epsg", 2962,
		      "NAD83(CSRS) / UTM zone 21N");
    add_proj4text (p, 0, "+proj=utm +zone=21 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / UTM zone 21N\",GEOGCS[\"NAD83(CSR");
    add_srs_wkt (p, 1,
		 "S)\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-57],PARAMETER[\"scale_factor\",0.9996],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"2962\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 12, "AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2964, "epsg", 2964, "NAD27 / Alaska Albers");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=55 +lat_2=65 +lat_0=50 +lon_0=-154 +x_0");
    add_proj4text (p, 1,
		   "=0 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_de");
    add_proj4text (p, 2, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Alaska Albers\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Albers_Conic_E");
    add_srs_wkt (p, 8,
		 "qual_Area\"],PARAMETER[\"standard_parallel_1\",55],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"standard_parallel_2\",65],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 10,
		 "_center\",50],PARAMETER[\"longitude_of_center\",-154],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_easting\",0],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"2964\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 2965, "epsg", 2965,
		      "NAD83 / Indiana East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=37.5 +lon_0=-85.66666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=99999.99989839978 +y_0=249999.9998983998 +");
    add_proj4text (p, 2, "ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Indiana East (ftUS)\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercato");
    add_srs_wkt (p, 8,
		 "r\"],PARAMETER[\"latitude_of_origin\",37.5],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-85.66666666666667],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.999966667],PARAMETER[\"false_easting\",32808");
    add_srs_wkt (p, 11,
		 "3.333],PARAMETER[\"false_northing\",820208.3330000002],A");
    add_srs_wkt (p, 12,
		 "UTHORITY[\"EPSG\",\"2965\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 13, "NORTH]]");
    p = add_epsg_def (first, last, 2966, "epsg", 2966,
		      "NAD83 / Indiana West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=37.5 +lon_0=-87.08333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=900000 +y_0=249999.9998983998 +ellps=GRS80");
    add_proj4text (p, 2, " +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Indiana West (ftUS)\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercato");
    add_srs_wkt (p, 8,
		 "r\"],PARAMETER[\"latitude_of_origin\",37.5],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-87.08333333333333],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.999966667],PARAMETER[\"false_easting\",29527");
    add_srs_wkt (p, 11,
		 "50],PARAMETER[\"false_northing\",820208.3330000002],AUTH");
    add_srs_wkt (p, 12,
		 "ORITY[\"EPSG\",\"2966\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 13, "TH]]");
    p = add_epsg_def (first, last, 2967, "epsg", 2967,
		      "NAD83(HARN) / Indiana East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=37.5 +lon_0=-85.66666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=99999.99989839978 +y_0=249999.9998983998 +");
    add_proj4text (p, 2, "ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Indiana East (ftUS)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",");
    add_srs_wkt (p, 7,
		 "0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECT");
    add_srs_wkt (p, 8,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 9,
		 "gin\",37.5],PARAMETER[\"central_meridian\",-85.666666666");
    add_srs_wkt (p, 10,
		 "66667],PARAMETER[\"scale_factor\",0.999966667],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_easting\",328083.333],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",820208.3330000002],AUTHORITY[\"EPSG\",\"2967\"],AXIS");
    add_srs_wkt (p, 13, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2968, "epsg", 2968,
		      "NAD83(HARN) / Indiana West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=37.5 +lon_0=-87.08333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=900000 +y_0=249999.9998983998 +ellps=GRS80");
    add_proj4text (p, 2, " +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Indiana West (ftUS)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",");
    add_srs_wkt (p, 7,
		 "0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECT");
    add_srs_wkt (p, 8,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 9,
		 "gin\",37.5],PARAMETER[\"central_meridian\",-87.083333333");
    add_srs_wkt (p, 10,
		 "33333],PARAMETER[\"scale_factor\",0.999966667],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_easting\",2952750],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",820208.3330000002],AUTHORITY[\"EPSG\",\"2968\"],AXIS[\"");
    add_srs_wkt (p, 13, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2969, "epsg", 2969,
		      "Fort Marigot / UTM zone 20N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +ellps=intl +towgs84=137,248,-430,0,0");
    add_proj4text (p, 1, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Fort Marigot / UTM zone 20N\",GEOGCS[\"Fort Mar");
    add_srs_wkt (p, 1,
		 "igot\",DATUM[\"Fort_Marigot\",SPHEROID[\"International 1");
    add_srs_wkt (p, 2,
		 "924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[");
    add_srs_wkt (p, 3,
		 "137,248,-430,0,0,0,0],AUTHORITY[\"EPSG\",\"6621\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4621\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",-63],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",0],AUTHORITY[\"EPSG\",\"2969\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 12, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2970, "epsg", 2970,
		      "Guadeloupe 1948 / UTM zone 20N");
    add_proj4text (p, 0, "+proj=utm +zone=20 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Guadeloupe 1948 / UTM zone 20N\",GEOGCS[\"Guade");
    add_srs_wkt (p, 1,
		 "loupe 1948\",DATUM[\"Guadeloupe_1948\",SPHEROID[\"Intern");
    add_srs_wkt (p, 2,
		 "ational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6622\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4622\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-63],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"2970\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 12, "g\",NORTH]]");
    p = add_epsg_def (first, last, 2971, "epsg", 2971, "CSG67 / UTM zone 22N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=22 +ellps=intl +towgs84=-186,230,110,0,0");
    add_proj4text (p, 1, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"CSG67 / UTM zone 22N\",GEOGCS[\"CSG67\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Centre_Spatial_Guyanais_1967\",SPHEROID[\"International ");
    add_srs_wkt (p, 2,
		 "1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84");
    add_srs_wkt (p, 3,
		 "[-186,230,110,0,0,0,0],AUTHORITY[\"EPSG\",\"6623\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4623\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",-51],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",0],AUTHORITY[\"EPSG\",\"2971\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 12, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2972, "epsg", 2972, "RGFG95 / UTM zone 22N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=22 +ellps=GRS80 +towgs84=2,2,-2,0,0,0,0 ");
    add_proj4text (p, 1, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGFG95 / UTM zone 22N\",GEOGCS[\"RGFG95\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Reseau_Geodesique_Francais_Guyane_1995\",SPHEROID[\"G");
    add_srs_wkt (p, 2,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "9\"]],TOWGS84[2,2,-2,0,0,0,0],AUTHORITY[\"EPSG\",\"6624\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4624\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-51],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"2972\"],AXIS[\"Eas");
    add_srs_wkt (p, 12, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2973, "epsg", 2973,
		      "Martinique 1938 / UTM zone 20N");
    add_proj4text (p, 0, "+proj=utm +zone=20 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Martinique 1938 / UTM zone 20N\",GEOGCS[\"Marti");
    add_srs_wkt (p, 1,
		 "nique 1938\",DATUM[\"Martinique_1938\",SPHEROID[\"Intern");
    add_srs_wkt (p, 2,
		 "ational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6625\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4625\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-63],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"2973\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 12, "g\",NORTH]]");
    p = add_epsg_def (first, last, 2975, "epsg", 2975, "RGR92 / UTM zone 40S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=40 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGR92 / UTM zone 40S\",GEOGCS[\"RGR92\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Reseau_Geodesique_de_la_Reunion_1992\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 2,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 3,
		 ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6627\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6,
		 "\"]],AUTHORITY[\"EPSG\",\"4627\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"ce");
    add_srs_wkt (p, 9,
		 "ntral_meridian\",57],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 11,
		 "rthing\",10000000],AUTHORITY[\"EPSG\",\"2975\"],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2976, "epsg", 2976,
		      "Tahiti 52 / UTM zone 6S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=6 +south +ellps=intl +towgs84=162,117,15");
    add_proj4text (p, 1, "4,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tahiti 52 / UTM zone 6S\",GEOGCS[\"Tahiti 52\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Tahiti_52\",SPHEROID[\"International 1924\",6378");
    add_srs_wkt (p, 2,
		 "388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[162,117,15");
    add_srs_wkt (p, 3,
		 "4,0,0,0,0],AUTHORITY[\"EPSG\",\"6628\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"4628\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETE");
    add_srs_wkt (p, 8,
		 "R[\"latitude_of_origin\",0],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 9,
		 "\",-147],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",500000],PARAMETER[\"false_northing\",100");
    add_srs_wkt (p, 11,
		 "00000],AUTHORITY[\"EPSG\",\"2976\"],AXIS[\"Easting\",EAS");
    add_srs_wkt (p, 12, "T],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2977, "epsg", 2977,
		      "Tahaa 54 / UTM zone 5S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=5 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tahaa 54 / UTM zone 5S\",GEOGCS[\"Tahaa 54\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"Tahaa_54\",SPHEROID[\"International 1924\",6378388");
    add_srs_wkt (p, 2,
		 ",297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "6629\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 4,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 5,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4629\"]],UNIT[\"met");
    add_srs_wkt (p, 6,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 7,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 8,
		 "AMETER[\"central_meridian\",-153],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 9,
		 "or\",0.9996],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"297");
    add_srs_wkt (p, 11,
		 "7\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2978, "epsg", 2978,
		      "IGN72 Nuku Hiva / UTM zone 7S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=7 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGN72 Nuku Hiva / UTM zone 7S\",GEOGCS[\"IGN72 ");
    add_srs_wkt (p, 1,
		 "Nuku Hiva\",DATUM[\"IGN72_Nuku_Hiva\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6630\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"4630\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",-141],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",10000000],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"2978\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2979, "epsg", 2979,
		      "K0 1949 / UTM zone 42S (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=42 +south +ellps=intl +towgs84=145,-187,");
    add_proj4text (p, 1, "103,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"K0 1949 / UTM zone 42S (deprecated)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "K0 1949\",DATUM[\"K0_1949\",SPHEROID[\"International 192");
    add_srs_wkt (p, 2,
		 "4\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[14");
    add_srs_wkt (p, 3,
		 "5,-187,103,0,0,0,0],AUTHORITY[\"EPSG\",\"6631\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4631\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",69],PARAMETER[\"scale_factor\",0.9996],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_easting\",500000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 11,
		 "ng\",10000000],AUTHORITY[\"EPSG\",\"2979\"],AXIS[\"Easti");
    add_srs_wkt (p, 12, "ng\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2980, "epsg", 2980,
		      "Combani 1950 / UTM zone 38S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=38 +south +ellps=intl +towgs84=-382,-59,");
    add_proj4text (p, 1, "-262,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Combani 1950 / UTM zone 38S\",GEOGCS[\"Combani ");
    add_srs_wkt (p, 1,
		 "1950\",DATUM[\"Combani_1950\",SPHEROID[\"International 1");
    add_srs_wkt (p, 2,
		 "924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[");
    add_srs_wkt (p, 3,
		 "-382,-59,-262,0,0,0,0],AUTHORITY[\"EPSG\",\"6632\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4632\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",45],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 11,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"2980\"],AXIS[\"Eas");
    add_srs_wkt (p, 12, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2981, "epsg", 2981,
		      "IGN56 Lifou / UTM zone 58S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGN56 Lifou / UTM zone 58S\",GEOGCS[\"IGN56 Lif");
    add_srs_wkt (p, 1,
		 "ou\",DATUM[\"IGN56_Lifou\",SPHEROID[\"International 1924");
    add_srs_wkt (p, 2,
		 "\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6633\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4633\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",0],PARAMETER[\"central_meridian\",165],PARAMETER[\"sc");
    add_srs_wkt (p, 9,
		 "ale_factor\",0.9996],PARAMETER[\"false_easting\",500000]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",10000000],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 11,
		 "\",\"2981\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NO");
    add_srs_wkt (p, 12, "RTH]]");
    p = add_epsg_def (first, last, 2982, "epsg", 2982,
		      "IGN72 Grand Terre / UTM zone 58S (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGN72 Grand Terre / UTM zone 58S (deprecated)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"IGN72 Grand Terre\",DATUM[\"IGN72_Grande_Terre");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6634\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.0174532925199433,AUTHORITY[\"EPSG\",\"9108\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4634\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",165],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 11,
		 "ing\",10000000],AUTHORITY[\"EPSG\",\"2982\"],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2983, "epsg", 2983,
		      "ST87 Ouvea / UTM zone 58S (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +south +ellps=intl +towgs84=-122.383,");
    add_proj4text (p, 1,
		   "-188.696,103.344,3.5107,-4.9668,-5.7047,4.4798 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ST87 Ouvea / UTM zone 58S (deprecated)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"ST87 Ouvea\",DATUM[\"ST87_Ouvea\",SPHEROID[\"Internat");
    add_srs_wkt (p, 2,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[-122.383,-188.696,103.344,3.5107,-4.9668,-5.7047,");
    add_srs_wkt (p, 4,
		 "4.4798],AUTHORITY[\"EPSG\",\"6635\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 5,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 6,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"4635\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 8,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 10,
		 "65],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"2983\"],AXIS[\"Easting\",EAST],AX");
    add_srs_wkt (p, 13, "IS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2984, "epsg", 2984,
		      "RGNC 1991 / Lambert New Caledonia (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-20.66666666666667 +lat_2=-22.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-21.5 +lon_0=166 +x_0=400000 +y_0=300000 +e");
    add_proj4text (p, 2, "llps=intl +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGNC 1991 / Lambert New Caledonia (deprecated)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"RGNC 1991\",DATUM[\"Reseau_Geodesique_Nouvelle");
    add_srs_wkt (p, 2,
		 "_Caledonie_1991\",SPHEROID[\"International 1924\",637838");
    add_srs_wkt (p, 3,
		 "8,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[0,0,0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0],AUTHORITY[\"EPSG\",\"6645\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 6,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"4645\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 8,
		 "\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"standard_parallel_1\",-20.66666666666667],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"standard_parallel_2\",-22.33333333333333],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "latitude_of_origin\",-21.5],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 12,
		 "\",166],PARAMETER[\"false_easting\",400000],PARAMETER[\"");
    add_srs_wkt (p, 13,
		 "false_northing\",300000],AUTHORITY[\"EPSG\",\"2984\"],AX");
    add_srs_wkt (p, 14, "IS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2987, "epsg", 2987,
		      "Saint Pierre et Miquelon 1950 / UTM zone 21N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +ellps=clrk66 +towgs84=30,430,368,0,0");
    add_proj4text (p, 1, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Saint Pierre et Miquelon 1950 / UTM zone 21N\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Saint Pierre et Miquelon 1950\",DATUM[\"Saint_P");
    add_srs_wkt (p, 2,
		 "ierre_et_Miquelon_1950\",SPHEROID[\"Clarke 1866\",637820");
    add_srs_wkt (p, 3,
		 "6.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],TOWG");
    add_srs_wkt (p, 4,
		 "S84[30,430,368,0,0,0,0],AUTHORITY[\"EPSG\",\"6638\"]],PR");
    add_srs_wkt (p, 5,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 7,
		 "\"]],AUTHORITY[\"EPSG\",\"4638\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 8,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 9,
		 "or\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"ce");
    add_srs_wkt (p, 10,
		 "ntral_meridian\",-57],PARAMETER[\"scale_factor\",0.9996]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_easting\",500000],PARAMETER[\"false_n");
    add_srs_wkt (p, 12,
		 "orthing\",0],AUTHORITY[\"EPSG\",\"2987\"],AXIS[\"Easting");
    add_srs_wkt (p, 13, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2988, "epsg", 2988, "MOP78 / UTM zone 1S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=1 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MOP78 / UTM zone 1S\",GEOGCS[\"MOP78\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "MOP78\",SPHEROID[\"International 1924\",6378388,297,AUTH");
    add_srs_wkt (p, 2,
		 "ORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6639\"]],");
    add_srs_wkt (p, 3,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 4,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 5,
		 "22\"]],AUTHORITY[\"EPSG\",\"4639\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 7,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "central_meridian\",-177],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"2988\"],AXIS");
    add_srs_wkt (p, 11, "[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2989, "epsg", 2989,
		      "RRAF 1991 / UTM zone 20N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RRAF 1991 / UTM zone 20N\",GEOGCS[\"RRAF 1991\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Reseau_de_Reference_des_Antilles_Francaises_199");
    add_srs_wkt (p, 2,
		 "1\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"6640\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 5,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4640\"]],UN");
    add_srs_wkt (p, 7,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 9,
		 ",0],PARAMETER[\"central_meridian\",-63],PARAMETER[\"scal");
    add_srs_wkt (p, 10,
		 "e_factor\",0.9996],PARAMETER[\"false_easting\",500000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2989");
    add_srs_wkt (p, 12, "\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2990, "epsg", 2990,
		      "Reunion 1947 / TM Reunion (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-21.11666666666667 +lon_0=55.53333333");
    add_proj4text (p, 1,
		   "333333 +k=1 +x_0=50000 +y_0=160000 +ellps=intl +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Reunion 1947 / TM Reunion (deprecated)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"Reunion 1947\",DATUM[\"Reunion_1947\",SPHEROID[\"Inte");
    add_srs_wkt (p, 2,
		 "rnational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6626\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4626\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"lat");
    add_srs_wkt (p, 8,
		 "itude_of_origin\",-21.11666666666667],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",55.53333333333333],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",1],PARAMETER[\"false_easting\",50000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",160000],AUTHORITY[\"EPSG\",\"2990\"],AXIS[");
    add_srs_wkt (p, 12, "\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2991, "epsg", 2991,
		      "NAD83 / Oregon Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43 +lat_2=45.5 +lat_0=41.75 +lon_0=-120");
    add_proj4text (p, 1,
		   ".5 +x_0=400000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m");
    add_proj4text (p, 2, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Oregon Lambert\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 8,
		 "rd_parallel_1\",43],PARAMETER[\"standard_parallel_2\",45");
    add_srs_wkt (p, 9,
		 ".5],PARAMETER[\"latitude_of_origin\",41.75],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "central_meridian\",-120.5],PARAMETER[\"false_easting\",4");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"2991\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2992, "epsg", 2992,
		      "NAD83 / Oregon Lambert (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43 +lat_2=45.5 +lat_0=41.75 +lon_0=-120");
    add_proj4text (p, 1,
		   ".5 +x_0=399999.9999984 +y_0=0 +ellps=GRS80 +datum=NAD83 ");
    add_proj4text (p, 2, "+units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Oregon Lambert (ft)\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"standard_parallel_1\",43],PARAMETER[\"standard_parall");
    add_srs_wkt (p, 9,
		 "el_2\",45.5],PARAMETER[\"latitude_of_origin\",41.75],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"central_meridian\",-120.5],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",1312335.958],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 12,
		 "HORITY[\"EPSG\",\"2992\"],AXIS[\"X\",EAST],AXIS[\"Y\",NO");
    add_srs_wkt (p, 13, "RTH]]");
    p = add_epsg_def (first, last, 2993, "epsg", 2993,
		      "NAD83(HARN) / Oregon Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43 +lat_2=45.5 +lat_0=41.75 +lon_0=-120");
    add_proj4text (p, 1,
		   ".5 +x_0=400000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Oregon Lambert\",GEOGCS[\"NAD83(H");
    add_srs_wkt (p, 1,
		 "ARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",43],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_2\",45.5],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",41.75],PARAMETER[\"central_meridian\",-120.5],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_easting\",400000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",0],AUTHORITY[\"EPSG\",\"2993\"],AXIS[\"X\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2994, "epsg", 2994,
		      "NAD83(HARN) / Oregon Lambert (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43 +lat_2=45.5 +lat_0=41.75 +lon_0=-120");
    add_proj4text (p, 1,
		   ".5 +x_0=399999.9999984 +y_0=0 +ellps=GRS80 +units=ft +no");
    add_proj4text (p, 2, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Oregon Lambert (ft)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"foot\",0.3048,AUT");
    add_srs_wkt (p, 7,
		 "HORITY[\"EPSG\",\"9002\"]],PROJECTION[\"Lambert_Conforma");
    add_srs_wkt (p, 8,
		 "l_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",43],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"standard_parallel_2\",45.5],PARAMETER[\"latitud");
    add_srs_wkt (p, 10,
		 "e_of_origin\",41.75],PARAMETER[\"central_meridian\",-120");
    add_srs_wkt (p, 11,
		 ".5],PARAMETER[\"false_easting\",1312335.958],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"2994\"],AXIS[\"");
    add_srs_wkt (p, 13, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 2995, "epsg", 2995,
		      "IGN53 Mare / UTM zone 58S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGN53 Mare / UTM zone 58S\",GEOGCS[\"IGN53 Mare");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"IGN53_Mare\",SPHEROID[\"International 1924\",");
    add_srs_wkt (p, 2,
		 "6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6641\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4641\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",165],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",0.9996],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11,
		 "\"2995\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH");
    add_srs_wkt (p, 12, "]]");
    p = add_epsg_def (first, last, 2996, "epsg", 2996,
		      "ST84 Ile des Pins / UTM zone 58S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ST84 Ile des Pins / UTM zone 58S\",GEOGCS[\"ST8");
    add_srs_wkt (p, 1,
		 "4 Ile des Pins\",DATUM[\"ST84_Ile_des_Pins\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "22\"]],AUTHORITY[\"EPSG\",\"6642\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4642\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "65],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"2996\"],AXIS[\"Easting\",EAST],AX");
    add_srs_wkt (p, 12, "IS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 2997, "epsg", 2997,
		      "ST71 Belep / UTM zone 58S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +south +ellps=intl +towgs84=-480.26,-");
    add_proj4text (p, 1,
		   "438.32,-643.429,16.3119,20.1721,-4.0349,-111.7 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ST71 Belep / UTM zone 58S\",GEOGCS[\"ST71 Belep");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"ST71_Belep\",SPHEROID[\"International 1924\",");
    add_srs_wkt (p, 2,
		 "6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-480.2");
    add_srs_wkt (p, 3,
		 "6,-438.32,-643.429,16.3119,20.1721,-4.0349,-111.7],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6643\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"464");
    add_srs_wkt (p, 7,
		 "3\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",0],PARAMETER[\"central_meridian\",165],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",10000000],AUTHORITY");
    add_srs_wkt (p, 12,
		 "[\"EPSG\",\"2997\"],AXIS[\"Easting\",EAST],AXIS[\"Northi");
    add_srs_wkt (p, 13, "ng\",NORTH]]");
    p = add_epsg_def (first, last, 2998, "epsg", 2998,
		      "NEA74 Noumea / UTM zone 58S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NEA74 Noumea / UTM zone 58S\",GEOGCS[\"NEA74 No");
    add_srs_wkt (p, 1,
		 "umea\",DATUM[\"NEA74_Noumea\",SPHEROID[\"International 1");
    add_srs_wkt (p, 2,
		 "924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6644\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4644\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",165],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",10000000],AUTHORITY[\"EP");
    add_srs_wkt (p, 11,
		 "SG\",\"2998\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",");
    add_srs_wkt (p, 12, "NORTH]]");
    p = add_epsg_def (first, last, 2999, "epsg", 2999,
		      "Grand Comoros / UTM zone 38S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=38 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Grand Comoros / UTM zone 38S\",GEOGCS[\"Grand C");
    add_srs_wkt (p, 1,
		 "omoros\",DATUM[\"Grand_Comoros\",SPHEROID[\"Internationa");
    add_srs_wkt (p, 2,
		 "l 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6646\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"464");
    add_srs_wkt (p, 6,
		 "6\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",45],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"scale_factor\",0.9996],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 10,
		 "00000],PARAMETER[\"false_northing\",10000000],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"2999\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 12, "g\",NORTH]]");
    p = add_epsg_def (first, last, 3000, "epsg", 3000, "Segara / NEIEZ");
    add_proj4text (p, 0,
		   "+proj=merc +lon_0=110 +k=0.997 +x_0=3900000 +y_0=900000 ");
    add_proj4text (p, 1, "+ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Segara / NEIEZ\",GEOGCS[\"Segara\",DATUM[\"Gunu");
    add_srs_wkt (p, 1,
		 "ng_Segara\",SPHEROID[\"Bessel 1841\",6377397.155,299.152");
    add_srs_wkt (p, 2,
		 "8128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "6613\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 4,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 5,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4613\"]],UNIT[\"met");
    add_srs_wkt (p, 6,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Mercat");
    add_srs_wkt (p, 7,
		 "or_1SP\"],PARAMETER[\"central_meridian\",110],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"scale_factor\",0.997],PARAMETER[\"false_easting\",3900");
    add_srs_wkt (p, 9,
		 "000],PARAMETER[\"false_northing\",900000],AUTHORITY[\"EP");
    add_srs_wkt (p, 10, "SG\",\"3000\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3001, "epsg", 3001, "Batavia / NEIEZ");
    add_proj4text (p, 0,
		   "+proj=merc +lon_0=110 +k=0.997 +x_0=3900000 +y_0=900000 ");
    add_proj4text (p, 1, "+ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Batavia / NEIEZ\",GEOGCS[\"Batavia\",DATUM[\"Ba");
    add_srs_wkt (p, 1,
		 "tavia\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6211");
    add_srs_wkt (p, 3,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4211\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Mercator_1");
    add_srs_wkt (p, 7,
		 "SP\"],PARAMETER[\"central_meridian\",110],PARAMETER[\"sc");
    add_srs_wkt (p, 8,
		 "ale_factor\",0.997],PARAMETER[\"false_easting\",3900000]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"false_northing\",900000],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 10, ",\"3001\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3002, "epsg", 3002, "Makassar / NEIEZ");
    add_proj4text (p, 0,
		   "+proj=merc +lon_0=110 +k=0.997 +x_0=3900000 +y_0=900000 ");
    add_proj4text (p, 1,
		   "+ellps=bessel +towgs84=-587.8,519.75,145.76,0,0,0,0 +uni");
    add_proj4text (p, 2, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Makassar / NEIEZ\",GEOGCS[\"Makassar\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Makassar\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528");
    add_srs_wkt (p, 2,
		 "128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-587.8,519.75,");
    add_srs_wkt (p, 3,
		 "145.76,0,0,0,0],AUTHORITY[\"EPSG\",\"6257\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4257\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Mercator_1SP\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"central_meridian\",110],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "997],PARAMETER[\"false_easting\",3900000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",900000],AUTHORITY[\"EPSG\",\"3002\"],AXIS");
    add_srs_wkt (p, 11, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3003, "epsg", 3003,
		      "Monte Mario / Italy zone 1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=9 +k=0.9996 +x_0=1500000 +y_");
    add_proj4text (p, 1, "0=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Monte Mario / Italy zone 1\",GEOGCS[\"Monte Mar");
    add_srs_wkt (p, 1,
		 "io\",DATUM[\"Monte_Mario\",SPHEROID[\"International 1924");
    add_srs_wkt (p, 2,
		 "\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6265\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4265\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",0],PARAMETER[\"central_meridian\",9],PARAMETER[\"scal");
    add_srs_wkt (p, 9,
		 "e_factor\",0.9996],PARAMETER[\"false_easting\",1500000],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"300");
    add_srs_wkt (p, 11, "3\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3004, "epsg", 3004,
		      "Monte Mario / Italy zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=0.9996 +x_0=2520000 +y");
    add_proj4text (p, 1, "_0=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Monte Mario / Italy zone 2\",GEOGCS[\"Monte Mar");
    add_srs_wkt (p, 1,
		 "io\",DATUM[\"Monte_Mario\",SPHEROID[\"International 1924");
    add_srs_wkt (p, 2,
		 "\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6265\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4265\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",0],PARAMETER[\"central_meridian\",15],PARAMETER[\"sca");
    add_srs_wkt (p, 9,
		 "le_factor\",0.9996],PARAMETER[\"false_easting\",2520000]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"30");
    add_srs_wkt (p, 11, "04\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3005, "epsg", 3005, "NAD83 / BC Albers");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=50 +lat_2=58.5 +lat_0=45 +lon_0=-126 +x");
    add_proj4text (p, 1,
		   "_0=1000000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no");
    add_proj4text (p, 2, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / BC Albers\",GEOGCS[\"NAD83\",DATUM[\"No");
    add_srs_wkt (p, 1,
		 "rth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137,");
    add_srs_wkt (p, 2,
		 "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Albers_Conic_Equal_Area\"],PARAMETER[\"standard_paral");
    add_srs_wkt (p, 8,
		 "lel_1\",50],PARAMETER[\"standard_parallel_2\",58.5],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"latitude_of_center\",45],PARAMETER[\"longitude_o");
    add_srs_wkt (p, 10,
		 "f_center\",-126],PARAMETER[\"false_easting\",1000000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3005\"");
    add_srs_wkt (p, 12, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3006, "epsg", 3006, "SWEREF99 TM");
    add_proj4text (p, 0,
		   "+proj=utm +zone=33 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SWEREF99 TM\",GEOGCS[\"SWEREF99\",DATUM[\"SWERE");
    add_srs_wkt (p, 1,
		 "F99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR");
    add_srs_wkt (p, 2,
		 "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4619\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",15],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "3006\"],AXIS[\"x\",NORTH],AXIS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 3007, "epsg", 3007, "SWEREF99 12 00");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=12 +k=1 +x_0=150000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SWEREF99 12 00\",GEOGCS[\"SWEREF99\",DATUM[\"SW");
    add_srs_wkt (p, 1,
		 "EREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4619");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",12],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",1],PARAMETER[\"false_easting\",150000]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"30");
    add_srs_wkt (p, 11, "07\"],AXIS[\"x\",NORTH],AXIS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 3008, "epsg", 3008, "SWEREF99 13 30");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=13.5 +k=1 +x_0=150000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SWEREF99 13 30\",GEOGCS[\"SWEREF99\",DATUM[\"SW");
    add_srs_wkt (p, 1,
		 "EREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4619");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",13.5],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",15000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "3008\"],AXIS[\"x\",NORTH],AXIS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 3009, "epsg", 3009, "SWEREF99 15 00");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=1 +x_0=150000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SWEREF99 15 00\",GEOGCS[\"SWEREF99\",DATUM[\"SW");
    add_srs_wkt (p, 1,
		 "EREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4619");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",15],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",1],PARAMETER[\"false_easting\",150000]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"30");
    add_srs_wkt (p, 11, "09\"],AXIS[\"x\",NORTH],AXIS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 3010, "epsg", 3010, "SWEREF99 16 30");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=16.5 +k=1 +x_0=150000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SWEREF99 16 30\",GEOGCS[\"SWEREF99\",DATUM[\"SW");
    add_srs_wkt (p, 1,
		 "EREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4619");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",16.5],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",15000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "3010\"],AXIS[\"x\",NORTH],AXIS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 3011, "epsg", 3011, "SWEREF99 18 00");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18 +k=1 +x_0=150000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SWEREF99 18 00\",GEOGCS[\"SWEREF99\",DATUM[\"SW");
    add_srs_wkt (p, 1,
		 "EREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4619");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",18],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",1],PARAMETER[\"false_easting\",150000]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"30");
    add_srs_wkt (p, 11, "11\"],AXIS[\"x\",NORTH],AXIS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 3012, "epsg", 3012, "SWEREF99 14 15");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=14.25 +k=1 +x_0=150000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SWEREF99 14 15\",GEOGCS[\"SWEREF99\",DATUM[\"SW");
    add_srs_wkt (p, 1,
		 "EREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4619");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",14.25],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",1500");
    add_srs_wkt (p, 10,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "3012\"],AXIS[\"x\",NORTH],AXIS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 3013, "epsg", 3013, "SWEREF99 15 45");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15.75 +k=1 +x_0=150000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SWEREF99 15 45\",GEOGCS[\"SWEREF99\",DATUM[\"SW");
    add_srs_wkt (p, 1,
		 "EREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4619");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",15.75],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",1500");
    add_srs_wkt (p, 10,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "3013\"],AXIS[\"x\",NORTH],AXIS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 3014, "epsg", 3014, "SWEREF99 17 15");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=17.25 +k=1 +x_0=150000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SWEREF99 17 15\",GEOGCS[\"SWEREF99\",DATUM[\"SW");
    add_srs_wkt (p, 1,
		 "EREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4619");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",17.25],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",1500");
    add_srs_wkt (p, 10,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "3014\"],AXIS[\"x\",NORTH],AXIS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 3015, "epsg", 3015, "SWEREF99 18 45");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18.75 +k=1 +x_0=150000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SWEREF99 18 45\",GEOGCS[\"SWEREF99\",DATUM[\"SW");
    add_srs_wkt (p, 1,
		 "EREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4619");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",18.75],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",1500");
    add_srs_wkt (p, 10,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "3015\"],AXIS[\"x\",NORTH],AXIS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 3016, "epsg", 3016, "SWEREF99 20 15");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=20.25 +k=1 +x_0=150000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SWEREF99 20 15\",GEOGCS[\"SWEREF99\",DATUM[\"SW");
    add_srs_wkt (p, 1,
		 "EREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4619");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",20.25],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",1500");
    add_srs_wkt (p, 10,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "3016\"],AXIS[\"x\",NORTH],AXIS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 3017, "epsg", 3017, "SWEREF99 21 45");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21.75 +k=1 +x_0=150000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SWEREF99 21 45\",GEOGCS[\"SWEREF99\",DATUM[\"SW");
    add_srs_wkt (p, 1,
		 "EREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4619");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",21.75],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",1500");
    add_srs_wkt (p, 10,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "3017\"],AXIS[\"x\",NORTH],AXIS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 3018, "epsg", 3018, "SWEREF99 23 15");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=23.25 +k=1 +x_0=150000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SWEREF99 23 15\",GEOGCS[\"SWEREF99\",DATUM[\"SW");
    add_srs_wkt (p, 1,
		 "EREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4619");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",23.25],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",1500");
    add_srs_wkt (p, 10,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "3018\"],AXIS[\"x\",NORTH],AXIS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 3019, "epsg", 3019, "RT90 7.5 gon V");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=11.30827777777778 +k=1 +x_0=");
    add_proj4text (p, 1, "1500000 +y_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RT90 7.5 gon V\",GEOGCS[\"RT90\",DATUM[\"Rikets");
    add_srs_wkt (p, 1,
		 "_koordinatsystem_1990\",SPHEROID[\"Bessel 1841\",6377397");
    add_srs_wkt (p, 2,
		 ".155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6124\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4124\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",11.3082777777777");
    add_srs_wkt (p, 9,
		 "8],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",1500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"3019\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3020, "epsg", 3020, "RT90 5 gon V");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=13.55827777777778 +k=1 +x_0=");
    add_proj4text (p, 1, "1500000 +y_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RT90 5 gon V\",GEOGCS[\"RT90\",DATUM[\"Rikets_k");
    add_srs_wkt (p, 1,
		 "oordinatsystem_1990\",SPHEROID[\"Bessel 1841\",6377397.1");
    add_srs_wkt (p, 2,
		 "55,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6124\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4124\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",0],PARAMETER[\"central_meridian\",13.55827777777778],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",1500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 11, "SG\",\"3020\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3021, "epsg", 3021, "RT90 2.5 gon V");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15.80827777777778 +k=1 +x_0=");
    add_proj4text (p, 1, "1500000 +y_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RT90 2.5 gon V\",GEOGCS[\"RT90\",DATUM[\"Rikets");
    add_srs_wkt (p, 1,
		 "_koordinatsystem_1990\",SPHEROID[\"Bessel 1841\",6377397");
    add_srs_wkt (p, 2,
		 ".155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6124\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4124\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",15.8082777777777");
    add_srs_wkt (p, 9,
		 "8],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",1500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"3021\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3022, "epsg", 3022, "RT90 0 gon");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18.05827777777778 +k=1 +x_0=");
    add_proj4text (p, 1, "1500000 +y_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RT90 0 gon\",GEOGCS[\"RT90\",DATUM[\"Rikets_koo");
    add_srs_wkt (p, 1,
		 "rdinatsystem_1990\",SPHEROID[\"Bessel 1841\",6377397.155");
    add_srs_wkt (p, 2,
		 ",299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6124\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4124\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",0],PARAMETER[\"central_meridian\",18.05827777777778],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 10,
		 "1500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 11, "G\",\"3022\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3023, "epsg", 3023, "RT90 2.5 gon O");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=20.30827777777778 +k=1 +x_0=");
    add_proj4text (p, 1, "1500000 +y_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RT90 2.5 gon O\",GEOGCS[\"RT90\",DATUM[\"Rikets");
    add_srs_wkt (p, 1,
		 "_koordinatsystem_1990\",SPHEROID[\"Bessel 1841\",6377397");
    add_srs_wkt (p, 2,
		 ".155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6124\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4124\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",20.3082777777777");
    add_srs_wkt (p, 9,
		 "8],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",1500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"3023\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3024, "epsg", 3024, "RT90 5 gon O");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=22.55827777777778 +k=1 +x_0=");
    add_proj4text (p, 1, "1500000 +y_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RT90 5 gon O\",GEOGCS[\"RT90\",DATUM[\"Rikets_k");
    add_srs_wkt (p, 1,
		 "oordinatsystem_1990\",SPHEROID[\"Bessel 1841\",6377397.1");
    add_srs_wkt (p, 2,
		 "55,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6124\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4124\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",0],PARAMETER[\"central_meridian\",22.55827777777778],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",1500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 11, "SG\",\"3024\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3025, "epsg", 3025, "RT38 7.5 gon V");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=11.30827777777778 +k=1 +x_0=");
    add_proj4text (p, 1, "1500000 +y_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RT38 7.5 gon V\",GEOGCS[\"RT38\",DATUM[\"Stockh");
    add_srs_wkt (p, 1,
		 "olm_1938\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528");
    add_srs_wkt (p, 2,
		 "128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "308\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4308\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transve");
    add_srs_wkt (p, 7,
		 "rse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"central_meridian\",11.30827777777778],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"scale_factor\",1],PARAMETER[\"false_easting\",1500000]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"30");
    add_srs_wkt (p, 11, "25\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3026, "epsg", 3026, "RT38 5 gon V");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=13.55827777777778 +k=1 +x_0=");
    add_proj4text (p, 1, "1500000 +y_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RT38 5 gon V\",GEOGCS[\"RT38\",DATUM[\"Stockhol");
    add_srs_wkt (p, 1,
		 "m_1938\",SPHEROID[\"Bessel 1841\",6377397.155,299.152812");
    add_srs_wkt (p, 2,
		 "8,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"630");
    add_srs_wkt (p, 3,
		 "8\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4308\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",13.55827777777778],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "cale_factor\",1],PARAMETER[\"false_easting\",1500000],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3026\"");
    add_srs_wkt (p, 11, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3027, "epsg", 3027, "RT38 2.5 gon V");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15.80827777777778 +k=1 +x_0=");
    add_proj4text (p, 1, "1500000 +y_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RT38 2.5 gon V\",GEOGCS[\"RT38\",DATUM[\"Stockh");
    add_srs_wkt (p, 1,
		 "olm_1938\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528");
    add_srs_wkt (p, 2,
		 "128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "308\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4308\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transve");
    add_srs_wkt (p, 7,
		 "rse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"central_meridian\",15.80827777777778],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"scale_factor\",1],PARAMETER[\"false_easting\",1500000]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"30");
    add_srs_wkt (p, 11, "27\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3028, "epsg", 3028, "RT38 0 gon");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18.05827777777778 +k=1 +x_0=");
    add_proj4text (p, 1, "1500000 +y_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RT38 0 gon\",GEOGCS[\"RT38\",DATUM[\"Stockholm_");
    add_srs_wkt (p, 1,
		 "1938\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6308\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4308\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",18.05827777777778],PARAMETER[\"scal");
    add_srs_wkt (p, 9,
		 "e_factor\",1],PARAMETER[\"false_easting\",1500000],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3028\"],");
    add_srs_wkt (p, 11, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3029, "epsg", 3029, "RT38 2.5 gon O");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=20.30827777777778 +k=1 +x_0=");
    add_proj4text (p, 1, "1500000 +y_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RT38 2.5 gon O\",GEOGCS[\"RT38\",DATUM[\"Stockh");
    add_srs_wkt (p, 1,
		 "olm_1938\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528");
    add_srs_wkt (p, 2,
		 "128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "308\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4308\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transve");
    add_srs_wkt (p, 7,
		 "rse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"central_meridian\",20.30827777777778],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"scale_factor\",1],PARAMETER[\"false_easting\",1500000]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"30");
    add_srs_wkt (p, 11, "29\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3030, "epsg", 3030, "RT38 5 gon O");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=22.55827777777778 +k=1 +x_0=");
    add_proj4text (p, 1, "1500000 +y_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RT38 5 gon O\",GEOGCS[\"RT38\",DATUM[\"Stockhol");
    add_srs_wkt (p, 1,
		 "m_1938\",SPHEROID[\"Bessel 1841\",6377397.155,299.152812");
    add_srs_wkt (p, 2,
		 "8,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"630");
    add_srs_wkt (p, 3,
		 "8\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4308\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",22.55827777777778],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "cale_factor\",1],PARAMETER[\"false_easting\",1500000],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3030\"");
    add_srs_wkt (p, 11, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3031, "epsg", 3031,
		      "WGS 84 / Antarctic Polar Stereographic");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-71 +lon_0=0 +k=1 +x_0=0 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=WGS84 +datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / Antarctic Polar Stereographic\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"WGS 84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",63781");
    add_srs_wkt (p, 2,
		 "37,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Polar_Stereographic\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",-71],PARAMETER[\"central_meridian\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",1],PARAMETER[\"false_easting\",0],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3031\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3032, "epsg", 3032,
		      "WGS 84 / Australian Antarctic Polar Stereographic");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-71 +lon_0=70 +k=1 +x_0=6");
    add_proj4text (p, 1,
		   "000000 +y_0=6000000 +ellps=WGS84 +datum=WGS84 +units=m +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / Australian Antarctic Polar Stereograph");
    add_srs_wkt (p, 1,
		 "ic\",GEOGCS[\"WGS 84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS");
    add_srs_wkt (p, 2,
		 " 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4326\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Polar_Stereographic\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",-71],PARAMETER[\"central_meridian\",70],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",6000000],PARAMETER[\"false_northing\",6000000],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"3032\"],AXIS[\"Easting\",UNKNOWN],AXIS[\"No");
    add_srs_wkt (p, 12, "rthing\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3033, "epsg", 3033,
		      "WGS 84 / Australian Antarctic Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.5 +lat_2=-74.5 +lat_0=-50 +lon_0=70");
    add_proj4text (p, 1,
		   " +x_0=6000000 +y_0=6000000 +ellps=WGS84 +datum=WGS84 +un");
    add_proj4text (p, 2, "its=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / Australian Antarctic Lambert\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"WGS 84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard");
    add_srs_wkt (p, 8,
		 "_parallel_1\",-68.5],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "74.5],PARAMETER[\"latitude_of_origin\",-50],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "central_meridian\",70],PARAMETER[\"false_easting\",60000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",6000000],AUTHORITY[\"EP");
    add_srs_wkt (p, 12,
		 "SG\",\"3033\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",");
    add_srs_wkt (p, 13, "NORTH]]");
    p = add_epsg_def (first, last, 3034, "epsg", 3034, "ETRS89 / ETRS-LCC");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=35 +lat_2=65 +lat_0=52 +lon_0=10 +x_0=4");
    add_proj4text (p, 1, "000000 +y_0=2800000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-LCC\",GEOGCS[\"ETRS89\",DATUM[\"E");
    add_srs_wkt (p, 1,
		 "uropean_Terrestrial_Reference_System_1989\",SPHEROID[\"G");
    add_srs_wkt (p, 2,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "9\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"standard_parallel_1\",35],PARAMETER[\"standard_p");
    add_srs_wkt (p, 9,
		 "arallel_2\",65],PARAMETER[\"latitude_of_origin\",52],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"central_meridian\",10],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 11,
		 "g\",4000000],PARAMETER[\"false_northing\",2800000],AUTHO");
    add_srs_wkt (p, 12,
		 "RITY[\"EPSG\",\"3034\"],AXIS[\"Northing\",NORTH],AXIS[\"");
    add_srs_wkt (p, 13, "Easting\",EAST]]");
    p = add_epsg_def (first, last, 3035, "epsg", 3035, "ETRS89 / ETRS-LAEA");
    add_proj4text (p, 0,
		   "+proj=laea +lat_0=52 +lon_0=10 +x_0=4321000 +y_0=3210000");
    add_proj4text (p, 1, " +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-LAEA\",GEOGCS[\"ETRS89\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Terrestrial_Reference_System_1989\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Lambert_Azimuthal_Equal_Area\"],PAR");
    add_srs_wkt (p, 8,
		 "AMETER[\"latitude_of_center\",52],PARAMETER[\"longitude_");
    add_srs_wkt (p, 9,
		 "of_center\",10],PARAMETER[\"false_easting\",4321000],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_northing\",3210000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "3035\"],AXIS[\"Y\",NORTH],AXIS[\"X\",EAST]]");
    p = add_epsg_def (first, last, 3036, "epsg", 3036, "Moznet / UTM zone 36S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=36 +south +ellps=WGS84 +towgs84=0,0,0,-0");
    add_proj4text (p, 1, ",-0,-0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Moznet / UTM zone 36S\",GEOGCS[\"Moznet\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Moznet_ITRF94\",SPHEROID[\"WGS 84\",6378137,298.25722");
    add_srs_wkt (p, 2,
		 "3563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,-0,-0,-");
    add_srs_wkt (p, 3,
		 "0,0],AUTHORITY[\"EPSG\",\"6130\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4130\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",33");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 10,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"3036\"],AXIS[\"Easting\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3037, "epsg", 3037, "Moznet / UTM zone 37S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=37 +south +ellps=WGS84 +towgs84=0,0,0,-0");
    add_proj4text (p, 1, ",-0,-0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Moznet / UTM zone 37S\",GEOGCS[\"Moznet\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Moznet_ITRF94\",SPHEROID[\"WGS 84\",6378137,298.25722");
    add_srs_wkt (p, 2,
		 "3563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,-0,-0,-");
    add_srs_wkt (p, 3,
		 "0,0],AUTHORITY[\"EPSG\",\"6130\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4130\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",39");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 10,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"3037\"],AXIS[\"Easting\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3038, "epsg", 3038, "ETRS89 / ETRS-TM26");
    add_proj4text (p, 0, "+proj=utm +zone=26 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-TM26\",GEOGCS[\"ETRS89\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Terrestrial_Reference_System_1989\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 9,
		 "27],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 11,
		 "RITY[\"EPSG\",\"3038\"],AXIS[\"Northing\",NORTH],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST]]");
    p = add_epsg_def (first, last, 3039, "epsg", 3039, "ETRS89 / ETRS-TM27");
    add_proj4text (p, 0, "+proj=utm +zone=27 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-TM27\",GEOGCS[\"ETRS89\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Terrestrial_Reference_System_1989\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 9,
		 "21],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 11,
		 "RITY[\"EPSG\",\"3039\"],AXIS[\"Northing\",NORTH],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST]]");
    p = add_epsg_def (first, last, 3040, "epsg", 3040, "ETRS89 / ETRS-TM28");
    add_proj4text (p, 0, "+proj=utm +zone=28 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-TM28\",GEOGCS[\"ETRS89\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Terrestrial_Reference_System_1989\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 9,
		 "15],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 11,
		 "RITY[\"EPSG\",\"3040\"],AXIS[\"Northing\",NORTH],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST]]");
    p = add_epsg_def (first, last, 3041, "epsg", 3041, "ETRS89 / ETRS-TM29");
    add_proj4text (p, 0, "+proj=utm +zone=29 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-TM29\",GEOGCS[\"ETRS89\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Terrestrial_Reference_System_1989\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 9,
		 "9],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"3041\"],AXIS[\"Northing\",NORTH],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST]]");
    p = add_epsg_def (first, last, 3042, "epsg", 3042, "ETRS89 / ETRS-TM30");
    add_proj4text (p, 0, "+proj=utm +zone=30 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-TM30\",GEOGCS[\"ETRS89\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Terrestrial_Reference_System_1989\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"3042\"],AXIS[\"Northing\",NORTH],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST]]");
    p = add_epsg_def (first, last, 3043, "epsg", 3043, "ETRS89 / ETRS-TM31");
    add_proj4text (p, 0, "+proj=utm +zone=31 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-TM31\",GEOGCS[\"ETRS89\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Terrestrial_Reference_System_1989\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",3");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 10,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 11,
		 "TY[\"EPSG\",\"3043\"],AXIS[\"Northing\",NORTH],AXIS[\"Ea");
    add_srs_wkt (p, 12, "sting\",EAST]]");
    p = add_epsg_def (first, last, 3044, "epsg", 3044, "ETRS89 / ETRS-TM32");
    add_proj4text (p, 0, "+proj=utm +zone=32 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-TM32\",GEOGCS[\"ETRS89\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Terrestrial_Reference_System_1989\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",9");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 10,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 11,
		 "TY[\"EPSG\",\"3044\"],AXIS[\"Northing\",NORTH],AXIS[\"Ea");
    add_srs_wkt (p, 12, "sting\",EAST]]");
    p = add_epsg_def (first, last, 3045, "epsg", 3045, "ETRS89 / ETRS-TM33");
    add_proj4text (p, 0, "+proj=utm +zone=33 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-TM33\",GEOGCS[\"ETRS89\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Terrestrial_Reference_System_1989\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "5],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"3045\"],AXIS[\"Northing\",NORTH],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST]]");
    p = add_epsg_def (first, last, 3046, "epsg", 3046, "ETRS89 / ETRS-TM34");
    add_proj4text (p, 0, "+proj=utm +zone=34 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-TM34\",GEOGCS[\"ETRS89\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Terrestrial_Reference_System_1989\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",2");
    add_srs_wkt (p, 9,
		 "1],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"3046\"],AXIS[\"Northing\",NORTH],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST]]");
    p = add_epsg_def (first, last, 3047, "epsg", 3047, "ETRS89 / ETRS-TM35");
    add_proj4text (p, 0, "+proj=utm +zone=35 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-TM35\",GEOGCS[\"ETRS89\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Terrestrial_Reference_System_1989\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",2");
    add_srs_wkt (p, 9,
		 "7],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"3047\"],AXIS[\"Northing\",NORTH],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST]]");
    p = add_epsg_def (first, last, 3048, "epsg", 3048, "ETRS89 / ETRS-TM36");
    add_proj4text (p, 0, "+proj=utm +zone=36 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-TM36\",GEOGCS[\"ETRS89\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Terrestrial_Reference_System_1989\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",3");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"3048\"],AXIS[\"Northing\",NORTH],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST]]");
    p = add_epsg_def (first, last, 3049, "epsg", 3049, "ETRS89 / ETRS-TM37");
    add_proj4text (p, 0, "+proj=utm +zone=37 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-TM37\",GEOGCS[\"ETRS89\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Terrestrial_Reference_System_1989\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",3");
    add_srs_wkt (p, 9,
		 "9],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"3049\"],AXIS[\"Northing\",NORTH],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST]]");
    p = add_epsg_def (first, last, 3050, "epsg", 3050, "ETRS89 / ETRS-TM38");
    add_proj4text (p, 0, "+proj=utm +zone=38 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-TM38\",GEOGCS[\"ETRS89\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Terrestrial_Reference_System_1989\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",4");
    add_srs_wkt (p, 9,
		 "5],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"3050\"],AXIS[\"Northing\",NORTH],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST]]");
    p = add_epsg_def (first, last, 3051, "epsg", 3051, "ETRS89 / ETRS-TM39");
    add_proj4text (p, 0, "+proj=utm +zone=39 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-TM39\",GEOGCS[\"ETRS89\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "European_Terrestrial_Reference_System_1989\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",5");
    add_srs_wkt (p, 9,
		 "1],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"3051\"],AXIS[\"Northing\",NORTH],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST]]");
    p = add_epsg_def (first, last, 3054, "epsg", 3054,
		      "Hjorsey 1955 / UTM zone 26N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=26 +ellps=intl +towgs84=-73,46,-86,0,0,0");
    add_proj4text (p, 1, ",0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Hjorsey 1955 / UTM zone 26N\",GEOGCS[\"Hjorsey ");
    add_srs_wkt (p, 1,
		 "1955\",DATUM[\"Hjorsey_1955\",SPHEROID[\"International 1");
    add_srs_wkt (p, 2,
		 "924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[");
    add_srs_wkt (p, 3,
		 "-73,46,-86,0,0,0,0],AUTHORITY[\"EPSG\",\"6658\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4658\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-27],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 11,
		 "ing\",0],AUTHORITY[\"EPSG\",\"3054\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3055, "epsg", 3055,
		      "Hjorsey 1955 / UTM zone 27N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=27 +ellps=intl +towgs84=-73,46,-86,0,0,0");
    add_proj4text (p, 1, ",0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Hjorsey 1955 / UTM zone 27N\",GEOGCS[\"Hjorsey ");
    add_srs_wkt (p, 1,
		 "1955\",DATUM[\"Hjorsey_1955\",SPHEROID[\"International 1");
    add_srs_wkt (p, 2,
		 "924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[");
    add_srs_wkt (p, 3,
		 "-73,46,-86,0,0,0,0],AUTHORITY[\"EPSG\",\"6658\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4658\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-21],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 11,
		 "ing\",0],AUTHORITY[\"EPSG\",\"3055\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3056, "epsg", 3056,
		      "Hjorsey 1955 / UTM zone 28N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=28 +ellps=intl +towgs84=-73,46,-86,0,0,0");
    add_proj4text (p, 1, ",0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Hjorsey 1955 / UTM zone 28N\",GEOGCS[\"Hjorsey ");
    add_srs_wkt (p, 1,
		 "1955\",DATUM[\"Hjorsey_1955\",SPHEROID[\"International 1");
    add_srs_wkt (p, 2,
		 "924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[");
    add_srs_wkt (p, 3,
		 "-73,46,-86,0,0,0,0],AUTHORITY[\"EPSG\",\"6658\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4658\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-15],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 11,
		 "ing\",0],AUTHORITY[\"EPSG\",\"3056\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3057, "epsg", 3057, "ISN93 / Lambert 1993");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=64.25 +lat_2=65.75 +lat_0=65 +lon_0=-19");
    add_proj4text (p, 1,
		   " +x_0=500000 +y_0=500000 +ellps=GRS80 +towgs84=0,0,0,0,0");
    add_proj4text (p, 2, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ISN93 / Lambert 1993\",GEOGCS[\"ISN93\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Islands_Network_1993\",SPHEROID[\"GRS 1980\",6378137,298");
    add_srs_wkt (p, 2,
		 ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0");
    add_srs_wkt (p, 3,
		 ",0,0,0],AUTHORITY[\"EPSG\",\"6659\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4659\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PAR");
    add_srs_wkt (p, 8,
		 "AMETER[\"standard_parallel_1\",64.25],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_2\",65.75],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 10,
		 "65],PARAMETER[\"central_meridian\",-19],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",500000],PARAMETER[\"false_northing\",500000]");
    add_srs_wkt (p, 12,
		 ",AUTHORITY[\"EPSG\",\"3057\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3058, "epsg", 3058,
		      "Helle 1954 / Jan Mayen Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-8.5 +k=1 +x_0=50000 +y_0=-7");
    add_proj4text (p, 1,
		   "800000 +ellps=intl +towgs84=982.609,552.753,-540.873,32.");
    add_proj4text (p, 2, "3934,-153.257,-96.2266,16.805 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Helle 1954 / Jan Mayen Grid\",GEOGCS[\"Helle 19");
    add_srs_wkt (p, 1,
		 "54\",DATUM[\"Helle_1954\",SPHEROID[\"International 1924\"");
    add_srs_wkt (p, 2,
		 ",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[982.6");
    add_srs_wkt (p, 3,
		 "09,552.753,-540.873,32.3934,-153.257,-96.2266,16.805],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6660\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4660\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-8.5],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",-7800000],AUTHORITY[\"");
    add_srs_wkt (p, 12, "EPSG\",\"3058\"],AXIS[\"x\",NORTH],AXIS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 3059, "epsg", 3059, "LKS92 / Latvia TM");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=24 +k=0.9996 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=-6000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"LKS92 / Latvia TM\",GEOGCS[\"LKS92\",DATUM[\"La");
    add_srs_wkt (p, 1,
		 "tvia_1992\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6661\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "661\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 8,
		 "of_origin\",0],PARAMETER[\"central_meridian\",24],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",500000],PARAMETER[\"false_northing\",-6000000],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"3059\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 3060, "epsg", 3060,
		      "IGN72 Grande Terre / UTM zone 58S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGN72 Grande Terre / UTM zone 58S\",GEOGCS[\"IG");
    add_srs_wkt (p, 1,
		 "N72 Grande Terre\",DATUM[\"IGN72_Grande_Terre\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7022\"]],AUTHORITY[\"EPSG\",\"6634\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4662\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",165],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_easting\",500000],PARAMETER[\"false_northing\",100000");
    add_srs_wkt (p, 11,
		 "00],AUTHORITY[\"EPSG\",\"3060\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 12, "AXIS[\"Northing\",NORTH]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_08 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 3061, "epsg", 3061,
		      "Porto Santo 1995 / UTM zone 28N");
    add_proj4text (p, 0, "+proj=utm +zone=28 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Porto Santo 1995 / UTM zone 28N\",GEOGCS[\"Port");
    add_srs_wkt (p, 1,
		 "o Santo 1995\",DATUM[\"Porto_Santo_1995\",SPHEROID[\"Int");
    add_srs_wkt (p, 2,
		 "ernational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6663\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4663\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"lat");
    add_srs_wkt (p, 8,
		 "itude_of_origin\",0],PARAMETER[\"central_meridian\",-15]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_ea");
    add_srs_wkt (p, 10,
		 "sting\",500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"3061\"],AXIS[\"Easting\",EAST],AXIS[\"North");
    add_srs_wkt (p, 12, "ing\",NORTH]]");
    p = add_epsg_def (first, last, 3062, "epsg", 3062,
		      "Azores Oriental 1995 / UTM zone 26N");
    add_proj4text (p, 0, "+proj=utm +zone=26 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Azores Oriental 1995 / UTM zone 26N\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Azores Oriental 1995\",DATUM[\"Azores_Oriental_Islands_1");
    add_srs_wkt (p, 2,
		 "995\",SPHEROID[\"International 1924\",6378388,297,AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6664\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6,
		 "\"]],AUTHORITY[\"EPSG\",\"4664\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"ce");
    add_srs_wkt (p, 9,
		 "ntral_meridian\",-27],PARAMETER[\"scale_factor\",0.9996]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_easting\",500000],PARAMETER[\"false_n");
    add_srs_wkt (p, 11,
		 "orthing\",0],AUTHORITY[\"EPSG\",\"3062\"],AXIS[\"Easting");
    add_srs_wkt (p, 12, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3063, "epsg", 3063,
		      "Azores Central 1995 / UTM zone 26N");
    add_proj4text (p, 0, "+proj=utm +zone=26 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Azores Central 1995 / UTM zone 26N\",GEOGCS[\"A");
    add_srs_wkt (p, 1,
		 "zores Central 1995\",DATUM[\"Azores_Central_Islands_1995");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6665\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4665\"]],UNIT[\"metre\",1,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-27],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 11,
		 "ing\",0],AUTHORITY[\"EPSG\",\"3063\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3064, "epsg", 3064, "IGM95 / UTM zone 32N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=32 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGM95 / UTM zone 32N\",GEOGCS[\"IGM95\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Istituto_Geografico_Militaire_1995\",SPHEROID[\"WGS 84\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOW");
    add_srs_wkt (p, 3,
		 "GS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6670\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4670\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",9],PARAMETER[\"scale_factor\",0.9996],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_easting\",500000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 11,
		 "g\",0],AUTHORITY[\"EPSG\",\"3064\"],AXIS[\"Easting\",EAS");
    add_srs_wkt (p, 12, "T],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3065, "epsg", 3065, "IGM95 / UTM zone 33N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=33 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGM95 / UTM zone 33N\",GEOGCS[\"IGM95\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Istituto_Geografico_Militaire_1995\",SPHEROID[\"WGS 84\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOW");
    add_srs_wkt (p, 3,
		 "GS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6670\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4670\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",15],PARAMETER[\"scale_factor\",0.9996],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_easting\",500000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 11,
		 "ng\",0],AUTHORITY[\"EPSG\",\"3065\"],AXIS[\"Easting\",EA");
    add_srs_wkt (p, 12, "ST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3066, "epsg", 3066, "ED50 / Jordan TM");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=37 +k=0.9998 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=-3000000 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / Jordan TM\",GEOGCS[\"ED50\",DATUM[\"Euro");
    add_srs_wkt (p, 1,
		 "pean_Datum_1950\",SPHEROID[\"International 1924\",637838");
    add_srs_wkt (p, 2,
		 "8,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 4,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 5,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"met");
    add_srs_wkt (p, 6,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 7,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 8,
		 "AMETER[\"central_meridian\",37],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 9,
		 "\",0.9998],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_northing\",-3000000],AUTHORITY[\"EPSG\",\"3066\"");
    add_srs_wkt (p, 11, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3067, "epsg", 3067, "ETRS89 / ETRS-TM35FIN");
    add_proj4text (p, 0, "+proj=utm +zone=35 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-TM35FIN\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",27],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"3067\"],AXIS[\"Easting\",EAST],AXIS[\"N");
    add_srs_wkt (p, 12, "orthing\",NORTH]]");
    p = add_epsg_def (first, last, 3068, "epsg", 3068, "DHDN / Soldner Berlin");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=52.41864827777778 +lon_0=13.6272036666");
    add_proj4text (p, 1,
		   "6667 +x_0=40000 +y_0=10000 +ellps=bessel +datum=potsdam ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DHDN / Soldner Berlin\",GEOGCS[\"DHDN\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Deutsches_Hauptdreiecksnetz\",SPHEROID[\"Bessel 1841\",6");
    add_srs_wkt (p, 2,
		 "377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6314\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "314\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Cassini_Soldner\"],PARAMETER[\"latitude_of_o");
    add_srs_wkt (p, 8,
		 "rigin\",52.41864827777778],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",13.62720366666667],PARAMETER[\"false_easting\",40000],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_northing\",10000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "3068\"],AXIS[\"x\",NORTH],AXIS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 3069, "epsg", 3069,
		      "NAD27 / Wisconsin Transverse Mercator");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-90 +k=0.9996 +x_0=500000 +y");
    add_proj4text (p, 1,
		   "_0=-4500000 +ellps=clrk66 +datum=NAD27 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Wisconsin Transverse Mercator\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD27\",DATUM[\"North_American_Datum_1927\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4267\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-90],PARAMETER[\"scale_factor\",0.9996],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",-4500000],AUTHORITY[\"EPSG\",\"3069\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3070, "epsg", 3070,
		      "NAD83 / Wisconsin Transverse Mercator");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-90 +k=0.9996 +x_0=520000 +y");
    add_proj4text (p, 1,
		   "_0=-4480000 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Wisconsin Transverse Mercator\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 9,
		 "90],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",520000],PARAMETER[\"false_northing\",-4480000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"3070\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 3071, "epsg", 3071,
		      "NAD83(HARN) / Wisconsin Transverse Mercator");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-90 +k=0.9996 +x_0=520000 +y");
    add_proj4text (p, 1, "_0=-4480000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Wisconsin Transverse Mercator\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Region");
    add_srs_wkt (p, 2,
		 "al_Network\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-90],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",520000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",-4480000],AUTHORITY[\"EPSG\",\"3071\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3072, "epsg", 3072,
		      "NAD83 / Maine CS2000 East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=43.83333333333334 +lon_0=-67.875 +k=0");
    add_proj4text (p, 1,
		   ".99998 +x_0=700000 +y_0=0 +ellps=GRS80 +datum=NAD83 +uni");
    add_proj4text (p, 2, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Maine CS2000 East\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 7,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 8,
		 "f_origin\",43.83333333333334],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-67.875],PARAMETER[\"scale_factor\",0.99998],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",700000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 11,
		 "\",0],AUTHORITY[\"EPSG\",\"3072\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3073, "epsg", 3073,
		      "NAD83 / Maine CS2000 Central (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=43 +lon_0=-69.125 +k=0.99998 +x_0=500");
    add_proj4text (p, 1,
		   "000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Maine CS2000 Central (deprecated)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83\",DATUM[\"North_American_Datum_1983\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",43],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",-69.125],PARAMETER[\"scale_factor\",0.99998],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_easting\",500000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 11,
		 "g\",0],AUTHORITY[\"EPSG\",\"3073\"],AXIS[\"X\",EAST],AXI");
    add_srs_wkt (p, 12, "S[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3074, "epsg", 3074,
		      "NAD83 / Maine CS2000 West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.83333333333334 +lon_0=-70.375 +k=0");
    add_proj4text (p, 1,
		   ".99998 +x_0=300000 +y_0=0 +ellps=GRS80 +datum=NAD83 +uni");
    add_proj4text (p, 2, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Maine CS2000 West\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 7,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 8,
		 "f_origin\",42.83333333333334],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-70.375],PARAMETER[\"scale_factor\",0.99998],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",300000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 11,
		 "\",0],AUTHORITY[\"EPSG\",\"3074\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3075, "epsg", 3075,
		      "NAD83(HARN) / Maine CS2000 East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=43.83333333333334 +lon_0=-67.875 +k=0");
    add_proj4text (p, 1,
		   ".99998 +x_0=700000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Maine CS2000 East\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",43.83333333333334],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"central_meridian\",-67.875],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 10,
		 "or\",0.99998],PARAMETER[\"false_easting\",700000],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3075\"],A");
    add_srs_wkt (p, 12, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3076, "epsg", 3076,
		      "NAD83(HARN) / Maine CS2000 Central (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=43 +lon_0=-69.125 +k=0.99998 +x_0=500");
    add_proj4text (p, 1, "000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Maine CS2000 Central (deprecated)");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Re");
    add_srs_wkt (p, 2,
		 "gional_Network\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 8,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",43],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"central_meridian\",-69.125],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.99998],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3076\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3077, "epsg", 3077,
		      "NAD83(HARN) / Maine CS2000 West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.83333333333334 +lon_0=-70.375 +k=0");
    add_proj4text (p, 1,
		   ".99998 +x_0=300000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Maine CS2000 West\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",42.83333333333334],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"central_meridian\",-70.375],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 10,
		 "or\",0.99998],PARAMETER[\"false_easting\",300000],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3077\"],A");
    add_srs_wkt (p, 12, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3078, "epsg", 3078,
		      "NAD83 / Michigan Oblique Mercator");
    add_proj4text (p, 0,
		   "+proj=omerc +lat_0=45.30916666666666 +lonc=-86 +alpha=33");
    add_proj4text (p, 1,
		   "7.25556 +k=0.9996 +x_0=2546731.496 +y_0=-4354009.816 +el");
    add_proj4text (p, 2, "lps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Michigan Oblique Mercator\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Hotine_Oblique_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_center\",45.30916666666666],PARAMETER[\"long");
    add_srs_wkt (p, 9,
		 "itude_of_center\",-86],PARAMETER[\"azimuth\",337.25556],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"rectified_grid_angle\",337.25556],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"scale_factor\",0.9996],PARAMETER[\"false_easting\",254");
    add_srs_wkt (p, 12,
		 "6731.496],PARAMETER[\"false_northing\",-4354009.816],AUT");
    add_srs_wkt (p, 13,
		 "HORITY[\"EPSG\",\"3078\"],AXIS[\"X\",EAST],AXIS[\"Y\",NO");
    add_srs_wkt (p, 14, "RTH]]");
    p = add_epsg_def (first, last, 3079, "epsg", 3079,
		      "NAD83(HARN) / Michigan Oblique Mercator");
    add_proj4text (p, 0,
		   "+proj=omerc +lat_0=45.30916666666666 +lonc=-86 +alpha=33");
    add_proj4text (p, 1,
		   "7.25556 +k=0.9996 +x_0=2546731.496 +y_0=-4354009.816 +el");
    add_proj4text (p, 2, "lps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Michigan Oblique Mercator\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_N");
    add_srs_wkt (p, 2,
		 "etwork\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]]");
    add_srs_wkt (p, 4,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 5,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 6,
		 "122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Hotine_Oblique");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_center\",45.3091666");
    add_srs_wkt (p, 9,
		 "6666666],PARAMETER[\"longitude_of_center\",-86],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"azimuth\",337.25556],PARAMETER[\"rectified_grid_angl");
    add_srs_wkt (p, 11,
		 "e\",337.25556],PARAMETER[\"scale_factor\",0.9996],PARAME");
    add_srs_wkt (p, 12,
		 "TER[\"false_easting\",2546731.496],PARAMETER[\"false_nor");
    add_srs_wkt (p, 13,
		 "thing\",-4354009.816],AUTHORITY[\"EPSG\",\"3079\"],AXIS[");
    add_srs_wkt (p, 14, "\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3080, "epsg", 3080, "NAD27 / Shackleford");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=27.41666666666667 +lat_2=34.91666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=31.16666666666667 +lon_0=-100 +x_0=914400 +y_");
    add_proj4text (p, 2,
		   "0=914400 +ellps=clrk66 +datum=NAD27 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Shackleford\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"standard_parallel_1\",27.41666666666667],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_2\",34.91666666666666],PARAMETER[\"lat");
    add_srs_wkt (p, 10,
		 "itude_of_origin\",31.16666666666667],PARAMETER[\"central");
    add_srs_wkt (p, 11,
		 "_meridian\",-100],PARAMETER[\"false_easting\",3000000],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_northing\",3000000],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 13, "\"3080\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3081, "epsg", 3081,
		      "NAD83 / Texas State Mapping System");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=27.41666666666667 +lat_2=34.91666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=31.16666666666667 +lon_0=-100 +x_0=1000000 +y");
    add_proj4text (p, 2,
		   "_0=1000000 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Texas State Mapping System\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GR");
    add_srs_wkt (p, 2,
		 "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"standard_parallel_1\",27.41666666666667],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"standard_parallel_2\",34.91666666666666],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "latitude_of_origin\",31.16666666666667],PARAMETER[\"cent");
    add_srs_wkt (p, 11,
		 "ral_meridian\",-100],PARAMETER[\"false_easting\",1000000");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_northing\",1000000],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 13, "\",\"3081\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3082, "epsg", 3082,
		      "NAD83 / Texas Centric Lambert Conformal");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=27.5 +lat_2=35 +lat_0=18 +lon_0=-100 +x");
    add_proj4text (p, 1,
		   "_0=1500000 +y_0=5000000 +ellps=GRS80 +datum=NAD83 +units");
    add_proj4text (p, 2, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Texas Centric Lambert Conformal\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"standard_parallel_1\",27.5],PARAMETER[\"stand");
    add_srs_wkt (p, 9,
		 "ard_parallel_2\",35],PARAMETER[\"latitude_of_origin\",18");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"central_meridian\",-100],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_easting\",1500000],PARAMETER[\"false_northing\",5000000");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"3082\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 13, "\",NORTH]]");
    p = add_epsg_def (first, last, 3083, "epsg", 3083,
		      "NAD83 / Texas Centric Albers Equal Area");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=27.5 +lat_2=35 +lat_0=18 +lon_0=-100 +x");
    add_proj4text (p, 1,
		   "_0=1500000 +y_0=6000000 +ellps=GRS80 +datum=NAD83 +units");
    add_proj4text (p, 2, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Texas Centric Albers Equal Area\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Albers_Conic_Equal_Area\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"standard_parallel_1\",27.5],PARAMETER[\"standard_");
    add_srs_wkt (p, 9,
		 "parallel_2\",35],PARAMETER[\"latitude_of_center\",18],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"longitude_of_center\",-100],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",1500000],PARAMETER[\"false_northing\",6000000]");
    add_srs_wkt (p, 12,
		 ",AUTHORITY[\"EPSG\",\"3083\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3084, "epsg", 3084,
		      "NAD83(HARN) / Texas Centric Lambert Conformal");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=27.5 +lat_2=35 +lat_0=18 +lon_0=-100 +x");
    add_proj4text (p, 1,
		   "_0=1500000 +y_0=5000000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Texas Centric Lambert Conformal\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regi");
    add_srs_wkt (p, 2,
		 "onal_Network\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 3,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"61");
    add_srs_wkt (p, 4,
		 "52\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Co");
    add_srs_wkt (p, 8,
		 "nformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",2");
    add_srs_wkt (p, 9,
		 "7.5],PARAMETER[\"standard_parallel_2\",35],PARAMETER[\"l");
    add_srs_wkt (p, 10,
		 "atitude_of_origin\",18],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 11,
		 "100],PARAMETER[\"false_easting\",1500000],PARAMETER[\"fa");
    add_srs_wkt (p, 12,
		 "lse_northing\",5000000],AUTHORITY[\"EPSG\",\"3084\"],AXI");
    add_srs_wkt (p, 13, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3085, "epsg", 3085,
		      "NAD83(HARN) / Texas Centric Albers Equal Area");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=27.5 +lat_2=35 +lat_0=18 +lon_0=-100 +x");
    add_proj4text (p, 1,
		   "_0=1500000 +y_0=6000000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Texas Centric Albers Equal Area\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regi");
    add_srs_wkt (p, 2,
		 "onal_Network\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 3,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"61");
    add_srs_wkt (p, 4,
		 "52\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Albers_Con");
    add_srs_wkt (p, 8,
		 "ic_Equal_Area\"],PARAMETER[\"standard_parallel_1\",27.5]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"standard_parallel_2\",35],PARAMETER[\"latit");
    add_srs_wkt (p, 10,
		 "ude_of_center\",18],PARAMETER[\"longitude_of_center\",-1");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_easting\",1500000],PARAMETER[\"fal");
    add_srs_wkt (p, 12,
		 "se_northing\",6000000],AUTHORITY[\"EPSG\",\"3085\"],AXIS");
    add_srs_wkt (p, 13, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3086, "epsg", 3086,
		      "NAD83 / Florida GDL Albers");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=24 +lat_2=31.5 +lat_0=24 +lon_0=-84 +x_");
    add_proj4text (p, 1,
		   "0=400000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_d");
    add_proj4text (p, 2, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Florida GDL Albers\",GEOGCS[\"NAD83\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Albers_Conic_Equal_Area\"],PARAMETER[\"stand");
    add_srs_wkt (p, 8,
		 "ard_parallel_1\",24],PARAMETER[\"standard_parallel_2\",3");
    add_srs_wkt (p, 9,
		 "1.5],PARAMETER[\"latitude_of_center\",24],PARAMETER[\"lo");
    add_srs_wkt (p, 10,
		 "ngitude_of_center\",-84],PARAMETER[\"false_easting\",400");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"3086\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3087, "epsg", 3087,
		      "NAD83(HARN) / Florida GDL Albers");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=24 +lat_2=31.5 +lat_0=24 +lon_0=-84 +x_");
    add_proj4text (p, 1, "0=400000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Florida GDL Albers\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Albers_Conic_Equal_Area\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",24],PARAMETER[\"stan");
    add_srs_wkt (p, 9,
		 "dard_parallel_2\",31.5],PARAMETER[\"latitude_of_center\"");
    add_srs_wkt (p, 10,
		 ",24],PARAMETER[\"longitude_of_center\",-84],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",400000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"3087\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3088, "epsg", 3088,
		      "NAD83 / Kentucky Single Zone");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.08333333333334 +lat_2=38.66666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=36.33333333333334 +lon_0=-85.75 +x_0=1500000 ");
    add_proj4text (p, 2,
		   "+y_0=1000000 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_proj4text (p, 3, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Kentucky Single Zone\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "standard_parallel_1\",37.08333333333334],PARAMETER[\"sta");
    add_srs_wkt (p, 9,
		 "ndard_parallel_2\",38.66666666666666],PARAMETER[\"latitu");
    add_srs_wkt (p, 10,
		 "de_of_origin\",36.33333333333334],PARAMETER[\"central_me");
    add_srs_wkt (p, 11,
		 "ridian\",-85.75],PARAMETER[\"false_easting\",1500000],PA");
    add_srs_wkt (p, 12,
		 "RAMETER[\"false_northing\",1000000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 13, "3088\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3089, "epsg", 3089,
		      "NAD83 / Kentucky Single Zone (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.08333333333334 +lat_2=38.66666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=36.33333333333334 +lon_0=-85.75 +x_0=1500000 ");
    add_proj4text (p, 2,
		   "+y_0=999999.9998983998 +ellps=GRS80 +datum=NAD83 +units=");
    add_proj4text (p, 3, "us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Kentucky Single Zone (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"G");
    add_srs_wkt (p, 2,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "9\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "37.08333333333334],PARAMETER[\"standard_parallel_2\",38.");
    add_srs_wkt (p, 10,
		 "66666666666666],PARAMETER[\"latitude_of_origin\",36.3333");
    add_srs_wkt (p, 11,
		 "3333333334],PARAMETER[\"central_meridian\",-85.75],PARAM");
    add_srs_wkt (p, 12,
		 "ETER[\"false_easting\",4921250],PARAMETER[\"false_northi");
    add_srs_wkt (p, 13,
		 "ng\",3280833.333],AUTHORITY[\"EPSG\",\"3089\"],AXIS[\"X\"");
    add_srs_wkt (p, 14, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3090, "epsg", 3090,
		      "NAD83(HARN) / Kentucky Single Zone");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.08333333333334 +lat_2=38.66666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=36.33333333333334 +lon_0=-85.75 +x_0=1500000 ");
    add_proj4text (p, 2, "+y_0=1000000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Kentucky Single Zone\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Networ");
    add_srs_wkt (p, 2,
		 "k\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_C");
    add_srs_wkt (p, 8,
		 "onic_2SP\"],PARAMETER[\"standard_parallel_1\",37.0833333");
    add_srs_wkt (p, 9,
		 "3333334],PARAMETER[\"standard_parallel_2\",38.6666666666");
    add_srs_wkt (p, 10,
		 "6666],PARAMETER[\"latitude_of_origin\",36.33333333333334");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"central_meridian\",-85.75],PARAMETER[\"fal");
    add_srs_wkt (p, 12,
		 "se_easting\",1500000],PARAMETER[\"false_northing\",10000");
    add_srs_wkt (p, 13,
		 "00],AUTHORITY[\"EPSG\",\"3090\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 14, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3091, "epsg", 3091,
		      "NAD83(HARN) / Kentucky Single Zone (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.08333333333334 +lat_2=38.66666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=36.33333333333334 +lon_0=-85.75 +x_0=1500000 ");
    add_proj4text (p, 2,
		   "+y_0=999999.9998983998 +ellps=GRS80 +units=us-ft +no_def");
    add_proj4text (p, 3, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Kentucky Single Zone (ftUS)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional");
    add_srs_wkt (p, 2,
		 "_Network\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey ");
    add_srs_wkt (p, 7,
		 "foot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",37.08333333333334],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",38.66666666666666],PARAMETER[\"latitu");
    add_srs_wkt (p, 11,
		 "de_of_origin\",36.33333333333334],PARAMETER[\"central_me");
    add_srs_wkt (p, 12,
		 "ridian\",-85.75],PARAMETER[\"false_easting\",4921250],PA");
    add_srs_wkt (p, 13,
		 "RAMETER[\"false_northing\",3280833.333],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 14, "\",\"3091\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3092, "epsg", 3092, "Tokyo / UTM zone 51N");
    add_proj4text (p, 0, "+proj=utm +zone=51 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / UTM zone 51N\",GEOGCS[\"Tokyo\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6301");
    add_srs_wkt (p, 3,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4301\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",123],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 9,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3092\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3093, "epsg", 3093, "Tokyo / UTM zone 52N");
    add_proj4text (p, 0, "+proj=utm +zone=52 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / UTM zone 52N\",GEOGCS[\"Tokyo\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6301");
    add_srs_wkt (p, 3,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4301\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",129],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 9,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3093\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3094, "epsg", 3094, "Tokyo / UTM zone 53N");
    add_proj4text (p, 0, "+proj=utm +zone=53 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / UTM zone 53N\",GEOGCS[\"Tokyo\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6301");
    add_srs_wkt (p, 3,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4301\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",135],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 9,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3094\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3095, "epsg", 3095, "Tokyo / UTM zone 54N");
    add_proj4text (p, 0, "+proj=utm +zone=54 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / UTM zone 54N\",GEOGCS[\"Tokyo\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6301");
    add_srs_wkt (p, 3,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4301\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",141],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 9,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3095\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3096, "epsg", 3096, "Tokyo / UTM zone 55N");
    add_proj4text (p, 0, "+proj=utm +zone=55 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / UTM zone 55N\",GEOGCS[\"Tokyo\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6301");
    add_srs_wkt (p, 3,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4301\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",147],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 9,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3096\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3097, "epsg", 3097,
		      "JGD2000 / UTM zone 51N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=51 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / UTM zone 51N\",GEOGCS[\"JGD2000\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Japanese_Geodetic_Datum_2000\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOW");
    add_srs_wkt (p, 3,
		 "GS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",123],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 11,
		 "ing\",0],AUTHORITY[\"EPSG\",\"3097\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3098, "epsg", 3098,
		      "JGD2000 / UTM zone 52N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=52 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / UTM zone 52N\",GEOGCS[\"JGD2000\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Japanese_Geodetic_Datum_2000\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOW");
    add_srs_wkt (p, 3,
		 "GS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",129],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 11,
		 "ing\",0],AUTHORITY[\"EPSG\",\"3098\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3099, "epsg", 3099,
		      "JGD2000 / UTM zone 53N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=53 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / UTM zone 53N\",GEOGCS[\"JGD2000\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Japanese_Geodetic_Datum_2000\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOW");
    add_srs_wkt (p, 3,
		 "GS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",135],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 11,
		 "ing\",0],AUTHORITY[\"EPSG\",\"3099\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3100, "epsg", 3100,
		      "JGD2000 / UTM zone 54N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=54 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / UTM zone 54N\",GEOGCS[\"JGD2000\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Japanese_Geodetic_Datum_2000\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOW");
    add_srs_wkt (p, 3,
		 "GS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",141],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 11,
		 "ing\",0],AUTHORITY[\"EPSG\",\"3100\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3101, "epsg", 3101,
		      "JGD2000 / UTM zone 55N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=55 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JGD2000 / UTM zone 55N\",GEOGCS[\"JGD2000\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Japanese_Geodetic_Datum_2000\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOW");
    add_srs_wkt (p, 3,
		 "GS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4612\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",147],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 11,
		 "ing\",0],AUTHORITY[\"EPSG\",\"3101\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3102, "epsg", 3102,
		      "American Samoa 1962 / American Samoa Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-14.26666666666667 +lat_0=-14.266666666");
    add_proj4text (p, 1,
		   "66667 +lon_0=-170 +k_0=1 +x_0=152400.3048006096 +y_0=951");
    add_proj4text (p, 2,
		   "69.31165862332 +ellps=clrk66 +towgs84=-115,118,426,0,0,0");
    add_proj4text (p, 3, ",0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"American Samoa 1962 / American Samoa Lambert\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"American Samoa 1962\",DATUM[\"American_Samoa_19");
    add_srs_wkt (p, 2,
		 "62\",SPHEROID[\"Clarke 1866\",6378206.4,294.978698213900");
    add_srs_wkt (p, 3,
		 "6,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-115,118,426,0,0");
    add_srs_wkt (p, 4,
		 ",0,0],AUTHORITY[\"EPSG\",\"6169\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4169\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 9,
		 "onformal_Conic_1SP\"],PARAMETER[\"latitude_of_origin\",-");
    add_srs_wkt (p, 10,
		 "14.26666666666667],PARAMETER[\"central_meridian\",-170],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",500000],PARAMETER[\"false_northing\",312234.65],AUTHORI");
    add_srs_wkt (p, 13,
		 "TY[\"EPSG\",\"3102\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]");
    add_srs_wkt (p, 14, "]");
    p = add_epsg_def (first, last, 3103, "epsg", 3103,
		      "Mauritania 1999 / UTM zone 28N (deprecated)");
    add_proj4text (p, 0, "+proj=utm +zone=28 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Mauritania 1999 / UTM zone 28N (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Mauritania 1999\",DATUM[\"Mauritania_1999\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6681\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4681\"]],UNIT[\"metre\",1,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-15],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 11,
		 "ing\",0],AUTHORITY[\"EPSG\",\"3103\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3104, "epsg", 3104,
		      "Mauritania 1999 / UTM zone 29N (deprecated)");
    add_proj4text (p, 0, "+proj=utm +zone=29 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Mauritania 1999 / UTM zone 29N (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Mauritania 1999\",DATUM[\"Mauritania_1999\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6681\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4681\"]],UNIT[\"metre\",1,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-9],PARAMETER[\"scale_factor\",0.9996],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_easting\",500000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 11,
		 "ng\",0],AUTHORITY[\"EPSG\",\"3104\"],AXIS[\"Easting\",EA");
    add_srs_wkt (p, 12, "ST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3105, "epsg", 3105,
		      "Mauritania 1999 / UTM zone 30N (deprecated)");
    add_proj4text (p, 0, "+proj=utm +zone=30 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Mauritania 1999 / UTM zone 30N (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Mauritania 1999\",DATUM[\"Mauritania_1999\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6681\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4681\"]],UNIT[\"metre\",1,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",-3],PARAMETER[\"scale_factor\",0.9996],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_easting\",500000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 11,
		 "ng\",0],AUTHORITY[\"EPSG\",\"3105\"],AXIS[\"Easting\",EA");
    add_srs_wkt (p, 12, "ST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3106, "epsg", 3106,
		      "Gulshan 303 / Bangladesh Transverse Mercator");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=90 +k=0.9996 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=0 +a=6377276.345 +b=6356075.41314024 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Gulshan 303 / Bangladesh Transverse Mercator\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Gulshan 303\",DATUM[\"Gulshan_303\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7015\"]],AUTHORITY[\"EPSG\",\"6682\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4682\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",90],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 10,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"3106\"],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3107, "epsg", 3107, "GDA94 / SA Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-28 +lat_2=-36 +lat_0=-32 +lon_0=135 +x");
    add_proj4text (p, 1,
		   "_0=1000000 +y_0=2000000 +ellps=GRS80 +towgs84=0,0,0,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDA94 / SA Lambert\",GEOGCS[\"GDA94\",DATUM[\"G");
    add_srs_wkt (p, 1,
		 "eocentric_Datum_of_Australia_1994\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOW");
    add_srs_wkt (p, 3,
		 "GS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4283\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Con");
    add_srs_wkt (p, 8,
		 "ic_2SP\"],PARAMETER[\"standard_parallel_1\",-28],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"standard_parallel_2\",-36],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 10,
		 "origin\",-32],PARAMETER[\"central_meridian\",135],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_easting\",1000000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",2000000],AUTHORITY[\"EPSG\",\"3107\"],AXIS[\"Easting");
    add_srs_wkt (p, 13, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3108, "epsg", 3108,
		      "ETRS89 / Guernsey Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=49.5 +lon_0=-2.416666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "99997 +x_0=47000 +y_0=50000 +ellps=GRS80 +units=m +no_de");
    add_proj4text (p, 2, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / Guernsey Grid\",GEOGCS[\"ETRS89\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"European_Terrestrial_Reference_System_1989\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETE");
    add_srs_wkt (p, 8,
		 "R[\"latitude_of_origin\",49.5],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",-2.416666666666667],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "99997],PARAMETER[\"false_easting\",47000],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_northing\",50000],AUTHORITY[\"EPSG\",\"3108\"],AXIS[");
    add_srs_wkt (p, 12, "\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3109, "epsg", 3109,
		      "ETRS89 / Jersey Transverse Mercator");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=49.225 +lon_0=-2.135 +k=0.99999990000");
    add_proj4text (p, 1,
		   "00001 +x_0=40000 +y_0=70000 +ellps=GRS80 +units=m +no_de");
    add_proj4text (p, 2, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / Jersey Transverse Mercator\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "ETRS89\",DATUM[\"European_Terrestrial_Reference_System_1");
    add_srs_wkt (p, 2,
		 "989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6,
		 "\"]],AUTHORITY[\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",49.225],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"central_meridian\",-2.135],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9999999],PARAMETER[\"false_easting\",40000],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_northing\",70000],AUTHORITY[\"EPSG\",\"3109\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3110, "epsg", 3110, "AGD66 / Vicgrid66");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-36 +lat_2=-38 +lat_0=-37 +lon_0=145 +x");
    add_proj4text (p, 1,
		   "_0=2500000 +y_0=4500000 +ellps=aust_SA +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD66 / Vicgrid66\",GEOGCS[\"AGD66\",DATUM[\"Au");
    add_srs_wkt (p, 1,
		 "stralian_Geodetic_Datum_1966\",SPHEROID[\"Australian Nat");
    add_srs_wkt (p, 2,
		 "ional Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 3,
		 "3\"]],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4202\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"standard_parallel_1\",-36],PARAMETER[\"standard_");
    add_srs_wkt (p, 9,
		 "parallel_2\",-38],PARAMETER[\"latitude_of_origin\",-37],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"central_meridian\",145],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",2500000],PARAMETER[\"false_northing\",4500000],A");
    add_srs_wkt (p, 12,
		 "UTHORITY[\"EPSG\",\"3110\"],AXIS[\"Easting\",EAST],AXIS[");
    add_srs_wkt (p, 13, "\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3111, "epsg", 3111, "GDA94 / Vicgrid94");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-36 +lat_2=-38 +lat_0=-37 +lon_0=145 +x");
    add_proj4text (p, 1,
		   "_0=2500000 +y_0=2500000 +ellps=GRS80 +towgs84=0,0,0,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDA94 / Vicgrid94\",GEOGCS[\"GDA94\",DATUM[\"Ge");
    add_srs_wkt (p, 1,
		 "ocentric_Datum_of_Australia_1994\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOW");
    add_srs_wkt (p, 3,
		 "GS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4283\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Con");
    add_srs_wkt (p, 8,
		 "ic_2SP\"],PARAMETER[\"standard_parallel_1\",-36],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"standard_parallel_2\",-38],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 10,
		 "origin\",-37],PARAMETER[\"central_meridian\",145],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_easting\",2500000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",2500000],AUTHORITY[\"EPSG\",\"3111\"],AXIS[\"Easting");
    add_srs_wkt (p, 13, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3112, "epsg", 3112,
		      "GDA94 / Geoscience Australia Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-18 +lat_2=-36 +lat_0=0 +lon_0=134 +x_0");
    add_proj4text (p, 1,
		   "=0 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDA94 / Geoscience Australia Lambert\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "GDA94\",DATUM[\"Geocentric_Datum_of_Australia_1994\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4283\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lamb");
    add_srs_wkt (p, 8,
		 "ert_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel");
    add_srs_wkt (p, 9,
		 "_1\",-18],PARAMETER[\"standard_parallel_2\",-36],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"latitude_of_origin\",0],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 11,
		 "n\",134],PARAMETER[\"false_easting\",0],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"3112\"],AXIS[\"East");
    add_srs_wkt (p, 13, "ing\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3113, "epsg", 3113, "GDA94 / BCSG02");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-28 +lon_0=153 +k=0.99999 +x_0=50000 ");
    add_proj4text (p, 1,
		   "+y_0=100000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m");
    add_proj4text (p, 2, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDA94 / BCSG02\",GEOGCS[\"GDA94\",DATUM[\"Geoce");
    add_srs_wkt (p, 1,
		 "ntric_Datum_of_Australia_1994\",SPHEROID[\"GRS 1980\",63");
    add_srs_wkt (p, 2,
		 "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8");
    add_srs_wkt (p, 3,
		 "4[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4283\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",-28],PARAMETER[\"central_");
    add_srs_wkt (p, 9,
		 "meridian\",153],PARAMETER[\"scale_factor\",0.99999],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_easting\",50000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 11,
		 "g\",100000],AUTHORITY[\"EPSG\",\"3113\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 12, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3114, "epsg", 3114,
		      "MAGNA-SIRGAS / Colombia Far West zone");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=4.596200416666666 +lon_0=-80.07750791");
    add_proj4text (p, 1,
		   "666666 +k=1 +x_0=1000000 +y_0=1000000 +ellps=GRS80 +towg");
    add_proj4text (p, 2, "s84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MAGNA-SIRGAS / Colombia Far West zone\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"MAGNA-SIRGAS\",DATUM[\"Marco_Geocentrico_Nacional_de_R");
    add_srs_wkt (p, 2,
		 "eferencia\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "686\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",4.596200416666666],PARAMETER[\"central_merid");
    add_srs_wkt (p, 10,
		 "ian\",-80.07750791666666],PARAMETER[\"scale_factor\",1],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_easting\",1000000],PARAMETER[\"false_n");
    add_srs_wkt (p, 12,
		 "orthing\",1000000],AUTHORITY[\"EPSG\",\"3114\"],AXIS[\"N");
    add_srs_wkt (p, 13, "orthing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 3115, "epsg", 3115,
		      "MAGNA-SIRGAS / Colombia West zone");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=4.596200416666666 +lon_0=-77.07750791");
    add_proj4text (p, 1,
		   "666666 +k=1 +x_0=1000000 +y_0=1000000 +ellps=GRS80 +towg");
    add_proj4text (p, 2, "s84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MAGNA-SIRGAS / Colombia West zone\",GEOGCS[\"MA");
    add_srs_wkt (p, 1,
		 "GNA-SIRGAS\",DATUM[\"Marco_Geocentrico_Nacional_de_Refer");
    add_srs_wkt (p, 2,
		 "encia\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 5,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4686\"");
    add_srs_wkt (p, 7,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 9,
		 "igin\",4.596200416666666],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 10,
		 ",-77.07750791666666],PARAMETER[\"scale_factor\",1],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_easting\",1000000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 12,
		 "ng\",1000000],AUTHORITY[\"EPSG\",\"3115\"],AXIS[\"Northi");
    add_srs_wkt (p, 13, "ng\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 3116, "epsg", 3116,
		      "MAGNA-SIRGAS / Colombia Bogota zone");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=4.596200416666666 +lon_0=-74.07750791");
    add_proj4text (p, 1,
		   "666666 +k=1 +x_0=1000000 +y_0=1000000 +ellps=GRS80 +towg");
    add_proj4text (p, 2, "s84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MAGNA-SIRGAS / Colombia Bogota zone\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "MAGNA-SIRGAS\",DATUM[\"Marco_Geocentrico_Nacional_de_Ref");
    add_srs_wkt (p, 2,
		 "erencia\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"468");
    add_srs_wkt (p, 7,
		 "6\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",4.596200416666666],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 10,
		 "n\",-74.07750791666666],PARAMETER[\"scale_factor\",1],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_easting\",1000000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 12,
		 "thing\",1000000],AUTHORITY[\"EPSG\",\"3116\"],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 3117, "epsg", 3117,
		      "MAGNA-SIRGAS / Colombia East Central zone");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=4.596200416666666 +lon_0=-71.07750791");
    add_proj4text (p, 1,
		   "666666 +k=1 +x_0=1000000 +y_0=1000000 +ellps=GRS80 +towg");
    add_proj4text (p, 2, "s84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MAGNA-SIRGAS / Colombia East Central zone\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"MAGNA-SIRGAS\",DATUM[\"Marco_Geocentrico_Nacional_");
    add_srs_wkt (p, 2,
		 "de_Referencia\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4686\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",4.596200416666666],PARAMETER[\"central_m");
    add_srs_wkt (p, 10,
		 "eridian\",-71.07750791666666],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 11,
		 ",1],PARAMETER[\"false_easting\",1000000],PARAMETER[\"fal");
    add_srs_wkt (p, 12,
		 "se_northing\",1000000],AUTHORITY[\"EPSG\",\"3117\"],AXIS");
    add_srs_wkt (p, 13, "[\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 3118, "epsg", 3118,
		      "MAGNA-SIRGAS / Colombia East zone");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=4.596200416666666 +lon_0=-68.07750791");
    add_proj4text (p, 1,
		   "666666 +k=1 +x_0=1000000 +y_0=1000000 +ellps=GRS80 +towg");
    add_proj4text (p, 2, "s84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MAGNA-SIRGAS / Colombia East zone\",GEOGCS[\"MA");
    add_srs_wkt (p, 1,
		 "GNA-SIRGAS\",DATUM[\"Marco_Geocentrico_Nacional_de_Refer");
    add_srs_wkt (p, 2,
		 "encia\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 5,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4686\"");
    add_srs_wkt (p, 7,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 9,
		 "igin\",4.596200416666666],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 10,
		 ",-68.07750791666666],PARAMETER[\"scale_factor\",1],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_easting\",1000000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 12,
		 "ng\",1000000],AUTHORITY[\"EPSG\",\"3118\"],AXIS[\"Northi");
    add_srs_wkt (p, 13, "ng\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 3119, "epsg", 3119,
		      "Douala 1948 / AEF west");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=10.5 +k=0.999 +x_0=1000000 +");
    add_proj4text (p, 1,
		   "y_0=1000000 +ellps=intl +towgs84=-206.1,-174.7,-87.7,0,0");
    add_proj4text (p, 2, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Douala 1948 / AEF west\",GEOGCS[\"Douala 1948\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Douala_1948\",SPHEROID[\"International 1924\",6");
    add_srs_wkt (p, 2,
		 "378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-206.1,");
    add_srs_wkt (p, 3,
		 "-174.7,-87.7,0,0,0,0],AUTHORITY[\"EPSG\",\"6192\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4192\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",10.5],PARAMETER[\"scale_factor\",0.999],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",1000000],PARAMETER[\"false_no");
    add_srs_wkt (p, 11,
		 "rthing\",1000000],AUTHORITY[\"EPSG\",\"3119\"],AXIS[\"Ea");
    add_srs_wkt (p, 12, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3120, "epsg", 3120,
		      "Pulkovo 1942(58) / Poland zone I");
    add_proj4text (p, 0,
		   "+proj=sterea +lat_0=50.625 +lon_0=21.08333333333333 +k=0");
    add_proj4text (p, 1,
		   ".9998 +x_0=4637000 +y_0=5467000 +ellps=krass +units=m +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / Poland zone I\",GEOGCS[\"Pul");
    add_srs_wkt (p, 1,
		 "kovo 1942(58)\",DATUM[\"Pulkovo_1942_58\",SPHEROID[\"Kra");
    add_srs_wkt (p, 2,
		 "ssowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4179\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Oblique_Stereographic\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",50.625],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 9,
		 "\",21.08333333333333],PARAMETER[\"scale_factor\",0.9998]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_easting\",4637000],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "northing\",5467000],AUTHORITY[\"EPSG\",\"3120\"],AXIS[\"");
    add_srs_wkt (p, 12, "X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3121, "epsg", 3121,
		      "PRS92 / Philippines zone 1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=0.99995 +x_0=500000 +");
    add_proj4text (p, 1,
		   "y_0=0 +ellps=clrk66 +towgs84=-127.62,-67.24,-47.04,-3.06");
    add_proj4text (p, 2, "8,4.903,1.578,-1.06 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PRS92 / Philippines zone 1\",GEOGCS[\"PRS92\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Philippine_Reference_System_1992\",SPHEROID[\"Cla");
    add_srs_wkt (p, 2,
		 "rke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7008\"]],TOWGS84[-127.62,-67.24,-47.04,-3.068,4.903,1");
    add_srs_wkt (p, 4,
		 ".578,-1.06],AUTHORITY[\"EPSG\",\"6683\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 5,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 6,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"4683\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 8,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 10,
		 "an\",117],PARAMETER[\"scale_factor\",0.99995],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"3121\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 13, "\",NORTH]]");
    p = add_epsg_def (first, last, 3122, "epsg", 3122,
		      "PRS92 / Philippines zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=119 +k=0.99995 +x_0=500000 +");
    add_proj4text (p, 1,
		   "y_0=0 +ellps=clrk66 +towgs84=-127.62,-67.24,-47.04,-3.06");
    add_proj4text (p, 2, "8,4.903,1.578,-1.06 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PRS92 / Philippines zone 2\",GEOGCS[\"PRS92\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Philippine_Reference_System_1992\",SPHEROID[\"Cla");
    add_srs_wkt (p, 2,
		 "rke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7008\"]],TOWGS84[-127.62,-67.24,-47.04,-3.068,4.903,1");
    add_srs_wkt (p, 4,
		 ".578,-1.06],AUTHORITY[\"EPSG\",\"6683\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 5,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 6,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"4683\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 8,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 10,
		 "an\",119],PARAMETER[\"scale_factor\",0.99995],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"3122\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 13, "\",NORTH]]");
    p = add_epsg_def (first, last, 3123, "epsg", 3123,
		      "PRS92 / Philippines zone 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=121 +k=0.99995 +x_0=500000 +");
    add_proj4text (p, 1,
		   "y_0=0 +ellps=clrk66 +towgs84=-127.62,-67.24,-47.04,-3.06");
    add_proj4text (p, 2, "8,4.903,1.578,-1.06 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PRS92 / Philippines zone 3\",GEOGCS[\"PRS92\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Philippine_Reference_System_1992\",SPHEROID[\"Cla");
    add_srs_wkt (p, 2,
		 "rke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7008\"]],TOWGS84[-127.62,-67.24,-47.04,-3.068,4.903,1");
    add_srs_wkt (p, 4,
		 ".578,-1.06],AUTHORITY[\"EPSG\",\"6683\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 5,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 6,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"4683\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 8,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 10,
		 "an\",121],PARAMETER[\"scale_factor\",0.99995],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"3123\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 13, "\",NORTH]]");
    p = add_epsg_def (first, last, 3124, "epsg", 3124,
		      "PRS92 / Philippines zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=0.99995 +x_0=500000 +");
    add_proj4text (p, 1,
		   "y_0=0 +ellps=clrk66 +towgs84=-127.62,-67.24,-47.04,-3.06");
    add_proj4text (p, 2, "8,4.903,1.578,-1.06 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PRS92 / Philippines zone 4\",GEOGCS[\"PRS92\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Philippine_Reference_System_1992\",SPHEROID[\"Cla");
    add_srs_wkt (p, 2,
		 "rke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7008\"]],TOWGS84[-127.62,-67.24,-47.04,-3.068,4.903,1");
    add_srs_wkt (p, 4,
		 ".578,-1.06],AUTHORITY[\"EPSG\",\"6683\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 5,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 6,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"4683\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 8,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 10,
		 "an\",123],PARAMETER[\"scale_factor\",0.99995],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"3124\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 13, "\",NORTH]]");
    p = add_epsg_def (first, last, 3125, "epsg", 3125,
		      "PRS92 / Philippines zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=125 +k=0.99995 +x_0=500000 +");
    add_proj4text (p, 1,
		   "y_0=0 +ellps=clrk66 +towgs84=-127.62,-67.24,-47.04,-3.06");
    add_proj4text (p, 2, "8,4.903,1.578,-1.06 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PRS92 / Philippines zone 5\",GEOGCS[\"PRS92\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Philippine_Reference_System_1992\",SPHEROID[\"Cla");
    add_srs_wkt (p, 2,
		 "rke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7008\"]],TOWGS84[-127.62,-67.24,-47.04,-3.068,4.903,1");
    add_srs_wkt (p, 4,
		 ".578,-1.06],AUTHORITY[\"EPSG\",\"6683\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 5,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 6,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"4683\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 8,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 10,
		 "an\",125],PARAMETER[\"scale_factor\",0.99995],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"3125\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 13, "\",NORTH]]");
    p = add_epsg_def (first, last, 3126, "epsg", 3126, "ETRS89 / ETRS-GK19FIN");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=19 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-GK19FIN\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",19],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3126\"],AXIS[\"Northing\",NORTH],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST]]");
    p = add_epsg_def (first, last, 3127, "epsg", 3127, "ETRS89 / ETRS-GK20FIN");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=20 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-GK20FIN\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",20],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3127\"],AXIS[\"Northing\",NORTH],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST]]");
    p = add_epsg_def (first, last, 3128, "epsg", 3128, "ETRS89 / ETRS-GK21FIN");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-GK21FIN\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",21],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3128\"],AXIS[\"Northing\",NORTH],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST]]");
    p = add_epsg_def (first, last, 3129, "epsg", 3129, "ETRS89 / ETRS-GK22FIN");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=22 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-GK22FIN\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",22],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3129\"],AXIS[\"Northing\",NORTH],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST]]");
    p = add_epsg_def (first, last, 3130, "epsg", 3130, "ETRS89 / ETRS-GK23FIN");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=23 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-GK23FIN\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",23],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3130\"],AXIS[\"Northing\",NORTH],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST]]");
    p = add_epsg_def (first, last, 3131, "epsg", 3131, "ETRS89 / ETRS-GK24FIN");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=24 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-GK24FIN\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",24],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3131\"],AXIS[\"Northing\",NORTH],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST]]");
    p = add_epsg_def (first, last, 3132, "epsg", 3132, "ETRS89 / ETRS-GK25FIN");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=25 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-GK25FIN\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",25],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3132\"],AXIS[\"Northing\",NORTH],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST]]");
    p = add_epsg_def (first, last, 3133, "epsg", 3133, "ETRS89 / ETRS-GK26FIN");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=26 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-GK26FIN\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",26],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3133\"],AXIS[\"Northing\",NORTH],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST]]");
    p = add_epsg_def (first, last, 3134, "epsg", 3134, "ETRS89 / ETRS-GK27FIN");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=27 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-GK27FIN\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",27],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3134\"],AXIS[\"Northing\",NORTH],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST]]");
    p = add_epsg_def (first, last, 3135, "epsg", 3135, "ETRS89 / ETRS-GK28FIN");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=28 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-GK28FIN\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",28],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3135\"],AXIS[\"Northing\",NORTH],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST]]");
    p = add_epsg_def (first, last, 3136, "epsg", 3136, "ETRS89 / ETRS-GK29FIN");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=29 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-GK29FIN\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",29],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3136\"],AXIS[\"Northing\",NORTH],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST]]");
    p = add_epsg_def (first, last, 3137, "epsg", 3137, "ETRS89 / ETRS-GK30FIN");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=30 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-GK30FIN\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",30],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3137\"],AXIS[\"Northing\",NORTH],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST]]");
    p = add_epsg_def (first, last, 3138, "epsg", 3138, "ETRS89 / ETRS-GK31FIN");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=31 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / ETRS-GK31FIN\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",31],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3138\"],AXIS[\"Northing\",NORTH],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST]]");
    p = add_epsg_def (first, last, 3140, "epsg", 3140,
		      "Viti Levu 1912 / Viti Levu Grid");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=-18 +lon_0=178 +x_0=109435.392 +y_0=14");
    add_proj4text (p, 1,
		   "1622.272 +a=6378306.3696 +b=6356571.996 +towgs84=51,391,");
    add_proj4text (p, 2, "-36,0,0,0,0 +to_meter=0.201168 +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Viti Levu 1912 / Viti Levu Grid\",GEOGCS[\"Viti");
    add_srs_wkt (p, 1,
		 " Levu 1912\",DATUM[\"Viti_Levu_1912\",SPHEROID[\"Clarke ");
    add_srs_wkt (p, 2,
		 "1880 (international foot)\",6378306.3696,293.46630765563");
    add_srs_wkt (p, 3,
		 "49,AUTHORITY[\"EPSG\",\"7055\"]],TOWGS84[51,391,-36,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6752\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4752\"]],UNIT[\"link\",0.201168,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 8,
		 ",\"9098\"]],PROJECTION[\"Cassini_Soldner\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "latitude_of_origin\",-18],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 10,
		 ",178],PARAMETER[\"false_easting\",544000],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_northing\",704000],AUTHORITY[\"EPSG\",\"3140\"],AXIS");
    add_srs_wkt (p, 12, "[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3141, "epsg", 3141,
		      "Fiji 1956 / UTM zone 60S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=60 +south +ellps=intl +towgs84=265.025,3");
    add_proj4text (p, 1, "84.929,-194.046,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Fiji 1956 / UTM zone 60S\",GEOGCS[\"Fiji 1956\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Fiji_1956\",SPHEROID[\"International 1924\",637");
    add_srs_wkt (p, 2,
		 "8388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[265.025,3");
    add_srs_wkt (p, 3,
		 "84.929,-194.046,0,0,0,0],AUTHORITY[\"EPSG\",\"6721\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 6,
		 "2\"]],AUTHORITY[\"EPSG\",\"4721\"]],UNIT[\"metre\",1,AUT");
    add_srs_wkt (p, 7,
		 "HORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merca");
    add_srs_wkt (p, 8,
		 "tor\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"c");
    add_srs_wkt (p, 9,
		 "entral_meridian\",177],PARAMETER[\"scale_factor\",0.9996");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"false_easting\",500000],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "northing\",10000000],AUTHORITY[\"EPSG\",\"3141\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3142, "epsg", 3142,
		      "Fiji 1956 / UTM zone 1S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=1 +south +ellps=intl +towgs84=265.025,38");
    add_proj4text (p, 1, "4.929,-194.046,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Fiji 1956 / UTM zone 1S\",GEOGCS[\"Fiji 1956\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Fiji_1956\",SPHEROID[\"International 1924\",6378");
    add_srs_wkt (p, 2,
		 "388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[265.025,38");
    add_srs_wkt (p, 3,
		 "4.929,-194.046,0,0,0,0],AUTHORITY[\"EPSG\",\"6721\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6,
		 "\"]],AUTHORITY[\"EPSG\",\"4721\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"ce");
    add_srs_wkt (p, 9,
		 "ntral_meridian\",-177],PARAMETER[\"scale_factor\",0.9996");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"false_easting\",500000],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "northing\",10000000],AUTHORITY[\"EPSG\",\"3142\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3143, "epsg", 3143,
		      "Fiji 1986 / Fiji Map Grid (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-17 +lon_0=178.75 +k=0.99985 +x_0=200");
    add_proj4text (p, 1, "0000 +y_0=4000000 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Fiji 1986 / Fiji Map Grid (deprecated)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"Fiji 1986\",DATUM[\"Fiji_Geodetic_Datum_1986\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6720\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4720\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"lat");
    add_srs_wkt (p, 8,
		 "itude_of_origin\",-17],PARAMETER[\"central_meridian\",17");
    add_srs_wkt (p, 9,
		 "8.75],PARAMETER[\"scale_factor\",0.99985],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_easting\",2000000],PARAMETER[\"false_northing\",4000");
    add_srs_wkt (p, 11,
		 "000],AUTHORITY[\"EPSG\",\"3143\"],AXIS[\"Easting\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3146, "epsg", 3146,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18 +k=1 +x_0=6500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 6\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",18],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",65000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "3146\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3147, "epsg", 3147,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger CM 18E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger CM 18E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",18],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "3147\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3148, "epsg", 3148,
		      "Indian 1960 / UTM zone 48N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=48 +a=6377276.345 +b=6356075.41314024 +u");
    add_proj4text (p, 1, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Indian 1960 / UTM zone 48N\",GEOGCS[\"Indian 19");
    add_srs_wkt (p, 1,
		 "60\",DATUM[\"Indian_1960\",SPHEROID[\"Everest 1830 (1937");
    add_srs_wkt (p, 2,
		 " Adjustment)\",6377276.345,300.8017,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7015\"]],AUTHORITY[\"EPSG\",\"6131\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4131\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",105],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_easting\",500000],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"3148\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3149, "epsg", 3149,
		      "Indian 1960 / UTM zone 49N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=49 +a=6377276.345 +b=6356075.41314024 +u");
    add_proj4text (p, 1, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Indian 1960 / UTM zone 49N\",GEOGCS[\"Indian 19");
    add_srs_wkt (p, 1,
		 "60\",DATUM[\"Indian_1960\",SPHEROID[\"Everest 1830 (1937");
    add_srs_wkt (p, 2,
		 " Adjustment)\",6377276.345,300.8017,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7015\"]],AUTHORITY[\"EPSG\",\"6131\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4131\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",111],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_easting\",500000],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"3149\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3150, "epsg", 3150,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18 +k=1 +x_0=6500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 6\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",18],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",6500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"3150\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3151, "epsg", 3151,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger CM 18E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger CM 18E\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",18],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "3151\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3152, "epsg", 3152, "ST74");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18.05779 +k=0.99999425 +x_0=");
    add_proj4text (p, 1,
		   "100178.1808 +y_0=-6500614.7836 +ellps=GRS80 +towgs84=0,0");
    add_proj4text (p, 2, ",0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ST74\",GEOGCS[\"SWEREF99\",DATUM[\"SWEREF99\",S");
    add_srs_wkt (p, 1,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 2,
		 "PSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4619\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 7,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"central_meridian\",18.05779],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "cale_factor\",0.99999425],PARAMETER[\"false_easting\",10");
    add_srs_wkt (p, 10,
		 "0178.1808],PARAMETER[\"false_northing\",-6500614.7836],A");
    add_srs_wkt (p, 11,
		 "UTHORITY[\"EPSG\",\"3152\"],AXIS[\"x\",NORTH],AXIS[\"y\"");
    add_srs_wkt (p, 12, ",EAST]]");
    p = add_epsg_def (first, last, 3153, "epsg", 3153,
		      "NAD83(CSRS) / BC Albers");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=50 +lat_2=58.5 +lat_0=45 +lon_0=-126 +x");
    add_proj4text (p, 1, "_0=1000000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / BC Albers\",GEOGCS[\"NAD83(CSRS)\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Albers_Conic_Equal_Area\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"standard_parallel_1\",50],PARAMETER[\"standard");
    add_srs_wkt (p, 9,
		 "_parallel_2\",58.5],PARAMETER[\"latitude_of_center\",45]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"longitude_of_center\",-126],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_easting\",1000000],PARAMETER[\"false_northing\",0],AU");
    add_srs_wkt (p, 12,
		 "THORITY[\"EPSG\",\"3153\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3154, "epsg", 3154,
		      "NAD83(CSRS) / UTM zone 7N");
    add_proj4text (p, 0, "+proj=utm +zone=7 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / UTM zone 7N\",GEOGCS[\"NAD83(CSRS");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-141],PARAMETER[\"scale_factor\",0.9996],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",500000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 11,
		 "\",0],AUTHORITY[\"EPSG\",\"3154\"],AXIS[\"Easting\",EAST");
    add_srs_wkt (p, 12, "],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3155, "epsg", 3155,
		      "NAD83(CSRS) / UTM zone 8N");
    add_proj4text (p, 0, "+proj=utm +zone=8 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / UTM zone 8N\",GEOGCS[\"NAD83(CSRS");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-135],PARAMETER[\"scale_factor\",0.9996],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",500000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 11,
		 "\",0],AUTHORITY[\"EPSG\",\"3155\"],AXIS[\"Easting\",EAST");
    add_srs_wkt (p, 12, "],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3156, "epsg", 3156,
		      "NAD83(CSRS) / UTM zone 9N");
    add_proj4text (p, 0, "+proj=utm +zone=9 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / UTM zone 9N\",GEOGCS[\"NAD83(CSRS");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-129],PARAMETER[\"scale_factor\",0.9996],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",500000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 11,
		 "\",0],AUTHORITY[\"EPSG\",\"3156\"],AXIS[\"Easting\",EAST");
    add_srs_wkt (p, 12, "],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3157, "epsg", 3157,
		      "NAD83(CSRS) / UTM zone 10N");
    add_proj4text (p, 0, "+proj=utm +zone=10 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / UTM zone 10N\",GEOGCS[\"NAD83(CSR");
    add_srs_wkt (p, 1,
		 "S)\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-123],PARAMETER[\"scale_factor\",0.9996],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",500000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 11,
		 "\",0],AUTHORITY[\"EPSG\",\"3157\"],AXIS[\"Easting\",EAST");
    add_srs_wkt (p, 12, "],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3158, "epsg", 3158,
		      "NAD83(CSRS) / UTM zone 14N");
    add_proj4text (p, 0, "+proj=utm +zone=14 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / UTM zone 14N\",GEOGCS[\"NAD83(CSR");
    add_srs_wkt (p, 1,
		 "S)\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-99],PARAMETER[\"scale_factor\",0.9996],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"3158\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 12, "AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3159, "epsg", 3159,
		      "NAD83(CSRS) / UTM zone 15N");
    add_proj4text (p, 0, "+proj=utm +zone=15 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / UTM zone 15N\",GEOGCS[\"NAD83(CSR");
    add_srs_wkt (p, 1,
		 "S)\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-93],PARAMETER[\"scale_factor\",0.9996],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"3159\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 12, "AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3160, "epsg", 3160,
		      "NAD83(CSRS) / UTM zone 16N");
    add_proj4text (p, 0, "+proj=utm +zone=16 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / UTM zone 16N\",GEOGCS[\"NAD83(CSR");
    add_srs_wkt (p, 1,
		 "S)\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-87],PARAMETER[\"scale_factor\",0.9996],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"3160\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 12, "AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3161, "epsg", 3161,
		      "NAD83 / Ontario MNR Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.5 +lat_2=53.5 +lat_0=0 +lon_0=-85 +x");
    add_proj4text (p, 1,
		   "_0=930000 +y_0=6430000 +ellps=GRS80 +datum=NAD83 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Ontario MNR Lambert\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 8,
		 "tandard_parallel_1\",44.5],PARAMETER[\"standard_parallel");
    add_srs_wkt (p, 9,
		 "_2\",53.5],PARAMETER[\"latitude_of_origin\",0],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"central_meridian\",-85],PARAMETER[\"false_easting\",9");
    add_srs_wkt (p, 11,
		 "30000],PARAMETER[\"false_northing\",6430000],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3161\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3162, "epsg", 3162,
		      "NAD83(CSRS) / Ontario MNR Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.5 +lat_2=53.5 +lat_0=0 +lon_0=-85 +x");
    add_proj4text (p, 1,
		   "_0=930000 +y_0=6430000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / Ontario MNR Lambert\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(CSRS)\",DATUM[\"NAD83_Canadian_Spatial_Reference_Sys");
    add_srs_wkt (p, 2,
		 "tem\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6,
		 "\"]],AUTHORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",44.5],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"standard_parallel_2\",53.5],PARAMETER[\"latitu");
    add_srs_wkt (p, 10,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",-85],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_easting\",930000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 12,
		 "hing\",6430000],AUTHORITY[\"EPSG\",\"3162\"],AXIS[\"East");
    add_srs_wkt (p, 13, "ing\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3163, "epsg", 3163,
		      "RGNC91-93 / Lambert New Caledonia");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-20.66666666666667 +lat_2=-22.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-21.5 +lon_0=166 +x_0=400000 +y_0=300000 +e");
    add_proj4text (p, 2, "llps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGNC91-93 / Lambert New Caledonia\",GEOGCS[\"RG");
    add_srs_wkt (p, 1,
		 "NC91-93\",DATUM[\"Reseau_Geodesique_de_Nouvelle_Caledoni");
    add_srs_wkt (p, 2,
		 "e_91_93\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6749\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"474");
    add_srs_wkt (p, 7,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"sta");
    add_srs_wkt (p, 9,
		 "ndard_parallel_1\",-20.66666666666667],PARAMETER[\"stand");
    add_srs_wkt (p, 10,
		 "ard_parallel_2\",-22.33333333333333],PARAMETER[\"latitud");
    add_srs_wkt (p, 11,
		 "e_of_origin\",-21.5],PARAMETER[\"central_meridian\",166]");
    add_srs_wkt (p, 12,
		 ",PARAMETER[\"false_easting\",400000],PARAMETER[\"false_n");
    add_srs_wkt (p, 13,
		 "orthing\",300000],AUTHORITY[\"EPSG\",\"3163\"],AXIS[\"X\"");
    add_srs_wkt (p, 14, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3164, "epsg", 3164,
		      "ST87 Ouvea / UTM zone 58S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +south +ellps=WGS84 +towgs84=-56.263,");
    add_proj4text (p, 1, "16.136,-22.856,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ST87 Ouvea / UTM zone 58S\",GEOGCS[\"ST87 Ouvea");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"ST87_Ouvea\",SPHEROID[\"WGS 84\",6378137,298.");
    add_srs_wkt (p, 2,
		 "257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[-56.263,");
    add_srs_wkt (p, 3,
		 "16.136,-22.856,0,0,0,0],AUTHORITY[\"EPSG\",\"6750\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6,
		 "\"]],AUTHORITY[\"EPSG\",\"4750\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"ce");
    add_srs_wkt (p, 9,
		 "ntral_meridian\",165],PARAMETER[\"scale_factor\",0.9996]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_easting\",500000],PARAMETER[\"false_n");
    add_srs_wkt (p, 11,
		 "orthing\",10000000],AUTHORITY[\"EPSG\",\"3164\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3165, "epsg", 3165,
		      "NEA74 Noumea / Noumea Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-22.24469175 +lat_2=-22.29469175 +lat_0");
    add_proj4text (p, 1,
		   "=-22.26969175 +lon_0=166.44242575 +x_0=0.66 +y_0=1.02 +e");
    add_proj4text (p, 2, "llps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NEA74 Noumea / Noumea Lambert\",GEOGCS[\"NEA74 ");
    add_srs_wkt (p, 1,
		 "Noumea\",DATUM[\"NEA74_Noumea\",SPHEROID[\"International");
    add_srs_wkt (p, 2,
		 " 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6644\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4644");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"stan");
    add_srs_wkt (p, 8,
		 "dard_parallel_1\",-22.24469175],PARAMETER[\"standard_par");
    add_srs_wkt (p, 9,
		 "allel_2\",-22.29469175],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 10,
		 ",-22.26969175],PARAMETER[\"central_meridian\",166.442425");
    add_srs_wkt (p, 11,
		 "75],PARAMETER[\"false_easting\",0.66],PARAMETER[\"false_");
    add_srs_wkt (p, 12,
		 "northing\",1.02],AUTHORITY[\"EPSG\",\"3165\"],AXIS[\"X\"");
    add_srs_wkt (p, 13, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3166, "epsg", 3166,
		      "NEA74 Noumea / Noumea Lambert 2");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-22.24472222222222 +lat_2=-22.294722222");
    add_proj4text (p, 1,
		   "22222 +lat_0=-22.26972222222222 +lon_0=166.4425 +x_0=8.3");
    add_proj4text (p, 2,
		   "13000000000001 +y_0=-2.354 +ellps=intl +units=m +no_defs");
    add_proj4text (p, 3, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NEA74 Noumea / Noumea Lambert 2\",GEOGCS[\"NEA7");
    add_srs_wkt (p, 1,
		 "4 Noumea\",DATUM[\"NEA74_Noumea\",SPHEROID[\"Internation");
    add_srs_wkt (p, 2,
		 "al 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6644\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"46");
    add_srs_wkt (p, 6,
		 "44\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 7,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 8,
		 "andard_parallel_1\",-22.24472222222222],PARAMETER[\"stan");
    add_srs_wkt (p, 9,
		 "dard_parallel_2\",-22.29472222222222],PARAMETER[\"latitu");
    add_srs_wkt (p, 10,
		 "de_of_origin\",-22.26972222222222],PARAMETER[\"central_m");
    add_srs_wkt (p, 11,
		 "eridian\",166.4425],PARAMETER[\"false_easting\",8.313],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_northing\",-2.354],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 13, "3166\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3167, "epsg", 3167,
		      "Kertau (RSO) / RSO Malaya (ch)");
    add_proj4text (p, 0,
		   "+proj=omerc +lat_0=4 +lonc=102.25 +alpha=323.0257905 +k=");
    add_proj4text (p, 1,
		   "0.99984 +x_0=40000 +y_0=0 +a=6377295.664 +b=6356094.6679");
    add_proj4text (p, 2, "15204 +to_meter=20.116756 +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kertau (RSO) / RSO Malaya (ch)\",GEOGCS[\"Kerta");
    add_srs_wkt (p, 1,
		 "u (RSO)\",DATUM[\"Kertau_RSO\",SPHEROID[\"Everest 1830 (");
    add_srs_wkt (p, 2,
		 "RSO 1969)\",6377295.664,300.8017,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "56\"]],AUTHORITY[\"EPSG\",\"6751\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4751\"]],UNIT[\"British chain (Sears 1922 trunca");
    add_srs_wkt (p, 7,
		 "ted)\",20.116756,AUTHORITY[\"EPSG\",\"9301\"]],PROJECTIO");
    add_srs_wkt (p, 8,
		 "N[\"Hotine_Oblique_Mercator\"],PARAMETER[\"latitude_of_c");
    add_srs_wkt (p, 9,
		 "enter\",4],PARAMETER[\"longitude_of_center\",102.25],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"azimuth\",323.0257905],PARAMETER[\"rectified_gr");
    add_srs_wkt (p, 11,
		 "id_angle\",323.1301023611111],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 12,
		 ",0.99984],PARAMETER[\"false_easting\",1988.392164223695]");
    add_srs_wkt (p, 13,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"31");
    add_srs_wkt (p, 14,
		 "67\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3168, "epsg", 3168,
		      "Kertau (RSO) / RSO Malaya (m)");
    add_proj4text (p, 0,
		   "+proj=omerc +lat_0=4 +lonc=102.25 +alpha=323.0257905 +k=");
    add_proj4text (p, 1,
		   "0.99984 +x_0=804670.24 +y_0=0 +a=6377295.664 +b=6356094.");
    add_proj4text (p, 2, "667915204 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kertau (RSO) / RSO Malaya (m)\",GEOGCS[\"Kertau");
    add_srs_wkt (p, 1,
		 " (RSO)\",DATUM[\"Kertau_RSO\",SPHEROID[\"Everest 1830 (R");
    add_srs_wkt (p, 2,
		 "SO 1969)\",6377295.664,300.8017,AUTHORITY[\"EPSG\",\"705");
    add_srs_wkt (p, 3,
		 "6\"]],AUTHORITY[\"EPSG\",\"6751\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4751\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Hotine_Oblique_Mercator\"],PARAMETE");
    add_srs_wkt (p, 8,
		 "R[\"latitude_of_center\",4],PARAMETER[\"longitude_of_cen");
    add_srs_wkt (p, 9,
		 "ter\",102.25],PARAMETER[\"azimuth\",323.0257905],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"rectified_grid_angle\",323.1301023611111],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"scale_factor\",0.99984],PARAMETER[\"false_easting\",8");
    add_srs_wkt (p, 12,
		 "04670.24],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 13,
		 "SG\",\"3168\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",");
    add_srs_wkt (p, 14, "NORTH]]");
    p = add_epsg_def (first, last, 3169, "epsg", 3169,
		      "RGNC91-93 / UTM zone 57S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=57 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGNC91-93 / UTM zone 57S\",GEOGCS[\"RGNC91-93\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Reseau_Geodesique_de_Nouvelle_Caledonie_91_93\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"6749\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4749\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 8,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 9,
		 "0],PARAMETER[\"central_meridian\",159],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.9996],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12,
		 "\"3169\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 3170, "epsg", 3170,
		      "RGNC91-93 / UTM zone 58S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGNC91-93 / UTM zone 58S\",GEOGCS[\"RGNC91-93\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Reseau_Geodesique_de_Nouvelle_Caledonie_91_93\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"6749\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4749\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 8,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 9,
		 "0],PARAMETER[\"central_meridian\",165],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.9996],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12,
		 "\"3170\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 3171, "epsg", 3171,
		      "RGNC91-93 / UTM zone 59S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=59 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGNC91-93 / UTM zone 59S\",GEOGCS[\"RGNC91-93\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Reseau_Geodesique_de_Nouvelle_Caledonie_91_93\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"6749\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4749\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 8,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 9,
		 "0],PARAMETER[\"central_meridian\",171],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.9996],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12,
		 "\"3171\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 3172, "epsg", 3172,
		      "IGN53 Mare / UTM zone 59S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=59 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGN53 Mare / UTM zone 59S\",GEOGCS[\"IGN53 Mare");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"IGN53_Mare\",SPHEROID[\"International 1924\",");
    add_srs_wkt (p, 2,
		 "6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6641\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4641\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",171],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",0.9996],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11,
		 "\"3172\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH");
    add_srs_wkt (p, 12, "]]");
    p = add_epsg_def (first, last, 3174, "epsg", 3174,
		      "NAD83 / Great Lakes Albers");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=42.122774 +lat_2=49.01518 +lat_0=45.568");
    add_proj4text (p, 1,
		   "977 +lon_0=-84.455955 +x_0=1000000 +y_0=1000000 +ellps=G");
    add_proj4text (p, 2, "RS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Great Lakes Albers\",GEOGCS[\"NAD83\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Albers_Conic_Equal_Area\"],PARAMETER[\"stand");
    add_srs_wkt (p, 8,
		 "ard_parallel_1\",42.122774],PARAMETER[\"standard_paralle");
    add_srs_wkt (p, 9,
		 "l_2\",49.01518],PARAMETER[\"latitude_of_center\",45.5689");
    add_srs_wkt (p, 10,
		 "77],PARAMETER[\"longitude_of_center\",-84.455955],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_easting\",1000000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",1000000],AUTHORITY[\"EPSG\",\"3174\"],AXIS[\"X\",EAS");
    add_srs_wkt (p, 13, "T],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3175, "epsg", 3175,
		      "NAD83 / Great Lakes and St Lawrence Albers");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=42.122774 +lat_2=49.01518 +lat_0=45.568");
    add_proj4text (p, 1,
		   "977 +lon_0=-83.248627 +x_0=1000000 +y_0=1000000 +ellps=G");
    add_proj4text (p, 2, "RS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Great Lakes and St Lawrence Albers\",GE");
    add_srs_wkt (p, 1,
		 "OGCS[\"NAD83\",DATUM[\"North_American_Datum_1983\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Albers_Conic_Equal_Area\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"standard_parallel_1\",42.122774],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_2\",49.01518],PARAMETER[\"latitude_of_c");
    add_srs_wkt (p, 10,
		 "enter\",45.568977],PARAMETER[\"longitude_of_center\",-83");
    add_srs_wkt (p, 11,
		 ".248627],PARAMETER[\"false_easting\",1000000],PARAMETER[");
    add_srs_wkt (p, 12,
		 "\"false_northing\",1000000],AUTHORITY[\"EPSG\",\"3175\"]");
    add_srs_wkt (p, 13, ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3176, "epsg", 3176,
		      "Indian 1960 / TM 106 NE");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=106 +k=0.9996 +x_0=500000 +y");
    add_proj4text (p, 1,
		   "_0=0 +a=6377276.345 +b=6356075.41314024 +units=m +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Indian 1960 / TM 106 NE\",GEOGCS[\"Indian 1960\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Indian_1960\",SPHEROID[\"Everest 1830 (1937 Adj");
    add_srs_wkt (p, 2,
		 "ustment)\",6377276.345,300.8017,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "5\"]],AUTHORITY[\"EPSG\",\"6131\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4131\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "06],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 11,
		 "RITY[\"EPSG\",\"3176\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 12, "rthing\",NORTH]]");
    p = add_epsg_def (first, last, 3177, "epsg", 3177, "LGD2006 / Libya TM");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=17 +k=0.9965000000000001 +x_");
    add_proj4text (p, 1,
		   "0=1000000 +y_0=0 +ellps=intl +towgs84=-208.406,-109.878,");
    add_proj4text (p, 2, "-2.5764,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"LGD2006 / Libya TM\",GEOGCS[\"LGD2006\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Libyan_Geodetic_Datum_2006\",SPHEROID[\"International 19");
    add_srs_wkt (p, 2,
		 "24\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-");
    add_srs_wkt (p, 3,
		 "208.406,-109.878,-2.5764,0,0,0,0],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 4,
		 "754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 5,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 6,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4754\"]],UNIT[\"metr");
    add_srs_wkt (p, 7,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transve");
    add_srs_wkt (p, 8,
		 "rse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"central_meridian\",17],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9965],PARAMETER[\"false_easting\",1000000],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_northing\",0],AUTHORITY[\"EPSG\",\"3177\"],AXIS[");
    add_srs_wkt (p, 12, "\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3178, "epsg", 3178, "GR96 / UTM zone 18N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GR96 / UTM zone 18N\",GEOGCS[\"GR96\",DATUM[\"G");
    add_srs_wkt (p, 1,
		 "reenland_1996\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 2,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4747\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-75],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"3178\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 12, "g\",NORTH]]");
    p = add_epsg_def (first, last, 3179, "epsg", 3179, "GR96 / UTM zone 19N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GR96 / UTM zone 19N\",GEOGCS[\"GR96\",DATUM[\"G");
    add_srs_wkt (p, 1,
		 "reenland_1996\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 2,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4747\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-69],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"3179\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 12, "g\",NORTH]]");
    p = add_epsg_def (first, last, 3180, "epsg", 3180, "GR96 / UTM zone 20N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GR96 / UTM zone 20N\",GEOGCS[\"GR96\",DATUM[\"G");
    add_srs_wkt (p, 1,
		 "reenland_1996\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 2,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4747\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-63],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"3180\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 12, "g\",NORTH]]");
    p = add_epsg_def (first, last, 3181, "epsg", 3181, "GR96 / UTM zone 21N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GR96 / UTM zone 21N\",GEOGCS[\"GR96\",DATUM[\"G");
    add_srs_wkt (p, 1,
		 "reenland_1996\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 2,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4747\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-57],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"3181\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 12, "g\",NORTH]]");
    p = add_epsg_def (first, last, 3182, "epsg", 3182, "GR96 / UTM zone 22N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=22 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GR96 / UTM zone 22N\",GEOGCS[\"GR96\",DATUM[\"G");
    add_srs_wkt (p, 1,
		 "reenland_1996\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 2,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4747\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-51],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"3182\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 12, "g\",NORTH]]");
    p = add_epsg_def (first, last, 3183, "epsg", 3183, "GR96 / UTM zone 23N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=23 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GR96 / UTM zone 23N\",GEOGCS[\"GR96\",DATUM[\"G");
    add_srs_wkt (p, 1,
		 "reenland_1996\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 2,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4747\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-45],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"3183\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 12, "g\",NORTH]]");
    p = add_epsg_def (first, last, 3184, "epsg", 3184, "GR96 / UTM zone 24N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=24 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GR96 / UTM zone 24N\",GEOGCS[\"GR96\",DATUM[\"G");
    add_srs_wkt (p, 1,
		 "reenland_1996\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 2,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4747\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-39],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"3184\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 12, "g\",NORTH]]");
    p = add_epsg_def (first, last, 3185, "epsg", 3185, "GR96 / UTM zone 25N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=25 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GR96 / UTM zone 25N\",GEOGCS[\"GR96\",DATUM[\"G");
    add_srs_wkt (p, 1,
		 "reenland_1996\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 2,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4747\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-33],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"3185\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 12, "g\",NORTH]]");
    p = add_epsg_def (first, last, 3186, "epsg", 3186, "GR96 / UTM zone 26N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=26 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GR96 / UTM zone 26N\",GEOGCS[\"GR96\",DATUM[\"G");
    add_srs_wkt (p, 1,
		 "reenland_1996\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 2,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4747\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-27],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"3186\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 12, "g\",NORTH]]");
    p = add_epsg_def (first, last, 3187, "epsg", 3187, "GR96 / UTM zone 27N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=27 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GR96 / UTM zone 27N\",GEOGCS[\"GR96\",DATUM[\"G");
    add_srs_wkt (p, 1,
		 "reenland_1996\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 2,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4747\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-21],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"3187\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 12, "g\",NORTH]]");
    p = add_epsg_def (first, last, 3188, "epsg", 3188, "GR96 / UTM zone 28N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=28 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GR96 / UTM zone 28N\",GEOGCS[\"GR96\",DATUM[\"G");
    add_srs_wkt (p, 1,
		 "reenland_1996\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 2,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4747\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-15],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"3188\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 12, "g\",NORTH]]");
    p = add_epsg_def (first, last, 3189, "epsg", 3189, "GR96 / UTM zone 29N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=29 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GR96 / UTM zone 29N\",GEOGCS[\"GR96\",DATUM[\"G");
    add_srs_wkt (p, 1,
		 "reenland_1996\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 2,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4747\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-9],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"3189\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 3190, "epsg", 3190,
		      "LGD2006 / Libya TM zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=9 +k=0.99995 +x_0=200000 +y_");
    add_proj4text (p, 1,
		   "0=0 +ellps=intl +towgs84=-208.406,-109.878,-2.5764,0,0,0");
    add_proj4text (p, 2, ",0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"LGD2006 / Libya TM zone 5\",GEOGCS[\"LGD2006\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Libyan_Geodetic_Datum_2006\",SPHEROID[\"Internat");
    add_srs_wkt (p, 2,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[-208.406,-109.878,-2.5764,0,0,0,0],AUTHORITY[\"EP");
    add_srs_wkt (p, 4,
		 "SG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4754\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 8,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 9,
		 "0],PARAMETER[\"central_meridian\",9],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",0.99995],PARAMETER[\"false_easting\",200000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3190\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3191, "epsg", 3191,
		      "LGD2006 / Libya TM zone 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=11 +k=0.99995 +x_0=200000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=intl +towgs84=-208.406,-109.878,-2.5764,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"LGD2006 / Libya TM zone 6\",GEOGCS[\"LGD2006\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Libyan_Geodetic_Datum_2006\",SPHEROID[\"Internat");
    add_srs_wkt (p, 2,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[-208.406,-109.878,-2.5764,0,0,0,0],AUTHORITY[\"EP");
    add_srs_wkt (p, 4,
		 "SG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4754\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 8,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 9,
		 "0],PARAMETER[\"central_meridian\",11],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.99995],PARAMETER[\"false_easting\",200000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3191\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3192, "epsg", 3192,
		      "LGD2006 / Libya TM zone 7");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=13 +k=0.99995 +x_0=200000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=intl +towgs84=-208.406,-109.878,-2.5764,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"LGD2006 / Libya TM zone 7\",GEOGCS[\"LGD2006\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Libyan_Geodetic_Datum_2006\",SPHEROID[\"Internat");
    add_srs_wkt (p, 2,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[-208.406,-109.878,-2.5764,0,0,0,0],AUTHORITY[\"EP");
    add_srs_wkt (p, 4,
		 "SG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4754\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 8,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 9,
		 "0],PARAMETER[\"central_meridian\",13],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.99995],PARAMETER[\"false_easting\",200000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3192\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_09 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 3193, "epsg", 3193,
		      "LGD2006 / Libya TM zone 8");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=0.99995 +x_0=200000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=intl +towgs84=-208.406,-109.878,-2.5764,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"LGD2006 / Libya TM zone 8\",GEOGCS[\"LGD2006\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Libyan_Geodetic_Datum_2006\",SPHEROID[\"Internat");
    add_srs_wkt (p, 2,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[-208.406,-109.878,-2.5764,0,0,0,0],AUTHORITY[\"EP");
    add_srs_wkt (p, 4,
		 "SG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4754\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 8,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 9,
		 "0],PARAMETER[\"central_meridian\",15],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.99995],PARAMETER[\"false_easting\",200000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3193\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3194, "epsg", 3194,
		      "LGD2006 / Libya TM zone 9");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=17 +k=0.99995 +x_0=200000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=intl +towgs84=-208.406,-109.878,-2.5764,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"LGD2006 / Libya TM zone 9\",GEOGCS[\"LGD2006\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Libyan_Geodetic_Datum_2006\",SPHEROID[\"Internat");
    add_srs_wkt (p, 2,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[-208.406,-109.878,-2.5764,0,0,0,0],AUTHORITY[\"EP");
    add_srs_wkt (p, 4,
		 "SG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4754\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 8,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 9,
		 "0],PARAMETER[\"central_meridian\",17],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.99995],PARAMETER[\"false_easting\",200000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3194\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3195, "epsg", 3195,
		      "LGD2006 / Libya TM zone 10");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=19 +k=0.99995 +x_0=200000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=intl +towgs84=-208.406,-109.878,-2.5764,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"LGD2006 / Libya TM zone 10\",GEOGCS[\"LGD2006\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Libyan_Geodetic_Datum_2006\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "TOWGS84[-208.406,-109.878,-2.5764,0,0,0,0],AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 5,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4754\"]],UN");
    add_srs_wkt (p, 7,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 9,
		 ",0],PARAMETER[\"central_meridian\",19],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.99995],PARAMETER[\"false_easting\",200000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3195");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3196, "epsg", 3196,
		      "LGD2006 / Libya TM zone 11");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=0.99995 +x_0=200000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=intl +towgs84=-208.406,-109.878,-2.5764,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"LGD2006 / Libya TM zone 11\",GEOGCS[\"LGD2006\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Libyan_Geodetic_Datum_2006\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "TOWGS84[-208.406,-109.878,-2.5764,0,0,0,0],AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 5,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4754\"]],UN");
    add_srs_wkt (p, 7,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 9,
		 ",0],PARAMETER[\"central_meridian\",21],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.99995],PARAMETER[\"false_easting\",200000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3196");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3197, "epsg", 3197,
		      "LGD2006 / Libya TM zone 12");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=23 +k=0.99995 +x_0=200000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=intl +towgs84=-208.406,-109.878,-2.5764,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"LGD2006 / Libya TM zone 12\",GEOGCS[\"LGD2006\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Libyan_Geodetic_Datum_2006\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "TOWGS84[-208.406,-109.878,-2.5764,0,0,0,0],AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 5,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4754\"]],UN");
    add_srs_wkt (p, 7,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 9,
		 ",0],PARAMETER[\"central_meridian\",23],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.99995],PARAMETER[\"false_easting\",200000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3197");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3198, "epsg", 3198,
		      "LGD2006 / Libya TM zone 13");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=25 +k=0.99995 +x_0=200000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=intl +towgs84=-208.406,-109.878,-2.5764,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"LGD2006 / Libya TM zone 13\",GEOGCS[\"LGD2006\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Libyan_Geodetic_Datum_2006\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "TOWGS84[-208.406,-109.878,-2.5764,0,0,0,0],AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 5,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4754\"]],UN");
    add_srs_wkt (p, 7,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 9,
		 ",0],PARAMETER[\"central_meridian\",25],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.99995],PARAMETER[\"false_easting\",200000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3198");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3199, "epsg", 3199,
		      "LGD2006 / UTM zone 32N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=32 +ellps=intl +towgs84=-208.406,-109.87");
    add_proj4text (p, 1, "8,-2.5764,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"LGD2006 / UTM zone 32N\",GEOGCS[\"LGD2006\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Libyan_Geodetic_Datum_2006\",SPHEROID[\"Internation");
    add_srs_wkt (p, 2,
		 "al 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWG");
    add_srs_wkt (p, 3,
		 "S84[-208.406,-109.878,-2.5764,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4754\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"central_meridian\",9],PARAMETER[\"scale_facto");
    add_srs_wkt (p, 10,
		 "r\",0.9996],PARAMETER[\"false_easting\",500000],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3199\"],AXI");
    add_srs_wkt (p, 12, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3200, "epsg", 3200, "FD58 / Iraq zone");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=32.5 +lat_0=32.5 +lon_0=45 +k_0=0.99878");
    add_proj4text (p, 1,
		   "64078000001 +x_0=1500000 +y_0=1166200 +ellps=clrk80 +uni");
    add_proj4text (p, 2, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"FD58 / Iraq zone\",GEOGCS[\"FD58\",DATUM[\"Fina");
    add_srs_wkt (p, 1,
		 "l_Datum_1958\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.14");
    add_srs_wkt (p, 2,
		 "5,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"6132\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4132\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"L");
    add_srs_wkt (p, 7,
		 "ambert_Conformal_Conic_1SP\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 8,
		 "igin\",32.5],PARAMETER[\"central_meridian\",45],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"scale_factor\",0.9987864078],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",1500000],PARAMETER[\"false_northing\",1166200],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"3200\"],AXIS[\"Easting\",EAST],AXIS[\"N");
    add_srs_wkt (p, 12, "orthing\",NORTH]]");
    p = add_epsg_def (first, last, 3201, "epsg", 3201,
		      "LGD2006 / UTM zone 33N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=33 +ellps=intl +towgs84=-208.406,-109.87");
    add_proj4text (p, 1, "8,-2.5764,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"LGD2006 / UTM zone 33N\",GEOGCS[\"LGD2006\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Libyan_Geodetic_Datum_2006\",SPHEROID[\"Internation");
    add_srs_wkt (p, 2,
		 "al 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWG");
    add_srs_wkt (p, 3,
		 "S84[-208.406,-109.878,-2.5764,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4754\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"central_meridian\",15],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 10,
		 "or\",0.9996],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3201\"],AX");
    add_srs_wkt (p, 12, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3202, "epsg", 3202,
		      "LGD2006 / UTM zone 34N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=34 +ellps=intl +towgs84=-208.406,-109.87");
    add_proj4text (p, 1, "8,-2.5764,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"LGD2006 / UTM zone 34N\",GEOGCS[\"LGD2006\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Libyan_Geodetic_Datum_2006\",SPHEROID[\"Internation");
    add_srs_wkt (p, 2,
		 "al 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWG");
    add_srs_wkt (p, 3,
		 "S84[-208.406,-109.878,-2.5764,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4754\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"central_meridian\",21],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 10,
		 "or\",0.9996],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3202\"],AX");
    add_srs_wkt (p, 12, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3203, "epsg", 3203,
		      "LGD2006 / UTM zone 35N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=35 +ellps=intl +towgs84=-208.406,-109.87");
    add_proj4text (p, 1, "8,-2.5764,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"LGD2006 / UTM zone 35N\",GEOGCS[\"LGD2006\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Libyan_Geodetic_Datum_2006\",SPHEROID[\"Internation");
    add_srs_wkt (p, 2,
		 "al 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWG");
    add_srs_wkt (p, 3,
		 "S84[-208.406,-109.878,-2.5764,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4754\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"central_meridian\",27],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 10,
		 "or\",0.9996],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3203\"],AX");
    add_srs_wkt (p, 12, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3204, "epsg", 3204,
		      "WGS 84 / SCAR IMW SP19-20");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-60.66666666666666 +lat_2=-63.333333333");
    add_proj4text (p, 1,
		   "33334 +lat_0=-90 +lon_0=-66 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SP19-20\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-60.66666666666666],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "63.33333333333334],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-66],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3204\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3205, "epsg", 3205,
		      "WGS 84 / SCAR IMW SP21-22");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-60.66666666666666 +lat_2=-63.333333333");
    add_proj4text (p, 1,
		   "33334 +lat_0=-90 +lon_0=-54 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SP21-22\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-60.66666666666666],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "63.33333333333334],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-54],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3205\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3206, "epsg", 3206,
		      "WGS 84 / SCAR IMW SP23-24");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-60.66666666666666 +lat_2=-63.333333333");
    add_proj4text (p, 1,
		   "33334 +lat_0=-90 +lon_0=-42 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SP23-24\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-60.66666666666666],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "63.33333333333334],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-42],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3206\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3207, "epsg", 3207,
		      "WGS 84 / SCAR IMW SQ01-02");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-64.66666666666667 +lat_2=-67.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-174 +x_0=0 +y_0=0 +ellps=WGS84 ");
    add_proj4text (p, 2, "+datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SQ01-02\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-64.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "67.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-174],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3207\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3208, "epsg", 3208,
		      "WGS 84 / SCAR IMW SQ19-20");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-64.66666666666667 +lat_2=-67.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-66 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SQ19-20\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-64.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "67.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-66],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3208\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3209, "epsg", 3209,
		      "WGS 84 / SCAR IMW SQ21-22");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-64.66666666666667 +lat_2=-67.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-54 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SQ21-22\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-64.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "67.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-54],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3209\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3210, "epsg", 3210,
		      "WGS 84 / SCAR IMW SQ37-38");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-64.66666666666667 +lat_2=-67.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=42 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SQ37-38\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-64.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "67.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",42],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3210\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3211, "epsg", 3211,
		      "WGS 84 / SCAR IMW SQ39-40");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-64.66666666666667 +lat_2=-67.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=54 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SQ39-40\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-64.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "67.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",54],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3211\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3212, "epsg", 3212,
		      "WGS 84 / SCAR IMW SQ41-42");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-64.66666666666667 +lat_2=-67.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=66 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SQ41-42\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-64.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "67.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",66],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3212\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3213, "epsg", 3213,
		      "WGS 84 / SCAR IMW SQ43-44");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-64.66666666666667 +lat_2=-67.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=78 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SQ43-44\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-64.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "67.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",78],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3213\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3214, "epsg", 3214,
		      "WGS 84 / SCAR IMW SQ45-46");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-64.66666666666667 +lat_2=-67.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=90 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SQ45-46\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-64.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "67.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",90],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3214\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3215, "epsg", 3215,
		      "WGS 84 / SCAR IMW SQ47-48");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-64.66666666666667 +lat_2=-67.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=102 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SQ47-48\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-64.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "67.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",102],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3215\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3216, "epsg", 3216,
		      "WGS 84 / SCAR IMW SQ49-50");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-64.66666666666667 +lat_2=-67.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=114 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SQ49-50\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-64.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "67.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",114],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3216\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3217, "epsg", 3217,
		      "WGS 84 / SCAR IMW SQ51-52");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-64.66666666666667 +lat_2=-67.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=126 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SQ51-52\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-64.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "67.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",126],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3217\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3218, "epsg", 3218,
		      "WGS 84 / SCAR IMW SQ53-54");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-64.66666666666667 +lat_2=-67.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=138 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SQ53-54\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-64.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "67.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",138],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3218\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3219, "epsg", 3219,
		      "WGS 84 / SCAR IMW SQ55-56");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-64.66666666666667 +lat_2=-67.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=150 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SQ55-56\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-64.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "67.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",150],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3219\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3220, "epsg", 3220,
		      "WGS 84 / SCAR IMW SQ57-58");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-64.66666666666667 +lat_2=-67.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=162 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SQ57-58\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-64.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "67.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",162],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3220\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3221, "epsg", 3221,
		      "WGS 84 / SCAR IMW SR13-14");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-102 +x_0=0 +y_0=0 +ellps=WGS84 ");
    add_proj4text (p, 2, "+datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR13-14\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-102],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3221\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3222, "epsg", 3222,
		      "WGS 84 / SCAR IMW SR15-16");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-90 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR15-16\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-90],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3222\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3223, "epsg", 3223,
		      "WGS 84 / SCAR IMW SR17-18");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-78 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR17-18\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-78],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3223\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3224, "epsg", 3224,
		      "WGS 84 / SCAR IMW SR19-20");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-66 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR19-20\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-66],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3224\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3225, "epsg", 3225,
		      "WGS 84 / SCAR IMW SR27-28");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-18 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR27-28\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-18],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3225\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3226, "epsg", 3226,
		      "WGS 84 / SCAR IMW SR29-30");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-6 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR29-30\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-6],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3226\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3227, "epsg", 3227,
		      "WGS 84 / SCAR IMW SR31-32");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=6 +x_0=0 +y_0=0 +ellps=WGS84 +da");
    add_proj4text (p, 2, "tum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR31-32\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",6],PARAMETER[\"false_eas");
    add_srs_wkt (p, 11,
		 "ting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 12,
		 "SG\",\"3227\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",");
    add_srs_wkt (p, 13, "NORTH]]");
    p = add_epsg_def (first, last, 3228, "epsg", 3228,
		      "WGS 84 / SCAR IMW SR33-34");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=18 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR33-34\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",18],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3228\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3229, "epsg", 3229,
		      "WGS 84 / SCAR IMW SR35-36");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=30 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR35-36\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",30],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3229\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3230, "epsg", 3230,
		      "WGS 84 / SCAR IMW SR37-38");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=42 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR37-38\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",42],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3230\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3231, "epsg", 3231,
		      "WGS 84 / SCAR IMW SR39-40");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=54 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR39-40\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",54],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3231\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3232, "epsg", 3232,
		      "WGS 84 / SCAR IMW SR41-42");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=66 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR41-42\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",66],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3232\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3233, "epsg", 3233,
		      "WGS 84 / SCAR IMW SR43-44");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=78 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR43-44\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",78],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3233\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3234, "epsg", 3234,
		      "WGS 84 / SCAR IMW SR45-46");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=90 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR45-46\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",90],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3234\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3235, "epsg", 3235,
		      "WGS 84 / SCAR IMW SR47-48");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=102 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR47-48\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",102],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3235\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3236, "epsg", 3236,
		      "WGS 84 / SCAR IMW SR49-50");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=114 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR49-50\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",114],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3236\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3237, "epsg", 3237,
		      "WGS 84 / SCAR IMW SR51-52");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=126 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR51-52\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",126],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3237\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3238, "epsg", 3238,
		      "WGS 84 / SCAR IMW SR53-54");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=138 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR53-54\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",138],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3238\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3239, "epsg", 3239,
		      "WGS 84 / SCAR IMW SR55-56");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=150 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR55-56\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",150],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3239\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3240, "epsg", 3240,
		      "WGS 84 / SCAR IMW SR57-58");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=162 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR57-58\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",162],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3240\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3241, "epsg", 3241,
		      "WGS 84 / SCAR IMW SR59-60");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-68.66666666666667 +lat_2=-71.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=174 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SR59-60\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-68.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "71.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",174],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3241\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3242, "epsg", 3242,
		      "WGS 84 / SCAR IMW SS04-06");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-72.66666666666667 +lat_2=-75.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-153 +x_0=0 +y_0=0 +ellps=WGS84 ");
    add_proj4text (p, 2, "+datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SS04-06\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-72.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "75.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-153],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3242\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3243, "epsg", 3243,
		      "WGS 84 / SCAR IMW SS07-09");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-72.66666666666667 +lat_2=-75.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-135 +x_0=0 +y_0=0 +ellps=WGS84 ");
    add_proj4text (p, 2, "+datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SS07-09\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-72.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "75.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-135],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3243\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3244, "epsg", 3244,
		      "WGS 84 / SCAR IMW SS10-12");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-72.66666666666667 +lat_2=-75.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-117 +x_0=0 +y_0=0 +ellps=WGS84 ");
    add_proj4text (p, 2, "+datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SS10-12\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-72.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "75.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-117],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3244\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3245, "epsg", 3245,
		      "WGS 84 / SCAR IMW SS13-15");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-72.66666666666667 +lat_2=-75.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-99 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SS13-15\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-72.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "75.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-99],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3245\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3246, "epsg", 3246,
		      "WGS 84 / SCAR IMW SS16-18");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-72.66666666666667 +lat_2=-75.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-81 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SS16-18\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-72.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "75.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-81],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3246\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3247, "epsg", 3247,
		      "WGS 84 / SCAR IMW SS19-21");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-72.66666666666667 +lat_2=-75.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-63 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SS19-21\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-72.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "75.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-63],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3247\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3248, "epsg", 3248,
		      "WGS 84 / SCAR IMW SS25-27");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-72.66666666666667 +lat_2=-75.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-27 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SS25-27\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-72.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "75.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-27],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3248\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3249, "epsg", 3249,
		      "WGS 84 / SCAR IMW SS28-30");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-72.66666666666667 +lat_2=-75.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-9 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SS28-30\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-72.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "75.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-9],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3249\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3250, "epsg", 3250,
		      "WGS 84 / SCAR IMW SS31-33");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-72.66666666666667 +lat_2=-75.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=9 +x_0=0 +y_0=0 +ellps=WGS84 +da");
    add_proj4text (p, 2, "tum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SS31-33\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-72.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "75.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",9],PARAMETER[\"false_eas");
    add_srs_wkt (p, 11,
		 "ting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 12,
		 "SG\",\"3250\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",");
    add_srs_wkt (p, 13, "NORTH]]");
    p = add_epsg_def (first, last, 3251, "epsg", 3251,
		      "WGS 84 / SCAR IMW SS34-36");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-72.66666666666667 +lat_2=-75.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=27 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SS34-36\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-72.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "75.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",27],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3251\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3252, "epsg", 3252,
		      "WGS 84 / SCAR IMW SS37-39");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-72.66666666666667 +lat_2=-75.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=45 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SS37-39\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-72.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "75.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",45],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3252\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3253, "epsg", 3253,
		      "WGS 84 / SCAR IMW SS40-42");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-72.66666666666667 +lat_2=-75.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=63 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SS40-42\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-72.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "75.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",63],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3253\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3254, "epsg", 3254,
		      "WGS 84 / SCAR IMW SS43-45");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-72.66666666666667 +lat_2=-75.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=81 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SS43-45\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-72.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "75.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",81],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3254\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3255, "epsg", 3255,
		      "WGS 84 / SCAR IMW SS46-48");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-72.66666666666667 +lat_2=-75.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=99 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SS46-48\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-72.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "75.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",99],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3255\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3256, "epsg", 3256,
		      "WGS 84 / SCAR IMW SS49-51");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-72.66666666666667 +lat_2=-75.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=117 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SS49-51\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-72.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "75.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",117],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3256\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3257, "epsg", 3257,
		      "WGS 84 / SCAR IMW SS52-54");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-72.66666666666667 +lat_2=-75.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=135 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SS52-54\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-72.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "75.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",135],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3257\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3258, "epsg", 3258,
		      "WGS 84 / SCAR IMW SS55-57");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-72.66666666666667 +lat_2=-75.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=153 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SS55-57\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-72.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "75.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",153],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3258\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3259, "epsg", 3259,
		      "WGS 84 / SCAR IMW SS58-60");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-72.66666666666667 +lat_2=-75.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=171 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SS58-60\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-72.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "75.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",171],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3259\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3260, "epsg", 3260,
		      "WGS 84 / SCAR IMW ST01-04");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-76.66666666666667 +lat_2=-79.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-168 +x_0=0 +y_0=0 +ellps=WGS84 ");
    add_proj4text (p, 2, "+datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW ST01-04\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-76.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "79.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-168],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3260\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3261, "epsg", 3261,
		      "WGS 84 / SCAR IMW ST05-08");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-76.66666666666667 +lat_2=-79.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-144 +x_0=0 +y_0=0 +ellps=WGS84 ");
    add_proj4text (p, 2, "+datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW ST05-08\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-76.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "79.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-144],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3261\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3262, "epsg", 3262,
		      "WGS 84 / SCAR IMW ST09-12");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-76.66666666666667 +lat_2=-79.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-120 +x_0=0 +y_0=0 +ellps=WGS84 ");
    add_proj4text (p, 2, "+datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW ST09-12\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-76.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "79.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-120],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3262\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3263, "epsg", 3263,
		      "WGS 84 / SCAR IMW ST13-16");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-76.66666666666667 +lat_2=-79.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-96 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW ST13-16\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-76.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "79.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-96],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3263\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3264, "epsg", 3264,
		      "WGS 84 / SCAR IMW ST17-20");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-76.66666666666667 +lat_2=-79.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-72 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW ST17-20\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-76.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "79.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-72],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3264\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3265, "epsg", 3265,
		      "WGS 84 / SCAR IMW ST21-24");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-76.66666666666667 +lat_2=-79.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-48 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW ST21-24\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-76.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "79.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-48],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3265\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3266, "epsg", 3266,
		      "WGS 84 / SCAR IMW ST25-28");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-76.66666666666667 +lat_2=-79.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=-24 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW ST25-28\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-76.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "79.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-24],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3266\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3267, "epsg", 3267,
		      "WGS 84 / SCAR IMW ST29-32");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-76.66666666666667 +lat_2=-79.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=0 +x_0=0 +y_0=0 +ellps=WGS84 +da");
    add_proj4text (p, 2, "tum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW ST29-32\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-76.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "79.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",0],PARAMETER[\"false_eas");
    add_srs_wkt (p, 11,
		 "ting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 12,
		 "SG\",\"3267\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",");
    add_srs_wkt (p, 13, "NORTH]]");
    p = add_epsg_def (first, last, 3268, "epsg", 3268,
		      "WGS 84 / SCAR IMW ST33-36");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-76.66666666666667 +lat_2=-79.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=24 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW ST33-36\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-76.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "79.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",24],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3268\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3269, "epsg", 3269,
		      "WGS 84 / SCAR IMW ST37-40");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-76.66666666666667 +lat_2=-79.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=48 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW ST37-40\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-76.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "79.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",48],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3269\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3270, "epsg", 3270,
		      "WGS 84 / SCAR IMW ST41-44");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-76.66666666666667 +lat_2=-79.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=72 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW ST41-44\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-76.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "79.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",72],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3270\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3271, "epsg", 3271,
		      "WGS 84 / SCAR IMW ST45-48");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-76.66666666666667 +lat_2=-79.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=96 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 2, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW ST45-48\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-76.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "79.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",96],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3271\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3272, "epsg", 3272,
		      "WGS 84 / SCAR IMW ST49-52");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-76.66666666666667 +lat_2=-79.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=120 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW ST49-52\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-76.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "79.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",120],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3272\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3273, "epsg", 3273,
		      "WGS 84 / SCAR IMW ST53-56");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-76.66666666666667 +lat_2=-79.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=144 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW ST53-56\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-76.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "79.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",144],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3273\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3274, "epsg", 3274,
		      "WGS 84 / SCAR IMW ST57-60");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-76.66666666666667 +lat_2=-79.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=168 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW ST57-60\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert");
    add_srs_wkt (p, 7,
		 "_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 8,
		 ",-76.66666666666667],PARAMETER[\"standard_parallel_2\",-");
    add_srs_wkt (p, 9,
		 "79.33333333333333],PARAMETER[\"latitude_of_origin\",-90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",168],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12,
		 "EPSG\",\"3274\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3275, "epsg", 3275,
		      "WGS 84 / SCAR IMW SU01-05");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-80.23861111111111 +lon_0");
    add_proj4text (p, 1,
		   "=-165 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +unit");
    add_proj4text (p, 2, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SU01-05\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_S");
    add_srs_wkt (p, 7,
		 "tereographic\"],PARAMETER[\"latitude_of_origin\",-80.238");
    add_srs_wkt (p, 8,
		 "61111111111],PARAMETER[\"central_meridian\",-165],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",0],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3275");
    add_srs_wkt (p, 11,
		 "\"],AXIS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 3276, "epsg", 3276,
		      "WGS 84 / SCAR IMW SU06-10");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-80.23861111111111 +lon_0");
    add_proj4text (p, 1,
		   "=-135 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +unit");
    add_proj4text (p, 2, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SU06-10\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_S");
    add_srs_wkt (p, 7,
		 "tereographic\"],PARAMETER[\"latitude_of_origin\",-80.238");
    add_srs_wkt (p, 8,
		 "61111111111],PARAMETER[\"central_meridian\",-135],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",0],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3276");
    add_srs_wkt (p, 11,
		 "\"],AXIS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 3277, "epsg", 3277,
		      "WGS 84 / SCAR IMW SU11-15");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-80.23861111111111 +lon_0");
    add_proj4text (p, 1,
		   "=-105 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +unit");
    add_proj4text (p, 2, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SU11-15\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_S");
    add_srs_wkt (p, 7,
		 "tereographic\"],PARAMETER[\"latitude_of_origin\",-80.238");
    add_srs_wkt (p, 8,
		 "61111111111],PARAMETER[\"central_meridian\",-105],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",0],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3277");
    add_srs_wkt (p, 11,
		 "\"],AXIS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 3278, "epsg", 3278,
		      "WGS 84 / SCAR IMW SU16-20");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-80.23861111111111 +lon_0");
    add_proj4text (p, 1,
		   "=-75 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units");
    add_proj4text (p, 2, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SU16-20\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_S");
    add_srs_wkt (p, 7,
		 "tereographic\"],PARAMETER[\"latitude_of_origin\",-80.238");
    add_srs_wkt (p, 8,
		 "61111111111],PARAMETER[\"central_meridian\",-75],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",0],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3278\"");
    add_srs_wkt (p, 11,
		 "],AXIS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3279, "epsg", 3279,
		      "WGS 84 / SCAR IMW SU21-25");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-80.23861111111111 +lon_0");
    add_proj4text (p, 1,
		   "=-45 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units");
    add_proj4text (p, 2, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SU21-25\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_S");
    add_srs_wkt (p, 7,
		 "tereographic\"],PARAMETER[\"latitude_of_origin\",-80.238");
    add_srs_wkt (p, 8,
		 "61111111111],PARAMETER[\"central_meridian\",-45],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",0],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3279\"");
    add_srs_wkt (p, 11,
		 "],AXIS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3280, "epsg", 3280,
		      "WGS 84 / SCAR IMW SU26-30");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-80.23861111111111 +lon_0");
    add_proj4text (p, 1,
		   "=-15 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units");
    add_proj4text (p, 2, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SU26-30\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_S");
    add_srs_wkt (p, 7,
		 "tereographic\"],PARAMETER[\"latitude_of_origin\",-80.238");
    add_srs_wkt (p, 8,
		 "61111111111],PARAMETER[\"central_meridian\",-15],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",0],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3280\"");
    add_srs_wkt (p, 11,
		 "],AXIS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3281, "epsg", 3281,
		      "WGS 84 / SCAR IMW SU31-35");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-80.23861111111111 +lon_0");
    add_proj4text (p, 1,
		   "=15 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SU31-35\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_S");
    add_srs_wkt (p, 7,
		 "tereographic\"],PARAMETER[\"latitude_of_origin\",-80.238");
    add_srs_wkt (p, 8,
		 "61111111111],PARAMETER[\"central_meridian\",15],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",0],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3281\"");
    add_srs_wkt (p, 11,
		 "],AXIS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3282, "epsg", 3282,
		      "WGS 84 / SCAR IMW SU36-40");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-80.23861111111111 +lon_0");
    add_proj4text (p, 1,
		   "=45 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SU36-40\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_S");
    add_srs_wkt (p, 7,
		 "tereographic\"],PARAMETER[\"latitude_of_origin\",-80.238");
    add_srs_wkt (p, 8,
		 "61111111111],PARAMETER[\"central_meridian\",45],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",0],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3282\"");
    add_srs_wkt (p, 11,
		 "],AXIS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3283, "epsg", 3283,
		      "WGS 84 / SCAR IMW SU41-45");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-80.23861111111111 +lon_0");
    add_proj4text (p, 1,
		   "=75 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SU41-45\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_S");
    add_srs_wkt (p, 7,
		 "tereographic\"],PARAMETER[\"latitude_of_origin\",-80.238");
    add_srs_wkt (p, 8,
		 "61111111111],PARAMETER[\"central_meridian\",75],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",0],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3283\"");
    add_srs_wkt (p, 11,
		 "],AXIS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3284, "epsg", 3284,
		      "WGS 84 / SCAR IMW SU46-50");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-80.23861111111111 +lon_0");
    add_proj4text (p, 1,
		   "=105 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units");
    add_proj4text (p, 2, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SU46-50\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_S");
    add_srs_wkt (p, 7,
		 "tereographic\"],PARAMETER[\"latitude_of_origin\",-80.238");
    add_srs_wkt (p, 8,
		 "61111111111],PARAMETER[\"central_meridian\",105],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",0],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3284\"");
    add_srs_wkt (p, 11,
		 "],AXIS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3285, "epsg", 3285,
		      "WGS 84 / SCAR IMW SU51-55");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-80.23861111111111 +lon_0");
    add_proj4text (p, 1,
		   "=135 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units");
    add_proj4text (p, 2, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SU51-55\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_S");
    add_srs_wkt (p, 7,
		 "tereographic\"],PARAMETER[\"latitude_of_origin\",-80.238");
    add_srs_wkt (p, 8,
		 "61111111111],PARAMETER[\"central_meridian\",135],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",0],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3285\"");
    add_srs_wkt (p, 11,
		 "],AXIS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3286, "epsg", 3286,
		      "WGS 84 / SCAR IMW SU56-60");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-80.23861111111111 +lon_0");
    add_proj4text (p, 1,
		   "=165 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units");
    add_proj4text (p, 2, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SU56-60\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_S");
    add_srs_wkt (p, 7,
		 "tereographic\"],PARAMETER[\"latitude_of_origin\",-80.238");
    add_srs_wkt (p, 8,
		 "61111111111],PARAMETER[\"central_meridian\",165],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",0],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3286\"");
    add_srs_wkt (p, 11,
		 "],AXIS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3287, "epsg", 3287,
		      "WGS 84 / SCAR IMW SV01-10");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-80.23861111111111 +lon_0");
    add_proj4text (p, 1,
		   "=-150 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +unit");
    add_proj4text (p, 2, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SV01-10\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_S");
    add_srs_wkt (p, 7,
		 "tereographic\"],PARAMETER[\"latitude_of_origin\",-80.238");
    add_srs_wkt (p, 8,
		 "61111111111],PARAMETER[\"central_meridian\",-150],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",0],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3287");
    add_srs_wkt (p, 11,
		 "\"],AXIS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 3288, "epsg", 3288,
		      "WGS 84 / SCAR IMW SV11-20");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-80.23861111111111 +lon_0");
    add_proj4text (p, 1,
		   "=-90 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units");
    add_proj4text (p, 2, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SV11-20\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_S");
    add_srs_wkt (p, 7,
		 "tereographic\"],PARAMETER[\"latitude_of_origin\",-80.238");
    add_srs_wkt (p, 8,
		 "61111111111],PARAMETER[\"central_meridian\",-90],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",0],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3288\"");
    add_srs_wkt (p, 11,
		 "],AXIS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3289, "epsg", 3289,
		      "WGS 84 / SCAR IMW SV21-30");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-80.23861111111111 +lon_0");
    add_proj4text (p, 1,
		   "=-30 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units");
    add_proj4text (p, 2, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SV21-30\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_S");
    add_srs_wkt (p, 7,
		 "tereographic\"],PARAMETER[\"latitude_of_origin\",-80.238");
    add_srs_wkt (p, 8,
		 "61111111111],PARAMETER[\"central_meridian\",-30],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",0],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3289\"");
    add_srs_wkt (p, 11,
		 "],AXIS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3290, "epsg", 3290,
		      "WGS 84 / SCAR IMW SV31-40");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-80.23861111111111 +lon_0");
    add_proj4text (p, 1,
		   "=30 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SV31-40\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_S");
    add_srs_wkt (p, 7,
		 "tereographic\"],PARAMETER[\"latitude_of_origin\",-80.238");
    add_srs_wkt (p, 8,
		 "61111111111],PARAMETER[\"central_meridian\",30],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",0],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3290\"");
    add_srs_wkt (p, 11,
		 "],AXIS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3291, "epsg", 3291,
		      "WGS 84 / SCAR IMW SV41-50");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-80.23861111111111 +lon_0");
    add_proj4text (p, 1,
		   "=90 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SV41-50\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_S");
    add_srs_wkt (p, 7,
		 "tereographic\"],PARAMETER[\"latitude_of_origin\",-80.238");
    add_srs_wkt (p, 8,
		 "61111111111],PARAMETER[\"central_meridian\",90],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",0],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3291\"");
    add_srs_wkt (p, 11,
		 "],AXIS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3292, "epsg", 3292,
		      "WGS 84 / SCAR IMW SV51-60");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-80.23861111111111 +lon_0");
    add_proj4text (p, 1,
		   "=150 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units");
    add_proj4text (p, 2, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SV51-60\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_S");
    add_srs_wkt (p, 7,
		 "tereographic\"],PARAMETER[\"latitude_of_origin\",-80.238");
    add_srs_wkt (p, 8,
		 "61111111111],PARAMETER[\"central_meridian\",150],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",0],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3292\"");
    add_srs_wkt (p, 11,
		 "],AXIS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3293, "epsg", 3293,
		      "WGS 84 / SCAR IMW SW01-60");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-80.23861111111111 +lon_0");
    add_proj4text (p, 1,
		   "=0 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=m");
    add_proj4text (p, 2, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / SCAR IMW SW01-60\",GEOGCS[\"WGS 84\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223");
    add_srs_wkt (p, 2,
		 "563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_S");
    add_srs_wkt (p, 7,
		 "tereographic\"],PARAMETER[\"latitude_of_origin\",-80.238");
    add_srs_wkt (p, 8,
		 "61111111111],PARAMETER[\"central_meridian\",0],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",1],PARAMETER[\"false_easting\",0],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3293\"]");
    add_srs_wkt (p, 11,
		 ",AXIS[\"Easting\",UNKNOWN],AXIS[\"Northing\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3294, "epsg", 3294,
		      "WGS 84 / USGS Transantarctic Mountains");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-76.66666666666667 +lat_2=-79.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-78 +lon_0=162 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 2, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / USGS Transantarctic Mountains\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"WGS 84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",63781");
    add_srs_wkt (p, 2,
		 "37,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standar");
    add_srs_wkt (p, 8,
		 "d_parallel_1\",-76.66666666666667],PARAMETER[\"standard_");
    add_srs_wkt (p, 9,
		 "parallel_2\",-79.33333333333333],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 10,
		 "_origin\",-78],PARAMETER[\"central_meridian\",162],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_easting\",0],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"3294\"],AXIS[\"Easting\",EAST],AX");
    add_srs_wkt (p, 13, "IS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3296, "epsg", 3296, "RGPF / UTM zone 5S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=5 +south +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGPF / UTM zone 5S\",GEOGCS[\"RGPF\",DATUM[\"Re");
    add_srs_wkt (p, 1,
		 "seau_Geodesique_de_la_Polynesie_Francaise\",SPHEROID[\"G");
    add_srs_wkt (p, 2,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "9\"]],AUTHORITY[\"EPSG\",\"6687\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4687\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 9,
		 "153],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",500000],PARAMETER[\"false_northing\",1000000");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"3296\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3297, "epsg", 3297, "RGPF / UTM zone 6S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=6 +south +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGPF / UTM zone 6S\",GEOGCS[\"RGPF\",DATUM[\"Re");
    add_srs_wkt (p, 1,
		 "seau_Geodesique_de_la_Polynesie_Francaise\",SPHEROID[\"G");
    add_srs_wkt (p, 2,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "9\"]],AUTHORITY[\"EPSG\",\"6687\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4687\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 9,
		 "147],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",500000],PARAMETER[\"false_northing\",1000000");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"3297\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3298, "epsg", 3298, "RGPF / UTM zone 7S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=7 +south +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGPF / UTM zone 7S\",GEOGCS[\"RGPF\",DATUM[\"Re");
    add_srs_wkt (p, 1,
		 "seau_Geodesique_de_la_Polynesie_Francaise\",SPHEROID[\"G");
    add_srs_wkt (p, 2,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "9\"]],AUTHORITY[\"EPSG\",\"6687\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4687\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 9,
		 "141],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",500000],PARAMETER[\"false_northing\",1000000");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"3298\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3299, "epsg", 3299, "RGPF / UTM zone 8S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=8 +south +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGPF / UTM zone 8S\",GEOGCS[\"RGPF\",DATUM[\"Re");
    add_srs_wkt (p, 1,
		 "seau_Geodesique_de_la_Polynesie_Francaise\",SPHEROID[\"G");
    add_srs_wkt (p, 2,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "9\"]],AUTHORITY[\"EPSG\",\"6687\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4687\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 9,
		 "135],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",500000],PARAMETER[\"false_northing\",1000000");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"3299\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3300, "epsg", 3300,
		      "Estonian Coordinate System of 1992");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=59.33333333333334 +lat_2=58 +lat_0=57.5");
    add_proj4text (p, 1,
		   "1755393055556 +lon_0=24 +x_0=500000 +y_0=6375000 +ellps=");
    add_proj4text (p, 2,
		   "GRS80 +towgs84=0.055,-0.541,-0.185,0.0183,-0.0003,-0.007");
    add_proj4text (p, 3, ",-0.014 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Estonian Coordinate System of 1992\",GEOGCS[\"E");
    add_srs_wkt (p, 1,
		 "ST92\",DATUM[\"Estonia_1992\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[");
    add_srs_wkt (p, 3,
		 "0.055,-0.541,-0.185,0.0183,-0.0003,-0.007,-0.014],AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"6133\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4133");
    add_srs_wkt (p, 7,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 8,
		 "ECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"stan");
    add_srs_wkt (p, 9,
		 "dard_parallel_1\",59.33333333333334],PARAMETER[\"standar");
    add_srs_wkt (p, 10,
		 "d_parallel_2\",58],PARAMETER[\"latitude_of_origin\",57.5");
    add_srs_wkt (p, 11,
		 "1755393055556],PARAMETER[\"central_meridian\",24],PARAME");
    add_srs_wkt (p, 12,
		 "TER[\"false_easting\",500000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 13,
		 "\",6375000],AUTHORITY[\"EPSG\",\"3300\"],AXIS[\"X\",NORT");
    add_srs_wkt (p, 14, "H],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3301, "epsg", 3301,
		      "Estonian Coordinate System of 1997");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=59.33333333333334 +lat_2=58 +lat_0=57.5");
    add_proj4text (p, 1,
		   "1755393055556 +lon_0=24 +x_0=500000 +y_0=6375000 +ellps=");
    add_proj4text (p, 2, "GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Estonian Coordinate System of 1997\",GEOGCS[\"E");
    add_srs_wkt (p, 1,
		 "ST97\",DATUM[\"Estonia_1997\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6180\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4180\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"standard_parallel_1\",59.33333333333334]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"standard_parallel_2\",58],PARAMETER[\"latit");
    add_srs_wkt (p, 10,
		 "ude_of_origin\",57.51755393055556],PARAMETER[\"central_m");
    add_srs_wkt (p, 11,
		 "eridian\",24],PARAMETER[\"false_easting\",500000],PARAME");
    add_srs_wkt (p, 12,
		 "TER[\"false_northing\",6375000],AUTHORITY[\"EPSG\",\"330");
    add_srs_wkt (p, 13, "1\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3302, "epsg", 3302,
		      "IGN63 Hiva Oa / UTM zone 7S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=7 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGN63 Hiva Oa / UTM zone 7S\",GEOGCS[\"IGN63 Hi");
    add_srs_wkt (p, 1,
		 "va Oa\",DATUM[\"IGN63_Hiva_Oa\",SPHEROID[\"International");
    add_srs_wkt (p, 2,
		 " 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6689\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4689");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",-141],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 10,
		 "500000],PARAMETER[\"false_northing\",10000000],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3302\"],AXIS[\"Easting\",EAST],AXIS[\"Northi");
    add_srs_wkt (p, 12, "ng\",NORTH]]");
    p = add_epsg_def (first, last, 3303, "epsg", 3303,
		      "Fatu Iva 72 / UTM zone 7S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=7 +south +ellps=intl +towgs84=347.103,10");
    add_proj4text (p, 1,
		   "78.12,2623.92,-33.8875,70.6773,-9.3943,186.074 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Fatu Iva 72 / UTM zone 7S\",GEOGCS[\"Fatu Iva 7");
    add_srs_wkt (p, 1,
		 "2\",DATUM[\"Fatu_Iva_72\",SPHEROID[\"International 1924\"");
    add_srs_wkt (p, 2,
		 ",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[347.1");
    add_srs_wkt (p, 3,
		 "03,1078.12,2623.92,-33.8875,70.6773,-9.3943,186.074],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6688\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "688\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",-141],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 11,
		 "\",500000],PARAMETER[\"false_northing\",10000000],AUTHOR");
    add_srs_wkt (p, 12,
		 "ITY[\"EPSG\",\"3303\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 3304, "epsg", 3304,
		      "Tahiti 79 / UTM zone 6S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=6 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tahiti 79 / UTM zone 6S\",GEOGCS[\"Tahiti 79\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Tahiti_79\",SPHEROID[\"International 1924\",6378");
    add_srs_wkt (p, 2,
		 "388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6690\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4690\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",-147],PARAMETER[\"scale_fa");
    add_srs_wkt (p, 9,
		 "ctor\",0.9996],PARAMETER[\"false_easting\",500000],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"3");
    add_srs_wkt (p, 11,
		 "304\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3305, "epsg", 3305,
		      "Moorea 87 / UTM zone 6S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=6 +south +ellps=intl +towgs84=215.525,14");
    add_proj4text (p, 1,
		   "9.593,176.229,-3.2624,-1.692,-1.1571,10.4773 +units=m +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Moorea 87 / UTM zone 6S\",GEOGCS[\"Moorea 87\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Moorea_87\",SPHEROID[\"International 1924\",6378");
    add_srs_wkt (p, 2,
		 "388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[215.525,14");
    add_srs_wkt (p, 3,
		 "9.593,176.229,-3.2624,-1.692,-1.1571,10.4773],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6691\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4691\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",-147],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",10000000],AUTHORITY[\"EP");
    add_srs_wkt (p, 12,
		 "SG\",\"3305\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",");
    add_srs_wkt (p, 13, "NORTH]]");
    p = add_epsg_def (first, last, 3306, "epsg", 3306,
		      "Maupiti 83 / UTM zone 5S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=5 +south +ellps=intl +towgs84=217.037,86");
    add_proj4text (p, 1, ".959,23.956,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Maupiti 83 / UTM zone 5S\",GEOGCS[\"Maupiti 83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Maupiti_83\",SPHEROID[\"International 1924\",63");
    add_srs_wkt (p, 2,
		 "78388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[217.037,");
    add_srs_wkt (p, 3,
		 "86.959,23.956,0,0,0,0],AUTHORITY[\"EPSG\",\"6692\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4692\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",-153],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 11,
		 "rthing\",10000000],AUTHORITY[\"EPSG\",\"3306\"],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3307, "epsg", 3307,
		      "Nakhl-e Ghanem / UTM zone 39N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=39 +ellps=WGS84 +towgs84=0,-0.15,0.68,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Nakhl-e Ghanem / UTM zone 39N\",GEOGCS[\"Nakhl-");
    add_srs_wkt (p, 1,
		 "e Ghanem\",DATUM[\"Nakhl_e_Ghanem\",SPHEROID[\"WGS 84\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWG");
    add_srs_wkt (p, 3,
		 "S84[0,-0.15,0.68,0,0,0,0],AUTHORITY[\"EPSG\",\"6693\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4693\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",51],PARAMETER[\"scale_factor\",0.9996");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"false_easting\",500000],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "northing\",0],AUTHORITY[\"EPSG\",\"3307\"],AXIS[\"Eastin");
    add_srs_wkt (p, 12, "g\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3308, "epsg", 3308, "GDA94 / NSW Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-30.75 +lat_2=-35.75 +lat_0=-33.25 +lon");
    add_proj4text (p, 1,
		   "_0=147 +x_0=9300000 +y_0=4500000 +ellps=GRS80 +towgs84=0");
    add_proj4text (p, 2, ",0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDA94 / NSW Lambert\",GEOGCS[\"GDA94\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Geocentric_Datum_of_Australia_1994\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4283\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_C");
    add_srs_wkt (p, 8,
		 "onic_2SP\"],PARAMETER[\"standard_parallel_1\",-30.75],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"standard_parallel_2\",-35.75],PARAMETER[\"lati");
    add_srs_wkt (p, 10,
		 "tude_of_origin\",-33.25],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 11,
		 "147],PARAMETER[\"false_easting\",9300000],PARAMETER[\"fa");
    add_srs_wkt (p, 12,
		 "lse_northing\",4500000],AUTHORITY[\"EPSG\",\"3308\"],AXI");
    add_srs_wkt (p, 13, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3309, "epsg", 3309,
		      "NAD27 / California Albers");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=34 +lat_2=40.5 +lat_0=0 +lon_0=-120 +x_");
    add_proj4text (p, 1,
		   "0=0 +y_0=-4000000 +ellps=clrk66 +datum=NAD27 +units=m +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / California Albers\",GEOGCS[\"NAD27\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866");
    add_srs_wkt (p, 2,
		 "\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Albers_Conic_Equal_Area\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"standard_parallel_1\",34],PARAMETER[\"standard_parall");
    add_srs_wkt (p, 9,
		 "el_2\",40.5],PARAMETER[\"latitude_of_center\",0],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"longitude_of_center\",-120],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",0],PARAMETER[\"false_northing\",-4000000],AUTHORITY");
    add_srs_wkt (p, 12,
		 "[\"EPSG\",\"3309\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3310, "epsg", 3310,
		      "NAD83 / California Albers");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=34 +lat_2=40.5 +lat_0=0 +lon_0=-120 +x_");
    add_proj4text (p, 1,
		   "0=0 +y_0=-4000000 +ellps=GRS80 +datum=NAD83 +units=m +no");
    add_proj4text (p, 2, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / California Albers\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 7,
		 "OJECTION[\"Albers_Conic_Equal_Area\"],PARAMETER[\"standa");
    add_srs_wkt (p, 8,
		 "rd_parallel_1\",34],PARAMETER[\"standard_parallel_2\",40");
    add_srs_wkt (p, 9,
		 ".5],PARAMETER[\"latitude_of_center\",0],PARAMETER[\"long");
    add_srs_wkt (p, 10,
		 "itude_of_center\",-120],PARAMETER[\"false_easting\",0],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",-4000000],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"3310\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3311, "epsg", 3311,
		      "NAD83(HARN) / California Albers");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=34 +lat_2=40.5 +lat_0=0 +lon_0=-120 +x_");
    add_proj4text (p, 1, "0=0 +y_0=-4000000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / California Albers\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Albers_Conic_Equal_Area\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",34],PARAMETER[\"stan");
    add_srs_wkt (p, 9,
		 "dard_parallel_2\",40.5],PARAMETER[\"latitude_of_center\"");
    add_srs_wkt (p, 10,
		 ",0],PARAMETER[\"longitude_of_center\",-120],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",0],PARAMETER[\"false_northing\",-4000000");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"3311\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 13, "\",NORTH]]");
    p = add_epsg_def (first, last, 3312, "epsg", 3312, "CSG67 / UTM zone 21N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +ellps=intl +towgs84=-186,230,110,0,0");
    add_proj4text (p, 1, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"CSG67 / UTM zone 21N\",GEOGCS[\"CSG67\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Centre_Spatial_Guyanais_1967\",SPHEROID[\"International ");
    add_srs_wkt (p, 2,
		 "1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84");
    add_srs_wkt (p, 3,
		 "[-186,230,110,0,0,0,0],AUTHORITY[\"EPSG\",\"6623\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4623\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",-57],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",0],AUTHORITY[\"EPSG\",\"3312\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 12, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3313, "epsg", 3313, "RGFG95 / UTM zone 21N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +ellps=GRS80 +towgs84=2,2,-2,0,0,0,0 ");
    add_proj4text (p, 1, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGFG95 / UTM zone 21N\",GEOGCS[\"RGFG95\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Reseau_Geodesique_Francais_Guyane_1995\",SPHEROID[\"G");
    add_srs_wkt (p, 2,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "9\"]],TOWGS84[2,2,-2,0,0,0,0],AUTHORITY[\"EPSG\",\"6624\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4624\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-57],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"3313\"],AXIS[\"Eas");
    add_srs_wkt (p, 12, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3314, "epsg", 3314,
		      "Katanga 1955 / Katanga Lambert (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-6.5 +lat_2=-11.5 +lat_0=0 +lon_0=26 +x");
    add_proj4text (p, 1,
		   "_0=0 +y_0=0 +ellps=clrk66 +towgs84=-103.746,-9.614,-255.");
    add_proj4text (p, 2, "95,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Katanga 1955 / Katanga Lambert (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Katanga 1955\",DATUM[\"Katanga_1955\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7008\"]],TOWGS84[-103.746,-9.614,-255.95,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6695\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4695\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",-6.5],PARAMETER[\"standard_para");
    add_srs_wkt (p, 10,
		 "llel_2\",-11.5],PARAMETER[\"latitude_of_origin\",0],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",26],PARAMETER[\"false_easting");
    add_srs_wkt (p, 12,
		 "\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 13,
		 ",\"3314\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORT");
    add_srs_wkt (p, 14, "H]]");
    p = add_epsg_def (first, last, 3315, "epsg", 3315,
		      "Katanga 1955 / Katanga TM (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-9 +lon_0=26 +k=0.9998 +x_0=0 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=clrk66 +towgs84=-103.746,-9.614,-255.95,0,0,0,0 +");
    add_proj4text (p, 2, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Katanga 1955 / Katanga TM (deprecated)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"Katanga 1955\",DATUM[\"Katanga_1955\",SPHEROID[\"Clar");
    add_srs_wkt (p, 2,
		 "ke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7008\"]],TOWGS84[-103.746,-9.614,-255.95,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6695\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "695\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",-9],PARAMETER[\"central_meridian\",26],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.9998],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12,
		 "3315\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 3316, "epsg", 3316,
		      "Kasai 1953 / Congo TM zone 22");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=22 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=10000000 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kasai 1953 / Congo TM zone 22\",GEOGCS[\"Kasai ");
    add_srs_wkt (p, 1,
		 "1953\",DATUM[\"Kasai_1953\",SPHEROID[\"Clarke 1880 (RGS)");
    add_srs_wkt (p, 2,
		 "\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6696\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "696\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 8,
		 "of_origin\",0],PARAMETER[\"central_meridian\",22],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"scale_factor\",0.9999],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",500000],PARAMETER[\"false_northing\",10000000],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"3316\"],AXIS[\"Easting\",EAST],AXIS[\"North");
    add_srs_wkt (p, 12, "ing\",NORTH]]");
    p = add_epsg_def (first, last, 3317, "epsg", 3317,
		      "Kasai 1953 / Congo TM zone 24");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=24 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=10000000 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kasai 1953 / Congo TM zone 24\",GEOGCS[\"Kasai ");
    add_srs_wkt (p, 1,
		 "1953\",DATUM[\"Kasai_1953\",SPHEROID[\"Clarke 1880 (RGS)");
    add_srs_wkt (p, 2,
		 "\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6696\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "696\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 8,
		 "of_origin\",0],PARAMETER[\"central_meridian\",24],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"scale_factor\",0.9999],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",500000],PARAMETER[\"false_northing\",10000000],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"3317\"],AXIS[\"Easting\",EAST],AXIS[\"North");
    add_srs_wkt (p, 12, "ing\",NORTH]]");
    p = add_epsg_def (first, last, 3318, "epsg", 3318,
		      "IGC 1962 / Congo TM zone 12");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=12 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=10000000 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGC 1962 / Congo TM zone 12\",GEOGCS[\"IGC 1962");
    add_srs_wkt (p, 1,
		 " 6th Parallel South\",DATUM[\"IGC_1962_Arc_of_the_6th_Pa");
    add_srs_wkt (p, 2,
		 "rallel_South\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.14");
    add_srs_wkt (p, 3,
		 "5,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4697\"]],UNIT[");
    add_srs_wkt (p, 7,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",12],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",0.9999],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12,
		 "3318\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 3319, "epsg", 3319,
		      "IGC 1962 / Congo TM zone 14");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=14 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=10000000 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGC 1962 / Congo TM zone 14\",GEOGCS[\"IGC 1962");
    add_srs_wkt (p, 1,
		 " 6th Parallel South\",DATUM[\"IGC_1962_Arc_of_the_6th_Pa");
    add_srs_wkt (p, 2,
		 "rallel_South\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.14");
    add_srs_wkt (p, 3,
		 "5,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4697\"]],UNIT[");
    add_srs_wkt (p, 7,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",14],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",0.9999],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12,
		 "3319\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 3320, "epsg", 3320,
		      "IGC 1962 / Congo TM zone 16");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=16 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=10000000 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGC 1962 / Congo TM zone 16\",GEOGCS[\"IGC 1962");
    add_srs_wkt (p, 1,
		 " 6th Parallel South\",DATUM[\"IGC_1962_Arc_of_the_6th_Pa");
    add_srs_wkt (p, 2,
		 "rallel_South\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.14");
    add_srs_wkt (p, 3,
		 "5,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4697\"]],UNIT[");
    add_srs_wkt (p, 7,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",16],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",0.9999],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12,
		 "3320\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 3321, "epsg", 3321,
		      "IGC 1962 / Congo TM zone 18");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=10000000 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGC 1962 / Congo TM zone 18\",GEOGCS[\"IGC 1962");
    add_srs_wkt (p, 1,
		 " 6th Parallel South\",DATUM[\"IGC_1962_Arc_of_the_6th_Pa");
    add_srs_wkt (p, 2,
		 "rallel_South\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.14");
    add_srs_wkt (p, 3,
		 "5,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4697\"]],UNIT[");
    add_srs_wkt (p, 7,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",18],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",0.9999],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12,
		 "3321\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 13, "");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_10 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 3322, "epsg", 3322,
		      "IGC 1962 / Congo TM zone 20");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=20 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=10000000 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGC 1962 / Congo TM zone 20\",GEOGCS[\"IGC 1962");
    add_srs_wkt (p, 1,
		 " 6th Parallel South\",DATUM[\"IGC_1962_Arc_of_the_6th_Pa");
    add_srs_wkt (p, 2,
		 "rallel_South\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.14");
    add_srs_wkt (p, 3,
		 "5,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4697\"]],UNIT[");
    add_srs_wkt (p, 7,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",20],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",0.9999],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12,
		 "3322\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 3323, "epsg", 3323,
		      "IGC 1962 / Congo TM zone 22");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=22 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=10000000 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGC 1962 / Congo TM zone 22\",GEOGCS[\"IGC 1962");
    add_srs_wkt (p, 1,
		 " 6th Parallel South\",DATUM[\"IGC_1962_Arc_of_the_6th_Pa");
    add_srs_wkt (p, 2,
		 "rallel_South\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.14");
    add_srs_wkt (p, 3,
		 "5,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4697\"]],UNIT[");
    add_srs_wkt (p, 7,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",22],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",0.9999],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12,
		 "3323\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 3324, "epsg", 3324,
		      "IGC 1962 / Congo TM zone 24");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=24 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=10000000 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGC 1962 / Congo TM zone 24\",GEOGCS[\"IGC 1962");
    add_srs_wkt (p, 1,
		 " 6th Parallel South\",DATUM[\"IGC_1962_Arc_of_the_6th_Pa");
    add_srs_wkt (p, 2,
		 "rallel_South\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.14");
    add_srs_wkt (p, 3,
		 "5,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4697\"]],UNIT[");
    add_srs_wkt (p, 7,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",24],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",0.9999],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12,
		 "3324\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 3325, "epsg", 3325,
		      "IGC 1962 / Congo TM zone 26");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=26 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=10000000 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGC 1962 / Congo TM zone 26\",GEOGCS[\"IGC 1962");
    add_srs_wkt (p, 1,
		 " 6th Parallel South\",DATUM[\"IGC_1962_Arc_of_the_6th_Pa");
    add_srs_wkt (p, 2,
		 "rallel_South\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.14");
    add_srs_wkt (p, 3,
		 "5,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4697\"]],UNIT[");
    add_srs_wkt (p, 7,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",26],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",0.9999],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12,
		 "3325\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 3326, "epsg", 3326,
		      "IGC 1962 / Congo TM zone 28");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=28 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=10000000 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGC 1962 / Congo TM zone 28\",GEOGCS[\"IGC 1962");
    add_srs_wkt (p, 1,
		 " 6th Parallel South\",DATUM[\"IGC_1962_Arc_of_the_6th_Pa");
    add_srs_wkt (p, 2,
		 "rallel_South\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.14");
    add_srs_wkt (p, 3,
		 "5,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4697\"]],UNIT[");
    add_srs_wkt (p, 7,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",28],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",0.9999],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12,
		 "3326\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 3327, "epsg", 3327,
		      "IGC 1962 / Congo TM zone 30");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=30 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=10000000 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGC 1962 / Congo TM zone 30\",GEOGCS[\"IGC 1962");
    add_srs_wkt (p, 1,
		 " 6th Parallel South\",DATUM[\"IGC_1962_Arc_of_the_6th_Pa");
    add_srs_wkt (p, 2,
		 "rallel_South\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.14");
    add_srs_wkt (p, 3,
		 "5,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4697\"]],UNIT[");
    add_srs_wkt (p, 7,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",30],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",0.9999],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12,
		 "3327\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 3328, "epsg", 3328,
		      "Pulkovo 1942(58) / GUGiK-80");
    add_proj4text (p, 0,
		   "+proj=sterea +lat_0=52.16666666666666 +lon_0=19.16666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999714 +x_0=500000 +y_0=500000 +ellps=krass ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / GUGiK-80\",GEOGCS[\"Pulkovo ");
    add_srs_wkt (p, 1,
		 "1942(58)\",DATUM[\"Pulkovo_1942_58\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4179\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Oblique_Stereographic\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",52.16666666666666],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",19.16666666666667],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".999714],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",500000],AUTHORITY[\"EPSG\",\"3328\"],AX");
    add_srs_wkt (p, 12, "IS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3329, "epsg", 3329,
		      "Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=1 +x_0=5500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 5");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo_1942_58\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4179\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",15],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",5500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"3329\"],AXIS[\"X\",NORTH],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",EAST]]");
    p = add_epsg_def (first, last, 3330, "epsg", 3330,
		      "Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18 +k=1 +x_0=6500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 6");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo_1942_58\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4179\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",18],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",6500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"3330\"],AXIS[\"X\",NORTH],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",EAST]]");
    p = add_epsg_def (first, last, 3331, "epsg", 3331,
		      "Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 7");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=1 +x_0=7500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 7");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo_1942_58\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4179\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",21],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",7500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"3331\"],AXIS[\"X\",NORTH],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",EAST]]");
    p = add_epsg_def (first, last, 3332, "epsg", 3332,
		      "Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 8");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=24 +k=1 +x_0=8500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 8");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo_1942_58\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4179\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",24],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",8500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"3332\"],AXIS[\"X\",NORTH],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",EAST]]");
    p = add_epsg_def (first, last, 3333, "epsg", 3333,
		      "Pulkovo 1942(58) / Gauss-Kruger zone 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=1 +x_0=3500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / Gauss-Kruger zone 3\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo_1942_58\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7024\"]],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4179\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",15],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",3500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"3333\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 3334, "epsg", 3334,
		      "Pulkovo 1942(58) / Gauss-Kruger zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=1 +x_0=4500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / Gauss-Kruger zone 4\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo_1942_58\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7024\"]],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4179\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",21],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",4500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"3334\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 3335, "epsg", 3335,
		      "Pulkovo 1942(58) / Gauss-Kruger zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=27 +k=1 +x_0=5500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / Gauss-Kruger zone 5\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo_1942_58\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7024\"]],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4179\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",27],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",5500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"3335\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 3336, "epsg", 3336,
		      "IGN 1962 Kerguelen / UTM zone 42S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=42 +south +ellps=intl +towgs84=145,-187,");
    add_proj4text (p, 1, "103,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGN 1962 Kerguelen / UTM zone 42S\",GEOGCS[\"IG");
    add_srs_wkt (p, 1,
		 "N 1962 Kerguelen\",DATUM[\"IGN_1962_Kerguelen\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7022\"]],TOWGS84[145,-187,103,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6698\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4698\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"central_meridian\",69],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 10,
		 "or\",0.9996],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"333");
    add_srs_wkt (p, 12,
		 "6\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3337, "epsg", 3337,
		      "Le Pouce 1934 / Mauritius Grid");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-20.19506944444445 +lat_0=-20.195069444");
    add_proj4text (p, 1,
		   "44445 +lon_0=57.52182777777778 +k_0=1 +x_0=1000000 +y_0=");
    add_proj4text (p, 2,
		   "1000000 +ellps=clrk80 +towgs84=-770.1,158.4,-498.2,0,0,0");
    add_proj4text (p, 3, ",0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Le Pouce 1934 / Mauritius Grid\",GEOGCS[\"Le Po");
    add_srs_wkt (p, 1,
		 "uce 1934\",DATUM[\"Le_Pouce_1934\",SPHEROID[\"Clarke 188");
    add_srs_wkt (p, 2,
		 "0 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[-770.1,158.4,-498.2,0,0,0,0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6699\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4699\"]],UNIT[");
    add_srs_wkt (p, 7,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"L");
    add_srs_wkt (p, 8,
		 "ambert_Conformal_Conic_1SP\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 9,
		 "igin\",-20.19506944444445],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 10,
		 ",57.52182777777778],PARAMETER[\"scale_factor\",1],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_easting\",1000000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",1000000],AUTHORITY[\"EPSG\",\"3337\"],AXIS[\"Easting");
    add_srs_wkt (p, 13, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3338, "epsg", 3338, "NAD83 / Alaska Albers");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=55 +lat_2=65 +lat_0=50 +lon_0=-154 +x_0");
    add_proj4text (p, 1,
		   "=0 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alaska Albers\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Albers_Conic_Equal_Area\"],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 8,
		 "rallel_1\",55],PARAMETER[\"standard_parallel_2\",65],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"latitude_of_center\",50],PARAMETER[\"longitude_");
    add_srs_wkt (p, 10,
		 "of_center\",-154],PARAMETER[\"false_easting\",0],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3338\"],AX");
    add_srs_wkt (p, 12, "IS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3339, "epsg", 3339,
		      "IGCB 1955 / Congo TM zone 12");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=12 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=10000000 +ellps=clrk80 +towgs84=-79.9,-158,-168.9,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGCB 1955 / Congo TM zone 12\",GEOGCS[\"IGCB 19");
    add_srs_wkt (p, 1,
		 "55\",DATUM[\"Institut_Geographique_du_Congo_Belge_1955\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"7012\"]],TOWGS84[-79.9,-158,-168.9,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6701\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4701\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 8,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 9,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",12");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"3339\"],AXIS[\"Easting\",EAST],AXIS");
    add_srs_wkt (p, 13, "[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3340, "epsg", 3340,
		      "IGCB 1955 / Congo TM zone 14");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=14 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=10000000 +ellps=clrk80 +towgs84=-79.9,-158,-168.9,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGCB 1955 / Congo TM zone 14\",GEOGCS[\"IGCB 19");
    add_srs_wkt (p, 1,
		 "55\",DATUM[\"Institut_Geographique_du_Congo_Belge_1955\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"7012\"]],TOWGS84[-79.9,-158,-168.9,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6701\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4701\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 8,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 9,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",14");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"3340\"],AXIS[\"Easting\",EAST],AXIS");
    add_srs_wkt (p, 13, "[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3341, "epsg", 3341,
		      "IGCB 1955 / Congo TM zone 16");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=16 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=10000000 +ellps=clrk80 +towgs84=-79.9,-158,-168.9,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGCB 1955 / Congo TM zone 16\",GEOGCS[\"IGCB 19");
    add_srs_wkt (p, 1,
		 "55\",DATUM[\"Institut_Geographique_du_Congo_Belge_1955\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"7012\"]],TOWGS84[-79.9,-158,-168.9,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6701\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4701\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 8,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 9,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",16");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"3341\"],AXIS[\"Easting\",EAST],AXIS");
    add_srs_wkt (p, 13, "[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3342, "epsg", 3342,
		      "IGCB 1955 / UTM zone 33S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=33 +south +ellps=clrk80 +towgs84=-79.9,-");
    add_proj4text (p, 1, "158,-168.9,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGCB 1955 / UTM zone 33S\",GEOGCS[\"IGCB 1955\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Institut_Geographique_du_Congo_Belge_1955\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7012\"]],TOWGS84[-79.9,-158,-168.9,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6701\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4701\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",15],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",500000],PARAMETER[\"false_northing\",10000000],AUTH");
    add_srs_wkt (p, 12,
		 "ORITY[\"EPSG\",\"3342\"],AXIS[\"Easting\",EAST],AXIS[\"N");
    add_srs_wkt (p, 13, "orthing\",NORTH]]");
    p = add_epsg_def (first, last, 3343, "epsg", 3343,
		      "Mauritania 1999 / UTM zone 28N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=28 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Mauritania 1999 / UTM zone 28N\",GEOGCS[\"Mauri");
    add_srs_wkt (p, 1,
		 "tania 1999\",DATUM[\"Mauritania_1999\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 2,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 3,
		 ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6702\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6,
		 "\"]],AUTHORITY[\"EPSG\",\"4702\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"ce");
    add_srs_wkt (p, 9,
		 "ntral_meridian\",-15],PARAMETER[\"scale_factor\",0.9996]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_easting\",500000],PARAMETER[\"false_n");
    add_srs_wkt (p, 11,
		 "orthing\",0],AUTHORITY[\"EPSG\",\"3343\"],AXIS[\"Easting");
    add_srs_wkt (p, 12, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3344, "epsg", 3344,
		      "Mauritania 1999 / UTM zone 29N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=29 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Mauritania 1999 / UTM zone 29N\",GEOGCS[\"Mauri");
    add_srs_wkt (p, 1,
		 "tania 1999\",DATUM[\"Mauritania_1999\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 2,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 3,
		 ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6702\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6,
		 "\"]],AUTHORITY[\"EPSG\",\"4702\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"ce");
    add_srs_wkt (p, 9,
		 "ntral_meridian\",-9],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 11,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"3344\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 12, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3345, "epsg", 3345,
		      "Mauritania 1999 / UTM zone 30N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=30 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Mauritania 1999 / UTM zone 30N\",GEOGCS[\"Mauri");
    add_srs_wkt (p, 1,
		 "tania 1999\",DATUM[\"Mauritania_1999\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 2,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 3,
		 ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6702\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6,
		 "\"]],AUTHORITY[\"EPSG\",\"4702\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"ce");
    add_srs_wkt (p, 9,
		 "ntral_meridian\",-3],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 11,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"3345\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 12, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3346, "epsg", 3346, "LKS94 / Lithuania TM");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=24 +k=0.9998 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"LKS94 / Lithuania TM\",GEOGCS[\"LKS94\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Lithuania_1994_ETRS89\",SPHEROID[\"GRS 1980\",6378137,29");
    add_srs_wkt (p, 2,
		 "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0],AUTHORITY[\"EPSG\",\"6126\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4669\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",24],PARAMETER[\"scale_factor\",0.9998],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"3346\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EA");
    add_srs_wkt (p, 12, "ST]]");
    p = add_epsg_def (first, last, 3347, "epsg", 3347,
		      "NAD83 / Statistics Canada Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49 +lat_2=77 +lat_0=63.390675 +lon_0=-9");
    add_proj4text (p, 1,
		   "1.86666666666666 +x_0=6200000 +y_0=3000000 +ellps=GRS80 ");
    add_proj4text (p, 2, "+datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Statistics Canada Lambert\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"standard_parallel_1\",49],PARAMETER[\"standard_para");
    add_srs_wkt (p, 9,
		 "llel_2\",77],PARAMETER[\"latitude_of_origin\",63.390675]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"central_meridian\",-91.86666666666666],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_easting\",6200000],PARAMETER[\"false_north");
    add_srs_wkt (p, 12,
		 "ing\",3000000],AUTHORITY[\"EPSG\",\"3347\"],AXIS[\"Easti");
    add_srs_wkt (p, 13, "ng\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3348, "epsg", 3348,
		      "NAD83(CSRS) / Statistics Canada Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49 +lat_2=77 +lat_0=63.390675 +lon_0=-9");
    add_proj4text (p, 1,
		   "1.86666666666666 +x_0=6200000 +y_0=3000000 +ellps=GRS80 ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / Statistics Canada Lambert\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(CSRS)\",DATUM[\"NAD83_Canadian_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Confo");
    add_srs_wkt (p, 8,
		 "rmal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",49],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"standard_parallel_2\",77],PARAMETER[\"latitu");
    add_srs_wkt (p, 10,
		 "de_of_origin\",63.390675],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 11,
		 ",-91.86666666666666],PARAMETER[\"false_easting\",6200000");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_northing\",3000000],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 13,
		 "\",\"3348\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NO");
    add_srs_wkt (p, 14, "RTH]]");
    p = add_epsg_def (first, last, 3349, "epsg", 3349,
		      "WGS 84 / PDC Mercator (deprecated)");
    add_proj4text (p, 0,
		   "+proj=merc +lon_0=-150 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +");
    add_proj4text (p, 1, "datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / PDC Mercator (deprecated)\",GEOGCS[\"W");
    add_srs_wkt (p, 1,
		 "GS 84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,2");
    add_srs_wkt (p, 2,
		 "98.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Mercator_1SP\"],PARAMETER[\"central_meridian\",-150],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 9,
		 ",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 10,
		 "3349\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 11, "");
    p = add_epsg_def (first, last, 3350, "epsg", 3350,
		      "Pulkovo 1942 / CS63 zone C0");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0.1 +lon_0=21.95 +k=1 +x_0=250000 +y_");
    add_proj4text (p, 1, "0=0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / CS63 zone C0\",GEOGCS[\"Pulkovo ");
    add_srs_wkt (p, 1,
		 "1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassowsky 1940");
    add_srs_wkt (p, 2,
		 "\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[2");
    add_srs_wkt (p, 3,
		 "3.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0.1]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"central_meridian\",21.95],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",1],PARAMETER[\"false_easting\",250000],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3350\"],AX");
    add_srs_wkt (p, 12, "IS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3351, "epsg", 3351,
		      "Pulkovo 1942 / CS63 zone C1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0.1 +lon_0=24.95 +k=1 +x_0=1250000 +y");
    add_proj4text (p, 1, "_0=0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / CS63 zone C1\",GEOGCS[\"Pulkovo ");
    add_srs_wkt (p, 1,
		 "1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassowsky 1940");
    add_srs_wkt (p, 2,
		 "\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[2");
    add_srs_wkt (p, 3,
		 "3.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0.1]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"central_meridian\",24.95],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",1],PARAMETER[\"false_easting\",1250000],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3351\"],A");
    add_srs_wkt (p, 12, "XIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3352, "epsg", 3352,
		      "Pulkovo 1942 / CS63 zone C2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0.1 +lon_0=27.95 +k=1 +x_0=2250000 +y");
    add_proj4text (p, 1, "_0=0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / CS63 zone C2\",GEOGCS[\"Pulkovo ");
    add_srs_wkt (p, 1,
		 "1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassowsky 1940");
    add_srs_wkt (p, 2,
		 "\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[2");
    add_srs_wkt (p, 3,
		 "3.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0.1]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"central_meridian\",27.95],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",1],PARAMETER[\"false_easting\",2250000],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3352\"],A");
    add_srs_wkt (p, 12, "XIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3353, "epsg", 3353,
		      "Mhast (onshore) / UTM zone 32S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=32 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Mhast (onshore) / UTM zone 32S\",GEOGCS[\"Mhast");
    add_srs_wkt (p, 1,
		 " (onshore)\",DATUM[\"Mhast_onshore\",SPHEROID[\"Internat");
    add_srs_wkt (p, 2,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6704\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4704\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",9],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",500000],PARAMETER[\"false_northing\",10000000],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"3353\"],AXIS[\"Easting\",EAST],AXIS[\"North");
    add_srs_wkt (p, 12, "ing\",NORTH]]");
    p = add_epsg_def (first, last, 3354, "epsg", 3354,
		      "Mhast (offshore) / UTM zone 32S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=32 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Mhast (offshore) / UTM zone 32S\",GEOGCS[\"Mhas");
    add_srs_wkt (p, 1,
		 "t (offshore)\",DATUM[\"Mhast_offshore\",SPHEROID[\"Inter");
    add_srs_wkt (p, 2,
		 "national 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6705\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4705\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",9],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",10000000],AUTHO");
    add_srs_wkt (p, 11,
		 "RITY[\"EPSG\",\"3354\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 12, "rthing\",NORTH]]");
    p = add_epsg_def (first, last, 3355, "epsg", 3355,
		      "Egypt Gulf of Suez S-650 TL / Red Belt");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=31 +k=1 +x_0=615000 +y_0=81");
    add_proj4text (p, 1,
		   "0000 +ellps=helmert +towgs84=-146.21,112.63,4.05,0,0,0,0");
    add_proj4text (p, 2, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Egypt Gulf of Suez S-650 TL / Red Belt\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"Egypt Gulf of Suez S-650 TL\",DATUM[\"Egypt_Gulf_of_S");
    add_srs_wkt (p, 2,
		 "uez_S_650_TL\",SPHEROID[\"Helmert 1906\",6378200,298.3,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7020\"]],TOWGS84[-146.21,112.63,4.05");
    add_srs_wkt (p, 4,
		 ",0,0,0,0],AUTHORITY[\"EPSG\",\"6706\"]],PRIMEM[\"Greenwi");
    add_srs_wkt (p, 5,
		 "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0");
    add_srs_wkt (p, 6,
		 "1745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY");
    add_srs_wkt (p, 7,
		 "[\"EPSG\",\"4706\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 8,
		 ",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"latitude_of_origin\",30],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 10,
		 "n\",31],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",615000],PARAMETER[\"false_northing\",810000],A");
    add_srs_wkt (p, 12,
		 "UTHORITY[\"EPSG\",\"3355\"],AXIS[\"Easting\",EAST],AXIS[");
    add_srs_wkt (p, 13, "\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3356, "epsg", 3356,
		      "Grand Cayman 1959 / UTM zone 17N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=17 +ellps=clrk66 +towgs84=67.8,106.1,138");
    add_proj4text (p, 1, ".8,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Grand Cayman 1959 / UTM zone 17N\",GEOGCS[\"Gra");
    add_srs_wkt (p, 1,
		 "nd Cayman 1959\",DATUM[\"Grand_Cayman_1959\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7008\"]],TOWGS84[67.8,106.1,138.8,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"6723\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4723");
    add_srs_wkt (p, 7,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 8,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 9,
		 "origin\",0],PARAMETER[\"central_meridian\",-81],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"scale_factor\",0.9996],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12,
		 ",\"3356\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORT");
    add_srs_wkt (p, 13, "H]]");
    p = add_epsg_def (first, last, 3357, "epsg", 3357,
		      "Little Cayman 1961 / UTM zone 17N");
    add_proj4text (p, 0, "+proj=utm +zone=17 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Little Cayman 1961 / UTM zone 17N\",GEOGCS[\"Li");
    add_srs_wkt (p, 1,
		 "ttle Cayman 1961\",DATUM[\"Little_Cayman_1961\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6726\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4726\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-81],PARAMETER[\"scale_factor\",0.9996],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"3357\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 12, "AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3358, "epsg", 3358,
		      "NAD83(HARN) / North Carolina");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.16666666666666 +lat_2=34.33333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=33.75 +lon_0=-79 +x_0=609601.22 +y_0=0 +ellps");
    add_proj4text (p, 2, "=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / North Carolina\",GEOGCS[\"NAD83(H");
    add_srs_wkt (p, 1,
		 "ARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",36.1666666666666");
    add_srs_wkt (p, 9,
		 "6],PARAMETER[\"standard_parallel_2\",34.33333333333334],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",33.75],PARAMETER[\"cent");
    add_srs_wkt (p, 11,
		 "ral_meridian\",-79],PARAMETER[\"false_easting\",609601.2");
    add_srs_wkt (p, 12,
		 "2],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 13, "3358\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3359, "epsg", 3359,
		      "NAD83(HARN) / North Carolina (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.16666666666666 +lat_2=34.33333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=33.75 +lon_0=-79 +x_0=609601.2192024385 +y_0=");
    add_proj4text (p, 2, "0 +ellps=GRS80 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / North Carolina (ftUS) (deprecated");
    add_srs_wkt (p, 1,
		 ")\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_R");
    add_srs_wkt (p, 2,
		 "egional_Network\",SPHEROID[\"GRS 1980\",6378137,298.2572");
    add_srs_wkt (p, 3,
		 "22101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"foo");
    add_srs_wkt (p, 7,
		 "t\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],PROJECTION[\"La");
    add_srs_wkt (p, 8,
		 "mbert_Conformal_Conic_2SP\"],PARAMETER[\"standard_parall");
    add_srs_wkt (p, 9,
		 "el_1\",36.16666666666666],PARAMETER[\"standard_parallel_");
    add_srs_wkt (p, 10,
		 "2\",34.33333333333334],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 11,
		 "33.75],PARAMETER[\"central_meridian\",-79],PARAMETER[\"f");
    add_srs_wkt (p, 12,
		 "alse_easting\",2000004.000008],PARAMETER[\"false_northin");
    add_srs_wkt (p, 13,
		 "g\",0],AUTHORITY[\"EPSG\",\"3359\"],AXIS[\"X\",EAST],AXI");
    add_srs_wkt (p, 14, "S[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3360, "epsg", 3360,
		      "NAD83(HARN) / South Carolina");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=34.83333333333334 +lat_2=32.5 +lat_0=31");
    add_proj4text (p, 1,
		   ".83333333333333 +lon_0=-81 +x_0=609600 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / South Carolina\",GEOGCS[\"NAD83(H");
    add_srs_wkt (p, 1,
		 "ARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",34.8333333333333");
    add_srs_wkt (p, 9,
		 "4],PARAMETER[\"standard_parallel_2\",32.5],PARAMETER[\"l");
    add_srs_wkt (p, 10,
		 "atitude_of_origin\",31.83333333333333],PARAMETER[\"centr");
    add_srs_wkt (p, 11,
		 "al_meridian\",-81],PARAMETER[\"false_easting\",609600],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3360");
    add_srs_wkt (p, 13, "\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3361, "epsg", 3361,
		      "NAD83(HARN) / South Carolina (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=34.83333333333334 +lat_2=32.5 +lat_0=31");
    add_proj4text (p, 1,
		   ".83333333333333 +lon_0=-81 +x_0=609600 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / South Carolina (ft)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"foot\",0.3048,AUT");
    add_srs_wkt (p, 7,
		 "HORITY[\"EPSG\",\"9002\"]],PROJECTION[\"Lambert_Conforma");
    add_srs_wkt (p, 8,
		 "l_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",34.8333");
    add_srs_wkt (p, 9,
		 "3333333334],PARAMETER[\"standard_parallel_2\",32.5],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"latitude_of_origin\",31.83333333333333],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"central_meridian\",-81],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 12,
		 "2000000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 13, "G\",\"3361\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3362, "epsg", 3362,
		      "NAD83(HARN) / Pennsylvania North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.95 +lat_2=40.88333333333333 +lat_0=4");
    add_proj4text (p, 1,
		   "0.16666666666666 +lon_0=-77.75 +x_0=600000 +y_0=0 +ellps");
    add_proj4text (p, 2, "=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Pennsylvania North\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",41.95],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_2\",40.88333333333333],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "latitude_of_origin\",40.16666666666666],PARAMETER[\"cent");
    add_srs_wkt (p, 11,
		 "ral_meridian\",-77.75],PARAMETER[\"false_easting\",60000");
    add_srs_wkt (p, 12,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 13, "3362\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3363, "epsg", 3363,
		      "NAD83(HARN) / Pennsylvania North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.95 +lat_2=40.88333333333333 +lat_0=4");
    add_proj4text (p, 1,
		   "0.16666666666666 +lon_0=-77.75 +x_0=600000 +y_0=0 +ellps");
    add_proj4text (p, 2, "=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Pennsylvania North (ftUS)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_N");
    add_srs_wkt (p, 2,
		 "etwork\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]]");
    add_srs_wkt (p, 4,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 5,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 6,
		 "122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey f");
    add_srs_wkt (p, 7,
		 "oot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_1\",41.95],PARAMETER[\"standard_paralle");
    add_srs_wkt (p, 10,
		 "l_2\",40.88333333333333],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 11,
		 ",40.16666666666666],PARAMETER[\"central_meridian\",-77.7");
    add_srs_wkt (p, 12,
		 "5],PARAMETER[\"false_easting\",1968500],PARAMETER[\"fals");
    add_srs_wkt (p, 13,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"3363\"],AXIS[\"X\",");
    add_srs_wkt (p, 14, "EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3364, "epsg", 3364,
		      "NAD83(HARN) / Pennsylvania South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.96666666666667 +lat_2=39.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=39.33333333333334 +lon_0=-77.75 +x_0=600000 +");
    add_proj4text (p, 2, "y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Pennsylvania South\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",40.9666666666666");
    add_srs_wkt (p, 9,
		 "7],PARAMETER[\"standard_parallel_2\",39.93333333333333],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",39.33333333333334],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-77.75],PARAMETER[\"false_eas");
    add_srs_wkt (p, 12,
		 "ting\",600000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 13,
		 "[\"EPSG\",\"3364\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3365, "epsg", 3365,
		      "NAD83(HARN) / Pennsylvania South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.96666666666667 +lat_2=39.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=39.33333333333334 +lon_0=-77.75 +x_0=600000 +");
    add_proj4text (p, 2, "y_0=0 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Pennsylvania South (ftUS)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_N");
    add_srs_wkt (p, 2,
		 "etwork\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]]");
    add_srs_wkt (p, 4,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 5,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 6,
		 "122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey f");
    add_srs_wkt (p, 7,
		 "oot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_1\",40.96666666666667],PARAMETER[\"stan");
    add_srs_wkt (p, 10,
		 "dard_parallel_2\",39.93333333333333],PARAMETER[\"latitud");
    add_srs_wkt (p, 11,
		 "e_of_origin\",39.33333333333334],PARAMETER[\"central_mer");
    add_srs_wkt (p, 12,
		 "idian\",-77.75],PARAMETER[\"false_easting\",1968500],PAR");
    add_srs_wkt (p, 13,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3365\"");
    add_srs_wkt (p, 14, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3366, "epsg", 3366,
		      "Hong Kong 1963 Grid System (deprecated)");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=22.31213333333334 +lon_0=114.178555555");
    add_proj4text (p, 1,
		   "5556 +x_0=40243.57775604237 +y_0=19069.93351512578 +a=63");
    add_proj4text (p, 2,
		   "78293.645208759 +b=6356617.987679838 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Hong Kong 1963 Grid System (deprecated)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"Hong Kong 1963\",DATUM[\"Hong_Kong_1963\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Clarke 1858\",6378293.645208759,294.2606763692569,AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"7007\"]],AUTHORITY[\"EPSG\",\"6738\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6,
		 "\"]],AUTHORITY[\"EPSG\",\"4738\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Cassini_Soldner\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",22.31213333333334],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"central_meridian\",114.1785555555556],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_easting\",40243.57775604237],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_northing\",19069.93351512578],AUTHORITY[\"EPSG\",\"3366");
    add_srs_wkt (p, 12, "\"],AXIS[\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 3367, "epsg", 3367,
		      "IGN Astro 1960 / UTM zone 28N");
    add_proj4text (p, 0, "+proj=utm +zone=28 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGN Astro 1960 / UTM zone 28N\",GEOGCS[\"IGN As");
    add_srs_wkt (p, 1,
		 "tro 1960\",DATUM[\"IGN_Astro_1960\",SPHEROID[\"Clarke 18");
    add_srs_wkt (p, 2,
		 "80 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6700\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4700\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",-1");
    add_srs_wkt (p, 9,
		 "5],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"3367\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 12, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 3368, "epsg", 3368,
		      "IGN Astro 1960 / UTM zone 29N");
    add_proj4text (p, 0, "+proj=utm +zone=29 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGN Astro 1960 / UTM zone 29N\",GEOGCS[\"IGN As");
    add_srs_wkt (p, 1,
		 "tro 1960\",DATUM[\"IGN_Astro_1960\",SPHEROID[\"Clarke 18");
    add_srs_wkt (p, 2,
		 "80 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6700\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4700\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",-9");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 10,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 11,
		 "TY[\"EPSG\",\"3368\"],AXIS[\"Easting\",EAST],AXIS[\"Nort");
    add_srs_wkt (p, 12, "hing\",NORTH]]");
    p = add_epsg_def (first, last, 3369, "epsg", 3369,
		      "IGN Astro 1960 / UTM zone 30N");
    add_proj4text (p, 0, "+proj=utm +zone=30 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGN Astro 1960 / UTM zone 30N\",GEOGCS[\"IGN As");
    add_srs_wkt (p, 1,
		 "tro 1960\",DATUM[\"IGN_Astro_1960\",SPHEROID[\"Clarke 18");
    add_srs_wkt (p, 2,
		 "80 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6700\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4700\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",-3");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 10,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 11,
		 "TY[\"EPSG\",\"3369\"],AXIS[\"Easting\",EAST],AXIS[\"Nort");
    add_srs_wkt (p, 12, "hing\",NORTH]]");
    p = add_epsg_def (first, last, 3370, "epsg", 3370, "NAD27 / UTM zone 59N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=59 +ellps=clrk66 +datum=NAD27 +units=m +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 59N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",171],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 11,
		 "PSG\",\"3370\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 3371, "epsg", 3371, "NAD27 / UTM zone 60N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=60 +ellps=clrk66 +datum=NAD27 +units=m +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 60N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",177],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 11,
		 "PSG\",\"3371\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 3372, "epsg", 3372, "NAD83 / UTM zone 59N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=59 +ellps=GRS80 +datum=NAD83 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 59N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",171],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "3372\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 3373, "epsg", 3373, "NAD83 / UTM zone 60N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=60 +ellps=GRS80 +datum=NAD83 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 60N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",177],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "3373\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 3374, "epsg", 3374, "FD54 / UTM zone 29N");
    add_proj4text (p, 0, "+proj=utm +zone=29 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"FD54 / UTM zone 29N\",GEOGCS[\"FD54\",DATUM[\"F");
    add_srs_wkt (p, 1,
		 "aroe_Datum_1954\",SPHEROID[\"International 1924\",637838");
    add_srs_wkt (p, 2,
		 "8,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "6741\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 4,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 5,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4741\"]],UNIT[\"met");
    add_srs_wkt (p, 6,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 7,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 8,
		 "AMETER[\"central_meridian\",-9],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 9,
		 "\",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3374\"],AXIS");
    add_srs_wkt (p, 11, "[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3375, "epsg", 3375,
		      "GDM2000 / Peninsula RSO");
    add_proj4text (p, 0,
		   "+proj=omerc +lat_0=4 +lonc=102.25 +alpha=323.02579646666");
    add_proj4text (p, 1,
		   "66 +k=0.99984 +x_0=804671 +y_0=0 +ellps=GRS80 +units=m +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDM2000 / Peninsula RSO\",GEOGCS[\"GDM2000\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"Geodetic_Datum_of_Malaysia_2000\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 2,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4742\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Hotine_Oblique_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_center\",4],PARAMETER[\"longitude_of_center\"");
    add_srs_wkt (p, 9,
		 ",102.25],PARAMETER[\"azimuth\",323.0257964666666],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"rectified_grid_angle\",323.1301023611111],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"scale_factor\",0.99984],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 12,
		 "804671],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 13,
		 "\",\"3375\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NO");
    add_srs_wkt (p, 14, "RTH]]");
    p = add_epsg_def (first, last, 3376, "epsg", 3376,
		      "GDM2000 / East Malaysia BRSO");
    add_proj4text (p, 0,
		   "+proj=omerc +lat_0=4 +lonc=115 +alpha=53.31580995 +k=0.9");
    add_proj4text (p, 1, "9984 +x_0=0 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDM2000 / East Malaysia BRSO\",GEOGCS[\"GDM2000");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"Geodetic_Datum_of_Malaysia_2000\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4742\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Hotine_Oblique_Mercator\"],PARAMETE");
    add_srs_wkt (p, 8,
		 "R[\"latitude_of_center\",4],PARAMETER[\"longitude_of_cen");
    add_srs_wkt (p, 9,
		 "ter\",115],PARAMETER[\"azimuth\",53.31580995],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"rectified_grid_angle\",53.13010236111111],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "scale_factor\",0.99984],PARAMETER[\"false_easting\",0],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3376");
    add_srs_wkt (p, 13, "\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3377, "epsg", 3377, "GDM2000 / Johor Grid");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=2.121679744444445 +lon_0=103.427936236");
    add_proj4text (p, 1,
		   "1111 +x_0=-14810.562 +y_0=8758.32 +ellps=GRS80 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDM2000 / Johor Grid\",GEOGCS[\"GDM2000\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Geodetic_Datum_of_Malaysia_2000\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4742\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Cassini_Soldner\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",2.121679744444445],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 9,
		 "\",103.4279362361111],PARAMETER[\"false_easting\",-14810");
    add_srs_wkt (p, 10,
		 ".562],PARAMETER[\"false_northing\",8758.32],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"3377\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 3378, "epsg", 3378,
		      "GDM2000 / Sembilan and Melaka Grid");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=2.682347636111111 +lon_0=101.974905041");
    add_proj4text (p, 1,
		   "6667 +x_0=3673.785 +y_0=-4240.573 +ellps=GRS80 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDM2000 / Sembilan and Melaka Grid\",GEOGCS[\"G");
    add_srs_wkt (p, 1,
		 "DM2000\",DATUM[\"Geodetic_Datum_of_Malaysia_2000\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4742\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Cassini_Soldner\"],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"latitude_of_origin\",2.682347636111111],PARAMETER[\"ce");
    add_srs_wkt (p, 9,
		 "ntral_meridian\",101.9749050416667],PARAMETER[\"false_ea");
    add_srs_wkt (p, 10,
		 "sting\",3673.785],PARAMETER[\"false_northing\",-4240.573");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"3378\"],AXIS[\"Easting\",EAST],AX");
    add_srs_wkt (p, 12, "IS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3379, "epsg", 3379, "GDM2000 / PahangGrid");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=3.769388088888889 +lon_0=102.368298983");
    add_proj4text (p, 1,
		   "3333 +x_0=-7368.228 +y_0=6485.858 +ellps=GRS80 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDM2000 / PahangGrid\",GEOGCS[\"GDM2000\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Geodetic_Datum_of_Malaysia_2000\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4742\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Cassini_Soldner\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",3.769388088888889],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 9,
		 "\",102.3682989833333],PARAMETER[\"false_easting\",-7368.");
    add_srs_wkt (p, 10,
		 "228],PARAMETER[\"false_northing\",6485.858],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"3379\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 3380, "epsg", 3380,
		      "GDM2000 / Selangor Grid");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=3.68464905 +lon_0=101.3891079138889 +x");
    add_proj4text (p, 1,
		   "_0=-34836.161 +y_0=56464.049 +ellps=GRS80 +units=m +no_d");
    add_proj4text (p, 2, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDM2000 / Selangor Grid\",GEOGCS[\"GDM2000\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"Geodetic_Datum_of_Malaysia_2000\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 2,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4742\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Cassini_Soldner\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 8,
		 "of_origin\",3.68464905],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "01.3891079138889],PARAMETER[\"false_easting\",-34836.161");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"false_northing\",56464.049],AUTHORITY[\"EP");
    add_srs_wkt (p, 11,
		 "SG\",\"3380\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",");
    add_srs_wkt (p, 12, "NORTH]]");
    p = add_epsg_def (first, last, 3381, "epsg", 3381,
		      "GDM2000 / Terengganu Grid");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=4.9762852 +lon_0=103.070275625 +x_0=19");
    add_proj4text (p, 1,
		   "594.245 +y_0=3371.895 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDM2000 / Terengganu Grid\",GEOGCS[\"GDM2000\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Geodetic_Datum_of_Malaysia_2000\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4742\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Cassini_Soldner\"],PARAMETER[\"latitud");
    add_srs_wkt (p, 8,
		 "e_of_origin\",4.9762852],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 9,
		 "103.070275625],PARAMETER[\"false_easting\",19594.245],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",3371.895],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11,
		 "\"3381\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH");
    add_srs_wkt (p, 12, "]]");
    p = add_epsg_def (first, last, 3382, "epsg", 3382, "GDM2000 / Pinang Grid");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=5.421517541666667 +lon_0=100.344376963");
    add_proj4text (p, 1,
		   "8889 +x_0=-23.414 +y_0=62.283 +ellps=GRS80 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDM2000 / Pinang Grid\",GEOGCS[\"GDM2000\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"Geodetic_Datum_of_Malaysia_2000\",SPHEROID[\"GRS 198");
    add_srs_wkt (p, 2,
		 "0\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"4742\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Cassini_Soldner\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 8,
		 "f_origin\",5.421517541666667],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",100.3443769638889],PARAMETER[\"false_easting\",-23.");
    add_srs_wkt (p, 10,
		 "414],PARAMETER[\"false_northing\",62.283],AUTHORITY[\"EP");
    add_srs_wkt (p, 11,
		 "SG\",\"3382\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",");
    add_srs_wkt (p, 12, "NORTH]]");
    p = add_epsg_def (first, last, 3383, "epsg", 3383,
		      "GDM2000 / Kedah and Perlis Grid");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=5.964672713888889 +lon_0=100.636371111");
    add_proj4text (p, 1, "1111 +x_0=0 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDM2000 / Kedah and Perlis Grid\",GEOGCS[\"GDM2");
    add_srs_wkt (p, 1,
		 "000\",DATUM[\"Geodetic_Datum_of_Malaysia_2000\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4742\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Cassini_Soldner\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",5.964672713888889],PARAMETER[\"centr");
    add_srs_wkt (p, 9,
		 "al_meridian\",100.6363711111111],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 11,
		 "\",\"3383\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NO");
    add_srs_wkt (p, 12, "RTH]]");
    p = add_epsg_def (first, last, 3384, "epsg", 3384, "GDM2000 / Perak Grid");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=4.859063022222222 +lon_0=100.815410586");
    add_proj4text (p, 1,
		   "1111 +x_0=-1.769 +y_0=133454.779 +ellps=GRS80 +units=m +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDM2000 / Perak Grid\",GEOGCS[\"GDM2000\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Geodetic_Datum_of_Malaysia_2000\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4742\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Cassini_Soldner\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",4.859063022222222],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 9,
		 "\",100.8154105861111],PARAMETER[\"false_easting\",-1.769");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"false_northing\",133454.779],AUTHORITY[\"E");
    add_srs_wkt (p, 11,
		 "PSG\",\"3384\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 3385, "epsg", 3385,
		      "GDM2000 / Kelantan Grid");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=5.972543658333334 +lon_0=102.295241669");
    add_proj4text (p, 1,
		   "4444 +x_0=13227.851 +y_0=8739.894 +ellps=GRS80 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDM2000 / Kelantan Grid\",GEOGCS[\"GDM2000\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"Geodetic_Datum_of_Malaysia_2000\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 2,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4742\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Cassini_Soldner\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 8,
		 "of_origin\",5.972543658333334],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",102.2952416694444],PARAMETER[\"false_easting\",132");
    add_srs_wkt (p, 10,
		 "27.851],PARAMETER[\"false_northing\",8739.894],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3385\"],AXIS[\"Easting\",EAST],AXIS[\"Northi");
    add_srs_wkt (p, 12, "ng\",NORTH]]");
    p = add_epsg_def (first, last, 3386, "epsg", 3386, "KKJ / Finland zone 0");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"KKJ / Finland zone 0\",GEOGCS[\"KKJ\",DATUM[\"K");
    add_srs_wkt (p, 1,
		 "artastokoordinaattijarjestelma_1966\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6123\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"4123\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",18],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 10,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"3386\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3387, "epsg", 3387, "KKJ / Finland zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=33 +k=1 +x_0=5500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"KKJ / Finland zone 5\",GEOGCS[\"KKJ\",DATUM[\"K");
    add_srs_wkt (p, 1,
		 "artastokoordinaattijarjestelma_1966\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6123\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"4123\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",0],PARAMETER[\"central_meridian\",33],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 10,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 11, "\",\"3387\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3388, "epsg", 3388,
		      "Pulkovo 1942 / Caspian Sea Mercator");
    add_proj4text (p, 0,
		   "+proj=merc +lon_0=51 +k=1 +x_0=0 +y_0=0 +ellps=krass +un");
    add_proj4text (p, 1, "its=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Caspian Sea Mercator\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Mercator_1SP\"],PARAMETER[\"central_meridian\",51],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "3388\"],AXIS[\"none\",NORTH],AXIS[\"none\",EAST]]");
    p = add_epsg_def (first, last, 3389, "epsg", 3389,
		      "Pulkovo 1942 / 3-degree Gauss-Kruger zone 60");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=180 +k=1 +x_0=60500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / 3-degree Gauss-Kruger zone 60\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",180],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",60");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"3389\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3390, "epsg", 3390,
		      "Pulkovo 1995 / 3-degree Gauss-Kruger zone 60");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=180 +k=1 +x_0=60500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / 3-degree Gauss-Kruger zone 60\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",180],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",60");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"3390\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3391, "epsg", 3391,
		      "Karbala 1979 / UTM zone 37N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=37 +ellps=clrk80 +towgs84=84.1,-320.1,21");
    add_proj4text (p, 1, "8.7,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Karbala 1979 / UTM zone 37N\",GEOGCS[\"Karbala ");
    add_srs_wkt (p, 1,
		 "1979\",DATUM[\"Karbala_1979\",SPHEROID[\"Clarke 1880 (RG");
    add_srs_wkt (p, 2,
		 "S)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[84.1,-320.1,218.7,0,0,0,0],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 4,
		 "743\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 5,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 6,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4743\"]],UNIT[\"metr");
    add_srs_wkt (p, 7,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transve");
    add_srs_wkt (p, 8,
		 "rse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"central_meridian\",39],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3391\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3392, "epsg", 3392,
		      "Karbala 1979 / UTM zone 38N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=38 +ellps=clrk80 +towgs84=84.1,-320.1,21");
    add_proj4text (p, 1, "8.7,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Karbala 1979 / UTM zone 38N\",GEOGCS[\"Karbala ");
    add_srs_wkt (p, 1,
		 "1979\",DATUM[\"Karbala_1979\",SPHEROID[\"Clarke 1880 (RG");
    add_srs_wkt (p, 2,
		 "S)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[84.1,-320.1,218.7,0,0,0,0],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 4,
		 "743\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 5,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 6,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4743\"]],UNIT[\"metr");
    add_srs_wkt (p, 7,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transve");
    add_srs_wkt (p, 8,
		 "rse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"central_meridian\",45],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3392\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3393, "epsg", 3393,
		      "Karbala 1979 / UTM zone 39N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=39 +ellps=clrk80 +towgs84=84.1,-320.1,21");
    add_proj4text (p, 1, "8.7,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Karbala 1979 / UTM zone 39N\",GEOGCS[\"Karbala ");
    add_srs_wkt (p, 1,
		 "1979\",DATUM[\"Karbala_1979\",SPHEROID[\"Clarke 1880 (RG");
    add_srs_wkt (p, 2,
		 "S)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[84.1,-320.1,218.7,0,0,0,0],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 4,
		 "743\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 5,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 6,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4743\"]],UNIT[\"metr");
    add_srs_wkt (p, 7,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transve");
    add_srs_wkt (p, 8,
		 "rse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"central_meridian\",51],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3393\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3394, "epsg", 3394,
		      "Nahrwan 1934 / Iraq zone");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=32.5 +lat_0=32.5 +lon_0=45 +k_0=0.99878");
    add_proj4text (p, 1,
		   "64078000001 +x_0=1500000 +y_0=1166200 +ellps=clrk80 +uni");
    add_proj4text (p, 2, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Nahrwan 1934 / Iraq zone\",GEOGCS[\"Nahrwan 193");
    add_srs_wkt (p, 1,
		 "4\",DATUM[\"Nahrwan_1934\",SPHEROID[\"Clarke 1880 (RGS)\"");
    add_srs_wkt (p, 2,
		 ",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6744\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"474");
    add_srs_wkt (p, 6,
		 "4\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Lambert_Conformal_Conic_1SP\"],PARAMETER[\"lat");
    add_srs_wkt (p, 8,
		 "itude_of_origin\",32.5],PARAMETER[\"central_meridian\",4");
    add_srs_wkt (p, 9,
		 "5],PARAMETER[\"scale_factor\",0.9987864078],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",1500000],PARAMETER[\"false_northing\",11");
    add_srs_wkt (p, 11,
		 "66200],AUTHORITY[\"EPSG\",\"3394\"],AXIS[\"Easting\",EAS");
    add_srs_wkt (p, 12, "T],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3395, "epsg", 3395,
		      "WGS 84 / World Mercator");
    add_proj4text (p, 0,
		   "+proj=merc +lon_0=0 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +dat");
    add_proj4text (p, 1, "um=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / World Mercator\",GEOGCS[\"WGS 84\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.25722356");
    add_srs_wkt (p, 2,
		 "3,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"632");
    add_srs_wkt (p, 3,
		 "6\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Mercator_1");
    add_srs_wkt (p, 7,
		 "SP\"],PARAMETER[\"central_meridian\",0],PARAMETER[\"scal");
    add_srs_wkt (p, 8,
		 "e_factor\",1],PARAMETER[\"false_easting\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3395\"],AXIS[\"");
    add_srs_wkt (p, 10, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3396, "epsg", 3396,
		      "PD/83 / 3-degree Gauss-Kruger zone 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=9 +k=1 +x_0=3500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PD/83 / 3-degree Gauss-Kruger zone 3\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "PD/83\",DATUM[\"Potsdam_Datum_83\",SPHEROID[\"Bessel 184");
    add_srs_wkt (p, 2,
		 "1\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6746\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4746\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",9],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",3");
    add_srs_wkt (p, 10,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 11, "\",\"3396\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3397, "epsg", 3397,
		      "PD/83 / 3-degree Gauss-Kruger zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=12 +k=1 +x_0=4500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PD/83 / 3-degree Gauss-Kruger zone 4\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "PD/83\",DATUM[\"Potsdam_Datum_83\",SPHEROID[\"Bessel 184");
    add_srs_wkt (p, 2,
		 "1\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6746\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4746\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",12],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 10,
		 "4500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 11, "G\",\"3397\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3398, "epsg", 3398,
		      "RD/83 / 3-degree Gauss-Kruger zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=12 +k=1 +x_0=4500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RD/83 / 3-degree Gauss-Kruger zone 4\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "RD/83\",DATUM[\"Rauenberg_Datum_83\",SPHEROID[\"Bessel 1");
    add_srs_wkt (p, 2,
		 "841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6745\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4745\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"lat");
    add_srs_wkt (p, 8,
		 "itude_of_origin\",0],PARAMETER[\"central_meridian\",12],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",4500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 11, "SG\",\"3398\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3399, "epsg", 3399,
		      "RD/83 / 3-degree Gauss-Kruger zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=1 +x_0=5500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RD/83 / 3-degree Gauss-Kruger zone 5\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "RD/83\",DATUM[\"Rauenberg_Datum_83\",SPHEROID[\"Bessel 1");
    add_srs_wkt (p, 2,
		 "841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6745\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4745\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"lat");
    add_srs_wkt (p, 8,
		 "itude_of_origin\",0],PARAMETER[\"central_meridian\",15],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",5500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 11, "SG\",\"3399\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3400, "epsg", 3400,
		      "NAD83 / Alberta 10-TM (Forest)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-115 +k=0.9992 +x_0=500000 +");
    add_proj4text (p, 1, "y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alberta 10-TM (Forest)\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 2,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-115],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"scale_factor\",0.9992],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3400\"],AXIS[\"Easting\",EAST],AXIS[\"Northi");
    add_srs_wkt (p, 12, "ng\",NORTH]]");
    p = add_epsg_def (first, last, 3401, "epsg", 3401,
		      "NAD83 / Alberta 10-TM (Resource)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-115 +k=0.9992 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alberta 10-TM (Resource)\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS ");
    add_srs_wkt (p, 2,
		 "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"lat");
    add_srs_wkt (p, 8,
		 "itude_of_origin\",0],PARAMETER[\"central_meridian\",-115");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"scale_factor\",0.9992],PARAMETER[\"false_e");
    add_srs_wkt (p, 10,
		 "asting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"3401\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 3402, "epsg", 3402,
		      "NAD83(CSRS) / Alberta 10-TM (Forest)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-115 +k=0.9992 +x_0=500000 +");
    add_proj4text (p, 1, "y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / Alberta 10-TM (Forest)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(CSRS)\",DATUM[\"NAD83_Canadian_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-115],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 10,
		 "92],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"3402\"],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3403, "epsg", 3403,
		      "NAD83(CSRS) / Alberta 10-TM (Resource)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-115 +k=0.9992 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / Alberta 10-TM (Resource)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(CSRS)\",DATUM[\"NAD83_Canadian_Spatial_Referenc");
    add_srs_wkt (p, 2,
		 "e_System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-115],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 10,
		 "9992],PARAMETER[\"false_easting\",0],PARAMETER[\"false_n");
    add_srs_wkt (p, 11,
		 "orthing\",0],AUTHORITY[\"EPSG\",\"3403\"],AXIS[\"Easting");
    add_srs_wkt (p, 12, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3404, "epsg", 3404,
		      "NAD83(HARN) / North Carolina (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.16666666666666 +lat_2=34.33333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=33.75 +lon_0=-79 +x_0=609601.2192024384 +y_0=");
    add_proj4text (p, 2, "0 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / North Carolina (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",36.16666666666666],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",34.33333333333334],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",33.75],PARAMETER[\"central_meridian\",-79],PARA");
    add_srs_wkt (p, 12,
		 "METER[\"false_easting\",2000000],PARAMETER[\"false_north");
    add_srs_wkt (p, 13,
		 "ing\",0],AUTHORITY[\"EPSG\",\"3404\"],AXIS[\"X\",EAST],A");
    add_srs_wkt (p, 14, "XIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3405, "epsg", 3405,
		      "VN-2000 / UTM zone 48N");
    add_proj4text (p, 0, "+proj=utm +zone=48 +ellps=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"VN-2000 / UTM zone 48N\",GEOGCS[\"VN-2000\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Vietnam_2000\",SPHEROID[\"WGS 84\",6378137,298.2572");
    add_srs_wkt (p, 2,
		 "23563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "6756\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 4,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 5,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4756\"]],UNIT[\"met");
    add_srs_wkt (p, 6,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 7,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 8,
		 "AMETER[\"central_meridian\",105],PARAMETER[\"scale_facto");
    add_srs_wkt (p, 9,
		 "r\",0.9996],PARAMETER[\"false_easting\",500000],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3405\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3406, "epsg", 3406,
		      "VN-2000 / UTM zone 49N");
    add_proj4text (p, 0, "+proj=utm +zone=49 +ellps=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"VN-2000 / UTM zone 49N\",GEOGCS[\"VN-2000\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Vietnam_2000\",SPHEROID[\"WGS 84\",6378137,298.2572");
    add_srs_wkt (p, 2,
		 "23563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "6756\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 4,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 5,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4756\"]],UNIT[\"met");
    add_srs_wkt (p, 6,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 7,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 8,
		 "AMETER[\"central_meridian\",111],PARAMETER[\"scale_facto");
    add_srs_wkt (p, 9,
		 "r\",0.9996],PARAMETER[\"false_easting\",500000],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3406\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3407, "epsg", 3407,
		      "Hong Kong 1963 Grid System");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=22.31213333333334 +lon_0=114.178555555");
    add_proj4text (p, 1,
		   "5556 +x_0=40243.57775604237 +y_0=19069.93351512578 +a=63");
    add_proj4text (p, 2,
		   "78293.645208759 +b=6356617.987679838 +to_meter=0.3047972");
    add_proj4text (p, 3, "654 +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Hong Kong 1963 Grid System\",GEOGCS[\"Hong Kong");
    add_srs_wkt (p, 1,
		 " 1963\",DATUM[\"Hong_Kong_1963\",SPHEROID[\"Clarke 1858\"");
    add_srs_wkt (p, 2,
		 ",6378293.645208759,294.2606763692569,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"7007\"]],AUTHORITY[\"EPSG\",\"6738\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"4738\"]],UNIT[\"Clarke's foot\",0.304797265");
    add_srs_wkt (p, 7,
		 "4,AUTHORITY[\"EPSG\",\"9005\"]],PROJECTION[\"Cassini_Sol");
    add_srs_wkt (p, 8,
		 "dner\"],PARAMETER[\"latitude_of_origin\",22.312133333333");
    add_srs_wkt (p, 9,
		 "34],PARAMETER[\"central_meridian\",114.1785555555556],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_easting\",132033.92],PARAMETER[\"false_n");
    add_srs_wkt (p, 11,
		 "orthing\",62565.96],AUTHORITY[\"EPSG\",\"3407\"],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 3408, "epsg", 3408, "unnamed");
    add_proj4text (p, 0,
		   "+proj=laea +lat_0=90 +lon_0=0 +x_0=0 +y_0=0 +a=6371228 +");
    add_proj4text (p, 1, "b=6371228 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"unnamed\",GEOGCS[\"unnamed ellipse\",DATUM[\"un");
    add_srs_wkt (p, 1,
		 "known\",SPHEROID[\"unnamed\",6371228,0]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 2,
		 "ich\",0],UNIT[\"degree\",0.0174532925199433]],PROJECTION");
    add_srs_wkt (p, 3,
		 "[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 4,
		 "of_center\",90],PARAMETER[\"longitude_of_center\",0],PAR");
    add_srs_wkt (p, 5,
		 "AMETER[\"false_easting\",0],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 6, ",0],UNIT[\"Meter\",1],AUTHORITY[\"EPSG\",\"3408\"]]");
    p = add_epsg_def (first, last, 3409, "epsg", 3409, "unnamed");
    add_proj4text (p, 0,
		   "+proj=laea +lat_0=-90 +lon_0=0 +x_0=0 +y_0=0 +a=6371228 ");
    add_proj4text (p, 1, "+b=6371228 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"unnamed\",GEOGCS[\"unnamed ellipse\",DATUM[\"un");
    add_srs_wkt (p, 1,
		 "known\",SPHEROID[\"unnamed\",6371228,0]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 2,
		 "ich\",0],UNIT[\"degree\",0.0174532925199433]],PROJECTION");
    add_srs_wkt (p, 3,
		 "[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 4,
		 "of_center\",-90],PARAMETER[\"longitude_of_center\",0],PA");
    add_srs_wkt (p, 5,
		 "RAMETER[\"false_easting\",0],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 6, ",0],UNIT[\"Meter\",1],AUTHORITY[\"EPSG\",\"3409\"]]");
    p = add_epsg_def (first, last, 3410, "epsg", 3410,
		      "NSIDC EASE-Grid Global");
    add_proj4text (p, 0,
		   "+proj=cea +lon_0=0 +lat_ts=30 +x_0=0 +y_0=0 +a=6371228 +");
    add_proj4text (p, 1, "b=6371228 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NSIDC EASE-Grid Global\",GEOGCS[\"Unspecified d");
    add_srs_wkt (p, 1,
		 "atum based upon the International 1924 Authalic Sphere\"");
    add_srs_wkt (p, 2,
		 ",DATUM[\"Not_specified_based_on_International_1924_Autha");
    add_srs_wkt (p, 3,
		 "lic_Sphere\",SPHEROID[\"International 1924 Authalic Sphe");
    add_srs_wkt (p, 4,
		 "re\",6371228,0,AUTHORITY[\"EPSG\",\"7057\"]],AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"6053\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 6,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4053\"]],U");
    add_srs_wkt (p, 8,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 9,
		 "[\"Cylindrical_Equal_Area\"],PARAMETER[\"standard_parall");
    add_srs_wkt (p, 10,
		 "el_1\",30],PARAMETER[\"central_meridian\",0],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",0],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 12,
		 "RITY[\"EPSG\",\"3410\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORT");
    add_srs_wkt (p, 13, "H]]");
    p = add_epsg_def (first, last, 3411, "epsg", 3411,
		      "NSIDC Sea Ice Polar Stereographic North");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=90 +lat_ts=70 +lon_0=-45 +k=1 +x_0=0 ");
    add_proj4text (p, 1, "+y_0=0 +a=6378273 +b=6356889.449 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NSIDC Sea Ice Polar Stereographic North\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"Unspecified datum based upon the Hughes 1980 ellipso");
    add_srs_wkt (p, 2,
		 "id\",DATUM[\"Not_specified_based_on_Hughes_1980_ellipsoi");
    add_srs_wkt (p, 3,
		 "d\",SPHEROID[\"Hughes 1980\",6378273,298.279411123061,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"7058\"]],AUTHORITY[\"EPSG\",\"6054\"]");
    add_srs_wkt (p, 5,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4054\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 8,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_Stereog");
    add_srs_wkt (p, 9,
		 "raphic\"],PARAMETER[\"latitude_of_origin\",70],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"central_meridian\",-45],PARAMETER[\"scale_factor\",1]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_easting\",0],PARAMETER[\"false_northi");
    add_srs_wkt (p, 12,
		 "ng\",0],AUTHORITY[\"EPSG\",\"3411\"],AXIS[\"X\",UNKNOWN]");
    add_srs_wkt (p, 13, ",AXIS[\"Y\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3412, "epsg", 3412,
		      "NSIDC Sea Ice Polar Stereographic South");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-70 +lon_0=0 +k=1 +x_0=0 ");
    add_proj4text (p, 1, "+y_0=0 +a=6378273 +b=6356889.449 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NSIDC Sea Ice Polar Stereographic South\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"Unspecified datum based upon the Hughes 1980 ellipso");
    add_srs_wkt (p, 2,
		 "id\",DATUM[\"Not_specified_based_on_Hughes_1980_ellipsoi");
    add_srs_wkt (p, 3,
		 "d\",SPHEROID[\"Hughes 1980\",6378273,298.279411123061,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"7058\"]],AUTHORITY[\"EPSG\",\"6054\"]");
    add_srs_wkt (p, 5,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4054\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 8,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_Stereog");
    add_srs_wkt (p, 9,
		 "raphic\"],PARAMETER[\"latitude_of_origin\",-70],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"central_meridian\",0],PARAMETER[\"scale_factor\",1],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_easting\",0],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",0],AUTHORITY[\"EPSG\",\"3412\"],AXIS[\"X\",UNKNOWN],");
    add_srs_wkt (p, 13, "AXIS[\"Y\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3413, "epsg", 3413,
		      "WGS 84 / NSIDC Sea Ice Polar Stereographic North");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=90 +lat_ts=70 +lon_0=-45 +k=1 +x_0=0 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=WGS84 +datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / NSIDC Sea Ice Polar Stereographic Nort");
    add_srs_wkt (p, 1,
		 "h\",GEOGCS[\"WGS 84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS ");
    add_srs_wkt (p, 2,
		 "84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4326\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Polar_Stereographic\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",70],PARAMETER[\"central_meridian\",-45],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "3413\"],AXIS[\"X\",UNKNOWN],AXIS[\"Y\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3414, "epsg", 3414, "SVY21 / Singapore TM");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=1.366666666666667 +lon_0=103.83333333");
    add_proj4text (p, 1,
		   "33333 +k=1 +x_0=28001.642 +y_0=38744.572 +ellps=WGS84 +u");
    add_proj4text (p, 2, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SVY21 / Singapore TM\",GEOGCS[\"SVY21\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "SVY21\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHOR");
    add_srs_wkt (p, 2,
		 "ITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6757\"]],PR");
    add_srs_wkt (p, 3,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 4,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 5,
		 "\"]],AUTHORITY[\"EPSG\",\"4757\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 6,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 7,
		 "or\"],PARAMETER[\"latitude_of_origin\",1.366666666666667");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"central_meridian\",103.8333333333333],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",28");
    add_srs_wkt (p, 10,
		 "001.642],PARAMETER[\"false_northing\",38744.572],AUTHORI");
    add_srs_wkt (p, 11,
		 "TY[\"EPSG\",\"3414\"],AXIS[\"Northing\",NORTH],AXIS[\"Ea");
    add_srs_wkt (p, 12, "sting\",EAST]]");
    p = add_epsg_def (first, last, 3415, "epsg", 3415,
		      "WGS 72BE / South China Sea Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=18 +lat_2=24 +lat_0=21 +lon_0=114 +x_0=");
    add_proj4text (p, 1,
		   "500000 +y_0=500000 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 2, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / South China Sea Lambert\",GEOGCS[\"W");
    add_srs_wkt (p, 1,
		 "GS 72BE\",DATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7043\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EP");
    add_srs_wkt (p, 4,
		 "SG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 8,
		 "Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_para");
    add_srs_wkt (p, 9,
		 "llel_1\",18],PARAMETER[\"standard_parallel_2\",24],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"latitude_of_origin\",21],PARAMETER[\"central_meri");
    add_srs_wkt (p, 11,
		 "dian\",114],PARAMETER[\"false_easting\",500000],PARAMETE");
    add_srs_wkt (p, 12,
		 "R[\"false_northing\",500000],AUTHORITY[\"EPSG\",\"3415\"");
    add_srs_wkt (p, 13, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3416, "epsg", 3416,
		      "ETRS89 / Austria Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49 +lat_2=46 +lat_0=47.5 +lon_0=13.3333");
    add_proj4text (p, 1,
		   "3333333333 +x_0=400000 +y_0=400000 +ellps=GRS80 +units=m");
    add_proj4text (p, 2, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / Austria Lambert\",GEOGCS[\"ETRS89\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"European_Terrestrial_Reference_System_1989\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",49],PARAMETER[\"stan");
    add_srs_wkt (p, 9,
		 "dard_parallel_2\",46],PARAMETER[\"latitude_of_origin\",4");
    add_srs_wkt (p, 10,
		 "7.5],PARAMETER[\"central_meridian\",13.33333333333333],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_easting\",400000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 12,
		 "thing\",400000],AUTHORITY[\"EPSG\",\"3416\"],AXIS[\"X\",");
    add_srs_wkt (p, 13, "NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3417, "epsg", 3417,
		      "NAD83 / Iowa North (ft US)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43.26666666666667 +lat_2=42.06666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=41.5 +lon_0=-93.5 +x_0=1500000 +y_0=999999.99");
    add_proj4text (p, 2,
		   "99898402 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_proj4text (p, 3, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Iowa North (ft US)\",GEOGCS[\"NAD83\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal_");
    add_srs_wkt (p, 8,
		 "Conic_2SP\"],PARAMETER[\"standard_parallel_1\",43.266666");
    add_srs_wkt (p, 9,
		 "66666667],PARAMETER[\"standard_parallel_2\",42.066666666");
    add_srs_wkt (p, 10,
		 "66667],PARAMETER[\"latitude_of_origin\",41.5],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"central_meridian\",-93.5],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 12,
		 "4921250],PARAMETER[\"false_northing\",3280833.333300001]");
    add_srs_wkt (p, 13,
		 ",AUTHORITY[\"EPSG\",\"3417\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 14, ",NORTH]]");
    p = add_epsg_def (first, last, 3418, "epsg", 3418,
		      "NAD83 / Iowa South (ft US)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.78333333333333 +lat_2=40.61666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=40 +lon_0=-93.5 +x_0=500000.00001016 +y_0=0 +");
    add_proj4text (p, 2, "ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Iowa South (ft US)\",GEOGCS[\"NAD83\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal_");
    add_srs_wkt (p, 8,
		 "Conic_2SP\"],PARAMETER[\"standard_parallel_1\",41.783333");
    add_srs_wkt (p, 9,
		 "33333333],PARAMETER[\"standard_parallel_2\",40.616666666");
    add_srs_wkt (p, 10,
		 "66667],PARAMETER[\"latitude_of_origin\",40],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "central_meridian\",-93.5],PARAMETER[\"false_easting\",16");
    add_srs_wkt (p, 12,
		 "40416.6667],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13, "EPSG\",\"3418\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3419, "epsg", 3419,
		      "NAD83 / Kansas North (ft US)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.78333333333333 +lat_2=38.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=38.33333333333334 +lon_0=-98 +x_0=399999.9999");
    add_proj4text (p, 2,
		   "8984 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_d");
    add_proj4text (p, 3, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Kansas North (ft US)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",39.78333");
    add_srs_wkt (p, 9,
		 "333333333],PARAMETER[\"standard_parallel_2\",38.71666666");
    add_srs_wkt (p, 10,
		 "666667],PARAMETER[\"latitude_of_origin\",38.333333333333");
    add_srs_wkt (p, 11,
		 "34],PARAMETER[\"central_meridian\",-98],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_easting\",1312333.3333],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 13,
		 "],AUTHORITY[\"EPSG\",\"3419\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 14, "\",NORTH]]");
    p = add_epsg_def (first, last, 3420, "epsg", 3420,
		      "NAD83 / Kansas South (ft US)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.56666666666667 +lat_2=37.26666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.66666666666666 +lon_0=-98.5 +x_0=399999.99");
    add_proj4text (p, 2,
		   "998984 +y_0=399999.99998984 +ellps=GRS80 +datum=NAD83 +u");
    add_proj4text (p, 3, "nits=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Kansas South (ft US)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",38.56666");
    add_srs_wkt (p, 9,
		 "666666667],PARAMETER[\"standard_parallel_2\",37.26666666");
    add_srs_wkt (p, 10,
		 "666667],PARAMETER[\"latitude_of_origin\",36.666666666666");
    add_srs_wkt (p, 11,
		 "66],PARAMETER[\"central_meridian\",-98.5],PARAMETER[\"fa");
    add_srs_wkt (p, 12,
		 "lse_easting\",1312333.3333],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 13,
		 ",1312333.3333],AUTHORITY[\"EPSG\",\"3420\"],AXIS[\"X\",E");
    add_srs_wkt (p, 14, "AST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3421, "epsg", 3421,
		      "NAD83 / Nevada East (ft US)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-115.5833333333333 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=200000.00001016 +y_0=8000000.000010163 +ellps=");
    add_proj4text (p, 2, "GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Nevada East (ft US)\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercato");
    add_srs_wkt (p, 8,
		 "r\"],PARAMETER[\"latitude_of_origin\",34.75],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-115.5833333333333],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.9999],PARAMETER[\"false_easting\",656166.666");
    add_srs_wkt (p, 11,
		 "7],PARAMETER[\"false_northing\",26246666.66670001],AUTHO");
    add_srs_wkt (p, 12,
		 "RITY[\"EPSG\",\"3421\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORT");
    add_srs_wkt (p, 13, "H]]");
    p = add_epsg_def (first, last, 3422, "epsg", 3422,
		      "NAD83 / Nevada Central (ft US)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-116.6666666666667 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=500000.00001016 +y_0=6000000 +ellps=GRS80 +dat");
    add_proj4text (p, 2, "um=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Nevada Central (ft US)\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 2,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",34.75],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-116.6666666666667],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",16404");
    add_srs_wkt (p, 11,
		 "16.6667],PARAMETER[\"false_northing\",19685000],AUTHORIT");
    add_srs_wkt (p, 12,
		 "Y[\"EPSG\",\"3422\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 3423, "epsg", 3423,
		      "NAD83 / Nevada West (ft US)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-118.5833333333333 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=800000.0000101599 +y_0=3999999.99998984 +ellps");
    add_proj4text (p, 2, "=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Nevada West (ft US)\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercato");
    add_srs_wkt (p, 8,
		 "r\"],PARAMETER[\"latitude_of_origin\",34.75],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-118.5833333333333],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.9999],PARAMETER[\"false_easting\",2624666.66");
    add_srs_wkt (p, 11,
		 "67],PARAMETER[\"false_northing\",13123333.3333],AUTHORIT");
    add_srs_wkt (p, 12,
		 "Y[\"EPSG\",\"3423\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 3424, "epsg", 3424,
		      "NAD83 / New Jersey (ft US)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38.83333333333334 +lon_0=-74.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=150000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=");
    add_proj4text (p, 2, "us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / New Jersey (ft US)\",GEOGCS[\"NAD83\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercato");
    add_srs_wkt (p, 8,
		 "r\"],PARAMETER[\"latitude_of_origin\",38.83333333333334]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"central_meridian\",-74.5],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.9999],PARAMETER[\"false_easting\",492125],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3424\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3425, "epsg", 3425,
		      "NAD83(HARN) / Iowa North (ft US)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43.26666666666667 +lat_2=42.06666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=41.5 +lon_0=-93.5 +x_0=1500000 +y_0=999999.99");
    add_proj4text (p, 2, "99898402 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Iowa North (ft US)\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",0.30");
    add_srs_wkt (p, 7,
		 "48006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 9,
		 "rallel_1\",43.26666666666667],PARAMETER[\"standard_paral");
    add_srs_wkt (p, 10,
		 "lel_2\",42.06666666666667],PARAMETER[\"latitude_of_origi");
    add_srs_wkt (p, 11,
		 "n\",41.5],PARAMETER[\"central_meridian\",-93.5],PARAMETE");
    add_srs_wkt (p, 12,
		 "R[\"false_easting\",4921250],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 13,
		 ",3280833.333300001],AUTHORITY[\"EPSG\",\"3425\"],AXIS[\"");
    add_srs_wkt (p, 14, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3426, "epsg", 3426,
		      "NAD83(HARN) / Iowa South (ft US)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.78333333333333 +lat_2=40.61666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=40 +lon_0=-93.5 +x_0=500000.00001016 +y_0=0 +");
    add_proj4text (p, 2, "ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Iowa South (ft US)\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",0.30");
    add_srs_wkt (p, 7,
		 "48006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 9,
		 "rallel_1\",41.78333333333333],PARAMETER[\"standard_paral");
    add_srs_wkt (p, 10,
		 "lel_2\",40.61666666666667],PARAMETER[\"latitude_of_origi");
    add_srs_wkt (p, 11,
		 "n\",40],PARAMETER[\"central_meridian\",-93.5],PARAMETER[");
    add_srs_wkt (p, 12,
		 "\"false_easting\",1640416.6667],PARAMETER[\"false_northi");
    add_srs_wkt (p, 13,
		 "ng\",0],AUTHORITY[\"EPSG\",\"3426\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 14, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3427, "epsg", 3427,
		      "NAD83(HARN) / Kansas North (ft US)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.78333333333333 +lat_2=38.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=38.33333333333334 +lon_0=-98 +x_0=399999.9999");
    add_proj4text (p, 2, "8984 +y_0=0 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Kansas North (ft US)\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Networ");
    add_srs_wkt (p, 2,
		 "k\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",39.78333333333333],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",38.71666666666667],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",38.33333333333334],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 12,
		 "\",-98],PARAMETER[\"false_easting\",1312333.3333],PARAME");
    add_srs_wkt (p, 13,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3427\"],A");
    add_srs_wkt (p, 14, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3428, "epsg", 3428,
		      "NAD83(HARN) / Kansas South (ft US)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.56666666666667 +lat_2=37.26666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.66666666666666 +lon_0=-98.5 +x_0=399999.99");
    add_proj4text (p, 2,
		   "998984 +y_0=399999.99998984 +ellps=GRS80 +units=us-ft +n");
    add_proj4text (p, 3, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Kansas South (ft US)\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Networ");
    add_srs_wkt (p, 2,
		 "k\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",38.56666666666667],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",37.26666666666667],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",36.66666666666666],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 12,
		 "\",-98.5],PARAMETER[\"false_easting\",1312333.3333],PARA");
    add_srs_wkt (p, 13,
		 "METER[\"false_northing\",1312333.3333],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 14, ",\"3428\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3429, "epsg", 3429,
		      "NAD83(HARN) / Nevada East (ft US)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-115.5833333333333 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=200000.00001016 +y_0=8000000.000010163 +ellps=");
    add_proj4text (p, 2, "GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Nevada East (ft US)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",");
    add_srs_wkt (p, 7,
		 "0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECT");
    add_srs_wkt (p, 8,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 9,
		 "gin\",34.75],PARAMETER[\"central_meridian\",-115.5833333");
    add_srs_wkt (p, 10,
		 "333333],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",656166.6667],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",26246666.66670001],AUTHORITY[\"EPSG\",\"3429\"],AXIS[\"");
    add_srs_wkt (p, 13, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3430, "epsg", 3430,
		      "NAD83(HARN) / Nevada Central (ft US)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-116.6666666666667 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=500000.00001016 +y_0=6000000 +ellps=GRS80 +uni");
    add_proj4text (p, 2, "ts=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Nevada Central (ft US)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 9,
		 "igin\",34.75],PARAMETER[\"central_meridian\",-116.666666");
    add_srs_wkt (p, 10,
		 "6666667],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",1640416.6667],PARAMETER[\"false_northing");
    add_srs_wkt (p, 12,
		 "\",19685000],AUTHORITY[\"EPSG\",\"3430\"],AXIS[\"X\",EAS");
    add_srs_wkt (p, 13, "T],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3431, "epsg", 3431,
		      "NAD83(HARN) / Nevada West (ft US)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-118.5833333333333 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=800000.0000101599 +y_0=3999999.99998984 +ellps");
    add_proj4text (p, 2, "=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Nevada West (ft US)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",");
    add_srs_wkt (p, 7,
		 "0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECT");
    add_srs_wkt (p, 8,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 9,
		 "gin\",34.75],PARAMETER[\"central_meridian\",-118.5833333");
    add_srs_wkt (p, 10,
		 "333333],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",2624666.6667],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",13123333.3333],AUTHORITY[\"EPSG\",\"3431\"],AXIS[\"X\",");
    add_srs_wkt (p, 13, "EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3432, "epsg", 3432,
		      "NAD83(HARN) / New Jersey (ft US)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38.83333333333334 +lon_0=-74.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=150000 +y_0=0 +ellps=GRS80 +units=us-ft +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / New Jersey (ft US)\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",0.30");
    add_srs_wkt (p, 7,
		 "48006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 9,
		 ",38.83333333333334],PARAMETER[\"central_meridian\",-74.5");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",492125],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"3432\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]");
    add_srs_wkt (p, 13, "]");
    p = add_epsg_def (first, last, 3433, "epsg", 3433,
		      "NAD83 / Arkansas North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.23333333333333 +lat_2=34.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=34.33333333333334 +lon_0=-92 +x_0=399999.9999");
    add_proj4text (p, 2,
		   "8984 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_d");
    add_proj4text (p, 3, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Arkansas North (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",36.23333");
    add_srs_wkt (p, 9,
		 "333333333],PARAMETER[\"standard_parallel_2\",34.93333333");
    add_srs_wkt (p, 10,
		 "333333],PARAMETER[\"latitude_of_origin\",34.333333333333");
    add_srs_wkt (p, 11,
		 "34],PARAMETER[\"central_meridian\",-92],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_easting\",1312333.3333],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 13,
		 "],AUTHORITY[\"EPSG\",\"3433\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 14, "\",NORTH]]");
    p = add_epsg_def (first, last, 3434, "epsg", 3434,
		      "NAD83 / Arkansas South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=34.76666666666667 +lat_2=33.3 +lat_0=32");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-92 +x_0=399999.99998984 +y_0=399");
    add_proj4text (p, 2,
		   "999.99998984 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_");
    add_proj4text (p, 3, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Arkansas South (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",34.76666");
    add_srs_wkt (p, 9,
		 "666666667],PARAMETER[\"standard_parallel_2\",33.3],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"latitude_of_origin\",32.66666666666666],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"central_meridian\",-92],PARAMETER[\"false_easting\",1");
    add_srs_wkt (p, 12,
		 "312333.3333],PARAMETER[\"false_northing\",1312333.3333],");
    add_srs_wkt (p, 13,
		 "AUTHORITY[\"EPSG\",\"3434\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 14, ",NORTH]]");
    p = add_epsg_def (first, last, 3435, "epsg", 3435,
		      "NAD83 / Illinois East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36.66666666666666 +lon_0=-88.33333333");
    add_proj4text (p, 1,
		   "333333 +k=0.9999749999999999 +x_0=300000.0000000001 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Illinois East (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",36.66666666666666");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-88.33333333333333],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",0.999975],PARAMETER[\"false_east");
    add_srs_wkt (p, 11,
		 "ing\",984250.0000000002],PARAMETER[\"false_northing\",0]");
    add_srs_wkt (p, 12,
		 ",AUTHORITY[\"EPSG\",\"3435\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3436, "epsg", 3436,
		      "NAD83 / Illinois West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36.66666666666666 +lon_0=-90.16666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999941177 +x_0=699999.9999898402 +y_0=0 +ell");
    add_proj4text (p, 2, "ps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Illinois West (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",36.66666666666666");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-90.16666666666667],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",0.999941177],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",2296583.333300001],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"3436\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3437, "epsg", 3437,
		      "NAD83 / New Hampshire (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.5 +lon_0=-71.66666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=300000.0000000001 +y_0=0 +ellps=GRS80 +dat");
    add_proj4text (p, 2, "um=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / New Hampshire (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",42.5],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-71.66666666666667],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.999966667],PARAMETER[\"false_easting\",98425");
    add_srs_wkt (p, 11,
		 "0.0000000002],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 12,
		 "\"EPSG\",\"3437\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3438, "epsg", 3438,
		      "NAD83 / Rhode Island (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.08333333333334 +lon_0=-71.5 +k=0.9");
    add_proj4text (p, 1,
		   "9999375 +x_0=99999.99998983997 +y_0=0 +ellps=GRS80 +datu");
    add_proj4text (p, 2, "m=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Rhode Island (ftUS)\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercato");
    add_srs_wkt (p, 8,
		 "r\"],PARAMETER[\"latitude_of_origin\",41.08333333333334]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"central_meridian\",-71.5],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.99999375],PARAMETER[\"false_easting\",328083");
    add_srs_wkt (p, 11,
		 ".3333],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"3438\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3439, "epsg", 3439, "PSD93 / UTM zone 39N");
    add_proj4text (p, 0, "+proj=utm +zone=39 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PSD93 / UTM zone 39N\",GEOGCS[\"PSD93\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "PDO_Survey_Datum_1993\",SPHEROID[\"Clarke 1880 (RGS)\",6");
    add_srs_wkt (p, 2,
		 "378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6134\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4134\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 8,
		 "igin\",0],PARAMETER[\"central_meridian\",51],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "3439\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 3440, "epsg", 3440, "PSD93 / UTM zone 40N");
    add_proj4text (p, 0, "+proj=utm +zone=40 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PSD93 / UTM zone 40N\",GEOGCS[\"PSD93\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "PDO_Survey_Datum_1993\",SPHEROID[\"Clarke 1880 (RGS)\",6");
    add_srs_wkt (p, 2,
		 "378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6134\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4134\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 8,
		 "igin\",0],PARAMETER[\"central_meridian\",57],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "3440\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 3441, "epsg", 3441,
		      "NAD83(HARN) / Arkansas North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.23333333333333 +lat_2=34.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=34.33333333333334 +lon_0=-92 +x_0=399999.9999");
    add_proj4text (p, 2, "8984 +y_0=0 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Arkansas North (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",36.23333333333333],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",34.93333333333333],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",34.33333333333334],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 12,
		 "\",-92],PARAMETER[\"false_easting\",1312333.3333],PARAME");
    add_srs_wkt (p, 13,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3441\"],A");
    add_srs_wkt (p, 14, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3442, "epsg", 3442,
		      "NAD83(HARN) / Arkansas South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=34.76666666666667 +lat_2=33.3 +lat_0=32");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-92 +x_0=399999.99998984 +y_0=399");
    add_proj4text (p, 2, "999.99998984 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Arkansas South (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",34.76666666666667],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",33.3],PARAMETER[\"latitude_of_origin\",32.6");
    add_srs_wkt (p, 11,
		 "6666666666666],PARAMETER[\"central_meridian\",-92],PARAM");
    add_srs_wkt (p, 12,
		 "ETER[\"false_easting\",1312333.3333],PARAMETER[\"false_n");
    add_srs_wkt (p, 13,
		 "orthing\",1312333.3333],AUTHORITY[\"EPSG\",\"3442\"],AXI");
    add_srs_wkt (p, 14, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3443, "epsg", 3443,
		      "NAD83(HARN) / Illinois East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36.66666666666666 +lon_0=-88.33333333");
    add_proj4text (p, 1,
		   "333333 +k=0.9999749999999999 +x_0=300000.0000000001 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Illinois East (ftUS)\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Networ");
    add_srs_wkt (p, 2,
		 "k\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 9,
		 "igin\",36.66666666666666],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 10,
		 ",-88.33333333333333],PARAMETER[\"scale_factor\",0.999975");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"false_easting\",984250.0000000002],PARAMET");
    add_srs_wkt (p, 12,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3443\"],AX");
    add_srs_wkt (p, 13, "IS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3444, "epsg", 3444,
		      "NAD83(HARN) / Illinois West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36.66666666666666 +lon_0=-90.16666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999941177 +x_0=699999.9999898402 +y_0=0 +ell");
    add_proj4text (p, 2, "ps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Illinois West (ftUS)\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Networ");
    add_srs_wkt (p, 2,
		 "k\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 9,
		 "igin\",36.66666666666666],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 10,
		 ",-90.16666666666667],PARAMETER[\"scale_factor\",0.999941");
    add_srs_wkt (p, 11,
		 "177],PARAMETER[\"false_easting\",2296583.333300001],PARA");
    add_srs_wkt (p, 12,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3444\"]");
    add_srs_wkt (p, 13, ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3445, "epsg", 3445,
		      "NAD83(HARN) / New Hampshire (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.5 +lon_0=-71.66666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=300000.0000000001 +y_0=0 +ellps=GRS80 +uni");
    add_proj4text (p, 2, "ts=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / New Hampshire (ftUS)\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Networ");
    add_srs_wkt (p, 2,
		 "k\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 9,
		 "igin\",42.5],PARAMETER[\"central_meridian\",-71.66666666");
    add_srs_wkt (p, 10,
		 "666667],PARAMETER[\"scale_factor\",0.999966667],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_easting\",984250.0000000002],PARAMETER[\"false");
    add_srs_wkt (p, 12,
		 "_northing\",0],AUTHORITY[\"EPSG\",\"3445\"],AXIS[\"X\",E");
    add_srs_wkt (p, 13, "AST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3446, "epsg", 3446,
		      "NAD83(HARN) / Rhode Island (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.08333333333334 +lon_0=-71.5 +k=0.9");
    add_proj4text (p, 1,
		   "9999375 +x_0=99999.99998983997 +y_0=0 +ellps=GRS80 +unit");
    add_proj4text (p, 2, "s=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Rhode Island (ftUS)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",");
    add_srs_wkt (p, 7,
		 "0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECT");
    add_srs_wkt (p, 8,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 9,
		 "gin\",41.08333333333334],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 10,
		 "-71.5],PARAMETER[\"scale_factor\",0.99999375],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_easting\",328083.3333],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",0],AUTHORITY[\"EPSG\",\"3446\"],AXIS[\"X\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3447, "epsg", 3447,
		      "ETRS89 / Belgian Lambert 2005");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49.83333333333334 +lat_2=51.16666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=50.797815 +lon_0=4.359215833333333 +x_0=15032");
    add_proj4text (p, 2, "8 +y_0=166262 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / Belgian Lambert 2005\",GEOGCS[\"ETRS89");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"European_Terrestrial_Reference_System_1989\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",49.8333333333333");
    add_srs_wkt (p, 9,
		 "4],PARAMETER[\"standard_parallel_2\",51.16666666666666],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",50.797815],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "central_meridian\",4.359215833333333],PARAMETER[\"false_");
    add_srs_wkt (p, 12,
		 "easting\",150328],PARAMETER[\"false_northing\",166262],A");
    add_srs_wkt (p, 13,
		 "UTHORITY[\"EPSG\",\"3447\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 14, "NORTH]]");
    p = add_epsg_def (first, last, 3448, "epsg", 3448,
		      "JAD2001 / Jamaica Metric Grid");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=18 +lat_0=18 +lon_0=-77 +k_0=1 +x_0=750");
    add_proj4text (p, 1,
		   "000 +y_0=650000 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +uni");
    add_proj4text (p, 2, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JAD2001 / Jamaica Metric Grid\",GEOGCS[\"JAD200");
    add_srs_wkt (p, 1,
		 "1\",DATUM[\"Jamaica_2001\",SPHEROID[\"WGS 84\",6378137,2");
    add_srs_wkt (p, 2,
		 "98.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0");
    add_srs_wkt (p, 3,
		 ",0,0,0,0],AUTHORITY[\"EPSG\",\"6758\"]],PRIMEM[\"Greenwi");
    add_srs_wkt (p, 4,
		 "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0");
    add_srs_wkt (p, 5,
		 "1745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"4758\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1SP\"],");
    add_srs_wkt (p, 8,
		 "PARAMETER[\"latitude_of_origin\",18],PARAMETER[\"central");
    add_srs_wkt (p, 9,
		 "_meridian\",-77],PARAMETER[\"scale_factor\",1],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",750000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "650000],AUTHORITY[\"EPSG\",\"3448\"],AXIS[\"Easting\",EA");
    add_srs_wkt (p, 12, "ST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3449, "epsg", 3449,
		      "JAD2001 / UTM zone 17N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=17 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JAD2001 / UTM zone 17N\",GEOGCS[\"JAD2001\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Jamaica_2001\",SPHEROID[\"WGS 84\",6378137,298.2572");
    add_srs_wkt (p, 2,
		 "23563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,");
    add_srs_wkt (p, 3,
		 "0],AUTHORITY[\"EPSG\",\"6758\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4758\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"lat");
    add_srs_wkt (p, 8,
		 "itude_of_origin\",0],PARAMETER[\"central_meridian\",-81]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_ea");
    add_srs_wkt (p, 10,
		 "sting\",500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"3449\"],AXIS[\"Easting\",EAST],AXIS[\"North");
    add_srs_wkt (p, 12, "ing\",NORTH]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_11 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 3450, "epsg", 3450,
		      "JAD2001 / UTM zone 18N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JAD2001 / UTM zone 18N\",GEOGCS[\"JAD2001\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Jamaica_2001\",SPHEROID[\"WGS 84\",6378137,298.2572");
    add_srs_wkt (p, 2,
		 "23563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,");
    add_srs_wkt (p, 3,
		 "0],AUTHORITY[\"EPSG\",\"6758\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4758\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"lat");
    add_srs_wkt (p, 8,
		 "itude_of_origin\",0],PARAMETER[\"central_meridian\",-75]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_ea");
    add_srs_wkt (p, 10,
		 "sting\",500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"3450\"],AXIS[\"Easting\",EAST],AXIS[\"North");
    add_srs_wkt (p, 12, "ing\",NORTH]]");
    p = add_epsg_def (first, last, 3451, "epsg", 3451,
		      "NAD83 / Louisiana North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=32.66666666666666 +lat_2=31.16666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=30.5 +lon_0=-92.5 +x_0=999999.9999898402 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Louisiana North (ftUS)\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 2,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",32.66");
    add_srs_wkt (p, 9,
		 "666666666666],PARAMETER[\"standard_parallel_2\",31.16666");
    add_srs_wkt (p, 10,
		 "666666667],PARAMETER[\"latitude_of_origin\",30.5],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"central_meridian\",-92.5],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 12,
		 "g\",3280833.333300001],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 13,
		 "UTHORITY[\"EPSG\",\"3451\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 14, "NORTH]]");
    p = add_epsg_def (first, last, 3452, "epsg", 3452,
		      "NAD83 / Louisiana South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.7 +lat_2=29.3 +lat_0=28.5 +lon_0=-91");
    add_proj4text (p, 1,
		   ".33333333333333 +x_0=999999.9999898402 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Louisiana South (ftUS)\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 2,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",30.7]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"standard_parallel_2\",29.3],PARAMETER[\"lat");
    add_srs_wkt (p, 10,
		 "itude_of_origin\",28.5],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 11,
		 "91.33333333333333],PARAMETER[\"false_easting\",3280833.3");
    add_srs_wkt (p, 12,
		 "33300001],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 13, "SG\",\"3452\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3453, "epsg", 3453,
		      "NAD83 / Louisiana Offshore (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=27.83333333333333 +lat_2=26.16666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=25.5 +lon_0=-91.33333333333333 +x_0=999999.99");
    add_proj4text (p, 2,
		   "99898402 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=us-ft +");
    add_proj4text (p, 3, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Louisiana Offshore (ftUS)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",27.");
    add_srs_wkt (p, 9,
		 "83333333333333],PARAMETER[\"standard_parallel_2\",26.166");
    add_srs_wkt (p, 10,
		 "66666666667],PARAMETER[\"latitude_of_origin\",25.5],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-91.33333333333333],PARAMETER");
    add_srs_wkt (p, 12,
		 "[\"false_easting\",3280833.333300001],PARAMETER[\"false_");
    add_srs_wkt (p, 13,
		 "northing\",0],AUTHORITY[\"EPSG\",\"3453\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 14, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3454, "epsg", 3454,
		      "NAD83 / South Dakota North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.4 +lat_2=42.83333333333334 +lat_0=42");
    add_proj4text (p, 1,
		   ".33333333333334 +lon_0=-100.3333333333333 +x_0=600000 +y");
    add_proj4text (p, 2,
		   "_0=0 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / South Dakota North (ftUS)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",44.");
    add_srs_wkt (p, 9,
		 "4],PARAMETER[\"standard_parallel_2\",42.83333333333334],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",42.33333333333334],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-100.3333333333333],PARAMETER");
    add_srs_wkt (p, 12,
		 "[\"false_easting\",1968500],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 13,
		 ",0],AUTHORITY[\"EPSG\",\"3454\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 14, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3455, "epsg", 3455,
		      "NAD83 / South Dakota South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.4 +lat_2=42.83333333333334 +lat_0=42");
    add_proj4text (p, 1,
		   ".33333333333334 +lon_0=-100.3333333333333 +x_0=600000 +y");
    add_proj4text (p, 2,
		   "_0=0 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / South Dakota South (ftUS)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",44.");
    add_srs_wkt (p, 9,
		 "4],PARAMETER[\"standard_parallel_2\",42.83333333333334],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",42.33333333333334],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-100.3333333333333],PARAMETER");
    add_srs_wkt (p, 12,
		 "[\"false_easting\",1968500],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 13,
		 ",0],AUTHORITY[\"EPSG\",\"3455\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 14, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3456, "epsg", 3456,
		      "NAD83(HARN) / Louisiana North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=32.66666666666666 +lat_2=31.16666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=30.5 +lon_0=-92.5 +x_0=999999.9999898402 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Louisiana North (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",32.66666666666666],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",31.16666666666667],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",30.5],PARAMETER[\"central_meridian\",-92.5],PAR");
    add_srs_wkt (p, 12,
		 "AMETER[\"false_easting\",3280833.333300001],PARAMETER[\"");
    add_srs_wkt (p, 13,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3456\"],AXIS[\"");
    add_srs_wkt (p, 14, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3457, "epsg", 3457,
		      "NAD83(HARN) / Louisiana South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.7 +lat_2=29.3 +lat_0=28.5 +lon_0=-91");
    add_proj4text (p, 1,
		   ".33333333333333 +x_0=999999.9999898402 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Louisiana South (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",30.7],PARAMETER[\"standard_parallel_2\",");
    add_srs_wkt (p, 10,
		 "29.3],PARAMETER[\"latitude_of_origin\",28.5],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "central_meridian\",-91.33333333333333],PARAMETER[\"false");
    add_srs_wkt (p, 12,
		 "_easting\",3280833.333300001],PARAMETER[\"false_northing");
    add_srs_wkt (p, 13,
		 "\",0],AUTHORITY[\"EPSG\",\"3457\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 14, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3458, "epsg", 3458,
		      "NAD83(HARN) / South Dakota North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.68333333333333 +lat_2=44.41666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=43.83333333333334 +lon_0=-100 +x_0=600000 +y_");
    add_proj4text (p, 2, "0=0 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / South Dakota North (ftUS)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_N");
    add_srs_wkt (p, 2,
		 "etwork\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]]");
    add_srs_wkt (p, 4,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 5,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 6,
		 "122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey f");
    add_srs_wkt (p, 7,
		 "oot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_1\",45.68333333333333],PARAMETER[\"stan");
    add_srs_wkt (p, 10,
		 "dard_parallel_2\",44.41666666666666],PARAMETER[\"latitud");
    add_srs_wkt (p, 11,
		 "e_of_origin\",43.83333333333334],PARAMETER[\"central_mer");
    add_srs_wkt (p, 12,
		 "idian\",-100],PARAMETER[\"false_easting\",1968500],PARAM");
    add_srs_wkt (p, 13,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3458\"],");
    add_srs_wkt (p, 14, "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3459, "epsg", 3459,
		      "NAD83(HARN) / South Dakota South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.4 +lat_2=42.83333333333334 +lat_0=42");
    add_proj4text (p, 1,
		   ".33333333333334 +lon_0=-100.3333333333333 +x_0=600000 +y");
    add_proj4text (p, 2, "_0=0 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / South Dakota South (ftUS)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_N");
    add_srs_wkt (p, 2,
		 "etwork\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]]");
    add_srs_wkt (p, 4,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 5,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 6,
		 "122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey f");
    add_srs_wkt (p, 7,
		 "oot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_1\",44.4],PARAMETER[\"standard_parallel");
    add_srs_wkt (p, 10,
		 "_2\",42.83333333333334],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 11,
		 ",42.33333333333334],PARAMETER[\"central_meridian\",-100.");
    add_srs_wkt (p, 12,
		 "3333333333333],PARAMETER[\"false_easting\",1968500],PARA");
    add_srs_wkt (p, 13,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3459\"]");
    add_srs_wkt (p, 14, ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3460, "epsg", 3460,
		      "Fiji 1986 / Fiji Map Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-17 +lon_0=178.75 +k=0.99985 +x_0=200");
    add_proj4text (p, 1, "0000 +y_0=4000000 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Fiji 1986 / Fiji Map Grid\",GEOGCS[\"Fiji 1986\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Fiji_Geodetic_Datum_1986\",SPHEROID[\"WGS 72\",");
    add_srs_wkt (p, 2,
		 "6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6720\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4720\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",-17],PARAMETER[\"central_meridian\",178.75],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",0.99985],PARAMETER[\"false_easting\",2");
    add_srs_wkt (p, 10,
		 "000000],PARAMETER[\"false_northing\",4000000],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"3460\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 12, "g\",NORTH]]");
    p = add_epsg_def (first, last, 3461, "epsg", 3461,
		      "Dabola 1981 / UTM zone 28N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=28 +a=6378249.2 +b=6356515 +towgs84=-83,");
    add_proj4text (p, 1, "37,124,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Dabola 1981 / UTM zone 28N\",GEOGCS[\"Dabola 19");
    add_srs_wkt (p, 1,
		 "81\",DATUM[\"Dabola_1981\",SPHEROID[\"Clarke 1880 (IGN)\"");
    add_srs_wkt (p, 2,
		 ",6378249.2,293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[-83,37,124,0,0,0,0],AUTHORITY[\"EPSG\",\"6155");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4155\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-15],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3461\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3462, "epsg", 3462,
		      "Dabola 1981 / UTM zone 29N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=29 +a=6378249.2 +b=6356515 +towgs84=-83,");
    add_proj4text (p, 1, "37,124,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Dabola 1981 / UTM zone 29N\",GEOGCS[\"Dabola 19");
    add_srs_wkt (p, 1,
		 "81\",DATUM[\"Dabola_1981\",SPHEROID[\"Clarke 1880 (IGN)\"");
    add_srs_wkt (p, 2,
		 ",6378249.2,293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[-83,37,124,0,0,0,0],AUTHORITY[\"EPSG\",\"6155");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4155\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-9],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"3462\"],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3463, "epsg", 3463,
		      "NAD83 / Maine CS2000 Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=43.5 +lon_0=-69.125 +k=0.99998 +x_0=5");
    add_proj4text (p, 1,
		   "00000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Maine CS2000 Central\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",43.5],PARAMETER[\"central_meridian\",-69.12");
    add_srs_wkt (p, 9,
		 "5],PARAMETER[\"scale_factor\",0.99998],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 11,
		 "RITY[\"EPSG\",\"3463\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORT");
    add_srs_wkt (p, 12, "H]]");
    p = add_epsg_def (first, last, 3464, "epsg", 3464,
		      "NAD83(HARN) / Maine CS2000 Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=43.5 +lon_0=-69.125 +k=0.99998 +x_0=5");
    add_proj4text (p, 1, "00000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Maine CS2000 Central\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Networ");
    add_srs_wkt (p, 2,
		 "k\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",43.5],PARAMETER[\"c");
    add_srs_wkt (p, 9,
		 "entral_meridian\",-69.125],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 10,
		 "99998],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"3464\"],AXIS[\"X");
    add_srs_wkt (p, 12, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3465, "epsg", 3465,
		      "NAD83(NSRS2007) / Alabama East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30.5 +lon_0=-85.83333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "9996 +x_0=200000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Alabama East\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",30.5],PARAMETER[\"central_meridian\",-85.833");
    add_srs_wkt (p, 10,
		 "33333333333],PARAMETER[\"scale_factor\",0.99996],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_easting\",200000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"3465\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3466, "epsg", 3466,
		      "NAD83(NSRS2007) / Alabama West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=-87.5 +k=0.999933333 +x_0=6");
    add_proj4text (p, 1,
		   "00000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Alabama West\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",30],PARAMETER[\"central_meridian\",-87.5],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.999933333],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",600000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 12,
		 "ITY[\"EPSG\",\"3466\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 3467, "epsg", 3467,
		      "NAD83(NSRS2007) / Alaska Albers");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=55 +lat_2=65 +lat_0=50 +lon_0=-154 +x_0");
    add_proj4text (p, 1,
		   "=0 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Alaska Albers\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Albers_Conic_Equal_Area\"],PARAMETER[\"stan");
    add_srs_wkt (p, 9,
		 "dard_parallel_1\",55],PARAMETER[\"standard_parallel_2\",");
    add_srs_wkt (p, 10,
		 "65],PARAMETER[\"latitude_of_center\",50],PARAMETER[\"lon");
    add_srs_wkt (p, 11,
		 "gitude_of_center\",-154],PARAMETER[\"false_easting\",0],");
    add_srs_wkt (p, 12,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"346");
    add_srs_wkt (p, 13, "7\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3468, "epsg", 3468,
		      "NAD83(NSRS2007) / Alaska zone 1");
    add_proj4text (p, 0,
		   "+proj=omerc +lat_0=57 +lonc=-133.6666666666667 +alpha=32");
    add_proj4text (p, 1,
		   "3.1301023611111 +k=0.9999 +x_0=5000000 +y_0=-5000000 +el");
    add_proj4text (p, 2, "lps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Alaska zone 1\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Hotine_Oblique_Mercator\"],PARAMETER[\"lati");
    add_srs_wkt (p, 9,
		 "tude_of_center\",57],PARAMETER[\"longitude_of_center\",-");
    add_srs_wkt (p, 10,
		 "133.6666666666667],PARAMETER[\"azimuth\",323.13010236111");
    add_srs_wkt (p, 11,
		 "11],PARAMETER[\"rectified_grid_angle\",323.1301023611111");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_e");
    add_srs_wkt (p, 13,
		 "asting\",5000000],PARAMETER[\"false_northing\",-5000000]");
    add_srs_wkt (p, 14,
		 ",AUTHORITY[\"EPSG\",\"3468\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 15, ",NORTH]]");
    p = add_epsg_def (first, last, 3469, "epsg", 3469,
		      "NAD83(NSRS2007) / Alaska zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-142 +k=0.9999 +x_0=500000 ");
    add_proj4text (p, 1,
		   "+y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Alaska zone 2\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",54],PARAMETER[\"central_meridian\",-142],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12, "EPSG\",\"3469\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3470, "epsg", 3470,
		      "NAD83(NSRS2007) / Alaska zone 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-146 +k=0.9999 +x_0=500000 ");
    add_proj4text (p, 1,
		   "+y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Alaska zone 3\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",54],PARAMETER[\"central_meridian\",-146],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12, "EPSG\",\"3470\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3471, "epsg", 3471,
		      "NAD83(NSRS2007) / Alaska zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-150 +k=0.9999 +x_0=500000 ");
    add_proj4text (p, 1,
		   "+y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Alaska zone 4\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",54],PARAMETER[\"central_meridian\",-150],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12, "EPSG\",\"3471\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3472, "epsg", 3472,
		      "NAD83(NSRS2007) / Alaska zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-154 +k=0.9999 +x_0=500000 ");
    add_proj4text (p, 1,
		   "+y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Alaska zone 5\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",54],PARAMETER[\"central_meridian\",-154],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12, "EPSG\",\"3472\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3473, "epsg", 3473,
		      "NAD83(NSRS2007) / Alaska zone 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-158 +k=0.9999 +x_0=500000 ");
    add_proj4text (p, 1,
		   "+y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Alaska zone 6\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",54],PARAMETER[\"central_meridian\",-158],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12, "EPSG\",\"3473\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3474, "epsg", 3474,
		      "NAD83(NSRS2007) / Alaska zone 7");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-162 +k=0.9999 +x_0=500000 ");
    add_proj4text (p, 1,
		   "+y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Alaska zone 7\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",54],PARAMETER[\"central_meridian\",-162],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12, "EPSG\",\"3474\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3475, "epsg", 3475,
		      "NAD83(NSRS2007) / Alaska zone 8");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-166 +k=0.9999 +x_0=500000 ");
    add_proj4text (p, 1,
		   "+y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Alaska zone 8\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",54],PARAMETER[\"central_meridian\",-166],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12, "EPSG\",\"3475\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3476, "epsg", 3476,
		      "NAD83(NSRS2007) / Alaska zone 9");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-170 +k=0.9999 +x_0=500000 ");
    add_proj4text (p, 1,
		   "+y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Alaska zone 9\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",54],PARAMETER[\"central_meridian\",-170],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 11,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 12, "EPSG\",\"3476\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3477, "epsg", 3477,
		      "NAD83(NSRS2007) / Alaska zone 10");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=53.83333333333334 +lat_2=51.83333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=51 +lon_0=-176 +x_0=1000000 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Alaska zone 10\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_");
    add_srs_wkt (p, 2,
		 "System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",53.83333333333334],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",51.83333333333334],PARAMETER[\"latitu");
    add_srs_wkt (p, 11,
		 "de_of_origin\",51],PARAMETER[\"central_meridian\",-176],");
    add_srs_wkt (p, 12,
		 "PARAMETER[\"false_easting\",1000000],PARAMETER[\"false_n");
    add_srs_wkt (p, 13,
		 "orthing\",0],AUTHORITY[\"EPSG\",\"3477\"],AXIS[\"X\",EAS");
    add_srs_wkt (p, 14, "T],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3478, "epsg", 3478,
		      "NAD83(NSRS2007) / Arizona Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-111.9166666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=213360 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0");
    add_proj4text (p, 2, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Arizona Central\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference");
    add_srs_wkt (p, 2,
		 "_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 3,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 6,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 8,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 9,
		 "de_of_origin\",31],PARAMETER[\"central_meridian\",-111.9");
    add_srs_wkt (p, 10,
		 "166666666667],PARAMETER[\"scale_factor\",0.9999],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_easting\",213360],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"3478\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3479, "epsg", 3479,
		      "NAD83(NSRS2007) / Arizona Central (ft)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-111.9166666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=213360 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0");
    add_proj4text (p, 2, " +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Arizona Central (ft)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refe");
    add_srs_wkt (p, 2,
		 "rence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25");
    add_srs_wkt (p, 3,
		 "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4759\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 8,
		 "\"9002\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"latitude_of_origin\",31],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 10,
		 "\",-111.9166666666667],PARAMETER[\"scale_factor\",0.9999");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"false_easting\",700000],PARAMETER[\"false_");
    add_srs_wkt (p, 12,
		 "northing\",0],AUTHORITY[\"EPSG\",\"3479\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 13, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3480, "epsg", 3480,
		      "NAD83(NSRS2007) / Arizona East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-110.1666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=213360 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0");
    add_proj4text (p, 2, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Arizona East\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",31],PARAMETER[\"central_meridian\",-110.1666");
    add_srs_wkt (p, 10,
		 "666666667],PARAMETER[\"scale_factor\",0.9999],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_easting\",213360],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"3480\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 13, "\",NORTH]]");
    p = add_epsg_def (first, last, 3481, "epsg", 3481,
		      "NAD83(NSRS2007) / Arizona East (ft)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-110.1666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=213360 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0");
    add_proj4text (p, 2, " +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Arizona East (ft)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 8,
		 "02\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 9,
		 "atitude_of_origin\",31],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 10,
		 "110.1666666666667],PARAMETER[\"scale_factor\",0.9999],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_easting\",700000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 12,
		 "hing\",0],AUTHORITY[\"EPSG\",\"3481\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 13, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3482, "epsg", 3482,
		      "NAD83(NSRS2007) / Arizona West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-113.75 +k=0.999933333 +x_0");
    add_proj4text (p, 1,
		   "=213360 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +unit");
    add_proj4text (p, 2, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Arizona West\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",31],PARAMETER[\"central_meridian\",-113.75],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"scale_factor\",0.999933333],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",213360],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 12,
		 "ORITY[\"EPSG\",\"3482\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 13, "TH]]");
    p = add_epsg_def (first, last, 3483, "epsg", 3483,
		      "NAD83(NSRS2007) / Arizona West (ft)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-113.75 +k=0.999933333 +x_0");
    add_proj4text (p, 1,
		   "=213360 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +unit");
    add_proj4text (p, 2, "s=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Arizona West (ft)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 8,
		 "02\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 9,
		 "atitude_of_origin\",31],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 10,
		 "113.75],PARAMETER[\"scale_factor\",0.999933333],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_easting\",700000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"3483\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3484, "epsg", 3484,
		      "NAD83(NSRS2007) / Arkansas North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.23333333333333 +lat_2=34.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=34.33333333333334 +lon_0=-92 +x_0=400000 +y_0");
    add_proj4text (p, 2,
		   "=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_proj4text (p, 3, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Arkansas North\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_");
    add_srs_wkt (p, 2,
		 "System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",36.23333333333333],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",34.93333333333333],PARAMETER[\"latitu");
    add_srs_wkt (p, 11,
		 "de_of_origin\",34.33333333333334],PARAMETER[\"central_me");
    add_srs_wkt (p, 12,
		 "ridian\",-92],PARAMETER[\"false_easting\",400000],PARAME");
    add_srs_wkt (p, 13,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3484\"],A");
    add_srs_wkt (p, 14, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3485, "epsg", 3485,
		      "NAD83(NSRS2007) / Arkansas North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.23333333333333 +lat_2=34.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=34.33333333333334 +lon_0=-92 +x_0=399999.9999");
    add_proj4text (p, 2,
		   "8984 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=u");
    add_proj4text (p, 3, "s-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Arkansas North (ftUS)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2");
    add_srs_wkt (p, 3,
		 "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 9,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 10,
		 "36.23333333333333],PARAMETER[\"standard_parallel_2\",34.");
    add_srs_wkt (p, 11,
		 "93333333333333],PARAMETER[\"latitude_of_origin\",34.3333");
    add_srs_wkt (p, 12,
		 "3333333334],PARAMETER[\"central_meridian\",-92],PARAMETE");
    add_srs_wkt (p, 13,
		 "R[\"false_easting\",1312333.3333],PARAMETER[\"false_nort");
    add_srs_wkt (p, 14,
		 "hing\",0],AUTHORITY[\"EPSG\",\"3485\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 15, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3486, "epsg", 3486,
		      "NAD83(NSRS2007) / Arkansas South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=34.76666666666667 +lat_2=33.3 +lat_0=32");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-92 +x_0=400000 +y_0=400000 +ellp");
    add_proj4text (p, 2, "s=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Arkansas South\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_");
    add_srs_wkt (p, 2,
		 "System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",34.76666666666667],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",33.3],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 11,
		 ",32.66666666666666],PARAMETER[\"central_meridian\",-92],");
    add_srs_wkt (p, 12,
		 "PARAMETER[\"false_easting\",400000],PARAMETER[\"false_no");
    add_srs_wkt (p, 13,
		 "rthing\",400000],AUTHORITY[\"EPSG\",\"3486\"],AXIS[\"X\"");
    add_srs_wkt (p, 14, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3487, "epsg", 3487,
		      "NAD83(NSRS2007) / Arkansas South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=34.76666666666667 +lat_2=33.3 +lat_0=32");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-92 +x_0=399999.99998984 +y_0=399");
    add_proj4text (p, 2,
		   "999.99998984 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 3, "us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Arkansas South (ftUS)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2");
    add_srs_wkt (p, 3,
		 "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 9,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 10,
		 "34.76666666666667],PARAMETER[\"standard_parallel_2\",33.");
    add_srs_wkt (p, 11,
		 "3],PARAMETER[\"latitude_of_origin\",32.66666666666666],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"central_meridian\",-92],PARAMETER[\"false_eas");
    add_srs_wkt (p, 13,
		 "ting\",1312333.3333],PARAMETER[\"false_northing\",131233");
    add_srs_wkt (p, 14,
		 "3.3333],AUTHORITY[\"EPSG\",\"3487\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 15, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3488, "epsg", 3488,
		      "NAD83(NSRS2007) / California Albers");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=34 +lat_2=40.5 +lat_0=0 +lon_0=-120 +x_");
    add_proj4text (p, 1,
		   "0=0 +y_0=-4000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +u");
    add_proj4text (p, 2, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / California Albers\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Albers_Conic_Equal_Area\"],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_1\",34],PARAMETER[\"standard_parallel_2");
    add_srs_wkt (p, 10,
		 "\",40.5],PARAMETER[\"latitude_of_center\",0],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "longitude_of_center\",-120],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 12,
		 "0],PARAMETER[\"false_northing\",-4000000],AUTHORITY[\"EP");
    add_srs_wkt (p, 13, "SG\",\"3488\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3489, "epsg", 3489,
		      "NAD83(NSRS2007) / California zone 1");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.66666666666666 +lat_2=40 +lat_0=39.3");
    add_proj4text (p, 1,
		   "3333333333334 +lon_0=-122 +x_0=2000000 +y_0=500000 +ellp");
    add_proj4text (p, 2, "s=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / California zone 1\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",41.66666666666666],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "standard_parallel_2\",40],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 11,
		 "\",39.33333333333334],PARAMETER[\"central_meridian\",-12");
    add_srs_wkt (p, 12,
		 "2],PARAMETER[\"false_easting\",2000000],PARAMETER[\"fals");
    add_srs_wkt (p, 13,
		 "e_northing\",500000],AUTHORITY[\"EPSG\",\"3489\"],AXIS[\"");
    add_srs_wkt (p, 14, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3490, "epsg", 3490,
		      "NAD83(NSRS2007) / California zone 1 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.66666666666666 +lat_2=40 +lat_0=39.3");
    add_proj4text (p, 1,
		   "3333333333334 +lon_0=-122 +x_0=2000000.0001016 +y_0=5000");
    add_proj4text (p, 2,
		   "00.0001016001 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units");
    add_proj4text (p, 3, "=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / California zone 1 (ftUS)\",GE");
    add_srs_wkt (p, 1,
		 "OGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_");
    add_srs_wkt (p, 2,
		 "Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137,29");
    add_srs_wkt (p, 3,
		 "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 5,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 6,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 7,
		 "\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096");
    add_srs_wkt (p, 8,
		 "012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lamber");
    add_srs_wkt (p, 9,
		 "t_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1");
    add_srs_wkt (p, 10,
		 "\",41.66666666666666],PARAMETER[\"standard_parallel_2\",");
    add_srs_wkt (p, 11,
		 "40],PARAMETER[\"latitude_of_origin\",39.33333333333334],");
    add_srs_wkt (p, 12,
		 "PARAMETER[\"central_meridian\",-122],PARAMETER[\"false_e");
    add_srs_wkt (p, 13,
		 "asting\",6561666.667],PARAMETER[\"false_northing\",16404");
    add_srs_wkt (p, 14,
		 "16.667],AUTHORITY[\"EPSG\",\"3490\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 15, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3491, "epsg", 3491,
		      "NAD83(NSRS2007) / California zone 2");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.83333333333334 +lat_2=38.33333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=37.66666666666666 +lon_0=-122 +x_0=2000000 +y");
    add_proj4text (p, 2,
		   "_0=500000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +");
    add_proj4text (p, 3, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / California zone 2\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",39.83333333333334],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "standard_parallel_2\",38.33333333333334],PARAMETER[\"lat");
    add_srs_wkt (p, 11,
		 "itude_of_origin\",37.66666666666666],PARAMETER[\"central");
    add_srs_wkt (p, 12,
		 "_meridian\",-122],PARAMETER[\"false_easting\",2000000],P");
    add_srs_wkt (p, 13,
		 "ARAMETER[\"false_northing\",500000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 14, "3491\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3492, "epsg", 3492,
		      "NAD83(NSRS2007) / California zone 2 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.83333333333334 +lat_2=38.33333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=37.66666666666666 +lon_0=-122 +x_0=2000000.00");
    add_proj4text (p, 2,
		   "01016 +y_0=500000.0001016001 +ellps=GRS80 +towgs84=0,0,0");
    add_proj4text (p, 3, ",0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / California zone 2 (ftUS)\",GE");
    add_srs_wkt (p, 1,
		 "OGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_");
    add_srs_wkt (p, 2,
		 "Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137,29");
    add_srs_wkt (p, 3,
		 "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 5,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 6,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 7,
		 "\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096");
    add_srs_wkt (p, 8,
		 "012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lamber");
    add_srs_wkt (p, 9,
		 "t_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1");
    add_srs_wkt (p, 10,
		 "\",39.83333333333334],PARAMETER[\"standard_parallel_2\",");
    add_srs_wkt (p, 11,
		 "38.33333333333334],PARAMETER[\"latitude_of_origin\",37.6");
    add_srs_wkt (p, 12,
		 "6666666666666],PARAMETER[\"central_meridian\",-122],PARA");
    add_srs_wkt (p, 13,
		 "METER[\"false_easting\",6561666.667],PARAMETER[\"false_n");
    add_srs_wkt (p, 14,
		 "orthing\",1640416.667],AUTHORITY[\"EPSG\",\"3492\"],AXIS");
    add_srs_wkt (p, 15, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3493, "epsg", 3493,
		      "NAD83(NSRS2007) / California zone 3");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.43333333333333 +lat_2=37.06666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.5 +lon_0=-120.5 +x_0=2000000 +y_0=500000 +");
    add_proj4text (p, 2,
		   "ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / California zone 3\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",38.43333333333333],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "standard_parallel_2\",37.06666666666667],PARAMETER[\"lat");
    add_srs_wkt (p, 11,
		 "itude_of_origin\",36.5],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 12,
		 "120.5],PARAMETER[\"false_easting\",2000000],PARAMETER[\"");
    add_srs_wkt (p, 13,
		 "false_northing\",500000],AUTHORITY[\"EPSG\",\"3493\"],AX");
    add_srs_wkt (p, 14, "IS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3494, "epsg", 3494,
		      "NAD83(NSRS2007) / California zone 3 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.43333333333333 +lat_2=37.06666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.5 +lon_0=-120.5 +x_0=2000000.0001016 +y_0=");
    add_proj4text (p, 2,
		   "500000.0001016001 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +u");
    add_proj4text (p, 3, "nits=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / California zone 3 (ftUS)\",GE");
    add_srs_wkt (p, 1,
		 "OGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_");
    add_srs_wkt (p, 2,
		 "Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137,29");
    add_srs_wkt (p, 3,
		 "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 5,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 6,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 7,
		 "\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096");
    add_srs_wkt (p, 8,
		 "012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lamber");
    add_srs_wkt (p, 9,
		 "t_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1");
    add_srs_wkt (p, 10,
		 "\",38.43333333333333],PARAMETER[\"standard_parallel_2\",");
    add_srs_wkt (p, 11,
		 "37.06666666666667],PARAMETER[\"latitude_of_origin\",36.5");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"central_meridian\",-120.5],PARAMETER[\"fal");
    add_srs_wkt (p, 13,
		 "se_easting\",6561666.667],PARAMETER[\"false_northing\",1");
    add_srs_wkt (p, 14,
		 "640416.667],AUTHORITY[\"EPSG\",\"3494\"],AXIS[\"X\",EAST");
    add_srs_wkt (p, 15, "],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3495, "epsg", 3495,
		      "NAD83(NSRS2007) / California zone 4");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.25 +lat_2=36 +lat_0=35.3333333333333");
    add_proj4text (p, 1,
		   "4 +lon_0=-119 +x_0=2000000 +y_0=500000 +ellps=GRS80 +tow");
    add_proj4text (p, 2, "gs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / California zone 4\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",37.25],PARAMETER[\"standard_par");
    add_srs_wkt (p, 10,
		 "allel_2\",36],PARAMETER[\"latitude_of_origin\",35.333333");
    add_srs_wkt (p, 11,
		 "33333334],PARAMETER[\"central_meridian\",-119],PARAMETER");
    add_srs_wkt (p, 12,
		 "[\"false_easting\",2000000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 13,
		 ",500000],AUTHORITY[\"EPSG\",\"3495\"],AXIS[\"X\",EAST],A");
    add_srs_wkt (p, 14, "XIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3496, "epsg", 3496,
		      "NAD83(NSRS2007) / California zone 4 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.25 +lat_2=36 +lat_0=35.3333333333333");
    add_proj4text (p, 1,
		   "4 +lon_0=-119 +x_0=2000000.0001016 +y_0=500000.000101600");
    add_proj4text (p, 2,
		   "1 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_d");
    add_proj4text (p, 3, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / California zone 4 (ftUS)\",GE");
    add_srs_wkt (p, 1,
		 "OGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_");
    add_srs_wkt (p, 2,
		 "Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137,29");
    add_srs_wkt (p, 3,
		 "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 5,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 6,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 7,
		 "\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096");
    add_srs_wkt (p, 8,
		 "012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lamber");
    add_srs_wkt (p, 9,
		 "t_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1");
    add_srs_wkt (p, 10,
		 "\",37.25],PARAMETER[\"standard_parallel_2\",36],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"latitude_of_origin\",35.33333333333334],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "central_meridian\",-119],PARAMETER[\"false_easting\",656");
    add_srs_wkt (p, 13,
		 "1666.667],PARAMETER[\"false_northing\",1640416.667],AUTH");
    add_srs_wkt (p, 14,
		 "ORITY[\"EPSG\",\"3496\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 15, "TH]]");
    p = add_epsg_def (first, last, 3497, "epsg", 3497,
		      "NAD83(NSRS2007) / California zone 5");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=35.46666666666667 +lat_2=34.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=33.5 +lon_0=-118 +x_0=2000000 +y_0=500000 +el");
    add_proj4text (p, 2, "lps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / California zone 5\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",35.46666666666667],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "standard_parallel_2\",34.03333333333333],PARAMETER[\"lat");
    add_srs_wkt (p, 11,
		 "itude_of_origin\",33.5],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 12,
		 "118],PARAMETER[\"false_easting\",2000000],PARAMETER[\"fa");
    add_srs_wkt (p, 13,
		 "lse_northing\",500000],AUTHORITY[\"EPSG\",\"3497\"],AXIS");
    add_srs_wkt (p, 14, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3498, "epsg", 3498,
		      "NAD83(NSRS2007) / California zone 5 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=35.46666666666667 +lat_2=34.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=33.5 +lon_0=-118 +x_0=2000000.0001016 +y_0=50");
    add_proj4text (p, 2,
		   "0000.0001016001 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +uni");
    add_proj4text (p, 3, "ts=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / California zone 5 (ftUS)\",GE");
    add_srs_wkt (p, 1,
		 "OGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_");
    add_srs_wkt (p, 2,
		 "Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137,29");
    add_srs_wkt (p, 3,
		 "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 5,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 6,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 7,
		 "\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096");
    add_srs_wkt (p, 8,
		 "012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lamber");
    add_srs_wkt (p, 9,
		 "t_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1");
    add_srs_wkt (p, 10,
		 "\",35.46666666666667],PARAMETER[\"standard_parallel_2\",");
    add_srs_wkt (p, 11,
		 "34.03333333333333],PARAMETER[\"latitude_of_origin\",33.5");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"central_meridian\",-118],PARAMETER[\"false");
    add_srs_wkt (p, 13,
		 "_easting\",6561666.667],PARAMETER[\"false_northing\",164");
    add_srs_wkt (p, 14,
		 "0416.667],AUTHORITY[\"EPSG\",\"3498\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 15, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3499, "epsg", 3499,
		      "NAD83(NSRS2007) / California zone 6");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.88333333333333 +lat_2=32.78333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=32.16666666666666 +lon_0=-116.25 +x_0=2000000");
    add_proj4text (p, 2,
		   " +y_0=500000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 3, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / California zone 6\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",33.88333333333333],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "standard_parallel_2\",32.78333333333333],PARAMETER[\"lat");
    add_srs_wkt (p, 11,
		 "itude_of_origin\",32.16666666666666],PARAMETER[\"central");
    add_srs_wkt (p, 12,
		 "_meridian\",-116.25],PARAMETER[\"false_easting\",2000000");
    add_srs_wkt (p, 13,
		 "],PARAMETER[\"false_northing\",500000],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 14, ",\"3499\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3500, "epsg", 3500,
		      "NAD83(NSRS2007) / California zone 6 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.88333333333333 +lat_2=32.78333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=32.16666666666666 +lon_0=-116.25 +x_0=2000000");
    add_proj4text (p, 2,
		   ".0001016 +y_0=500000.0001016001 +ellps=GRS80 +towgs84=0,");
    add_proj4text (p, 3, "0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / California zone 6 (ftUS)\",GE");
    add_srs_wkt (p, 1,
		 "OGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_");
    add_srs_wkt (p, 2,
		 "Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137,29");
    add_srs_wkt (p, 3,
		 "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 5,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 6,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 7,
		 "\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096");
    add_srs_wkt (p, 8,
		 "012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lamber");
    add_srs_wkt (p, 9,
		 "t_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1");
    add_srs_wkt (p, 10,
		 "\",33.88333333333333],PARAMETER[\"standard_parallel_2\",");
    add_srs_wkt (p, 11,
		 "32.78333333333333],PARAMETER[\"latitude_of_origin\",32.1");
    add_srs_wkt (p, 12,
		 "6666666666666],PARAMETER[\"central_meridian\",-116.25],P");
    add_srs_wkt (p, 13,
		 "ARAMETER[\"false_easting\",6561666.667],PARAMETER[\"fals");
    add_srs_wkt (p, 14,
		 "e_northing\",1640416.667],AUTHORITY[\"EPSG\",\"3500\"],A");
    add_srs_wkt (p, 15, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3501, "epsg", 3501,
		      "NAD83(NSRS2007) / Colorado Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.75 +lat_2=38.45 +lat_0=37.8333333333");
    add_proj4text (p, 1,
		   "3334 +lon_0=-105.5 +x_0=914401.8289 +y_0=304800.6096 +el");
    add_proj4text (p, 2, "lps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Colorado Central\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referenc");
    add_srs_wkt (p, 2,
		 "e_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",39.75],PARAMETER[\"standard_par");
    add_srs_wkt (p, 10,
		 "allel_2\",38.45],PARAMETER[\"latitude_of_origin\",37.833");
    add_srs_wkt (p, 11,
		 "33333333334],PARAMETER[\"central_meridian\",-105.5],PARA");
    add_srs_wkt (p, 12,
		 "METER[\"false_easting\",914401.8289],PARAMETER[\"false_n");
    add_srs_wkt (p, 13,
		 "orthing\",304800.6096],AUTHORITY[\"EPSG\",\"3501\"],AXIS");
    add_srs_wkt (p, 14, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3502, "epsg", 3502,
		      "NAD83(NSRS2007) / Colorado Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.75 +lat_2=38.45 +lat_0=37.8333333333");
    add_proj4text (p, 1,
		   "3334 +lon_0=-105.5 +x_0=914401.8288036576 +y_0=304800.60");
    add_proj4text (p, 2,
		   "96012192 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-f");
    add_proj4text (p, 3, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Colorado Central (ftUS)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_R");
    add_srs_wkt (p, 2,
		 "eference_System_2007\",SPHEROID[\"GRS 1980\",6378137,298");
    add_srs_wkt (p, 3,
		 ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 5,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 6,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.304800609601");
    add_srs_wkt (p, 8,
		 "2192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_");
    add_srs_wkt (p, 9,
		 "Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 10,
		 ",39.75],PARAMETER[\"standard_parallel_2\",38.45],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"latitude_of_origin\",37.83333333333334],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "central_meridian\",-105.5],PARAMETER[\"false_easting\",3");
    add_srs_wkt (p, 13,
		 "000000],PARAMETER[\"false_northing\",1000000],AUTHORITY[");
    add_srs_wkt (p, 14,
		 "\"EPSG\",\"3502\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3503, "epsg", 3503,
		      "NAD83(NSRS2007) / Colorado North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.78333333333333 +lat_2=39.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=39.33333333333334 +lon_0=-105.5 +x_0=914401.8");
    add_proj4text (p, 2,
		   "289 +y_0=304800.6096 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0");
    add_proj4text (p, 3, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Colorado North\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_");
    add_srs_wkt (p, 2,
		 "System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",40.78333333333333],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",39.71666666666667],PARAMETER[\"latitu");
    add_srs_wkt (p, 11,
		 "de_of_origin\",39.33333333333334],PARAMETER[\"central_me");
    add_srs_wkt (p, 12,
		 "ridian\",-105.5],PARAMETER[\"false_easting\",914401.8289");
    add_srs_wkt (p, 13,
		 "],PARAMETER[\"false_northing\",304800.6096],AUTHORITY[\"");
    add_srs_wkt (p, 14, "EPSG\",\"3503\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3504, "epsg", 3504,
		      "NAD83(NSRS2007) / Colorado North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.78333333333333 +lat_2=39.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=39.33333333333334 +lon_0=-105.5 +x_0=914401.8");
    add_proj4text (p, 2,
		   "288036576 +y_0=304800.6096012192 +ellps=GRS80 +towgs84=0");
    add_proj4text (p, 3, ",0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Colorado North (ftUS)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2");
    add_srs_wkt (p, 3,
		 "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 9,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 10,
		 "40.78333333333333],PARAMETER[\"standard_parallel_2\",39.");
    add_srs_wkt (p, 11,
		 "71666666666667],PARAMETER[\"latitude_of_origin\",39.3333");
    add_srs_wkt (p, 12,
		 "3333333334],PARAMETER[\"central_meridian\",-105.5],PARAM");
    add_srs_wkt (p, 13,
		 "ETER[\"false_easting\",3000000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 14,
		 "ng\",1000000],AUTHORITY[\"EPSG\",\"3504\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 15, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3505, "epsg", 3505,
		      "NAD83(NSRS2007) / Colorado South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.43333333333333 +lat_2=37.23333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=36.66666666666666 +lon_0=-105.5 +x_0=914401.8");
    add_proj4text (p, 2,
		   "289 +y_0=304800.6096 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0");
    add_proj4text (p, 3, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Colorado South\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_");
    add_srs_wkt (p, 2,
		 "System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",38.43333333333333],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",37.23333333333333],PARAMETER[\"latitu");
    add_srs_wkt (p, 11,
		 "de_of_origin\",36.66666666666666],PARAMETER[\"central_me");
    add_srs_wkt (p, 12,
		 "ridian\",-105.5],PARAMETER[\"false_easting\",914401.8289");
    add_srs_wkt (p, 13,
		 "],PARAMETER[\"false_northing\",304800.6096],AUTHORITY[\"");
    add_srs_wkt (p, 14, "EPSG\",\"3505\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3506, "epsg", 3506,
		      "NAD83(NSRS2007) / Colorado South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.43333333333333 +lat_2=37.23333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=36.66666666666666 +lon_0=-105.5 +x_0=914401.8");
    add_proj4text (p, 2,
		   "288036576 +y_0=304800.6096012192 +ellps=GRS80 +towgs84=0");
    add_proj4text (p, 3, ",0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Colorado South (ftUS)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2");
    add_srs_wkt (p, 3,
		 "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 9,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 10,
		 "38.43333333333333],PARAMETER[\"standard_parallel_2\",37.");
    add_srs_wkt (p, 11,
		 "23333333333333],PARAMETER[\"latitude_of_origin\",36.6666");
    add_srs_wkt (p, 12,
		 "6666666666],PARAMETER[\"central_meridian\",-105.5],PARAM");
    add_srs_wkt (p, 13,
		 "ETER[\"false_easting\",3000000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 14,
		 "ng\",1000000],AUTHORITY[\"EPSG\",\"3506\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 15, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3507, "epsg", 3507,
		      "NAD83(NSRS2007) / Connecticut");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.86666666666667 +lat_2=41.2 +lat_0=40");
    add_proj4text (p, 1,
		   ".83333333333334 +lon_0=-72.75 +x_0=304800.6096 +y_0=1524");
    add_proj4text (p, 2,
		   "00.3048 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no");
    add_proj4text (p, 3, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Connecticut\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sys");
    add_srs_wkt (p, 2,
		 "tem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"47");
    add_srs_wkt (p, 7,
		 "59\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 9,
		 "andard_parallel_1\",41.86666666666667],PARAMETER[\"stand");
    add_srs_wkt (p, 10,
		 "ard_parallel_2\",41.2],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 11,
		 "40.83333333333334],PARAMETER[\"central_meridian\",-72.75");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_easting\",304800.6096],PARAMETER[\"f");
    add_srs_wkt (p, 13,
		 "alse_northing\",152400.3048],AUTHORITY[\"EPSG\",\"3507\"");
    add_srs_wkt (p, 14, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3508, "epsg", 3508,
		      "NAD83(NSRS2007) / Connecticut (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.86666666666667 +lat_2=41.2 +lat_0=40");
    add_proj4text (p, 1,
		   ".83333333333334 +lon_0=-72.75 +x_0=304800.6096012192 +y_");
    add_proj4text (p, 2,
		   "0=152400.3048006096 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 ");
    add_proj4text (p, 3, "+units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Connecticut (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 8,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 9,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",41.86");
    add_srs_wkt (p, 10,
		 "666666666667],PARAMETER[\"standard_parallel_2\",41.2],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"latitude_of_origin\",40.83333333333334],PARAME");
    add_srs_wkt (p, 12,
		 "TER[\"central_meridian\",-72.75],PARAMETER[\"false_easti");
    add_srs_wkt (p, 13,
		 "ng\",1000000],PARAMETER[\"false_northing\",500000],AUTHO");
    add_srs_wkt (p, 14,
		 "RITY[\"EPSG\",\"3508\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORT");
    add_srs_wkt (p, 15, "H]]");
    p = add_epsg_def (first, last, 3509, "epsg", 3509,
		      "NAD83(NSRS2007) / Delaware");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38 +lon_0=-75.41666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "995 +x_0=200000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0");
    add_proj4text (p, 2, ",0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Delaware\",GEOGCS[\"NAD83(NSR");
    add_srs_wkt (p, 1,
		 "S2007)\",DATUM[\"NAD83_National_Spatial_Reference_System");
    add_srs_wkt (p, 2,
		 "_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 5,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4759\"");
    add_srs_wkt (p, 7,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 9,
		 "igin\",38],PARAMETER[\"central_meridian\",-75.4166666666");
    add_srs_wkt (p, 10,
		 "6667],PARAMETER[\"scale_factor\",0.999995],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",200000],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 12,
		 "UTHORITY[\"EPSG\",\"3509\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 13, "NORTH]]");
    p = add_epsg_def (first, last, 3510, "epsg", 3510,
		      "NAD83(NSRS2007) / Delaware (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38 +lon_0=-75.41666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "995 +x_0=200000.0001016002 +y_0=0 +ellps=GRS80 +towgs84=");
    add_proj4text (p, 2, "0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Delaware (ftUS)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference");
    add_srs_wkt (p, 2,
		 "_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 3,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 6,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4759\"]],UNIT[\"US survey foot\",0.3048006096012192,AU");
    add_srs_wkt (p, 8,
		 "THORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 9,
		 "ator\"],PARAMETER[\"latitude_of_origin\",38],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "central_meridian\",-75.41666666666667],PARAMETER[\"scale");
    add_srs_wkt (p, 11,
		 "_factor\",0.999995],PARAMETER[\"false_easting\",656166.6");
    add_srs_wkt (p, 12,
		 "67],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 13, "3510\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3511, "epsg", 3511,
		      "NAD83(NSRS2007) / Florida East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=24.33333333333333 +lon_0=-81 +k=0.999");
    add_proj4text (p, 1,
		   "941177 +x_0=200000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 2, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Florida East\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",24.33333333333333],PARAMETER[\"central_merid");
    add_srs_wkt (p, 10,
		 "ian\",-81],PARAMETER[\"scale_factor\",0.999941177],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_easting\",200000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",0],AUTHORITY[\"EPSG\",\"3511\"],AXIS[\"X\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3512, "epsg", 3512,
		      "NAD83(NSRS2007) / Florida East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=24.33333333333333 +lon_0=-81 +k=0.999");
    add_proj4text (p, 1,
		   "941177 +x_0=200000.0001016002 +y_0=0 +ellps=GRS80 +towgs");
    add_proj4text (p, 2, "84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Florida East (ftUS)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"US survey foot\",0.304800609601219");
    add_srs_wkt (p, 8,
		 "2,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_");
    add_srs_wkt (p, 9,
		 "Mercator\"],PARAMETER[\"latitude_of_origin\",24.33333333");
    add_srs_wkt (p, 10,
		 "333333],PARAMETER[\"central_meridian\",-81],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "scale_factor\",0.999941177],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 12,
		 "656166.667],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13, "EPSG\",\"3512\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3513, "epsg", 3513,
		      "NAD83(NSRS2007) / Florida GDL Albers");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=24 +lat_2=31.5 +lat_0=24 +lon_0=-84 +x_");
    add_proj4text (p, 1,
		   "0=400000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +uni");
    add_proj4text (p, 2, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Florida GDL Albers\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Albers_Conic_Equal_Area\"],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_1\",24],PARAMETER[\"standard_parallel_2");
    add_srs_wkt (p, 10,
		 "\",31.5],PARAMETER[\"latitude_of_center\",24],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"longitude_of_center\",-84],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",400000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 13, "G\",\"3513\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3514, "epsg", 3514,
		      "NAD83(NSRS2007) / Florida North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.75 +lat_2=29.58333333333333 +lat_0=2");
    add_proj4text (p, 1,
		   "9 +lon_0=-84.5 +x_0=600000 +y_0=0 +ellps=GRS80 +towgs84=");
    add_proj4text (p, 2, "0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Florida North\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",30.75],PARAMETER[\"standard_parall");
    add_srs_wkt (p, 10,
		 "el_2\",29.58333333333333],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 11,
		 "\",29],PARAMETER[\"central_meridian\",-84.5],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_easting\",600000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 13,
		 "AUTHORITY[\"EPSG\",\"3514\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 14, ",NORTH]]");
    p = add_epsg_def (first, last, 3515, "epsg", 3515,
		      "NAD83(NSRS2007) / Florida North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.75 +lat_2=29.58333333333333 +lat_0=2");
    add_proj4text (p, 1,
		   "9 +lon_0=-84.5 +x_0=600000 +y_0=0 +ellps=GRS80 +towgs84=");
    add_proj4text (p, 2, "0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Florida North (ftUS)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refe");
    add_srs_wkt (p, 2,
		 "rence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25");
    add_srs_wkt (p, 3,
		 "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4759\"]],UNIT[\"US survey foot\",0.30480060960121");
    add_srs_wkt (p, 8,
		 "92,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Co");
    add_srs_wkt (p, 9,
		 "nformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",3");
    add_srs_wkt (p, 10,
		 "0.75],PARAMETER[\"standard_parallel_2\",29.5833333333333");
    add_srs_wkt (p, 11,
		 "3],PARAMETER[\"latitude_of_origin\",29],PARAMETER[\"cent");
    add_srs_wkt (p, 12,
		 "ral_meridian\",-84.5],PARAMETER[\"false_easting\",196850");
    add_srs_wkt (p, 13,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 14, "3515\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3516, "epsg", 3516,
		      "NAD83(NSRS2007) / Florida West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=24.33333333333333 +lon_0=-82 +k=0.999");
    add_proj4text (p, 1,
		   "941177 +x_0=200000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 2, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Florida West\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",24.33333333333333],PARAMETER[\"central_merid");
    add_srs_wkt (p, 10,
		 "ian\",-82],PARAMETER[\"scale_factor\",0.999941177],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_easting\",200000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",0],AUTHORITY[\"EPSG\",\"3516\"],AXIS[\"X\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3517, "epsg", 3517,
		      "NAD83(NSRS2007) / Florida West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=24.33333333333333 +lon_0=-82 +k=0.999");
    add_proj4text (p, 1,
		   "941177 +x_0=200000.0001016002 +y_0=0 +ellps=GRS80 +towgs");
    add_proj4text (p, 2, "84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Florida West (ftUS)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"US survey foot\",0.304800609601219");
    add_srs_wkt (p, 8,
		 "2,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_");
    add_srs_wkt (p, 9,
		 "Mercator\"],PARAMETER[\"latitude_of_origin\",24.33333333");
    add_srs_wkt (p, 10,
		 "333333],PARAMETER[\"central_meridian\",-82],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "scale_factor\",0.999941177],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 12,
		 "656166.667],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13, "EPSG\",\"3517\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3518, "epsg", 3518,
		      "NAD83(NSRS2007) / Georgia East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=-82.16666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=200000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0");
    add_proj4text (p, 2, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Georgia East\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",30],PARAMETER[\"central_meridian\",-82.16666");
    add_srs_wkt (p, 10,
		 "666666667],PARAMETER[\"scale_factor\",0.9999],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_easting\",200000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"3518\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 13, "\",NORTH]]");
    p = add_epsg_def (first, last, 3519, "epsg", 3519,
		      "NAD83(NSRS2007) / Georgia East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=-82.16666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=200000.0001016002 +y_0=0 +ellps=GRS80 +towgs84=0,");
    add_proj4text (p, 2, "0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Georgia East (ftUS)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"US survey foot\",0.304800609601219");
    add_srs_wkt (p, 8,
		 "2,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_");
    add_srs_wkt (p, 9,
		 "Mercator\"],PARAMETER[\"latitude_of_origin\",30],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"central_meridian\",-82.16666666666667],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",65616");
    add_srs_wkt (p, 12,
		 "6.667],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 13, ",\"3519\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3520, "epsg", 3520,
		      "NAD83(NSRS2007) / Georgia West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=-84.16666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=700000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0");
    add_proj4text (p, 2, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Georgia West\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",30],PARAMETER[\"central_meridian\",-84.16666");
    add_srs_wkt (p, 10,
		 "666666667],PARAMETER[\"scale_factor\",0.9999],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_easting\",700000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"3520\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 13, "\",NORTH]]");
    p = add_epsg_def (first, last, 3521, "epsg", 3521,
		      "NAD83(NSRS2007) / Georgia West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=-84.16666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=699999.9998983998 +y_0=0 +ellps=GRS80 +towgs84=0,");
    add_proj4text (p, 2, "0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Georgia West (ftUS)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"US survey foot\",0.304800609601219");
    add_srs_wkt (p, 8,
		 "2,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_");
    add_srs_wkt (p, 9,
		 "Mercator\"],PARAMETER[\"latitude_of_origin\",30],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"central_meridian\",-84.16666666666667],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",22965");
    add_srs_wkt (p, 12,
		 "83.333],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 13, "\",\"3521\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3522, "epsg", 3522,
		      "NAD83(NSRS2007) / Idaho Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-114 +k=0.99");
    add_proj4text (p, 1,
		   "99473679999999 +x_0=500000 +y_0=0 +ellps=GRS80 +towgs84=");
    add_proj4text (p, 2, "0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Idaho Central\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",41.66666666666666],PARAMETER[\"central_meri");
    add_srs_wkt (p, 10,
		 "dian\",-114],PARAMETER[\"scale_factor\",0.999947368],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 12,
		 "ing\",0],AUTHORITY[\"EPSG\",\"3522\"],AXIS[\"X\",EAST],A");
    add_srs_wkt (p, 13, "XIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3523, "epsg", 3523,
		      "NAD83(NSRS2007) / Idaho Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-114 +k=0.99");
    add_proj4text (p, 1,
		   "99473679999999 +x_0=500000.0001016001 +y_0=0 +ellps=GRS8");
    add_proj4text (p, 2, "0 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Idaho Central (ftUS)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refe");
    add_srs_wkt (p, 2,
		 "rence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25");
    add_srs_wkt (p, 3,
		 "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4759\"]],UNIT[\"US survey foot\",0.30480060960121");
    add_srs_wkt (p, 8,
		 "92,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 9,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",41.6666666");
    add_srs_wkt (p, 10,
		 "6666666],PARAMETER[\"central_meridian\",-114],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"scale_factor\",0.999947368],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",1640416.667],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 13,
		 "\"EPSG\",\"3523\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3524, "epsg", 3524,
		      "NAD83(NSRS2007) / Idaho East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-112.1666666");
    add_proj4text (p, 1,
		   "666667 +k=0.9999473679999999 +x_0=200000 +y_0=0 +ellps=G");
    add_proj4text (p, 2, "RS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Idaho East\",GEOGCS[\"NAD83(N");
    add_srs_wkt (p, 1,
		 "SRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Syst");
    add_srs_wkt (p, 2,
		 "em_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"475");
    add_srs_wkt (p, 7,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",41.66666666666666],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 10,
		 "n\",-112.1666666666667],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 11,
		 "947368],PARAMETER[\"false_easting\",200000],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3524\"],AXIS[\"");
    add_srs_wkt (p, 13, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3525, "epsg", 3525,
		      "NAD83(NSRS2007) / Idaho East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-112.1666666");
    add_proj4text (p, 1,
		   "666667 +k=0.9999473679999999 +x_0=200000.0001016002 +y_0");
    add_proj4text (p, 2,
		   "=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_");
    add_proj4text (p, 3, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Idaho East (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 8,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 9,
		 "cator\"],PARAMETER[\"latitude_of_origin\",41.66666666666");
    add_srs_wkt (p, 10,
		 "666],PARAMETER[\"central_meridian\",-112.1666666666667],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"scale_factor\",0.999947368],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_easting\",656166.667],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 13,
		 "AUTHORITY[\"EPSG\",\"3525\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 14, ",NORTH]]");
    p = add_epsg_def (first, last, 3526, "epsg", 3526,
		      "NAD83(NSRS2007) / Idaho West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-115.75 +k=0");
    add_proj4text (p, 1,
		   ".999933333 +x_0=800000 +y_0=0 +ellps=GRS80 +towgs84=0,0,");
    add_proj4text (p, 2, "0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Idaho West\",GEOGCS[\"NAD83(N");
    add_srs_wkt (p, 1,
		 "SRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Syst");
    add_srs_wkt (p, 2,
		 "em_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"475");
    add_srs_wkt (p, 7,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",41.66666666666666],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 10,
		 "n\",-115.75],PARAMETER[\"scale_factor\",0.999933333],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_easting\",800000],PARAMETER[\"false_north");
    add_srs_wkt (p, 12,
		 "ing\",0],AUTHORITY[\"EPSG\",\"3526\"],AXIS[\"X\",EAST],A");
    add_srs_wkt (p, 13, "XIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3527, "epsg", 3527,
		      "NAD83(NSRS2007) / Idaho West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-115.75 +k=0");
    add_proj4text (p, 1,
		   ".999933333 +x_0=800000.0001016001 +y_0=0 +ellps=GRS80 +t");
    add_proj4text (p, 2, "owgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Idaho West (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 8,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 9,
		 "cator\"],PARAMETER[\"latitude_of_origin\",41.66666666666");
    add_srs_wkt (p, 10,
		 "666],PARAMETER[\"central_meridian\",-115.75],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "scale_factor\",0.999933333],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 12,
		 "2624666.667],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13, "EPSG\",\"3527\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3528, "epsg", 3528,
		      "NAD83(NSRS2007) / Illinois East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36.66666666666666 +lon_0=-88.33333333");
    add_proj4text (p, 1,
		   "333333 +k=0.9999749999999999 +x_0=300000 +y_0=0 +ellps=G");
    add_proj4text (p, 2, "RS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Illinois East\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",36.66666666666666],PARAMETER[\"central_meri");
    add_srs_wkt (p, 10,
		 "dian\",-88.33333333333333],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 11,
		 "999975],PARAMETER[\"false_easting\",300000],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3528\"],AXIS[\"");
    add_srs_wkt (p, 13, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3529, "epsg", 3529,
		      "NAD83(NSRS2007) / Illinois East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36.66666666666666 +lon_0=-88.33333333");
    add_proj4text (p, 1,
		   "333333 +k=0.9999749999999999 +x_0=300000.0000000001 +y_0");
    add_proj4text (p, 2,
		   "=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_");
    add_proj4text (p, 3, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Illinois East (ftUS)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refe");
    add_srs_wkt (p, 2,
		 "rence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25");
    add_srs_wkt (p, 3,
		 "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4759\"]],UNIT[\"US survey foot\",0.30480060960121");
    add_srs_wkt (p, 8,
		 "92,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 9,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",36.6666666");
    add_srs_wkt (p, 10,
		 "6666666],PARAMETER[\"central_meridian\",-88.333333333333");
    add_srs_wkt (p, 11,
		 "33],PARAMETER[\"scale_factor\",0.999975],PARAMETER[\"fal");
    add_srs_wkt (p, 12,
		 "se_easting\",984250.0000000002],PARAMETER[\"false_northi");
    add_srs_wkt (p, 13,
		 "ng\",0],AUTHORITY[\"EPSG\",\"3529\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 14, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3530, "epsg", 3530,
		      "NAD83(NSRS2007) / Illinois West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36.66666666666666 +lon_0=-90.16666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999941177 +x_0=700000 +y_0=0 +ellps=GRS80 +t");
    add_proj4text (p, 2, "owgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Illinois West\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",36.66666666666666],PARAMETER[\"central_meri");
    add_srs_wkt (p, 10,
		 "dian\",-90.16666666666667],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 11,
		 "999941177],PARAMETER[\"false_easting\",700000],PARAMETER");
    add_srs_wkt (p, 12,
		 "[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3530\"],AXIS");
    add_srs_wkt (p, 13, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3531, "epsg", 3531,
		      "NAD83(NSRS2007) / Illinois West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36.66666666666666 +lon_0=-90.16666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999941177 +x_0=699999.9999898402 +y_0=0 +ell");
    add_proj4text (p, 2,
		   "ps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Illinois West (ftUS)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refe");
    add_srs_wkt (p, 2,
		 "rence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25");
    add_srs_wkt (p, 3,
		 "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4759\"]],UNIT[\"US survey foot\",0.30480060960121");
    add_srs_wkt (p, 8,
		 "92,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 9,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",36.6666666");
    add_srs_wkt (p, 10,
		 "6666666],PARAMETER[\"central_meridian\",-90.166666666666");
    add_srs_wkt (p, 11,
		 "67],PARAMETER[\"scale_factor\",0.999941177],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_easting\",2296583.333300001],PARAMETER[\"false_nor");
    add_srs_wkt (p, 13,
		 "thing\",0],AUTHORITY[\"EPSG\",\"3531\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 14, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3532, "epsg", 3532,
		      "NAD83(NSRS2007) / Indiana East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=37.5 +lon_0=-85.66666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=100000 +y_0=250000 +ellps=GRS80 +towgs84=0");
    add_proj4text (p, 2, ",0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Indiana East\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",37.5],PARAMETER[\"central_meridian\",-85.666");
    add_srs_wkt (p, 10,
		 "66666666667],PARAMETER[\"scale_factor\",0.999966667],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_easting\",100000],PARAMETER[\"false_north");
    add_srs_wkt (p, 12,
		 "ing\",250000],AUTHORITY[\"EPSG\",\"3532\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 13, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3533, "epsg", 3533,
		      "NAD83(NSRS2007) / Indiana East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=37.5 +lon_0=-85.66666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=99999.99989839978 +y_0=249999.9998983998 +");
    add_proj4text (p, 2,
		   "ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_proj4text (p, 3, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Indiana East (ftUS)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"US survey foot\",0.304800609601219");
    add_srs_wkt (p, 8,
		 "2,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_");
    add_srs_wkt (p, 9,
		 "Mercator\"],PARAMETER[\"latitude_of_origin\",37.5],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"central_meridian\",-85.66666666666667],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"scale_factor\",0.999966667],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",328083.333],PARAMETER[\"false_northing\",820208.3330000");
    add_srs_wkt (p, 13,
		 "002],AUTHORITY[\"EPSG\",\"3533\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 14, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3534, "epsg", 3534,
		      "NAD83(NSRS2007) / Indiana West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=37.5 +lon_0=-87.08333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=900000 +y_0=250000 +ellps=GRS80 +towgs84=0");
    add_proj4text (p, 2, ",0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Indiana West\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",37.5],PARAMETER[\"central_meridian\",-87.083");
    add_srs_wkt (p, 10,
		 "33333333333],PARAMETER[\"scale_factor\",0.999966667],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_easting\",900000],PARAMETER[\"false_north");
    add_srs_wkt (p, 12,
		 "ing\",250000],AUTHORITY[\"EPSG\",\"3534\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 13, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3535, "epsg", 3535,
		      "NAD83(NSRS2007) / Indiana West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=37.5 +lon_0=-87.08333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=900000 +y_0=249999.9998983998 +ellps=GRS80");
    add_proj4text (p, 2, " +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Indiana West (ftUS)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"US survey foot\",0.304800609601219");
    add_srs_wkt (p, 8,
		 "2,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_");
    add_srs_wkt (p, 9,
		 "Mercator\"],PARAMETER[\"latitude_of_origin\",37.5],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"central_meridian\",-87.08333333333333],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"scale_factor\",0.999966667],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",2952750],PARAMETER[\"false_northing\",820208.3330000002");
    add_srs_wkt (p, 13,
		 "],AUTHORITY[\"EPSG\",\"3535\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 14, "\",NORTH]]");
    p = add_epsg_def (first, last, 3536, "epsg", 3536,
		      "NAD83(NSRS2007) / Iowa North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43.26666666666667 +lat_2=42.06666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=41.5 +lon_0=-93.5 +x_0=1500000 +y_0=1000000 +");
    add_proj4text (p, 2,
		   "ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Iowa North\",GEOGCS[\"NAD83(N");
    add_srs_wkt (p, 1,
		 "SRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Syst");
    add_srs_wkt (p, 2,
		 "em_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"475");
    add_srs_wkt (p, 7,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"sta");
    add_srs_wkt (p, 9,
		 "ndard_parallel_1\",43.26666666666667],PARAMETER[\"standa");
    add_srs_wkt (p, 10,
		 "rd_parallel_2\",42.06666666666667],PARAMETER[\"latitude_");
    add_srs_wkt (p, 11,
		 "of_origin\",41.5],PARAMETER[\"central_meridian\",-93.5],");
    add_srs_wkt (p, 12,
		 "PARAMETER[\"false_easting\",1500000],PARAMETER[\"false_n");
    add_srs_wkt (p, 13,
		 "orthing\",1000000],AUTHORITY[\"EPSG\",\"3536\"],AXIS[\"X");
    add_srs_wkt (p, 14, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3537, "epsg", 3537,
		      "NAD83(NSRS2007) / Iowa North (ft US)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43.26666666666667 +lat_2=42.06666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=41.5 +lon_0=-93.5 +x_0=1500000 +y_0=999999.99");
    add_proj4text (p, 2,
		   "99898402 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-f");
    add_proj4text (p, 3, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Iowa North (ft US)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 8,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 9,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",43.26");
    add_srs_wkt (p, 10,
		 "666666666667],PARAMETER[\"standard_parallel_2\",42.06666");
    add_srs_wkt (p, 11,
		 "666666667],PARAMETER[\"latitude_of_origin\",41.5],PARAME");
    add_srs_wkt (p, 12,
		 "TER[\"central_meridian\",-93.5],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 13,
		 "g\",4921250],PARAMETER[\"false_northing\",3280833.333300");
    add_srs_wkt (p, 14,
		 "001],AUTHORITY[\"EPSG\",\"3537\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 15, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3538, "epsg", 3538,
		      "NAD83(NSRS2007) / Iowa South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.78333333333333 +lat_2=40.61666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=40 +lon_0=-93.5 +x_0=500000 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Iowa South\",GEOGCS[\"NAD83(N");
    add_srs_wkt (p, 1,
		 "SRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Syst");
    add_srs_wkt (p, 2,
		 "em_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"475");
    add_srs_wkt (p, 7,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"sta");
    add_srs_wkt (p, 9,
		 "ndard_parallel_1\",41.78333333333333],PARAMETER[\"standa");
    add_srs_wkt (p, 10,
		 "rd_parallel_2\",40.61666666666667],PARAMETER[\"latitude_");
    add_srs_wkt (p, 11,
		 "of_origin\",40],PARAMETER[\"central_meridian\",-93.5],PA");
    add_srs_wkt (p, 12,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 13,
		 "hing\",0],AUTHORITY[\"EPSG\",\"3538\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 14, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3539, "epsg", 3539,
		      "NAD83(NSRS2007) / Iowa South (ft US)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.78333333333333 +lat_2=40.61666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=40 +lon_0=-93.5 +x_0=500000.00001016 +y_0=0 +");
    add_proj4text (p, 2,
		   "ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_proj4text (p, 3, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Iowa South (ft US)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 8,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 9,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",41.78");
    add_srs_wkt (p, 10,
		 "333333333333],PARAMETER[\"standard_parallel_2\",40.61666");
    add_srs_wkt (p, 11,
		 "666666667],PARAMETER[\"latitude_of_origin\",40],PARAMETE");
    add_srs_wkt (p, 12,
		 "R[\"central_meridian\",-93.5],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 13,
		 ",1640416.6667],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 14,
		 "[\"EPSG\",\"3539\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3540, "epsg", 3540,
		      "NAD83(NSRS2007) / Kansas North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.78333333333333 +lat_2=38.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=38.33333333333334 +lon_0=-98 +x_0=400000 +y_0");
    add_proj4text (p, 2,
		   "=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_proj4text (p, 3, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Kansas North\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_1\",39.78333333333333],PARAMETER[\"stan");
    add_srs_wkt (p, 10,
		 "dard_parallel_2\",38.71666666666667],PARAMETER[\"latitud");
    add_srs_wkt (p, 11,
		 "e_of_origin\",38.33333333333334],PARAMETER[\"central_mer");
    add_srs_wkt (p, 12,
		 "idian\",-98],PARAMETER[\"false_easting\",400000],PARAMET");
    add_srs_wkt (p, 13,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3540\"],AX");
    add_srs_wkt (p, 14, "IS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3541, "epsg", 3541,
		      "NAD83(NSRS2007) / Kansas North (ft US)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.78333333333333 +lat_2=38.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=38.33333333333334 +lon_0=-98 +x_0=399999.9999");
    add_proj4text (p, 2,
		   "8984 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=u");
    add_proj4text (p, 3, "s-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Kansas North (ft US)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refe");
    add_srs_wkt (p, 2,
		 "rence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25");
    add_srs_wkt (p, 3,
		 "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4759\"]],UNIT[\"US survey foot\",0.30480060960121");
    add_srs_wkt (p, 8,
		 "92,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Co");
    add_srs_wkt (p, 9,
		 "nformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",3");
    add_srs_wkt (p, 10,
		 "9.78333333333333],PARAMETER[\"standard_parallel_2\",38.7");
    add_srs_wkt (p, 11,
		 "1666666666667],PARAMETER[\"latitude_of_origin\",38.33333");
    add_srs_wkt (p, 12,
		 "333333334],PARAMETER[\"central_meridian\",-98],PARAMETER");
    add_srs_wkt (p, 13,
		 "[\"false_easting\",1312333.3333],PARAMETER[\"false_north");
    add_srs_wkt (p, 14,
		 "ing\",0],AUTHORITY[\"EPSG\",\"3541\"],AXIS[\"X\",EAST],A");
    add_srs_wkt (p, 15, "XIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3542, "epsg", 3542,
		      "NAD83(NSRS2007) / Kansas South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.56666666666667 +lat_2=37.26666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.66666666666666 +lon_0=-98.5 +x_0=400000 +y");
    add_proj4text (p, 2,
		   "_0=400000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +");
    add_proj4text (p, 3, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Kansas South\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_1\",38.56666666666667],PARAMETER[\"stan");
    add_srs_wkt (p, 10,
		 "dard_parallel_2\",37.26666666666667],PARAMETER[\"latitud");
    add_srs_wkt (p, 11,
		 "e_of_origin\",36.66666666666666],PARAMETER[\"central_mer");
    add_srs_wkt (p, 12,
		 "idian\",-98.5],PARAMETER[\"false_easting\",400000],PARAM");
    add_srs_wkt (p, 13,
		 "ETER[\"false_northing\",400000],AUTHORITY[\"EPSG\",\"354");
    add_srs_wkt (p, 14, "2\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3543, "epsg", 3543,
		      "NAD83(NSRS2007) / Kansas South (ft US)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.56666666666667 +lat_2=37.26666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.66666666666666 +lon_0=-98.5 +x_0=399999.99");
    add_proj4text (p, 2,
		   "998984 +y_0=399999.99998984 +ellps=GRS80 +towgs84=0,0,0,");
    add_proj4text (p, 3, "0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Kansas South (ft US)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refe");
    add_srs_wkt (p, 2,
		 "rence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25");
    add_srs_wkt (p, 3,
		 "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4759\"]],UNIT[\"US survey foot\",0.30480060960121");
    add_srs_wkt (p, 8,
		 "92,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Co");
    add_srs_wkt (p, 9,
		 "nformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",3");
    add_srs_wkt (p, 10,
		 "8.56666666666667],PARAMETER[\"standard_parallel_2\",37.2");
    add_srs_wkt (p, 11,
		 "6666666666667],PARAMETER[\"latitude_of_origin\",36.66666");
    add_srs_wkt (p, 12,
		 "666666666],PARAMETER[\"central_meridian\",-98.5],PARAMET");
    add_srs_wkt (p, 13,
		 "ER[\"false_easting\",1312333.3333],PARAMETER[\"false_nor");
    add_srs_wkt (p, 14,
		 "thing\",1312333.3333],AUTHORITY[\"EPSG\",\"3543\"],AXIS[");
    add_srs_wkt (p, 15, "\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3544, "epsg", 3544,
		      "NAD83(NSRS2007) / Kentucky North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.96666666666667 +lat_2=38.96666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=37.5 +lon_0=-84.25 +x_0=500000 +y_0=0 +ellps=");
    add_proj4text (p, 2, "GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Kentucky North\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_");
    add_srs_wkt (p, 2,
		 "System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",37.96666666666667],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",38.96666666666667],PARAMETER[\"latitu");
    add_srs_wkt (p, 11,
		 "de_of_origin\",37.5],PARAMETER[\"central_meridian\",-84.");
    add_srs_wkt (p, 12,
		 "25],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 13,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"3544\"],AXIS[\"X\",");
    add_srs_wkt (p, 14, "EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3545, "epsg", 3545,
		      "NAD83(NSRS2007) / Kentucky North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.96666666666667 +lat_2=38.96666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=37.5 +lon_0=-84.25 +x_0=500000.0001016001 +y_");
    add_proj4text (p, 2,
		   "0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no");
    add_proj4text (p, 3, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Kentucky North (ftUS)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2");
    add_srs_wkt (p, 3,
		 "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 9,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 10,
		 "37.96666666666667],PARAMETER[\"standard_parallel_2\",38.");
    add_srs_wkt (p, 11,
		 "96666666666667],PARAMETER[\"latitude_of_origin\",37.5],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"central_meridian\",-84.25],PARAMETER[\"false_");
    add_srs_wkt (p, 13,
		 "easting\",1640416.667],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 14,
		 "UTHORITY[\"EPSG\",\"3545\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 15, "NORTH]]");
    p = add_epsg_def (first, last, 3546, "epsg", 3546,
		      "NAD83(NSRS2007) / Kentucky Single Zone");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.08333333333334 +lat_2=38.66666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=36.33333333333334 +lon_0=-85.75 +x_0=1500000 ");
    add_proj4text (p, 2,
		   "+y_0=1000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 3, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Kentucky Single Zone\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refe");
    add_srs_wkt (p, 2,
		 "rence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25");
    add_srs_wkt (p, 3,
		 "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 8,
		 "01\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"standard_parallel_1\",37.08333333333334],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"standard_parallel_2\",38.66666666666666],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "latitude_of_origin\",36.33333333333334],PARAMETER[\"cent");
    add_srs_wkt (p, 12,
		 "ral_meridian\",-85.75],PARAMETER[\"false_easting\",15000");
    add_srs_wkt (p, 13,
		 "00],PARAMETER[\"false_northing\",1000000],AUTHORITY[\"EP");
    add_srs_wkt (p, 14, "SG\",\"3546\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3547, "epsg", 3547,
		      "NAD83(NSRS2007) / Kentucky Single Zone (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.08333333333334 +lat_2=38.66666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=36.33333333333334 +lon_0=-85.75 +x_0=1500000 ");
    add_proj4text (p, 2,
		   "+y_0=999999.9998983998 +ellps=GRS80 +towgs84=0,0,0,0,0,0");
    add_proj4text (p, 3, ",0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Kentucky Single Zone (ftUS)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spati");
    add_srs_wkt (p, 2,
		 "al_Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 3,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 5,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 6,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006");
    add_srs_wkt (p, 8,
		 "096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lam");
    add_srs_wkt (p, 9,
		 "bert_Conformal_Conic_2SP\"],PARAMETER[\"standard_paralle");
    add_srs_wkt (p, 10,
		 "l_1\",37.08333333333334],PARAMETER[\"standard_parallel_2");
    add_srs_wkt (p, 11,
		 "\",38.66666666666666],PARAMETER[\"latitude_of_origin\",3");
    add_srs_wkt (p, 12,
		 "6.33333333333334],PARAMETER[\"central_meridian\",-85.75]");
    add_srs_wkt (p, 13,
		 ",PARAMETER[\"false_easting\",4921250],PARAMETER[\"false_");
    add_srs_wkt (p, 14,
		 "northing\",3280833.333],AUTHORITY[\"EPSG\",\"3547\"],AXI");
    add_srs_wkt (p, 15, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3548, "epsg", 3548,
		      "NAD83(NSRS2007) / Kentucky South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.93333333333333 +lat_2=36.73333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=36.33333333333334 +lon_0=-85.75 +x_0=500000 +");
    add_proj4text (p, 2,
		   "y_0=500000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m ");
    add_proj4text (p, 3, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Kentucky South\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_");
    add_srs_wkt (p, 2,
		 "System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",37.93333333333333],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",36.73333333333333],PARAMETER[\"latitu");
    add_srs_wkt (p, 11,
		 "de_of_origin\",36.33333333333334],PARAMETER[\"central_me");
    add_srs_wkt (p, 12,
		 "ridian\",-85.75],PARAMETER[\"false_easting\",500000],PAR");
    add_srs_wkt (p, 13,
		 "AMETER[\"false_northing\",500000],AUTHORITY[\"EPSG\",\"3");
    add_srs_wkt (p, 14, "548\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3549, "epsg", 3549,
		      "NAD83(NSRS2007) / Kentucky South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.93333333333333 +lat_2=36.73333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=36.33333333333334 +lon_0=-85.75 +x_0=500000.0");
    add_proj4text (p, 2,
		   "001016001 +y_0=500000.0001016001 +ellps=GRS80 +towgs84=0");
    add_proj4text (p, 3, ",0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Kentucky South (ftUS)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2");
    add_srs_wkt (p, 3,
		 "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 9,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 10,
		 "37.93333333333333],PARAMETER[\"standard_parallel_2\",36.");
    add_srs_wkt (p, 11,
		 "73333333333333],PARAMETER[\"latitude_of_origin\",36.3333");
    add_srs_wkt (p, 12,
		 "3333333334],PARAMETER[\"central_meridian\",-85.75],PARAM");
    add_srs_wkt (p, 13,
		 "ETER[\"false_easting\",1640416.667],PARAMETER[\"false_no");
    add_srs_wkt (p, 14,
		 "rthing\",1640416.667],AUTHORITY[\"EPSG\",\"3549\"],AXIS[");
    add_srs_wkt (p, 15, "\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3550, "epsg", 3550,
		      "NAD83(NSRS2007) / Louisiana North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=32.66666666666666 +lat_2=31.16666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=30.5 +lon_0=-92.5 +x_0=1000000 +y_0=0 +ellps=");
    add_proj4text (p, 2, "GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Louisiana North\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference");
    add_srs_wkt (p, 2,
		 "_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 3,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 6,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 8,
		 "],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"standard_parallel_1\",32.66666666666666],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "tandard_parallel_2\",31.16666666666667],PARAMETER[\"lati");
    add_srs_wkt (p, 11,
		 "tude_of_origin\",30.5],PARAMETER[\"central_meridian\",-9");
    add_srs_wkt (p, 12,
		 "2.5],PARAMETER[\"false_easting\",1000000],PARAMETER[\"fa");
    add_srs_wkt (p, 13,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"3550\"],AXIS[\"X\"");
    add_srs_wkt (p, 14, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3551, "epsg", 3551,
		      "NAD83(NSRS2007) / Louisiana North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=32.66666666666666 +lat_2=31.16666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=30.5 +lon_0=-92.5 +x_0=999999.9999898402 +y_0");
    add_proj4text (p, 2,
		   "=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_");
    add_proj4text (p, 3, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Louisiana North (ftUS)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Re");
    add_srs_wkt (p, 2,
		 "ference_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.");
    add_srs_wkt (p, 3,
		 "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 9,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 10,
		 "32.66666666666666],PARAMETER[\"standard_parallel_2\",31.");
    add_srs_wkt (p, 11,
		 "16666666666667],PARAMETER[\"latitude_of_origin\",30.5],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"central_meridian\",-92.5],PARAMETER[\"false_e");
    add_srs_wkt (p, 13,
		 "asting\",3280833.333300001],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 14,
		 ",0],AUTHORITY[\"EPSG\",\"3551\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 15, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3552, "epsg", 3552,
		      "NAD83(NSRS2007) / Louisiana South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.7 +lat_2=29.3 +lat_0=28.5 +lon_0=-91");
    add_proj4text (p, 1,
		   ".33333333333333 +x_0=1000000 +y_0=0 +ellps=GRS80 +towgs8");
    add_proj4text (p, 2, "4=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Louisiana South\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference");
    add_srs_wkt (p, 2,
		 "_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 3,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 6,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 8,
		 "],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"standard_parallel_1\",30.7],PARAMETER[\"standard_paral");
    add_srs_wkt (p, 10,
		 "lel_2\",29.3],PARAMETER[\"latitude_of_origin\",28.5],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"central_meridian\",-91.33333333333333],PARAMETE");
    add_srs_wkt (p, 12,
		 "R[\"false_easting\",1000000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 13,
		 ",0],AUTHORITY[\"EPSG\",\"3552\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 14, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3553, "epsg", 3553,
		      "NAD83(NSRS2007) / Louisiana South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.7 +lat_2=29.3 +lat_0=28.5 +lon_0=-91");
    add_proj4text (p, 1,
		   ".33333333333333 +x_0=999999.9999898402 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Louisiana South (ftUS)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Re");
    add_srs_wkt (p, 2,
		 "ference_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.");
    add_srs_wkt (p, 3,
		 "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 9,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 10,
		 "30.7],PARAMETER[\"standard_parallel_2\",29.3],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"latitude_of_origin\",28.5],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 12,
		 "n\",-91.33333333333333],PARAMETER[\"false_easting\",3280");
    add_srs_wkt (p, 13,
		 "833.333300001],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 14,
		 "[\"EPSG\",\"3553\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3554, "epsg", 3554,
		      "NAD83(NSRS2007) / Maine CS2000 Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=43.5 +lon_0=-69.125 +k=0.99998 +x_0=5");
    add_proj4text (p, 1,
		   "00000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Maine CS2000 Central\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refe");
    add_srs_wkt (p, 2,
		 "rence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25");
    add_srs_wkt (p, 3,
		 "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 8,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 9,
		 "atitude_of_origin\",43.5],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 10,
		 ",-69.125],PARAMETER[\"scale_factor\",0.99998],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"3554\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 13, "\",NORTH]]");
    p = add_epsg_def (first, last, 3555, "epsg", 3555,
		      "NAD83(NSRS2007) / Maine CS2000 East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=43.83333333333334 +lon_0=-67.875 +k=0");
    add_proj4text (p, 1,
		   ".99998 +x_0=700000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 2, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Maine CS2000 East\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",43.83333333333334],PARAMETER[\"central_m");
    add_srs_wkt (p, 10,
		 "eridian\",-67.875],PARAMETER[\"scale_factor\",0.99998],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_easting\",700000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 12,
		 "thing\",0],AUTHORITY[\"EPSG\",\"3555\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 13, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3556, "epsg", 3556,
		      "NAD83(NSRS2007) / Maine CS2000 West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.83333333333334 +lon_0=-70.375 +k=0");
    add_proj4text (p, 1,
		   ".99998 +x_0=300000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 2, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Maine CS2000 West\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",42.83333333333334],PARAMETER[\"central_m");
    add_srs_wkt (p, 10,
		 "eridian\",-70.375],PARAMETER[\"scale_factor\",0.99998],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_easting\",300000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 12,
		 "thing\",0],AUTHORITY[\"EPSG\",\"3556\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 13, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3557, "epsg", 3557,
		      "NAD83(NSRS2007) / Maine East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=43.66666666666666 +lon_0=-68.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=300000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0");
    add_proj4text (p, 2, ",0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Maine East\",GEOGCS[\"NAD83(N");
    add_srs_wkt (p, 1,
		 "SRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Syst");
    add_srs_wkt (p, 2,
		 "em_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"475");
    add_srs_wkt (p, 7,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",43.66666666666666],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 10,
		 "n\",-68.5],PARAMETER[\"scale_factor\",0.9999],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_easting\",300000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"3557\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 13, "\",NORTH]]");
    p = add_epsg_def (first, last, 3558, "epsg", 3558,
		      "NAD83(NSRS2007) / Maine West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.83333333333334 +lon_0=-70.16666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999966667 +x_0=900000 +y_0=0 +ellps=GRS80 +t");
    add_proj4text (p, 2, "owgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Maine West\",GEOGCS[\"NAD83(N");
    add_srs_wkt (p, 1,
		 "SRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Syst");
    add_srs_wkt (p, 2,
		 "em_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"475");
    add_srs_wkt (p, 7,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",42.83333333333334],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 10,
		 "n\",-70.16666666666667],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 11,
		 "966667],PARAMETER[\"false_easting\",900000],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3558\"],AXIS[\"");
    add_srs_wkt (p, 13, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3559, "epsg", 3559,
		      "NAD83(NSRS2007) / Maryland");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.45 +lat_2=38.3 +lat_0=37.66666666666");
    add_proj4text (p, 1,
		   "666 +lon_0=-77 +x_0=400000 +y_0=0 +ellps=GRS80 +towgs84=");
    add_proj4text (p, 2, "0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Maryland\",GEOGCS[\"NAD83(NSR");
    add_srs_wkt (p, 1,
		 "S2007)\",DATUM[\"NAD83_National_Spatial_Reference_System");
    add_srs_wkt (p, 2,
		 "_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 5,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4759\"");
    add_srs_wkt (p, 7,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",39.45],PARAMETER[\"standard_parallel_2\"");
    add_srs_wkt (p, 10,
		 ",38.3],PARAMETER[\"latitude_of_origin\",37.6666666666666");
    add_srs_wkt (p, 11,
		 "6],PARAMETER[\"central_meridian\",-77],PARAMETER[\"false");
    add_srs_wkt (p, 12,
		 "_easting\",400000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 13,
		 "RITY[\"EPSG\",\"3559\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORT");
    add_srs_wkt (p, 14, "H]]");
    p = add_epsg_def (first, last, 3560, "epsg", 3560,
		      "NAD83 / Utah North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.78333333333333 +lat_2=40.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=40.33333333333334 +lon_0=-111.5 +x_0=500000.0");
    add_proj4text (p, 2,
		   "0001016 +y_0=999999.9999898402 +ellps=GRS80 +datum=NAD83");
    add_proj4text (p, 3, " +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Utah North (ftUS)\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal_C");
    add_srs_wkt (p, 8,
		 "onic_2SP\"],PARAMETER[\"standard_parallel_1\",41.7833333");
    add_srs_wkt (p, 9,
		 "3333333],PARAMETER[\"standard_parallel_2\",40.7166666666");
    add_srs_wkt (p, 10,
		 "6667],PARAMETER[\"latitude_of_origin\",40.33333333333334");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"central_meridian\",-111.5],PARAMETER[\"fal");
    add_srs_wkt (p, 12,
		 "se_easting\",1640416.6667],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 13,
		 "3280833.333300001],AUTHORITY[\"EPSG\",\"3560\"],AXIS[\"X");
    add_srs_wkt (p, 14, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3561, "epsg", 3561,
		      "Old Hawaiian / Hawaii zone 1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=18.83333333333333 +lon_0=-155.5 +k=0.");
    add_proj4text (p, 1,
		   "999966667 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +u");
    add_proj4text (p, 2, "nits=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Old Hawaiian / Hawaii zone 1\",GEOGCS[\"Old Haw");
    add_srs_wkt (p, 1,
		 "aiian\",DATUM[\"Old_Hawaiian\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6135\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4135\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",18.83333333333");
    add_srs_wkt (p, 9,
		 "333],PARAMETER[\"central_meridian\",-155.5],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",0.999966667],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"3561\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3562, "epsg", 3562,
		      "Old Hawaiian / Hawaii zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=20.33333333333333 +lon_0=-156.6666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999966667 +x_0=152400.3048006096 +y_0=0 +ell");
    add_proj4text (p, 2, "ps=clrk66 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Old Hawaiian / Hawaii zone 2\",GEOGCS[\"Old Haw");
    add_srs_wkt (p, 1,
		 "aiian\",DATUM[\"Old_Hawaiian\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6135\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4135\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",20.33333333333");
    add_srs_wkt (p, 9,
		 "333],PARAMETER[\"central_meridian\",-156.6666666666667],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"scale_factor\",0.999966667],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 12,
		 "ORITY[\"EPSG\",\"3562\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 13, "TH]]");
    p = add_epsg_def (first, last, 3563, "epsg", 3563,
		      "Old Hawaiian / Hawaii zone 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=21.16666666666667 +lon_0=-158 +k=0.99");
    add_proj4text (p, 1,
		   "999 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +units=u");
    add_proj4text (p, 2, "s-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Old Hawaiian / Hawaii zone 3\",GEOGCS[\"Old Haw");
    add_srs_wkt (p, 1,
		 "aiian\",DATUM[\"Old_Hawaiian\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6135\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4135\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",21.16666666666");
    add_srs_wkt (p, 9,
		 "667],PARAMETER[\"central_meridian\",-158],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",0.99999],PARAMETER[\"false_easting\",500000");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3");
    add_srs_wkt (p, 12, "563\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3564, "epsg", 3564,
		      "Old Hawaiian / Hawaii zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=21.83333333333333 +lon_0=-159.5 +k=0.");
    add_proj4text (p, 1,
		   "99999 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +units");
    add_proj4text (p, 2, "=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Old Hawaiian / Hawaii zone 4\",GEOGCS[\"Old Haw");
    add_srs_wkt (p, 1,
		 "aiian\",DATUM[\"Old_Hawaiian\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6135\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4135\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",21.83333333333");
    add_srs_wkt (p, 9,
		 "333],PARAMETER[\"central_meridian\",-159.5],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",0.99999],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "3564\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3565, "epsg", 3565,
		      "Old Hawaiian / Hawaii zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=21.66666666666667 +lon_0=-160.1666666");
    add_proj4text (p, 1,
		   "666667 +k=1 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 ");
    add_proj4text (p, 2, "+units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Old Hawaiian / Hawaii zone 5\",GEOGCS[\"Old Haw");
    add_srs_wkt (p, 1,
		 "aiian\",DATUM[\"Old_Hawaiian\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6135\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4135\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",21.66666666666");
    add_srs_wkt (p, 9,
		 "667],PARAMETER[\"central_meridian\",-160.1666666666667],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12, "G\",\"3565\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3566, "epsg", 3566,
		      "NAD83 / Utah Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.65 +lat_2=39.01666666666667 +lat_0=3");
    add_proj4text (p, 1,
		   "8.33333333333334 +lon_0=-111.5 +x_0=500000.00001016 +y_0");
    add_proj4text (p, 2,
		   "=2000000.00001016 +ellps=GRS80 +datum=NAD83 +units=us-ft");
    add_proj4text (p, 3, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Utah Central (ftUS)\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal_");
    add_srs_wkt (p, 8,
		 "Conic_2SP\"],PARAMETER[\"standard_parallel_1\",40.65],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"standard_parallel_2\",39.01666666666667],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"latitude_of_origin\",38.33333333333334],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"central_meridian\",-111.5],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",1640416.6667],PARAMETER[\"false_northing\",6561666.6667");
    add_srs_wkt (p, 13,
		 "00001],AUTHORITY[\"EPSG\",\"3566\"],AXIS[\"X\",EAST],AXI");
    add_srs_wkt (p, 14, "S[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3567, "epsg", 3567,
		      "NAD83 / Utah South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.35 +lat_2=37.21666666666667 +lat_0=3");
    add_proj4text (p, 1,
		   "6.66666666666666 +lon_0=-111.5 +x_0=500000.00001016 +y_0");
    add_proj4text (p, 2,
		   "=3000000 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_proj4text (p, 3, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Utah South (ftUS)\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal_C");
    add_srs_wkt (p, 8,
		 "onic_2SP\"],PARAMETER[\"standard_parallel_1\",38.35],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"standard_parallel_2\",37.21666666666667],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"latitude_of_origin\",36.66666666666666],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"central_meridian\",-111.5],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",1640416.6667],PARAMETER[\"false_northing\",9842500.0000");
    add_srs_wkt (p, 13,
		 "00002],AUTHORITY[\"EPSG\",\"3567\"],AXIS[\"X\",EAST],AXI");
    add_srs_wkt (p, 14, "S[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3568, "epsg", 3568,
		      "NAD83(HARN) / Utah North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.78333333333333 +lat_2=40.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=40.33333333333334 +lon_0=-111.5 +x_0=500000.0");
    add_proj4text (p, 2,
		   "0001016 +y_0=999999.9999898402 +ellps=GRS80 +units=us-ft");
    add_proj4text (p, 3, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Utah North (ftUS)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",0.30");
    add_srs_wkt (p, 7,
		 "48006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 9,
		 "rallel_1\",41.78333333333333],PARAMETER[\"standard_paral");
    add_srs_wkt (p, 10,
		 "lel_2\",40.71666666666667],PARAMETER[\"latitude_of_origi");
    add_srs_wkt (p, 11,
		 "n\",40.33333333333334],PARAMETER[\"central_meridian\",-1");
    add_srs_wkt (p, 12,
		 "11.5],PARAMETER[\"false_easting\",1640416.6667],PARAMETE");
    add_srs_wkt (p, 13,
		 "R[\"false_northing\",3280833.333300001],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 14, "\",\"3568\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3569, "epsg", 3569,
		      "NAD83(HARN) / Utah Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.65 +lat_2=39.01666666666667 +lat_0=3");
    add_proj4text (p, 1,
		   "8.33333333333334 +lon_0=-111.5 +x_0=500000.00001016 +y_0");
    add_proj4text (p, 2,
		   "=2000000.00001016 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Utah Central (ftUS)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",");
    add_srs_wkt (p, 7,
		 "0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECT");
    add_srs_wkt (p, 8,
		 "ION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standar");
    add_srs_wkt (p, 9,
		 "d_parallel_1\",40.65],PARAMETER[\"standard_parallel_2\",");
    add_srs_wkt (p, 10,
		 "39.01666666666667],PARAMETER[\"latitude_of_origin\",38.3");
    add_srs_wkt (p, 11,
		 "3333333333334],PARAMETER[\"central_meridian\",-111.5],PA");
    add_srs_wkt (p, 12,
		 "RAMETER[\"false_easting\",1640416.6667],PARAMETER[\"fals");
    add_srs_wkt (p, 13,
		 "e_northing\",6561666.666700001],AUTHORITY[\"EPSG\",\"356");
    add_srs_wkt (p, 14, "9\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3570, "epsg", 3570,
		      "NAD83(HARN) / Utah South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.35 +lat_2=37.21666666666667 +lat_0=3");
    add_proj4text (p, 1,
		   "6.66666666666666 +lon_0=-111.5 +x_0=500000.00001016 +y_0");
    add_proj4text (p, 2, "=3000000 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Utah South (ftUS)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",0.30");
    add_srs_wkt (p, 7,
		 "48006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 9,
		 "rallel_1\",38.35],PARAMETER[\"standard_parallel_2\",37.2");
    add_srs_wkt (p, 10,
		 "1666666666667],PARAMETER[\"latitude_of_origin\",36.66666");
    add_srs_wkt (p, 11,
		 "666666666],PARAMETER[\"central_meridian\",-111.5],PARAME");
    add_srs_wkt (p, 12,
		 "TER[\"false_easting\",1640416.6667],PARAMETER[\"false_no");
    add_srs_wkt (p, 13,
		 "rthing\",9842500.000000002],AUTHORITY[\"EPSG\",\"3570\"]");
    add_srs_wkt (p, 14, ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3571, "epsg", 3571,
		      "WGS 84 / North Pole LAEA Bering Sea");
    add_proj4text (p, 0,
		   "+proj=laea +lat_0=90 +lon_0=180 +x_0=0 +y_0=0 +ellps=WGS");
    add_proj4text (p, 1, "84 +datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / North Pole LAEA Bering Sea\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "WGS 84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,");
    add_srs_wkt (p, 2,
		 "298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 8,
		 "of_center\",90],PARAMETER[\"longitude_of_center\",180],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",0],PARAMETER[\"false_northing");
    add_srs_wkt (p, 10,
		 "\",0],AUTHORITY[\"EPSG\",\"3571\"],AXIS[\"X\",UNKNOWN],A");
    add_srs_wkt (p, 11, "XIS[\"Y\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3572, "epsg", 3572,
		      "WGS 84 / North Pole LAEA Alaska");
    add_proj4text (p, 0,
		   "+proj=laea +lat_0=90 +lon_0=-150 +x_0=0 +y_0=0 +ellps=WG");
    add_proj4text (p, 1, "S84 +datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / North Pole LAEA Alaska\",GEOGCS[\"WGS ");
    add_srs_wkt (p, 1,
		 "84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.");
    add_srs_wkt (p, 2,
		 "257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"L");
    add_srs_wkt (p, 7,
		 "ambert_Azimuthal_Equal_Area\"],PARAMETER[\"latitude_of_c");
    add_srs_wkt (p, 8,
		 "enter\",90],PARAMETER[\"longitude_of_center\",-150],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"false_easting\",0],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 10,
		 "0],AUTHORITY[\"EPSG\",\"3572\"],AXIS[\"X\",UNKNOWN],AXIS");
    add_srs_wkt (p, 11, "[\"Y\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3573, "epsg", 3573,
		      "WGS 84 / North Pole LAEA Canada");
    add_proj4text (p, 0,
		   "+proj=laea +lat_0=90 +lon_0=-100 +x_0=0 +y_0=0 +ellps=WG");
    add_proj4text (p, 1, "S84 +datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / North Pole LAEA Canada\",GEOGCS[\"WGS ");
    add_srs_wkt (p, 1,
		 "84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.");
    add_srs_wkt (p, 2,
		 "257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"L");
    add_srs_wkt (p, 7,
		 "ambert_Azimuthal_Equal_Area\"],PARAMETER[\"latitude_of_c");
    add_srs_wkt (p, 8,
		 "enter\",90],PARAMETER[\"longitude_of_center\",-100],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"false_easting\",0],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 10,
		 "0],AUTHORITY[\"EPSG\",\"3573\"],AXIS[\"X\",UNKNOWN],AXIS");
    add_srs_wkt (p, 11, "[\"Y\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3574, "epsg", 3574,
		      "WGS 84 / North Pole LAEA Atlantic");
    add_proj4text (p, 0,
		   "+proj=laea +lat_0=90 +lon_0=-40 +x_0=0 +y_0=0 +ellps=WGS");
    add_proj4text (p, 1, "84 +datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / North Pole LAEA Atlantic\",GEOGCS[\"WG");
    add_srs_wkt (p, 1,
		 "S 84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,29");
    add_srs_wkt (p, 2,
		 "8.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "center\",90],PARAMETER[\"longitude_of_center\",-40],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"false_easting\",0],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 10,
		 "0],AUTHORITY[\"EPSG\",\"3574\"],AXIS[\"X\",UNKNOWN],AXIS");
    add_srs_wkt (p, 11, "[\"Y\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3575, "epsg", 3575,
		      "WGS 84 / North Pole LAEA Europe");
    add_proj4text (p, 0,
		   "+proj=laea +lat_0=90 +lon_0=10 +x_0=0 +y_0=0 +ellps=WGS8");
    add_proj4text (p, 1, "4 +datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / North Pole LAEA Europe\",GEOGCS[\"WGS ");
    add_srs_wkt (p, 1,
		 "84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.");
    add_srs_wkt (p, 2,
		 "257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"L");
    add_srs_wkt (p, 7,
		 "ambert_Azimuthal_Equal_Area\"],PARAMETER[\"latitude_of_c");
    add_srs_wkt (p, 8,
		 "enter\",90],PARAMETER[\"longitude_of_center\",10],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"false_easting\",0],PARAMETER[\"false_northing\",0]");
    add_srs_wkt (p, 10,
		 ",AUTHORITY[\"EPSG\",\"3575\"],AXIS[\"X\",UNKNOWN],AXIS[\"");
    add_srs_wkt (p, 11, "Y\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3576, "epsg", 3576,
		      "WGS 84 / North Pole LAEA Russia");
    add_proj4text (p, 0,
		   "+proj=laea +lat_0=90 +lon_0=90 +x_0=0 +y_0=0 +ellps=WGS8");
    add_proj4text (p, 1, "4 +datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / North Pole LAEA Russia\",GEOGCS[\"WGS ");
    add_srs_wkt (p, 1,
		 "84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.");
    add_srs_wkt (p, 2,
		 "257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"L");
    add_srs_wkt (p, 7,
		 "ambert_Azimuthal_Equal_Area\"],PARAMETER[\"latitude_of_c");
    add_srs_wkt (p, 8,
		 "enter\",90],PARAMETER[\"longitude_of_center\",90],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"false_easting\",0],PARAMETER[\"false_northing\",0]");
    add_srs_wkt (p, 10,
		 ",AUTHORITY[\"EPSG\",\"3576\"],AXIS[\"X\",UNKNOWN],AXIS[\"");
    add_srs_wkt (p, 11, "Y\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3577, "epsg", 3577,
		      "GDA94 / Australian Albers");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=-18 +lat_2=-36 +lat_0=0 +lon_0=132 +x_0");
    add_proj4text (p, 1,
		   "=0 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDA94 / Australian Albers\",GEOGCS[\"GDA94\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"Geocentric_Datum_of_Australia_1994\",SPHEROID[\"GR");
    add_srs_wkt (p, 2,
		 "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019");
    add_srs_wkt (p, 3,
		 "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4283\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Albers_Conic_");
    add_srs_wkt (p, 8,
		 "Equal_Area\"],PARAMETER[\"standard_parallel_1\",-18],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"standard_parallel_2\",-36],PARAMETER[\"latitude");
    add_srs_wkt (p, 10,
		 "_of_center\",0],PARAMETER[\"longitude_of_center\",132],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_easting\",0],PARAMETER[\"false_northing");
    add_srs_wkt (p, 12,
		 "\",0],AUTHORITY[\"EPSG\",\"3577\"],AXIS[\"Easting\",EAST");
    add_srs_wkt (p, 13, "],AXIS[\"Northing\",NORTH]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_12 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 3578, "epsg", 3578, "NAD83 / Yukon Albers");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=61.66666666666666 +lat_2=68 +lat_0=59 +");
    add_proj4text (p, 1,
		   "lon_0=-132.5 +x_0=500000 +y_0=500000 +ellps=GRS80 +datum");
    add_proj4text (p, 2, "=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Yukon Albers\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Albers_Conic_Equal_Area\"],PARAMETER[\"standard_par");
    add_srs_wkt (p, 8,
		 "allel_1\",61.66666666666666],PARAMETER[\"standard_parall");
    add_srs_wkt (p, 9,
		 "el_2\",68],PARAMETER[\"latitude_of_center\",59],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"longitude_of_center\",-132.5],PARAMETER[\"false_east");
    add_srs_wkt (p, 11,
		 "ing\",500000],PARAMETER[\"false_northing\",500000],AUTHO");
    add_srs_wkt (p, 12,
		 "RITY[\"EPSG\",\"3578\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 13, "rthing\",NORTH]]");
    p = add_epsg_def (first, last, 3579, "epsg", 3579,
		      "NAD83(CSRS) / Yukon Albers");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=61.66666666666666 +lat_2=68 +lat_0=59 +");
    add_proj4text (p, 1,
		   "lon_0=-132.5 +x_0=500000 +y_0=500000 +ellps=GRS80 +units");
    add_proj4text (p, 2, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / Yukon Albers\",GEOGCS[\"NAD83(CSR");
    add_srs_wkt (p, 1,
		 "S)\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Albers_Conic_Equal_Area\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",61.66666666666666],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"standard_parallel_2\",68],PARAMETER[\"latitud");
    add_srs_wkt (p, 10,
		 "e_of_center\",59],PARAMETER[\"longitude_of_center\",-132");
    add_srs_wkt (p, 11,
		 ".5],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_northing\",500000],AUTHORITY[\"EPSG\",\"3579\"],AXIS[\"");
    add_srs_wkt (p, 13, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3580, "epsg", 3580, "NAD83 / NWT Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=62 +lat_2=70 +lat_0=0 +lon_0=-112 +x_0=");
    add_proj4text (p, 1,
		   "0 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / NWT Lambert\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard");
    add_srs_wkt (p, 8,
		 "_parallel_1\",62],PARAMETER[\"standard_parallel_2\",70],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_");
    add_srs_wkt (p, 10,
		 "meridian\",-112],PARAMETER[\"false_easting\",0],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3580\"],AXI");
    add_srs_wkt (p, 12, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3581, "epsg", 3581,
		      "NAD83(CSRS) / NWT Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=62 +lat_2=70 +lat_0=0 +lon_0=-112 +x_0=");
    add_proj4text (p, 1, "0 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / NWT Lambert\",GEOGCS[\"NAD83(CSRS");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",62],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_2\",70],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 10,
		 "\",0],PARAMETER[\"central_meridian\",-112],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",0],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 12,
		 "ITY[\"EPSG\",\"3581\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 3582, "epsg", 3582,
		      "NAD83(NSRS2007) / Maryland (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.45 +lat_2=38.3 +lat_0=37.66666666666");
    add_proj4text (p, 1,
		   "666 +lon_0=-77 +x_0=399999.9998983998 +y_0=0 +ellps=GRS8");
    add_proj4text (p, 2, "0 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Maryland (ftUS)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference");
    add_srs_wkt (p, 2,
		 "_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 3,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 6,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4759\"]],UNIT[\"US survey foot\",0.3048006096012192,AU");
    add_srs_wkt (p, 8,
		 "THORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conform");
    add_srs_wkt (p, 9,
		 "al_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",39.45]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"standard_parallel_2\",38.3],PARAMETER[\"lat");
    add_srs_wkt (p, 11,
		 "itude_of_origin\",37.66666666666666],PARAMETER[\"central");
    add_srs_wkt (p, 12,
		 "_meridian\",-77],PARAMETER[\"false_easting\",1312333.333");
    add_srs_wkt (p, 13,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3");
    add_srs_wkt (p, 14, "582\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3583, "epsg", 3583,
		      "NAD83(NSRS2007) / Massachusetts Island");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.48333333333333 +lat_2=41.28333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=41 +lon_0=-70.5 +x_0=500000 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Massachusetts Island\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refe");
    add_srs_wkt (p, 2,
		 "rence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25");
    add_srs_wkt (p, 3,
		 "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 8,
		 "01\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"standard_parallel_1\",41.48333333333333],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"standard_parallel_2\",41.28333333333333],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "latitude_of_origin\",41],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 12,
		 "-70.5],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 13,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"3583\"],AXIS[\"X");
    add_srs_wkt (p, 14, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3584, "epsg", 3584,
		      "NAD83(NSRS2007) / Massachusetts Island (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.48333333333333 +lat_2=41.28333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=41 +lon_0=-70.5 +x_0=500000.0001016001 +y_0=0");
    add_proj4text (p, 2,
		   " +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_de");
    add_proj4text (p, 3, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Massachusetts Island (ftUS)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spati");
    add_srs_wkt (p, 2,
		 "al_Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 3,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 5,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 6,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006");
    add_srs_wkt (p, 8,
		 "096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lam");
    add_srs_wkt (p, 9,
		 "bert_Conformal_Conic_2SP\"],PARAMETER[\"standard_paralle");
    add_srs_wkt (p, 10,
		 "l_1\",41.48333333333333],PARAMETER[\"standard_parallel_2");
    add_srs_wkt (p, 11,
		 "\",41.28333333333333],PARAMETER[\"latitude_of_origin\",4");
    add_srs_wkt (p, 12,
		 "1],PARAMETER[\"central_meridian\",-70.5],PARAMETER[\"fal");
    add_srs_wkt (p, 13,
		 "se_easting\",1640416.667],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 14,
		 "],AUTHORITY[\"EPSG\",\"3584\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 15, "\",NORTH]]");
    p = add_epsg_def (first, last, 3585, "epsg", 3585,
		      "NAD83(NSRS2007) / Massachusetts Mainland");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=42.68333333333333 +lat_2=41.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=41 +lon_0=-71.5 +x_0=200000 +y_0=750000 +ellp");
    add_proj4text (p, 2, "s=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Massachusetts Mainland\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Re");
    add_srs_wkt (p, 2,
		 "ference_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.");
    add_srs_wkt (p, 3,
		 "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 8,
		 "001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"standard_parallel_1\",42.68333333333333],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"standard_parallel_2\",41.71666666666667],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"latitude_of_origin\",41],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 12,
		 ",-71.5],PARAMETER[\"false_easting\",200000],PARAMETER[\"");
    add_srs_wkt (p, 13,
		 "false_northing\",750000],AUTHORITY[\"EPSG\",\"3585\"],AX");
    add_srs_wkt (p, 14, "IS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3586, "epsg", 3586,
		      "NAD83(NSRS2007) / Massachusetts Mainland (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=42.68333333333333 +lat_2=41.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=41 +lon_0=-71.5 +x_0=200000.0001016002 +y_0=7");
    add_proj4text (p, 2,
		   "50000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +");
    add_proj4text (p, 3, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Massachusetts Mainland (ftUS)");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spa");
    add_srs_wkt (p, 2,
		 "tial_Reference_System_2007\",SPHEROID[\"GRS 1980\",63781");
    add_srs_wkt (p, 3,
		 "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0");
    add_srs_wkt (p, 4,
		 ",0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 5,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 6,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.30480");
    add_srs_wkt (p, 8,
		 "06096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"L");
    add_srs_wkt (p, 9,
		 "ambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_paral");
    add_srs_wkt (p, 10,
		 "lel_1\",42.68333333333333],PARAMETER[\"standard_parallel");
    add_srs_wkt (p, 11,
		 "_2\",41.71666666666667],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 12,
		 ",41],PARAMETER[\"central_meridian\",-71.5],PARAMETER[\"f");
    add_srs_wkt (p, 13,
		 "alse_easting\",656166.667],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 14,
		 "2460625],AUTHORITY[\"EPSG\",\"3586\"],AXIS[\"X\",EAST],A");
    add_srs_wkt (p, 15, "XIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3587, "epsg", 3587,
		      "NAD83(NSRS2007) / Michigan Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.7 +lat_2=44.18333333333333 +lat_0=43");
    add_proj4text (p, 1,
		   ".31666666666667 +lon_0=-84.36666666666666 +x_0=6000000 +");
    add_proj4text (p, 2,
		   "y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_d");
    add_proj4text (p, 3, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Michigan Central\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referenc");
    add_srs_wkt (p, 2,
		 "e_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",45.7],PARAMETER[\"standard_para");
    add_srs_wkt (p, 10,
		 "llel_2\",44.18333333333333],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 11,
		 "in\",43.31666666666667],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 12,
		 "84.36666666666666],PARAMETER[\"false_easting\",6000000],");
    add_srs_wkt (p, 13,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"358");
    add_srs_wkt (p, 14, "7\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3588, "epsg", 3588,
		      "NAD83(NSRS2007) / Michigan Central (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.7 +lat_2=44.18333333333333 +lat_0=43");
    add_proj4text (p, 1,
		   ".31666666666667 +lon_0=-84.36666666666666 +x_0=5999999.9");
    add_proj4text (p, 2,
		   "99976001 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +uni");
    add_proj4text (p, 3, "ts=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Michigan Central (ft)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2");
    add_srs_wkt (p, 3,
		 "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 8,
		 ",\"9002\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"standard_parallel_1\",45.7],PARAMETER[\"stan");
    add_srs_wkt (p, 10,
		 "dard_parallel_2\",44.18333333333333],PARAMETER[\"latitud");
    add_srs_wkt (p, 11,
		 "e_of_origin\",43.31666666666667],PARAMETER[\"central_mer");
    add_srs_wkt (p, 12,
		 "idian\",-84.36666666666666],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 13,
		 "19685039.37],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 14, "EPSG\",\"3588\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3589, "epsg", 3589,
		      "NAD83(NSRS2007) / Michigan North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.08333333333334 +lat_2=45.48333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=44.78333333333333 +lon_0=-87 +x_0=8000000 +y_");
    add_proj4text (p, 2,
		   "0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_def");
    add_proj4text (p, 3, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Michigan North\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_");
    add_srs_wkt (p, 2,
		 "System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",47.08333333333334],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",45.48333333333333],PARAMETER[\"latitu");
    add_srs_wkt (p, 11,
		 "de_of_origin\",44.78333333333333],PARAMETER[\"central_me");
    add_srs_wkt (p, 12,
		 "ridian\",-87],PARAMETER[\"false_easting\",8000000],PARAM");
    add_srs_wkt (p, 13,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3589\"],");
    add_srs_wkt (p, 14, "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3590, "epsg", 3590,
		      "NAD83(NSRS2007) / Michigan North (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.08333333333334 +lat_2=45.48333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=44.78333333333333 +lon_0=-87 +x_0=7999999.999");
    add_proj4text (p, 2,
		   "968001 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units");
    add_proj4text (p, 3, "=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Michigan North (ft)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 8,
		 "9002\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"standard_parallel_1\",47.08333333333334],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"standard_parallel_2\",45.48333333333333],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"latitude_of_origin\",44.78333333333333],PARAMETER[\"c");
    add_srs_wkt (p, 12,
		 "entral_meridian\",-87],PARAMETER[\"false_easting\",26246");
    add_srs_wkt (p, 13,
		 "719.16],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 14, "\",\"3590\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3591, "epsg", 3591,
		      "NAD83(NSRS2007) / Michigan Oblique Mercator");
    add_proj4text (p, 0,
		   "+proj=omerc +lat_0=45.30916666666666 +lonc=-86 +alpha=33");
    add_proj4text (p, 1,
		   "7.25556 +k=0.9996 +x_0=2546731.496 +y_0=-4354009.816 +el");
    add_proj4text (p, 2, "lps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Michigan Oblique Mercator\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial");
    add_srs_wkt (p, 2,
		 "_Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137,2");
    add_srs_wkt (p, 3,
		 "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwi");
    add_srs_wkt (p, 5,
		 "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0");
    add_srs_wkt (p, 6,
		 "1745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY");
    add_srs_wkt (p, 7,
		 "[\"EPSG\",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 8,
		 ",\"9001\"]],PROJECTION[\"Hotine_Oblique_Mercator\"],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"latitude_of_center\",45.30916666666666],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"longitude_of_center\",-86],PARAMETER[\"azimuth\",337");
    add_srs_wkt (p, 11,
		 ".25556],PARAMETER[\"rectified_grid_angle\",337.25556],PA");
    add_srs_wkt (p, 12,
		 "RAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_easti");
    add_srs_wkt (p, 13,
		 "ng\",2546731.496],PARAMETER[\"false_northing\",-4354009.");
    add_srs_wkt (p, 14,
		 "816],AUTHORITY[\"EPSG\",\"3591\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 15, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3592, "epsg", 3592,
		      "NAD83(NSRS2007) / Michigan South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43.66666666666666 +lat_2=42.1 +lat_0=41");
    add_proj4text (p, 1,
		   ".5 +lon_0=-84.36666666666666 +x_0=4000000 +y_0=0 +ellps=");
    add_proj4text (p, 2, "GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Michigan South\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_");
    add_srs_wkt (p, 2,
		 "System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",43.66666666666666],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",42.1],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 11,
		 ",41.5],PARAMETER[\"central_meridian\",-84.36666666666666");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_easting\",4000000],PARAMETER[\"false");
    add_srs_wkt (p, 13,
		 "_northing\",0],AUTHORITY[\"EPSG\",\"3592\"],AXIS[\"X\",E");
    add_srs_wkt (p, 14, "AST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3593, "epsg", 3593,
		      "NAD83(NSRS2007) / Michigan South (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43.66666666666666 +lat_2=42.1 +lat_0=41");
    add_proj4text (p, 1,
		   ".5 +lon_0=-84.36666666666666 +x_0=3999999.999984 +y_0=0 ");
    add_proj4text (p, 2,
		   "+ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Michigan South (ft)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 8,
		 "9002\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"standard_parallel_1\",43.66666666666666],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"standard_parallel_2\",42.1],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 11,
		 "f_origin\",41.5],PARAMETER[\"central_meridian\",-84.3666");
    add_srs_wkt (p, 12,
		 "6666666666],PARAMETER[\"false_easting\",13123359.58],PAR");
    add_srs_wkt (p, 13,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3593\"");
    add_srs_wkt (p, 14, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3594, "epsg", 3594,
		      "NAD83(NSRS2007) / Minnesota Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.05 +lat_2=45.61666666666667 +lat_0=4");
    add_proj4text (p, 1,
		   "5 +lon_0=-94.25 +x_0=800000 +y_0=100000 +ellps=GRS80 +to");
    add_proj4text (p, 2, "wgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Minnesota Central\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",47.05],PARAMETER[\"standard_par");
    add_srs_wkt (p, 10,
		 "allel_2\",45.61666666666667],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 11,
		 "gin\",45],PARAMETER[\"central_meridian\",-94.25],PARAMET");
    add_srs_wkt (p, 12,
		 "ER[\"false_easting\",800000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 13,
		 ",100000],AUTHORITY[\"EPSG\",\"3594\"],AXIS[\"X\",EAST],A");
    add_srs_wkt (p, 14, "XIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3595, "epsg", 3595,
		      "NAD83(NSRS2007) / Minnesota North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.63333333333333 +lat_2=47.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=46.5 +lon_0=-93.09999999999999 +x_0=800000 +y");
    add_proj4text (p, 2,
		   "_0=100000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +");
    add_proj4text (p, 3, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Minnesota North\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference");
    add_srs_wkt (p, 2,
		 "_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 3,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 6,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 8,
		 "],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"standard_parallel_1\",48.63333333333333],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "tandard_parallel_2\",47.03333333333333],PARAMETER[\"lati");
    add_srs_wkt (p, 11,
		 "tude_of_origin\",46.5],PARAMETER[\"central_meridian\",-9");
    add_srs_wkt (p, 12,
		 "3.1],PARAMETER[\"false_easting\",800000],PARAMETER[\"fal");
    add_srs_wkt (p, 13,
		 "se_northing\",100000],AUTHORITY[\"EPSG\",\"3595\"],AXIS[");
    add_srs_wkt (p, 14, "\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3596, "epsg", 3596,
		      "NAD83(NSRS2007) / Minnesota South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.21666666666667 +lat_2=43.78333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=43 +lon_0=-94 +x_0=800000 +y_0=100000 +ellps=");
    add_proj4text (p, 2, "GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Minnesota South\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference");
    add_srs_wkt (p, 2,
		 "_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 3,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 6,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 8,
		 "],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"standard_parallel_1\",45.21666666666667],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "tandard_parallel_2\",43.78333333333333],PARAMETER[\"lati");
    add_srs_wkt (p, 11,
		 "tude_of_origin\",43],PARAMETER[\"central_meridian\",-94]");
    add_srs_wkt (p, 12,
		 ",PARAMETER[\"false_easting\",800000],PARAMETER[\"false_n");
    add_srs_wkt (p, 13,
		 "orthing\",100000],AUTHORITY[\"EPSG\",\"3596\"],AXIS[\"X\"");
    add_srs_wkt (p, 14, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3597, "epsg", 3597,
		      "NAD83(NSRS2007) / Mississippi East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=29.5 +lon_0=-88.83333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "9995 +x_0=300000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Mississippi East\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referenc");
    add_srs_wkt (p, 2,
		 "e_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",29.5],PARAMETER[\"central_meridian\",-88");
    add_srs_wkt (p, 10,
		 ".83333333333333],PARAMETER[\"scale_factor\",0.99995],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_easting\",300000],PARAMETER[\"false_north");
    add_srs_wkt (p, 12,
		 "ing\",0],AUTHORITY[\"EPSG\",\"3597\"],AXIS[\"X\",EAST],A");
    add_srs_wkt (p, 13, "XIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3598, "epsg", 3598,
		      "NAD83(NSRS2007) / Mississippi East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=29.5 +lon_0=-88.83333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "9995 +x_0=300000.0000000001 +y_0=0 +ellps=GRS80 +towgs84");
    add_proj4text (p, 2, "=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Mississippi East (ftUS)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_R");
    add_srs_wkt (p, 2,
		 "eference_System_2007\",SPHEROID[\"GRS 1980\",6378137,298");
    add_srs_wkt (p, 3,
		 ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 5,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 6,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.304800609601");
    add_srs_wkt (p, 8,
		 "2192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transver");
    add_srs_wkt (p, 9,
		 "se_Mercator\"],PARAMETER[\"latitude_of_origin\",29.5],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"central_meridian\",-88.83333333333333],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"scale_factor\",0.99995],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",984250.0000000002],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 13,
		 "ORITY[\"EPSG\",\"3598\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 14, "TH]]");
    p = add_epsg_def (first, last, 3599, "epsg", 3599,
		      "NAD83(NSRS2007) / Mississippi West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=29.5 +lon_0=-90.33333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "9995 +x_0=700000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Mississippi West\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referenc");
    add_srs_wkt (p, 2,
		 "e_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",29.5],PARAMETER[\"central_meridian\",-90");
    add_srs_wkt (p, 10,
		 ".33333333333333],PARAMETER[\"scale_factor\",0.99995],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_easting\",700000],PARAMETER[\"false_north");
    add_srs_wkt (p, 12,
		 "ing\",0],AUTHORITY[\"EPSG\",\"3599\"],AXIS[\"X\",EAST],A");
    add_srs_wkt (p, 13, "XIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3600, "epsg", 3600,
		      "NAD83(NSRS2007) / Mississippi West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=29.5 +lon_0=-90.33333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "9995 +x_0=699999.9998983998 +y_0=0 +ellps=GRS80 +towgs84");
    add_proj4text (p, 2, "=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Mississippi West (ftUS)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_R");
    add_srs_wkt (p, 2,
		 "eference_System_2007\",SPHEROID[\"GRS 1980\",6378137,298");
    add_srs_wkt (p, 3,
		 ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 5,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 6,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.304800609601");
    add_srs_wkt (p, 8,
		 "2192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transver");
    add_srs_wkt (p, 9,
		 "se_Mercator\"],PARAMETER[\"latitude_of_origin\",29.5],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"central_meridian\",-90.33333333333333],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"scale_factor\",0.99995],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",2296583.333],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 13,
		 "\"EPSG\",\"3600\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3601, "epsg", 3601,
		      "NAD83(NSRS2007) / Missouri Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=35.83333333333334 +lon_0=-92.5 +k=0.9");
    add_proj4text (p, 1,
		   "99933333 +x_0=500000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,");
    add_proj4text (p, 2, "0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Missouri Central\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referenc");
    add_srs_wkt (p, 2,
		 "e_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",35.83333333333334],PARAMETER[\"central_m");
    add_srs_wkt (p, 10,
		 "eridian\",-92.5],PARAMETER[\"scale_factor\",0.999933333]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_easting\",500000],PARAMETER[\"false_n");
    add_srs_wkt (p, 12,
		 "orthing\",0],AUTHORITY[\"EPSG\",\"3601\"],AXIS[\"X\",EAS");
    add_srs_wkt (p, 13, "T],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3602, "epsg", 3602,
		      "NAD83(NSRS2007) / Missouri East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=35.83333333333334 +lon_0=-90.5 +k=0.9");
    add_proj4text (p, 1,
		   "99933333 +x_0=250000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,");
    add_proj4text (p, 2, "0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Missouri East\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",35.83333333333334],PARAMETER[\"central_meri");
    add_srs_wkt (p, 10,
		 "dian\",-90.5],PARAMETER[\"scale_factor\",0.999933333],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_easting\",250000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 12,
		 "hing\",0],AUTHORITY[\"EPSG\",\"3602\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 13, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3603, "epsg", 3603,
		      "NAD83(NSRS2007) / Missouri West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36.16666666666666 +lon_0=-94.5 +k=0.9");
    add_proj4text (p, 1,
		   "99941177 +x_0=850000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,");
    add_proj4text (p, 2, "0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Missouri West\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",36.16666666666666],PARAMETER[\"central_meri");
    add_srs_wkt (p, 10,
		 "dian\",-94.5],PARAMETER[\"scale_factor\",0.999941177],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_easting\",850000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 12,
		 "hing\",0],AUTHORITY[\"EPSG\",\"3603\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 13, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3604, "epsg", 3604,
		      "NAD83(NSRS2007) / Montana");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49 +lat_2=45 +lat_0=44.25 +lon_0=-109.5");
    add_proj4text (p, 1,
		   " +x_0=600000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Montana\",GEOGCS[\"NAD83(NSRS");
    add_srs_wkt (p, 1,
		 "2007)\",DATUM[\"NAD83_National_Spatial_Reference_System_");
    add_srs_wkt (p, 2,
		 "2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4759\"]");
    add_srs_wkt (p, 7,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 8,
		 "ION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standar");
    add_srs_wkt (p, 9,
		 "d_parallel_1\",49],PARAMETER[\"standard_parallel_2\",45]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"latitude_of_origin\",44.25],PARAMETER[\"cen");
    add_srs_wkt (p, 11,
		 "tral_meridian\",-109.5],PARAMETER[\"false_easting\",6000");
    add_srs_wkt (p, 12,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 13, "3604\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3605, "epsg", 3605,
		      "NAD83(NSRS2007) / Montana (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49 +lat_2=45 +lat_0=44.25 +lon_0=-109.5");
    add_proj4text (p, 1,
		   " +x_0=599999.9999976 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,");
    add_proj4text (p, 2, "0,0,0,0 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Montana (ft)\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",49],PARAMETER[\"standard_parall");
    add_srs_wkt (p, 10,
		 "el_2\",45],PARAMETER[\"latitude_of_origin\",44.25],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"central_meridian\",-109.5],PARAMETER[\"false_east");
    add_srs_wkt (p, 12,
		 "ing\",1968503.937],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 13,
		 "RITY[\"EPSG\",\"3605\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORT");
    add_srs_wkt (p, 14, "H]]");
    p = add_epsg_def (first, last, 3606, "epsg", 3606,
		      "NAD83(NSRS2007) / Nebraska");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43 +lat_2=40 +lat_0=39.83333333333334 +");
    add_proj4text (p, 1,
		   "lon_0=-100 +x_0=500000 +y_0=0 +ellps=GRS80 +towgs84=0,0,");
    add_proj4text (p, 2, "0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Nebraska\",GEOGCS[\"NAD83(NSR");
    add_srs_wkt (p, 1,
		 "S2007)\",DATUM[\"NAD83_National_Spatial_Reference_System");
    add_srs_wkt (p, 2,
		 "_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 5,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4759\"");
    add_srs_wkt (p, 7,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",43],PARAMETER[\"standard_parallel_2\",40");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"latitude_of_origin\",39.83333333333334],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"central_meridian\",-100],PARAMETER[\"false_eas");
    add_srs_wkt (p, 12,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 13,
		 "[\"EPSG\",\"3606\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3607, "epsg", 3607,
		      "NAD83(NSRS2007) / Nevada Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-116.6666666666667 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=500000 +y_0=6000000 +ellps=GRS80 +towgs84=0,0,");
    add_proj4text (p, 2, "0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Nevada Central\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_");
    add_srs_wkt (p, 2,
		 "System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",34.75],PARAMETER[\"central_meridian\",-116.");
    add_srs_wkt (p, 10,
		 "6666666666667],PARAMETER[\"scale_factor\",0.9999],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_easting\",500000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 12,
		 "\",6000000],AUTHORITY[\"EPSG\",\"3607\"],AXIS[\"X\",EAST");
    add_srs_wkt (p, 13, "],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3608, "epsg", 3608,
		      "NAD83(NSRS2007) / Nevada Central (ft US)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-116.6666666666667 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=500000.00001016 +y_0=6000000 +ellps=GRS80 +tow");
    add_proj4text (p, 2, "gs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Nevada Central (ft US)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Re");
    add_srs_wkt (p, 2,
		 "ference_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.");
    add_srs_wkt (p, 3,
		 "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transvers");
    add_srs_wkt (p, 9,
		 "e_Mercator\"],PARAMETER[\"latitude_of_origin\",34.75],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"central_meridian\",-116.6666666666667],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 12,
		 "1640416.6667],PARAMETER[\"false_northing\",19685000],AUT");
    add_srs_wkt (p, 13,
		 "HORITY[\"EPSG\",\"3608\"],AXIS[\"X\",EAST],AXIS[\"Y\",NO");
    add_srs_wkt (p, 14, "RTH]]");
    p = add_epsg_def (first, last, 3609, "epsg", 3609,
		      "NAD83(NSRS2007) / Nevada East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-115.5833333333333 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=200000 +y_0=8000000 +ellps=GRS80 +towgs84=0,0,");
    add_proj4text (p, 2, "0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Nevada East\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sys");
    add_srs_wkt (p, 2,
		 "tem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"47");
    add_srs_wkt (p, 7,
		 "59\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",34.75],PARAMETER[\"central_meridian\",-115.58");
    add_srs_wkt (p, 10,
		 "33333333333],PARAMETER[\"scale_factor\",0.9999],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_easting\",200000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",8000000],AUTHORITY[\"EPSG\",\"3609\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 13, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3610, "epsg", 3610,
		      "NAD83(NSRS2007) / Nevada East (ft US)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-115.5833333333333 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=200000.00001016 +y_0=8000000.000010163 +ellps=");
    add_proj4text (p, 2, "GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Nevada East (ft US)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"US survey foot\",0.304800609601219");
    add_srs_wkt (p, 8,
		 "2,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_");
    add_srs_wkt (p, 9,
		 "Mercator\"],PARAMETER[\"latitude_of_origin\",34.75],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"central_meridian\",-115.5833333333333],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",65");
    add_srs_wkt (p, 12,
		 "6166.6667],PARAMETER[\"false_northing\",26246666.6667000");
    add_srs_wkt (p, 13,
		 "1],AUTHORITY[\"EPSG\",\"3610\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 14, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3611, "epsg", 3611,
		      "NAD83(NSRS2007) / Nevada West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-118.5833333333333 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=800000 +y_0=4000000 +ellps=GRS80 +towgs84=0,0,");
    add_proj4text (p, 2, "0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Nevada West\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sys");
    add_srs_wkt (p, 2,
		 "tem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"47");
    add_srs_wkt (p, 7,
		 "59\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",34.75],PARAMETER[\"central_meridian\",-118.58");
    add_srs_wkt (p, 10,
		 "33333333333],PARAMETER[\"scale_factor\",0.9999],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_easting\",800000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",4000000],AUTHORITY[\"EPSG\",\"3611\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 13, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3612, "epsg", 3612,
		      "NAD83(NSRS2007) / Nevada West (ft US)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-118.5833333333333 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=800000.0000101599 +y_0=3999999.99998984 +ellps");
    add_proj4text (p, 2, "=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Nevada West (ft US)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"US survey foot\",0.304800609601219");
    add_srs_wkt (p, 8,
		 "2,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_");
    add_srs_wkt (p, 9,
		 "Mercator\"],PARAMETER[\"latitude_of_origin\",34.75],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"central_meridian\",-118.5833333333333],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",26");
    add_srs_wkt (p, 12,
		 "24666.6667],PARAMETER[\"false_northing\",13123333.3333],");
    add_srs_wkt (p, 13,
		 "AUTHORITY[\"EPSG\",\"3612\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 14, ",NORTH]]");
    p = add_epsg_def (first, last, 3613, "epsg", 3613,
		      "NAD83(NSRS2007) / New Hampshire");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.5 +lon_0=-71.66666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=300000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,");
    add_proj4text (p, 2, "0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / New Hampshire\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",42.5],PARAMETER[\"central_meridian\",-71.66");
    add_srs_wkt (p, 10,
		 "666666666667],PARAMETER[\"scale_factor\",0.999966667],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_easting\",300000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 12,
		 "hing\",0],AUTHORITY[\"EPSG\",\"3613\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 13, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3614, "epsg", 3614,
		      "NAD83(NSRS2007) / New Hampshire (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.5 +lon_0=-71.66666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=300000.0000000001 +y_0=0 +ellps=GRS80 +tow");
    add_proj4text (p, 2, "gs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / New Hampshire (ftUS)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refe");
    add_srs_wkt (p, 2,
		 "rence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25");
    add_srs_wkt (p, 3,
		 "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4759\"]],UNIT[\"US survey foot\",0.30480060960121");
    add_srs_wkt (p, 8,
		 "92,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 9,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",42.5],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"central_meridian\",-71.66666666666667],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"scale_factor\",0.999966667],PARAMETER[\"false_easting");
    add_srs_wkt (p, 12,
		 "\",984250.0000000002],PARAMETER[\"false_northing\",0],AU");
    add_srs_wkt (p, 13,
		 "THORITY[\"EPSG\",\"3614\"],AXIS[\"X\",EAST],AXIS[\"Y\",N");
    add_srs_wkt (p, 14, "ORTH]]");
    p = add_epsg_def (first, last, 3615, "epsg", 3615,
		      "NAD83(NSRS2007) / New Jersey");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38.83333333333334 +lon_0=-74.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=150000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0");
    add_proj4text (p, 2, ",0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / New Jersey\",GEOGCS[\"NAD83(N");
    add_srs_wkt (p, 1,
		 "SRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Syst");
    add_srs_wkt (p, 2,
		 "em_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"475");
    add_srs_wkt (p, 7,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",38.83333333333334],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 10,
		 "n\",-74.5],PARAMETER[\"scale_factor\",0.9999],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_easting\",150000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"3615\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 13, "\",NORTH]]");
    p = add_epsg_def (first, last, 3616, "epsg", 3616,
		      "NAD83(NSRS2007) / New Jersey (ft US)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38.83333333333334 +lon_0=-74.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=150000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0");
    add_proj4text (p, 2, ",0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / New Jersey (ft US)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 8,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 9,
		 "cator\"],PARAMETER[\"latitude_of_origin\",38.83333333333");
    add_srs_wkt (p, 10,
		 "334],PARAMETER[\"central_meridian\",-74.5],PARAMETER[\"s");
    add_srs_wkt (p, 11,
		 "cale_factor\",0.9999],PARAMETER[\"false_easting\",492125");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3");
    add_srs_wkt (p, 13, "616\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3617, "epsg", 3617,
		      "NAD83(NSRS2007) / New Mexico Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-106.25 +k=0.9999 +x_0=5000");
    add_proj4text (p, 1,
		   "00 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / New Mexico Central\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",31],PARAMETER[\"central_meridian\",-106.");
    add_srs_wkt (p, 10,
		 "25],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_easting\",500000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 12,
		 "RITY[\"EPSG\",\"3617\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORT");
    add_srs_wkt (p, 13, "H]]");
    p = add_epsg_def (first, last, 3618, "epsg", 3618,
		      "NAD83(NSRS2007) / New Mexico Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-106.25 +k=0.9999 +x_0=5000");
    add_proj4text (p, 1,
		   "00.0001016001 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0");
    add_proj4text (p, 2, " +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / New Mexico Central (ftUS)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial");
    add_srs_wkt (p, 2,
		 "_Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137,2");
    add_srs_wkt (p, 3,
		 "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwi");
    add_srs_wkt (p, 5,
		 "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0");
    add_srs_wkt (p, 6,
		 "1745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY");
    add_srs_wkt (p, 7,
		 "[\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.304800609");
    add_srs_wkt (p, 8,
		 "6012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Trans");
    add_srs_wkt (p, 9,
		 "verse_Mercator\"],PARAMETER[\"latitude_of_origin\",31],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"central_meridian\",-106.25],PARAMETER[\"scale");
    add_srs_wkt (p, 11,
		 "_factor\",0.9999],PARAMETER[\"false_easting\",1640416.66");
    add_srs_wkt (p, 12,
		 "7],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 13, "3618\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3619, "epsg", 3619,
		      "NAD83(NSRS2007) / New Mexico East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-104.3333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "909091 +x_0=165000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 2, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / New Mexico East\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference");
    add_srs_wkt (p, 2,
		 "_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 3,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 6,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 8,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 9,
		 "de_of_origin\",31],PARAMETER[\"central_meridian\",-104.3");
    add_srs_wkt (p, 10,
		 "333333333333],PARAMETER[\"scale_factor\",0.999909091],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_easting\",165000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 12,
		 "hing\",0],AUTHORITY[\"EPSG\",\"3619\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 13, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3620, "epsg", 3620,
		      "NAD83(NSRS2007) / New Mexico East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-104.3333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "909091 +x_0=165000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 2, "0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / New Mexico East (ftUS)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Re");
    add_srs_wkt (p, 2,
		 "ference_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.");
    add_srs_wkt (p, 3,
		 "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transvers");
    add_srs_wkt (p, 9,
		 "e_Mercator\"],PARAMETER[\"latitude_of_origin\",31],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"central_meridian\",-104.3333333333333],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"scale_factor\",0.999909091],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",541337.5],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 13, "PSG\",\"3620\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3621, "epsg", 3621,
		      "NAD83(NSRS2007) / New Mexico West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-107.8333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "916667 +x_0=830000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 2, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / New Mexico West\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference");
    add_srs_wkt (p, 2,
		 "_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 3,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 6,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 8,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 9,
		 "de_of_origin\",31],PARAMETER[\"central_meridian\",-107.8");
    add_srs_wkt (p, 10,
		 "333333333333],PARAMETER[\"scale_factor\",0.999916667],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_easting\",830000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 12,
		 "hing\",0],AUTHORITY[\"EPSG\",\"3621\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 13, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3622, "epsg", 3622,
		      "NAD83(NSRS2007) / New Mexico West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-107.8333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "916667 +x_0=830000.0001016001 +y_0=0 +ellps=GRS80 +towgs");
    add_proj4text (p, 2, "84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / New Mexico West (ftUS)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Re");
    add_srs_wkt (p, 2,
		 "ference_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.");
    add_srs_wkt (p, 3,
		 "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transvers");
    add_srs_wkt (p, 9,
		 "e_Mercator\"],PARAMETER[\"latitude_of_origin\",31],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"central_meridian\",-107.8333333333333],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"scale_factor\",0.999916667],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",2723091.667],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 13,
		 "\"EPSG\",\"3622\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3623, "epsg", 3623,
		      "NAD83(NSRS2007) / New York Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40 +lon_0=-76.58333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "9375 +x_0=250000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / New York Central\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referenc");
    add_srs_wkt (p, 2,
		 "e_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",40],PARAMETER[\"central_meridian\",-76.5");
    add_srs_wkt (p, 10,
		 "8333333333333],PARAMETER[\"scale_factor\",0.9999375],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_easting\",250000],PARAMETER[\"false_north");
    add_srs_wkt (p, 12,
		 "ing\",0],AUTHORITY[\"EPSG\",\"3623\"],AXIS[\"X\",EAST],A");
    add_srs_wkt (p, 13, "XIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3624, "epsg", 3624,
		      "NAD83(NSRS2007) / New York Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40 +lon_0=-76.58333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "9375 +x_0=249999.9998983998 +y_0=0 +ellps=GRS80 +towgs84");
    add_proj4text (p, 2, "=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / New York Central (ftUS)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_R");
    add_srs_wkt (p, 2,
		 "eference_System_2007\",SPHEROID[\"GRS 1980\",6378137,298");
    add_srs_wkt (p, 3,
		 ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 5,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 6,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.304800609601");
    add_srs_wkt (p, 8,
		 "2192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transver");
    add_srs_wkt (p, 9,
		 "se_Mercator\"],PARAMETER[\"latitude_of_origin\",40],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"central_meridian\",-76.58333333333333],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"scale_factor\",0.9999375],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",820208.3330000002],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 13,
		 "ORITY[\"EPSG\",\"3624\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 14, "TH]]");
    p = add_epsg_def (first, last, 3625, "epsg", 3625,
		      "NAD83(NSRS2007) / New York East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38.83333333333334 +lon_0=-74.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=150000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0");
    add_proj4text (p, 2, ",0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / New York East\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",38.83333333333334],PARAMETER[\"central_meri");
    add_srs_wkt (p, 10,
		 "dian\",-74.5],PARAMETER[\"scale_factor\",0.9999],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_easting\",150000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"3625\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3626, "epsg", 3626,
		      "NAD83(NSRS2007) / New York East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38.83333333333334 +lon_0=-74.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=150000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0");
    add_proj4text (p, 2, ",0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / New York East (ftUS)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refe");
    add_srs_wkt (p, 2,
		 "rence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25");
    add_srs_wkt (p, 3,
		 "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4759\"]],UNIT[\"US survey foot\",0.30480060960121");
    add_srs_wkt (p, 8,
		 "92,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 9,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",38.8333333");
    add_srs_wkt (p, 10,
		 "3333334],PARAMETER[\"central_meridian\",-74.5],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",49");
    add_srs_wkt (p, 12,
		 "2125],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 13, ",\"3626\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3627, "epsg", 3627,
		      "NAD83(NSRS2007) / New York Long Island");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.03333333333333 +lat_2=40.66666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=40.16666666666666 +lon_0=-74 +x_0=300000 +y_0");
    add_proj4text (p, 2,
		   "=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_proj4text (p, 3, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / New York Long Island\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refe");
    add_srs_wkt (p, 2,
		 "rence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25");
    add_srs_wkt (p, 3,
		 "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 8,
		 "01\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"standard_parallel_1\",41.03333333333333],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"standard_parallel_2\",40.66666666666666],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "latitude_of_origin\",40.16666666666666],PARAMETER[\"cent");
    add_srs_wkt (p, 12,
		 "ral_meridian\",-74],PARAMETER[\"false_easting\",300000],");
    add_srs_wkt (p, 13,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"362");
    add_srs_wkt (p, 14, "7\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3628, "epsg", 3628,
		      "NAD83(NSRS2007) / New York Long Island (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.03333333333333 +lat_2=40.66666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=40.16666666666666 +lon_0=-74 +x_0=300000.0000");
    add_proj4text (p, 2,
		   "000001 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units");
    add_proj4text (p, 3, "=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / New York Long Island (ftUS)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spati");
    add_srs_wkt (p, 2,
		 "al_Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 3,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 5,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 6,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006");
    add_srs_wkt (p, 8,
		 "096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lam");
    add_srs_wkt (p, 9,
		 "bert_Conformal_Conic_2SP\"],PARAMETER[\"standard_paralle");
    add_srs_wkt (p, 10,
		 "l_1\",41.03333333333333],PARAMETER[\"standard_parallel_2");
    add_srs_wkt (p, 11,
		 "\",40.66666666666666],PARAMETER[\"latitude_of_origin\",4");
    add_srs_wkt (p, 12,
		 "0.16666666666666],PARAMETER[\"central_meridian\",-74],PA");
    add_srs_wkt (p, 13,
		 "RAMETER[\"false_easting\",984250.0000000002],PARAMETER[\"");
    add_srs_wkt (p, 14,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3628\"],AXIS[\"");
    add_srs_wkt (p, 15, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3629, "epsg", 3629,
		      "NAD83(NSRS2007) / New York West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40 +lon_0=-78.58333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "9375 +x_0=350000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / New York West\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",40],PARAMETER[\"central_meridian\",-78.5833");
    add_srs_wkt (p, 10,
		 "3333333333],PARAMETER[\"scale_factor\",0.9999375],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_easting\",350000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 12,
		 "\",0],AUTHORITY[\"EPSG\",\"3629\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 13, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3630, "epsg", 3630,
		      "NAD83(NSRS2007) / New York West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40 +lon_0=-78.58333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "9375 +x_0=350000.0001016001 +y_0=0 +ellps=GRS80 +towgs84");
    add_proj4text (p, 2, "=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / New York West (ftUS)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refe");
    add_srs_wkt (p, 2,
		 "rence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25");
    add_srs_wkt (p, 3,
		 "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4759\"]],UNIT[\"US survey foot\",0.30480060960121");
    add_srs_wkt (p, 8,
		 "92,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 9,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",40],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"central_meridian\",-78.58333333333333],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "scale_factor\",0.9999375],PARAMETER[\"false_easting\",11");
    add_srs_wkt (p, 12,
		 "48291.667],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 13, "PSG\",\"3630\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3631, "epsg", 3631,
		      "NAD83(NSRS2007) / North Carolina");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.16666666666666 +lat_2=34.33333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=33.75 +lon_0=-79 +x_0=609601.22 +y_0=0 +ellps");
    add_proj4text (p, 2, "=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / North Carolina\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_");
    add_srs_wkt (p, 2,
		 "System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",36.16666666666666],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",34.33333333333334],PARAMETER[\"latitu");
    add_srs_wkt (p, 11,
		 "de_of_origin\",33.75],PARAMETER[\"central_meridian\",-79");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_easting\",609601.22],PARAMETER[\"fal");
    add_srs_wkt (p, 13,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"3631\"],AXIS[\"X\"");
    add_srs_wkt (p, 14, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3632, "epsg", 3632,
		      "NAD83(NSRS2007) / North Carolina (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.16666666666666 +lat_2=34.33333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=33.75 +lon_0=-79 +x_0=609601.2192024384 +y_0=");
    add_proj4text (p, 2,
		   "0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_d");
    add_proj4text (p, 3, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / North Carolina (ftUS)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2");
    add_srs_wkt (p, 3,
		 "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 9,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 10,
		 "36.16666666666666],PARAMETER[\"standard_parallel_2\",34.");
    add_srs_wkt (p, 11,
		 "33333333333334],PARAMETER[\"latitude_of_origin\",33.75],");
    add_srs_wkt (p, 12,
		 "PARAMETER[\"central_meridian\",-79],PARAMETER[\"false_ea");
    add_srs_wkt (p, 13,
		 "sting\",2000000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 14,
		 "TY[\"EPSG\",\"3632\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]");
    add_srs_wkt (p, 15, "]");
    p = add_epsg_def (first, last, 3633, "epsg", 3633,
		      "NAD83(NSRS2007) / North Dakota North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.73333333333333 +lat_2=47.43333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=47 +lon_0=-100.5 +x_0=600000 +y_0=0 +ellps=GR");
    add_proj4text (p, 2, "S80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / North Dakota North\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",48.73333333333333],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "standard_parallel_2\",47.43333333333333],PARAMETER[\"lat");
    add_srs_wkt (p, 11,
		 "itude_of_origin\",47],PARAMETER[\"central_meridian\",-10");
    add_srs_wkt (p, 12,
		 "0.5],PARAMETER[\"false_easting\",600000],PARAMETER[\"fal");
    add_srs_wkt (p, 13,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"3633\"],AXIS[\"X\"");
    add_srs_wkt (p, 14, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3634, "epsg", 3634,
		      "NAD83(NSRS2007) / North Dakota North (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.73333333333333 +lat_2=47.43333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=47 +lon_0=-100.5 +x_0=599999.9999976 +y_0=0 +");
    add_proj4text (p, 2,
		   "ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / North Dakota North (ft)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_R");
    add_srs_wkt (p, 2,
		 "eference_System_2007\",SPHEROID[\"GRS 1980\",6378137,298");
    add_srs_wkt (p, 3,
		 ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 5,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 6,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"4759\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 8,
		 ",\"9002\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"standard_parallel_1\",48.73333333333333],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"standard_parallel_2\",47.43333333333333],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"latitude_of_origin\",47],PARAMETER[\"central_merid");
    add_srs_wkt (p, 12,
		 "ian\",-100.5],PARAMETER[\"false_easting\",1968503.937],P");
    add_srs_wkt (p, 13,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3634");
    add_srs_wkt (p, 14, "\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3635, "epsg", 3635,
		      "NAD83(NSRS2007) / North Dakota South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.48333333333333 +lat_2=46.18333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=45.66666666666666 +lon_0=-100.5 +x_0=600000 +");
    add_proj4text (p, 2,
		   "y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_d");
    add_proj4text (p, 3, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / North Dakota South\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",47.48333333333333],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "standard_parallel_2\",46.18333333333333],PARAMETER[\"lat");
    add_srs_wkt (p, 11,
		 "itude_of_origin\",45.66666666666666],PARAMETER[\"central");
    add_srs_wkt (p, 12,
		 "_meridian\",-100.5],PARAMETER[\"false_easting\",600000],");
    add_srs_wkt (p, 13,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"363");
    add_srs_wkt (p, 14, "5\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3636, "epsg", 3636,
		      "NAD83(NSRS2007) / North Dakota South (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.48333333333333 +lat_2=46.18333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=45.66666666666666 +lon_0=-100.5 +x_0=599999.9");
    add_proj4text (p, 2,
		   "999976 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units");
    add_proj4text (p, 3, "=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / North Dakota South (ft)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_R");
    add_srs_wkt (p, 2,
		 "eference_System_2007\",SPHEROID[\"GRS 1980\",6378137,298");
    add_srs_wkt (p, 3,
		 ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 5,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 6,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"4759\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 8,
		 ",\"9002\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"standard_parallel_1\",47.48333333333333],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"standard_parallel_2\",46.18333333333333],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"latitude_of_origin\",45.66666666666666],PARAMETER[");
    add_srs_wkt (p, 12,
		 "\"central_meridian\",-100.5],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 13,
		 ",1968503.937],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 14,
		 "\"EPSG\",\"3636\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3637, "epsg", 3637,
		      "NAD83(NSRS2007) / Ohio North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.7 +lat_2=40.43333333333333 +lat_0=39");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-82.5 +x_0=600000 +y_0=0 +ellps=G");
    add_proj4text (p, 2, "RS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Ohio North\",GEOGCS[\"NAD83(N");
    add_srs_wkt (p, 1,
		 "SRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Syst");
    add_srs_wkt (p, 2,
		 "em_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"475");
    add_srs_wkt (p, 7,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"sta");
    add_srs_wkt (p, 9,
		 "ndard_parallel_1\",41.7],PARAMETER[\"standard_parallel_2");
    add_srs_wkt (p, 10,
		 "\",40.43333333333333],PARAMETER[\"latitude_of_origin\",3");
    add_srs_wkt (p, 11,
		 "9.66666666666666],PARAMETER[\"central_meridian\",-82.5],");
    add_srs_wkt (p, 12,
		 "PARAMETER[\"false_easting\",600000],PARAMETER[\"false_no");
    add_srs_wkt (p, 13,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"3637\"],AXIS[\"X\",EAST");
    add_srs_wkt (p, 14, "],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3638, "epsg", 3638,
		      "NAD83(NSRS2007) / Ohio South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.03333333333333 +lat_2=38.73333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=38 +lon_0=-82.5 +x_0=600000 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Ohio South\",GEOGCS[\"NAD83(N");
    add_srs_wkt (p, 1,
		 "SRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Syst");
    add_srs_wkt (p, 2,
		 "em_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"475");
    add_srs_wkt (p, 7,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"sta");
    add_srs_wkt (p, 9,
		 "ndard_parallel_1\",40.03333333333333],PARAMETER[\"standa");
    add_srs_wkt (p, 10,
		 "rd_parallel_2\",38.73333333333333],PARAMETER[\"latitude_");
    add_srs_wkt (p, 11,
		 "of_origin\",38],PARAMETER[\"central_meridian\",-82.5],PA");
    add_srs_wkt (p, 12,
		 "RAMETER[\"false_easting\",600000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 13,
		 "hing\",0],AUTHORITY[\"EPSG\",\"3638\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 14, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3639, "epsg", 3639,
		      "NAD83(NSRS2007) / Oklahoma North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.76666666666667 +lat_2=35.56666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=35 +lon_0=-98 +x_0=600000 +y_0=0 +ellps=GRS80");
    add_proj4text (p, 2, " +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Oklahoma North\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_");
    add_srs_wkt (p, 2,
		 "System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",36.76666666666667],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",35.56666666666667],PARAMETER[\"latitu");
    add_srs_wkt (p, 11,
		 "de_of_origin\",35],PARAMETER[\"central_meridian\",-98],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_easting\",600000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 13,
		 "thing\",0],AUTHORITY[\"EPSG\",\"3639\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 14, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3640, "epsg", 3640,
		      "NAD83(NSRS2007) / Oklahoma North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.76666666666667 +lat_2=35.56666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=35 +lon_0=-98 +x_0=600000 +y_0=0 +ellps=GRS80");
    add_proj4text (p, 2, " +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Oklahoma North (ftUS)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2");
    add_srs_wkt (p, 3,
		 "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 9,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 10,
		 "36.76666666666667],PARAMETER[\"standard_parallel_2\",35.");
    add_srs_wkt (p, 11,
		 "56666666666667],PARAMETER[\"latitude_of_origin\",35],PAR");
    add_srs_wkt (p, 12,
		 "AMETER[\"central_meridian\",-98],PARAMETER[\"false_easti");
    add_srs_wkt (p, 13,
		 "ng\",1968500],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 14,
		 "\"EPSG\",\"3640\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3641, "epsg", 3641,
		      "NAD83(NSRS2007) / Oklahoma South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=35.23333333333333 +lat_2=33.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=33.33333333333334 +lon_0=-98 +x_0=600000 +y_0");
    add_proj4text (p, 2,
		   "=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_proj4text (p, 3, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Oklahoma South\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_");
    add_srs_wkt (p, 2,
		 "System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",35.23333333333333],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",33.93333333333333],PARAMETER[\"latitu");
    add_srs_wkt (p, 11,
		 "de_of_origin\",33.33333333333334],PARAMETER[\"central_me");
    add_srs_wkt (p, 12,
		 "ridian\",-98],PARAMETER[\"false_easting\",600000],PARAME");
    add_srs_wkt (p, 13,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3641\"],A");
    add_srs_wkt (p, 14, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3642, "epsg", 3642,
		      "NAD83(NSRS2007) / Oklahoma South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=35.23333333333333 +lat_2=33.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=33.33333333333334 +lon_0=-98 +x_0=600000 +y_0");
    add_proj4text (p, 2,
		   "=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_");
    add_proj4text (p, 3, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Oklahoma South (ftUS)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2");
    add_srs_wkt (p, 3,
		 "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 9,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 10,
		 "35.23333333333333],PARAMETER[\"standard_parallel_2\",33.");
    add_srs_wkt (p, 11,
		 "93333333333333],PARAMETER[\"latitude_of_origin\",33.3333");
    add_srs_wkt (p, 12,
		 "3333333334],PARAMETER[\"central_meridian\",-98],PARAMETE");
    add_srs_wkt (p, 13,
		 "R[\"false_easting\",1968500],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 14,
		 ",0],AUTHORITY[\"EPSG\",\"3642\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 15, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3643, "epsg", 3643,
		      "NAD83(NSRS2007) / Oregon Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43 +lat_2=45.5 +lat_0=41.75 +lon_0=-120");
    add_proj4text (p, 1,
		   ".5 +x_0=400000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,");
    add_proj4text (p, 2, "0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Oregon Lambert\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_");
    add_srs_wkt (p, 2,
		 "System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",43],PARAMETER[\"standard_parallel_");
    add_srs_wkt (p, 10,
		 "2\",45.5],PARAMETER[\"latitude_of_origin\",41.75],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"central_meridian\",-120.5],PARAMETER[\"false_easti");
    add_srs_wkt (p, 12,
		 "ng\",400000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13, "EPSG\",\"3643\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3644, "epsg", 3644,
		      "NAD83(NSRS2007) / Oregon Lambert (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43 +lat_2=45.5 +lat_0=41.75 +lon_0=-120");
    add_proj4text (p, 1,
		   ".5 +x_0=399999.9999984 +y_0=0 +ellps=GRS80 +towgs84=0,0,");
    add_proj4text (p, 2, "0,0,0,0,0 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Oregon Lambert (ft)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 8,
		 "9002\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"standard_parallel_1\",43],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",45.5],PARAMETER[\"latitude_of_origin\",41.7");
    add_srs_wkt (p, 11,
		 "5],PARAMETER[\"central_meridian\",-120.5],PARAMETER[\"fa");
    add_srs_wkt (p, 12,
		 "lse_easting\",1312335.958],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 13,
		 "0],AUTHORITY[\"EPSG\",\"3644\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 14, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3645, "epsg", 3645,
		      "NAD83(NSRS2007) / Oregon North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46 +lat_2=44.33333333333334 +lat_0=43.6");
    add_proj4text (p, 1,
		   "6666666666666 +lon_0=-120.5 +x_0=2500000 +y_0=0 +ellps=G");
    add_proj4text (p, 2, "RS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Oregon North\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_1\",46],PARAMETER[\"standard_parallel_2");
    add_srs_wkt (p, 10,
		 "\",44.33333333333334],PARAMETER[\"latitude_of_origin\",4");
    add_srs_wkt (p, 11,
		 "3.66666666666666],PARAMETER[\"central_meridian\",-120.5]");
    add_srs_wkt (p, 12,
		 ",PARAMETER[\"false_easting\",2500000],PARAMETER[\"false_");
    add_srs_wkt (p, 13,
		 "northing\",0],AUTHORITY[\"EPSG\",\"3645\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 14, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3646, "epsg", 3646,
		      "NAD83(NSRS2007) / Oregon North (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46 +lat_2=44.33333333333334 +lat_0=43.6");
    add_proj4text (p, 1,
		   "6666666666666 +lon_0=-120.5 +x_0=2500000.0001424 +y_0=0 ");
    add_proj4text (p, 2,
		   "+ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Oregon North (ft)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 8,
		 "02\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"standard_parallel_1\",46],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 10,
		 "rallel_2\",44.33333333333334],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 11,
		 "igin\",43.66666666666666],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 12,
		 ",-120.5],PARAMETER[\"false_easting\",8202099.738],PARAME");
    add_srs_wkt (p, 13,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3646\"],A");
    add_srs_wkt (p, 14, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3647, "epsg", 3647,
		      "NAD83(NSRS2007) / Oregon South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44 +lat_2=42.33333333333334 +lat_0=41.6");
    add_proj4text (p, 1,
		   "6666666666666 +lon_0=-120.5 +x_0=1500000 +y_0=0 +ellps=G");
    add_proj4text (p, 2, "RS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Oregon South\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_1\",44],PARAMETER[\"standard_parallel_2");
    add_srs_wkt (p, 10,
		 "\",42.33333333333334],PARAMETER[\"latitude_of_origin\",4");
    add_srs_wkt (p, 11,
		 "1.66666666666666],PARAMETER[\"central_meridian\",-120.5]");
    add_srs_wkt (p, 12,
		 ",PARAMETER[\"false_easting\",1500000],PARAMETER[\"false_");
    add_srs_wkt (p, 13,
		 "northing\",0],AUTHORITY[\"EPSG\",\"3647\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 14, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3648, "epsg", 3648,
		      "NAD83(NSRS2007) / Oregon South (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44 +lat_2=42.33333333333334 +lat_0=41.6");
    add_proj4text (p, 1,
		   "6666666666666 +lon_0=-120.5 +x_0=1500000.0001464 +y_0=0 ");
    add_proj4text (p, 2,
		   "+ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Oregon South (ft)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 8,
		 "02\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"standard_parallel_1\",44],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 10,
		 "rallel_2\",42.33333333333334],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 11,
		 "igin\",41.66666666666666],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 12,
		 ",-120.5],PARAMETER[\"false_easting\",4921259.843],PARAME");
    add_srs_wkt (p, 13,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3648\"],A");
    add_srs_wkt (p, 14, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3649, "epsg", 3649,
		      "NAD83(NSRS2007) / Pennsylvania North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.95 +lat_2=40.88333333333333 +lat_0=4");
    add_proj4text (p, 1,
		   "0.16666666666666 +lon_0=-77.75 +x_0=600000 +y_0=0 +ellps");
    add_proj4text (p, 2, "=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Pennsylvania North\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",41.95],PARAMETER[\"standard_par");
    add_srs_wkt (p, 10,
		 "allel_2\",40.88333333333333],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 11,
		 "gin\",40.16666666666666],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 12,
		 "-77.75],PARAMETER[\"false_easting\",600000],PARAMETER[\"");
    add_srs_wkt (p, 13,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3649\"],AXIS[\"");
    add_srs_wkt (p, 14, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3650, "epsg", 3650,
		      "NAD83(NSRS2007) / Pennsylvania North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.95 +lat_2=40.88333333333333 +lat_0=4");
    add_proj4text (p, 1,
		   "0.16666666666666 +lon_0=-77.75 +x_0=600000 +y_0=0 +ellps");
    add_proj4text (p, 2, "=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Pennsylvania North (ftUS)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial");
    add_srs_wkt (p, 2,
		 "_Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137,2");
    add_srs_wkt (p, 3,
		 "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwi");
    add_srs_wkt (p, 5,
		 "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0");
    add_srs_wkt (p, 6,
		 "1745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY");
    add_srs_wkt (p, 7,
		 "[\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.304800609");
    add_srs_wkt (p, 8,
		 "6012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambe");
    add_srs_wkt (p, 9,
		 "rt_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_");
    add_srs_wkt (p, 10,
		 "1\",41.95],PARAMETER[\"standard_parallel_2\",40.88333333");
    add_srs_wkt (p, 11,
		 "333333],PARAMETER[\"latitude_of_origin\",40.166666666666");
    add_srs_wkt (p, 12,
		 "66],PARAMETER[\"central_meridian\",-77.75],PARAMETER[\"f");
    add_srs_wkt (p, 13,
		 "alse_easting\",1968500],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 14,
		 "AUTHORITY[\"EPSG\",\"3650\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 15, ",NORTH]]");
    p = add_epsg_def (first, last, 3651, "epsg", 3651,
		      "NAD83(NSRS2007) / Pennsylvania South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.96666666666667 +lat_2=39.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=39.33333333333334 +lon_0=-77.75 +x_0=600000 +");
    add_proj4text (p, 2,
		   "y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_d");
    add_proj4text (p, 3, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Pennsylvania South\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",40.96666666666667],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "standard_parallel_2\",39.93333333333333],PARAMETER[\"lat");
    add_srs_wkt (p, 11,
		 "itude_of_origin\",39.33333333333334],PARAMETER[\"central");
    add_srs_wkt (p, 12,
		 "_meridian\",-77.75],PARAMETER[\"false_easting\",600000],");
    add_srs_wkt (p, 13,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"365");
    add_srs_wkt (p, 14, "1\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3652, "epsg", 3652,
		      "NAD83(NSRS2007) / Pennsylvania South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.96666666666667 +lat_2=39.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=39.33333333333334 +lon_0=-77.75 +x_0=600000 +");
    add_proj4text (p, 2,
		   "y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +");
    add_proj4text (p, 3, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Pennsylvania South (ftUS)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial");
    add_srs_wkt (p, 2,
		 "_Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137,2");
    add_srs_wkt (p, 3,
		 "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwi");
    add_srs_wkt (p, 5,
		 "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0");
    add_srs_wkt (p, 6,
		 "1745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY");
    add_srs_wkt (p, 7,
		 "[\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.304800609");
    add_srs_wkt (p, 8,
		 "6012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambe");
    add_srs_wkt (p, 9,
		 "rt_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_");
    add_srs_wkt (p, 10,
		 "1\",40.96666666666667],PARAMETER[\"standard_parallel_2\"");
    add_srs_wkt (p, 11,
		 ",39.93333333333333],PARAMETER[\"latitude_of_origin\",39.");
    add_srs_wkt (p, 12,
		 "33333333333334],PARAMETER[\"central_meridian\",-77.75],P");
    add_srs_wkt (p, 13,
		 "ARAMETER[\"false_easting\",1968500],PARAMETER[\"false_no");
    add_srs_wkt (p, 14,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"3652\"],AXIS[\"X\",EAST");
    add_srs_wkt (p, 15, "],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3653, "epsg", 3653,
		      "NAD83(NSRS2007) / Rhode Island");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.08333333333334 +lon_0=-71.5 +k=0.9");
    add_proj4text (p, 1,
		   "9999375 +x_0=100000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0");
    add_proj4text (p, 2, ",0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Rhode Island\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",41.08333333333334],PARAMETER[\"central_merid");
    add_srs_wkt (p, 10,
		 "ian\",-71.5],PARAMETER[\"scale_factor\",0.99999375],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_easting\",100000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 12,
		 "ng\",0],AUTHORITY[\"EPSG\",\"3653\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 13, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3654, "epsg", 3654,
		      "NAD83(NSRS2007) / Rhode Island (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.08333333333334 +lon_0=-71.5 +k=0.9");
    add_proj4text (p, 1,
		   "9999375 +x_0=99999.99998983997 +y_0=0 +ellps=GRS80 +towg");
    add_proj4text (p, 2, "s84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Rhode Island (ftUS)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"US survey foot\",0.304800609601219");
    add_srs_wkt (p, 8,
		 "2,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_");
    add_srs_wkt (p, 9,
		 "Mercator\"],PARAMETER[\"latitude_of_origin\",41.08333333");
    add_srs_wkt (p, 10,
		 "333334],PARAMETER[\"central_meridian\",-71.5],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"scale_factor\",0.99999375],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",328083.3333],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 13,
		 "\"EPSG\",\"3654\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3655, "epsg", 3655,
		      "NAD83(NSRS2007) / South Carolina");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=34.83333333333334 +lat_2=32.5 +lat_0=31");
    add_proj4text (p, 1,
		   ".83333333333333 +lon_0=-81 +x_0=609600 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / South Carolina\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_");
    add_srs_wkt (p, 2,
		 "System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",34.83333333333334],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",32.5],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 11,
		 ",31.83333333333333],PARAMETER[\"central_meridian\",-81],");
    add_srs_wkt (p, 12,
		 "PARAMETER[\"false_easting\",609600],PARAMETER[\"false_no");
    add_srs_wkt (p, 13,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"3655\"],AXIS[\"X\",EAST");
    add_srs_wkt (p, 14, "],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3656, "epsg", 3656,
		      "NAD83(NSRS2007) / South Carolina (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=34.83333333333334 +lat_2=32.5 +lat_0=31");
    add_proj4text (p, 1,
		   ".83333333333333 +lon_0=-81 +x_0=609600 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +towgs84=0,0,0,0,0,0,0 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / South Carolina (ft)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 8,
		 "9002\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"standard_parallel_1\",34.83333333333334],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"standard_parallel_2\",32.5],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 11,
		 "f_origin\",31.83333333333333],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 12,
		 "an\",-81],PARAMETER[\"false_easting\",2000000],PARAMETER");
    add_srs_wkt (p, 13,
		 "[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3656\"],AXIS");
    add_srs_wkt (p, 14, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3657, "epsg", 3657,
		      "NAD83(NSRS2007) / South Dakota North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.68333333333333 +lat_2=44.41666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=43.83333333333334 +lon_0=-100 +x_0=600000 +y_");
    add_proj4text (p, 2,
		   "0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_def");
    add_proj4text (p, 3, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / South Dakota North\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",45.68333333333333],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "standard_parallel_2\",44.41666666666666],PARAMETER[\"lat");
    add_srs_wkt (p, 11,
		 "itude_of_origin\",43.83333333333334],PARAMETER[\"central");
    add_srs_wkt (p, 12,
		 "_meridian\",-100],PARAMETER[\"false_easting\",600000],PA");
    add_srs_wkt (p, 13,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3657\"");
    add_srs_wkt (p, 14, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3658, "epsg", 3658,
		      "NAD83(NSRS2007) / South Dakota North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.68333333333333 +lat_2=44.41666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=43.83333333333334 +lon_0=-100 +x_0=600000 +y_");
    add_proj4text (p, 2,
		   "0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no");
    add_proj4text (p, 3, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / South Dakota North (ftUS)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial");
    add_srs_wkt (p, 2,
		 "_Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137,2");
    add_srs_wkt (p, 3,
		 "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwi");
    add_srs_wkt (p, 5,
		 "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0");
    add_srs_wkt (p, 6,
		 "1745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY");
    add_srs_wkt (p, 7,
		 "[\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.304800609");
    add_srs_wkt (p, 8,
		 "6012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambe");
    add_srs_wkt (p, 9,
		 "rt_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_");
    add_srs_wkt (p, 10,
		 "1\",45.68333333333333],PARAMETER[\"standard_parallel_2\"");
    add_srs_wkt (p, 11,
		 ",44.41666666666666],PARAMETER[\"latitude_of_origin\",43.");
    add_srs_wkt (p, 12,
		 "83333333333334],PARAMETER[\"central_meridian\",-100],PAR");
    add_srs_wkt (p, 13,
		 "AMETER[\"false_easting\",1968500],PARAMETER[\"false_nort");
    add_srs_wkt (p, 14,
		 "hing\",0],AUTHORITY[\"EPSG\",\"3658\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 15, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3659, "epsg", 3659,
		      "NAD83(NSRS2007) / South Dakota South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.4 +lat_2=42.83333333333334 +lat_0=42");
    add_proj4text (p, 1,
		   ".33333333333334 +lon_0=-100.3333333333333 +x_0=600000 +y");
    add_proj4text (p, 2,
		   "_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_de");
    add_proj4text (p, 3, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / South Dakota South\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",44.4],PARAMETER[\"standard_para");
    add_srs_wkt (p, 10,
		 "llel_2\",42.83333333333334],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 11,
		 "in\",42.33333333333334],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 12,
		 "100.3333333333333],PARAMETER[\"false_easting\",600000],P");
    add_srs_wkt (p, 13,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3659");
    add_srs_wkt (p, 14, "\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3660, "epsg", 3660,
		      "NAD83(NSRS2007) / South Dakota South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.4 +lat_2=42.83333333333334 +lat_0=42");
    add_proj4text (p, 1,
		   ".33333333333334 +lon_0=-100.3333333333333 +x_0=600000 +y");
    add_proj4text (p, 2,
		   "_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +n");
    add_proj4text (p, 3, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / South Dakota South (ftUS)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial");
    add_srs_wkt (p, 2,
		 "_Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137,2");
    add_srs_wkt (p, 3,
		 "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwi");
    add_srs_wkt (p, 5,
		 "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0");
    add_srs_wkt (p, 6,
		 "1745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY");
    add_srs_wkt (p, 7,
		 "[\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.304800609");
    add_srs_wkt (p, 8,
		 "6012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambe");
    add_srs_wkt (p, 9,
		 "rt_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_");
    add_srs_wkt (p, 10,
		 "1\",44.4],PARAMETER[\"standard_parallel_2\",42.833333333");
    add_srs_wkt (p, 11,
		 "33334],PARAMETER[\"latitude_of_origin\",42.3333333333333");
    add_srs_wkt (p, 12,
		 "4],PARAMETER[\"central_meridian\",-100.3333333333333],PA");
    add_srs_wkt (p, 13,
		 "RAMETER[\"false_easting\",1968500],PARAMETER[\"false_nor");
    add_srs_wkt (p, 14,
		 "thing\",0],AUTHORITY[\"EPSG\",\"3660\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 15, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3661, "epsg", 3661,
		      "NAD83(NSRS2007) / Tennessee");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.41666666666666 +lat_2=35.25 +lat_0=3");
    add_proj4text (p, 1,
		   "4.33333333333334 +lon_0=-86 +x_0=600000 +y_0=0 +ellps=GR");
    add_proj4text (p, 2, "S80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Tennessee\",GEOGCS[\"NAD83(NS");
    add_srs_wkt (p, 1,
		 "RS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Syste");
    add_srs_wkt (p, 2,
		 "m_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4759");
    add_srs_wkt (p, 7,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 8,
		 "ECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"stan");
    add_srs_wkt (p, 9,
		 "dard_parallel_1\",36.41666666666666],PARAMETER[\"standar");
    add_srs_wkt (p, 10,
		 "d_parallel_2\",35.25],PARAMETER[\"latitude_of_origin\",3");
    add_srs_wkt (p, 11,
		 "4.33333333333334],PARAMETER[\"central_meridian\",-86],PA");
    add_srs_wkt (p, 12,
		 "RAMETER[\"false_easting\",600000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 13,
		 "hing\",0],AUTHORITY[\"EPSG\",\"3661\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 14, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3662, "epsg", 3662,
		      "NAD83(NSRS2007) / Tennessee (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.41666666666666 +lat_2=35.25 +lat_0=3");
    add_proj4text (p, 1,
		   "4.33333333333334 +lon_0=-86 +x_0=600000 +y_0=0 +ellps=GR");
    add_proj4text (p, 2, "S80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Tennessee (ftUS)\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referenc");
    add_srs_wkt (p, 2,
		 "e_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 8,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 9,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",36.41");
    add_srs_wkt (p, 10,
		 "666666666666],PARAMETER[\"standard_parallel_2\",35.25],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"latitude_of_origin\",34.33333333333334],PARAM");
    add_srs_wkt (p, 12,
		 "ETER[\"central_meridian\",-86],PARAMETER[\"false_easting");
    add_srs_wkt (p, 13,
		 "\",1968500],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 14, "EPSG\",\"3662\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3663, "epsg", 3663,
		      "NAD83(NSRS2007) / Texas Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=31.88333333333333 +lat_2=30.11666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=29.66666666666667 +lon_0=-100.3333333333333 +");
    add_proj4text (p, 2,
		   "x_0=700000 +y_0=3000000 +ellps=GRS80 +towgs84=0,0,0,0,0,");
    add_proj4text (p, 3, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Texas Central\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_S");
    add_srs_wkt (p, 2,
		 "ystem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",31.88333333333333],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",30.11666666666667],PARAMETER[\"latitu");
    add_srs_wkt (p, 11,
		 "de_of_origin\",29.66666666666667],PARAMETER[\"central_me");
    add_srs_wkt (p, 12,
		 "ridian\",-100.3333333333333],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 13,
		 ",700000],PARAMETER[\"false_northing\",3000000],AUTHORITY");
    add_srs_wkt (p, 14,
		 "[\"EPSG\",\"3663\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3664, "epsg", 3664,
		      "NAD83(NSRS2007) / Texas Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=31.88333333333333 +lat_2=30.11666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=29.66666666666667 +lon_0=-100.3333333333333 +");
    add_proj4text (p, 2,
		   "x_0=699999.9998983998 +y_0=3000000 +ellps=GRS80 +towgs84");
    add_proj4text (p, 3, "=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Texas Central (ftUS)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refe");
    add_srs_wkt (p, 2,
		 "rence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25");
    add_srs_wkt (p, 3,
		 "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4759\"]],UNIT[\"US survey foot\",0.30480060960121");
    add_srs_wkt (p, 8,
		 "92,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Co");
    add_srs_wkt (p, 9,
		 "nformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",3");
    add_srs_wkt (p, 10,
		 "1.88333333333333],PARAMETER[\"standard_parallel_2\",30.1");
    add_srs_wkt (p, 11,
		 "1666666666667],PARAMETER[\"latitude_of_origin\",29.66666");
    add_srs_wkt (p, 12,
		 "666666667],PARAMETER[\"central_meridian\",-100.333333333");
    add_srs_wkt (p, 13,
		 "3333],PARAMETER[\"false_easting\",2296583.333],PARAMETER");
    add_srs_wkt (p, 14,
		 "[\"false_northing\",9842500.000000002],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 15, ",\"3664\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3665, "epsg", 3665,
		      "NAD83(NSRS2007) / Texas Centric Albers Equal Area");
    add_proj4text (p, 0,
		   "+proj=aea +lat_1=27.5 +lat_2=35 +lat_0=18 +lon_0=-100 +x");
    add_proj4text (p, 1,
		   "_0=1500000 +y_0=6000000 +ellps=GRS80 +towgs84=0,0,0,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Texas Centric Albers Equal Ar");
    add_srs_wkt (p, 1,
		 "ea\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_S");
    add_srs_wkt (p, 2,
		 "patial_Reference_System_2007\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 3,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84");
    add_srs_wkt (p, 4,
		 "[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"G");
    add_srs_wkt (p, 5,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 6,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 7,
		 "HORITY[\"EPSG\",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 8,
		 "EPSG\",\"9001\"]],PROJECTION[\"Albers_Conic_Equal_Area\"");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"standard_parallel_1\",27.5],PARAMETER[\"st");
    add_srs_wkt (p, 10,
		 "andard_parallel_2\",35],PARAMETER[\"latitude_of_center\"");
    add_srs_wkt (p, 11,
		 ",18],PARAMETER[\"longitude_of_center\",-100],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_easting\",1500000],PARAMETER[\"false_northing\",60");
    add_srs_wkt (p, 13,
		 "00000],AUTHORITY[\"EPSG\",\"3665\"],AXIS[\"X\",EAST],AXI");
    add_srs_wkt (p, 14, "S[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3666, "epsg", 3666,
		      "NAD83(NSRS2007) / Texas Centric Lambert Conformal");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=27.5 +lat_2=35 +lat_0=18 +lon_0=-100 +x");
    add_proj4text (p, 1,
		   "_0=1500000 +y_0=5000000 +ellps=GRS80 +towgs84=0,0,0,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Texas Centric Lambert Conform");
    add_srs_wkt (p, 1,
		 "al\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_S");
    add_srs_wkt (p, 2,
		 "patial_Reference_System_2007\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 3,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84");
    add_srs_wkt (p, 4,
		 "[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"G");
    add_srs_wkt (p, 5,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 6,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 7,
		 "HORITY[\"EPSG\",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 8,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 9,
		 "SP\"],PARAMETER[\"standard_parallel_1\",27.5],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"standard_parallel_2\",35],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 11,
		 "in\",18],PARAMETER[\"central_meridian\",-100],PARAMETER[");
    add_srs_wkt (p, 12,
		 "\"false_easting\",1500000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 13,
		 "5000000],AUTHORITY[\"EPSG\",\"3666\"],AXIS[\"X\",EAST],A");
    add_srs_wkt (p, 14, "XIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3667, "epsg", 3667,
		      "NAD83(NSRS2007) / Texas North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.18333333333333 +lat_2=34.65 +lat_0=3");
    add_proj4text (p, 1,
		   "4 +lon_0=-101.5 +x_0=200000 +y_0=1000000 +ellps=GRS80 +t");
    add_proj4text (p, 2, "owgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Texas North\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sys");
    add_srs_wkt (p, 2,
		 "tem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"47");
    add_srs_wkt (p, 7,
		 "59\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 9,
		 "andard_parallel_1\",36.18333333333333],PARAMETER[\"stand");
    add_srs_wkt (p, 10,
		 "ard_parallel_2\",34.65],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 11,
		 ",34],PARAMETER[\"central_meridian\",-101.5],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_easting\",200000],PARAMETER[\"false_northing\",100");
    add_srs_wkt (p, 13,
		 "0000],AUTHORITY[\"EPSG\",\"3667\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 14, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3668, "epsg", 3668,
		      "NAD83(NSRS2007) / Texas North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.18333333333333 +lat_2=34.65 +lat_0=3");
    add_proj4text (p, 1,
		   "4 +lon_0=-101.5 +x_0=200000.0001016002 +y_0=999999.99989");
    add_proj4text (p, 2,
		   "83998 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +");
    add_proj4text (p, 3, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Texas North (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 8,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 9,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",36.18");
    add_srs_wkt (p, 10,
		 "333333333333],PARAMETER[\"standard_parallel_2\",34.65],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"latitude_of_origin\",34],PARAMETER[\"central_");
    add_srs_wkt (p, 12,
		 "meridian\",-101.5],PARAMETER[\"false_easting\",656166.66");
    add_srs_wkt (p, 13,
		 "7],PARAMETER[\"false_northing\",3280833.333],AUTHORITY[\"");
    add_srs_wkt (p, 14, "EPSG\",\"3668\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3669, "epsg", 3669,
		      "NAD83(NSRS2007) / Texas North Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.96666666666667 +lat_2=32.13333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=31.66666666666667 +lon_0=-98.5 +x_0=600000 +y");
    add_proj4text (p, 2,
		   "_0=2000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m ");
    add_proj4text (p, 3, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Texas North Central\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"standard_parallel_1\",33.96666666666667],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"standard_parallel_2\",32.13333333333333],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "latitude_of_origin\",31.66666666666667],PARAMETER[\"cent");
    add_srs_wkt (p, 12,
		 "ral_meridian\",-98.5],PARAMETER[\"false_easting\",600000");
    add_srs_wkt (p, 13,
		 "],PARAMETER[\"false_northing\",2000000],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 14, "\",\"3669\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3670, "epsg", 3670,
		      "NAD83(NSRS2007) / Texas North Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.96666666666667 +lat_2=32.13333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=31.66666666666667 +lon_0=-98.5 +x_0=600000 +y");
    add_proj4text (p, 2,
		   "_0=2000000.0001016 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 3, "units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Texas North Central (ftUS)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatia");
    add_srs_wkt (p, 2,
		 "l_Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137,");
    add_srs_wkt (p, 3,
		 "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 5,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 6,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.30480060");
    add_srs_wkt (p, 8,
		 "96012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lamb");
    add_srs_wkt (p, 9,
		 "ert_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel");
    add_srs_wkt (p, 10,
		 "_1\",33.96666666666667],PARAMETER[\"standard_parallel_2\"");
    add_srs_wkt (p, 11,
		 ",32.13333333333333],PARAMETER[\"latitude_of_origin\",31.");
    add_srs_wkt (p, 12,
		 "66666666666667],PARAMETER[\"central_meridian\",-98.5],PA");
    add_srs_wkt (p, 13,
		 "RAMETER[\"false_easting\",1968500],PARAMETER[\"false_nor");
    add_srs_wkt (p, 14,
		 "thing\",6561666.667],AUTHORITY[\"EPSG\",\"3670\"],AXIS[\"");
    add_srs_wkt (p, 15, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3671, "epsg", 3671,
		      "NAD83(NSRS2007) / Texas South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=27.83333333333333 +lat_2=26.16666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=25.66666666666667 +lon_0=-98.5 +x_0=300000 +y");
    add_proj4text (p, 2,
		   "_0=5000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m ");
    add_proj4text (p, 3, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Texas South\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sys");
    add_srs_wkt (p, 2,
		 "tem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"47");
    add_srs_wkt (p, 7,
		 "59\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 9,
		 "andard_parallel_1\",27.83333333333333],PARAMETER[\"stand");
    add_srs_wkt (p, 10,
		 "ard_parallel_2\",26.16666666666667],PARAMETER[\"latitude");
    add_srs_wkt (p, 11,
		 "_of_origin\",25.66666666666667],PARAMETER[\"central_meri");
    add_srs_wkt (p, 12,
		 "dian\",-98.5],PARAMETER[\"false_easting\",300000],PARAME");
    add_srs_wkt (p, 13,
		 "TER[\"false_northing\",5000000],AUTHORITY[\"EPSG\",\"367");
    add_srs_wkt (p, 14, "1\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3672, "epsg", 3672,
		      "NAD83(NSRS2007) / Texas South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=27.83333333333333 +lat_2=26.16666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=25.66666666666667 +lon_0=-98.5 +x_0=300000.00");
    add_proj4text (p, 2,
		   "00000001 +y_0=5000000.0001016 +ellps=GRS80 +towgs84=0,0,");
    add_proj4text (p, 3, "0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Texas South (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 8,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 9,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",27.83");
    add_srs_wkt (p, 10,
		 "333333333333],PARAMETER[\"standard_parallel_2\",26.16666");
    add_srs_wkt (p, 11,
		 "666666667],PARAMETER[\"latitude_of_origin\",25.666666666");
    add_srs_wkt (p, 12,
		 "66667],PARAMETER[\"central_meridian\",-98.5],PARAMETER[\"");
    add_srs_wkt (p, 13,
		 "false_easting\",984250.0000000002],PARAMETER[\"false_nor");
    add_srs_wkt (p, 14,
		 "thing\",16404166.667],AUTHORITY[\"EPSG\",\"3672\"],AXIS[");
    add_srs_wkt (p, 15, "\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3673, "epsg", 3673,
		      "NAD83(NSRS2007) / Texas South Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.28333333333333 +lat_2=28.38333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=27.83333333333333 +lon_0=-99 +x_0=600000 +y_0");
    add_proj4text (p, 2,
		   "=4000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +n");
    add_proj4text (p, 3, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Texas South Central\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"standard_parallel_1\",30.28333333333333],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"standard_parallel_2\",28.38333333333333],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "latitude_of_origin\",27.83333333333333],PARAMETER[\"cent");
    add_srs_wkt (p, 12,
		 "ral_meridian\",-99],PARAMETER[\"false_easting\",600000],");
    add_srs_wkt (p, 13,
		 "PARAMETER[\"false_northing\",4000000],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 14, ",\"3673\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3674, "epsg", 3674,
		      "NAD83(NSRS2007) / Texas South Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.28333333333333 +lat_2=28.38333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=27.83333333333333 +lon_0=-99 +x_0=600000 +y_0");
    add_proj4text (p, 2,
		   "=3999999.9998984 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +un");
    add_proj4text (p, 3, "its=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Texas South Central (ftUS)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatia");
    add_srs_wkt (p, 2,
		 "l_Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137,");
    add_srs_wkt (p, 3,
		 "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 5,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 6,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.30480060");
    add_srs_wkt (p, 8,
		 "96012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lamb");
    add_srs_wkt (p, 9,
		 "ert_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel");
    add_srs_wkt (p, 10,
		 "_1\",30.28333333333333],PARAMETER[\"standard_parallel_2\"");
    add_srs_wkt (p, 11,
		 ",28.38333333333333],PARAMETER[\"latitude_of_origin\",27.");
    add_srs_wkt (p, 12,
		 "83333333333333],PARAMETER[\"central_meridian\",-99],PARA");
    add_srs_wkt (p, 13,
		 "METER[\"false_easting\",1968500],PARAMETER[\"false_north");
    add_srs_wkt (p, 14,
		 "ing\",13123333.333],AUTHORITY[\"EPSG\",\"3674\"],AXIS[\"");
    add_srs_wkt (p, 15, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3675, "epsg", 3675,
		      "NAD83(NSRS2007) / Utah Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.65 +lat_2=39.01666666666667 +lat_0=3");
    add_proj4text (p, 1,
		   "8.33333333333334 +lon_0=-111.5 +x_0=500000 +y_0=2000000 ");
    add_proj4text (p, 2,
		   "+ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Utah Central\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_1\",40.65],PARAMETER[\"standard_paralle");
    add_srs_wkt (p, 10,
		 "l_2\",39.01666666666667],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 11,
		 ",38.33333333333334],PARAMETER[\"central_meridian\",-111.");
    add_srs_wkt (p, 12,
		 "5],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 13,
		 "_northing\",2000000],AUTHORITY[\"EPSG\",\"3675\"],AXIS[\"");
    add_srs_wkt (p, 14, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3676, "epsg", 3676,
		      "NAD83(NSRS2007) / Utah Central (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.65 +lat_2=39.01666666666667 +lat_0=3");
    add_proj4text (p, 1,
		   "8.33333333333334 +lon_0=-111.5 +x_0=500000.0001504 +y_0=");
    add_proj4text (p, 2,
		   "1999999.999992 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +unit");
    add_proj4text (p, 3, "s=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Utah Central (ft)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 8,
		 "02\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"standard_parallel_1\",40.65],PARAMETER[\"standard");
    add_srs_wkt (p, 10,
		 "_parallel_2\",39.01666666666667],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 11,
		 "_origin\",38.33333333333334],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 12,
		 "n\",-111.5],PARAMETER[\"false_easting\",1640419.948],PAR");
    add_srs_wkt (p, 13,
		 "AMETER[\"false_northing\",6561679.79],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 14, ",\"3676\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3677, "epsg", 3677,
		      "NAD83(NSRS2007) / Utah Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.65 +lat_2=39.01666666666667 +lat_0=3");
    add_proj4text (p, 1,
		   "8.33333333333334 +lon_0=-111.5 +x_0=500000.00001016 +y_0");
    add_proj4text (p, 2,
		   "=2000000.00001016 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +u");
    add_proj4text (p, 3, "nits=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Utah Central (ftUS)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"US survey foot\",0.304800609601219");
    add_srs_wkt (p, 8,
		 "2,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Con");
    add_srs_wkt (p, 9,
		 "formal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",40");
    add_srs_wkt (p, 10,
		 ".65],PARAMETER[\"standard_parallel_2\",39.01666666666667");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"latitude_of_origin\",38.33333333333334],PA");
    add_srs_wkt (p, 12,
		 "RAMETER[\"central_meridian\",-111.5],PARAMETER[\"false_e");
    add_srs_wkt (p, 13,
		 "asting\",1640416.6667],PARAMETER[\"false_northing\",6561");
    add_srs_wkt (p, 14,
		 "666.666700001],AUTHORITY[\"EPSG\",\"3677\"],AXIS[\"X\",E");
    add_srs_wkt (p, 15, "AST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3678, "epsg", 3678,
		      "NAD83(NSRS2007) / Utah North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.78333333333333 +lat_2=40.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=40.33333333333334 +lon_0=-111.5 +x_0=500000 +");
    add_proj4text (p, 2,
		   "y_0=1000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m");
    add_proj4text (p, 3, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Utah North\",GEOGCS[\"NAD83(N");
    add_srs_wkt (p, 1,
		 "SRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Syst");
    add_srs_wkt (p, 2,
		 "em_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"475");
    add_srs_wkt (p, 7,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"sta");
    add_srs_wkt (p, 9,
		 "ndard_parallel_1\",41.78333333333333],PARAMETER[\"standa");
    add_srs_wkt (p, 10,
		 "rd_parallel_2\",40.71666666666667],PARAMETER[\"latitude_");
    add_srs_wkt (p, 11,
		 "of_origin\",40.33333333333334],PARAMETER[\"central_merid");
    add_srs_wkt (p, 12,
		 "ian\",-111.5],PARAMETER[\"false_easting\",500000],PARAME");
    add_srs_wkt (p, 13,
		 "TER[\"false_northing\",1000000],AUTHORITY[\"EPSG\",\"367");
    add_srs_wkt (p, 14, "8\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3679, "epsg", 3679,
		      "NAD83(NSRS2007) / Utah North (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.78333333333333 +lat_2=40.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=40.33333333333334 +lon_0=-111.5 +x_0=500000.0");
    add_proj4text (p, 2,
		   "001504 +y_0=999999.9999960001 +ellps=GRS80 +towgs84=0,0,");
    add_proj4text (p, 3, "0,0,0,0,0 +units=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Utah North (ft)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference");
    add_srs_wkt (p, 2,
		 "_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 3,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 6,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4759\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "2\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"standard_parallel_1\",41.78333333333333],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"standard_parallel_2\",40.71666666666667],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "latitude_of_origin\",40.33333333333334],PARAMETER[\"cent");
    add_srs_wkt (p, 12,
		 "ral_meridian\",-111.5],PARAMETER[\"false_easting\",16404");
    add_srs_wkt (p, 13,
		 "19.948],PARAMETER[\"false_northing\",3280839.895],AUTHOR");
    add_srs_wkt (p, 14,
		 "ITY[\"EPSG\",\"3679\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH");
    add_srs_wkt (p, 15, "]]");
    p = add_epsg_def (first, last, 3680, "epsg", 3680,
		      "NAD83(NSRS2007) / Utah North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.78333333333333 +lat_2=40.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=40.33333333333334 +lon_0=-111.5 +x_0=500000.0");
    add_proj4text (p, 2,
		   "0001016 +y_0=999999.9999898402 +ellps=GRS80 +towgs84=0,0");
    add_proj4text (p, 3, ",0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Utah North (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 8,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 9,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",41.78");
    add_srs_wkt (p, 10,
		 "333333333333],PARAMETER[\"standard_parallel_2\",40.71666");
    add_srs_wkt (p, 11,
		 "666666667],PARAMETER[\"latitude_of_origin\",40.333333333");
    add_srs_wkt (p, 12,
		 "33334],PARAMETER[\"central_meridian\",-111.5],PARAMETER[");
    add_srs_wkt (p, 13,
		 "\"false_easting\",1640416.6667],PARAMETER[\"false_northi");
    add_srs_wkt (p, 14,
		 "ng\",3280833.333300001],AUTHORITY[\"EPSG\",\"3680\"],AXI");
    add_srs_wkt (p, 15, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3681, "epsg", 3681,
		      "NAD83(NSRS2007) / Utah South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.35 +lat_2=37.21666666666667 +lat_0=3");
    add_proj4text (p, 1,
		   "6.66666666666666 +lon_0=-111.5 +x_0=500000 +y_0=3000000 ");
    add_proj4text (p, 2,
		   "+ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Utah South\",GEOGCS[\"NAD83(N");
    add_srs_wkt (p, 1,
		 "SRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Syst");
    add_srs_wkt (p, 2,
		 "em_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"475");
    add_srs_wkt (p, 7,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"sta");
    add_srs_wkt (p, 9,
		 "ndard_parallel_1\",38.35],PARAMETER[\"standard_parallel_");
    add_srs_wkt (p, 10,
		 "2\",37.21666666666667],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 11,
		 "36.66666666666666],PARAMETER[\"central_meridian\",-111.5");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_easting\",500000],PARAMETER[\"false_");
    add_srs_wkt (p, 13,
		 "northing\",3000000],AUTHORITY[\"EPSG\",\"3681\"],AXIS[\"");
    add_srs_wkt (p, 14, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3682, "epsg", 3682,
		      "NAD83(NSRS2007) / Utah South (ft)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.35 +lat_2=37.21666666666667 +lat_0=3");
    add_proj4text (p, 1,
		   "6.66666666666666 +lon_0=-111.5 +x_0=500000.0001504 +y_0=");
    add_proj4text (p, 2,
		   "2999999.999988 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +unit");
    add_proj4text (p, 3, "s=ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Utah South (ft)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference");
    add_srs_wkt (p, 2,
		 "_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 3,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 6,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4759\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "2\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"standard_parallel_1\",38.35],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",37.21666666666667],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",36.66666666666666],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 12,
		 "\",-111.5],PARAMETER[\"false_easting\",1640419.948],PARA");
    add_srs_wkt (p, 13,
		 "METER[\"false_northing\",9842519.685],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 14, ",\"3682\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3683, "epsg", 3683,
		      "NAD83(NSRS2007) / Utah South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.35 +lat_2=37.21666666666667 +lat_0=3");
    add_proj4text (p, 1,
		   "6.66666666666666 +lon_0=-111.5 +x_0=500000.00001016 +y_0");
    add_proj4text (p, 2,
		   "=3000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-f");
    add_proj4text (p, 3, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Utah South (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 8,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 9,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",38.35");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"standard_parallel_2\",37.21666666666667],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"latitude_of_origin\",36.66666666666666],PARAM");
    add_srs_wkt (p, 12,
		 "ETER[\"central_meridian\",-111.5],PARAMETER[\"false_east");
    add_srs_wkt (p, 13,
		 "ing\",1640416.6667],PARAMETER[\"false_northing\",9842500");
    add_srs_wkt (p, 14,
		 ".000000002],AUTHORITY[\"EPSG\",\"3683\"],AXIS[\"X\",EAST");
    add_srs_wkt (p, 15, "],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3684, "epsg", 3684,
		      "NAD83(NSRS2007) / Vermont");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.5 +lon_0=-72.5 +k=0.999964286 +x_0");
    add_proj4text (p, 1,
		   "=500000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +unit");
    add_proj4text (p, 2, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Vermont\",GEOGCS[\"NAD83(NSRS");
    add_srs_wkt (p, 1,
		 "2007)\",DATUM[\"NAD83_National_Spatial_Reference_System_");
    add_srs_wkt (p, 2,
		 "2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4759\"]");
    add_srs_wkt (p, 7,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 8,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 9,
		 "gin\",42.5],PARAMETER[\"central_meridian\",-72.5],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",0.999964286],PARAMETER[\"false_east");
    add_srs_wkt (p, 11,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 12,
		 "\"EPSG\",\"3684\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3685, "epsg", 3685,
		      "NAD83(NSRS2007) / Virginia North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.2 +lat_2=38.03333333333333 +lat_0=37");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-78.5 +x_0=3500000 +y_0=2000000 +");
    add_proj4text (p, 2,
		   "ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Virginia North\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_");
    add_srs_wkt (p, 2,
		 "System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",39.2],PARAMETER[\"standard_paralle");
    add_srs_wkt (p, 10,
		 "l_2\",38.03333333333333],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 11,
		 ",37.66666666666666],PARAMETER[\"central_meridian\",-78.5");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_easting\",3500000],PARAMETER[\"false");
    add_srs_wkt (p, 13,
		 "_northing\",2000000],AUTHORITY[\"EPSG\",\"3685\"],AXIS[\"");
    add_srs_wkt (p, 14, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3686, "epsg", 3686,
		      "NAD83(NSRS2007) / Virginia North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.2 +lat_2=38.03333333333333 +lat_0=37");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-78.5 +x_0=3500000.0001016 +y_0=2");
    add_proj4text (p, 2,
		   "000000.0001016 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +unit");
    add_proj4text (p, 3, "s=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Virginia North (ftUS)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2");
    add_srs_wkt (p, 3,
		 "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 9,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 10,
		 "39.2],PARAMETER[\"standard_parallel_2\",38.0333333333333");
    add_srs_wkt (p, 11,
		 "3],PARAMETER[\"latitude_of_origin\",37.66666666666666],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"central_meridian\",-78.5],PARAMETER[\"false_e");
    add_srs_wkt (p, 13,
		 "asting\",11482916.667],PARAMETER[\"false_northing\",6561");
    add_srs_wkt (p, 14,
		 "666.667],AUTHORITY[\"EPSG\",\"3686\"],AXIS[\"X\",EAST],A");
    add_srs_wkt (p, 15, "XIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3687, "epsg", 3687,
		      "NAD83(NSRS2007) / Virginia South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.96666666666667 +lat_2=36.76666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.33333333333334 +lon_0=-78.5 +x_0=3500000 +");
    add_proj4text (p, 2,
		   "y_0=1000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m");
    add_proj4text (p, 3, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Virginia South\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_");
    add_srs_wkt (p, 2,
		 "System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",37.96666666666667],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",36.76666666666667],PARAMETER[\"latitu");
    add_srs_wkt (p, 11,
		 "de_of_origin\",36.33333333333334],PARAMETER[\"central_me");
    add_srs_wkt (p, 12,
		 "ridian\",-78.5],PARAMETER[\"false_easting\",3500000],PAR");
    add_srs_wkt (p, 13,
		 "AMETER[\"false_northing\",1000000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 14, "3687\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3688, "epsg", 3688,
		      "NAD83(NSRS2007) / Virginia South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.96666666666667 +lat_2=36.76666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.33333333333334 +lon_0=-78.5 +x_0=3500000.0");
    add_proj4text (p, 2,
		   "001016 +y_0=999999.9998983998 +ellps=GRS80 +towgs84=0,0,");
    add_proj4text (p, 3, "0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Virginia South (ftUS)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Ref");
    add_srs_wkt (p, 2,
		 "erence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2");
    add_srs_wkt (p, 3,
		 "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 9,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 10,
		 "37.96666666666667],PARAMETER[\"standard_parallel_2\",36.");
    add_srs_wkt (p, 11,
		 "76666666666667],PARAMETER[\"latitude_of_origin\",36.3333");
    add_srs_wkt (p, 12,
		 "3333333334],PARAMETER[\"central_meridian\",-78.5],PARAME");
    add_srs_wkt (p, 13,
		 "TER[\"false_easting\",11482916.667],PARAMETER[\"false_no");
    add_srs_wkt (p, 14,
		 "rthing\",3280833.333],AUTHORITY[\"EPSG\",\"3688\"],AXIS[");
    add_srs_wkt (p, 15, "\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3689, "epsg", 3689,
		      "NAD83(NSRS2007) / Washington North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.73333333333333 +lat_2=47.5 +lat_0=47");
    add_proj4text (p, 1,
		   " +lon_0=-120.8333333333333 +x_0=500000 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Washington North\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referenc");
    add_srs_wkt (p, 2,
		 "e_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",48.73333333333333],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "standard_parallel_2\",47.5],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 11,
		 "in\",47],PARAMETER[\"central_meridian\",-120.83333333333");
    add_srs_wkt (p, 12,
		 "33],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 13,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"3689\"],AXIS[\"X\",");
    add_srs_wkt (p, 14, "EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3690, "epsg", 3690,
		      "NAD83(NSRS2007) / Washington North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.73333333333333 +lat_2=47.5 +lat_0=47");
    add_proj4text (p, 1,
		   " +lon_0=-120.8333333333333 +x_0=500000.0001016001 +y_0=0");
    add_proj4text (p, 2,
		   " +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_de");
    add_proj4text (p, 3, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Washington North (ftUS)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_R");
    add_srs_wkt (p, 2,
		 "eference_System_2007\",SPHEROID[\"GRS 1980\",6378137,298");
    add_srs_wkt (p, 3,
		 ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 5,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 6,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.304800609601");
    add_srs_wkt (p, 8,
		 "2192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_");
    add_srs_wkt (p, 9,
		 "Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 10,
		 ",48.73333333333333],PARAMETER[\"standard_parallel_2\",47");
    add_srs_wkt (p, 11,
		 ".5],PARAMETER[\"latitude_of_origin\",47],PARAMETER[\"cen");
    add_srs_wkt (p, 12,
		 "tral_meridian\",-120.8333333333333],PARAMETER[\"false_ea");
    add_srs_wkt (p, 13,
		 "sting\",1640416.667],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 14,
		 "HORITY[\"EPSG\",\"3690\"],AXIS[\"X\",EAST],AXIS[\"Y\",NO");
    add_srs_wkt (p, 15, "RTH]]");
    p = add_epsg_def (first, last, 3691, "epsg", 3691,
		      "NAD83(NSRS2007) / Washington South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.33333333333334 +lat_2=45.83333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=45.33333333333334 +lon_0=-120.5 +x_0=500000 +");
    add_proj4text (p, 2,
		   "y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_d");
    add_proj4text (p, 3, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Washington South\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referenc");
    add_srs_wkt (p, 2,
		 "e_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",47.33333333333334],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "standard_parallel_2\",45.83333333333334],PARAMETER[\"lat");
    add_srs_wkt (p, 11,
		 "itude_of_origin\",45.33333333333334],PARAMETER[\"central");
    add_srs_wkt (p, 12,
		 "_meridian\",-120.5],PARAMETER[\"false_easting\",500000],");
    add_srs_wkt (p, 13,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"369");
    add_srs_wkt (p, 14, "1\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3692, "epsg", 3692,
		      "NAD83(NSRS2007) / Washington South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.33333333333334 +lat_2=45.83333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=45.33333333333334 +lon_0=-120.5 +x_0=500000.0");
    add_proj4text (p, 2,
		   "001016001 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +un");
    add_proj4text (p, 3, "its=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Washington South (ftUS)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_R");
    add_srs_wkt (p, 2,
		 "eference_System_2007\",SPHEROID[\"GRS 1980\",6378137,298");
    add_srs_wkt (p, 3,
		 ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 5,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 6,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.304800609601");
    add_srs_wkt (p, 8,
		 "2192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_");
    add_srs_wkt (p, 9,
		 "Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 10,
		 ",47.33333333333334],PARAMETER[\"standard_parallel_2\",45");
    add_srs_wkt (p, 11,
		 ".83333333333334],PARAMETER[\"latitude_of_origin\",45.333");
    add_srs_wkt (p, 12,
		 "33333333334],PARAMETER[\"central_meridian\",-120.5],PARA");
    add_srs_wkt (p, 13,
		 "METER[\"false_easting\",1640416.667],PARAMETER[\"false_n");
    add_srs_wkt (p, 14,
		 "orthing\",0],AUTHORITY[\"EPSG\",\"3692\"],AXIS[\"X\",EAS");
    add_srs_wkt (p, 15, "T],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3693, "epsg", 3693,
		      "NAD83(NSRS2007) / West Virginia North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.25 +lat_2=39 +lat_0=38.5 +lon_0=-79.");
    add_proj4text (p, 1,
		   "5 +x_0=600000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0");
    add_proj4text (p, 2, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / West Virginia North\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"standard_parallel_1\",40.25],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",39],PARAMETER[\"latitude_of_origin\",38.5],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"central_meridian\",-79.5],PARAMETER[\"false_");
    add_srs_wkt (p, 12,
		 "easting\",600000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 13,
		 "ITY[\"EPSG\",\"3693\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH");
    add_srs_wkt (p, 14, "]]");
    p = add_epsg_def (first, last, 3694, "epsg", 3694,
		      "NAD83(NSRS2007) / West Virginia South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.88333333333333 +lat_2=37.48333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=37 +lon_0=-81 +x_0=600000 +y_0=0 +ellps=GRS80");
    add_proj4text (p, 2, " +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / West Virginia South\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"standard_parallel_1\",38.88333333333333],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"standard_parallel_2\",37.48333333333333],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "latitude_of_origin\",37],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 12,
		 "-81],PARAMETER[\"false_easting\",600000],PARAMETER[\"fal");
    add_srs_wkt (p, 13,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"3694\"],AXIS[\"X\"");
    add_srs_wkt (p, 14, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3695, "epsg", 3695,
		      "NAD83(NSRS2007) / Wisconsin Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.5 +lat_2=44.25 +lat_0=43.83333333333");
    add_proj4text (p, 1,
		   "334 +lon_0=-90 +x_0=600000 +y_0=0 +ellps=GRS80 +towgs84=");
    add_proj4text (p, 2, "0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Wisconsin Central\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",45.5],PARAMETER[\"standard_para");
    add_srs_wkt (p, 10,
		 "llel_2\",44.25],PARAMETER[\"latitude_of_origin\",43.8333");
    add_srs_wkt (p, 11,
		 "3333333334],PARAMETER[\"central_meridian\",-90],PARAMETE");
    add_srs_wkt (p, 12,
		 "R[\"false_easting\",600000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 13,
		 ",0],AUTHORITY[\"EPSG\",\"3695\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 14, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3696, "epsg", 3696,
		      "NAD83(NSRS2007) / Wisconsin Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.5 +lat_2=44.25 +lat_0=43.83333333333");
    add_proj4text (p, 1,
		   "334 +lon_0=-90 +x_0=600000 +y_0=0 +ellps=GRS80 +towgs84=");
    add_proj4text (p, 2, "0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Wisconsin Central (ftUS)\",GE");
    add_srs_wkt (p, 1,
		 "OGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_");
    add_srs_wkt (p, 2,
		 "Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137,29");
    add_srs_wkt (p, 3,
		 "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 5,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 6,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 7,
		 "\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096");
    add_srs_wkt (p, 8,
		 "012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lamber");
    add_srs_wkt (p, 9,
		 "t_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1");
    add_srs_wkt (p, 10,
		 "\",45.5],PARAMETER[\"standard_parallel_2\",44.25],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"latitude_of_origin\",43.83333333333334],PARAMETER[");
    add_srs_wkt (p, 12,
		 "\"central_meridian\",-90],PARAMETER[\"false_easting\",19");
    add_srs_wkt (p, 13,
		 "68500],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 14, ",\"3696\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3697, "epsg", 3697,
		      "NAD83(NSRS2007) / Wisconsin North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46.76666666666667 +lat_2=45.56666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=45.16666666666666 +lon_0=-90 +x_0=600000 +y_0");
    add_proj4text (p, 2,
		   "=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_proj4text (p, 3, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Wisconsin North\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference");
    add_srs_wkt (p, 2,
		 "_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 3,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 6,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 8,
		 "],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"standard_parallel_1\",46.76666666666667],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "tandard_parallel_2\",45.56666666666667],PARAMETER[\"lati");
    add_srs_wkt (p, 11,
		 "tude_of_origin\",45.16666666666666],PARAMETER[\"central_");
    add_srs_wkt (p, 12,
		 "meridian\",-90],PARAMETER[\"false_easting\",600000],PARA");
    add_srs_wkt (p, 13,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3697\"]");
    add_srs_wkt (p, 14, ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3698, "epsg", 3698,
		      "NAD83(NSRS2007) / Wisconsin North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46.76666666666667 +lat_2=45.56666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=45.16666666666666 +lon_0=-90 +x_0=600000 +y_0");
    add_proj4text (p, 2,
		   "=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_");
    add_proj4text (p, 3, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Wisconsin North (ftUS)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Re");
    add_srs_wkt (p, 2,
		 "ference_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.");
    add_srs_wkt (p, 3,
		 "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 9,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 10,
		 "46.76666666666667],PARAMETER[\"standard_parallel_2\",45.");
    add_srs_wkt (p, 11,
		 "56666666666667],PARAMETER[\"latitude_of_origin\",45.1666");
    add_srs_wkt (p, 12,
		 "6666666666],PARAMETER[\"central_meridian\",-90],PARAMETE");
    add_srs_wkt (p, 13,
		 "R[\"false_easting\",1968500],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 14,
		 ",0],AUTHORITY[\"EPSG\",\"3698\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 15, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3699, "epsg", 3699,
		      "NAD83(NSRS2007) / Wisconsin South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.06666666666667 +lat_2=42.73333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=42 +lon_0=-90 +x_0=600000 +y_0=0 +ellps=GRS80");
    add_proj4text (p, 2, " +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Wisconsin South\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference");
    add_srs_wkt (p, 2,
		 "_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 3,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 6,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 8,
		 "],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"standard_parallel_1\",44.06666666666667],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "tandard_parallel_2\",42.73333333333333],PARAMETER[\"lati");
    add_srs_wkt (p, 11,
		 "tude_of_origin\",42],PARAMETER[\"central_meridian\",-90]");
    add_srs_wkt (p, 12,
		 ",PARAMETER[\"false_easting\",600000],PARAMETER[\"false_n");
    add_srs_wkt (p, 13,
		 "orthing\",0],AUTHORITY[\"EPSG\",\"3699\"],AXIS[\"X\",EAS");
    add_srs_wkt (p, 14, "T],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3700, "epsg", 3700,
		      "NAD83(NSRS2007) / Wisconsin South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.06666666666667 +lat_2=42.73333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=42 +lon_0=-90 +x_0=600000 +y_0=0 +ellps=GRS80");
    add_proj4text (p, 2, " +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Wisconsin South (ftUS)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Re");
    add_srs_wkt (p, 2,
		 "ference_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.");
    add_srs_wkt (p, 3,
		 "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 9,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 10,
		 "44.06666666666667],PARAMETER[\"standard_parallel_2\",42.");
    add_srs_wkt (p, 11,
		 "73333333333333],PARAMETER[\"latitude_of_origin\",42],PAR");
    add_srs_wkt (p, 12,
		 "AMETER[\"central_meridian\",-90],PARAMETER[\"false_easti");
    add_srs_wkt (p, 13,
		 "ng\",1968500],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 14,
		 "\"EPSG\",\"3700\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3701, "epsg", 3701,
		      "NAD83(NSRS2007) / Wisconsin Transverse Mercator");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-90 +k=0.9996 +x_0=520000 +y");
    add_proj4text (p, 1,
		   "_0=-4480000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m");
    add_proj4text (p, 2, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Wisconsin Transverse Mercator");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spa");
    add_srs_wkt (p, 2,
		 "tial_Reference_System_2007\",SPHEROID[\"GRS 1980\",63781");
    add_srs_wkt (p, 3,
		 "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0");
    add_srs_wkt (p, 4,
		 ",0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 5,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 6,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 8,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 10,
		 "dian\",-90],PARAMETER[\"scale_factor\",0.9996],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_easting\",520000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 12,
		 "-4480000],AUTHORITY[\"EPSG\",\"3701\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 13, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3702, "epsg", 3702,
		      "NAD83(NSRS2007) / Wyoming East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-105.1666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "999375 +x_0=200000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 2, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Wyoming East\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",40.5],PARAMETER[\"central_meridian\",-105.16");
    add_srs_wkt (p, 10,
		 "66666666667],PARAMETER[\"scale_factor\",0.9999375],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_easting\",200000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",0],AUTHORITY[\"EPSG\",\"3702\"],AXIS[\"X\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3703, "epsg", 3703,
		      "NAD83(NSRS2007) / Wyoming East Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-107.3333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "999375 +x_0=400000 +y_0=100000 +ellps=GRS80 +towgs84=0,0");
    add_proj4text (p, 2, ",0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Wyoming East Central\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refe");
    add_srs_wkt (p, 2,
		 "rence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25");
    add_srs_wkt (p, 3,
		 "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 8,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 9,
		 "atitude_of_origin\",40.5],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 10,
		 ",-107.3333333333333],PARAMETER[\"scale_factor\",0.999937");
    add_srs_wkt (p, 11,
		 "5],PARAMETER[\"false_easting\",400000],PARAMETER[\"false");
    add_srs_wkt (p, 12,
		 "_northing\",100000],AUTHORITY[\"EPSG\",\"3703\"],AXIS[\"");
    add_srs_wkt (p, 13, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3704, "epsg", 3704,
		      "NAD83(NSRS2007) / Wyoming West Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-108.75 +k=0.9999375 +x_0");
    add_proj4text (p, 1,
		   "=600000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +unit");
    add_proj4text (p, 2, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Wyoming West Central\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refe");
    add_srs_wkt (p, 2,
		 "rence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25");
    add_srs_wkt (p, 3,
		 "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 8,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 9,
		 "atitude_of_origin\",40.5],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 10,
		 ",-108.75],PARAMETER[\"scale_factor\",0.9999375],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_easting\",600000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"3704\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3705, "epsg", 3705,
		      "NAD83(NSRS2007) / Wyoming West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-110.0833333333333 +k=0.9");
    add_proj4text (p, 1,
		   "999375 +x_0=800000 +y_0=100000 +ellps=GRS80 +towgs84=0,0");
    add_proj4text (p, 2, ",0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Wyoming West\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",40.5],PARAMETER[\"central_meridian\",-110.08");
    add_srs_wkt (p, 10,
		 "33333333333],PARAMETER[\"scale_factor\",0.9999375],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_easting\",800000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",100000],AUTHORITY[\"EPSG\",\"3705\"],AXIS[\"X\",EAST");
    add_srs_wkt (p, 13, "],AXIS[\"Y\",NORTH]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_13 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 3706, "epsg", 3706,
		      "NAD83(NSRS2007) / UTM zone 59N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=59 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 59N\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",171],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12,
		 "G\",\"3706\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 3707, "epsg", 3707,
		      "NAD83(NSRS2007) / UTM zone 60N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=60 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 60N\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",177],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12,
		 "G\",\"3707\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 3708, "epsg", 3708,
		      "NAD83(NSRS2007) / UTM zone 1N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=1 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +u");
    add_proj4text (p, 1, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 1N\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sys");
    add_srs_wkt (p, 2,
		 "tem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"47");
    add_srs_wkt (p, 7,
		 "59\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",-177],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12,
		 "G\",\"3708\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 3709, "epsg", 3709,
		      "NAD83(NSRS2007) / UTM zone 2N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=2 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +u");
    add_proj4text (p, 1, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 2N\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sys");
    add_srs_wkt (p, 2,
		 "tem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"47");
    add_srs_wkt (p, 7,
		 "59\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",-171],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12,
		 "G\",\"3709\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 3710, "epsg", 3710,
		      "NAD83(NSRS2007) / UTM zone 3N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=3 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +u");
    add_proj4text (p, 1, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 3N\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sys");
    add_srs_wkt (p, 2,
		 "tem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"47");
    add_srs_wkt (p, 7,
		 "59\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",-165],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12,
		 "G\",\"3710\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 3711, "epsg", 3711,
		      "NAD83(NSRS2007) / UTM zone 4N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=4 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +u");
    add_proj4text (p, 1, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 4N\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sys");
    add_srs_wkt (p, 2,
		 "tem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"47");
    add_srs_wkt (p, 7,
		 "59\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",-159],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12,
		 "G\",\"3711\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 3712, "epsg", 3712,
		      "NAD83(NSRS2007) / UTM zone 5N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=5 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +u");
    add_proj4text (p, 1, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 5N\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sys");
    add_srs_wkt (p, 2,
		 "tem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"47");
    add_srs_wkt (p, 7,
		 "59\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",-153],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12,
		 "G\",\"3712\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 3713, "epsg", 3713,
		      "NAD83(NSRS2007) / UTM zone 6N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=6 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +u");
    add_proj4text (p, 1, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 6N\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sys");
    add_srs_wkt (p, 2,
		 "tem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"47");
    add_srs_wkt (p, 7,
		 "59\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",-147],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12,
		 "G\",\"3713\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 3714, "epsg", 3714,
		      "NAD83(NSRS2007) / UTM zone 7N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=7 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +u");
    add_proj4text (p, 1, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 7N\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sys");
    add_srs_wkt (p, 2,
		 "tem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"47");
    add_srs_wkt (p, 7,
		 "59\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",-141],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12,
		 "G\",\"3714\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 3715, "epsg", 3715,
		      "NAD83(NSRS2007) / UTM zone 8N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=8 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +u");
    add_proj4text (p, 1, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 8N\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sys");
    add_srs_wkt (p, 2,
		 "tem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"47");
    add_srs_wkt (p, 7,
		 "59\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",-135],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12,
		 "G\",\"3715\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 3716, "epsg", 3716,
		      "NAD83(NSRS2007) / UTM zone 9N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=9 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +u");
    add_proj4text (p, 1, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 9N\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sys");
    add_srs_wkt (p, 2,
		 "tem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"47");
    add_srs_wkt (p, 7,
		 "59\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",-129],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12,
		 "G\",\"3716\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 3717, "epsg", 3717,
		      "NAD83(NSRS2007) / UTM zone 10N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=10 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 10N\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",-123],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 11,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3717\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3718, "epsg", 3718,
		      "NAD83(NSRS2007) / UTM zone 11N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=11 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 11N\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",-117],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 11,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3718\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3719, "epsg", 3719,
		      "NAD83(NSRS2007) / UTM zone 12N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=12 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 12N\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",-111],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 11,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3719\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3720, "epsg", 3720,
		      "NAD83(NSRS2007) / UTM zone 13N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=13 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 13N\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",-105],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 11,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12,
		 "PSG\",\"3720\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3721, "epsg", 3721,
		      "NAD83(NSRS2007) / UTM zone 14N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=14 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 14N\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",-99],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12,
		 "G\",\"3721\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 3722, "epsg", 3722,
		      "NAD83(NSRS2007) / UTM zone 15N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=15 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 15N\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",-93],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12,
		 "G\",\"3722\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 3723, "epsg", 3723,
		      "NAD83(NSRS2007) / UTM zone 16N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=16 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 16N\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",-87],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12,
		 "G\",\"3723\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 3724, "epsg", 3724,
		      "NAD83(NSRS2007) / UTM zone 17N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=17 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 17N\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",-81],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12,
		 "G\",\"3724\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 3725, "epsg", 3725,
		      "NAD83(NSRS2007) / UTM zone 18N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 18N\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",-75],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12,
		 "G\",\"3725\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 3726, "epsg", 3726,
		      "NAD83(NSRS2007) / UTM zone 19N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / UTM zone 19N\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",-69],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12,
		 "G\",\"3726\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 3727, "epsg", 3727,
		      "Reunion 1947 / TM Reunion");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-21.11666666666667 +lon_0=55.53333333");
    add_proj4text (p, 1,
		   "333333 +k=1 +x_0=160000 +y_0=50000 +ellps=intl +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Reunion 1947 / TM Reunion\",GEOGCS[\"Reunion 19");
    add_srs_wkt (p, 1,
		 "47\",DATUM[\"Reunion_1947\",SPHEROID[\"International 192");
    add_srs_wkt (p, 2,
		 "4\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6626\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4626\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",-21.11666666666667],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 9,
		 "55.53333333333333],PARAMETER[\"scale_factor\",1],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",160000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",50000],AUTHORITY[\"EPSG\",\"3727\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 12, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3728, "epsg", 3728,
		      "NAD83(NSRS2007) / Ohio North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.7 +lat_2=40.43333333333333 +lat_0=39");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-82.5 +x_0=600000 +y_0=0 +ellps=G");
    add_proj4text (p, 2, "RS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Ohio North (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 8,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 9,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",41.7]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"standard_parallel_2\",40.43333333333333],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"latitude_of_origin\",39.66666666666666],PARAME");
    add_srs_wkt (p, 12,
		 "TER[\"central_meridian\",-82.5],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 13,
		 "g\",1968500],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 14, "EPSG\",\"3728\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3729, "epsg", 3729,
		      "NAD83(NSRS2007) / Ohio South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.03333333333333 +lat_2=38.73333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=38 +lon_0=-82.5 +x_0=600000 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Ohio South (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 8,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 9,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",40.03");
    add_srs_wkt (p, 10,
		 "333333333333],PARAMETER[\"standard_parallel_2\",38.73333");
    add_srs_wkt (p, 11,
		 "333333333],PARAMETER[\"latitude_of_origin\",38],PARAMETE");
    add_srs_wkt (p, 12,
		 "R[\"central_meridian\",-82.5],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 13,
		 ",1968500],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 14, "SG\",\"3729\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3730, "epsg", 3730,
		      "NAD83(NSRS2007) / Wyoming East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-105.1666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "999375 +x_0=200000.00001016 +y_0=0 +ellps=GRS80 +towgs84");
    add_proj4text (p, 2, "=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Wyoming East (ftUS)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"US survey foot\",0.304800609601219");
    add_srs_wkt (p, 8,
		 "2,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_");
    add_srs_wkt (p, 9,
		 "Mercator\"],PARAMETER[\"latitude_of_origin\",40.5],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"central_meridian\",-105.1666666666667],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"scale_factor\",0.9999375],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 12,
		 "656166.6667],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13, "EPSG\",\"3730\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3731, "epsg", 3731,
		      "NAD83(NSRS2007) / Wyoming East Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-107.3333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "999375 +x_0=399999.99998984 +y_0=99999.99998983997 +ellp");
    add_proj4text (p, 2,
		   "s=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Wyoming East Central (ftUS)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spati");
    add_srs_wkt (p, 2,
		 "al_Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 3,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 5,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 6,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006");
    add_srs_wkt (p, 8,
		 "096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Tra");
    add_srs_wkt (p, 9,
		 "nsverse_Mercator\"],PARAMETER[\"latitude_of_origin\",40.");
    add_srs_wkt (p, 10,
		 "5],PARAMETER[\"central_meridian\",-107.3333333333333],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"scale_factor\",0.9999375],PARAMETER[\"false_ea");
    add_srs_wkt (p, 12,
		 "sting\",1312333.3333],PARAMETER[\"false_northing\",32808");
    add_srs_wkt (p, 13,
		 "3.3333],AUTHORITY[\"EPSG\",\"3731\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 14, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3732, "epsg", 3732,
		      "NAD83(NSRS2007) / Wyoming West Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-108.75 +k=0.9999375 +x_0");
    add_proj4text (p, 1,
		   "=600000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +unit");
    add_proj4text (p, 2, "s=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Wyoming West Central (ftUS)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spati");
    add_srs_wkt (p, 2,
		 "al_Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 3,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 5,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 6,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006");
    add_srs_wkt (p, 8,
		 "096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Tra");
    add_srs_wkt (p, 9,
		 "nsverse_Mercator\"],PARAMETER[\"latitude_of_origin\",40.");
    add_srs_wkt (p, 10,
		 "5],PARAMETER[\"central_meridian\",-108.75],PARAMETER[\"s");
    add_srs_wkt (p, 11,
		 "cale_factor\",0.9999375],PARAMETER[\"false_easting\",196");
    add_srs_wkt (p, 12,
		 "8500],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 13, ",\"3732\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3733, "epsg", 3733,
		      "NAD83(NSRS2007) / Wyoming West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-110.0833333333333 +k=0.9");
    add_proj4text (p, 1,
		   "999375 +x_0=800000.0000101599 +y_0=99999.99998983997 +el");
    add_proj4text (p, 2,
		   "lps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Wyoming West (ftUS)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4759\"]],UNIT[\"US survey foot\",0.304800609601219");
    add_srs_wkt (p, 8,
		 "2,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_");
    add_srs_wkt (p, 9,
		 "Mercator\"],PARAMETER[\"latitude_of_origin\",40.5],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"central_meridian\",-110.0833333333333],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"scale_factor\",0.9999375],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 12,
		 "2624666.6667],PARAMETER[\"false_northing\",328083.3333],");
    add_srs_wkt (p, 13,
		 "AUTHORITY[\"EPSG\",\"3733\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 14, ",NORTH]]");
    p = add_epsg_def (first, last, 3734, "epsg", 3734,
		      "NAD83 / Ohio North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.7 +lat_2=40.43333333333333 +lat_0=39");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-82.5 +x_0=600000 +y_0=0 +ellps=G");
    add_proj4text (p, 2, "RS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Ohio North (ftUS)\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal_C");
    add_srs_wkt (p, 8,
		 "onic_2SP\"],PARAMETER[\"standard_parallel_1\",41.7],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"standard_parallel_2\",40.43333333333333],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"latitude_of_origin\",39.66666666666666],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "central_meridian\",-82.5],PARAMETER[\"false_easting\",19");
    add_srs_wkt (p, 12,
		 "68500],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 13, ",\"3734\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3735, "epsg", 3735,
		      "NAD83 / Ohio South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.03333333333333 +lat_2=38.73333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=38 +lon_0=-82.5 +x_0=600000 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Ohio South (ftUS)\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal_C");
    add_srs_wkt (p, 8,
		 "onic_2SP\"],PARAMETER[\"standard_parallel_1\",40.0333333");
    add_srs_wkt (p, 9,
		 "3333333],PARAMETER[\"standard_parallel_2\",38.7333333333");
    add_srs_wkt (p, 10,
		 "3333],PARAMETER[\"latitude_of_origin\",38],PARAMETER[\"c");
    add_srs_wkt (p, 11,
		 "entral_meridian\",-82.5],PARAMETER[\"false_easting\",196");
    add_srs_wkt (p, 12,
		 "8500],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 13, ",\"3735\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3736, "epsg", 3736,
		      "NAD83 / Wyoming East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-105.1666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "999375 +x_0=200000.00001016 +y_0=0 +ellps=GRS80 +datum=N");
    add_proj4text (p, 2, "AD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Wyoming East (ftUS)\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercato");
    add_srs_wkt (p, 8,
		 "r\"],PARAMETER[\"latitude_of_origin\",40.5],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-105.1666666666667],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.9999375],PARAMETER[\"false_easting\",656166.");
    add_srs_wkt (p, 11,
		 "6667],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"3736\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3737, "epsg", 3737,
		      "NAD83 / Wyoming East Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-107.3333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "999375 +x_0=399999.99998984 +y_0=99999.99998983997 +ellp");
    add_proj4text (p, 2, "s=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Wyoming East Central (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"G");
    add_srs_wkt (p, 2,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "9\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transvers");
    add_srs_wkt (p, 8,
		 "e_Mercator\"],PARAMETER[\"latitude_of_origin\",40.5],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"central_meridian\",-107.3333333333333],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"scale_factor\",0.9999375],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",1312333.3333],PARAMETER[\"false_northing\",328083.3333]");
    add_srs_wkt (p, 12,
		 ",AUTHORITY[\"EPSG\",\"3737\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 3738, "epsg", 3738,
		      "NAD83 / Wyoming West Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-108.75 +k=0.9999375 +x_0");
    add_proj4text (p, 1,
		   "=600000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=us-ft +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Wyoming West Central (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"G");
    add_srs_wkt (p, 2,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "9\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transvers");
    add_srs_wkt (p, 8,
		 "e_Mercator\"],PARAMETER[\"latitude_of_origin\",40.5],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"central_meridian\",-108.75],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",0.9999375],PARAMETER[\"false_easting\",1968500],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"373");
    add_srs_wkt (p, 12, "8\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3739, "epsg", 3739,
		      "NAD83 / Wyoming West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-110.0833333333333 +k=0.9");
    add_proj4text (p, 1,
		   "999375 +x_0=800000.0000101599 +y_0=99999.99998983997 +el");
    add_proj4text (p, 2, "lps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Wyoming West (ftUS)\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercato");
    add_srs_wkt (p, 8,
		 "r\"],PARAMETER[\"latitude_of_origin\",40.5],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-110.0833333333333],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.9999375],PARAMETER[\"false_easting\",2624666");
    add_srs_wkt (p, 11,
		 ".6667],PARAMETER[\"false_northing\",328083.3333],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"3739\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]");
    add_srs_wkt (p, 13, "]");
    p = add_epsg_def (first, last, 3740, "epsg", 3740,
		      "NAD83(HARN) / UTM zone 10N");
    add_proj4text (p, 0, "+proj=utm +zone=10 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / UTM zone 10N\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-123],PARAMETER[\"scale_factor\",0.9996],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"3740\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 12, "AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3741, "epsg", 3741,
		      "NAD83(HARN) / UTM zone 11N");
    add_proj4text (p, 0, "+proj=utm +zone=11 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / UTM zone 11N\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-117],PARAMETER[\"scale_factor\",0.9996],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"3741\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 12, "AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3742, "epsg", 3742,
		      "NAD83(HARN) / UTM zone 12N");
    add_proj4text (p, 0, "+proj=utm +zone=12 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / UTM zone 12N\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-111],PARAMETER[\"scale_factor\",0.9996],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"3742\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 12, "AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3743, "epsg", 3743,
		      "NAD83(HARN) / UTM zone 13N");
    add_proj4text (p, 0, "+proj=utm +zone=13 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / UTM zone 13N\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-105],PARAMETER[\"scale_factor\",0.9996],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"3743\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 12, "AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3744, "epsg", 3744,
		      "NAD83(HARN) / UTM zone 14N");
    add_proj4text (p, 0, "+proj=utm +zone=14 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / UTM zone 14N\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-99],PARAMETER[\"scale_factor\",0.9996],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",500000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"3744\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3745, "epsg", 3745,
		      "NAD83(HARN) / UTM zone 15N");
    add_proj4text (p, 0, "+proj=utm +zone=15 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / UTM zone 15N\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-93],PARAMETER[\"scale_factor\",0.9996],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",500000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"3745\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3746, "epsg", 3746,
		      "NAD83(HARN) / UTM zone 16N");
    add_proj4text (p, 0, "+proj=utm +zone=16 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / UTM zone 16N\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-87],PARAMETER[\"scale_factor\",0.9996],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",500000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"3746\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3747, "epsg", 3747,
		      "NAD83(HARN) / UTM zone 17N");
    add_proj4text (p, 0, "+proj=utm +zone=17 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / UTM zone 17N\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-81],PARAMETER[\"scale_factor\",0.9996],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",500000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"3747\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3748, "epsg", 3748,
		      "NAD83(HARN) / UTM zone 18N");
    add_proj4text (p, 0, "+proj=utm +zone=18 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / UTM zone 18N\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-75],PARAMETER[\"scale_factor\",0.9996],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",500000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"3748\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3749, "epsg", 3749,
		      "NAD83(HARN) / UTM zone 19N");
    add_proj4text (p, 0, "+proj=utm +zone=19 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / UTM zone 19N\",GEOGCS[\"NAD83(HAR");
    add_srs_wkt (p, 1,
		 "N)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-69],PARAMETER[\"scale_factor\",0.9996],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",500000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"3749\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3750, "epsg", 3750,
		      "NAD83(HARN) / UTM zone 4N");
    add_proj4text (p, 0, "+proj=utm +zone=4 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / UTM zone 4N\",GEOGCS[\"NAD83(HARN");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-159],PARAMETER[\"scale_factor\",0.9996],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"3750\"],AXIS[\"Easting\",EAST],AX");
    add_srs_wkt (p, 12, "IS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3751, "epsg", 3751,
		      "NAD83(HARN) / UTM zone 5N");
    add_proj4text (p, 0, "+proj=utm +zone=5 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / UTM zone 5N\",GEOGCS[\"NAD83(HARN");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-153],PARAMETER[\"scale_factor\",0.9996],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"3751\"],AXIS[\"Easting\",EAST],AX");
    add_srs_wkt (p, 12, "IS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3752, "epsg", 3752,
		      "WGS 84 / Mercator 41 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=merc +lon_0=100 +lat_ts=-41 +x_0=0 +y_0=0 +ellps=W");
    add_proj4text (p, 1, "GS84 +datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / Mercator 41 (deprecated)\",GEOGCS[\"WG");
    add_srs_wkt (p, 1,
		 "S 84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,29");
    add_srs_wkt (p, 2,
		 "8.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Mercator_1SP\"],PARAMETER[\"latitude_of_origin\",-41],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"central_meridian\",100],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 9,
		 "or\",1],PARAMETER[\"false_easting\",0],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_northing\",0],AUTHORITY[\"EPSG\",\"3752\"],AXIS[\"X\",E");
    add_srs_wkt (p, 11, "AST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3753, "epsg", 3753,
		      "NAD83(HARN) / Ohio North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.7 +lat_2=40.43333333333333 +lat_0=39");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-82.5 +x_0=600000 +y_0=0 +ellps=G");
    add_proj4text (p, 2, "RS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Ohio North (ftUS)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",0.30");
    add_srs_wkt (p, 7,
		 "48006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 9,
		 "rallel_1\",41.7],PARAMETER[\"standard_parallel_2\",40.43");
    add_srs_wkt (p, 10,
		 "333333333333],PARAMETER[\"latitude_of_origin\",39.666666");
    add_srs_wkt (p, 11,
		 "66666666],PARAMETER[\"central_meridian\",-82.5],PARAMETE");
    add_srs_wkt (p, 12,
		 "R[\"false_easting\",1968500],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 13,
		 ",0],AUTHORITY[\"EPSG\",\"3753\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 14, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3754, "epsg", 3754,
		      "NAD83(HARN) / Ohio South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.03333333333333 +lat_2=38.73333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=38 +lon_0=-82.5 +x_0=600000 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Ohio South (ftUS)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",0.30");
    add_srs_wkt (p, 7,
		 "48006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 9,
		 "rallel_1\",40.03333333333333],PARAMETER[\"standard_paral");
    add_srs_wkt (p, 10,
		 "lel_2\",38.73333333333333],PARAMETER[\"latitude_of_origi");
    add_srs_wkt (p, 11,
		 "n\",38],PARAMETER[\"central_meridian\",-82.5],PARAMETER[");
    add_srs_wkt (p, 12,
		 "\"false_easting\",1968500],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 13,
		 "0],AUTHORITY[\"EPSG\",\"3754\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 14, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3755, "epsg", 3755,
		      "NAD83(HARN) / Wyoming East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-105.1666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "999375 +x_0=200000.00001016 +y_0=0 +ellps=GRS80 +units=u");
    add_proj4text (p, 2, "s-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Wyoming East (ftUS)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",");
    add_srs_wkt (p, 7,
		 "0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECT");
    add_srs_wkt (p, 8,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 9,
		 "gin\",40.5],PARAMETER[\"central_meridian\",-105.16666666");
    add_srs_wkt (p, 10,
		 "66667],PARAMETER[\"scale_factor\",0.9999375],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",656166.6667],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"3755\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3756, "epsg", 3756,
		      "NAD83(HARN) / Wyoming East Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-107.3333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "999375 +x_0=399999.99998984 +y_0=99999.99998983997 +ellp");
    add_proj4text (p, 2, "s=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Wyoming East Central (ftUS)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional");
    add_srs_wkt (p, 2,
		 "_Network\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey ");
    add_srs_wkt (p, 7,
		 "foot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",40.5],PARAMETER[\"central_meridian\",-107.3");
    add_srs_wkt (p, 10,
		 "333333333333],PARAMETER[\"scale_factor\",0.9999375],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_easting\",1312333.3333],PARAMETER[\"false_");
    add_srs_wkt (p, 12,
		 "northing\",328083.3333],AUTHORITY[\"EPSG\",\"3756\"],AXI");
    add_srs_wkt (p, 13, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3757, "epsg", 3757,
		      "NAD83(HARN) / Wyoming West Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-108.75 +k=0.9999375 +x_0");
    add_proj4text (p, 1, "=600000 +y_0=0 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Wyoming West Central (ftUS)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional");
    add_srs_wkt (p, 2,
		 "_Network\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey ");
    add_srs_wkt (p, 7,
		 "foot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",40.5],PARAMETER[\"central_meridian\",-108.7");
    add_srs_wkt (p, 10,
		 "5],PARAMETER[\"scale_factor\",0.9999375],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_easting\",1968500],PARAMETER[\"false_northing\",0],AU");
    add_srs_wkt (p, 12,
		 "THORITY[\"EPSG\",\"3757\"],AXIS[\"X\",EAST],AXIS[\"Y\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 3758, "epsg", 3758,
		      "NAD83(HARN) / Wyoming West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-110.0833333333333 +k=0.9");
    add_proj4text (p, 1,
		   "999375 +x_0=800000.0000101599 +y_0=99999.99998983997 +el");
    add_proj4text (p, 2, "lps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Wyoming West (ftUS)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",");
    add_srs_wkt (p, 7,
		 "0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECT");
    add_srs_wkt (p, 8,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 9,
		 "gin\",40.5],PARAMETER[\"central_meridian\",-110.08333333");
    add_srs_wkt (p, 10,
		 "33333],PARAMETER[\"scale_factor\",0.9999375],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",2624666.6667],PARAMETER[\"false_northing");
    add_srs_wkt (p, 12,
		 "\",328083.3333],AUTHORITY[\"EPSG\",\"3758\"],AXIS[\"X\",");
    add_srs_wkt (p, 13, "EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3759, "epsg", 3759,
		      "NAD83 / Hawaii zone 3 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=21.16666666666667 +lon_0=-158 +k=0.99");
    add_proj4text (p, 1,
		   "999 +x_0=500000.00001016 +y_0=0 +ellps=GRS80 +datum=NAD8");
    add_proj4text (p, 2, "3 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Hawaii zone 3 (ftUS)\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",21.16666666666667");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-158],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.99999],PARAMETER[\"false_easting\",1640416.6");
    add_srs_wkt (p, 11,
		 "667],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"3759\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3760, "epsg", 3760,
		      "NAD83(HARN) / Hawaii zone 3 (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=21.16666666666667 +lon_0=-158 +k=0.99");
    add_proj4text (p, 1,
		   "999 +x_0=500000.00001016 +y_0=0 +ellps=GRS80 +units=us-f");
    add_proj4text (p, 2, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Hawaii zone 3 (ftUS)\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Networ");
    add_srs_wkt (p, 2,
		 "k\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 9,
		 "igin\",21.16666666666667],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 10,
		 ",-158],PARAMETER[\"scale_factor\",0.99999],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",1640416.6667],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"3760\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 3761, "epsg", 3761,
		      "NAD83(CSRS) / UTM zone 22N");
    add_proj4text (p, 0, "+proj=utm +zone=22 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / UTM zone 22N\",GEOGCS[\"NAD83(CSR");
    add_srs_wkt (p, 1,
		 "S)\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-51],PARAMETER[\"scale_factor\",0.9996],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"3761\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 12, "AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3762, "epsg", 3762,
		      "WGS 84 / South Georgia Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-54 +lat_2=-54.75 +lat_0=-55 +lon_0=-37");
    add_proj4text (p, 1,
		   " +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=m +no_de");
    add_proj4text (p, 2, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / South Georgia Lambert\",GEOGCS[\"WGS 8");
    add_srs_wkt (p, 1,
		 "4\",DATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.2");
    add_srs_wkt (p, 2,
		 "57223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lamb");
    add_srs_wkt (p, 7,
		 "ert_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel");
    add_srs_wkt (p, 8,
		 "_1\",-54],PARAMETER[\"standard_parallel_2\",-54.75],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"latitude_of_origin\",-55],PARAMETER[\"central_me");
    add_srs_wkt (p, 10,
		 "ridian\",-37],PARAMETER[\"false_easting\",0],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3762\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3763, "epsg", 3763,
		      "ETRS89 / Portugal TM06");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=39.66825833333333 +lon_0=-8.133108333");
    add_proj4text (p, 1,
		   "333334 +k=1 +x_0=0 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / Portugal TM06\",GEOGCS[\"ETRS89\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"European_Terrestrial_Reference_System_1989\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETE");
    add_srs_wkt (p, 8,
		 "R[\"latitude_of_origin\",39.66825833333333],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-8.133108333333334],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",1],PARAMETER[\"false_easting\",0],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3763\"],AXIS[\"");
    add_srs_wkt (p, 12, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3764, "epsg", 3764,
		      "NZGD2000 / Chatham Island Circuit 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-44 +lon_0=-176.5 +k=1 +x_0=400000 +y");
    add_proj4text (p, 1,
		   "_0=800000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Chatham Island Circuit 2000\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NZGD2000\",DATUM[\"New_Zealand_Geodetic_Datum_2000\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 8,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 9,
		 "-44],PARAMETER[\"central_meridian\",-176.5],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",400000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",800000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12,
		 "3764\"],AXIS[\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 3765, "epsg", 3765, "HTRS96 / Croatia TM");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=16.5 +k=0.9999 +x_0=500000 +");
    add_proj4text (p, 1,
		   "y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_d");
    add_proj4text (p, 2, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"HTRS96 / Croatia TM\",GEOGCS[\"HTRS96\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Croatian_Terrestrial_Reference_System\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 2,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 3,
		 "],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6761\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 6,
		 "2\"]],AUTHORITY[\"EPSG\",\"4761\"]],UNIT[\"metre\",1,AUT");
    add_srs_wkt (p, 7,
		 "HORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merca");
    add_srs_wkt (p, 8,
		 "tor\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"c");
    add_srs_wkt (p, 9,
		 "entral_meridian\",16.5],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 10,
		 "9],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_northing\",0],AUTHORITY[\"EPSG\",\"3765\"],AXIS[\"Easti");
    add_srs_wkt (p, 12, "ng\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3766, "epsg", 3766, "HTRS96 / Croatia LCC");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.91666666666666 +lat_2=43.08333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=0 +lon_0=16.5 +x_0=0 +y_0=0 +ellps=GRS80 +tow");
    add_proj4text (p, 2, "gs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"HTRS96 / Croatia LCC\",GEOGCS[\"HTRS96\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"Croatian_Terrestrial_Reference_System\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6761\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4761\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conform");
    add_srs_wkt (p, 8,
		 "al_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",45.916");
    add_srs_wkt (p, 9,
		 "66666666666],PARAMETER[\"standard_parallel_2\",43.083333");
    add_srs_wkt (p, 10,
		 "33333334],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"central_meridian\",16.5],PARAMETER[\"false_easting\",0");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3");
    add_srs_wkt (p, 13,
		 "766\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3767, "epsg", 3767, "HTRS96 / UTM zone 33N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=33 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"HTRS96 / UTM zone 33N\",GEOGCS[\"HTRS96\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Croatian_Terrestrial_Reference_System\",SPHEROID[\"GR");
    add_srs_wkt (p, 2,
		 "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019");
    add_srs_wkt (p, 3,
		 "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6761\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4761\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",15],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 10,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"3767\"],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3768, "epsg", 3768, "HTRS96 / UTM zone 34N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=34 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"HTRS96 / UTM zone 34N\",GEOGCS[\"HTRS96\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Croatian_Terrestrial_Reference_System\",SPHEROID[\"GR");
    add_srs_wkt (p, 2,
		 "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019");
    add_srs_wkt (p, 3,
		 "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6761\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4761\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",21],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 10,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"3768\"],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3769, "epsg", 3769,
		      "Bermuda 1957 / UTM zone 20N");
    add_proj4text (p, 0, "+proj=utm +zone=20 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Bermuda 1957 / UTM zone 20N\",GEOGCS[\"Bermuda ");
    add_srs_wkt (p, 1,
		 "1957\",DATUM[\"Bermuda_1957\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6216\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4216\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-63],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"3769\"],AXIS[\"Easting\",EAST],AXIS[\"Northin");
    add_srs_wkt (p, 12, "g\",NORTH]]");
    p = add_epsg_def (first, last, 3770, "epsg", 3770,
		      "BDA2000 / Bermuda 2000 National Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=32 +lon_0=-64.75 +k=1 +x_0=550000 +y_");
    add_proj4text (p, 1,
		   "0=100000 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=m +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"BDA2000 / Bermuda 2000 National Grid\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "BDA2000\",DATUM[\"Bermuda_2000\",SPHEROID[\"WGS 84\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84");
    add_srs_wkt (p, 3,
		 "[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6762\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4762\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",32],PARAMETER[\"central_m");
    add_srs_wkt (p, 9,
		 "eridian\",-64.75],PARAMETER[\"scale_factor\",1],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_easting\",550000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",100000],AUTHORITY[\"EPSG\",\"3770\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3771, "epsg", 3771,
		      "NAD27 / Alberta 3TM ref merid 111 W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-111 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=clrk66 +datum=NAD27 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Alberta 3TM ref merid 111 W\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD27\",DATUM[\"North_American_Datum_1927\",SPHEROID[\"C");
    add_srs_wkt (p, 2,
		 "larke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-111],PARAMETER[\"scale_factor\",0.9999],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_easting\",0],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 11,
		 "UTHORITY[\"EPSG\",\"3771\"],AXIS[\"Easting\",EAST],AXIS[");
    add_srs_wkt (p, 12, "\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3772, "epsg", 3772,
		      "NAD27 / Alberta 3TM ref merid 114 W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-114 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=clrk66 +datum=NAD27 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Alberta 3TM ref merid 114 W\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD27\",DATUM[\"North_American_Datum_1927\",SPHEROID[\"C");
    add_srs_wkt (p, 2,
		 "larke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-114],PARAMETER[\"scale_factor\",0.9999],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_easting\",0],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 11,
		 "UTHORITY[\"EPSG\",\"3772\"],AXIS[\"Easting\",EAST],AXIS[");
    add_srs_wkt (p, 12, "\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3773, "epsg", 3773,
		      "NAD27 / Alberta 3TM ref merid 117 W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-117 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=clrk66 +datum=NAD27 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Alberta 3TM ref merid 117 W\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD27\",DATUM[\"North_American_Datum_1927\",SPHEROID[\"C");
    add_srs_wkt (p, 2,
		 "larke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-117],PARAMETER[\"scale_factor\",0.9999],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_easting\",0],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 11,
		 "UTHORITY[\"EPSG\",\"3773\"],AXIS[\"Easting\",EAST],AXIS[");
    add_srs_wkt (p, 12, "\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3774, "epsg", 3774,
		      "NAD27 / Alberta 3TM ref merid 120 W (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-120 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=clrk66 +datum=NAD27 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Alberta 3TM ref merid 120 W (deprecated");
    add_srs_wkt (p, 1,
		 ")\",GEOGCS[\"NAD27\",DATUM[\"North_American_Datum_1927\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982139006,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6267\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4267\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-120],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 10,
		 "9999],PARAMETER[\"false_easting\",0],PARAMETER[\"false_n");
    add_srs_wkt (p, 11,
		 "orthing\",0],AUTHORITY[\"EPSG\",\"3774\"],AXIS[\"Easting");
    add_srs_wkt (p, 12, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3775, "epsg", 3775,
		      "NAD83 / Alberta 3TM ref merid 111 W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-111 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alberta 3TM ref merid 111 W\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"G");
    add_srs_wkt (p, 2,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "9\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 9,
		 "111],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",0],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3775\"],AXIS[\"Easting\",EAST],AXIS[\"Northi");
    add_srs_wkt (p, 12, "ng\",NORTH]]");
    p = add_epsg_def (first, last, 3776, "epsg", 3776,
		      "NAD83 / Alberta 3TM ref merid 114 W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-114 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alberta 3TM ref merid 114 W\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"G");
    add_srs_wkt (p, 2,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "9\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 9,
		 "114],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",0],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3776\"],AXIS[\"Easting\",EAST],AXIS[\"Northi");
    add_srs_wkt (p, 12, "ng\",NORTH]]");
    p = add_epsg_def (first, last, 3777, "epsg", 3777,
		      "NAD83 / Alberta 3TM ref merid 117 W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-117 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alberta 3TM ref merid 117 W\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"G");
    add_srs_wkt (p, 2,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "9\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 9,
		 "117],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",0],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3777\"],AXIS[\"Easting\",EAST],AXIS[\"Northi");
    add_srs_wkt (p, 12, "ng\",NORTH]]");
    p = add_epsg_def (first, last, 3778, "epsg", 3778,
		      "NAD83 / Alberta 3TM ref merid 120 W (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-120 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alberta 3TM ref merid 120 W (deprecated");
    add_srs_wkt (p, 1,
		 ")\",GEOGCS[\"NAD83\",DATUM[\"North_American_Datum_1983\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-120],PARAMETER[\"scale_factor\",0.9999],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",0],PARAMETER[\"false_northing\",0]");
    add_srs_wkt (p, 11,
		 ",AUTHORITY[\"EPSG\",\"3778\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 12, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3779, "epsg", 3779,
		      "NAD83(CSRS) / Alberta 3TM ref merid 111 W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-111 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / Alberta 3TM ref merid 111 W\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(CSRS)\",DATUM[\"NAD83_Canadian_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"614");
    add_srs_wkt (p, 4,
		 "0\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-111],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9999],PARAMETER[\"false_easting\",0],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"3779\"],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3780, "epsg", 3780,
		      "NAD83(CSRS) / Alberta 3TM ref merid 114 W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-114 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / Alberta 3TM ref merid 114 W\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(CSRS)\",DATUM[\"NAD83_Canadian_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"614");
    add_srs_wkt (p, 4,
		 "0\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-114],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9999],PARAMETER[\"false_easting\",0],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"3780\"],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3781, "epsg", 3781,
		      "NAD83(CSRS) / Alberta 3TM ref merid 117 W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-117 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / Alberta 3TM ref merid 117 W\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(CSRS)\",DATUM[\"NAD83_Canadian_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"614");
    add_srs_wkt (p, 4,
		 "0\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-117],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9999],PARAMETER[\"false_easting\",0],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"3781\"],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3782, "epsg", 3782,
		      "NAD83(CSRS) / Alberta 3TM ref merid 120 W (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-120 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / Alberta 3TM ref merid 120 W (depr");
    add_srs_wkt (p, 1,
		 "ecated)\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83_Canadian_");
    add_srs_wkt (p, 2,
		 "Spatial_Reference_System\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 3,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4617\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",-120],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",0.9999],PARAMETER[\"false_easting\",0],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3782\"");
    add_srs_wkt (p, 12, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3783, "epsg", 3783,
		      "Pitcairn 2006 / Pitcairn TM 2006");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-25.06855261111111 +lon_0=-130.112967");
    add_proj4text (p, 1,
		   "1111111 +k=1 +x_0=14200 +y_0=15500 +ellps=WGS84 +towgs84");
    add_proj4text (p, 2, "=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pitcairn 2006 / Pitcairn TM 2006\",GEOGCS[\"Pit");
    add_srs_wkt (p, 1,
		 "cairn 2006\",DATUM[\"Pitcairn_2006\",SPHEROID[\"WGS 84\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOW");
    add_srs_wkt (p, 3,
		 "GS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6763\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4763\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",-25.06855261111111],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"central_meridian\",-130.1129671111111],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",1420");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",15500],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12,
		 ",\"3783\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORT");
    add_srs_wkt (p, 13, "H]]");
    p = add_epsg_def (first, last, 3784, "epsg", 3784,
		      "Pitcairn 1967 / UTM zone 9S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=9 +south +ellps=intl +towgs84=185,165,42");
    add_proj4text (p, 1, ",0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pitcairn 1967 / UTM zone 9S\",GEOGCS[\"Pitcairn");
    add_srs_wkt (p, 1,
		 " 1967\",DATUM[\"Pitcairn_1967\",SPHEROID[\"International");
    add_srs_wkt (p, 2,
		 " 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS8");
    add_srs_wkt (p, 3,
		 "4[185,165,42,0,0,0,0],AUTHORITY[\"EPSG\",\"6729\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4729\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",-129],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 11,
		 "rthing\",10000000],AUTHORITY[\"EPSG\",\"3784\"],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3785, "epsg", 3785,
		      "Popular Visualisation CRS / Mercator (deprecated)");
    add_proj4text (p, 0,
		   "+proj=merc +a=6378137 +b=6378137 +lat_ts=0.0 +lon_0=0.0 ");
    add_proj4text (p, 1,
		   "+x_0=0.0 +y_0=0 +k=1.0 +units=m +nadgrids=@null +wktext ");
    add_proj4text (p, 2, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Popular Visualisation CRS / Mercator (deprecate");
    add_srs_wkt (p, 1,
		 "d)\",GEOGCS[\"Popular Visualisation CRS\",DATUM[\"Popula");
    add_srs_wkt (p, 2,
		 "r_Visualisation_Datum\",SPHEROID[\"Popular Visualisation");
    add_srs_wkt (p, 3,
		 " Sphere\",6378137,0,AUTHORITY[\"EPSG\",\"7059\"]],TOWGS8");
    add_srs_wkt (p, 4,
		 "4[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6055\"]],PRIMEM[\"");
    add_srs_wkt (p, 5,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 6,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"4055\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 8,
		 "EPSG\",\"9001\"]],PROJECTION[\"Mercator_1SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"central_meridian\",0],PARAMETER[\"scale_factor\",1],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",0],PARAMETER[\"false_northing");
    add_srs_wkt (p, 11,
		 "\",0],EXTENSION[\"PROJ4\",\"+proj=merc +a=6378137 +b=637");
    add_srs_wkt (p, 12,
		 "8137 +lat_ts=0.0 +lon_0=0.0 +x_0=0.0 +y_0=0 +k=1.0 +unit");
    add_srs_wkt (p, 13,
		 "s=m +nadgrids=@null +wktext  +no_defs\"],AUTHORITY[\"EPS");
    add_srs_wkt (p, 14, "G\",\"3785\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3786, "epsg", 3786,
		      "World Equidistant Cylindrical (Sphere) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=eqc +lat_ts=0 +lat_0=0 +lon_0=0 +x_0=0 +y_0=0 +a=6");
    add_proj4text (p, 1, "371007 +b=6371007 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"World Equidistant Cylindrical (Sphere) (depreca");
    add_srs_wkt (p, 1,
		 "ted)\",GEOGCS[\"Unspecified datum based upon the GRS 198");
    add_srs_wkt (p, 2,
		 "0 Authalic Sphere\",DATUM[\"Not_specified_based_on_GRS_1");
    add_srs_wkt (p, 3,
		 "980_Authalic_Sphere\",SPHEROID[\"GRS 1980 Authalic Spher");
    add_srs_wkt (p, 4,
		 "e\",6371007,0,AUTHORITY[\"EPSG\",\"7048\"]],AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"6047\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 6,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4047\"]],U");
    add_srs_wkt (p, 8,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 9,
		 "[\"Equirectangular\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"central_meridian\",0],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12, "PSG\",\"3786\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3787, "epsg", 3787,
		      "MGI / Slovene National Grid (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=-5000000 +ellps=bessel +datum=hermannskogel +units=m +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / Slovene National Grid (deprecated)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"MGI\",DATUM[\"Militar_Geographische_Institute\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7004\"]],TOWGS84[577.326,90.129,463.919,5.137");
    add_srs_wkt (p, 4,
		 ",1.474,5.297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIME");
    add_srs_wkt (p, 5,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 6,
		 "egree\",0.0174532925199433,AUTHORITY[\"EPSG\",\"9108\"]]");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"4312\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 8,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 10,
		 "l_meridian\",15],PARAMETER[\"scale_factor\",0.9999],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_easting\",500000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 12,
		 "ng\",-5000000],AUTHORITY[\"EPSG\",\"3787\"],AXIS[\"Y\",E");
    add_srs_wkt (p, 13, "AST],AXIS[\"X\",NORTH]]");
    p = add_epsg_def (first, last, 3788, "epsg", 3788,
		      "NZGD2000 / Auckland Islands TM 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=166 +k=1 +x_0=3500000 +y_0=1");
    add_proj4text (p, 1,
		   "0000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no");
    add_proj4text (p, 2, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Auckland Islands TM 2000\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NZGD2000\",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"central_meridian\",166],PARAMETER[\"scale_fac");
    add_srs_wkt (p, 10,
		 "tor\",1],PARAMETER[\"false_easting\",3500000],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"3788\"");
    add_srs_wkt (p, 12, "],AXIS[\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 3789, "epsg", 3789,
		      "NZGD2000 / Campbell Island TM 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=169 +k=1 +x_0=3500000 +y_0=1");
    add_proj4text (p, 1,
		   "0000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no");
    add_proj4text (p, 2, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Campbell Island TM 2000\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "ZGD2000\",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"central_meridian\",169],PARAMETER[\"scale_fac");
    add_srs_wkt (p, 10,
		 "tor\",1],PARAMETER[\"false_easting\",3500000],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"3789\"");
    add_srs_wkt (p, 12, "],AXIS[\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 3790, "epsg", 3790,
		      "NZGD2000 / Antipodes Islands TM 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=179 +k=1 +x_0=3500000 +y_0=1");
    add_proj4text (p, 1,
		   "0000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no");
    add_proj4text (p, 2, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Antipodes Islands TM 2000\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NZGD2000\",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"central_meridian\",179],PARAMETER[\"scale_fac");
    add_srs_wkt (p, 10,
		 "tor\",1],PARAMETER[\"false_easting\",3500000],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"3790\"");
    add_srs_wkt (p, 12, "],AXIS[\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 3791, "epsg", 3791,
		      "NZGD2000 / Raoul Island TM 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-178 +k=1 +x_0=3500000 +y_0=");
    add_proj4text (p, 1,
		   "10000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Raoul Island TM 2000\",GEOGCS[\"NZGD");
    add_srs_wkt (p, 1,
		 "2000\",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 4,
		 "167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 5,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 6,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metr");
    add_srs_wkt (p, 7,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transve");
    add_srs_wkt (p, 8,
		 "rse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"central_meridian\",-178],PARAMETER[\"scale_facto");
    add_srs_wkt (p, 10,
		 "r\",1],PARAMETER[\"false_easting\",3500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"3791\"],");
    add_srs_wkt (p, 12, "AXIS[\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 3793, "epsg", 3793,
		      "NZGD2000 / Chatham Islands TM 2000");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-176.5 +k=1 +x_0=3500000 +y_");
    add_proj4text (p, 1,
		   "0=10000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / Chatham Islands TM 2000\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "ZGD2000\",DATUM[\"New_Zealand_Geodetic_Datum_2000\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"central_meridian\",-176.5],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",1],PARAMETER[\"false_easting\",3500000],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"379");
    add_srs_wkt (p, 12,
		 "3\"],AXIS[\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 3794, "epsg", 3794,
		      "Slovenia 1996 / Slovene National Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=-5000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Slovenia 1996 / Slovene National Grid\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"Slovenia 1996\",DATUM[\"Slovenia_Geodetic_Datum_1996\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"6765\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4765\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 8,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 9,
		 "0],PARAMETER[\"central_meridian\",15],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.9999],PARAMETER[\"false_easting\",500000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",-5000000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12,
		 "3794\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 3795, "epsg", 3795, "NAD27 / Cuba Norte");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=23 +lat_2=21.7 +lat_0=22.35 +lon_0=-81 ");
    add_proj4text (p, 1,
		   "+x_0=500000 +y_0=280296.016 +ellps=clrk66 +datum=NAD27 +");
    add_proj4text (p, 2, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Cuba Norte\",GEOGCS[\"NAD27\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6378");
    add_srs_wkt (p, 2,
		 "206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "standard_parallel_1\",23],PARAMETER[\"standard_parallel_");
    add_srs_wkt (p, 9,
		 "2\",21.7],PARAMETER[\"latitude_of_origin\",22.35],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"central_meridian\",-81],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",280296.016],AUTHOR");
    add_srs_wkt (p, 12,
		 "ITY[\"EPSG\",\"3795\"],AXIS[\"Y\",NORTH],AXIS[\"X\",EAST");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 3796, "epsg", 3796, "NAD27 / Cuba Sur");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=21.3 +lat_2=20.13333333333333 +lat_0=20");
    add_proj4text (p, 1,
		   ".71666666666667 +lon_0=-76.83333333333333 +x_0=500000 +y");
    add_proj4text (p, 2,
		   "_0=229126.939 +ellps=clrk66 +datum=NAD27 +units=m +no_de");
    add_proj4text (p, 3, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Cuba Sur\",GEOGCS[\"NAD27\",DATUM[\"Nor");
    add_srs_wkt (p, 1,
		 "th_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637820");
    add_srs_wkt (p, 2,
		 "6.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "67\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 7,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 8,
		 "andard_parallel_1\",21.3],PARAMETER[\"standard_parallel_");
    add_srs_wkt (p, 9,
		 "2\",20.13333333333333],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 10,
		 "20.71666666666667],PARAMETER[\"central_meridian\",-76.83");
    add_srs_wkt (p, 11,
		 "333333333333],PARAMETER[\"false_easting\",500000],PARAME");
    add_srs_wkt (p, 12,
		 "TER[\"false_northing\",229126.939],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 13, "3796\"],AXIS[\"Y\",NORTH],AXIS[\"X\",EAST]]");
    p = add_epsg_def (first, last, 3797, "epsg", 3797, "NAD27 / MTQ Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=50 +lat_2=46 +lat_0=44 +lon_0=-70 +x_0=");
    add_proj4text (p, 1,
		   "800000 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=m +no_de");
    add_proj4text (p, 2, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / MTQ Lambert\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "standard_parallel_1\",50],PARAMETER[\"standard_parallel_");
    add_srs_wkt (p, 9,
		 "2\",46],PARAMETER[\"latitude_of_origin\",44],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "central_meridian\",-70],PARAMETER[\"false_easting\",8000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "3797\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3798, "epsg", 3798, "NAD83 / MTQ Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=50 +lat_2=46 +lat_0=44 +lon_0=-70 +x_0=");
    add_proj4text (p, 1,
		   "800000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / MTQ Lambert\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard");
    add_srs_wkt (p, 8,
		 "_parallel_1\",50],PARAMETER[\"standard_parallel_2\",46],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"latitude_of_origin\",44],PARAMETER[\"central");
    add_srs_wkt (p, 10,
		 "_meridian\",-70],PARAMETER[\"false_easting\",800000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3798\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3799, "epsg", 3799,
		      "NAD83(CSRS) / MTQ Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=50 +lat_2=46 +lat_0=44 +lon_0=-70 +x_0=");
    add_proj4text (p, 1, "800000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / MTQ Lambert\",GEOGCS[\"NAD83(CSRS");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",50],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_2\",46],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 10,
		 "\",44],PARAMETER[\"central_meridian\",-70],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",800000],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 12,
		 "UTHORITY[\"EPSG\",\"3799\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 13, "NORTH]]");
    p = add_epsg_def (first, last, 3800, "epsg", 3800,
		      "NAD27 / Alberta 3TM ref merid 120 W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-120 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=clrk66 +datum=NAD27 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Alberta 3TM ref merid 120 W\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD27\",DATUM[\"North_American_Datum_1927\",SPHEROID[\"C");
    add_srs_wkt (p, 2,
		 "larke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-120],PARAMETER[\"scale_factor\",0.9999],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_easting\",0],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 11,
		 "UTHORITY[\"EPSG\",\"3800\"],AXIS[\"Easting\",EAST],AXIS[");
    add_srs_wkt (p, 12, "\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3801, "epsg", 3801,
		      "NAD83 / Alberta 3TM ref merid 120 W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-120 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alberta 3TM ref merid 120 W\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"G");
    add_srs_wkt (p, 2,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "9\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 9,
		 "120],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",0],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3801\"],AXIS[\"Easting\",EAST],AXIS[\"Northi");
    add_srs_wkt (p, 12, "ng\",NORTH]]");
    p = add_epsg_def (first, last, 3802, "epsg", 3802,
		      "NAD83(CSRS) / Alberta 3TM ref merid 120 W");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-120 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / Alberta 3TM ref merid 120 W\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"NAD83(CSRS)\",DATUM[\"NAD83_Canadian_Spatial_Refer");
    add_srs_wkt (p, 2,
		 "ence_System\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"614");
    add_srs_wkt (p, 4,
		 "0\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-120],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9999],PARAMETER[\"false_easting\",0],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"3802\"],AXIS[\"East");
    add_srs_wkt (p, 12, "ing\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3812, "epsg", 3812,
		      "ETRS89 / Belgian Lambert 2008");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49.83333333333334 +lat_2=51.16666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=50.797815 +lon_0=4.359215833333333 +x_0=64932");
    add_proj4text (p, 2, "8 +y_0=665262 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / Belgian Lambert 2008\",GEOGCS[\"ETRS89");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"European_Terrestrial_Reference_System_1989\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",49.8333333333333");
    add_srs_wkt (p, 9,
		 "4],PARAMETER[\"standard_parallel_2\",51.16666666666666],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",50.797815],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "central_meridian\",4.359215833333333],PARAMETER[\"false_");
    add_srs_wkt (p, 12,
		 "easting\",649328],PARAMETER[\"false_northing\",665262],A");
    add_srs_wkt (p, 13,
		 "UTHORITY[\"EPSG\",\"3812\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 14, "NORTH]]");
    p = add_epsg_def (first, last, 3814, "epsg", 3814,
		      "NAD83 / Mississippi TM");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=32.5 +lon_0=-89.75 +k=0.9998335 +x_0=");
    add_proj4text (p, 1,
		   "500000 +y_0=1300000 +ellps=GRS80 +datum=NAD83 +units=m +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Mississippi TM\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 8,
		 "igin\",32.5],PARAMETER[\"central_meridian\",-89.75],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9998335],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",1300000],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"3814\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 12, "TH]]");
    p = add_epsg_def (first, last, 3815, "epsg", 3815,
		      "NAD83(HARN) / Mississippi TM");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=32.5 +lon_0=-89.75 +k=0.9998335 +x_0=");
    add_proj4text (p, 1, "500000 +y_0=1300000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Mississippi TM\",GEOGCS[\"NAD83(H");
    add_srs_wkt (p, 1,
		 "ARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",32.5],PARAMETER[\"central");
    add_srs_wkt (p, 9,
		 "_meridian\",-89.75],PARAMETER[\"scale_factor\",0.9998335");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"false_easting\",500000],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "northing\",1300000],AUTHORITY[\"EPSG\",\"3815\"],AXIS[\"");
    add_srs_wkt (p, 12, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3816, "epsg", 3816,
		      "NAD83(NSRS2007) / Mississippi TM");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=32.5 +lon_0=-89.75 +k=0.9998335 +x_0=");
    add_proj4text (p, 1,
		   "500000 +y_0=1300000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Mississippi TM\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference_");
    add_srs_wkt (p, 2,
		 "System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",32.5],PARAMETER[\"central_meridian\",-89.75");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9998335],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",500000],PARAMETER[\"false_northing\",1300000");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"3816\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 13, "\",NORTH]]");
    p = add_epsg_def (first, last, 3819, "epsg", 3819, "HD1909");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=bessel +towgs84=595.48,121.69,515.3");
    add_proj4text (p, 1, "5,4.115,-2.9383,0.853,-3.408 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"HD1909\",DATUM[\"Hungarian_Datum_1909\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EP");
    add_srs_wkt (p, 2,
		 "SG\",\"7004\"]],TOWGS84[595.48,121.69,515.35,4.115,-2.93");
    add_srs_wkt (p, 3,
		 "83,0.853,-3.408],AUTHORITY[\"EPSG\",\"1024\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6, "THORITY[\"EPSG\",\"3819\"]]");
    p = add_epsg_def (first, last, 3821, "epsg", 3821, "TWD67");
    add_proj4text (p, 0, "+proj=longlat +ellps=aust_SA +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"TWD67\",DATUM[\"Taiwan_Datum_1967\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7050\"]],AUTHORITY[\"EPSG\",\"1025\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 3,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 4,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 5, "\"EPSG\",\"3821\"]]");
    p = add_epsg_def (first, last, 3824, "epsg", 3824, "TWD97");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"TWD97\",DATUM[\"Taiwan_Datum_1997\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 2,
		 "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1026\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5, "9122\"]],AUTHORITY[\"EPSG\",\"3824\"]]");
    p = add_epsg_def (first, last, 3825, "epsg", 3825, "TWD97 / TM2 zone 119");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=119 +k=0.9999 +x_0=250000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_de");
    add_proj4text (p, 2, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"TWD97 / TM2 zone 119\",GEOGCS[\"TWD97\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Taiwan_Datum_1997\",SPHEROID[\"GRS 1980\",6378137,298.25");
    add_srs_wkt (p, 2,
		 "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,");
    add_srs_wkt (p, 3,
		 "0,0],AUTHORITY[\"EPSG\",\"1026\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"3824\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",11");
    add_srs_wkt (p, 9,
		 "9],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",250000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"3825\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH");
    add_srs_wkt (p, 12, "]]");
    p = add_epsg_def (first, last, 3826, "epsg", 3826, "TWD97 / TM2 zone 121");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=121 +k=0.9999 +x_0=250000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_de");
    add_proj4text (p, 2, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"TWD97 / TM2 zone 121\",GEOGCS[\"TWD97\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Taiwan_Datum_1997\",SPHEROID[\"GRS 1980\",6378137,298.25");
    add_srs_wkt (p, 2,
		 "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,");
    add_srs_wkt (p, 3,
		 "0,0],AUTHORITY[\"EPSG\",\"1026\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"3824\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",12");
    add_srs_wkt (p, 9,
		 "1],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",250000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"3826\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH");
    add_srs_wkt (p, 12, "]]");
    p = add_epsg_def (first, last, 3827, "epsg", 3827, "TWD67 / TM2 zone 119");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=119 +k=0.9999 +x_0=250000 +y");
    add_proj4text (p, 1, "_0=0 +ellps=aust_SA +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"TWD67 / TM2 zone 119\",GEOGCS[\"TWD67\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Taiwan_Datum_1967\",SPHEROID[\"GRS 1967 Modified\",63781");
    add_srs_wkt (p, 2,
		 "60,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"1025\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"3821\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 7,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"central_meridian\",119],PARAMETER[\"scale_");
    add_srs_wkt (p, 9,
		 "factor\",0.9999],PARAMETER[\"false_easting\",250000],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3827\"");
    add_srs_wkt (p, 11, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3828, "epsg", 3828, "TWD67 / TM2 zone 121");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=121 +k=0.9999 +x_0=250000 +y");
    add_proj4text (p, 1, "_0=0 +ellps=aust_SA +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"TWD67 / TM2 zone 121\",GEOGCS[\"TWD67\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Taiwan_Datum_1967\",SPHEROID[\"GRS 1967 Modified\",63781");
    add_srs_wkt (p, 2,
		 "60,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"1025\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"3821\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 7,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"central_meridian\",121],PARAMETER[\"scale_");
    add_srs_wkt (p, 9,
		 "factor\",0.9999],PARAMETER[\"false_easting\",250000],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3828\"");
    add_srs_wkt (p, 11, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3829, "epsg", 3829,
		      "Hu Tzu Shan / UTM zone 51N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=51 +ellps=intl +towgs84=-637,-549,-203,0");
    add_proj4text (p, 1, ",0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Hu Tzu Shan / UTM zone 51N\",GEOGCS[\"Hu Tzu Sh");
    add_srs_wkt (p, 1,
		 "an 1950\",DATUM[\"Hu_Tzu_Shan_1950\",SPHEROID[\"Internat");
    add_srs_wkt (p, 2,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[-637,-549,-203,0,0,0,0],AUTHORITY[\"EPSG\",\"6236");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4236\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",123],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3829\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3832, "epsg", 3832, "WGS 84 / PDC Mercator");
    add_proj4text (p, 0,
		   "+proj=merc +lon_0=150 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 1, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / PDC Mercator\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Mercator_1SP\"");
    add_srs_wkt (p, 7,
		 "],PARAMETER[\"central_meridian\",150],PARAMETER[\"scale_");
    add_srs_wkt (p, 8,
		 "factor\",1],PARAMETER[\"false_easting\",0],PARAMETER[\"f");
    add_srs_wkt (p, 9,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"3832\"],AXIS[\"E");
    add_srs_wkt (p, 10, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3833, "epsg", 3833,
		      "Pulkovo 1942(58) / Gauss-Kruger zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=9 +k=1 +x_0=2500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / Gauss-Kruger zone 2\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo_1942_58\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7024\"]],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4179\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",9],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",2500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3833\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3834, "epsg", 3834,
		      "Pulkovo 1942(83) / Gauss-Kruger zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=9 +k=1 +x_0=2500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(83) / Gauss-Kruger zone 2\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo_1942_83\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7024\"]],AUTHORITY[\"EPSG\",\"6178\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4178\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",9],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",2500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"3834\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3835, "epsg", 3835,
		      "Pulkovo 1942(83) / Gauss-Kruger zone 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=1 +x_0=3500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(83) / Gauss-Kruger zone 3\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo_1942_83\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7024\"]],AUTHORITY[\"EPSG\",\"6178\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4178\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",15],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",3500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"3835\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 3836, "epsg", 3836,
		      "Pulkovo 1942(83) / Gauss-Kruger zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=1 +x_0=4500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(83) / Gauss-Kruger zone 4\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo_1942_83\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7024\"]],AUTHORITY[\"EPSG\",\"6178\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4178\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",21],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",4500000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"3836\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 3837, "epsg", 3837,
		      "Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=9 +k=1 +x_0=3500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 3");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo_1942_58\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4179\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",9],PARAMETER[\"scale_factor\",1],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_easting\",3500000],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 11,
		 "UTHORITY[\"EPSG\",\"3837\"],AXIS[\"X\",NORTH],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",EAST]]");
    p = add_epsg_def (first, last, 3838, "epsg", 3838,
		      "Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=12 +k=1 +x_0=4500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 4");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo_1942_58\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4179\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",12],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",4500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"3838\"],AXIS[\"X\",NORTH],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",EAST]]");
    p = add_epsg_def (first, last, 3839, "epsg", 3839,
		      "Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 9");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=27 +k=1 +x_0=9500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 9");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo_1942_58\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4179\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",27],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",9500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"3839\"],AXIS[\"X\",NORTH],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",EAST]]");
    p = add_epsg_def (first, last, 3840, "epsg", 3840,
		      "Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 10");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=30 +k=1 +x_0=10500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 1");
    add_srs_wkt (p, 1,
		 "0\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo_1942_58\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4179\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",30],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",10500000],PARAMETER[\"false_northing\",0]");
    add_srs_wkt (p, 11,
		 ",AUTHORITY[\"EPSG\",\"3840\"],AXIS[\"X\",NORTH],AXIS[\"Y");
    add_srs_wkt (p, 12, "\",EAST]]");
    p = add_epsg_def (first, last, 3841, "epsg", 3841,
		      "Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18 +k=1 +x_0=6500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 6");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo_1942_83\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6178\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4178\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",18],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",6500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"3841\"],AXIS[\"X\",NORTH],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",EAST]]");
    p = add_epsg_def (first, last, 3842, "epsg", 3842,
		      "Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 7");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18 +k=1 +x_0=6500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 7");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo_1942_83\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6178\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4178\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",18],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",6500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"3842\"],AXIS[\"X\",NORTH],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",EAST]]");
    p = add_epsg_def (first, last, 3843, "epsg", 3843,
		      "Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 8");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18 +k=1 +x_0=6500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 8");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo_1942_83\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6178\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4178\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",18],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",6500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"3843\"],AXIS[\"X\",NORTH],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",EAST]]");
    p = add_epsg_def (first, last, 3844, "epsg", 3844,
		      "Pulkovo 1942(58) / Stereo70");
    add_proj4text (p, 0,
		   "+proj=sterea +lat_0=46 +lon_0=25 +k=0.99975 +x_0=500000 ");
    add_proj4text (p, 1, "+y_0=500000 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942(58) / Stereo70\",GEOGCS[\"Pulkovo ");
    add_srs_wkt (p, 1,
		 "1942(58)\",DATUM[\"Pulkovo_1942_58\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4179\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Oblique_Stereographic\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",46],PARAMETER[\"central_meridian\",25],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"scale_factor\",0.99975],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",500000],AUTHO");
    add_srs_wkt (p, 11,
		 "RITY[\"EPSG\",\"3844\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAS");
    add_srs_wkt (p, 12, "T]]");
    p = add_epsg_def (first, last, 3845, "epsg", 3845,
		      "SWEREF99 / RT90 7.5 gon V emulation");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=11.30625 +k=1.000006 +x_0=15");
    add_proj4text (p, 1,
		   "00025.141 +y_0=-667.282 +ellps=GRS80 +towgs84=0,0,0,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SWEREF99 / RT90 7.5 gon V emulation\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "SWEREF99\",DATUM[\"SWEREF99\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6619\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4619\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",11.30625],PARAMETER[\"scale_factor\",1.000006],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_easting\",1500025.141],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_northing\",-667.282],AUTHORITY[\"EPSG\",\"3845\"],AXIS[");
    add_srs_wkt (p, 12, "\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3846, "epsg", 3846,
		      "SWEREF99 / RT90 5 gon V emulation");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=13.55626666666667 +k=1.00000");
    add_proj4text (p, 1,
		   "58 +x_0=1500044.695 +y_0=-667.13 +ellps=GRS80 +towgs84=0");
    add_proj4text (p, 2, ",0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SWEREF99 / RT90 5 gon V emulation\",GEOGCS[\"SW");
    add_srs_wkt (p, 1,
		 "EREF99\",DATUM[\"SWEREF99\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6619\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 4,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 5,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"4619\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"latitude_of_origin\",0],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",13.55626666666667],PARAMETER[\"scale_factor\",1.00");
    add_srs_wkt (p, 10,
		 "00058],PARAMETER[\"false_easting\",1500044.695],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_northing\",-667.13],AUTHORITY[\"EPSG\",\"3846\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3847, "epsg", 3847,
		      "SWEREF99 / RT90 2.5 gon V emulation");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15.80628452944445 +k=1.00000");
    add_proj4text (p, 1,
		   "561024 +x_0=1500064.274 +y_0=-667.711 +ellps=GRS80 +towg");
    add_proj4text (p, 2, "s84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SWEREF99 / RT90 2.5 gon V emulation\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "SWEREF99\",DATUM[\"SWEREF99\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6619\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4619\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",15.80628452944445],PARAMETER[\"scale_factor\",1.0");
    add_srs_wkt (p, 10,
		 "0000561024],PARAMETER[\"false_easting\",1500064.274],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",-667.711],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "3847\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3848, "epsg", 3848,
		      "SWEREF99 / RT90 0 gon emulation");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18.0563 +k=1.0000054 +x_0=15");
    add_proj4text (p, 1,
		   "00083.521 +y_0=-668.8440000000001 +ellps=GRS80 +towgs84=");
    add_proj4text (p, 2, "0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SWEREF99 / RT90 0 gon emulation\",GEOGCS[\"SWER");
    add_srs_wkt (p, 1,
		 "EF99\",DATUM[\"SWEREF99\",SPHEROID[\"GRS 1980\",6378137,");
    add_srs_wkt (p, 2,
		 "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0],AUTHORITY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"4619\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETE");
    add_srs_wkt (p, 8,
		 "R[\"latitude_of_origin\",0],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 9,
		 "\",18.0563],PARAMETER[\"scale_factor\",1.0000054],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",1500083.521],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",-668.844],AUTHORITY[\"EPSG\",\"3848\"],AXIS[\"X\"");
    add_srs_wkt (p, 12, ",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3849, "epsg", 3849,
		      "SWEREF99 / RT90 2.5 gon O emulation");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=20.30631666666667 +k=1.00000");
    add_proj4text (p, 1,
		   "52 +x_0=1500102.765 +y_0=-670.706 +ellps=GRS80 +towgs84=");
    add_proj4text (p, 2, "0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SWEREF99 / RT90 2.5 gon O emulation\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "SWEREF99\",DATUM[\"SWEREF99\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6619\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4619\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",20.30631666666667],PARAMETER[\"scale_factor\",1.0");
    add_srs_wkt (p, 10,
		 "000052],PARAMETER[\"false_easting\",1500102.765],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",-670.706],AUTHORITY[\"EPSG\",\"384");
    add_srs_wkt (p, 12, "9\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3850, "epsg", 3850,
		      "SWEREF99 / RT90 5 gon O emulation");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=22.55633333333333 +k=1.00000");
    add_proj4text (p, 1,
		   "49 +x_0=1500121.846 +y_0=-672.557 +ellps=GRS80 +towgs84=");
    add_proj4text (p, 2, "0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SWEREF99 / RT90 5 gon O emulation\",GEOGCS[\"SW");
    add_srs_wkt (p, 1,
		 "EREF99\",DATUM[\"SWEREF99\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6619\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 4,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 5,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"4619\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"latitude_of_origin\",0],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",22.55633333333333],PARAMETER[\"scale_factor\",1.00");
    add_srs_wkt (p, 10,
		 "00049],PARAMETER[\"false_easting\",1500121.846],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_northing\",-672.557],AUTHORITY[\"EPSG\",\"3850");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3851, "epsg", 3851, "NZGD2000 / NZCS2000");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-37.5 +lat_2=-44.5 +lat_0=-41 +lon_0=17");
    add_proj4text (p, 1,
		   "3 +x_0=3000000 +y_0=7000000 +ellps=GRS80 +towgs84=0,0,0,");
    add_proj4text (p, 2, "0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD2000 / NZCS2000\",GEOGCS[\"NZGD2000\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"New_Zealand_Geodetic_Datum_2000\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4167\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_C");
    add_srs_wkt (p, 8,
		 "onic_2SP\"],PARAMETER[\"standard_parallel_1\",-37.5],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"standard_parallel_2\",-44.5],PARAMETER[\"latitu");
    add_srs_wkt (p, 10,
		 "de_of_origin\",-41],PARAMETER[\"central_meridian\",173],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_easting\",3000000],PARAMETER[\"false_n");
    add_srs_wkt (p, 12,
		 "orthing\",7000000],AUTHORITY[\"EPSG\",\"3851\"],AXIS[\"N");
    add_srs_wkt (p, 13, "orthing\",NORTH],AXIS[\"Easting\",EAST]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_14 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 3852, "epsg", 3852, "RSRGD2000 / DGLC2000");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-76.66666666666667 +lat_2=-79.333333333");
    add_proj4text (p, 1,
		   "33333 +lat_0=-90 +lon_0=157 +x_0=500000 +y_0=0 +ellps=GR");
    add_proj4text (p, 2, "S80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RSRGD2000 / DGLC2000\",GEOGCS[\"RSRGD2000\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Ross_Sea_Region_Geodetic_Datum_2000\",SPHEROID[\"GR");
    add_srs_wkt (p, 2,
		 "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019");
    add_srs_wkt (p, 3,
		 "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6764\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4764\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Confo");
    add_srs_wkt (p, 8,
		 "rmal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",-76.");
    add_srs_wkt (p, 9,
		 "66666666666667],PARAMETER[\"standard_parallel_2\",-79.33");
    add_srs_wkt (p, 10,
		 "333333333333],PARAMETER[\"latitude_of_origin\",-90],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",157],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 12,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13,
		 "EPSG\",\"3852\"],AXIS[\"Northing\",NORTH],AXIS[\"Easting");
    add_srs_wkt (p, 14, "\",EAST]]");
    p = add_epsg_def (first, last, 3854, "epsg", 3854, "County ST74");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18.05787 +k=0.99999506 +x_0=");
    add_proj4text (p, 1,
		   "100182.7406 +y_0=-6500620.1207 +ellps=GRS80 +towgs84=0,0");
    add_proj4text (p, 2, ",0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"County ST74\",GEOGCS[\"SWEREF99\",DATUM[\"SWERE");
    add_srs_wkt (p, 1,
		 "F99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR");
    add_srs_wkt (p, 2,
		 "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4619\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",18.05787],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"scale_factor\",0.99999506],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",100182.7406],PARAMETER[\"false_northing\",-6500620.");
    add_srs_wkt (p, 11,
		 "1207],AUTHORITY[\"EPSG\",\"3854\"],AXIS[\"x\",NORTH],AXI");
    add_srs_wkt (p, 12, "S[\"y\",EAST]]");
    p = add_epsg_def (first, last, 3857, "epsg", 3857,
		      "WGS 84 / Pseudo-Mercator");
    add_proj4text (p, 0,
		   "+proj=merc +a=6378137 +b=6378137 +lat_ts=0.0 +lon_0=0.0 ");
    add_proj4text (p, 1,
		   "+x_0=0.0 +y_0=0 +k=1.0 +units=m +nadgrids=@null +wktext ");
    add_proj4text (p, 2, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / Pseudo-Mercator\",GEOGCS[\"WGS 84\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.2572235");
    add_srs_wkt (p, 2,
		 "63,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"63");
    add_srs_wkt (p, 3,
		 "26\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Mercator_1");
    add_srs_wkt (p, 7,
		 "SP\"],PARAMETER[\"central_meridian\",0],PARAMETER[\"scal");
    add_srs_wkt (p, 8,
		 "e_factor\",1],PARAMETER[\"false_easting\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "false_northing\",0],EXTENSION[\"PROJ4\",\"+proj=merc +a=");
    add_srs_wkt (p, 10,
		 "6378137 +b=6378137 +lat_ts=0.0 +lon_0=0.0 +x_0=0.0 +y_0=");
    add_srs_wkt (p, 11,
		 "0 +k=1.0 +units=m +nadgrids=@null +wktext  +no_defs\"],A");
    add_srs_wkt (p, 12,
		 "UTHORITY[\"EPSG\",\"3857\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 13, "NORTH]]");
    p = add_epsg_def (first, last, 3889, "epsg", 3889, "IGRS");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"IGRS\",DATUM[\"Iraqi_Geospatial_Reference_Syste");
    add_srs_wkt (p, 1,
		 "m\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"1029\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"3889\"]]");
    p = add_epsg_def (first, last, 3890, "epsg", 3890, "IGRS / UTM zone 37N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=37 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGRS / UTM zone 37N\",GEOGCS[\"IGRS\",DATUM[\"I");
    add_srs_wkt (p, 1,
		 "raqi_Geospatial_Reference_System\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOW");
    add_srs_wkt (p, 3,
		 "GS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1029\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"3889\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",39],PARAMETER[\"scale_factor\",0.9996],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_easting\",500000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 11,
		 "ng\",0],AUTHORITY[\"EPSG\",\"3890\"],AXIS[\"Easting\",EA");
    add_srs_wkt (p, 12, "ST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3891, "epsg", 3891, "IGRS / UTM zone 38N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=38 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGRS / UTM zone 38N\",GEOGCS[\"IGRS\",DATUM[\"I");
    add_srs_wkt (p, 1,
		 "raqi_Geospatial_Reference_System\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOW");
    add_srs_wkt (p, 3,
		 "GS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1029\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"3889\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",45],PARAMETER[\"scale_factor\",0.9996],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_easting\",500000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 11,
		 "ng\",0],AUTHORITY[\"EPSG\",\"3891\"],AXIS[\"Easting\",EA");
    add_srs_wkt (p, 12, "ST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3892, "epsg", 3892, "IGRS / UTM zone 39N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=39 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"IGRS / UTM zone 39N\",GEOGCS[\"IGRS\",DATUM[\"I");
    add_srs_wkt (p, 1,
		 "raqi_Geospatial_Reference_System\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOW");
    add_srs_wkt (p, 3,
		 "GS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1029\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"3889\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 9,
		 "l_meridian\",51],PARAMETER[\"scale_factor\",0.9996],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_easting\",500000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 11,
		 "ng\",0],AUTHORITY[\"EPSG\",\"3892\"],AXIS[\"Easting\",EA");
    add_srs_wkt (p, 12, "ST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3893, "epsg", 3893,
		      "ED50 / Iraq National Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=29.02626833333333 +lon_0=46.5 +k=0.99");
    add_proj4text (p, 1, "94 +x_0=800000 +y_0=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / Iraq National Grid\",GEOGCS[\"ED50\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"European_Datum_1950\",SPHEROID[\"International 1924");
    add_srs_wkt (p, 2,
		 "\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",29.02626833333333],PARAMETER[\"central_meridian\",46.");
    add_srs_wkt (p, 9,
		 "5],PARAMETER[\"scale_factor\",0.9994],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",800000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"3893\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 12, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 3906, "epsg", 3906, "MGI 1901");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"MGI 1901\",DATUM[\"MGI_1901\",SPHEROID[\"Bessel");
    add_srs_wkt (p, 1,
		 " 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 2,
		 "4\"]],AUTHORITY[\"EPSG\",\"1031\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 3,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 4,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 5, "PSG\",\"3906\"]]");
    p = add_epsg_def (first, last, 3907, "epsg", 3907,
		      "MGI 1901 / Balkans zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=0.9999 +x_0=5500000 +y");
    add_proj4text (p, 1, "_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI 1901 / Balkans zone 5\",GEOGCS[\"MGI 1901\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"MGI_1901\",SPHEROID[\"Bessel 1841\",6377397.155");
    add_srs_wkt (p, 2,
		 ",299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"1031\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"3906\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",0],PARAMETER[\"central_meridian\",15],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",0.9999],PARAMETER[\"false_easting\",5500000],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3907");
    add_srs_wkt (p, 11, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3908, "epsg", 3908,
		      "MGI 1901 / Balkans zone 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18 +k=0.9999 +x_0=6500000 +y");
    add_proj4text (p, 1, "_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI 1901 / Balkans zone 6\",GEOGCS[\"MGI 1901\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"MGI_1901\",SPHEROID[\"Bessel 1841\",6377397.155");
    add_srs_wkt (p, 2,
		 ",299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"1031\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"3906\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",0],PARAMETER[\"central_meridian\",18],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",0.9999],PARAMETER[\"false_easting\",6500000],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3908");
    add_srs_wkt (p, 11, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3909, "epsg", 3909,
		      "MGI 1901 / Balkans zone 7");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=0.9999 +x_0=7500000 +y");
    add_proj4text (p, 1, "_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI 1901 / Balkans zone 7\",GEOGCS[\"MGI 1901\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"MGI_1901\",SPHEROID[\"Bessel 1841\",6377397.155");
    add_srs_wkt (p, 2,
		 ",299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"1031\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"3906\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",0],PARAMETER[\"central_meridian\",21],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",0.9999],PARAMETER[\"false_easting\",7500000],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3909");
    add_srs_wkt (p, 11, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3910, "epsg", 3910,
		      "MGI 1901 / Balkans zone 8");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=24 +k=0.9999 +x_0=8500000 +y");
    add_proj4text (p, 1, "_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI 1901 / Balkans zone 8\",GEOGCS[\"MGI 1901\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"MGI_1901\",SPHEROID[\"Bessel 1841\",6377397.155");
    add_srs_wkt (p, 2,
		 ",299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"1031\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"3906\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",0],PARAMETER[\"central_meridian\",24],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",0.9999],PARAMETER[\"false_easting\",8500000],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3910");
    add_srs_wkt (p, 11, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3911, "epsg", 3911,
		      "MGI 1901 / Slovenia Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI 1901 / Slovenia Grid\",GEOGCS[\"MGI 1901\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"MGI_1901\",SPHEROID[\"Bessel 1841\",6377397.155,");
    add_srs_wkt (p, 2,
		 "299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"1031\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"3906\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",15],PARAMETER[\"scale_");
    add_srs_wkt (p, 9,
		 "factor\",0.9999],PARAMETER[\"false_easting\",500000],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3911\"");
    add_srs_wkt (p, 11, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 3912, "epsg", 3912,
		      "MGI 1901 / Slovene National Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=-5000000 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI 1901 / Slovene National Grid\",GEOGCS[\"MGI");
    add_srs_wkt (p, 1,
		 " 1901\",DATUM[\"MGI_1901\",SPHEROID[\"Bessel 1841\",6377");
    add_srs_wkt (p, 2,
		 "397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"1031\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"3906");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",0],PARAMETER[\"central_meridian\",15],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 10,
		 "0000],PARAMETER[\"false_northing\",-5000000],AUTHORITY[\"");
    add_srs_wkt (p, 11, "EPSG\",\"3912\"],AXIS[\"Y\",EAST],AXIS[\"X\",NORTH]]");
    p = add_epsg_def (first, last, 3920, "epsg", 3920,
		      "Puerto Rico / UTM zone 20N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +ellps=clrk66 +towgs84=11,72,-101,0,0");
    add_proj4text (p, 1, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Puerto Rico / UTM zone 20N\",GEOGCS[\"Puerto Ri");
    add_srs_wkt (p, 1,
		 "co\",DATUM[\"Puerto_Rico\",SPHEROID[\"Clarke 1866\",6378");
    add_srs_wkt (p, 2,
		 "206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[11,72,-101,0,0,0,0],AUTHORITY[\"EPSG\",\"6139\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4139\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-63],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 10,
		 "6],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_northing\",0],AUTHORITY[\"EPSG\",\"3920\"],AXIS[\"Easti");
    add_srs_wkt (p, 12, "ng\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3942, "epsg", 3942, "RGF93 / CC42");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.25 +lat_2=42.75 +lat_0=42 +lon_0=3 +");
    add_proj4text (p, 1,
		   "x_0=1700000 +y_0=1200000 +ellps=GRS80 +towgs84=0,0,0,0,0");
    add_proj4text (p, 2, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGF93 / CC42\",GEOGCS[\"RGF93\",DATUM[\"Reseau_");
    add_srs_wkt (p, 1,
		 "Geodesique_Francais_1993\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0");
    add_srs_wkt (p, 3,
		 ",0,0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4171\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",41.25],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_2\",42.75],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",42],PARAMETER[\"central_meridian\",3],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",1700000],PARAMETER[\"false_northing\",120");
    add_srs_wkt (p, 12,
		 "0000],AUTHORITY[\"EPSG\",\"3942\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 13, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3943, "epsg", 3943, "RGF93 / CC43");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=42.25 +lat_2=43.75 +lat_0=43 +lon_0=3 +");
    add_proj4text (p, 1,
		   "x_0=1700000 +y_0=2200000 +ellps=GRS80 +towgs84=0,0,0,0,0");
    add_proj4text (p, 2, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGF93 / CC43\",GEOGCS[\"RGF93\",DATUM[\"Reseau_");
    add_srs_wkt (p, 1,
		 "Geodesique_Francais_1993\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0");
    add_srs_wkt (p, 3,
		 ",0,0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4171\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",42.25],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_2\",43.75],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",43],PARAMETER[\"central_meridian\",3],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",1700000],PARAMETER[\"false_northing\",220");
    add_srs_wkt (p, 12,
		 "0000],AUTHORITY[\"EPSG\",\"3943\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 13, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3944, "epsg", 3944, "RGF93 / CC44");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43.25 +lat_2=44.75 +lat_0=44 +lon_0=3 +");
    add_proj4text (p, 1,
		   "x_0=1700000 +y_0=3200000 +ellps=GRS80 +towgs84=0,0,0,0,0");
    add_proj4text (p, 2, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGF93 / CC44\",GEOGCS[\"RGF93\",DATUM[\"Reseau_");
    add_srs_wkt (p, 1,
		 "Geodesique_Francais_1993\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0");
    add_srs_wkt (p, 3,
		 ",0,0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4171\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",43.25],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_2\",44.75],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",44],PARAMETER[\"central_meridian\",3],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",1700000],PARAMETER[\"false_northing\",320");
    add_srs_wkt (p, 12,
		 "0000],AUTHORITY[\"EPSG\",\"3944\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 13, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3945, "epsg", 3945, "RGF93 / CC45");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.25 +lat_2=45.75 +lat_0=45 +lon_0=3 +");
    add_proj4text (p, 1,
		   "x_0=1700000 +y_0=4200000 +ellps=GRS80 +towgs84=0,0,0,0,0");
    add_proj4text (p, 2, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGF93 / CC45\",GEOGCS[\"RGF93\",DATUM[\"Reseau_");
    add_srs_wkt (p, 1,
		 "Geodesique_Francais_1993\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0");
    add_srs_wkt (p, 3,
		 ",0,0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4171\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",44.25],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_2\",45.75],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",45],PARAMETER[\"central_meridian\",3],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",1700000],PARAMETER[\"false_northing\",420");
    add_srs_wkt (p, 12,
		 "0000],AUTHORITY[\"EPSG\",\"3945\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 13, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3946, "epsg", 3946, "RGF93 / CC46");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.25 +lat_2=46.75 +lat_0=46 +lon_0=3 +");
    add_proj4text (p, 1,
		   "x_0=1700000 +y_0=5200000 +ellps=GRS80 +towgs84=0,0,0,0,0");
    add_proj4text (p, 2, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGF93 / CC46\",GEOGCS[\"RGF93\",DATUM[\"Reseau_");
    add_srs_wkt (p, 1,
		 "Geodesique_Francais_1993\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0");
    add_srs_wkt (p, 3,
		 ",0,0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4171\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",45.25],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_2\",46.75],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",46],PARAMETER[\"central_meridian\",3],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",1700000],PARAMETER[\"false_northing\",520");
    add_srs_wkt (p, 12,
		 "0000],AUTHORITY[\"EPSG\",\"3946\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 13, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3947, "epsg", 3947, "RGF93 / CC47");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46.25 +lat_2=47.75 +lat_0=47 +lon_0=3 +");
    add_proj4text (p, 1,
		   "x_0=1700000 +y_0=6200000 +ellps=GRS80 +towgs84=0,0,0,0,0");
    add_proj4text (p, 2, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGF93 / CC47\",GEOGCS[\"RGF93\",DATUM[\"Reseau_");
    add_srs_wkt (p, 1,
		 "Geodesique_Francais_1993\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0");
    add_srs_wkt (p, 3,
		 ",0,0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4171\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",46.25],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_2\",47.75],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",47],PARAMETER[\"central_meridian\",3],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",1700000],PARAMETER[\"false_northing\",620");
    add_srs_wkt (p, 12,
		 "0000],AUTHORITY[\"EPSG\",\"3947\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 13, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3948, "epsg", 3948, "RGF93 / CC48");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.25 +lat_2=48.75 +lat_0=48 +lon_0=3 +");
    add_proj4text (p, 1,
		   "x_0=1700000 +y_0=7200000 +ellps=GRS80 +towgs84=0,0,0,0,0");
    add_proj4text (p, 2, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGF93 / CC48\",GEOGCS[\"RGF93\",DATUM[\"Reseau_");
    add_srs_wkt (p, 1,
		 "Geodesique_Francais_1993\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0");
    add_srs_wkt (p, 3,
		 ",0,0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4171\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",47.25],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_2\",48.75],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",48],PARAMETER[\"central_meridian\",3],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",1700000],PARAMETER[\"false_northing\",720");
    add_srs_wkt (p, 12,
		 "0000],AUTHORITY[\"EPSG\",\"3948\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 13, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3949, "epsg", 3949, "RGF93 / CC49");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.25 +lat_2=49.75 +lat_0=49 +lon_0=3 +");
    add_proj4text (p, 1,
		   "x_0=1700000 +y_0=8200000 +ellps=GRS80 +towgs84=0,0,0,0,0");
    add_proj4text (p, 2, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGF93 / CC49\",GEOGCS[\"RGF93\",DATUM[\"Reseau_");
    add_srs_wkt (p, 1,
		 "Geodesique_Francais_1993\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0");
    add_srs_wkt (p, 3,
		 ",0,0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4171\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",48.25],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_2\",49.75],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",49],PARAMETER[\"central_meridian\",3],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",1700000],PARAMETER[\"false_northing\",820");
    add_srs_wkt (p, 12,
		 "0000],AUTHORITY[\"EPSG\",\"3949\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 13, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3950, "epsg", 3950, "RGF93 / CC50");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49.25 +lat_2=50.75 +lat_0=50 +lon_0=3 +");
    add_proj4text (p, 1,
		   "x_0=1700000 +y_0=9200000 +ellps=GRS80 +towgs84=0,0,0,0,0");
    add_proj4text (p, 2, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGF93 / CC50\",GEOGCS[\"RGF93\",DATUM[\"Reseau_");
    add_srs_wkt (p, 1,
		 "Geodesique_Francais_1993\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0");
    add_srs_wkt (p, 3,
		 ",0,0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4171\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"standard_parallel_1\",49.25],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "tandard_parallel_2\",50.75],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",50],PARAMETER[\"central_meridian\",3],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",1700000],PARAMETER[\"false_northing\",920");
    add_srs_wkt (p, 12,
		 "0000],AUTHORITY[\"EPSG\",\"3950\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 13, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3968, "epsg", 3968,
		      "NAD83 / Virginia Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37 +lat_2=39.5 +lat_0=36 +lon_0=-79.5 +");
    add_proj4text (p, 1,
		   "x_0=0 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Virginia Lambert\",GEOGCS[\"NAD83\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6");
    add_srs_wkt (p, 2,
		 "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"426");
    add_srs_wkt (p, 6,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"sta");
    add_srs_wkt (p, 8,
		 "ndard_parallel_1\",37],PARAMETER[\"standard_parallel_2\"");
    add_srs_wkt (p, 9,
		 ",39.5],PARAMETER[\"latitude_of_origin\",36],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "central_meridian\",-79.5],PARAMETER[\"false_easting\",0]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"39");
    add_srs_wkt (p, 12, "68\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3969, "epsg", 3969,
		      "NAD83(HARN) / Virginia Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37 +lat_2=39.5 +lat_0=36 +lon_0=-79.5 +");
    add_proj4text (p, 1, "x_0=0 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Virginia Lambert\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",37],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_2\",39.5],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",36],PARAMETER[\"central_meridian\",-79.5],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_easting\",0],PARAMETER[\"false_northing\",0],AU");
    add_srs_wkt (p, 12,
		 "THORITY[\"EPSG\",\"3969\"],AXIS[\"X\",EAST],AXIS[\"Y\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 3970, "epsg", 3970,
		      "NAD83(NSRS2007) / Virginia Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37 +lat_2=39.5 +lat_0=36 +lon_0=-79.5 +");
    add_proj4text (p, 1,
		   "x_0=0 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Virginia Lambert\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referenc");
    add_srs_wkt (p, 2,
		 "e_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.257222");
    add_srs_wkt (p, 3,
		 "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",37],PARAMETER[\"standard_parall");
    add_srs_wkt (p, 10,
		 "el_2\",39.5],PARAMETER[\"latitude_of_origin\",36],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"central_meridian\",-79.5],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 12,
		 "g\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 13, ",\"3970\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3975, "epsg", 3975,
		      "WGS 84 / NSIDC EASE-Grid Global");
    add_proj4text (p, 0,
		   "+proj=cea +lon_0=0 +lat_ts=30 +x_0=0 +y_0=0 +ellps=WGS84");
    add_proj4text (p, 1, " +datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / NSIDC EASE-Grid Global\",GEOGCS[\"WGS ");
    add_srs_wkt (p, 1,
		 "84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.");
    add_srs_wkt (p, 2,
		 "257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"C");
    add_srs_wkt (p, 7,
		 "ylindrical_Equal_Area\"],PARAMETER[\"standard_parallel_1");
    add_srs_wkt (p, 8,
		 "\",30],PARAMETER[\"central_meridian\",0],PARAMETER[\"fal");
    add_srs_wkt (p, 9,
		 "se_easting\",0],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 10,
		 "Y[\"EPSG\",\"3975\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    add_srs_wkt (p, 11, "");
    p = add_epsg_def (first, last, 3976, "epsg", 3976,
		      "WGS 84 / NSIDC Sea Ice Polar Stereographic South");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-70 +lon_0=0 +k=1 +x_0=0 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=WGS84 +datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / NSIDC Sea Ice Polar Stereographic Sout");
    add_srs_wkt (p, 1,
		 "h\",GEOGCS[\"WGS 84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS ");
    add_srs_wkt (p, 2,
		 "84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4326\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Polar_Stereographic\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",-70],PARAMETER[\"central_meridian\",0],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "3976\"],AXIS[\"X\",UNKNOWN],AXIS[\"Y\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3978, "epsg", 3978,
		      "NAD83 / Canada Atlas Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49 +lat_2=77 +lat_0=49 +lon_0=-95 +x_0=");
    add_proj4text (p, 1,
		   "0 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Canada Atlas Lambert\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "standard_parallel_1\",49],PARAMETER[\"standard_parallel_");
    add_srs_wkt (p, 9,
		 "2\",77],PARAMETER[\"latitude_of_origin\",49],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "central_meridian\",-95],PARAMETER[\"false_easting\",0],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3978");
    add_srs_wkt (p, 12, "\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3979, "epsg", 3979,
		      "NAD83(CSRS) / Canada Atlas Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49 +lat_2=77 +lat_0=49 +lon_0=-95 +x_0=");
    add_proj4text (p, 1, "0 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / Canada Atlas Lambert\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83(CSRS)\",DATUM[\"NAD83_Canadian_Spatial_Reference_Sy");
    add_srs_wkt (p, 2,
		 "stem\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 6,
		 "2\"]],AUTHORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUT");
    add_srs_wkt (p, 7,
		 "HORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conforma");
    add_srs_wkt (p, 8,
		 "l_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",49],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"standard_parallel_2\",77],PARAMETER[\"latitude_");
    add_srs_wkt (p, 10,
		 "of_origin\",49],PARAMETER[\"central_meridian\",-95],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_easting\",0],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 12,
		 "0],AUTHORITY[\"EPSG\",\"3979\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 13, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 3985, "epsg", 3985,
		      "Katanga 1955 / Katanga Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=-6.5 +lat_2=-11.5 +lat_0=9 +lon_0=26 +x");
    add_proj4text (p, 1,
		   "_0=500000 +y_0=500000 +ellps=clrk66 +towgs84=-103.746,-9");
    add_proj4text (p, 2, ".614,-255.95,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Katanga 1955 / Katanga Lambert\",GEOGCS[\"Katan");
    add_srs_wkt (p, 1,
		 "ga 1955\",DATUM[\"Katanga_1955\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[-103.746,-9.614,-255.95,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6695\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4695\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_p");
    add_srs_wkt (p, 9,
		 "arallel_1\",-6.5],PARAMETER[\"standard_parallel_2\",-11.");
    add_srs_wkt (p, 10,
		 "5],PARAMETER[\"latitude_of_origin\",9],PARAMETER[\"centr");
    add_srs_wkt (p, 11,
		 "al_meridian\",26],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 12,
		 "RAMETER[\"false_northing\",500000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 13, "3985\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3986, "epsg", 3986,
		      "Katanga 1955 / Katanga Gauss zone A");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-9 +lon_0=30 +k=1 +x_0=200000 +y_0=50");
    add_proj4text (p, 1,
		   "0000 +ellps=clrk66 +towgs84=-103.746,-9.614,-255.95,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Katanga 1955 / Katanga Gauss zone A\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Katanga 1955\",DATUM[\"Katanga_1955\",SPHEROID[\"Clarke ");
    add_srs_wkt (p, 2,
		 "1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7008\"]],TOWGS84[-103.746,-9.614,-255.95,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"6695\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4695");
    add_srs_wkt (p, 7,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 8,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 9,
		 "origin\",-9],PARAMETER[\"central_meridian\",30],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",200000");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"false_northing\",500000],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"3986\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3987, "epsg", 3987,
		      "Katanga 1955 / Katanga Gauss zone B");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-9 +lon_0=28 +k=1 +x_0=200000 +y_0=50");
    add_proj4text (p, 1,
		   "0000 +ellps=clrk66 +towgs84=-103.746,-9.614,-255.95,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Katanga 1955 / Katanga Gauss zone B\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Katanga 1955\",DATUM[\"Katanga_1955\",SPHEROID[\"Clarke ");
    add_srs_wkt (p, 2,
		 "1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7008\"]],TOWGS84[-103.746,-9.614,-255.95,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"6695\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4695");
    add_srs_wkt (p, 7,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 8,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 9,
		 "origin\",-9],PARAMETER[\"central_meridian\",28],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",200000");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"false_northing\",500000],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"3987\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3988, "epsg", 3988,
		      "Katanga 1955 / Katanga Gauss zone C");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-9 +lon_0=26 +k=1 +x_0=200000 +y_0=50");
    add_proj4text (p, 1,
		   "0000 +ellps=clrk66 +towgs84=-103.746,-9.614,-255.95,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Katanga 1955 / Katanga Gauss zone C\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Katanga 1955\",DATUM[\"Katanga_1955\",SPHEROID[\"Clarke ");
    add_srs_wkt (p, 2,
		 "1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7008\"]],TOWGS84[-103.746,-9.614,-255.95,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"6695\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4695");
    add_srs_wkt (p, 7,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 8,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 9,
		 "origin\",-9],PARAMETER[\"central_meridian\",26],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",200000");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"false_northing\",500000],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"3988\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3989, "epsg", 3989,
		      "Katanga 1955 / Katanga Gauss zone D");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-9 +lon_0=24 +k=1 +x_0=200000 +y_0=50");
    add_proj4text (p, 1,
		   "0000 +ellps=clrk66 +towgs84=-103.746,-9.614,-255.95,0,0,");
    add_proj4text (p, 2, "0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Katanga 1955 / Katanga Gauss zone D\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Katanga 1955\",DATUM[\"Katanga_1955\",SPHEROID[\"Clarke ");
    add_srs_wkt (p, 2,
		 "1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7008\"]],TOWGS84[-103.746,-9.614,-255.95,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"6695\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4695");
    add_srs_wkt (p, 7,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 8,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 9,
		 "origin\",-9],PARAMETER[\"central_meridian\",24],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",200000");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"false_northing\",500000],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"3989\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3991, "epsg", 3991,
		      "Puerto Rico State Plane CS of 1927");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=18.43333333333333 +lat_2=18.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=17.83333333333333 +lon_0=-66.43333333333334 +");
    add_proj4text (p, 2,
		   "x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +towgs84=11,7");
    add_proj4text (p, 3, "2,-101,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Puerto Rico State Plane CS of 1927\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "uerto Rico\",DATUM[\"Puerto_Rico\",SPHEROID[\"Clarke 186");
    add_srs_wkt (p, 2,
		 "6\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 3,
		 "8\"]],TOWGS84[11,72,-101,0,0,0,0],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 4,
		 "139\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 5,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 6,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4139\"]],UNIT[\"US s");
    add_srs_wkt (p, 7,
		 "urvey foot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "3\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"standard_parallel_1\",18.43333333333333],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"standard_parallel_2\",18.03333333333333],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "latitude_of_origin\",17.83333333333333],PARAMETER[\"cent");
    add_srs_wkt (p, 12,
		 "ral_meridian\",-66.43333333333334],PARAMETER[\"false_eas");
    add_srs_wkt (p, 13,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 14,
		 "[\"EPSG\",\"3991\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3992, "epsg", 3992,
		      "Puerto Rico / St. Croix");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=18.43333333333333 +lat_2=18.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=17.83333333333333 +lon_0=-66.43333333333334 +");
    add_proj4text (p, 2,
		   "x_0=152400.3048006096 +y_0=30480.06096012192 +ellps=clrk");
    add_proj4text (p, 3,
		   "66 +towgs84=11,72,-101,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Puerto Rico / St. Croix\",GEOGCS[\"Puerto Rico\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Puerto_Rico\",SPHEROID[\"Clarke 1866\",6378206.");
    add_srs_wkt (p, 2,
		 "4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS8");
    add_srs_wkt (p, 3,
		 "4[11,72,-101,0,0,0,0],AUTHORITY[\"EPSG\",\"6139\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4139\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",18.43333333333333],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",18.03333333333333],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",17.83333333333333],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 12,
		 "\",-66.43333333333334],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 13,
		 "0],PARAMETER[\"false_northing\",100000],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 14, "\",\"3992\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3994, "epsg", 3994, "WGS 84 / Mercator 41");
    add_proj4text (p, 0,
		   "+proj=merc +lon_0=100 +k=1 +x_0=0 +y_0=0 +ellps=WGS84 +d");
    add_proj4text (p, 1, "atum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / Mercator 41\",GEOGCS[\"WGS 84\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Mercator_1SP\"");
    add_srs_wkt (p, 7,
		 "],PARAMETER[\"central_meridian\",100],PARAMETER[\"scale_");
    add_srs_wkt (p, 8,
		 "factor\",1],PARAMETER[\"false_easting\",0],PARAMETER[\"f");
    add_srs_wkt (p, 9,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"3994\"],AXIS[\"X");
    add_srs_wkt (p, 10, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 3995, "epsg", 3995,
		      "WGS 84 / Arctic Polar Stereographic");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=90 +lat_ts=71 +lon_0=0 +k=1 +x_0=0 +y");
    add_proj4text (p, 1, "_0=0 +ellps=WGS84 +datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / Arctic Polar Stereographic\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "WGS 84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,");
    add_srs_wkt (p, 2,
		 "298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Polar_Stereographic\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",71],PARAMETER[\"central_meridian\",0],PARAMETER[\"sca");
    add_srs_wkt (p, 9,
		 "le_factor\",1],PARAMETER[\"false_easting\",0],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_northing\",0],AUTHORITY[\"EPSG\",\"3995\"],AXIS[");
    add_srs_wkt (p, 11, "\"X\",UNKNOWN],AXIS[\"Y\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3996, "epsg", 3996,
		      "WGS 84 / IBCAO Polar Stereographic");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=90 +lat_ts=75 +lon_0=0 +k=1 +x_0=0 +y");
    add_proj4text (p, 1, "_0=0 +ellps=WGS84 +datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / IBCAO Polar Stereographic\",GEOGCS[\"W");
    add_srs_wkt (p, 1,
		 "GS 84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,2");
    add_srs_wkt (p, 2,
		 "98.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Polar_Stereographic\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",75],PARAMETER[\"central_meridian\",0],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",1],PARAMETER[\"false_easting\",0],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"3996\"],AXIS[\"");
    add_srs_wkt (p, 11, "X\",UNKNOWN],AXIS[\"Y\",UNKNOWN]]");
    p = add_epsg_def (first, last, 3997, "epsg", 3997,
		      "WGS 84 / Dubai Local TM");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=55.33333333333334 +k=1 +x_0=");
    add_proj4text (p, 1,
		   "500000 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=m +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / Dubai Local TM\",GEOGCS[\"WGS 84\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.25722356");
    add_srs_wkt (p, 2,
		 "3,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"632");
    add_srs_wkt (p, 3,
		 "6\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",55.33333333333334],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "cale_factor\",1],PARAMETER[\"false_easting\",500000],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3997\"");
    add_srs_wkt (p, 11, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 4001, "epsg", 4001,
		      "Unknown datum based upon the Airy 1830 ellipsoid");
    add_proj4text (p, 0, "+proj=longlat +ellps=airy +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Airy 1830 ellipsoi");
    add_srs_wkt (p, 1,
		 "d\",DATUM[\"Not_specified_based_on_Airy_1830_ellipsoid\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Airy 1830\",6377563.396,299.3249646,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7001\"]],AUTHORITY[\"EPSG\",\"6001\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6, "]],AUTHORITY[\"EPSG\",\"4001\"]]");
    p = add_epsg_def (first, last, 4002, "epsg", 4002,
		      "Unknown datum based upon the Airy Modified 1849 ellipsoid");
    add_proj4text (p, 0, "+proj=longlat +ellps=mod_airy +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Airy Modified 1849");
    add_srs_wkt (p, 1,
		 " ellipsoid\",DATUM[\"Not_specified_based_on_Airy_Modifie");
    add_srs_wkt (p, 2,
		 "d_1849_ellipsoid\",SPHEROID[\"Airy Modified 1849\",63773");
    add_srs_wkt (p, 3,
		 "40.189,299.3249646,AUTHORITY[\"EPSG\",\"7002\"]],AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"6002\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 5,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4002\"");
    add_srs_wkt (p, 7, "]]");
    p = add_epsg_def (first, last, 4003, "epsg", 4003,
		      "Unknown datum based upon the Australian National Spheroid");
    add_proj4text (p, 0, "+proj=longlat +ellps=aust_SA +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Australian Nationa");
    add_srs_wkt (p, 1,
		 "l Spheroid\",DATUM[\"Not_specified_based_on_Australian_N");
    add_srs_wkt (p, 2,
		 "ational_Spheroid\",SPHEROID[\"Australian National Sphero");
    add_srs_wkt (p, 3,
		 "id\",6378160,298.25,AUTHORITY[\"EPSG\",\"7003\"]],AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"6003\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4003");
    add_srs_wkt (p, 7, "\"]]");
    p = add_epsg_def (first, last, 4004, "epsg", 4004,
		      "Unknown datum based upon the Bessel 1841 ellipsoid");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Bessel 1841 ellips");
    add_srs_wkt (p, 1,
		 "oid\",DATUM[\"Not_specified_based_on_Bessel_1841_ellipso");
    add_srs_wkt (p, 2,
		 "id\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6004\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6, "9122\"]],AUTHORITY[\"EPSG\",\"4004\"]]");
    p = add_epsg_def (first, last, 4005, "epsg", 4005,
		      "Unknown datum based upon the Bessel Modified ellipsoid");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377492.018 +b=6356173.508712696 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Bessel Modified el");
    add_srs_wkt (p, 1,
		 "lipsoid\",DATUM[\"Not_specified_based_on_Bessel_Modified");
    add_srs_wkt (p, 2,
		 "_ellipsoid\",SPHEROID[\"Bessel Modified\",6377492.018,29");
    add_srs_wkt (p, 3,
		 "9.1528128,AUTHORITY[\"EPSG\",\"7005\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6005\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6, "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4005\"]]");
    p = add_epsg_def (first, last, 4006, "epsg", 4006,
		      "Unknown datum based upon the Bessel Namibia ellipsoid");
    add_proj4text (p, 0, "+proj=longlat +ellps=bess_nam +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Bessel Namibia ell");
    add_srs_wkt (p, 1,
		 "ipsoid\",DATUM[\"Not_specified_based_on_Bessel_Namibia_e");
    add_srs_wkt (p, 2,
		 "llipsoid\",SPHEROID[\"Bessel Namibia (GLM)\",6377483.865");
    add_srs_wkt (p, 3,
		 "280419,299.1528128,AUTHORITY[\"EPSG\",\"7046\"]],AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"6006\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 5,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4006\"");
    add_srs_wkt (p, 7, "]]");
    p = add_epsg_def (first, last, 4007, "epsg", 4007,
		      "Unknown datum based upon the Clarke 1858 ellipsoid");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378293.645208759 +b=6356617.987679838 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Clarke 1858 ellips");
    add_srs_wkt (p, 1,
		 "oid\",DATUM[\"Not_specified_based_on_Clarke_1858_ellipso");
    add_srs_wkt (p, 2,
		 "id\",SPHEROID[\"Clarke 1858\",6378293.645208759,294.2606");
    add_srs_wkt (p, 3,
		 "763692569,AUTHORITY[\"EPSG\",\"7007\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6007\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6, "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4007\"]]");
    p = add_epsg_def (first, last, 4008, "epsg", 4008,
		      "Unknown datum based upon the Clarke 1866 ellipsoid");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk66 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Clarke 1866 ellips");
    add_srs_wkt (p, 1,
		 "oid\",DATUM[\"Not_specified_based_on_Clarke_1866_ellipso");
    add_srs_wkt (p, 2,
		 "id\",SPHEROID[\"Clarke 1866\",6378206.4,294.978698213900");
    add_srs_wkt (p, 3,
		 "6,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"600");
    add_srs_wkt (p, 4,
		 "8\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4008\"]]");
    p = add_epsg_def (first, last, 4009, "epsg", 4009,
		      "Unknown datum based upon the Clarke 1866 Michigan ellipsoid");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378450.047548896 +b=6356826.621488444 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Clarke 1866 Michig");
    add_srs_wkt (p, 1,
		 "an ellipsoid\",DATUM[\"Not_specified_based_on_Clarke_186");
    add_srs_wkt (p, 2,
		 "6_Michigan_ellipsoid\",SPHEROID[\"Clarke 1866 Michigan\"");
    add_srs_wkt (p, 3,
		 ",6378450.047548896,294.9786971646739,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"7009\"]],AUTHORITY[\"EPSG\",\"6009\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 5,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 6,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 7, "Y[\"EPSG\",\"4009\"]]");
    p = add_epsg_def (first, last, 4010, "epsg", 4010,
		      "Unknown datum based upon the Clarke 1880 (Benoit) ellipsoid");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378300.789 +b=6356566.435 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Clarke 1880 (Benoi");
    add_srs_wkt (p, 1,
		 "t) ellipsoid\",DATUM[\"Not_specified_based_on_Clarke_188");
    add_srs_wkt (p, 2,
		 "0_Benoit_ellipsoid\",SPHEROID[\"Clarke 1880 (Benoit)\",6");
    add_srs_wkt (p, 3,
		 "378300.789,293.4663155389802,AUTHORITY[\"EPSG\",\"7010\"");
    add_srs_wkt (p, 4,
		 "]],AUTHORITY[\"EPSG\",\"6010\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 6,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7, "\",\"4010\"]]");
    p = add_epsg_def (first, last, 4011, "epsg", 4011,
		      "Unknown datum based upon the Clarke 1880 (IGN) ellipsoid");
    add_proj4text (p, 0, "+proj=longlat +a=6378249.2 +b=6356515 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Clarke 1880 (IGN) ");
    add_srs_wkt (p, 1,
		 "ellipsoid\",DATUM[\"Not_specified_based_on_Clarke_1880_I");
    add_srs_wkt (p, 2,
		 "GN_ellipsoid\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,");
    add_srs_wkt (p, 3,
		 "293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"6011\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4011\"]");
    add_srs_wkt (p, 7, "]");
    p = add_epsg_def (first, last, 4012, "epsg", 4012,
		      "Unknown datum based upon the Clarke 1880 (RGS) ellipsoid");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Clarke 1880 (RGS) ");
    add_srs_wkt (p, 1,
		 "ellipsoid\",DATUM[\"Not_specified_based_on_Clarke_1880_R");
    add_srs_wkt (p, 2,
		 "GS_ellipsoid\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.14");
    add_srs_wkt (p, 3,
		 "5,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6012\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6, "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4012\"]]");
    p = add_epsg_def (first, last, 4013, "epsg", 4013,
		      "Unknown datum based upon the Clarke 1880 (Arc) ellipsoid");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378249.145 +b=6356514.966398753 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Clarke 1880 (Arc) ");
    add_srs_wkt (p, 1,
		 "ellipsoid\",DATUM[\"Not_specified_based_on_Clarke_1880_A");
    add_srs_wkt (p, 2,
		 "rc_ellipsoid\",SPHEROID[\"Clarke 1880 (Arc)\",6378249.14");
    add_srs_wkt (p, 3,
		 "5,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6013\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4013\"]]");
    p = add_epsg_def (first, last, 4014, "epsg", 4014,
		      "Unknown datum based upon the Clarke 1880 (SGA 1922) ellipsoid");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378249.2 +b=6356514.996941779 +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Clarke 1880 (SGA 1");
    add_srs_wkt (p, 1,
		 "922) ellipsoid\",DATUM[\"Not_specified_based_on_Clarke_1");
    add_srs_wkt (p, 2,
		 "880_SGA_1922_ellipsoid\",SPHEROID[\"Clarke 1880 (SGA 192");
    add_srs_wkt (p, 3,
		 "2)\",6378249.2,293.46598,AUTHORITY[\"EPSG\",\"7014\"]],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6014\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7, "4014\"]]");
    p = add_epsg_def (first, last, 4015, "epsg", 4015,
		      "Unknown datum based upon the Everest 1830 (1937 Adjustment) ellipsoid");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377276.345 +b=6356075.41314024 +no_def");
    add_proj4text (p, 1, "s");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Everest 1830 (1937");
    add_srs_wkt (p, 1,
		 " Adjustment) ellipsoid\",DATUM[\"Not_specified_based_on_");
    add_srs_wkt (p, 2,
		 "Everest_1830_1937_Adjustment_ellipsoid\",SPHEROID[\"Ever");
    add_srs_wkt (p, 3,
		 "est 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"7015\"]],AUTHORITY[\"EPSG\",\"6015\"]],PR");
    add_srs_wkt (p, 5,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 7, "\"]],AUTHORITY[\"EPSG\",\"4015\"]]");
    p = add_epsg_def (first, last, 4016, "epsg", 4016,
		      "Unknown datum based upon the Everest 1830 (1967 Definition) ellipsoid");
    add_proj4text (p, 0, "+proj=longlat +ellps=evrstSS +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Everest 1830 (1967");
    add_srs_wkt (p, 1,
		 " Definition) ellipsoid\",DATUM[\"Not_specified_based_on_");
    add_srs_wkt (p, 2,
		 "Everest_1830_1967_Definition_ellipsoid\",SPHEROID[\"Ever");
    add_srs_wkt (p, 3,
		 "est 1830 (1967 Definition)\",6377298.556,300.8017,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"7016\"]],AUTHORITY[\"EPSG\",\"6016\"]],PR");
    add_srs_wkt (p, 5,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 7, "\"]],AUTHORITY[\"EPSG\",\"4016\"]]");
    p = add_epsg_def (first, last, 4018, "epsg", 4018,
		      "Unknown datum based upon the Everest 1830 Modified ellipsoid");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377304.063 +b=6356103.038993155 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Everest 1830 Modif");
    add_srs_wkt (p, 1,
		 "ied ellipsoid\",DATUM[\"Not_specified_based_on_Everest_1");
    add_srs_wkt (p, 2,
		 "830_Modified_ellipsoid\",SPHEROID[\"Everest 1830 Modifie");
    add_srs_wkt (p, 3,
		 "d\",6377304.063,300.8017,AUTHORITY[\"EPSG\",\"7018\"]],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6018\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7, "4018\"]]");
    p = add_epsg_def (first, last, 4019, "epsg", 4019,
		      "Unknown datum based upon the GRS 1980 ellipsoid");
    add_proj4text (p, 0, "+proj=longlat +ellps=GRS80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the GRS 1980 ellipsoid");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"Not_specified_based_on_GRS_1980_ellipsoid\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6019\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6, "THORITY[\"EPSG\",\"4019\"]]");
    p = add_epsg_def (first, last, 4020, "epsg", 4020,
		      "Unknown datum based upon the Helmert 1906 ellipsoid");
    add_proj4text (p, 0, "+proj=longlat +ellps=helmert +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Helmert 1906 ellip");
    add_srs_wkt (p, 1,
		 "soid\",DATUM[\"Not_specified_based_on_Helmert_1906_ellip");
    add_srs_wkt (p, 2,
		 "soid\",SPHEROID[\"Helmert 1906\",6378200,298.3,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7020\"]],AUTHORITY[\"EPSG\",\"6020\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6, "],AUTHORITY[\"EPSG\",\"4020\"]]");
    p = add_epsg_def (first, last, 4021, "epsg", 4021,
		      "Unknown datum based upon the Indonesian National Spheroid");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378160 +b=6356774.50408554 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Indonesian Nationa");
    add_srs_wkt (p, 1,
		 "l Spheroid\",DATUM[\"Not_specified_based_on_Indonesian_N");
    add_srs_wkt (p, 2,
		 "ational_Spheroid\",SPHEROID[\"Indonesian National Sphero");
    add_srs_wkt (p, 3,
		 "id\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\"]],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6021\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"402");
    add_srs_wkt (p, 7, "1\"]]");
    p = add_epsg_def (first, last, 4022, "epsg", 4022,
		      "Unknown datum based upon the International 1924 ellipsoid");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the International 1924");
    add_srs_wkt (p, 1,
		 " ellipsoid\",DATUM[\"Not_specified_based_on_Internationa");
    add_srs_wkt (p, 2,
		 "l_1924_ellipsoid\",SPHEROID[\"International 1924\",63783");
    add_srs_wkt (p, 3,
		 "88,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6022\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4022\"]]");
    p = add_epsg_def (first, last, 4023, "epsg", 4023, "MOLDREF99");
    add_proj4text (p, 0, "+proj=longlat +ellps=GRS80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"MOLDREF99\",DATUM[\"MOLDREF99\",SPHEROID[\"GRS ");
    add_srs_wkt (p, 1,
		 "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 2,
		 "]],AUTHORITY[\"EPSG\",\"1032\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 4,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 5, "\",\"4023\"]]");
    p = add_epsg_def (first, last, 4024, "epsg", 4024,
		      "Unknown datum based upon the Krassowsky 1940 ellipsoid");
    add_proj4text (p, 0, "+proj=longlat +ellps=krass +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Krassowsky 1940 el");
    add_srs_wkt (p, 1,
		 "lipsoid\",DATUM[\"Not_specified_based_on_Krassowsky_1940");
    add_srs_wkt (p, 2,
		 "_ellipsoid\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6024\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6, "9122\"]],AUTHORITY[\"EPSG\",\"4024\"]]");
    p = add_epsg_def (first, last, 4025, "epsg", 4025,
		      "Unknown datum based upon the NWL 9D ellipsoid");
    add_proj4text (p, 0, "+proj=longlat +ellps=WGS66 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the NWL 9D ellipsoid\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Not_specified_based_on_NWL_9D_ellipsoid\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"NWL 9D\",6378145,298.25,AUTHORITY[\"EPSG\",\"7025\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6025\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6, "\",\"4025\"]]");
    p = add_epsg_def (first, last, 4026, "epsg", 4026,
		      "MOLDREF99 / Moldova TM");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=28.4 +k=0.9999400000000001 +");
    add_proj4text (p, 1,
		   "x_0=200000 +y_0=-5000000 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MOLDREF99 / Moldova TM\",GEOGCS[\"MOLDREF99\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"MOLDREF99\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 2,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"1032\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 4,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4023\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",28.4],PARAMETER[\"scale_fa");
    add_srs_wkt (p, 9,
		 "ctor\",0.99994],PARAMETER[\"false_easting\",200000],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_northing\",-5000000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "4026\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 4027, "epsg", 4027,
		      "Unknown datum based upon the Plessis 1817 ellipsoid");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6376523 +b=6355862.933255573 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Plessis 1817 ellip");
    add_srs_wkt (p, 1,
		 "soid\",DATUM[\"Not_specified_based_on_Plessis_1817_ellip");
    add_srs_wkt (p, 2,
		 "soid\",SPHEROID[\"Plessis 1817\",6376523,308.64,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7027\"]],AUTHORITY[\"EPSG\",\"6027\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6, "]],AUTHORITY[\"EPSG\",\"4027\"]]");
    p = add_epsg_def (first, last, 4028, "epsg", 4028,
		      "Unknown datum based upon the Struve 1860 ellipsoid");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378298.3 +b=6356657.142669561 +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Struve 1860 ellips");
    add_srs_wkt (p, 1,
		 "oid\",DATUM[\"Not_specified_based_on_Struve_1860_ellipso");
    add_srs_wkt (p, 2,
		 "id\",SPHEROID[\"Struve 1860\",6378298.3,294.73,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7028\"]],AUTHORITY[\"EPSG\",\"6028\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6, "],AUTHORITY[\"EPSG\",\"4028\"]]");
    p = add_epsg_def (first, last, 4029, "epsg", 4029,
		      "Unknown datum based upon the War Office ellipsoid");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378300 +b=6356751.689189189 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the War Office ellipso");
    add_srs_wkt (p, 1,
		 "id\",DATUM[\"Not_specified_based_on_War_Office_ellipsoid");
    add_srs_wkt (p, 2,
		 "\",SPHEROID[\"War Office\",6378300,296,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7029\"]],AUTHORITY[\"EPSG\",\"6029\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6, "TY[\"EPSG\",\"4029\"]]");
    p = add_epsg_def (first, last, 4030, "epsg", 4030,
		      "Unknown datum based upon the WGS 84 ellipsoid");
    add_proj4text (p, 0, "+proj=longlat +ellps=WGS84 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the WGS 84 ellipsoid\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Not_specified_based_on_WGS_84_ellipsoid\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"7030\"]],AUTHORITY[\"EPSG\",\"6030\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6, "Y[\"EPSG\",\"4030\"]]");
    p = add_epsg_def (first, last, 4031, "epsg", 4031,
		      "Unknown datum based upon the GEM 10C ellipsoid");
    add_proj4text (p, 0, "+proj=longlat +ellps=WGS84 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the GEM 10C ellipsoid\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Not_specified_based_on_GEM_10C_ellipsoid\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GEM 10C\",6378137,298.257223563,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7031\"]],AUTHORITY[\"EPSG\",\"6031\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6, "TY[\"EPSG\",\"4031\"]]");
    p = add_epsg_def (first, last, 4032, "epsg", 4032,
		      "Unknown datum based upon the OSU86F ellipsoid");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378136.2 +b=6356751.516927429 +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the OSU86F ellipsoid\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Not_specified_based_on_OSU86F_ellipsoid\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"OSU86F\",6378136.2,298.257223563,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7032\"]],AUTHORITY[\"EPSG\",\"6032\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6, "TY[\"EPSG\",\"4032\"]]");
    p = add_epsg_def (first, last, 4033, "epsg", 4033,
		      "Unknown datum based upon the OSU91A ellipsoid");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378136.3 +b=6356751.616592146 +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the OSU91A ellipsoid\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Not_specified_based_on_OSU91A_ellipsoid\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"OSU91A\",6378136.3,298.257223563,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7033\"]],AUTHORITY[\"EPSG\",\"6033\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6, "TY[\"EPSG\",\"4033\"]]");
    p = add_epsg_def (first, last, 4034, "epsg", 4034,
		      "Unknown datum based upon the Clarke 1880 ellipsoid");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378249.144808011 +b=6356514.966204134 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Clarke 1880 ellips");
    add_srs_wkt (p, 1,
		 "oid\",DATUM[\"Not_specified_based_on_Clarke_1880_ellipso");
    add_srs_wkt (p, 2,
		 "id\",SPHEROID[\"Clarke 1880\",6378249.144808011,293.4663");
    add_srs_wkt (p, 3,
		 "076556349,AUTHORITY[\"EPSG\",\"7034\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6034\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6, "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4034\"]]");
    p = add_epsg_def (first, last, 4035, "epsg", 4035,
		      "Unknown datum based upon the Authalic Sphere");
    add_proj4text (p, 0, "+proj=longlat +a=6371000 +b=6371000 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Authalic Sphere\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Not_specified_based_on_Authalic_Sphere\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"Sphere\",6371000,0,AUTHORITY[\"EPSG\",\"7035\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6035\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "33,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"40");
    add_srs_wkt (p, 6, "35\"]]");
    p = add_epsg_def (first, last, 4036, "epsg", 4036,
		      "Unknown datum based upon the GRS 1967 ellipsoid");
    add_proj4text (p, 0, "+proj=longlat +ellps=GRS67 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the GRS 1967 ellipsoid");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"Not_specified_based_on_GRS_1967_ellipsoid\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1967\",6378160,298.247167427,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7036\"]],AUTHORITY[\"EPSG\",\"6036\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6, "THORITY[\"EPSG\",\"4036\"]]");
    p = add_epsg_def (first, last, 4037, "epsg", 4037, "WGS 84 / TMzn35N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=35 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / TMzn35N\",GEOGCS[\"WGS 84\",DATUM[\"WG");
    add_srs_wkt (p, 1,
		 "S_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHO");
    add_srs_wkt (p, 2,
		 "RITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],P");
    add_srs_wkt (p, 3,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 4,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 5,
		 "2\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merca");
    add_srs_wkt (p, 7,
		 "tor\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"c");
    add_srs_wkt (p, 8,
		 "entral_meridian\",27],PARAMETER[\"scale_factor\",0.9996]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"false_easting\",500000],PARAMETER[\"false_n");
    add_srs_wkt (p, 10,
		 "orthing\",0],AUTHORITY[\"EPSG\",\"4037\"],AXIS[\"Northin");
    add_srs_wkt (p, 11, "g\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 4038, "epsg", 4038, "WGS 84 / TMzn36N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=36 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / TMzn36N\",GEOGCS[\"WGS 84\",DATUM[\"WG");
    add_srs_wkt (p, 1,
		 "S_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHO");
    add_srs_wkt (p, 2,
		 "RITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],P");
    add_srs_wkt (p, 3,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 4,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 5,
		 "2\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merca");
    add_srs_wkt (p, 7,
		 "tor\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"c");
    add_srs_wkt (p, 8,
		 "entral_meridian\",33],PARAMETER[\"scale_factor\",0.9996]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"false_easting\",500000],PARAMETER[\"false_n");
    add_srs_wkt (p, 10,
		 "orthing\",0],AUTHORITY[\"EPSG\",\"4038\"],AXIS[\"Northin");
    add_srs_wkt (p, 11, "g\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 4041, "epsg", 4041,
		      "Unknown datum based upon the Average Terrestrial System 1977 ellipsoid");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378135 +b=6356750.304921594 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Average Terrestria");
    add_srs_wkt (p, 1,
		 "l System 1977 ellipsoid\",DATUM[\"Not_specified_based_on");
    add_srs_wkt (p, 2,
		 "_Average_Terrestrial_System_1977_ellipsoid\",SPHEROID[\"");
    add_srs_wkt (p, 3,
		 "Average Terrestrial System 1977\",6378135,298.257,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"7041\"]],AUTHORITY[\"EPSG\",\"6041\"]],PR");
    add_srs_wkt (p, 5,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 7, "\"]],AUTHORITY[\"EPSG\",\"4041\"]]");
    p = add_epsg_def (first, last, 4042, "epsg", 4042,
		      "Unknown datum based upon the Everest (1830 Definition) ellipsoid");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377299.36559538 +b=6356098.359005156 +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Everest (1830 Defi");
    add_srs_wkt (p, 1,
		 "nition) ellipsoid\",DATUM[\"Not_specified_based_on_Evere");
    add_srs_wkt (p, 2,
		 "st_1830_Definition_ellipsoid\",SPHEROID[\"Everest (1830 ");
    add_srs_wkt (p, 3,
		 "Definition)\",6377299.36559538,300.8017255433552,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"7042\"]],AUTHORITY[\"EPSG\",\"6042\"]],PRI");
    add_srs_wkt (p, 5,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 6,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 7, "]],AUTHORITY[\"EPSG\",\"4042\"]]");
    p = add_epsg_def (first, last, 4043, "epsg", 4043,
		      "Unknown datum based upon the WGS 72 ellipsoid");
    add_proj4text (p, 0, "+proj=longlat +ellps=WGS72 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the WGS 72 ellipsoid\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Not_specified_based_on_WGS_72_ellipsoid\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6043\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6, "\",\"4043\"]]");
    p = add_epsg_def (first, last, 4044, "epsg", 4044,
		      "Unknown datum based upon the Everest 1830 (1962 Definition) ellipsoid");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377301.243 +b=6356100.230165384 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Everest 1830 (1962");
    add_srs_wkt (p, 1,
		 " Definition) ellipsoid\",DATUM[\"Not_specified_based_on_");
    add_srs_wkt (p, 2,
		 "Everest_1830_1962_Definition_ellipsoid\",SPHEROID[\"Ever");
    add_srs_wkt (p, 3,
		 "est 1830 (1962 Definition)\",6377301.243,300.8017255,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"7044\"]],AUTHORITY[\"EPSG\",\"6044\"]]");
    add_srs_wkt (p, 5,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7, "122\"]],AUTHORITY[\"EPSG\",\"4044\"]]");
    p = add_epsg_def (first, last, 4045, "epsg", 4045,
		      "Unknown datum based upon the Everest 1830 (1975 Definition) ellipsoid");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377299.151 +b=6356098.145120132 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unknown datum based upon the Everest 1830 (1975");
    add_srs_wkt (p, 1,
		 " Definition) ellipsoid\",DATUM[\"Not_specified_based_on_");
    add_srs_wkt (p, 2,
		 "Everest_1830_1975_Definition_ellipsoid\",SPHEROID[\"Ever");
    add_srs_wkt (p, 3,
		 "est 1830 (1975 Definition)\",6377299.151,300.8017255,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"7045\"]],AUTHORITY[\"EPSG\",\"6045\"]]");
    add_srs_wkt (p, 5,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7, "122\"]],AUTHORITY[\"EPSG\",\"4045\"]]");
    p = add_epsg_def (first, last, 4046, "epsg", 4046, "RGRDC 2005");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"RGRDC 2005\",DATUM[\"Reseau_Geodesique_de_la_RD");
    add_srs_wkt (p, 1,
		 "C_2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4046");
    add_srs_wkt (p, 6, "\"]]");
    p = add_epsg_def (first, last, 4047, "epsg", 4047,
		      "Unspecified datum based upon the GRS 1980 Authalic Sphere");
    add_proj4text (p, 0, "+proj=longlat +a=6371007 +b=6371007 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unspecified datum based upon the GRS 1980 Autha");
    add_srs_wkt (p, 1,
		 "lic Sphere\",DATUM[\"Not_specified_based_on_GRS_1980_Aut");
    add_srs_wkt (p, 2,
		 "halic_Sphere\",SPHEROID[\"GRS 1980 Authalic Sphere\",637");
    add_srs_wkt (p, 3,
		 "1007,0,AUTHORITY[\"EPSG\",\"7048\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6047\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4047\"]]");
    p = add_epsg_def (first, last, 4048, "epsg", 4048,
		      "RGRDC 2005 / Congo TM zone 12");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=12 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=10000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGRDC 2005 / Congo TM zone 12\",GEOGCS[\"RGRDC ");
    add_srs_wkt (p, 1,
		 "2005\",DATUM[\"Reseau_Geodesique_de_la_RDC_2005\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4046\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 8,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"central_meridian\",12],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",0.9999],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"4048\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 4049, "epsg", 4049,
		      "RGRDC 2005 / Congo TM zone 14");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=14 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=10000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGRDC 2005 / Congo TM zone 14\",GEOGCS[\"RGRDC ");
    add_srs_wkt (p, 1,
		 "2005\",DATUM[\"Reseau_Geodesique_de_la_RDC_2005\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4046\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 8,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"central_meridian\",14],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",0.9999],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"4049\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 4050, "epsg", 4050,
		      "RGRDC 2005 / Congo TM zone 16");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=16 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=10000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGRDC 2005 / Congo TM zone 16\",GEOGCS[\"RGRDC ");
    add_srs_wkt (p, 1,
		 "2005\",DATUM[\"Reseau_Geodesique_de_la_RDC_2005\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4046\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 8,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"central_meridian\",16],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",0.9999],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"4050\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 4051, "epsg", 4051,
		      "RGRDC 2005 / Congo TM zone 18");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=10000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGRDC 2005 / Congo TM zone 18\",GEOGCS[\"RGRDC ");
    add_srs_wkt (p, 1,
		 "2005\",DATUM[\"Reseau_Geodesique_de_la_RDC_2005\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4046\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 8,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"central_meridian\",18],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",0.9999],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"4051\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 4052, "epsg", 4052,
		      "Unspecified datum based upon the Clarke 1866 Authalic Sphere");
    add_proj4text (p, 0, "+proj=longlat +a=6370997 +b=6370997 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unspecified datum based upon the Clarke 1866 Au");
    add_srs_wkt (p, 1,
		 "thalic Sphere\",DATUM[\"Not_specified_based_on_Clarke_18");
    add_srs_wkt (p, 2,
		 "66_Authalic_Sphere\",SPHEROID[\"Clarke 1866 Authalic Sph");
    add_srs_wkt (p, 3,
		 "ere\",6370997,0,AUTHORITY[\"EPSG\",\"7052\"]],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6052\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4052\"]]");
    add_srs_wkt (p, 7, "");
    p = add_epsg_def (first, last, 4053, "epsg", 4053,
		      "Unspecified datum based upon the International 1924 Authalic Sphere");
    add_proj4text (p, 0, "+proj=longlat +a=6371228 +b=6371228 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unspecified datum based upon the International ");
    add_srs_wkt (p, 1,
		 "1924 Authalic Sphere\",DATUM[\"Not_specified_based_on_In");
    add_srs_wkt (p, 2,
		 "ternational_1924_Authalic_Sphere\",SPHEROID[\"Internatio");
    add_srs_wkt (p, 3,
		 "nal 1924 Authalic Sphere\",6371228,0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"7057\"]],AUTHORITY[\"EPSG\",\"6053\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 5,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 6,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 7, "Y[\"EPSG\",\"4053\"]]");
    p = add_epsg_def (first, last, 4054, "epsg", 4054,
		      "Unspecified datum based upon the Hughes 1980 ellipsoid");
    add_proj4text (p, 0, "+proj=longlat +a=6378273 +b=6356889.449 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Unspecified datum based upon the Hughes 1980 el");
    add_srs_wkt (p, 1,
		 "lipsoid\",DATUM[\"Not_specified_based_on_Hughes_1980_ell");
    add_srs_wkt (p, 2,
		 "ipsoid\",SPHEROID[\"Hughes 1980\",6378273,298.2794111230");
    add_srs_wkt (p, 3,
		 "61,AUTHORITY[\"EPSG\",\"7058\"]],AUTHORITY[\"EPSG\",\"60");
    add_srs_wkt (p, 4,
		 "54\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4054\"]]");
    p = add_epsg_def (first, last, 4055, "epsg", 4055,
		      "Popular Visualisation CRS");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378137 +b=6378137 +towgs84=0,0,0,0,0,0");
    add_proj4text (p, 1, ",0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Popular Visualisation CRS\",DATUM[\"Popular_Vis");
    add_srs_wkt (p, 1,
		 "ualisation_Datum\",SPHEROID[\"Popular Visualisation Sphe");
    add_srs_wkt (p, 2,
		 "re\",6378137,0,AUTHORITY[\"EPSG\",\"7059\"]],TOWGS84[0,0");
    add_srs_wkt (p, 3,
		 ",0,0,0,0,0],AUTHORITY[\"EPSG\",\"6055\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6, "TY[\"EPSG\",\"4055\"]]");
    p = add_epsg_def (first, last, 4056, "epsg", 4056,
		      "RGRDC 2005 / Congo TM zone 20");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=20 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=10000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGRDC 2005 / Congo TM zone 20\",GEOGCS[\"RGRDC ");
    add_srs_wkt (p, 1,
		 "2005\",DATUM[\"Reseau_Geodesique_de_la_RDC_2005\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4046\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 8,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"central_meridian\",20],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",0.9999],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"4056\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 4057, "epsg", 4057,
		      "RGRDC 2005 / Congo TM zone 22");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=22 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=10000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGRDC 2005 / Congo TM zone 22\",GEOGCS[\"RGRDC ");
    add_srs_wkt (p, 1,
		 "2005\",DATUM[\"Reseau_Geodesique_de_la_RDC_2005\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4046\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 8,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"central_meridian\",22],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",0.9999],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"4057\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 4058, "epsg", 4058,
		      "RGRDC 2005 / Congo TM zone 24");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=24 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=10000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGRDC 2005 / Congo TM zone 24\",GEOGCS[\"RGRDC ");
    add_srs_wkt (p, 1,
		 "2005\",DATUM[\"Reseau_Geodesique_de_la_RDC_2005\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4046\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 8,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"central_meridian\",24],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",0.9999],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"4058\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 4059, "epsg", 4059,
		      "RGRDC 2005 / Congo TM zone 26");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=26 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=10000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGRDC 2005 / Congo TM zone 26\",GEOGCS[\"RGRDC ");
    add_srs_wkt (p, 1,
		 "2005\",DATUM[\"Reseau_Geodesique_de_la_RDC_2005\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4046\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 8,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"central_meridian\",26],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",0.9999],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"4059\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 4060, "epsg", 4060,
		      "RGRDC 2005 / Congo TM zone 28");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=28 +k=0.9999 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=10000000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGRDC 2005 / Congo TM zone 28\",GEOGCS[\"RGRDC ");
    add_srs_wkt (p, 1,
		 "2005\",DATUM[\"Reseau_Geodesique_de_la_RDC_2005\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4046\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 8,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"central_meridian\",28],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",0.9999],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"4060\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 4061, "epsg", 4061,
		      "RGRDC 2005 / UTM zone 33S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=33 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGRDC 2005 / UTM zone 33S\",GEOGCS[\"RGRDC 2005");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"Reseau_Geodesique_de_la_RDC_2005\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1033\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4046\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",15],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 10,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"4061\"],AXIS");
    add_srs_wkt (p, 12, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 4062, "epsg", 4062,
		      "RGRDC 2005 / UTM zone 34S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=34 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGRDC 2005 / UTM zone 34S\",GEOGCS[\"RGRDC 2005");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"Reseau_Geodesique_de_la_RDC_2005\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1033\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4046\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",21],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 10,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"4062\"],AXIS");
    add_srs_wkt (p, 12, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 4063, "epsg", 4063,
		      "RGRDC 2005 / UTM zone 35S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=35 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RGRDC 2005 / UTM zone 35S\",GEOGCS[\"RGRDC 2005");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"Reseau_Geodesique_de_la_RDC_2005\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1033\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4046\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",27],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 10,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"4063\"],AXIS");
    add_srs_wkt (p, 12, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 4071, "epsg", 4071, "Chua / UTM zone 23S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=23 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Chua / UTM zone 23S\",GEOGCS[\"Chua\",DATUM[\"C");
    add_srs_wkt (p, 1,
		 "hua\",SPHEROID[\"International 1924\",6378388,297,AUTHOR");
    add_srs_wkt (p, 2,
		 "ITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6224\"]],PR");
    add_srs_wkt (p, 3,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 4,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 5,
		 "\"]],AUTHORITY[\"EPSG\",\"4224\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 6,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 7,
		 "or\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"ce");
    add_srs_wkt (p, 8,
		 "ntral_meridian\",-45],PARAMETER[\"scale_factor\",0.9996]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"false_easting\",500000],PARAMETER[\"false_n");
    add_srs_wkt (p, 10,
		 "orthing\",10000000],AUTHORITY[\"EPSG\",\"4071\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 4075, "epsg", 4075, "SREF98");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"SREF98\",DATUM[\"Serbian_Reference_Network_1998");
    add_srs_wkt (p, 1,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 2,
		 "[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"1034\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4075\"]]");
    p = add_epsg_def (first, last, 4081, "epsg", 4081, "REGCAN95");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"REGCAN95\",DATUM[\"Red_Geodesica_de_Canarias_19");
    add_srs_wkt (p, 1,
		 "95\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"1035\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4081\"]]");
    add_srs_wkt (p, 6, "");
    p = add_epsg_def (first, last, 4082, "epsg", 4082,
		      "REGCAN95 / UTM zone 27N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=27 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"REGCAN95 / UTM zone 27N\",GEOGCS[\"REGCAN95\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Red_Geodesica_de_Canarias_1995\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 2,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 3,
		 "],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1035\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 6,
		 "2\"]],AUTHORITY[\"EPSG\",\"4081\"]],UNIT[\"metre\",1,AUT");
    add_srs_wkt (p, 7,
		 "HORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merca");
    add_srs_wkt (p, 8,
		 "tor\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"c");
    add_srs_wkt (p, 9,
		 "entral_meridian\",-21],PARAMETER[\"scale_factor\",0.9996");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"false_easting\",500000],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "northing\",0],AUTHORITY[\"EPSG\",\"4082\"],AXIS[\"Eastin");
    add_srs_wkt (p, 12, "g\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 4083, "epsg", 4083,
		      "REGCAN95 / UTM zone 28N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=28 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"REGCAN95 / UTM zone 28N\",GEOGCS[\"REGCAN95\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Red_Geodesica_de_Canarias_1995\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 2,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 3,
		 "],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1035\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 6,
		 "2\"]],AUTHORITY[\"EPSG\",\"4081\"]],UNIT[\"metre\",1,AUT");
    add_srs_wkt (p, 7,
		 "HORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merca");
    add_srs_wkt (p, 8,
		 "tor\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"c");
    add_srs_wkt (p, 9,
		 "entral_meridian\",-15],PARAMETER[\"scale_factor\",0.9996");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"false_easting\",500000],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "northing\",0],AUTHORITY[\"EPSG\",\"4083\"],AXIS[\"Eastin");
    add_srs_wkt (p, 12, "g\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 4120, "epsg", 4120, "Greek");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Greek\",DATUM[\"Greek\",SPHEROID[\"Bessel 1841\"");
    add_srs_wkt (p, 1,
		 ",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"6120\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 4,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5, "4120\"]]");
    p = add_epsg_def (first, last, 4121, "epsg", 4121, "GGRS87");
    add_proj4text (p, 0, "+proj=longlat +ellps=GRS80 +datum=GGRS87 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"GGRS87\",DATUM[\"Greek_Geodetic_Reference_Syste");
    add_srs_wkt (p, 1,
		 "m_1987\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7019\"]],TOWGS84[-199.87,74.79,246.62,");
    add_srs_wkt (p, 3,
		 "0,0,0,0],AUTHORITY[\"EPSG\",\"6121\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6, "\"EPSG\",\"4121\"]]");
    p = add_epsg_def (first, last, 4122, "epsg", 4122, "ATS77");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378135 +b=6356750.304921594 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"ATS77\",DATUM[\"Average_Terrestrial_System_1977");
    add_srs_wkt (p, 1,
		 "\",SPHEROID[\"Average Terrestrial System 1977\",6378135,");
    add_srs_wkt (p, 2,
		 "298.257,AUTHORITY[\"EPSG\",\"7041\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6122\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4122\"]]");
    p = add_epsg_def (first, last, 4123, "epsg", 4123, "KKJ");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"KKJ\",DATUM[\"Kartastokoordinaattijarjestelma_1");
    add_srs_wkt (p, 1,
		 "966\",SPHEROID[\"International 1924\",6378388,297,AUTHOR");
    add_srs_wkt (p, 2,
		 "ITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6123\"]],PR");
    add_srs_wkt (p, 3,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 4,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 5, "\"]],AUTHORITY[\"EPSG\",\"4123\"]]");
    p = add_epsg_def (first, last, 4124, "epsg", 4124, "RT90");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"RT90\",DATUM[\"Rikets_koordinatsystem_1990\",SP");
    add_srs_wkt (p, 1,
		 "HEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY");
    add_srs_wkt (p, 2,
		 "[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6124\"]],PRIME");
    add_srs_wkt (p, 3,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 4,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 5, "],AUTHORITY[\"EPSG\",\"4124\"]]");
    p = add_epsg_def (first, last, 4125, "epsg", 4125, "Samboja");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=bessel +towgs84=-404.78,685.68,45.4");
    add_proj4text (p, 1, "7,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Samboja\",DATUM[\"Samboja\",SPHEROID[\"Bessel 1");
    add_srs_wkt (p, 1,
		 "841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"");
    add_srs_wkt (p, 2,
		 "]],TOWGS84[-404.78,685.68,45.47,0,0,0,0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"6125\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.0174532925199433,AUTHORITY");
    add_srs_wkt (p, 5, "[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"4125\"]]");
    p = add_epsg_def (first, last, 4126, "epsg", 4126, "LKS94 (ETRS89)");
    add_proj4text (p, 0, "+proj=longlat +ellps=GRS80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"LKS94 (ETRS89)\",DATUM[\"Lithuania_1994_ETRS89\"");
    add_srs_wkt (p, 1,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6126\"]],PRIMEM[\"");
    add_srs_wkt (p, 3,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 4,
		 "e\",0.0174532925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUT");
    add_srs_wkt (p, 5, "HORITY[\"EPSG\",\"4126\"]]");
    p = add_epsg_def (first, last, 4127, "epsg", 4127, "Tete");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk66 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Tete\",DATUM[\"Tete\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 1,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 2,
		 "],AUTHORITY[\"EPSG\",\"6127\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 4,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"4127\"]]");
    p = add_epsg_def (first, last, 4128, "epsg", 4128, "Madzansua");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk66 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Madzansua\",DATUM[\"Madzansua\",SPHEROID[\"Clar");
    add_srs_wkt (p, 1,
		 "ke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 2,
		 ",\"7008\"]],AUTHORITY[\"EPSG\",\"6128\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 3,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 4,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 5, "TY[\"EPSG\",\"4128\"]]");
    p = add_epsg_def (first, last, 4129, "epsg", 4129, "Observatario");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk66 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Observatario\",DATUM[\"Observatario\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6129\"]],PRIMEM[\"");
    add_srs_wkt (p, 3,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 4,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 5, "THORITY[\"EPSG\",\"4129\"]]");
    p = add_epsg_def (first, last, 4130, "epsg", 4130, "Moznet");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=WGS84 +towgs84=0,0,0,-0,-0,-0,0 +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Moznet\",DATUM[\"Moznet_ITRF94\",SPHEROID[\"WGS");
    add_srs_wkt (p, 1,
		 " 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]");
    add_srs_wkt (p, 2,
		 "],TOWGS84[0,0,0,-0,-0,-0,0],AUTHORITY[\"EPSG\",\"6130\"]");
    add_srs_wkt (p, 3,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 4,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5, "9122\"]],AUTHORITY[\"EPSG\",\"4130\"]]");
    p = add_epsg_def (first, last, 4131, "epsg", 4131, "Indian 1960");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377276.345 +b=6356075.41314024 +no_def");
    add_proj4text (p, 1, "s");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Indian 1960\",DATUM[\"Indian_1960\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AU");
    add_srs_wkt (p, 2,
		 "THORITY[\"EPSG\",\"7015\"]],AUTHORITY[\"EPSG\",\"6131\"]");
    add_srs_wkt (p, 3,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 4,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5, "9122\"]],AUTHORITY[\"EPSG\",\"4131\"]]");
    p = add_epsg_def (first, last, 4132, "epsg", 4132, "FD58");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"FD58\",DATUM[\"Final_Datum_1958\",SPHEROID[\"Cl");
    add_srs_wkt (p, 1,
		 "arke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 2,
		 ",\"7012\"]],AUTHORITY[\"EPSG\",\"6132\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 3,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 4,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 5, "TY[\"EPSG\",\"4132\"]]");
    p = add_epsg_def (first, last, 4133, "epsg", 4133, "EST92");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0.055,-0.541,-0.185,");
    add_proj4text (p, 1, "0.0183,-0.0003,-0.007,-0.014 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"EST92\",DATUM[\"Estonia_1992\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 1,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 2,
		 "],TOWGS84[0.055,-0.541,-0.185,0.0183,-0.0003,-0.007,-0.0");
    add_srs_wkt (p, 3,
		 "14],AUTHORITY[\"EPSG\",\"6133\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 5,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 6, "G\",\"4133\"]]");
    p = add_epsg_def (first, last, 4134, "epsg", 4134, "PSD93");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"PSD93\",DATUM[\"PDO_Survey_Datum_1993\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6134\"]],PRIMEM[\"");
    add_srs_wkt (p, 3,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 4,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 5, "THORITY[\"EPSG\",\"4134\"]]");
    p = add_epsg_def (first, last, 4135, "epsg", 4135, "Old Hawaiian");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk66 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Old Hawaiian\",DATUM[\"Old_Hawaiian\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6135\"]],PRIMEM[\"");
    add_srs_wkt (p, 3,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 4,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 5, "THORITY[\"EPSG\",\"4135\"]]");
    p = add_epsg_def (first, last, 4136, "epsg", 4136, "St. Lawrence Island");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk66 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"St. Lawrence Island\",DATUM[\"St_Lawrence_Islan");
    add_srs_wkt (p, 1,
		 "d\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982139006");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6136");
    add_srs_wkt (p, 3,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4136\"]]");
    p = add_epsg_def (first, last, 4137, "epsg", 4137, "St. Paul Island");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk66 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"St. Paul Island\",DATUM[\"St_Paul_Island\",SPHE");
    add_srs_wkt (p, 1,
		 "ROID[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6137\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5, "]],AUTHORITY[\"EPSG\",\"4137\"]]");
    p = add_epsg_def (first, last, 4138, "epsg", 4138, "St. George Island");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk66 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"St. George Island\",DATUM[\"St_George_Island\",");
    add_srs_wkt (p, 1,
		 "SPHEROID[\"Clarke 1866\",6378206.4,294.9786982139006,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6138\"]]");
    add_srs_wkt (p, 3,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 4,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 5, "122\"]],AUTHORITY[\"EPSG\",\"4138\"]]");
    p = add_epsg_def (first, last, 4139, "epsg", 4139, "Puerto Rico");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk66 +towgs84=11,72,-101,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Puerto Rico\",DATUM[\"Puerto_Rico\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EP");
    add_srs_wkt (p, 2,
		 "SG\",\"7008\"]],TOWGS84[11,72,-101,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6139\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4139\"]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_15 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 4140, "epsg", 4140, "NAD83(CSRS98)");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"NAD83(CSRS98)\",DATUM[\"NAD83_Canadian_Spatial_");
    add_srs_wkt (p, 1,
		 "Reference_System\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 2,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 3,
		 ",0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 5,
		 "925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6, "\",\"4140\"]]");
    p = add_epsg_def (first, last, 4141, "epsg", 4141, "Israel");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=-48,55,52,0,0,0,0 +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Israel\",DATUM[\"Israel\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 1,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOW");
    add_srs_wkt (p, 2,
		 "GS84[-48,55,52,0,0,0,0],AUTHORITY[\"EPSG\",\"6141\"]],PR");
    add_srs_wkt (p, 3,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 4,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 5, "\"]],AUTHORITY[\"EPSG\",\"4141\"]]");
    p = add_epsg_def (first, last, 4142, "epsg", 4142, "Locodjo 1965");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=-125,53,467,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Locodjo 1965\",DATUM[\"Locodjo_1965\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EP");
    add_srs_wkt (p, 2,
		 "SG\",\"7012\"]],TOWGS84[-125,53,467,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6142\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4142\"]]");
    p = add_epsg_def (first, last, 4143, "epsg", 4143, "Abidjan 1987");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=-124.76,53,466.79,0");
    add_proj4text (p, 1, ",0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Abidjan 1987\",DATUM[\"Abidjan_1987\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EP");
    add_srs_wkt (p, 2,
		 "SG\",\"7012\"]],TOWGS84[-124.76,53,466.79,0,0,0,0],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6143\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"414");
    add_srs_wkt (p, 6, "3\"]]");
    p = add_epsg_def (first, last, 4144, "epsg", 4144, "Kalianpur 1937");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377276.345 +b=6356075.41314024 +no_def");
    add_proj4text (p, 1, "s");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Kalianpur 1937\",DATUM[\"Kalianpur_1937\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"Everest 1830 (1937 Adjustment)\",6377276.345,300.8");
    add_srs_wkt (p, 2,
		 "017,AUTHORITY[\"EPSG\",\"7015\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "144\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5, "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4144\"]]");
    p = add_epsg_def (first, last, 4145, "epsg", 4145, "Kalianpur 1962");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377301.243 +b=6356100.230165384 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Kalianpur 1962\",DATUM[\"Kalianpur_1962\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"Everest 1830 (1962 Definition)\",6377301.243,300.8");
    add_srs_wkt (p, 2,
		 "017255,AUTHORITY[\"EPSG\",\"7044\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"6145\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 4,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4145\"]]");
    p = add_epsg_def (first, last, 4146, "epsg", 4146, "Kalianpur 1975");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377299.151 +b=6356098.145120132 +towgs");
    add_proj4text (p, 1, "84=295,736,257,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Kalianpur 1975\",DATUM[\"Kalianpur_1975\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"Everest 1830 (1975 Definition)\",6377299.151,300.8");
    add_srs_wkt (p, 2,
		 "017255,AUTHORITY[\"EPSG\",\"7045\"]],TOWGS84[295,736,257");
    add_srs_wkt (p, 3,
		 ",0,0,0,0],AUTHORITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwi");
    add_srs_wkt (p, 4,
		 "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0");
    add_srs_wkt (p, 5,
		 "1745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY");
    add_srs_wkt (p, 6, "[\"EPSG\",\"4146\"]]");
    p = add_epsg_def (first, last, 4147, "epsg", 4147, "Hanoi 1972");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=krass +towgs84=-17.51,-108.32,-62.3");
    add_proj4text (p, 1, "9,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Hanoi 1972\",DATUM[\"Hanoi_1972\",SPHEROID[\"Kr");
    add_srs_wkt (p, 1,
		 "assowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"");
    add_srs_wkt (p, 2,
		 "]],TOWGS84[-17.51,-108.32,-62.39,0,0,0,0],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6147\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4147\"]]");
    p = add_epsg_def (first, last, 4148, "epsg", 4148, "Hartebeesthoek94");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Hartebeesthoek94\",DATUM[\"Hartebeesthoek94\",S");
    add_srs_wkt (p, 1,
		 "PHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6148\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4148\"]]");
    p = add_epsg_def (first, last, 4149, "epsg", 4149, "CH1903");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=bessel +towgs84=674.374,15.056,405.");
    add_proj4text (p, 1, "346,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"CH1903\",DATUM[\"CH1903\",SPHEROID[\"Bessel 184");
    add_srs_wkt (p, 1,
		 "1\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]");
    add_srs_wkt (p, 2,
		 "],TOWGS84[674.374,15.056,405.346,0,0,0,0],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6149\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4149\"]]");
    p = add_epsg_def (first, last, 4150, "epsg", 4150, "CH1903+");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=bessel +towgs84=674.374,15.056,405.");
    add_proj4text (p, 1, "346,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"CH1903+\",DATUM[\"CH1903\",SPHEROID[\"Bessel 18");
    add_srs_wkt (p, 1,
		 "41\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"");
    add_srs_wkt (p, 2,
		 "]],TOWGS84[674.374,15.056,405.346,0,0,0,0],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6150\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5, "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4150\"]]");
    p = add_epsg_def (first, last, 4151, "epsg", 4151, "CHTRF95");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"CHTRF95\",DATUM[\"Swiss_Terrestrial_Reference_F");
    add_srs_wkt (p, 1,
		 "rame_1995\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6151\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6, "151\"]]");
    p = add_epsg_def (first, last, 4152, "epsg", 4152, "NAD83(HARN)");
    add_proj4text (p, 0, "+proj=longlat +ellps=GRS80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regio");
    add_srs_wkt (p, 1,
		 "nal_Network\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 2,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"615");
    add_srs_wkt (p, 3,
		 "2\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4152\"]]");
    p = add_epsg_def (first, last, 4153, "epsg", 4153, "Rassadiran");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-133.63,-157.5,-158.6");
    add_proj4text (p, 1, "2,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Rassadiran\",DATUM[\"Rassadiran\",SPHEROID[\"In");
    add_srs_wkt (p, 1,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 2,
		 "\"]],TOWGS84[-133.63,-157.5,-158.62,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6153\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4153\"]]");
    p = add_epsg_def (first, last, 4154, "epsg", 4154, "ED50(ED77)");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"ED50(ED77)\",DATUM[\"European_Datum_1950_1977\"");
    add_srs_wkt (p, 1,
		 ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6154\"]],PRIMEM[\"");
    add_srs_wkt (p, 3,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 4,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 5, "THORITY[\"EPSG\",\"4154\"]]");
    p = add_epsg_def (first, last, 4155, "epsg", 4155, "Dabola 1981");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378249.2 +b=6356515 +towgs84=-83,37,12");
    add_proj4text (p, 1, "4,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Dabola 1981\",DATUM[\"Dabola_1981\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "Clarke 1880 (IGN)\",6378249.2,293.4660212936265,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7011\"]],TOWGS84[-83,37,124,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6155\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4155");
    add_srs_wkt (p, 6, "\"]]");
    p = add_epsg_def (first, last, 4156, "epsg", 4156, "S-JTSK");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"S-JTSK\",DATUM[\"Jednotne_Trigonometricke_Site_");
    add_srs_wkt (p, 1,
		 "Katastralni\",SPHEROID[\"Bessel 1841\",6377397.155,299.1");
    add_srs_wkt (p, 2,
		 "528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"6156\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 4,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4156\"]]");
    p = add_epsg_def (first, last, 4157, "epsg", 4157, "Mount Dillon");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378293.645208759 +b=6356617.987679838 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Mount Dillon\",DATUM[\"Mount_Dillon\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"Clarke 1858\",6378293.645208759,294.2606763692569,AUTH");
    add_srs_wkt (p, 2,
		 "ORITY[\"EPSG\",\"7007\"]],AUTHORITY[\"EPSG\",\"6157\"]],");
    add_srs_wkt (p, 3,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 4,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 5, "22\"]],AUTHORITY[\"EPSG\",\"4157\"]]");
    p = add_epsg_def (first, last, 4158, "epsg", 4158, "Naparima 1955");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Naparima 1955\",DATUM[\"Naparima_1955\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 2,
		 "\"7022\"]],AUTHORITY[\"EPSG\",\"6158\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 3,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 4,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"4158\"]]");
    p = add_epsg_def (first, last, 4159, "epsg", 4159, "ELD79");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"ELD79\",DATUM[\"European_Libyan_Datum_1979\",SP");
    add_srs_wkt (p, 1,
		 "HEROID[\"International 1924\",6378388,297,AUTHORITY[\"EP");
    add_srs_wkt (p, 2,
		 "SG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6159\"]],PRIMEM[\"G");
    add_srs_wkt (p, 3,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 4,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 5, "HORITY[\"EPSG\",\"4159\"]]");
    p = add_epsg_def (first, last, 4160, "epsg", 4160, "Chos Malal 1914");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Chos Malal 1914\",DATUM[\"Chos_Malal_1914\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"International 1924\",6378388,297,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7022\"]],AUTHORITY[\"EPSG\",\"6160\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 3,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 4,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 5, "RITY[\"EPSG\",\"4160\"]]");
    p = add_epsg_def (first, last, 4161, "epsg", 4161, "Pampa del Castillo");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=27.5,14,186.4,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Pampa del Castillo\",DATUM[\"Pampa_del_Castillo");
    add_srs_wkt (p, 1,
		 "\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY");
    add_srs_wkt (p, 2,
		 "[\"EPSG\",\"7022\"]],TOWGS84[27.5,14,186.4,0,0,0,0],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6161\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"41");
    add_srs_wkt (p, 6, "61\"]]");
    p = add_epsg_def (first, last, 4162, "epsg", 4162, "Korean 1985");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Korean 1985\",DATUM[\"Korean_Datum_1985\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6162\"]],PRIMEM[\"");
    add_srs_wkt (p, 3,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 4,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 5, "THORITY[\"EPSG\",\"4162\"]]");
    p = add_epsg_def (first, last, 4163, "epsg", 4163, "Yemen NGN96");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Yemen NGN96\",DATUM[\"Yemen_National_Geodetic_N");
    add_srs_wkt (p, 1,
		 "etwork_1996\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6163\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6, "163\"]]");
    p = add_epsg_def (first, last, 4164, "epsg", 4164, "South Yemen");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=krass +towgs84=-76,-138,67,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"South Yemen\",DATUM[\"South_Yemen\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 2,
		 "4\"]],TOWGS84[-76,-138,67,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "6164\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 4,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 5, "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4164\"]]");
    p = add_epsg_def (first, last, 4165, "epsg", 4165, "Bissau");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-173,253,27,0,0,0,0 +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Bissau\",DATUM[\"Bissau\",SPHEROID[\"Internatio");
    add_srs_wkt (p, 1,
		 "nal 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOW");
    add_srs_wkt (p, 2,
		 "GS84[-173,253,27,0,0,0,0],AUTHORITY[\"EPSG\",\"6165\"]],");
    add_srs_wkt (p, 3,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 4,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 5, "22\"]],AUTHORITY[\"EPSG\",\"4165\"]]");
    p = add_epsg_def (first, last, 4166, "epsg", 4166, "Korean 1995");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Korean 1995\",DATUM[\"Korean_Datum_1995\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 2,
		 "\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "166\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5, "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4166\"]]");
    p = add_epsg_def (first, last, 4167, "epsg", 4167, "NZGD2000");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"NZGD2000\",DATUM[\"New_Zealand_Geodetic_Datum_2");
    add_srs_wkt (p, 1,
		 "000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR");
    add_srs_wkt (p, 2,
		 "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4167\"]");
    add_srs_wkt (p, 6, "]");
    p = add_epsg_def (first, last, 4168, "epsg", 4168, "Accra");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378300 +b=6356751.689189189 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Accra\",DATUM[\"Accra\",SPHEROID[\"War Office\"");
    add_srs_wkt (p, 1,
		 ",6378300,296,AUTHORITY[\"EPSG\",\"7029\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 2,
		 "PSG\",\"6168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 4, "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4168\"]]");
    p = add_epsg_def (first, last, 4169, "epsg", 4169, "American Samoa 1962");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk66 +towgs84=-115,118,426,0,0,0,");
    add_proj4text (p, 1, "0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"American Samoa 1962\",DATUM[\"American_Samoa_19");
    add_srs_wkt (p, 1,
		 "62\",SPHEROID[\"Clarke 1866\",6378206.4,294.978698213900");
    add_srs_wkt (p, 2,
		 "6,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-115,118,426,0,0");
    add_srs_wkt (p, 3,
		 ",0,0],AUTHORITY[\"EPSG\",\"6169\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6, "PSG\",\"4169\"]]");
    p = add_epsg_def (first, last, 4170, "epsg", 4170, "SIRGAS 1995");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"SIRGAS 1995\",DATUM[\"Sistema_de_Referencia_Geo");
    add_srs_wkt (p, 1,
		 "centrico_para_America_del_Sur_1995\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6170\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6, "]],AUTHORITY[\"EPSG\",\"4170\"]]");
    p = add_epsg_def (first, last, 4171, "epsg", 4171, "RGF93");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"RGF93\",DATUM[\"Reseau_Geodesique_Francais_1993");
    add_srs_wkt (p, 1,
		 "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY");
    add_srs_wkt (p, 2,
		 "[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6171\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4171\"]]");
    p = add_epsg_def (first, last, 4172, "epsg", 4172, "POSGAR");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"POSGAR\",DATUM[\"Posiciones_Geodesicas_Argentin");
    add_srs_wkt (p, 1,
		 "as\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6172\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199433,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"4172\"]]");
    p = add_epsg_def (first, last, 4173, "epsg", 4173, "IRENET95");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"IRENET95\",DATUM[\"IRENET95\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 1,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 2,
		 ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6173\"]],PR");
    add_srs_wkt (p, 3,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 4,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 5, "\"]],AUTHORITY[\"EPSG\",\"4173\"]]");
    p = add_epsg_def (first, last, 4174, "epsg", 4174, "Sierra Leone 1924");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378300 +b=6356751.689189189 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Sierra Leone 1924\",DATUM[\"Sierra_Leone_Colony");
    add_srs_wkt (p, 1,
		 "_1924\",SPHEROID[\"War Office\",6378300,296,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7029\"]],AUTHORITY[\"EPSG\",\"6174\"]],PRIMEM[\"");
    add_srs_wkt (p, 3,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 4,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 5, "THORITY[\"EPSG\",\"4174\"]]");
    p = add_epsg_def (first, last, 4175, "epsg", 4175, "Sierra Leone 1968");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=-88,4,101,0,0,0,0 +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Sierra Leone 1968\",DATUM[\"Sierra_Leone_1968\"");
    add_srs_wkt (p, 1,
		 ",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTH");
    add_srs_wkt (p, 2,
		 "ORITY[\"EPSG\",\"7012\"]],TOWGS84[-88,4,101,0,0,0,0],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6175\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6, "175\"]]");
    p = add_epsg_def (first, last, 4176, "epsg", 4176, "Australian Antarctic");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Australian Antarctic\",DATUM[\"Australian_Antar");
    add_srs_wkt (p, 1,
		 "ctic_Datum_1998\",SPHEROID[\"GRS 1980\",6378137,298.2572");
    add_srs_wkt (p, 2,
		 "22101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,");
    add_srs_wkt (p, 3,
		 "0],AUTHORITY[\"EPSG\",\"6176\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6, "\",\"4176\"]]");
    p = add_epsg_def (first, last, 4178, "epsg", 4178, "Pulkovo 1942(83)");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=krass +towgs84=24,-123,-94,0.02,-0.");
    add_proj4text (p, 1, "25,-0.13,1.1 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo_1942_83\",SP");
    add_srs_wkt (p, 1,
		 "HEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7024\"]],AUTHORITY[\"EPSG\",\"6178\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 3,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 4,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 5, "RITY[\"EPSG\",\"4178\"]]");
    p = add_epsg_def (first, last, 4179, "epsg", 4179, "Pulkovo 1942(58)");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=krass +towgs84=33.4,-146.6,-76.3,-0");
    add_proj4text (p, 1, ".359,-0.053,0.844,-0.84 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo_1942_58\",SP");
    add_srs_wkt (p, 1,
		 "HEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7024\"]],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 3,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 4,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 5, "RITY[\"EPSG\",\"4179\"]]");
    p = add_epsg_def (first, last, 4180, "epsg", 4180, "EST97");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"EST97\",DATUM[\"Estonia_1997\",SPHEROID[\"GRS 1");
    add_srs_wkt (p, 1,
		 "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]");
    add_srs_wkt (p, 2,
		 "],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6180\"]],P");
    add_srs_wkt (p, 3,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 4,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 5, "2\"]],AUTHORITY[\"EPSG\",\"4180\"]]");
    p = add_epsg_def (first, last, 4181, "epsg", 4181, "Luxembourg 1930");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-193,13.7,-39.3,-0.41");
    add_proj4text (p, 1, ",-2.933,2.688,0.43 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Luxembourg 1930\",DATUM[\"Luxembourg_1930\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"International 1924\",6378388,297,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7022\"]],TOWGS84[-193,13.7,-39.3,-0.41,-2.933,2.68");
    add_srs_wkt (p, 3,
		 "8,0.43],AUTHORITY[\"EPSG\",\"6181\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6, "EPSG\",\"4181\"]]");
    p = add_epsg_def (first, last, 4182, "epsg", 4182,
		      "Azores Occidental 1939");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Azores Occidental 1939\",DATUM[\"Azores_Occiden");
    add_srs_wkt (p, 1,
		 "tal_Islands_1939\",SPHEROID[\"International 1924\",63783");
    add_srs_wkt (p, 2,
		 "88,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"6182\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 4,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4182\"]]");
    p = add_epsg_def (first, last, 4183, "epsg", 4183, "Azores Central 1948");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-104,167,-38,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Azores Central 1948\",DATUM[\"Azores_Central_Is");
    add_srs_wkt (p, 1,
		 "lands_1948\",SPHEROID[\"International 1924\",6378388,297");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-104,167,-38,0,0,");
    add_srs_wkt (p, 3,
		 "0,0],AUTHORITY[\"EPSG\",\"6183\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6, "SG\",\"4183\"]]");
    p = add_epsg_def (first, last, 4184, "epsg", 4184, "Azores Oriental 1940");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-203,141,53,0,0,0,0 +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Azores Oriental 1940\",DATUM[\"Azores_Oriental_");
    add_srs_wkt (p, 1,
		 "Islands_1940\",SPHEROID[\"International 1924\",6378388,2");
    add_srs_wkt (p, 2,
		 "97,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-203,141,53,0,0");
    add_srs_wkt (p, 3,
		 ",0,0],AUTHORITY[\"EPSG\",\"6184\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6, "PSG\",\"4184\"]]");
    p = add_epsg_def (first, last, 4185, "epsg", 4185, "Madeira 1936");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Madeira 1936\",DATUM[\"Madeira_1936\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7022\"]],AUTHORITY[\"EPSG\",\"6185\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 3,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 4,
		 "74532925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"4185\"]]");
    p = add_epsg_def (first, last, 4188, "epsg", 4188, "OSNI 1952");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=airy +towgs84=482.5,-130.6,564.6,-1");
    add_proj4text (p, 1, ".042,-0.214,-0.631,8.15 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"OSNI 1952\",DATUM[\"OSNI_1952\",SPHEROID[\"Airy");
    add_srs_wkt (p, 1,
		 " 1830\",6377563.396,299.3249646,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 2,
		 "1\"]],TOWGS84[482.5,-130.6,564.6,-1.042,-0.214,-0.631,8.");
    add_srs_wkt (p, 3,
		 "15],AUTHORITY[\"EPSG\",\"6188\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 5,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 6, "G\",\"4188\"]]");
    p = add_epsg_def (first, last, 4189, "epsg", 4189, "REGVEN");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"REGVEN\",DATUM[\"Red_Geodesica_Venezolana\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6189\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4189\"]]");
    p = add_epsg_def (first, last, 4190, "epsg", 4190, "POSGAR 98");
    add_proj4text (p, 0, "+proj=longlat +ellps=GRS80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"POSGAR 98\",DATUM[\"Posiciones_Geodesicas_Argen");
    add_srs_wkt (p, 1,
		 "tinas_1998\",SPHEROID[\"GRS 1980\",6378137,298.257222101");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6190");
    add_srs_wkt (p, 3,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4190\"]]");
    p = add_epsg_def (first, last, 4191, "epsg", 4191, "Albanian 1987");
    add_proj4text (p, 0, "+proj=longlat +ellps=krass +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Albanian 1987\",DATUM[\"Albanian_1987\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7024\"]],AUTHORITY[\"EPSG\",\"6191\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 3,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 4,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 5, "\"EPSG\",\"4191\"]]");
    p = add_epsg_def (first, last, 4192, "epsg", 4192, "Douala 1948");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-206.1,-174.7,-87.7,0");
    add_proj4text (p, 1, ",0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Douala 1948\",DATUM[\"Douala_1948\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 2,
		 "22\"]],TOWGS84[-206.1,-174.7,-87.7,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6192\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4192\"]]");
    p = add_epsg_def (first, last, 4193, "epsg", 4193, "Manoca 1962");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378249.2 +b=6356515 +towgs84=-70.9,-15");
    add_proj4text (p, 1, "1.8,-41.4,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Manoca 1962\",DATUM[\"Manoca_1962\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "Clarke 1880 (IGN)\",6378249.2,293.4660212936265,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7011\"]],TOWGS84[-70.9,-151.8,-41.4,0,0,0,0");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6193\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6, ",\"4193\"]]");
    p = add_epsg_def (first, last, 4194, "epsg", 4194, "Qornoq 1927");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Qornoq 1927\",DATUM[\"Qornoq_1927\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 2,
		 "22\"]],AUTHORITY[\"EPSG\",\"6194\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 3,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 4,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 5, "PSG\",\"4194\"]]");
    p = add_epsg_def (first, last, 4195, "epsg", 4195, "Scoresbysund 1952");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=105,326,-102.5,0,0,0.");
    add_proj4text (p, 1, "814,-0.6 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Scoresbysund 1952\",DATUM[\"Scoresbysund_1952\"");
    add_srs_wkt (p, 1,
		 ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7022\"]],TOWGS84[105,326,-102.5,0,0,0.814,-0.6]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6195\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6, ",\"4195\"]]");
    p = add_epsg_def (first, last, 4196, "epsg", 4196, "Ammassalik 1958");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-45,417,-3.5,0,0,0.81");
    add_proj4text (p, 1, "4,-0.6 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Ammassalik 1958\",DATUM[\"Ammassalik_1958\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"International 1924\",6378388,297,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7022\"]],TOWGS84[-45,417,-3.5,0,0,0.814,-0.6],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6196\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"41");
    add_srs_wkt (p, 6, "96\"]]");
    p = add_epsg_def (first, last, 4197, "epsg", 4197, "Garoua");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Garoua\",DATUM[\"Garoua\",SPHEROID[\"Clarke 188");
    add_srs_wkt (p, 1,
		 "0 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"");
    add_srs_wkt (p, 2,
		 "]],AUTHORITY[\"EPSG\",\"6197\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 4,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 5, "\",\"4197\"]]");
    p = add_epsg_def (first, last, 4198, "epsg", 4198, "Kousseri");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Kousseri\",DATUM[\"Kousseri\",SPHEROID[\"Clarke");
    add_srs_wkt (p, 1,
		 " 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 2,
		 "012\"]],AUTHORITY[\"EPSG\",\"6198\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 3,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 4,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"4198\"]]");
    p = add_epsg_def (first, last, 4199, "epsg", 4199, "Egypt 1930");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Egypt 1930\",DATUM[\"Egypt_1930\",SPHEROID[\"In");
    add_srs_wkt (p, 1,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 2,
		 "\"]],AUTHORITY[\"EPSG\",\"6199\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 3,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 4,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 5, "SG\",\"4199\"]]");
    p = add_epsg_def (first, last, 4200, "epsg", 4200, "Pulkovo 1995");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=krass +towgs84=24.82,-131.21,-82.66");
    add_proj4text (p, 1, ",-0,-0,0.16,-0.12 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 2,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6, "4200\"]]");
    p = add_epsg_def (first, last, 4201, "epsg", 4201, "Adindan");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Adindan\",DATUM[\"Adindan\",SPHEROID[\"Clarke 1");
    add_srs_wkt (p, 1,
		 "880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 2,
		 "2\"]],AUTHORITY[\"EPSG\",\"6201\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 3,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 4,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 5, "PSG\",\"4201\"]]");
    p = add_epsg_def (first, last, 4202, "epsg", 4202, "AGD66");
    add_proj4text (p, 0, "+proj=longlat +ellps=aust_SA +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"AGD66\",DATUM[\"Australian_Geodetic_Datum_1966\"");
    add_srs_wkt (p, 1,
		 ",SPHEROID[\"Australian National Spheroid\",6378160,298.2");
    add_srs_wkt (p, 2,
		 "5,AUTHORITY[\"EPSG\",\"7003\"]],AUTHORITY[\"EPSG\",\"620");
    add_srs_wkt (p, 3,
		 "2\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4202\"]]");
    p = add_epsg_def (first, last, 4203, "epsg", 4203, "AGD84");
    add_proj4text (p, 0, "+proj=longlat +ellps=aust_SA +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"AGD84\",DATUM[\"Australian_Geodetic_Datum_1984\"");
    add_srs_wkt (p, 1,
		 ",SPHEROID[\"Australian National Spheroid\",6378160,298.2");
    add_srs_wkt (p, 2,
		 "5,AUTHORITY[\"EPSG\",\"7003\"]],AUTHORITY[\"EPSG\",\"620");
    add_srs_wkt (p, 3,
		 "3\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4203\"]]");
    p = add_epsg_def (first, last, 4204, "epsg", 4204, "Ain el Abd");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Ain el Abd\",DATUM[\"Ain_el_Abd_1970\",SPHEROID");
    add_srs_wkt (p, 1,
		 "[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7022\"]],AUTHORITY[\"EPSG\",\"6204\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 3,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 4,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 5, "\"EPSG\",\"4204\"]]");
    p = add_epsg_def (first, last, 4205, "epsg", 4205, "Afgooye");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=krass +towgs84=-43,-163,45,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Afgooye\",DATUM[\"Afgooye\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 1,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 2,
		 "WGS84[-43,-163,45,0,0,0,0],AUTHORITY[\"EPSG\",\"6205\"]]");
    add_srs_wkt (p, 3,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 4,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 5, "122\"]],AUTHORITY[\"EPSG\",\"4205\"]]");
    p = add_epsg_def (first, last, 4206, "epsg", 4206, "Agadez");
    add_proj4text (p, 0, "+proj=longlat +a=6378249.2 +b=6356515 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Agadez\",DATUM[\"Agadez\",SPHEROID[\"Clarke 188");
    add_srs_wkt (p, 1,
		 "0 (IGN)\",6378249.2,293.4660212936265,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 2,
		 ",\"7011\"]],AUTHORITY[\"EPSG\",\"6206\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 3,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 4,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 5, "TY[\"EPSG\",\"4206\"]]");
    p = add_epsg_def (first, last, 4207, "epsg", 4207, "Lisbon");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Lisbon\",DATUM[\"Lisbon_1937\",SPHEROID[\"Inter");
    add_srs_wkt (p, 1,
		 "national 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]");
    add_srs_wkt (p, 2,
		 "],AUTHORITY[\"EPSG\",\"6207\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 4,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"4207\"]]");
    p = add_epsg_def (first, last, 4208, "epsg", 4208, "Aratu");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Aratu\",DATUM[\"Aratu\",SPHEROID[\"Internationa");
    add_srs_wkt (p, 1,
		 "l 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHO");
    add_srs_wkt (p, 2,
		 "RITY[\"EPSG\",\"6208\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 4,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"420");
    add_srs_wkt (p, 5, "8\"]]");
    p = add_epsg_def (first, last, 4209, "epsg", 4209, "Arc 1950");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378249.145 +b=6356514.966398753 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Arc 1950\",DATUM[\"Arc_1950\",SPHEROID[\"Clarke");
    add_srs_wkt (p, 1,
		 " 1880 (Arc)\",6378249.145,293.4663077,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 2,
		 ",\"7013\"]],AUTHORITY[\"EPSG\",\"6209\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 3,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 4,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 5, "TY[\"EPSG\",\"4209\"]]");
    p = add_epsg_def (first, last, 4210, "epsg", 4210, "Arc 1960");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Arc 1960\",DATUM[\"Arc_1960\",SPHEROID[\"Clarke");
    add_srs_wkt (p, 1,
		 " 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 2,
		 "012\"]],AUTHORITY[\"EPSG\",\"6210\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 3,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 4,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"4210\"]]");
    p = add_epsg_def (first, last, 4211, "epsg", 4211, "Batavia");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Batavia\",DATUM[\"Batavia\",SPHEROID[\"Bessel 1");
    add_srs_wkt (p, 1,
		 "841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"");
    add_srs_wkt (p, 2,
		 "]],AUTHORITY[\"EPSG\",\"6211\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 4,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 5, "\",\"4211\"]]");
    p = add_epsg_def (first, last, 4212, "epsg", 4212, "Barbados 1938");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=31.95,300.99,419.19");
    add_proj4text (p, 1, ",0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Barbados 1938\",DATUM[\"Barbados_1938\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7012\"]],TOWGS84[31.95,300.99,419.19,0,0,0,0],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6212\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6, "4212\"]]");
    p = add_epsg_def (first, last, 4213, "epsg", 4213, "Beduaram");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378249.2 +b=6356515 +towgs84=-106,-87,");
    add_proj4text (p, 1, "188,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Beduaram\",DATUM[\"Beduaram\",SPHEROID[\"Clarke");
    add_srs_wkt (p, 1,
		 " 1880 (IGN)\",6378249.2,293.4660212936265,AUTHORITY[\"EP");
    add_srs_wkt (p, 2,
		 "SG\",\"7011\"]],TOWGS84[-106,-87,188,0,0,0,0],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6213\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4213\"]]");
    add_srs_wkt (p, 6, "");
    p = add_epsg_def (first, last, 4214, "epsg", 4214, "Beijing 1954");
    add_proj4text (p, 0, "+proj=longlat +ellps=krass +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 2,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 3,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 4,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"4214\"]]");
    p = add_epsg_def (first, last, 4215, "epsg", 4215, "Belge 1950");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Belge 1950\",DATUM[\"Reseau_National_Belge_1950");
    add_srs_wkt (p, 1,
		 "\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY");
    add_srs_wkt (p, 2,
		 "[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6215\"]],PRIME");
    add_srs_wkt (p, 3,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 4,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 5, "],AUTHORITY[\"EPSG\",\"4215\"]]");
    p = add_epsg_def (first, last, 4216, "epsg", 4216, "Bermuda 1957");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk66 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Bermuda 1957\",DATUM[\"Bermuda_1957\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6216\"]],PRIMEM[\"");
    add_srs_wkt (p, 3,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 4,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 5, "THORITY[\"EPSG\",\"4216\"]]");
    p = add_epsg_def (first, last, 4218, "epsg", 4218, "Bogota 1975");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=307,304,-318,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Bogota 1975\",DATUM[\"Bogota_1975\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 2,
		 "22\"]],TOWGS84[307,304,-318,0,0,0,0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"6218\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 4,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4218\"]]");
    p = add_epsg_def (first, last, 4219, "epsg", 4219, "Bukit Rimpah");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=bessel +towgs84=-384,664,-48,0,0,0,");
    add_proj4text (p, 1, "0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Bukit Rimpah\",DATUM[\"Bukit_Rimpah\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 2,
		 "\",\"7004\"]],TOWGS84[-384,664,-48,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6219\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4219\"]]");
    p = add_epsg_def (first, last, 4220, "epsg", 4220, "Camacupa");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Camacupa\",DATUM[\"Camacupa\",SPHEROID[\"Clarke");
    add_srs_wkt (p, 1,
		 " 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 2,
		 "012\"]],AUTHORITY[\"EPSG\",\"6220\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 3,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 4,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"4220\"]]");
    p = add_epsg_def (first, last, 4221, "epsg", 4221, "Campo Inchauspe");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Campo Inchauspe\",DATUM[\"Campo_Inchauspe\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"International 1924\",6378388,297,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7022\"]],AUTHORITY[\"EPSG\",\"6221\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 3,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 4,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 5, "RITY[\"EPSG\",\"4221\"]]");
    p = add_epsg_def (first, last, 4222, "epsg", 4222, "Cape");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378249.145 +b=6356514.966398753 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Cape\",DATUM[\"Cape\",SPHEROID[\"Clarke 1880 (A");
    add_srs_wkt (p, 1,
		 "rc)\",6378249.145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"");
    add_srs_wkt (p, 2,
		 "]],AUTHORITY[\"EPSG\",\"6222\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 4,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 5, "\",\"4222\"]]");
    p = add_epsg_def (first, last, 4223, "epsg", 4223, "Carthage");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378249.2 +b=6356515 +datum=carthage +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Carthage\",DATUM[\"Carthage\",SPHEROID[\"Clarke");
    add_srs_wkt (p, 1,
		 " 1880 (IGN)\",6378249.2,293.4660212936265,AUTHORITY[\"EP");
    add_srs_wkt (p, 2,
		 "SG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6223\"]],PRIMEM[\"G");
    add_srs_wkt (p, 3,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 4,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 5, "HORITY[\"EPSG\",\"4223\"]]");
    p = add_epsg_def (first, last, 4224, "epsg", 4224, "Chua");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Chua\",DATUM[\"Chua\",SPHEROID[\"International ");
    add_srs_wkt (p, 1,
		 "1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"6224\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4224\"");
    add_srs_wkt (p, 5, "]]");
    p = add_epsg_def (first, last, 4225, "epsg", 4225, "Corrego Alegre");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-206,172,-6,0,0,0,0 +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Corrego Alegre\",DATUM[\"Corrego_Alegre\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 2,
		 ",\"7022\"]],TOWGS84[-206,172,-6,0,0,0,0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"6225\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4225\"]]");
    p = add_epsg_def (first, last, 4226, "epsg", 4226, "Cote d'Ivoire");
    add_proj4text (p, 0, "+proj=longlat +a=6378249.2 +b=6356515 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Cote d'Ivoire\",DATUM[\"Cote_d_Ivoire\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936265,AUTH");
    add_srs_wkt (p, 2,
		 "ORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6226\"]],");
    add_srs_wkt (p, 3,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 4,
		 "T[\"degree\",0.0174532925199433,AUTHORITY[\"EPSG\",\"910");
    add_srs_wkt (p, 5, "8\"]],AUTHORITY[\"EPSG\",\"4226\"]]");
    p = add_epsg_def (first, last, 4227, "epsg", 4227, "Deir ez Zor");
    add_proj4text (p, 0, "+proj=longlat +a=6378249.2 +b=6356515 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Deir ez Zor\",DATUM[\"Deir_ez_Zor\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "Clarke 1880 (IGN)\",6378249.2,293.4660212936265,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6227\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5, "]],AUTHORITY[\"EPSG\",\"4227\"]]");
    p = add_epsg_def (first, last, 4228, "epsg", 4228, "Douala");
    add_proj4text (p, 0, "+proj=longlat +a=6378249.2 +b=6356515 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Douala\",DATUM[\"Douala\",SPHEROID[\"Clarke 188");
    add_srs_wkt (p, 1,
		 "0 (IGN)\",6378249.2,293.4660212936265,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 2,
		 ",\"7011\"]],AUTHORITY[\"EPSG\",\"6228\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 3,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 4,
		 ".0174532925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"4228\"]]");
    p = add_epsg_def (first, last, 4229, "epsg", 4229, "Egypt 1907");
    add_proj4text (p, 0, "+proj=longlat +ellps=helmert +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Egypt 1907\",DATUM[\"Egypt_1907\",SPHEROID[\"He");
    add_srs_wkt (p, 1,
		 "lmert 1906\",6378200,298.3,AUTHORITY[\"EPSG\",\"7020\"]]");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"6229\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 4,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"4229\"]]");
    p = add_epsg_def (first, last, 4230, "epsg", 4230, "ED50");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"ED50\",DATUM[\"European_Datum_1950\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 2,
		 "22\"]],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 3,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 4,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 5, "PSG\",\"4230\"]]");
    p = add_epsg_def (first, last, 4231, "epsg", 4231, "ED87");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"ED87\",DATUM[\"European_Datum_1987\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 2,
		 "22\"]],AUTHORITY[\"EPSG\",\"6231\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 3,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 4,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 5, "PSG\",\"4231\"]]");
    p = add_epsg_def (first, last, 4232, "epsg", 4232, "Fahud");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Fahud\",DATUM[\"Fahud\",SPHEROID[\"Clarke 1880 ");
    add_srs_wkt (p, 1,
		 "(RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]");
    add_srs_wkt (p, 2,
		 "],AUTHORITY[\"EPSG\",\"6232\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 4,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"4232\"]]");
    p = add_epsg_def (first, last, 4233, "epsg", 4233, "Gandajika 1970");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-133,-321,50,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Gandajika 1970\",DATUM[\"Gandajika_1970\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 2,
		 ",\"7022\"]],TOWGS84[-133,-321,50,0,0,0,0],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6233\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4233\"]]");
    p = add_epsg_def (first, last, 4234, "epsg", 4234, "Garoua");
    add_proj4text (p, 0, "+proj=longlat +a=6378249.2 +b=6356515 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Garoua\",DATUM[\"Garoua\",SPHEROID[\"Clarke 188");
    add_srs_wkt (p, 1,
		 "0 (IGN)\",6378249.2,293.4660212936265,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 2,
		 ",\"7011\"]],AUTHORITY[\"EPSG\",\"6234\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 3,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 4,
		 ".0174532925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"4234\"]]");
    p = add_epsg_def (first, last, 4235, "epsg", 4235, "Guyane Francaise");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Guyane Francaise\",DATUM[\"Guyane_Francaise\",S");
    add_srs_wkt (p, 1,
		 "PHEROID[\"International 1924\",6378388,297,AUTHORITY[\"E");
    add_srs_wkt (p, 2,
		 "PSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6235\"]],PRIMEM[\"");
    add_srs_wkt (p, 3,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 4,
		 "e\",0.0174532925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUT");
    add_srs_wkt (p, 5, "HORITY[\"EPSG\",\"4235\"]]");
    p = add_epsg_def (first, last, 4236, "epsg", 4236, "Hu Tzu Shan 1950");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-637,-549,-203,0,0,0,");
    add_proj4text (p, 1, "0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Hu Tzu Shan 1950\",DATUM[\"Hu_Tzu_Shan_1950\",S");
    add_srs_wkt (p, 1,
		 "PHEROID[\"International 1924\",6378388,297,AUTHORITY[\"E");
    add_srs_wkt (p, 2,
		 "PSG\",\"7022\"]],TOWGS84[-637,-549,-203,0,0,0,0],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6236\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4236\"");
    add_srs_wkt (p, 6, "]]");
    p = add_epsg_def (first, last, 4237, "epsg", 4237, "HD72");
    add_proj4text (p, 0, "+proj=longlat +ellps=GRS67 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"HD72\",DATUM[\"Hungarian_Datum_1972\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"GRS 1967\",6378160,298.247167427,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7036\"]],AUTHORITY[\"EPSG\",\"6237\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 3,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 4,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 5, "\"EPSG\",\"4237\"]]");
    p = add_epsg_def (first, last, 4238, "epsg", 4238, "ID74");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378160 +b=6356774.50408554 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"ID74\",DATUM[\"Indonesian_Datum_1974\",SPHEROID");
    add_srs_wkt (p, 1,
		 "[\"Indonesian National Spheroid\",6378160,298.247,AUTHOR");
    add_srs_wkt (p, 2,
		 "ITY[\"EPSG\",\"7021\"]],AUTHORITY[\"EPSG\",\"6238\"]],PR");
    add_srs_wkt (p, 3,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 4,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 5, "\"]],AUTHORITY[\"EPSG\",\"4238\"]]");
    p = add_epsg_def (first, last, 4239, "epsg", 4239, "Indian 1954");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377276.345 +b=6356075.41314024 +towgs8");
    add_proj4text (p, 1, "4=217,823,299,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Indian 1954\",DATUM[\"Indian_1954\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AU");
    add_srs_wkt (p, 2,
		 "THORITY[\"EPSG\",\"7015\"]],TOWGS84[217,823,299,0,0,0,0]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6239\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6, ",\"4239\"]]");
    p = add_epsg_def (first, last, 4240, "epsg", 4240, "Indian 1975");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377276.345 +b=6356075.41314024 +no_def");
    add_proj4text (p, 1, "s");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Indian 1975\",DATUM[\"Indian_1975\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AU");
    add_srs_wkt (p, 2,
		 "THORITY[\"EPSG\",\"7015\"]],AUTHORITY[\"EPSG\",\"6240\"]");
    add_srs_wkt (p, 3,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 4,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5, "9122\"]],AUTHORITY[\"EPSG\",\"4240\"]]");
    p = add_epsg_def (first, last, 4241, "epsg", 4241, "Jamaica 1875");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378249.144808011 +b=6356514.966204134 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Jamaica 1875\",DATUM[\"Jamaica_1875\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"Clarke 1880\",6378249.144808011,293.4663076556349,AUTH");
    add_srs_wkt (p, 2,
		 "ORITY[\"EPSG\",\"7034\"]],AUTHORITY[\"EPSG\",\"6241\"]],");
    add_srs_wkt (p, 3,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 4,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 5, "22\"]],AUTHORITY[\"EPSG\",\"4241\"]]");
    p = add_epsg_def (first, last, 4242, "epsg", 4242, "JAD69");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk66 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"JAD69\",DATUM[\"Jamaica_1969\",SPHEROID[\"Clark");
    add_srs_wkt (p, 1,
		 "e 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 2,
		 "\"7008\"]],AUTHORITY[\"EPSG\",\"6242\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 3,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 4,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"4242\"]]");
    p = add_epsg_def (first, last, 4243, "epsg", 4243, "Kalianpur 1880");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377299.36559538 +b=6356098.359005156 +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Kalianpur 1880\",DATUM[\"Kalianpur_1880\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"Everest (1830 Definition)\",6377299.36559538,300.8");
    add_srs_wkt (p, 2,
		 "017255433552,AUTHORITY[\"EPSG\",\"7042\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6243\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5, "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4243\"]]");
    p = add_epsg_def (first, last, 4244, "epsg", 4244, "Kandawala");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377276.345 +b=6356075.41314024 +towgs8");
    add_proj4text (p, 1, "4=-97,787,86,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Kandawala\",DATUM[\"Kandawala\",SPHEROID[\"Ever");
    add_srs_wkt (p, 1,
		 "est 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHOR");
    add_srs_wkt (p, 2,
		 "ITY[\"EPSG\",\"7015\"]],TOWGS84[-97,787,86,0,0,0,0],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6244\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6, "44\"]]");
    p = add_epsg_def (first, last, 4245, "epsg", 4245, "Kertau 1968");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377304.063 +b=6356103.038993155 +towgs");
    add_proj4text (p, 1, "84=-11,851,5,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Kertau 1968\",DATUM[\"Kertau_1968\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "Everest 1830 Modified\",6377304.063,300.8017,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7018\"]],TOWGS84[-11,851,5,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6245\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4245\"]]");
    p = add_epsg_def (first, last, 4246, "epsg", 4246, "KOC");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=-294.7,-200.1,525.5");
    add_proj4text (p, 1, ",0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"KOC\",DATUM[\"Kuwait_Oil_Company\",SPHEROID[\"C");
    add_srs_wkt (p, 1,
		 "larke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 2,
		 ",\"7012\"]],TOWGS84[-294.7,-200.1,525.5,0,0,0,0],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6246\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4246\"");
    add_srs_wkt (p, 6, "]]");
    p = add_epsg_def (first, last, 4247, "epsg", 4247, "La Canoa");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-273.5,110.6,-357.9,0");
    add_proj4text (p, 1, ",0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"La Canoa\",DATUM[\"La_Canoa\",SPHEROID[\"Intern");
    add_srs_wkt (p, 1,
		 "ational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]]");
    add_srs_wkt (p, 2,
		 ",TOWGS84[-273.5,110.6,-357.9,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6247\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4247\"]]");
    p = add_epsg_def (first, last, 4248, "epsg", 4248, "PSAD56");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"PSAD56\",DATUM[\"Provisional_South_American_Dat");
    add_srs_wkt (p, 1,
		 "um_1956\",SPHEROID[\"International 1924\",6378388,297,AU");
    add_srs_wkt (p, 2,
		 "THORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6248\"]");
    add_srs_wkt (p, 3,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 4,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5, "9122\"]],AUTHORITY[\"EPSG\",\"4248\"]]");
    p = add_epsg_def (first, last, 4249, "epsg", 4249, "Lake");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Lake\",DATUM[\"Lake\",SPHEROID[\"International ");
    add_srs_wkt (p, 1,
		 "1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"6249\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4249\"");
    add_srs_wkt (p, 5, "]]");
    p = add_epsg_def (first, last, 4250, "epsg", 4250, "Leigon");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=-130,29,364,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Leigon\",DATUM[\"Leigon\",SPHEROID[\"Clarke 188");
    add_srs_wkt (p, 1,
		 "0 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"");
    add_srs_wkt (p, 2,
		 "]],TOWGS84[-130,29,364,0,0,0,0],AUTHORITY[\"EPSG\",\"625");
    add_srs_wkt (p, 3,
		 "0\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4250\"]]");
    p = add_epsg_def (first, last, 4251, "epsg", 4251, "Liberia 1964");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=-90,40,88,0,0,0,0 +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Liberia 1964\",DATUM[\"Liberia_1964\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EP");
    add_srs_wkt (p, 2,
		 "SG\",\"7012\"]],TOWGS84[-90,40,88,0,0,0,0],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6251\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5, "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4251\"]]");
    p = add_epsg_def (first, last, 4252, "epsg", 4252, "Lome");
    add_proj4text (p, 0, "+proj=longlat +a=6378249.2 +b=6356515 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Lome\",DATUM[\"Lome\",SPHEROID[\"Clarke 1880 (I");
    add_srs_wkt (p, 1,
		 "GN)\",6378249.2,293.4660212936265,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 2,
		 "011\"]],AUTHORITY[\"EPSG\",\"6252\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 3,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 4,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"4252\"]]");
    p = add_epsg_def (first, last, 4253, "epsg", 4253, "Luzon 1911");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk66 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Luzon 1911\",DATUM[\"Luzon_1911\",SPHEROID[\"Cl");
    add_srs_wkt (p, 1,
		 "arke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 2,
		 "\",\"7008\"]],AUTHORITY[\"EPSG\",\"6253\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 3,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 4,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 5, "RITY[\"EPSG\",\"4253\"]]");
    p = add_epsg_def (first, last, 4254, "epsg", 4254, "Hito XVIII 1963");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Hito XVIII 1963\",DATUM[\"Hito_XVIII_1963\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"International 1924\",6378388,297,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7022\"]],AUTHORITY[\"EPSG\",\"6254\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 3,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 4,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 5, "RITY[\"EPSG\",\"4254\"]]");
    p = add_epsg_def (first, last, 4255, "epsg", 4255, "Herat North");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-333,-222,114,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Herat North\",DATUM[\"Herat_North\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 2,
		 "22\"]],TOWGS84[-333,-222,114,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6255\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4255\"]]");
    p = add_epsg_def (first, last, 4256, "epsg", 4256, "Mahe 1971");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=41,-220,-134,0,0,0,");
    add_proj4text (p, 1, "0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Mahe 1971\",DATUM[\"Mahe_1971\",SPHEROID[\"Clar");
    add_srs_wkt (p, 1,
		 "ke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7012\"]],TOWGS84[41,-220,-134,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6256\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4256\"]]");
    p = add_epsg_def (first, last, 4257, "epsg", 4257, "Makassar");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=bessel +towgs84=-587.8,519.75,145.7");
    add_proj4text (p, 1, "6,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Makassar\",DATUM[\"Makassar\",SPHEROID[\"Bessel");
    add_srs_wkt (p, 1,
		 " 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 2,
		 "4\"]],TOWGS84[-587.8,519.75,145.76,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6257\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4257\"]]");
    p = add_epsg_def (first, last, 4258, "epsg", 4258, "ETRS89");
    add_proj4text (p, 0, "+proj=longlat +ellps=GRS80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"ETRS89\",DATUM[\"European_Terrestrial_Reference");
    add_srs_wkt (p, 1,
		 "_System_1989\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 2,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"62");
    add_srs_wkt (p, 3,
		 "58\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4258\"]]");
    p = add_epsg_def (first, last, 4259, "epsg", 4259, "Malongo 1987");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Malongo 1987\",DATUM[\"Malongo_1987\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7022\"]],AUTHORITY[\"EPSG\",\"6259\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 3,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 4,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 5, "\"EPSG\",\"4259\"]]");
    p = add_epsg_def (first, last, 4260, "epsg", 4260, "Manoca");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=-70.9,-151.8,-41.4,");
    add_proj4text (p, 1, "0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Manoca\",DATUM[\"Manoca\",SPHEROID[\"Clarke 188");
    add_srs_wkt (p, 1,
		 "0 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"");
    add_srs_wkt (p, 2,
		 "]],TOWGS84[-70.9,-151.8,-41.4,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6260\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.0174532925199433,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"4260\"]]");
    p = add_epsg_def (first, last, 4261, "epsg", 4261, "Merchich");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378249.2 +b=6356515 +towgs84=31,146,47");
    add_proj4text (p, 1, ",0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Merchich\",DATUM[\"Merchich\",SPHEROID[\"Clarke");
    add_srs_wkt (p, 1,
		 " 1880 (IGN)\",6378249.2,293.4660212936265,AUTHORITY[\"EP");
    add_srs_wkt (p, 2,
		 "SG\",\"7011\"]],TOWGS84[31,146,47,0,0,0,0],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6261\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5, "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4261\"]]");
    p = add_epsg_def (first, last, 4262, "epsg", 4262, "Massawa");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=bessel +towgs84=639,405,60,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Massawa\",DATUM[\"Massawa\",SPHEROID[\"Bessel 1");
    add_srs_wkt (p, 1,
		 "841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"");
    add_srs_wkt (p, 2,
		 "]],TOWGS84[639,405,60,0,0,0,0],AUTHORITY[\"EPSG\",\"6262");
    add_srs_wkt (p, 3,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4262\"]]");
    p = add_epsg_def (first, last, 4263, "epsg", 4263, "Minna");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Minna\",DATUM[\"Minna\",SPHEROID[\"Clarke 1880 ");
    add_srs_wkt (p, 1,
		 "(RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]");
    add_srs_wkt (p, 2,
		 "],AUTHORITY[\"EPSG\",\"6263\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 4,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"4263\"]]");
    p = add_epsg_def (first, last, 4264, "epsg", 4264, "Mhast");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-252.95,-4.11,-96.38,");
    add_proj4text (p, 1, "0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Mhast\",DATUM[\"Mhast\",SPHEROID[\"Internationa");
    add_srs_wkt (p, 1,
		 "l 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS");
    add_srs_wkt (p, 2,
		 "84[-252.95,-4.11,-96.38,0,0,0,0],AUTHORITY[\"EPSG\",\"62");
    add_srs_wkt (p, 3,
		 "64\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4264\"]]");
    p = add_epsg_def (first, last, 4265, "epsg", 4265, "Monte Mario");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Monte Mario\",DATUM[\"Monte_Mario\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 2,
		 "22\"]],AUTHORITY[\"EPSG\",\"6265\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 3,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 4,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 5, "PSG\",\"4265\"]]");
    p = add_epsg_def (first, last, 4266, "epsg", 4266, "M'poraloko");
    add_proj4text (p, 0, "+proj=longlat +a=6378249.2 +b=6356515 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"M'poraloko\",DATUM[\"M_poraloko\",SPHEROID[\"Cl");
    add_srs_wkt (p, 1,
		 "arke 1880 (IGN)\",6378249.2,293.4660212936265,AUTHORITY[");
    add_srs_wkt (p, 2,
		 "\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6266\"]],PRIMEM");
    add_srs_wkt (p, 3,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 4,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 5, ",AUTHORITY[\"EPSG\",\"4266\"]]");
    p = add_epsg_def (first, last, 4267, "epsg", 4267, "NAD27");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk66 +datum=NAD27 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"NAD27\",DATUM[\"North_American_Datum_1927\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHOR");
    add_srs_wkt (p, 2,
		 "ITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PR");
    add_srs_wkt (p, 3,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 4,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 5, "\"]],AUTHORITY[\"EPSG\",\"4267\"]]");
    p = add_epsg_def (first, last, 4268, "epsg", 4268, "NAD27 Michigan");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378450.047548896 +b=6356826.621488444 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"NAD27 Michigan\",DATUM[\"NAD_Michigan\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"Clarke 1866 Michigan\",6378450.047548896,294.9786971");
    add_srs_wkt (p, 2,
		 "646739,AUTHORITY[\"EPSG\",\"7009\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"6268\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 4,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4268\"]]");
    p = add_epsg_def (first, last, 4269, "epsg", 4269, "NAD83");
    add_proj4text (p, 0, "+proj=longlat +ellps=GRS80 +datum=NAD83 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"NAD83\",DATUM[\"North_American_Datum_1983\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 3,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 4,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 5, "RITY[\"EPSG\",\"4269\"]]");
    p = add_epsg_def (first, last, 4270, "epsg", 4270, "Nahrwan 1967");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Nahrwan 1967\",DATUM[\"Nahrwan_1967\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EP");
    add_srs_wkt (p, 2,
		 "SG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6270\"]],PRIMEM[\"G");
    add_srs_wkt (p, 3,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 4,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 5, "HORITY[\"EPSG\",\"4270\"]]");
    p = add_epsg_def (first, last, 4271, "epsg", 4271, "Naparima 1972");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Naparima 1972\",DATUM[\"Naparima_1972\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 2,
		 "\"7022\"]],AUTHORITY[\"EPSG\",\"6271\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 3,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 4,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"4271\"]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_16 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
    struct epsg_defs *p;
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    p = add_epsg_def (first, last, 4272, "epsg", 4272, "NZGD49");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +datum=nzgd49 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"NZGD49\",DATUM[\"New_Zealand_Geodetic_Datum_194");
    add_srs_wkt (p, 1,
		 "9\",SPHEROID[\"International 1924\",6378388,297,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0");
    add_srs_wkt (p, 3,
		 ".1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6, "THORITY[\"EPSG\",\"4272\"]]");
    p = add_epsg_def (first, last, 4273, "epsg", 4273, "NGO 1948");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377492.018 +b=6356173.508712696 +towgs");
    add_proj4text (p, 1, "84=278.3,93,474.5,7.889,0.05,-6.61,6.21 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"NGO 1948\",DATUM[\"NGO_1948\",SPHEROID[\"Bessel");
    add_srs_wkt (p, 1,
		 " Modified\",6377492.018,299.1528128,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7005\"]],TOWGS84[278.3,93,474.5,7.889,0.05,-6.61,6.21],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6273\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6, "4273\"]]");
    p = add_epsg_def (first, last, 4274, "epsg", 4274, "Datum 73");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Datum 73\",DATUM[\"Datum_73\",SPHEROID[\"Intern");
    add_srs_wkt (p, 1,
		 "ational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]]");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"6274\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 4,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"4274\"]]");
    p = add_epsg_def (first, last, 4275, "epsg", 4275, "NTF");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378249.2 +b=6356515 +towgs84=-168,-60,");
    add_proj4text (p, 1, "320,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"NTF\",DATUM[\"Nouvelle_Triangulation_Francaise\"");
    add_srs_wkt (p, 1,
		 ",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936");
    add_srs_wkt (p, 2,
		 "265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-168,-60,320,0");
    add_srs_wkt (p, 3,
		 ",0,0,0],AUTHORITY[\"EPSG\",\"6275\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6, "EPSG\",\"4275\"]]");
    p = add_epsg_def (first, last, 4276, "epsg", 4276, "NSWC 9Z-2");
    add_proj4text (p, 0, "+proj=longlat +ellps=WGS66 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"NSWC 9Z-2\",DATUM[\"NSWC_9Z_2\",SPHEROID[\"NWL ");
    add_srs_wkt (p, 1,
		 "9D\",6378145,298.25,AUTHORITY[\"EPSG\",\"7025\"]],AUTHOR");
    add_srs_wkt (p, 2,
		 "ITY[\"EPSG\",\"6276\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4276");
    add_srs_wkt (p, 5, "\"]]");
    p = add_epsg_def (first, last, 4277, "epsg", 4277, "OSGB 1936");
    add_proj4text (p, 0, "+proj=longlat +ellps=airy +datum=OSGB36 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"OSGB 1936\",DATUM[\"OSGB_1936\",SPHEROID[\"Airy");
    add_srs_wkt (p, 1,
		 " 1830\",6377563.396,299.3249646,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 2,
		 "1\"]],AUTHORITY[\"EPSG\",\"6277\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 3,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 4,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 5, "PSG\",\"4277\"]]");
    p = add_epsg_def (first, last, 4278, "epsg", 4278, "OSGB70");
    add_proj4text (p, 0, "+proj=longlat +ellps=airy +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"OSGB70\",DATUM[\"OSGB_1970_SN\",SPHEROID[\"Airy");
    add_srs_wkt (p, 1,
		 " 1830\",6377563.396,299.3249646,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 2,
		 "1\"]],AUTHORITY[\"EPSG\",\"6278\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 3,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 4,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 5, "PSG\",\"4278\"]]");
    p = add_epsg_def (first, last, 4279, "epsg", 4279, "OS(SN)80");
    add_proj4text (p, 0, "+proj=longlat +ellps=airy +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"OS(SN)80\",DATUM[\"OS_SN_1980\",SPHEROID[\"Airy");
    add_srs_wkt (p, 1,
		 " 1830\",6377563.396,299.3249646,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 2,
		 "1\"]],AUTHORITY[\"EPSG\",\"6279\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 3,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 4,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 5, "PSG\",\"4279\"]]");
    p = add_epsg_def (first, last, 4280, "epsg", 4280, "Padang");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Padang\",DATUM[\"Padang_1884\",SPHEROID[\"Besse");
    add_srs_wkt (p, 1,
		 "l 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 2,
		 "04\"]],AUTHORITY[\"EPSG\",\"6280\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 3,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 4,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 5, "PSG\",\"4280\"]]");
    p = add_epsg_def (first, last, 4281, "epsg", 4281, "Palestine 1923");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378300.789 +b=6356566.435 +towgs84=-27");
    add_proj4text (p, 1,
		   "5.722,94.7824,340.894,-8.001,-4.42,-11.821,1 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Palestine 1923\",DATUM[\"Palestine_1923\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"Clarke 1880 (Benoit)\",6378300.789,293.46631553898");
    add_srs_wkt (p, 2,
		 "02,AUTHORITY[\"EPSG\",\"7010\"]],TOWGS84[-275.722,94.782");
    add_srs_wkt (p, 3,
		 "4,340.894,-8.001,-4.42,-11.821,1],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 4,
		 "281\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 5,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 6, "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4281\"]]");
    p = add_epsg_def (first, last, 4282, "epsg", 4282, "Pointe Noire");
    add_proj4text (p, 0, "+proj=longlat +a=6378249.2 +b=6356515 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Pointe Noire\",DATUM[\"Congo_1960_Pointe_Noire\"");
    add_srs_wkt (p, 1,
		 ",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936");
    add_srs_wkt (p, 2,
		 "265,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "282\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5, "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4282\"]]");
    p = add_epsg_def (first, last, 4283, "epsg", 4283, "GDA94");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"GDA94\",DATUM[\"Geocentric_Datum_of_Australia_1");
    add_srs_wkt (p, 1,
		 "994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR");
    add_srs_wkt (p, 2,
		 "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4283\"]");
    add_srs_wkt (p, 6, "]");
    p = add_epsg_def (first, last, 4284, "epsg", 4284, "Pulkovo 1942");
    add_proj4text (p, 0, "+proj=longlat +ellps=krass +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 2,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6, "4284\"]]");
    p = add_epsg_def (first, last, 4285, "epsg", 4285, "Qatar 1974");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Qatar 1974\",DATUM[\"Qatar_1974\",SPHEROID[\"In");
    add_srs_wkt (p, 1,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 2,
		 "\"]],AUTHORITY[\"EPSG\",\"6285\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 3,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 4,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 5, "SG\",\"4285\"]]");
    p = add_epsg_def (first, last, 4286, "epsg", 4286, "Qatar 1948");
    add_proj4text (p, 0, "+proj=longlat +ellps=helmert +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Qatar 1948\",DATUM[\"Qatar_1948\",SPHEROID[\"He");
    add_srs_wkt (p, 1,
		 "lmert 1906\",6378200,298.3,AUTHORITY[\"EPSG\",\"7020\"]]");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"6286\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 4,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"4286\"]]");
    p = add_epsg_def (first, last, 4287, "epsg", 4287, "Qornoq");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=164,138,-189,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Qornoq\",DATUM[\"Qornoq\",SPHEROID[\"Internatio");
    add_srs_wkt (p, 1,
		 "nal 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOW");
    add_srs_wkt (p, 2,
		 "GS84[164,138,-189,0,0,0,0],AUTHORITY[\"EPSG\",\"6287\"]]");
    add_srs_wkt (p, 3,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 4,
		 "IT[\"degree\",0.0174532925199433,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 5, "08\"]],AUTHORITY[\"EPSG\",\"4287\"]]");
    p = add_epsg_def (first, last, 4288, "epsg", 4288, "Loma Quintana");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Loma Quintana\",DATUM[\"Loma_Quintana\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 2,
		 "\"7022\"]],AUTHORITY[\"EPSG\",\"6288\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 3,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 4,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"4288\"]]");
    p = add_epsg_def (first, last, 4289, "epsg", 4289, "Amersfoort");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Amersfoort\",DATUM[\"Amersfoort\",SPHEROID[\"Be");
    add_srs_wkt (p, 1,
		 "ssel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7004\"]],AUTHORITY[\"EPSG\",\"6289\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 3,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 4,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 5, "\"EPSG\",\"4289\"]]");
    p = add_epsg_def (first, last, 4291, "epsg", 4291, "SAD69");
    add_proj4text (p, 0, "+proj=longlat +ellps=GRS67 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"SAD69\",DATUM[\"South_American_Datum_1969\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"GRS 1967\",6378160,298.247167427,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7036\"]],AUTHORITY[\"EPSG\",\"6291\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 3,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 4,
		 ",0.0174532925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHOR");
    add_srs_wkt (p, 5, "ITY[\"EPSG\",\"4291\"]]");
    p = add_epsg_def (first, last, 4292, "epsg", 4292, "Sapper Hill 1943");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-355,21,72,0,0,0,0 +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Sapper Hill 1943\",DATUM[\"Sapper_Hill_1943\",S");
    add_srs_wkt (p, 1,
		 "PHEROID[\"International 1924\",6378388,297,AUTHORITY[\"E");
    add_srs_wkt (p, 2,
		 "PSG\",\"7022\"]],TOWGS84[-355,21,72,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6292\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4292\"]]");
    p = add_epsg_def (first, last, 4293, "epsg", 4293, "Schwarzeck");
    add_proj4text (p, 0, "+proj=longlat +ellps=bess_nam +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Schwarzeck\",DATUM[\"Schwarzeck\",SPHEROID[\"Be");
    add_srs_wkt (p, 1,
		 "ssel Namibia (GLM)\",6377483.865280419,299.1528128,AUTHO");
    add_srs_wkt (p, 2,
		 "RITY[\"EPSG\",\"7046\"]],AUTHORITY[\"EPSG\",\"6293\"]],P");
    add_srs_wkt (p, 3,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 4,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 5, "2\"]],AUTHORITY[\"EPSG\",\"4293\"]]");
    p = add_epsg_def (first, last, 4294, "epsg", 4294, "Segora");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Segora\",DATUM[\"Segora\",SPHEROID[\"Bessel 184");
    add_srs_wkt (p, 1,
		 "1\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]");
    add_srs_wkt (p, 2,
		 "],AUTHORITY[\"EPSG\",\"6294\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 4,
		 "5199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"4294\"]]");
    p = add_epsg_def (first, last, 4295, "epsg", 4295, "Serindung");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Serindung\",DATUM[\"Serindung\",SPHEROID[\"Bess");
    add_srs_wkt (p, 1,
		 "el 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 2,
		 "004\"]],AUTHORITY[\"EPSG\",\"6295\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 3,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 4,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"4295\"]]");
    p = add_epsg_def (first, last, 4296, "epsg", 4296, "Sudan");
    add_proj4text (p, 0, "+proj=longlat +a=6378249.2 +b=6356515 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Sudan\",DATUM[\"Sudan\",SPHEROID[\"Clarke 1880 ");
    add_srs_wkt (p, 1,
		 "(IGN)\",6378249.2,293.4660212936265,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7011\"]],AUTHORITY[\"EPSG\",\"6296\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 3,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 4,
		 "74532925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"4296\"]]");
    p = add_epsg_def (first, last, 4297, "epsg", 4297, "Tananarive");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-189,-242,-91,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Tananarive\",DATUM[\"Tananarive_1925\",SPHEROID");
    add_srs_wkt (p, 1,
		 "[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7022\"]],TOWGS84[-189,-242,-91,0,0,0,0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"6297\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 5, "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4297\"]]");
    p = add_epsg_def (first, last, 4298, "epsg", 4298, "Timbalai 1948");
    add_proj4text (p, 0, "+proj=longlat +ellps=evrstSS +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Timbalai 1948\",DATUM[\"Timbalai_1948\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"Everest 1830 (1967 Definition)\",6377298.556,300.801");
    add_srs_wkt (p, 2,
		 "7,AUTHORITY[\"EPSG\",\"7016\"]],AUTHORITY[\"EPSG\",\"629");
    add_srs_wkt (p, 3,
		 "8\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4298\"]]");
    p = add_epsg_def (first, last, 4299, "epsg", 4299, "TM65");
    add_proj4text (p, 0, "+proj=longlat +ellps=mod_airy +datum=ire65 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"TM65\",DATUM[\"TM65\",SPHEROID[\"Airy Modified ");
    add_srs_wkt (p, 1,
		 "1849\",6377340.189,299.3249646,AUTHORITY[\"EPSG\",\"7002");
    add_srs_wkt (p, 2,
		 "\"]],AUTHORITY[\"EPSG\",\"6299\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 3,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 4,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 5, "SG\",\"4299\"]]");
    p = add_epsg_def (first, last, 4300, "epsg", 4300, "TM75");
    add_proj4text (p, 0, "+proj=longlat +ellps=mod_airy +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"TM75\",DATUM[\"Geodetic_Datum_of_1965\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"Airy Modified 1849\",6377340.189,299.3249646,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7002\"]],AUTHORITY[\"EPSG\",\"6300\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5, "]],AUTHORITY[\"EPSG\",\"4300\"]]");
    p = add_epsg_def (first, last, 4301, "epsg", 4301, "Tokyo");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\"");
    add_srs_wkt (p, 1,
		 ",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 4,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5, "4301\"]]");
    p = add_epsg_def (first, last, 4302, "epsg", 4302, "Trinidad 1903");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378293.645208759 +b=6356617.987679838 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Trinidad 1903\",DATUM[\"Trinidad_1903\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"Clarke 1858\",6378293.645208759,294.2606763692569,AU");
    add_srs_wkt (p, 2,
		 "THORITY[\"EPSG\",\"7007\"]],AUTHORITY[\"EPSG\",\"6302\"]");
    add_srs_wkt (p, 3,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 4,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5, "9122\"]],AUTHORITY[\"EPSG\",\"4302\"]]");
    p = add_epsg_def (first, last, 4303, "epsg", 4303, "TC(1948)");
    add_proj4text (p, 0, "+proj=longlat +ellps=helmert +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"TC(1948)\",DATUM[\"Trucial_Coast_1948\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"Helmert 1906\",6378200,298.3,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 2,
		 "20\"]],AUTHORITY[\"EPSG\",\"6303\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 3,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 4,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 5, "PSG\",\"4303\"]]");
    p = add_epsg_def (first, last, 4304, "epsg", 4304, "Voirol 1875");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378249.2 +b=6356515 +towgs84=-73,-247,");
    add_proj4text (p, 1, "227,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Voirol 1875\",DATUM[\"Voirol_1875\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "Clarke 1880 (IGN)\",6378249.2,293.4660212936265,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7011\"]],TOWGS84[-73,-247,227,0,0,0,0],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6304\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"43");
    add_srs_wkt (p, 6, "04\"]]");
    p = add_epsg_def (first, last, 4306, "epsg", 4306, "Bern 1938");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Bern 1938\",DATUM[\"Bern_1938\",SPHEROID[\"Bess");
    add_srs_wkt (p, 1,
		 "el 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 2,
		 "004\"]],AUTHORITY[\"EPSG\",\"6306\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 3,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 4,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"4306\"]]");
    p = add_epsg_def (first, last, 4307, "epsg", 4307, "Nord Sahara 1959");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Nord Sahara 1959\",DATUM[\"Nord_Sahara_1959\",S");
    add_srs_wkt (p, 1,
		 "PHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHOR");
    add_srs_wkt (p, 2,
		 "ITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6307\"]],PR");
    add_srs_wkt (p, 3,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 4,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 5, "\"]],AUTHORITY[\"EPSG\",\"4307\"]]");
    p = add_epsg_def (first, last, 4308, "epsg", 4308, "RT38");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"RT38\",DATUM[\"Stockholm_1938\",SPHEROID[\"Bess");
    add_srs_wkt (p, 1,
		 "el 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 2,
		 "004\"]],AUTHORITY[\"EPSG\",\"6308\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 3,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 4,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"4308\"]]");
    p = add_epsg_def (first, last, 4309, "epsg", 4309, "Yacare");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-155,171,37,0,0,0,0 +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Yacare\",DATUM[\"Yacare\",SPHEROID[\"Internatio");
    add_srs_wkt (p, 1,
		 "nal 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOW");
    add_srs_wkt (p, 2,
		 "GS84[-155,171,37,0,0,0,0],AUTHORITY[\"EPSG\",\"6309\"]],");
    add_srs_wkt (p, 3,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 4,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 5, "22\"]],AUTHORITY[\"EPSG\",\"4309\"]]");
    p = add_epsg_def (first, last, 4310, "epsg", 4310, "Yoff");
    add_proj4text (p, 0, "+proj=longlat +a=6378249.2 +b=6356515 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Yoff\",DATUM[\"Yoff\",SPHEROID[\"Clarke 1880 (I");
    add_srs_wkt (p, 1,
		 "GN)\",6378249.2,293.4660212936265,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 2,
		 "011\"]],AUTHORITY[\"EPSG\",\"6310\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 3,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 4,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"4310\"]]");
    p = add_epsg_def (first, last, 4311, "epsg", 4311, "Zanderij");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-265,120,-358,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Zanderij\",DATUM[\"Zanderij\",SPHEROID[\"Intern");
    add_srs_wkt (p, 1,
		 "ational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]]");
    add_srs_wkt (p, 2,
		 ",TOWGS84[-265,120,-358,0,0,0,0],AUTHORITY[\"EPSG\",\"631");
    add_srs_wkt (p, 3,
		 "1\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4311\"]]");
    p = add_epsg_def (first, last, 4312, "epsg", 4312, "MGI");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=bessel +datum=hermannskogel +no_def");
    add_proj4text (p, 1, "s");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"MGI\",DATUM[\"Militar_Geographische_Institute\"");
    add_srs_wkt (p, 1,
		 ",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHOR");
    add_srs_wkt (p, 2,
		 "ITY[\"EPSG\",\"7004\"]],TOWGS84[577.326,90.129,463.919,5");
    add_srs_wkt (p, 3,
		 ".137,1.474,5.297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.0174532925199433,AUTHORITY[\"EPSG\",\"9108");
    add_srs_wkt (p, 6, "\"]],AUTHORITY[\"EPSG\",\"4312\"]]");
    p = add_epsg_def (first, last, 4313, "epsg", 4313, "Belge 1972");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=106.869,-52.2978,103.");
    add_proj4text (p, 1, "724,-0.33657,0.456955,-1.84218,1 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Belge 1972\",DATUM[\"Reseau_National_Belge_1972");
    add_srs_wkt (p, 1,
		 "\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY");
    add_srs_wkt (p, 2,
		 "[\"EPSG\",\"7022\"]],TOWGS84[106.869,-52.2978,103.724,-0");
    add_srs_wkt (p, 3,
		 ".33657,0.456955,-1.84218,1],AUTHORITY[\"EPSG\",\"6313\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6, "9122\"]],AUTHORITY[\"EPSG\",\"4313\"]]");
    p = add_epsg_def (first, last, 4314, "epsg", 4314, "DHDN");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +datum=potsdam +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"DHDN\",DATUM[\"Deutsches_Hauptdreiecksnetz\",SP");
    add_srs_wkt (p, 1,
		 "HEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY");
    add_srs_wkt (p, 2,
		 "[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6314\"]],PRIME");
    add_srs_wkt (p, 3,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 4,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 5, "],AUTHORITY[\"EPSG\",\"4314\"]]");
    p = add_epsg_def (first, last, 4315, "epsg", 4315, "Conakry 1905");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378249.2 +b=6356515 +towgs84=-23,259,-");
    add_proj4text (p, 1, "9,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Conakry 1905\",DATUM[\"Conakry_1905\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"Clarke 1880 (IGN)\",6378249.2,293.4660212936265,AUTHOR");
    add_srs_wkt (p, 2,
		 "ITY[\"EPSG\",\"7011\"]],TOWGS84[-23,259,-9,0,0,0,0],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6315\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"43");
    add_srs_wkt (p, 6, "15\"]]");
    p = add_epsg_def (first, last, 4316, "epsg", 4316, "Dealul Piscului 1930");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Dealul Piscului 1930\",DATUM[\"Dealul_Piscului_");
    add_srs_wkt (p, 1,
		 "1930\",SPHEROID[\"International 1924\",6378388,297,AUTHO");
    add_srs_wkt (p, 2,
		 "RITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6316\"]],P");
    add_srs_wkt (p, 3,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 4,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 5, "2\"]],AUTHORITY[\"EPSG\",\"4316\"]]");
    p = add_epsg_def (first, last, 4317, "epsg", 4317, "Dealul Piscului 1970");
    add_proj4text (p, 0, "+proj=longlat +ellps=krass +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Dealul Piscului 1970\",DATUM[\"Dealul_Piscului_");
    add_srs_wkt (p, 1,
		 "1970\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHOR");
    add_srs_wkt (p, 2,
		 "ITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6317\"]],PR");
    add_srs_wkt (p, 3,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 4,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 5, "\"]],AUTHORITY[\"EPSG\",\"4317\"]]");
    p = add_epsg_def (first, last, 4318, "epsg", 4318, "NGN");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=WGS84 +towgs84=-3.2,-5.7,2.8,0,0,0,");
    add_proj4text (p, 1, "0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"NGN\",DATUM[\"National_Geodetic_Network\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 2,
		 "\"7030\"]],TOWGS84[-3.2,-5.7,2.8,0,0,0,0],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6318\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4318\"]]");
    p = add_epsg_def (first, last, 4319, "epsg", 4319, "KUDAMS");
    add_proj4text (p, 0, "+proj=longlat +ellps=GRS80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"KUDAMS\",DATUM[\"Kuwait_Utility\",SPHEROID[\"GR");
    add_srs_wkt (p, 1,
		 "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019");
    add_srs_wkt (p, 2,
		 "\"]],AUTHORITY[\"EPSG\",\"6319\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 3,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 4,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 5, "SG\",\"4319\"]]");
    p = add_epsg_def (first, last, 4322, "epsg", 4322, "WGS 72");
    add_proj4text (p, 0, "+proj=longlat +ellps=WGS72 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"WGS 72\",DATUM[\"WGS_1972\",SPHEROID[\"WGS 72\"");
    add_srs_wkt (p, 1,
		 ",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],AUTHORITY[");
    add_srs_wkt (p, 2,
		 "\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4322\"]]");
    add_srs_wkt (p, 5, "");
    p = add_epsg_def (first, last, 4324, "epsg", 4324, "WGS 72BE");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.814,-0");
    add_proj4text (p, 1, ".38 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"WGS 72BE\",DATUM[\"WGS_1972_Transit_Broadcast_E");
    add_srs_wkt (p, 1,
		 "phemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[");
    add_srs_wkt (p, 2,
		 "\"EPSG\",\"7043\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6, "324\"]]");

#endif /* loading unconditionally the 4326 EPSG def */
    p = add_epsg_def (first, last, 4326, "epsg", 4326, "WGS 84");
    add_proj4text (p, 0, "+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"WGS 84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS 84\"");
    add_srs_wkt (p, 1,
		 ",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 4,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 5, "326\"]]");
#if OMIT_EPSG == 0		/* resuming conditional EPSG initialization */

    p = add_epsg_def (first, last, 4600, "epsg", 4600, "Anguilla 1957");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Anguilla 1957\",DATUM[\"Anguilla_1957\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6600\"]],PRIMEM[\"");
    add_srs_wkt (p, 3,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 4,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 5, "THORITY[\"EPSG\",\"4600\"]]");
    p = add_epsg_def (first, last, 4601, "epsg", 4601, "Antigua 1943");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Antigua 1943\",DATUM[\"Antigua_1943\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EP");
    add_srs_wkt (p, 2,
		 "SG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6601\"]],PRIMEM[\"G");
    add_srs_wkt (p, 3,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 4,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 5, "HORITY[\"EPSG\",\"4601\"]]");
    p = add_epsg_def (first, last, 4602, "epsg", 4602, "Dominica 1945");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=725,685,536,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Dominica 1945\",DATUM[\"Dominica_1945\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7012\"]],TOWGS84[725,685,536,0,0,0,0],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6602\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4602\"]");
    add_srs_wkt (p, 6, "]");
    p = add_epsg_def (first, last, 4603, "epsg", 4603, "Grenada 1953");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=72,213.7,93,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Grenada 1953\",DATUM[\"Grenada_1953\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EP");
    add_srs_wkt (p, 2,
		 "SG\",\"7012\"]],TOWGS84[72,213.7,93,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6603\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4603\"]]");
    p = add_epsg_def (first, last, 4604, "epsg", 4604, "Montserrat 1958");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=174,359,365,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Montserrat 1958\",DATUM[\"Montserrat_1958\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7012\"]],TOWGS84[174,359,365,0,0,0,0],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6604\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"460");
    add_srs_wkt (p, 6, "4\"]]");
    p = add_epsg_def (first, last, 4605, "epsg", 4605, "St. Kitts 1955");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"St. Kitts 1955\",DATUM[\"St_Kitts_1955\",SPHERO");
    add_srs_wkt (p, 1,
		 "ID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6605\"]],PRIMEM[\"");
    add_srs_wkt (p, 3,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 4,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 5, "THORITY[\"EPSG\",\"4605\"]]");
    p = add_epsg_def (first, last, 4606, "epsg", 4606, "St. Lucia 1955");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=-149,128,296,0,0,0,");
    add_proj4text (p, 1, "0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"St. Lucia 1955\",DATUM[\"St_Lucia_1955\",SPHERO");
    add_srs_wkt (p, 1,
		 "ID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7012\"]],TOWGS84[-149,128,296,0,0,0,0],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6606\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4606\"]");
    add_srs_wkt (p, 6, "]");
    p = add_epsg_def (first, last, 4607, "epsg", 4607, "St. Vincent 1945");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=195.671,332.517,274");
    add_proj4text (p, 1, ".607,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"St. Vincent 1945\",DATUM[\"St_Vincent_1945\",SP");
    add_srs_wkt (p, 1,
		 "HEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7012\"]],TOWGS84[195.671,332.517,274.607,0");
    add_srs_wkt (p, 3,
		 ",0,0,0],AUTHORITY[\"EPSG\",\"6607\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6, "EPSG\",\"4607\"]]");
    p = add_epsg_def (first, last, 4608, "epsg", 4608, "NAD27(76)");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk66 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"NAD27(76)\",DATUM[\"North_American_Datum_1927_1");
    add_srs_wkt (p, 1,
		 "976\",SPHEROID[\"Clarke 1866\",6378206.4,294.97869821390");
    add_srs_wkt (p, 2,
		 "06,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"66");
    add_srs_wkt (p, 3,
		 "08\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4608\"]]");
    p = add_epsg_def (first, last, 4609, "epsg", 4609, "NAD27(CGQ77)");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk66 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"NAD27(CGQ77)\",DATUM[\"North_American_Datum_192");
    add_srs_wkt (p, 1,
		 "7_CGQ77\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982");
    add_srs_wkt (p, 2,
		 "139006,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"6609\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 4,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4609\"]]");
    p = add_epsg_def (first, last, 4610, "epsg", 4610, "Xian 1980");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378140 +b=6356755.288157528 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Xian 1980\",DATUM[\"Xian_1980\",SPHEROID[\"Xian");
    add_srs_wkt (p, 1,
		 " 1980\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AU");
    add_srs_wkt (p, 2,
		 "THORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 4,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5, "4610\"]]");
    p = add_epsg_def (first, last, 4611, "epsg", 4611, "Hong Kong 1980");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-162.619,-276.959,-16");
    add_proj4text (p, 1, "1.764,0.067753,-2.24365,-1.15883,-1.09425 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Hong Kong 1980\",DATUM[\"Hong_Kong_1980\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 2,
		 ",\"7022\"]],TOWGS84[-162.619,-276.959,-161.764,0.067753,");
    add_srs_wkt (p, 3,
		 "-2.24365,-1.15883,-1.09425],AUTHORITY[\"EPSG\",\"6611\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6, "9122\"]],AUTHORITY[\"EPSG\",\"4611\"]]");
    p = add_epsg_def (first, last, 4612, "epsg", 4612, "JGD2000");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"JGD2000\",DATUM[\"Japanese_Geodetic_Datum_2000\"");
    add_srs_wkt (p, 1,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4612\"]]");
    p = add_epsg_def (first, last, 4613, "epsg", 4613, "Segara");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Segara\",DATUM[\"Gunung_Segara\",SPHEROID[\"Bes");
    add_srs_wkt (p, 1,
		 "sel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7004\"]],AUTHORITY[\"EPSG\",\"6613\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 3,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 4,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 5, "\"EPSG\",\"4613\"]]");
    p = add_epsg_def (first, last, 4614, "epsg", 4614, "QND95");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-119.425,-303.659,-11");
    add_proj4text (p, 1, ".0006,1.1643,0.174458,1.09626,3.65706 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"QND95\",DATUM[\"Qatar_National_Datum_1995\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"International 1924\",6378388,297,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7022\"]],TOWGS84[-119.425,-303.659,-11.0006,1.1643");
    add_srs_wkt (p, 3,
		 ",0.174458,1.09626,3.65706],AUTHORITY[\"EPSG\",\"6614\"]]");
    add_srs_wkt (p, 4,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 5,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 6, "122\"]],AUTHORITY[\"EPSG\",\"4614\"]]");
    p = add_epsg_def (first, last, 4615, "epsg", 4615, "Porto Santo");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-499,-249,314,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Porto Santo\",DATUM[\"Porto_Santo_1936\",SPHERO");
    add_srs_wkt (p, 1,
		 "ID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 2,
		 ",\"7022\"]],TOWGS84[-499,-249,314,0,0,0,0],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6615\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5, "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4615\"]]");
    p = add_epsg_def (first, last, 4616, "epsg", 4616, "Selvagem Grande");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Selvagem Grande\",DATUM[\"Selvagem_Grande\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"International 1924\",6378388,297,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7022\"]],AUTHORITY[\"EPSG\",\"6616\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 3,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 4,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 5, "RITY[\"EPSG\",\"4616\"]]");
    p = add_epsg_def (first, last, 4617, "epsg", 4617, "NAD83(CSRS)");
    add_proj4text (p, 0, "+proj=longlat +ellps=GRS80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83_Canadian_Spatial_Re");
    add_srs_wkt (p, 1,
		 "ference_System\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 2,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 4,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 5, "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4617\"]]");
    p = add_epsg_def (first, last, 4618, "epsg", 4618, "SAD69");
    add_proj4text (p, 0, "+proj=longlat +ellps=aust_SA +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"SAD69\",DATUM[\"South_American_Datum_1969\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"E");
    add_srs_wkt (p, 2,
		 "PSG\",\"7050\"]],AUTHORITY[\"EPSG\",\"6618\"]],PRIMEM[\"");
    add_srs_wkt (p, 3,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 4,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 5, "THORITY[\"EPSG\",\"4618\"]]");
    p = add_epsg_def (first, last, 4619, "epsg", 4619, "SWEREF99");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"SWEREF99\",DATUM[\"SWEREF99\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 1,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 2,
		 ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6619\"]],PR");
    add_srs_wkt (p, 3,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 4,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 5, "\"]],AUTHORITY[\"EPSG\",\"4619\"]]");
    p = add_epsg_def (first, last, 4620, "epsg", 4620, "Point 58");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=-106,-129,165,0,0,0");
    add_proj4text (p, 1, ",0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Point 58\",DATUM[\"Point_58\",SPHEROID[\"Clarke");
    add_srs_wkt (p, 1,
		 " 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 2,
		 "012\"]],TOWGS84[-106,-129,165,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6620\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4620\"]]");
    p = add_epsg_def (first, last, 4621, "epsg", 4621, "Fort Marigot");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=137,248,-430,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Fort Marigot\",DATUM[\"Fort_Marigot\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7022\"]],TOWGS84[137,248,-430,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6621\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4621\"]]");
    p = add_epsg_def (first, last, 4622, "epsg", 4622, "Guadeloupe 1948");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Guadeloupe 1948\",DATUM[\"Guadeloupe_1948\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"International 1924\",6378388,297,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7022\"]],AUTHORITY[\"EPSG\",\"6622\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 3,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 4,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 5, "RITY[\"EPSG\",\"4622\"]]");
    p = add_epsg_def (first, last, 4623, "epsg", 4623, "CSG67");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-186,230,110,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"CSG67\",DATUM[\"Centre_Spatial_Guyanais_1967\",");
    add_srs_wkt (p, 1,
		 "SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7022\"]],TOWGS84[-186,230,110,0,0,0,0],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6623\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4623\"]");
    add_srs_wkt (p, 6, "]");
    p = add_epsg_def (first, last, 4624, "epsg", 4624, "RGFG95");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=2,2,-2,0,0,0,0 +no_d");
    add_proj4text (p, 1, "efs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"RGFG95\",DATUM[\"Reseau_Geodesique_Francais_Guy");
    add_srs_wkt (p, 1,
		 "ane_1995\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[2,2,-2,0,0,0,0],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6624\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6, "624\"]]");
    p = add_epsg_def (first, last, 4625, "epsg", 4625, "Martinique 1938");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Martinique 1938\",DATUM[\"Martinique_1938\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"International 1924\",6378388,297,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7022\"]],AUTHORITY[\"EPSG\",\"6625\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 3,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 4,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 5, "RITY[\"EPSG\",\"4625\"]]");
    p = add_epsg_def (first, last, 4626, "epsg", 4626, "Reunion 1947");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Reunion 1947\",DATUM[\"Reunion_1947\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7022\"]],AUTHORITY[\"EPSG\",\"6626\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 3,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 4,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 5, "\"EPSG\",\"4626\"]]");
    p = add_epsg_def (first, last, 4627, "epsg", 4627, "RGR92");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"RGR92\",DATUM[\"Reseau_Geodesique_de_la_Reunion");
    add_srs_wkt (p, 1,
		 "_1992\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH");
    add_srs_wkt (p, 2,
		 "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6627\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4627\"");
    add_srs_wkt (p, 6, "]]");
    p = add_epsg_def (first, last, 4628, "epsg", 4628, "Tahiti 52");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=162,117,154,0,0,0,0 +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Tahiti 52\",DATUM[\"Tahiti_52\",SPHEROID[\"Inte");
    add_srs_wkt (p, 1,
		 "rnational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"");
    add_srs_wkt (p, 2,
		 "]],TOWGS84[162,117,154,0,0,0,0],AUTHORITY[\"EPSG\",\"662");
    add_srs_wkt (p, 3,
		 "8\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4628\"]]");
    p = add_epsg_def (first, last, 4629, "epsg", 4629, "Tahaa 54");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Tahaa 54\",DATUM[\"Tahaa_54\",SPHEROID[\"Intern");
    add_srs_wkt (p, 1,
		 "ational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]]");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"6629\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 4,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"4629\"]]");
    p = add_epsg_def (first, last, 4630, "epsg", 4630, "IGN72 Nuku Hiva");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"IGN72 Nuku Hiva\",DATUM[\"IGN72_Nuku_Hiva\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"International 1924\",6378388,297,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7022\"]],AUTHORITY[\"EPSG\",\"6630\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 3,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 4,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 5, "RITY[\"EPSG\",\"4630\"]]");
    p = add_epsg_def (first, last, 4631, "epsg", 4631, "K0 1949");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=145,-187,103,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"K0 1949\",DATUM[\"K0_1949\",SPHEROID[\"Internat");
    add_srs_wkt (p, 1,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],T");
    add_srs_wkt (p, 2,
		 "OWGS84[145,-187,103,0,0,0,0],AUTHORITY[\"EPSG\",\"6631\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5, "9122\"]],AUTHORITY[\"EPSG\",\"4631\"]]");
    p = add_epsg_def (first, last, 4632, "epsg", 4632, "Combani 1950");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-382,-59,-262,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Combani 1950\",DATUM[\"Combani_1950\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7022\"]],TOWGS84[-382,-59,-262,0,0,0,0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"6632\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 5, "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4632\"]]");
    p = add_epsg_def (first, last, 4633, "epsg", 4633, "IGN56 Lifou");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"IGN56 Lifou\",DATUM[\"IGN56_Lifou\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 2,
		 "22\"]],AUTHORITY[\"EPSG\",\"6633\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 3,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 4,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 5, "PSG\",\"4633\"]]");
    p = add_epsg_def (first, last, 4634, "epsg", 4634, "IGN72 Grand Terre");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"IGN72 Grand Terre\",DATUM[\"IGN72_Grande_Terre\"");
    add_srs_wkt (p, 1,
		 ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6634\"]],PRIMEM[\"");
    add_srs_wkt (p, 3,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 4,
		 "e\",0.0174532925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUT");
    add_srs_wkt (p, 5, "HORITY[\"EPSG\",\"4634\"]]");
    p = add_epsg_def (first, last, 4635, "epsg", 4635, "ST87 Ouvea");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-122.383,-188.696,103");
    add_proj4text (p, 1, ".344,3.5107,-4.9668,-5.7047,4.4798 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"ST87 Ouvea\",DATUM[\"ST87_Ouvea\",SPHEROID[\"In");
    add_srs_wkt (p, 1,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 2,
		 "\"]],TOWGS84[-122.383,-188.696,103.344,3.5107,-4.9668,-5");
    add_srs_wkt (p, 3,
		 ".7047,4.4798],AUTHORITY[\"EPSG\",\"6635\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6, "RITY[\"EPSG\",\"4635\"]]");
    p = add_epsg_def (first, last, 4636, "epsg", 4636, "Petrels 1972");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=365,194,166,0,0,0,0 +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Petrels 1972\",DATUM[\"Petrels_1972\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7022\"]],TOWGS84[365,194,166,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6636\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4636\"]]");
    p = add_epsg_def (first, last, 4637, "epsg", 4637, "Perroud 1950");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=325,154,172,0,0,0,0 +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Perroud 1950\",DATUM[\"Pointe_Geologie_Perroud_");
    add_srs_wkt (p, 1,
		 "1950\",SPHEROID[\"International 1924\",6378388,297,AUTHO");
    add_srs_wkt (p, 2,
		 "RITY[\"EPSG\",\"7022\"]],TOWGS84[325,154,172,0,0,0,0],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6637\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6, "4637\"]]");
    p = add_epsg_def (first, last, 4638, "epsg", 4638,
		      "Saint Pierre et Miquelon 1950");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk66 +towgs84=30,430,368,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Saint Pierre et Miquelon 1950\",DATUM[\"Saint_P");
    add_srs_wkt (p, 1,
		 "ierre_et_Miquelon_1950\",SPHEROID[\"Clarke 1866\",637820");
    add_srs_wkt (p, 2,
		 "6.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],TOWG");
    add_srs_wkt (p, 3,
		 "S84[30,430,368,0,0,0,0],AUTHORITY[\"EPSG\",\"6638\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6, "\"]],AUTHORITY[\"EPSG\",\"4638\"]]");
    p = add_epsg_def (first, last, 4639, "epsg", 4639, "MOP78");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"MOP78\",DATUM[\"MOP78\",SPHEROID[\"Internationa");
    add_srs_wkt (p, 1,
		 "l 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHO");
    add_srs_wkt (p, 2,
		 "RITY[\"EPSG\",\"6639\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 4,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"463");
    add_srs_wkt (p, 5, "9\"]]");
    p = add_epsg_def (first, last, 4640, "epsg", 4640, "RRAF 1991");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"RRAF 1991\",DATUM[\"Reseau_de_Reference_des_Ant");
    add_srs_wkt (p, 1,
		 "illes_Francaises_1991\",SPHEROID[\"WGS 84\",6378137,298.");
    add_srs_wkt (p, 2,
		 "257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,");
    add_srs_wkt (p, 3,
		 "0,0,0],AUTHORITY[\"EPSG\",\"6640\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6, "PSG\",\"4640\"]]");
    p = add_epsg_def (first, last, 4641, "epsg", 4641, "IGN53 Mare");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"IGN53 Mare\",DATUM[\"IGN53_Mare\",SPHEROID[\"In");
    add_srs_wkt (p, 1,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 2,
		 "\"]],AUTHORITY[\"EPSG\",\"6641\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 3,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 4,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 5, "SG\",\"4641\"]]");
    p = add_epsg_def (first, last, 4642, "epsg", 4642, "ST84 Ile des Pins");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"ST84 Ile des Pins\",DATUM[\"ST84_Ile_des_Pins\"");
    add_srs_wkt (p, 1,
		 ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6642\"]],PRIMEM[\"");
    add_srs_wkt (p, 3,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 4,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 5, "THORITY[\"EPSG\",\"4642\"]]");
    p = add_epsg_def (first, last, 4643, "epsg", 4643, "ST71 Belep");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-480.26,-438.32,-643.");
    add_proj4text (p, 1, "429,16.3119,20.1721,-4.0349,-111.7 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"ST71 Belep\",DATUM[\"ST71_Belep\",SPHEROID[\"In");
    add_srs_wkt (p, 1,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 2,
		 "\"]],TOWGS84[-480.26,-438.32,-643.429,16.3119,20.1721,-4");
    add_srs_wkt (p, 3,
		 ".0349,-111.7],AUTHORITY[\"EPSG\",\"6643\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6, "RITY[\"EPSG\",\"4643\"]]");
    p = add_epsg_def (first, last, 4644, "epsg", 4644, "NEA74 Noumea");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"NEA74 Noumea\",DATUM[\"NEA74_Noumea\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7022\"]],AUTHORITY[\"EPSG\",\"6644\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 3,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 4,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 5, "\"EPSG\",\"4644\"]]");
    p = add_epsg_def (first, last, 4645, "epsg", 4645, "RGNC 1991");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=0,0,0,0,0,0,0 +no_def");
    add_proj4text (p, 1, "s");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"RGNC 1991\",DATUM[\"Reseau_Geodesique_Nouvelle_");
    add_srs_wkt (p, 1,
		 "Caledonie_1991\",SPHEROID[\"International 1924\",6378388");
    add_srs_wkt (p, 2,
		 ",297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6645\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6, ",\"4645\"]]");
    p = add_epsg_def (first, last, 4646, "epsg", 4646, "Grand Comoros");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Grand Comoros\",DATUM[\"Grand_Comoros\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 2,
		 "\"7022\"]],AUTHORITY[\"EPSG\",\"6646\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 3,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 4,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"4646\"]]");
    p = add_epsg_def (first, last, 4657, "epsg", 4657, "Reykjavik 1900");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377019.27 +b=6355762.5391 +towgs84=-28");
    add_proj4text (p, 1, ",199,5,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Reykjavik 1900\",DATUM[\"Reykjavik_1900\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"Danish 1876\",6377019.27,300,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7051\"]],TOWGS84[-28,199,5,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "6657\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 4,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 5, "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4657\"]]");
    p = add_epsg_def (first, last, 4658, "epsg", 4658, "Hjorsey 1955");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-73,46,-86,0,0,0,0 +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Hjorsey 1955\",DATUM[\"Hjorsey_1955\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7022\"]],TOWGS84[-73,46,-86,0,0,0,0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"6658\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 4,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4658\"]]");
    p = add_epsg_def (first, last, 4659, "epsg", 4659, "ISN93");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"ISN93\",DATUM[\"Islands_Network_1993\",SPHEROID");
    add_srs_wkt (p, 1,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"665");
    add_srs_wkt (p, 3,
		 "9\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4659\"]]");
    p = add_epsg_def (first, last, 4660, "epsg", 4660, "Helle 1954");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=982.609,552.753,-540.");
    add_proj4text (p, 1, "873,32.3934,-153.257,-96.2266,16.805 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Helle 1954\",DATUM[\"Helle_1954\",SPHEROID[\"In");
    add_srs_wkt (p, 1,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 2,
		 "\"]],TOWGS84[982.609,552.753,-540.873,32.3934,-153.257,-");
    add_srs_wkt (p, 3,
		 "96.2266,16.805],AUTHORITY[\"EPSG\",\"6660\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6, "HORITY[\"EPSG\",\"4660\"]]");
    p = add_epsg_def (first, last, 4661, "epsg", 4661, "LKS92");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"LKS92\",DATUM[\"Latvia_1992\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 1,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 2,
		 ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6661\"]],PR");
    add_srs_wkt (p, 3,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 4,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 5, "\"]],AUTHORITY[\"EPSG\",\"4661\"]]");
    p = add_epsg_def (first, last, 4662, "epsg", 4662, "IGN72 Grande Terre");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"IGN72 Grande Terre\",DATUM[\"IGN72_Grande_Terre");
    add_srs_wkt (p, 1,
		 "\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY");
    add_srs_wkt (p, 2,
		 "[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6634\"]],PRIME");
    add_srs_wkt (p, 3,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 4,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 5, "],AUTHORITY[\"EPSG\",\"4662\"]]");
    p = add_epsg_def (first, last, 4663, "epsg", 4663, "Porto Santo 1995");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Porto Santo 1995\",DATUM[\"Porto_Santo_1995\",S");
    add_srs_wkt (p, 1,
		 "PHEROID[\"International 1924\",6378388,297,AUTHORITY[\"E");
    add_srs_wkt (p, 2,
		 "PSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6663\"]],PRIMEM[\"");
    add_srs_wkt (p, 3,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 4,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 5, "THORITY[\"EPSG\",\"4663\"]]");
    p = add_epsg_def (first, last, 4664, "epsg", 4664, "Azores Oriental 1995");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Azores Oriental 1995\",DATUM[\"Azores_Oriental_");
    add_srs_wkt (p, 1,
		 "Islands_1995\",SPHEROID[\"International 1924\",6378388,2");
    add_srs_wkt (p, 2,
		 "97,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"66");
    add_srs_wkt (p, 3,
		 "64\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4664\"]]");
    p = add_epsg_def (first, last, 4665, "epsg", 4665, "Azores Central 1995");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Azores Central 1995\",DATUM[\"Azores_Central_Is");
    add_srs_wkt (p, 1,
		 "lands_1995\",SPHEROID[\"International 1924\",6378388,297");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6665");
    add_srs_wkt (p, 3,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4665\"]]");
    p = add_epsg_def (first, last, 4666, "epsg", 4666, "Lisbon 1890");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Lisbon 1890\",DATUM[\"Lisbon_1890\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 2,
		 ",\"7004\"]],AUTHORITY[\"EPSG\",\"6666\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 3,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 4,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 5, "TY[\"EPSG\",\"4666\"]]");
    p = add_epsg_def (first, last, 4667, "epsg", 4667, "IKBD-92");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"IKBD-92\",DATUM[\"Iraq_Kuwait_Boundary_Datum_19");
    add_srs_wkt (p, 1,
		 "92\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY");
    add_srs_wkt (p, 2,
		 "[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6667\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4667\"]]");
    p = add_epsg_def (first, last, 4668, "epsg", 4668, "ED79");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-86,-98,-119,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"ED79\",DATUM[\"European_Datum_1979\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 2,
		 "22\"]],TOWGS84[-86,-98,-119,0,0,0,0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"6668\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 4,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4668\"]]");
    p = add_epsg_def (first, last, 4669, "epsg", 4669, "LKS94");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"LKS94\",DATUM[\"Lithuania_1994_ETRS89\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 2,
		 "\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "126\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5, "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4669\"]]");
    p = add_epsg_def (first, last, 4670, "epsg", 4670, "IGM95");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"IGM95\",DATUM[\"Istituto_Geografico_Militaire_1");
    add_srs_wkt (p, 1,
		 "995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6670\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4670\"]]");
    p = add_epsg_def (first, last, 4671, "epsg", 4671, "Voirol 1879");
    add_proj4text (p, 0, "+proj=longlat +a=6378249.2 +b=6356515 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Voirol 1879\",DATUM[\"Voirol_1879\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "Clarke 1880 (IGN)\",6378249.2,293.4660212936265,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6671\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5, "]],AUTHORITY[\"EPSG\",\"4671\"]]");
    p = add_epsg_def (first, last, 4672, "epsg", 4672, "Chatham Islands 1971");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=175,-38,113,0,0,0,0 +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Chatham Islands 1971\",DATUM[\"Chatham_Islands_");
    add_srs_wkt (p, 1,
		 "Datum_1971\",SPHEROID[\"International 1924\",6378388,297");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[175,-38,113,0,0,0");
    add_srs_wkt (p, 3,
		 ",0],AUTHORITY[\"EPSG\",\"6672\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 5,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 6, "G\",\"4672\"]]");
    p = add_epsg_def (first, last, 4673, "epsg", 4673, "Chatham Islands 1979");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=174.05,-25.49,112.57,");
    add_proj4text (p, 1, "-0,-0,0.554,0.2263 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Chatham Islands 1979\",DATUM[\"Chatham_Islands_");
    add_srs_wkt (p, 1,
		 "Datum_1979\",SPHEROID[\"International 1924\",6378388,297");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[174.05,-25.49,112");
    add_srs_wkt (p, 3,
		 ".57,-0,-0,0.554,0.2263],AUTHORITY[\"EPSG\",\"6673\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6, "\"]],AUTHORITY[\"EPSG\",\"4673\"]]");
    p = add_epsg_def (first, last, 4674, "epsg", 4674, "SIRGAS 2000");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema_de_Referencia_Geo");
    add_srs_wkt (p, 1,
		 "centrico_para_America_del_Sur_2000\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6, "]],AUTHORITY[\"EPSG\",\"4674\"]]");
    p = add_epsg_def (first, last, 4675, "epsg", 4675, "Guam 1963");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk66 +towgs84=-100,-248,259,0,0,0");
    add_proj4text (p, 1, ",0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Guam 1963\",DATUM[\"Guam_1963\",SPHEROID[\"Clar");
    add_srs_wkt (p, 1,
		 "ke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 2,
		 ",\"7008\"]],TOWGS84[-100,-248,259,0,0,0,0],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6675\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5, "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4675\"]]");
    p = add_epsg_def (first, last, 4676, "epsg", 4676, "Vientiane 1982");
    add_proj4text (p, 0, "+proj=longlat +ellps=krass +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Vientiane 1982\",DATUM[\"Vientiane_1982\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 2,
		 ",\"7024\"]],AUTHORITY[\"EPSG\",\"6676\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 3,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 4,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 5, "TY[\"EPSG\",\"4676\"]]");
    p = add_epsg_def (first, last, 4677, "epsg", 4677, "Lao 1993");
    add_proj4text (p, 0, "+proj=longlat +ellps=krass +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Lao 1993\",DATUM[\"Lao_1993\",SPHEROID[\"Krasso");
    add_srs_wkt (p, 1,
		 "wsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"6677\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 4,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5, "\"4677\"]]");
    p = add_epsg_def (first, last, 4678, "epsg", 4678, "Lao 1997");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=krass +towgs84=44.585,-131.212,-39.");
    add_proj4text (p, 1, "544,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Lao 1997\",DATUM[\"Lao_National_Datum_1997\",SP");
    add_srs_wkt (p, 1,
		 "HEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7024\"]],TOWGS84[44.585,-131.212,-39.544,0,0,0,0],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6678\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6, "\"4678\"]]");
    p = add_epsg_def (first, last, 4679, "epsg", 4679, "Jouik 1961");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=-80.01,253.26,291.1");
    add_proj4text (p, 1, "9,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Jouik 1961\",DATUM[\"Jouik_1961\",SPHEROID[\"Cl");
    add_srs_wkt (p, 1,
		 "arke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 2,
		 ",\"7012\"]],TOWGS84[-80.01,253.26,291.19,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6679\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4679");
    add_srs_wkt (p, 6, "\"]]");
    p = add_epsg_def (first, last, 4680, "epsg", 4680, "Nouakchott 1965");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=124.5,-63.5,-281,0,");
    add_proj4text (p, 1, "0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Nouakchott 1965\",DATUM[\"Nouakchott_1965\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7012\"]],TOWGS84[124.5,-63.5,-281,0,0,0,0],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6680\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6, "\"4680\"]]");
    p = add_epsg_def (first, last, 4681, "epsg", 4681, "Mauritania 1999");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Mauritania 1999\",DATUM[\"Mauritania_1999\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6681\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5, "]],AUTHORITY[\"EPSG\",\"4681\"]]");
    p = add_epsg_def (first, last, 4682, "epsg", 4682, "Gulshan 303");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377276.345 +b=6356075.41314024 +no_def");
    add_proj4text (p, 1, "s");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Gulshan 303\",DATUM[\"Gulshan_303\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AU");
    add_srs_wkt (p, 2,
		 "THORITY[\"EPSG\",\"7015\"]],AUTHORITY[\"EPSG\",\"6682\"]");
    add_srs_wkt (p, 3,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 4,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5, "9122\"]],AUTHORITY[\"EPSG\",\"4682\"]]");
    p = add_epsg_def (first, last, 4683, "epsg", 4683, "PRS92");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk66 +towgs84=-127.62,-67.24,-47.");
    add_proj4text (p, 1, "04,-3.068,4.903,1.578,-1.06 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"PRS92\",DATUM[\"Philippine_Reference_System_199");
    add_srs_wkt (p, 1,
		 "2\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982139006");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-127.62,-67.24,-4");
    add_srs_wkt (p, 3,
		 "7.04,-3.068,4.903,1.578,-1.06],AUTHORITY[\"EPSG\",\"6683");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4683\"]]");
    p = add_epsg_def (first, last, 4684, "epsg", 4684, "Gan 1970");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-133,-321,50,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Gan 1970\",DATUM[\"Gan_1970\",SPHEROID[\"Intern");
    add_srs_wkt (p, 1,
		 "ational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]]");
    add_srs_wkt (p, 2,
		 ",TOWGS84[-133,-321,50,0,0,0,0],AUTHORITY[\"EPSG\",\"6684");
    add_srs_wkt (p, 3,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4684\"]]");
    p = add_epsg_def (first, last, 4685, "epsg", 4685, "Gandajika");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Gandajika\",DATUM[\"Gandajika\",SPHEROID[\"Inte");
    add_srs_wkt (p, 1,
		 "rnational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"");
    add_srs_wkt (p, 2,
		 "]],AUTHORITY[\"EPSG\",\"6685\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 4,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 5, "\",\"4685\"]]");
    p = add_epsg_def (first, last, 4686, "epsg", 4686, "MAGNA-SIRGAS");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco_Geocentrico_Nacion");
    add_srs_wkt (p, 1,
		 "al_de_Referencia\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 2,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 3,
		 ",0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 5,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 6, "G\",\"4686\"]]");
    p = add_epsg_def (first, last, 4687, "epsg", 4687, "RGPF");
    add_proj4text (p, 0, "+proj=longlat +ellps=GRS80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"RGPF\",DATUM[\"Reseau_Geodesique_de_la_Polynesi");
    add_srs_wkt (p, 1,
		 "e_Francaise\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 2,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"668");
    add_srs_wkt (p, 3,
		 "7\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4687\"]]");
    p = add_epsg_def (first, last, 4688, "epsg", 4688, "Fatu Iva 72");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=347.103,1078.12,2623.");
    add_proj4text (p, 1, "92,-33.8875,70.6773,-9.3943,186.074 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Fatu Iva 72\",DATUM[\"Fatu_Iva_72\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 2,
		 "22\"]],TOWGS84[347.103,1078.12,2623.92,-33.8875,70.6773,");
    add_srs_wkt (p, 3,
		 "-9.3943,186.074],AUTHORITY[\"EPSG\",\"6688\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6, "THORITY[\"EPSG\",\"4688\"]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_17 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 4689, "epsg", 4689, "IGN63 Hiva Oa");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"IGN63 Hiva Oa\",DATUM[\"IGN63_Hiva_Oa\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 2,
		 "\"7022\"]],AUTHORITY[\"EPSG\",\"6689\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 3,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 4,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"4689\"]]");
    p = add_epsg_def (first, last, 4690, "epsg", 4690, "Tahiti 79");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Tahiti 79\",DATUM[\"Tahiti_79\",SPHEROID[\"Inte");
    add_srs_wkt (p, 1,
		 "rnational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"");
    add_srs_wkt (p, 2,
		 "]],AUTHORITY[\"EPSG\",\"6690\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 4,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 5, "\",\"4690\"]]");
    p = add_epsg_def (first, last, 4691, "epsg", 4691, "Moorea 87");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=215.525,149.593,176.2");
    add_proj4text (p, 1, "29,-3.2624,-1.692,-1.1571,10.4773 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Moorea 87\",DATUM[\"Moorea_87\",SPHEROID[\"Inte");
    add_srs_wkt (p, 1,
		 "rnational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"");
    add_srs_wkt (p, 2,
		 "]],TOWGS84[215.525,149.593,176.229,-3.2624,-1.692,-1.157");
    add_srs_wkt (p, 3,
		 "1,10.4773],AUTHORITY[\"EPSG\",\"6691\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6, "Y[\"EPSG\",\"4691\"]]");
    p = add_epsg_def (first, last, 4692, "epsg", 4692, "Maupiti 83");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=217.037,86.959,23.956");
    add_proj4text (p, 1, ",0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Maupiti 83\",DATUM[\"Maupiti_83\",SPHEROID[\"In");
    add_srs_wkt (p, 1,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 2,
		 "\"]],TOWGS84[217.037,86.959,23.956,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6692\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4692\"]]");
    p = add_epsg_def (first, last, 4693, "epsg", 4693, "Nakhl-e Ghanem");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=WGS84 +towgs84=0,-0.15,0.68,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Nakhl-e Ghanem\",DATUM[\"Nakhl_e_Ghanem\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 2,
		 "\"7030\"]],TOWGS84[0,-0.15,0.68,0,0,0,0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"6693\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4693\"]]");
    p = add_epsg_def (first, last, 4694, "epsg", 4694, "POSGAR 94");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"POSGAR 94\",DATUM[\"Posiciones_Geodesicas_Argen");
    add_srs_wkt (p, 1,
		 "tinas_1994\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6694\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"46");
    add_srs_wkt (p, 6, "94\"]]");
    p = add_epsg_def (first, last, 4695, "epsg", 4695, "Katanga 1955");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk66 +towgs84=-103.746,-9.614,-25");
    add_proj4text (p, 1, "5.95,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Katanga 1955\",DATUM[\"Katanga_1955\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7008\"]],TOWGS84[-103.746,-9.614,-255.95,0,0,0,");
    add_srs_wkt (p, 3,
		 "0],AUTHORITY[\"EPSG\",\"6695\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6, "\",\"4695\"]]");
    p = add_epsg_def (first, last, 4696, "epsg", 4696, "Kasai 1953");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Kasai 1953\",DATUM[\"Kasai_1953\",SPHEROID[\"Cl");
    add_srs_wkt (p, 1,
		 "arke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 2,
		 ",\"7012\"]],AUTHORITY[\"EPSG\",\"6696\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 3,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 4,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 5, "TY[\"EPSG\",\"4696\"]]");
    p = add_epsg_def (first, last, 4697, "epsg", 4697,
		      "IGC 1962 6th Parallel South");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"IGC 1962 6th Parallel South\",DATUM[\"IGC_1962_");
    add_srs_wkt (p, 1,
		 "Arc_of_the_6th_Parallel_South\",SPHEROID[\"Clarke 1880 (");
    add_srs_wkt (p, 2,
		 "RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6697\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6, ",\"4697\"]]");
    p = add_epsg_def (first, last, 4698, "epsg", 4698, "IGN 1962 Kerguelen");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=145,-187,103,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"IGN 1962 Kerguelen\",DATUM[\"IGN_1962_Kerguelen");
    add_srs_wkt (p, 1,
		 "\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY");
    add_srs_wkt (p, 2,
		 "[\"EPSG\",\"7022\"]],TOWGS84[145,-187,103,0,0,0,0],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6698\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"469");
    add_srs_wkt (p, 6, "8\"]]");
    p = add_epsg_def (first, last, 4699, "epsg", 4699, "Le Pouce 1934");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=-770.1,158.4,-498.2");
    add_proj4text (p, 1, ",0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Le Pouce 1934\",DATUM[\"Le_Pouce_1934\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7012\"]],TOWGS84[-770.1,158.4,-498.2,0,0,0,0],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6699\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6, "4699\"]]");
    p = add_epsg_def (first, last, 4700, "epsg", 4700, "IGN Astro 1960");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"IGN Astro 1960\",DATUM[\"IGN_Astro_1960\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[");
    add_srs_wkt (p, 2,
		 "\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6700\"]],PRIMEM");
    add_srs_wkt (p, 3,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 4,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 5, ",AUTHORITY[\"EPSG\",\"4700\"]]");
    p = add_epsg_def (first, last, 4701, "epsg", 4701, "IGCB 1955");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=-79.9,-158,-168.9,0");
    add_proj4text (p, 1, ",0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"IGCB 1955\",DATUM[\"Institut_Geographique_du_Co");
    add_srs_wkt (p, 1,
		 "ngo_Belge_1955\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.");
    add_srs_wkt (p, 2,
		 "145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-79.9,");
    add_srs_wkt (p, 3,
		 "-158,-168.9,0,0,0,0],AUTHORITY[\"EPSG\",\"6701\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6, "],AUTHORITY[\"EPSG\",\"4701\"]]");
    p = add_epsg_def (first, last, 4702, "epsg", 4702, "Mauritania 1999");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Mauritania 1999\",DATUM[\"Mauritania_1999\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6702\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4702\"]]");
    p = add_epsg_def (first, last, 4703, "epsg", 4703, "Mhast 1951");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Mhast 1951\",DATUM[\"Missao_Hidrografico_Angola");
    add_srs_wkt (p, 1,
		 "_y_Sao_Tome_1951\",SPHEROID[\"Clarke 1880 (RGS)\",637824");
    add_srs_wkt (p, 2,
		 "9.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6703\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4703\"]]");
    p = add_epsg_def (first, last, 4704, "epsg", 4704, "Mhast (onshore)");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Mhast (onshore)\",DATUM[\"Mhast_onshore\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 2,
		 ",\"7022\"]],AUTHORITY[\"EPSG\",\"6704\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 3,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 4,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 5, "TY[\"EPSG\",\"4704\"]]");
    p = add_epsg_def (first, last, 4705, "epsg", 4705, "Mhast (offshore)");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Mhast (offshore)\",DATUM[\"Mhast_offshore\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"International 1924\",6378388,297,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7022\"]],AUTHORITY[\"EPSG\",\"6705\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 3,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 4,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 5, "RITY[\"EPSG\",\"4705\"]]");
    p = add_epsg_def (first, last, 4706, "epsg", 4706,
		      "Egypt Gulf of Suez S-650 TL");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=helmert +towgs84=-146.21,112.63,4.0");
    add_proj4text (p, 1, "5,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Egypt Gulf of Suez S-650 TL\",DATUM[\"Egypt_Gul");
    add_srs_wkt (p, 1,
		 "f_of_Suez_S_650_TL\",SPHEROID[\"Helmert 1906\",6378200,2");
    add_srs_wkt (p, 2,
		 "98.3,AUTHORITY[\"EPSG\",\"7020\"]],TOWGS84[-146.21,112.6");
    add_srs_wkt (p, 3,
		 "3,4.05,0,0,0,0],AUTHORITY[\"EPSG\",\"6706\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6, "HORITY[\"EPSG\",\"4706\"]]");
    p = add_epsg_def (first, last, 4707, "epsg", 4707, "Tern Island 1961");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=114,-116,-333,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Tern Island 1961\",DATUM[\"Tern_Island_1961\",S");
    add_srs_wkt (p, 1,
		 "PHEROID[\"International 1924\",6378388,297,AUTHORITY[\"E");
    add_srs_wkt (p, 2,
		 "PSG\",\"7022\"]],TOWGS84[114,-116,-333,0,0,0,0],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6707\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4707\"]");
    add_srs_wkt (p, 6, "]");
    p = add_epsg_def (first, last, 4708, "epsg", 4708, "Cocos Islands 1965");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=aust_SA +towgs84=-491,-22,435,0,0,0");
    add_proj4text (p, 1, ",0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Cocos Islands 1965\",DATUM[\"Cocos_Islands_1965");
    add_srs_wkt (p, 1,
		 "\",SPHEROID[\"Australian National Spheroid\",6378160,298");
    add_srs_wkt (p, 2,
		 ".25,AUTHORITY[\"EPSG\",\"7003\"]],TOWGS84[-491,-22,435,0");
    add_srs_wkt (p, 3,
		 ",0,0,0],AUTHORITY[\"EPSG\",\"6708\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6, "EPSG\",\"4708\"]]");
    p = add_epsg_def (first, last, 4709, "epsg", 4709, "Iwo Jima 1945");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=145,75,-272,0,0,0,0 +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Iwo Jima 1945\",DATUM[\"Iwo_Jima_1945\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 2,
		 "\"7022\"]],TOWGS84[145,75,-272,0,0,0,0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"6709\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 5, "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4709\"]]");
    p = add_epsg_def (first, last, 4710, "epsg", 4710, "St. Helena 1971");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-320,550,-494,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"St. Helena 1971\",DATUM[\"St_Helena_1971\",SPHE");
    add_srs_wkt (p, 1,
		 "ROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 2,
		 "\",\"7022\"]],TOWGS84[-320,550,-494,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6710\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4710\"]]");
    p = add_epsg_def (first, last, 4711, "epsg", 4711, "Marcus Island 1952");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=124,-234,-25,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Marcus Island 1952\",DATUM[\"Marcus_Island_1952");
    add_srs_wkt (p, 1,
		 "\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY");
    add_srs_wkt (p, 2,
		 "[\"EPSG\",\"7022\"]],TOWGS84[124,-234,-25,0,0,0,0],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6711\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"471");
    add_srs_wkt (p, 6, "1\"]]");
    p = add_epsg_def (first, last, 4712, "epsg", 4712, "Ascension Island 1958");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-205,107,53,0,0,0,0 +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Ascension Island 1958\",DATUM[\"Ascension_Islan");
    add_srs_wkt (p, 1,
		 "d_1958\",SPHEROID[\"International 1924\",6378388,297,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7022\"]],TOWGS84[-205,107,53,0,0,0,0],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6712\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6, "\"4712\"]]");
    p = add_epsg_def (first, last, 4713, "epsg", 4713, "Ayabelle Lighthouse");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=-79,-129,145,0,0,0,");
    add_proj4text (p, 1, "0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Ayabelle Lighthouse\",DATUM[\"Ayabelle_Lighthou");
    add_srs_wkt (p, 1,
		 "se\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-79,-129,145,0,0,0");
    add_srs_wkt (p, 3,
		 ",0],AUTHORITY[\"EPSG\",\"6713\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 5,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 6, "G\",\"4713\"]]");
    p = add_epsg_def (first, last, 4714, "epsg", 4714, "Bellevue");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-127,-769,472,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Bellevue\",DATUM[\"Bellevue\",SPHEROID[\"Intern");
    add_srs_wkt (p, 1,
		 "ational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]]");
    add_srs_wkt (p, 2,
		 ",TOWGS84[-127,-769,472,0,0,0,0],AUTHORITY[\"EPSG\",\"671");
    add_srs_wkt (p, 3,
		 "4\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4714\"]]");
    p = add_epsg_def (first, last, 4715, "epsg", 4715, "Camp Area Astro");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-104,-129,239,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Camp Area Astro\",DATUM[\"Camp_Area_Astro\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"International 1924\",6378388,297,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7022\"]],TOWGS84[-104,-129,239,0,0,0,0],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6715\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4715\"]]");
    add_srs_wkt (p, 6, "");
    p = add_epsg_def (first, last, 4716, "epsg", 4716, "Phoenix Islands 1966");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=298,-304,-375,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Phoenix Islands 1966\",DATUM[\"Phoenix_Islands_");
    add_srs_wkt (p, 1,
		 "1966\",SPHEROID[\"International 1924\",6378388,297,AUTHO");
    add_srs_wkt (p, 2,
		 "RITY[\"EPSG\",\"7022\"]],TOWGS84[298,-304,-375,0,0,0,0],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6716\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6, "\"4716\"]]");
    p = add_epsg_def (first, last, 4717, "epsg", 4717, "Cape Canaveral");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk66 +towgs84=-2,151,181,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Cape Canaveral\",DATUM[\"Cape_Canaveral\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7008\"]],TOWGS84[-2,151,181,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6717\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4717");
    add_srs_wkt (p, 6, "\"]]");
    p = add_epsg_def (first, last, 4718, "epsg", 4718, "Solomon 1968");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Solomon 1968\",DATUM[\"Solomon_1968\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7022\"]],AUTHORITY[\"EPSG\",\"6718\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 3,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 4,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 5, "\"EPSG\",\"4718\"]]");
    p = add_epsg_def (first, last, 4719, "epsg", 4719, "Easter Island 1967");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=211,147,111,0,0,0,0 +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Easter Island 1967\",DATUM[\"Easter_Island_1967");
    add_srs_wkt (p, 1,
		 "\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY");
    add_srs_wkt (p, 2,
		 "[\"EPSG\",\"7022\"]],TOWGS84[211,147,111,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6719\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4719");
    add_srs_wkt (p, 6, "\"]]");
    p = add_epsg_def (first, last, 4720, "epsg", 4720, "Fiji 1986");
    add_proj4text (p, 0, "+proj=longlat +ellps=WGS72 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Fiji 1986\",DATUM[\"Fiji_Geodetic_Datum_1986\",");
    add_srs_wkt (p, 1,
		 "SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7043\"]],AUTHORITY[\"EPSG\",\"6720\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 3,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 4,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 5, "\"EPSG\",\"4720\"]]");
    p = add_epsg_def (first, last, 4721, "epsg", 4721, "Fiji 1956");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=265.025,384.929,-194.");
    add_proj4text (p, 1, "046,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Fiji 1956\",DATUM[\"Fiji_1956\",SPHEROID[\"Inte");
    add_srs_wkt (p, 1,
		 "rnational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"");
    add_srs_wkt (p, 2,
		 "]],TOWGS84[265.025,384.929,-194.046,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6721\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4721\"]]");
    p = add_epsg_def (first, last, 4722, "epsg", 4722, "South Georgia 1968");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-794,119,-298,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"South Georgia 1968\",DATUM[\"South_Georgia_1968");
    add_srs_wkt (p, 1,
		 "\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY");
    add_srs_wkt (p, 2,
		 "[\"EPSG\",\"7022\"]],TOWGS84[-794,119,-298,0,0,0,0],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6722\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"47");
    add_srs_wkt (p, 6, "22\"]]");
    p = add_epsg_def (first, last, 4723, "epsg", 4723, "Grand Cayman 1959");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk66 +towgs84=67.8,106.1,138.8,0,");
    add_proj4text (p, 1, "0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Grand Cayman 1959\",DATUM[\"Grand_Cayman_1959\"");
    add_srs_wkt (p, 1,
		 ",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982139006,AU");
    add_srs_wkt (p, 2,
		 "THORITY[\"EPSG\",\"7008\"]],TOWGS84[67.8,106.1,138.8,0,0");
    add_srs_wkt (p, 3,
		 ",0,0],AUTHORITY[\"EPSG\",\"6723\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6, "PSG\",\"4723\"]]");
    p = add_epsg_def (first, last, 4724, "epsg", 4724, "Diego Garcia 1969");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=208,-435,-229,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Diego Garcia 1969\",DATUM[\"Diego_Garcia_1969\"");
    add_srs_wkt (p, 1,
		 ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7022\"]],TOWGS84[208,-435,-229,0,0,0,0],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6724\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4724\"");
    add_srs_wkt (p, 6, "]]");
    p = add_epsg_def (first, last, 4725, "epsg", 4725, "Johnston Island 1961");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=189,-79,-202,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Johnston Island 1961\",DATUM[\"Johnston_Island_");
    add_srs_wkt (p, 1,
		 "1961\",SPHEROID[\"International 1924\",6378388,297,AUTHO");
    add_srs_wkt (p, 2,
		 "RITY[\"EPSG\",\"7022\"]],TOWGS84[189,-79,-202,0,0,0,0],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6725\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6, "4725\"]]");
    p = add_epsg_def (first, last, 4726, "epsg", 4726, "Little Cayman 1961");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk66 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Little Cayman 1961\",DATUM[\"Little_Cayman_1961");
    add_srs_wkt (p, 1,
		 "\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982139006,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6726\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5, "9122\"]],AUTHORITY[\"EPSG\",\"4726\"]]");
    p = add_epsg_def (first, last, 4727, "epsg", 4727, "Midway 1961");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Midway 1961\",DATUM[\"Midway_1961\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 2,
		 "22\"]],AUTHORITY[\"EPSG\",\"6727\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 3,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 4,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 5, "PSG\",\"4727\"]]");
    p = add_epsg_def (first, last, 4728, "epsg", 4728,
		      "Pico de las Nieves 1984");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-307,-92,127,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Pico de las Nieves 1984\",DATUM[\"Pico_de_las_N");
    add_srs_wkt (p, 1,
		 "ieves_1984\",SPHEROID[\"International 1924\",6378388,297");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-307,-92,127,0,0,");
    add_srs_wkt (p, 3,
		 "0,0],AUTHORITY[\"EPSG\",\"6728\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6, "SG\",\"4728\"]]");
    p = add_epsg_def (first, last, 4729, "epsg", 4729, "Pitcairn 1967");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=185,165,42,0,0,0,0 +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Pitcairn 1967\",DATUM[\"Pitcairn_1967\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 2,
		 "\"7022\"]],TOWGS84[185,165,42,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6729\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4729\"]]");
    p = add_epsg_def (first, last, 4730, "epsg", 4730, "Santo 1965");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=170,42,84,0,0,0,0 +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Santo 1965\",DATUM[\"Santo_1965\",SPHEROID[\"In");
    add_srs_wkt (p, 1,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 2,
		 "\"]],TOWGS84[170,42,84,0,0,0,0],AUTHORITY[\"EPSG\",\"673");
    add_srs_wkt (p, 3,
		 "0\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4730\"]]");
    p = add_epsg_def (first, last, 4731, "epsg", 4731, "Viti Levu 1916");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=51,391,-36,0,0,0,0 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Viti Levu 1916\",DATUM[\"Viti_Levu_1916\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[");
    add_srs_wkt (p, 2,
		 "\"EPSG\",\"7012\"]],TOWGS84[51,391,-36,0,0,0,0],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6731\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4731\"]");
    add_srs_wkt (p, 6, "]");
    p = add_epsg_def (first, last, 4732, "epsg", 4732, "Marshall Islands 1960");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378270 +b=6356794.343434343 +towgs84=1");
    add_proj4text (p, 1, "02,52,-38,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Marshall Islands 1960\",DATUM[\"Marshall_Island");
    add_srs_wkt (p, 1,
		 "s_1960\",SPHEROID[\"Hough 1960\",6378270,297,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7053\"]],TOWGS84[102,52,-38,0,0,0,0],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6732\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4732\"]]");
    add_srs_wkt (p, 6, "");
    p = add_epsg_def (first, last, 4733, "epsg", 4733, "Wake Island 1952");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=276,-57,149,0,0,0,0 +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Wake Island 1952\",DATUM[\"Wake_Island_1952\",S");
    add_srs_wkt (p, 1,
		 "PHEROID[\"International 1924\",6378388,297,AUTHORITY[\"E");
    add_srs_wkt (p, 2,
		 "PSG\",\"7022\"]],TOWGS84[276,-57,149,0,0,0,0],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6733\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4733\"]]");
    add_srs_wkt (p, 6, "");
    p = add_epsg_def (first, last, 4734, "epsg", 4734, "Tristan 1968");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-632,438,-609,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Tristan 1968\",DATUM[\"Tristan_1968\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7022\"]],TOWGS84[-632,438,-609,0,0,0,0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"6734\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 5, "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4734\"]]");
    p = add_epsg_def (first, last, 4735, "epsg", 4735, "Kusaie 1951");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=647,1777,-1124,0,0,0,");
    add_proj4text (p, 1, "0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Kusaie 1951\",DATUM[\"Kusaie_1951\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 2,
		 "22\"]],TOWGS84[647,1777,-1124,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6735\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4735\"]]");
    p = add_epsg_def (first, last, 4736, "epsg", 4736, "Deception Island");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=260,12,-147,0,0,0,0");
    add_proj4text (p, 1, " +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Deception Island\",DATUM[\"Deception_Island\",S");
    add_srs_wkt (p, 1,
		 "PHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHOR");
    add_srs_wkt (p, 2,
		 "ITY[\"EPSG\",\"7012\"]],TOWGS84[260,12,-147,0,0,0,0],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6736\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6, "736\"]]");
    p = add_epsg_def (first, last, 4737, "epsg", 4737, "Korea 2000");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Korea 2000\",DATUM[\"Geocentric_datum_of_Korea\"");
    add_srs_wkt (p, 1,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"6737\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4737\"]]");
    p = add_epsg_def (first, last, 4738, "epsg", 4738, "Hong Kong 1963");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378293.645208759 +b=6356617.987679838 ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Hong Kong 1963\",DATUM[\"Hong_Kong_1963\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"Clarke 1858\",6378293.645208759,294.2606763692569,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7007\"]],AUTHORITY[\"EPSG\",\"6738\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5, "9122\"]],AUTHORITY[\"EPSG\",\"4738\"]]");
    p = add_epsg_def (first, last, 4739, "epsg", 4739, "Hong Kong 1963(67)");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-156,-271,-189,0,0,0,");
    add_proj4text (p, 1, "0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Hong Kong 1963(67)\",DATUM[\"Hong_Kong_1963_67\"");
    add_srs_wkt (p, 1,
		 ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7022\"]],TOWGS84[-156,-271,-189,0,0,0,0],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6739\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4739");
    add_srs_wkt (p, 6, "\"]]");
    p = add_epsg_def (first, last, 4740, "epsg", 4740, "PZ-90");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378136 +b=6356751.361745712 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"PZ-90\",DATUM[\"Parametrop_Zemp_1990\",SPHEROID");
    add_srs_wkt (p, 1,
		 "[\"PZ-90\",6378136,298.257839303,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 2,
		 "54\"]],AUTHORITY[\"EPSG\",\"6740\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 3,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 4,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 5, "PSG\",\"4740\"]]");
    p = add_epsg_def (first, last, 4741, "epsg", 4741, "FD54");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"FD54\",DATUM[\"Faroe_Datum_1954\",SPHEROID[\"In");
    add_srs_wkt (p, 1,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 2,
		 "\"]],AUTHORITY[\"EPSG\",\"6741\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 3,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 4,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 5, "SG\",\"4741\"]]");
    p = add_epsg_def (first, last, 4742, "epsg", 4742, "GDM2000");
    add_proj4text (p, 0, "+proj=longlat +ellps=GRS80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"GDM2000\",DATUM[\"Geodetic_Datum_of_Malaysia_20");
    add_srs_wkt (p, 1,
		 "00\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6742\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5, "]],AUTHORITY[\"EPSG\",\"4742\"]]");
    p = add_epsg_def (first, last, 4743, "epsg", 4743, "Karbala 1979");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=clrk80 +towgs84=84.1,-320.1,218.7,0");
    add_proj4text (p, 1, ",0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Karbala 1979\",DATUM[\"Karbala_1979\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EP");
    add_srs_wkt (p, 2,
		 "SG\",\"7012\"]],TOWGS84[84.1,-320.1,218.7,0,0,0,0],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6743\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"474");
    add_srs_wkt (p, 6, "3\"]]");
    p = add_epsg_def (first, last, 4744, "epsg", 4744, "Nahrwan 1934");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Nahrwan 1934\",DATUM[\"Nahrwan_1934\",SPHEROID[");
    add_srs_wkt (p, 1,
		 "\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EP");
    add_srs_wkt (p, 2,
		 "SG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6744\"]],PRIMEM[\"G");
    add_srs_wkt (p, 3,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 4,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 5, "HORITY[\"EPSG\",\"4744\"]]");
    p = add_epsg_def (first, last, 4745, "epsg", 4745, "RD/83");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"RD/83\",DATUM[\"Rauenberg_Datum_83\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 2,
		 ",\"7004\"]],AUTHORITY[\"EPSG\",\"6745\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 3,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 4,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 5, "TY[\"EPSG\",\"4745\"]]");
    p = add_epsg_def (first, last, 4746, "epsg", 4746, "PD/83");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"PD/83\",DATUM[\"Potsdam_Datum_83\",SPHEROID[\"B");
    add_srs_wkt (p, 1,
		 "essel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 2,
		 "\"7004\"]],AUTHORITY[\"EPSG\",\"6746\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 3,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 4,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"4746\"]]");
    p = add_epsg_def (first, last, 4747, "epsg", 4747, "GR96");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"GR96\",DATUM[\"Greenland_1996\",SPHEROID[\"GRS ");
    add_srs_wkt (p, 1,
		 "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 2,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],");
    add_srs_wkt (p, 3,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 4,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 5, "22\"]],AUTHORITY[\"EPSG\",\"4747\"]]");
    p = add_epsg_def (first, last, 4748, "epsg", 4748, "Vanua Levu 1915");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378306.3696 +b=6356571.996 +towgs84=51");
    add_proj4text (p, 1, ",391,-36,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Vanua Levu 1915\",DATUM[\"Vanua_Levu_1915\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"Clarke 1880 (international foot)\",6378306.3696,");
    add_srs_wkt (p, 2,
		 "293.4663076556349,AUTHORITY[\"EPSG\",\"7055\"]],TOWGS84[");
    add_srs_wkt (p, 3,
		 "51,391,-36,0,0,0,0],AUTHORITY[\"EPSG\",\"6748\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6, ",AUTHORITY[\"EPSG\",\"4748\"]]");
    p = add_epsg_def (first, last, 4749, "epsg", 4749, "RGNC91-93");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"RGNC91-93\",DATUM[\"Reseau_Geodesique_de_Nouvel");
    add_srs_wkt (p, 1,
		 "le_Caledonie_91_93\",SPHEROID[\"GRS 1980\",6378137,298.2");
    add_srs_wkt (p, 2,
		 "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0");
    add_srs_wkt (p, 3,
		 ",0,0],AUTHORITY[\"EPSG\",\"6749\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6, "PSG\",\"4749\"]]");
    p = add_epsg_def (first, last, 4750, "epsg", 4750, "ST87 Ouvea");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=WGS84 +towgs84=-56.263,16.136,-22.8");
    add_proj4text (p, 1, "56,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"ST87 Ouvea\",DATUM[\"ST87_Ouvea\",SPHEROID[\"WG");
    add_srs_wkt (p, 1,
		 "S 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"");
    add_srs_wkt (p, 2,
		 "]],TOWGS84[-56.263,16.136,-22.856,0,0,0,0],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6750\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5, "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4750\"]]");
    p = add_epsg_def (first, last, 4751, "epsg", 4751, "Kertau (RSO)");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377295.664 +b=6356094.667915204 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Kertau (RSO)\",DATUM[\"Kertau_RSO\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "Everest 1830 (RSO 1969)\",6377295.664,300.8017,AUTHORITY");
    add_srs_wkt (p, 2,
		 "[\"EPSG\",\"7056\"]],AUTHORITY[\"EPSG\",\"6751\"]],PRIME");
    add_srs_wkt (p, 3,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 4,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 5, "],AUTHORITY[\"EPSG\",\"4751\"]]");
    p = add_epsg_def (first, last, 4752, "epsg", 4752, "Viti Levu 1912");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378306.3696 +b=6356571.996 +towgs84=51");
    add_proj4text (p, 1, ",391,-36,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Viti Levu 1912\",DATUM[\"Viti_Levu_1912\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"Clarke 1880 (international foot)\",6378306.3696,29");
    add_srs_wkt (p, 2,
		 "3.4663076556349,AUTHORITY[\"EPSG\",\"7055\"]],TOWGS84[51");
    add_srs_wkt (p, 3,
		 ",391,-36,0,0,0,0],AUTHORITY[\"EPSG\",\"6752\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6, "THORITY[\"EPSG\",\"4752\"]]");
    p = add_epsg_def (first, last, 4753, "epsg", 4753, "fk89");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"fk89\",DATUM[\"fk89\",SPHEROID[\"International ");
    add_srs_wkt (p, 1,
		 "1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"6753\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4753\"");
    add_srs_wkt (p, 5, "]]");
    p = add_epsg_def (first, last, 4754, "epsg", 4754, "LGD2006");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-208.406,-109.878,-2.");
    add_proj4text (p, 1, "5764,0,0,0,0 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"LGD2006\",DATUM[\"Libyan_Geodetic_Datum_2006\",");
    add_srs_wkt (p, 1,
		 "SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7022\"]],TOWGS84[-208.406,-109.878,-2.5764,0,0,");
    add_srs_wkt (p, 3,
		 "0,0],AUTHORITY[\"EPSG\",\"6754\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6, "SG\",\"4754\"]]");
    p = add_epsg_def (first, last, 4755, "epsg", 4755, "DGN95");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"DGN95\",DATUM[\"Datum_Geodesi_Nasional_1995\",S");
    add_srs_wkt (p, 1,
		 "PHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPS");
    add_srs_wkt (p, 2,
		 "G\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4755\"]]");
    p = add_epsg_def (first, last, 4756, "epsg", 4756, "VN-2000");
    add_proj4text (p, 0, "+proj=longlat +ellps=WGS84 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"VN-2000\",DATUM[\"Vietnam_2000\",SPHEROID[\"WGS");
    add_srs_wkt (p, 1,
		 " 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]");
    add_srs_wkt (p, 2,
		 "],AUTHORITY[\"EPSG\",\"6756\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 4,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"4756\"]]");
    p = add_epsg_def (first, last, 4757, "epsg", 4757, "SVY21");
    add_proj4text (p, 0, "+proj=longlat +ellps=WGS84 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"SVY21\",DATUM[\"SVY21\",SPHEROID[\"WGS 84\",637");
    add_srs_wkt (p, 1,
		 "8137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"6757\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4757\"");
    add_srs_wkt (p, 5, "]]");
    p = add_epsg_def (first, last, 4758, "epsg", 4758, "JAD2001");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"JAD2001\",DATUM[\"Jamaica_2001\",SPHEROID[\"WGS");
    add_srs_wkt (p, 1,
		 " 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]");
    add_srs_wkt (p, 2,
		 "],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6758\"]],P");
    add_srs_wkt (p, 3,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 4,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 5, "2\"]],AUTHORITY[\"EPSG\",\"4758\"]]");
    p = add_epsg_def (first, last, 4759, "epsg", 4759, "NAD83(NSRS2007)");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatia");
    add_srs_wkt (p, 1,
		 "l_Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137,");
    add_srs_wkt (p, 2,
		 "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6, "Y[\"EPSG\",\"4759\"]]");
    p = add_epsg_def (first, last, 4760, "epsg", 4760, "WGS 66");
    add_proj4text (p, 0, "+proj=longlat +ellps=WGS66 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"WGS 66\",DATUM[\"World_Geodetic_System_1966\",S");
    add_srs_wkt (p, 1,
		 "PHEROID[\"NWL 9D\",6378145,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 2,
		 "025\"]],AUTHORITY[\"EPSG\",\"6760\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 3,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 4,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"4760\"]]");
    p = add_epsg_def (first, last, 4761, "epsg", 4761, "HTRS96");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"HTRS96\",DATUM[\"Croatian_Terrestrial_Reference");
    add_srs_wkt (p, 1,
		 "_System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU");
    add_srs_wkt (p, 2,
		 "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6761\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"476");
    add_srs_wkt (p, 6, "1\"]]");
    p = add_epsg_def (first, last, 4762, "epsg", 4762, "BDA2000");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"BDA2000\",DATUM[\"Bermuda_2000\",SPHEROID[\"WGS");
    add_srs_wkt (p, 1,
		 " 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]");
    add_srs_wkt (p, 2,
		 "],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6762\"]],P");
    add_srs_wkt (p, 3,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 4,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 5, "2\"]],AUTHORITY[\"EPSG\",\"4762\"]]");
    p = add_epsg_def (first, last, 4763, "epsg", 4763, "Pitcairn 2006");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Pitcairn 2006\",DATUM[\"Pitcairn_2006\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 2,
		 "7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"676");
    add_srs_wkt (p, 3,
		 "3\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4763\"]]");
    p = add_epsg_def (first, last, 4764, "epsg", 4764, "RSRGD2000");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"RSRGD2000\",DATUM[\"Ross_Sea_Region_Geodetic_Da");
    add_srs_wkt (p, 1,
		 "tum_2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6764\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"47");
    add_srs_wkt (p, 6, "64\"]]");
    p = add_epsg_def (first, last, 4765, "epsg", 4765, "Slovenia 1996");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Slovenia 1996\",DATUM[\"Slovenia_Geodetic_Datum");
    add_srs_wkt (p, 1,
		 "_1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH");
    add_srs_wkt (p, 2,
		 "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6765\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4765\"");
    add_srs_wkt (p, 6, "]]");
    p = add_epsg_def (first, last, 4801, "epsg", 4801, "Bern 1898 (Bern)");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +pm=bern +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Bern 1898 (Bern)\",DATUM[\"CH1903_Bern\",SPHERO");
    add_srs_wkt (p, 1,
		 "ID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"E");
    add_srs_wkt (p, 2,
		 "PSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6801\"]],PRIMEM[\"");
    add_srs_wkt (p, 3,
		 "Bern\",7.439583333333333,AUTHORITY[\"EPSG\",\"8907\"]],U");
    add_srs_wkt (p, 4,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5, "9122\"]],AUTHORITY[\"EPSG\",\"4801\"]]");
    p = add_epsg_def (first, last, 4802, "epsg", 4802, "Bogota 1975 (Bogota)");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +pm=bogota +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Bogota 1975 (Bogota)\",DATUM[\"Bogota_1975_Bogo");
    add_srs_wkt (p, 1,
		 "ta\",SPHEROID[\"International 1924\",6378388,297,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6802\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Bogota\",-74.08091666666667,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 4,
		 "904\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4802\"]]");
    p = add_epsg_def (first, last, 4803, "epsg", 4803, "Lisbon (Lisbon)");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +pm=lisbon +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Lisbon (Lisbon)\",DATUM[\"Lisbon_1937_Lisbon\",");
    add_srs_wkt (p, 1,
		 "SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6803\"]],PRIMEM[\"");
    add_srs_wkt (p, 3,
		 "Lisbon\",-9.131906111111112,AUTHORITY[\"EPSG\",\"8902\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9122\"]],AUTHORITY[\"EPSG\",\"4803\"]]");
    p = add_epsg_def (first, last, 4804, "epsg", 4804, "Makassar (Jakarta)");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=bessel +towgs84=-587.8,519.75,145.7");
    add_proj4text (p, 1, "6,0,0,0,0 +pm=jakarta +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Makassar (Jakarta)\",DATUM[\"Makassar_Jakarta\"");
    add_srs_wkt (p, 1,
		 ",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHOR");
    add_srs_wkt (p, 2,
		 "ITY[\"EPSG\",\"7004\"]],TOWGS84[-587.8,519.75,145.76,0,0");
    add_srs_wkt (p, 3,
		 ",0,0],AUTHORITY[\"EPSG\",\"6804\"]],PRIMEM[\"Jakarta\",1");
    add_srs_wkt (p, 4,
		 "06.8077194444444,AUTHORITY[\"EPSG\",\"8908\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6, ",AUTHORITY[\"EPSG\",\"4804\"]]");
    p = add_epsg_def (first, last, 4805, "epsg", 4805, "MGI (Ferro)");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +pm=ferro +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"MGI (Ferro)\",DATUM[\"Militar_Geographische_Ins");
    add_srs_wkt (p, 1,
		 "titut_Ferro\",SPHEROID[\"Bessel 1841\",6377397.155,299.1");
    add_srs_wkt (p, 2,
		 "528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"6805\"]],PRIMEM[\"Ferro\",-17.66666666666667,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8909\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4805");
    add_srs_wkt (p, 6, "\"]]");
    p = add_epsg_def (first, last, 4806, "epsg", 4806, "Monte Mario (Rome)");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +pm=rome +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Monte Mario (Rome)\",DATUM[\"Monte_Mario_Rome\"");
    add_srs_wkt (p, 1,
		 ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"");
    add_srs_wkt (p, 2,
		 "EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6806\"]],PRIMEM[\"");
    add_srs_wkt (p, 3,
		 "Rome\",12.45233333333333,AUTHORITY[\"EPSG\",\"8906\"]],U");
    add_srs_wkt (p, 4,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5, "9122\"]],AUTHORITY[\"EPSG\",\"4806\"]]");
    p = add_epsg_def (first, last, 4807, "epsg", 4807, "NTF (Paris)");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378249.2 +b=6356515 +towgs84=-168,-60,");
    add_proj4text (p, 1, "320,0,0,0,0 +pm=paris +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"NTF (Paris)\",DATUM[\"Nouvelle_Triangulation_Fr");
    add_srs_wkt (p, 1,
		 "ancaise_Paris\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2");
    add_srs_wkt (p, 2,
		 ",293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84");
    add_srs_wkt (p, 3,
		 "[-168,-60,320,0,0,0,0],AUTHORITY[\"EPSG\",\"6807\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Paris\",2.33722917,AUTHORITY[\"EPSG\",\"8903\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"grad\",0.01570796326794897,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6, "05\"]],AUTHORITY[\"EPSG\",\"4807\"]]");
    p = add_epsg_def (first, last, 4808, "epsg", 4808, "Padang (Jakarta)");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +pm=jakarta +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Padang (Jakarta)\",DATUM[\"Padang_1884_Jakarta\"");
    add_srs_wkt (p, 1,
		 ",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHOR");
    add_srs_wkt (p, 2,
		 "ITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6808\"]],PR");
    add_srs_wkt (p, 3,
		 "IMEM[\"Jakarta\",106.8077194444444,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8908\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4808\"]]");
    p = add_epsg_def (first, last, 4809, "epsg", 4809, "Belge 1950 (Brussels)");
    add_proj4text (p, 0, "+proj=longlat +ellps=intl +pm=brussels +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Belge 1950 (Brussels)\",DATUM[\"Reseau_National");
    add_srs_wkt (p, 1,
		 "_Belge_1950_Brussels\",SPHEROID[\"International 1924\",6");
    add_srs_wkt (p, 2,
		 "378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"6809\"]],PRIMEM[\"Brussels\",4.367975,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8910\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4809\"]");
    add_srs_wkt (p, 6, "]");
    p = add_epsg_def (first, last, 4810, "epsg", 4810, "Tananarive (Paris)");
    add_proj4text (p, 0,
		   "+proj=longlat +ellps=intl +towgs84=-189,-242,-91,0,0,0,0");
    add_proj4text (p, 1, " +pm=paris +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Tananarive (Paris)\",DATUM[\"Tananarive_1925_Pa");
    add_srs_wkt (p, 1,
		 "ris\",SPHEROID[\"International 1924\",6378388,297,AUTHOR");
    add_srs_wkt (p, 2,
		 "ITY[\"EPSG\",\"7022\"]],TOWGS84[-189,-242,-91,0,0,0,0],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6810\"]],PRIMEM[\"Paris\",2.33722917");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.015707963");
    add_srs_wkt (p, 5,
		 "26794897,AUTHORITY[\"EPSG\",\"9105\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6, ",\"4810\"]]");
    p = add_epsg_def (first, last, 4811, "epsg", 4811, "Voirol 1875 (Paris)");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378249.2 +b=6356515 +towgs84=-73,-247,");
    add_proj4text (p, 1, "227,0,0,0,0 +pm=paris +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Voirol 1875 (Paris)\",DATUM[\"Voirol_1875_Paris");
    add_srs_wkt (p, 1,
		 "\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.46602129");
    add_srs_wkt (p, 2,
		 "36265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-73,-247,227");
    add_srs_wkt (p, 3,
		 ",0,0,0,0],AUTHORITY[\"EPSG\",\"6811\"]],PRIMEM[\"Paris\"");
    add_srs_wkt (p, 4,
		 ",2.33722917,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",");
    add_srs_wkt (p, 5,
		 "0.01570796326794897,AUTHORITY[\"EPSG\",\"9105\"]],AUTHOR");
    add_srs_wkt (p, 6, "ITY[\"EPSG\",\"4811\"]]");
    p = add_epsg_def (first, last, 4813, "epsg", 4813, "Batavia (Jakarta)");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +pm=jakarta +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Batavia (Jakarta)\",DATUM[\"Batavia_Jakarta\",S");
    add_srs_wkt (p, 1,
		 "PHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6813\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Jakarta\",106.8077194444444,AUTHORITY[\"EPSG\",\"89");
    add_srs_wkt (p, 4,
		 "08\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"E");
    add_srs_wkt (p, 5, "PSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4813\"]]");
    p = add_epsg_def (first, last, 4814, "epsg", 4814, "RT38 (Stockholm)");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +pm=stockholm +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"RT38 (Stockholm)\",DATUM[\"Stockholm_1938_Stock");
    add_srs_wkt (p, 1,
		 "holm\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6814\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Stockholm\",18.05827777777778,AUTHORITY[\"EP");
    add_srs_wkt (p, 4,
		 "SG\",\"8911\"]],UNIT[\"degree\",0.01745329251994328,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4814\"]]");
    p = add_epsg_def (first, last, 4815, "epsg", 4815, "Greek (Athens)");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +pm=athens +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Greek (Athens)\",DATUM[\"Greek_Athens\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EP");
    add_srs_wkt (p, 2,
		 "SG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6815\"]],PRIMEM[\"A");
    add_srs_wkt (p, 3,
		 "thens\",23.7163375,AUTHORITY[\"EPSG\",\"8912\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5, "]],AUTHORITY[\"EPSG\",\"4815\"]]");
    p = add_epsg_def (first, last, 4816, "epsg", 4816, "Carthage (Paris)");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378249.2 +b=6356515 +pm=paris +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Carthage (Paris)\",DATUM[\"Carthage_Paris\",SPH");
    add_srs_wkt (p, 1,
		 "EROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936265,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6816\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Paris\",2.33722917,AUTHORITY[\"EPSG\",\"8903");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"grad\",0.01570796326794897,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5, ",\"9105\"]],AUTHORITY[\"EPSG\",\"4816\"]]");
    p = add_epsg_def (first, last, 4817, "epsg", 4817, "NGO 1948 (Oslo)");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6377492.018 +b=6356173.508712696 +towgs");
    add_proj4text (p, 1,
		   "84=278.3,93,474.5,7.889,0.05,-6.61,6.21 +pm=oslo +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"NGO 1948 (Oslo)\",DATUM[\"NGO_1948_Oslo\",SPHER");
    add_srs_wkt (p, 1,
		 "OID[\"Bessel Modified\",6377492.018,299.1528128,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7005\"]],TOWGS84[278.3,93,474.5,7.889,0.05,");
    add_srs_wkt (p, 3,
		 "-6.61,6.21],AUTHORITY[\"EPSG\",\"6817\"]],PRIMEM[\"Oslo\"");
    add_srs_wkt (p, 4,
		 ",10.72291666666667,AUTHORITY[\"EPSG\",\"8913\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6, "]],AUTHORITY[\"EPSG\",\"4817\"]]");
    p = add_epsg_def (first, last, 4818, "epsg", 4818, "S-JTSK (Ferro)");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +pm=ferro +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"S-JTSK (Ferro)\",DATUM[\"S_JTSK_Ferro\",SPHEROI");
    add_srs_wkt (p, 1,
		 "D[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EP");
    add_srs_wkt (p, 2,
		 "SG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6818\"]],PRIMEM[\"F");
    add_srs_wkt (p, 3,
		 "erro\",-17.66666666666667,AUTHORITY[\"EPSG\",\"8909\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5, "9122\"]],AUTHORITY[\"EPSG\",\"4818\"]]");
    p = add_epsg_def (first, last, 4819, "epsg", 4819,
		      "Nord Sahara 1959 (Paris)");
    add_proj4text (p, 0, "+proj=longlat +ellps=clrk80 +pm=paris +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Nord Sahara 1959 (Paris)\",DATUM[\"Nord_Sahara_");
    add_srs_wkt (p, 1,
		 "1959_Paris\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,");
    add_srs_wkt (p, 2,
		 "293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6819\"]],PRIMEM[\"Paris\",2.33722917,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8903\"]],UNIT[\"grad\",0.01570796326794897,AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"9105\"]],AUTHORITY[\"EPSG\",\"4819\"]]");
    p = add_epsg_def (first, last, 4820, "epsg", 4820, "Segara (Jakarta)");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +pm=jakarta +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Segara (Jakarta)\",DATUM[\"Gunung_Segara_Jakart");
    add_srs_wkt (p, 1,
		 "a\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6820\"]]");
    add_srs_wkt (p, 3,
		 ",PRIMEM[\"Jakarta\",106.8077194444444,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8908\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4820\"]]");
    p = add_epsg_def (first, last, 4821, "epsg", 4821, "Voirol 1879 (Paris)");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378249.2 +b=6356515 +pm=paris +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Voirol 1879 (Paris)\",DATUM[\"Voirol_1879_Paris");
    add_srs_wkt (p, 1,
		 "\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.46602129");
    add_srs_wkt (p, 2,
		 "36265,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "6821\"]],PRIMEM[\"Paris\",2.33722917,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"8903\"]],UNIT[\"grad\",0.01570796326794897,AUTHORITY[\"");
    add_srs_wkt (p, 5, "EPSG\",\"9105\"]],AUTHORITY[\"EPSG\",\"4821\"]]");
    p = add_epsg_def (first, last, 4901, "epsg", 4901, "unnamed ellipse");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6376523 +b=6355862.933255573 +pm=2.3372");
    add_proj4text (p, 1, "291666985 +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"unnamed ellipse\",DATUM[\"unknown\",SPHEROID[\"");
    add_srs_wkt (p, 1,
		 "unnamed\",6376523,308.6399999999991]],PRIMEM[\"unnamed\"");
    add_srs_wkt (p, 2,
		 ",2.3372291666985],UNIT[\"degree\",0.0174532925199433],AU");
    add_srs_wkt (p, 3, "THORITY[\"EPSG\",\"4901\"]]");
    p = add_epsg_def (first, last, 4902, "epsg", 4902, "NDG (Paris)");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6376523 +b=6355862.933255573 +pm=paris ");
    add_proj4text (p, 1, "+no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"NDG (Paris)\",DATUM[\"Nord_de_Guerre_Paris\",SP");
    add_srs_wkt (p, 1,
		 "HEROID[\"Plessis 1817\",6376523,308.64,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 2,
		 ",\"7027\"]],AUTHORITY[\"EPSG\",\"6902\"]],PRIMEM[\"Paris");
    add_srs_wkt (p, 3,
		 "\",2.33722917,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\"");
    add_srs_wkt (p, 4,
		 ",0.01570796326794897,AUTHORITY[\"EPSG\",\"9105\"]],AUTHO");
    add_srs_wkt (p, 5, "RITY[\"EPSG\",\"4902\"]]");
    p = add_epsg_def (first, last, 4903, "epsg", 4903, "Madrid 1870 (Madrid)");
    add_proj4text (p, 0,
		   "+proj=longlat +a=6378298.3 +b=6356657.142669561 +pm=madr");
    add_proj4text (p, 1, "id +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Madrid 1870 (Madrid)\",DATUM[\"Madrid_1870_Madr");
    add_srs_wkt (p, 1,
		 "id\",SPHEROID[\"Struve 1860\",6378298.3,294.73,AUTHORITY");
    add_srs_wkt (p, 2,
		 "[\"EPSG\",\"7028\"]],AUTHORITY[\"EPSG\",\"6903\"]],PRIME");
    add_srs_wkt (p, 3,
		 "M[\"Madrid\",-3.687938888888889,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 4,
		 "5\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 5, "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4903\"]]");
    p = add_epsg_def (first, last, 4904, "epsg", 4904, "Lisbon 1890 (Lisbon)");
    add_proj4text (p, 0, "+proj=longlat +ellps=bessel +pm=lisbon +no_defs");
    add_srs_wkt (p, 0,
		 "GEOGCS[\"Lisbon 1890 (Lisbon)\",DATUM[\"Lisbon_1890_Lisb");
    add_srs_wkt (p, 1,
		 "on\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AU");
    add_srs_wkt (p, 2,
		 "THORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6904\"]");
    add_srs_wkt (p, 3,
		 "],PRIMEM[\"Lisbon\",-9.131906111111112,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8902\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5, "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4904\"]]");
    p = add_epsg_def (first, last, 20004, "epsg", 20004,
		      "Pulkovo 1995 / Gauss-Kruger zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=1 +x_0=4500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 4\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",21],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",4500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"20004\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20005, "epsg", 20005,
		      "Pulkovo 1995 / Gauss-Kruger zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=27 +k=1 +x_0=5500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 5\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",27],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",5500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"20005\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20006, "epsg", 20006,
		      "Pulkovo 1995 / Gauss-Kruger zone 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=33 +k=1 +x_0=6500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 6\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",33],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",6500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"20006\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20007, "epsg", 20007,
		      "Pulkovo 1995 / Gauss-Kruger zone 7");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=39 +k=1 +x_0=7500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 7\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",39],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",7500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"20007\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20008, "epsg", 20008,
		      "Pulkovo 1995 / Gauss-Kruger zone 8");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=45 +k=1 +x_0=8500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 8\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",45],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",8500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"20008\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20009, "epsg", 20009,
		      "Pulkovo 1995 / Gauss-Kruger zone 9");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=51 +k=1 +x_0=9500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 9\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",51],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",9500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"20009\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20010, "epsg", 20010,
		      "Pulkovo 1995 / Gauss-Kruger zone 10");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=57 +k=1 +x_0=10500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 10\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",57],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",1],PARAMETER[\"false_easting\",10500000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2001");
    add_srs_wkt (p, 12, "0\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20011, "epsg", 20011,
		      "Pulkovo 1995 / Gauss-Kruger zone 11");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=63 +k=1 +x_0=11500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 11\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",63],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",1],PARAMETER[\"false_easting\",11500000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2001");
    add_srs_wkt (p, 12, "1\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20012, "epsg", 20012,
		      "Pulkovo 1995 / Gauss-Kruger zone 12");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=69 +k=1 +x_0=12500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 12\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",69],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",1],PARAMETER[\"false_easting\",12500000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2001");
    add_srs_wkt (p, 12, "2\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20013, "epsg", 20013,
		      "Pulkovo 1995 / Gauss-Kruger zone 13");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=75 +k=1 +x_0=13500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 13\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",75],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",1],PARAMETER[\"false_easting\",13500000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2001");
    add_srs_wkt (p, 12, "3\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20014, "epsg", 20014,
		      "Pulkovo 1995 / Gauss-Kruger zone 14");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=81 +k=1 +x_0=14500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 14\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",81],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",1],PARAMETER[\"false_easting\",14500000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2001");
    add_srs_wkt (p, 12, "4\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20015, "epsg", 20015,
		      "Pulkovo 1995 / Gauss-Kruger zone 15");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=87 +k=1 +x_0=15500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 15\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",87],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",1],PARAMETER[\"false_easting\",15500000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2001");
    add_srs_wkt (p, 12, "5\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20016, "epsg", 20016,
		      "Pulkovo 1995 / Gauss-Kruger zone 16");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=93 +k=1 +x_0=16500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 16\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",93],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",1],PARAMETER[\"false_easting\",16500000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2001");
    add_srs_wkt (p, 12, "6\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20017, "epsg", 20017,
		      "Pulkovo 1995 / Gauss-Kruger zone 17");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=99 +k=1 +x_0=17500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 17\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",99],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",1],PARAMETER[\"false_easting\",17500000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2001");
    add_srs_wkt (p, 12, "7\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20018, "epsg", 20018,
		      "Pulkovo 1995 / Gauss-Kruger zone 18");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=18500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 18\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",105],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",18500000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"200");
    add_srs_wkt (p, 12, "18\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20019, "epsg", 20019,
		      "Pulkovo 1995 / Gauss-Kruger zone 19");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=19500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 19\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",111],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",19500000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"200");
    add_srs_wkt (p, 12, "19\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20020, "epsg", 20020,
		      "Pulkovo 1995 / Gauss-Kruger zone 20");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=1 +x_0=20500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 20\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",117],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",20500000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"200");
    add_srs_wkt (p, 12, "20\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20021, "epsg", 20021,
		      "Pulkovo 1995 / Gauss-Kruger zone 21");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=1 +x_0=21500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 21\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",123],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",21500000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"200");
    add_srs_wkt (p, 12, "21\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20022, "epsg", 20022,
		      "Pulkovo 1995 / Gauss-Kruger zone 22");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=129 +k=1 +x_0=22500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 22\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",129],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",22500000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"200");
    add_srs_wkt (p, 12, "22\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20023, "epsg", 20023,
		      "Pulkovo 1995 / Gauss-Kruger zone 23");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=135 +k=1 +x_0=23500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 23\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",135],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",23500000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"200");
    add_srs_wkt (p, 12, "23\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20024, "epsg", 20024,
		      "Pulkovo 1995 / Gauss-Kruger zone 24");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=141 +k=1 +x_0=24500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 24\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",141],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",24500000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"200");
    add_srs_wkt (p, 12, "24\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20025, "epsg", 20025,
		      "Pulkovo 1995 / Gauss-Kruger zone 25");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=147 +k=1 +x_0=25500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 25\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",147],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",25500000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"200");
    add_srs_wkt (p, 12, "25\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20026, "epsg", 20026,
		      "Pulkovo 1995 / Gauss-Kruger zone 26");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=153 +k=1 +x_0=26500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 26\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",153],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",26500000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"200");
    add_srs_wkt (p, 12, "26\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20027, "epsg", 20027,
		      "Pulkovo 1995 / Gauss-Kruger zone 27");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=159 +k=1 +x_0=27500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 27\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",159],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",27500000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"200");
    add_srs_wkt (p, 12, "27\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20028, "epsg", 20028,
		      "Pulkovo 1995 / Gauss-Kruger zone 28");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=165 +k=1 +x_0=28500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 28\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",165],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",28500000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"200");
    add_srs_wkt (p, 12, "28\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20029, "epsg", 20029,
		      "Pulkovo 1995 / Gauss-Kruger zone 29");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=171 +k=1 +x_0=29500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 29\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",171],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",29500000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"200");
    add_srs_wkt (p, 12, "29\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20030, "epsg", 20030,
		      "Pulkovo 1995 / Gauss-Kruger zone 30");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=177 +k=1 +x_0=30500000 +y_0=");
    add_proj4text (p, 1,
		   "0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,");
    add_proj4text (p, 2, "-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 30\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",177],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",30500000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"200");
    add_srs_wkt (p, 12, "30\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_18 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 20031, "epsg", 20031,
		      "Pulkovo 1995 / Gauss-Kruger zone 31");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-177 +k=1 +x_0=31500000 +y_0");
    add_proj4text (p, 1,
		   "=0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16");
    add_proj4text (p, 2, ",-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 31\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",-177],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",31500000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"200");
    add_srs_wkt (p, 12, "31\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20032, "epsg", 20032,
		      "Pulkovo 1995 / Gauss-Kruger zone 32");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-171 +k=1 +x_0=32500000 +y_0");
    add_proj4text (p, 1,
		   "=0 +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16");
    add_proj4text (p, 2, ",-0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger zone 32\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4200\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",0],PARAMETER[\"central_meridian\",-171],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",1],PARAMETER[\"false_easting\",32500000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"200");
    add_srs_wkt (p, 12, "32\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20064, "epsg", 20064,
		      "Pulkovo 1995 / Gauss-Kruger 4N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 4N (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",21],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "20064\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20065, "epsg", 20065,
		      "Pulkovo 1995 / Gauss-Kruger 5N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=27 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 5N (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",27],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "20065\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20066, "epsg", 20066,
		      "Pulkovo 1995 / Gauss-Kruger 6N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=33 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 6N (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",33],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "20066\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20067, "epsg", 20067,
		      "Pulkovo 1995 / Gauss-Kruger 7N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=39 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 7N (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",39],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "20067\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20068, "epsg", 20068,
		      "Pulkovo 1995 / Gauss-Kruger 8N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=45 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 8N (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",45],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "20068\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20069, "epsg", 20069,
		      "Pulkovo 1995 / Gauss-Kruger 9N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=51 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 9N (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",51],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",5000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "20069\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20070, "epsg", 20070,
		      "Pulkovo 1995 / Gauss-Kruger 10N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=57 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 10N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",57],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"20070\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20071, "epsg", 20071,
		      "Pulkovo 1995 / Gauss-Kruger 11N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=63 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 11N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",63],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"20071\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20072, "epsg", 20072,
		      "Pulkovo 1995 / Gauss-Kruger 12N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=69 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 12N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",69],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"20072\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20073, "epsg", 20073,
		      "Pulkovo 1995 / Gauss-Kruger 13N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=75 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 13N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",75],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"20073\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20074, "epsg", 20074,
		      "Pulkovo 1995 / Gauss-Kruger 14N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=81 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 14N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",81],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"20074\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20075, "epsg", 20075,
		      "Pulkovo 1995 / Gauss-Kruger 15N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=87 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 15N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",87],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"20075\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20076, "epsg", 20076,
		      "Pulkovo 1995 / Gauss-Kruger 16N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=93 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 16N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",93],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"20076\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20077, "epsg", 20077,
		      "Pulkovo 1995 / Gauss-Kruger 17N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=99 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1,
		   "ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0.");
    add_proj4text (p, 2, "12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 17N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",99],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"20077\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20078, "epsg", 20078,
		      "Pulkovo 1995 / Gauss-Kruger 18N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 18N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",105],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"20078\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20079, "epsg", 20079,
		      "Pulkovo 1995 / Gauss-Kruger 19N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 19N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",111],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"20079\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20080, "epsg", 20080,
		      "Pulkovo 1995 / Gauss-Kruger 20N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 20N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",117],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"20080\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20081, "epsg", 20081,
		      "Pulkovo 1995 / Gauss-Kruger 21N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 21N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",123],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"20081\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20082, "epsg", 20082,
		      "Pulkovo 1995 / Gauss-Kruger 22N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=129 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 22N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",129],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"20082\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20083, "epsg", 20083,
		      "Pulkovo 1995 / Gauss-Kruger 23N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=135 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 23N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",135],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"20083\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20084, "epsg", 20084,
		      "Pulkovo 1995 / Gauss-Kruger 24N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=141 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 24N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",141],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"20084\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20085, "epsg", 20085,
		      "Pulkovo 1995 / Gauss-Kruger 25N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=147 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 25N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",147],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"20085\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20086, "epsg", 20086,
		      "Pulkovo 1995 / Gauss-Kruger 26N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=153 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 26N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",153],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"20086\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20087, "epsg", 20087,
		      "Pulkovo 1995 / Gauss-Kruger 27N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=159 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 27N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",159],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"20087\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20088, "epsg", 20088,
		      "Pulkovo 1995 / Gauss-Kruger 28N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=165 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 28N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",165],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"20088\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20089, "epsg", 20089,
		      "Pulkovo 1995 / Gauss-Kruger 29N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=171 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 29N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",171],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"20089\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20090, "epsg", 20090,
		      "Pulkovo 1995 / Gauss-Kruger 30N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=177 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-0");
    add_proj4text (p, 2, ".12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 30N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",177],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"20090\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20091, "epsg", 20091,
		      "Pulkovo 1995 / Gauss-Kruger 31N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-177 +k=1 +x_0=500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 31N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-177],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"20091\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20092, "epsg", 20092,
		      "Pulkovo 1995 / Gauss-Kruger 32N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-171 +k=1 +x_0=500000 +y_0=0");
    add_proj4text (p, 1,
		   " +ellps=krass +towgs84=24.82,-131.21,-82.66,-0,-0,0.16,-");
    add_proj4text (p, 2, "0.12 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1995 / Gauss-Kruger 32N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo_1995\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[24.82,-131.21,-82.66,-0,-0,0.16,-0.12],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4200\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-171],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"20092\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 20135, "epsg", 20135,
		      "Adindan / UTM zone 35N");
    add_proj4text (p, 0, "+proj=utm +zone=35 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Adindan / UTM zone 35N\",GEOGCS[\"Adindan\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Adindan\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.14");
    add_srs_wkt (p, 2,
		 "5,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"6201\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4201\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 7,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"central_meridian\",27],PARAMETER[\"scale_f");
    add_srs_wkt (p, 9,
		 "actor\",0.9996],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"20135\"");
    add_srs_wkt (p, 11, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20136, "epsg", 20136,
		      "Adindan / UTM zone 36N");
    add_proj4text (p, 0, "+proj=utm +zone=36 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Adindan / UTM zone 36N\",GEOGCS[\"Adindan\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Adindan\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.14");
    add_srs_wkt (p, 2,
		 "5,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"6201\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4201\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 7,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"central_meridian\",33],PARAMETER[\"scale_f");
    add_srs_wkt (p, 9,
		 "actor\",0.9996],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"20136\"");
    add_srs_wkt (p, 11, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20137, "epsg", 20137,
		      "Adindan / UTM zone 37N");
    add_proj4text (p, 0, "+proj=utm +zone=37 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Adindan / UTM zone 37N\",GEOGCS[\"Adindan\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Adindan\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.14");
    add_srs_wkt (p, 2,
		 "5,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"6201\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4201\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 7,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"central_meridian\",39],PARAMETER[\"scale_f");
    add_srs_wkt (p, 9,
		 "actor\",0.9996],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"20137\"");
    add_srs_wkt (p, 11, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20138, "epsg", 20138,
		      "Adindan / UTM zone 38N");
    add_proj4text (p, 0, "+proj=utm +zone=38 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Adindan / UTM zone 38N\",GEOGCS[\"Adindan\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Adindan\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.14");
    add_srs_wkt (p, 2,
		 "5,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"6201\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4201\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 7,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"central_meridian\",45],PARAMETER[\"scale_f");
    add_srs_wkt (p, 9,
		 "actor\",0.9996],PARAMETER[\"false_easting\",500000],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"20138\"");
    add_srs_wkt (p, 11, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20248, "epsg", 20248, "AGD66 / AMG zone 48");
    add_proj4text (p, 0,
		   "+proj=utm +zone=48 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD66 / AMG zone 48\",GEOGCS[\"AGD66\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1966\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4202\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "05],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20248\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20249, "epsg", 20249, "AGD66 / AMG zone 49");
    add_proj4text (p, 0,
		   "+proj=utm +zone=49 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD66 / AMG zone 49\",GEOGCS[\"AGD66\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1966\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4202\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "11],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20249\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20250, "epsg", 20250, "AGD66 / AMG zone 50");
    add_proj4text (p, 0,
		   "+proj=utm +zone=50 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD66 / AMG zone 50\",GEOGCS[\"AGD66\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1966\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4202\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "17],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20250\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20251, "epsg", 20251, "AGD66 / AMG zone 51");
    add_proj4text (p, 0,
		   "+proj=utm +zone=51 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD66 / AMG zone 51\",GEOGCS[\"AGD66\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1966\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4202\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "23],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20251\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20252, "epsg", 20252, "AGD66 / AMG zone 52");
    add_proj4text (p, 0,
		   "+proj=utm +zone=52 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD66 / AMG zone 52\",GEOGCS[\"AGD66\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1966\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4202\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "29],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20252\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20253, "epsg", 20253, "AGD66 / AMG zone 53");
    add_proj4text (p, 0,
		   "+proj=utm +zone=53 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD66 / AMG zone 53\",GEOGCS[\"AGD66\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1966\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4202\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "35],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20253\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20254, "epsg", 20254, "AGD66 / AMG zone 54");
    add_proj4text (p, 0,
		   "+proj=utm +zone=54 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD66 / AMG zone 54\",GEOGCS[\"AGD66\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1966\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4202\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "41],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20254\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20255, "epsg", 20255, "AGD66 / AMG zone 55");
    add_proj4text (p, 0,
		   "+proj=utm +zone=55 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD66 / AMG zone 55\",GEOGCS[\"AGD66\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1966\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4202\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "47],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20255\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20256, "epsg", 20256, "AGD66 / AMG zone 56");
    add_proj4text (p, 0,
		   "+proj=utm +zone=56 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD66 / AMG zone 56\",GEOGCS[\"AGD66\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1966\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4202\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "53],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20256\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20257, "epsg", 20257, "AGD66 / AMG zone 57");
    add_proj4text (p, 0,
		   "+proj=utm +zone=57 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD66 / AMG zone 57\",GEOGCS[\"AGD66\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1966\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4202\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "59],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20257\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20258, "epsg", 20258, "AGD66 / AMG zone 58");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD66 / AMG zone 58\",GEOGCS[\"AGD66\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1966\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4202\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "65],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20258\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20348, "epsg", 20348, "AGD84 / AMG zone 48");
    add_proj4text (p, 0,
		   "+proj=utm +zone=48 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD84 / AMG zone 48\",GEOGCS[\"AGD84\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1984\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4203\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "05],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20348\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20349, "epsg", 20349, "AGD84 / AMG zone 49");
    add_proj4text (p, 0,
		   "+proj=utm +zone=49 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD84 / AMG zone 49\",GEOGCS[\"AGD84\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1984\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4203\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "11],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20349\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20350, "epsg", 20350, "AGD84 / AMG zone 50");
    add_proj4text (p, 0,
		   "+proj=utm +zone=50 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD84 / AMG zone 50\",GEOGCS[\"AGD84\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1984\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4203\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "17],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20350\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20351, "epsg", 20351, "AGD84 / AMG zone 51");
    add_proj4text (p, 0,
		   "+proj=utm +zone=51 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD84 / AMG zone 51\",GEOGCS[\"AGD84\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1984\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4203\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "23],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20351\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20352, "epsg", 20352, "AGD84 / AMG zone 52");
    add_proj4text (p, 0,
		   "+proj=utm +zone=52 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD84 / AMG zone 52\",GEOGCS[\"AGD84\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1984\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4203\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "29],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20352\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20353, "epsg", 20353, "AGD84 / AMG zone 53");
    add_proj4text (p, 0,
		   "+proj=utm +zone=53 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD84 / AMG zone 53\",GEOGCS[\"AGD84\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1984\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4203\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "35],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20353\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20354, "epsg", 20354, "AGD84 / AMG zone 54");
    add_proj4text (p, 0,
		   "+proj=utm +zone=54 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD84 / AMG zone 54\",GEOGCS[\"AGD84\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1984\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4203\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "41],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20354\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20355, "epsg", 20355, "AGD84 / AMG zone 55");
    add_proj4text (p, 0,
		   "+proj=utm +zone=55 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD84 / AMG zone 55\",GEOGCS[\"AGD84\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1984\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4203\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "47],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20355\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20356, "epsg", 20356, "AGD84 / AMG zone 56");
    add_proj4text (p, 0,
		   "+proj=utm +zone=56 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD84 / AMG zone 56\",GEOGCS[\"AGD84\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1984\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4203\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "53],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20356\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20357, "epsg", 20357, "AGD84 / AMG zone 57");
    add_proj4text (p, 0,
		   "+proj=utm +zone=57 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD84 / AMG zone 57\",GEOGCS[\"AGD84\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1984\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4203\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "59],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20357\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20358, "epsg", 20358, "AGD84 / AMG zone 58");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"AGD84 / AMG zone 58\",GEOGCS[\"AGD84\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Australian_Geodetic_Datum_1984\",SPHEROID[\"Australian N");
    add_srs_wkt (p, 2,
		 "ational Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "003\"]],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4203\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "65],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",10000000");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"20358\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20436, "epsg", 20436,
		      "Ain el Abd / UTM zone 36N");
    add_proj4text (p, 0, "+proj=utm +zone=36 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Ain el Abd / UTM zone 36N\",GEOGCS[\"Ain el Abd");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"Ain_el_Abd_1970\",SPHEROID[\"International 19");
    add_srs_wkt (p, 2,
		 "24\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6204\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4204\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",33],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "20436\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 20437, "epsg", 20437,
		      "Ain el Abd / UTM zone 37N");
    add_proj4text (p, 0, "+proj=utm +zone=37 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Ain el Abd / UTM zone 37N\",GEOGCS[\"Ain el Abd");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"Ain_el_Abd_1970\",SPHEROID[\"International 19");
    add_srs_wkt (p, 2,
		 "24\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6204\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4204\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",39],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "20437\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 20438, "epsg", 20438,
		      "Ain el Abd / UTM zone 38N");
    add_proj4text (p, 0, "+proj=utm +zone=38 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Ain el Abd / UTM zone 38N\",GEOGCS[\"Ain el Abd");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"Ain_el_Abd_1970\",SPHEROID[\"International 19");
    add_srs_wkt (p, 2,
		 "24\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6204\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4204\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",45],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "20438\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 20439, "epsg", 20439,
		      "Ain el Abd / UTM zone 39N");
    add_proj4text (p, 0, "+proj=utm +zone=39 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Ain el Abd / UTM zone 39N\",GEOGCS[\"Ain el Abd");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"Ain_el_Abd_1970\",SPHEROID[\"International 19");
    add_srs_wkt (p, 2,
		 "24\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6204\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4204\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",51],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "20439\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 20440, "epsg", 20440,
		      "Ain el Abd / UTM zone 40N");
    add_proj4text (p, 0, "+proj=utm +zone=40 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Ain el Abd / UTM zone 40N\",GEOGCS[\"Ain el Abd");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"Ain_el_Abd_1970\",SPHEROID[\"International 19");
    add_srs_wkt (p, 2,
		 "24\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6204\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4204\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",57],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "20440\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 20499, "epsg", 20499,
		      "Ain el Abd / Bahrain Grid");
    add_proj4text (p, 0, "+proj=utm +zone=39 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Ain el Abd / Bahrain Grid\",GEOGCS[\"Ain el Abd");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"Ain_el_Abd_1970\",SPHEROID[\"International 19");
    add_srs_wkt (p, 2,
		 "24\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6204\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4204\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",51],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "20499\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 20538, "epsg", 20538,
		      "Afgooye / UTM zone 38N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=38 +ellps=krass +towgs84=-43,-163,45,0,0");
    add_proj4text (p, 1, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Afgooye / UTM zone 38N\",GEOGCS[\"Afgooye\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Afgooye\",SPHEROID[\"Krassowsky 1940\",6378245,298.");
    add_srs_wkt (p, 2,
		 "3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[-43,-163,45,0,0,");
    add_srs_wkt (p, 3,
		 "0,0],AUTHORITY[\"EPSG\",\"6205\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4205\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",45");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 10,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 11,
		 "TY[\"EPSG\",\"20538\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 12, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 20539, "epsg", 20539,
		      "Afgooye / UTM zone 39N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=39 +ellps=krass +towgs84=-43,-163,45,0,0");
    add_proj4text (p, 1, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Afgooye / UTM zone 39N\",GEOGCS[\"Afgooye\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Afgooye\",SPHEROID[\"Krassowsky 1940\",6378245,298.");
    add_srs_wkt (p, 2,
		 "3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[-43,-163,45,0,0,");
    add_srs_wkt (p, 3,
		 "0,0],AUTHORITY[\"EPSG\",\"6205\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4205\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",51");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 10,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 11,
		 "TY[\"EPSG\",\"20539\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 12, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 20790, "epsg", 20790,
		      "Lisbon (Lisbon) / Portuguese National Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=39.66666666666666 +lon_0=1 +k=1 +x_0=");
    add_proj4text (p, 1,
		   "200000 +y_0=300000 +ellps=intl +pm=lisbon +units=m +no_d");
    add_proj4text (p, 2, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Lisbon (Lisbon) / Portuguese National Grid\",GE");
    add_srs_wkt (p, 1,
		 "OGCS[\"Lisbon (Lisbon)\",DATUM[\"Lisbon_1937_Lisbon\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"International 1924\",6378388,297,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6803\"]],PRIMEM[\"L");
    add_srs_wkt (p, 4,
		 "isbon\",-9.131906111111112,AUTHORITY[\"EPSG\",\"8902\"]]");
    add_srs_wkt (p, 5,
		 ",UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"9122\"]],AUTHORITY[\"EPSG\",\"4803\"]],UNIT[\"metre\",");
    add_srs_wkt (p, 7,
		 "1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_");
    add_srs_wkt (p, 8,
		 "Mercator\"],PARAMETER[\"latitude_of_origin\",39.66666666");
    add_srs_wkt (p, 9,
		 "666666],PARAMETER[\"central_meridian\",1],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",200000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",300000],AUTHORITY[\"EPSG\",\"20");
    add_srs_wkt (p, 12, "790\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 20791, "epsg", 20791,
		      "Lisbon (Lisbon) / Portuguese Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=39.66666666666666 +lon_0=1 +k=1 +x_0=");
    add_proj4text (p, 1, "0 +y_0=0 +ellps=intl +pm=lisbon +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Lisbon (Lisbon) / Portuguese Grid\",GEOGCS[\"Li");
    add_srs_wkt (p, 1,
		 "sbon (Lisbon)\",DATUM[\"Lisbon_1937_Lisbon\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "22\"]],AUTHORITY[\"EPSG\",\"6803\"]],PRIMEM[\"Lisbon\",-");
    add_srs_wkt (p, 4,
		 "9.131906111111112,AUTHORITY[\"EPSG\",\"8902\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4803\"]],UNIT[\"metre\",1,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",39.66666666666666],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"central_meridian\",1],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",1],PARAMETER[\"false_easting\",0],PARAMETER[\"false_n");
    add_srs_wkt (p, 11,
		 "orthing\",0],AUTHORITY[\"EPSG\",\"20791\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 12, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 20822, "epsg", 20822,
		      "Aratu / UTM zone 22S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=22 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Aratu / UTM zone 22S\",GEOGCS[\"Aratu\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Aratu\",SPHEROID[\"International 1924\",6378388,297,AUTH");
    add_srs_wkt (p, 2,
		 "ORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6208\"]],");
    add_srs_wkt (p, 3,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 4,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 5,
		 "22\"]],AUTHORITY[\"EPSG\",\"4208\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 7,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "central_meridian\",-51],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 9,
		 "6],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_northing\",10000000],AUTHORITY[\"EPSG\",\"20822\"],AXIS");
    add_srs_wkt (p, 11, "[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20823, "epsg", 20823,
		      "Aratu / UTM zone 23S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=23 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Aratu / UTM zone 23S\",GEOGCS[\"Aratu\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Aratu\",SPHEROID[\"International 1924\",6378388,297,AUTH");
    add_srs_wkt (p, 2,
		 "ORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6208\"]],");
    add_srs_wkt (p, 3,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 4,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 5,
		 "22\"]],AUTHORITY[\"EPSG\",\"4208\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 7,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "central_meridian\",-45],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 9,
		 "6],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_northing\",10000000],AUTHORITY[\"EPSG\",\"20823\"],AXIS");
    add_srs_wkt (p, 11, "[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20824, "epsg", 20824,
		      "Aratu / UTM zone 24S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=24 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Aratu / UTM zone 24S\",GEOGCS[\"Aratu\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Aratu\",SPHEROID[\"International 1924\",6378388,297,AUTH");
    add_srs_wkt (p, 2,
		 "ORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6208\"]],");
    add_srs_wkt (p, 3,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 4,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 5,
		 "22\"]],AUTHORITY[\"EPSG\",\"4208\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 7,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "central_meridian\",-39],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 9,
		 "6],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_northing\",10000000],AUTHORITY[\"EPSG\",\"20824\"],AXIS");
    add_srs_wkt (p, 11, "[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 20934, "epsg", 20934,
		      "Arc 1950 / UTM zone 34S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=34 +south +a=6378249.145 +b=6356514.9663");
    add_proj4text (p, 1, "98753 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Arc 1950 / UTM zone 34S\",GEOGCS[\"Arc 1950\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Arc_1950\",SPHEROID[\"Clarke 1880 (Arc)\",6378249");
    add_srs_wkt (p, 2,
		 ".145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6209\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4209\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",21],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",10000000],AUTHORITY[\"EP");
    add_srs_wkt (p, 11,
		 "SG\",\"20934\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 20935, "epsg", 20935,
		      "Arc 1950 / UTM zone 35S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=35 +south +a=6378249.145 +b=6356514.9663");
    add_proj4text (p, 1, "98753 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Arc 1950 / UTM zone 35S\",GEOGCS[\"Arc 1950\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Arc_1950\",SPHEROID[\"Clarke 1880 (Arc)\",6378249");
    add_srs_wkt (p, 2,
		 ".145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6209\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4209\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",27],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",10000000],AUTHORITY[\"EP");
    add_srs_wkt (p, 11,
		 "SG\",\"20935\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 20936, "epsg", 20936,
		      "Arc 1950 / UTM zone 36S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=36 +south +a=6378249.145 +b=6356514.9663");
    add_proj4text (p, 1, "98753 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Arc 1950 / UTM zone 36S\",GEOGCS[\"Arc 1950\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Arc_1950\",SPHEROID[\"Clarke 1880 (Arc)\",6378249");
    add_srs_wkt (p, 2,
		 ".145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6209\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4209\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",33],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",10000000],AUTHORITY[\"EP");
    add_srs_wkt (p, 11,
		 "SG\",\"20936\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 21035, "epsg", 21035,
		      "Arc 1960 / UTM zone 35S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=35 +south +ellps=clrk80 +units=m +no_def");
    add_proj4text (p, 1, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Arc 1960 / UTM zone 35S\",GEOGCS[\"Arc 1960\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Arc_1960\",SPHEROID[\"Clarke 1880 (RGS)\",6378249");
    add_srs_wkt (p, 2,
		 ".145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6210\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4210\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",0],PARAMETER[\"central_meridian\",27],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",0.9996],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11,
		 "\"21035\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORT");
    add_srs_wkt (p, 12, "H]]");
    p = add_epsg_def (first, last, 21036, "epsg", 21036,
		      "Arc 1960 / UTM zone 36S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=36 +south +ellps=clrk80 +units=m +no_def");
    add_proj4text (p, 1, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Arc 1960 / UTM zone 36S\",GEOGCS[\"Arc 1960\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Arc_1960\",SPHEROID[\"Clarke 1880 (RGS)\",6378249");
    add_srs_wkt (p, 2,
		 ".145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6210\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4210\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",0],PARAMETER[\"central_meridian\",33],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",0.9996],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11,
		 "\"21036\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORT");
    add_srs_wkt (p, 12, "H]]");
    p = add_epsg_def (first, last, 21037, "epsg", 21037,
		      "Arc 1960 / UTM zone 37S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=37 +south +ellps=clrk80 +units=m +no_def");
    add_proj4text (p, 1, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Arc 1960 / UTM zone 37S\",GEOGCS[\"Arc 1960\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Arc_1960\",SPHEROID[\"Clarke 1880 (RGS)\",6378249");
    add_srs_wkt (p, 2,
		 ".145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6210\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4210\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",0],PARAMETER[\"central_meridian\",39],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",0.9996],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11,
		 "\"21037\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORT");
    add_srs_wkt (p, 12, "H]]");
    p = add_epsg_def (first, last, 21095, "epsg", 21095,
		      "Arc 1960 / UTM zone 35N");
    add_proj4text (p, 0, "+proj=utm +zone=35 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Arc 1960 / UTM zone 35N\",GEOGCS[\"Arc 1960\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Arc_1960\",SPHEROID[\"Clarke 1880 (RGS)\",6378249");
    add_srs_wkt (p, 2,
		 ".145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6210\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4210\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",0],PARAMETER[\"central_meridian\",27],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",0.9996],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"21095");
    add_srs_wkt (p, 11, "\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 21096, "epsg", 21096,
		      "Arc 1960 / UTM zone 36N");
    add_proj4text (p, 0, "+proj=utm +zone=36 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Arc 1960 / UTM zone 36N\",GEOGCS[\"Arc 1960\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Arc_1960\",SPHEROID[\"Clarke 1880 (RGS)\",6378249");
    add_srs_wkt (p, 2,
		 ".145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6210\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4210\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",0],PARAMETER[\"central_meridian\",33],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",0.9996],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"21096");
    add_srs_wkt (p, 11, "\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 21097, "epsg", 21097,
		      "Arc 1960 / UTM zone 37N");
    add_proj4text (p, 0, "+proj=utm +zone=37 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Arc 1960 / UTM zone 37N\",GEOGCS[\"Arc 1960\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Arc_1960\",SPHEROID[\"Clarke 1880 (RGS)\",6378249");
    add_srs_wkt (p, 2,
		 ".145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6210\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4210\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",0],PARAMETER[\"central_meridian\",39],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",0.9996],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"21097");
    add_srs_wkt (p, 11, "\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 21100, "epsg", 21100,
		      "Batavia (Jakarta) / NEIEZ (deprecated)");
    add_proj4text (p, 0,
		   "+proj=merc +lon_0=110 +k=0.997 +x_0=3900000 +y_0=900000 ");
    add_proj4text (p, 1, "+ellps=bessel +pm=jakarta +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Batavia (Jakarta) / NEIEZ (deprecated)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"Batavia (Jakarta)\",DATUM[\"Batavia_Jakarta\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6813\"]],PRIMEM[\"J");
    add_srs_wkt (p, 4,
		 "akarta\",106.8077194444444,AUTHORITY[\"EPSG\",\"8908\"]]");
    add_srs_wkt (p, 5,
		 ",UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"9122\"]],AUTHORITY[\"EPSG\",\"4813\"]],UNIT[\"metre\",");
    add_srs_wkt (p, 7,
		 "1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Mercator_1S");
    add_srs_wkt (p, 8,
		 "P\"],PARAMETER[\"central_meridian\",110],PARAMETER[\"sca");
    add_srs_wkt (p, 9,
		 "le_factor\",0.997],PARAMETER[\"false_easting\",3900000],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"false_northing\",900000],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"21100\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 21148, "epsg", 21148,
		      "Batavia / UTM zone 48S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=48 +south +ellps=bessel +units=m +no_def");
    add_proj4text (p, 1, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Batavia / UTM zone 48S\",GEOGCS[\"Batavia\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Batavia\",SPHEROID[\"Bessel 1841\",6377397.155,299.");
    add_srs_wkt (p, 2,
		 "1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6211\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4211\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",105],PARAMETER[\"scale_fac");
    add_srs_wkt (p, 9,
		 "tor\",0.9996],PARAMETER[\"false_easting\",500000],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"21");
    add_srs_wkt (p, 11,
		 "148\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 21149, "epsg", 21149,
		      "Batavia / UTM zone 49S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=49 +south +ellps=bessel +units=m +no_def");
    add_proj4text (p, 1, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Batavia / UTM zone 49S\",GEOGCS[\"Batavia\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Batavia\",SPHEROID[\"Bessel 1841\",6377397.155,299.");
    add_srs_wkt (p, 2,
		 "1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6211\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4211\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",111],PARAMETER[\"scale_fac");
    add_srs_wkt (p, 9,
		 "tor\",0.9996],PARAMETER[\"false_easting\",500000],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"21");
    add_srs_wkt (p, 11,
		 "149\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 21150, "epsg", 21150,
		      "Batavia / UTM zone 50S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=50 +south +ellps=bessel +units=m +no_def");
    add_proj4text (p, 1, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Batavia / UTM zone 50S\",GEOGCS[\"Batavia\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Batavia\",SPHEROID[\"Bessel 1841\",6377397.155,299.");
    add_srs_wkt (p, 2,
		 "1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6211\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4211\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",117],PARAMETER[\"scale_fac");
    add_srs_wkt (p, 9,
		 "tor\",0.9996],PARAMETER[\"false_easting\",500000],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"21");
    add_srs_wkt (p, 11,
		 "150\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 21291, "epsg", 21291,
		      "Barbados 1938 / British West Indies Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-62 +k=0.9995000000000001 +x");
    add_proj4text (p, 1,
		   "_0=400000 +y_0=0 +ellps=clrk80 +towgs84=31.95,300.99,419");
    add_proj4text (p, 2, ".19,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Barbados 1938 / British West Indies Grid\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"Barbados 1938\",DATUM[\"Barbados_1938\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7012\"]],TOWGS84[31.95,300.99,419.19,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6212\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"421");
    add_srs_wkt (p, 7,
		 "2\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",0],PARAMETER[\"central_meridian\",-62],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",0.9995],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 11,
		 "400000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12,
		 "\",\"21291\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 13, "ORTH]]");
    p = add_epsg_def (first, last, 21292, "epsg", 21292,
		      "Barbados 1938 / Barbados National Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=13.17638888888889 +lon_0=-59.55972222");
    add_proj4text (p, 1,
		   "222222 +k=0.9999986 +x_0=30000 +y_0=75000 +ellps=clrk80 ");
    add_proj4text (p, 2,
		   "+towgs84=31.95,300.99,419.19,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Barbados 1938 / Barbados National Grid\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"Barbados 1938\",DATUM[\"Barbados_1938\",SPHEROID[\"Cl");
    add_srs_wkt (p, 2,
		 "arke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7012\"]],TOWGS84[31.95,300.99,419.19,0,0,0,0],AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"6212\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 5,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4212\"");
    add_srs_wkt (p, 7,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 9,
		 "igin\",13.17638888888889],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 10,
		 ",-59.55972222222222],PARAMETER[\"scale_factor\",0.999998");
    add_srs_wkt (p, 11,
		 "6],PARAMETER[\"false_easting\",30000],PARAMETER[\"false_");
    add_srs_wkt (p, 12,
		 "northing\",75000],AUTHORITY[\"EPSG\",\"21292\"],AXIS[\"E");
    add_srs_wkt (p, 13, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 21413, "epsg", 21413,
		      "Beijing 1954 / Gauss-Kruger zone 13");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=75 +k=1 +x_0=13500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger zone 13\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",75],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",135");
    add_srs_wkt (p, 10,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"21413\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21414, "epsg", 21414,
		      "Beijing 1954 / Gauss-Kruger zone 14");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=81 +k=1 +x_0=14500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger zone 14\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",81],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",145");
    add_srs_wkt (p, 10,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"21414\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21415, "epsg", 21415,
		      "Beijing 1954 / Gauss-Kruger zone 15");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=87 +k=1 +x_0=15500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger zone 15\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",87],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",155");
    add_srs_wkt (p, 10,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"21415\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21416, "epsg", 21416,
		      "Beijing 1954 / Gauss-Kruger zone 16");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=93 +k=1 +x_0=16500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger zone 16\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",93],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",165");
    add_srs_wkt (p, 10,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"21416\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21417, "epsg", 21417,
		      "Beijing 1954 / Gauss-Kruger zone 17");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=99 +k=1 +x_0=17500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger zone 17\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",99],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",175");
    add_srs_wkt (p, 10,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"21417\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21418, "epsg", 21418,
		      "Beijing 1954 / Gauss-Kruger zone 18");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=18500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger zone 18\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",105],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",18");
    add_srs_wkt (p, 10,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 11, "\",\"21418\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21419, "epsg", 21419,
		      "Beijing 1954 / Gauss-Kruger zone 19");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=19500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger zone 19\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",111],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",19");
    add_srs_wkt (p, 10,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 11, "\",\"21419\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21420, "epsg", 21420,
		      "Beijing 1954 / Gauss-Kruger zone 20");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=1 +x_0=20500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger zone 20\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",117],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",20");
    add_srs_wkt (p, 10,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 11, "\",\"21420\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21421, "epsg", 21421,
		      "Beijing 1954 / Gauss-Kruger zone 21");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=1 +x_0=21500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger zone 21\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",123],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",21");
    add_srs_wkt (p, 10,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 11, "\",\"21421\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21422, "epsg", 21422,
		      "Beijing 1954 / Gauss-Kruger zone 22");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=129 +k=1 +x_0=22500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger zone 22\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",129],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",22");
    add_srs_wkt (p, 10,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 11, "\",\"21422\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21423, "epsg", 21423,
		      "Beijing 1954 / Gauss-Kruger zone 23");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=135 +k=1 +x_0=23500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger zone 23\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",135],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",23");
    add_srs_wkt (p, 10,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 11, "\",\"21423\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21453, "epsg", 21453,
		      "Beijing 1954 / Gauss-Kruger CM 75E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=75 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger CM 75E\",GEOGCS[\"B");
    add_srs_wkt (p, 1,
		 "eijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",75],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 10,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"21453\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21454, "epsg", 21454,
		      "Beijing 1954 / Gauss-Kruger CM 81E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=81 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger CM 81E\",GEOGCS[\"B");
    add_srs_wkt (p, 1,
		 "eijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",81],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 10,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"21454\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21455, "epsg", 21455,
		      "Beijing 1954 / Gauss-Kruger CM 87E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=87 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger CM 87E\",GEOGCS[\"B");
    add_srs_wkt (p, 1,
		 "eijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",87],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 10,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"21455\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21456, "epsg", 21456,
		      "Beijing 1954 / Gauss-Kruger CM 93E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=93 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger CM 93E\",GEOGCS[\"B");
    add_srs_wkt (p, 1,
		 "eijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",93],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 10,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"21456\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21457, "epsg", 21457,
		      "Beijing 1954 / Gauss-Kruger CM 99E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=99 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger CM 99E\",GEOGCS[\"B");
    add_srs_wkt (p, 1,
		 "eijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",99],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 10,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"21457\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21458, "epsg", 21458,
		      "Beijing 1954 / Gauss-Kruger CM 105E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger CM 105E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",105],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 10,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"21458\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21459, "epsg", 21459,
		      "Beijing 1954 / Gauss-Kruger CM 111E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger CM 111E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",111],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 10,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"21459\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21460, "epsg", 21460,
		      "Beijing 1954 / Gauss-Kruger CM 117E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger CM 117E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",117],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 10,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"21460\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21461, "epsg", 21461,
		      "Beijing 1954 / Gauss-Kruger CM 123E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger CM 123E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",123],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 10,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"21461\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21462, "epsg", 21462,
		      "Beijing 1954 / Gauss-Kruger CM 129E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=129 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger CM 129E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",129],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 10,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"21462\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21463, "epsg", 21463,
		      "Beijing 1954 / Gauss-Kruger CM 135E");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=135 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger CM 135E\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",135],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 10,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"21463\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21473, "epsg", 21473,
		      "Beijing 1954 / Gauss-Kruger 13N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=75 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger 13N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",7");
    add_srs_wkt (p, 9,
		 "5],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"21473\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21474, "epsg", 21474,
		      "Beijing 1954 / Gauss-Kruger 14N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=81 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger 14N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",8");
    add_srs_wkt (p, 9,
		 "1],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"21474\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21475, "epsg", 21475,
		      "Beijing 1954 / Gauss-Kruger 15N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=87 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger 15N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",8");
    add_srs_wkt (p, 9,
		 "7],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"21475\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21476, "epsg", 21476,
		      "Beijing 1954 / Gauss-Kruger 16N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=93 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger 16N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",9");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"21476\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21477, "epsg", 21477,
		      "Beijing 1954 / Gauss-Kruger 17N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=99 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger 17N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",9");
    add_srs_wkt (p, 9,
		 "9],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"21477\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21478, "epsg", 21478,
		      "Beijing 1954 / Gauss-Kruger 18N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger 18N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "05],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"21478\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21479, "epsg", 21479,
		      "Beijing 1954 / Gauss-Kruger 19N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger 19N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "11],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"21479\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21480, "epsg", 21480,
		      "Beijing 1954 / Gauss-Kruger 20N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger 20N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "17],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"21480\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21481, "epsg", 21481,
		      "Beijing 1954 / Gauss-Kruger 21N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger 21N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "23],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"21481\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21482, "epsg", 21482,
		      "Beijing 1954 / Gauss-Kruger 22N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=129 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger 22N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "29],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"21482\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21483, "epsg", 21483,
		      "Beijing 1954 / Gauss-Kruger 23N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=135 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Beijing 1954 / Gauss-Kruger 23N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Beijing 1954\",DATUM[\"Beijing_1954\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4214\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 9,
		 "35],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"21483\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 21500, "epsg", 21500,
		      "Belge 1950 (Brussels) / Belge Lambert 50");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49.83333333333334 +lat_2=51.16666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=90 +lon_0=0 +x_0=150000 +y_0=5400000 +ellps=i");
    add_proj4text (p, 2, "ntl +pm=brussels +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Belge 1950 (Brussels) / Belge Lambert 50\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"Belge 1950 (Brussels)\",DATUM[\"Reseau_National_Bel");
    add_srs_wkt (p, 2,
		 "ge_1950_Brussels\",SPHEROID[\"International 1924\",63783");
    add_srs_wkt (p, 3,
		 "88,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6809\"]],PRIMEM[\"Brussels\",4.367975,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 5,
		 "\",\"8910\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4809\"]],UN");
    add_srs_wkt (p, 7,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 9,
		 "rallel_1\",49.83333333333334],PARAMETER[\"standard_paral");
    add_srs_wkt (p, 10,
		 "lel_2\",51.16666666666666],PARAMETER[\"latitude_of_origi");
    add_srs_wkt (p, 11,
		 "n\",90],PARAMETER[\"central_meridian\",0],PARAMETER[\"fa");
    add_srs_wkt (p, 12,
		 "lse_easting\",150000],PARAMETER[\"false_northing\",54000");
    add_srs_wkt (p, 13,
		 "00],AUTHORITY[\"EPSG\",\"21500\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 14, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 21780, "epsg", 21780,
		      "Bern 1898 (Bern) / LV03C");
    add_proj4text (p, 0,
		   "+proj=somerc +lat_0=46.95240555555556 +lon_0=0 +k_0=1 +x");
    add_proj4text (p, 1,
		   "_0=0 +y_0=0 +ellps=bessel +pm=bern +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Bern 1898 (Bern) / LV03C\",GEOGCS[\"Bern 1898 (");
    add_srs_wkt (p, 1,
		 "Bern)\",DATUM[\"CH1903_Bern\",SPHEROID[\"Bessel 1841\",6");
    add_srs_wkt (p, 2,
		 "377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6801\"]],PRIMEM[\"Bern\",7.43958333333");
    add_srs_wkt (p, 4,
		 "3333,AUTHORITY[\"EPSG\",\"8907\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4801\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "9001\"]],PROJECTION[\"Hotine_Oblique_Mercator\"],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"latitude_of_center\",46.95240555555556],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "longitude_of_center\",0],PARAMETER[\"azimuth\",90],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"rectified_grid_angle\",90],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 11,
		 "or\",1],PARAMETER[\"false_easting\",0],PARAMETER[\"false");
    add_srs_wkt (p, 12,
		 "_northing\",0],AUTHORITY[\"EPSG\",\"21780\"],AXIS[\"Y\",");
    add_srs_wkt (p, 13, "EAST],AXIS[\"X\",NORTH]]");
    p = add_epsg_def (first, last, 21781, "epsg", 21781, "CH1903 / LV03");
    add_proj4text (p, 0,
		   "+proj=somerc +lat_0=46.95240555555556 +lon_0=7.439583333");
    add_proj4text (p, 1,
		   "333333 +k_0=1 +x_0=600000 +y_0=200000 +ellps=bessel +tow");
    add_proj4text (p, 2,
		   "gs84=674.374,15.056,405.346,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"CH1903 / LV03\",GEOGCS[\"CH1903\",DATUM[\"CH190");
    add_srs_wkt (p, 1,
		 "3\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7004\"]],TOWGS84[674.374,15.056,405.34");
    add_srs_wkt (p, 3,
		 "6,0,0,0,0],AUTHORITY[\"EPSG\",\"6149\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"4149\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"9001\"]],PROJECTION[\"Hotine_Oblique_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_center\",46.95240555555556],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"longitude_of_center\",7.439583333333333],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "azimuth\",90],PARAMETER[\"rectified_grid_angle\",90],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",6");
    add_srs_wkt (p, 12,
		 "00000],PARAMETER[\"false_northing\",200000],AUTHORITY[\"");
    add_srs_wkt (p, 13,
		 "EPSG\",\"21781\"],AXIS[\"Y\",EAST],AXIS[\"X\",NORTH]]");
    p = add_epsg_def (first, last, 21782, "epsg", 21782, "CH1903 / LV03C-G");
    add_proj4text (p, 0,
		   "+proj=somerc +lat_0=46.95240555555556 +lon_0=7.439583333");
    add_proj4text (p, 1,
		   "333333 +k_0=1 +x_0=0 +y_0=0 +ellps=bessel +towgs84=674.3");
    add_proj4text (p, 2, "74,15.056,405.346,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"CH1903 / LV03C-G\",GEOGCS[\"CH1903\",DATUM[\"CH");
    add_srs_wkt (p, 1,
		 "1903\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[674.374,15.056,405");
    add_srs_wkt (p, 3,
		 ".346,0,0,0,0],AUTHORITY[\"EPSG\",\"6149\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4149\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Hotine_Oblique_Mercator\"],");
    add_srs_wkt (p, 8,
		 "PARAMETER[\"latitude_of_center\",46.95240555555556],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"longitude_of_center\",7.439583333333333],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"azimuth\",90],PARAMETER[\"rectified_grid_angle\",90");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 12,
		 "g\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 13, ",\"21782\"],AXIS[\"Y\",EAST],AXIS[\"X\",NORTH]]");
    p = add_epsg_def (first, last, 21817, "epsg", 21817,
		      "Bogota 1975 / UTM zone 17N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=17 +ellps=intl +towgs84=307,304,-318,0,0");
    add_proj4text (p, 1, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Bogota 1975 / UTM zone 17N (deprecated)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"Bogota 1975\",DATUM[\"Bogota_1975\",SPHEROID[\"Inter");
    add_srs_wkt (p, 2,
		 "national 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]");
    add_srs_wkt (p, 3,
		 "],TOWGS84[307,304,-318,0,0,0,0],AUTHORITY[\"EPSG\",\"621");
    add_srs_wkt (p, 4,
		 "8\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4218\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-81],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"21817\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 21818, "epsg", 21818,
		      "Bogota 1975 / UTM zone 18N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +ellps=intl +towgs84=307,304,-318,0,0");
    add_proj4text (p, 1, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Bogota 1975 / UTM zone 18N\",GEOGCS[\"Bogota 19");
    add_srs_wkt (p, 1,
		 "75\",DATUM[\"Bogota_1975\",SPHEROID[\"International 1924");
    add_srs_wkt (p, 2,
		 "\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[307");
    add_srs_wkt (p, 3,
		 ",304,-318,0,0,0,0],AUTHORITY[\"EPSG\",\"6218\"]],PRIMEM[");
    add_srs_wkt (p, 4,
		 "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg");
    add_srs_wkt (p, 5,
		 "ree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"4218\"]],UNIT[\"metre\",1,AUTHORITY");
    add_srs_wkt (p, 7,
		 "[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"]");
    add_srs_wkt (p, 8,
		 ",PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"central");
    add_srs_wkt (p, 9,
		 "_meridian\",-75],PARAMETER[\"scale_factor\",0.9996],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_easting\",500000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 11,
		 "ng\",0],AUTHORITY[\"EPSG\",\"21818\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 21891, "epsg", 21891,
		      "Bogota 1975 / Colombia West zone (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=4.599047222222222 +lon_0=-77.08091666");
    add_proj4text (p, 1,
		   "666667 +k=1 +x_0=1000000 +y_0=1000000 +ellps=intl +towgs");
    add_proj4text (p, 2, "84=307,304,-318,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Bogota 1975 / Colombia West zone (deprecated)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"Bogota 1975\",DATUM[\"Bogota_1975\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "22\"]],TOWGS84[307,304,-318,0,0,0,0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6218\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4218\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",4.59");
    add_srs_wkt (p, 9,
		 "9047222222222],PARAMETER[\"central_meridian\",-77.080916");
    add_srs_wkt (p, 10,
		 "66666667],PARAMETER[\"scale_factor\",1],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",1000000],PARAMETER[\"false_northing\",100000");
    add_srs_wkt (p, 12,
		 "0],AUTHORITY[\"EPSG\",\"21891\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 21892, "epsg", 21892,
		      "Bogota 1975 / Colombia Bogota zone (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=4.599047222222222 +lon_0=-74.08091666");
    add_proj4text (p, 1,
		   "666667 +k=1 +x_0=1000000 +y_0=1000000 +ellps=intl +towgs");
    add_proj4text (p, 2, "84=307,304,-318,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Bogota 1975 / Colombia Bogota zone (deprecated)");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"Bogota 1975\",DATUM[\"Bogota_1975\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7022\"]],TOWGS84[307,304,-318,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6218\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4218\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",4.59");
    add_srs_wkt (p, 9,
		 "9047222222222],PARAMETER[\"central_meridian\",-74.080916");
    add_srs_wkt (p, 10,
		 "66666667],PARAMETER[\"scale_factor\",1],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",1000000],PARAMETER[\"false_northing\",100000");
    add_srs_wkt (p, 12,
		 "0],AUTHORITY[\"EPSG\",\"21892\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 21893, "epsg", 21893,
		      "Bogota 1975 / Colombia East Central zone (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=4.599047222222222 +lon_0=-71.08091666");
    add_proj4text (p, 1,
		   "666667 +k=1 +x_0=1000000 +y_0=1000000 +ellps=intl +towgs");
    add_proj4text (p, 2, "84=307,304,-318,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Bogota 1975 / Colombia East Central zone (depre");
    add_srs_wkt (p, 1,
		 "cated)\",GEOGCS[\"Bogota 1975\",DATUM[\"Bogota_1975\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"International 1924\",6378388,297,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7022\"]],TOWGS84[307,304,-318,0,0,0,0],AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"6218\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 5,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4218\"]]");
    add_srs_wkt (p, 7,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 9,
		 "in\",4.599047222222222],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 10,
		 "71.08091666666667],PARAMETER[\"scale_factor\",1],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_easting\",1000000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 12,
		 "\",1000000],AUTHORITY[\"EPSG\",\"21893\"],AXIS[\"X\",EAS");
    add_srs_wkt (p, 13, "T],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 21894, "epsg", 21894,
		      "Bogota 1975 / Colombia East (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=4.599047222222222 +lon_0=-68.08091666");
    add_proj4text (p, 1,
		   "666667 +k=1 +x_0=1000000 +y_0=1000000 +ellps=intl +towgs");
    add_proj4text (p, 2, "84=307,304,-318,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Bogota 1975 / Colombia East (deprecated)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"Bogota 1975\",DATUM[\"Bogota_1975\",SPHEROID[\"Inte");
    add_srs_wkt (p, 2,
		 "rnational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[307,304,-318,0,0,0,0],AUTHORITY[\"EPSG\",\"62");
    add_srs_wkt (p, 4,
		 "18\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4218\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",4.59904722");
    add_srs_wkt (p, 9,
		 "2222222],PARAMETER[\"central_meridian\",-68.080916666666");
    add_srs_wkt (p, 10,
		 "67],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 11,
		 "ing\",1000000],PARAMETER[\"false_northing\",1000000],AUT");
    add_srs_wkt (p, 12,
		 "HORITY[\"EPSG\",\"21894\"],AXIS[\"X\",EAST],AXIS[\"Y\",N");
    add_srs_wkt (p, 13, "ORTH]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_19 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 21896, "epsg", 21896,
		      "Bogota 1975 / Colombia West zone");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=4.599047222222222 +lon_0=-77.08091666");
    add_proj4text (p, 1,
		   "666667 +k=1 +x_0=1000000 +y_0=1000000 +ellps=intl +towgs");
    add_proj4text (p, 2, "84=307,304,-318,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Bogota 1975 / Colombia West zone\",GEOGCS[\"Bog");
    add_srs_wkt (p, 1,
		 "ota 1975\",DATUM[\"Bogota_1975\",SPHEROID[\"Internationa");
    add_srs_wkt (p, 2,
		 "l 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS");
    add_srs_wkt (p, 3,
		 "84[307,304,-318,0,0,0,0],AUTHORITY[\"EPSG\",\"6218\"]],P");
    add_srs_wkt (p, 4,
		 "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT");
    add_srs_wkt (p, 5,
		 "[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"912");
    add_srs_wkt (p, 6,
		 "2\"]],AUTHORITY[\"EPSG\",\"4218\"]],UNIT[\"metre\",1,AUT");
    add_srs_wkt (p, 7,
		 "HORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merca");
    add_srs_wkt (p, 8,
		 "tor\"],PARAMETER[\"latitude_of_origin\",4.59904722222222");
    add_srs_wkt (p, 9,
		 "2],PARAMETER[\"central_meridian\",-77.08091666666667],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 11,
		 "1000000],PARAMETER[\"false_northing\",1000000],AUTHORITY");
    add_srs_wkt (p, 12,
		 "[\"EPSG\",\"21896\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 21897, "epsg", 21897,
		      "Bogota 1975 / Colombia Bogota zone");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=4.599047222222222 +lon_0=-74.08091666");
    add_proj4text (p, 1,
		   "666667 +k=1 +x_0=1000000 +y_0=1000000 +ellps=intl +towgs");
    add_proj4text (p, 2, "84=307,304,-318,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Bogota 1975 / Colombia Bogota zone\",GEOGCS[\"B");
    add_srs_wkt (p, 1,
		 "ogota 1975\",DATUM[\"Bogota_1975\",SPHEROID[\"Internatio");
    add_srs_wkt (p, 2,
		 "nal 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOW");
    add_srs_wkt (p, 3,
		 "GS84[307,304,-318,0,0,0,0],AUTHORITY[\"EPSG\",\"6218\"]]");
    add_srs_wkt (p, 4,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 5,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 6,
		 "122\"]],AUTHORITY[\"EPSG\",\"4218\"]],UNIT[\"metre\",1,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",4.599047222222");
    add_srs_wkt (p, 9,
		 "222],PARAMETER[\"central_meridian\",-74.08091666666667],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",1000000],PARAMETER[\"false_northing\",1000000],AUTHORIT");
    add_srs_wkt (p, 12,
		 "Y[\"EPSG\",\"21897\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]");
    add_srs_wkt (p, 13, "]");
    p = add_epsg_def (first, last, 21898, "epsg", 21898,
		      "Bogota 1975 / Colombia East Central zone");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=4.599047222222222 +lon_0=-71.08091666");
    add_proj4text (p, 1,
		   "666667 +k=1 +x_0=1000000 +y_0=1000000 +ellps=intl +towgs");
    add_proj4text (p, 2, "84=307,304,-318,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Bogota 1975 / Colombia East Central zone\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"Bogota 1975\",DATUM[\"Bogota_1975\",SPHEROID[\"Inte");
    add_srs_wkt (p, 2,
		 "rnational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[307,304,-318,0,0,0,0],AUTHORITY[\"EPSG\",\"62");
    add_srs_wkt (p, 4,
		 "18\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4218\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",4.59904722");
    add_srs_wkt (p, 9,
		 "2222222],PARAMETER[\"central_meridian\",-71.080916666666");
    add_srs_wkt (p, 10,
		 "67],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_east");
    add_srs_wkt (p, 11,
		 "ing\",1000000],PARAMETER[\"false_northing\",1000000],AUT");
    add_srs_wkt (p, 12,
		 "HORITY[\"EPSG\",\"21898\"],AXIS[\"X\",NORTH],AXIS[\"Y\",");
    add_srs_wkt (p, 13, "EAST]]");
    p = add_epsg_def (first, last, 21899, "epsg", 21899,
		      "Bogota 1975 / Colombia East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=4.599047222222222 +lon_0=-68.08091666");
    add_proj4text (p, 1,
		   "666667 +k=1 +x_0=1000000 +y_0=1000000 +ellps=intl +towgs");
    add_proj4text (p, 2, "84=307,304,-318,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Bogota 1975 / Colombia East\",GEOGCS[\"Bogota 1");
    add_srs_wkt (p, 1,
		 "975\",DATUM[\"Bogota_1975\",SPHEROID[\"International 192");
    add_srs_wkt (p, 2,
		 "4\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[30");
    add_srs_wkt (p, 3,
		 "7,304,-318,0,0,0,0],AUTHORITY[\"EPSG\",\"6218\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4218\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",4.599047222222222],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"central_meridian\",-68.08091666666667],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",10000");
    add_srs_wkt (p, 11,
		 "00],PARAMETER[\"false_northing\",1000000],AUTHORITY[\"EP");
    add_srs_wkt (p, 12, "SG\",\"21899\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 22032, "epsg", 22032,
		      "Camacupa / UTM zone 32S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=32 +south +ellps=clrk80 +units=m +no_def");
    add_proj4text (p, 1, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Camacupa / UTM zone 32S\",GEOGCS[\"Camacupa\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Camacupa\",SPHEROID[\"Clarke 1880 (RGS)\",6378249");
    add_srs_wkt (p, 2,
		 ".145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6220\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4220\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",0],PARAMETER[\"central_meridian\",9],PARAMETER[\"scale_");
    add_srs_wkt (p, 9,
		 "factor\",0.9996],PARAMETER[\"false_easting\",500000],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "22032\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 22033, "epsg", 22033,
		      "Camacupa / UTM zone 33S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=33 +south +ellps=clrk80 +units=m +no_def");
    add_proj4text (p, 1, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Camacupa / UTM zone 33S\",GEOGCS[\"Camacupa\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Camacupa\",SPHEROID[\"Clarke 1880 (RGS)\",6378249");
    add_srs_wkt (p, 2,
		 ".145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6220\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4220\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",0],PARAMETER[\"central_meridian\",15],PARAMETER[\"scale");
    add_srs_wkt (p, 9,
		 "_factor\",0.9996],PARAMETER[\"false_easting\",500000],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11,
		 "\"22033\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORT");
    add_srs_wkt (p, 12, "H]]");
    p = add_epsg_def (first, last, 22091, "epsg", 22091,
		      "Camacupa / TM 11.30 SE");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=11.5 +k=0.9996 +x_0=500000 +");
    add_proj4text (p, 1, "y_0=10000000 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Camacupa / TM 11.30 SE\",GEOGCS[\"Camacupa\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"Camacupa\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.");
    add_srs_wkt (p, 2,
		 "145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6220\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4220\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",11.5],PARAMETER[\"scal");
    add_srs_wkt (p, 9,
		 "e_factor\",0.9996],PARAMETER[\"false_easting\",500000],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_northing\",10000000],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11,
		 ",\"22091\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NOR");
    add_srs_wkt (p, 12, "TH]]");
    p = add_epsg_def (first, last, 22092, "epsg", 22092, "Camacupa / TM 12 SE");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=12 +k=0.9996 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=10000000 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Camacupa / TM 12 SE\",GEOGCS[\"Camacupa\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Camacupa\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145");
    add_srs_wkt (p, 2,
		 ",293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6220\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4220\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",12],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 9,
		 "or\",0.9996],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"220");
    add_srs_wkt (p, 11,
		 "92\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 22171, "epsg", 22171,
		      "POSGAR 98 / Argentina 1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-72 +k=1 +x_0=1500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"POSGAR 98 / Argentina 1\",GEOGCS[\"POSGAR 98\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Posiciones_Geodesicas_Argentinas_1998\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6190\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4190\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",-90],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 9,
		 "n\",-72],PARAMETER[\"scale_factor\",1],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",1500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"22171\"],AXIS[\"X\",NORTH],AXIS[\"Y\",E");
    add_srs_wkt (p, 12, "AST]]");
    p = add_epsg_def (first, last, 22172, "epsg", 22172,
		      "POSGAR 98 / Argentina 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-69 +k=1 +x_0=2500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"POSGAR 98 / Argentina 2\",GEOGCS[\"POSGAR 98\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Posiciones_Geodesicas_Argentinas_1998\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6190\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4190\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",-90],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 9,
		 "n\",-69],PARAMETER[\"scale_factor\",1],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",2500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"22172\"],AXIS[\"X\",NORTH],AXIS[\"Y\",E");
    add_srs_wkt (p, 12, "AST]]");
    p = add_epsg_def (first, last, 22173, "epsg", 22173,
		      "POSGAR 98 / Argentina 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-66 +k=1 +x_0=3500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"POSGAR 98 / Argentina 3\",GEOGCS[\"POSGAR 98\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Posiciones_Geodesicas_Argentinas_1998\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6190\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4190\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",-90],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 9,
		 "n\",-66],PARAMETER[\"scale_factor\",1],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",3500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"22173\"],AXIS[\"X\",NORTH],AXIS[\"Y\",E");
    add_srs_wkt (p, 12, "AST]]");
    p = add_epsg_def (first, last, 22174, "epsg", 22174,
		      "POSGAR 98 / Argentina 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-63 +k=1 +x_0=4500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"POSGAR 98 / Argentina 4\",GEOGCS[\"POSGAR 98\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Posiciones_Geodesicas_Argentinas_1998\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6190\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4190\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",-90],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 9,
		 "n\",-63],PARAMETER[\"scale_factor\",1],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",4500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"22174\"],AXIS[\"X\",NORTH],AXIS[\"Y\",E");
    add_srs_wkt (p, 12, "AST]]");
    p = add_epsg_def (first, last, 22175, "epsg", 22175,
		      "POSGAR 98 / Argentina 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-60 +k=1 +x_0=5500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"POSGAR 98 / Argentina 5\",GEOGCS[\"POSGAR 98\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Posiciones_Geodesicas_Argentinas_1998\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6190\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4190\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",-90],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 9,
		 "n\",-60],PARAMETER[\"scale_factor\",1],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",5500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"22175\"],AXIS[\"X\",NORTH],AXIS[\"Y\",E");
    add_srs_wkt (p, 12, "AST]]");
    p = add_epsg_def (first, last, 22176, "epsg", 22176,
		      "POSGAR 98 / Argentina 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-57 +k=1 +x_0=6500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"POSGAR 98 / Argentina 6\",GEOGCS[\"POSGAR 98\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Posiciones_Geodesicas_Argentinas_1998\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6190\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4190\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",-90],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 9,
		 "n\",-57],PARAMETER[\"scale_factor\",1],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",6500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"22176\"],AXIS[\"X\",NORTH],AXIS[\"Y\",E");
    add_srs_wkt (p, 12, "AST]]");
    p = add_epsg_def (first, last, 22177, "epsg", 22177,
		      "POSGAR 98 / Argentina 7");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-54 +k=1 +x_0=7500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"POSGAR 98 / Argentina 7\",GEOGCS[\"POSGAR 98\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Posiciones_Geodesicas_Argentinas_1998\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6190\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4190\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",-90],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 9,
		 "n\",-54],PARAMETER[\"scale_factor\",1],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",7500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"22177\"],AXIS[\"X\",NORTH],AXIS[\"Y\",E");
    add_srs_wkt (p, 12, "AST]]");
    p = add_epsg_def (first, last, 22181, "epsg", 22181,
		      "POSGAR 94 / Argentina 1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-72 +k=1 +x_0=1500000 +y_0");
    add_proj4text (p, 1,
		   "=0 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"POSGAR 94 / Argentina 1\",GEOGCS[\"POSGAR 94\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Posiciones_Geodesicas_Argentinas_1994\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6694");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4694\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",-90],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"central_meridian\",-72],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",1],PARAMETER[\"false_easting\",1500000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"22181\"],AXIS[\"X\"");
    add_srs_wkt (p, 12, ",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 22182, "epsg", 22182,
		      "POSGAR 94 / Argentina 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-69 +k=1 +x_0=2500000 +y_0");
    add_proj4text (p, 1,
		   "=0 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"POSGAR 94 / Argentina 2\",GEOGCS[\"POSGAR 94\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Posiciones_Geodesicas_Argentinas_1994\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6694");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4694\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",-90],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"central_meridian\",-69],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",1],PARAMETER[\"false_easting\",2500000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"22182\"],AXIS[\"X\"");
    add_srs_wkt (p, 12, ",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 22183, "epsg", 22183,
		      "POSGAR 94 / Argentina 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-66 +k=1 +x_0=3500000 +y_0");
    add_proj4text (p, 1,
		   "=0 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"POSGAR 94 / Argentina 3\",GEOGCS[\"POSGAR 94\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Posiciones_Geodesicas_Argentinas_1994\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6694");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4694\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",-90],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"central_meridian\",-66],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",1],PARAMETER[\"false_easting\",3500000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"22183\"],AXIS[\"X\"");
    add_srs_wkt (p, 12, ",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 22184, "epsg", 22184,
		      "POSGAR 94 / Argentina 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-63 +k=1 +x_0=4500000 +y_0");
    add_proj4text (p, 1,
		   "=0 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"POSGAR 94 / Argentina 4\",GEOGCS[\"POSGAR 94\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Posiciones_Geodesicas_Argentinas_1994\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6694");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4694\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",-90],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"central_meridian\",-63],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",1],PARAMETER[\"false_easting\",4500000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"22184\"],AXIS[\"X\"");
    add_srs_wkt (p, 12, ",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 22185, "epsg", 22185,
		      "POSGAR 94 / Argentina 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-60 +k=1 +x_0=5500000 +y_0");
    add_proj4text (p, 1,
		   "=0 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"POSGAR 94 / Argentina 5\",GEOGCS[\"POSGAR 94\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Posiciones_Geodesicas_Argentinas_1994\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6694");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4694\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",-90],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"central_meridian\",-60],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",1],PARAMETER[\"false_easting\",5500000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"22185\"],AXIS[\"X\"");
    add_srs_wkt (p, 12, ",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 22186, "epsg", 22186,
		      "POSGAR 94 / Argentina 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-57 +k=1 +x_0=6500000 +y_0");
    add_proj4text (p, 1,
		   "=0 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"POSGAR 94 / Argentina 6\",GEOGCS[\"POSGAR 94\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Posiciones_Geodesicas_Argentinas_1994\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6694");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4694\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",-90],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"central_meridian\",-57],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",1],PARAMETER[\"false_easting\",6500000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"22186\"],AXIS[\"X\"");
    add_srs_wkt (p, 12, ",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 22187, "epsg", 22187,
		      "POSGAR 94 / Argentina 7");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-54 +k=1 +x_0=7500000 +y_0");
    add_proj4text (p, 1,
		   "=0 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"POSGAR 94 / Argentina 7\",GEOGCS[\"POSGAR 94\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Posiciones_Geodesicas_Argentinas_1994\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6694");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4694\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",-90],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"central_meridian\",-54],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",1],PARAMETER[\"false_easting\",7500000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"22187\"],AXIS[\"X\"");
    add_srs_wkt (p, 12, ",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 22191, "epsg", 22191,
		      "Campo Inchauspe / Argentina 1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-72 +k=1 +x_0=1500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Campo Inchauspe / Argentina 1\",GEOGCS[\"Campo ");
    add_srs_wkt (p, 1,
		 "Inchauspe\",DATUM[\"Campo_Inchauspe\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6221\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"4221\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",-90],PARAMETER[\"central_meridian\",-72],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",1500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 11, "SG\",\"22191\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 22192, "epsg", 22192,
		      "Campo Inchauspe / Argentina 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-69 +k=1 +x_0=2500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Campo Inchauspe / Argentina 2\",GEOGCS[\"Campo ");
    add_srs_wkt (p, 1,
		 "Inchauspe\",DATUM[\"Campo_Inchauspe\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6221\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"4221\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",-90],PARAMETER[\"central_meridian\",-69],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",2500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 11, "SG\",\"22192\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 22193, "epsg", 22193,
		      "Campo Inchauspe / Argentina 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-66 +k=1 +x_0=3500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Campo Inchauspe / Argentina 3\",GEOGCS[\"Campo ");
    add_srs_wkt (p, 1,
		 "Inchauspe\",DATUM[\"Campo_Inchauspe\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6221\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"4221\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",-90],PARAMETER[\"central_meridian\",-66],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",3500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 11, "SG\",\"22193\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 22194, "epsg", 22194,
		      "Campo Inchauspe / Argentina 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-63 +k=1 +x_0=4500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Campo Inchauspe / Argentina 4\",GEOGCS[\"Campo ");
    add_srs_wkt (p, 1,
		 "Inchauspe\",DATUM[\"Campo_Inchauspe\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6221\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"4221\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",-90],PARAMETER[\"central_meridian\",-63],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",4500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 11, "SG\",\"22194\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 22195, "epsg", 22195,
		      "Campo Inchauspe / Argentina 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-60 +k=1 +x_0=5500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Campo Inchauspe / Argentina 5\",GEOGCS[\"Campo ");
    add_srs_wkt (p, 1,
		 "Inchauspe\",DATUM[\"Campo_Inchauspe\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6221\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"4221\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",-90],PARAMETER[\"central_meridian\",-60],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",5500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 11, "SG\",\"22195\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 22196, "epsg", 22196,
		      "Campo Inchauspe / Argentina 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-57 +k=1 +x_0=6500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Campo Inchauspe / Argentina 6\",GEOGCS[\"Campo ");
    add_srs_wkt (p, 1,
		 "Inchauspe\",DATUM[\"Campo_Inchauspe\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6221\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"4221\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",-90],PARAMETER[\"central_meridian\",-57],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",6500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 11, "SG\",\"22196\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 22197, "epsg", 22197,
		      "Campo Inchauspe / Argentina 7");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-90 +lon_0=-54 +k=1 +x_0=7500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Campo Inchauspe / Argentina 7\",GEOGCS[\"Campo ");
    add_srs_wkt (p, 1,
		 "Inchauspe\",DATUM[\"Campo_Inchauspe\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"6221\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 5,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 6,
		 "\"4221\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]");
    add_srs_wkt (p, 7,
		 "],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitu");
    add_srs_wkt (p, 8,
		 "de_of_origin\",-90],PARAMETER[\"central_meridian\",-54],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",7500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 11, "SG\",\"22197\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 22234, "epsg", 22234, "Cape / UTM zone 34S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=34 +south +a=6378249.145 +b=6356514.9663");
    add_proj4text (p, 1, "98753 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Cape / UTM zone 34S\",GEOGCS[\"Cape\",DATUM[\"C");
    add_srs_wkt (p, 1,
		 "ape\",SPHEROID[\"Clarke 1880 (Arc)\",6378249.145,293.466");
    add_srs_wkt (p, 2,
		 "3077,AUTHORITY[\"EPSG\",\"7013\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "6222\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 4,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 5,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4222\"]],UNIT[\"met");
    add_srs_wkt (p, 6,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 7,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 8,
		 "AMETER[\"central_meridian\",21],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 9,
		 "\",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"22234");
    add_srs_wkt (p, 11, "\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 22235, "epsg", 22235, "Cape / UTM zone 35S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=35 +south +a=6378249.145 +b=6356514.9663");
    add_proj4text (p, 1, "98753 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Cape / UTM zone 35S\",GEOGCS[\"Cape\",DATUM[\"C");
    add_srs_wkt (p, 1,
		 "ape\",SPHEROID[\"Clarke 1880 (Arc)\",6378249.145,293.466");
    add_srs_wkt (p, 2,
		 "3077,AUTHORITY[\"EPSG\",\"7013\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "6222\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 4,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 5,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4222\"]],UNIT[\"met");
    add_srs_wkt (p, 6,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 7,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 8,
		 "AMETER[\"central_meridian\",27],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 9,
		 "\",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"22235");
    add_srs_wkt (p, 11, "\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 22236, "epsg", 22236, "Cape / UTM zone 36S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=36 +south +a=6378249.145 +b=6356514.9663");
    add_proj4text (p, 1, "98753 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Cape / UTM zone 36S\",GEOGCS[\"Cape\",DATUM[\"C");
    add_srs_wkt (p, 1,
		 "ape\",SPHEROID[\"Clarke 1880 (Arc)\",6378249.145,293.466");
    add_srs_wkt (p, 2,
		 "3077,AUTHORITY[\"EPSG\",\"7013\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "6222\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 4,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 5,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4222\"]],UNIT[\"met");
    add_srs_wkt (p, 6,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 7,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 8,
		 "AMETER[\"central_meridian\",33],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 9,
		 "\",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_northing\",10000000],AUTHORITY[\"EPSG\",\"22236");
    add_srs_wkt (p, 11, "\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 22332, "epsg", 22332,
		      "Carthage / UTM zone 32N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=32 +a=6378249.2 +b=6356515 +datum=cartha");
    add_proj4text (p, 1, "ge +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Carthage / UTM zone 32N\",GEOGCS[\"Carthage\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Carthage\",SPHEROID[\"Clarke 1880 (IGN)\",6378249");
    add_srs_wkt (p, 2,
		 ".2,293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6223\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"422");
    add_srs_wkt (p, 6,
		 "3\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",9],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",0.9996],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 10,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11,
		 ",\"22332\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NOR");
    add_srs_wkt (p, 12, "TH]]");
    p = add_epsg_def (first, last, 22391, "epsg", 22391,
		      "Carthage / Nord Tunisie");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36 +lat_0=36 +lon_0=9.9 +k_0=0.99962554");
    add_proj4text (p, 1,
		   "4 +x_0=500000 +y_0=300000 +a=6378249.2 +b=6356515 +datum");
    add_proj4text (p, 2, "=carthage +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Carthage / Nord Tunisie\",GEOGCS[\"Carthage\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Carthage\",SPHEROID[\"Clarke 1880 (IGN)\",6378249");
    add_srs_wkt (p, 2,
		 ".2,293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6223\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"422");
    add_srs_wkt (p, 6,
		 "3\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Lambert_Conformal_Conic_1SP\"],PARAMETER[\"lat");
    add_srs_wkt (p, 8,
		 "itude_of_origin\",36],PARAMETER[\"central_meridian\",9.9");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"scale_factor\",0.999625544],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_easting\",500000],PARAMETER[\"false_northing\",30000");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"22391\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 22392, "epsg", 22392,
		      "Carthage / Sud Tunisie");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.3 +lat_0=33.3 +lon_0=9.9 +k_0=0.9996");
    add_proj4text (p, 1,
		   "25769 +x_0=500000 +y_0=300000 +a=6378249.2 +b=6356515 +d");
    add_proj4text (p, 2, "atum=carthage +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Carthage / Sud Tunisie\",GEOGCS[\"Carthage\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"Carthage\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.");
    add_srs_wkt (p, 2,
		 "2,293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6223\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4223");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Lambert_Conformal_Conic_1SP\"],PARAMETER[\"lati");
    add_srs_wkt (p, 8,
		 "tude_of_origin\",33.3],PARAMETER[\"central_meridian\",9.");
    add_srs_wkt (p, 9,
		 "9],PARAMETER[\"scale_factor\",0.999625769],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",500000],PARAMETER[\"false_northing\",3000");
    add_srs_wkt (p, 11,
		 "00],AUTHORITY[\"EPSG\",\"22392\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 12, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 22521, "epsg", 22521,
		      "Corrego Alegre / UTM zone 21S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +south +ellps=intl +towgs84=-206,172,");
    add_proj4text (p, 1, "-6,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Corrego Alegre / UTM zone 21S\",GEOGCS[\"Correg");
    add_srs_wkt (p, 1,
		 "o Alegre\",DATUM[\"Corrego_Alegre\",SPHEROID[\"Internati");
    add_srs_wkt (p, 2,
		 "onal 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[-206,172,-6,0,0,0,0],AUTHORITY[\"EPSG\",\"6225\"]]");
    add_srs_wkt (p, 4,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 5,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 6,
		 "122\"]],AUTHORITY[\"EPSG\",\"4225\"]],UNIT[\"metre\",1,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-57],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 10,
		 "6],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_northing\",10000000],AUTHORITY[\"EPSG\",\"22521\"],AXIS");
    add_srs_wkt (p, 12, "[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 22522, "epsg", 22522,
		      "Corrego Alegre / UTM zone 22S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=22 +south +ellps=intl +towgs84=-206,172,");
    add_proj4text (p, 1, "-6,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Corrego Alegre / UTM zone 22S\",GEOGCS[\"Correg");
    add_srs_wkt (p, 1,
		 "o Alegre\",DATUM[\"Corrego_Alegre\",SPHEROID[\"Internati");
    add_srs_wkt (p, 2,
		 "onal 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[-206,172,-6,0,0,0,0],AUTHORITY[\"EPSG\",\"6225\"]]");
    add_srs_wkt (p, 4,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 5,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 6,
		 "122\"]],AUTHORITY[\"EPSG\",\"4225\"]],UNIT[\"metre\",1,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-51],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 10,
		 "6],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_northing\",10000000],AUTHORITY[\"EPSG\",\"22522\"],AXIS");
    add_srs_wkt (p, 12, "[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 22523, "epsg", 22523,
		      "Corrego Alegre / UTM zone 23S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=23 +south +ellps=intl +towgs84=-206,172,");
    add_proj4text (p, 1, "-6,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Corrego Alegre / UTM zone 23S\",GEOGCS[\"Correg");
    add_srs_wkt (p, 1,
		 "o Alegre\",DATUM[\"Corrego_Alegre\",SPHEROID[\"Internati");
    add_srs_wkt (p, 2,
		 "onal 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[-206,172,-6,0,0,0,0],AUTHORITY[\"EPSG\",\"6225\"]]");
    add_srs_wkt (p, 4,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 5,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 6,
		 "122\"]],AUTHORITY[\"EPSG\",\"4225\"]],UNIT[\"metre\",1,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-45],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 10,
		 "6],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_northing\",10000000],AUTHORITY[\"EPSG\",\"22523\"],AXIS");
    add_srs_wkt (p, 12, "[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 22524, "epsg", 22524,
		      "Corrego Alegre / UTM zone 24S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=24 +south +ellps=intl +towgs84=-206,172,");
    add_proj4text (p, 1, "-6,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Corrego Alegre / UTM zone 24S\",GEOGCS[\"Correg");
    add_srs_wkt (p, 1,
		 "o Alegre\",DATUM[\"Corrego_Alegre\",SPHEROID[\"Internati");
    add_srs_wkt (p, 2,
		 "onal 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[-206,172,-6,0,0,0,0],AUTHORITY[\"EPSG\",\"6225\"]]");
    add_srs_wkt (p, 4,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 5,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 6,
		 "122\"]],AUTHORITY[\"EPSG\",\"4225\"]],UNIT[\"metre\",1,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-39],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 10,
		 "6],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_northing\",10000000],AUTHORITY[\"EPSG\",\"22524\"],AXIS");
    add_srs_wkt (p, 12, "[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 22525, "epsg", 22525,
		      "Corrego Alegre / UTM zone 25S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=25 +south +ellps=intl +towgs84=-206,172,");
    add_proj4text (p, 1, "-6,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Corrego Alegre / UTM zone 25S\",GEOGCS[\"Correg");
    add_srs_wkt (p, 1,
		 "o Alegre\",DATUM[\"Corrego_Alegre\",SPHEROID[\"Internati");
    add_srs_wkt (p, 2,
		 "onal 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[-206,172,-6,0,0,0,0],AUTHORITY[\"EPSG\",\"6225\"]]");
    add_srs_wkt (p, 4,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 5,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 6,
		 "122\"]],AUTHORITY[\"EPSG\",\"4225\"]],UNIT[\"metre\",1,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-33],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 10,
		 "6],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_northing\",10000000],AUTHORITY[\"EPSG\",\"22525\"],AXIS");
    add_srs_wkt (p, 12, "[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 22700, "epsg", 22700,
		      "Deir ez Zor / Levant Zone");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=34.65 +lat_0=34.65 +lon_0=37.35 +k_0=0.");
    add_proj4text (p, 1,
		   "9996256 +x_0=300000 +y_0=300000 +a=6378249.2 +b=6356515 ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Deir ez Zor / Levant Zone\",GEOGCS[\"Deir ez Zo");
    add_srs_wkt (p, 1,
		 "r\",DATUM[\"Deir_ez_Zor\",SPHEROID[\"Clarke 1880 (IGN)\"");
    add_srs_wkt (p, 2,
		 ",6378249.2,293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6227\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4227\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Lambert_Conformal_Conic_1SP\"],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"latitude_of_origin\",34.65],PARAMETER[\"central_mer");
    add_srs_wkt (p, 9,
		 "idian\",37.35],PARAMETER[\"scale_factor\",0.9996256],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_easting\",300000],PARAMETER[\"false_north");
    add_srs_wkt (p, 11,
		 "ing\",300000],AUTHORITY[\"EPSG\",\"22700\"],AXIS[\"X\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 22770, "epsg", 22770,
		      "Deir ez Zor / Syria Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=34.65 +lat_0=34.65 +lon_0=37.35 +k_0=0.");
    add_proj4text (p, 1,
		   "9996256 +x_0=300000 +y_0=300000 +a=6378249.2 +b=6356515 ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Deir ez Zor / Syria Lambert\",GEOGCS[\"Deir ez ");
    add_srs_wkt (p, 1,
		 "Zor\",DATUM[\"Deir_ez_Zor\",SPHEROID[\"Clarke 1880 (IGN)");
    add_srs_wkt (p, 2,
		 "\",6378249.2,293.4660212936265,AUTHORITY[\"EPSG\",\"7011");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6227\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4227\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Lambert_Conformal_Conic_1SP\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"latitude_of_origin\",34.65],PARAMETER[\"central_m");
    add_srs_wkt (p, 9,
		 "eridian\",37.35],PARAMETER[\"scale_factor\",0.9996256],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",300000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",300000],AUTHORITY[\"EPSG\",\"22770\"],AXIS[\"X\"");
    add_srs_wkt (p, 12, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 22780, "epsg", 22780,
		      "Deir ez Zor / Levant Stereographic");
    add_proj4text (p, 0,
		   "+proj=sterea +lat_0=34.2 +lon_0=39.15 +k=0.9995341 +x_0=");
    add_proj4text (p, 1, "0 +y_0=0 +a=6378249.2 +b=6356515 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Deir ez Zor / Levant Stereographic\",GEOGCS[\"D");
    add_srs_wkt (p, 1,
		 "eir ez Zor\",DATUM[\"Deir_ez_Zor\",SPHEROID[\"Clarke 188");
    add_srs_wkt (p, 2,
		 "0 (IGN)\",6378249.2,293.4660212936265,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7011\"]],AUTHORITY[\"EPSG\",\"6227\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4227\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Oblique_Stereographic\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",34.2],PARAMETER[\"central_m");
    add_srs_wkt (p, 9,
		 "eridian\",39.15],PARAMETER[\"scale_factor\",0.9995341],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",0],PARAMETER[\"false_northing");
    add_srs_wkt (p, 11,
		 "\",0],AUTHORITY[\"EPSG\",\"22780\"],AXIS[\"X\",EAST],AXI");
    add_srs_wkt (p, 12, "S[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 22832, "epsg", 22832,
		      "Douala / UTM zone 32N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=32 +a=6378249.2 +b=6356515 +units=m +no_");
    add_proj4text (p, 1, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Douala / UTM zone 32N (deprecated)\",GEOGCS[\"D");
    add_srs_wkt (p, 1,
		 "ouala\",DATUM[\"Douala\",SPHEROID[\"Clarke 1880 (IGN)\",");
    add_srs_wkt (p, 2,
		 "6378249.2,293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6228\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "5199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4228\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",9],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"22832\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 22991, "epsg", 22991,
		      "Egypt 1907 / Blue Belt");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=35 +k=1 +x_0=300000 +y_0=11");
    add_proj4text (p, 1, "00000 +ellps=helmert +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Egypt 1907 / Blue Belt\",GEOGCS[\"Egypt 1907\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Egypt_1907\",SPHEROID[\"Helmert 1906\",6378200,2");
    add_srs_wkt (p, 2,
		 "98.3,AUTHORITY[\"EPSG\",\"7020\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "6229\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 4,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 5,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4229\"]],UNIT[\"met");
    add_srs_wkt (p, 6,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 7,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",30],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"central_meridian\",35],PARAMETER[\"scale_facto");
    add_srs_wkt (p, 9,
		 "r\",1],PARAMETER[\"false_easting\",300000],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_northing\",1100000],AUTHORITY[\"EPSG\",\"22991\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 22992, "epsg", 22992,
		      "Egypt 1907 / Red Belt");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=31 +k=1 +x_0=615000 +y_0=81");
    add_proj4text (p, 1, "0000 +ellps=helmert +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Egypt 1907 / Red Belt\",GEOGCS[\"Egypt 1907\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Egypt_1907\",SPHEROID[\"Helmert 1906\",6378200,29");
    add_srs_wkt (p, 2,
		 "8.3,AUTHORITY[\"EPSG\",\"7020\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "229\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4229\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transve");
    add_srs_wkt (p, 7,
		 "rse_Mercator\"],PARAMETER[\"latitude_of_origin\",30],PAR");
    add_srs_wkt (p, 8,
		 "AMETER[\"central_meridian\",31],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 9,
		 "\",1],PARAMETER[\"false_easting\",615000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",810000],AUTHORITY[\"EPSG\",\"22992\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 22993, "epsg", 22993,
		      "Egypt 1907 / Purple Belt");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=27 +k=1 +x_0=700000 +y_0=20");
    add_proj4text (p, 1, "0000 +ellps=helmert +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Egypt 1907 / Purple Belt\",GEOGCS[\"Egypt 1907\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Egypt_1907\",SPHEROID[\"Helmert 1906\",6378200,");
    add_srs_wkt (p, 2,
		 "298.3,AUTHORITY[\"EPSG\",\"7020\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "6229\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 4,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 5,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4229\"]],UNIT[\"met");
    add_srs_wkt (p, 6,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 7,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",30],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"central_meridian\",27],PARAMETER[\"scale_facto");
    add_srs_wkt (p, 9,
		 "r\",1],PARAMETER[\"false_easting\",700000],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_northing\",200000],AUTHORITY[\"EPSG\",\"22993\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 22994, "epsg", 22994,
		      "Egypt 1907 / Extended Purple Belt");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=27 +k=1 +x_0=700000 +y_0=12");
    add_proj4text (p, 1, "00000 +ellps=helmert +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Egypt 1907 / Extended Purple Belt\",GEOGCS[\"Eg");
    add_srs_wkt (p, 1,
		 "ypt 1907\",DATUM[\"Egypt_1907\",SPHEROID[\"Helmert 1906\"");
    add_srs_wkt (p, 2,
		 ",6378200,298.3,AUTHORITY[\"EPSG\",\"7020\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6229\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4229\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",30],PARAMETER[\"central_meridian\",27],PARAMETER[\"sc");
    add_srs_wkt (p, 9,
		 "ale_factor\",1],PARAMETER[\"false_easting\",700000],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_northing\",1200000],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 11,
		 "2994\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 23028, "epsg", 23028, "ED50 / UTM zone 28N");
    add_proj4text (p, 0, "+proj=utm +zone=28 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / UTM zone 28N\",GEOGCS[\"ED50\",DATUM[\"E");
    add_srs_wkt (p, 1,
		 "uropean_Datum_1950\",SPHEROID[\"International 1924\",637");
    add_srs_wkt (p, 2,
		 "8388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",-15],PARAMETER[\"scale_fac");
    add_srs_wkt (p, 9,
		 "tor\",0.9996],PARAMETER[\"false_easting\",500000],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23028\"],");
    add_srs_wkt (p, 11, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23029, "epsg", 23029, "ED50 / UTM zone 29N");
    add_proj4text (p, 0, "+proj=utm +zone=29 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / UTM zone 29N\",GEOGCS[\"ED50\",DATUM[\"E");
    add_srs_wkt (p, 1,
		 "uropean_Datum_1950\",SPHEROID[\"International 1924\",637");
    add_srs_wkt (p, 2,
		 "8388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",-9],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 9,
		 "or\",0.9996],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23029\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23030, "epsg", 23030, "ED50 / UTM zone 30N");
    add_proj4text (p, 0, "+proj=utm +zone=30 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / UTM zone 30N\",GEOGCS[\"ED50\",DATUM[\"E");
    add_srs_wkt (p, 1,
		 "uropean_Datum_1950\",SPHEROID[\"International 1924\",637");
    add_srs_wkt (p, 2,
		 "8388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",-3],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 9,
		 "or\",0.9996],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23030\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23031, "epsg", 23031, "ED50 / UTM zone 31N");
    add_proj4text (p, 0, "+proj=utm +zone=31 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / UTM zone 31N\",GEOGCS[\"ED50\",DATUM[\"E");
    add_srs_wkt (p, 1,
		 "uropean_Datum_1950\",SPHEROID[\"International 1924\",637");
    add_srs_wkt (p, 2,
		 "8388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",3],PARAMETER[\"scale_facto");
    add_srs_wkt (p, 9,
		 "r\",0.9996],PARAMETER[\"false_easting\",500000],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23031\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23032, "epsg", 23032, "ED50 / UTM zone 32N");
    add_proj4text (p, 0, "+proj=utm +zone=32 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / UTM zone 32N\",GEOGCS[\"ED50\",DATUM[\"E");
    add_srs_wkt (p, 1,
		 "uropean_Datum_1950\",SPHEROID[\"International 1924\",637");
    add_srs_wkt (p, 2,
		 "8388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",9],PARAMETER[\"scale_facto");
    add_srs_wkt (p, 9,
		 "r\",0.9996],PARAMETER[\"false_easting\",500000],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23032\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23033, "epsg", 23033, "ED50 / UTM zone 33N");
    add_proj4text (p, 0, "+proj=utm +zone=33 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / UTM zone 33N\",GEOGCS[\"ED50\",DATUM[\"E");
    add_srs_wkt (p, 1,
		 "uropean_Datum_1950\",SPHEROID[\"International 1924\",637");
    add_srs_wkt (p, 2,
		 "8388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",15],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 9,
		 "or\",0.9996],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23033\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23034, "epsg", 23034, "ED50 / UTM zone 34N");
    add_proj4text (p, 0, "+proj=utm +zone=34 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / UTM zone 34N\",GEOGCS[\"ED50\",DATUM[\"E");
    add_srs_wkt (p, 1,
		 "uropean_Datum_1950\",SPHEROID[\"International 1924\",637");
    add_srs_wkt (p, 2,
		 "8388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",21],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 9,
		 "or\",0.9996],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23034\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23035, "epsg", 23035, "ED50 / UTM zone 35N");
    add_proj4text (p, 0, "+proj=utm +zone=35 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / UTM zone 35N\",GEOGCS[\"ED50\",DATUM[\"E");
    add_srs_wkt (p, 1,
		 "uropean_Datum_1950\",SPHEROID[\"International 1924\",637");
    add_srs_wkt (p, 2,
		 "8388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",27],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 9,
		 "or\",0.9996],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23035\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23036, "epsg", 23036, "ED50 / UTM zone 36N");
    add_proj4text (p, 0, "+proj=utm +zone=36 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / UTM zone 36N\",GEOGCS[\"ED50\",DATUM[\"E");
    add_srs_wkt (p, 1,
		 "uropean_Datum_1950\",SPHEROID[\"International 1924\",637");
    add_srs_wkt (p, 2,
		 "8388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",33],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 9,
		 "or\",0.9996],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23036\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23037, "epsg", 23037, "ED50 / UTM zone 37N");
    add_proj4text (p, 0, "+proj=utm +zone=37 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / UTM zone 37N\",GEOGCS[\"ED50\",DATUM[\"E");
    add_srs_wkt (p, 1,
		 "uropean_Datum_1950\",SPHEROID[\"International 1924\",637");
    add_srs_wkt (p, 2,
		 "8388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",39],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 9,
		 "or\",0.9996],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23037\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23038, "epsg", 23038, "ED50 / UTM zone 38N");
    add_proj4text (p, 0, "+proj=utm +zone=38 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / UTM zone 38N\",GEOGCS[\"ED50\",DATUM[\"E");
    add_srs_wkt (p, 1,
		 "uropean_Datum_1950\",SPHEROID[\"International 1924\",637");
    add_srs_wkt (p, 2,
		 "8388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",45],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 9,
		 "or\",0.9996],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23038\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23090, "epsg", 23090, "ED50 / TM 0 N");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=0 +k=0.9996 +x_0=500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / TM 0 N\",GEOGCS[\"ED50\",DATUM[\"Europea");
    add_srs_wkt (p, 1,
		 "n_Datum_1950\",SPHEROID[\"International 1924\",6378388,2");
    add_srs_wkt (p, 2,
		 "97,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"62");
    add_srs_wkt (p, 3,
		 "30\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",0],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"23090\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23095, "epsg", 23095, "ED50 / TM 5 NE");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=5 +k=0.9996 +x_0=500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ED50 / TM 5 NE\",GEOGCS[\"ED50\",DATUM[\"Europe");
    add_srs_wkt (p, 1,
		 "an_Datum_1950\",SPHEROID[\"International 1924\",6378388,");
    add_srs_wkt (p, 2,
		 "297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4230\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transve");
    add_srs_wkt (p, 7,
		 "rse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"central_meridian\",5],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 9,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"23095\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23239, "epsg", 23239,
		      "Fahud / UTM zone 39N");
    add_proj4text (p, 0, "+proj=utm +zone=39 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Fahud / UTM zone 39N\",GEOGCS[\"Fahud\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Fahud\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.4");
    add_srs_wkt (p, 2,
		 "65,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"62");
    add_srs_wkt (p, 3,
		 "32\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4232\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",51],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 9,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"23239\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23240, "epsg", 23240,
		      "Fahud / UTM zone 40N");
    add_proj4text (p, 0, "+proj=utm +zone=40 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Fahud / UTM zone 40N\",GEOGCS[\"Fahud\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Fahud\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.4");
    add_srs_wkt (p, 2,
		 "65,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"62");
    add_srs_wkt (p, 3,
		 "32\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4232\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",57],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 9,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"23240\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23433, "epsg", 23433,
		      "Garoua / UTM zone 33N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=33 +a=6378249.2 +b=6356515 +units=m +no_");
    add_proj4text (p, 1, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Garoua / UTM zone 33N (deprecated)\",GEOGCS[\"G");
    add_srs_wkt (p, 1,
		 "aroua\",DATUM[\"Garoua\",SPHEROID[\"Clarke 1880 (IGN)\",");
    add_srs_wkt (p, 2,
		 "6378249.2,293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6234\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "5199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4234\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",15],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"23433\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 23700, "epsg", 23700, "HD72 / EOV");
    add_proj4text (p, 0,
		   "+proj=somerc +lat_0=47.14439372222222 +lon_0=19.04857177");
    add_proj4text (p, 1,
		   "777778 +k_0=0.99993 +x_0=650000 +y_0=200000 +ellps=GRS67");
    add_proj4text (p, 2, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"HD72 / EOV\",GEOGCS[\"HD72\",DATUM[\"Hungarian_");
    add_srs_wkt (p, 1,
		 "Datum_1972\",SPHEROID[\"GRS 1967\",6378160,298.247167427");
    add_srs_wkt (p, 2,
		 ",AUTHORITY[\"EPSG\",\"7036\"]],AUTHORITY[\"EPSG\",\"6237");
    add_srs_wkt (p, 3,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 4,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4237\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Hotine_Obl");
    add_srs_wkt (p, 7,
		 "ique_Mercator\"],PARAMETER[\"latitude_of_center\",47.144");
    add_srs_wkt (p, 8,
		 "39372222222],PARAMETER[\"longitude_of_center\",19.048571");
    add_srs_wkt (p, 9,
		 "77777778],PARAMETER[\"azimuth\",90],PARAMETER[\"rectifie");
    add_srs_wkt (p, 10,
		 "d_grid_angle\",90],PARAMETER[\"scale_factor\",0.99993],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_easting\",650000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 12,
		 "thing\",200000],AUTHORITY[\"EPSG\",\"23700\"],AXIS[\"Y\"");
    add_srs_wkt (p, 13, ",EAST],AXIS[\"X\",NORTH]]");
    p = add_epsg_def (first, last, 23830, "epsg", 23830,
		      "DGN95 / Indonesia TM-3 zone 46.2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=94.5 +k=0.9999 +x_0=200000 +");
    add_proj4text (p, 1,
		   "y_0=1500000 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=m");
    add_proj4text (p, 2, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / Indonesia TM-3 zone 46.2\",GEOGCS[\"DGN");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Datum_Geodesi_Nasional_1995\",SPHEROID[\"WG");
    add_srs_wkt (p, 2,
		 "S 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",94.5],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 10,
		 "99],PARAMETER[\"false_easting\",200000],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",1500000],AUTHORITY[\"EPSG\",\"23830\"],AXIS");
    add_srs_wkt (p, 12, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 23831, "epsg", 23831,
		      "DGN95 / Indonesia TM-3 zone 47.1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=97.5 +k=0.9999 +x_0=200000 +");
    add_proj4text (p, 1,
		   "y_0=1500000 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=m");
    add_proj4text (p, 2, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / Indonesia TM-3 zone 47.1\",GEOGCS[\"DGN");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Datum_Geodesi_Nasional_1995\",SPHEROID[\"WG");
    add_srs_wkt (p, 2,
		 "S 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",97.5],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 10,
		 "99],PARAMETER[\"false_easting\",200000],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",1500000],AUTHORITY[\"EPSG\",\"23831\"],AXIS");
    add_srs_wkt (p, 12, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 23832, "epsg", 23832,
		      "DGN95 / Indonesia TM-3 zone 47.2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=100.5 +k=0.9999 +x_0=200000 ");
    add_proj4text (p, 1,
		   "+y_0=1500000 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / Indonesia TM-3 zone 47.2\",GEOGCS[\"DGN");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Datum_Geodesi_Nasional_1995\",SPHEROID[\"WG");
    add_srs_wkt (p, 2,
		 "S 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",100.5],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "999],PARAMETER[\"false_easting\",200000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",1500000],AUTHORITY[\"EPSG\",\"23832\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 23833, "epsg", 23833,
		      "DGN95 / Indonesia TM-3 zone 48.1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=103.5 +k=0.9999 +x_0=200000 ");
    add_proj4text (p, 1,
		   "+y_0=1500000 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / Indonesia TM-3 zone 48.1\",GEOGCS[\"DGN");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Datum_Geodesi_Nasional_1995\",SPHEROID[\"WG");
    add_srs_wkt (p, 2,
		 "S 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",103.5],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "999],PARAMETER[\"false_easting\",200000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",1500000],AUTHORITY[\"EPSG\",\"23833\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 23834, "epsg", 23834,
		      "DGN95 / Indonesia TM-3 zone 48.2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=106.5 +k=0.9999 +x_0=200000 ");
    add_proj4text (p, 1,
		   "+y_0=1500000 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / Indonesia TM-3 zone 48.2\",GEOGCS[\"DGN");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Datum_Geodesi_Nasional_1995\",SPHEROID[\"WG");
    add_srs_wkt (p, 2,
		 "S 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",106.5],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "999],PARAMETER[\"false_easting\",200000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",1500000],AUTHORITY[\"EPSG\",\"23834\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 23835, "epsg", 23835,
		      "DGN95 / Indonesia TM-3 zone 49.1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=109.5 +k=0.9999 +x_0=200000 ");
    add_proj4text (p, 1,
		   "+y_0=1500000 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / Indonesia TM-3 zone 49.1\",GEOGCS[\"DGN");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Datum_Geodesi_Nasional_1995\",SPHEROID[\"WG");
    add_srs_wkt (p, 2,
		 "S 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",109.5],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "999],PARAMETER[\"false_easting\",200000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",1500000],AUTHORITY[\"EPSG\",\"23835\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 23836, "epsg", 23836,
		      "DGN95 / Indonesia TM-3 zone 49.2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=112.5 +k=0.9999 +x_0=200000 ");
    add_proj4text (p, 1,
		   "+y_0=1500000 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / Indonesia TM-3 zone 49.2\",GEOGCS[\"DGN");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Datum_Geodesi_Nasional_1995\",SPHEROID[\"WG");
    add_srs_wkt (p, 2,
		 "S 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",112.5],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "999],PARAMETER[\"false_easting\",200000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",1500000],AUTHORITY[\"EPSG\",\"23836\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 23837, "epsg", 23837,
		      "DGN95 / Indonesia TM-3 zone 50.1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=115.5 +k=0.9999 +x_0=200000 ");
    add_proj4text (p, 1,
		   "+y_0=1500000 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / Indonesia TM-3 zone 50.1\",GEOGCS[\"DGN");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Datum_Geodesi_Nasional_1995\",SPHEROID[\"WG");
    add_srs_wkt (p, 2,
		 "S 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",115.5],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "999],PARAMETER[\"false_easting\",200000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",1500000],AUTHORITY[\"EPSG\",\"23837\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 23838, "epsg", 23838,
		      "DGN95 / Indonesia TM-3 zone 50.2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=118.5 +k=0.9999 +x_0=200000 ");
    add_proj4text (p, 1,
		   "+y_0=1500000 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / Indonesia TM-3 zone 50.2\",GEOGCS[\"DGN");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Datum_Geodesi_Nasional_1995\",SPHEROID[\"WG");
    add_srs_wkt (p, 2,
		 "S 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",118.5],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "999],PARAMETER[\"false_easting\",200000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",1500000],AUTHORITY[\"EPSG\",\"23838\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 23839, "epsg", 23839,
		      "DGN95 / Indonesia TM-3 zone 51.1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=121.5 +k=0.9999 +x_0=200000 ");
    add_proj4text (p, 1,
		   "+y_0=1500000 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / Indonesia TM-3 zone 51.1\",GEOGCS[\"DGN");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Datum_Geodesi_Nasional_1995\",SPHEROID[\"WG");
    add_srs_wkt (p, 2,
		 "S 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",121.5],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "999],PARAMETER[\"false_easting\",200000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",1500000],AUTHORITY[\"EPSG\",\"23839\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 23840, "epsg", 23840,
		      "DGN95 / Indonesia TM-3 zone 51.2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=124.5 +k=0.9999 +x_0=200000 ");
    add_proj4text (p, 1,
		   "+y_0=1500000 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / Indonesia TM-3 zone 51.2\",GEOGCS[\"DGN");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Datum_Geodesi_Nasional_1995\",SPHEROID[\"WG");
    add_srs_wkt (p, 2,
		 "S 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",124.5],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "999],PARAMETER[\"false_easting\",200000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",1500000],AUTHORITY[\"EPSG\",\"23840\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 23841, "epsg", 23841,
		      "DGN95 / Indonesia TM-3 zone 52.1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=127.5 +k=0.9999 +x_0=200000 ");
    add_proj4text (p, 1,
		   "+y_0=1500000 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / Indonesia TM-3 zone 52.1\",GEOGCS[\"DGN");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Datum_Geodesi_Nasional_1995\",SPHEROID[\"WG");
    add_srs_wkt (p, 2,
		 "S 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",127.5],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "999],PARAMETER[\"false_easting\",200000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",1500000],AUTHORITY[\"EPSG\",\"23841\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 23842, "epsg", 23842,
		      "DGN95 / Indonesia TM-3 zone 52.2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=130.5 +k=0.9999 +x_0=200000 ");
    add_proj4text (p, 1,
		   "+y_0=1500000 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / Indonesia TM-3 zone 52.2\",GEOGCS[\"DGN");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Datum_Geodesi_Nasional_1995\",SPHEROID[\"WG");
    add_srs_wkt (p, 2,
		 "S 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",130.5],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "999],PARAMETER[\"false_easting\",200000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",1500000],AUTHORITY[\"EPSG\",\"23842\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 23843, "epsg", 23843,
		      "DGN95 / Indonesia TM-3 zone 53.1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=133.5 +k=0.9999 +x_0=200000 ");
    add_proj4text (p, 1,
		   "+y_0=1500000 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / Indonesia TM-3 zone 53.1\",GEOGCS[\"DGN");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Datum_Geodesi_Nasional_1995\",SPHEROID[\"WG");
    add_srs_wkt (p, 2,
		 "S 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",133.5],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "999],PARAMETER[\"false_easting\",200000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",1500000],AUTHORITY[\"EPSG\",\"23843\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 23844, "epsg", 23844,
		      "DGN95 / Indonesia TM-3 zone 53.2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=136.5 +k=0.9999 +x_0=200000 ");
    add_proj4text (p, 1,
		   "+y_0=1500000 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / Indonesia TM-3 zone 53.2\",GEOGCS[\"DGN");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Datum_Geodesi_Nasional_1995\",SPHEROID[\"WG");
    add_srs_wkt (p, 2,
		 "S 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",136.5],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "999],PARAMETER[\"false_easting\",200000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",1500000],AUTHORITY[\"EPSG\",\"23844\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 23845, "epsg", 23845,
		      "DGN95 / Indonesia TM-3 zone 54.1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=139.5 +k=0.9999 +x_0=200000 ");
    add_proj4text (p, 1,
		   "+y_0=1500000 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / Indonesia TM-3 zone 54.1\",GEOGCS[\"DGN");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Datum_Geodesi_Nasional_1995\",SPHEROID[\"WG");
    add_srs_wkt (p, 2,
		 "S 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 8,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",139.5],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "999],PARAMETER[\"false_easting\",200000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",1500000],AUTHORITY[\"EPSG\",\"23845\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 23846, "epsg", 23846, "ID74 / UTM zone 46N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=46 +a=6378160 +b=6356774.50408554 +units");
    add_proj4text (p, 1, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ID74 / UTM zone 46N\",GEOGCS[\"ID74\",DATUM[\"I");
    add_srs_wkt (p, 1,
		 "ndonesian_Datum_1974\",SPHEROID[\"Indonesian National Sp");
    add_srs_wkt (p, 2,
		 "heroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4238\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",93],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 11,
		 "G\",\"23846\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",");
    add_srs_wkt (p, 12, "NORTH]]");
    p = add_epsg_def (first, last, 23847, "epsg", 23847, "ID74 / UTM zone 47N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=47 +a=6378160 +b=6356774.50408554 +units");
    add_proj4text (p, 1, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ID74 / UTM zone 47N\",GEOGCS[\"ID74\",DATUM[\"I");
    add_srs_wkt (p, 1,
		 "ndonesian_Datum_1974\",SPHEROID[\"Indonesian National Sp");
    add_srs_wkt (p, 2,
		 "heroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4238\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",99],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 11,
		 "G\",\"23847\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",");
    add_srs_wkt (p, 12, "NORTH]]");
    p = add_epsg_def (first, last, 23848, "epsg", 23848, "ID74 / UTM zone 48N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=48 +a=6378160 +b=6356774.50408554 +units");
    add_proj4text (p, 1, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ID74 / UTM zone 48N\",GEOGCS[\"ID74\",DATUM[\"I");
    add_srs_wkt (p, 1,
		 "ndonesian_Datum_1974\",SPHEROID[\"Indonesian National Sp");
    add_srs_wkt (p, 2,
		 "heroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4238\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",105],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 11,
		 "PSG\",\"23848\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 23849, "epsg", 23849, "ID74 / UTM zone 49N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=49 +a=6378160 +b=6356774.50408554 +units");
    add_proj4text (p, 1, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ID74 / UTM zone 49N\",GEOGCS[\"ID74\",DATUM[\"I");
    add_srs_wkt (p, 1,
		 "ndonesian_Datum_1974\",SPHEROID[\"Indonesian National Sp");
    add_srs_wkt (p, 2,
		 "heroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4238\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",111],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 11,
		 "PSG\",\"23849\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 23850, "epsg", 23850, "ID74 / UTM zone 50N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=50 +a=6378160 +b=6356774.50408554 +units");
    add_proj4text (p, 1, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ID74 / UTM zone 50N\",GEOGCS[\"ID74\",DATUM[\"I");
    add_srs_wkt (p, 1,
		 "ndonesian_Datum_1974\",SPHEROID[\"Indonesian National Sp");
    add_srs_wkt (p, 2,
		 "heroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4238\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",117],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 11,
		 "PSG\",\"23850\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 23851, "epsg", 23851, "ID74 / UTM zone 51N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=51 +a=6378160 +b=6356774.50408554 +units");
    add_proj4text (p, 1, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ID74 / UTM zone 51N\",GEOGCS[\"ID74\",DATUM[\"I");
    add_srs_wkt (p, 1,
		 "ndonesian_Datum_1974\",SPHEROID[\"Indonesian National Sp");
    add_srs_wkt (p, 2,
		 "heroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4238\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",123],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 11,
		 "PSG\",\"23851\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 23852, "epsg", 23852, "ID74 / UTM zone 52N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=52 +a=6378160 +b=6356774.50408554 +units");
    add_proj4text (p, 1, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ID74 / UTM zone 52N\",GEOGCS[\"ID74\",DATUM[\"I");
    add_srs_wkt (p, 1,
		 "ndonesian_Datum_1974\",SPHEROID[\"Indonesian National Sp");
    add_srs_wkt (p, 2,
		 "heroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4238\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",129],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 11,
		 "PSG\",\"23852\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 23853, "epsg", 23853,
		      "ID74 / UTM zone 53N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=53 +a=6378160 +b=6356774.50408554 +units");
    add_proj4text (p, 1, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ID74 / UTM zone 53N (deprecated)\",GEOGCS[\"ID7");
    add_srs_wkt (p, 1,
		 "4\",DATUM[\"Indonesian_Datum_1974\",SPHEROID[\"Indonesia");
    add_srs_wkt (p, 2,
		 "n National Spheroid\",6378160,298.247,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7021\"]],AUTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4238\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",135],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"23853\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 12, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23866, "epsg", 23866,
		      "DGN95 / UTM zone 46N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=46 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / UTM zone 46N\",GEOGCS[\"DGN95\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS 84\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 4,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 5,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"latitude_of_origin\",0],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",93],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"23866\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 12, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23867, "epsg", 23867,
		      "DGN95 / UTM zone 47N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=47 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / UTM zone 47N\",GEOGCS[\"DGN95\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS 84\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 4,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 5,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"latitude_of_origin\",0],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",99],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"23867\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 12, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23868, "epsg", 23868,
		      "DGN95 / UTM zone 48N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=48 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / UTM zone 48N\",GEOGCS[\"DGN95\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS 84\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 4,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 5,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"latitude_of_origin\",0],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",105],PARAMETER[\"scale_factor\",0.9996],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"23868\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23869, "epsg", 23869,
		      "DGN95 / UTM zone 49N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=49 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / UTM zone 49N\",GEOGCS[\"DGN95\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS 84\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 4,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 5,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"latitude_of_origin\",0],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",111],PARAMETER[\"scale_factor\",0.9996],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"23869\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23870, "epsg", 23870,
		      "DGN95 / UTM zone 50N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=50 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / UTM zone 50N\",GEOGCS[\"DGN95\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS 84\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 4,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 5,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"latitude_of_origin\",0],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",117],PARAMETER[\"scale_factor\",0.9996],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"23870\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23871, "epsg", 23871,
		      "DGN95 / UTM zone 51N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=51 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / UTM zone 51N\",GEOGCS[\"DGN95\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS 84\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 4,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 5,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"latitude_of_origin\",0],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",123],PARAMETER[\"scale_factor\",0.9996],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"23871\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23872, "epsg", 23872,
		      "DGN95 / UTM zone 52N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=52 +ellps=WGS84 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / UTM zone 52N\",GEOGCS[\"DGN95\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS 84\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 4,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 5,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"latitude_of_origin\",0],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",129],PARAMETER[\"scale_factor\",0.9996],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"23872\"],AXIS[\"Easting\",EAST],A");
    add_srs_wkt (p, 12, "XIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23877, "epsg", 23877,
		      "DGN95 / UTM zone 47S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=47 +south +ellps=WGS84 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / UTM zone 47S\",GEOGCS[\"DGN95\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS 84\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 4,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 5,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"latitude_of_origin\",0],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",99],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",500000],PARAMETER[\"false_northing\",100");
    add_srs_wkt (p, 11,
		 "00000],AUTHORITY[\"EPSG\",\"23877\"],AXIS[\"Easting\",EA");
    add_srs_wkt (p, 12, "ST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23878, "epsg", 23878,
		      "DGN95 / UTM zone 48S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=48 +south +ellps=WGS84 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / UTM zone 48S\",GEOGCS[\"DGN95\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS 84\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 4,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 5,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"latitude_of_origin\",0],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",105],PARAMETER[\"scale_factor\",0.9996],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",1");
    add_srs_wkt (p, 11,
		 "0000000],AUTHORITY[\"EPSG\",\"23878\"],AXIS[\"Easting\",");
    add_srs_wkt (p, 12, "EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23879, "epsg", 23879,
		      "DGN95 / UTM zone 49S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=49 +south +ellps=WGS84 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / UTM zone 49S\",GEOGCS[\"DGN95\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS 84\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 4,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 5,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"latitude_of_origin\",0],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",111],PARAMETER[\"scale_factor\",0.9996],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",1");
    add_srs_wkt (p, 11,
		 "0000000],AUTHORITY[\"EPSG\",\"23879\"],AXIS[\"Easting\",");
    add_srs_wkt (p, 12, "EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23880, "epsg", 23880,
		      "DGN95 / UTM zone 50S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=50 +south +ellps=WGS84 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / UTM zone 50S\",GEOGCS[\"DGN95\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS 84\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 4,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 5,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"latitude_of_origin\",0],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",117],PARAMETER[\"scale_factor\",0.9996],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",1");
    add_srs_wkt (p, 11,
		 "0000000],AUTHORITY[\"EPSG\",\"23880\"],AXIS[\"Easting\",");
    add_srs_wkt (p, 12, "EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23881, "epsg", 23881,
		      "DGN95 / UTM zone 51S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=51 +south +ellps=WGS84 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / UTM zone 51S\",GEOGCS[\"DGN95\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS 84\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 4,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 5,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"latitude_of_origin\",0],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",123],PARAMETER[\"scale_factor\",0.9996],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",1");
    add_srs_wkt (p, 11,
		 "0000000],AUTHORITY[\"EPSG\",\"23881\"],AXIS[\"Easting\",");
    add_srs_wkt (p, 12, "EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23882, "epsg", 23882,
		      "DGN95 / UTM zone 52S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=52 +south +ellps=WGS84 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / UTM zone 52S\",GEOGCS[\"DGN95\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS 84\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 4,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 5,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"latitude_of_origin\",0],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",129],PARAMETER[\"scale_factor\",0.9996],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",1");
    add_srs_wkt (p, 11,
		 "0000000],AUTHORITY[\"EPSG\",\"23882\"],AXIS[\"Easting\",");
    add_srs_wkt (p, 12, "EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23883, "epsg", 23883,
		      "DGN95 / UTM zone 53S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=53 +south +ellps=WGS84 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / UTM zone 53S\",GEOGCS[\"DGN95\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS 84\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 4,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 5,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"latitude_of_origin\",0],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",135],PARAMETER[\"scale_factor\",0.9996],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",1");
    add_srs_wkt (p, 11,
		 "0000000],AUTHORITY[\"EPSG\",\"23883\"],AXIS[\"Easting\",");
    add_srs_wkt (p, 12, "EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23884, "epsg", 23884,
		      "DGN95 / UTM zone 54S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=54 +south +ellps=WGS84 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DGN95 / UTM zone 54S\",GEOGCS[\"DGN95\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS 84\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,");
    add_srs_wkt (p, 3,
		 "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 4,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 5,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"4755\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"latitude_of_origin\",0],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",141],PARAMETER[\"scale_factor\",0.9996],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",1");
    add_srs_wkt (p, 11,
		 "0000000],AUTHORITY[\"EPSG\",\"23884\"],AXIS[\"Easting\",");
    add_srs_wkt (p, 12, "EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23886, "epsg", 23886,
		      "ID74 / UTM zone 46S (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=46 +south +a=6378160 +b=6356774.50408554");
    add_proj4text (p, 1, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ID74 / UTM zone 46S (deprecated)\",GEOGCS[\"ID7");
    add_srs_wkt (p, 1,
		 "4\",DATUM[\"Indonesian_Datum_1974\",SPHEROID[\"Indonesia");
    add_srs_wkt (p, 2,
		 "n National Spheroid\",6378160,298.247,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7021\"]],AUTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4238\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",93],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",500000],PARAMETER[\"false_northing\",100");
    add_srs_wkt (p, 11,
		 "00000],AUTHORITY[\"EPSG\",\"23886\"],AXIS[\"Easting\",EA");
    add_srs_wkt (p, 12, "ST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23887, "epsg", 23887, "ID74 / UTM zone 47S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=47 +south +a=6378160 +b=6356774.50408554");
    add_proj4text (p, 1, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ID74 / UTM zone 47S\",GEOGCS[\"ID74\",DATUM[\"I");
    add_srs_wkt (p, 1,
		 "ndonesian_Datum_1974\",SPHEROID[\"Indonesian National Sp");
    add_srs_wkt (p, 2,
		 "heroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4238\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",99],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",500000],PARAMETER[\"false_northing\",10000000],AUTHORIT");
    add_srs_wkt (p, 11,
		 "Y[\"EPSG\",\"23887\"],AXIS[\"Easting\",EAST],AXIS[\"Nort");
    add_srs_wkt (p, 12, "hing\",NORTH]]");
    p = add_epsg_def (first, last, 23888, "epsg", 23888, "ID74 / UTM zone 48S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=48 +south +a=6378160 +b=6356774.50408554");
    add_proj4text (p, 1, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ID74 / UTM zone 48S\",GEOGCS[\"ID74\",DATUM[\"I");
    add_srs_wkt (p, 1,
		 "ndonesian_Datum_1974\",SPHEROID[\"Indonesian National Sp");
    add_srs_wkt (p, 2,
		 "heroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4238\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",105],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",10000000],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"23888\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 12, "rthing\",NORTH]]");
    p = add_epsg_def (first, last, 23889, "epsg", 23889, "ID74 / UTM zone 49S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=49 +south +a=6378160 +b=6356774.50408554");
    add_proj4text (p, 1, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ID74 / UTM zone 49S\",GEOGCS[\"ID74\",DATUM[\"I");
    add_srs_wkt (p, 1,
		 "ndonesian_Datum_1974\",SPHEROID[\"Indonesian National Sp");
    add_srs_wkt (p, 2,
		 "heroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4238\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",111],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",10000000],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"23889\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 12, "rthing\",NORTH]]");
    p = add_epsg_def (first, last, 23890, "epsg", 23890, "ID74 / UTM zone 50S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=50 +south +a=6378160 +b=6356774.50408554");
    add_proj4text (p, 1, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ID74 / UTM zone 50S\",GEOGCS[\"ID74\",DATUM[\"I");
    add_srs_wkt (p, 1,
		 "ndonesian_Datum_1974\",SPHEROID[\"Indonesian National Sp");
    add_srs_wkt (p, 2,
		 "heroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4238\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",117],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",10000000],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"23890\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 12, "rthing\",NORTH]]");
    p = add_epsg_def (first, last, 23891, "epsg", 23891, "ID74 / UTM zone 51S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=51 +south +a=6378160 +b=6356774.50408554");
    add_proj4text (p, 1, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ID74 / UTM zone 51S\",GEOGCS[\"ID74\",DATUM[\"I");
    add_srs_wkt (p, 1,
		 "ndonesian_Datum_1974\",SPHEROID[\"Indonesian National Sp");
    add_srs_wkt (p, 2,
		 "heroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4238\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",123],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",10000000],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"23891\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 12, "rthing\",NORTH]]");
    p = add_epsg_def (first, last, 23892, "epsg", 23892, "ID74 / UTM zone 52S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=52 +south +a=6378160 +b=6356774.50408554");
    add_proj4text (p, 1, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ID74 / UTM zone 52S\",GEOGCS[\"ID74\",DATUM[\"I");
    add_srs_wkt (p, 1,
		 "ndonesian_Datum_1974\",SPHEROID[\"Indonesian National Sp");
    add_srs_wkt (p, 2,
		 "heroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4238\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",129],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",10000000],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"23892\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 12, "rthing\",NORTH]]");
    p = add_epsg_def (first, last, 23893, "epsg", 23893, "ID74 / UTM zone 53S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=53 +south +a=6378160 +b=6356774.50408554");
    add_proj4text (p, 1, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ID74 / UTM zone 53S\",GEOGCS[\"ID74\",DATUM[\"I");
    add_srs_wkt (p, 1,
		 "ndonesian_Datum_1974\",SPHEROID[\"Indonesian National Sp");
    add_srs_wkt (p, 2,
		 "heroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4238\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",135],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",10000000],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"23893\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 12, "rthing\",NORTH]]");
    p = add_epsg_def (first, last, 23894, "epsg", 23894, "ID74 / UTM zone 54S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=54 +south +a=6378160 +b=6356774.50408554");
    add_proj4text (p, 1, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ID74 / UTM zone 54S\",GEOGCS[\"ID74\",DATUM[\"I");
    add_srs_wkt (p, 1,
		 "ndonesian_Datum_1974\",SPHEROID[\"Indonesian National Sp");
    add_srs_wkt (p, 2,
		 "heroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4238\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",141],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",10000000],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"23894\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 12, "rthing\",NORTH]]");
    p = add_epsg_def (first, last, 23946, "epsg", 23946,
		      "Indian 1954 / UTM zone 46N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=46 +a=6377276.345 +b=6356075.41314024 +t");
    add_proj4text (p, 1, "owgs84=217,823,299,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Indian 1954 / UTM zone 46N\",GEOGCS[\"Indian 19");
    add_srs_wkt (p, 1,
		 "54\",DATUM[\"Indian_1954\",SPHEROID[\"Everest 1830 (1937");
    add_srs_wkt (p, 2,
		 " Adjustment)\",6377276.345,300.8017,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7015\"]],TOWGS84[217,823,299,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6239\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4239\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"central_meridian\",93],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 10,
		 "or\",0.9996],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23946\"],A");
    add_srs_wkt (p, 12, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23947, "epsg", 23947,
		      "Indian 1954 / UTM zone 47N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=47 +a=6377276.345 +b=6356075.41314024 +t");
    add_proj4text (p, 1, "owgs84=217,823,299,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Indian 1954 / UTM zone 47N\",GEOGCS[\"Indian 19");
    add_srs_wkt (p, 1,
		 "54\",DATUM[\"Indian_1954\",SPHEROID[\"Everest 1830 (1937");
    add_srs_wkt (p, 2,
		 " Adjustment)\",6377276.345,300.8017,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7015\"]],TOWGS84[217,823,299,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6239\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4239\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"central_meridian\",99],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 10,
		 "or\",0.9996],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23947\"],A");
    add_srs_wkt (p, 12, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 23948, "epsg", 23948,
		      "Indian 1954 / UTM zone 48N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=48 +a=6377276.345 +b=6356075.41314024 +t");
    add_proj4text (p, 1, "owgs84=217,823,299,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Indian 1954 / UTM zone 48N\",GEOGCS[\"Indian 19");
    add_srs_wkt (p, 1,
		 "54\",DATUM[\"Indian_1954\",SPHEROID[\"Everest 1830 (1937");
    add_srs_wkt (p, 2,
		 " Adjustment)\",6377276.345,300.8017,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7015\"]],TOWGS84[217,823,299,0,0,0,0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6239\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4239\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"central_meridian\",105],PARAMETER[\"scale_fac");
    add_srs_wkt (p, 10,
		 "tor\",0.9996],PARAMETER[\"false_easting\",500000],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"23948\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24047, "epsg", 24047,
		      "Indian 1975 / UTM zone 47N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=47 +a=6377276.345 +b=6356075.41314024 +u");
    add_proj4text (p, 1, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Indian 1975 / UTM zone 47N\",GEOGCS[\"Indian 19");
    add_srs_wkt (p, 1,
		 "75\",DATUM[\"Indian_1975\",SPHEROID[\"Everest 1830 (1937");
    add_srs_wkt (p, 2,
		 " Adjustment)\",6377276.345,300.8017,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7015\"]],AUTHORITY[\"EPSG\",\"6240\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4240\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",99],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"24047\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24048, "epsg", 24048,
		      "Indian 1975 / UTM zone 48N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=48 +a=6377276.345 +b=6356075.41314024 +u");
    add_proj4text (p, 1, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Indian 1975 / UTM zone 48N\",GEOGCS[\"Indian 19");
    add_srs_wkt (p, 1,
		 "75\",DATUM[\"Indian_1975\",SPHEROID[\"Everest 1830 (1937");
    add_srs_wkt (p, 2,
		 " Adjustment)\",6377276.345,300.8017,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7015\"]],AUTHORITY[\"EPSG\",\"6240\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4240\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",105],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_easting\",500000],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"24048\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24100, "epsg", 24100,
		      "Jamaica 1875 / Jamaica (Old Grid)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=18 +lat_0=18 +lon_0=-77 +k_0=1 +x_0=167");
    add_proj4text (p, 1,
		   "638.49597 +y_0=121918.90616 +a=6378249.144808011 +b=6356");
    add_proj4text (p, 2, "514.966204134 +to_meter=0.3047972654 +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Jamaica 1875 / Jamaica (Old Grid)\",GEOGCS[\"Ja");
    add_srs_wkt (p, 1,
		 "maica 1875\",DATUM[\"Jamaica_1875\",SPHEROID[\"Clarke 18");
    add_srs_wkt (p, 2,
		 "80\",6378249.144808011,293.4663076556349,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7034\"]],AUTHORITY[\"EPSG\",\"6241\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4241\"]],UNIT[\"Clarke's foot\",0.304797");
    add_srs_wkt (p, 7,
		 "2654,AUTHORITY[\"EPSG\",\"9005\"]],PROJECTION[\"Lambert_");
    add_srs_wkt (p, 8,
		 "Conformal_Conic_1SP\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 9,
		 "18],PARAMETER[\"central_meridian\",-77],PARAMETER[\"scal");
    add_srs_wkt (p, 10,
		 "e_factor\",1],PARAMETER[\"false_easting\",550000],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",400000],AUTHORITY[\"EPSG\",\"2410");
    add_srs_wkt (p, 12,
		 "0\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24200, "epsg", 24200,
		      "JAD69 / Jamaica National Grid");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=18 +lat_0=18 +lon_0=-77 +k_0=1 +x_0=250");
    add_proj4text (p, 1, "000 +y_0=150000 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"JAD69 / Jamaica National Grid\",GEOGCS[\"JAD69\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Jamaica_1969\",SPHEROID[\"Clarke 1866\",6378206");
    add_srs_wkt (p, 2,
		 ".4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6242\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"424");
    add_srs_wkt (p, 6,
		 "2\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Lambert_Conformal_Conic_1SP\"],PARAMETER[\"lat");
    add_srs_wkt (p, 8,
		 "itude_of_origin\",18],PARAMETER[\"central_meridian\",-77");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",250000],PARAMETER[\"false_northing\",150000],AUTHORI");
    add_srs_wkt (p, 11,
		 "TY[\"EPSG\",\"24200\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 12, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 24305, "epsg", 24305,
		      "Kalianpur 1937 / UTM zone 45N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=45 +a=6377276.345 +b=6356075.41314024 +u");
    add_proj4text (p, 1, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1937 / UTM zone 45N\",GEOGCS[\"Kalian");
    add_srs_wkt (p, 1,
		 "pur 1937\",DATUM[\"Kalianpur_1937\",SPHEROID[\"Everest 1");
    add_srs_wkt (p, 2,
		 "830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7015\"]],AUTHORITY[\"EPSG\",\"6144\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4144\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",87],PARAMETER[\"scale_factor\",0.9996],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"24305\"],AXIS[\"Easting\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24306, "epsg", 24306,
		      "Kalianpur 1937 / UTM zone 46N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=46 +a=6377276.345 +b=6356075.41314024 +u");
    add_proj4text (p, 1, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1937 / UTM zone 46N\",GEOGCS[\"Kalian");
    add_srs_wkt (p, 1,
		 "pur 1937\",DATUM[\"Kalianpur_1937\",SPHEROID[\"Everest 1");
    add_srs_wkt (p, 2,
		 "830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7015\"]],AUTHORITY[\"EPSG\",\"6144\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4144\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",93],PARAMETER[\"scale_factor\",0.9996],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"24306\"],AXIS[\"Easting\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24311, "epsg", 24311,
		      "Kalianpur 1962 / UTM zone 41N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=41 +a=6377301.243 +b=6356100.230165384 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1962 / UTM zone 41N\",GEOGCS[\"Kalian");
    add_srs_wkt (p, 1,
		 "pur 1962\",DATUM[\"Kalianpur_1962\",SPHEROID[\"Everest 1");
    add_srs_wkt (p, 2,
		 "830 (1962 Definition)\",6377301.243,300.8017255,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7044\"]],AUTHORITY[\"EPSG\",\"6145\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4145\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",63],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 11,
		 "hing\",0],AUTHORITY[\"EPSG\",\"24311\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 12, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24312, "epsg", 24312,
		      "Kalianpur 1962 / UTM zone 42N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=42 +a=6377301.243 +b=6356100.230165384 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1962 / UTM zone 42N\",GEOGCS[\"Kalian");
    add_srs_wkt (p, 1,
		 "pur 1962\",DATUM[\"Kalianpur_1962\",SPHEROID[\"Everest 1");
    add_srs_wkt (p, 2,
		 "830 (1962 Definition)\",6377301.243,300.8017255,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7044\"]],AUTHORITY[\"EPSG\",\"6145\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4145\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",69],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 11,
		 "hing\",0],AUTHORITY[\"EPSG\",\"24312\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 12, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24313, "epsg", 24313,
		      "Kalianpur 1962 / UTM zone 43N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=43 +a=6377301.243 +b=6356100.230165384 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1962 / UTM zone 43N\",GEOGCS[\"Kalian");
    add_srs_wkt (p, 1,
		 "pur 1962\",DATUM[\"Kalianpur_1962\",SPHEROID[\"Everest 1");
    add_srs_wkt (p, 2,
		 "830 (1962 Definition)\",6377301.243,300.8017255,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7044\"]],AUTHORITY[\"EPSG\",\"6145\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4145\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",75],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 11,
		 "hing\",0],AUTHORITY[\"EPSG\",\"24313\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 12, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24342, "epsg", 24342,
		      "Kalianpur 1975 / UTM zone 42N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=42 +a=6377299.151 +b=6356098.145120132 +");
    add_proj4text (p, 1, "towgs84=295,736,257,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1975 / UTM zone 42N\",GEOGCS[\"Kalian");
    add_srs_wkt (p, 1,
		 "pur 1975\",DATUM[\"Kalianpur_1975\",SPHEROID[\"Everest 1");
    add_srs_wkt (p, 2,
		 "830 (1975 Definition)\",6377299.151,300.8017255,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"414");
    add_srs_wkt (p, 7,
		 "6\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",0],PARAMETER[\"central_meridian\",69],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"scale_factor\",0.9996],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12,
		 ",\"24342\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NOR");
    add_srs_wkt (p, 13, "TH]]");
    p = add_epsg_def (first, last, 24343, "epsg", 24343,
		      "Kalianpur 1975 / UTM zone 43N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=43 +a=6377299.151 +b=6356098.145120132 +");
    add_proj4text (p, 1, "towgs84=295,736,257,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1975 / UTM zone 43N\",GEOGCS[\"Kalian");
    add_srs_wkt (p, 1,
		 "pur 1975\",DATUM[\"Kalianpur_1975\",SPHEROID[\"Everest 1");
    add_srs_wkt (p, 2,
		 "830 (1975 Definition)\",6377299.151,300.8017255,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"414");
    add_srs_wkt (p, 7,
		 "6\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",0],PARAMETER[\"central_meridian\",75],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"scale_factor\",0.9996],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12,
		 ",\"24343\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NOR");
    add_srs_wkt (p, 13, "TH]]");
    p = add_epsg_def (first, last, 24344, "epsg", 24344,
		      "Kalianpur 1975 / UTM zone 44N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=44 +a=6377299.151 +b=6356098.145120132 +");
    add_proj4text (p, 1, "towgs84=295,736,257,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1975 / UTM zone 44N\",GEOGCS[\"Kalian");
    add_srs_wkt (p, 1,
		 "pur 1975\",DATUM[\"Kalianpur_1975\",SPHEROID[\"Everest 1");
    add_srs_wkt (p, 2,
		 "830 (1975 Definition)\",6377299.151,300.8017255,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"414");
    add_srs_wkt (p, 7,
		 "6\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",0],PARAMETER[\"central_meridian\",81],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"scale_factor\",0.9996],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12,
		 ",\"24344\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NOR");
    add_srs_wkt (p, 13, "TH]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_20 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 24345, "epsg", 24345,
		      "Kalianpur 1975 / UTM zone 45N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=45 +a=6377299.151 +b=6356098.145120132 +");
    add_proj4text (p, 1, "towgs84=295,736,257,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1975 / UTM zone 45N\",GEOGCS[\"Kalian");
    add_srs_wkt (p, 1,
		 "pur 1975\",DATUM[\"Kalianpur_1975\",SPHEROID[\"Everest 1");
    add_srs_wkt (p, 2,
		 "830 (1975 Definition)\",6377299.151,300.8017255,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"414");
    add_srs_wkt (p, 7,
		 "6\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",0],PARAMETER[\"central_meridian\",87],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"scale_factor\",0.9996],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12,
		 ",\"24345\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NOR");
    add_srs_wkt (p, 13, "TH]]");
    p = add_epsg_def (first, last, 24346, "epsg", 24346,
		      "Kalianpur 1975 / UTM zone 46N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=46 +a=6377299.151 +b=6356098.145120132 +");
    add_proj4text (p, 1, "towgs84=295,736,257,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1975 / UTM zone 46N\",GEOGCS[\"Kalian");
    add_srs_wkt (p, 1,
		 "pur 1975\",DATUM[\"Kalianpur_1975\",SPHEROID[\"Everest 1");
    add_srs_wkt (p, 2,
		 "830 (1975 Definition)\",6377299.151,300.8017255,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"414");
    add_srs_wkt (p, 7,
		 "6\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",0],PARAMETER[\"central_meridian\",93],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"scale_factor\",0.9996],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12,
		 ",\"24346\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NOR");
    add_srs_wkt (p, 13, "TH]]");
    p = add_epsg_def (first, last, 24347, "epsg", 24347,
		      "Kalianpur 1975 / UTM zone 47N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=47 +a=6377299.151 +b=6356098.145120132 +");
    add_proj4text (p, 1, "towgs84=295,736,257,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1975 / UTM zone 47N\",GEOGCS[\"Kalian");
    add_srs_wkt (p, 1,
		 "pur 1975\",DATUM[\"Kalianpur_1975\",SPHEROID[\"Everest 1");
    add_srs_wkt (p, 2,
		 "830 (1975 Definition)\",6377299.151,300.8017255,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"414");
    add_srs_wkt (p, 7,
		 "6\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",0],PARAMETER[\"central_meridian\",99],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"scale_factor\",0.9996],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12,
		 ",\"24347\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NOR");
    add_srs_wkt (p, 13, "TH]]");
    p = add_epsg_def (first, last, 24370, "epsg", 24370,
		      "Kalianpur 1880 / India zone 0");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.5 +lat_0=39.5 +lon_0=68 +k_0=0.99846");
    add_proj4text (p, 1,
		   "154 +x_0=2153865.73916853 +y_0=2368292.194628102 +a=6377");
    add_proj4text (p, 2,
		   "299.36559538 +b=6356098.359005156 +to_meter=0.9143985307");
    add_proj4text (p, 3, "444408 +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1880 / India zone 0\",GEOGCS[\"Kalian");
    add_srs_wkt (p, 1,
		 "pur 1880\",DATUM[\"Kalianpur_1880\",SPHEROID[\"Everest (");
    add_srs_wkt (p, 2,
		 "1830 Definition)\",6377299.36559538,300.8017255433552,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7042\"]],AUTHORITY[\"EPSG\",\"6243\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4243\"]],UNIT[\"Indian yar");
    add_srs_wkt (p, 7,
		 "d\",0.9143985307444408,AUTHORITY[\"EPSG\",\"9084\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Lambert_Conformal_Conic_1SP\"],PARAMETER[\"lat");
    add_srs_wkt (p, 9,
		 "itude_of_origin\",39.5],PARAMETER[\"central_meridian\",6");
    add_srs_wkt (p, 10,
		 "8],PARAMETER[\"scale_factor\",0.99846154],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_easting\",2355500],PARAMETER[\"false_northing\",2590");
    add_srs_wkt (p, 12,
		 "000],AUTHORITY[\"EPSG\",\"24370\"],AXIS[\"Easting\",EAST");
    add_srs_wkt (p, 13, "],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24371, "epsg", 24371,
		      "Kalianpur 1880 / India zone I");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=32.5 +lat_0=32.5 +lon_0=68 +k_0=0.99878");
    add_proj4text (p, 1,
		   "641 +x_0=2743195.592233322 +y_0=914398.5307444407 +a=637");
    add_proj4text (p, 2,
		   "7299.36559538 +b=6356098.359005156 +to_meter=0.914398530");
    add_proj4text (p, 3, "7444408 +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1880 / India zone I\",GEOGCS[\"Kalian");
    add_srs_wkt (p, 1,
		 "pur 1880\",DATUM[\"Kalianpur_1880\",SPHEROID[\"Everest (");
    add_srs_wkt (p, 2,
		 "1830 Definition)\",6377299.36559538,300.8017255433552,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7042\"]],AUTHORITY[\"EPSG\",\"6243\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4243\"]],UNIT[\"Indian yar");
    add_srs_wkt (p, 7,
		 "d\",0.9143985307444408,AUTHORITY[\"EPSG\",\"9084\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Lambert_Conformal_Conic_1SP\"],PARAMETER[\"lat");
    add_srs_wkt (p, 9,
		 "itude_of_origin\",32.5],PARAMETER[\"central_meridian\",6");
    add_srs_wkt (p, 10,
		 "8],PARAMETER[\"scale_factor\",0.99878641],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_easting\",3000000],PARAMETER[\"false_northing\",1000");
    add_srs_wkt (p, 12,
		 "000],AUTHORITY[\"EPSG\",\"24371\"],AXIS[\"Easting\",EAST");
    add_srs_wkt (p, 13, "],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24372, "epsg", 24372,
		      "Kalianpur 1880 / India zone IIa");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=26 +lat_0=26 +lon_0=74 +k_0=0.99878641 ");
    add_proj4text (p, 1,
		   "+x_0=2743195.592233322 +y_0=914398.5307444407 +a=6377299");
    add_proj4text (p, 2,
		   ".36559538 +b=6356098.359005156 +to_meter=0.9143985307444");
    add_proj4text (p, 3, "408 +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1880 / India zone IIa\",GEOGCS[\"Kali");
    add_srs_wkt (p, 1,
		 "anpur 1880\",DATUM[\"Kalianpur_1880\",SPHEROID[\"Everest");
    add_srs_wkt (p, 2,
		 " (1830 Definition)\",6377299.36559538,300.8017255433552,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7042\"]],AUTHORITY[\"EPSG\",\"6243\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4243\"]],UNIT[\"Indian yar");
    add_srs_wkt (p, 7,
		 "d\",0.9143985307444408,AUTHORITY[\"EPSG\",\"9084\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Lambert_Conformal_Conic_1SP\"],PARAMETER[\"lat");
    add_srs_wkt (p, 9,
		 "itude_of_origin\",26],PARAMETER[\"central_meridian\",74]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"scale_factor\",0.99878641],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",3000000],PARAMETER[\"false_northing\",100000");
    add_srs_wkt (p, 12,
		 "0],AUTHORITY[\"EPSG\",\"24372\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 13, "AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24373, "epsg", 24373,
		      "Kalianpur 1880 / India zone III");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=19 +lat_0=19 +lon_0=80 +k_0=0.99878641 ");
    add_proj4text (p, 1,
		   "+x_0=2743195.592233322 +y_0=914398.5307444407 +a=6377299");
    add_proj4text (p, 2,
		   ".36559538 +b=6356098.359005156 +to_meter=0.9143985307444");
    add_proj4text (p, 3, "408 +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1880 / India zone III\",GEOGCS[\"Kali");
    add_srs_wkt (p, 1,
		 "anpur 1880\",DATUM[\"Kalianpur_1880\",SPHEROID[\"Everest");
    add_srs_wkt (p, 2,
		 " (1830 Definition)\",6377299.36559538,300.8017255433552,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7042\"]],AUTHORITY[\"EPSG\",\"6243\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4243\"]],UNIT[\"Indian yar");
    add_srs_wkt (p, 7,
		 "d\",0.9143985307444408,AUTHORITY[\"EPSG\",\"9084\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Lambert_Conformal_Conic_1SP\"],PARAMETER[\"lat");
    add_srs_wkt (p, 9,
		 "itude_of_origin\",19],PARAMETER[\"central_meridian\",80]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"scale_factor\",0.99878641],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",3000000],PARAMETER[\"false_northing\",100000");
    add_srs_wkt (p, 12,
		 "0],AUTHORITY[\"EPSG\",\"24373\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 13, "AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24374, "epsg", 24374,
		      "Kalianpur 1880 / India zone IV");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=12 +lat_0=12 +lon_0=80 +k_0=0.99878641 ");
    add_proj4text (p, 1,
		   "+x_0=2743195.592233322 +y_0=914398.5307444407 +a=6377299");
    add_proj4text (p, 2,
		   ".36559538 +b=6356098.359005156 +to_meter=0.9143985307444");
    add_proj4text (p, 3, "408 +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1880 / India zone IV\",GEOGCS[\"Kalia");
    add_srs_wkt (p, 1,
		 "npur 1880\",DATUM[\"Kalianpur_1880\",SPHEROID[\"Everest ");
    add_srs_wkt (p, 2,
		 "(1830 Definition)\",6377299.36559538,300.8017255433552,A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"7042\"]],AUTHORITY[\"EPSG\",\"6243\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4243\"]],UNIT[\"Indian yar");
    add_srs_wkt (p, 7,
		 "d\",0.9143985307444408,AUTHORITY[\"EPSG\",\"9084\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Lambert_Conformal_Conic_1SP\"],PARAMETER[\"lat");
    add_srs_wkt (p, 9,
		 "itude_of_origin\",12],PARAMETER[\"central_meridian\",80]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"scale_factor\",0.99878641],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",3000000],PARAMETER[\"false_northing\",100000");
    add_srs_wkt (p, 12,
		 "0],AUTHORITY[\"EPSG\",\"24374\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 13, "AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24375, "epsg", 24375,
		      "Kalianpur 1937 / India zone IIb");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=26 +lat_0=26 +lon_0=90 +k_0=0.99878641 ");
    add_proj4text (p, 1,
		   "+x_0=2743185.69 +y_0=914395.23 +a=6377276.345 +b=6356075");
    add_proj4text (p, 2, ".41314024 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1937 / India zone IIb\",GEOGCS[\"Kali");
    add_srs_wkt (p, 1,
		 "anpur 1937\",DATUM[\"Kalianpur_1937\",SPHEROID[\"Everest");
    add_srs_wkt (p, 2,
		 " 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7015\"]],AUTHORITY[\"EPSG\",\"6144\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4144\"]],UNIT[\"metre\",1,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Co");
    add_srs_wkt (p, 8,
		 "nic_1SP\"],PARAMETER[\"latitude_of_origin\",26],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",90],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 10,
		 "99878641],PARAMETER[\"false_easting\",2743185.69],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",914395.23],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 12,
		 "4375\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 24376, "epsg", 24376,
		      "Kalianpur 1962 / India zone I");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=32.5 +lat_0=32.5 +lon_0=68 +k_0=0.99878");
    add_proj4text (p, 1,
		   "641 +x_0=2743196.4 +y_0=914398.8 +a=6377301.243 +b=63561");
    add_proj4text (p, 2, "00.230165384 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1962 / India zone I\",GEOGCS[\"Kalian");
    add_srs_wkt (p, 1,
		 "pur 1962\",DATUM[\"Kalianpur_1962\",SPHEROID[\"Everest 1");
    add_srs_wkt (p, 2,
		 "830 (1962 Definition)\",6377301.243,300.8017255,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7044\"]],AUTHORITY[\"EPSG\",\"6145\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4145\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_C");
    add_srs_wkt (p, 8,
		 "onic_1SP\"],PARAMETER[\"latitude_of_origin\",32.5],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"central_meridian\",68],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.99878641],PARAMETER[\"false_easting\",2743196.4],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",914398.8],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12,
		 "24376\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 13, "]");
    p = add_epsg_def (first, last, 24377, "epsg", 24377,
		      "Kalianpur 1962 / India zone IIa");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=26 +lat_0=26 +lon_0=74 +k_0=0.99878641 ");
    add_proj4text (p, 1,
		   "+x_0=2743196.4 +y_0=914398.8 +a=6377301.243 +b=6356100.2");
    add_proj4text (p, 2, "30165384 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1962 / India zone IIa\",GEOGCS[\"Kali");
    add_srs_wkt (p, 1,
		 "anpur 1962\",DATUM[\"Kalianpur_1962\",SPHEROID[\"Everest");
    add_srs_wkt (p, 2,
		 " 1830 (1962 Definition)\",6377301.243,300.8017255,AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"7044\"]],AUTHORITY[\"EPSG\",\"6145\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6,
		 "\"]],AUTHORITY[\"EPSG\",\"4145\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_1SP\"],PARAMETER[\"latitude_of_origin\",26],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"central_meridian\",74],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.99878641],PARAMETER[\"false_easting\",2743196.4],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",914398.8],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12,
		 "24377\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 13, "]");
    p = add_epsg_def (first, last, 24378, "epsg", 24378,
		      "Kalianpur 1975 / India zone I");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=32.5 +lat_0=32.5 +lon_0=68 +k_0=0.99878");
    add_proj4text (p, 1,
		   "641 +x_0=2743195.5 +y_0=914398.5 +a=6377299.151 +b=63560");
    add_proj4text (p, 2,
		   "98.145120132 +towgs84=295,736,257,0,0,0,0 +units=m +no_d");
    add_proj4text (p, 3, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1975 / India zone I\",GEOGCS[\"Kalian");
    add_srs_wkt (p, 1,
		 "pur 1975\",DATUM[\"Kalianpur_1975\",SPHEROID[\"Everest 1");
    add_srs_wkt (p, 2,
		 "830 (1975 Definition)\",6377299.151,300.8017255,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"414");
    add_srs_wkt (p, 7,
		 "6\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Lambert_Conformal_Conic_1SP\"],PARAMETER[\"lat");
    add_srs_wkt (p, 9,
		 "itude_of_origin\",32.5],PARAMETER[\"central_meridian\",6");
    add_srs_wkt (p, 10,
		 "8],PARAMETER[\"scale_factor\",0.99878641],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_easting\",2743195.5],PARAMETER[\"false_northing\",91");
    add_srs_wkt (p, 12,
		 "4398.5],AUTHORITY[\"EPSG\",\"24378\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 13, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24379, "epsg", 24379,
		      "Kalianpur 1975 / India zone IIa");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=26 +lat_0=26 +lon_0=74 +k_0=0.99878641 ");
    add_proj4text (p, 1,
		   "+x_0=2743195.5 +y_0=914398.5 +a=6377299.151 +b=6356098.1");
    add_proj4text (p, 2,
		   "45120132 +towgs84=295,736,257,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1975 / India zone IIa\",GEOGCS[\"Kali");
    add_srs_wkt (p, 1,
		 "anpur 1975\",DATUM[\"Kalianpur_1975\",SPHEROID[\"Everest");
    add_srs_wkt (p, 2,
		 " 1830 (1975 Definition)\",6377299.151,300.8017255,AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "146\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Lambert_Conformal_Conic_1SP\"],PARAMETER[\"l");
    add_srs_wkt (p, 9,
		 "atitude_of_origin\",26],PARAMETER[\"central_meridian\",7");
    add_srs_wkt (p, 10,
		 "4],PARAMETER[\"scale_factor\",0.99878641],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_easting\",2743195.5],PARAMETER[\"false_northing\",91");
    add_srs_wkt (p, 12,
		 "4398.5],AUTHORITY[\"EPSG\",\"24379\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 13, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24380, "epsg", 24380,
		      "Kalianpur 1975 / India zone IIb");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=26 +lat_0=26 +lon_0=90 +k_0=0.99878641 ");
    add_proj4text (p, 1,
		   "+x_0=2743195.5 +y_0=914398.5 +a=6377299.151 +b=6356098.1");
    add_proj4text (p, 2,
		   "45120132 +towgs84=295,736,257,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1975 / India zone IIb\",GEOGCS[\"Kali");
    add_srs_wkt (p, 1,
		 "anpur 1975\",DATUM[\"Kalianpur_1975\",SPHEROID[\"Everest");
    add_srs_wkt (p, 2,
		 " 1830 (1975 Definition)\",6377299.151,300.8017255,AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "146\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Lambert_Conformal_Conic_1SP\"],PARAMETER[\"l");
    add_srs_wkt (p, 9,
		 "atitude_of_origin\",26],PARAMETER[\"central_meridian\",9");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"scale_factor\",0.99878641],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_easting\",2743195.5],PARAMETER[\"false_northing\",91");
    add_srs_wkt (p, 12,
		 "4398.5],AUTHORITY[\"EPSG\",\"24380\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 13, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24381, "epsg", 24381,
		      "Kalianpur 1975 / India zone III");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=19 +lat_0=19 +lon_0=80 +k_0=0.99878641 ");
    add_proj4text (p, 1,
		   "+x_0=2743195.5 +y_0=914398.5 +a=6377299.151 +b=6356098.1");
    add_proj4text (p, 2,
		   "45120132 +towgs84=295,736,257,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1975 / India zone III\",GEOGCS[\"Kali");
    add_srs_wkt (p, 1,
		 "anpur 1975\",DATUM[\"Kalianpur_1975\",SPHEROID[\"Everest");
    add_srs_wkt (p, 2,
		 " 1830 (1975 Definition)\",6377299.151,300.8017255,AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "146\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Lambert_Conformal_Conic_1SP\"],PARAMETER[\"l");
    add_srs_wkt (p, 9,
		 "atitude_of_origin\",19],PARAMETER[\"central_meridian\",8");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"scale_factor\",0.99878641],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_easting\",2743195.5],PARAMETER[\"false_northing\",91");
    add_srs_wkt (p, 12,
		 "4398.5],AUTHORITY[\"EPSG\",\"24381\"],AXIS[\"Easting\",E");
    add_srs_wkt (p, 13, "AST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24382, "epsg", 24382,
		      "Kalianpur 1880 / India zone IIb");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=26 +lat_0=26 +lon_0=90 +k_0=0.99878641 ");
    add_proj4text (p, 1,
		   "+x_0=2743195.592233322 +y_0=914398.5307444407 +a=6377299");
    add_proj4text (p, 2,
		   ".36559538 +b=6356098.359005156 +to_meter=0.9143985307444");
    add_proj4text (p, 3, "408 +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1880 / India zone IIb\",GEOGCS[\"Kali");
    add_srs_wkt (p, 1,
		 "anpur 1880\",DATUM[\"Kalianpur_1880\",SPHEROID[\"Everest");
    add_srs_wkt (p, 2,
		 " (1830 Definition)\",6377299.36559538,300.8017255433552,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7042\"]],AUTHORITY[\"EPSG\",\"6243\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4243\"]],UNIT[\"Indian yar");
    add_srs_wkt (p, 7,
		 "d\",0.9143985307444408,AUTHORITY[\"EPSG\",\"9084\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Lambert_Conformal_Conic_1SP\"],PARAMETER[\"lat");
    add_srs_wkt (p, 9,
		 "itude_of_origin\",26],PARAMETER[\"central_meridian\",90]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"scale_factor\",0.99878641],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",3000000],PARAMETER[\"false_northing\",100000");
    add_srs_wkt (p, 12,
		 "0],AUTHORITY[\"EPSG\",\"24382\"],AXIS[\"Easting\",EAST],");
    add_srs_wkt (p, 13, "AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24383, "epsg", 24383,
		      "Kalianpur 1975 / India zone IV");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=12 +lat_0=12 +lon_0=80 +k_0=0.99878641 ");
    add_proj4text (p, 1,
		   "+x_0=2743195.5 +y_0=914398.5 +a=6377299.151 +b=6356098.1");
    add_proj4text (p, 2,
		   "45120132 +towgs84=295,736,257,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kalianpur 1975 / India zone IV\",GEOGCS[\"Kalia");
    add_srs_wkt (p, 1,
		 "npur 1975\",DATUM[\"Kalianpur_1975\",SPHEROID[\"Everest ");
    add_srs_wkt (p, 2,
		 "1830 (1975 Definition)\",6377299.151,300.8017255,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"41");
    add_srs_wkt (p, 7,
		 "46\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Lambert_Conformal_Conic_1SP\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",12],PARAMETER[\"central_meridian\",80");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.99878641],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_easting\",2743195.5],PARAMETER[\"false_northing\",914");
    add_srs_wkt (p, 12,
		 "398.5],AUTHORITY[\"EPSG\",\"24383\"],AXIS[\"Easting\",EA");
    add_srs_wkt (p, 13, "ST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24500, "epsg", 24500,
		      "Kertau 1968 / Singapore Grid");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=1.287646666666667 +lon_0=103.853002222");
    add_proj4text (p, 1,
		   "2222 +x_0=30000 +y_0=30000 +a=6377304.063 +b=6356103.038");
    add_proj4text (p, 2, "993155 +towgs84=-11,851,5,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kertau 1968 / Singapore Grid\",GEOGCS[\"Kertau ");
    add_srs_wkt (p, 1,
		 "1968\",DATUM[\"Kertau_1968\",SPHEROID[\"Everest 1830 Mod");
    add_srs_wkt (p, 2,
		 "ified\",6377304.063,300.8017,AUTHORITY[\"EPSG\",\"7018\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[-11,851,5,0,0,0,0],AUTHORITY[\"EPSG\",\"6245\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4245\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Cassini_Soldn");
    add_srs_wkt (p, 8,
		 "er\"],PARAMETER[\"latitude_of_origin\",1.287646666666667");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",103.8530022222222],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_easting\",30000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 11,
		 "g\",30000],AUTHORITY[\"EPSG\",\"24500\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 12, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24547, "epsg", 24547,
		      "Kertau 1968 / UTM zone 47N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=47 +a=6377304.063 +b=6356103.038993155 +");
    add_proj4text (p, 1, "towgs84=-11,851,5,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kertau 1968 / UTM zone 47N\",GEOGCS[\"Kertau 19");
    add_srs_wkt (p, 1,
		 "68\",DATUM[\"Kertau_1968\",SPHEROID[\"Everest 1830 Modif");
    add_srs_wkt (p, 2,
		 "ied\",6377304.063,300.8017,AUTHORITY[\"EPSG\",\"7018\"]]");
    add_srs_wkt (p, 3,
		 ",TOWGS84[-11,851,5,0,0,0,0],AUTHORITY[\"EPSG\",\"6245\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4245\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",99],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 10,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"24547\"],AXIS[\"Eas");
    add_srs_wkt (p, 12, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24548, "epsg", 24548,
		      "Kertau 1968 / UTM zone 48N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=48 +a=6377304.063 +b=6356103.038993155 +");
    add_proj4text (p, 1, "towgs84=-11,851,5,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kertau 1968 / UTM zone 48N\",GEOGCS[\"Kertau 19");
    add_srs_wkt (p, 1,
		 "68\",DATUM[\"Kertau_1968\",SPHEROID[\"Everest 1830 Modif");
    add_srs_wkt (p, 2,
		 "ied\",6377304.063,300.8017,AUTHORITY[\"EPSG\",\"7018\"]]");
    add_srs_wkt (p, 3,
		 ",TOWGS84[-11,851,5,0,0,0,0],AUTHORITY[\"EPSG\",\"6245\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4245\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",105],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"24548\"],AXIS[\"Ea");
    add_srs_wkt (p, 12, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24571, "epsg", 24571,
		      "Kertau / R.S.O. Malaya (ch) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=omerc +lat_0=4 +lonc=102.25 +alpha=323.0257905 +k=");
    add_proj4text (p, 1,
		   "0.99984 +x_0=804671.2997750348 +y_0=0 +a=6377304.063 +b=");
    add_proj4text (p, 2,
		   "6356103.038993155 +towgs84=-11,851,5,0,0,0,0 +to_meter=2");
    add_proj4text (p, 3, "0.11678249437587 +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Kertau / R.S.O. Malaya (ch) (deprecated)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"Kertau 1968\",DATUM[\"Kertau_1968\",SPHEROID[\"Ever");
    add_srs_wkt (p, 2,
		 "est 1830 Modified\",6377304.063,300.8017,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7018\"]],TOWGS84[-11,851,5,0,0,0,0],AUTHORITY[\"EP");
    add_srs_wkt (p, 4,
		 "SG\",\"6245\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4245\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"British chain (Benoit 1895 B)\",20.11678249437587,AUT");
    add_srs_wkt (p, 8,
		 "HORITY[\"EPSG\",\"9062\"]],PROJECTION[\"Hotine_Oblique_M");
    add_srs_wkt (p, 9,
		 "ercator\"],PARAMETER[\"latitude_of_center\",4],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"longitude_of_center\",102.25],PARAMETER[\"azimuth\",3");
    add_srs_wkt (p, 11,
		 "23.0257905],PARAMETER[\"rectified_grid_angle\",323.13010");
    add_srs_wkt (p, 12,
		 "23611111],PARAMETER[\"scale_factor\",0.99984],PARAMETER[");
    add_srs_wkt (p, 13,
		 "\"false_easting\",40000],PARAMETER[\"false_northing\",0]");
    add_srs_wkt (p, 14,
		 ",AUTHORITY[\"EPSG\",\"24571\"],AXIS[\"Easting\",EAST],AX");
    add_srs_wkt (p, 15, "IS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24600, "epsg", 24600, "KOC Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=32.5 +lat_0=32.5 +lon_0=45 +k_0=0.99878");
    add_proj4text (p, 1,
		   "64078000001 +x_0=1500000 +y_0=1166200 +ellps=clrk80 +tow");
    add_proj4text (p, 2, "gs84=-294.7,-200.1,525.5,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"KOC Lambert\",GEOGCS[\"KOC\",DATUM[\"Kuwait_Oil");
    add_srs_wkt (p, 1,
		 "_Company\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,29");
    add_srs_wkt (p, 2,
		 "3.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-294.7,-200.");
    add_srs_wkt (p, 3,
		 "1,525.5,0,0,0,0],AUTHORITY[\"EPSG\",\"6246\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4246\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"latitude_of_origin\",32.5],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",45],PARAMETER[\"scale_factor\",0.9987");
    add_srs_wkt (p, 10,
		 "864078],PARAMETER[\"false_easting\",1500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",1166200],AUTHORITY[\"EPSG\",\"24600\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24718, "epsg", 24718,
		      "La Canoa / UTM zone 18N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +ellps=intl +towgs84=-273.5,110.6,-35");
    add_proj4text (p, 1, "7.9,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"La Canoa / UTM zone 18N\",GEOGCS[\"La Canoa\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"La_Canoa\",SPHEROID[\"International 1924\",637838");
    add_srs_wkt (p, 2,
		 "8,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-273.5,110.6");
    add_srs_wkt (p, 3,
		 ",-357.9,0,0,0,0],AUTHORITY[\"EPSG\",\"6247\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4247\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-75],PARAMETER[\"scale_factor\",0.9996],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"24718\"],AXIS[\"Easting\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24719, "epsg", 24719,
		      "La Canoa / UTM zone 19N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +ellps=intl +towgs84=-273.5,110.6,-35");
    add_proj4text (p, 1, "7.9,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"La Canoa / UTM zone 19N\",GEOGCS[\"La Canoa\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"La_Canoa\",SPHEROID[\"International 1924\",637838");
    add_srs_wkt (p, 2,
		 "8,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-273.5,110.6");
    add_srs_wkt (p, 3,
		 ",-357.9,0,0,0,0],AUTHORITY[\"EPSG\",\"6247\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4247\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-69],PARAMETER[\"scale_factor\",0.9996],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"24719\"],AXIS[\"Easting\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24720, "epsg", 24720,
		      "La Canoa / UTM zone 20N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +ellps=intl +towgs84=-273.5,110.6,-35");
    add_proj4text (p, 1, "7.9,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"La Canoa / UTM zone 20N\",GEOGCS[\"La Canoa\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"La_Canoa\",SPHEROID[\"International 1924\",637838");
    add_srs_wkt (p, 2,
		 "8,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-273.5,110.6");
    add_srs_wkt (p, 3,
		 ",-357.9,0,0,0,0],AUTHORITY[\"EPSG\",\"6247\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4247\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-63],PARAMETER[\"scale_factor\",0.9996],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"24720\"],AXIS[\"Easting\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24817, "epsg", 24817,
		      "PSAD56 / UTM zone 17N");
    add_proj4text (p, 0, "+proj=utm +zone=17 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PSAD56 / UTM zone 17N\",GEOGCS[\"PSAD56\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Provisional_South_American_Datum_1956\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4248\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",-8");
    add_srs_wkt (p, 9,
		 "1],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"24817\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 12, "rthing\",NORTH]]");
    p = add_epsg_def (first, last, 24818, "epsg", 24818,
		      "PSAD56 / UTM zone 18N");
    add_proj4text (p, 0, "+proj=utm +zone=18 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PSAD56 / UTM zone 18N\",GEOGCS[\"PSAD56\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Provisional_South_American_Datum_1956\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4248\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",-7");
    add_srs_wkt (p, 9,
		 "5],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"24818\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 12, "rthing\",NORTH]]");
    p = add_epsg_def (first, last, 24819, "epsg", 24819,
		      "PSAD56 / UTM zone 19N");
    add_proj4text (p, 0, "+proj=utm +zone=19 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PSAD56 / UTM zone 19N\",GEOGCS[\"PSAD56\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Provisional_South_American_Datum_1956\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4248\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",-6");
    add_srs_wkt (p, 9,
		 "9],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"24819\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 12, "rthing\",NORTH]]");
    p = add_epsg_def (first, last, 24820, "epsg", 24820,
		      "PSAD56 / UTM zone 20N");
    add_proj4text (p, 0, "+proj=utm +zone=20 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PSAD56 / UTM zone 20N\",GEOGCS[\"PSAD56\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Provisional_South_American_Datum_1956\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4248\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",-6");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"24820\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 12, "rthing\",NORTH]]");
    p = add_epsg_def (first, last, 24821, "epsg", 24821,
		      "PSAD56 / UTM zone 21N");
    add_proj4text (p, 0, "+proj=utm +zone=21 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PSAD56 / UTM zone 21N\",GEOGCS[\"PSAD56\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Provisional_South_American_Datum_1956\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4248\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",-5");
    add_srs_wkt (p, 9,
		 "7],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"24821\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 12, "rthing\",NORTH]]");
    p = add_epsg_def (first, last, 24877, "epsg", 24877,
		      "PSAD56 / UTM zone 17S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=17 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PSAD56 / UTM zone 17S\",GEOGCS[\"PSAD56\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Provisional_South_American_Datum_1956\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4248\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",-8");
    add_srs_wkt (p, 9,
		 "1],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",10000000]");
    add_srs_wkt (p, 11,
		 ",AUTHORITY[\"EPSG\",\"24877\"],AXIS[\"Easting\",EAST],AX");
    add_srs_wkt (p, 12, "IS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24878, "epsg", 24878,
		      "PSAD56 / UTM zone 18S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PSAD56 / UTM zone 18S\",GEOGCS[\"PSAD56\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Provisional_South_American_Datum_1956\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4248\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",-7");
    add_srs_wkt (p, 9,
		 "5],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",10000000]");
    add_srs_wkt (p, 11,
		 ",AUTHORITY[\"EPSG\",\"24878\"],AXIS[\"Easting\",EAST],AX");
    add_srs_wkt (p, 12, "IS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24879, "epsg", 24879,
		      "PSAD56 / UTM zone 19S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PSAD56 / UTM zone 19S\",GEOGCS[\"PSAD56\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Provisional_South_American_Datum_1956\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4248\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",-6");
    add_srs_wkt (p, 9,
		 "9],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",10000000]");
    add_srs_wkt (p, 11,
		 ",AUTHORITY[\"EPSG\",\"24879\"],AXIS[\"Easting\",EAST],AX");
    add_srs_wkt (p, 12, "IS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24880, "epsg", 24880,
		      "PSAD56 / UTM zone 20S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PSAD56 / UTM zone 20S\",GEOGCS[\"PSAD56\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Provisional_South_American_Datum_1956\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4248\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",-6");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",10000000]");
    add_srs_wkt (p, 11,
		 ",AUTHORITY[\"EPSG\",\"24880\"],AXIS[\"Easting\",EAST],AX");
    add_srs_wkt (p, 12, "IS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24881, "epsg", 24881,
		      "PSAD56 / UTM zone 21S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PSAD56 / UTM zone 21S\",GEOGCS[\"PSAD56\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Provisional_South_American_Datum_1956\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4248\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",-5");
    add_srs_wkt (p, 9,
		 "7],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",10000000]");
    add_srs_wkt (p, 11,
		 ",AUTHORITY[\"EPSG\",\"24881\"],AXIS[\"Easting\",EAST],AX");
    add_srs_wkt (p, 12, "IS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24882, "epsg", 24882,
		      "PSAD56 / UTM zone 22S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=22 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PSAD56 / UTM zone 22S\",GEOGCS[\"PSAD56\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Provisional_South_American_Datum_1956\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4248\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",-5");
    add_srs_wkt (p, 9,
		 "1],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",500000],PARAMETER[\"false_northing\",10000000]");
    add_srs_wkt (p, 11,
		 ",AUTHORITY[\"EPSG\",\"24882\"],AXIS[\"Easting\",EAST],AX");
    add_srs_wkt (p, 12, "IS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 24891, "epsg", 24891,
		      "PSAD56 / Peru west zone");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-6 +lon_0=-80.5 +k=0.99983008 +x_0=22");
    add_proj4text (p, 1, "2000 +y_0=1426834.743 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PSAD56 / Peru west zone\",GEOGCS[\"PSAD56\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Provisional_South_American_Datum_1956\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "22\"]],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4248\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",-6],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 9,
		 "-80.5],PARAMETER[\"scale_factor\",0.99983008],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",222000],PARAMETER[\"false_northing\",1");
    add_srs_wkt (p, 11,
		 "426834.743],AUTHORITY[\"EPSG\",\"24891\"],AXIS[\"X\",EAS");
    add_srs_wkt (p, 12, "T],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 24892, "epsg", 24892,
		      "PSAD56 / Peru central zone");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-9.5 +lon_0=-76 +k=0.99932994 +x_0=72");
    add_proj4text (p, 1, "0000 +y_0=1039979.159 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PSAD56 / Peru central zone\",GEOGCS[\"PSAD56\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"Provisional_South_American_Datum_1956\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7022\"]],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4248\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",-9.5],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-76],PARAMETER[\"scale_factor\",0.99932994],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",720000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",1039979.159],AUTHORITY[\"EPSG\",\"24892\"],AXIS[\"X\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 24893, "epsg", 24893,
		      "PSAD56 / Peru east zone");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-9.5 +lon_0=-70.5 +k=0.99952992 +x_0=");
    add_proj4text (p, 1,
		   "1324000 +y_0=1040084.558 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"PSAD56 / Peru east zone\",GEOGCS[\"PSAD56\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Provisional_South_American_Datum_1956\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "22\"]],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4248\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",-9.5],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",-70.5],PARAMETER[\"scale_factor\",0.99952992],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",1324000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",1040084.558],AUTHORITY[\"EPSG\",\"24893\"],AXIS[\"X\",E");
    add_srs_wkt (p, 12, "AST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 25000, "epsg", 25000,
		      "Leigon / Ghana Metre Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=4.666666666666667 +lon_0=-1 +k=0.9997");
    add_proj4text (p, 1,
		   "5 +x_0=274319.51 +y_0=0 +ellps=clrk80 +towgs84=-130,29,3");
    add_proj4text (p, 2, "64,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Leigon / Ghana Metre Grid\",GEOGCS[\"Leigon\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Leigon\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.1");
    add_srs_wkt (p, 2,
		 "45,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-130,29");
    add_srs_wkt (p, 3,
		 ",364,0,0,0,0],AUTHORITY[\"EPSG\",\"6250\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4250\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",4.666666666666667],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-1],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 10,
		 "99975],PARAMETER[\"false_easting\",274319.51],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_northing\",0],AUTHORITY[\"EPSG\",\"25000\"],AXIS");
    add_srs_wkt (p, 12, "[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 25231, "epsg", 25231, "Lome / UTM zone 31N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=31 +a=6378249.2 +b=6356515 +units=m +no_");
    add_proj4text (p, 1, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Lome / UTM zone 31N\",GEOGCS[\"Lome\",DATUM[\"L");
    add_srs_wkt (p, 1,
		 "ome\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.46602");
    add_srs_wkt (p, 2,
		 "12936265,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6252\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4252\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",3],PARAMETER[\"scale_facto");
    add_srs_wkt (p, 9,
		 "r\",0.9996],PARAMETER[\"false_easting\",500000],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"25231\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 25391, "epsg", 25391,
		      "Luzon 1911 / Philippines zone I");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=0.99995 +x_0=500000 +");
    add_proj4text (p, 1, "y_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Luzon 1911 / Philippines zone I\",GEOGCS[\"Luzo");
    add_srs_wkt (p, 1,
		 "n 1911\",DATUM[\"Luzon_1911\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6253\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4253\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",117],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.99995],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"25391\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 25392, "epsg", 25392,
		      "Luzon 1911 / Philippines zone II");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=119 +k=0.99995 +x_0=500000 +");
    add_proj4text (p, 1, "y_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Luzon 1911 / Philippines zone II\",GEOGCS[\"Luz");
    add_srs_wkt (p, 1,
		 "on 1911\",DATUM[\"Luzon_1911\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6253\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4253\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",119],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.99995],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"25392\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 25393, "epsg", 25393,
		      "Luzon 1911 / Philippines zone III");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=121 +k=0.99995 +x_0=500000 +");
    add_proj4text (p, 1, "y_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Luzon 1911 / Philippines zone III\",GEOGCS[\"Lu");
    add_srs_wkt (p, 1,
		 "zon 1911\",DATUM[\"Luzon_1911\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6253\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4253\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"lat");
    add_srs_wkt (p, 8,
		 "itude_of_origin\",0],PARAMETER[\"central_meridian\",121]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"scale_factor\",0.99995],PARAMETER[\"false_e");
    add_srs_wkt (p, 10,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 11,
		 "TY[\"EPSG\",\"25393\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH");
    add_srs_wkt (p, 12, "]]");
    p = add_epsg_def (first, last, 25394, "epsg", 25394,
		      "Luzon 1911 / Philippines zone IV");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=0.99995 +x_0=500000 +");
    add_proj4text (p, 1, "y_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Luzon 1911 / Philippines zone IV\",GEOGCS[\"Luz");
    add_srs_wkt (p, 1,
		 "on 1911\",DATUM[\"Luzon_1911\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6253\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4253\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",123],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.99995],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"25394\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 25395, "epsg", 25395,
		      "Luzon 1911 / Philippines zone V");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=125 +k=0.99995 +x_0=500000 +");
    add_proj4text (p, 1, "y_0=0 +ellps=clrk66 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Luzon 1911 / Philippines zone V\",GEOGCS[\"Luzo");
    add_srs_wkt (p, 1,
		 "n 1911\",DATUM[\"Luzon_1911\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6253\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4253\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",125],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.99995],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"25395\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 25700, "epsg", 25700,
		      "Makassar (Jakarta) / NEIEZ (deprecated)");
    add_proj4text (p, 0,
		   "+proj=merc +lon_0=110 +k=0.997 +x_0=3900000 +y_0=900000 ");
    add_proj4text (p, 1,
		   "+ellps=bessel +towgs84=-587.8,519.75,145.76,0,0,0,0 +pm=");
    add_proj4text (p, 2, "jakarta +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Makassar (Jakarta) / NEIEZ (deprecated)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"Makassar (Jakarta)\",DATUM[\"Makassar_Jakarta\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7004\"]],TOWGS84[-587.8,519.75,145.76,0,0,0,0],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6804\"]],PRIMEM[\"Jakarta\",106.807");
    add_srs_wkt (p, 5,
		 "7194444444,AUTHORITY[\"EPSG\",\"8908\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 6,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"4804\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 8,
		 "SG\",\"9001\"]],PROJECTION[\"Mercator_1SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",110],PARAMETER[\"scale_factor\",0.997");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"false_easting\",3900000],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_northing\",900000],AUTHORITY[\"EPSG\",\"25700\"],AXIS[\"");
    add_srs_wkt (p, 12, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 25828, "epsg", 25828,
		      "ETRS89 / UTM zone 28N");
    add_proj4text (p, 0, "+proj=utm +zone=28 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / UTM zone 28N\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",-15],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_easting\",500000],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"25828\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 25829, "epsg", 25829,
		      "ETRS89 / UTM zone 29N");
    add_proj4text (p, 0, "+proj=utm +zone=29 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / UTM zone 29N\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",-9],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"25829\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 25830, "epsg", 25830,
		      "ETRS89 / UTM zone 30N");
    add_proj4text (p, 0, "+proj=utm +zone=30 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / UTM zone 30N\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",-3],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"25830\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 25831, "epsg", 25831,
		      "ETRS89 / UTM zone 31N");
    add_proj4text (p, 0, "+proj=utm +zone=31 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / UTM zone 31N\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",3],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 11,
		 "RITY[\"EPSG\",\"25831\"],AXIS[\"Easting\",EAST],AXIS[\"N");
    add_srs_wkt (p, 12, "orthing\",NORTH]]");
    p = add_epsg_def (first, last, 25832, "epsg", 25832,
		      "ETRS89 / UTM zone 32N");
    add_proj4text (p, 0, "+proj=utm +zone=32 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / UTM zone 32N\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",9],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 11,
		 "RITY[\"EPSG\",\"25832\"],AXIS[\"Easting\",EAST],AXIS[\"N");
    add_srs_wkt (p, 12, "orthing\",NORTH]]");
    p = add_epsg_def (first, last, 25833, "epsg", 25833,
		      "ETRS89 / UTM zone 33N");
    add_proj4text (p, 0, "+proj=utm +zone=33 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / UTM zone 33N\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",15],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"25833\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 25834, "epsg", 25834,
		      "ETRS89 / UTM zone 34N");
    add_proj4text (p, 0, "+proj=utm +zone=34 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / UTM zone 34N\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",21],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"25834\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 25835, "epsg", 25835,
		      "ETRS89 / UTM zone 35N");
    add_proj4text (p, 0, "+proj=utm +zone=35 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / UTM zone 35N\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",27],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"25835\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 25836, "epsg", 25836,
		      "ETRS89 / UTM zone 36N");
    add_proj4text (p, 0, "+proj=utm +zone=36 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / UTM zone 36N\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",33],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"25836\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 25837, "epsg", 25837,
		      "ETRS89 / UTM zone 37N");
    add_proj4text (p, 0, "+proj=utm +zone=37 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / UTM zone 37N\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",39],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"25837\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 25838, "epsg", 25838,
		      "ETRS89 / UTM zone 38N");
    add_proj4text (p, 0, "+proj=utm +zone=38 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / UTM zone 38N\",GEOGCS[\"ETRS89\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"European_Terrestrial_Reference_System_1989\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",45],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"25838\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 25884, "epsg", 25884,
		      "ETRS89 / TM Baltic93");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=24 +k=0.9996 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"ETRS89 / TM Baltic93\",GEOGCS[\"ETRS89\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"European_Terrestrial_Reference_System_1989\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4258\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 9,
		 ",24],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"25884\"],AXIS[\"X\",NORTH],AXIS[\"Y\",E");
    add_srs_wkt (p, 12, "AST]]");
    p = add_epsg_def (first, last, 25932, "epsg", 25932,
		      "Malongo 1987 / UTM zone 32S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=32 +south +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Malongo 1987 / UTM zone 32S\",GEOGCS[\"Malongo ");
    add_srs_wkt (p, 1,
		 "1987\",DATUM[\"Malongo_1987\",SPHEROID[\"International 1");
    add_srs_wkt (p, 2,
		 "924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6259\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4259\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",9],PARAMETER[\"s");
    add_srs_wkt (p, 9,
		 "cale_factor\",0.9996],PARAMETER[\"false_easting\",500000");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"false_northing\",10000000],AUTHORITY[\"EPS");
    add_srs_wkt (p, 11,
		 "G\",\"25932\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",");
    add_srs_wkt (p, 12, "NORTH]]");
    p = add_epsg_def (first, last, 26191, "epsg", 26191,
		      "Merchich / Nord Maroc");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.3 +lat_0=33.3 +lon_0=-5.4 +k_0=0.999");
    add_proj4text (p, 1,
		   "625769 +x_0=500000 +y_0=300000 +a=6378249.2 +b=6356515 +");
    add_proj4text (p, 2, "towgs84=31,146,47,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Merchich / Nord Maroc\",GEOGCS[\"Merchich\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Merchich\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2");
    add_srs_wkt (p, 2,
		 ",293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84");
    add_srs_wkt (p, 3,
		 "[31,146,47,0,0,0,0],AUTHORITY[\"EPSG\",\"6261\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4261\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Con");
    add_srs_wkt (p, 8,
		 "ic_1SP\"],PARAMETER[\"latitude_of_origin\",33.3],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-5.4],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.999625769],PARAMETER[\"false_easting\",500000],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",300000],AUTHORITY[\"EPSG\",\"2619");
    add_srs_wkt (p, 12, "1\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26192, "epsg", 26192,
		      "Merchich / Sud Maroc");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=29.7 +lat_0=29.7 +lon_0=-5.4 +k_0=0.999");
    add_proj4text (p, 1,
		   "6155960000001 +x_0=500000 +y_0=300000 +a=6378249.2 +b=63");
    add_proj4text (p, 2, "56515 +towgs84=31,146,47,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Merchich / Sud Maroc\",GEOGCS[\"Merchich\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"Merchich\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,");
    add_srs_wkt (p, 2,
		 "293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[");
    add_srs_wkt (p, 3,
		 "31,146,47,0,0,0,0],AUTHORITY[\"EPSG\",\"6261\"]],PRIMEM[");
    add_srs_wkt (p, 4,
		 "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg");
    add_srs_wkt (p, 5,
		 "ree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"4261\"]],UNIT[\"metre\",1,AUTHORITY");
    add_srs_wkt (p, 7,
		 "[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Coni");
    add_srs_wkt (p, 8,
		 "c_1SP\"],PARAMETER[\"latitude_of_origin\",29.7],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-5.4],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.999615596],PARAMETER[\"false_easting\",500000],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_northing\",300000],AUTHORITY[\"EPSG\",\"26192");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26193, "epsg", 26193,
		      "Merchich / Sahara (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=26.1 +lat_0=26.1 +lon_0=-5.4 +k_0=0.999");
    add_proj4text (p, 1,
		   "6 +x_0=1200000 +y_0=400000 +a=6378249.2 +b=6356515 +towg");
    add_proj4text (p, 2, "s84=31,146,47,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Merchich / Sahara (deprecated)\",GEOGCS[\"Merch");
    add_srs_wkt (p, 1,
		 "ich\",DATUM[\"Merchich\",SPHEROID[\"Clarke 1880 (IGN)\",");
    add_srs_wkt (p, 2,
		 "6378249.2,293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"]");
    add_srs_wkt (p, 3,
		 "],TOWGS84[31,146,47,0,0,0,0],AUTHORITY[\"EPSG\",\"6261\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4261\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Confo");
    add_srs_wkt (p, 8,
		 "rmal_Conic_1SP\"],PARAMETER[\"latitude_of_origin\",26.1]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"central_meridian\",-5.4],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.9996],PARAMETER[\"false_easting\",1200000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",400000],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "26193\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26194, "epsg", 26194,
		      "Merchich / Sahara Nord");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=26.1 +lat_0=26.1 +lon_0=-5.4 +k_0=0.999");
    add_proj4text (p, 1,
		   "616304 +x_0=1200000 +y_0=400000 +a=6378249.2 +b=6356515 ");
    add_proj4text (p, 2, "+towgs84=31,146,47,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Merchich / Sahara Nord\",GEOGCS[\"Merchich\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"Merchich\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.");
    add_srs_wkt (p, 2,
		 "2,293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS8");
    add_srs_wkt (p, 3,
		 "4[31,146,47,0,0,0,0],AUTHORITY[\"EPSG\",\"6261\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4261\"]],UNIT[\"metre\",1,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Co");
    add_srs_wkt (p, 8,
		 "nic_1SP\"],PARAMETER[\"latitude_of_origin\",26.1],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"central_meridian\",-5.4],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.999616304],PARAMETER[\"false_easting\",1200000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",400000],AUTHORITY[\"EPSG\",\"261");
    add_srs_wkt (p, 12, "94\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26195, "epsg", 26195,
		      "Merchich / Sahara Sud");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=22.5 +lat_0=22.5 +lon_0=-5.4 +k_0=0.999");
    add_proj4text (p, 1,
		   "616437 +x_0=1500000 +y_0=400000 +a=6378249.2 +b=6356515 ");
    add_proj4text (p, 2, "+towgs84=31,146,47,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Merchich / Sahara Sud\",GEOGCS[\"Merchich\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Merchich\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2");
    add_srs_wkt (p, 2,
		 ",293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84");
    add_srs_wkt (p, 3,
		 "[31,146,47,0,0,0,0],AUTHORITY[\"EPSG\",\"6261\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4261\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Con");
    add_srs_wkt (p, 8,
		 "ic_1SP\"],PARAMETER[\"latitude_of_origin\",22.5],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-5.4],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.999616437],PARAMETER[\"false_easting\",1500000],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_northing\",400000],AUTHORITY[\"EPSG\",\"261");
    add_srs_wkt (p, 12, "95\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26237, "epsg", 26237,
		      "Massawa / UTM zone 37N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=37 +ellps=bessel +towgs84=639,405,60,0,0");
    add_proj4text (p, 1, ",0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Massawa / UTM zone 37N\",GEOGCS[\"Massawa\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Massawa\",SPHEROID[\"Bessel 1841\",6377397.155,299.");
    add_srs_wkt (p, 2,
		 "1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[639,405,60");
    add_srs_wkt (p, 3,
		 ",0,0,0,0],AUTHORITY[\"EPSG\",\"6262\"]],PRIMEM[\"Greenwi");
    add_srs_wkt (p, 4,
		 "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0");
    add_srs_wkt (p, 5,
		 "1745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"4262\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETE");
    add_srs_wkt (p, 8,
		 "R[\"latitude_of_origin\",0],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 9,
		 "\",39],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_easting\",500000],PARAMETER[\"false_northing\",0],AU");
    add_srs_wkt (p, 11,
		 "THORITY[\"EPSG\",\"26237\"],AXIS[\"Easting\",EAST],AXIS[");
    add_srs_wkt (p, 12, "\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 26331, "epsg", 26331,
		      "Minna / UTM zone 31N");
    add_proj4text (p, 0, "+proj=utm +zone=31 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Minna / UTM zone 31N\",GEOGCS[\"Minna\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Minna\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.4");
    add_srs_wkt (p, 2,
		 "65,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"62");
    add_srs_wkt (p, 3,
		 "63\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4263\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",3],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"26331\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 26332, "epsg", 26332,
		      "Minna / UTM zone 32N");
    add_proj4text (p, 0, "+proj=utm +zone=32 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Minna / UTM zone 32N\",GEOGCS[\"Minna\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Minna\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.4");
    add_srs_wkt (p, 2,
		 "65,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"62");
    add_srs_wkt (p, 3,
		 "63\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4263\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 6,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 7,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"central_meridian\",9],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"26332\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 26391, "epsg", 26391,
		      "Minna / Nigeria West Belt");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=4 +lon_0=4.5 +k=0.99975 +x_0=230738.2");
    add_proj4text (p, 1, "6 +y_0=0 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Minna / Nigeria West Belt\",GEOGCS[\"Minna\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"Minna\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145");
    add_srs_wkt (p, 2,
		 ",293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6263\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4263\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",4],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",4.5],PARAMETER[\"scale_fac");
    add_srs_wkt (p, 9,
		 "tor\",0.99975],PARAMETER[\"false_easting\",230738.26],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"26391");
    add_srs_wkt (p, 11, "\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 26392, "epsg", 26392,
		      "Minna / Nigeria Mid Belt");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=4 +lon_0=8.5 +k=0.99975 +x_0=670553.9");
    add_proj4text (p, 1, "8 +y_0=0 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Minna / Nigeria Mid Belt\",GEOGCS[\"Minna\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Minna\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,");
    add_srs_wkt (p, 2,
		 "293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6263\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4263\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",4],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",8.5],PARAMETER[\"scale_fac");
    add_srs_wkt (p, 9,
		 "tor\",0.99975],PARAMETER[\"false_easting\",670553.98],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"26392");
    add_srs_wkt (p, 11, "\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 26393, "epsg", 26393,
		      "Minna / Nigeria East Belt");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=4 +lon_0=12.5 +k=0.99975 +x_0=1110369");
    add_proj4text (p, 1, ".7 +y_0=0 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Minna / Nigeria East Belt\",GEOGCS[\"Minna\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"Minna\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145");
    add_srs_wkt (p, 2,
		 ",293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6263\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4263\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",4],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",12.5],PARAMETER[\"scale_fa");
    add_srs_wkt (p, 9,
		 "ctor\",0.99975],PARAMETER[\"false_easting\",1110369.7],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2639");
    add_srs_wkt (p, 11,
		 "3\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 26432, "epsg", 26432,
		      "Mhast / UTM zone 32S (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=32 +south +ellps=intl +towgs84=-252.95,-");
    add_proj4text (p, 1, "4.11,-96.38,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Mhast / UTM zone 32S (deprecated)\",GEOGCS[\"Mh");
    add_srs_wkt (p, 1,
		 "ast\",DATUM[\"Mhast\",SPHEROID[\"International 1924\",63");
    add_srs_wkt (p, 2,
		 "78388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-252.95,");
    add_srs_wkt (p, 3,
		 "-4.11,-96.38,0,0,0,0],AUTHORITY[\"EPSG\",\"6264\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4264\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",9],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 11,
		 "ing\",10000000],AUTHORITY[\"EPSG\",\"26432\"],AXIS[\"Eas");
    add_srs_wkt (p, 12, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 26591, "epsg", 26591,
		      "Monte Mario (Rome) / Italy zone 1 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-3.45233333333333 +k=0.9996 ");
    add_proj4text (p, 1,
		   "+x_0=1500000 +y_0=0 +ellps=intl +pm=rome +units=m +no_de");
    add_proj4text (p, 2, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Monte Mario (Rome) / Italy zone 1 (deprecated)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"Monte Mario (Rome)\",DATUM[\"Monte_Mario_Rome\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6806\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Rome\",12.45233333333333,AUTHORITY[\"EPSG\",\"8906\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4806\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-3.45233333333333],PARAMETER[\"scal");
    add_srs_wkt (p, 10,
		 "e_factor\",0.9996],PARAMETER[\"false_easting\",1500000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"265");
    add_srs_wkt (p, 12, "91\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26592, "epsg", 26592,
		      "Monte Mario (Rome) / Italy zone 2 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=2.54766666666666 +k=0.9996 +");
    add_proj4text (p, 1,
		   "x_0=2520000 +y_0=0 +ellps=intl +pm=rome +units=m +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Monte Mario (Rome) / Italy zone 2 (deprecated)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"Monte Mario (Rome)\",DATUM[\"Monte_Mario_Rome\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6806\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Rome\",12.45233333333333,AUTHORITY[\"EPSG\",\"8906\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4806\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 8,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",2.54766666666666],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.9996],PARAMETER[\"false_easting\",2520000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2659");
    add_srs_wkt (p, 12, "2\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26632, "epsg", 26632,
		      "M'poraloko / UTM zone 32N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=32 +a=6378249.2 +b=6356515 +units=m +no_");
    add_proj4text (p, 1, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"M'poraloko / UTM zone 32N\",GEOGCS[\"M'poraloko");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"M_poraloko\",SPHEROID[\"Clarke 1880 (IGN)\",6");
    add_srs_wkt (p, 2,
		 "378249.2,293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6266\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4266\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",9],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"26632\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 26692, "epsg", 26692,
		      "M'poraloko / UTM zone 32S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=32 +south +a=6378249.2 +b=6356515 +units");
    add_proj4text (p, 1, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"M'poraloko / UTM zone 32S\",GEOGCS[\"M'poraloko");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"M_poraloko\",SPHEROID[\"Clarke 1880 (IGN)\",6");
    add_srs_wkt (p, 2,
		 "378249.2,293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6266\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4266\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",9],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",10000000],AUTHO");
    add_srs_wkt (p, 11,
		 "RITY[\"EPSG\",\"26692\"],AXIS[\"Easting\",EAST],AXIS[\"N");
    add_srs_wkt (p, 12, "orthing\",NORTH]]");
    p = add_epsg_def (first, last, 26701, "epsg", 26701, "NAD27 / UTM zone 1N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=1 +ellps=clrk66 +datum=NAD27 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 1N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-177],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"26701\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 26702, "epsg", 26702, "NAD27 / UTM zone 2N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=2 +ellps=clrk66 +datum=NAD27 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 2N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-171],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"26702\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 26703, "epsg", 26703, "NAD27 / UTM zone 3N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=3 +ellps=clrk66 +datum=NAD27 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 3N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-165],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"26703\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 26704, "epsg", 26704, "NAD27 / UTM zone 4N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=4 +ellps=clrk66 +datum=NAD27 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 4N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-159],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"26704\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 26705, "epsg", 26705, "NAD27 / UTM zone 5N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=5 +ellps=clrk66 +datum=NAD27 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 5N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-153],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"26705\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 26706, "epsg", 26706, "NAD27 / UTM zone 6N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=6 +ellps=clrk66 +datum=NAD27 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 6N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-147],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"26706\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 26707, "epsg", 26707, "NAD27 / UTM zone 7N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=7 +ellps=clrk66 +datum=NAD27 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 7N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-141],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"26707\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 26708, "epsg", 26708, "NAD27 / UTM zone 8N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=8 +ellps=clrk66 +datum=NAD27 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 8N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-135],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"26708\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 26709, "epsg", 26709, "NAD27 / UTM zone 9N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=9 +ellps=clrk66 +datum=NAD27 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 9N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-129],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"26709\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 26710, "epsg", 26710,
		      "NAD27 / UTM zone 10N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=10 +ellps=clrk66 +datum=NAD27 +units=m +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 10N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-123],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"26710\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 26711, "epsg", 26711,
		      "NAD27 / UTM zone 11N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=11 +ellps=clrk66 +datum=NAD27 +units=m +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 11N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-117],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"26711\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 26712, "epsg", 26712,
		      "NAD27 / UTM zone 12N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=12 +ellps=clrk66 +datum=NAD27 +units=m +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 12N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-111],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"26712\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 26713, "epsg", 26713,
		      "NAD27 / UTM zone 13N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=13 +ellps=clrk66 +datum=NAD27 +units=m +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 13N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-105],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9996],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"26713\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 26714, "epsg", 26714,
		      "NAD27 / UTM zone 14N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=14 +ellps=clrk66 +datum=NAD27 +units=m +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 14N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-99],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 11,
		 "PSG\",\"26714\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 26715, "epsg", 26715,
		      "NAD27 / UTM zone 15N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=15 +ellps=clrk66 +datum=NAD27 +units=m +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 15N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-93],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 11,
		 "PSG\",\"26715\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 26716, "epsg", 26716,
		      "NAD27 / UTM zone 16N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=16 +ellps=clrk66 +datum=NAD27 +units=m +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 16N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-87],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 11,
		 "PSG\",\"26716\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 26717, "epsg", 26717,
		      "NAD27 / UTM zone 17N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=17 +ellps=clrk66 +datum=NAD27 +units=m +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 17N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-81],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 11,
		 "PSG\",\"26717\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 26718, "epsg", 26718,
		      "NAD27 / UTM zone 18N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +ellps=clrk66 +datum=NAD27 +units=m +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 18N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-75],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 11,
		 "PSG\",\"26718\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 26719, "epsg", 26719,
		      "NAD27 / UTM zone 19N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +ellps=clrk66 +datum=NAD27 +units=m +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 19N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-69],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 11,
		 "PSG\",\"26719\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 26720, "epsg", 26720,
		      "NAD27 / UTM zone 20N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +ellps=clrk66 +datum=NAD27 +units=m +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 20N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-63],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 11,
		 "PSG\",\"26720\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 26721, "epsg", 26721,
		      "NAD27 / UTM zone 21N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +ellps=clrk66 +datum=NAD27 +units=m +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 21N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-57],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 11,
		 "PSG\",\"26721\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 26722, "epsg", 26722,
		      "NAD27 / UTM zone 22N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=22 +ellps=clrk66 +datum=NAD27 +units=m +");
    add_proj4text (p, 1, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / UTM zone 22N\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-51],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9996],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 11,
		 "PSG\",\"26722\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 26729, "epsg", 26729,
		      "NAD27 / Alabama East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30.5 +lon_0=-85.83333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "9996 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +datum=");
    add_proj4text (p, 2, "NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Alabama East\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",30.5],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-85.83333333333333],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.99996],PARAMETER[\"false_easting\",500000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2672");
    add_srs_wkt (p, 12, "9\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26730, "epsg", 26730,
		      "NAD27 / Alabama West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=-87.5 +k=0.999933333 +x_0=1");
    add_proj4text (p, 1,
		   "52400.3048006096 +y_0=0 +ellps=clrk66 +datum=NAD27 +unit");
    add_proj4text (p, 2, "s=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Alabama West\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",30],PARAMETER[\"c");
    add_srs_wkt (p, 9,
		 "entral_meridian\",-87.5],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 10,
		 "9933333],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"26730\"],AXIS[\"");
    add_srs_wkt (p, 12, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26731, "epsg", 26731,
		      "NAD27 / Alaska zone 1");
    add_proj4text (p, 0,
		   "+proj=omerc +lat_0=57 +lonc=-133.6666666666667 +alpha=32");
    add_proj4text (p, 1,
		   "3.1301023611111 +k=0.9999 +x_0=5000000.001016002 +y_0=-5");
    add_proj4text (p, 2,
		   "000000.001016002 +ellps=clrk66 +datum=NAD27 +units=us-ft");
    add_proj4text (p, 3, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Alaska zone 1\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Hotine_Oblique");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_center\",57],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"longitude_of_center\",-133.6666666666667],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"azimuth\",323.1301023611111],PARAMETER[\"rectified_g");
    add_srs_wkt (p, 11,
		 "rid_angle\",323.1301023611111],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 12,
		 ",0.9999],PARAMETER[\"false_easting\",16404166.67],PARAME");
    add_srs_wkt (p, 13,
		 "TER[\"false_northing\",-16404166.67],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 14, "\"26731\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26732, "epsg", 26732,
		      "NAD27 / Alaska zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-142 +k=0.9999 +x_0=152400.");
    add_proj4text (p, 1,
		   "3048006096 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-f");
    add_proj4text (p, 2, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Alaska zone 2\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",54],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-142],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 10,
		 "9999],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"26732\"],AXIS[\"X");
    add_srs_wkt (p, 12, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26733, "epsg", 26733,
		      "NAD27 / Alaska zone 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-146 +k=0.9999 +x_0=152400.");
    add_proj4text (p, 1,
		   "3048006096 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-f");
    add_proj4text (p, 2, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Alaska zone 3\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",54],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-146],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 10,
		 "9999],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"26733\"],AXIS[\"X");
    add_srs_wkt (p, 12, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26734, "epsg", 26734,
		      "NAD27 / Alaska zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-150 +k=0.9999 +x_0=152400.");
    add_proj4text (p, 1,
		   "3048006096 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-f");
    add_proj4text (p, 2, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Alaska zone 4\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",54],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-150],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 10,
		 "9999],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"26734\"],AXIS[\"X");
    add_srs_wkt (p, 12, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26735, "epsg", 26735,
		      "NAD27 / Alaska zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-154 +k=0.9999 +x_0=152400.");
    add_proj4text (p, 1,
		   "3048006096 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-f");
    add_proj4text (p, 2, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Alaska zone 5\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",54],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-154],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 10,
		 "9999],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"26735\"],AXIS[\"X");
    add_srs_wkt (p, 12, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26736, "epsg", 26736,
		      "NAD27 / Alaska zone 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-158 +k=0.9999 +x_0=152400.");
    add_proj4text (p, 1,
		   "3048006096 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-f");
    add_proj4text (p, 2, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Alaska zone 6\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",54],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-158],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 10,
		 "9999],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"26736\"],AXIS[\"X");
    add_srs_wkt (p, 12, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26737, "epsg", 26737,
		      "NAD27 / Alaska zone 7");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-162 +k=0.9999 +x_0=213360.");
    add_proj4text (p, 1,
		   "4267208534 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-f");
    add_proj4text (p, 2, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Alaska zone 7\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",54],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-162],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 10,
		 "9999],PARAMETER[\"false_easting\",700000],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"26737\"],AXIS[\"X");
    add_srs_wkt (p, 12, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26738, "epsg", 26738,
		      "NAD27 / Alaska zone 8");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-166 +k=0.9999 +x_0=152400.");
    add_proj4text (p, 1,
		   "3048006096 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-f");
    add_proj4text (p, 2, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Alaska zone 8\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",54],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-166],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 10,
		 "9999],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"26738\"],AXIS[\"X");
    add_srs_wkt (p, 12, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26739, "epsg", 26739,
		      "NAD27 / Alaska zone 9");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-170 +k=0.9999 +x_0=182880.");
    add_proj4text (p, 1,
		   "3657607315 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-f");
    add_proj4text (p, 2, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Alaska zone 9\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",54],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-170],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 10,
		 "9999],PARAMETER[\"false_easting\",600000],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"26739\"],AXIS[\"X");
    add_srs_wkt (p, 12, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26740, "epsg", 26740,
		      "NAD27 / Alaska zone 10");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=53.83333333333334 +lat_2=51.83333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=51 +lon_0=-176 +x_0=914401.8288036576 +y_0=0 ");
    add_proj4text (p, 2, "+ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Alaska zone 10\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",53.83");
    add_srs_wkt (p, 9,
		 "333333333334],PARAMETER[\"standard_parallel_2\",51.83333");
    add_srs_wkt (p, 10,
		 "333333334],PARAMETER[\"latitude_of_origin\",51],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"central_meridian\",-176],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",3000000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 13, "SG\",\"26740\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26741, "epsg", 26741,
		      "NAD27 / California zone I");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.66666666666666 +lat_2=40 +lat_0=39.3");
    add_proj4text (p, 1,
		   "3333333333334 +lon_0=-122 +x_0=609601.2192024384 +y_0=0 ");
    add_proj4text (p, 2, "+ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / California zone I\",GEOGCS[\"NAD27\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866");
    add_srs_wkt (p, 2,
		 "\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4267\"]],UNIT[\"US survey foot\",0.30480060960121");
    add_srs_wkt (p, 7,
		 "92,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Co");
    add_srs_wkt (p, 8,
		 "nformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",4");
    add_srs_wkt (p, 9,
		 "1.66666666666666],PARAMETER[\"standard_parallel_2\",40],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",39.33333333333334],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-122],PARAMETER[\"false_easti");
    add_srs_wkt (p, 12,
		 "ng\",2000000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 13,
		 "\"EPSG\",\"26741\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26742, "epsg", 26742,
		      "NAD27 / California zone II");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.83333333333334 +lat_2=38.33333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=37.66666666666666 +lon_0=-122 +x_0=609601.219");
    add_proj4text (p, 2,
		   "2024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +");
    add_proj4text (p, 3, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / California zone II\",GEOGCS[\"NAD27\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 186");
    add_srs_wkt (p, 2,
		 "6\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 3,
		 "8\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "39.83333333333334],PARAMETER[\"standard_parallel_2\",38.");
    add_srs_wkt (p, 10,
		 "33333333333334],PARAMETER[\"latitude_of_origin\",37.6666");
    add_srs_wkt (p, 11,
		 "6666666666],PARAMETER[\"central_meridian\",-122],PARAMET");
    add_srs_wkt (p, 12,
		 "ER[\"false_easting\",2000000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 13,
		 "\",0],AUTHORITY[\"EPSG\",\"26742\"],AXIS[\"X\",EAST],AXI");
    add_srs_wkt (p, 14, "S[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26743, "epsg", 26743,
		      "NAD27 / California zone III");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.43333333333333 +lat_2=37.06666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.5 +lon_0=-120.5 +x_0=609601.2192024384 +y_");
    add_proj4text (p, 2,
		   "0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / California zone III\",GEOGCS[\"NAD27\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 18");
    add_srs_wkt (p, 2,
		 "66\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "08\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "38.43333333333333],PARAMETER[\"standard_parallel_2\",37.");
    add_srs_wkt (p, 10,
		 "06666666666667],PARAMETER[\"latitude_of_origin\",36.5],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"central_meridian\",-120.5],PARAMETER[\"false_");
    add_srs_wkt (p, 12,
		 "easting\",2000000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 13,
		 "RITY[\"EPSG\",\"26743\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 14, "TH]]");
    p = add_epsg_def (first, last, 26744, "epsg", 26744,
		      "NAD27 / California zone IV");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.25 +lat_2=36 +lat_0=35.3333333333333");
    add_proj4text (p, 1,
		   "4 +lon_0=-119 +x_0=609601.2192024384 +y_0=0 +ellps=clrk6");
    add_proj4text (p, 2, "6 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / California zone IV\",GEOGCS[\"NAD27\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 186");
    add_srs_wkt (p, 2,
		 "6\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 3,
		 "8\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "37.25],PARAMETER[\"standard_parallel_2\",36],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "latitude_of_origin\",35.33333333333334],PARAMETER[\"cent");
    add_srs_wkt (p, 11,
		 "ral_meridian\",-119],PARAMETER[\"false_easting\",2000000");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 13, "6744\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26745, "epsg", 26745,
		      "NAD27 / California zone V");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=35.46666666666667 +lat_2=34.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=33.5 +lon_0=-118 +x_0=609601.2192024384 +y_0=");
    add_proj4text (p, 2, "0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / California zone V\",GEOGCS[\"NAD27\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866");
    add_srs_wkt (p, 2,
		 "\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4267\"]],UNIT[\"US survey foot\",0.30480060960121");
    add_srs_wkt (p, 7,
		 "92,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Co");
    add_srs_wkt (p, 8,
		 "nformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",3");
    add_srs_wkt (p, 9,
		 "5.46666666666667],PARAMETER[\"standard_parallel_2\",34.0");
    add_srs_wkt (p, 10,
		 "3333333333333],PARAMETER[\"latitude_of_origin\",33.5],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"central_meridian\",-118],PARAMETER[\"false_eas");
    add_srs_wkt (p, 12,
		 "ting\",2000000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 13,
		 "Y[\"EPSG\",\"26745\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]");
    add_srs_wkt (p, 14, "]");
    p = add_epsg_def (first, last, 26746, "epsg", 26746,
		      "NAD27 / California zone VI");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.88333333333333 +lat_2=32.78333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=32.16666666666666 +lon_0=-116.25 +x_0=609601.");
    add_proj4text (p, 2,
		   "2192024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-f");
    add_proj4text (p, 3, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / California zone VI\",GEOGCS[\"NAD27\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 186");
    add_srs_wkt (p, 2,
		 "6\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 3,
		 "8\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "33.88333333333333],PARAMETER[\"standard_parallel_2\",32.");
    add_srs_wkt (p, 10,
		 "78333333333333],PARAMETER[\"latitude_of_origin\",32.1666");
    add_srs_wkt (p, 11,
		 "6666666666],PARAMETER[\"central_meridian\",-116.25],PARA");
    add_srs_wkt (p, 12,
		 "METER[\"false_easting\",2000000],PARAMETER[\"false_north");
    add_srs_wkt (p, 13,
		 "ing\",0],AUTHORITY[\"EPSG\",\"26746\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 14, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26747, "epsg", 26747,
		      "NAD27 / California zone VII (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=34.41666666666666 +lat_2=33.86666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=34.13333333333333 +lon_0=-118.3333333333333 +");
    add_proj4text (p, 2,
		   "x_0=1276106.450596901 +y_0=127079.524511049 +ellps=clrk6");
    add_proj4text (p, 3, "6 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / California zone VII (deprecated)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NAD27\",DATUM[\"North_American_Datum_1927\",SPHEROI");
    add_srs_wkt (p, 2,
		 "D[\"Clarke 1866\",6378206.4,294.9786982139006,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM");
    add_srs_wkt (p, 4,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 5,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"4267\"]],UNIT[\"US survey foot\",0");
    add_srs_wkt (p, 7,
		 ".3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTI");
    add_srs_wkt (p, 8,
		 "ON[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard");
    add_srs_wkt (p, 9,
		 "_parallel_1\",34.41666666666666],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 10,
		 "rallel_2\",33.86666666666667],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 11,
		 "igin\",34.13333333333333],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 12,
		 ",-118.3333333333333],PARAMETER[\"false_easting\",4186692");
    add_srs_wkt (p, 13,
		 ".58],PARAMETER[\"false_northing\",416926.74],AUTHORITY[\"");
    add_srs_wkt (p, 14,
		 "EPSG\",\"26747\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26748, "epsg", 26748,
		      "NAD27 / Arizona East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-110.1666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +datum=NAD");
    add_proj4text (p, 2, "27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Arizona East\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",31],PARAMETER[\"c");
    add_srs_wkt (p, 9,
		 "entral_meridian\",-110.1666666666667],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.9999],PARAMETER[\"false_easting\",500000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"26748\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26749, "epsg", 26749,
		      "NAD27 / Arizona Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-111.9166666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +datum=NAD");
    add_proj4text (p, 2, "27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Arizona Central\",GEOGCS[\"NAD27\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_M");
    add_srs_wkt (p, 8,
		 "ercator\"],PARAMETER[\"latitude_of_origin\",31],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-111.9166666666667],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",0.9999],PARAMETER[\"false_easting\",500000");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 12, "6749\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26750, "epsg", 26750,
		      "NAD27 / Arizona West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-113.75 +k=0.999933333 +x_0");
    add_proj4text (p, 1,
		   "=152400.3048006096 +y_0=0 +ellps=clrk66 +datum=NAD27 +un");
    add_proj4text (p, 2, "its=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Arizona West\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",31],PARAMETER[\"c");
    add_srs_wkt (p, 9,
		 "entral_meridian\",-113.75],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 10,
		 "999933333],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_northing\",0],AUTHORITY[\"EPSG\",\"26750\"],AXI");
    add_srs_wkt (p, 12, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26751, "epsg", 26751,
		      "NAD27 / Arkansas North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.23333333333333 +lat_2=34.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=34.33333333333334 +lon_0=-92 +x_0=609601.2192");
    add_proj4text (p, 2,
		   "024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +n");
    add_proj4text (p, 3, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Arkansas North\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",36.23");
    add_srs_wkt (p, 9,
		 "333333333333],PARAMETER[\"standard_parallel_2\",34.93333");
    add_srs_wkt (p, 10,
		 "333333333],PARAMETER[\"latitude_of_origin\",34.333333333");
    add_srs_wkt (p, 11,
		 "33334],PARAMETER[\"central_meridian\",-92],PARAMETER[\"f");
    add_srs_wkt (p, 12,
		 "alse_easting\",2000000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 13,
		 "AUTHORITY[\"EPSG\",\"26751\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 14, ",NORTH]]");
    p = add_epsg_def (first, last, 26752, "epsg", 26752,
		      "NAD27 / Arkansas South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=34.76666666666667 +lat_2=33.3 +lat_0=32");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-92 +x_0=609601.2192024384 +y_0=0");
    add_proj4text (p, 2, " +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Arkansas South\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",34.76");
    add_srs_wkt (p, 9,
		 "666666666667],PARAMETER[\"standard_parallel_2\",33.3],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"latitude_of_origin\",32.66666666666666],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"central_meridian\",-92],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",2000000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 13, "SG\",\"26752\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26753, "epsg", 26753,
		      "NAD27 / Colorado North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.71666666666667 +lat_2=40.78333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=39.33333333333334 +lon_0=-105.5 +x_0=609601.2");
    add_proj4text (p, 2,
		   "192024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft");
    add_proj4text (p, 3, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Colorado North\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",39.71");
    add_srs_wkt (p, 9,
		 "666666666667],PARAMETER[\"standard_parallel_2\",40.78333");
    add_srs_wkt (p, 10,
		 "333333333],PARAMETER[\"latitude_of_origin\",39.333333333");
    add_srs_wkt (p, 11,
		 "33334],PARAMETER[\"central_meridian\",-105.5],PARAMETER[");
    add_srs_wkt (p, 12,
		 "\"false_easting\",2000000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 13,
		 "0],AUTHORITY[\"EPSG\",\"26753\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 14, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 26754, "epsg", 26754,
		      "NAD27 / Colorado Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.75 +lat_2=38.45 +lat_0=37.8333333333");
    add_proj4text (p, 1,
		   "3334 +lon_0=-105.5 +x_0=609601.2192024384 +y_0=0 +ellps=");
    add_proj4text (p, 2, "clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Colorado Central\",GEOGCS[\"NAD27\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",39.");
    add_srs_wkt (p, 9,
		 "75],PARAMETER[\"standard_parallel_2\",38.45],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "latitude_of_origin\",37.83333333333334],PARAMETER[\"cent");
    add_srs_wkt (p, 11,
		 "ral_meridian\",-105.5],PARAMETER[\"false_easting\",20000");
    add_srs_wkt (p, 12,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 13, "26754\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26755, "epsg", 26755,
		      "NAD27 / Colorado South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.43333333333333 +lat_2=37.23333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=36.66666666666666 +lon_0=-105.5 +x_0=609601.2");
    add_proj4text (p, 2,
		   "192024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft");
    add_proj4text (p, 3, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Colorado South\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",38.43");
    add_srs_wkt (p, 9,
		 "333333333333],PARAMETER[\"standard_parallel_2\",37.23333");
    add_srs_wkt (p, 10,
		 "333333333],PARAMETER[\"latitude_of_origin\",36.666666666");
    add_srs_wkt (p, 11,
		 "66666],PARAMETER[\"central_meridian\",-105.5],PARAMETER[");
    add_srs_wkt (p, 12,
		 "\"false_easting\",2000000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 13,
		 "0],AUTHORITY[\"EPSG\",\"26755\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 14, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 26756, "epsg", 26756, "NAD27 / Connecticut");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.86666666666667 +lat_2=41.2 +lat_0=40");
    add_proj4text (p, 1,
		   ".83333333333334 +lon_0=-72.75 +x_0=182880.3657607315 +y_");
    add_proj4text (p, 2,
		   "0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Connecticut\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",41.86666");
    add_srs_wkt (p, 9,
		 "666666667],PARAMETER[\"standard_parallel_2\",41.2],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"latitude_of_origin\",40.83333333333334],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"central_meridian\",-72.75],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",600000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 13, "G\",\"26756\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26757, "epsg", 26757, "NAD27 / Delaware");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38 +lon_0=-75.41666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "995 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +datum=N");
    add_proj4text (p, 2, "AD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Delaware\",GEOGCS[\"NAD27\",DATUM[\"Nor");
    add_srs_wkt (p, 1,
		 "th_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637820");
    add_srs_wkt (p, 2,
		 "6.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "67\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",38],PARAMETER[\"cen");
    add_srs_wkt (p, 9,
		 "tral_meridian\",-75.41666666666667],PARAMETER[\"scale_fa");
    add_srs_wkt (p, 10,
		 "ctor\",0.999995],PARAMETER[\"false_easting\",500000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"26757\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26758, "epsg", 26758,
		      "NAD27 / Florida East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=24.33333333333333 +lon_0=-81 +k=0.999");
    add_proj4text (p, 1,
		   "941177 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +datu");
    add_proj4text (p, 2, "m=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Florida East\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",24.33333333333333");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-81],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.999941177],PARAMETER[\"false_easting\",500000");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 12, "6758\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_21 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 26759, "epsg", 26759,
		      "NAD27 / Florida West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=24.33333333333333 +lon_0=-82 +k=0.999");
    add_proj4text (p, 1,
		   "941177 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +datu");
    add_proj4text (p, 2, "m=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Florida West\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",24.33333333333333");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-82],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.999941177],PARAMETER[\"false_easting\",500000");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 12, "6759\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26760, "epsg", 26760,
		      "NAD27 / Florida North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.75 +lat_2=29.58333333333333 +lat_0=2");
    add_proj4text (p, 1,
		   "9 +lon_0=-84.5 +x_0=609601.2192024384 +y_0=0 +ellps=clrk");
    add_proj4text (p, 2, "66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Florida North\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",30.75");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"standard_parallel_2\",29.58333333333333],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"latitude_of_origin\",29],PARAMETER[\"central_");
    add_srs_wkt (p, 11,
		 "meridian\",-84.5],PARAMETER[\"false_easting\",2000000],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2676");
    add_srs_wkt (p, 13, "0\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26766, "epsg", 26766,
		      "NAD27 / Georgia East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=-82.16666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +datum=NAD");
    add_proj4text (p, 2, "27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Georgia East\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",30],PARAMETER[\"c");
    add_srs_wkt (p, 9,
		 "entral_meridian\",-82.16666666666667],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.9999],PARAMETER[\"false_easting\",500000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"26766\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26767, "epsg", 26767,
		      "NAD27 / Georgia West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=-84.16666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +datum=NAD");
    add_proj4text (p, 2, "27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Georgia West\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",30],PARAMETER[\"c");
    add_srs_wkt (p, 9,
		 "entral_meridian\",-84.16666666666667],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.9999],PARAMETER[\"false_easting\",500000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"26767\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26768, "epsg", 26768, "NAD27 / Idaho East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-112.1666666");
    add_proj4text (p, 1,
		   "666667 +k=0.9999473679999999 +x_0=152400.3048006096 +y_0");
    add_proj4text (p, 2, "=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Idaho East\",GEOGCS[\"NAD27\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6378");
    add_srs_wkt (p, 2,
		 "206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",41.66666666666666");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-112.1666666666667],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",0.999947368],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"26768\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 26769, "epsg", 26769,
		      "NAD27 / Idaho Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-114 +k=0.99");
    add_proj4text (p, 1,
		   "99473679999999 +x_0=152400.3048006096 +y_0=0 +ellps=clrk");
    add_proj4text (p, 2, "66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Idaho Central\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",41.66666666666");
    add_srs_wkt (p, 9,
		 "666],PARAMETER[\"central_meridian\",-114],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",0.999947368],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"26769\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26770, "epsg", 26770, "NAD27 / Idaho West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-115.75 +k=0");
    add_proj4text (p, 1,
		   ".999933333 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +");
    add_proj4text (p, 2, "datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Idaho West\",GEOGCS[\"NAD27\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6378");
    add_srs_wkt (p, 2,
		 "206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",41.66666666666666");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-115.75],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",0.999933333],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"26770\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26771, "epsg", 26771,
		      "NAD27 / Illinois East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36.66666666666666 +lon_0=-88.33333333");
    add_proj4text (p, 1,
		   "333333 +k=0.9999749999999999 +x_0=152400.3048006096 +y_0");
    add_proj4text (p, 2, "=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Illinois East\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",36.66666666666");
    add_srs_wkt (p, 9,
		 "666],PARAMETER[\"central_meridian\",-88.33333333333333],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"scale_factor\",0.999975],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"26771\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 26772, "epsg", 26772,
		      "NAD27 / Illinois West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36.66666666666666 +lon_0=-90.16666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999941177 +x_0=152400.3048006096 +y_0=0 +ell");
    add_proj4text (p, 2, "ps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Illinois West\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",36.66666666666");
    add_srs_wkt (p, 9,
		 "666],PARAMETER[\"central_meridian\",-90.16666666666667],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"scale_factor\",0.999941177],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 12,
		 "ORITY[\"EPSG\",\"26772\"],AXIS[\"X\",EAST],AXIS[\"Y\",NO");
    add_srs_wkt (p, 13, "RTH]]");
    p = add_epsg_def (first, last, 26773, "epsg", 26773,
		      "NAD27 / Indiana East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=37.5 +lon_0=-85.66666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +da");
    add_proj4text (p, 2, "tum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Indiana East\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",37.5],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-85.66666666666667],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.999966667],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "26773\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26774, "epsg", 26774,
		      "NAD27 / Indiana West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=37.5 +lon_0=-87.08333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +da");
    add_proj4text (p, 2, "tum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Indiana West\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",37.5],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-87.08333333333333],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.999966667],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "26774\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26775, "epsg", 26775, "NAD27 / Iowa North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43.26666666666667 +lat_2=42.06666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=41.5 +lon_0=-93.5 +x_0=609601.2192024384 +y_0");
    add_proj4text (p, 2, "=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Iowa North\",GEOGCS[\"NAD27\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6378");
    add_srs_wkt (p, 2,
		 "206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",43.26666");
    add_srs_wkt (p, 9,
		 "666666667],PARAMETER[\"standard_parallel_2\",42.06666666");
    add_srs_wkt (p, 10,
		 "666667],PARAMETER[\"latitude_of_origin\",41.5],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"central_meridian\",-93.5],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",2000000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 13, "SG\",\"26775\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26776, "epsg", 26776, "NAD27 / Iowa South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.78333333333333 +lat_2=40.61666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=40 +lon_0=-93.5 +x_0=609601.2192024384 +y_0=0");
    add_proj4text (p, 2, " +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Iowa South\",GEOGCS[\"NAD27\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6378");
    add_srs_wkt (p, 2,
		 "206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",41.78333");
    add_srs_wkt (p, 9,
		 "333333333],PARAMETER[\"standard_parallel_2\",40.61666666");
    add_srs_wkt (p, 10,
		 "666667],PARAMETER[\"latitude_of_origin\",40],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "central_meridian\",-93.5],PARAMETER[\"false_easting\",20");
    add_srs_wkt (p, 12,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 13, ",\"26776\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26777, "epsg", 26777,
		      "NAD27 / Kansas North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.78333333333333 +lat_2=38.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=38.33333333333334 +lon_0=-98 +x_0=609601.2192");
    add_proj4text (p, 2,
		   "024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +n");
    add_proj4text (p, 3, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Kansas North\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",39.78333");
    add_srs_wkt (p, 9,
		 "333333333],PARAMETER[\"standard_parallel_2\",38.71666666");
    add_srs_wkt (p, 10,
		 "666667],PARAMETER[\"latitude_of_origin\",38.333333333333");
    add_srs_wkt (p, 11,
		 "34],PARAMETER[\"central_meridian\",-98],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_easting\",2000000],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 13,
		 "HORITY[\"EPSG\",\"26777\"],AXIS[\"X\",EAST],AXIS[\"Y\",N");
    add_srs_wkt (p, 14, "ORTH]]");
    p = add_epsg_def (first, last, 26778, "epsg", 26778,
		      "NAD27 / Kansas South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.56666666666667 +lat_2=37.26666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.66666666666666 +lon_0=-98.5 +x_0=609601.21");
    add_proj4text (p, 2,
		   "92024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft ");
    add_proj4text (p, 3, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Kansas South\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",38.56666");
    add_srs_wkt (p, 9,
		 "666666667],PARAMETER[\"standard_parallel_2\",37.26666666");
    add_srs_wkt (p, 10,
		 "666667],PARAMETER[\"latitude_of_origin\",36.666666666666");
    add_srs_wkt (p, 11,
		 "66],PARAMETER[\"central_meridian\",-98.5],PARAMETER[\"fa");
    add_srs_wkt (p, 12,
		 "lse_easting\",2000000],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 13,
		 "UTHORITY[\"EPSG\",\"26778\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 14, ",NORTH]]");
    p = add_epsg_def (first, last, 26779, "epsg", 26779,
		      "NAD27 / Kentucky North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.96666666666667 +lat_2=38.96666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=37.5 +lon_0=-84.25 +x_0=609601.2192024384 +y_");
    add_proj4text (p, 2,
		   "0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Kentucky North\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",37.96");
    add_srs_wkt (p, 9,
		 "666666666667],PARAMETER[\"standard_parallel_2\",38.96666");
    add_srs_wkt (p, 10,
		 "666666667],PARAMETER[\"latitude_of_origin\",37.5],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"central_meridian\",-84.25],PARAMETER[\"false_easti");
    add_srs_wkt (p, 12,
		 "ng\",2000000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 13,
		 "\"EPSG\",\"26779\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26780, "epsg", 26780,
		      "NAD27 / Kentucky South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.73333333333333 +lat_2=37.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=36.33333333333334 +lon_0=-85.75 +x_0=609601.2");
    add_proj4text (p, 2,
		   "192024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft");
    add_proj4text (p, 3, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Kentucky South\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",36.73");
    add_srs_wkt (p, 9,
		 "333333333333],PARAMETER[\"standard_parallel_2\",37.93333");
    add_srs_wkt (p, 10,
		 "333333333],PARAMETER[\"latitude_of_origin\",36.333333333");
    add_srs_wkt (p, 11,
		 "33334],PARAMETER[\"central_meridian\",-85.75],PARAMETER[");
    add_srs_wkt (p, 12,
		 "\"false_easting\",2000000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 13,
		 "0],AUTHORITY[\"EPSG\",\"26780\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 14, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 26781, "epsg", 26781,
		      "NAD27 / Louisiana North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=31.16666666666667 +lat_2=32.66666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=30.66666666666667 +lon_0=-92.5 +x_0=609601.21");
    add_proj4text (p, 2,
		   "92024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft ");
    add_proj4text (p, 3, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Louisiana North\",GEOGCS[\"NAD27\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",31.");
    add_srs_wkt (p, 9,
		 "16666666666667],PARAMETER[\"standard_parallel_2\",32.666");
    add_srs_wkt (p, 10,
		 "66666666666],PARAMETER[\"latitude_of_origin\",30.6666666");
    add_srs_wkt (p, 11,
		 "6666667],PARAMETER[\"central_meridian\",-92.5],PARAMETER");
    add_srs_wkt (p, 12,
		 "[\"false_easting\",2000000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 13,
		 ",0],AUTHORITY[\"EPSG\",\"26781\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 14, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26782, "epsg", 26782,
		      "NAD27 / Louisiana South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=29.3 +lat_2=30.7 +lat_0=28.666666666666");
    add_proj4text (p, 1,
		   "67 +lon_0=-91.33333333333333 +x_0=609601.2192024384 +y_0");
    add_proj4text (p, 2, "=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Louisiana South\",GEOGCS[\"NAD27\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",29.");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"standard_parallel_2\",30.7],PARAMETER[\"l");
    add_srs_wkt (p, 10,
		 "atitude_of_origin\",28.66666666666667],PARAMETER[\"centr");
    add_srs_wkt (p, 11,
		 "al_meridian\",-91.33333333333333],PARAMETER[\"false_east");
    add_srs_wkt (p, 12,
		 "ing\",2000000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 13,
		 "[\"EPSG\",\"26782\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    add_srs_wkt (p, 14, "");
    p = add_epsg_def (first, last, 26783, "epsg", 26783, "NAD27 / Maine East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=43.83333333333334 +lon_0=-68.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +datum=N");
    add_proj4text (p, 2, "AD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Maine East\",GEOGCS[\"NAD27\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6378");
    add_srs_wkt (p, 2,
		 "206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",43.83333333333334");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-68.5],PARAMETER[\"scal");
    add_srs_wkt (p, 10,
		 "e_factor\",0.9999],PARAMETER[\"false_easting\",500000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2678");
    add_srs_wkt (p, 12, "3\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26784, "epsg", 26784, "NAD27 / Maine West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.83333333333334 +lon_0=-70.16666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999966667 +x_0=152400.3048006096 +y_0=0 +ell");
    add_proj4text (p, 2, "ps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Maine West\",GEOGCS[\"NAD27\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6378");
    add_srs_wkt (p, 2,
		 "206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",42.83333333333334");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-70.16666666666667],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",0.999966667],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"26784\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 26785, "epsg", 26785, "NAD27 / Maryland");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.3 +lat_2=39.45 +lat_0=37.83333333333");
    add_proj4text (p, 1,
		   "334 +lon_0=-77 +x_0=243840.4876809754 +y_0=0 +ellps=clrk");
    add_proj4text (p, 2, "66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Maryland\",GEOGCS[\"NAD27\",DATUM[\"Nor");
    add_srs_wkt (p, 1,
		 "th_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637820");
    add_srs_wkt (p, 2,
		 "6.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "67\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal_C");
    add_srs_wkt (p, 8,
		 "onic_2SP\"],PARAMETER[\"standard_parallel_1\",38.3],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"standard_parallel_2\",39.45],PARAMETER[\"latitud");
    add_srs_wkt (p, 10,
		 "e_of_origin\",37.83333333333334],PARAMETER[\"central_mer");
    add_srs_wkt (p, 11,
		 "idian\",-77],PARAMETER[\"false_easting\",800000.00000000");
    add_srs_wkt (p, 12,
		 "02],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 13, "26785\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26786, "epsg", 26786,
		      "NAD27 / Massachusetts Mainland");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.71666666666667 +lat_2=42.68333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=41 +lon_0=-71.5 +x_0=182880.3657607315 +y_0=0");
    add_proj4text (p, 2, " +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Massachusetts Mainland\",GEOGCS[\"NAD27");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke");
    add_srs_wkt (p, 2,
		 " 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096");
    add_srs_wkt (p, 7,
		 "012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lamber");
    add_srs_wkt (p, 8,
		 "t_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1");
    add_srs_wkt (p, 9,
		 "\",41.71666666666667],PARAMETER[\"standard_parallel_2\",");
    add_srs_wkt (p, 10,
		 "42.68333333333333],PARAMETER[\"latitude_of_origin\",41],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"central_meridian\",-71.5],PARAMETER[\"false_");
    add_srs_wkt (p, 12,
		 "easting\",600000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 13,
		 "ITY[\"EPSG\",\"26786\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORT");
    add_srs_wkt (p, 14, "H]]");
    p = add_epsg_def (first, last, 26787, "epsg", 26787,
		      "NAD27 / Massachusetts Island");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.28333333333333 +lat_2=41.48333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=41 +lon_0=-70.5 +x_0=60960.12192024384 +y_0=0");
    add_proj4text (p, 2, " +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Massachusetts Island\",GEOGCS[\"NAD27\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1");
    add_srs_wkt (p, 2,
		 "866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4267\"]],UNIT[\"US survey foot\",0.304800609601");
    add_srs_wkt (p, 7,
		 "2192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_");
    add_srs_wkt (p, 8,
		 "Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 9,
		 ",41.28333333333333],PARAMETER[\"standard_parallel_2\",41");
    add_srs_wkt (p, 10,
		 ".48333333333333],PARAMETER[\"latitude_of_origin\",41],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"central_meridian\",-70.5],PARAMETER[\"false_ea");
    add_srs_wkt (p, 12,
		 "sting\",200000],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 13,
		 "Y[\"EPSG\",\"26787\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]");
    add_srs_wkt (p, 14, "]");
    p = add_epsg_def (first, last, 26791, "epsg", 26791,
		      "NAD27 / Minnesota North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.03333333333333 +lat_2=48.63333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=46.5 +lon_0=-93.09999999999999 +x_0=609601.21");
    add_proj4text (p, 2,
		   "92024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft ");
    add_proj4text (p, 3, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Minnesota North\",GEOGCS[\"NAD27\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",47.");
    add_srs_wkt (p, 9,
		 "03333333333333],PARAMETER[\"standard_parallel_2\",48.633");
    add_srs_wkt (p, 10,
		 "33333333333],PARAMETER[\"latitude_of_origin\",46.5],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-93.1],PARAMETER[\"false_east");
    add_srs_wkt (p, 12,
		 "ing\",2000000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 13,
		 "[\"EPSG\",\"26791\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    add_srs_wkt (p, 14, "");
    p = add_epsg_def (first, last, 26792, "epsg", 26792,
		      "NAD27 / Minnesota Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.61666666666667 +lat_2=47.05 +lat_0=4");
    add_proj4text (p, 1,
		   "5 +lon_0=-94.25 +x_0=609601.2192024384 +y_0=0 +ellps=clr");
    add_proj4text (p, 2, "k66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Minnesota Central\",GEOGCS[\"NAD27\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866");
    add_srs_wkt (p, 2,
		 "\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4267\"]],UNIT[\"US survey foot\",0.30480060960121");
    add_srs_wkt (p, 7,
		 "92,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Co");
    add_srs_wkt (p, 8,
		 "nformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",4");
    add_srs_wkt (p, 9,
		 "5.61666666666667],PARAMETER[\"standard_parallel_2\",47.0");
    add_srs_wkt (p, 10,
		 "5],PARAMETER[\"latitude_of_origin\",45],PARAMETER[\"cent");
    add_srs_wkt (p, 11,
		 "ral_meridian\",-94.25],PARAMETER[\"false_easting\",20000");
    add_srs_wkt (p, 12,
		 "00],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 13, "26792\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26793, "epsg", 26793,
		      "NAD27 / Minnesota South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43.78333333333333 +lat_2=45.21666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=43 +lon_0=-94 +x_0=609601.2192024384 +y_0=0 +");
    add_proj4text (p, 2, "ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Minnesota South\",GEOGCS[\"NAD27\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",43.");
    add_srs_wkt (p, 9,
		 "78333333333333],PARAMETER[\"standard_parallel_2\",45.216");
    add_srs_wkt (p, 10,
		 "66666666667],PARAMETER[\"latitude_of_origin\",43],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"central_meridian\",-94],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",2000000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 13, "SG\",\"26793\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26794, "epsg", 26794,
		      "NAD27 / Mississippi East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=29.66666666666667 +lon_0=-88.83333333");
    add_proj4text (p, 1,
		   "333333 +k=0.99996 +x_0=152400.3048006096 +y_0=0 +ellps=c");
    add_proj4text (p, 2, "lrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Mississippi East\",GEOGCS[\"NAD27\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_M");
    add_srs_wkt (p, 8,
		 "ercator\"],PARAMETER[\"latitude_of_origin\",29.666666666");
    add_srs_wkt (p, 9,
		 "66667],PARAMETER[\"central_meridian\",-88.83333333333333");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.99996],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 12,
		 "ITY[\"EPSG\",\"26794\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORT");
    add_srs_wkt (p, 13, "H]]");
    p = add_epsg_def (first, last, 26795, "epsg", 26795,
		      "NAD27 / Mississippi West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30.5 +lon_0=-90.33333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "99941177 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +da");
    add_proj4text (p, 2, "tum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Mississippi West\",GEOGCS[\"NAD27\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_M");
    add_srs_wkt (p, 8,
		 "ercator\"],PARAMETER[\"latitude_of_origin\",30.5],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"central_meridian\",-90.33333333333333],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",0.999941177],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"26795\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26796, "epsg", 26796,
		      "NAD27 / Missouri East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=35.83333333333334 +lon_0=-90.5 +k=0.9");
    add_proj4text (p, 1,
		   "99933333 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +da");
    add_proj4text (p, 2, "tum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Missouri East\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",35.83333333333");
    add_srs_wkt (p, 9,
		 "334],PARAMETER[\"central_meridian\",-90.5],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",0.999933333],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"26796\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26797, "epsg", 26797,
		      "NAD27 / Missouri Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=35.83333333333334 +lon_0=-92.5 +k=0.9");
    add_proj4text (p, 1,
		   "99933333 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +da");
    add_proj4text (p, 2, "tum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Missouri Central\",GEOGCS[\"NAD27\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_M");
    add_srs_wkt (p, 8,
		 "ercator\"],PARAMETER[\"latitude_of_origin\",35.833333333");
    add_srs_wkt (p, 9,
		 "33334],PARAMETER[\"central_meridian\",-92.5],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",0.999933333],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"26797\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26798, "epsg", 26798,
		      "NAD27 / Missouri West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36.16666666666666 +lon_0=-94.5 +k=0.9");
    add_proj4text (p, 1,
		   "99941177 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +da");
    add_proj4text (p, 2, "tum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Missouri West\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",36.16666666666");
    add_srs_wkt (p, 9,
		 "666],PARAMETER[\"central_meridian\",-94.5],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",0.999941177],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"26798\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26799, "epsg", 26799,
		      "NAD27 / California zone VII");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=34.41666666666666 +lat_2=33.86666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=34.13333333333333 +lon_0=-118.3333333333333 +");
    add_proj4text (p, 2,
		   "x_0=1276106.450596901 +y_0=1268253.006858014 +ellps=clrk");
    add_proj4text (p, 3, "66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / California zone VII\",GEOGCS[\"NAD27\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 18");
    add_srs_wkt (p, 2,
		 "66\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "08\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "34.41666666666666],PARAMETER[\"standard_parallel_2\",33.");
    add_srs_wkt (p, 10,
		 "86666666666667],PARAMETER[\"latitude_of_origin\",34.1333");
    add_srs_wkt (p, 11,
		 "3333333333],PARAMETER[\"central_meridian\",-118.33333333");
    add_srs_wkt (p, 12,
		 "33333],PARAMETER[\"false_easting\",4186692.58],PARAMETER");
    add_srs_wkt (p, 13,
		 "[\"false_northing\",4160926.74],AUTHORITY[\"EPSG\",\"267");
    add_srs_wkt (p, 14, "99\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26801, "epsg", 26801,
		      "NAD Michigan / Michigan East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.5 +lon_0=-83.66666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "99942857 +x_0=152400.3048006096 +y_0=0 +a=6378450.047548");
    add_proj4text (p, 2, "896 +b=6356826.621488444 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD Michigan / Michigan East\",GEOGCS[\"NAD27 M");
    add_srs_wkt (p, 1,
		 "ichigan\",DATUM[\"NAD_Michigan\",SPHEROID[\"Clarke 1866 ");
    add_srs_wkt (p, 2,
		 "Michigan\",6378450.047548896,294.9786971646739,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7009\"]],AUTHORITY[\"EPSG\",\"6268\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4268\"]],UNIT[\"US survey foot\",");
    add_srs_wkt (p, 7,
		 "0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECT");
    add_srs_wkt (p, 8,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 9,
		 "gin\",41.5],PARAMETER[\"central_meridian\",-83.666666666");
    add_srs_wkt (p, 10,
		 "66667],PARAMETER[\"scale_factor\",0.999942857],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_easting\",500000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 12,
		 "0],AUTHORITY[\"EPSG\",\"26801\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 26802, "epsg", 26802,
		      "NAD Michigan / Michigan Old Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.5 +lon_0=-85.75 +k=0.999909091 +x_");
    add_proj4text (p, 1,
		   "0=152400.3048006096 +y_0=0 +a=6378450.047548896 +b=63568");
    add_proj4text (p, 2, "26.621488444 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD Michigan / Michigan Old Central\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD27 Michigan\",DATUM[\"NAD_Michigan\",SPHEROID[\"Clark");
    add_srs_wkt (p, 2,
		 "e 1866 Michigan\",6378450.047548896,294.9786971646739,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7009\"]],AUTHORITY[\"EPSG\",\"6268\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4268\"]],UNIT[\"US survey ");
    add_srs_wkt (p, 7,
		 "foot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",41.5],PARAMETER[\"central_meridian\",-85.75");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.999909091],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_easting\",500000],PARAMETER[\"false_northing\",0],AU");
    add_srs_wkt (p, 12,
		 "THORITY[\"EPSG\",\"26802\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 13, "NORTH]]");
    p = add_epsg_def (first, last, 26803, "epsg", 26803,
		      "NAD Michigan / Michigan West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.5 +lon_0=-88.75 +k=0.999909091 +x_");
    add_proj4text (p, 1,
		   "0=152400.3048006096 +y_0=0 +a=6378450.047548896 +b=63568");
    add_proj4text (p, 2, "26.621488444 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD Michigan / Michigan West\",GEOGCS[\"NAD27 M");
    add_srs_wkt (p, 1,
		 "ichigan\",DATUM[\"NAD_Michigan\",SPHEROID[\"Clarke 1866 ");
    add_srs_wkt (p, 2,
		 "Michigan\",6378450.047548896,294.9786971646739,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7009\"]],AUTHORITY[\"EPSG\",\"6268\"]],PRIME");
    add_srs_wkt (p, 4,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 5,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 6,
		 "],AUTHORITY[\"EPSG\",\"4268\"]],UNIT[\"US survey foot\",");
    add_srs_wkt (p, 7,
		 "0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECT");
    add_srs_wkt (p, 8,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 9,
		 "gin\",41.5],PARAMETER[\"central_meridian\",-88.75],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.999909091],PARAMETER[\"false_eas");
    add_srs_wkt (p, 11,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 12,
		 "[\"EPSG\",\"26803\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 26811, "epsg", 26811,
		      "NAD Michigan / Michigan North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.48333333333333 +lat_2=47.08333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=44.78333333333333 +lon_0=-87 +x_0=609601.2192");
    add_proj4text (p, 2,
		   "024384 +y_0=0 +a=6378450.047548896 +b=6356826.621488444 ");
    add_proj4text (p, 3, "+units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD Michigan / Michigan North\",GEOGCS[\"NAD27 ");
    add_srs_wkt (p, 1,
		 "Michigan\",DATUM[\"NAD_Michigan\",SPHEROID[\"Clarke 1866");
    add_srs_wkt (p, 2,
		 " Michigan\",6378450.047548896,294.9786971646739,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7009\"]],AUTHORITY[\"EPSG\",\"6268\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4268\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",45.48333333333333],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",47.08333333333334],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",44.78333333333333],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 12,
		 "\",-87],PARAMETER[\"false_easting\",2000000],PARAMETER[\"");
    add_srs_wkt (p, 13,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"26811\"],AXIS[\"");
    add_srs_wkt (p, 14, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26812, "epsg", 26812,
		      "NAD Michigan / Michigan Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.18333333333333 +lat_2=45.7 +lat_0=43");
    add_proj4text (p, 1,
		   ".31666666666667 +lon_0=-84.33333333333333 +x_0=609601.21");
    add_proj4text (p, 2,
		   "92024384 +y_0=0 +a=6378450.047548896 +b=6356826.62148844");
    add_proj4text (p, 3, "4 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD Michigan / Michigan Central\",GEOGCS[\"NAD2");
    add_srs_wkt (p, 1,
		 "7 Michigan\",DATUM[\"NAD_Michigan\",SPHEROID[\"Clarke 18");
    add_srs_wkt (p, 2,
		 "66 Michigan\",6378450.047548896,294.9786971646739,AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"7009\"]],AUTHORITY[\"EPSG\",\"6268\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6,
		 "\"]],AUTHORITY[\"EPSG\",\"4268\"]],UNIT[\"US survey foot");
    add_srs_wkt (p, 7,
		 "\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJ");
    add_srs_wkt (p, 8,
		 "ECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"stan");
    add_srs_wkt (p, 9,
		 "dard_parallel_1\",44.18333333333333],PARAMETER[\"standar");
    add_srs_wkt (p, 10,
		 "d_parallel_2\",45.7],PARAMETER[\"latitude_of_origin\",43");
    add_srs_wkt (p, 11,
		 ".31666666666667],PARAMETER[\"central_meridian\",-84.3333");
    add_srs_wkt (p, 12,
		 "3333333333],PARAMETER[\"false_easting\",2000000],PARAMET");
    add_srs_wkt (p, 13,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"26812\"],A");
    add_srs_wkt (p, 14, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26813, "epsg", 26813,
		      "NAD Michigan / Michigan South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=42.1 +lat_2=43.66666666666666 +lat_0=41");
    add_proj4text (p, 1,
		   ".5 +lon_0=-84.33333333333333 +x_0=609601.2192024384 +y_0");
    add_proj4text (p, 2,
		   "=0 +a=6378450.047548896 +b=6356826.621488444 +units=us-f");
    add_proj4text (p, 3, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD Michigan / Michigan South\",GEOGCS[\"NAD27 ");
    add_srs_wkt (p, 1,
		 "Michigan\",DATUM[\"NAD_Michigan\",SPHEROID[\"Clarke 1866");
    add_srs_wkt (p, 2,
		 " Michigan\",6378450.047548896,294.9786971646739,AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"7009\"]],AUTHORITY[\"EPSG\",\"6268\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4268\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",42.1],PARAMETER[\"standard_parallel_2\",");
    add_srs_wkt (p, 10,
		 "43.66666666666666],PARAMETER[\"latitude_of_origin\",41.5");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"central_meridian\",-84.33333333333333],PAR");
    add_srs_wkt (p, 12,
		 "AMETER[\"false_easting\",2000000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 13,
		 "hing\",0],AUTHORITY[\"EPSG\",\"26813\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 14, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26814, "epsg", 26814,
		      "NAD83 / Maine East (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=43.66666666666666 +lon_0=-68.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=300000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Maine East (ftUS) (deprecated)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",43.66666666666666],PARAMETER[\"c");
    add_srs_wkt (p, 9,
		 "entral_meridian\",-68.5],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 10,
		 "99],PARAMETER[\"false_easting\",300000],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"26814\"],AXIS[\"X\"");
    add_srs_wkt (p, 12, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26815, "epsg", 26815,
		      "NAD83 / Maine West (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.83333333333334 +lon_0=-70.16666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999966667 +x_0=900000 +y_0=0 +ellps=GRS80 +d");
    add_proj4text (p, 2, "atum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Maine West (ftUS) (deprecated)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7019\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",42.83333333333334],PARAMETER[\"c");
    add_srs_wkt (p, 9,
		 "entral_meridian\",-70.16666666666667],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.999966667],PARAMETER[\"false_easting\",900000");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 12, "6815\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26819, "epsg", 26819,
		      "NAD83 / Minnesota North (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.63333333333333 +lat_2=47.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=46.5 +lon_0=-93.09999999999999 +x_0=800000.00");
    add_proj4text (p, 2,
		   "00101601 +y_0=99999.99998984 +ellps=GRS80 +datum=NAD83 +");
    add_proj4text (p, 3, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Minnesota North (ftUS) (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"NAD83\",DATUM[\"North_American_Datum_1983\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"standard_parallel_1\",48.63333333333333]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"standard_parallel_2\",47.03333333333333],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"latitude_of_origin\",46.5],PARAMETER[\"central");
    add_srs_wkt (p, 11,
		 "_meridian\",-93.1],PARAMETER[\"false_easting\",800000.00");
    add_srs_wkt (p, 12,
		 "00101601],PARAMETER[\"false_northing\",99999.99998984],A");
    add_srs_wkt (p, 13,
		 "UTHORITY[\"EPSG\",\"26819\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 14, ",NORTH]]");
    p = add_epsg_def (first, last, 26820, "epsg", 26820,
		      "NAD83 / Minnesota Central (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.05 +lat_2=45.61666666666667 +lat_0=4");
    add_proj4text (p, 1,
		   "5 +lon_0=-94.25 +x_0=800000.0000101601 +y_0=99999.999989");
    add_proj4text (p, 2, "84 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Minnesota Central (ftUS) (deprecated)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"NAD83\",DATUM[\"North_American_Datum_1983\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",47.05],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_2\",45.61666666666667],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "latitude_of_origin\",45],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 11,
		 "-94.25],PARAMETER[\"false_easting\",800000.0000101601],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_northing\",99999.99998984],AUTHORITY[\"");
    add_srs_wkt (p, 13,
		 "EPSG\",\"26820\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26821, "epsg", 26821,
		      "NAD83 / Minnesota South (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.21666666666667 +lat_2=43.78333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=43 +lon_0=-94 +x_0=800000.0000101601 +y_0=999");
    add_proj4text (p, 2,
		   "99.99998984 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Minnesota South (ftUS) (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"NAD83\",DATUM[\"North_American_Datum_1983\",SPHE");
    add_srs_wkt (p, 2,
		 "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"standard_parallel_1\",45.21666666666667]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"standard_parallel_2\",43.78333333333333],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"latitude_of_origin\",43],PARAMETER[\"central_m");
    add_srs_wkt (p, 11,
		 "eridian\",-94],PARAMETER[\"false_easting\",800000.000010");
    add_srs_wkt (p, 12,
		 "1601],PARAMETER[\"false_northing\",99999.99998984],AUTHO");
    add_srs_wkt (p, 13,
		 "RITY[\"EPSG\",\"26821\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 14, "TH]]");
    p = add_epsg_def (first, last, 26822, "epsg", 26822,
		      "NAD83 / Nebraska (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43 +lat_2=40 +lat_0=39.83333333333334 +");
    add_proj4text (p, 1,
		   "lon_0=-100 +x_0=500000.0000101601 +y_0=0 +ellps=GRS80 +d");
    add_proj4text (p, 2, "atum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Nebraska (ftUS) (deprecated)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"G");
    add_srs_wkt (p, 2,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "9\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"standard_parallel_1\",43],PARAMETER[\"standard_p");
    add_srs_wkt (p, 9,
		 "arallel_2\",40],PARAMETER[\"latitude_of_origin\",39.8333");
    add_srs_wkt (p, 10,
		 "3333333334],PARAMETER[\"central_meridian\",-100],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_easting\",500000.0000101601],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"26822\"],AXIS[\"X\"");
    add_srs_wkt (p, 13, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26823, "epsg", 26823,
		      "NAD83 / West Virginia North (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.25 +lat_2=39 +lat_0=38.5 +lon_0=-79.");
    add_proj4text (p, 1,
		   "5 +x_0=1968500 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m");
    add_proj4text (p, 2, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / West Virginia North (ftUS) (deprecated)");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"NAD83\",DATUM[\"North_American_Datum_1983\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",40.25],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_2\",39],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 10,
		 "gin\",38.5],PARAMETER[\"central_meridian\",-79.5],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_easting\",1968500],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",0],AUTHORITY[\"EPSG\",\"26823\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 13, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26824, "epsg", 26824,
		      "NAD83 / West Virginia South (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.88333333333333 +lat_2=37.48333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=37 +lon_0=-81 +x_0=1968500 +y_0=0 +ellps=GRS8");
    add_proj4text (p, 2, "0 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / West Virginia South (ftUS) (deprecated)");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"NAD83\",DATUM[\"North_American_Datum_1983\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"standard_parallel_1\",38.8833333333333");
    add_srs_wkt (p, 9,
		 "3],PARAMETER[\"standard_parallel_2\",37.48333333333333],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",37],PARAMETER[\"central");
    add_srs_wkt (p, 11,
		 "_meridian\",-81],PARAMETER[\"false_easting\",1968500],PA");
    add_srs_wkt (p, 12,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"26824");
    add_srs_wkt (p, 13, "\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26825, "epsg", 26825,
		      "NAD83(HARN) / Maine East (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=43.66666666666666 +lon_0=-68.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=300000 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Maine East (ftUS) (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regio");
    add_srs_wkt (p, 2,
		 "nal_Network\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"615");
    add_srs_wkt (p, 4,
		 "2\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",43.6666666");
    add_srs_wkt (p, 9,
		 "6666666],PARAMETER[\"central_meridian\",-68.5],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",30");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"26825\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26826, "epsg", 26826,
		      "NAD83(HARN) / Maine West (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.83333333333334 +lon_0=-70.16666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999966667 +x_0=900000 +y_0=0 +ellps=GRS80 +u");
    add_proj4text (p, 2, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Maine West (ftUS) (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regio");
    add_srs_wkt (p, 2,
		 "nal_Network\",SPHEROID[\"GRS 1980\",6378137,298.25722210");
    add_srs_wkt (p, 3,
		 "1,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"615");
    add_srs_wkt (p, 4,
		 "2\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 5,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",42.8333333");
    add_srs_wkt (p, 9,
		 "3333334],PARAMETER[\"central_meridian\",-70.166666666666");
    add_srs_wkt (p, 10,
		 "67],PARAMETER[\"scale_factor\",0.999966667],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",900000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"26826\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 26830, "epsg", 26830,
		      "NAD83(HARN) / Minnesota North (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.63333333333333 +lat_2=47.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=46.5 +lon_0=-93.09999999999999 +x_0=800000.00");
    add_proj4text (p, 2,
		   "00101601 +y_0=99999.99998984 +ellps=GRS80 +units=m +no_d");
    add_proj4text (p, 3, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Minnesota North (ftUS) (deprecate");
    add_srs_wkt (p, 1,
		 "d)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_");
    add_srs_wkt (p, 2,
		 "Regional_Network\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lamb");
    add_srs_wkt (p, 8,
		 "ert_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel");
    add_srs_wkt (p, 9,
		 "_1\",48.63333333333333],PARAMETER[\"standard_parallel_2\"");
    add_srs_wkt (p, 10,
		 ",47.03333333333333],PARAMETER[\"latitude_of_origin\",46.");
    add_srs_wkt (p, 11,
		 "5],PARAMETER[\"central_meridian\",-93.1],PARAMETER[\"fal");
    add_srs_wkt (p, 12,
		 "se_easting\",800000.0000101601],PARAMETER[\"false_northi");
    add_srs_wkt (p, 13,
		 "ng\",99999.99998984],AUTHORITY[\"EPSG\",\"26830\"],AXIS[");
    add_srs_wkt (p, 14, "\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26831, "epsg", 26831,
		      "NAD83(HARN) / Minnesota Central (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.05 +lat_2=45.61666666666667 +lat_0=4");
    add_proj4text (p, 1,
		   "5 +lon_0=-94.25 +x_0=800000.0000101601 +y_0=99999.999989");
    add_proj4text (p, 2, "84 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Minnesota Central (ftUS) (depreca");
    add_srs_wkt (p, 1,
		 "ted)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accurac");
    add_srs_wkt (p, 2,
		 "y_Regional_Network\",SPHEROID[\"GRS 1980\",6378137,298.2");
    add_srs_wkt (p, 3,
		 "57222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lamb");
    add_srs_wkt (p, 8,
		 "ert_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel");
    add_srs_wkt (p, 9,
		 "_1\",47.05],PARAMETER[\"standard_parallel_2\",45.6166666");
    add_srs_wkt (p, 10,
		 "6666667],PARAMETER[\"latitude_of_origin\",45],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"central_meridian\",-94.25],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",800000.0000101601],PARAMETER[\"false_northing\",99999.9");
    add_srs_wkt (p, 13,
		 "9998984],AUTHORITY[\"EPSG\",\"26831\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 14, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26832, "epsg", 26832,
		      "NAD83(HARN) / Minnesota South (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.21666666666667 +lat_2=43.78333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=43 +lon_0=-94 +x_0=800000.0000101601 +y_0=999");
    add_proj4text (p, 2, "99.99998984 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Minnesota South (ftUS) (deprecate");
    add_srs_wkt (p, 1,
		 "d)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_");
    add_srs_wkt (p, 2,
		 "Regional_Network\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lamb");
    add_srs_wkt (p, 8,
		 "ert_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel");
    add_srs_wkt (p, 9,
		 "_1\",45.21666666666667],PARAMETER[\"standard_parallel_2\"");
    add_srs_wkt (p, 10,
		 ",43.78333333333333],PARAMETER[\"latitude_of_origin\",43]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"central_meridian\",-94],PARAMETER[\"false_e");
    add_srs_wkt (p, 12,
		 "asting\",800000.0000101601],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 13,
		 ",99999.99998984],AUTHORITY[\"EPSG\",\"26832\"],AXIS[\"X\"");
    add_srs_wkt (p, 14, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26833, "epsg", 26833,
		      "NAD83(HARN) / Nebraska (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43 +lat_2=40 +lat_0=39.83333333333334 +");
    add_proj4text (p, 1,
		   "lon_0=-100 +x_0=500000.0000101601 +y_0=0 +ellps=GRS80 +u");
    add_proj4text (p, 2, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Nebraska (ftUS) (deprecated)\",GE");
    add_srs_wkt (p, 1,
		 "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regiona");
    add_srs_wkt (p, 2,
		 "l_Network\",SPHEROID[\"GRS 1980\",6378137,298.257222101,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 5,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Confo");
    add_srs_wkt (p, 8,
		 "rmal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",43],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"standard_parallel_2\",40],PARAMETER[\"latitu");
    add_srs_wkt (p, 10,
		 "de_of_origin\",39.83333333333334],PARAMETER[\"central_me");
    add_srs_wkt (p, 11,
		 "ridian\",-100],PARAMETER[\"false_easting\",500000.000010");
    add_srs_wkt (p, 12,
		 "1601],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 13, ",\"26833\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26834, "epsg", 26834,
		      "NAD83(HARN) / West Virginia North (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.25 +lat_2=39 +lat_0=38.5 +lon_0=-79.");
    add_proj4text (p, 1,
		   "5 +x_0=1968500 +y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / West Virginia North (ftUS) (depre");
    add_srs_wkt (p, 1,
		 "cated)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accur");
    add_srs_wkt (p, 2,
		 "acy_Regional_Network\",SPHEROID[\"GRS 1980\",6378137,298");
    add_srs_wkt (p, 3,
		 ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 8,
		 "Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_para");
    add_srs_wkt (p, 9,
		 "llel_1\",40.25],PARAMETER[\"standard_parallel_2\",39],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"latitude_of_origin\",38.5],PARAMETER[\"central");
    add_srs_wkt (p, 11,
		 "_meridian\",-79.5],PARAMETER[\"false_easting\",1968500],");
    add_srs_wkt (p, 12,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"268");
    add_srs_wkt (p, 13, "34\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26835, "epsg", 26835,
		      "NAD83(HARN) / West Virginia South (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.88333333333333 +lat_2=37.48333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=37 +lon_0=-81 +x_0=1968500 +y_0=0 +ellps=GRS8");
    add_proj4text (p, 2, "0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / West Virginia South (ftUS) (depre");
    add_srs_wkt (p, 1,
		 "cated)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accur");
    add_srs_wkt (p, 2,
		 "acy_Regional_Network\",SPHEROID[\"GRS 1980\",6378137,298");
    add_srs_wkt (p, 3,
		 ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 8,
		 "Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_para");
    add_srs_wkt (p, 9,
		 "llel_1\",38.88333333333333],PARAMETER[\"standard_paralle");
    add_srs_wkt (p, 10,
		 "l_2\",37.48333333333333],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 11,
		 ",37],PARAMETER[\"central_meridian\",-81],PARAMETER[\"fal");
    add_srs_wkt (p, 12,
		 "se_easting\",1968500],PARAMETER[\"false_northing\",0],AU");
    add_srs_wkt (p, 13,
		 "THORITY[\"EPSG\",\"26835\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 14, "NORTH]]");
    p = add_epsg_def (first, last, 26836, "epsg", 26836,
		      "NAD83(NSRS2007) / Maine East (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=43.66666666666666 +lon_0=-68.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=300000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0");
    add_proj4text (p, 2, ",0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Maine East (ftUS) (deprecated");
    add_srs_wkt (p, 1,
		 ")\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Sp");
    add_srs_wkt (p, 2,
		 "atial_Reference_System_2007\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 3,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[");
    add_srs_wkt (p, 4,
		 "0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 5,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 6,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 8,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"latitude_of_origin\",43.66666666666666],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"central_meridian\",-68.5],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 11,
		 ",0.9999],PARAMETER[\"false_easting\",300000],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"26836\"],AXIS[\"");
    add_srs_wkt (p, 13, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26837, "epsg", 26837,
		      "NAD83(NSRS2007) / Maine West (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.83333333333334 +lon_0=-70.16666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999966667 +x_0=900000 +y_0=0 +ellps=GRS80 +t");
    add_proj4text (p, 2, "owgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Maine West (ftUS) (deprecated");
    add_srs_wkt (p, 1,
		 ")\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Sp");
    add_srs_wkt (p, 2,
		 "atial_Reference_System_2007\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 3,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[");
    add_srs_wkt (p, 4,
		 "0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 5,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 6,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 8,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"latitude_of_origin\",42.83333333333334],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"central_meridian\",-70.16666666666667],PARAMETER[\"s");
    add_srs_wkt (p, 11,
		 "cale_factor\",0.999966667],PARAMETER[\"false_easting\",9");
    add_srs_wkt (p, 12,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 13, ",\"26837\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26841, "epsg", 26841,
		      "NAD83(NSRS2007) / Minnesota North (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.63333333333333 +lat_2=47.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=46.5 +lon_0=-93.09999999999999 +x_0=800000.00");
    add_proj4text (p, 2,
		   "00101601 +y_0=99999.99998984 +ellps=GRS80 +towgs84=0,0,0");
    add_proj4text (p, 3, ",0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Minnesota North (ftUS) (depre");
    add_srs_wkt (p, 1,
		 "cated)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_Nation");
    add_srs_wkt (p, 2,
		 "al_Spatial_Reference_System_2007\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 3,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOW");
    add_srs_wkt (p, 4,
		 "GS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM");
    add_srs_wkt (p, 5,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 6,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"4759\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 8,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Con");
    add_srs_wkt (p, 9,
		 "ic_2SP\"],PARAMETER[\"standard_parallel_1\",48.633333333");
    add_srs_wkt (p, 10,
		 "33333],PARAMETER[\"standard_parallel_2\",47.033333333333");
    add_srs_wkt (p, 11,
		 "33],PARAMETER[\"latitude_of_origin\",46.5],PARAMETER[\"c");
    add_srs_wkt (p, 12,
		 "entral_meridian\",-93.1],PARAMETER[\"false_easting\",800");
    add_srs_wkt (p, 13,
		 "000.0000101601],PARAMETER[\"false_northing\",99999.99998");
    add_srs_wkt (p, 14,
		 "984],AUTHORITY[\"EPSG\",\"26841\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 15, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26842, "epsg", 26842,
		      "NAD83(NSRS2007) / Minnesota Central (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.05 +lat_2=45.61666666666667 +lat_0=4");
    add_proj4text (p, 1,
		   "5 +lon_0=-94.25 +x_0=800000.0000101601 +y_0=99999.999989");
    add_proj4text (p, 2,
		   "84 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_proj4text (p, 3, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Minnesota Central (ftUS) (dep");
    add_srs_wkt (p, 1,
		 "recated)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_Nati");
    add_srs_wkt (p, 2,
		 "onal_Spatial_Reference_System_2007\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 3,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T");
    add_srs_wkt (p, 4,
		 "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIM");
    add_srs_wkt (p, 5,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 6,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 7,
		 "]],AUTHORITY[\"EPSG\",\"4759\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 8,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_C");
    add_srs_wkt (p, 9,
		 "onic_2SP\"],PARAMETER[\"standard_parallel_1\",47.05],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"standard_parallel_2\",45.61666666666667],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"latitude_of_origin\",45],PARAMETER[\"central_merid");
    add_srs_wkt (p, 12,
		 "ian\",-94.25],PARAMETER[\"false_easting\",800000.0000101");
    add_srs_wkt (p, 13,
		 "601],PARAMETER[\"false_northing\",99999.99998984],AUTHOR");
    add_srs_wkt (p, 14,
		 "ITY[\"EPSG\",\"26842\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORT");
    add_srs_wkt (p, 15, "H]]");
    p = add_epsg_def (first, last, 26843, "epsg", 26843,
		      "NAD83(NSRS2007) / Minnesota South (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.21666666666667 +lat_2=43.78333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=43 +lon_0=-94 +x_0=800000.0000101601 +y_0=999");
    add_proj4text (p, 2,
		   "99.99998984 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m");
    add_proj4text (p, 3, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Minnesota South (ftUS) (depre");
    add_srs_wkt (p, 1,
		 "cated)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_Nation");
    add_srs_wkt (p, 2,
		 "al_Spatial_Reference_System_2007\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 3,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOW");
    add_srs_wkt (p, 4,
		 "GS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM");
    add_srs_wkt (p, 5,
		 "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de");
    add_srs_wkt (p, 6,
		 "gree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]]");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"4759\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 8,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Con");
    add_srs_wkt (p, 9,
		 "ic_2SP\"],PARAMETER[\"standard_parallel_1\",45.216666666");
    add_srs_wkt (p, 10,
		 "66667],PARAMETER[\"standard_parallel_2\",43.783333333333");
    add_srs_wkt (p, 11,
		 "33],PARAMETER[\"latitude_of_origin\",43],PARAMETER[\"cen");
    add_srs_wkt (p, 12,
		 "tral_meridian\",-94],PARAMETER[\"false_easting\",800000.");
    add_srs_wkt (p, 13,
		 "0000101601],PARAMETER[\"false_northing\",99999.99998984]");
    add_srs_wkt (p, 14,
		 ",AUTHORITY[\"EPSG\",\"26843\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 15, "\",NORTH]]");
    p = add_epsg_def (first, last, 26844, "epsg", 26844,
		      "NAD83(NSRS2007) / Nebraska (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43 +lat_2=40 +lat_0=39.83333333333334 +");
    add_proj4text (p, 1,
		   "lon_0=-100 +x_0=500000.0000101601 +y_0=0 +ellps=GRS80 +t");
    add_proj4text (p, 2, "owgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Nebraska (ftUS) (deprecated)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spati");
    add_srs_wkt (p, 2,
		 "al_Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 3,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0");
    add_srs_wkt (p, 4,
		 ",0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 5,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 6,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"4759\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 8,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"standard_parallel_1\",43],PARAMETER[\"stan");
    add_srs_wkt (p, 10,
		 "dard_parallel_2\",40],PARAMETER[\"latitude_of_origin\",3");
    add_srs_wkt (p, 11,
		 "9.83333333333334],PARAMETER[\"central_meridian\",-100],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_easting\",500000.0000101601],PARAMETER[");
    add_srs_wkt (p, 13,
		 "\"false_northing\",0],AUTHORITY[\"EPSG\",\"26844\"],AXIS");
    add_srs_wkt (p, 14, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26845, "epsg", 26845,
		      "NAD83(NSRS2007) / West Virginia North (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.25 +lat_2=39 +lat_0=38.5 +lon_0=-79.");
    add_proj4text (p, 1,
		   "5 +x_0=1968500 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,");
    add_proj4text (p, 2, "0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / West Virginia North (ftUS) (d");
    add_srs_wkt (p, 1,
		 "eprecated)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_Na");
    add_srs_wkt (p, 2,
		 "tional_Spatial_Reference_System_2007\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 3,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 4,
		 ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PR");
    add_srs_wkt (p, 5,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 7,
		 "\"]],AUTHORITY[\"EPSG\",\"4759\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 8,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 9,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",40.25],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"standard_parallel_2\",39],PARAMETER[\"latitud");
    add_srs_wkt (p, 11,
		 "e_of_origin\",38.5],PARAMETER[\"central_meridian\",-79.5");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_easting\",1968500],PARAMETER[\"false");
    add_srs_wkt (p, 13,
		 "_northing\",0],AUTHORITY[\"EPSG\",\"26845\"],AXIS[\"X\",");
    add_srs_wkt (p, 14, "EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26846, "epsg", 26846,
		      "NAD83(NSRS2007) / West Virginia South (ftUS) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.88333333333333 +lat_2=37.48333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=37 +lon_0=-81 +x_0=1968500 +y_0=0 +ellps=GRS8");
    add_proj4text (p, 2, "0 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / West Virginia South (ftUS) (d");
    add_srs_wkt (p, 1,
		 "eprecated)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_Na");
    add_srs_wkt (p, 2,
		 "tional_Spatial_Reference_System_2007\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 3,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 4,
		 ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PR");
    add_srs_wkt (p, 5,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 7,
		 "\"]],AUTHORITY[\"EPSG\",\"4759\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 8,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 9,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",38.88333");
    add_srs_wkt (p, 10,
		 "333333333],PARAMETER[\"standard_parallel_2\",37.48333333");
    add_srs_wkt (p, 11,
		 "333333],PARAMETER[\"latitude_of_origin\",37],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "central_meridian\",-81],PARAMETER[\"false_easting\",1968");
    add_srs_wkt (p, 13,
		 "500],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 14, "\"26846\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26847, "epsg", 26847,
		      "NAD83 / Maine East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=43.66666666666666 +lon_0=-68.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=300000.0000000001 +y_0=0 +ellps=GRS80 +datum=NA");
    add_proj4text (p, 2, "D83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Maine East (ftUS)\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",43.66666666666666],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"central_meridian\",-68.5],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.9999],PARAMETER[\"false_easting\",984250.0000");
    add_srs_wkt (p, 11,
		 "000002],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"26847\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26848, "epsg", 26848,
		      "NAD83 / Maine West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.83333333333334 +lon_0=-70.16666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999966667 +x_0=900000 +y_0=0 +ellps=GRS80 +d");
    add_proj4text (p, 2, "atum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Maine West (ftUS)\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",42.83333333333334],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"central_meridian\",-70.16666666666667],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.999966667],PARAMETER[\"false_eas");
    add_srs_wkt (p, 11,
		 "ting\",2952750],PARAMETER[\"false_northing\",0],AUTHORIT");
    add_srs_wkt (p, 12,
		 "Y[\"EPSG\",\"26848\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]");
    add_srs_wkt (p, 13, "]");
    p = add_epsg_def (first, last, 26849, "epsg", 26849,
		      "NAD83 / Minnesota North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.63333333333333 +lat_2=47.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=46.5 +lon_0=-93.09999999999999 +x_0=800000.00");
    add_proj4text (p, 2,
		   "00101599 +y_0=99999.99998983997 +ellps=GRS80 +datum=NAD8");
    add_proj4text (p, 3, "3 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Minnesota North (ftUS)\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 2,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",48.63");
    add_srs_wkt (p, 9,
		 "333333333333],PARAMETER[\"standard_parallel_2\",47.03333");
    add_srs_wkt (p, 10,
		 "333333333],PARAMETER[\"latitude_of_origin\",46.5],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"central_meridian\",-93.1],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 12,
		 "g\",2624666.6667],PARAMETER[\"false_northing\",328083.33");
    add_srs_wkt (p, 13,
		 "33],AUTHORITY[\"EPSG\",\"26849\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 14, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26850, "epsg", 26850,
		      "NAD83 / Minnesota Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.05 +lat_2=45.61666666666667 +lat_0=4");
    add_proj4text (p, 1,
		   "5 +lon_0=-94.25 +x_0=800000.0000101599 +y_0=99999.999989");
    add_proj4text (p, 2,
		   "83997 +ellps=GRS80 +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Minnesota Central (ftUS)\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS ");
    add_srs_wkt (p, 2,
		 "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",47.");
    add_srs_wkt (p, 9,
		 "05],PARAMETER[\"standard_parallel_2\",45.61666666666667]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"latitude_of_origin\",45],PARAMETER[\"centra");
    add_srs_wkt (p, 11,
		 "l_meridian\",-94.25],PARAMETER[\"false_easting\",2624666");
    add_srs_wkt (p, 12,
		 ".6667],PARAMETER[\"false_northing\",328083.3333],AUTHORI");
    add_srs_wkt (p, 13,
		 "TY[\"EPSG\",\"26850\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH");
    add_srs_wkt (p, 14, "]]");
    p = add_epsg_def (first, last, 26851, "epsg", 26851,
		      "NAD83 / Minnesota South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.21666666666667 +lat_2=43.78333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=43 +lon_0=-94 +x_0=800000.0000101599 +y_0=999");
    add_proj4text (p, 2,
		   "99.99998983997 +ellps=GRS80 +datum=NAD83 +units=us-ft +n");
    add_proj4text (p, 3, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Minnesota South (ftUS)\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 2,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4269\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",45.21");
    add_srs_wkt (p, 9,
		 "666666666667],PARAMETER[\"standard_parallel_2\",43.78333");
    add_srs_wkt (p, 10,
		 "333333333],PARAMETER[\"latitude_of_origin\",43],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"central_meridian\",-94],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 12,
		 "2624666.6667],PARAMETER[\"false_northing\",328083.3333],");
    add_srs_wkt (p, 13,
		 "AUTHORITY[\"EPSG\",\"26851\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 14, ",NORTH]]");
    p = add_epsg_def (first, last, 26852, "epsg", 26852,
		      "NAD83 / Nebraska (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43 +lat_2=40 +lat_0=39.83333333333334 +");
    add_proj4text (p, 1,
		   "lon_0=-100 +x_0=500000.00001016 +y_0=0 +ellps=GRS80 +dat");
    add_proj4text (p, 2, "um=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Nebraska (ftUS)\",GEOGCS[\"NAD83\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",63");
    add_srs_wkt (p, 2,
		 "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal_Con");
    add_srs_wkt (p, 8,
		 "ic_2SP\"],PARAMETER[\"standard_parallel_1\",43],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"standard_parallel_2\",40],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 10,
		 "igin\",39.83333333333334],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 11,
		 ",-100],PARAMETER[\"false_easting\",1640416.6667],PARAMET");
    add_srs_wkt (p, 12,
		 "ER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"26852\"],A");
    add_srs_wkt (p, 13, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26853, "epsg", 26853,
		      "NAD83 / West Virginia North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.25 +lat_2=39 +lat_0=38.5 +lon_0=-79.");
    add_proj4text (p, 1,
		   "5 +x_0=600000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=us");
    add_proj4text (p, 2, "-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / West Virginia North (ftUS)\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GR");
    add_srs_wkt (p, 2,
		 "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4269\"]],UNIT[\"US survey foot\",0.30480060960121");
    add_srs_wkt (p, 7,
		 "92,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Co");
    add_srs_wkt (p, 8,
		 "nformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",4");
    add_srs_wkt (p, 9,
		 "0.25],PARAMETER[\"standard_parallel_2\",39],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "latitude_of_origin\",38.5],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 11,
		 ",-79.5],PARAMETER[\"false_easting\",1968500],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"26853\"],AXIS[\"");
    add_srs_wkt (p, 13, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26854, "epsg", 26854,
		      "NAD83 / West Virginia South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.88333333333333 +lat_2=37.48333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=37 +lon_0=-81 +x_0=600000 +y_0=0 +ellps=GRS80");
    add_proj4text (p, 2, " +datum=NAD83 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / West Virginia South (ftUS)\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "AD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GR");
    add_srs_wkt (p, 2,
		 "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4269\"]],UNIT[\"US survey foot\",0.30480060960121");
    add_srs_wkt (p, 7,
		 "92,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Co");
    add_srs_wkt (p, 8,
		 "nformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",3");
    add_srs_wkt (p, 9,
		 "8.88333333333333],PARAMETER[\"standard_parallel_2\",37.4");
    add_srs_wkt (p, 10,
		 "8333333333333],PARAMETER[\"latitude_of_origin\",37],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"central_meridian\",-81],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 12,
		 "g\",1968500],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13,
		 "EPSG\",\"26854\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26855, "epsg", 26855,
		      "NAD83(HARN) / Maine East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=43.66666666666666 +lon_0=-68.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=300000.0000000001 +y_0=0 +ellps=GRS80 +units=us");
    add_proj4text (p, 2, "-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Maine East (ftUS)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",0.30");
    add_srs_wkt (p, 7,
		 "48006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 9,
		 ",43.66666666666666],PARAMETER[\"central_meridian\",-68.5");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",984250.0000000002],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"26855\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 13, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26856, "epsg", 26856,
		      "NAD83(HARN) / Maine West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.83333333333334 +lon_0=-70.16666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999966667 +x_0=900000 +y_0=0 +ellps=GRS80 +u");
    add_proj4text (p, 2, "nits=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Maine West (ftUS)\",GEOGCS[\"NAD8");
    add_srs_wkt (p, 1,
		 "3(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",0.30");
    add_srs_wkt (p, 7,
		 "48006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 9,
		 ",42.83333333333334],PARAMETER[\"central_meridian\",-70.1");
    add_srs_wkt (p, 10,
		 "6666666666667],PARAMETER[\"scale_factor\",0.999966667],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_easting\",2952750],PARAMETER[\"false_no");
    add_srs_wkt (p, 12,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"26856\"],AXIS[\"X\",EAS");
    add_srs_wkt (p, 13, "T],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26857, "epsg", 26857,
		      "NAD83(HARN) / Minnesota North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.63333333333333 +lat_2=47.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=46.5 +lon_0=-93.09999999999999 +x_0=800000.00");
    add_proj4text (p, 2,
		   "00101599 +y_0=99999.99998983997 +ellps=GRS80 +units=us-f");
    add_proj4text (p, 3, "t +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Minnesota North (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",48.63333333333333],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",47.03333333333333],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",46.5],PARAMETER[\"central_meridian\",-93.1],PAR");
    add_srs_wkt (p, 12,
		 "AMETER[\"false_easting\",2624666.6667],PARAMETER[\"false");
    add_srs_wkt (p, 13,
		 "_northing\",328083.3333],AUTHORITY[\"EPSG\",\"26857\"],A");
    add_srs_wkt (p, 14, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26858, "epsg", 26858,
		      "NAD83(HARN) / Minnesota Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.05 +lat_2=45.61666666666667 +lat_0=4");
    add_proj4text (p, 1,
		   "5 +lon_0=-94.25 +x_0=800000.0000101599 +y_0=99999.999989");
    add_proj4text (p, 2, "83997 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Minnesota Central (ftUS)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Ne");
    add_srs_wkt (p, 2,
		 "twork\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],");
    add_srs_wkt (p, 4,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 5,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 6,
		 "22\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey fo");
    add_srs_wkt (p, 7,
		 "ot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 9,
		 "andard_parallel_1\",47.05],PARAMETER[\"standard_parallel");
    add_srs_wkt (p, 10,
		 "_2\",45.61666666666667],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 11,
		 ",45],PARAMETER[\"central_meridian\",-94.25],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_easting\",2624666.6667],PARAMETER[\"false_northing");
    add_srs_wkt (p, 13,
		 "\",328083.3333],AUTHORITY[\"EPSG\",\"26858\"],AXIS[\"X\"");
    add_srs_wkt (p, 14, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26859, "epsg", 26859,
		      "NAD83(HARN) / Minnesota South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.21666666666667 +lat_2=43.78333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=43 +lon_0=-94 +x_0=800000.0000101599 +y_0=999");
    add_proj4text (p, 2, "99.99998983997 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Minnesota South (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Netwo");
    add_srs_wkt (p, 2,
		 "rk\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRI");
    add_srs_wkt (p, 4,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\"");
    add_srs_wkt (p, 7,
		 ",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJEC");
    add_srs_wkt (p, 8,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_1\",45.21666666666667],PARAMETER[\"standard_");
    add_srs_wkt (p, 10,
		 "parallel_2\",43.78333333333333],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 11,
		 "origin\",43],PARAMETER[\"central_meridian\",-94],PARAMET");
    add_srs_wkt (p, 12,
		 "ER[\"false_easting\",2624666.6667],PARAMETER[\"false_nor");
    add_srs_wkt (p, 13,
		 "thing\",328083.3333],AUTHORITY[\"EPSG\",\"26859\"],AXIS[");
    add_srs_wkt (p, 14, "\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26860, "epsg", 26860,
		      "NAD83(HARN) / Nebraska (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43 +lat_2=40 +lat_0=39.83333333333334 +");
    add_proj4text (p, 1,
		   "lon_0=-100 +x_0=500000.00001016 +y_0=0 +ellps=GRS80 +uni");
    add_proj4text (p, 2, "ts=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / Nebraska (ftUS)\",GEOGCS[\"NAD83(");
    add_srs_wkt (p, 1,
		 "HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_Network\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey foot\",0.30");
    add_srs_wkt (p, 7,
		 "48006096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[");
    add_srs_wkt (p, 8,
		 "\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 9,
		 "rallel_1\",43],PARAMETER[\"standard_parallel_2\",40],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"latitude_of_origin\",39.83333333333334],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"central_meridian\",-100],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",1640416.6667],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 13,
		 "[\"EPSG\",\"26860\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    add_srs_wkt (p, 14, "");
    p = add_epsg_def (first, last, 26861, "epsg", 26861,
		      "NAD83(HARN) / West Virginia North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.25 +lat_2=39 +lat_0=38.5 +lon_0=-79.");
    add_proj4text (p, 1,
		   "5 +x_0=600000 +y_0=0 +ellps=GRS80 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / West Virginia North (ftUS)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_");
    add_srs_wkt (p, 2,
		 "Network\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey ");
    add_srs_wkt (p, 7,
		 "foot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",40.25],PARAMETER[\"standard_parall");
    add_srs_wkt (p, 10,
		 "el_2\",39],PARAMETER[\"latitude_of_origin\",38.5],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"central_meridian\",-79.5],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 12,
		 "g\",1968500],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13,
		 "EPSG\",\"26861\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26862, "epsg", 26862,
		      "NAD83(HARN) / West Virginia South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.88333333333333 +lat_2=37.48333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=37 +lon_0=-81 +x_0=600000 +y_0=0 +ellps=GRS80");
    add_proj4text (p, 2, " +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(HARN) / West Virginia South (ftUS)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NAD83(HARN)\",DATUM[\"NAD83_High_Accuracy_Regional_");
    add_srs_wkt (p, 2,
		 "Network\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6152\"]");
    add_srs_wkt (p, 4,
		 "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U");
    add_srs_wkt (p, 5,
		 "NIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4152\"]],UNIT[\"US survey ");
    add_srs_wkt (p, 7,
		 "foot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_1\",38.88333333333333],PARAMETER[\"sta");
    add_srs_wkt (p, 10,
		 "ndard_parallel_2\",37.48333333333333],PARAMETER[\"latitu");
    add_srs_wkt (p, 11,
		 "de_of_origin\",37],PARAMETER[\"central_meridian\",-81],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_easting\",1968500],PARAMETER[\"false_no");
    add_srs_wkt (p, 13,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"26862\"],AXIS[\"X\",EAS");
    add_srs_wkt (p, 14, "T],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26863, "epsg", 26863,
		      "NAD83(NSRS2007) / Maine East (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=43.66666666666666 +lon_0=-68.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=300000.0000000001 +y_0=0 +ellps=GRS80 +towgs84=");
    add_proj4text (p, 2, "0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Maine East (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 8,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 9,
		 "cator\"],PARAMETER[\"latitude_of_origin\",43.66666666666");
    add_srs_wkt (p, 10,
		 "666],PARAMETER[\"central_meridian\",-68.5],PARAMETER[\"s");
    add_srs_wkt (p, 11,
		 "cale_factor\",0.9999],PARAMETER[\"false_easting\",984250");
    add_srs_wkt (p, 12,
		 ".0000000002],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13,
		 "EPSG\",\"26863\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26864, "epsg", 26864,
		      "NAD83(NSRS2007) / Maine West (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.83333333333334 +lon_0=-70.16666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999966667 +x_0=900000 +y_0=0 +ellps=GRS80 +t");
    add_proj4text (p, 2, "owgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Maine West (ftUS)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Referen");
    add_srs_wkt (p, 2,
		 "ce_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.25722");
    add_srs_wkt (p, 3,
		 "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 8,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 9,
		 "cator\"],PARAMETER[\"latitude_of_origin\",42.83333333333");
    add_srs_wkt (p, 10,
		 "334],PARAMETER[\"central_meridian\",-70.16666666666667],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"scale_factor\",0.999966667],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_easting\",2952750],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 13,
		 "HORITY[\"EPSG\",\"26864\"],AXIS[\"X\",EAST],AXIS[\"Y\",N");
    add_srs_wkt (p, 14, "ORTH]]");
    p = add_epsg_def (first, last, 26865, "epsg", 26865,
		      "NAD83(NSRS2007) / Minnesota North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.63333333333333 +lat_2=47.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=46.5 +lon_0=-93.09999999999999 +x_0=800000.00");
    add_proj4text (p, 2,
		   "00101599 +y_0=99999.99998983997 +ellps=GRS80 +towgs84=0,");
    add_proj4text (p, 3, "0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Minnesota North (ftUS)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Re");
    add_srs_wkt (p, 2,
		 "ference_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.");
    add_srs_wkt (p, 3,
		 "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 9,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 10,
		 "48.63333333333333],PARAMETER[\"standard_parallel_2\",47.");
    add_srs_wkt (p, 11,
		 "03333333333333],PARAMETER[\"latitude_of_origin\",46.5],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"central_meridian\",-93.1],PARAMETER[\"false_e");
    add_srs_wkt (p, 13,
		 "asting\",2624666.6667],PARAMETER[\"false_northing\",3280");
    add_srs_wkt (p, 14,
		 "83.3333],AUTHORITY[\"EPSG\",\"26865\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 15, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26866, "epsg", 26866,
		      "NAD83(NSRS2007) / Minnesota Central (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.05 +lat_2=45.61666666666667 +lat_0=4");
    add_proj4text (p, 1,
		   "5 +lon_0=-94.25 +x_0=800000.0000101599 +y_0=99999.999989");
    add_proj4text (p, 2,
		   "83997 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=us-ft +");
    add_proj4text (p, 3, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Minnesota Central (ftUS)\",GE");
    add_srs_wkt (p, 1,
		 "OGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_");
    add_srs_wkt (p, 2,
		 "Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137,29");
    add_srs_wkt (p, 3,
		 "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 5,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 6,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 7,
		 "\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096");
    add_srs_wkt (p, 8,
		 "012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lamber");
    add_srs_wkt (p, 9,
		 "t_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1");
    add_srs_wkt (p, 10,
		 "\",47.05],PARAMETER[\"standard_parallel_2\",45.616666666");
    add_srs_wkt (p, 11,
		 "66667],PARAMETER[\"latitude_of_origin\",45],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "central_meridian\",-94.25],PARAMETER[\"false_easting\",2");
    add_srs_wkt (p, 13,
		 "624666.6667],PARAMETER[\"false_northing\",328083.3333],A");
    add_srs_wkt (p, 14,
		 "UTHORITY[\"EPSG\",\"26866\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 15, ",NORTH]]");
    p = add_epsg_def (first, last, 26867, "epsg", 26867,
		      "NAD83(NSRS2007) / Minnesota South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.21666666666667 +lat_2=43.78333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=43 +lon_0=-94 +x_0=800000.0000101599 +y_0=999");
    add_proj4text (p, 2,
		   "99.99998983997 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +unit");
    add_proj4text (p, 3, "s=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Minnesota South (ftUS)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Re");
    add_srs_wkt (p, 2,
		 "ference_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.");
    add_srs_wkt (p, 3,
		 "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 5,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 6,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 7,
		 "PSG\",\"4759\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 8,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 9,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 10,
		 "45.21666666666667],PARAMETER[\"standard_parallel_2\",43.");
    add_srs_wkt (p, 11,
		 "78333333333333],PARAMETER[\"latitude_of_origin\",43],PAR");
    add_srs_wkt (p, 12,
		 "AMETER[\"central_meridian\",-94],PARAMETER[\"false_easti");
    add_srs_wkt (p, 13,
		 "ng\",2624666.6667],PARAMETER[\"false_northing\",328083.3");
    add_srs_wkt (p, 14,
		 "333],AUTHORITY[\"EPSG\",\"26867\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 15, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26868, "epsg", 26868,
		      "NAD83(NSRS2007) / Nebraska (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43 +lat_2=40 +lat_0=39.83333333333334 +");
    add_proj4text (p, 1,
		   "lon_0=-100 +x_0=500000.00001016 +y_0=0 +ellps=GRS80 +tow");
    add_proj4text (p, 2, "gs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / Nebraska (ftUS)\",GEOGCS[\"NA");
    add_srs_wkt (p, 1,
		 "D83(NSRS2007)\",DATUM[\"NAD83_National_Spatial_Reference");
    add_srs_wkt (p, 2,
		 "_System_2007\",SPHEROID[\"GRS 1980\",6378137,298.2572221");
    add_srs_wkt (p, 3,
		 "01,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 6,
		 "994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 7,
		 "\"4759\"]],UNIT[\"US survey foot\",0.3048006096012192,AU");
    add_srs_wkt (p, 8,
		 "THORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conform");
    add_srs_wkt (p, 9,
		 "al_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",43],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"standard_parallel_2\",40],PARAMETER[\"latitude");
    add_srs_wkt (p, 11,
		 "_of_origin\",39.83333333333334],PARAMETER[\"central_meri");
    add_srs_wkt (p, 12,
		 "dian\",-100],PARAMETER[\"false_easting\",1640416.6667],P");
    add_srs_wkt (p, 13,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2686");
    add_srs_wkt (p, 14, "8\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26869, "epsg", 26869,
		      "NAD83(NSRS2007) / West Virginia North (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.25 +lat_2=39 +lat_0=38.5 +lon_0=-79.");
    add_proj4text (p, 1,
		   "5 +x_0=600000 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0");
    add_proj4text (p, 2, " +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / West Virginia North (ftUS)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatia");
    add_srs_wkt (p, 2,
		 "l_Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137,");
    add_srs_wkt (p, 3,
		 "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 5,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 6,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.30480060");
    add_srs_wkt (p, 8,
		 "96012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lamb");
    add_srs_wkt (p, 9,
		 "ert_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel");
    add_srs_wkt (p, 10,
		 "_1\",40.25],PARAMETER[\"standard_parallel_2\",39],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"latitude_of_origin\",38.5],PARAMETER[\"central_mer");
    add_srs_wkt (p, 12,
		 "idian\",-79.5],PARAMETER[\"false_easting\",1968500],PARA");
    add_srs_wkt (p, 13,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"26869\"");
    add_srs_wkt (p, 14, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26870, "epsg", 26870,
		      "NAD83(NSRS2007) / West Virginia South (ftUS)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.88333333333333 +lat_2=37.48333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=37 +lon_0=-81 +x_0=600000 +y_0=0 +ellps=GRS80");
    add_proj4text (p, 2, " +towgs84=0,0,0,0,0,0,0 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(NSRS2007) / West Virginia South (ftUS)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83_National_Spatia");
    add_srs_wkt (p, 2,
		 "l_Reference_System_2007\",SPHEROID[\"GRS 1980\",6378137,");
    add_srs_wkt (p, 3,
		 "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,");
    add_srs_wkt (p, 4,
		 "0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 5,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 6,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"4759\"]],UNIT[\"US survey foot\",0.30480060");
    add_srs_wkt (p, 8,
		 "96012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lamb");
    add_srs_wkt (p, 9,
		 "ert_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel");
    add_srs_wkt (p, 10,
		 "_1\",38.88333333333333],PARAMETER[\"standard_parallel_2\"");
    add_srs_wkt (p, 11,
		 ",37.48333333333333],PARAMETER[\"latitude_of_origin\",37]");
    add_srs_wkt (p, 12,
		 ",PARAMETER[\"central_meridian\",-81],PARAMETER[\"false_e");
    add_srs_wkt (p, 13,
		 "asting\",1968500],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 14,
		 "ITY[\"EPSG\",\"26870\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORT");
    add_srs_wkt (p, 15, "H]]");
    p = add_epsg_def (first, last, 26891, "epsg", 26891,
		      "NAD83(CSRS) / MTM zone 11");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-82.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / MTM zone 11\",GEOGCS[\"NAD83(CSRS");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-82.5],PARAMETER[\"scale_factor\",0.9999],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_easting\",304800],PARAMETER[\"false_northin");
    add_srs_wkt (p, 11,
		 "g\",0],AUTHORITY[\"EPSG\",\"26891\"],AXIS[\"Easting\",EA");
    add_srs_wkt (p, 12, "ST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 26892, "epsg", 26892,
		      "NAD83(CSRS) / MTM zone 12");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-81 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / MTM zone 12\",GEOGCS[\"NAD83(CSRS");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-81],PARAMETER[\"scale_factor\",0.9999],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",304800],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"26892\"],AXIS[\"Easting\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 26893, "epsg", 26893,
		      "NAD83(CSRS) / MTM zone 13");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-84 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / MTM zone 13\",GEOGCS[\"NAD83(CSRS");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-84],PARAMETER[\"scale_factor\",0.9999],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",304800],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"26893\"],AXIS[\"Easting\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 26894, "epsg", 26894,
		      "NAD83(CSRS) / MTM zone 14");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-87 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / MTM zone 14\",GEOGCS[\"NAD83(CSRS");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-87],PARAMETER[\"scale_factor\",0.9999],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",304800],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"26894\"],AXIS[\"Easting\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 26895, "epsg", 26895,
		      "NAD83(CSRS) / MTM zone 15");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-90 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / MTM zone 15\",GEOGCS[\"NAD83(CSRS");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-90],PARAMETER[\"scale_factor\",0.9999],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",304800],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"26895\"],AXIS[\"Easting\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 26896, "epsg", 26896,
		      "NAD83(CSRS) / MTM zone 16");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-93 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / MTM zone 16\",GEOGCS[\"NAD83(CSRS");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-93],PARAMETER[\"scale_factor\",0.9999],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",304800],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"26896\"],AXIS[\"Easting\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 26897, "epsg", 26897,
		      "NAD83(CSRS) / MTM zone 17");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-96 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / MTM zone 17\",GEOGCS[\"NAD83(CSRS");
    add_srs_wkt (p, 1,
		 ")\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SP");
    add_srs_wkt (p, 2,
		 "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",-96],PARAMETER[\"scale_factor\",0.9999],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",304800],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"26897\"],AXIS[\"Easting\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 26898, "epsg", 26898,
		      "NAD83(CSRS) / MTM zone 1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-53 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / MTM zone 1\",GEOGCS[\"NAD83(CSRS)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-53],PARAMETER[\"scale_factor\",0.9999],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",304800],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"26898\"],AXIS[\"E(X)\",EAST],AXI");
    add_srs_wkt (p, 12, "S[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 26899, "epsg", 26899,
		      "NAD83(CSRS) / MTM zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-56 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83(CSRS) / MTM zone 2\",GEOGCS[\"NAD83(CSRS)");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"NAD83_Canadian_Spatial_Reference_System\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7019\"]],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4617\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",0],PARAMETER[\"central_meri");
    add_srs_wkt (p, 9,
		 "dian\",-56],PARAMETER[\"scale_factor\",0.9999],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",304800],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"26899\"],AXIS[\"E(X)\",EAST],AXI");
    add_srs_wkt (p, 12, "S[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 26901, "epsg", 26901, "NAD83 / UTM zone 1N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=1 +ellps=GRS80 +datum=NAD83 +units=m +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 1N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-177],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26901\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26902, "epsg", 26902, "NAD83 / UTM zone 2N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=2 +ellps=GRS80 +datum=NAD83 +units=m +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 2N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-171],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26902\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26903, "epsg", 26903, "NAD83 / UTM zone 3N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=3 +ellps=GRS80 +datum=NAD83 +units=m +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 3N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-165],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26903\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26904, "epsg", 26904, "NAD83 / UTM zone 4N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=4 +ellps=GRS80 +datum=NAD83 +units=m +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 4N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-159],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26904\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26905, "epsg", 26905, "NAD83 / UTM zone 5N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=5 +ellps=GRS80 +datum=NAD83 +units=m +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 5N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-153],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26905\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26906, "epsg", 26906, "NAD83 / UTM zone 6N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=6 +ellps=GRS80 +datum=NAD83 +units=m +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 6N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-147],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26906\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26907, "epsg", 26907, "NAD83 / UTM zone 7N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=7 +ellps=GRS80 +datum=NAD83 +units=m +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 7N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-141],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26907\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26908, "epsg", 26908, "NAD83 / UTM zone 8N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=8 +ellps=GRS80 +datum=NAD83 +units=m +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 8N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-135],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26908\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26909, "epsg", 26909, "NAD83 / UTM zone 9N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=9 +ellps=GRS80 +datum=NAD83 +units=m +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 9N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-129],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26909\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26910, "epsg", 26910,
		      "NAD83 / UTM zone 10N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=10 +ellps=GRS80 +datum=NAD83 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 10N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-123],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26910\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26911, "epsg", 26911,
		      "NAD83 / UTM zone 11N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=11 +ellps=GRS80 +datum=NAD83 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 11N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-117],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26911\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26912, "epsg", 26912,
		      "NAD83 / UTM zone 12N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=12 +ellps=GRS80 +datum=NAD83 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 12N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-111],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26912\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26913, "epsg", 26913,
		      "NAD83 / UTM zone 13N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=13 +ellps=GRS80 +datum=NAD83 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 13N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-105],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26913\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26914, "epsg", 26914,
		      "NAD83 / UTM zone 14N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=14 +ellps=GRS80 +datum=NAD83 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 14N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-99],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26914\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26915, "epsg", 26915,
		      "NAD83 / UTM zone 15N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=15 +ellps=GRS80 +datum=NAD83 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 15N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-93],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26915\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26916, "epsg", 26916,
		      "NAD83 / UTM zone 16N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=16 +ellps=GRS80 +datum=NAD83 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 16N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-87],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26916\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26917, "epsg", 26917,
		      "NAD83 / UTM zone 17N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=17 +ellps=GRS80 +datum=NAD83 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 17N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-81],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26917\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26918, "epsg", 26918,
		      "NAD83 / UTM zone 18N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +ellps=GRS80 +datum=NAD83 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 18N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-75],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26918\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26919, "epsg", 26919,
		      "NAD83 / UTM zone 19N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +ellps=GRS80 +datum=NAD83 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 19N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-69],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26919\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26920, "epsg", 26920,
		      "NAD83 / UTM zone 20N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +ellps=GRS80 +datum=NAD83 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 20N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-63],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26920\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26921, "epsg", 26921,
		      "NAD83 / UTM zone 21N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +ellps=GRS80 +datum=NAD83 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 21N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-57],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26921\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26922, "epsg", 26922,
		      "NAD83 / UTM zone 22N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=22 +ellps=GRS80 +datum=NAD83 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 22N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-51],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26922\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26923, "epsg", 26923,
		      "NAD83 / UTM zone 23N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=23 +ellps=GRS80 +datum=NAD83 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / UTM zone 23N\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-45],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "26923\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 26929, "epsg", 26929,
		      "NAD83 / Alabama East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30.5 +lon_0=-85.83333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "9996 +x_0=200000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units");
    add_proj4text (p, 2, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alabama East\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",30.5],PARAMETER[\"central_meridian\",-85.8333333333");
    add_srs_wkt (p, 9,
		 "3333],PARAMETER[\"scale_factor\",0.99996],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_easting\",200000],PARAMETER[\"false_northing\",0],AU");
    add_srs_wkt (p, 11,
		 "THORITY[\"EPSG\",\"26929\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 12, "NORTH]]");
    p = add_epsg_def (first, last, 26930, "epsg", 26930,
		      "NAD83 / Alabama West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=-87.5 +k=0.999933333 +x_0=6");
    add_proj4text (p, 1,
		   "00000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_proj4text (p, 2, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alabama West\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",30],PARAMETER[\"central_meridian\",-87.5],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",0.999933333],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",600000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 11, "PSG\",\"26930\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26931, "epsg", 26931,
		      "NAD83 / Alaska zone 1");
    add_proj4text (p, 0,
		   "+proj=omerc +lat_0=57 +lonc=-133.6666666666667 +alpha=32");
    add_proj4text (p, 1,
		   "3.1301023611111 +k=0.9999 +x_0=5000000 +y_0=-5000000 +el");
    add_proj4text (p, 2, "lps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alaska zone 1\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Hotine_Oblique_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_center\",57],PARAMETER[\"longitude_of_center\",-133.666");
    add_srs_wkt (p, 9,
		 "6666666667],PARAMETER[\"azimuth\",323.1301023611111],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"rectified_grid_angle\",323.1301023611111],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"scale_factor\",0.9999],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",5000000],PARAMETER[\"false_northing\",-5000000],AUTHORI");
    add_srs_wkt (p, 13,
		 "TY[\"EPSG\",\"26931\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH");
    add_srs_wkt (p, 14, "]]");
    p = add_epsg_def (first, last, 26932, "epsg", 26932,
		      "NAD83 / Alaska zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-142 +k=0.9999 +x_0=500000 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alaska zone 2\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",54],PARAMETER[\"central_meridian\",-142],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 10,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"26932\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26933, "epsg", 26933,
		      "NAD83 / Alaska zone 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-146 +k=0.9999 +x_0=500000 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alaska zone 3\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",54],PARAMETER[\"central_meridian\",-146],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 10,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"26933\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26934, "epsg", 26934,
		      "NAD83 / Alaska zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-150 +k=0.9999 +x_0=500000 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alaska zone 4\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",54],PARAMETER[\"central_meridian\",-150],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 10,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"26934\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26935, "epsg", 26935,
		      "NAD83 / Alaska zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-154 +k=0.9999 +x_0=500000 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alaska zone 5\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",54],PARAMETER[\"central_meridian\",-154],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 10,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"26935\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26936, "epsg", 26936,
		      "NAD83 / Alaska zone 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-158 +k=0.9999 +x_0=500000 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alaska zone 6\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",54],PARAMETER[\"central_meridian\",-158],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 10,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"26936\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26937, "epsg", 26937,
		      "NAD83 / Alaska zone 7");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-162 +k=0.9999 +x_0=500000 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alaska zone 7\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",54],PARAMETER[\"central_meridian\",-162],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 10,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"26937\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_22 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 26938, "epsg", 26938,
		      "NAD83 / Alaska zone 8");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-166 +k=0.9999 +x_0=500000 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alaska zone 8\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",54],PARAMETER[\"central_meridian\",-166],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 10,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"26938\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26939, "epsg", 26939,
		      "NAD83 / Alaska zone 9");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=54 +lon_0=-170 +k=0.9999 +x_0=500000 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alaska zone 9\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",54],PARAMETER[\"central_meridian\",-170],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 10,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"26939\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26940, "epsg", 26940,
		      "NAD83 / Alaska zone 10");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=53.83333333333334 +lat_2=51.83333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=51 +lon_0=-176 +x_0=1000000 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Alaska zone 10\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 8,
		 "rd_parallel_1\",53.83333333333334],PARAMETER[\"standard_");
    add_srs_wkt (p, 9,
		 "parallel_2\",51.83333333333334],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 10,
		 "origin\",51],PARAMETER[\"central_meridian\",-176],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_easting\",1000000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",0],AUTHORITY[\"EPSG\",\"26940\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 13, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26941, "epsg", 26941,
		      "NAD83 / California zone 1");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.66666666666666 +lat_2=40 +lat_0=39.3");
    add_proj4text (p, 1,
		   "3333333333334 +lon_0=-122 +x_0=2000000 +y_0=500000 +ellp");
    add_proj4text (p, 2, "s=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / California zone 1\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 7,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 8,
		 "andard_parallel_1\",41.66666666666666],PARAMETER[\"stand");
    add_srs_wkt (p, 9,
		 "ard_parallel_2\",40],PARAMETER[\"latitude_of_origin\",39");
    add_srs_wkt (p, 10,
		 ".33333333333334],PARAMETER[\"central_meridian\",-122],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_easting\",2000000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 12,
		 "thing\",500000],AUTHORITY[\"EPSG\",\"26941\"],AXIS[\"X\"");
    add_srs_wkt (p, 13, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26942, "epsg", 26942,
		      "NAD83 / California zone 2");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.83333333333334 +lat_2=38.33333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=37.66666666666666 +lon_0=-122 +x_0=2000000 +y");
    add_proj4text (p, 2,
		   "_0=500000 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / California zone 2\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 7,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 8,
		 "andard_parallel_1\",39.83333333333334],PARAMETER[\"stand");
    add_srs_wkt (p, 9,
		 "ard_parallel_2\",38.33333333333334],PARAMETER[\"latitude");
    add_srs_wkt (p, 10,
		 "_of_origin\",37.66666666666666],PARAMETER[\"central_meri");
    add_srs_wkt (p, 11,
		 "dian\",-122],PARAMETER[\"false_easting\",2000000],PARAME");
    add_srs_wkt (p, 12,
		 "TER[\"false_northing\",500000],AUTHORITY[\"EPSG\",\"2694");
    add_srs_wkt (p, 13, "2\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26943, "epsg", 26943,
		      "NAD83 / California zone 3");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.43333333333333 +lat_2=37.06666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.5 +lon_0=-120.5 +x_0=2000000 +y_0=500000 +");
    add_proj4text (p, 2, "ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / California zone 3\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 7,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 8,
		 "andard_parallel_1\",38.43333333333333],PARAMETER[\"stand");
    add_srs_wkt (p, 9,
		 "ard_parallel_2\",37.06666666666667],PARAMETER[\"latitude");
    add_srs_wkt (p, 10,
		 "_of_origin\",36.5],PARAMETER[\"central_meridian\",-120.5");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"false_easting\",2000000],PARAMETER[\"false");
    add_srs_wkt (p, 12,
		 "_northing\",500000],AUTHORITY[\"EPSG\",\"26943\"],AXIS[\"");
    add_srs_wkt (p, 13, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26944, "epsg", 26944,
		      "NAD83 / California zone 4");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.25 +lat_2=36 +lat_0=35.3333333333333");
    add_proj4text (p, 1,
		   "4 +lon_0=-119 +x_0=2000000 +y_0=500000 +ellps=GRS80 +dat");
    add_proj4text (p, 2, "um=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / California zone 4\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 7,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 8,
		 "andard_parallel_1\",37.25],PARAMETER[\"standard_parallel");
    add_srs_wkt (p, 9,
		 "_2\",36],PARAMETER[\"latitude_of_origin\",35.33333333333");
    add_srs_wkt (p, 10,
		 "334],PARAMETER[\"central_meridian\",-119],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_easting\",2000000],PARAMETER[\"false_northing\",5000");
    add_srs_wkt (p, 12,
		 "00],AUTHORITY[\"EPSG\",\"26944\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 13, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26945, "epsg", 26945,
		      "NAD83 / California zone 5");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=35.46666666666667 +lat_2=34.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=33.5 +lon_0=-118 +x_0=2000000 +y_0=500000 +el");
    add_proj4text (p, 2, "lps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / California zone 5\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 7,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 8,
		 "andard_parallel_1\",35.46666666666667],PARAMETER[\"stand");
    add_srs_wkt (p, 9,
		 "ard_parallel_2\",34.03333333333333],PARAMETER[\"latitude");
    add_srs_wkt (p, 10,
		 "_of_origin\",33.5],PARAMETER[\"central_meridian\",-118],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_easting\",2000000],PARAMETER[\"false_n");
    add_srs_wkt (p, 12,
		 "orthing\",500000],AUTHORITY[\"EPSG\",\"26945\"],AXIS[\"X");
    add_srs_wkt (p, 13, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26946, "epsg", 26946,
		      "NAD83 / California zone 6");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.88333333333333 +lat_2=32.78333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=32.16666666666666 +lon_0=-116.25 +x_0=2000000");
    add_proj4text (p, 2,
		   " +y_0=500000 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_proj4text (p, 3, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / California zone 6\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 7,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 8,
		 "andard_parallel_1\",33.88333333333333],PARAMETER[\"stand");
    add_srs_wkt (p, 9,
		 "ard_parallel_2\",32.78333333333333],PARAMETER[\"latitude");
    add_srs_wkt (p, 10,
		 "_of_origin\",32.16666666666666],PARAMETER[\"central_meri");
    add_srs_wkt (p, 11,
		 "dian\",-116.25],PARAMETER[\"false_easting\",2000000],PAR");
    add_srs_wkt (p, 12,
		 "AMETER[\"false_northing\",500000],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 13, "6946\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26948, "epsg", 26948,
		      "NAD83 / Arizona East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-110.1666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=213360 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Arizona East\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",31],PARAMETER[\"central_meridian\",-110.16666666666");
    add_srs_wkt (p, 9,
		 "67],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",213360],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 11,
		 "RITY[\"EPSG\",\"26948\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 12, "TH]]");
    p = add_epsg_def (first, last, 26949, "epsg", 26949,
		      "NAD83 / Arizona Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-111.9166666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=213360 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Arizona Central\",GEOGCS[\"NAD83\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",63");
    add_srs_wkt (p, 2,
		 "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",31],PARAMETER[\"central_meridian\",-111.9166666");
    add_srs_wkt (p, 9,
		 "666667],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",213360],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 11,
		 "UTHORITY[\"EPSG\",\"26949\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 26950, "epsg", 26950,
		      "NAD83 / Arizona West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-113.75 +k=0.999933333 +x_0");
    add_proj4text (p, 1,
		   "=213360 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_de");
    add_proj4text (p, 2, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Arizona West\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",31],PARAMETER[\"central_meridian\",-113.75],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",0.999933333],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",213360],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"26950\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26951, "epsg", 26951,
		      "NAD83 / Arkansas North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.23333333333333 +lat_2=34.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=34.33333333333334 +lon_0=-92 +x_0=400000 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Arkansas North\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 8,
		 "rd_parallel_1\",36.23333333333333],PARAMETER[\"standard_");
    add_srs_wkt (p, 9,
		 "parallel_2\",34.93333333333333],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 10,
		 "origin\",34.33333333333334],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 11,
		 "\",-92],PARAMETER[\"false_easting\",400000],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"26951\"],AXIS[\"");
    add_srs_wkt (p, 13, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26952, "epsg", 26952,
		      "NAD83 / Arkansas South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=34.76666666666667 +lat_2=33.3 +lat_0=32");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-92 +x_0=400000 +y_0=400000 +ellp");
    add_proj4text (p, 2, "s=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Arkansas South\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 8,
		 "rd_parallel_1\",34.76666666666667],PARAMETER[\"standard_");
    add_srs_wkt (p, 9,
		 "parallel_2\",33.3],PARAMETER[\"latitude_of_origin\",32.6");
    add_srs_wkt (p, 10,
		 "6666666666666],PARAMETER[\"central_meridian\",-92],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_easting\",400000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",400000],AUTHORITY[\"EPSG\",\"26952\"],AXIS[\"X\",EAS");
    add_srs_wkt (p, 13, "T],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26953, "epsg", 26953,
		      "NAD83 / Colorado North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.78333333333333 +lat_2=39.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=39.33333333333334 +lon_0=-105.5 +x_0=914401.8");
    add_proj4text (p, 2,
		   "289 +y_0=304800.6096 +ellps=GRS80 +datum=NAD83 +units=m ");
    add_proj4text (p, 3, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Colorado North\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 8,
		 "rd_parallel_1\",40.78333333333333],PARAMETER[\"standard_");
    add_srs_wkt (p, 9,
		 "parallel_2\",39.71666666666667],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 10,
		 "origin\",39.33333333333334],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 11,
		 "\",-105.5],PARAMETER[\"false_easting\",914401.8289],PARA");
    add_srs_wkt (p, 12,
		 "METER[\"false_northing\",304800.6096],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 13, ",\"26953\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26954, "epsg", 26954,
		      "NAD83 / Colorado Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.75 +lat_2=38.45 +lat_0=37.8333333333");
    add_proj4text (p, 1,
		   "3334 +lon_0=-105.5 +x_0=914401.8289 +y_0=304800.6096 +el");
    add_proj4text (p, 2, "lps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Colorado Central\",GEOGCS[\"NAD83\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6");
    add_srs_wkt (p, 2,
		 "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"426");
    add_srs_wkt (p, 6,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"sta");
    add_srs_wkt (p, 8,
		 "ndard_parallel_1\",39.75],PARAMETER[\"standard_parallel_");
    add_srs_wkt (p, 9,
		 "2\",38.45],PARAMETER[\"latitude_of_origin\",37.833333333");
    add_srs_wkt (p, 10,
		 "33334],PARAMETER[\"central_meridian\",-105.5],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_easting\",914401.8289],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",304800.6096],AUTHORITY[\"EPSG\",\"26954\"],AXIS[\"X\"");
    add_srs_wkt (p, 13, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26955, "epsg", 26955,
		      "NAD83 / Colorado South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.43333333333333 +lat_2=37.23333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=36.66666666666666 +lon_0=-105.5 +x_0=914401.8");
    add_proj4text (p, 2,
		   "289 +y_0=304800.6096 +ellps=GRS80 +datum=NAD83 +units=m ");
    add_proj4text (p, 3, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Colorado South\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 8,
		 "rd_parallel_1\",38.43333333333333],PARAMETER[\"standard_");
    add_srs_wkt (p, 9,
		 "parallel_2\",37.23333333333333],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 10,
		 "origin\",36.66666666666666],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 11,
		 "\",-105.5],PARAMETER[\"false_easting\",914401.8289],PARA");
    add_srs_wkt (p, 12,
		 "METER[\"false_northing\",304800.6096],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 13, ",\"26955\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26956, "epsg", 26956, "NAD83 / Connecticut");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.86666666666667 +lat_2=41.2 +lat_0=40");
    add_proj4text (p, 1,
		   ".83333333333334 +lon_0=-72.75 +x_0=304800.6096 +y_0=1524");
    add_proj4text (p, 2, "00.3048 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Connecticut\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard");
    add_srs_wkt (p, 8,
		 "_parallel_1\",41.86666666666667],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 9,
		 "rallel_2\",41.2],PARAMETER[\"latitude_of_origin\",40.833");
    add_srs_wkt (p, 10,
		 "33333333334],PARAMETER[\"central_meridian\",-72.75],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_easting\",304800.6096],PARAMETER[\"false_n");
    add_srs_wkt (p, 12,
		 "orthing\",152400.3048],AUTHORITY[\"EPSG\",\"26956\"],AXI");
    add_srs_wkt (p, 13, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26957, "epsg", 26957, "NAD83 / Delaware");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38 +lon_0=-75.41666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "995 +x_0=200000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Delaware\",GEOGCS[\"NAD83\",DATUM[\"Nor");
    add_srs_wkt (p, 1,
		 "th_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137,2");
    add_srs_wkt (p, 2,
		 "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",38],PARAMETER[\"central_meridian\",-75.41666666666667],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"scale_factor\",0.999995],PARAMETER[\"false_e");
    add_srs_wkt (p, 10,
		 "asting\",200000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 11,
		 "TY[\"EPSG\",\"26957\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH");
    add_srs_wkt (p, 12, "]]");
    p = add_epsg_def (first, last, 26958, "epsg", 26958,
		      "NAD83 / Florida East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=24.33333333333333 +lon_0=-81 +k=0.999");
    add_proj4text (p, 1,
		   "941177 +x_0=200000 +y_0=0 +ellps=GRS80 +datum=NAD83 +uni");
    add_proj4text (p, 2, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Florida East\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",24.33333333333333],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 9,
		 "81],PARAMETER[\"scale_factor\",0.999941177],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",200000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"26958\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 26959, "epsg", 26959,
		      "NAD83 / Florida West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=24.33333333333333 +lon_0=-82 +k=0.999");
    add_proj4text (p, 1,
		   "941177 +x_0=200000 +y_0=0 +ellps=GRS80 +datum=NAD83 +uni");
    add_proj4text (p, 2, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Florida West\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",24.33333333333333],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 9,
		 "82],PARAMETER[\"scale_factor\",0.999941177],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",200000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"26959\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 26960, "epsg", 26960,
		      "NAD83 / Florida North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.75 +lat_2=29.58333333333333 +lat_0=2");
    add_proj4text (p, 1,
		   "9 +lon_0=-84.5 +x_0=600000 +y_0=0 +ellps=GRS80 +datum=NA");
    add_proj4text (p, 2, "D83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Florida North\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standar");
    add_srs_wkt (p, 8,
		 "d_parallel_1\",30.75],PARAMETER[\"standard_parallel_2\",");
    add_srs_wkt (p, 9,
		 "29.58333333333333],PARAMETER[\"latitude_of_origin\",29],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"central_meridian\",-84.5],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",600000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 12,
		 "ITY[\"EPSG\",\"26960\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORT");
    add_srs_wkt (p, 13, "H]]");
    p = add_epsg_def (first, last, 26961, "epsg", 26961,
		      "NAD83 / Hawaii zone 1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=18.83333333333333 +lon_0=-155.5 +k=0.");
    add_proj4text (p, 1,
		   "999966667 +x_0=500000 +y_0=0 +ellps=GRS80 +datum=NAD83 +");
    add_proj4text (p, 2, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Hawaii zone 1\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",18.83333333333333],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 9,
		 "-155.5],PARAMETER[\"scale_factor\",0.999966667],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"26961\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 12, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26962, "epsg", 26962,
		      "NAD83 / Hawaii zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=20.33333333333333 +lon_0=-156.6666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999966667 +x_0=500000 +y_0=0 +ellps=GRS80 +d");
    add_proj4text (p, 2, "atum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Hawaii zone 2\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",20.33333333333333],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 9,
		 "-156.6666666666667],PARAMETER[\"scale_factor\",0.9999666");
    add_srs_wkt (p, 10,
		 "67],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"26962\"],AXIS[\"X\"");
    add_srs_wkt (p, 12, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26963, "epsg", 26963,
		      "NAD83 / Hawaii zone 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=21.16666666666667 +lon_0=-158 +k=0.99");
    add_proj4text (p, 1,
		   "999 +x_0=500000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Hawaii zone 3\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",21.16666666666667],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 9,
		 "-158],PARAMETER[\"scale_factor\",0.99999],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_easting\",500000],PARAMETER[\"false_northing\",0],AU");
    add_srs_wkt (p, 11,
		 "THORITY[\"EPSG\",\"26963\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 12, "NORTH]]");
    p = add_epsg_def (first, last, 26964, "epsg", 26964,
		      "NAD83 / Hawaii zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=21.83333333333333 +lon_0=-159.5 +k=0.");
    add_proj4text (p, 1,
		   "99999 +x_0=500000 +y_0=0 +ellps=GRS80 +datum=NAD83 +unit");
    add_proj4text (p, 2, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Hawaii zone 4\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",21.83333333333333],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 9,
		 "-159.5],PARAMETER[\"scale_factor\",0.99999],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"26964\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 26965, "epsg", 26965,
		      "NAD83 / Hawaii zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=21.66666666666667 +lon_0=-160.1666666");
    add_proj4text (p, 1,
		   "666667 +k=1 +x_0=500000 +y_0=0 +ellps=GRS80 +datum=NAD83");
    add_proj4text (p, 2, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Hawaii zone 5\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",21.66666666666667],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 9,
		 "-160.1666666666667],PARAMETER[\"scale_factor\",1],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",500000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 11,
		 "\",0],AUTHORITY[\"EPSG\",\"26965\"],AXIS[\"X\",EAST],AXI");
    add_srs_wkt (p, 12, "S[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26966, "epsg", 26966,
		      "NAD83 / Georgia East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=-82.16666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=200000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Georgia East\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",30],PARAMETER[\"central_meridian\",-82.166666666666");
    add_srs_wkt (p, 9,
		 "67],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",200000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 11,
		 "RITY[\"EPSG\",\"26966\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 12, "TH]]");
    p = add_epsg_def (first, last, 26967, "epsg", 26967,
		      "NAD83 / Georgia West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=30 +lon_0=-84.16666666666667 +k=0.999");
    add_proj4text (p, 1,
		   "9 +x_0=700000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Georgia West\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",30],PARAMETER[\"central_meridian\",-84.166666666666");
    add_srs_wkt (p, 9,
		 "67],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",700000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 11,
		 "RITY[\"EPSG\",\"26967\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 12, "TH]]");
    p = add_epsg_def (first, last, 26968, "epsg", 26968, "NAD83 / Idaho East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-112.1666666");
    add_proj4text (p, 1,
		   "666667 +k=0.9999473679999999 +x_0=200000 +y_0=0 +ellps=G");
    add_proj4text (p, 2, "RS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Idaho East\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",41.66666666666666],PARAMETER[\"central_meridian\",-11");
    add_srs_wkt (p, 9,
		 "2.1666666666667],PARAMETER[\"scale_factor\",0.999947368]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_easting\",200000],PARAMETER[\"false_n");
    add_srs_wkt (p, 11,
		 "orthing\",0],AUTHORITY[\"EPSG\",\"26968\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 12, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26969, "epsg", 26969,
		      "NAD83 / Idaho Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-114 +k=0.99");
    add_proj4text (p, 1,
		   "99473679999999 +x_0=500000 +y_0=0 +ellps=GRS80 +datum=NA");
    add_proj4text (p, 2, "D83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Idaho Central\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",41.66666666666666],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 9,
		 "-114],PARAMETER[\"scale_factor\",0.999947368],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"26969\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 26970, "epsg", 26970, "NAD83 / Idaho West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.66666666666666 +lon_0=-115.75 +k=0");
    add_proj4text (p, 1,
		   ".999933333 +x_0=800000 +y_0=0 +ellps=GRS80 +datum=NAD83 ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Idaho West\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",41.66666666666666],PARAMETER[\"central_meridian\",-11");
    add_srs_wkt (p, 9,
		 "5.75],PARAMETER[\"scale_factor\",0.999933333],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",800000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"26970\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 26971, "epsg", 26971,
		      "NAD83 / Illinois East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36.66666666666666 +lon_0=-88.33333333");
    add_proj4text (p, 1,
		   "333333 +k=0.9999749999999999 +x_0=300000 +y_0=0 +ellps=G");
    add_proj4text (p, 2, "RS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Illinois East\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",36.66666666666666],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 9,
		 "-88.33333333333333],PARAMETER[\"scale_factor\",0.999975]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_easting\",300000],PARAMETER[\"false_n");
    add_srs_wkt (p, 11,
		 "orthing\",0],AUTHORITY[\"EPSG\",\"26971\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 12, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26972, "epsg", 26972,
		      "NAD83 / Illinois West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36.66666666666666 +lon_0=-90.16666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999941177 +x_0=700000 +y_0=0 +ellps=GRS80 +d");
    add_proj4text (p, 2, "atum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Illinois West\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",36.66666666666666],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 9,
		 "-90.16666666666667],PARAMETER[\"scale_factor\",0.9999411");
    add_srs_wkt (p, 10,
		 "77],PARAMETER[\"false_easting\",700000],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"26972\"],AXIS[\"X\"");
    add_srs_wkt (p, 12, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26973, "epsg", 26973,
		      "NAD83 / Indiana East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=37.5 +lon_0=-85.66666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=100000 +y_0=250000 +ellps=GRS80 +datum=NAD");
    add_proj4text (p, 2, "83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Indiana East\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",37.5],PARAMETER[\"central_meridian\",-85.6666666666");
    add_srs_wkt (p, 9,
		 "6667],PARAMETER[\"scale_factor\",0.999966667],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",100000],PARAMETER[\"false_northing\",2");
    add_srs_wkt (p, 11,
		 "50000],AUTHORITY[\"EPSG\",\"26973\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 12, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26974, "epsg", 26974,
		      "NAD83 / Indiana West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=37.5 +lon_0=-87.08333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=900000 +y_0=250000 +ellps=GRS80 +datum=NAD");
    add_proj4text (p, 2, "83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Indiana West\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",37.5],PARAMETER[\"central_meridian\",-87.0833333333");
    add_srs_wkt (p, 9,
		 "3333],PARAMETER[\"scale_factor\",0.999966667],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"false_easting\",900000],PARAMETER[\"false_northing\",2");
    add_srs_wkt (p, 11,
		 "50000],AUTHORITY[\"EPSG\",\"26974\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 12, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26975, "epsg", 26975, "NAD83 / Iowa North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43.26666666666667 +lat_2=42.06666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=41.5 +lon_0=-93.5 +x_0=1500000 +y_0=1000000 +");
    add_proj4text (p, 2, "ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Iowa North\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_p");
    add_srs_wkt (p, 8,
		 "arallel_1\",43.26666666666667],PARAMETER[\"standard_para");
    add_srs_wkt (p, 9,
		 "llel_2\",42.06666666666667],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",41.5],PARAMETER[\"central_meridian\",-93.5],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_easting\",1500000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 12,
		 "\",1000000],AUTHORITY[\"EPSG\",\"26975\"],AXIS[\"X\",EAS");
    add_srs_wkt (p, 13, "T],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26976, "epsg", 26976, "NAD83 / Iowa South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.78333333333333 +lat_2=40.61666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=40 +lon_0=-93.5 +x_0=500000 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Iowa South\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_p");
    add_srs_wkt (p, 8,
		 "arallel_1\",41.78333333333333],PARAMETER[\"standard_para");
    add_srs_wkt (p, 9,
		 "llel_2\",40.61666666666667],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",40],PARAMETER[\"central_meridian\",-93.5],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_easting\",500000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 12,
		 "0],AUTHORITY[\"EPSG\",\"26976\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 26977, "epsg", 26977,
		      "NAD83 / Kansas North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.78333333333333 +lat_2=38.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=38.33333333333334 +lon_0=-98 +x_0=400000 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Kansas North\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard");
    add_srs_wkt (p, 8,
		 "_parallel_1\",39.78333333333333],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 9,
		 "rallel_2\",38.71666666666667],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 10,
		 "igin\",38.33333333333334],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 11,
		 ",-98],PARAMETER[\"false_easting\",400000],PARAMETER[\"fa");
    add_srs_wkt (p, 12,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"26977\"],AXIS[\"X");
    add_srs_wkt (p, 13, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26978, "epsg", 26978,
		      "NAD83 / Kansas South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.56666666666667 +lat_2=37.26666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.66666666666666 +lon_0=-98.5 +x_0=400000 +y");
    add_proj4text (p, 2,
		   "_0=400000 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Kansas South\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard");
    add_srs_wkt (p, 8,
		 "_parallel_1\",38.56666666666667],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 9,
		 "rallel_2\",37.26666666666667],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 10,
		 "igin\",36.66666666666666],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 11,
		 ",-98.5],PARAMETER[\"false_easting\",400000],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",400000],AUTHORITY[\"EPSG\",\"26978\"],A");
    add_srs_wkt (p, 13, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26979, "epsg", 26979,
		      "NAD83 / Kentucky North (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.96666666666667 +lat_2=37.96666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=37.5 +lon_0=-84.25 +x_0=500000 +y_0=0 +ellps=");
    add_proj4text (p, 2, "GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Kentucky North (deprecated)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"G");
    add_srs_wkt (p, 2,
		 "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701");
    add_srs_wkt (p, 3,
		 "9\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"standard_parallel_1\",37.96666666666667],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"standard_parallel_2\",37.96666666666667],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"latitude_of_origin\",37.5],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 11,
		 "n\",-84.25],PARAMETER[\"false_easting\",500000],PARAMETE");
    add_srs_wkt (p, 12,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"26979\"],AX");
    add_srs_wkt (p, 13, "IS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26980, "epsg", 26980,
		      "NAD83 / Kentucky South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.93333333333333 +lat_2=36.73333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=36.33333333333334 +lon_0=-85.75 +x_0=500000 +");
    add_proj4text (p, 2,
		   "y_0=500000 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Kentucky South\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 8,
		 "rd_parallel_1\",37.93333333333333],PARAMETER[\"standard_");
    add_srs_wkt (p, 9,
		 "parallel_2\",36.73333333333333],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 10,
		 "origin\",36.33333333333334],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 11,
		 "\",-85.75],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 12,
		 "[\"false_northing\",500000],AUTHORITY[\"EPSG\",\"26980\"");
    add_srs_wkt (p, 13, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26981, "epsg", 26981,
		      "NAD83 / Louisiana North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=32.66666666666666 +lat_2=31.16666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=30.5 +lon_0=-92.5 +x_0=1000000 +y_0=0 +ellps=");
    add_proj4text (p, 2, "GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Louisiana North\",GEOGCS[\"NAD83\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",63");
    add_srs_wkt (p, 2,
		 "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"stan");
    add_srs_wkt (p, 8,
		 "dard_parallel_1\",32.66666666666666],PARAMETER[\"standar");
    add_srs_wkt (p, 9,
		 "d_parallel_2\",31.16666666666667],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 10,
		 "f_origin\",30.5],PARAMETER[\"central_meridian\",-92.5],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_easting\",1000000],PARAMETER[\"false_no");
    add_srs_wkt (p, 12,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"26981\"],AXIS[\"X\",EAS");
    add_srs_wkt (p, 13, "T],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26982, "epsg", 26982,
		      "NAD83 / Louisiana South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.7 +lat_2=29.3 +lat_0=28.5 +lon_0=-91");
    add_proj4text (p, 1,
		   ".33333333333333 +x_0=1000000 +y_0=0 +ellps=GRS80 +datum=");
    add_proj4text (p, 2, "NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Louisiana South\",GEOGCS[\"NAD83\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",63");
    add_srs_wkt (p, 2,
		 "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"stan");
    add_srs_wkt (p, 8,
		 "dard_parallel_1\",30.7],PARAMETER[\"standard_parallel_2\"");
    add_srs_wkt (p, 9,
		 ",29.3],PARAMETER[\"latitude_of_origin\",28.5],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"central_meridian\",-91.33333333333333],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_easting\",1000000],PARAMETER[\"false_northing\",0],AU");
    add_srs_wkt (p, 12,
		 "THORITY[\"EPSG\",\"26982\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 13, "NORTH]]");
    p = add_epsg_def (first, last, 26983, "epsg", 26983, "NAD83 / Maine East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=43.66666666666666 +lon_0=-68.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=300000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Maine East\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",43.66666666666666],PARAMETER[\"central_meridian\",-68");
    add_srs_wkt (p, 9,
		 ".5],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",300000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 11,
		 "RITY[\"EPSG\",\"26983\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 12, "TH]]");
    p = add_epsg_def (first, last, 26984, "epsg", 26984, "NAD83 / Maine West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.83333333333334 +lon_0=-70.16666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999966667 +x_0=900000 +y_0=0 +ellps=GRS80 +d");
    add_proj4text (p, 2, "atum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Maine West\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",42.83333333333334],PARAMETER[\"central_meridian\",-70");
    add_srs_wkt (p, 9,
		 ".16666666666667],PARAMETER[\"scale_factor\",0.999966667]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_easting\",900000],PARAMETER[\"false_n");
    add_srs_wkt (p, 11,
		 "orthing\",0],AUTHORITY[\"EPSG\",\"26984\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 12, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26985, "epsg", 26985, "NAD83 / Maryland");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.45 +lat_2=38.3 +lat_0=37.66666666666");
    add_proj4text (p, 1,
		   "666 +lon_0=-77 +x_0=400000 +y_0=0 +ellps=GRS80 +datum=NA");
    add_proj4text (p, 2, "D83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Maryland\",GEOGCS[\"NAD83\",DATUM[\"Nor");
    add_srs_wkt (p, 1,
		 "th_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137,2");
    add_srs_wkt (p, 2,
		 "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 8,
		 "rallel_1\",39.45],PARAMETER[\"standard_parallel_2\",38.3");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"latitude_of_origin\",37.66666666666666],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"central_meridian\",-77],PARAMETER[\"false_east");
    add_srs_wkt (p, 11,
		 "ing\",400000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 12,
		 "\"EPSG\",\"26985\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26986, "epsg", 26986,
		      "NAD83 / Massachusetts Mainland");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=42.68333333333333 +lat_2=41.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=41 +lon_0=-71.5 +x_0=200000 +y_0=750000 +ellp");
    add_proj4text (p, 2, "s=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Massachusetts Mainland\",GEOGCS[\"NAD83");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 19");
    add_srs_wkt (p, 2,
		 "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"standard_parallel_1\",42.68333333333333],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "standard_parallel_2\",41.71666666666667],PARAMETER[\"lat");
    add_srs_wkt (p, 10,
		 "itude_of_origin\",41],PARAMETER[\"central_meridian\",-71");
    add_srs_wkt (p, 11,
		 ".5],PARAMETER[\"false_easting\",200000],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_northing\",750000],AUTHORITY[\"EPSG\",\"26986\"],AXIS[");
    add_srs_wkt (p, 13, "\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26987, "epsg", 26987,
		      "NAD83 / Massachusetts Island");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.48333333333333 +lat_2=41.28333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=41 +lon_0=-70.5 +x_0=500000 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Massachusetts Island\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "standard_parallel_1\",41.48333333333333],PARAMETER[\"sta");
    add_srs_wkt (p, 9,
		 "ndard_parallel_2\",41.28333333333333],PARAMETER[\"latitu");
    add_srs_wkt (p, 10,
		 "de_of_origin\",41],PARAMETER[\"central_meridian\",-70.5]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_easting\",500000],PARAMETER[\"false_n");
    add_srs_wkt (p, 12,
		 "orthing\",0],AUTHORITY[\"EPSG\",\"26987\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 13, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26988, "epsg", 26988,
		      "NAD83 / Michigan North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.08333333333334 +lat_2=45.48333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=44.78333333333333 +lon_0=-87 +x_0=8000000 +y_");
    add_proj4text (p, 2, "0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Michigan North\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 8,
		 "rd_parallel_1\",47.08333333333334],PARAMETER[\"standard_");
    add_srs_wkt (p, 9,
		 "parallel_2\",45.48333333333333],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 10,
		 "origin\",44.78333333333333],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 11,
		 "\",-87],PARAMETER[\"false_easting\",8000000],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"26988\"],AXIS[\"");
    add_srs_wkt (p, 13, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26989, "epsg", 26989,
		      "NAD83 / Michigan Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.7 +lat_2=44.18333333333333 +lat_0=43");
    add_proj4text (p, 1,
		   ".31666666666667 +lon_0=-84.36666666666666 +x_0=6000000 +");
    add_proj4text (p, 2, "y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Michigan Central\",GEOGCS[\"NAD83\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6");
    add_srs_wkt (p, 2,
		 "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"426");
    add_srs_wkt (p, 6,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"sta");
    add_srs_wkt (p, 8,
		 "ndard_parallel_1\",45.7],PARAMETER[\"standard_parallel_2");
    add_srs_wkt (p, 9,
		 "\",44.18333333333333],PARAMETER[\"latitude_of_origin\",4");
    add_srs_wkt (p, 10,
		 "3.31666666666667],PARAMETER[\"central_meridian\",-84.366");
    add_srs_wkt (p, 11,
		 "66666666666],PARAMETER[\"false_easting\",6000000],PARAME");
    add_srs_wkt (p, 12,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"26989\"],");
    add_srs_wkt (p, 13, "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26990, "epsg", 26990,
		      "NAD83 / Michigan South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43.66666666666666 +lat_2=42.1 +lat_0=41");
    add_proj4text (p, 1,
		   ".5 +lon_0=-84.36666666666666 +x_0=4000000 +y_0=0 +ellps=");
    add_proj4text (p, 2, "GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Michigan South\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 8,
		 "rd_parallel_1\",43.66666666666666],PARAMETER[\"standard_");
    add_srs_wkt (p, 9,
		 "parallel_2\",42.1],PARAMETER[\"latitude_of_origin\",41.5");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"central_meridian\",-84.36666666666666],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_easting\",4000000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 12,
		 "hing\",0],AUTHORITY[\"EPSG\",\"26990\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 13, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26991, "epsg", 26991,
		      "NAD83 / Minnesota North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.63333333333333 +lat_2=47.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=46.5 +lon_0=-93.09999999999999 +x_0=800000 +y");
    add_proj4text (p, 2,
		   "_0=100000 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Minnesota North\",GEOGCS[\"NAD83\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",63");
    add_srs_wkt (p, 2,
		 "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"stan");
    add_srs_wkt (p, 8,
		 "dard_parallel_1\",48.63333333333333],PARAMETER[\"standar");
    add_srs_wkt (p, 9,
		 "d_parallel_2\",47.03333333333333],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 10,
		 "f_origin\",46.5],PARAMETER[\"central_meridian\",-93.1],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_easting\",800000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 12,
		 "thing\",100000],AUTHORITY[\"EPSG\",\"26991\"],AXIS[\"X\"");
    add_srs_wkt (p, 13, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26992, "epsg", 26992,
		      "NAD83 / Minnesota Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.05 +lat_2=45.61666666666667 +lat_0=4");
    add_proj4text (p, 1,
		   "5 +lon_0=-94.25 +x_0=800000 +y_0=100000 +ellps=GRS80 +da");
    add_proj4text (p, 2, "tum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Minnesota Central\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 7,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 8,
		 "andard_parallel_1\",47.05],PARAMETER[\"standard_parallel");
    add_srs_wkt (p, 9,
		 "_2\",45.61666666666667],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 10,
		 ",45],PARAMETER[\"central_meridian\",-94.25],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",800000],PARAMETER[\"false_northing\",100");
    add_srs_wkt (p, 12,
		 "000],AUTHORITY[\"EPSG\",\"26992\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 13, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26993, "epsg", 26993,
		      "NAD83 / Minnesota South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.21666666666667 +lat_2=43.78333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=43 +lon_0=-94 +x_0=800000 +y_0=100000 +ellps=");
    add_proj4text (p, 2, "GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Minnesota South\",GEOGCS[\"NAD83\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",63");
    add_srs_wkt (p, 2,
		 "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"stan");
    add_srs_wkt (p, 8,
		 "dard_parallel_1\",45.21666666666667],PARAMETER[\"standar");
    add_srs_wkt (p, 9,
		 "d_parallel_2\",43.78333333333333],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 10,
		 "f_origin\",43],PARAMETER[\"central_meridian\",-94],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_easting\",800000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",100000],AUTHORITY[\"EPSG\",\"26993\"],AXIS[\"X\",EAS");
    add_srs_wkt (p, 13, "T],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26994, "epsg", 26994,
		      "NAD83 / Mississippi East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=29.5 +lon_0=-88.83333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "9995 +x_0=300000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units");
    add_proj4text (p, 2, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Mississippi East\",GEOGCS[\"NAD83\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6");
    add_srs_wkt (p, 2,
		 "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"426");
    add_srs_wkt (p, 6,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",29.5],PARAMETER[\"central_meridian\",-88.83333");
    add_srs_wkt (p, 9,
		 "333333333],PARAMETER[\"scale_factor\",0.99995],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",300000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"26994\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 26995, "epsg", 26995,
		      "NAD83 / Mississippi West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=29.5 +lon_0=-90.33333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "9995 +x_0=700000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units");
    add_proj4text (p, 2, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Mississippi West\",GEOGCS[\"NAD83\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6");
    add_srs_wkt (p, 2,
		 "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"426");
    add_srs_wkt (p, 6,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",29.5],PARAMETER[\"central_meridian\",-90.33333");
    add_srs_wkt (p, 9,
		 "333333333],PARAMETER[\"scale_factor\",0.99995],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",700000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"26995\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 26996, "epsg", 26996,
		      "NAD83 / Missouri East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=35.83333333333334 +lon_0=-90.5 +k=0.9");
    add_proj4text (p, 1,
		   "99933333 +x_0=250000 +y_0=0 +ellps=GRS80 +datum=NAD83 +u");
    add_proj4text (p, 2, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Missouri East\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",35.83333333333334],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 9,
		 "-90.5],PARAMETER[\"scale_factor\",0.999933333],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",250000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"26996\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 26997, "epsg", 26997,
		      "NAD83 / Missouri Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=35.83333333333334 +lon_0=-92.5 +k=0.9");
    add_proj4text (p, 1,
		   "99933333 +x_0=500000 +y_0=0 +ellps=GRS80 +datum=NAD83 +u");
    add_proj4text (p, 2, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Missouri Central\",GEOGCS[\"NAD83\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6");
    add_srs_wkt (p, 2,
		 "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"426");
    add_srs_wkt (p, 6,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",35.83333333333334],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 9,
		 "n\",-92.5],PARAMETER[\"scale_factor\",0.999933333],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_easting\",500000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 11,
		 "g\",0],AUTHORITY[\"EPSG\",\"26997\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 12, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 26998, "epsg", 26998,
		      "NAD83 / Missouri West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36.16666666666666 +lon_0=-94.5 +k=0.9");
    add_proj4text (p, 1,
		   "99941177 +x_0=850000 +y_0=0 +ellps=GRS80 +datum=NAD83 +u");
    add_proj4text (p, 2, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Missouri West\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",36.16666666666666],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 9,
		 "-94.5],PARAMETER[\"scale_factor\",0.999941177],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",850000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"26998\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 27037, "epsg", 27037,
		      "Nahrwan 1967 / UTM zone 37N");
    add_proj4text (p, 0, "+proj=utm +zone=37 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Nahrwan 1967 / UTM zone 37N\",GEOGCS[\"Nahrwan ");
    add_srs_wkt (p, 1,
		 "1967\",DATUM[\"Nahrwan_1967\",SPHEROID[\"Clarke 1880 (RG");
    add_srs_wkt (p, 2,
		 "S)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6270\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4270\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",39],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 11,
		 "G\",\"27037\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",");
    add_srs_wkt (p, 12, "NORTH]]");
    p = add_epsg_def (first, last, 27038, "epsg", 27038,
		      "Nahrwan 1967 / UTM zone 38N");
    add_proj4text (p, 0, "+proj=utm +zone=38 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Nahrwan 1967 / UTM zone 38N\",GEOGCS[\"Nahrwan ");
    add_srs_wkt (p, 1,
		 "1967\",DATUM[\"Nahrwan_1967\",SPHEROID[\"Clarke 1880 (RG");
    add_srs_wkt (p, 2,
		 "S)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6270\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4270\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",45],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 11,
		 "G\",\"27038\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",");
    add_srs_wkt (p, 12, "NORTH]]");
    p = add_epsg_def (first, last, 27039, "epsg", 27039,
		      "Nahrwan 1967 / UTM zone 39N");
    add_proj4text (p, 0, "+proj=utm +zone=39 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Nahrwan 1967 / UTM zone 39N\",GEOGCS[\"Nahrwan ");
    add_srs_wkt (p, 1,
		 "1967\",DATUM[\"Nahrwan_1967\",SPHEROID[\"Clarke 1880 (RG");
    add_srs_wkt (p, 2,
		 "S)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6270\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4270\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",51],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 11,
		 "G\",\"27039\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",");
    add_srs_wkt (p, 12, "NORTH]]");
    p = add_epsg_def (first, last, 27040, "epsg", 27040,
		      "Nahrwan 1967 / UTM zone 40N");
    add_proj4text (p, 0, "+proj=utm +zone=40 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Nahrwan 1967 / UTM zone 40N\",GEOGCS[\"Nahrwan ");
    add_srs_wkt (p, 1,
		 "1967\",DATUM[\"Nahrwan_1967\",SPHEROID[\"Clarke 1880 (RG");
    add_srs_wkt (p, 2,
		 "S)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6270\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4270\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",57],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 11,
		 "G\",\"27040\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",");
    add_srs_wkt (p, 12, "NORTH]]");
    p = add_epsg_def (first, last, 27120, "epsg", 27120,
		      "Naparima 1972 / UTM zone 20N");
    add_proj4text (p, 0, "+proj=utm +zone=20 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Naparima 1972 / UTM zone 20N\",GEOGCS[\"Naparim");
    add_srs_wkt (p, 1,
		 "a 1972\",DATUM[\"Naparima_1972\",SPHEROID[\"Internationa");
    add_srs_wkt (p, 2,
		 "l 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6271\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"427");
    add_srs_wkt (p, 6,
		 "1\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",-63],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 10,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 11,
		 "\",\"27120\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",N");
    add_srs_wkt (p, 12, "ORTH]]");
    p = add_epsg_def (first, last, 27200, "epsg", 27200,
		      "NZGD49 / New Zealand Map Grid");
    add_proj4text (p, 0,
		   "+proj=nzmg +lat_0=-41 +lon_0=173 +x_0=2510000 +y_0=60231");
    add_proj4text (p, 1, "50 +ellps=intl +datum=nzgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / New Zealand Map Grid\",GEOGCS[\"NZGD49");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "22\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.599");
    add_srs_wkt (p, 4,
		 "3],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 6,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 8,
		 "\"]],PROJECTION[\"New_Zealand_Map_Grid\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",-41],PARAMETER[\"central_meridian\",1");
    add_srs_wkt (p, 10,
		 "73],PARAMETER[\"false_easting\",2510000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",6023150],AUTHORITY[\"EPSG\",\"27200\"],AXI");
    add_srs_wkt (p, 12, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 27205, "epsg", 27205,
		      "NZGD49 / Mount Eden Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-36.87986527777778 +lon_0=174.7643393");
    add_proj4text (p, 1,
		   "611111 +k=0.9999 +x_0=300000 +y_0=700000 +ellps=intl +da");
    add_proj4text (p, 2, "tum=nzgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Mount Eden Circuit\",GEOGCS[\"NZGD49\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",-36.87986527777778],PARAMETER[\"central_");
    add_srs_wkt (p, 10,
		 "meridian\",174.7643393611111],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 11,
		 ",0.9999],PARAMETER[\"false_easting\",300000],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",700000],AUTHORITY[\"EPSG\",\"27205\"],A");
    add_srs_wkt (p, 13, "XIS[\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27206, "epsg", 27206,
		      "NZGD49 / Bay of Plenty Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-37.76124980555556 +lon_0=176.4661972");
    add_proj4text (p, 1,
		   "5 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=nzgd49");
    add_proj4text (p, 2, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Bay of Plenty Circuit\",GEOGCS[\"NZGD4");
    add_srs_wkt (p, 1,
		 "9\",DATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "22\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.599");
    add_srs_wkt (p, 4,
		 "3],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 6,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 8,
		 "\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"lat");
    add_srs_wkt (p, 9,
		 "itude_of_origin\",-37.76124980555556],PARAMETER[\"centra");
    add_srs_wkt (p, 10,
		 "l_meridian\",176.46619725],PARAMETER[\"scale_factor\",1]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_easting\",300000],PARAMETER[\"false_n");
    add_srs_wkt (p, 12,
		 "orthing\",700000],AUTHORITY[\"EPSG\",\"27206\"],AXIS[\"N");
    add_srs_wkt (p, 13, "orthing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27207, "epsg", 27207,
		      "NZGD49 / Poverty Bay Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-38.62470277777778 +lon_0=177.8856362");
    add_proj4text (p, 1,
		   "777778 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Poverty Bay Circuit\",GEOGCS[\"NZGD49\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",-38.62470277777778],PARAMETER[\"central_");
    add_srs_wkt (p, 10,
		 "meridian\",177.8856362777778],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 11,
		 ",1],PARAMETER[\"false_easting\",300000],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_northing\",700000],AUTHORITY[\"EPSG\",\"27207\"],AXIS[");
    add_srs_wkt (p, 13, "\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27208, "epsg", 27208,
		      "NZGD49 / Hawkes Bay Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-39.65092930555556 +lon_0=176.6736805");
    add_proj4text (p, 1,
		   "277778 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Hawkes Bay Circuit\",GEOGCS[\"NZGD49\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",-39.65092930555556],PARAMETER[\"central_");
    add_srs_wkt (p, 10,
		 "meridian\",176.6736805277778],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 11,
		 ",1],PARAMETER[\"false_easting\",300000],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_northing\",700000],AUTHORITY[\"EPSG\",\"27208\"],AXIS[");
    add_srs_wkt (p, 13, "\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27209, "epsg", 27209,
		      "NZGD49 / Taranaki Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-39.13575830555556 +lon_0=174.2280117");
    add_proj4text (p, 1,
		   "5 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=nzgd49");
    add_proj4text (p, 2, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Taranaki Circuit\",GEOGCS[\"NZGD49\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"Inte");
    add_srs_wkt (p, 2,
		 "rnational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",-39.13575830555556],PARAMETER[\"central_mer");
    add_srs_wkt (p, 10,
		 "idian\",174.22801175],PARAMETER[\"scale_factor\",1],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_easting\",300000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 12,
		 "ng\",700000],AUTHORITY[\"EPSG\",\"27209\"],AXIS[\"Northi");
    add_srs_wkt (p, 13, "ng\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27210, "epsg", 27210,
		      "NZGD49 / Tuhirangi Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-39.51247038888889 +lon_0=175.6400368");
    add_proj4text (p, 1,
		   "055556 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Tuhirangi Circuit\",GEOGCS[\"NZGD49\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"Int");
    add_srs_wkt (p, 2,
		 "ernational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",-39.51247038888889],PARAMETER[\"central_mer");
    add_srs_wkt (p, 10,
		 "idian\",175.6400368055556],PARAMETER[\"scale_factor\",1]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_easting\",300000],PARAMETER[\"false_n");
    add_srs_wkt (p, 12,
		 "orthing\",700000],AUTHORITY[\"EPSG\",\"27210\"],AXIS[\"N");
    add_srs_wkt (p, 13, "orthing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27211, "epsg", 27211,
		      "NZGD49 / Wanganui Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-40.24194713888889 +lon_0=175.4880996");
    add_proj4text (p, 1,
		   "111111 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Wanganui Circuit\",GEOGCS[\"NZGD49\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"Inte");
    add_srs_wkt (p, 2,
		 "rnational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",-40.24194713888889],PARAMETER[\"central_mer");
    add_srs_wkt (p, 10,
		 "idian\",175.4880996111111],PARAMETER[\"scale_factor\",1]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_easting\",300000],PARAMETER[\"false_n");
    add_srs_wkt (p, 12,
		 "orthing\",700000],AUTHORITY[\"EPSG\",\"27211\"],AXIS[\"N");
    add_srs_wkt (p, 13, "orthing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27212, "epsg", 27212,
		      "NZGD49 / Wairarapa Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-40.92553263888889 +lon_0=175.6473496");
    add_proj4text (p, 1,
		   "666667 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Wairarapa Circuit\",GEOGCS[\"NZGD49\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"Int");
    add_srs_wkt (p, 2,
		 "ernational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",-40.92553263888889],PARAMETER[\"central_mer");
    add_srs_wkt (p, 10,
		 "idian\",175.6473496666667],PARAMETER[\"scale_factor\",1]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_easting\",300000],PARAMETER[\"false_n");
    add_srs_wkt (p, 12,
		 "orthing\",700000],AUTHORITY[\"EPSG\",\"27212\"],AXIS[\"N");
    add_srs_wkt (p, 13, "orthing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27213, "epsg", 27213,
		      "NZGD49 / Wellington Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-41.30131963888888 +lon_0=174.7766231");
    add_proj4text (p, 1,
		   "111111 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Wellington Circuit\",GEOGCS[\"NZGD49\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",-41.30131963888888],PARAMETER[\"central_");
    add_srs_wkt (p, 10,
		 "meridian\",174.7766231111111],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 11,
		 ",1],PARAMETER[\"false_easting\",300000],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_northing\",700000],AUTHORITY[\"EPSG\",\"27213\"],AXIS[");
    add_srs_wkt (p, 13, "\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27214, "epsg", 27214,
		      "NZGD49 / Collingwood Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-40.71475905555556 +lon_0=172.6720465");
    add_proj4text (p, 1,
		   " +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=nzgd49 ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Collingwood Circuit\",GEOGCS[\"NZGD49\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",-40.71475905555556],PARAMETER[\"central_");
    add_srs_wkt (p, 10,
		 "meridian\",172.6720465],PARAMETER[\"scale_factor\",1],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_easting\",300000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 12,
		 "hing\",700000],AUTHORITY[\"EPSG\",\"27214\"],AXIS[\"Nort");
    add_srs_wkt (p, 13, "hing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27215, "epsg", 27215,
		      "NZGD49 / Nelson Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-41.27454472222222 +lon_0=173.2993168");
    add_proj4text (p, 1,
		   "055555 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Nelson Circuit\",GEOGCS[\"NZGD49\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"Intern");
    add_srs_wkt (p, 2,
		 "ational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]]");
    add_srs_wkt (p, 3,
		 ",TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",-41.27454472222222],PARAMETER[\"central_meri");
    add_srs_wkt (p, 10,
		 "dian\",173.2993168055555],PARAMETER[\"scale_factor\",1],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_easting\",300000],PARAMETER[\"false_no");
    add_srs_wkt (p, 12,
		 "rthing\",700000],AUTHORITY[\"EPSG\",\"27215\"],AXIS[\"No");
    add_srs_wkt (p, 13, "rthing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27216, "epsg", 27216,
		      "NZGD49 / Karamea Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-41.28991152777778 +lon_0=172.1090281");
    add_proj4text (p, 1,
		   "944444 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Karamea Circuit\",GEOGCS[\"NZGD49\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"Inter");
    add_srs_wkt (p, 2,
		 "national 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]");
    add_srs_wkt (p, 3,
		 "],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",-41.28991152777778],PARAMETER[\"central_mer");
    add_srs_wkt (p, 10,
		 "idian\",172.1090281944444],PARAMETER[\"scale_factor\",1]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_easting\",300000],PARAMETER[\"false_n");
    add_srs_wkt (p, 12,
		 "orthing\",700000],AUTHORITY[\"EPSG\",\"27216\"],AXIS[\"N");
    add_srs_wkt (p, 13, "orthing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27217, "epsg", 27217,
		      "NZGD49 / Buller Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-41.81080286111111 +lon_0=171.5812600");
    add_proj4text (p, 1,
		   "555556 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Buller Circuit\",GEOGCS[\"NZGD49\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"Intern");
    add_srs_wkt (p, 2,
		 "ational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]]");
    add_srs_wkt (p, 3,
		 ",TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",-41.81080286111111],PARAMETER[\"central_meri");
    add_srs_wkt (p, 10,
		 "dian\",171.5812600555556],PARAMETER[\"scale_factor\",1],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_easting\",300000],PARAMETER[\"false_no");
    add_srs_wkt (p, 12,
		 "rthing\",700000],AUTHORITY[\"EPSG\",\"27217\"],AXIS[\"No");
    add_srs_wkt (p, 13, "rthing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27218, "epsg", 27218,
		      "NZGD49 / Grey Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-42.33369427777778 +lon_0=171.5497713");
    add_proj4text (p, 1,
		   "055556 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Grey Circuit\",GEOGCS[\"NZGD49\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"Internat");
    add_srs_wkt (p, 2,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"427");
    add_srs_wkt (p, 7,
		 "2\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",-42.33369427777778],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 10,
		 "an\",171.5497713055556],PARAMETER[\"scale_factor\",1],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_easting\",300000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 12,
		 "hing\",700000],AUTHORITY[\"EPSG\",\"27218\"],AXIS[\"Nort");
    add_srs_wkt (p, 13, "hing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27219, "epsg", 27219,
		      "NZGD49 / Amuri Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-42.68911658333333 +lon_0=173.0101333");
    add_proj4text (p, 1,
		   "888889 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Amuri Circuit\",GEOGCS[\"NZGD49\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "72\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",-42.68911658333333],PARAMETER[\"central_merid");
    add_srs_wkt (p, 10,
		 "ian\",173.0101333888889],PARAMETER[\"scale_factor\",1],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_easting\",300000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 12,
		 "thing\",700000],AUTHORITY[\"EPSG\",\"27219\"],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27220, "epsg", 27220,
		      "NZGD49 / Marlborough Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-41.54448666666666 +lon_0=173.8020741");
    add_proj4text (p, 1,
		   "111111 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Marlborough Circuit\",GEOGCS[\"NZGD49\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",-41.54448666666666],PARAMETER[\"central_");
    add_srs_wkt (p, 10,
		 "meridian\",173.8020741111111],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 11,
		 ",1],PARAMETER[\"false_easting\",300000],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_northing\",700000],AUTHORITY[\"EPSG\",\"27220\"],AXIS[");
    add_srs_wkt (p, 13, "\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27221, "epsg", 27221,
		      "NZGD49 / Hokitika Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-42.88632236111111 +lon_0=170.9799935");
    add_proj4text (p, 1,
		   " +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=nzgd49 ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Hokitika Circuit\",GEOGCS[\"NZGD49\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"Inte");
    add_srs_wkt (p, 2,
		 "rnational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",-42.88632236111111],PARAMETER[\"central_mer");
    add_srs_wkt (p, 10,
		 "idian\",170.9799935],PARAMETER[\"scale_factor\",1],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_easting\",300000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",700000],AUTHORITY[\"EPSG\",\"27221\"],AXIS[\"Northin");
    add_srs_wkt (p, 13, "g\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27222, "epsg", 27222,
		      "NZGD49 / Okarito Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-43.11012813888889 +lon_0=170.2609258");
    add_proj4text (p, 1,
		   "333333 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Okarito Circuit\",GEOGCS[\"NZGD49\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"Inter");
    add_srs_wkt (p, 2,
		 "national 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]");
    add_srs_wkt (p, 3,
		 "],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",-43.11012813888889],PARAMETER[\"central_mer");
    add_srs_wkt (p, 10,
		 "idian\",170.2609258333333],PARAMETER[\"scale_factor\",1]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_easting\",300000],PARAMETER[\"false_n");
    add_srs_wkt (p, 12,
		 "orthing\",700000],AUTHORITY[\"EPSG\",\"27222\"],AXIS[\"N");
    add_srs_wkt (p, 13, "orthing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27223, "epsg", 27223,
		      "NZGD49 / Jacksons Bay Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-43.97780288888889 +lon_0=168.606267 ");
    add_proj4text (p, 1,
		   "+k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=nzgd49 +");
    add_proj4text (p, 2, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Jacksons Bay Circuit\",GEOGCS[\"NZGD49");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "22\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.599");
    add_srs_wkt (p, 4,
		 "3],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 6,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 8,
		 "\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"lat");
    add_srs_wkt (p, 9,
		 "itude_of_origin\",-43.97780288888889],PARAMETER[\"centra");
    add_srs_wkt (p, 10,
		 "l_meridian\",168.606267],PARAMETER[\"scale_factor\",1],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_easting\",300000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 12,
		 "thing\",700000],AUTHORITY[\"EPSG\",\"27223\"],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27224, "epsg", 27224,
		      "NZGD49 / Mount Pleasant Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-43.59063758333333 +lon_0=172.7271935");
    add_proj4text (p, 1,
		   "833333 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Mount Pleasant Circuit\",GEOGCS[\"NZGD");
    add_srs_wkt (p, 1,
		 "49\",DATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5");
    add_srs_wkt (p, 4,
		 "993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 8,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 9,
		 "atitude_of_origin\",-43.59063758333333],PARAMETER[\"cent");
    add_srs_wkt (p, 10,
		 "ral_meridian\",172.7271935833333],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 11,
		 "or\",1],PARAMETER[\"false_easting\",300000],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",700000],AUTHORITY[\"EPSG\",\"27224\"],A");
    add_srs_wkt (p, 13, "XIS[\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27225, "epsg", 27225,
		      "NZGD49 / Gawler Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-43.74871155555556 +lon_0=171.3607484");
    add_proj4text (p, 1,
		   "722222 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Gawler Circuit\",GEOGCS[\"NZGD49\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"Intern");
    add_srs_wkt (p, 2,
		 "ational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]]");
    add_srs_wkt (p, 3,
		 ",TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",-43.74871155555556],PARAMETER[\"central_meri");
    add_srs_wkt (p, 10,
		 "dian\",171.3607484722222],PARAMETER[\"scale_factor\",1],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_easting\",300000],PARAMETER[\"false_no");
    add_srs_wkt (p, 12,
		 "rthing\",700000],AUTHORITY[\"EPSG\",\"27225\"],AXIS[\"No");
    add_srs_wkt (p, 13, "rthing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27226, "epsg", 27226,
		      "NZGD49 / Timaru Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-44.40222036111111 +lon_0=171.0572508");
    add_proj4text (p, 1,
		   "333333 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Timaru Circuit\",GEOGCS[\"NZGD49\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"Intern");
    add_srs_wkt (p, 2,
		 "ational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]]");
    add_srs_wkt (p, 3,
		 ",TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",-44.40222036111111],PARAMETER[\"central_meri");
    add_srs_wkt (p, 10,
		 "dian\",171.0572508333333],PARAMETER[\"scale_factor\",1],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_easting\",300000],PARAMETER[\"false_no");
    add_srs_wkt (p, 12,
		 "rthing\",700000],AUTHORITY[\"EPSG\",\"27226\"],AXIS[\"No");
    add_srs_wkt (p, 13, "rthing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27227, "epsg", 27227,
		      "NZGD49 / Lindis Peak Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-44.73526797222222 +lon_0=169.4677550");
    add_proj4text (p, 1,
		   "833333 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Lindis Peak Circuit\",GEOGCS[\"NZGD49\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",-44.73526797222222],PARAMETER[\"central_");
    add_srs_wkt (p, 10,
		 "meridian\",169.4677550833333],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 11,
		 ",1],PARAMETER[\"false_easting\",300000],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_northing\",700000],AUTHORITY[\"EPSG\",\"27227\"],AXIS[");
    add_srs_wkt (p, 13, "\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27228, "epsg", 27228,
		      "NZGD49 / Mount Nicholas Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-45.13290258333333 +lon_0=168.3986411");
    add_proj4text (p, 1,
		   "944444 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Mount Nicholas Circuit\",GEOGCS[\"NZGD");
    add_srs_wkt (p, 1,
		 "49\",DATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5");
    add_srs_wkt (p, 4,
		 "993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 5,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 6,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 8,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 9,
		 "atitude_of_origin\",-45.13290258333333],PARAMETER[\"cent");
    add_srs_wkt (p, 10,
		 "ral_meridian\",168.3986411944444],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 11,
		 "or\",1],PARAMETER[\"false_easting\",300000],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",700000],AUTHORITY[\"EPSG\",\"27228\"],A");
    add_srs_wkt (p, 13, "XIS[\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27229, "epsg", 27229,
		      "NZGD49 / Mount York Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-45.56372616666666 +lon_0=167.7388617");
    add_proj4text (p, 1,
		   "777778 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Mount York Circuit\",GEOGCS[\"NZGD49\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",-45.56372616666666],PARAMETER[\"central_");
    add_srs_wkt (p, 10,
		 "meridian\",167.7388617777778],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 11,
		 ",1],PARAMETER[\"false_easting\",300000],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_northing\",700000],AUTHORITY[\"EPSG\",\"27229\"],AXIS[");
    add_srs_wkt (p, 13, "\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27230, "epsg", 27230,
		      "NZGD49 / Observation Point Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-45.81619661111111 +lon_0=170.6285951");
    add_proj4text (p, 1,
		   "666667 +k=1 +x_0=300000 +y_0=700000 +ellps=intl +datum=n");
    add_proj4text (p, 2, "zgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Observation Point Circuit\",GEOGCS[\"N");
    add_srs_wkt (p, 1,
		 "ZGD49\",DATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-");
    add_srs_wkt (p, 4,
		 "4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 5,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 6,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 8,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "latitude_of_origin\",-45.81619661111111],PARAMETER[\"cen");
    add_srs_wkt (p, 10,
		 "tral_meridian\",170.6285951666667],PARAMETER[\"scale_fac");
    add_srs_wkt (p, 11,
		 "tor\",1],PARAMETER[\"false_easting\",300000],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",700000],AUTHORITY[\"EPSG\",\"27230\"],A");
    add_srs_wkt (p, 13, "XIS[\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27231, "epsg", 27231,
		      "NZGD49 / North Taieri Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-45.86151336111111 +lon_0=170.2825891");
    add_proj4text (p, 1,
		   "111111 +k=0.99996 +x_0=300000 +y_0=700000 +ellps=intl +d");
    add_proj4text (p, 2, "atum=nzgd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / North Taieri Circuit\",GEOGCS[\"NZGD49");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "22\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.599");
    add_srs_wkt (p, 4,
		 "3],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 6,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"4272\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 8,
		 "\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"lat");
    add_srs_wkt (p, 9,
		 "itude_of_origin\",-45.86151336111111],PARAMETER[\"centra");
    add_srs_wkt (p, 10,
		 "l_meridian\",170.2825891111111],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 11,
		 "\",0.99996],PARAMETER[\"false_easting\",300000],PARAMETE");
    add_srs_wkt (p, 12,
		 "R[\"false_northing\",700000],AUTHORITY[\"EPSG\",\"27231\"");
    add_srs_wkt (p, 13, "],AXIS[\"Northing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27232, "epsg", 27232,
		      "NZGD49 / Bluff Circuit");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-46.60000961111111 +lon_0=168.342872 ");
    add_proj4text (p, 1,
		   "+k=1 +x_0=300002.66 +y_0=699999.58 +ellps=intl +datum=nz");
    add_proj4text (p, 2, "gd49 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / Bluff Circuit\",GEOGCS[\"NZGD49\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"Interna");
    add_srs_wkt (p, 2,
		 "tional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],");
    add_srs_wkt (p, 3,
		 "TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "72\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",-46.60000961111111],PARAMETER[\"central_merid");
    add_srs_wkt (p, 10,
		 "ian\",168.342872],PARAMETER[\"scale_factor\",1],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_easting\",300002.66],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",699999.58],AUTHORITY[\"EPSG\",\"27232\"],AXIS[\"Nort");
    add_srs_wkt (p, 13, "hing\",NORTH],AXIS[\"Easting\",EAST]]");
    p = add_epsg_def (first, last, 27258, "epsg", 27258,
		      "NZGD49 / UTM zone 58S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +south +ellps=intl +datum=nzgd49 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / UTM zone 58S\",GEOGCS[\"NZGD49\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"Internat");
    add_srs_wkt (p, 2,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"427");
    add_srs_wkt (p, 7,
		 "2\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",0],PARAMETER[\"central_meridian\",165],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",10000000],AUTHORITY");
    add_srs_wkt (p, 12,
		 "[\"EPSG\",\"27258\"],AXIS[\"Easting\",EAST],AXIS[\"North");
    add_srs_wkt (p, 13, "ing\",NORTH]]");
    p = add_epsg_def (first, last, 27259, "epsg", 27259,
		      "NZGD49 / UTM zone 59S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=59 +south +ellps=intl +datum=nzgd49 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / UTM zone 59S\",GEOGCS[\"NZGD49\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"Internat");
    add_srs_wkt (p, 2,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"427");
    add_srs_wkt (p, 7,
		 "2\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",0],PARAMETER[\"central_meridian\",171],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",10000000],AUTHORITY");
    add_srs_wkt (p, 12,
		 "[\"EPSG\",\"27259\"],AXIS[\"Easting\",EAST],AXIS[\"North");
    add_srs_wkt (p, 13, "ing\",NORTH]]");
    p = add_epsg_def (first, last, 27260, "epsg", 27260,
		      "NZGD49 / UTM zone 60S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=60 +south +ellps=intl +datum=nzgd49 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / UTM zone 60S\",GEOGCS[\"NZGD49\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"Internat");
    add_srs_wkt (p, 2,
		 "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 6,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"427");
    add_srs_wkt (p, 7,
		 "2\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 8,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 9,
		 "_origin\",0],PARAMETER[\"central_meridian\",177],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",10000000],AUTHORITY");
    add_srs_wkt (p, 12,
		 "[\"EPSG\",\"27260\"],AXIS[\"Easting\",EAST],AXIS[\"North");
    add_srs_wkt (p, 13, "ing\",NORTH]]");
    p = add_epsg_def (first, last, 27291, "epsg", 27291,
		      "NZGD49 / North Island Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-39 +lon_0=175.5 +k=1 +x_0=274319.524");
    add_proj4text (p, 1,
		   "3848086 +y_0=365759.3658464114 +ellps=intl +datum=nzgd49");
    add_proj4text (p, 2, " +to_meter=0.9143984146160287 +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / North Island Grid\",GEOGCS[\"NZGD49\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"Int");
    add_srs_wkt (p, 2,
		 "ernational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4272\"]],UNIT[\"British yard (Sears 1922)\",0.9143984146");
    add_srs_wkt (p, 8,
		 "160287,AUTHORITY[\"EPSG\",\"9040\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 9,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",-39],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"central_meridian\",175.5],PARAMETER[\"scale_f");
    add_srs_wkt (p, 11,
		 "actor\",1],PARAMETER[\"false_easting\",300000],PARAMETER");
    add_srs_wkt (p, 12,
		 "[\"false_northing\",400000],AUTHORITY[\"EPSG\",\"27291\"");
    add_srs_wkt (p, 13, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 27292, "epsg", 27292,
		      "NZGD49 / South Island Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=-44 +lon_0=171.5 +k=1 +x_0=457199.207");
    add_proj4text (p, 1,
		   "3080143 +y_0=457199.2073080143 +ellps=intl +datum=nzgd49");
    add_proj4text (p, 2, " +to_meter=0.9143984146160287 +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NZGD49 / South Island Grid\",GEOGCS[\"NZGD49\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"New_Zealand_Geodetic_Datum_1949\",SPHEROID[\"Int");
    add_srs_wkt (p, 2,
		 "ernational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 5,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 6,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4272\"]],UNIT[\"British yard (Sears 1922)\",0.9143984146");
    add_srs_wkt (p, 8,
		 "160287,AUTHORITY[\"EPSG\",\"9040\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 9,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",-44],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"central_meridian\",171.5],PARAMETER[\"scale_f");
    add_srs_wkt (p, 11,
		 "actor\",1],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 12,
		 "[\"false_northing\",500000],AUTHORITY[\"EPSG\",\"27292\"");
    add_srs_wkt (p, 13, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 27391, "epsg", 27391,
		      "NGO 1948 (Oslo) / NGO zone I");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=58 +lon_0=-4.666666666666667 +k=1 +x_");
    add_proj4text (p, 1,
		   "0=0 +y_0=0 +a=6377492.018 +b=6356173.508712696 +towgs84=");
    add_proj4text (p, 2,
		   "278.3,93,474.5,7.889,0.05,-6.61,6.21 +pm=oslo +units=m +");
    add_proj4text (p, 3, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NGO 1948 (Oslo) / NGO zone I\",GEOGCS[\"NGO 194");
    add_srs_wkt (p, 1,
		 "8 (Oslo)\",DATUM[\"NGO_1948_Oslo\",SPHEROID[\"Bessel Mod");
    add_srs_wkt (p, 2,
		 "ified\",6377492.018,299.1528128,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 3,
		 "5\"]],TOWGS84[278.3,93,474.5,7.889,0.05,-6.61,6.21],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6817\"]],PRIMEM[\"Oslo\",10.72291666666");
    add_srs_wkt (p, 5,
		 "667,AUTHORITY[\"EPSG\",\"8913\"]],UNIT[\"degree\",0.0174");
    add_srs_wkt (p, 6,
		 "5329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"4817\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 8,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "latitude_of_origin\",58],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 10,
		 "-4.666666666666667],PARAMETER[\"scale_factor\",1],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_easting\",0],PARAMETER[\"false_northing\",0]");
    add_srs_wkt (p, 12,
		 ",AUTHORITY[\"EPSG\",\"27391\"],AXIS[\"x\",NORTH],AXIS[\"");
    add_srs_wkt (p, 13, "y\",EAST]]");
    p = add_epsg_def (first, last, 27392, "epsg", 27392,
		      "NGO 1948 (Oslo) / NGO zone II");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=58 +lon_0=-2.333333333333333 +k=1 +x_");
    add_proj4text (p, 1,
		   "0=0 +y_0=0 +a=6377492.018 +b=6356173.508712696 +towgs84=");
    add_proj4text (p, 2,
		   "278.3,93,474.5,7.889,0.05,-6.61,6.21 +pm=oslo +units=m +");
    add_proj4text (p, 3, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NGO 1948 (Oslo) / NGO zone II\",GEOGCS[\"NGO 19");
    add_srs_wkt (p, 1,
		 "48 (Oslo)\",DATUM[\"NGO_1948_Oslo\",SPHEROID[\"Bessel Mo");
    add_srs_wkt (p, 2,
		 "dified\",6377492.018,299.1528128,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "05\"]],TOWGS84[278.3,93,474.5,7.889,0.05,-6.61,6.21],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6817\"]],PRIMEM[\"Oslo\",10.7229166666");
    add_srs_wkt (p, 5,
		 "6667,AUTHORITY[\"EPSG\",\"8913\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 6,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"4817\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 8,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "latitude_of_origin\",58],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 10,
		 "-2.333333333333333],PARAMETER[\"scale_factor\",1],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_easting\",0],PARAMETER[\"false_northing\",0]");
    add_srs_wkt (p, 12,
		 ",AUTHORITY[\"EPSG\",\"27392\"],AXIS[\"x\",NORTH],AXIS[\"");
    add_srs_wkt (p, 13, "y\",EAST]]");
    p = add_epsg_def (first, last, 27393, "epsg", 27393,
		      "NGO 1948 (Oslo) / NGO zone III");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=58 +lon_0=0 +k=1 +x_0=0 +y_0=0 +a=637");
    add_proj4text (p, 1,
		   "7492.018 +b=6356173.508712696 +towgs84=278.3,93,474.5,7.");
    add_proj4text (p, 2, "889,0.05,-6.61,6.21 +pm=oslo +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NGO 1948 (Oslo) / NGO zone III\",GEOGCS[\"NGO 1");
    add_srs_wkt (p, 1,
		 "948 (Oslo)\",DATUM[\"NGO_1948_Oslo\",SPHEROID[\"Bessel M");
    add_srs_wkt (p, 2,
		 "odified\",6377492.018,299.1528128,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "005\"]],TOWGS84[278.3,93,474.5,7.889,0.05,-6.61,6.21],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6817\"]],PRIMEM[\"Oslo\",10.722916666");
    add_srs_wkt (p, 5,
		 "66667,AUTHORITY[\"EPSG\",\"8913\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 6,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 7,
		 "\"EPSG\",\"4817\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 8,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"latitude_of_origin\",58],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 10,
		 "\",0],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_ea");
    add_srs_wkt (p, 11,
		 "sting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12, "PSG\",\"27393\"],AXIS[\"x\",NORTH],AXIS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 27394, "epsg", 27394,
		      "NGO 1948 (Oslo) / NGO zone IV");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=58 +lon_0=2.5 +k=1 +x_0=0 +y_0=0 +a=6");
    add_proj4text (p, 1,
		   "377492.018 +b=6356173.508712696 +towgs84=278.3,93,474.5,");
    add_proj4text (p, 2, "7.889,0.05,-6.61,6.21 +pm=oslo +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NGO 1948 (Oslo) / NGO zone IV\",GEOGCS[\"NGO 19");
    add_srs_wkt (p, 1,
		 "48 (Oslo)\",DATUM[\"NGO_1948_Oslo\",SPHEROID[\"Bessel Mo");
    add_srs_wkt (p, 2,
		 "dified\",6377492.018,299.1528128,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "05\"]],TOWGS84[278.3,93,474.5,7.889,0.05,-6.61,6.21],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6817\"]],PRIMEM[\"Oslo\",10.7229166666");
    add_srs_wkt (p, 5,
		 "6667,AUTHORITY[\"EPSG\",\"8913\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 6,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"4817\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 8,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "latitude_of_origin\",58],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 10,
		 "2.5],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_eas");
    add_srs_wkt (p, 11,
		 "ting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 12, "SG\",\"27394\"],AXIS[\"x\",NORTH],AXIS[\"y\",EAST]]");
    p = add_epsg_def (first, last, 27395, "epsg", 27395,
		      "NGO 1948 (Oslo) / NGO zone V");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=58 +lon_0=6.166666666666667 +k=1 +x_0");
    add_proj4text (p, 1,
		   "=0 +y_0=0 +a=6377492.018 +b=6356173.508712696 +towgs84=2");
    add_proj4text (p, 2,
		   "78.3,93,474.5,7.889,0.05,-6.61,6.21 +pm=oslo +units=m +n");
    add_proj4text (p, 3, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NGO 1948 (Oslo) / NGO zone V\",GEOGCS[\"NGO 194");
    add_srs_wkt (p, 1,
		 "8 (Oslo)\",DATUM[\"NGO_1948_Oslo\",SPHEROID[\"Bessel Mod");
    add_srs_wkt (p, 2,
		 "ified\",6377492.018,299.1528128,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 3,
		 "5\"]],TOWGS84[278.3,93,474.5,7.889,0.05,-6.61,6.21],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6817\"]],PRIMEM[\"Oslo\",10.72291666666");
    add_srs_wkt (p, 5,
		 "667,AUTHORITY[\"EPSG\",\"8913\"]],UNIT[\"degree\",0.0174");
    add_srs_wkt (p, 6,
		 "5329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"4817\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 8,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "latitude_of_origin\",58],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 10,
		 "6.166666666666667],PARAMETER[\"scale_factor\",1],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_easting\",0],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"27395\"],AXIS[\"x\",NORTH],AXIS[\"y");
    add_srs_wkt (p, 13, "\",EAST]]");
    p = add_epsg_def (first, last, 27396, "epsg", 27396,
		      "NGO 1948 (Oslo) / NGO zone VI");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=58 +lon_0=10.16666666666667 +k=1 +x_0");
    add_proj4text (p, 1,
		   "=0 +y_0=0 +a=6377492.018 +b=6356173.508712696 +towgs84=2");
    add_proj4text (p, 2,
		   "78.3,93,474.5,7.889,0.05,-6.61,6.21 +pm=oslo +units=m +n");
    add_proj4text (p, 3, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NGO 1948 (Oslo) / NGO zone VI\",GEOGCS[\"NGO 19");
    add_srs_wkt (p, 1,
		 "48 (Oslo)\",DATUM[\"NGO_1948_Oslo\",SPHEROID[\"Bessel Mo");
    add_srs_wkt (p, 2,
		 "dified\",6377492.018,299.1528128,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "05\"]],TOWGS84[278.3,93,474.5,7.889,0.05,-6.61,6.21],AUT");
    add_srs_wkt (p, 4,
		 "HORITY[\"EPSG\",\"6817\"]],PRIMEM[\"Oslo\",10.7229166666");
    add_srs_wkt (p, 5,
		 "6667,AUTHORITY[\"EPSG\",\"8913\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 6,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"4817\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 8,
		 "9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "latitude_of_origin\",58],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 10,
		 "10.16666666666667],PARAMETER[\"scale_factor\",1],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_easting\",0],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"27396\"],AXIS[\"x\",NORTH],AXIS[\"y");
    add_srs_wkt (p, 13, "\",EAST]]");
    p = add_epsg_def (first, last, 27397, "epsg", 27397,
		      "NGO 1948 (Oslo) / NGO zone VII");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=58 +lon_0=14.16666666666667 +k=1 +x_0");
    add_proj4text (p, 1,
		   "=0 +y_0=0 +a=6377492.018 +b=6356173.508712696 +towgs84=2");
    add_proj4text (p, 2,
		   "78.3,93,474.5,7.889,0.05,-6.61,6.21 +pm=oslo +units=m +n");
    add_proj4text (p, 3, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NGO 1948 (Oslo) / NGO zone VII\",GEOGCS[\"NGO 1");
    add_srs_wkt (p, 1,
		 "948 (Oslo)\",DATUM[\"NGO_1948_Oslo\",SPHEROID[\"Bessel M");
    add_srs_wkt (p, 2,
		 "odified\",6377492.018,299.1528128,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "005\"]],TOWGS84[278.3,93,474.5,7.889,0.05,-6.61,6.21],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6817\"]],PRIMEM[\"Oslo\",10.722916666");
    add_srs_wkt (p, 5,
		 "66667,AUTHORITY[\"EPSG\",\"8913\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 6,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 7,
		 "\"EPSG\",\"4817\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 8,
		 "\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"latitude_of_origin\",58],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 10,
		 "\",14.16666666666667],PARAMETER[\"scale_factor\",1],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_easting\",0],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 12,
		 "0],AUTHORITY[\"EPSG\",\"27397\"],AXIS[\"x\",NORTH],AXIS[");
    add_srs_wkt (p, 13, "\"y\",EAST]]");
    p = add_epsg_def (first, last, 27398, "epsg", 27398,
		      "NGO 1948 (Oslo) / NGO zone VIII");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=58 +lon_0=18.33333333333333 +k=1 +x_0");
    add_proj4text (p, 1,
		   "=0 +y_0=0 +a=6377492.018 +b=6356173.508712696 +towgs84=2");
    add_proj4text (p, 2,
		   "78.3,93,474.5,7.889,0.05,-6.61,6.21 +pm=oslo +units=m +n");
    add_proj4text (p, 3, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NGO 1948 (Oslo) / NGO zone VIII\",GEOGCS[\"NGO ");
    add_srs_wkt (p, 1,
		 "1948 (Oslo)\",DATUM[\"NGO_1948_Oslo\",SPHEROID[\"Bessel ");
    add_srs_wkt (p, 2,
		 "Modified\",6377492.018,299.1528128,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7005\"]],TOWGS84[278.3,93,474.5,7.889,0.05,-6.61,6.21],A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"6817\"]],PRIMEM[\"Oslo\",10.72291666");
    add_srs_wkt (p, 5,
		 "666667,AUTHORITY[\"EPSG\",\"8913\"]],UNIT[\"degree\",0.0");
    add_srs_wkt (p, 6,
		 "1745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY");
    add_srs_wkt (p, 7,
		 "[\"EPSG\",\"4817\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 8,
		 ",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"latitude_of_origin\",58],PARAMETER[\"central_meridia");
    add_srs_wkt (p, 10,
		 "n\",18.33333333333333],PARAMETER[\"scale_factor\",1],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_easting\",0],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"27398\"],AXIS[\"x\",NORTH],AXIS");
    add_srs_wkt (p, 13, "[\"y\",EAST]]");
    p = add_epsg_def (first, last, 27429, "epsg", 27429,
		      "Datum 73 / UTM zone 29N");
    add_proj4text (p, 0, "+proj=utm +zone=29 +ellps=intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Datum 73 / UTM zone 29N\",GEOGCS[\"Datum 73\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Datum_73\",SPHEROID[\"International 1924\",637838");
    add_srs_wkt (p, 2,
		 "8,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "6274\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 4,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 5,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4274\"]],UNIT[\"met");
    add_srs_wkt (p, 6,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 7,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 8,
		 "AMETER[\"central_meridian\",-9],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 9,
		 "\",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_northing\",0],AUTHORITY[\"EPSG\",\"27429\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 27492, "epsg", 27492,
		      "Datum 73 / Modified Portuguese Grid (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=39.66666666666666 +lon_0=-8.131906111");
    add_proj4text (p, 1,
		   "111112 +k=1 +x_0=180.598 +y_0=-86.98999999999999 +ellps=");
    add_proj4text (p, 2, "intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Datum 73 / Modified Portuguese Grid (deprecated");
    add_srs_wkt (p, 1,
		 ")\",GEOGCS[\"Datum 73\",DATUM[\"Datum_73\",SPHEROID[\"In");
    add_srs_wkt (p, 2,
		 "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6274\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4274\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",39.66666666666666],PARAMETER[\"centr");
    add_srs_wkt (p, 9,
		 "al_meridian\",-8.131906111111112],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 10,
		 "or\",1],PARAMETER[\"false_easting\",180.598],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",-86.99],AUTHORITY[\"EPSG\",\"27492\"],A");
    add_srs_wkt (p, 12, "XIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 27493, "epsg", 27493,
		      "Datum 73 / Modified Portuguese Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=39.66666666666666 +lon_0=-8.131906111");
    add_proj4text (p, 1,
		   "111112 +k=1 +x_0=180.598 +y_0=-86.98999999999999 +ellps=");
    add_proj4text (p, 2, "intl +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Datum 73 / Modified Portuguese Grid\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Datum 73\",DATUM[\"Datum_73\",SPHEROID[\"International 1");
    add_srs_wkt (p, 2,
		 "924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6274\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4274\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",39.66666666666666],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 9,
		 "-8.131906111111112],PARAMETER[\"scale_factor\",1],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_easting\",180.598],PARAMETER[\"false_northin");
    add_srs_wkt (p, 11,
		 "g\",-86.99],AUTHORITY[\"EPSG\",\"27493\"],AXIS[\"X\",EAS");
    add_srs_wkt (p, 12, "T],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 27500, "epsg", 27500, "unnamed");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49.5 +lat_0=49.5 +lon_0=5.4 +k_0=0.9995");
    add_proj4text (p, 1,
		   "0908 +x_0=500000 +y_0=300000 +a=6376523 +b=6355862.93325");
    add_proj4text (p, 2, "5573 +pm=2.3372291666985 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"unnamed\",GEOGCS[\"unnamed ellipse\",DATUM[\"un");
    add_srs_wkt (p, 1,
		 "known\",SPHEROID[\"unnamed\",6376523,308.6399999999991]]");
    add_srs_wkt (p, 2,
		 ",PRIMEM[\"unnamed\",2.3372291666985],UNIT[\"degree\",0.0");
    add_srs_wkt (p, 3,
		 "174532925199433]],PROJECTION[\"Lambert_Conformal_Conic_1");
    add_srs_wkt (p, 4,
		 "SP\"],PARAMETER[\"latitude_of_origin\",49.5],PARAMETER[\"");
    add_srs_wkt (p, 5,
		 "central_meridian\",5.4],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 6,
		 "50908],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 7,
		 "alse_northing\",300000],UNIT[\"Meter\",1],AUTHORITY[\"EP");
    add_srs_wkt (p, 8, "SG\",\"27500\"]]");
    p = add_epsg_def (first, last, 27561, "epsg", 27561,
		      "NTF (Paris) / Lambert Nord France");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49.50000000000001 +lat_0=49.50000000000");
    add_proj4text (p, 1,
		   "001 +lon_0=0 +k_0=0.999877341 +x_0=600000 +y_0=200000 +a");
    add_proj4text (p, 2,
		   "=6378249.2 +b=6356515 +towgs84=-168,-60,320,0,0,0,0 +pm=");
    add_proj4text (p, 3, "paris +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NTF (Paris) / Lambert Nord France\",GEOGCS[\"NT");
    add_srs_wkt (p, 1,
		 "F (Paris)\",DATUM[\"Nouvelle_Triangulation_Francaise_Par");
    add_srs_wkt (p, 2,
		 "is\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.466021");
    add_srs_wkt (p, 3,
		 "2936265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-168,-60,3");
    add_srs_wkt (p, 4,
		 "20,0,0,0,0],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Paris");
    add_srs_wkt (p, 5,
		 "\",2.33722917,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\"");
    add_srs_wkt (p, 6,
		 ",0.01570796326794897,AUTHORITY[\"EPSG\",\"9105\"]],AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"4807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 8,
		 "SG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1SP");
    add_srs_wkt (p, 9,
		 "\"],PARAMETER[\"latitude_of_origin\",55],PARAMETER[\"cen");
    add_srs_wkt (p, 10,
		 "tral_meridian\",0],PARAMETER[\"scale_factor\",0.99987734");
    add_srs_wkt (p, 11,
		 "1],PARAMETER[\"false_easting\",600000],PARAMETER[\"false");
    add_srs_wkt (p, 12,
		 "_northing\",200000],AUTHORITY[\"EPSG\",\"27561\"],AXIS[\"");
    add_srs_wkt (p, 13, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 27562, "epsg", 27562,
		      "NTF (Paris) / Lambert Centre France");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46.8 +lat_0=46.8 +lon_0=0 +k_0=0.999877");
    add_proj4text (p, 1,
		   "42 +x_0=600000 +y_0=200000 +a=6378249.2 +b=6356515 +towg");
    add_proj4text (p, 2,
		   "s84=-168,-60,320,0,0,0,0 +pm=paris +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NTF (Paris) / Lambert Centre France\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NTF (Paris)\",DATUM[\"Nouvelle_Triangulation_Francaise_P");
    add_srs_wkt (p, 2,
		 "aris\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660");
    add_srs_wkt (p, 3,
		 "212936265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-168,-60");
    add_srs_wkt (p, 4,
		 ",320,0,0,0,0],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Par");
    add_srs_wkt (p, 5,
		 "is\",2.33722917,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"gra");
    add_srs_wkt (p, 6,
		 "d\",0.01570796326794897,AUTHORITY[\"EPSG\",\"9105\"]],AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"4807\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 8,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1");
    add_srs_wkt (p, 9,
		 "SP\"],PARAMETER[\"latitude_of_origin\",52],PARAMETER[\"c");
    add_srs_wkt (p, 10,
		 "entral_meridian\",0],PARAMETER[\"scale_factor\",0.999877");
    add_srs_wkt (p, 11,
		 "42],PARAMETER[\"false_easting\",600000],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_northing\",200000],AUTHORITY[\"EPSG\",\"27562\"],AXIS[");
    add_srs_wkt (p, 13, "\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 27563, "epsg", 27563,
		      "NTF (Paris) / Lambert Sud France");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.10000000000001 +lat_0=44.10000000000");
    add_proj4text (p, 1,
		   "001 +lon_0=0 +k_0=0.999877499 +x_0=600000 +y_0=200000 +a");
    add_proj4text (p, 2,
		   "=6378249.2 +b=6356515 +towgs84=-168,-60,320,0,0,0,0 +pm=");
    add_proj4text (p, 3, "paris +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NTF (Paris) / Lambert Sud France\",GEOGCS[\"NTF");
    add_srs_wkt (p, 1,
		 " (Paris)\",DATUM[\"Nouvelle_Triangulation_Francaise_Pari");
    add_srs_wkt (p, 2,
		 "s\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212");
    add_srs_wkt (p, 3,
		 "936265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-168,-60,32");
    add_srs_wkt (p, 4,
		 "0,0,0,0,0],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Paris\"");
    add_srs_wkt (p, 5,
		 ",2.33722917,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",");
    add_srs_wkt (p, 6,
		 "0.01570796326794897,AUTHORITY[\"EPSG\",\"9105\"]],AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"4807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 8,
		 "G\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1SP\"");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"latitude_of_origin\",49],PARAMETER[\"centr");
    add_srs_wkt (p, 10,
		 "al_meridian\",0],PARAMETER[\"scale_factor\",0.999877499]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_easting\",600000],PARAMETER[\"false_n");
    add_srs_wkt (p, 12,
		 "orthing\",200000],AUTHORITY[\"EPSG\",\"27563\"],AXIS[\"X");
    add_srs_wkt (p, 13, "\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 27564, "epsg", 27564,
		      "NTF (Paris) / Lambert Corse");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=42.16500000000001 +lat_0=42.16500000000");
    add_proj4text (p, 1,
		   "001 +lon_0=0 +k_0=0.99994471 +x_0=234.358 +y_0=185861.36");
    add_proj4text (p, 2,
		   "9 +a=6378249.2 +b=6356515 +towgs84=-168,-60,320,0,0,0,0 ");
    add_proj4text (p, 3, "+pm=paris +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NTF (Paris) / Lambert Corse\",GEOGCS[\"NTF (Par");
    add_srs_wkt (p, 1,
		 "is)\",DATUM[\"Nouvelle_Triangulation_Francaise_Paris\",S");
    add_srs_wkt (p, 2,
		 "PHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.466021293626");
    add_srs_wkt (p, 3,
		 "5,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-168,-60,320,0,0");
    add_srs_wkt (p, 4,
		 ",0,0],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Paris\",2.3");
    add_srs_wkt (p, 5,
		 "3722917,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.01");
    add_srs_wkt (p, 6,
		 "570796326794897,AUTHORITY[\"EPSG\",\"9105\"]],AUTHORITY[");
    add_srs_wkt (p, 7,
		 "\"EPSG\",\"4807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 8,
		 "\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1SP\"],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"latitude_of_origin\",46.85],PARAMETER[\"centr");
    add_srs_wkt (p, 10,
		 "al_meridian\",0],PARAMETER[\"scale_factor\",0.99994471],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_easting\",234.358],PARAMETER[\"false_n");
    add_srs_wkt (p, 12,
		 "orthing\",185861.369],AUTHORITY[\"EPSG\",\"27564\"],AXIS");
    add_srs_wkt (p, 13, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 27571, "epsg", 27571,
		      "NTF (Paris) / Lambert zone I");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49.50000000000001 +lat_0=49.50000000000");
    add_proj4text (p, 1,
		   "001 +lon_0=0 +k_0=0.999877341 +x_0=600000 +y_0=1200000 +");
    add_proj4text (p, 2,
		   "a=6378249.2 +b=6356515 +towgs84=-168,-60,320,0,0,0,0 +pm");
    add_proj4text (p, 3, "=paris +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NTF (Paris) / Lambert zone I\",GEOGCS[\"NTF (Pa");
    add_srs_wkt (p, 1,
		 "ris)\",DATUM[\"Nouvelle_Triangulation_Francaise_Paris\",");
    add_srs_wkt (p, 2,
		 "SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.46602129362");
    add_srs_wkt (p, 3,
		 "65,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-168,-60,320,0,");
    add_srs_wkt (p, 4,
		 "0,0,0],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Paris\",2.");
    add_srs_wkt (p, 5,
		 "33722917,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.0");
    add_srs_wkt (p, 6,
		 "1570796326794897,AUTHORITY[\"EPSG\",\"9105\"]],AUTHORITY");
    add_srs_wkt (p, 7,
		 "[\"EPSG\",\"4807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 8,
		 ",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1SP\"],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"latitude_of_origin\",55],PARAMETER[\"central");
    add_srs_wkt (p, 10,
		 "_meridian\",0],PARAMETER[\"scale_factor\",0.999877341],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_easting\",600000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 12,
		 "thing\",1200000],AUTHORITY[\"EPSG\",\"27571\"],AXIS[\"X\"");
    add_srs_wkt (p, 13, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 27572, "epsg", 27572,
		      "NTF (Paris) / Lambert zone II");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46.8 +lat_0=46.8 +lon_0=0 +k_0=0.999877");
    add_proj4text (p, 1,
		   "42 +x_0=600000 +y_0=2200000 +a=6378249.2 +b=6356515 +tow");
    add_proj4text (p, 2,
		   "gs84=-168,-60,320,0,0,0,0 +pm=paris +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NTF (Paris) / Lambert zone II\",GEOGCS[\"NTF (P");
    add_srs_wkt (p, 1,
		 "aris)\",DATUM[\"Nouvelle_Triangulation_Francaise_Paris\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936");
    add_srs_wkt (p, 3,
		 "265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-168,-60,320,0");
    add_srs_wkt (p, 4,
		 ",0,0,0],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Paris\",2");
    add_srs_wkt (p, 5,
		 ".33722917,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.");
    add_srs_wkt (p, 6,
		 "01570796326794897,AUTHORITY[\"EPSG\",\"9105\"]],AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"4807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 8,
		 ",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1SP\"],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"latitude_of_origin\",52],PARAMETER[\"central");
    add_srs_wkt (p, 10,
		 "_meridian\",0],PARAMETER[\"scale_factor\",0.99987742],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_easting\",600000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 12,
		 "hing\",2200000],AUTHORITY[\"EPSG\",\"27572\"],AXIS[\"X\"");
    add_srs_wkt (p, 13, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 27573, "epsg", 27573,
		      "NTF (Paris) / Lambert zone III");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.10000000000001 +lat_0=44.10000000000");
    add_proj4text (p, 1,
		   "001 +lon_0=0 +k_0=0.999877499 +x_0=600000 +y_0=3200000 +");
    add_proj4text (p, 2,
		   "a=6378249.2 +b=6356515 +towgs84=-168,-60,320,0,0,0,0 +pm");
    add_proj4text (p, 3, "=paris +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NTF (Paris) / Lambert zone III\",GEOGCS[\"NTF (");
    add_srs_wkt (p, 1,
		 "Paris)\",DATUM[\"Nouvelle_Triangulation_Francaise_Paris\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936");
    add_srs_wkt (p, 3,
		 "265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-168,-60,320,0");
    add_srs_wkt (p, 4,
		 ",0,0,0],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Paris\",2");
    add_srs_wkt (p, 5,
		 ".33722917,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.");
    add_srs_wkt (p, 6,
		 "01570796326794897,AUTHORITY[\"EPSG\",\"9105\"]],AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"4807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 8,
		 ",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1SP\"],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"latitude_of_origin\",49],PARAMETER[\"central");
    add_srs_wkt (p, 10,
		 "_meridian\",0],PARAMETER[\"scale_factor\",0.999877499],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_easting\",600000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 12,
		 "thing\",3200000],AUTHORITY[\"EPSG\",\"27573\"],AXIS[\"X\"");
    add_srs_wkt (p, 13, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 27574, "epsg", 27574,
		      "NTF (Paris) / Lambert zone IV");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=42.16500000000001 +lat_0=42.16500000000");
    add_proj4text (p, 1,
		   "001 +lon_0=0 +k_0=0.99994471 +x_0=234.358 +y_0=4185861.3");
    add_proj4text (p, 2,
		   "69 +a=6378249.2 +b=6356515 +towgs84=-168,-60,320,0,0,0,0");
    add_proj4text (p, 3, " +pm=paris +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NTF (Paris) / Lambert zone IV\",GEOGCS[\"NTF (P");
    add_srs_wkt (p, 1,
		 "aris)\",DATUM[\"Nouvelle_Triangulation_Francaise_Paris\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936");
    add_srs_wkt (p, 3,
		 "265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-168,-60,320,0");
    add_srs_wkt (p, 4,
		 ",0,0,0],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Paris\",2");
    add_srs_wkt (p, 5,
		 ".33722917,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.");
    add_srs_wkt (p, 6,
		 "01570796326794897,AUTHORITY[\"EPSG\",\"9105\"]],AUTHORIT");
    add_srs_wkt (p, 7,
		 "Y[\"EPSG\",\"4807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 8,
		 ",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1SP\"],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"latitude_of_origin\",46.85],PARAMETER[\"cent");
    add_srs_wkt (p, 10,
		 "ral_meridian\",0],PARAMETER[\"scale_factor\",0.99994471]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_easting\",234.358],PARAMETER[\"false_");
    add_srs_wkt (p, 12,
		 "northing\",4185861.369],AUTHORITY[\"EPSG\",\"27574\"],AX");
    add_srs_wkt (p, 13, "IS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 27581, "epsg", 27581,
		      "NTF (Paris) / France I (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49.50000000000001 +lat_0=49.50000000000");
    add_proj4text (p, 1,
		   "001 +lon_0=0 +k_0=0.999877341 +x_0=600000 +y_0=1200000 +");
    add_proj4text (p, 2,
		   "a=6378249.2 +b=6356515 +towgs84=-168,-60,320,0,0,0,0 +pm");
    add_proj4text (p, 3, "=paris +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NTF (Paris) / France I (deprecated)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NTF (Paris)\",DATUM[\"Nouvelle_Triangulation_Francaise_P");
    add_srs_wkt (p, 2,
		 "aris\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660");
    add_srs_wkt (p, 3,
		 "212936265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-168,-60");
    add_srs_wkt (p, 4,
		 ",320,0,0,0,0],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Par");
    add_srs_wkt (p, 5,
		 "is\",2.33722917,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"gra");
    add_srs_wkt (p, 6,
		 "d\",0.01570796326794897,AUTHORITY[\"EPSG\",\"9105\"]],AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"4807\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 8,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1");
    add_srs_wkt (p, 9,
		 "SP\"],PARAMETER[\"latitude_of_origin\",55],PARAMETER[\"c");
    add_srs_wkt (p, 10,
		 "entral_meridian\",0],PARAMETER[\"scale_factor\",0.999877");
    add_srs_wkt (p, 11,
		 "341],PARAMETER[\"false_easting\",600000],PARAMETER[\"fal");
    add_srs_wkt (p, 12,
		 "se_northing\",1200000],AUTHORITY[\"EPSG\",\"27581\"],AXI");
    add_srs_wkt (p, 13, "S[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 27582, "epsg", 27582,
		      "NTF (Paris) / France II (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46.8 +lat_0=46.8 +lon_0=0 +k_0=0.999877");
    add_proj4text (p, 1,
		   "42 +x_0=600000 +y_0=2200000 +a=6378249.2 +b=6356515 +tow");
    add_proj4text (p, 2,
		   "gs84=-168,-60,320,0,0,0,0 +pm=paris +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NTF (Paris) / France II (deprecated)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NTF (Paris)\",DATUM[\"Nouvelle_Triangulation_Francaise_P");
    add_srs_wkt (p, 2,
		 "aris\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660");
    add_srs_wkt (p, 3,
		 "212936265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-168,-60");
    add_srs_wkt (p, 4,
		 ",320,0,0,0,0],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Par");
    add_srs_wkt (p, 5,
		 "is\",2.33722917,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"gra");
    add_srs_wkt (p, 6,
		 "d\",0.01570796326794897,AUTHORITY[\"EPSG\",\"9105\"]],AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"4807\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 8,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1");
    add_srs_wkt (p, 9,
		 "SP\"],PARAMETER[\"latitude_of_origin\",52],PARAMETER[\"c");
    add_srs_wkt (p, 10,
		 "entral_meridian\",0],PARAMETER[\"scale_factor\",0.999877");
    add_srs_wkt (p, 11,
		 "42],PARAMETER[\"false_easting\",600000],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_northing\",2200000],AUTHORITY[\"EPSG\",\"27582\"],AXIS");
    add_srs_wkt (p, 13, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 27583, "epsg", 27583,
		      "NTF (Paris) / France III (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.10000000000001 +lat_0=44.10000000000");
    add_proj4text (p, 1,
		   "001 +lon_0=0 +k_0=0.999877499 +x_0=600000 +y_0=3200000 +");
    add_proj4text (p, 2,
		   "a=6378249.2 +b=6356515 +towgs84=-168,-60,320,0,0,0,0 +pm");
    add_proj4text (p, 3, "=paris +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NTF (Paris) / France III (deprecated)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NTF (Paris)\",DATUM[\"Nouvelle_Triangulation_Francaise");
    add_srs_wkt (p, 2,
		 "_Paris\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.46");
    add_srs_wkt (p, 3,
		 "60212936265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-168,-");
    add_srs_wkt (p, 4,
		 "60,320,0,0,0,0],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"P");
    add_srs_wkt (p, 5,
		 "aris\",2.33722917,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"g");
    add_srs_wkt (p, 6,
		 "rad\",0.01570796326794897,AUTHORITY[\"EPSG\",\"9105\"]],");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"4807\"]],UNIT[\"metre\",1,AUTHORITY");
    add_srs_wkt (p, 8,
		 "[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Coni");
    add_srs_wkt (p, 9,
		 "c_1SP\"],PARAMETER[\"latitude_of_origin\",49],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"central_meridian\",0],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 11,
		 "877499],PARAMETER[\"false_easting\",600000],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",3200000],AUTHORITY[\"EPSG\",\"27583\"],");
    add_srs_wkt (p, 13, "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 27584, "epsg", 27584,
		      "NTF (Paris) / France IV (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=42.16500000000001 +lat_0=42.16500000000");
    add_proj4text (p, 1,
		   "001 +lon_0=0 +k_0=0.99994471 +x_0=234.358 +y_0=4185861.3");
    add_proj4text (p, 2,
		   "69 +a=6378249.2 +b=6356515 +towgs84=-168,-60,320,0,0,0,0");
    add_proj4text (p, 3, " +pm=paris +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NTF (Paris) / France IV (deprecated)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NTF (Paris)\",DATUM[\"Nouvelle_Triangulation_Francaise_P");
    add_srs_wkt (p, 2,
		 "aris\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660");
    add_srs_wkt (p, 3,
		 "212936265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-168,-60");
    add_srs_wkt (p, 4,
		 ",320,0,0,0,0],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Par");
    add_srs_wkt (p, 5,
		 "is\",2.33722917,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"gra");
    add_srs_wkt (p, 6,
		 "d\",0.01570796326794897,AUTHORITY[\"EPSG\",\"9105\"]],AU");
    add_srs_wkt (p, 7,
		 "THORITY[\"EPSG\",\"4807\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 8,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1");
    add_srs_wkt (p, 9,
		 "SP\"],PARAMETER[\"latitude_of_origin\",46.85],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"central_meridian\",0],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 11,
		 "94471],PARAMETER[\"false_easting\",234.358],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",4185861.369],AUTHORITY[\"EPSG\",\"27584");
    add_srs_wkt (p, 13, "\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 27591, "epsg", 27591,
		      "NTF (Paris) / Nord France (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49.50000000000001 +lat_0=49.50000000000");
    add_proj4text (p, 1,
		   "001 +lon_0=0 +k_0=0.999877341 +x_0=600000 +y_0=200000 +a");
    add_proj4text (p, 2,
		   "=6378249.2 +b=6356515 +towgs84=-168,-60,320,0,0,0,0 +pm=");
    add_proj4text (p, 3, "paris +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NTF (Paris) / Nord France (deprecated)\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"NTF (Paris)\",DATUM[\"Nouvelle_Triangulation_Francais");
    add_srs_wkt (p, 2,
		 "e_Paris\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4");
    add_srs_wkt (p, 3,
		 "660212936265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-168,");
    add_srs_wkt (p, 4,
		 "-60,320,0,0,0,0],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"");
    add_srs_wkt (p, 5,
		 "Paris\",2.33722917,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"");
    add_srs_wkt (p, 6,
		 "grad\",0.01570796326794897,AUTHORITY[\"EPSG\",\"9105\"]]");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"4807\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 8,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Con");
    add_srs_wkt (p, 9,
		 "ic_1SP\"],PARAMETER[\"latitude_of_origin\",55],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"central_meridian\",0],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 11,
		 "9877341],PARAMETER[\"false_easting\",600000],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",200000],AUTHORITY[\"EPSG\",\"27591\"],A");
    add_srs_wkt (p, 13, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 27592, "epsg", 27592,
		      "NTF (Paris) / Centre France (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46.8 +lat_0=46.8 +lon_0=0 +k_0=0.999877");
    add_proj4text (p, 1,
		   "42 +x_0=600000 +y_0=200000 +a=6378249.2 +b=6356515 +towg");
    add_proj4text (p, 2,
		   "s84=-168,-60,320,0,0,0,0 +pm=paris +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NTF (Paris) / Centre France (deprecated)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"NTF (Paris)\",DATUM[\"Nouvelle_Triangulation_Franca");
    add_srs_wkt (p, 2,
		 "ise_Paris\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293");
    add_srs_wkt (p, 3,
		 ".4660212936265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-16");
    add_srs_wkt (p, 4,
		 "8,-60,320,0,0,0,0],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[");
    add_srs_wkt (p, 5,
		 "\"Paris\",2.33722917,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[");
    add_srs_wkt (p, 6,
		 "\"grad\",0.01570796326794897,AUTHORITY[\"EPSG\",\"9105\"");
    add_srs_wkt (p, 7,
		 "]],AUTHORITY[\"EPSG\",\"4807\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 8,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_C");
    add_srs_wkt (p, 9,
		 "onic_1SP\"],PARAMETER[\"latitude_of_origin\",52],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"central_meridian\",0],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 11,
		 "99987742],PARAMETER[\"false_easting\",600000],PARAMETER[");
    add_srs_wkt (p, 12,
		 "\"false_northing\",200000],AUTHORITY[\"EPSG\",\"27592\"]");
    add_srs_wkt (p, 13, ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 27593, "epsg", 27593,
		      "NTF (Paris) / Sud France (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.10000000000001 +lat_0=44.10000000000");
    add_proj4text (p, 1,
		   "001 +lon_0=0 +k_0=0.999877499 +x_0=600000 +y_0=200000 +a");
    add_proj4text (p, 2,
		   "=6378249.2 +b=6356515 +towgs84=-168,-60,320,0,0,0,0 +pm=");
    add_proj4text (p, 3, "paris +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NTF (Paris) / Sud France (deprecated)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"NTF (Paris)\",DATUM[\"Nouvelle_Triangulation_Francaise");
    add_srs_wkt (p, 2,
		 "_Paris\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.46");
    add_srs_wkt (p, 3,
		 "60212936265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-168,-");
    add_srs_wkt (p, 4,
		 "60,320,0,0,0,0],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"P");
    add_srs_wkt (p, 5,
		 "aris\",2.33722917,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"g");
    add_srs_wkt (p, 6,
		 "rad\",0.01570796326794897,AUTHORITY[\"EPSG\",\"9105\"]],");
    add_srs_wkt (p, 7,
		 "AUTHORITY[\"EPSG\",\"4807\"]],UNIT[\"metre\",1,AUTHORITY");
    add_srs_wkt (p, 8,
		 "[\"EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Coni");
    add_srs_wkt (p, 9,
		 "c_1SP\"],PARAMETER[\"latitude_of_origin\",49],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"central_meridian\",0],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 11,
		 "877499],PARAMETER[\"false_easting\",600000],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",200000],AUTHORITY[\"EPSG\",\"27593\"],A");
    add_srs_wkt (p, 13, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 27594, "epsg", 27594,
		      "NTF (Paris) / Corse (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=42.16500000000001 +lat_0=42.16500000000");
    add_proj4text (p, 1,
		   "001 +lon_0=0 +k_0=0.99994471 +x_0=234.358 +y_0=185861.36");
    add_proj4text (p, 2,
		   "9 +a=6378249.2 +b=6356515 +towgs84=-168,-60,320,0,0,0,0 ");
    add_proj4text (p, 3, "+pm=paris +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NTF (Paris) / Corse (deprecated)\",GEOGCS[\"NTF");
    add_srs_wkt (p, 1,
		 " (Paris)\",DATUM[\"Nouvelle_Triangulation_Francaise_Pari");
    add_srs_wkt (p, 2,
		 "s\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212");
    add_srs_wkt (p, 3,
		 "936265,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-168,-60,32");
    add_srs_wkt (p, 4,
		 "0,0,0,0,0],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Paris\"");
    add_srs_wkt (p, 5,
		 ",2.33722917,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",");
    add_srs_wkt (p, 6,
		 "0.01570796326794897,AUTHORITY[\"EPSG\",\"9105\"]],AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"4807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 8,
		 "G\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1SP\"");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"latitude_of_origin\",46.85],PARAMETER[\"ce");
    add_srs_wkt (p, 10,
		 "ntral_meridian\",0],PARAMETER[\"scale_factor\",0.9999447");
    add_srs_wkt (p, 11,
		 "1],PARAMETER[\"false_easting\",234.358],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_northing\",185861.369],AUTHORITY[\"EPSG\",\"27594\"],A");
    add_srs_wkt (p, 13, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 27700, "epsg", 27700,
		      "OSGB 1936 / British National Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=49 +lon_0=-2 +k=0.9996012717 +x_0=400");
    add_proj4text (p, 1,
		   "000 +y_0=-100000 +ellps=airy +datum=OSGB36 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"OSGB 1936 / British National Grid\",GEOGCS[\"OS");
    add_srs_wkt (p, 1,
		 "GB 1936\",DATUM[\"OSGB_1936\",SPHEROID[\"Airy 1830\",637");
    add_srs_wkt (p, 2,
		 "7563.396,299.3249646,AUTHORITY[\"EPSG\",\"7001\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6277\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"427");
    add_srs_wkt (p, 6,
		 "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",49],PARAMETER[\"central_meridian\",-2],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",0.9996012717],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",400000],PARAMETER[\"false_northing\",-100000],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"27700\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_23 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 28191, "epsg", 28191,
		      "Palestine 1923 / Palestine Grid");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=31.73409694444445 +lon_0=35.2120805555");
    add_proj4text (p, 1,
		   "5556 +x_0=170251.555 +y_0=126867.909 +a=6378300.789 +b=6");
    add_proj4text (p, 2,
		   "356566.435 +towgs84=-275.722,94.7824,340.894,-8.001,-4.4");
    add_proj4text (p, 3, "2,-11.821,1 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Palestine 1923 / Palestine Grid\",GEOGCS[\"Pale");
    add_srs_wkt (p, 1,
		 "stine 1923\",DATUM[\"Palestine_1923\",SPHEROID[\"Clarke ");
    add_srs_wkt (p, 2,
		 "1880 (Benoit)\",6378300.789,293.4663155389802,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7010\"]],TOWGS84[-275.722,94.7824,340.894,-8.");
    add_srs_wkt (p, 4,
		 "001,-4.42,-11.821,1],AUTHORITY[\"EPSG\",\"6281\"]],PRIME");
    add_srs_wkt (p, 5,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 6,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 7,
		 "],AUTHORITY[\"EPSG\",\"4281\"]],UNIT[\"metre\",1,AUTHORI");
    add_srs_wkt (p, 8,
		 "TY[\"EPSG\",\"9001\"]],PROJECTION[\"Cassini_Soldner\"],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"latitude_of_origin\",31.73409694444445],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"central_meridian\",35.21208055555556],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",170251.555],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",126867.909],AUTHORITY[\"EPSG\",\"28191\"],AXIS[\"Eastin");
    add_srs_wkt (p, 13, "g\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 28192, "epsg", 28192,
		      "Palestine 1923 / Palestine Belt");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31.73409694444445 +lon_0=35.212080555");
    add_proj4text (p, 1,
		   "55556 +k=1 +x_0=170251.555 +y_0=1126867.909 +a=6378300.7");
    add_proj4text (p, 2,
		   "89 +b=6356566.435 +towgs84=-275.722,94.7824,340.894,-8.0");
    add_proj4text (p, 3, "01,-4.42,-11.821,1 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Palestine 1923 / Palestine Belt\",GEOGCS[\"Pale");
    add_srs_wkt (p, 1,
		 "stine 1923\",DATUM[\"Palestine_1923\",SPHEROID[\"Clarke ");
    add_srs_wkt (p, 2,
		 "1880 (Benoit)\",6378300.789,293.4663155389802,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7010\"]],TOWGS84[-275.722,94.7824,340.894,-8.");
    add_srs_wkt (p, 4,
		 "001,-4.42,-11.821,1],AUTHORITY[\"EPSG\",\"6281\"]],PRIME");
    add_srs_wkt (p, 5,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 6,
		 "egree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]");
    add_srs_wkt (p, 7,
		 "],AUTHORITY[\"EPSG\",\"4281\"]],UNIT[\"metre\",1,AUTHORI");
    add_srs_wkt (p, 8,
		 "TY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"latitude_of_origin\",31.73409694444445],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"central_meridian\",35.21208055555556],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",170251");
    add_srs_wkt (p, 12,
		 ".555],PARAMETER[\"false_northing\",1126867.909],AUTHORIT");
    add_srs_wkt (p, 13,
		 "Y[\"EPSG\",\"28192\"],AXIS[\"Easting\",EAST],AXIS[\"Nort");
    add_srs_wkt (p, 14, "hing\",NORTH]]");
    p = add_epsg_def (first, last, 28193, "epsg", 28193,
		      "Palestine 1923 / Israeli CS Grid");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=31.73409694444445 +lon_0=35.2120805555");
    add_proj4text (p, 1,
		   "5556 +x_0=170251.555 +y_0=1126867.909 +a=6378300.789 +b=");
    add_proj4text (p, 2,
		   "6356566.435 +towgs84=-275.722,94.7824,340.894,-8.001,-4.");
    add_proj4text (p, 3, "42,-11.821,1 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Palestine 1923 / Israeli CS Grid\",GEOGCS[\"Pal");
    add_srs_wkt (p, 1,
		 "estine 1923\",DATUM[\"Palestine_1923\",SPHEROID[\"Clarke");
    add_srs_wkt (p, 2,
		 " 1880 (Benoit)\",6378300.789,293.4663155389802,AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"7010\"]],TOWGS84[-275.722,94.7824,340.894,-8");
    add_srs_wkt (p, 4,
		 ".001,-4.42,-11.821,1],AUTHORITY[\"EPSG\",\"6281\"]],PRIM");
    add_srs_wkt (p, 5,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 6,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 7,
		 "]],AUTHORITY[\"EPSG\",\"4281\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 8,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Cassini_Soldner\"],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"latitude_of_origin\",31.73409694444445],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"central_meridian\",35.21208055555556],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_easting\",170251.555],PARAMETER[\"false_northing");
    add_srs_wkt (p, 12,
		 "\",1126867.909],AUTHORITY[\"EPSG\",\"28193\"],AXIS[\"Eas");
    add_srs_wkt (p, 13, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 28232, "epsg", 28232,
		      "Pointe Noire / UTM zone 32S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=32 +south +a=6378249.2 +b=6356515 +units");
    add_proj4text (p, 1, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pointe Noire / UTM zone 32S\",GEOGCS[\"Pointe N");
    add_srs_wkt (p, 1,
		 "oire\",DATUM[\"Congo_1960_Pointe_Noire\",SPHEROID[\"Clar");
    add_srs_wkt (p, 2,
		 "ke 1880 (IGN)\",6378249.2,293.4660212936265,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6282\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4282\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",9],PARAMETER[\"scale_factor\",0.9996],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",500000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "10000000],AUTHORITY[\"EPSG\",\"28232\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 12, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 28348, "epsg", 28348, "GDA94 / MGA zone 48");
    add_proj4text (p, 0,
		   "+proj=utm +zone=48 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDA94 / MGA zone 48\",GEOGCS[\"GDA94\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Geocentric_Datum_of_Australia_1994\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4283\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",105],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"28348\"],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 28349, "epsg", 28349, "GDA94 / MGA zone 49");
    add_proj4text (p, 0,
		   "+proj=utm +zone=49 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDA94 / MGA zone 49\",GEOGCS[\"GDA94\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Geocentric_Datum_of_Australia_1994\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4283\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",111],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"28349\"],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 28350, "epsg", 28350, "GDA94 / MGA zone 50");
    add_proj4text (p, 0,
		   "+proj=utm +zone=50 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDA94 / MGA zone 50\",GEOGCS[\"GDA94\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Geocentric_Datum_of_Australia_1994\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4283\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",117],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"28350\"],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 28351, "epsg", 28351, "GDA94 / MGA zone 51");
    add_proj4text (p, 0,
		   "+proj=utm +zone=51 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDA94 / MGA zone 51\",GEOGCS[\"GDA94\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Geocentric_Datum_of_Australia_1994\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4283\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",123],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"28351\"],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 28352, "epsg", 28352, "GDA94 / MGA zone 52");
    add_proj4text (p, 0,
		   "+proj=utm +zone=52 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDA94 / MGA zone 52\",GEOGCS[\"GDA94\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Geocentric_Datum_of_Australia_1994\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4283\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",129],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"28352\"],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 28353, "epsg", 28353, "GDA94 / MGA zone 53");
    add_proj4text (p, 0,
		   "+proj=utm +zone=53 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDA94 / MGA zone 53\",GEOGCS[\"GDA94\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Geocentric_Datum_of_Australia_1994\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4283\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",135],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"28353\"],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 28354, "epsg", 28354, "GDA94 / MGA zone 54");
    add_proj4text (p, 0,
		   "+proj=utm +zone=54 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDA94 / MGA zone 54\",GEOGCS[\"GDA94\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Geocentric_Datum_of_Australia_1994\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4283\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",141],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"28354\"],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 28355, "epsg", 28355, "GDA94 / MGA zone 55");
    add_proj4text (p, 0,
		   "+proj=utm +zone=55 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDA94 / MGA zone 55\",GEOGCS[\"GDA94\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Geocentric_Datum_of_Australia_1994\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4283\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",147],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"28355\"],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 28356, "epsg", 28356, "GDA94 / MGA zone 56");
    add_proj4text (p, 0,
		   "+proj=utm +zone=56 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDA94 / MGA zone 56\",GEOGCS[\"GDA94\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Geocentric_Datum_of_Australia_1994\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4283\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",153],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"28356\"],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 28357, "epsg", 28357, "GDA94 / MGA zone 57");
    add_proj4text (p, 0,
		   "+proj=utm +zone=57 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDA94 / MGA zone 57\",GEOGCS[\"GDA94\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Geocentric_Datum_of_Australia_1994\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4283\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",159],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"28357\"],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 28358, "epsg", 28358, "GDA94 / MGA zone 58");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"GDA94 / MGA zone 58\",GEOGCS[\"GDA94\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Geocentric_Datum_of_Australia_1994\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4283\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",165],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"28358\"],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 28402, "epsg", 28402,
		      "Pulkovo 1942 / Gauss-Kruger zone 2 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=9 +k=1 +x_0=2500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 2 (deprecated)");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"7024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",9],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",2");
    add_srs_wkt (p, 11,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 12, "\",\"28402\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28403, "epsg", 28403,
		      "Pulkovo 1942 / Gauss-Kruger zone 3 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=1 +x_0=3500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 3 (deprecated)");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHERO");
    add_srs_wkt (p, 2,
		 "ID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"7024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12]");
    add_srs_wkt (p, 4,
		 ",AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 6,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",15],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 11,
		 "3500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12, "G\",\"28403\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28404, "epsg", 28404,
		      "Pulkovo 1942 / Gauss-Kruger zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=1 +x_0=4500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 4\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",21],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",4500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28404\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28405, "epsg", 28405,
		      "Pulkovo 1942 / Gauss-Kruger zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=27 +k=1 +x_0=5500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 5\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",27],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",5500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28405\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28406, "epsg", 28406,
		      "Pulkovo 1942 / Gauss-Kruger zone 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=33 +k=1 +x_0=6500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 6\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",33],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",6500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28406\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28407, "epsg", 28407,
		      "Pulkovo 1942 / Gauss-Kruger zone 7");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=39 +k=1 +x_0=7500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 7\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",39],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",7500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28407\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28408, "epsg", 28408,
		      "Pulkovo 1942 / Gauss-Kruger zone 8");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=45 +k=1 +x_0=8500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 8\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",45],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",8500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28408\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28409, "epsg", 28409,
		      "Pulkovo 1942 / Gauss-Kruger zone 9");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=51 +k=1 +x_0=9500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 9\",GEOGCS[\"P");
    add_srs_wkt (p, 1,
		 "ulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassows");
    add_srs_wkt (p, 2,
		 "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",51],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",9500000],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28409\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28410, "epsg", 28410,
		      "Pulkovo 1942 / Gauss-Kruger zone 10");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=57 +k=1 +x_0=10500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 10\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",57],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",10500000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28410\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28411, "epsg", 28411,
		      "Pulkovo 1942 / Gauss-Kruger zone 11");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=63 +k=1 +x_0=11500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 11\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",63],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",11500000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28411\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28412, "epsg", 28412,
		      "Pulkovo 1942 / Gauss-Kruger zone 12");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=69 +k=1 +x_0=12500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 12\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",69],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",12500000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28412\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28413, "epsg", 28413,
		      "Pulkovo 1942 / Gauss-Kruger zone 13");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=75 +k=1 +x_0=13500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 13\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",75],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",13500000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28413\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28414, "epsg", 28414,
		      "Pulkovo 1942 / Gauss-Kruger zone 14");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=81 +k=1 +x_0=14500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 14\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",81],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",14500000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28414\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28415, "epsg", 28415,
		      "Pulkovo 1942 / Gauss-Kruger zone 15");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=87 +k=1 +x_0=15500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 15\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",87],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",15500000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28415\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28416, "epsg", 28416,
		      "Pulkovo 1942 / Gauss-Kruger zone 16");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=93 +k=1 +x_0=16500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 16\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",93],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",16500000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28416\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28417, "epsg", 28417,
		      "Pulkovo 1942 / Gauss-Kruger zone 17");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=99 +k=1 +x_0=17500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 17\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",99],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",1],PARAMETER[\"false_easting\",17500000],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28417\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28418, "epsg", 28418,
		      "Pulkovo 1942 / Gauss-Kruger zone 18");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=18500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 18\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",105],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",18500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28418");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28419, "epsg", 28419,
		      "Pulkovo 1942 / Gauss-Kruger zone 19");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=19500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 19\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",111],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",19500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28419");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28420, "epsg", 28420,
		      "Pulkovo 1942 / Gauss-Kruger zone 20");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=1 +x_0=20500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 20\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",117],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",20500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28420");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28421, "epsg", 28421,
		      "Pulkovo 1942 / Gauss-Kruger zone 21");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=1 +x_0=21500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 21\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",123],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",21500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28421");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28422, "epsg", 28422,
		      "Pulkovo 1942 / Gauss-Kruger zone 22");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=129 +k=1 +x_0=22500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 22\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",129],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",22500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28422");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28423, "epsg", 28423,
		      "Pulkovo 1942 / Gauss-Kruger zone 23");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=135 +k=1 +x_0=23500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 23\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",135],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",23500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28423");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28424, "epsg", 28424,
		      "Pulkovo 1942 / Gauss-Kruger zone 24");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=141 +k=1 +x_0=24500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 24\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",141],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",24500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28424");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28425, "epsg", 28425,
		      "Pulkovo 1942 / Gauss-Kruger zone 25");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=147 +k=1 +x_0=25500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 25\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",147],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",25500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28425");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28426, "epsg", 28426,
		      "Pulkovo 1942 / Gauss-Kruger zone 26");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=153 +k=1 +x_0=26500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 26\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",153],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",26500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28426");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28427, "epsg", 28427,
		      "Pulkovo 1942 / Gauss-Kruger zone 27");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=159 +k=1 +x_0=27500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 27\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",159],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",27500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28427");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28428, "epsg", 28428,
		      "Pulkovo 1942 / Gauss-Kruger zone 28");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=165 +k=1 +x_0=28500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 28\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",165],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",28500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28428");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28429, "epsg", 28429,
		      "Pulkovo 1942 / Gauss-Kruger zone 29");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=171 +k=1 +x_0=29500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 29\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",171],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",29500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28429");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28430, "epsg", 28430,
		      "Pulkovo 1942 / Gauss-Kruger zone 30");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=177 +k=1 +x_0=30500000 +y_0=");
    add_proj4text (p, 1, "0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 30\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",177],PARAMETER[\"sc");
    add_srs_wkt (p, 10,
		 "ale_factor\",1],PARAMETER[\"false_easting\",30500000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"28430");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28431, "epsg", 28431,
		      "Pulkovo 1942 / Gauss-Kruger zone 31");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-177 +k=1 +x_0=31500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 31\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",-177],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",1],PARAMETER[\"false_easting\",31500000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2843");
    add_srs_wkt (p, 12, "1\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28432, "epsg", 28432,
		      "Pulkovo 1942 / Gauss-Kruger zone 32");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-171 +k=1 +x_0=32500000 +y_0");
    add_proj4text (p, 1, "=0 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger zone 32\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassow");
    add_srs_wkt (p, 2,
		 "sky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 9,
		 "\",0],PARAMETER[\"central_meridian\",-171],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",1],PARAMETER[\"false_easting\",32500000],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2843");
    add_srs_wkt (p, 12, "2\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28462, "epsg", 28462,
		      "Pulkovo 1942 / Gauss-Kruger 2N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=9 +k=1 +x_0=500000 +y_0=0 +e");
    add_proj4text (p, 1, "llps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 2N (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",9],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"scale_factor\",1],PARAMETER[\"false_easting\",500000");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"2");
    add_srs_wkt (p, 12, "8462\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28463, "epsg", 28463,
		      "Pulkovo 1942 / Gauss-Kruger 3N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 3N (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",15],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "28463\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28464, "epsg", 28464,
		      "Pulkovo 1942 / Gauss-Kruger 4N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 4N (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",21],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "28464\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28465, "epsg", 28465,
		      "Pulkovo 1942 / Gauss-Kruger 5N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=27 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 5N (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",27],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "28465\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28466, "epsg", 28466,
		      "Pulkovo 1942 / Gauss-Kruger 6N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=33 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 6N (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",33],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "28466\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28467, "epsg", 28467,
		      "Pulkovo 1942 / Gauss-Kruger 7N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=39 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 7N (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",39],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "28467\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28468, "epsg", 28468,
		      "Pulkovo 1942 / Gauss-Kruger 8N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=45 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 8N (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",45],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "28468\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28469, "epsg", 28469,
		      "Pulkovo 1942 / Gauss-Kruger 9N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=51 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 9N (deprecated)\",G");
    add_srs_wkt (p, 1,
		 "EOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "4\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 5,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 6,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 7,
		 "84\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 8,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",0],PARAMETER[\"central_meridian\",51],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"scale_factor\",1],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "28469\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28470, "epsg", 28470,
		      "Pulkovo 1942 / Gauss-Kruger 10N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=57 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 10N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",57],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"28470\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28471, "epsg", 28471,
		      "Pulkovo 1942 / Gauss-Kruger 11N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=63 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 11N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",63],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"28471\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28472, "epsg", 28472,
		      "Pulkovo 1942 / Gauss-Kruger 12N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=69 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 12N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",69],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"28472\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28473, "epsg", 28473,
		      "Pulkovo 1942 / Gauss-Kruger 13N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=75 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 13N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",75],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"28473\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28474, "epsg", 28474,
		      "Pulkovo 1942 / Gauss-Kruger 14N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=81 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 14N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",81],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"28474\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28475, "epsg", 28475,
		      "Pulkovo 1942 / Gauss-Kruger 15N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=87 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 15N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",87],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"28475\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28476, "epsg", 28476,
		      "Pulkovo 1942 / Gauss-Kruger 16N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=93 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 16N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",93],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"28476\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28477, "epsg", 28477,
		      "Pulkovo 1942 / Gauss-Kruger 17N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=99 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 17N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",99],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",1],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"28477\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28478, "epsg", 28478,
		      "Pulkovo 1942 / Gauss-Kruger 18N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=105 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 18N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",105],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"28478\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28479, "epsg", 28479,
		      "Pulkovo 1942 / Gauss-Kruger 19N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=111 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 19N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",111],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"28479\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28480, "epsg", 28480,
		      "Pulkovo 1942 / Gauss-Kruger 20N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=117 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 20N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",117],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"28480\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28481, "epsg", 28481,
		      "Pulkovo 1942 / Gauss-Kruger 21N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=123 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 21N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",123],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"28481\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28482, "epsg", 28482,
		      "Pulkovo 1942 / Gauss-Kruger 22N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=129 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 22N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",129],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"28482\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28483, "epsg", 28483,
		      "Pulkovo 1942 / Gauss-Kruger 23N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=135 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 23N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",135],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"28483\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28484, "epsg", 28484,
		      "Pulkovo 1942 / Gauss-Kruger 24N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=141 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 24N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",141],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"28484\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28485, "epsg", 28485,
		      "Pulkovo 1942 / Gauss-Kruger 25N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=147 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 25N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",147],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"28485\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28486, "epsg", 28486,
		      "Pulkovo 1942 / Gauss-Kruger 26N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=153 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 26N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",153],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"28486\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28487, "epsg", 28487,
		      "Pulkovo 1942 / Gauss-Kruger 27N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=159 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 27N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",159],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"28487\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28488, "epsg", 28488,
		      "Pulkovo 1942 / Gauss-Kruger 28N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=165 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 28N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",165],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"28488\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28489, "epsg", 28489,
		      "Pulkovo 1942 / Gauss-Kruger 29N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=171 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 29N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",171],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"28489\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28490, "epsg", 28490,
		      "Pulkovo 1942 / Gauss-Kruger 30N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=177 +k=1 +x_0=500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 30N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",177],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"scale_factor\",1],PARAMETER[\"false_easting\",50");
    add_srs_wkt (p, 11,
		 "0000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"28490\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28491, "epsg", 28491,
		      "Pulkovo 1942 / Gauss-Kruger 31N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-177 +k=1 +x_0=500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 31N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-177],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"28491\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28492, "epsg", 28492,
		      "Pulkovo 1942 / Gauss-Kruger 32N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-171 +k=1 +x_0=500000 +y_0=0");
    add_proj4text (p, 1, " +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Pulkovo 1942 / Gauss-Kruger 32N (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "024\"]],TOWGS84[23.9,-141.3,-80.9,-0,0.37,0.85,-0.12],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4284\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-171],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"28492\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 28600, "epsg", 28600,
		      "Qatar 1974 / Qatar National Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=24.45 +lon_0=51.21666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "9999 +x_0=200000 +y_0=300000 +ellps=intl +units=m +no_de");
    add_proj4text (p, 2, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Qatar 1974 / Qatar National Grid\",GEOGCS[\"Qat");
    add_srs_wkt (p, 1,
		 "ar 1974\",DATUM[\"Qatar_1974\",SPHEROID[\"International ");
    add_srs_wkt (p, 2,
		 "1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6285\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4285\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 8,
		 "igin\",24.45],PARAMETER[\"central_meridian\",51.21666666");
    add_srs_wkt (p, 9,
		 "666667],PARAMETER[\"scale_factor\",0.99999],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",200000],PARAMETER[\"false_northing\",300");
    add_srs_wkt (p, 11,
		 "000],AUTHORITY[\"EPSG\",\"28600\"],AXIS[\"Easting\",EAST");
    add_srs_wkt (p, 12, "],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 28991, "epsg", 28991, "Amersfoort / RD Old");
    add_proj4text (p, 0,
		   "+proj=sterea +lat_0=52.15616055555555 +lon_0=5.387638888");
    add_proj4text (p, 1,
		   "88889 +k=0.9999079 +x_0=0 +y_0=0 +ellps=bessel +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Amersfoort / RD Old\",GEOGCS[\"Amersfoort\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Amersfoort\",SPHEROID[\"Bessel 1841\",6377397.155,2");
    add_srs_wkt (p, 2,
		 "99.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"6289\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4289\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Oblique_Stereographic\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",52.15616055555555],PARAMETER[\"central_meridian\",5.387");
    add_srs_wkt (p, 9,
		 "63888888889],PARAMETER[\"scale_factor\",0.9999079],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_easting\",0],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 11,
		 "],AUTHORITY[\"EPSG\",\"28991\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 28992, "epsg", 28992, "Amersfoort / RD New");
    add_proj4text (p, 0,
		   "+proj=sterea +lat_0=52.15616055555555 +lon_0=5.387638888");
    add_proj4text (p, 1,
		   "88889 +k=0.9999079 +x_0=155000 +y_0=463000 +ellps=bessel");
    add_proj4text (p, 2, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Amersfoort / RD New\",GEOGCS[\"Amersfoort\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"Amersfoort\",SPHEROID[\"Bessel 1841\",6377397.155,2");
    add_srs_wkt (p, 2,
		 "99.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"6289\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4289\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Oblique_Stereographic\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",52.15616055555555],PARAMETER[\"central_meridian\",5.387");
    add_srs_wkt (p, 9,
		 "63888888889],PARAMETER[\"scale_factor\",0.9999079],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_easting\",155000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 11,
		 "g\",463000],AUTHORITY[\"EPSG\",\"28992\"],AXIS[\"X\",EAS");
    add_srs_wkt (p, 12, "T],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 29100, "epsg", 29100,
		      "SAD69 / Brazil Polyconic (deprecated)");
    add_proj4text (p, 0,
		   "+proj=poly +lat_0=0 +lon_0=-54 +x_0=5000000 +y_0=1000000");
    add_proj4text (p, 1, "0 +ellps=GRS67 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / Brazil Polyconic (deprecated)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"SAD69\",DATUM[\"South_American_Datum_1969\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "GRS 1967\",6378160,298.247167427,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "36\"]],AUTHORITY[\"EPSG\",\"6291\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "32925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4291\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Polyconic\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",0],PARAMETER[\"central_meridian\",-54],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"false_easting\",5000000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 10,
		 "\",10000000],AUTHORITY[\"EPSG\",\"29100\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 11, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 29101, "epsg", 29101,
		      "SAD69 / Brazil Polyconic");
    add_proj4text (p, 0,
		   "+proj=poly +lat_0=0 +lon_0=-54 +x_0=5000000 +y_0=1000000");
    add_proj4text (p, 1, "0 +ellps=aust_SA +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / Brazil Polyconic\",GEOGCS[\"SAD69\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"South_American_Datum_1969\",SPHEROID[\"GRS 1967 Mod");
    add_srs_wkt (p, 2,
		 "ified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "618\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Polyconic\"],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 8,
		 ",0],PARAMETER[\"central_meridian\",-54],PARAMETER[\"fals");
    add_srs_wkt (p, 9,
		 "e_easting\",5000000],PARAMETER[\"false_northing\",100000");
    add_srs_wkt (p, 10,
		 "00],AUTHORITY[\"EPSG\",\"29101\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 11, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 29118, "epsg", 29118,
		      "SAD69 / UTM zone 18N (deprecated)");
    add_proj4text (p, 0, "+proj=utm +zone=18 +ellps=GRS67 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 18N (deprecated)\",GEOGCS[\"SA");
    add_srs_wkt (p, 1,
		 "D69\",DATUM[\"South_American_Datum_1969\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1967\",6378160,298.247167427,AUTHORITY[\"EPSG\",\"7036\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6291\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "25199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4291\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-75],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"29118\"],AXIS[\"Easting\",EAST],AXIS[\"Northi");
    add_srs_wkt (p, 12, "ng\",NORTH]]");
    p = add_epsg_def (first, last, 29119, "epsg", 29119,
		      "SAD69 / UTM zone 19N (deprecated)");
    add_proj4text (p, 0, "+proj=utm +zone=19 +ellps=GRS67 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 19N (deprecated)\",GEOGCS[\"SA");
    add_srs_wkt (p, 1,
		 "D69\",DATUM[\"South_American_Datum_1969\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1967\",6378160,298.247167427,AUTHORITY[\"EPSG\",\"7036\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6291\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "25199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4291\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-69],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"29119\"],AXIS[\"Easting\",EAST],AXIS[\"Northi");
    add_srs_wkt (p, 12, "ng\",NORTH]]");
    p = add_epsg_def (first, last, 29120, "epsg", 29120,
		      "SAD69 / UTM zone 20N (deprecated)");
    add_proj4text (p, 0, "+proj=utm +zone=20 +ellps=GRS67 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 20N (deprecated)\",GEOGCS[\"SA");
    add_srs_wkt (p, 1,
		 "D69\",DATUM[\"South_American_Datum_1969\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1967\",6378160,298.247167427,AUTHORITY[\"EPSG\",\"7036\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6291\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "25199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4291\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-63],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"29120\"],AXIS[\"Easting\",EAST],AXIS[\"Northi");
    add_srs_wkt (p, 12, "ng\",NORTH]]");
    p = add_epsg_def (first, last, 29121, "epsg", 29121,
		      "SAD69 / UTM zone 21N (deprecated)");
    add_proj4text (p, 0, "+proj=utm +zone=21 +ellps=GRS67 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 21N (deprecated)\",GEOGCS[\"SA");
    add_srs_wkt (p, 1,
		 "D69\",DATUM[\"South_American_Datum_1969\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1967\",6378160,298.247167427,AUTHORITY[\"EPSG\",\"7036\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6291\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "25199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4291\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-57],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"29121\"],AXIS[\"Easting\",EAST],AXIS[\"Northi");
    add_srs_wkt (p, 12, "ng\",NORTH]]");
    p = add_epsg_def (first, last, 29122, "epsg", 29122,
		      "SAD69 / UTM zone 22N (deprecated)");
    add_proj4text (p, 0, "+proj=utm +zone=22 +ellps=GRS67 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 22N (deprecated)\",GEOGCS[\"SA");
    add_srs_wkt (p, 1,
		 "D69\",DATUM[\"South_American_Datum_1969\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1967\",6378160,298.247167427,AUTHORITY[\"EPSG\",\"7036\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6291\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "25199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4291\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-51],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"29122\"],AXIS[\"Easting\",EAST],AXIS[\"Northi");
    add_srs_wkt (p, 12, "ng\",NORTH]]");
    p = add_epsg_def (first, last, 29168, "epsg", 29168,
		      "SAD69 / UTM zone 18N");
    add_proj4text (p, 0, "+proj=utm +zone=18 +ellps=aust_SA +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 18N\",GEOGCS[\"SAD69\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "South_American_Datum_1969\",SPHEROID[\"GRS 1967 Modified");
    add_srs_wkt (p, 2,
		 "\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4618\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",-75],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "29168\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 29169, "epsg", 29169,
		      "SAD69 / UTM zone 19N");
    add_proj4text (p, 0, "+proj=utm +zone=19 +ellps=aust_SA +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 19N\",GEOGCS[\"SAD69\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "South_American_Datum_1969\",SPHEROID[\"GRS 1967 Modified");
    add_srs_wkt (p, 2,
		 "\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4618\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",-69],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "29169\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 29170, "epsg", 29170,
		      "SAD69 / UTM zone 20N");
    add_proj4text (p, 0, "+proj=utm +zone=20 +ellps=aust_SA +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 20N\",GEOGCS[\"SAD69\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "South_American_Datum_1969\",SPHEROID[\"GRS 1967 Modified");
    add_srs_wkt (p, 2,
		 "\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4618\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",-63],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "29170\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 29171, "epsg", 29171,
		      "SAD69 / UTM zone 21N");
    add_proj4text (p, 0, "+proj=utm +zone=21 +ellps=aust_SA +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 21N\",GEOGCS[\"SAD69\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "South_American_Datum_1969\",SPHEROID[\"GRS 1967 Modified");
    add_srs_wkt (p, 2,
		 "\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4618\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",-57],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "29171\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 29172, "epsg", 29172,
		      "SAD69 / UTM zone 22N");
    add_proj4text (p, 0, "+proj=utm +zone=22 +ellps=aust_SA +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 22N\",GEOGCS[\"SAD69\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "South_American_Datum_1969\",SPHEROID[\"GRS 1967 Modified");
    add_srs_wkt (p, 2,
		 "\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4618\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",-51],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "29172\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 29177, "epsg", 29177,
		      "SAD69 / UTM zone 17S (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=17 +south +ellps=GRS67 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 17S (deprecated)\",GEOGCS[\"SA");
    add_srs_wkt (p, 1,
		 "D69\",DATUM[\"South_American_Datum_1969\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1967\",6378160,298.247167427,AUTHORITY[\"EPSG\",\"7036\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6291\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "25199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4291\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-81],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",10000000],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"29177\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 29178, "epsg", 29178,
		      "SAD69 / UTM zone 18S (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +south +ellps=GRS67 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 18S (deprecated)\",GEOGCS[\"SA");
    add_srs_wkt (p, 1,
		 "D69\",DATUM[\"South_American_Datum_1969\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1967\",6378160,298.247167427,AUTHORITY[\"EPSG\",\"7036\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6291\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "25199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4291\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-75],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",10000000],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"29178\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 29179, "epsg", 29179,
		      "SAD69 / UTM zone 19S (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +south +ellps=GRS67 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 19S (deprecated)\",GEOGCS[\"SA");
    add_srs_wkt (p, 1,
		 "D69\",DATUM[\"South_American_Datum_1969\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1967\",6378160,298.247167427,AUTHORITY[\"EPSG\",\"7036\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6291\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "25199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4291\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-69],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",10000000],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"29179\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 29180, "epsg", 29180,
		      "SAD69 / UTM zone 20S (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +south +ellps=GRS67 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 20S (deprecated)\",GEOGCS[\"SA");
    add_srs_wkt (p, 1,
		 "D69\",DATUM[\"South_American_Datum_1969\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1967\",6378160,298.247167427,AUTHORITY[\"EPSG\",\"7036\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6291\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "25199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4291\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-63],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",10000000],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"29180\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 29181, "epsg", 29181,
		      "SAD69 / UTM zone 21S (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +south +ellps=GRS67 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 21S (deprecated)\",GEOGCS[\"SA");
    add_srs_wkt (p, 1,
		 "D69\",DATUM[\"South_American_Datum_1969\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1967\",6378160,298.247167427,AUTHORITY[\"EPSG\",\"7036\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6291\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "25199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4291\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-57],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",10000000],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"29181\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 29182, "epsg", 29182,
		      "SAD69 / UTM zone 22S (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=22 +south +ellps=GRS67 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 22S (deprecated)\",GEOGCS[\"SA");
    add_srs_wkt (p, 1,
		 "D69\",DATUM[\"South_American_Datum_1969\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1967\",6378160,298.247167427,AUTHORITY[\"EPSG\",\"7036\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6291\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "25199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4291\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-51],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",10000000],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"29182\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 29183, "epsg", 29183,
		      "SAD69 / UTM zone 23S (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=23 +south +ellps=GRS67 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 23S (deprecated)\",GEOGCS[\"SA");
    add_srs_wkt (p, 1,
		 "D69\",DATUM[\"South_American_Datum_1969\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1967\",6378160,298.247167427,AUTHORITY[\"EPSG\",\"7036\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6291\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "25199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4291\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-45],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",10000000],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"29183\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 29184, "epsg", 29184,
		      "SAD69 / UTM zone 24S (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=24 +south +ellps=GRS67 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 24S (deprecated)\",GEOGCS[\"SA");
    add_srs_wkt (p, 1,
		 "D69\",DATUM[\"South_American_Datum_1969\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1967\",6378160,298.247167427,AUTHORITY[\"EPSG\",\"7036\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6291\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "25199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4291\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-39],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",10000000],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"29184\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 29185, "epsg", 29185,
		      "SAD69 / UTM zone 25S (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=25 +south +ellps=GRS67 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 25S (deprecated)\",GEOGCS[\"SA");
    add_srs_wkt (p, 1,
		 "D69\",DATUM[\"South_American_Datum_1969\",SPHEROID[\"GRS");
    add_srs_wkt (p, 2,
		 " 1967\",6378160,298.247167427,AUTHORITY[\"EPSG\",\"7036\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6291\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "25199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4291\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-33],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",500000],PARAMETER[\"false_northing\",10000000],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"29185\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 29187, "epsg", 29187,
		      "SAD69 / UTM zone 17S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=17 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 17S\",GEOGCS[\"SAD69\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "South_American_Datum_1969\",SPHEROID[\"GRS 1967 Modified");
    add_srs_wkt (p, 2,
		 "\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4618\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",-81],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",10000000],AUTHORITY[\"EP");
    add_srs_wkt (p, 11,
		 "SG\",\"29187\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 29188, "epsg", 29188,
		      "SAD69 / UTM zone 18S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 18S\",GEOGCS[\"SAD69\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "South_American_Datum_1969\",SPHEROID[\"GRS 1967 Modified");
    add_srs_wkt (p, 2,
		 "\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4618\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",-75],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",10000000],AUTHORITY[\"EP");
    add_srs_wkt (p, 11,
		 "SG\",\"29188\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 29189, "epsg", 29189,
		      "SAD69 / UTM zone 19S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 19S\",GEOGCS[\"SAD69\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "South_American_Datum_1969\",SPHEROID[\"GRS 1967 Modified");
    add_srs_wkt (p, 2,
		 "\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4618\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",-69],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",10000000],AUTHORITY[\"EP");
    add_srs_wkt (p, 11,
		 "SG\",\"29189\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 29190, "epsg", 29190,
		      "SAD69 / UTM zone 20S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 20S\",GEOGCS[\"SAD69\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "South_American_Datum_1969\",SPHEROID[\"GRS 1967 Modified");
    add_srs_wkt (p, 2,
		 "\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4618\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",-63],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",10000000],AUTHORITY[\"EP");
    add_srs_wkt (p, 11,
		 "SG\",\"29190\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 29191, "epsg", 29191,
		      "SAD69 / UTM zone 21S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 21S\",GEOGCS[\"SAD69\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "South_American_Datum_1969\",SPHEROID[\"GRS 1967 Modified");
    add_srs_wkt (p, 2,
		 "\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4618\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",-57],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",10000000],AUTHORITY[\"EP");
    add_srs_wkt (p, 11,
		 "SG\",\"29191\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 29192, "epsg", 29192,
		      "SAD69 / UTM zone 22S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=22 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 22S\",GEOGCS[\"SAD69\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "South_American_Datum_1969\",SPHEROID[\"GRS 1967 Modified");
    add_srs_wkt (p, 2,
		 "\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4618\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",-51],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",10000000],AUTHORITY[\"EP");
    add_srs_wkt (p, 11,
		 "SG\",\"29192\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 29193, "epsg", 29193,
		      "SAD69 / UTM zone 23S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=23 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 23S\",GEOGCS[\"SAD69\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "South_American_Datum_1969\",SPHEROID[\"GRS 1967 Modified");
    add_srs_wkt (p, 2,
		 "\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4618\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",-45],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",10000000],AUTHORITY[\"EP");
    add_srs_wkt (p, 11,
		 "SG\",\"29193\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 29194, "epsg", 29194,
		      "SAD69 / UTM zone 24S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=24 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 24S\",GEOGCS[\"SAD69\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "South_American_Datum_1969\",SPHEROID[\"GRS 1967 Modified");
    add_srs_wkt (p, 2,
		 "\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4618\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",-39],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",10000000],AUTHORITY[\"EP");
    add_srs_wkt (p, 11,
		 "SG\",\"29194\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 29195, "epsg", 29195,
		      "SAD69 / UTM zone 25S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=25 +south +ellps=aust_SA +units=m +no_de");
    add_proj4text (p, 1, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SAD69 / UTM zone 25S\",GEOGCS[\"SAD69\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "South_American_Datum_1969\",SPHEROID[\"GRS 1967 Modified");
    add_srs_wkt (p, 2,
		 "\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4618\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",-33],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9996],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",10000000],AUTHORITY[\"EP");
    add_srs_wkt (p, 11,
		 "SG\",\"29195\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 29220, "epsg", 29220,
		      "Sapper Hill 1943 / UTM zone 20S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +south +ellps=intl +towgs84=-355,21,7");
    add_proj4text (p, 1, "2,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Sapper Hill 1943 / UTM zone 20S\",GEOGCS[\"Sapp");
    add_srs_wkt (p, 1,
		 "er Hill 1943\",DATUM[\"Sapper_Hill_1943\",SPHEROID[\"Int");
    add_srs_wkt (p, 2,
		 "ernational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[-355,21,72,0,0,0,0],AUTHORITY[\"EPSG\",\"6292");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4292\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-63],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"29220\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 29221, "epsg", 29221,
		      "Sapper Hill 1943 / UTM zone 21S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +south +ellps=intl +towgs84=-355,21,7");
    add_proj4text (p, 1, "2,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Sapper Hill 1943 / UTM zone 21S\",GEOGCS[\"Sapp");
    add_srs_wkt (p, 1,
		 "er Hill 1943\",DATUM[\"Sapper_Hill_1943\",SPHEROID[\"Int");
    add_srs_wkt (p, 2,
		 "ernational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[-355,21,72,0,0,0,0],AUTHORITY[\"EPSG\",\"6292");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4292\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-57],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"29221\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 29333, "epsg", 29333,
		      "Schwarzeck / UTM zone 33S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=33 +south +ellps=bess_nam +units=m +no_d");
    add_proj4text (p, 1, "efs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Schwarzeck / UTM zone 33S\",GEOGCS[\"Schwarzeck");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"Schwarzeck\",SPHEROID[\"Bessel Namibia (GLM)\"");
    add_srs_wkt (p, 2,
		 ",6377483.865280419,299.1528128,AUTHORITY[\"EPSG\",\"7046");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6293\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4293\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 7,
		 "01\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"l");
    add_srs_wkt (p, 8,
		 "atitude_of_origin\",0],PARAMETER[\"central_meridian\",15");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 10,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"29333\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 12, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 29635, "epsg", 29635,
		      "Sudan / UTM zone 35N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=35 +a=6378249.2 +b=6356515 +units=m +no_");
    add_proj4text (p, 1, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Sudan / UTM zone 35N (deprecated)\",GEOGCS[\"Su");
    add_srs_wkt (p, 1,
		 "dan\",DATUM[\"Sudan\",SPHEROID[\"Clarke 1880 (IGN)\",637");
    add_srs_wkt (p, 2,
		 "8249.2,293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6296\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "9433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4296\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",27],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 11,
		 "G\",\"29635\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",");
    add_srs_wkt (p, 12, "NORTH]]");
    p = add_epsg_def (first, last, 29636, "epsg", 29636,
		      "Sudan / UTM zone 36N (deprecated)");
    add_proj4text (p, 0,
		   "+proj=utm +zone=36 +a=6378249.2 +b=6356515 +units=m +no_");
    add_proj4text (p, 1, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Sudan / UTM zone 36N (deprecated)\",GEOGCS[\"Su");
    add_srs_wkt (p, 1,
		 "dan\",DATUM[\"Sudan\",SPHEROID[\"Clarke 1880 (IGN)\",637");
    add_srs_wkt (p, 2,
		 "8249.2,293.4660212936265,AUTHORITY[\"EPSG\",\"7011\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6296\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "9433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4296\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",33],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 11,
		 "G\",\"29636\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",");
    add_srs_wkt (p, 12, "NORTH]]");
    p = add_epsg_def (first, last, 29700, "epsg", 29700,
		      "Tananarive (Paris) / Laborde Grid (deprecated)");
    add_proj4text (p, 0,
		   "+proj=omerc +lat_0=-18.9 +lonc=44.10000000000001 +alpha=");
    add_proj4text (p, 1,
		   "18.9 +k=0.9995000000000001 +x_0=400000 +y_0=800000 +ellp");
    add_proj4text (p, 2,
		   "s=intl +towgs84=-189,-242,-91,0,0,0,0 +pm=paris +units=m");
    add_proj4text (p, 3, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tananarive (Paris) / Laborde Grid (deprecated)\"");
    add_srs_wkt (p, 1,
		 ",GEOGCS[\"Tananarive (Paris)\",DATUM[\"Tananarive_1925_P");
    add_srs_wkt (p, 2,
		 "aris\",SPHEROID[\"International 1924\",6378388,297,AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"7022\"]],TOWGS84[-189,-242,-91,0,0,0,0],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6810\"]],PRIMEM[\"Paris\",2.3372291");
    add_srs_wkt (p, 5,
		 "7,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.01570796");
    add_srs_wkt (p, 6,
		 "326794897,AUTHORITY[\"EPSG\",\"9105\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"4810\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 8,
		 "\"]],PROJECTION[\"Hotine_Oblique_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "latitude_of_center\",-21],PARAMETER[\"longitude_of_cente");
    add_srs_wkt (p, 10,
		 "r\",49],PARAMETER[\"azimuth\",21],PARAMETER[\"rectified_");
    add_srs_wkt (p, 11,
		 "grid_angle\",21],PARAMETER[\"scale_factor\",0.9995],PARA");
    add_srs_wkt (p, 12,
		 "METER[\"false_easting\",400000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 13,
		 "ng\",800000],AUTHORITY[\"EPSG\",\"29700\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 14, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 29702, "epsg", 29702,
		      "Tananarive (Paris) / Laborde Grid approximation");
    add_proj4text (p, 0,
		   "+proj=omerc +lat_0=-18.9 +lonc=44.10000000000001 +alpha=");
    add_proj4text (p, 1,
		   "18.9 +k=0.9995000000000001 +x_0=400000 +y_0=800000 +ellp");
    add_proj4text (p, 2,
		   "s=intl +towgs84=-189,-242,-91,0,0,0,0 +pm=paris +units=m");
    add_proj4text (p, 3, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tananarive (Paris) / Laborde Grid approximation");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"Tananarive (Paris)\",DATUM[\"Tananarive_1925");
    add_srs_wkt (p, 2,
		 "_Paris\",SPHEROID[\"International 1924\",6378388,297,AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"7022\"]],TOWGS84[-189,-242,-91,0,0,0,0");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6810\"]],PRIMEM[\"Paris\",2.33722");
    add_srs_wkt (p, 5,
		 "917,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.015707");
    add_srs_wkt (p, 6,
		 "96326794897,AUTHORITY[\"EPSG\",\"9105\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"4810\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90");
    add_srs_wkt (p, 8,
		 "01\"]],PROJECTION[\"Hotine_Oblique_Mercator\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"latitude_of_center\",-21],PARAMETER[\"longitude_of_ce");
    add_srs_wkt (p, 10,
		 "nter\",49],PARAMETER[\"azimuth\",21],PARAMETER[\"rectifi");
    add_srs_wkt (p, 11,
		 "ed_grid_angle\",21],PARAMETER[\"scale_factor\",0.9995],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_easting\",400000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 13,
		 "thing\",800000],AUTHORITY[\"EPSG\",\"29702\"],AXIS[\"X\"");
    add_srs_wkt (p, 14, ",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 29738, "epsg", 29738,
		      "Tananarive / UTM zone 38S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=38 +south +ellps=intl +towgs84=-189,-242");
    add_proj4text (p, 1, ",-91,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tananarive / UTM zone 38S\",GEOGCS[\"Tananarive");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"Tananarive_1925\",SPHEROID[\"International 19");
    add_srs_wkt (p, 2,
		 "24\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-");
    add_srs_wkt (p, 3,
		 "189,-242,-91,0,0,0,0],AUTHORITY[\"EPSG\",\"6297\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4297\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",45],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 11,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"29738\"],AXIS[\"Ea");
    add_srs_wkt (p, 12, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 29739, "epsg", 29739,
		      "Tananarive / UTM zone 39S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=39 +south +ellps=intl +towgs84=-189,-242");
    add_proj4text (p, 1, ",-91,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tananarive / UTM zone 39S\",GEOGCS[\"Tananarive");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"Tananarive_1925\",SPHEROID[\"International 19");
    add_srs_wkt (p, 2,
		 "24\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-");
    add_srs_wkt (p, 3,
		 "189,-242,-91,0,0,0,0],AUTHORITY[\"EPSG\",\"6297\"]],PRIM");
    add_srs_wkt (p, 4,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4297\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",51],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 11,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"29739\"],AXIS[\"Ea");
    add_srs_wkt (p, 12, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 29849, "epsg", 29849,
		      "Timbalai 1948 / UTM zone 49N");
    add_proj4text (p, 0, "+proj=utm +zone=49 +ellps=evrstSS +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Timbalai 1948 / UTM zone 49N\",GEOGCS[\"Timbala");
    add_srs_wkt (p, 1,
		 "i 1948\",DATUM[\"Timbalai_1948\",SPHEROID[\"Everest 1830");
    add_srs_wkt (p, 2,
		 " (1967 Definition)\",6377298.556,300.8017,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7016\"]],AUTHORITY[\"EPSG\",\"6298\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4298\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",111],PARAMETER[\"scale_factor\",0.9996],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"29849\"],AXIS[\"Easting\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 29850, "epsg", 29850,
		      "Timbalai 1948 / UTM zone 50N");
    add_proj4text (p, 0, "+proj=utm +zone=50 +ellps=evrstSS +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Timbalai 1948 / UTM zone 50N\",GEOGCS[\"Timbala");
    add_srs_wkt (p, 1,
		 "i 1948\",DATUM[\"Timbalai_1948\",SPHEROID[\"Everest 1830");
    add_srs_wkt (p, 2,
		 " (1967 Definition)\",6377298.556,300.8017,AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"7016\"]],AUTHORITY[\"EPSG\",\"6298\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4298\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",117],PARAMETER[\"scale_factor\",0.9996],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"29850\"],AXIS[\"Easting\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 29871, "epsg", 29871,
		      "Timbalai 1948 / RSO Borneo (ch)");
    add_proj4text (p, 0,
		   "+proj=omerc +lat_0=4 +lonc=115 +alpha=53.31582047222222 ");
    add_proj4text (p, 1,
		   "+k=0.99984 +x_0=590476.8714630401 +y_0=442857.653094361 ");
    add_proj4text (p, 2, "+ellps=evrstSS +to_meter=20.11676512155263 +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Timbalai 1948 / RSO Borneo (ch)\",GEOGCS[\"Timb");
    add_srs_wkt (p, 1,
		 "alai 1948\",DATUM[\"Timbalai_1948\",SPHEROID[\"Everest 1");
    add_srs_wkt (p, 2,
		 "830 (1967 Definition)\",6377298.556,300.8017,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7016\"]],AUTHORITY[\"EPSG\",\"6298\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4298\"]],UNIT[\"British chain (Sears ");
    add_srs_wkt (p, 7,
		 "1922)\",20.11676512155263,AUTHORITY[\"EPSG\",\"9042\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Hotine_Oblique_Mercator\"],PARAMETER[\"lati");
    add_srs_wkt (p, 9,
		 "tude_of_center\",4],PARAMETER[\"longitude_of_center\",11");
    add_srs_wkt (p, 10,
		 "5],PARAMETER[\"azimuth\",53.31582047222222],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "rectified_grid_angle\",53.13010236111111],PARAMETER[\"sc");
    add_srs_wkt (p, 12,
		 "ale_factor\",0.99984],PARAMETER[\"false_easting\",29352.");
    add_srs_wkt (p, 13,
		 "4763],PARAMETER[\"false_northing\",22014.3572],AUTHORITY");
    add_srs_wkt (p, 14,
		 "[\"EPSG\",\"29871\"],AXIS[\"Easting\",EAST],AXIS[\"North");
    add_srs_wkt (p, 15, "ing\",NORTH]]");
    p = add_epsg_def (first, last, 29872, "epsg", 29872,
		      "Timbalai 1948 / RSO Borneo (ft)");
    add_proj4text (p, 0,
		   "+proj=omerc +lat_0=4 +lonc=115 +alpha=53.31582047222222 ");
    add_proj4text (p, 1,
		   "+k=0.99984 +x_0=590476.8727431979 +y_0=442857.6545573985");
    add_proj4text (p, 2,
		   " +ellps=evrstSS +to_meter=0.3047994715386762 +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Timbalai 1948 / RSO Borneo (ft)\",GEOGCS[\"Timb");
    add_srs_wkt (p, 1,
		 "alai 1948\",DATUM[\"Timbalai_1948\",SPHEROID[\"Everest 1");
    add_srs_wkt (p, 2,
		 "830 (1967 Definition)\",6377298.556,300.8017,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7016\"]],AUTHORITY[\"EPSG\",\"6298\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4298\"]],UNIT[\"British foot (Sears 1");
    add_srs_wkt (p, 7,
		 "922)\",0.3047994715386762,AUTHORITY[\"EPSG\",\"9041\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Hotine_Oblique_Mercator\"],PARAMETER[\"lati");
    add_srs_wkt (p, 9,
		 "tude_of_center\",4],PARAMETER[\"longitude_of_center\",11");
    add_srs_wkt (p, 10,
		 "5],PARAMETER[\"azimuth\",53.31582047222222],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "rectified_grid_angle\",53.13010236111111],PARAMETER[\"sc");
    add_srs_wkt (p, 12,
		 "ale_factor\",0.99984],PARAMETER[\"false_easting\",193726");
    add_srs_wkt (p, 13,
		 "3.44],PARAMETER[\"false_northing\",1452947.58],AUTHORITY");
    add_srs_wkt (p, 14,
		 "[\"EPSG\",\"29872\"],AXIS[\"Easting\",EAST],AXIS[\"North");
    add_srs_wkt (p, 15, "ing\",NORTH]]");
    p = add_epsg_def (first, last, 29873, "epsg", 29873,
		      "Timbalai 1948 / RSO Borneo (m)");
    add_proj4text (p, 0,
		   "+proj=omerc +lat_0=4 +lonc=115 +alpha=53.31582047222222 ");
    add_proj4text (p, 1,
		   "+k=0.99984 +x_0=590476.87 +y_0=442857.65 +ellps=evrstSS ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Timbalai 1948 / RSO Borneo (m)\",GEOGCS[\"Timba");
    add_srs_wkt (p, 1,
		 "lai 1948\",DATUM[\"Timbalai_1948\",SPHEROID[\"Everest 18");
    add_srs_wkt (p, 2,
		 "30 (1967 Definition)\",6377298.556,300.8017,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7016\"]],AUTHORITY[\"EPSG\",\"6298\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4298\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Hotine_Oblique_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_center\",4],PARAMETER[\"longit");
    add_srs_wkt (p, 9,
		 "ude_of_center\",115],PARAMETER[\"azimuth\",53.3158204722");
    add_srs_wkt (p, 10,
		 "2222],PARAMETER[\"rectified_grid_angle\",53.130102361111");
    add_srs_wkt (p, 11,
		 "11],PARAMETER[\"scale_factor\",0.99984],PARAMETER[\"fals");
    add_srs_wkt (p, 12,
		 "e_easting\",590476.87],PARAMETER[\"false_northing\",4428");
    add_srs_wkt (p, 13,
		 "57.65],AUTHORITY[\"EPSG\",\"29873\"],AXIS[\"Easting\",EA");
    add_srs_wkt (p, 14, "ST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 29900, "epsg", 29900,
		      "TM65 / Irish National Grid (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=53.5 +lon_0=-8 +k=1.000035 +x_0=20000");
    add_proj4text (p, 1,
		   "0 +y_0=250000 +ellps=mod_airy +datum=ire65 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"TM65 / Irish National Grid (deprecated)\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"TM65\",DATUM[\"TM65\",SPHEROID[\"Airy Modified 1849\"");
    add_srs_wkt (p, 2,
		 ",6377340.189,299.3249646,AUTHORITY[\"EPSG\",\"7002\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6299\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4299\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",53.5],PARAMETER[\"central_meridian\",-8],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"scale_factor\",1.000035],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",200000],PARAMETER[\"false_northing\",250000],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"29900\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 29901, "epsg", 29901,
		      "OSNI 1952 / Irish National Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=53.5 +lon_0=-8 +k=1 +x_0=200000 +y_0=");
    add_proj4text (p, 1,
		   "250000 +ellps=airy +towgs84=482.5,-130.6,564.6,-1.042,-0");
    add_proj4text (p, 2, ".214,-0.631,8.15 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"OSNI 1952 / Irish National Grid\",GEOGCS[\"OSNI");
    add_srs_wkt (p, 1,
		 " 1952\",DATUM[\"OSNI_1952\",SPHEROID[\"Airy 1830\",63775");
    add_srs_wkt (p, 2,
		 "63.396,299.3249646,AUTHORITY[\"EPSG\",\"7001\"]],TOWGS84");
    add_srs_wkt (p, 3,
		 "[482.5,-130.6,564.6,-1.042,-0.214,-0.631,8.15],AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"6188\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4188\"]");
    add_srs_wkt (p, 7,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 8,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 9,
		 "gin\",53.5],PARAMETER[\"central_meridian\",-8],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"scale_factor\",1],PARAMETER[\"false_easting\",200000]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_northing\",250000],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12,
		 ",\"29901\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NOR");
    add_srs_wkt (p, 13, "TH]]");
    p = add_epsg_def (first, last, 29902, "epsg", 29902, "TM65 / Irish Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=53.5 +lon_0=-8 +k=1.000035 +x_0=20000");
    add_proj4text (p, 1,
		   "0 +y_0=250000 +ellps=mod_airy +datum=ire65 +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"TM65 / Irish Grid\",GEOGCS[\"TM65\",DATUM[\"TM6");
    add_srs_wkt (p, 1,
		 "5\",SPHEROID[\"Airy Modified 1849\",6377340.189,299.3249");
    add_srs_wkt (p, 2,
		 "646,AUTHORITY[\"EPSG\",\"7002\"]],AUTHORITY[\"EPSG\",\"6");
    add_srs_wkt (p, 3,
		 "299\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901");
    add_srs_wkt (p, 4,
		 "\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"9122\"]],AUTHORITY[\"EPSG\",\"4299\"]],UNIT[\"metr");
    add_srs_wkt (p, 6,
		 "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transve");
    add_srs_wkt (p, 7,
		 "rse_Mercator\"],PARAMETER[\"latitude_of_origin\",53.5],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",-8],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 9,
		 "or\",1.000035],PARAMETER[\"false_easting\",200000],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_northing\",250000],AUTHORITY[\"EPSG\",\"299");
    add_srs_wkt (p, 11,
		 "02\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 29903, "epsg", 29903, "TM75 / Irish Grid");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=53.5 +lon_0=-8 +k=1.000035 +x_0=20000");
    add_proj4text (p, 1, "0 +y_0=250000 +ellps=mod_airy +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"TM75 / Irish Grid\",GEOGCS[\"TM75\",DATUM[\"Geo");
    add_srs_wkt (p, 1,
		 "detic_Datum_of_1965\",SPHEROID[\"Airy Modified 1849\",63");
    add_srs_wkt (p, 2,
		 "77340.189,299.3249646,AUTHORITY[\"EPSG\",\"7002\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6300\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"43");
    add_srs_wkt (p, 6,
		 "00\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 7,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 8,
		 "f_origin\",53.5],PARAMETER[\"central_meridian\",-8],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",1.000035],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",200000],PARAMETER[\"false_northing\",250000],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"29903\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 12, "rthing\",NORTH]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_24 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 30161, "epsg", 30161,
		      "Tokyo / Japan Plane Rectangular CS I");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=33 +lon_0=129.5 +k=0.9999 +x_0=0 +y_0");
    add_proj4text (p, 1, "=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / Japan Plane Rectangular CS I\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397");
    add_srs_wkt (p, 2,
		 ".155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4301\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",33],PARAMETER[\"central_meridian\",129.5],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",0");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3");
    add_srs_wkt (p, 11, "0161\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 30162, "epsg", 30162,
		      "Tokyo / Japan Plane Rectangular CS II");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=33 +lon_0=131 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / Japan Plane Rectangular CS II\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",63773");
    add_srs_wkt (p, 2,
		 "97.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4301\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 8,
		 "igin\",33],PARAMETER[\"central_meridian\",131],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",0]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"30");
    add_srs_wkt (p, 11, "162\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 30163, "epsg", 30163,
		      "Tokyo / Japan Plane Rectangular CS III");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36 +lon_0=132.1666666666667 +k=0.9999");
    add_proj4text (p, 1, " +x_0=0 +y_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / Japan Plane Rectangular CS III\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377");
    add_srs_wkt (p, 2,
		 "397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4301");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",36],PARAMETER[\"central_meridian\",132.16666666");
    add_srs_wkt (p, 9,
		 "66667],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_easting\",0],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 11,
		 "TY[\"EPSG\",\"30163\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST");
    add_srs_wkt (p, 12, "]]");
    p = add_epsg_def (first, last, 30164, "epsg", 30164,
		      "Tokyo / Japan Plane Rectangular CS IV");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=33 +lon_0=133.5 +k=0.9999 +x_0=0 +y_0");
    add_proj4text (p, 1, "=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / Japan Plane Rectangular CS IV\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",63773");
    add_srs_wkt (p, 2,
		 "97.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4301\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 8,
		 "igin\",33],PARAMETER[\"central_meridian\",133.5],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "30164\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 30165, "epsg", 30165,
		      "Tokyo / Japan Plane Rectangular CS V");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36 +lon_0=134.3333333333333 +k=0.9999");
    add_proj4text (p, 1, " +x_0=0 +y_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / Japan Plane Rectangular CS V\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397");
    add_srs_wkt (p, 2,
		 ".155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4301\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",36],PARAMETER[\"central_meridian\",134.33333333333");
    add_srs_wkt (p, 9,
		 "33],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"30165\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 30166, "epsg", 30166,
		      "Tokyo / Japan Plane Rectangular CS VI");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36 +lon_0=136 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / Japan Plane Rectangular CS VI\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",63773");
    add_srs_wkt (p, 2,
		 "97.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4301\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 8,
		 "igin\",36],PARAMETER[\"central_meridian\",136],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",0]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"30");
    add_srs_wkt (p, 11, "166\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 30167, "epsg", 30167,
		      "Tokyo / Japan Plane Rectangular CS VII");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36 +lon_0=137.1666666666667 +k=0.9999");
    add_proj4text (p, 1, " +x_0=0 +y_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / Japan Plane Rectangular CS VII\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377");
    add_srs_wkt (p, 2,
		 "397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4301");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",36],PARAMETER[\"central_meridian\",137.16666666");
    add_srs_wkt (p, 9,
		 "66667],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_easting\",0],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 11,
		 "TY[\"EPSG\",\"30167\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST");
    add_srs_wkt (p, 12, "]]");
    p = add_epsg_def (first, last, 30168, "epsg", 30168,
		      "Tokyo / Japan Plane Rectangular CS VIII");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36 +lon_0=138.5 +k=0.9999 +x_0=0 +y_0");
    add_proj4text (p, 1, "=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / Japan Plane Rectangular CS VIII\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",637");
    add_srs_wkt (p, 2,
		 "7397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"430");
    add_srs_wkt (p, 6,
		 "1\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",36],PARAMETER[\"central_meridian\",138.5],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9999],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"30168\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 30169, "epsg", 30169,
		      "Tokyo / Japan Plane Rectangular CS IX");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=36 +lon_0=139.8333333333333 +k=0.9999");
    add_proj4text (p, 1, " +x_0=0 +y_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / Japan Plane Rectangular CS IX\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",63773");
    add_srs_wkt (p, 2,
		 "97.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4301\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 8,
		 "igin\",36],PARAMETER[\"central_meridian\",139.8333333333");
    add_srs_wkt (p, 9,
		 "333],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_easting\",0],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"30169\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 30170, "epsg", 30170,
		      "Tokyo / Japan Plane Rectangular CS X");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40 +lon_0=140.8333333333333 +k=0.9999");
    add_proj4text (p, 1, " +x_0=0 +y_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / Japan Plane Rectangular CS X\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397");
    add_srs_wkt (p, 2,
		 ".155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4301\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",40],PARAMETER[\"central_meridian\",140.83333333333");
    add_srs_wkt (p, 9,
		 "33],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"30170\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 30171, "epsg", 30171,
		      "Tokyo / Japan Plane Rectangular CS XI");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=44 +lon_0=140.25 +k=0.9999 +x_0=0 +y_");
    add_proj4text (p, 1, "0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / Japan Plane Rectangular CS XI\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",63773");
    add_srs_wkt (p, 2,
		 "97.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4301\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 8,
		 "igin\",44],PARAMETER[\"central_meridian\",140.25],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"scale_factor\",0.9999],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "30171\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 30172, "epsg", 30172,
		      "Tokyo / Japan Plane Rectangular CS XII");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=44 +lon_0=142.25 +k=0.9999 +x_0=0 +y_");
    add_proj4text (p, 1, "0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / Japan Plane Rectangular CS XII\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377");
    add_srs_wkt (p, 2,
		 "397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4301");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",44],PARAMETER[\"central_meridian\",142.25],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9999],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"30172\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 30173, "epsg", 30173,
		      "Tokyo / Japan Plane Rectangular CS XIII");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=44 +lon_0=144.25 +k=0.9999 +x_0=0 +y_");
    add_proj4text (p, 1, "0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / Japan Plane Rectangular CS XIII\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",637");
    add_srs_wkt (p, 2,
		 "7397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"430");
    add_srs_wkt (p, 6,
		 "1\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",44],PARAMETER[\"central_meridian\",144.25],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"scale_factor\",0.9999],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 10,
		 "g\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 11, ",\"30173\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 30174, "epsg", 30174,
		      "Tokyo / Japan Plane Rectangular CS XIV");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=26 +lon_0=142 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / Japan Plane Rectangular CS XIV\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377");
    add_srs_wkt (p, 2,
		 "397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4301");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",26],PARAMETER[\"central_meridian\",142],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "30174\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 30175, "epsg", 30175,
		      "Tokyo / Japan Plane Rectangular CS XV");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=26 +lon_0=127.5 +k=0.9999 +x_0=0 +y_0");
    add_proj4text (p, 1, "=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / Japan Plane Rectangular CS XV\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",63773");
    add_srs_wkt (p, 2,
		 "97.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4301\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 8,
		 "igin\",26],PARAMETER[\"central_meridian\",127.5],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "30175\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 30176, "epsg", 30176,
		      "Tokyo / Japan Plane Rectangular CS XVI");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=26 +lon_0=124 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / Japan Plane Rectangular CS XVI\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377");
    add_srs_wkt (p, 2,
		 "397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4301");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",26],PARAMETER[\"central_meridian\",124],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "30176\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 30177, "epsg", 30177,
		      "Tokyo / Japan Plane Rectangular CS XVII");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=26 +lon_0=131 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / Japan Plane Rectangular CS XVII\",GEOGC");
    add_srs_wkt (p, 1,
		 "S[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",637");
    add_srs_wkt (p, 2,
		 "7397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"430");
    add_srs_wkt (p, 6,
		 "1\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",26],PARAMETER[\"central_meridian\",131],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"scale_factor\",0.9999],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "30177\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 30178, "epsg", 30178,
		      "Tokyo / Japan Plane Rectangular CS XVIII");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=20 +lon_0=136 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / Japan Plane Rectangular CS XVIII\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",63");
    add_srs_wkt (p, 2,
		 "77397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"43");
    add_srs_wkt (p, 6,
		 "01\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 7,
		 "OJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 8,
		 "f_origin\",20],PARAMETER[\"central_meridian\",136],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"scale_factor\",0.9999],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 10,
		 ",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "30178\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 30179, "epsg", 30179,
		      "Tokyo / Japan Plane Rectangular CS XIX");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=26 +lon_0=154 +k=0.9999 +x_0=0 +y_0=0");
    add_proj4text (p, 1, " +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Tokyo / Japan Plane Rectangular CS XIX\",GEOGCS");
    add_srs_wkt (p, 1,
		 "[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377");
    add_srs_wkt (p, 2,
		 "397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4301");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",26],PARAMETER[\"central_meridian\",154],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "30179\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 30200, "epsg", 30200,
		      "Trinidad 1903 / Trinidad Grid");
    add_proj4text (p, 0,
		   "+proj=cass +lat_0=10.44166666666667 +lon_0=-61.333333333");
    add_proj4text (p, 1,
		   "33334 +x_0=86501.46392051999 +y_0=65379.0134283 +a=63782");
    add_proj4text (p, 2,
		   "93.645208759 +b=6356617.987679838 +to_meter=0.2011661951");
    add_proj4text (p, 3, "64 +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Trinidad 1903 / Trinidad Grid\",GEOGCS[\"Trinid");
    add_srs_wkt (p, 1,
		 "ad 1903\",DATUM[\"Trinidad_1903\",SPHEROID[\"Clarke 1858");
    add_srs_wkt (p, 2,
		 "\",6378293.645208759,294.2606763692569,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7007\"]],AUTHORITY[\"EPSG\",\"6302\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4302\"]],UNIT[\"Clarke's link\",0.20116619");
    add_srs_wkt (p, 7,
		 "5164,AUTHORITY[\"EPSG\",\"9039\"]],PROJECTION[\"Cassini_");
    add_srs_wkt (p, 8,
		 "Soldner\"],PARAMETER[\"latitude_of_origin\",10.441666666");
    add_srs_wkt (p, 9,
		 "66667],PARAMETER[\"central_meridian\",-61.33333333333334");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"false_easting\",430000],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "northing\",325000],AUTHORITY[\"EPSG\",\"30200\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 30339, "epsg", 30339,
		      "TC(1948) / UTM zone 39N");
    add_proj4text (p, 0, "+proj=utm +zone=39 +ellps=helmert +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"TC(1948) / UTM zone 39N\",GEOGCS[\"TC(1948)\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Trucial_Coast_1948\",SPHEROID[\"Helmert 1906\",63");
    add_srs_wkt (p, 2,
		 "78200,298.3,AUTHORITY[\"EPSG\",\"7020\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6303\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4303\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",51],PARAMETER[\"scale_");
    add_srs_wkt (p, 9,
		 "factor\",0.9996],PARAMETER[\"false_easting\",500000],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"30339\"");
    add_srs_wkt (p, 11, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 30340, "epsg", 30340,
		      "TC(1948) / UTM zone 40N");
    add_proj4text (p, 0, "+proj=utm +zone=40 +ellps=helmert +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"TC(1948) / UTM zone 40N\",GEOGCS[\"TC(1948)\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Trucial_Coast_1948\",SPHEROID[\"Helmert 1906\",63");
    add_srs_wkt (p, 2,
		 "78200,298.3,AUTHORITY[\"EPSG\",\"7020\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6303\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4303\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",57],PARAMETER[\"scale_");
    add_srs_wkt (p, 9,
		 "factor\",0.9996],PARAMETER[\"false_easting\",500000],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"30340\"");
    add_srs_wkt (p, 11, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 30491, "epsg", 30491,
		      "Voirol 1875 / Nord Algerie (ancienne)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36 +lat_0=36 +lon_0=2.7 +k_0=0.99962554");
    add_proj4text (p, 1,
		   "4 +x_0=500000 +y_0=300000 +a=6378249.2 +b=6356515 +towgs");
    add_proj4text (p, 2, "84=-73,-247,227,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Voirol 1875 / Nord Algerie (ancienne)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"Voirol 1875\",DATUM[\"Voirol_1875\",SPHEROID[\"Clarke ");
    add_srs_wkt (p, 2,
		 "1880 (IGN)\",6378249.2,293.4660212936265,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7011\"]],TOWGS84[-73,-247,227,0,0,0,0],AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"6304\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 5,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4304\"]],U");
    add_srs_wkt (p, 7,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 8,
		 "[\"Lambert_Conformal_Conic_1SP\"],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 9,
		 "f_origin\",36],PARAMETER[\"central_meridian\",2.7],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.999625544],PARAMETER[\"false_eas");
    add_srs_wkt (p, 11,
		 "ting\",500000],PARAMETER[\"false_northing\",300000],AUTH");
    add_srs_wkt (p, 12,
		 "ORITY[\"EPSG\",\"30491\"],AXIS[\"X\",EAST],AXIS[\"Y\",NO");
    add_srs_wkt (p, 13, "RTH]]");
    p = add_epsg_def (first, last, 30492, "epsg", 30492,
		      "Voirol 1875 / Sud Algerie (ancienne)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.3 +lat_0=33.3 +lon_0=2.7 +k_0=0.9996");
    add_proj4text (p, 1,
		   "25769 +x_0=500000 +y_0=300000 +a=6378249.2 +b=6356515 +t");
    add_proj4text (p, 2, "owgs84=-73,-247,227,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Voirol 1875 / Sud Algerie (ancienne)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Voirol 1875\",DATUM[\"Voirol_1875\",SPHEROID[\"Clarke 18");
    add_srs_wkt (p, 2,
		 "80 (IGN)\",6378249.2,293.4660212936265,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7011\"]],TOWGS84[-73,-247,227,0,0,0,0],AUTHORITY[\"EP");
    add_srs_wkt (p, 4,
		 "SG\",\"6304\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4304\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 8,
		 "Lambert_Conformal_Conic_1SP\"],PARAMETER[\"latitude_of_o");
    add_srs_wkt (p, 9,
		 "rigin\",33.3],PARAMETER[\"central_meridian\",2.7],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",0.999625769],PARAMETER[\"false_east");
    add_srs_wkt (p, 11,
		 "ing\",500000],PARAMETER[\"false_northing\",300000],AUTHO");
    add_srs_wkt (p, 12,
		 "RITY[\"EPSG\",\"30492\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 13, "TH]]");
    p = add_epsg_def (first, last, 30493, "epsg", 30493,
		      "Voirol 1879 / Nord Algerie (ancienne)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36 +lat_0=36 +lon_0=2.7 +k_0=0.99962554");
    add_proj4text (p, 1,
		   "4 +x_0=500000 +y_0=300000 +a=6378249.2 +b=6356515 +units");
    add_proj4text (p, 2, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Voirol 1879 / Nord Algerie (ancienne)\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"Voirol 1879\",DATUM[\"Voirol_1879\",SPHEROID[\"Clarke ");
    add_srs_wkt (p, 2,
		 "1880 (IGN)\",6378249.2,293.4660212936265,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7011\"]],AUTHORITY[\"EPSG\",\"6671\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4671\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1SP");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",36],PARAMETER[\"cen");
    add_srs_wkt (p, 9,
		 "tral_meridian\",2.7],PARAMETER[\"scale_factor\",0.999625");
    add_srs_wkt (p, 10,
		 "544],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",300000],AUTHORITY[\"EPSG\",\"30493\"],AXIS");
    add_srs_wkt (p, 12, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 30494, "epsg", 30494,
		      "Voirol 1879 / Sud Algerie (ancienne)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.3 +lat_0=33.3 +lon_0=2.7 +k_0=0.9996");
    add_proj4text (p, 1,
		   "25769 +x_0=500000 +y_0=300000 +a=6378249.2 +b=6356515 +u");
    add_proj4text (p, 2, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Voirol 1879 / Sud Algerie (ancienne)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Voirol 1879\",DATUM[\"Voirol_1879\",SPHEROID[\"Clarke 18");
    add_srs_wkt (p, 2,
		 "80 (IGN)\",6378249.2,293.4660212936265,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7011\"]],AUTHORITY[\"EPSG\",\"6671\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4671\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1SP\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",33.3],PARAMETER[\"cen");
    add_srs_wkt (p, 9,
		 "tral_meridian\",2.7],PARAMETER[\"scale_factor\",0.999625");
    add_srs_wkt (p, 10,
		 "769],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",300000],AUTHORITY[\"EPSG\",\"30494\"],AXIS");
    add_srs_wkt (p, 12, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 30729, "epsg", 30729,
		      "Nord Sahara 1959 / UTM zone 29N");
    add_proj4text (p, 0, "+proj=utm +zone=29 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Nord Sahara 1959 / UTM zone 29N\",GEOGCS[\"Nord");
    add_srs_wkt (p, 1,
		 " Sahara 1959\",DATUM[\"Nord_Sahara_1959\",SPHEROID[\"Cla");
    add_srs_wkt (p, 2,
		 "rke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"7012\"]],AUTHORITY[\"EPSG\",\"6307\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"4307\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETE");
    add_srs_wkt (p, 8,
		 "R[\"latitude_of_origin\",0],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 9,
		 "\",-9],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_easting\",500000],PARAMETER[\"false_northing\",0],AU");
    add_srs_wkt (p, 11,
		 "THORITY[\"EPSG\",\"30729\"],AXIS[\"Easting\",EAST],AXIS[");
    add_srs_wkt (p, 12, "\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 30730, "epsg", 30730,
		      "Nord Sahara 1959 / UTM zone 30N");
    add_proj4text (p, 0, "+proj=utm +zone=30 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Nord Sahara 1959 / UTM zone 30N\",GEOGCS[\"Nord");
    add_srs_wkt (p, 1,
		 " Sahara 1959\",DATUM[\"Nord_Sahara_1959\",SPHEROID[\"Cla");
    add_srs_wkt (p, 2,
		 "rke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"7012\"]],AUTHORITY[\"EPSG\",\"6307\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"4307\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETE");
    add_srs_wkt (p, 8,
		 "R[\"latitude_of_origin\",0],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 9,
		 "\",-3],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_easting\",500000],PARAMETER[\"false_northing\",0],AU");
    add_srs_wkt (p, 11,
		 "THORITY[\"EPSG\",\"30730\"],AXIS[\"Easting\",EAST],AXIS[");
    add_srs_wkt (p, 12, "\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 30731, "epsg", 30731,
		      "Nord Sahara 1959 / UTM zone 31N");
    add_proj4text (p, 0, "+proj=utm +zone=31 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Nord Sahara 1959 / UTM zone 31N\",GEOGCS[\"Nord");
    add_srs_wkt (p, 1,
		 " Sahara 1959\",DATUM[\"Nord_Sahara_1959\",SPHEROID[\"Cla");
    add_srs_wkt (p, 2,
		 "rke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"7012\"]],AUTHORITY[\"EPSG\",\"6307\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"4307\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETE");
    add_srs_wkt (p, 8,
		 "R[\"latitude_of_origin\",0],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 9,
		 "\",3],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_easting\",500000],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"30731\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 30732, "epsg", 30732,
		      "Nord Sahara 1959 / UTM zone 32N");
    add_proj4text (p, 0, "+proj=utm +zone=32 +ellps=clrk80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Nord Sahara 1959 / UTM zone 32N\",GEOGCS[\"Nord");
    add_srs_wkt (p, 1,
		 " Sahara 1959\",DATUM[\"Nord_Sahara_1959\",SPHEROID[\"Cla");
    add_srs_wkt (p, 2,
		 "rke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 3,
		 "\"7012\"]],AUTHORITY[\"EPSG\",\"6307\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"4307\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETE");
    add_srs_wkt (p, 8,
		 "R[\"latitude_of_origin\",0],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 9,
		 "\",9],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_easting\",500000],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"30732\"],AXIS[\"Easting\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Northing\",NORTH]]");
    p = add_epsg_def (first, last, 30791, "epsg", 30791,
		      "Nord Sahara 1959 / Voirol Unifie Nord");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36 +lat_0=36 +lon_0=2.7 +k_0=0.99962554");
    add_proj4text (p, 1,
		   "4 +x_0=500135 +y_0=300090 +ellps=clrk80 +units=m +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Nord Sahara 1959 / Voirol Unifie Nord\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"Nord Sahara 1959\",DATUM[\"Nord_Sahara_1959\",SPHEROID");
    add_srs_wkt (p, 2,
		 "[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6307\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4307\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1");
    add_srs_wkt (p, 8,
		 "SP\"],PARAMETER[\"latitude_of_origin\",36],PARAMETER[\"c");
    add_srs_wkt (p, 9,
		 "entral_meridian\",2.7],PARAMETER[\"scale_factor\",0.9996");
    add_srs_wkt (p, 10,
		 "25544],PARAMETER[\"false_easting\",500135],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",300090],AUTHORITY[\"EPSG\",\"30791\"],AX");
    add_srs_wkt (p, 12, "IS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 30792, "epsg", 30792,
		      "Nord Sahara 1959 / Voirol Unifie Sud");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.3 +lat_0=33.3 +lon_0=2.7 +k_0=0.9996");
    add_proj4text (p, 1,
		   "25769 +x_0=500135 +y_0=300090 +ellps=clrk80 +units=m +no");
    add_proj4text (p, 2, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Nord Sahara 1959 / Voirol Unifie Sud\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "Nord Sahara 1959\",DATUM[\"Nord_Sahara_1959\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7012\"]],AUTHORITY[\"EPSG\",\"6307\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 4,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4307\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 7,
		 "SG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_1SP");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",33.3],PARAMETER[\"c");
    add_srs_wkt (p, 9,
		 "entral_meridian\",2.7],PARAMETER[\"scale_factor\",0.9996");
    add_srs_wkt (p, 10,
		 "25769],PARAMETER[\"false_easting\",500135],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",300090],AUTHORITY[\"EPSG\",\"30792\"],AX");
    add_srs_wkt (p, 12, "IS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 30800, "epsg", 30800,
		      "RT38 2.5 gon W (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15.80827777777778 +k=1 +x_0=");
    add_proj4text (p, 1, "1500000 +y_0=0 +ellps=bessel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"RT38 2.5 gon W (deprecated)\",GEOGCS[\"RT38\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Stockholm_1938\",SPHEROID[\"Bessel 1841\",6377397");
    add_srs_wkt (p, 2,
		 ".155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY");
    add_srs_wkt (p, 3,
		 "[\"EPSG\",\"6308\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4308\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",0],PARAMETER[\"central_meridian\",15.8082777777777");
    add_srs_wkt (p, 9,
		 "8],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",1500000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 11,
		 "\"EPSG\",\"30800\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 31028, "epsg", 31028, "Yoff / UTM zone 28N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=28 +a=6378249.2 +b=6356515 +units=m +no_");
    add_proj4text (p, 1, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Yoff / UTM zone 28N\",GEOGCS[\"Yoff\",DATUM[\"Y");
    add_srs_wkt (p, 1,
		 "off\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.46602");
    add_srs_wkt (p, 2,
		 "12936265,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"6310\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 5,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4310\"]],UNIT[\"m");
    add_srs_wkt (p, 6,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 7,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 8,
		 "ARAMETER[\"central_meridian\",-15],PARAMETER[\"scale_fac");
    add_srs_wkt (p, 9,
		 "tor\",0.9996],PARAMETER[\"false_easting\",500000],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"31028\"],");
    add_srs_wkt (p, 11, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31121, "epsg", 31121,
		      "Zanderij / UTM zone 21N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +ellps=intl +towgs84=-265,120,-358,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Zanderij / UTM zone 21N\",GEOGCS[\"Zanderij\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Zanderij\",SPHEROID[\"International 1924\",637838");
    add_srs_wkt (p, 2,
		 "8,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-265,120,-35");
    add_srs_wkt (p, 3,
		 "8,0,0,0,0],AUTHORITY[\"EPSG\",\"6311\"]],PRIMEM[\"Greenw");
    add_srs_wkt (p, 4,
		 "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.");
    add_srs_wkt (p, 5,
		 "01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"4311\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETE");
    add_srs_wkt (p, 8,
		 "R[\"latitude_of_origin\",0],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 9,
		 "\",-57],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",500000],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 11,
		 "UTHORITY[\"EPSG\",\"31121\"],AXIS[\"Easting\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31154, "epsg", 31154, "Zanderij / TM 54 NW");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-54 +k=0.9996 +x_0=500000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=intl +towgs84=-265,120,-358,0,0,0,0 +units=m");
    add_proj4text (p, 2, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Zanderij / TM 54 NW\",GEOGCS[\"Zanderij\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Zanderij\",SPHEROID[\"International 1924\",6378388,29");
    add_srs_wkt (p, 2,
		 "7,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-265,120,-358,0,");
    add_srs_wkt (p, 3,
		 "0,0,0],AUTHORITY[\"EPSG\",\"6311\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4311\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 7,
		 "001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "latitude_of_origin\",0],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 9,
		 "54],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",500000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 11,
		 "RITY[\"EPSG\",\"31154\"],AXIS[\"Easting\",EAST],AXIS[\"N");
    add_srs_wkt (p, 12, "orthing\",NORTH]]");
    p = add_epsg_def (first, last, 31170, "epsg", 31170,
		      "Zanderij / Suriname Old TM");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-55.68333333333333 +k=0.9996");
    add_proj4text (p, 1,
		   " +x_0=500000 +y_0=0 +ellps=intl +towgs84=-265,120,-358,0");
    add_proj4text (p, 2, ",0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Zanderij / Suriname Old TM\",GEOGCS[\"Zanderij\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"Zanderij\",SPHEROID[\"International 1924\",6378");
    add_srs_wkt (p, 2,
		 "388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-265,120,-");
    add_srs_wkt (p, 3,
		 "358,0,0,0,0],AUTHORITY[\"EPSG\",\"6311\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 4,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 5,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"4311\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAM");
    add_srs_wkt (p, 8,
		 "ETER[\"latitude_of_origin\",0],PARAMETER[\"central_merid");
    add_srs_wkt (p, 9,
		 "ian\",-55.68333333333333],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 10,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 11,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"31170\"],AXIS[\"Ea");
    add_srs_wkt (p, 12, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31171, "epsg", 31171,
		      "Zanderij / Suriname TM");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-55.68333333333333 +k=0.9999");
    add_proj4text (p, 1,
		   " +x_0=500000 +y_0=0 +ellps=intl +towgs84=-265,120,-358,0");
    add_proj4text (p, 2, ",0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Zanderij / Suriname TM\",GEOGCS[\"Zanderij\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"Zanderij\",SPHEROID[\"International 1924\",6378388");
    add_srs_wkt (p, 2,
		 ",297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-265,120,-358");
    add_srs_wkt (p, 3,
		 ",0,0,0,0],AUTHORITY[\"EPSG\",\"6311\"]],PRIMEM[\"Greenwi");
    add_srs_wkt (p, 4,
		 "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0");
    add_srs_wkt (p, 5,
		 "1745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"4311\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETE");
    add_srs_wkt (p, 8,
		 "R[\"latitude_of_origin\",0],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 9,
		 "\",-55.68333333333333],PARAMETER[\"scale_factor\",0.9999");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"false_easting\",500000],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "northing\",0],AUTHORITY[\"EPSG\",\"31171\"],AXIS[\"Easti");
    add_srs_wkt (p, 12, "ng\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31251, "epsg", 31251,
		      "MGI (Ferro) / Austria GK West Zone");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=28 +k=1 +x_0=0 +y_0=-5000000");
    add_proj4text (p, 1, " +ellps=bessel +pm=ferro +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI (Ferro) / Austria GK West Zone\",GEOGCS[\"M");
    add_srs_wkt (p, 1,
		 "GI (Ferro)\",DATUM[\"Militar_Geographische_Institut_Ferr");
    add_srs_wkt (p, 2,
		 "o\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6805\"]]");
    add_srs_wkt (p, 4,
		 ",PRIMEM[\"Ferro\",-17.66666666666667,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8909\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4805\"]],UNIT[");
    add_srs_wkt (p, 7,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",28],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",1],PARAMETER[\"false_easting\",0],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_northing\",-5000000],AUTHORITY[\"EPSG\",\"31251\"],A");
    add_srs_wkt (p, 12, "XIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 31252, "epsg", 31252,
		      "MGI (Ferro) / Austria GK Central Zone");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=31 +k=1 +x_0=0 +y_0=-5000000");
    add_proj4text (p, 1, " +ellps=bessel +pm=ferro +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI (Ferro) / Austria GK Central Zone\",GEOGCS[");
    add_srs_wkt (p, 1,
		 "\"MGI (Ferro)\",DATUM[\"Militar_Geographische_Institut_F");
    add_srs_wkt (p, 2,
		 "erro\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,");
    add_srs_wkt (p, 3,
		 "AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6805\"");
    add_srs_wkt (p, 4,
		 "]],PRIMEM[\"Ferro\",-17.66666666666667,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"8909\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 6,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4805\"]],UNIT");
    add_srs_wkt (p, 7,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 8,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 9,
		 "0],PARAMETER[\"central_meridian\",31],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",1],PARAMETER[\"false_easting\",0],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",-5000000],AUTHORITY[\"EPSG\",\"31252\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 31253, "epsg", 31253,
		      "MGI (Ferro) / Austria GK East Zone");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=34 +k=1 +x_0=0 +y_0=-5000000");
    add_proj4text (p, 1, " +ellps=bessel +pm=ferro +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI (Ferro) / Austria GK East Zone\",GEOGCS[\"M");
    add_srs_wkt (p, 1,
		 "GI (Ferro)\",DATUM[\"Militar_Geographische_Institut_Ferr");
    add_srs_wkt (p, 2,
		 "o\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6805\"]]");
    add_srs_wkt (p, 4,
		 ",PRIMEM[\"Ferro\",-17.66666666666667,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8909\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4805\"]],UNIT[");
    add_srs_wkt (p, 7,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",34],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",1],PARAMETER[\"false_easting\",0],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_northing\",-5000000],AUTHORITY[\"EPSG\",\"31253\"],A");
    add_srs_wkt (p, 12, "XIS[\"X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 31254, "epsg", 31254,
		      "MGI / Austria GK West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=10.33333333333333 +k=1 +x_0=");
    add_proj4text (p, 1,
		   "0 +y_0=-5000000 +ellps=bessel +datum=hermannskogel +unit");
    add_proj4text (p, 2, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / Austria GK West\",GEOGCS[\"MGI\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Militar_Geographische_Institute\",SPHEROID[\"Bessel 1841");
    add_srs_wkt (p, 2,
		 "\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]]");
    add_srs_wkt (p, 3,
		 ",TOWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "5199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",10.333");
    add_srs_wkt (p, 10,
		 "33333333333],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",0],PARAMETER[\"false_northing\",-5000000]");
    add_srs_wkt (p, 12,
		 ",AUTHORITY[\"EPSG\",\"31254\"],AXIS[\"X\",NORTH],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",EAST]]");
    p = add_epsg_def (first, last, 31255, "epsg", 31255,
		      "MGI / Austria GK Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=13.33333333333333 +k=1 +x_0=");
    add_proj4text (p, 1,
		   "0 +y_0=-5000000 +ellps=bessel +datum=hermannskogel +unit");
    add_proj4text (p, 2, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / Austria GK Central\",GEOGCS[\"MGI\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"Militar_Geographische_Institute\",SPHEROID[\"Bessel 1");
    add_srs_wkt (p, 2,
		 "841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"");
    add_srs_wkt (p, 3,
		 "]],TOWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.42");
    add_srs_wkt (p, 4,
		 "32],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"4312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 8,
		 "\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"lat");
    add_srs_wkt (p, 9,
		 "itude_of_origin\",0],PARAMETER[\"central_meridian\",13.3");
    add_srs_wkt (p, 10,
		 "3333333333333],PARAMETER[\"scale_factor\",1],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",0],PARAMETER[\"false_northing\",-5000000");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"31255\"],AXIS[\"X\",NORTH],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",EAST]]");
    p = add_epsg_def (first, last, 31256, "epsg", 31256,
		      "MGI / Austria GK East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=16.33333333333333 +k=1 +x_0=");
    add_proj4text (p, 1,
		   "0 +y_0=-5000000 +ellps=bessel +datum=hermannskogel +unit");
    add_proj4text (p, 2, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / Austria GK East\",GEOGCS[\"MGI\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Militar_Geographische_Institute\",SPHEROID[\"Bessel 1841");
    add_srs_wkt (p, 2,
		 "\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]]");
    add_srs_wkt (p, 3,
		 ",TOWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "5199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",16.333");
    add_srs_wkt (p, 10,
		 "33333333333],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",0],PARAMETER[\"false_northing\",-5000000]");
    add_srs_wkt (p, 12,
		 ",AUTHORITY[\"EPSG\",\"31256\"],AXIS[\"X\",NORTH],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",EAST]]");
    p = add_epsg_def (first, last, 31257, "epsg", 31257,
		      "MGI / Austria GK M28");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=10.33333333333333 +k=1 +x_0=");
    add_proj4text (p, 1,
		   "150000 +y_0=-5000000 +ellps=bessel +datum=hermannskogel ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / Austria GK M28\",GEOGCS[\"MGI\",DATUM[\"M");
    add_srs_wkt (p, 1,
		 "ilitar_Geographische_Institute\",SPHEROID[\"Bessel 1841\"");
    add_srs_wkt (p, 2,
		 ",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 6,
		 "99433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",10.333333");
    add_srs_wkt (p, 10,
		 "33333333],PARAMETER[\"scale_factor\",1],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",150000],PARAMETER[\"false_northing\",-500000");
    add_srs_wkt (p, 12,
		 "0],AUTHORITY[\"EPSG\",\"31257\"],AXIS[\"X\",NORTH],AXIS[");
    add_srs_wkt (p, 13, "\"Y\",EAST]]");
    p = add_epsg_def (first, last, 31258, "epsg", 31258,
		      "MGI / Austria GK M31");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=13.33333333333333 +k=1 +x_0=");
    add_proj4text (p, 1,
		   "450000 +y_0=-5000000 +ellps=bessel +datum=hermannskogel ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / Austria GK M31\",GEOGCS[\"MGI\",DATUM[\"M");
    add_srs_wkt (p, 1,
		 "ilitar_Geographische_Institute\",SPHEROID[\"Bessel 1841\"");
    add_srs_wkt (p, 2,
		 ",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 6,
		 "99433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",13.333333");
    add_srs_wkt (p, 10,
		 "33333333],PARAMETER[\"scale_factor\",1],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",450000],PARAMETER[\"false_northing\",-500000");
    add_srs_wkt (p, 12,
		 "0],AUTHORITY[\"EPSG\",\"31258\"],AXIS[\"X\",NORTH],AXIS[");
    add_srs_wkt (p, 13, "\"Y\",EAST]]");
    p = add_epsg_def (first, last, 31259, "epsg", 31259,
		      "MGI / Austria GK M34");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=16.33333333333333 +k=1 +x_0=");
    add_proj4text (p, 1,
		   "750000 +y_0=-5000000 +ellps=bessel +datum=hermannskogel ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / Austria GK M34\",GEOGCS[\"MGI\",DATUM[\"M");
    add_srs_wkt (p, 1,
		 "ilitar_Geographische_Institute\",SPHEROID[\"Bessel 1841\"");
    add_srs_wkt (p, 2,
		 ",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],T");
    add_srs_wkt (p, 3,
		 "OWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232],");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251");
    add_srs_wkt (p, 6,
		 "99433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 7,
		 "4312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 8,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 9,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",16.333333");
    add_srs_wkt (p, 10,
		 "33333333],PARAMETER[\"scale_factor\",1],PARAMETER[\"fals");
    add_srs_wkt (p, 11,
		 "e_easting\",750000],PARAMETER[\"false_northing\",-500000");
    add_srs_wkt (p, 12,
		 "0],AUTHORITY[\"EPSG\",\"31259\"],AXIS[\"X\",NORTH],AXIS[");
    add_srs_wkt (p, 13, "\"Y\",EAST]]");
    p = add_epsg_def (first, last, 31265, "epsg", 31265,
		      "MGI / 3-degree Gauss zone 5 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=1 +x_0=5500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=bessel +datum=hermannskogel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / 3-degree Gauss zone 5 (deprecated)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"MGI\",DATUM[\"Militar_Geographische_Institute\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7004\"]],TOWGS84[577.326,90.129,463.919,5.137");
    add_srs_wkt (p, 4,
		 ",1.474,5.297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIME");
    add_srs_wkt (p, 5,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 6,
		 "egree\",0.0174532925199433,AUTHORITY[\"EPSG\",\"9108\"]]");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"4312\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 8,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 10,
		 "l_meridian\",15],PARAMETER[\"scale_factor\",1],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_easting\",5500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"31265\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 13, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 31266, "epsg", 31266,
		      "MGI / 3-degree Gauss zone 6 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18 +k=1 +x_0=6500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=bessel +datum=hermannskogel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / 3-degree Gauss zone 6 (deprecated)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"MGI\",DATUM[\"Militar_Geographische_Institute\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7004\"]],TOWGS84[577.326,90.129,463.919,5.137");
    add_srs_wkt (p, 4,
		 ",1.474,5.297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIME");
    add_srs_wkt (p, 5,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 6,
		 "egree\",0.0174532925199433,AUTHORITY[\"EPSG\",\"9108\"]]");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"4312\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 8,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 10,
		 "l_meridian\",18],PARAMETER[\"scale_factor\",1],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_easting\",6500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"31266\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 13, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 31267, "epsg", 31267,
		      "MGI / 3-degree Gauss zone 7 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=1 +x_0=7500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=bessel +datum=hermannskogel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / 3-degree Gauss zone 7 (deprecated)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"MGI\",DATUM[\"Militar_Geographische_Institute\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7004\"]],TOWGS84[577.326,90.129,463.919,5.137");
    add_srs_wkt (p, 4,
		 ",1.474,5.297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIME");
    add_srs_wkt (p, 5,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 6,
		 "egree\",0.0174532925199433,AUTHORITY[\"EPSG\",\"9108\"]]");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"4312\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 8,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 10,
		 "l_meridian\",21],PARAMETER[\"scale_factor\",1],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_easting\",7500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"31267\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 13, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 31268, "epsg", 31268,
		      "MGI / 3-degree Gauss zone 8 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=24 +k=1 +x_0=8500000 +y_0=0 ");
    add_proj4text (p, 1,
		   "+ellps=bessel +datum=hermannskogel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / 3-degree Gauss zone 8 (deprecated)\",GEOG");
    add_srs_wkt (p, 1,
		 "CS[\"MGI\",DATUM[\"Militar_Geographische_Institute\",SPH");
    add_srs_wkt (p, 2,
		 "EROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"7004\"]],TOWGS84[577.326,90.129,463.919,5.137");
    add_srs_wkt (p, 4,
		 ",1.474,5.297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIME");
    add_srs_wkt (p, 5,
		 "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d");
    add_srs_wkt (p, 6,
		 "egree\",0.0174532925199433,AUTHORITY[\"EPSG\",\"9108\"]]");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"4312\"]],UNIT[\"metre\",1,AUTHORIT");
    add_srs_wkt (p, 8,
		 "Y[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"centra");
    add_srs_wkt (p, 10,
		 "l_meridian\",24],PARAMETER[\"scale_factor\",1],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_easting\",8500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"31268\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 13, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 31275, "epsg", 31275,
		      "MGI / Balkans zone 5 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=0.9999 +x_0=5500000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=bessel +datum=hermannskogel +units=m +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / Balkans zone 5 (deprecated)\",GEOGCS[\"MG");
    add_srs_wkt (p, 1,
		 "I\",DATUM[\"Militar_Geographische_Institute\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7004\"]],TOWGS84[577.326,90.129,463.919,5.137,1.474,5");
    add_srs_wkt (p, 4,
		 ".297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 5,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 6,
		 "0.0174532925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"4312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 8,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 10,
		 "an\",15],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",5500000],PARAMETER[\"false_northing\",0]");
    add_srs_wkt (p, 12,
		 ",AUTHORITY[\"EPSG\",\"31275\"],AXIS[\"X\",NORTH],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",EAST]]");
    p = add_epsg_def (first, last, 31276, "epsg", 31276,
		      "MGI / Balkans zone 6 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=18 +k=0.9999 +x_0=6500000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=bessel +datum=hermannskogel +units=m +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / Balkans zone 6 (deprecated)\",GEOGCS[\"MG");
    add_srs_wkt (p, 1,
		 "I\",DATUM[\"Militar_Geographische_Institute\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7004\"]],TOWGS84[577.326,90.129,463.919,5.137,1.474,5");
    add_srs_wkt (p, 4,
		 ".297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 5,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 6,
		 "0.0174532925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"4312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 8,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 10,
		 "an\",18],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",6500000],PARAMETER[\"false_northing\",0]");
    add_srs_wkt (p, 12,
		 ",AUTHORITY[\"EPSG\",\"31276\"],AXIS[\"X\",NORTH],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",EAST]]");
    p = add_epsg_def (first, last, 31277, "epsg", 31277,
		      "MGI / Balkans zone 7 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=0.9999 +x_0=7500000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=bessel +datum=hermannskogel +units=m +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / Balkans zone 7 (deprecated)\",GEOGCS[\"MG");
    add_srs_wkt (p, 1,
		 "I\",DATUM[\"Militar_Geographische_Institute\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7004\"]],TOWGS84[577.326,90.129,463.919,5.137,1.474,5");
    add_srs_wkt (p, 4,
		 ".297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 5,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 6,
		 "0.0174532925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"4312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 8,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 10,
		 "an\",21],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",7500000],PARAMETER[\"false_northing\",0]");
    add_srs_wkt (p, 12,
		 ",AUTHORITY[\"EPSG\",\"31277\"],AXIS[\"X\",NORTH],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",EAST]]");
    p = add_epsg_def (first, last, 31278, "epsg", 31278,
		      "MGI / Balkans zone 8 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=21 +k=0.9999 +x_0=7500000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=bessel +datum=hermannskogel +units=m +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / Balkans zone 8 (deprecated)\",GEOGCS[\"MG");
    add_srs_wkt (p, 1,
		 "I\",DATUM[\"Militar_Geographische_Institute\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7004\"]],TOWGS84[577.326,90.129,463.919,5.137,1.474,5");
    add_srs_wkt (p, 4,
		 ".297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 5,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 6,
		 "0.0174532925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"4312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 8,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 10,
		 "an\",21],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",7500000],PARAMETER[\"false_northing\",0]");
    add_srs_wkt (p, 12,
		 ",AUTHORITY[\"EPSG\",\"31278\"],AXIS[\"X\",NORTH],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",EAST]]");
    p = add_epsg_def (first, last, 31279, "epsg", 31279,
		      "MGI / Balkans zone 8 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=24 +k=0.9999 +x_0=8500000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=bessel +datum=hermannskogel +units=m +no_def");
    add_proj4text (p, 2, "s");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / Balkans zone 8 (deprecated)\",GEOGCS[\"MG");
    add_srs_wkt (p, 1,
		 "I\",DATUM[\"Militar_Geographische_Institute\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7004\"]],TOWGS84[577.326,90.129,463.919,5.137,1.474,5");
    add_srs_wkt (p, 4,
		 ".297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 5,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 6,
		 "0.0174532925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"4312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 8,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 9,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 10,
		 "an\",24],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_easting\",8500000],PARAMETER[\"false_northing\",0]");
    add_srs_wkt (p, 12,
		 ",AUTHORITY[\"EPSG\",\"31279\"],AXIS[\"X\",NORTH],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",EAST]]");
    p = add_epsg_def (first, last, 31281, "epsg", 31281,
		      "MGI (Ferro) / Austria West Zone");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=28 +k=1 +x_0=0 +y_0=0 +ellps");
    add_proj4text (p, 1, "=bessel +pm=ferro +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI (Ferro) / Austria West Zone\",GEOGCS[\"MGI ");
    add_srs_wkt (p, 1,
		 "(Ferro)\",DATUM[\"Militar_Geographische_Institut_Ferro\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6805\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Ferro\",-17.66666666666667,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "909\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4805\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"central_meridian\",28],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 10,
		 "or\",1],PARAMETER[\"false_easting\",0],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_northing\",0],AUTHORITY[\"EPSG\",\"31281\"],AXIS[\"X\",");
    add_srs_wkt (p, 12, "NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 31282, "epsg", 31282,
		      "MGI (Ferro) / Austria Central Zone");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=31 +k=1 +x_0=0 +y_0=0 +ellps");
    add_proj4text (p, 1, "=bessel +pm=ferro +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI (Ferro) / Austria Central Zone\",GEOGCS[\"M");
    add_srs_wkt (p, 1,
		 "GI (Ferro)\",DATUM[\"Militar_Geographische_Institut_Ferr");
    add_srs_wkt (p, 2,
		 "o\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6805\"]]");
    add_srs_wkt (p, 4,
		 ",PRIMEM[\"Ferro\",-17.66666666666667,AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 5,
		 "\"8909\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY");
    add_srs_wkt (p, 6,
		 "[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4805\"]],UNIT[");
    add_srs_wkt (p, 7,
		 "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",31],PARAMETER[\"scale_f");
    add_srs_wkt (p, 10,
		 "actor\",1],PARAMETER[\"false_easting\",0],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"31282\"],AXIS[\"X");
    add_srs_wkt (p, 12, "\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 31283, "epsg", 31283,
		      "MGI (Ferro) / Austria East Zone");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=34 +k=1 +x_0=0 +y_0=0 +ellps");
    add_proj4text (p, 1, "=bessel +pm=ferro +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI (Ferro) / Austria East Zone\",GEOGCS[\"MGI ");
    add_srs_wkt (p, 1,
		 "(Ferro)\",DATUM[\"Militar_Geographische_Institut_Ferro\"");
    add_srs_wkt (p, 2,
		 ",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6805\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Ferro\",-17.66666666666667,AUTHORITY[\"EPSG\",\"8");
    add_srs_wkt (p, 5,
		 "909\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4805\"]],UNIT[\"m");
    add_srs_wkt (p, 7,
		 "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Tran");
    add_srs_wkt (p, 8,
		 "sverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"central_meridian\",34],PARAMETER[\"scale_fact");
    add_srs_wkt (p, 10,
		 "or\",1],PARAMETER[\"false_easting\",0],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_northing\",0],AUTHORITY[\"EPSG\",\"31283\"],AXIS[\"X\",");
    add_srs_wkt (p, 12, "NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 31284, "epsg", 31284, "MGI / Austria M28");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=10.33333333333333 +k=1 +x_0=");
    add_proj4text (p, 1,
		   "150000 +y_0=0 +ellps=bessel +datum=hermannskogel +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / Austria M28\",GEOGCS[\"MGI\",DATUM[\"Mili");
    add_srs_wkt (p, 1,
		 "tar_Geographische_Institute\",SPHEROID[\"Bessel 1841\",6");
    add_srs_wkt (p, 2,
		 "377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOW");
    add_srs_wkt (p, 3,
		 "GS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",10.3333333");
    add_srs_wkt (p, 10,
		 "3333333],PARAMETER[\"scale_factor\",1],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_easting\",150000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 12,
		 "RITY[\"EPSG\",\"31284\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EA");
    add_srs_wkt (p, 13, "ST]]");
    p = add_epsg_def (first, last, 31285, "epsg", 31285, "MGI / Austria M31");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=13.33333333333333 +k=1 +x_0=");
    add_proj4text (p, 1,
		   "450000 +y_0=0 +ellps=bessel +datum=hermannskogel +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / Austria M31\",GEOGCS[\"MGI\",DATUM[\"Mili");
    add_srs_wkt (p, 1,
		 "tar_Geographische_Institute\",SPHEROID[\"Bessel 1841\",6");
    add_srs_wkt (p, 2,
		 "377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOW");
    add_srs_wkt (p, 3,
		 "GS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",13.3333333");
    add_srs_wkt (p, 10,
		 "3333333],PARAMETER[\"scale_factor\",1],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_easting\",450000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 12,
		 "RITY[\"EPSG\",\"31285\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EA");
    add_srs_wkt (p, 13, "ST]]");
    p = add_epsg_def (first, last, 31286, "epsg", 31286, "MGI / Austria M34");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=16.33333333333333 +k=1 +x_0=");
    add_proj4text (p, 1,
		   "750000 +y_0=0 +ellps=bessel +datum=hermannskogel +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / Austria M34\",GEOGCS[\"MGI\",DATUM[\"Mili");
    add_srs_wkt (p, 1,
		 "tar_Geographische_Institute\",SPHEROID[\"Bessel 1841\",6");
    add_srs_wkt (p, 2,
		 "377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOW");
    add_srs_wkt (p, 3,
		 "GS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232],AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 6,
		 "433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",16.3333333");
    add_srs_wkt (p, 10,
		 "3333333],PARAMETER[\"scale_factor\",1],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_easting\",750000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 12,
		 "RITY[\"EPSG\",\"31286\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EA");
    add_srs_wkt (p, 13, "ST]]");
    p = add_epsg_def (first, last, 31287, "epsg", 31287,
		      "MGI / Austria Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49 +lat_2=46 +lat_0=47.5 +lon_0=13.3333");
    add_proj4text (p, 1,
		   "3333333333 +x_0=400000 +y_0=400000 +ellps=bessel +datum=");
    add_proj4text (p, 2, "hermannskogel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / Austria Lambert\",GEOGCS[\"MGI\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Militar_Geographische_Institute\",SPHEROID[\"Bessel 1841");
    add_srs_wkt (p, 2,
		 "\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]]");
    add_srs_wkt (p, 3,
		 ",TOWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "5199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 9,
		 "[\"standard_parallel_1\",49],PARAMETER[\"standard_parall");
    add_srs_wkt (p, 10,
		 "el_2\",46],PARAMETER[\"latitude_of_origin\",47.5],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"central_meridian\",13.33333333333333],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_easting\",400000],PARAMETER[\"false_northing\",400");
    add_srs_wkt (p, 13,
		 "000],AUTHORITY[\"EPSG\",\"31287\"],AXIS[\"X\",NORTH],AXI");
    add_srs_wkt (p, 14, "S[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 31288, "epsg", 31288, "MGI (Ferro) / M28");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=28 +k=1 +x_0=150000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=bessel +pm=ferro +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI (Ferro) / M28\",GEOGCS[\"MGI (Ferro)\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"Militar_Geographische_Institut_Ferro\",SPHEROID[\"Be");
    add_srs_wkt (p, 2,
		 "ssel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7004\"]],AUTHORITY[\"EPSG\",\"6805\"]],PRIMEM[\"Ferro\",");
    add_srs_wkt (p, 4,
		 "-17.66666666666667,AUTHORITY[\"EPSG\",\"8909\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4805\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",28],PARAMETER[\"scale_factor\",1],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",150000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"31288\"],AXIS[\"X\",NORTH],AXIS");
    add_srs_wkt (p, 12, "[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 31289, "epsg", 31289, "MGI (Ferro) / M31");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=31 +k=1 +x_0=450000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=bessel +pm=ferro +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI (Ferro) / M31\",GEOGCS[\"MGI (Ferro)\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"Militar_Geographische_Institut_Ferro\",SPHEROID[\"Be");
    add_srs_wkt (p, 2,
		 "ssel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7004\"]],AUTHORITY[\"EPSG\",\"6805\"]],PRIMEM[\"Ferro\",");
    add_srs_wkt (p, 4,
		 "-17.66666666666667,AUTHORITY[\"EPSG\",\"8909\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4805\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",31],PARAMETER[\"scale_factor\",1],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",450000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"31289\"],AXIS[\"X\",NORTH],AXIS");
    add_srs_wkt (p, 12, "[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 31290, "epsg", 31290, "MGI (Ferro) / M34");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=34 +k=1 +x_0=750000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=bessel +pm=ferro +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI (Ferro) / M34\",GEOGCS[\"MGI (Ferro)\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"Militar_Geographische_Institut_Ferro\",SPHEROID[\"Be");
    add_srs_wkt (p, 2,
		 "ssel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7004\"]],AUTHORITY[\"EPSG\",\"6805\"]],PRIMEM[\"Ferro\",");
    add_srs_wkt (p, 4,
		 "-17.66666666666667,AUTHORITY[\"EPSG\",\"8909\"]],UNIT[\"");
    add_srs_wkt (p, 5,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 6,
		 "]],AUTHORITY[\"EPSG\",\"4805\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 8,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 9,
		 "ral_meridian\",34],PARAMETER[\"scale_factor\",1],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"false_easting\",750000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"31290\"],AXIS[\"X\",NORTH],AXIS");
    add_srs_wkt (p, 12, "[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 31291, "epsg", 31291,
		      "MGI (Ferro) / Austria West Zone (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=28 +k=1 +x_0=0 +y_0=0 +ellps");
    add_proj4text (p, 1, "=bessel +pm=ferro +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI (Ferro) / Austria West Zone (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"MGI (Ferro)\",DATUM[\"Militar_Geographische_Ins");
    add_srs_wkt (p, 2,
		 "titut_Ferro\",SPHEROID[\"Bessel 1841\",6377397.155,299.1");
    add_srs_wkt (p, 3,
		 "528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6805\"]],PRIMEM[\"Ferro\",-17.66666666666667,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"8909\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4805");
    add_srs_wkt (p, 7,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 8,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 9,
		 "origin\",0],PARAMETER[\"central_meridian\",28],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"scale_factor\",1],PARAMETER[\"false_easting\",0],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"31291\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 31292, "epsg", 31292,
		      "MGI (Ferro) / Austria Central Zone (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=31 +k=1 +x_0=0 +y_0=0 +ellps");
    add_proj4text (p, 1, "=bessel +pm=ferro +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI (Ferro) / Austria Central Zone (deprecated)");
    add_srs_wkt (p, 1,
		 "\",GEOGCS[\"MGI (Ferro)\",DATUM[\"Militar_Geographische_");
    add_srs_wkt (p, 2,
		 "Institut_Ferro\",SPHEROID[\"Bessel 1841\",6377397.155,29");
    add_srs_wkt (p, 3,
		 "9.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"6805\"]],PRIMEM[\"Ferro\",-17.66666666666667,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"8909\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 6,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 7,
		 "805\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 8,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 9,
		 "of_origin\",0],PARAMETER[\"central_meridian\",31],PARAME");
    add_srs_wkt (p, 10,
		 "TER[\"scale_factor\",1],PARAMETER[\"false_easting\",0],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3129");
    add_srs_wkt (p, 12, "2\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 31293, "epsg", 31293,
		      "MGI (Ferro) / Austria East Zone (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=34 +k=1 +x_0=0 +y_0=0 +ellps");
    add_proj4text (p, 1, "=bessel +pm=ferro +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI (Ferro) / Austria East Zone (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"MGI (Ferro)\",DATUM[\"Militar_Geographische_Ins");
    add_srs_wkt (p, 2,
		 "titut_Ferro\",SPHEROID[\"Bessel 1841\",6377397.155,299.1");
    add_srs_wkt (p, 3,
		 "528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 4,
		 "\"6805\"]],PRIMEM[\"Ferro\",-17.66666666666667,AUTHORITY");
    add_srs_wkt (p, 5,
		 "[\"EPSG\",\"8909\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 6,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4805");
    add_srs_wkt (p, 7,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 8,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 9,
		 "origin\",0],PARAMETER[\"central_meridian\",34],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"scale_factor\",1],PARAMETER[\"false_easting\",0],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"31293\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 31294, "epsg", 31294,
		      "MGI / M28 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=10.33333333333333 +k=1 +x_0=");
    add_proj4text (p, 1,
		   "150000 +y_0=0 +ellps=bessel +datum=hermannskogel +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / M28 (deprecated)\",GEOGCS[\"MGI\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Militar_Geographische_Institute\",SPHEROID[\"Bessel 1841");
    add_srs_wkt (p, 2,
		 "\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]]");
    add_srs_wkt (p, 3,
		 ",TOWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "5199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",10.333");
    add_srs_wkt (p, 10,
		 "33333333333],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",150000],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 12,
		 "UTHORITY[\"EPSG\",\"31294\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 31295, "epsg", 31295,
		      "MGI / M31 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=13.33333333333333 +k=1 +x_0=");
    add_proj4text (p, 1,
		   "450000 +y_0=0 +ellps=bessel +datum=hermannskogel +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / M31 (deprecated)\",GEOGCS[\"MGI\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Militar_Geographische_Institute\",SPHEROID[\"Bessel 1841");
    add_srs_wkt (p, 2,
		 "\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]]");
    add_srs_wkt (p, 3,
		 ",TOWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "5199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",13.333");
    add_srs_wkt (p, 10,
		 "33333333333],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",450000],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 12,
		 "UTHORITY[\"EPSG\",\"31295\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 31296, "epsg", 31296,
		      "MGI / M34 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=16.33333333333333 +k=1 +x_0=");
    add_proj4text (p, 1,
		   "750000 +y_0=0 +ellps=bessel +datum=hermannskogel +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / M34 (deprecated)\",GEOGCS[\"MGI\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "Militar_Geographische_Institute\",SPHEROID[\"Bessel 1841");
    add_srs_wkt (p, 2,
		 "\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]]");
    add_srs_wkt (p, 3,
		 ",TOWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232");
    add_srs_wkt (p, 4,
		 "],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 5,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 6,
		 "5199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 7,
		 ",\"4312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 8,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 9,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",16.333");
    add_srs_wkt (p, 10,
		 "33333333333],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",750000],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 12,
		 "UTHORITY[\"EPSG\",\"31296\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 31297, "epsg", 31297,
		      "MGI / Austria Lambert (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49 +lat_2=46 +lat_0=47.5 +lon_0=13.3333");
    add_proj4text (p, 1,
		   "3333333333 +x_0=400000 +y_0=400000 +ellps=bessel +datum=");
    add_proj4text (p, 2, "hermannskogel +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"MGI / Austria Lambert (deprecated)\",GEOGCS[\"M");
    add_srs_wkt (p, 1,
		 "GI\",DATUM[\"Militar_Geographische_Institute\",SPHEROID[");
    add_srs_wkt (p, 2,
		 "\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 3,
		 "\",\"7004\"]],TOWGS84[577.326,90.129,463.919,5.137,1.474");
    add_srs_wkt (p, 4,
		 ",5.297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 5,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 6,
		 ",0.0174532925199433,AUTHORITY[\"EPSG\",\"9108\"]],AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"4312\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 8,
		 "G\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"standard_parallel_1\",49],PARAMETER[\"stan");
    add_srs_wkt (p, 10,
		 "dard_parallel_2\",46],PARAMETER[\"latitude_of_origin\",4");
    add_srs_wkt (p, 11,
		 "7.5],PARAMETER[\"central_meridian\",13.33333333333333],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_easting\",400000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 13,
		 "thing\",400000],AUTHORITY[\"EPSG\",\"31297\"],AXIS[\"X\"");
    add_srs_wkt (p, 14, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 31300, "epsg", 31300,
		      "Belge 1972 / Belge Lambert 72");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49.83333333333334 +lat_2=51.16666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=90 +lon_0=4.356939722222222 +x_0=150000.01256");
    add_proj4text (p, 2,
		   " +y_0=5400088.4378 +ellps=intl +towgs84=106.869,-52.2978");
    add_proj4text (p, 3,
		   ",103.724,-0.33657,0.456955,-1.84218,1 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Belge 1972 / Belge Lambert 72\",GEOGCS[\"Belge ");
    add_srs_wkt (p, 1,
		 "1972\",DATUM[\"Reseau_National_Belge_1972\",SPHEROID[\"I");
    add_srs_wkt (p, 2,
		 "nternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"702");
    add_srs_wkt (p, 3,
		 "2\"]],TOWGS84[106.869,-52.2978,103.724,-0.33657,0.456955");
    add_srs_wkt (p, 4,
		 ",-1.84218,1],AUTHORITY[\"EPSG\",\"6313\"]],PRIMEM[\"Gree");
    add_srs_wkt (p, 5,
		 "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",");
    add_srs_wkt (p, 6,
		 "0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHOR");
    add_srs_wkt (p, 7,
		 "ITY[\"EPSG\",\"4313\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS");
    add_srs_wkt (p, 8,
		 "G\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP_");
    add_srs_wkt (p, 9,
		 "Belgium\"],PARAMETER[\"standard_parallel_1\",49.83333333");
    add_srs_wkt (p, 10,
		 "333334],PARAMETER[\"standard_parallel_2\",51.16666666666");
    add_srs_wkt (p, 11,
		 "666],PARAMETER[\"latitude_of_origin\",90],PARAMETER[\"ce");
    add_srs_wkt (p, 12,
		 "ntral_meridian\",4.356939722222222],PARAMETER[\"false_ea");
    add_srs_wkt (p, 13,
		 "sting\",150000.01256],PARAMETER[\"false_northing\",54000");
    add_srs_wkt (p, 14,
		 "88.4378],AUTHORITY[\"EPSG\",\"31300\"],AXIS[\"X\",EAST],");
    add_srs_wkt (p, 15, "AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 31370, "epsg", 31370,
		      "Belge 1972 / Belgian Lambert 72");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=51.16666723333333 +lat_2=49.8333339 +la");
    add_proj4text (p, 1,
		   "t_0=90 +lon_0=4.367486666666666 +x_0=150000.013 +y_0=540");
    add_proj4text (p, 2,
		   "0088.438 +ellps=intl +towgs84=106.869,-52.2978,103.724,-");
    add_proj4text (p, 3, "0.33657,0.456955,-1.84218,1 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Belge 1972 / Belgian Lambert 72\",GEOGCS[\"Belg");
    add_srs_wkt (p, 1,
		 "e 1972\",DATUM[\"Reseau_National_Belge_1972\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "22\"]],TOWGS84[106.869,-52.2978,103.724,-0.33657,0.45695");
    add_srs_wkt (p, 4,
		 "5,-1.84218,1],AUTHORITY[\"EPSG\",\"6313\"]],PRIMEM[\"Gre");
    add_srs_wkt (p, 5,
		 "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 6,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 7,
		 "RITY[\"EPSG\",\"4313\"]],UNIT[\"metre\",1,AUTHORITY[\"EP");
    add_srs_wkt (p, 8,
		 "SG\",\"9001\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP");
    add_srs_wkt (p, 9,
		 "\"],PARAMETER[\"standard_parallel_1\",51.16666723333333]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"standard_parallel_2\",49.8333339],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"latitude_of_origin\",90],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 12,
		 "\",4.367486666666666],PARAMETER[\"false_easting\",150000");
    add_srs_wkt (p, 13,
		 ".013],PARAMETER[\"false_northing\",5400088.438],AUTHORIT");
    add_srs_wkt (p, 14,
		 "Y[\"EPSG\",\"31370\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]");
    add_srs_wkt (p, 15, "]");
    p = add_epsg_def (first, last, 31461, "epsg", 31461,
		      "DHDN / 3-degree Gauss zone 1 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=3 +k=1 +x_0=1500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=bessel +datum=potsdam +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DHDN / 3-degree Gauss zone 1 (deprecated)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"DHDN\",DATUM[\"Deutsches_Hauptdreiecksnetz\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6314\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4314\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",3],PARAMETER[\"scale_factor\",1],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_easting\",1500000],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 11,
		 "UTHORITY[\"EPSG\",\"31461\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 31462, "epsg", 31462,
		      "DHDN / 3-degree Gauss zone 2 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=6 +k=1 +x_0=2500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=bessel +datum=potsdam +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DHDN / 3-degree Gauss zone 2 (deprecated)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"DHDN\",DATUM[\"Deutsches_Hauptdreiecksnetz\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6314\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4314\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",6],PARAMETER[\"scale_factor\",1],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_easting\",2500000],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 11,
		 "UTHORITY[\"EPSG\",\"31462\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 31463, "epsg", 31463,
		      "DHDN / 3-degree Gauss zone 3 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=9 +k=1 +x_0=3500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=bessel +datum=potsdam +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DHDN / 3-degree Gauss zone 3 (deprecated)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"DHDN\",DATUM[\"Deutsches_Hauptdreiecksnetz\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6314\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4314\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",9],PARAMETER[\"scale_factor\",1],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_easting\",3500000],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 11,
		 "UTHORITY[\"EPSG\",\"31463\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 31464, "epsg", 31464,
		      "DHDN / 3-degree Gauss zone 4 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=12 +k=1 +x_0=4500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=bessel +datum=potsdam +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DHDN / 3-degree Gauss zone 4 (deprecated)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"DHDN\",DATUM[\"Deutsches_Hauptdreiecksnetz\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6314\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4314\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",12],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",4500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"31464\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 31465, "epsg", 31465,
		      "DHDN / 3-degree Gauss zone 5 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=1 +x_0=5500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=bessel +datum=potsdam +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DHDN / 3-degree Gauss zone 5 (deprecated)\",GEO");
    add_srs_wkt (p, 1,
		 "GCS[\"DHDN\",DATUM[\"Deutsches_Hauptdreiecksnetz\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6314\"]],PRIMEM[\"");
    add_srs_wkt (p, 4,
		 "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre");
    add_srs_wkt (p, 5,
		 "e\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"4314\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",15],PARAMETER[\"scale_factor\",1],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",5500000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"31465\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 31466, "epsg", 31466,
		      "DHDN / 3-degree Gauss-Kruger zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=6 +k=1 +x_0=2500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=bessel +datum=potsdam +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DHDN / 3-degree Gauss-Kruger zone 2\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "DHDN\",DATUM[\"Deutsches_Hauptdreiecksnetz\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7004\"]],AUTHORITY[\"EPSG\",\"6314\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4314\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",6],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",2500000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 11,
		 "RITY[\"EPSG\",\"31466\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EA");
    add_srs_wkt (p, 12, "ST]]");
    p = add_epsg_def (first, last, 31467, "epsg", 31467,
		      "DHDN / 3-degree Gauss-Kruger zone 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=9 +k=1 +x_0=3500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=bessel +datum=potsdam +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DHDN / 3-degree Gauss-Kruger zone 3\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "DHDN\",DATUM[\"Deutsches_Hauptdreiecksnetz\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7004\"]],AUTHORITY[\"EPSG\",\"6314\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4314\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",9],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "easting\",3500000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 11,
		 "RITY[\"EPSG\",\"31467\"],AXIS[\"X\",NORTH],AXIS[\"Y\",EA");
    add_srs_wkt (p, 12, "ST]]");
    p = add_epsg_def (first, last, 31468, "epsg", 31468,
		      "DHDN / 3-degree Gauss-Kruger zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=12 +k=1 +x_0=4500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=bessel +datum=potsdam +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DHDN / 3-degree Gauss-Kruger zone 4\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "DHDN\",DATUM[\"Deutsches_Hauptdreiecksnetz\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7004\"]],AUTHORITY[\"EPSG\",\"6314\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4314\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",12],PARAMETER[\"scale_factor\",1],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",4500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"31468\"],AXIS[\"X\",NORTH],AXIS[\"Y\",E");
    add_srs_wkt (p, 12, "AST]]");
    p = add_epsg_def (first, last, 31469, "epsg", 31469,
		      "DHDN / 3-degree Gauss-Kruger zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=1 +x_0=5500000 +y_0=0 ");
    add_proj4text (p, 1, "+ellps=bessel +datum=potsdam +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"DHDN / 3-degree Gauss-Kruger zone 5\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "DHDN\",DATUM[\"Deutsches_Hauptdreiecksnetz\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7004\"]],AUTHORITY[\"EPSG\",\"6314\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4314\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PARAME");
    add_srs_wkt (p, 8,
		 "TER[\"latitude_of_origin\",0],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",15],PARAMETER[\"scale_factor\",1],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",5500000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 11,
		 "ORITY[\"EPSG\",\"31469\"],AXIS[\"X\",NORTH],AXIS[\"Y\",E");
    add_srs_wkt (p, 12, "AST]]");
    p = add_epsg_def (first, last, 31528, "epsg", 31528,
		      "Conakry 1905 / UTM zone 28N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=28 +a=6378249.2 +b=6356515 +towgs84=-23,");
    add_proj4text (p, 1, "259,-9,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Conakry 1905 / UTM zone 28N\",GEOGCS[\"Conakry ");
    add_srs_wkt (p, 1,
		 "1905\",DATUM[\"Conakry_1905\",SPHEROID[\"Clarke 1880 (IG");
    add_srs_wkt (p, 2,
		 "N)\",6378249.2,293.4660212936265,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "11\"]],TOWGS84[-23,259,-9,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "6315\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4315\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 8,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"central_meridian\",-15],PARAMETER[\"scale_facto");
    add_srs_wkt (p, 10,
		 "r\",0.9996],PARAMETER[\"false_easting\",500000],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_northing\",0],AUTHORITY[\"EPSG\",\"31528\"],AX");
    add_srs_wkt (p, 12, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31529, "epsg", 31529,
		      "Conakry 1905 / UTM zone 29N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=29 +a=6378249.2 +b=6356515 +towgs84=-23,");
    add_proj4text (p, 1, "259,-9,0,0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Conakry 1905 / UTM zone 29N\",GEOGCS[\"Conakry ");
    add_srs_wkt (p, 1,
		 "1905\",DATUM[\"Conakry_1905\",SPHEROID[\"Clarke 1880 (IG");
    add_srs_wkt (p, 2,
		 "N)\",6378249.2,293.4660212936265,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "11\"]],TOWGS84[-23,259,-9,0,0,0,0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 4,
		 "6315\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890");
    add_srs_wkt (p, 5,
		 "1\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4315\"]],UNIT[\"met");
    add_srs_wkt (p, 7,
		 "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transv");
    add_srs_wkt (p, 8,
		 "erse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"central_meridian\",-9],PARAMETER[\"scale_factor");
    add_srs_wkt (p, 10,
		 "\",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_northing\",0],AUTHORITY[\"EPSG\",\"31529\"],AXI");
    add_srs_wkt (p, 12, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31600, "epsg", 31600,
		      "Dealul Piscului 1930 / Stereo 33");
    add_proj4text (p, 0,
		   "+proj=sterea +lat_0=45.9 +lon_0=25.39246588888889 +k=0.9");
    add_proj4text (p, 1,
		   "996667 +x_0=500000 +y_0=500000 +ellps=intl +units=m +no_");
    add_proj4text (p, 2, "defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Dealul Piscului 1930 / Stereo 33\",GEOGCS[\"Dea");
    add_srs_wkt (p, 1,
		 "lul Piscului 1930\",DATUM[\"Dealul_Piscului_1930\",SPHER");
    add_srs_wkt (p, 2,
		 "OID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 3,
		 ",\"7022\"]],AUTHORITY[\"EPSG\",\"6316\"]],PRIMEM[\"Green");
    add_srs_wkt (p, 4,
		 "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0");
    add_srs_wkt (p, 5,
		 ".01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORI");
    add_srs_wkt (p, 6,
		 "TY[\"EPSG\",\"4316\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 7,
		 "\",\"9001\"]],PROJECTION[\"Oblique_Stereographic\"],PARA");
    add_srs_wkt (p, 8,
		 "METER[\"latitude_of_origin\",45.9],PARAMETER[\"central_m");
    add_srs_wkt (p, 9,
		 "eridian\",25.39246588888889],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996667],PARAMETER[\"false_easting\",500000],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_northing\",500000],AUTHORITY[\"EPSG\",\"31600\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 31700, "epsg", 31700,
		      "Dealul Piscului 1970/ Stereo 70 (deprecated)");
    add_proj4text (p, 0,
		   "+proj=sterea +lat_0=46 +lon_0=25 +k=0.99975 +x_0=500000 ");
    add_proj4text (p, 1, "+y_0=500000 +ellps=krass +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"Dealul Piscului 1970/ Stereo 70 (deprecated)\",");
    add_srs_wkt (p, 1,
		 "GEOGCS[\"Dealul Piscului 1970\",DATUM[\"Dealul_Piscului_");
    add_srs_wkt (p, 2,
		 "1970\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6317\"]],PR");
    add_srs_wkt (p, 4,
		 "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[");
    add_srs_wkt (p, 5,
		 "\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122");
    add_srs_wkt (p, 6,
		 "\"]],AUTHORITY[\"EPSG\",\"4317\"]],UNIT[\"metre\",1,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Oblique_Stereogra");
    add_srs_wkt (p, 8,
		 "phic\"],PARAMETER[\"latitude_of_origin\",46],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",25],PARAMETER[\"scale_factor\",0.9997");
    add_srs_wkt (p, 10,
		 "5],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 11,
		 "_northing\",500000],AUTHORITY[\"EPSG\",\"31700\"],AXIS[\"");
    add_srs_wkt (p, 12, "X\",NORTH],AXIS[\"Y\",EAST]]");
    p = add_epsg_def (first, last, 31838, "epsg", 31838, "NGN / UTM zone 38N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=38 +ellps=WGS84 +towgs84=-3.2,-5.7,2.8,0");
    add_proj4text (p, 1, ",0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NGN / UTM zone 38N\",GEOGCS[\"NGN\",DATUM[\"Nat");
    add_srs_wkt (p, 1,
		 "ional_Geodetic_Network\",SPHEROID[\"WGS 84\",6378137,298");
    add_srs_wkt (p, 2,
		 ".257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[-3.2,-5");
    add_srs_wkt (p, 3,
		 ".7,2.8,0,0,0,0],AUTHORITY[\"EPSG\",\"6318\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4318\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",45],PARAMETER[\"scale_factor\",0.9996],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"31838\"],AXIS[\"Easting\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31839, "epsg", 31839, "NGN / UTM zone 39N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=39 +ellps=WGS84 +towgs84=-3.2,-5.7,2.8,0");
    add_proj4text (p, 1, ",0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NGN / UTM zone 39N\",GEOGCS[\"NGN\",DATUM[\"Nat");
    add_srs_wkt (p, 1,
		 "ional_Geodetic_Network\",SPHEROID[\"WGS 84\",6378137,298");
    add_srs_wkt (p, 2,
		 ".257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[-3.2,-5");
    add_srs_wkt (p, 3,
		 ".7,2.8,0,0,0,0],AUTHORITY[\"EPSG\",\"6318\"]],PRIMEM[\"G");
    add_srs_wkt (p, 4,
		 "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree");
    add_srs_wkt (p, 5,
		 "\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUT");
    add_srs_wkt (p, 6,
		 "HORITY[\"EPSG\",\"4318\"]],UNIT[\"metre\",1,AUTHORITY[\"");
    add_srs_wkt (p, 7,
		 "EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator\"],PA");
    add_srs_wkt (p, 8,
		 "RAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_me");
    add_srs_wkt (p, 9,
		 "ridian\",51],PARAMETER[\"scale_factor\",0.9996],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_easting\",500000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"31839\"],AXIS[\"Easting\",EAST]");
    add_srs_wkt (p, 12, ",AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31900, "epsg", 31900,
		      "KUDAMS / KTM (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=48 +k=0.9996 +x_0=500000 +y_");
    add_proj4text (p, 1, "0=0 +ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"KUDAMS / KTM (deprecated)\",GEOGCS[\"KUDAMS\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"Kuwait_Utility\",SPHEROID[\"GRS 1980\",6378137,29");
    add_srs_wkt (p, 2,
		 "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6319\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4319\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "0],PARAMETER[\"central_meridian\",48],PARAMETER[\"scale_");
    add_srs_wkt (p, 9,
		 "factor\",0.9996],PARAMETER[\"false_easting\",500000],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"31900\"");
    add_srs_wkt (p, 11, "],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31901, "epsg", 31901, "KUDAMS / KTM");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=48 +k=1 +x_0=500000 +y_0=0 +");
    add_proj4text (p, 1, "ellps=GRS80 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"KUDAMS / KTM\",GEOGCS[\"KUDAMS\",DATUM[\"Kuwait");
    add_srs_wkt (p, 1,
		 "_Utility\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6319\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4319\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",48],PARAMETER[\"scale_factor\",1],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"31901\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31965, "epsg", 31965,
		      "SIRGAS 2000 / UTM zone 11N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=11 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 11N\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-11");
    add_srs_wkt (p, 10,
		 "7],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 12,
		 "ITY[\"EPSG\",\"31965\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 13, "rthing\",NORTH]]");
    p = add_epsg_def (first, last, 31966, "epsg", 31966,
		      "SIRGAS 2000 / UTM zone 12N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=12 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 12N\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-11");
    add_srs_wkt (p, 10,
		 "1],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 12,
		 "ITY[\"EPSG\",\"31966\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 13, "rthing\",NORTH]]");
    p = add_epsg_def (first, last, 31967, "epsg", 31967,
		      "SIRGAS 2000 / UTM zone 13N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=13 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 13N\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-10");
    add_srs_wkt (p, 10,
		 "5],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",500000],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 12,
		 "ITY[\"EPSG\",\"31967\"],AXIS[\"Easting\",EAST],AXIS[\"No");
    add_srs_wkt (p, 13, "rthing\",NORTH]]");
    p = add_epsg_def (first, last, 31968, "epsg", 31968,
		      "SIRGAS 2000 / UTM zone 14N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=14 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 14N\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-99");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"31968\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 31969, "epsg", 31969,
		      "SIRGAS 2000 / UTM zone 15N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=15 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 15N\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-93");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"31969\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 31970, "epsg", 31970,
		      "SIRGAS 2000 / UTM zone 16N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=16 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 16N\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-87");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"31970\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 31971, "epsg", 31971,
		      "SIRGAS 2000 / UTM zone 17N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=17 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 17N\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-81");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"31971\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 31972, "epsg", 31972,
		      "SIRGAS 2000 / UTM zone 18N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 18N\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-75");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"31972\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 31973, "epsg", 31973,
		      "SIRGAS 2000 / UTM zone 19N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 19N\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-69");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"31973\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 31974, "epsg", 31974,
		      "SIRGAS 2000 / UTM zone 20N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 20N\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-63");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"31974\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 31975, "epsg", 31975,
		      "SIRGAS 2000 / UTM zone 21N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 21N\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-57");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"31975\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 31976, "epsg", 31976,
		      "SIRGAS 2000 / UTM zone 22N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=22 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 22N\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-51");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"31976\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 31977, "epsg", 31977,
		      "SIRGAS 2000 / UTM zone 17S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=17 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 17S\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-81");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"31977\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31978, "epsg", 31978,
		      "SIRGAS 2000 / UTM zone 18S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 18S\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-75");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"31978\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31979, "epsg", 31979,
		      "SIRGAS 2000 / UTM zone 19S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 19S\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-69");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"31979\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31980, "epsg", 31980,
		      "SIRGAS 2000 / UTM zone 20S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 20S\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-63");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"31980\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31981, "epsg", 31981,
		      "SIRGAS 2000 / UTM zone 21S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 21S\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-57");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"31981\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31982, "epsg", 31982,
		      "SIRGAS 2000 / UTM zone 22S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=22 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 22S\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-51");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"31982\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31983, "epsg", 31983,
		      "SIRGAS 2000 / UTM zone 23S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=23 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 23S\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-45");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"31983\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31984, "epsg", 31984,
		      "SIRGAS 2000 / UTM zone 24S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=24 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 24S\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-39");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"31984\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31985, "epsg", 31985,
		      "SIRGAS 2000 / UTM zone 25S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=25 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 2000 / UTM zone 25S\",GEOGCS[\"SIRGAS 20");
    add_srs_wkt (p, 1,
		 "00\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_2000\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4674\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-33");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"31985\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31986, "epsg", 31986,
		      "SIRGAS 1995 / UTM zone 17N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=17 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 1995 / UTM zone 17N\",GEOGCS[\"SIRGAS 19");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_1995\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4170\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-81");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"31986\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 31987, "epsg", 31987,
		      "SIRGAS 1995 / UTM zone 18N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 1995 / UTM zone 18N\",GEOGCS[\"SIRGAS 19");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_1995\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4170\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-75");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"31987\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 31988, "epsg", 31988,
		      "SIRGAS 1995 / UTM zone 19N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 1995 / UTM zone 19N\",GEOGCS[\"SIRGAS 19");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_1995\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4170\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-69");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"31988\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 31989, "epsg", 31989,
		      "SIRGAS 1995 / UTM zone 20N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 1995 / UTM zone 20N\",GEOGCS[\"SIRGAS 19");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_1995\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4170\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-63");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"31989\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 31990, "epsg", 31990,
		      "SIRGAS 1995 / UTM zone 21N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 1995 / UTM zone 21N\",GEOGCS[\"SIRGAS 19");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_1995\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4170\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-57");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"31990\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 31991, "epsg", 31991,
		      "SIRGAS 1995 / UTM zone 22N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=22 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +");
    add_proj4text (p, 1, "units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 1995 / UTM zone 22N\",GEOGCS[\"SIRGAS 19");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_1995\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4170\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-51");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"31991\"],AXIS[\"Easting\",EAST],AXIS[\"Nor");
    add_srs_wkt (p, 13, "thing\",NORTH]]");
    p = add_epsg_def (first, last, 31992, "epsg", 31992,
		      "SIRGAS 1995 / UTM zone 17S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=17 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 1995 / UTM zone 17S\",GEOGCS[\"SIRGAS 19");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_1995\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4170\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-81");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"31992\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31993, "epsg", 31993,
		      "SIRGAS 1995 / UTM zone 18S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 1995 / UTM zone 18S\",GEOGCS[\"SIRGAS 19");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_1995\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4170\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-75");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"31993\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31994, "epsg", 31994,
		      "SIRGAS 1995 / UTM zone 19S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 1995 / UTM zone 19S\",GEOGCS[\"SIRGAS 19");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_1995\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4170\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-69");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"31994\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31995, "epsg", 31995,
		      "SIRGAS 1995 / UTM zone 20S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 1995 / UTM zone 20S\",GEOGCS[\"SIRGAS 19");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_1995\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4170\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-63");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"31995\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31996, "epsg", 31996,
		      "SIRGAS 1995 / UTM zone 21S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 1995 / UTM zone 21S\",GEOGCS[\"SIRGAS 19");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_1995\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4170\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-57");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"31996\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31997, "epsg", 31997,
		      "SIRGAS 1995 / UTM zone 22S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=22 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 1995 / UTM zone 22S\",GEOGCS[\"SIRGAS 19");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_1995\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4170\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-51");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"31997\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31998, "epsg", 31998,
		      "SIRGAS 1995 / UTM zone 23S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=23 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 1995 / UTM zone 23S\",GEOGCS[\"SIRGAS 19");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_1995\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4170\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-45");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"31998\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 31999, "epsg", 31999,
		      "SIRGAS 1995 / UTM zone 24S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=24 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 1995 / UTM zone 24S\",GEOGCS[\"SIRGAS 19");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_1995\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4170\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-39");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"31999\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32000, "epsg", 32000,
		      "SIRGAS 1995 / UTM zone 25S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=25 +south +ellps=GRS80 +towgs84=0,0,0,0,");
    add_proj4text (p, 1, "0,0,0 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"SIRGAS 1995 / UTM zone 25S\",GEOGCS[\"SIRGAS 19");
    add_srs_wkt (p, 1,
		 "95\",DATUM[\"Sistema_de_Referencia_Geocentrico_para_Amer");
    add_srs_wkt (p, 2,
		 "ica_del_Sur_1995\",SPHEROID[\"GRS 1980\",6378137,298.257");
    add_srs_wkt (p, 3,
		 "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0");
    add_srs_wkt (p, 4,
		 ",0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Greenwich\",0");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532");
    add_srs_wkt (p, 6,
		 "9251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPS");
    add_srs_wkt (p, 7,
		 "G\",\"4170\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900");
    add_srs_wkt (p, 8,
		 "1\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"la");
    add_srs_wkt (p, 9,
		 "titude_of_origin\",0],PARAMETER[\"central_meridian\",-33");
    add_srs_wkt (p, 10,
		 "],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",10000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"32000\"],AXIS[\"Easting\",EAST],AXI");
    add_srs_wkt (p, 13, "S[\"Northing\",NORTH]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_25 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 32001, "epsg", 32001,
		      "NAD27 / Montana North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.71666666666667 +lat_2=47.85 +lat_0=4");
    add_proj4text (p, 1,
		   "7 +lon_0=-109.5 +x_0=609601.2192024384 +y_0=0 +ellps=clr");
    add_proj4text (p, 2, "k66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Montana North\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",48.71");
    add_srs_wkt (p, 9,
		 "666666666667],PARAMETER[\"standard_parallel_2\",47.85],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"latitude_of_origin\",47],PARAMETER[\"central_");
    add_srs_wkt (p, 11,
		 "meridian\",-109.5],PARAMETER[\"false_easting\",2000000],");
    add_srs_wkt (p, 12,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"320");
    add_srs_wkt (p, 13, "01\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32002, "epsg", 32002,
		      "NAD27 / Montana Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.88333333333333 +lat_2=46.45 +lat_0=4");
    add_proj4text (p, 1,
		   "5.83333333333334 +lon_0=-109.5 +x_0=609601.2192024384 +y");
    add_proj4text (p, 2,
		   "_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Montana Central\",GEOGCS[\"NAD27\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",47.");
    add_srs_wkt (p, 9,
		 "88333333333333],PARAMETER[\"standard_parallel_2\",46.45]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"latitude_of_origin\",45.83333333333334],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"central_meridian\",-109.5],PARAMETER[\"false_ea");
    add_srs_wkt (p, 12,
		 "sting\",2000000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 13,
		 "TY[\"EPSG\",\"32002\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH");
    add_srs_wkt (p, 14, "]]");
    p = add_epsg_def (first, last, 32003, "epsg", 32003,
		      "NAD27 / Montana South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46.4 +lat_2=44.86666666666667 +lat_0=44");
    add_proj4text (p, 1,
		   " +lon_0=-109.5 +x_0=609601.2192024384 +y_0=0 +ellps=clrk");
    add_proj4text (p, 2, "66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Montana South\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",46.4]");
    add_srs_wkt (p, 9,
		 ",PARAMETER[\"standard_parallel_2\",44.86666666666667],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"latitude_of_origin\",44],PARAMETER[\"central_m");
    add_srs_wkt (p, 11,
		 "eridian\",-109.5],PARAMETER[\"false_easting\",2000000],P");
    add_srs_wkt (p, 12,
		 "ARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3200");
    add_srs_wkt (p, 13, "3\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32005, "epsg", 32005,
		      "NAD27 / Nebraska North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.85 +lat_2=42.81666666666667 +lat_0=4");
    add_proj4text (p, 1,
		   "1.33333333333334 +lon_0=-100 +x_0=609601.2192024384 +y_0");
    add_proj4text (p, 2, "=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Nebraska North\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",41.85");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"standard_parallel_2\",42.81666666666667],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"latitude_of_origin\",41.33333333333334],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"central_meridian\",-100],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 12,
		 "g\",2000000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13,
		 "EPSG\",\"32005\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32006, "epsg", 32006,
		      "NAD27 / Nebraska South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.28333333333333 +lat_2=41.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=39.66666666666666 +lon_0=-99.5 +x_0=609601.21");
    add_proj4text (p, 2,
		   "92024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft ");
    add_proj4text (p, 3, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Nebraska South\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",40.28");
    add_srs_wkt (p, 9,
		 "333333333333],PARAMETER[\"standard_parallel_2\",41.71666");
    add_srs_wkt (p, 10,
		 "666666667],PARAMETER[\"latitude_of_origin\",39.666666666");
    add_srs_wkt (p, 11,
		 "66666],PARAMETER[\"central_meridian\",-99.5],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_easting\",2000000],PARAMETER[\"false_northing\",0]");
    add_srs_wkt (p, 13,
		 ",AUTHORITY[\"EPSG\",\"32006\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 14, "\",NORTH]]");
    p = add_epsg_def (first, last, 32007, "epsg", 32007, "NAD27 / Nevada East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-115.5833333333333 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +datum=");
    add_proj4text (p, 2, "NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Nevada East\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",34.75],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-115.5833333333333],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",0.9999],PARAMETER[\"false_easting\",500000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"320");
    add_srs_wkt (p, 12, "07\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32008, "epsg", 32008,
		      "NAD27 / Nevada Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-116.6666666666667 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +datum=");
    add_proj4text (p, 2, "NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Nevada Central\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",34.75],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-116.6666666666667],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "32008\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32009, "epsg", 32009, "NAD27 / Nevada West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-118.5833333333333 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +datum=");
    add_proj4text (p, 2, "NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Nevada West\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",34.75],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-118.5833333333333],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",0.9999],PARAMETER[\"false_easting\",500000],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"320");
    add_srs_wkt (p, 12, "09\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32010, "epsg", 32010,
		      "NAD27 / New Hampshire");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.5 +lon_0=-71.66666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +da");
    add_proj4text (p, 2, "tum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / New Hampshire\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",42.5],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-71.66666666666667],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",0.999966667],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"32010\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32011, "epsg", 32011, "NAD27 / New Jersey");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38.83333333333334 +lon_0=-74.66666666");
    add_proj4text (p, 1,
		   "666667 +k=0.9999749999999999 +x_0=609601.2192024384 +y_0");
    add_proj4text (p, 2, "=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / New Jersey\",GEOGCS[\"NAD27\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6378");
    add_srs_wkt (p, 2,
		 "206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",38.83333333333334");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-74.66666666666667],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",0.999975],PARAMETER[\"false_east");
    add_srs_wkt (p, 11,
		 "ing\",2000000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 12,
		 "[\"EPSG\",\"32011\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 32012, "epsg", 32012,
		      "NAD27 / New Mexico East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-104.3333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "909091 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +datu");
    add_proj4text (p, 2, "m=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / New Mexico East\",GEOGCS[\"NAD27\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_M");
    add_srs_wkt (p, 8,
		 "ercator\"],PARAMETER[\"latitude_of_origin\",31],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-104.3333333333333],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",0.999909091],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"32012\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32013, "epsg", 32013,
		      "NAD27 / New Mexico Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-106.25 +k=0.9999 +x_0=1524");
    add_proj4text (p, 1,
		   "00.3048006096 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=u");
    add_proj4text (p, 2, "s-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / New Mexico Central\",GEOGCS[\"NAD27\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 186");
    add_srs_wkt (p, 2,
		 "6\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 3,
		 "8\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transvers");
    add_srs_wkt (p, 8,
		 "e_Mercator\"],PARAMETER[\"latitude_of_origin\",31],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"central_meridian\",-106.25],PARAMETER[\"scale_fac");
    add_srs_wkt (p, 10,
		 "tor\",0.9999],PARAMETER[\"false_easting\",500000],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"32013\"],");
    add_srs_wkt (p, 12, "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32014, "epsg", 32014,
		      "NAD27 / New Mexico West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-107.8333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "916667 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +datu");
    add_proj4text (p, 2, "m=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / New Mexico West\",GEOGCS[\"NAD27\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_M");
    add_srs_wkt (p, 8,
		 "ercator\"],PARAMETER[\"latitude_of_origin\",31],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-107.8333333333333],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",0.999916667],PARAMETER[\"false_easting\",5");
    add_srs_wkt (p, 11,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 12, ",\"32014\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32015, "epsg", 32015,
		      "NAD27 / New York East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40 +lon_0=-74.33333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "966667 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +datu");
    add_proj4text (p, 2, "m=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / New York East\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",40],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-74.33333333333333],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",0.999966667],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"32015\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32016, "epsg", 32016,
		      "NAD27 / New York Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40 +lon_0=-76.58333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "9375 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +datum=");
    add_proj4text (p, 2, "NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / New York Central\",GEOGCS[\"NAD27\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_M");
    add_srs_wkt (p, 8,
		 "ercator\"],PARAMETER[\"latitude_of_origin\",40],PARAMETE");
    add_srs_wkt (p, 9,
		 "R[\"central_meridian\",-76.58333333333333],PARAMETER[\"s");
    add_srs_wkt (p, 10,
		 "cale_factor\",0.9999375],PARAMETER[\"false_easting\",500");
    add_srs_wkt (p, 11,
		 "000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 12, "\"32016\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32017, "epsg", 32017,
		      "NAD27 / New York West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40 +lon_0=-78.58333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "9375 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +datum=");
    add_proj4text (p, 2, "NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / New York West\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",40],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"central_meridian\",-78.58333333333333],PARAMETER[\"sca");
    add_srs_wkt (p, 10,
		 "le_factor\",0.9999375],PARAMETER[\"false_easting\",50000");
    add_srs_wkt (p, 11,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "32017\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32018, "epsg", 32018,
		      "NAD27 / New York Long Island");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.03333333333333 +lat_2=40.66666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=40.5 +lon_0=-74 +x_0=304800.6096012192 +y_0=0");
    add_proj4text (p, 2, " +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / New York Long Island\",GEOGCS[\"NAD27\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1");
    add_srs_wkt (p, 2,
		 "866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4267\"]],UNIT[\"US survey foot\",0.304800609601");
    add_srs_wkt (p, 7,
		 "2192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_");
    add_srs_wkt (p, 8,
		 "Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 9,
		 ",41.03333333333333],PARAMETER[\"standard_parallel_2\",40");
    add_srs_wkt (p, 10,
		 ".66666666666666],PARAMETER[\"latitude_of_origin\",40.5],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"central_meridian\",-74],PARAMETER[\"false_ea");
    add_srs_wkt (p, 12,
		 "sting\",1000000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 13,
		 "TY[\"EPSG\",\"32018\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH");
    add_srs_wkt (p, 14, "]]");
    p = add_epsg_def (first, last, 32019, "epsg", 32019,
		      "NAD27 / North Carolina");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=34.33333333333334 +lat_2=36.16666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=33.75 +lon_0=-79 +x_0=609601.2192024384 +y_0=");
    add_proj4text (p, 2, "0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / North Carolina\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",34.33");
    add_srs_wkt (p, 9,
		 "333333333334],PARAMETER[\"standard_parallel_2\",36.16666");
    add_srs_wkt (p, 10,
		 "666666666],PARAMETER[\"latitude_of_origin\",33.75],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"central_meridian\",-79],PARAMETER[\"false_easting");
    add_srs_wkt (p, 12,
		 "\",2000000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13,
		 "EPSG\",\"32019\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32020, "epsg", 32020,
		      "NAD27 / North Dakota North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.43333333333333 +lat_2=48.73333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=47 +lon_0=-100.5 +x_0=609601.2192024384 +y_0=");
    add_proj4text (p, 2, "0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / North Dakota North\",GEOGCS[\"NAD27\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 186");
    add_srs_wkt (p, 2,
		 "6\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 3,
		 "8\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "47.43333333333333],PARAMETER[\"standard_parallel_2\",48.");
    add_srs_wkt (p, 10,
		 "73333333333333],PARAMETER[\"latitude_of_origin\",47],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"central_meridian\",-100.5],PARAMETER[\"false_ea");
    add_srs_wkt (p, 12,
		 "sting\",2000000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 13,
		 "TY[\"EPSG\",\"32020\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH");
    add_srs_wkt (p, 14, "]]");
    p = add_epsg_def (first, last, 32021, "epsg", 32021,
		      "NAD27 / North Dakota South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46.18333333333333 +lat_2=47.48333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=45.66666666666666 +lon_0=-100.5 +x_0=609601.2");
    add_proj4text (p, 2,
		   "192024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft");
    add_proj4text (p, 3, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / North Dakota South\",GEOGCS[\"NAD27\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 186");
    add_srs_wkt (p, 2,
		 "6\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 3,
		 "8\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "46.18333333333333],PARAMETER[\"standard_parallel_2\",47.");
    add_srs_wkt (p, 10,
		 "48333333333333],PARAMETER[\"latitude_of_origin\",45.6666");
    add_srs_wkt (p, 11,
		 "6666666666],PARAMETER[\"central_meridian\",-100.5],PARAM");
    add_srs_wkt (p, 12,
		 "ETER[\"false_easting\",2000000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 13,
		 "ng\",0],AUTHORITY[\"EPSG\",\"32021\"],AXIS[\"X\",EAST],A");
    add_srs_wkt (p, 14, "XIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32022, "epsg", 32022, "NAD27 / Ohio North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.43333333333333 +lat_2=41.7 +lat_0=39");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-82.5 +x_0=609601.2192024384 +y_0");
    add_proj4text (p, 2, "=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Ohio North\",GEOGCS[\"NAD27\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6378");
    add_srs_wkt (p, 2,
		 "206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",40.43333");
    add_srs_wkt (p, 9,
		 "333333333],PARAMETER[\"standard_parallel_2\",41.7],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"latitude_of_origin\",39.66666666666666],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"central_meridian\",-82.5],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",2000000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 13, "SG\",\"32022\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32023, "epsg", 32023, "NAD27 / Ohio South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.73333333333333 +lat_2=40.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=38 +lon_0=-82.5 +x_0=609601.2192024384 +y_0=0");
    add_proj4text (p, 2, " +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Ohio South\",GEOGCS[\"NAD27\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6378");
    add_srs_wkt (p, 2,
		 "206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",38.73333");
    add_srs_wkt (p, 9,
		 "333333333],PARAMETER[\"standard_parallel_2\",40.03333333");
    add_srs_wkt (p, 10,
		 "333333],PARAMETER[\"latitude_of_origin\",38],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "central_meridian\",-82.5],PARAMETER[\"false_easting\",20");
    add_srs_wkt (p, 12,
		 "00000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 13, ",\"32023\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32024, "epsg", 32024,
		      "NAD27 / Oklahoma North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=35.56666666666667 +lat_2=36.76666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=35 +lon_0=-98 +x_0=609601.2192024384 +y_0=0 +");
    add_proj4text (p, 2, "ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Oklahoma North\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",35.56");
    add_srs_wkt (p, 9,
		 "666666666667],PARAMETER[\"standard_parallel_2\",36.76666");
    add_srs_wkt (p, 10,
		 "666666667],PARAMETER[\"latitude_of_origin\",35],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"central_meridian\",-98],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 12,
		 "2000000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 13, "G\",\"32024\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32025, "epsg", 32025,
		      "NAD27 / Oklahoma South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.93333333333333 +lat_2=35.23333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=33.33333333333334 +lon_0=-98 +x_0=609601.2192");
    add_proj4text (p, 2,
		   "024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +n");
    add_proj4text (p, 3, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Oklahoma South\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",33.93");
    add_srs_wkt (p, 9,
		 "333333333333],PARAMETER[\"standard_parallel_2\",35.23333");
    add_srs_wkt (p, 10,
		 "333333333],PARAMETER[\"latitude_of_origin\",33.333333333");
    add_srs_wkt (p, 11,
		 "33334],PARAMETER[\"central_meridian\",-98],PARAMETER[\"f");
    add_srs_wkt (p, 12,
		 "alse_easting\",2000000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 13,
		 "AUTHORITY[\"EPSG\",\"32025\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 14, ",NORTH]]");
    p = add_epsg_def (first, last, 32026, "epsg", 32026,
		      "NAD27 / Oregon North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.33333333333334 +lat_2=46 +lat_0=43.6");
    add_proj4text (p, 1,
		   "6666666666666 +lon_0=-120.5 +x_0=609601.2192024384 +y_0=");
    add_proj4text (p, 2, "0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Oregon North\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",44.33333");
    add_srs_wkt (p, 9,
		 "333333334],PARAMETER[\"standard_parallel_2\",46],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"latitude_of_origin\",43.66666666666666],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "central_meridian\",-120.5],PARAMETER[\"false_easting\",2");
    add_srs_wkt (p, 12,
		 "000000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 13, "\",\"32026\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32027, "epsg", 32027,
		      "NAD27 / Oregon South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=42.33333333333334 +lat_2=44 +lat_0=41.6");
    add_proj4text (p, 1,
		   "6666666666666 +lon_0=-120.5 +x_0=609601.2192024384 +y_0=");
    add_proj4text (p, 2, "0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Oregon South\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",42.33333");
    add_srs_wkt (p, 9,
		 "333333334],PARAMETER[\"standard_parallel_2\",44],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"latitude_of_origin\",41.66666666666666],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "central_meridian\",-120.5],PARAMETER[\"false_easting\",2");
    add_srs_wkt (p, 12,
		 "000000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 13, "\",\"32027\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32028, "epsg", 32028,
		      "NAD27 / Pennsylvania North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.88333333333333 +lat_2=41.95 +lat_0=4");
    add_proj4text (p, 1,
		   "0.16666666666666 +lon_0=-77.75 +x_0=609601.2192024384 +y");
    add_proj4text (p, 2,
		   "_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Pennsylvania North\",GEOGCS[\"NAD27\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 186");
    add_srs_wkt (p, 2,
		 "6\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 3,
		 "8\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "40.88333333333333],PARAMETER[\"standard_parallel_2\",41.");
    add_srs_wkt (p, 10,
		 "95],PARAMETER[\"latitude_of_origin\",40.16666666666666],");
    add_srs_wkt (p, 11,
		 "PARAMETER[\"central_meridian\",-77.75],PARAMETER[\"false");
    add_srs_wkt (p, 12,
		 "_easting\",2000000],PARAMETER[\"false_northing\",0],AUTH");
    add_srs_wkt (p, 13,
		 "ORITY[\"EPSG\",\"32028\"],AXIS[\"X\",EAST],AXIS[\"Y\",NO");
    add_srs_wkt (p, 14, "RTH]]");
    p = add_epsg_def (first, last, 32029, "epsg", 32029,
		      "NAD27 / Pennsylvania South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.93333333333333 +lat_2=40.8 +lat_0=39");
    add_proj4text (p, 1,
		   ".33333333333334 +lon_0=-77.75 +x_0=609601.2192024384 +y_");
    add_proj4text (p, 2,
		   "0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Pennsylvania South\",GEOGCS[\"NAD27\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 186");
    add_srs_wkt (p, 2,
		 "6\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 3,
		 "8\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "39.93333333333333],PARAMETER[\"standard_parallel_2\",40.");
    add_srs_wkt (p, 10,
		 "8],PARAMETER[\"latitude_of_origin\",39.33333333333334],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"central_meridian\",-77.75],PARAMETER[\"false_");
    add_srs_wkt (p, 12,
		 "easting\",2000000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 13,
		 "RITY[\"EPSG\",\"32029\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 14, "TH]]");
    p = add_epsg_def (first, last, 32030, "epsg", 32030,
		      "NAD27 / Rhode Island");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.08333333333334 +lon_0=-71.5 +k=0.9");
    add_proj4text (p, 1,
		   "999938 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +datu");
    add_proj4text (p, 2, "m=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Rhode Island\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",41.08333333333334");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-71.5],PARAMETER[\"scal");
    add_srs_wkt (p, 10,
		 "e_factor\",0.9999938],PARAMETER[\"false_easting\",500000");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"3");
    add_srs_wkt (p, 12, "2030\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32031, "epsg", 32031,
		      "NAD27 / South Carolina North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.76666666666667 +lat_2=34.96666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=33 +lon_0=-81 +x_0=609601.2192024384 +y_0=0 +");
    add_proj4text (p, 2, "ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / South Carolina North\",GEOGCS[\"NAD27\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1");
    add_srs_wkt (p, 2,
		 "866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4267\"]],UNIT[\"US survey foot\",0.304800609601");
    add_srs_wkt (p, 7,
		 "2192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_");
    add_srs_wkt (p, 8,
		 "Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 9,
		 ",33.76666666666667],PARAMETER[\"standard_parallel_2\",34");
    add_srs_wkt (p, 10,
		 ".96666666666667],PARAMETER[\"latitude_of_origin\",33],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"central_meridian\",-81],PARAMETER[\"false_east");
    add_srs_wkt (p, 12,
		 "ing\",2000000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 13,
		 "[\"EPSG\",\"32031\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    add_srs_wkt (p, 14, "");
    p = add_epsg_def (first, last, 32033, "epsg", 32033,
		      "NAD27 / South Carolina South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=32.33333333333334 +lat_2=33.66666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=31.83333333333333 +lon_0=-81 +x_0=609601.2192");
    add_proj4text (p, 2,
		   "024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +n");
    add_proj4text (p, 3, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / South Carolina South\",GEOGCS[\"NAD27\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1");
    add_srs_wkt (p, 2,
		 "866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4267\"]],UNIT[\"US survey foot\",0.304800609601");
    add_srs_wkt (p, 7,
		 "2192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_");
    add_srs_wkt (p, 8,
		 "Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\"");
    add_srs_wkt (p, 9,
		 ",32.33333333333334],PARAMETER[\"standard_parallel_2\",33");
    add_srs_wkt (p, 10,
		 ".66666666666666],PARAMETER[\"latitude_of_origin\",31.833");
    add_srs_wkt (p, 11,
		 "33333333333],PARAMETER[\"central_meridian\",-81],PARAMET");
    add_srs_wkt (p, 12,
		 "ER[\"false_easting\",2000000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 13,
		 "\",0],AUTHORITY[\"EPSG\",\"32033\"],AXIS[\"X\",EAST],AXI");
    add_srs_wkt (p, 14, "S[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32034, "epsg", 32034,
		      "NAD27 / South Dakota North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.41666666666666 +lat_2=45.68333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=43.83333333333334 +lon_0=-100 +x_0=609601.219");
    add_proj4text (p, 2,
		   "2024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +");
    add_proj4text (p, 3, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / South Dakota North\",GEOGCS[\"NAD27\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 186");
    add_srs_wkt (p, 2,
		 "6\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 3,
		 "8\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "44.41666666666666],PARAMETER[\"standard_parallel_2\",45.");
    add_srs_wkt (p, 10,
		 "68333333333333],PARAMETER[\"latitude_of_origin\",43.8333");
    add_srs_wkt (p, 11,
		 "3333333334],PARAMETER[\"central_meridian\",-100],PARAMET");
    add_srs_wkt (p, 12,
		 "ER[\"false_easting\",2000000],PARAMETER[\"false_northing");
    add_srs_wkt (p, 13,
		 "\",0],AUTHORITY[\"EPSG\",\"32034\"],AXIS[\"X\",EAST],AXI");
    add_srs_wkt (p, 14, "S[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32035, "epsg", 32035,
		      "NAD27 / South Dakota South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=42.83333333333334 +lat_2=44.4 +lat_0=42");
    add_proj4text (p, 1,
		   ".33333333333334 +lon_0=-100.3333333333333 +x_0=609601.21");
    add_proj4text (p, 2,
		   "92024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft ");
    add_proj4text (p, 3, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / South Dakota South\",GEOGCS[\"NAD27\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 186");
    add_srs_wkt (p, 2,
		 "6\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 3,
		 "8\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "42.83333333333334],PARAMETER[\"standard_parallel_2\",44.");
    add_srs_wkt (p, 10,
		 "4],PARAMETER[\"latitude_of_origin\",42.33333333333334],P");
    add_srs_wkt (p, 11,
		 "ARAMETER[\"central_meridian\",-100.3333333333333],PARAME");
    add_srs_wkt (p, 12,
		 "TER[\"false_easting\",2000000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 13,
		 "g\",0],AUTHORITY[\"EPSG\",\"32035\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 14, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32036, "epsg", 32036,
		      "NAD27 / Tennessee (deprecated)");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=35.25 +lat_2=36.41666666666666 +lat_0=3");
    add_proj4text (p, 1,
		   "4.66666666666666 +lon_0=-86 +x_0=30480.06096012192 +y_0=");
    add_proj4text (p, 2, "0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Tennessee (deprecated)\",GEOGCS[\"NAD27");
    add_srs_wkt (p, 1,
		 "\",DATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke");
    add_srs_wkt (p, 2,
		 " 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 3,
		 "7008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwic");
    add_srs_wkt (p, 4,
		 "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01");
    add_srs_wkt (p, 5,
		 "745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[");
    add_srs_wkt (p, 6,
		 "\"EPSG\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096");
    add_srs_wkt (p, 7,
		 "012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lamber");
    add_srs_wkt (p, 8,
		 "t_Conformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1");
    add_srs_wkt (p, 9,
		 "\",35.25],PARAMETER[\"standard_parallel_2\",36.416666666");
    add_srs_wkt (p, 10,
		 "66666],PARAMETER[\"latitude_of_origin\",34.6666666666666");
    add_srs_wkt (p, 11,
		 "6],PARAMETER[\"central_meridian\",-86],PARAMETER[\"false");
    add_srs_wkt (p, 12,
		 "_easting\",100000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 13,
		 "RITY[\"EPSG\",\"32036\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 14, "TH]]");
    p = add_epsg_def (first, last, 32037, "epsg", 32037, "NAD27 / Texas North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=34.65 +lat_2=36.18333333333333 +lat_0=3");
    add_proj4text (p, 1,
		   "4 +lon_0=-101.5 +x_0=609601.2192024384 +y_0=0 +ellps=clr");
    add_proj4text (p, 2, "k66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Texas North\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",34.65],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"standard_parallel_2\",36.18333333333333],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"latitude_of_origin\",34],PARAMETER[\"central_mer");
    add_srs_wkt (p, 11,
		 "idian\",-101.5],PARAMETER[\"false_easting\",2000000],PAR");
    add_srs_wkt (p, 12,
		 "AMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"32037\"");
    add_srs_wkt (p, 13, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32038, "epsg", 32038,
		      "NAD27 / Texas North Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=32.13333333333333 +lat_2=33.96666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=31.66666666666667 +lon_0=-97.5 +x_0=609601.21");
    add_proj4text (p, 2,
		   "92024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft ");
    add_proj4text (p, 3, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Texas North Central\",GEOGCS[\"NAD27\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 18");
    add_srs_wkt (p, 2,
		 "66\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "08\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "32.13333333333333],PARAMETER[\"standard_parallel_2\",33.");
    add_srs_wkt (p, 10,
		 "96666666666667],PARAMETER[\"latitude_of_origin\",31.6666");
    add_srs_wkt (p, 11,
		 "6666666667],PARAMETER[\"central_meridian\",-97.5],PARAME");
    add_srs_wkt (p, 12,
		 "TER[\"false_easting\",2000000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 13,
		 "g\",0],AUTHORITY[\"EPSG\",\"32038\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 14, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32039, "epsg", 32039,
		      "NAD27 / Texas Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.11666666666667 +lat_2=31.88333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=29.66666666666667 +lon_0=-100.3333333333333 +");
    add_proj4text (p, 2,
		   "x_0=609601.2192024384 +y_0=0 +ellps=clrk66 +datum=NAD27 ");
    add_proj4text (p, 3, "+units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Texas Central\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",30.11");
    add_srs_wkt (p, 9,
		 "666666666667],PARAMETER[\"standard_parallel_2\",31.88333");
    add_srs_wkt (p, 10,
		 "333333333],PARAMETER[\"latitude_of_origin\",29.666666666");
    add_srs_wkt (p, 11,
		 "66667],PARAMETER[\"central_meridian\",-100.3333333333333");
    add_srs_wkt (p, 12,
		 "],PARAMETER[\"false_easting\",2000000],PARAMETER[\"false");
    add_srs_wkt (p, 13,
		 "_northing\",0],AUTHORITY[\"EPSG\",\"32039\"],AXIS[\"X\",");
    add_srs_wkt (p, 14, "EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32040, "epsg", 32040,
		      "NAD27 / Texas South Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=28.38333333333333 +lat_2=30.28333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=27.83333333333333 +lon_0=-99 +x_0=609601.2192");
    add_proj4text (p, 2,
		   "024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +n");
    add_proj4text (p, 3, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Texas South Central\",GEOGCS[\"NAD27\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 18");
    add_srs_wkt (p, 2,
		 "66\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "08\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "28.38333333333333],PARAMETER[\"standard_parallel_2\",30.");
    add_srs_wkt (p, 10,
		 "28333333333333],PARAMETER[\"latitude_of_origin\",27.8333");
    add_srs_wkt (p, 11,
		 "3333333333],PARAMETER[\"central_meridian\",-99],PARAMETE");
    add_srs_wkt (p, 12,
		 "R[\"false_easting\",2000000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 13,
		 ",0],AUTHORITY[\"EPSG\",\"32040\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 14, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32041, "epsg", 32041, "NAD27 / Texas South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=26.16666666666667 +lat_2=27.83333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=25.66666666666667 +lon_0=-98.5 +x_0=609601.21");
    add_proj4text (p, 2,
		   "92024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft ");
    add_proj4text (p, 3, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Texas South\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",26.16666");
    add_srs_wkt (p, 9,
		 "666666667],PARAMETER[\"standard_parallel_2\",27.83333333");
    add_srs_wkt (p, 10,
		 "333333],PARAMETER[\"latitude_of_origin\",25.666666666666");
    add_srs_wkt (p, 11,
		 "67],PARAMETER[\"central_meridian\",-98.5],PARAMETER[\"fa");
    add_srs_wkt (p, 12,
		 "lse_easting\",2000000],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 13,
		 "UTHORITY[\"EPSG\",\"32041\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 14, ",NORTH]]");
    p = add_epsg_def (first, last, 32042, "epsg", 32042, "NAD27 / Utah North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.71666666666667 +lat_2=41.78333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=40.33333333333334 +lon_0=-111.5 +x_0=609601.2");
    add_proj4text (p, 2,
		   "192024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft");
    add_proj4text (p, 3, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Utah North\",GEOGCS[\"NAD27\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6378");
    add_srs_wkt (p, 2,
		 "206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",40.71666");
    add_srs_wkt (p, 9,
		 "666666667],PARAMETER[\"standard_parallel_2\",41.78333333");
    add_srs_wkt (p, 10,
		 "333333],PARAMETER[\"latitude_of_origin\",40.333333333333");
    add_srs_wkt (p, 11,
		 "34],PARAMETER[\"central_meridian\",-111.5],PARAMETER[\"f");
    add_srs_wkt (p, 12,
		 "alse_easting\",2000000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 13,
		 "AUTHORITY[\"EPSG\",\"32042\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 14, ",NORTH]]");
    p = add_epsg_def (first, last, 32043, "epsg", 32043,
		      "NAD27 / Utah Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.01666666666667 +lat_2=40.65 +lat_0=3");
    add_proj4text (p, 1,
		   "8.33333333333334 +lon_0=-111.5 +x_0=609601.2192024384 +y");
    add_proj4text (p, 2,
		   "_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Utah Central\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",39.01666");
    add_srs_wkt (p, 9,
		 "666666667],PARAMETER[\"standard_parallel_2\",40.65],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"latitude_of_origin\",38.33333333333334],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"central_meridian\",-111.5],PARAMETER[\"false_easting");
    add_srs_wkt (p, 12,
		 "\",2000000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13,
		 "EPSG\",\"32043\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32044, "epsg", 32044, "NAD27 / Utah South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.21666666666667 +lat_2=38.35 +lat_0=3");
    add_proj4text (p, 1,
		   "6.66666666666666 +lon_0=-111.5 +x_0=609601.2192024384 +y");
    add_proj4text (p, 2,
		   "_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Utah South\",GEOGCS[\"NAD27\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6378");
    add_srs_wkt (p, 2,
		 "206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conformal");
    add_srs_wkt (p, 8,
		 "_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",37.21666");
    add_srs_wkt (p, 9,
		 "666666667],PARAMETER[\"standard_parallel_2\",38.35],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"latitude_of_origin\",36.66666666666666],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"central_meridian\",-111.5],PARAMETER[\"false_easting");
    add_srs_wkt (p, 12,
		 "\",2000000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13,
		 "EPSG\",\"32044\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32045, "epsg", 32045, "NAD27 / Vermont");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.5 +lon_0=-72.5 +k=0.999964286 +x_0");
    add_proj4text (p, 1,
		   "=152400.3048006096 +y_0=0 +ellps=clrk66 +datum=NAD27 +un");
    add_proj4text (p, 2, "its=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Vermont\",GEOGCS[\"NAD27\",DATUM[\"Nort");
    add_srs_wkt (p, 1,
		 "h_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6378206");
    add_srs_wkt (p, 2,
		 ".4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"426");
    add_srs_wkt (p, 6,
		 "7\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTHORI");
    add_srs_wkt (p, 7,
		 "TY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercator\"");
    add_srs_wkt (p, 8,
		 "],PARAMETER[\"latitude_of_origin\",42.5],PARAMETER[\"cen");
    add_srs_wkt (p, 9,
		 "tral_meridian\",-72.5],PARAMETER[\"scale_factor\",0.9999");
    add_srs_wkt (p, 10,
		 "64286],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32045\"],AXIS[\"");
    add_srs_wkt (p, 12, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32046, "epsg", 32046,
		      "NAD27 / Virginia North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.03333333333333 +lat_2=39.2 +lat_0=37");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-78.5 +x_0=609601.2192024384 +y_0");
    add_proj4text (p, 2, "=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Virginia North\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",38.03");
    add_srs_wkt (p, 9,
		 "333333333333],PARAMETER[\"standard_parallel_2\",39.2],PA");
    add_srs_wkt (p, 10,
		 "RAMETER[\"latitude_of_origin\",37.66666666666666],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"central_meridian\",-78.5],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 12,
		 "g\",2000000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13,
		 "EPSG\",\"32046\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32047, "epsg", 32047,
		      "NAD27 / Virginia South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.76666666666667 +lat_2=37.96666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.33333333333334 +lon_0=-78.5 +x_0=609601.21");
    add_proj4text (p, 2,
		   "92024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft ");
    add_proj4text (p, 3, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Virginia South\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Confor");
    add_srs_wkt (p, 8,
		 "mal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",36.76");
    add_srs_wkt (p, 9,
		 "666666666667],PARAMETER[\"standard_parallel_2\",37.96666");
    add_srs_wkt (p, 10,
		 "666666667],PARAMETER[\"latitude_of_origin\",36.333333333");
    add_srs_wkt (p, 11,
		 "33334],PARAMETER[\"central_meridian\",-78.5],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_easting\",2000000],PARAMETER[\"false_northing\",0]");
    add_srs_wkt (p, 13,
		 ",AUTHORITY[\"EPSG\",\"32047\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 14, "\",NORTH]]");
    p = add_epsg_def (first, last, 32048, "epsg", 32048,
		      "NAD27 / Washington North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.5 +lat_2=48.73333333333333 +lat_0=47");
    add_proj4text (p, 1,
		   " +lon_0=-120.8333333333333 +x_0=609601.2192024384 +y_0=0");
    add_proj4text (p, 2, " +ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Washington North\",GEOGCS[\"NAD27\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",47.");
    add_srs_wkt (p, 9,
		 "5],PARAMETER[\"standard_parallel_2\",48.73333333333333],");
    add_srs_wkt (p, 10,
		 "PARAMETER[\"latitude_of_origin\",47],PARAMETER[\"central");
    add_srs_wkt (p, 11,
		 "_meridian\",-120.8333333333333],PARAMETER[\"false_eastin");
    add_srs_wkt (p, 12,
		 "g\",2000000],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 13,
		 "EPSG\",\"32048\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32049, "epsg", 32049,
		      "NAD27 / Washington South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.83333333333334 +lat_2=47.33333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=45.33333333333334 +lon_0=-120.5 +x_0=609601.2");
    add_proj4text (p, 2,
		   "192024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft");
    add_proj4text (p, 3, " +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Washington South\",GEOGCS[\"NAD27\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",45.");
    add_srs_wkt (p, 9,
		 "83333333333334],PARAMETER[\"standard_parallel_2\",47.333");
    add_srs_wkt (p, 10,
		 "33333333334],PARAMETER[\"latitude_of_origin\",45.3333333");
    add_srs_wkt (p, 11,
		 "3333334],PARAMETER[\"central_meridian\",-120.5],PARAMETE");
    add_srs_wkt (p, 12,
		 "R[\"false_easting\",2000000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 13,
		 ",0],AUTHORITY[\"EPSG\",\"32049\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 14, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32050, "epsg", 32050,
		      "NAD27 / West Virginia North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39 +lat_2=40.25 +lat_0=38.5 +lon_0=-79.");
    add_proj4text (p, 1,
		   "5 +x_0=609601.2192024384 +y_0=0 +ellps=clrk66 +datum=NAD");
    add_proj4text (p, 2, "27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / West Virginia North\",GEOGCS[\"NAD27\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 18");
    add_srs_wkt (p, 2,
		 "66\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "08\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "39],PARAMETER[\"standard_parallel_2\",40.25],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "latitude_of_origin\",38.5],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 11,
		 ",-79.5],PARAMETER[\"false_easting\",2000000],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32050\"],AXIS[\"");
    add_srs_wkt (p, 13, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32051, "epsg", 32051,
		      "NAD27 / West Virginia South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.48333333333333 +lat_2=38.88333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=37 +lon_0=-81 +x_0=609601.2192024384 +y_0=0 +");
    add_proj4text (p, 2, "ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / West Virginia South\",GEOGCS[\"NAD27\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 18");
    add_srs_wkt (p, 2,
		 "66\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"70");
    add_srs_wkt (p, 3,
		 "08\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "37.48333333333333],PARAMETER[\"standard_parallel_2\",38.");
    add_srs_wkt (p, 10,
		 "88333333333333],PARAMETER[\"latitude_of_origin\",37],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"central_meridian\",-81],PARAMETER[\"false_easti");
    add_srs_wkt (p, 12,
		 "ng\",2000000],PARAMETER[\"false_northing\",0],AUTHORITY[");
    add_srs_wkt (p, 13,
		 "\"EPSG\",\"32051\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32052, "epsg", 32052,
		      "NAD27 / Wisconsin North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.56666666666667 +lat_2=46.76666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=45.16666666666666 +lon_0=-90 +x_0=609601.2192");
    add_proj4text (p, 2,
		   "024384 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +n");
    add_proj4text (p, 3, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Wisconsin North\",GEOGCS[\"NAD27\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",45.");
    add_srs_wkt (p, 9,
		 "56666666666667],PARAMETER[\"standard_parallel_2\",46.766");
    add_srs_wkt (p, 10,
		 "66666666667],PARAMETER[\"latitude_of_origin\",45.1666666");
    add_srs_wkt (p, 11,
		 "6666666],PARAMETER[\"central_meridian\",-90],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_easting\",2000000],PARAMETER[\"false_northing\",0]");
    add_srs_wkt (p, 13,
		 ",AUTHORITY[\"EPSG\",\"32052\"],AXIS[\"X\",EAST],AXIS[\"Y");
    add_srs_wkt (p, 14, "\",NORTH]]");
    p = add_epsg_def (first, last, 32053, "epsg", 32053,
		      "NAD27 / Wisconsin Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.25 +lat_2=45.5 +lat_0=43.83333333333");
    add_proj4text (p, 1,
		   "334 +lon_0=-90 +x_0=609601.2192024384 +y_0=0 +ellps=clrk");
    add_proj4text (p, 2, "66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Wisconsin Central\",GEOGCS[\"NAD27\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866");
    add_srs_wkt (p, 2,
		 "\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008");
    add_srs_wkt (p, 3,
		 "\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",");
    add_srs_wkt (p, 4,
		 "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453");
    add_srs_wkt (p, 5,
		 "29251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 6,
		 "SG\",\"4267\"]],UNIT[\"US survey foot\",0.30480060960121");
    add_srs_wkt (p, 7,
		 "92,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Co");
    add_srs_wkt (p, 8,
		 "nformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",4");
    add_srs_wkt (p, 9,
		 "4.25],PARAMETER[\"standard_parallel_2\",45.5],PARAMETER[");
    add_srs_wkt (p, 10,
		 "\"latitude_of_origin\",43.83333333333334],PARAMETER[\"ce");
    add_srs_wkt (p, 11,
		 "ntral_meridian\",-90],PARAMETER[\"false_easting\",200000");
    add_srs_wkt (p, 12,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 13, "32053\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32054, "epsg", 32054,
		      "NAD27 / Wisconsin South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=42.73333333333333 +lat_2=44.06666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=42 +lon_0=-90 +x_0=609601.2192024384 +y_0=0 +");
    add_proj4text (p, 2, "ellps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Wisconsin South\",GEOGCS[\"NAD27\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192");
    add_srs_wkt (p, 7,
		 ",AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_Conf");
    add_srs_wkt (p, 8,
		 "ormal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",42.");
    add_srs_wkt (p, 9,
		 "73333333333333],PARAMETER[\"standard_parallel_2\",44.066");
    add_srs_wkt (p, 10,
		 "66666666667],PARAMETER[\"latitude_of_origin\",42],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"central_meridian\",-90],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 12,
		 ",2000000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 13, "SG\",\"32054\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32055, "epsg", 32055,
		      "NAD27 / Wyoming East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.66666666666666 +lon_0=-105.1666666");
    add_proj4text (p, 1,
		   "666667 +k=0.999941177 +x_0=152400.3048006096 +y_0=0 +ell");
    add_proj4text (p, 2, "ps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Wyoming East\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",40.66666666666666");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-105.1666666666667],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",0.999941177],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"32055\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 32056, "epsg", 32056,
		      "NAD27 / Wyoming East Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.66666666666666 +lon_0=-107.3333333");
    add_proj4text (p, 1,
		   "333333 +k=0.999941177 +x_0=152400.3048006096 +y_0=0 +ell");
    add_proj4text (p, 2, "ps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Wyoming East Central\",GEOGCS[\"NAD27\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1");
    add_srs_wkt (p, 2,
		 "866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4267\"]],UNIT[\"US survey foot\",0.304800609601");
    add_srs_wkt (p, 7,
		 "2192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transver");
    add_srs_wkt (p, 8,
		 "se_Mercator\"],PARAMETER[\"latitude_of_origin\",40.66666");
    add_srs_wkt (p, 9,
		 "666666666],PARAMETER[\"central_meridian\",-107.333333333");
    add_srs_wkt (p, 10,
		 "3333],PARAMETER[\"scale_factor\",0.999941177],PARAMETER[");
    add_srs_wkt (p, 11,
		 "\"false_easting\",500000],PARAMETER[\"false_northing\",0");
    add_srs_wkt (p, 12,
		 "],AUTHORITY[\"EPSG\",\"32056\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 32057, "epsg", 32057,
		      "NAD27 / Wyoming West Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.66666666666666 +lon_0=-108.75 +k=0");
    add_proj4text (p, 1,
		   ".999941177 +x_0=152400.3048006096 +y_0=0 +ellps=clrk66 +");
    add_proj4text (p, 2, "datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Wyoming West Central\",GEOGCS[\"NAD27\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1");
    add_srs_wkt (p, 2,
		 "866\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7");
    add_srs_wkt (p, 3,
		 "008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich");
    add_srs_wkt (p, 4,
		 "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017");
    add_srs_wkt (p, 5,
		 "45329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"");
    add_srs_wkt (p, 6,
		 "EPSG\",\"4267\"]],UNIT[\"US survey foot\",0.304800609601");
    add_srs_wkt (p, 7,
		 "2192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transver");
    add_srs_wkt (p, 8,
		 "se_Mercator\"],PARAMETER[\"latitude_of_origin\",40.66666");
    add_srs_wkt (p, 9,
		 "666666666],PARAMETER[\"central_meridian\",-108.75],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"scale_factor\",0.999941177],PARAMETER[\"false_eas");
    add_srs_wkt (p, 11,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 12,
		 "[\"EPSG\",\"32057\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    add_srs_wkt (p, 13, "");
    p = add_epsg_def (first, last, 32058, "epsg", 32058,
		      "NAD27 / Wyoming West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.66666666666666 +lon_0=-110.0833333");
    add_proj4text (p, 1,
		   "333333 +k=0.999941177 +x_0=152400.3048006096 +y_0=0 +ell");
    add_proj4text (p, 2, "ps=clrk66 +datum=NAD27 +units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Wyoming West\",GEOGCS[\"NAD27\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",637");
    add_srs_wkt (p, 2,
		 "8206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"US survey foot\",0.3048006096012192,AUTH");
    add_srs_wkt (p, 7,
		 "ORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercat");
    add_srs_wkt (p, 8,
		 "or\"],PARAMETER[\"latitude_of_origin\",40.66666666666666");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-110.0833333333333],PAR");
    add_srs_wkt (p, 10,
		 "AMETER[\"scale_factor\",0.999941177],PARAMETER[\"false_e");
    add_srs_wkt (p, 11,
		 "asting\",500000],PARAMETER[\"false_northing\",0],AUTHORI");
    add_srs_wkt (p, 12,
		 "TY[\"EPSG\",\"32058\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH");
    add_srs_wkt (p, 13, "]]");
    p = add_epsg_def (first, last, 32061, "epsg", 32061,
		      "NAD27 / Guatemala Norte");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=16.81666666666667 +lat_0=16.81666666666");
    add_proj4text (p, 1,
		   "667 +lon_0=-90.33333333333333 +k_0=0.99992226 +x_0=50000");
    add_proj4text (p, 2,
		   "0 +y_0=292209.579 +ellps=clrk66 +datum=NAD27 +units=m +n");
    add_proj4text (p, 3, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Guatemala Norte\",GEOGCS[\"NAD27\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\"");
    add_srs_wkt (p, 2,
		 ",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Lambert_Conformal_Conic_1SP\"],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"latitude_of_origin\",16.81666666666667],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-90.33333333333333],PARAMETER[\"scale");
    add_srs_wkt (p, 10,
		 "_factor\",0.99992226],PARAMETER[\"false_easting\",500000");
    add_srs_wkt (p, 11,
		 "],PARAMETER[\"false_northing\",292209.579],AUTHORITY[\"E");
    add_srs_wkt (p, 12, "PSG\",\"32061\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32062, "epsg", 32062,
		      "NAD27 / Guatemala Sur");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=14.9 +lat_0=14.9 +lon_0=-90.33333333333");
    add_proj4text (p, 1,
		   "333 +k_0=0.99989906 +x_0=500000 +y_0=325992.681 +ellps=c");
    add_proj4text (p, 2, "lrk66 +datum=NAD27 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Guatemala Sur\",GEOGCS[\"NAD27\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6");
    add_srs_wkt (p, 2,
		 "378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]]");
    add_srs_wkt (p, 3,
		 ",AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AU");
    add_srs_wkt (p, 4,
		 "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925");
    add_srs_wkt (p, 5,
		 "1994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_1SP\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"latitude_of_origin\",14.9],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 9,
		 "an\",-90.33333333333333],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 10,
		 "989906],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",325992.681],AUTHORITY[\"EPSG\",\"32062\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32064, "epsg", 32064,
		      "NAD27 / BLM 14N (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-99 +k=0.9996 +x_0=500000.00");
    add_proj4text (p, 1,
		   "1016002 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / BLM 14N (ftUS)\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-99],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 10,
		 "6],PARAMETER[\"false_easting\",1640416.67],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32064\"],AXIS[\"");
    add_srs_wkt (p, 12, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32065, "epsg", 32065,
		      "NAD27 / BLM 15N (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-93 +k=0.9996 +x_0=500000.00");
    add_proj4text (p, 1,
		   "1016002 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / BLM 15N (ftUS)\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-93],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 10,
		 "6],PARAMETER[\"false_easting\",1640416.67],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32065\"],AXIS[\"");
    add_srs_wkt (p, 12, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32066, "epsg", 32066,
		      "NAD27 / BLM 16N (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-87 +k=0.9996 +x_0=500000.00");
    add_proj4text (p, 1,
		   "1016002 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / BLM 16N (ftUS)\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-87],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 10,
		 "6],PARAMETER[\"false_easting\",1640416.67],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32066\"],AXIS[\"");
    add_srs_wkt (p, 12, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32067, "epsg", 32067,
		      "NAD27 / BLM 17N (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-81 +k=0.9996 +x_0=500000.00");
    add_proj4text (p, 1,
		   "1016002 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / BLM 17N (ftUS)\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012192,A");
    add_srs_wkt (p, 7,
		 "UTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mer");
    add_srs_wkt (p, 8,
		 "cator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "central_meridian\",-81],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 10,
		 "6],PARAMETER[\"false_easting\",1640416.67],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32067\"],AXIS[\"");
    add_srs_wkt (p, 12, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32074, "epsg", 32074,
		      "NAD27 / BLM 14N (feet) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-99 +k=0.9996 +x_0=500000.00");
    add_proj4text (p, 1,
		   "1016002 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / BLM 14N (feet) (deprecated)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD27\",DATUM[\"North_American_Datum_1927\",SPHEROID[\"C");
    add_srs_wkt (p, 2,
		 "larke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4267\"]],UNIT[\"US survey foot\",0.30480");
    add_srs_wkt (p, 7,
		 "06096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-99],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.9996],PARAMETER[\"false_easting\",1640416.67]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"32");
    add_srs_wkt (p, 12, "074\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32075, "epsg", 32075,
		      "NAD27 / BLM 15N (feet) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-93 +k=0.9996 +x_0=500000.00");
    add_proj4text (p, 1,
		   "1016002 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / BLM 15N (feet) (deprecated)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD27\",DATUM[\"North_American_Datum_1927\",SPHEROID[\"C");
    add_srs_wkt (p, 2,
		 "larke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4267\"]],UNIT[\"US survey foot\",0.30480");
    add_srs_wkt (p, 7,
		 "06096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-93],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.9996],PARAMETER[\"false_easting\",1640416.67]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"32");
    add_srs_wkt (p, 12, "075\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32076, "epsg", 32076,
		      "NAD27 / BLM 16N (feet) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-87 +k=0.9996 +x_0=500000.00");
    add_proj4text (p, 1,
		   "1016002 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / BLM 16N (feet) (deprecated)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD27\",DATUM[\"North_American_Datum_1927\",SPHEROID[\"C");
    add_srs_wkt (p, 2,
		 "larke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4267\"]],UNIT[\"US survey foot\",0.30480");
    add_srs_wkt (p, 7,
		 "06096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-87],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.9996],PARAMETER[\"false_easting\",1640416.67]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"32");
    add_srs_wkt (p, 12, "076\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32077, "epsg", 32077,
		      "NAD27 / BLM 17N (feet) (deprecated)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-81 +k=0.9996 +x_0=500000.00");
    add_proj4text (p, 1,
		   "1016002 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=us-ft +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / BLM 17N (feet) (deprecated)\",GEOGCS[\"");
    add_srs_wkt (p, 1,
		 "NAD27\",DATUM[\"North_American_Datum_1927\",SPHEROID[\"C");
    add_srs_wkt (p, 2,
		 "larke 1866\",6378206.4,294.9786982139006,AUTHORITY[\"EPS");
    add_srs_wkt (p, 3,
		 "G\",\"7008\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Gr");
    add_srs_wkt (p, 4,
		 "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\"");
    add_srs_wkt (p, 5,
		 ",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHO");
    add_srs_wkt (p, 6,
		 "RITY[\"EPSG\",\"4267\"]],UNIT[\"US survey foot\",0.30480");
    add_srs_wkt (p, 7,
		 "06096012192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"T");
    add_srs_wkt (p, 8,
		 "ransverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"central_meridian\",-81],PARAMETER[\"scale_");
    add_srs_wkt (p, 10,
		 "factor\",0.9996],PARAMETER[\"false_easting\",1640416.67]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"32");
    add_srs_wkt (p, 12, "077\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32081, "epsg", 32081, "NAD27 / MTM zone 1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-53 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=clrk66 +datum=NAD27 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / MTM zone 1\",GEOGCS[\"NAD27\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6378");
    add_srs_wkt (p, 2,
		 "206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-53],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9999],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",304800],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 11,
		 "PSG\",\"32081\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 32082, "epsg", 32082, "NAD27 / MTM zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-56 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=clrk66 +datum=NAD27 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / MTM zone 2\",GEOGCS[\"NAD27\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6378");
    add_srs_wkt (p, 2,
		 "206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-56],PARA");
    add_srs_wkt (p, 9,
		 "METER[\"scale_factor\",0.9999],PARAMETER[\"false_easting");
    add_srs_wkt (p, 10,
		 "\",304800],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 11,
		 "PSG\",\"32082\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 32083, "epsg", 32083, "NAD27 / MTM zone 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-58.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=clrk66 +datum=NAD27 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / MTM zone 3\",GEOGCS[\"NAD27\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6378");
    add_srs_wkt (p, 2,
		 "206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-58.5],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",304800],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"32083\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 32084, "epsg", 32084, "NAD27 / MTM zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-61.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=clrk66 +datum=NAD27 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / MTM zone 4\",GEOGCS[\"NAD27\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6378");
    add_srs_wkt (p, 2,
		 "206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-61.5],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",304800],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"32084\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 32085, "epsg", 32085, "NAD27 / MTM zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-64.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=clrk66 +datum=NAD27 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / MTM zone 5\",GEOGCS[\"NAD27\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6378");
    add_srs_wkt (p, 2,
		 "206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-64.5],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",304800],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"32085\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 32086, "epsg", 32086, "NAD27 / MTM zone 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-67.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=clrk66 +datum=NAD27 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / MTM zone 6\",GEOGCS[\"NAD27\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1927\",SPHEROID[\"Clarke 1866\",6378");
    add_srs_wkt (p, 2,
		 "206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]],AU");
    add_srs_wkt (p, 3,
		 "THORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHO");
    add_srs_wkt (p, 4,
		 "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199");
    add_srs_wkt (p, 5,
		 "4328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",0],PARAMETER[\"central_meridian\",-67.5],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easti");
    add_srs_wkt (p, 10,
		 "ng\",304800],PARAMETER[\"false_northing\",0],AUTHORITY[\"");
    add_srs_wkt (p, 11,
		 "EPSG\",\"32086\"],AXIS[\"Easting\",EAST],AXIS[\"Northing");
    add_srs_wkt (p, 12, "\",NORTH]]");
    p = add_epsg_def (first, last, 32098, "epsg", 32098,
		      "NAD27 / Quebec Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=60 +lat_2=46 +lat_0=44 +lon_0=-68.5 +x_");
    add_proj4text (p, 1,
		   "0=0 +y_0=0 +ellps=clrk66 +datum=NAD27 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Quebec Lambert\",GEOGCS[\"NAD27\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1927\",SPHEROID[\"Clarke 1866\",");
    add_srs_wkt (p, 2,
		 "6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"7008\"]");
    add_srs_wkt (p, 3,
		 "],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,A");
    add_srs_wkt (p, 4,
		 "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292");
    add_srs_wkt (p, 5,
		 "51994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"4267\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"standard_parallel_1\",60],PARAMETER[\"standard_parall");
    add_srs_wkt (p, 9,
		 "el_2\",46],PARAMETER[\"latitude_of_origin\",44],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"central_meridian\",-68.5],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 12, "32098\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32099, "epsg", 32099,
		      "NAD27 / Louisiana Offshore");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=27.83333333333333 +lat_2=26.16666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=25.66666666666667 +lon_0=-91.33333333333333 +");
    add_proj4text (p, 2,
		   "x_0=609601.2192024384 +y_0=0 +ellps=clrk66 +datum=NAD27 ");
    add_proj4text (p, 3, "+units=us-ft +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD27 / Louisiana Offshore\",GEOGCS[\"NAD27\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1927\",SPHEROID[\"Clarke 186");
    add_srs_wkt (p, 2,
		 "6\",6378206.4,294.9786982139006,AUTHORITY[\"EPSG\",\"700");
    add_srs_wkt (p, 3,
		 "8\"]],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\"");
    add_srs_wkt (p, 4,
		 ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745");
    add_srs_wkt (p, 5,
		 "329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 6,
		 "PSG\",\"4267\"]],UNIT[\"US survey foot\",0.3048006096012");
    add_srs_wkt (p, 7,
		 "192,AUTHORITY[\"EPSG\",\"9003\"]],PROJECTION[\"Lambert_C");
    add_srs_wkt (p, 8,
		 "onformal_Conic_2SP\"],PARAMETER[\"standard_parallel_1\",");
    add_srs_wkt (p, 9,
		 "27.83333333333333],PARAMETER[\"standard_parallel_2\",26.");
    add_srs_wkt (p, 10,
		 "16666666666667],PARAMETER[\"latitude_of_origin\",25.6666");
    add_srs_wkt (p, 11,
		 "6666666667],PARAMETER[\"central_meridian\",-91.333333333");
    add_srs_wkt (p, 12,
		 "33333],PARAMETER[\"false_easting\",2000000],PARAMETER[\"");
    add_srs_wkt (p, 13,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32099\"],AXIS[\"");
    add_srs_wkt (p, 14, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32100, "epsg", 32100, "NAD83 / Montana");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=49 +lat_2=45 +lat_0=44.25 +lon_0=-109.5");
    add_proj4text (p, 1,
		   " +x_0=600000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +");
    add_proj4text (p, 2, "no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Montana\",GEOGCS[\"NAD83\",DATUM[\"Nort");
    add_srs_wkt (p, 1,
		 "h_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137,29");
    add_srs_wkt (p, 2,
		 "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_para");
    add_srs_wkt (p, 8,
		 "llel_1\",49],PARAMETER[\"standard_parallel_2\",45],PARAM");
    add_srs_wkt (p, 9,
		 "ETER[\"latitude_of_origin\",44.25],PARAMETER[\"central_m");
    add_srs_wkt (p, 10,
		 "eridian\",-109.5],PARAMETER[\"false_easting\",600000],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"32100");
    add_srs_wkt (p, 12, "\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32104, "epsg", 32104, "NAD83 / Nebraska");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=43 +lat_2=40 +lat_0=39.83333333333334 +");
    add_proj4text (p, 1,
		   "lon_0=-100 +x_0=500000 +y_0=0 +ellps=GRS80 +datum=NAD83 ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Nebraska\",GEOGCS[\"NAD83\",DATUM[\"Nor");
    add_srs_wkt (p, 1,
		 "th_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137,2");
    add_srs_wkt (p, 2,
		 "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 8,
		 "rallel_1\",43],PARAMETER[\"standard_parallel_2\",40],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"latitude_of_origin\",39.83333333333334],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"central_meridian\",-100],PARAMETER[\"false_easting\"");
    add_srs_wkt (p, 11,
		 ",500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPS");
    add_srs_wkt (p, 12, "G\",\"32104\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32107, "epsg", 32107, "NAD83 / Nevada East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-115.5833333333333 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=200000 +y_0=8000000 +ellps=GRS80 +datum=NAD83 ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Nevada East\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",34.75],PARAMETER[\"central_meridian\",-115.58333333");
    add_srs_wkt (p, 9,
		 "33333],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_easting\",200000],PARAMETER[\"false_northing\",80000");
    add_srs_wkt (p, 11,
		 "00],AUTHORITY[\"EPSG\",\"32107\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 12, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32108, "epsg", 32108,
		      "NAD83 / Nevada Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-116.6666666666667 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=500000 +y_0=6000000 +ellps=GRS80 +datum=NAD83 ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Nevada Central\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 8,
		 "igin\",34.75],PARAMETER[\"central_meridian\",-116.666666");
    add_srs_wkt (p, 9,
		 "6666667],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",500000],PARAMETER[\"false_northing\",600");
    add_srs_wkt (p, 11,
		 "0000],AUTHORITY[\"EPSG\",\"32108\"],AXIS[\"X\",EAST],AXI");
    add_srs_wkt (p, 12, "S[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32109, "epsg", 32109, "NAD83 / Nevada West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=34.75 +lon_0=-118.5833333333333 +k=0.");
    add_proj4text (p, 1,
		   "9999 +x_0=800000 +y_0=4000000 +ellps=GRS80 +datum=NAD83 ");
    add_proj4text (p, 2, "+units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Nevada West\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",34.75],PARAMETER[\"central_meridian\",-118.58333333");
    add_srs_wkt (p, 9,
		 "33333],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_easting\",800000],PARAMETER[\"false_northing\",40000");
    add_srs_wkt (p, 11,
		 "00],AUTHORITY[\"EPSG\",\"32109\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 12, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32110, "epsg", 32110,
		      "NAD83 / New Hampshire");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.5 +lon_0=-71.66666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "99966667 +x_0=300000 +y_0=0 +ellps=GRS80 +datum=NAD83 +u");
    add_proj4text (p, 2, "nits=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / New Hampshire\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",42.5],PARAMETER[\"central_meridian\",-71.666666666");
    add_srs_wkt (p, 9,
		 "66667],PARAMETER[\"scale_factor\",0.999966667],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",300000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"32110\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 32111, "epsg", 32111, "NAD83 / New Jersey");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38.83333333333334 +lon_0=-74.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=150000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / New Jersey\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",38.83333333333334],PARAMETER[\"central_meridian\",-74");
    add_srs_wkt (p, 9,
		 ".5],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_easting\",150000],PARAMETER[\"false_northing\",0],AUTHO");
    add_srs_wkt (p, 11,
		 "RITY[\"EPSG\",\"32111\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 12, "TH]]");
    p = add_epsg_def (first, last, 32112, "epsg", 32112,
		      "NAD83 / New Mexico East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-104.3333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "909091 +x_0=165000 +y_0=0 +ellps=GRS80 +datum=NAD83 +uni");
    add_proj4text (p, 2, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / New Mexico East\",GEOGCS[\"NAD83\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",63");
    add_srs_wkt (p, 2,
		 "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",31],PARAMETER[\"central_meridian\",-104.3333333");
    add_srs_wkt (p, 9,
		 "333333],PARAMETER[\"scale_factor\",0.999909091],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_easting\",165000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"32112\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 12, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32113, "epsg", 32113,
		      "NAD83 / New Mexico Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-106.25 +k=0.9999 +x_0=5000");
    add_proj4text (p, 1,
		   "00 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / New Mexico Central\",GEOGCS[\"NAD83\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_");
    add_srs_wkt (p, 8,
		 "of_origin\",31],PARAMETER[\"central_meridian\",-106.25],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_eas");
    add_srs_wkt (p, 10,
		 "ting\",500000],PARAMETER[\"false_northing\",0],AUTHORITY");
    add_srs_wkt (p, 11,
		 "[\"EPSG\",\"32113\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    add_srs_wkt (p, 12, "");
    p = add_epsg_def (first, last, 32114, "epsg", 32114,
		      "NAD83 / New Mexico West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=31 +lon_0=-107.8333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "916667 +x_0=830000 +y_0=0 +ellps=GRS80 +datum=NAD83 +uni");
    add_proj4text (p, 2, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / New Mexico West\",GEOGCS[\"NAD83\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",63");
    add_srs_wkt (p, 2,
		 "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 8,
		 "origin\",31],PARAMETER[\"central_meridian\",-107.8333333");
    add_srs_wkt (p, 9,
		 "333333],PARAMETER[\"scale_factor\",0.999916667],PARAMETE");
    add_srs_wkt (p, 10,
		 "R[\"false_easting\",830000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 11,
		 ",0],AUTHORITY[\"EPSG\",\"32114\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 12, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32115, "epsg", 32115,
		      "NAD83 / New York East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=38.83333333333334 +lon_0=-74.5 +k=0.9");
    add_proj4text (p, 1,
		   "999 +x_0=150000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / New York East\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",38.83333333333334],PARAMETER[\"central_meridian\",");
    add_srs_wkt (p, 9,
		 "-74.5],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_easting\",150000],PARAMETER[\"false_northing\",0],AU");
    add_srs_wkt (p, 11,
		 "THORITY[\"EPSG\",\"32115\"],AXIS[\"X\",EAST],AXIS[\"Y\",");
    add_srs_wkt (p, 12, "NORTH]]");
    p = add_epsg_def (first, last, 32116, "epsg", 32116,
		      "NAD83 / New York Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40 +lon_0=-76.58333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "9375 +x_0=250000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units");
    add_proj4text (p, 2, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / New York Central\",GEOGCS[\"NAD83\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6");
    add_srs_wkt (p, 2,
		 "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"426");
    add_srs_wkt (p, 6,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of");
    add_srs_wkt (p, 8,
		 "_origin\",40],PARAMETER[\"central_meridian\",-76.5833333");
    add_srs_wkt (p, 9,
		 "3333333],PARAMETER[\"scale_factor\",0.9999375],PARAMETER");
    add_srs_wkt (p, 10,
		 "[\"false_easting\",250000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 11,
		 "0],AUTHORITY[\"EPSG\",\"32116\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 12, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 32117, "epsg", 32117,
		      "NAD83 / New York West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40 +lon_0=-78.58333333333333 +k=0.999");
    add_proj4text (p, 1,
		   "9375 +x_0=350000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units");
    add_proj4text (p, 2, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / New York West\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_ori");
    add_srs_wkt (p, 8,
		 "gin\",40],PARAMETER[\"central_meridian\",-78.58333333333");
    add_srs_wkt (p, 9,
		 "333],PARAMETER[\"scale_factor\",0.9999375],PARAMETER[\"f");
    add_srs_wkt (p, 10,
		 "alse_easting\",350000],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 11,
		 "UTHORITY[\"EPSG\",\"32117\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 32118, "epsg", 32118,
		      "NAD83 / New York Long Island");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.03333333333333 +lat_2=40.66666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=40.16666666666666 +lon_0=-74 +x_0=300000 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / New York Long Island\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "standard_parallel_1\",41.03333333333333],PARAMETER[\"sta");
    add_srs_wkt (p, 9,
		 "ndard_parallel_2\",40.66666666666666],PARAMETER[\"latitu");
    add_srs_wkt (p, 10,
		 "de_of_origin\",40.16666666666666],PARAMETER[\"central_me");
    add_srs_wkt (p, 11,
		 "ridian\",-74],PARAMETER[\"false_easting\",300000],PARAME");
    add_srs_wkt (p, 12,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"32118\"],");
    add_srs_wkt (p, 13, "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32119, "epsg", 32119,
		      "NAD83 / North Carolina");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.16666666666666 +lat_2=34.33333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=33.75 +lon_0=-79 +x_0=609601.22 +y_0=0 +ellps");
    add_proj4text (p, 2, "=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / North Carolina\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 8,
		 "rd_parallel_1\",36.16666666666666],PARAMETER[\"standard_");
    add_srs_wkt (p, 9,
		 "parallel_2\",34.33333333333334],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 10,
		 "origin\",33.75],PARAMETER[\"central_meridian\",-79],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_easting\",609601.22],PARAMETER[\"false_nor");
    add_srs_wkt (p, 12,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32119\"],AXIS[\"X\",EAST");
    add_srs_wkt (p, 13, "],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32120, "epsg", 32120,
		      "NAD83 / North Dakota North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.73333333333333 +lat_2=47.43333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=47 +lon_0=-100.5 +x_0=600000 +y_0=0 +ellps=GR");
    add_proj4text (p, 2, "S80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / North Dakota North\",GEOGCS[\"NAD83\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 8,
		 "tandard_parallel_1\",48.73333333333333],PARAMETER[\"stan");
    add_srs_wkt (p, 9,
		 "dard_parallel_2\",47.43333333333333],PARAMETER[\"latitud");
    add_srs_wkt (p, 10,
		 "e_of_origin\",47],PARAMETER[\"central_meridian\",-100.5]");
    add_srs_wkt (p, 11,
		 ",PARAMETER[\"false_easting\",600000],PARAMETER[\"false_n");
    add_srs_wkt (p, 12,
		 "orthing\",0],AUTHORITY[\"EPSG\",\"32120\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 13, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32121, "epsg", 32121,
		      "NAD83 / North Dakota South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.48333333333333 +lat_2=46.18333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=45.66666666666666 +lon_0=-100.5 +x_0=600000 +");
    add_proj4text (p, 2, "y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / North Dakota South\",GEOGCS[\"NAD83\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 8,
		 "tandard_parallel_1\",47.48333333333333],PARAMETER[\"stan");
    add_srs_wkt (p, 9,
		 "dard_parallel_2\",46.18333333333333],PARAMETER[\"latitud");
    add_srs_wkt (p, 10,
		 "e_of_origin\",45.66666666666666],PARAMETER[\"central_mer");
    add_srs_wkt (p, 11,
		 "idian\",-100.5],PARAMETER[\"false_easting\",600000],PARA");
    add_srs_wkt (p, 12,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"32121\"");
    add_srs_wkt (p, 13, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32122, "epsg", 32122, "NAD83 / Ohio North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.7 +lat_2=40.43333333333333 +lat_0=39");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-82.5 +x_0=600000 +y_0=0 +ellps=G");
    add_proj4text (p, 2, "RS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Ohio North\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_p");
    add_srs_wkt (p, 8,
		 "arallel_1\",41.7],PARAMETER[\"standard_parallel_2\",40.4");
    add_srs_wkt (p, 9,
		 "3333333333333],PARAMETER[\"latitude_of_origin\",39.66666");
    add_srs_wkt (p, 10,
		 "666666666],PARAMETER[\"central_meridian\",-82.5],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_easting\",600000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"32122\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 13, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32123, "epsg", 32123, "NAD83 / Ohio South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.03333333333333 +lat_2=38.73333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=38 +lon_0=-82.5 +x_0=600000 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Ohio South\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_p");
    add_srs_wkt (p, 8,
		 "arallel_1\",40.03333333333333],PARAMETER[\"standard_para");
    add_srs_wkt (p, 9,
		 "llel_2\",38.73333333333333],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",38],PARAMETER[\"central_meridian\",-82.5],PARAMETER");
    add_srs_wkt (p, 11,
		 "[\"false_easting\",600000],PARAMETER[\"false_northing\",");
    add_srs_wkt (p, 12,
		 "0],AUTHORITY[\"EPSG\",\"32123\"],AXIS[\"X\",EAST],AXIS[\"");
    add_srs_wkt (p, 13, "Y\",NORTH]]");
    p = add_epsg_def (first, last, 32124, "epsg", 32124,
		      "NAD83 / Oklahoma North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.76666666666667 +lat_2=35.56666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=35 +lon_0=-98 +x_0=600000 +y_0=0 +ellps=GRS80");
    add_proj4text (p, 2, " +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Oklahoma North\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 8,
		 "rd_parallel_1\",36.76666666666667],PARAMETER[\"standard_");
    add_srs_wkt (p, 9,
		 "parallel_2\",35.56666666666667],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 10,
		 "origin\",35],PARAMETER[\"central_meridian\",-98],PARAMET");
    add_srs_wkt (p, 11,
		 "ER[\"false_easting\",600000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"32124\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 13, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32125, "epsg", 32125,
		      "NAD83 / Oklahoma South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=35.23333333333333 +lat_2=33.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=33.33333333333334 +lon_0=-98 +x_0=600000 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Oklahoma South\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 8,
		 "rd_parallel_1\",35.23333333333333],PARAMETER[\"standard_");
    add_srs_wkt (p, 9,
		 "parallel_2\",33.93333333333333],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 10,
		 "origin\",33.33333333333334],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 11,
		 "\",-98],PARAMETER[\"false_easting\",600000],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32125\"],AXIS[\"");
    add_srs_wkt (p, 13, "X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32126, "epsg", 32126,
		      "NAD83 / Oregon North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46 +lat_2=44.33333333333334 +lat_0=43.6");
    add_proj4text (p, 1,
		   "6666666666666 +lon_0=-120.5 +x_0=2500000 +y_0=0 +ellps=G");
    add_proj4text (p, 2, "RS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Oregon North\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard");
    add_srs_wkt (p, 8,
		 "_parallel_1\",46],PARAMETER[\"standard_parallel_2\",44.3");
    add_srs_wkt (p, 9,
		 "3333333333334],PARAMETER[\"latitude_of_origin\",43.66666");
    add_srs_wkt (p, 10,
		 "666666666],PARAMETER[\"central_meridian\",-120.5],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_easting\",2500000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",0],AUTHORITY[\"EPSG\",\"32126\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 13, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32127, "epsg", 32127,
		      "NAD83 / Oregon South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44 +lat_2=42.33333333333334 +lat_0=41.6");
    add_proj4text (p, 1,
		   "6666666666666 +lon_0=-120.5 +x_0=1500000 +y_0=0 +ellps=G");
    add_proj4text (p, 2, "RS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Oregon South\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard");
    add_srs_wkt (p, 8,
		 "_parallel_1\",44],PARAMETER[\"standard_parallel_2\",42.3");
    add_srs_wkt (p, 9,
		 "3333333333334],PARAMETER[\"latitude_of_origin\",41.66666");
    add_srs_wkt (p, 10,
		 "666666666],PARAMETER[\"central_meridian\",-120.5],PARAME");
    add_srs_wkt (p, 11,
		 "TER[\"false_easting\",1500000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",0],AUTHORITY[\"EPSG\",\"32127\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 13, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32128, "epsg", 32128,
		      "NAD83 / Pennsylvania North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.95 +lat_2=40.88333333333333 +lat_0=4");
    add_proj4text (p, 1,
		   "0.16666666666666 +lon_0=-77.75 +x_0=600000 +y_0=0 +ellps");
    add_proj4text (p, 2, "=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Pennsylvania North\",GEOGCS[\"NAD83\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 8,
		 "tandard_parallel_1\",41.95],PARAMETER[\"standard_paralle");
    add_srs_wkt (p, 9,
		 "l_2\",40.88333333333333],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 10,
		 ",40.16666666666666],PARAMETER[\"central_meridian\",-77.7");
    add_srs_wkt (p, 11,
		 "5],PARAMETER[\"false_easting\",600000],PARAMETER[\"false");
    add_srs_wkt (p, 12,
		 "_northing\",0],AUTHORITY[\"EPSG\",\"32128\"],AXIS[\"X\",");
    add_srs_wkt (p, 13, "EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32129, "epsg", 32129,
		      "NAD83 / Pennsylvania South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.96666666666667 +lat_2=39.93333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=39.33333333333334 +lon_0=-77.75 +x_0=600000 +");
    add_proj4text (p, 2, "y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Pennsylvania South\",GEOGCS[\"NAD83\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 8,
		 "tandard_parallel_1\",40.96666666666667],PARAMETER[\"stan");
    add_srs_wkt (p, 9,
		 "dard_parallel_2\",39.93333333333333],PARAMETER[\"latitud");
    add_srs_wkt (p, 10,
		 "e_of_origin\",39.33333333333334],PARAMETER[\"central_mer");
    add_srs_wkt (p, 11,
		 "idian\",-77.75],PARAMETER[\"false_easting\",600000],PARA");
    add_srs_wkt (p, 12,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"32129\"");
    add_srs_wkt (p, 13, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32130, "epsg", 32130,
		      "NAD83 / Rhode Island");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=41.08333333333334 +lon_0=-71.5 +k=0.9");
    add_proj4text (p, 1,
		   "9999375 +x_0=100000 +y_0=0 +ellps=GRS80 +datum=NAD83 +un");
    add_proj4text (p, 2, "its=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Rhode Island\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",41.08333333333334],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 9,
		 "71.5],PARAMETER[\"scale_factor\",0.99999375],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",100000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"32130\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 32133, "epsg", 32133,
		      "NAD83 / South Carolina");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=34.83333333333334 +lat_2=32.5 +lat_0=31");
    add_proj4text (p, 1,
		   ".83333333333333 +lon_0=-81 +x_0=609600 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / South Carolina\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 8,
		 "rd_parallel_1\",34.83333333333334],PARAMETER[\"standard_");
    add_srs_wkt (p, 9,
		 "parallel_2\",32.5],PARAMETER[\"latitude_of_origin\",31.8");
    add_srs_wkt (p, 10,
		 "3333333333333],PARAMETER[\"central_meridian\",-81],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_easting\",609600],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",0],AUTHORITY[\"EPSG\",\"32133\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 13, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32134, "epsg", 32134,
		      "NAD83 / South Dakota North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.68333333333333 +lat_2=44.41666666666");
    add_proj4text (p, 1,
		   "666 +lat_0=43.83333333333334 +lon_0=-100 +x_0=600000 +y_");
    add_proj4text (p, 2, "0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / South Dakota North\",GEOGCS[\"NAD83\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 8,
		 "tandard_parallel_1\",45.68333333333333],PARAMETER[\"stan");
    add_srs_wkt (p, 9,
		 "dard_parallel_2\",44.41666666666666],PARAMETER[\"latitud");
    add_srs_wkt (p, 10,
		 "e_of_origin\",43.83333333333334],PARAMETER[\"central_mer");
    add_srs_wkt (p, 11,
		 "idian\",-100],PARAMETER[\"false_easting\",600000],PARAME");
    add_srs_wkt (p, 12,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"32134\"],");
    add_srs_wkt (p, 13, "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32135, "epsg", 32135,
		      "NAD83 / South Dakota South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.4 +lat_2=42.83333333333334 +lat_0=42");
    add_proj4text (p, 1,
		   ".33333333333334 +lon_0=-100.3333333333333 +x_0=600000 +y");
    add_proj4text (p, 2, "_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / South Dakota South\",GEOGCS[\"NAD83\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 8,
		 "tandard_parallel_1\",44.4],PARAMETER[\"standard_parallel");
    add_srs_wkt (p, 9,
		 "_2\",42.83333333333334],PARAMETER[\"latitude_of_origin\"");
    add_srs_wkt (p, 10,
		 ",42.33333333333334],PARAMETER[\"central_meridian\",-100.");
    add_srs_wkt (p, 11,
		 "3333333333333],PARAMETER[\"false_easting\",600000],PARAM");
    add_srs_wkt (p, 12,
		 "ETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"32135\"]");
    add_srs_wkt (p, 13, ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32136, "epsg", 32136, "NAD83 / Tennessee");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.41666666666666 +lat_2=35.25 +lat_0=3");
    add_proj4text (p, 1,
		   "4.33333333333334 +lon_0=-86 +x_0=600000 +y_0=0 +ellps=GR");
    add_proj4text (p, 2, "S80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Tennessee\",GEOGCS[\"NAD83\",DATUM[\"No");
    add_srs_wkt (p, 1,
		 "rth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137,");
    add_srs_wkt (p, 2,
		 "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_p");
    add_srs_wkt (p, 8,
		 "arallel_1\",36.41666666666666],PARAMETER[\"standard_para");
    add_srs_wkt (p, 9,
		 "llel_2\",35.25],PARAMETER[\"latitude_of_origin\",34.3333");
    add_srs_wkt (p, 10,
		 "3333333334],PARAMETER[\"central_meridian\",-86],PARAMETE");
    add_srs_wkt (p, 11,
		 "R[\"false_easting\",600000],PARAMETER[\"false_northing\"");
    add_srs_wkt (p, 12,
		 ",0],AUTHORITY[\"EPSG\",\"32136\"],AXIS[\"X\",EAST],AXIS[");
    add_srs_wkt (p, 13, "\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32137, "epsg", 32137, "NAD83 / Texas North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=36.18333333333333 +lat_2=34.65 +lat_0=3");
    add_proj4text (p, 1,
		   "4 +lon_0=-101.5 +x_0=200000 +y_0=1000000 +ellps=GRS80 +d");
    add_proj4text (p, 2, "atum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Texas North\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard");
    add_srs_wkt (p, 8,
		 "_parallel_1\",36.18333333333333],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 9,
		 "rallel_2\",34.65],PARAMETER[\"latitude_of_origin\",34],P");
    add_srs_wkt (p, 10,
		 "ARAMETER[\"central_meridian\",-101.5],PARAMETER[\"false_");
    add_srs_wkt (p, 11,
		 "easting\",200000],PARAMETER[\"false_northing\",1000000],");
    add_srs_wkt (p, 12,
		 "AUTHORITY[\"EPSG\",\"32137\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 32138, "epsg", 32138,
		      "NAD83 / Texas North Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=33.96666666666667 +lat_2=32.13333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=31.66666666666667 +lon_0=-98.5 +x_0=600000 +y");
    add_proj4text (p, 2,
		   "_0=2000000 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Texas North Central\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 8,
		 "tandard_parallel_1\",33.96666666666667],PARAMETER[\"stan");
    add_srs_wkt (p, 9,
		 "dard_parallel_2\",32.13333333333333],PARAMETER[\"latitud");
    add_srs_wkt (p, 10,
		 "e_of_origin\",31.66666666666667],PARAMETER[\"central_mer");
    add_srs_wkt (p, 11,
		 "idian\",-98.5],PARAMETER[\"false_easting\",600000],PARAM");
    add_srs_wkt (p, 12,
		 "ETER[\"false_northing\",2000000],AUTHORITY[\"EPSG\",\"32");
    add_srs_wkt (p, 13, "138\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32139, "epsg", 32139,
		      "NAD83 / Texas Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=31.88333333333333 +lat_2=30.11666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=29.66666666666667 +lon_0=-100.3333333333333 +");
    add_proj4text (p, 2,
		   "x_0=700000 +y_0=3000000 +ellps=GRS80 +datum=NAD83 +units");
    add_proj4text (p, 3, "=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Texas Central\",GEOGCS[\"NAD83\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378");
    add_srs_wkt (p, 2,
		 "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT");
    add_srs_wkt (p, 3,
		 "Y[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"");
    add_srs_wkt (p, 4,
		 "EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AU");
    add_srs_wkt (p, 5,
		 "THORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]");
    add_srs_wkt (p, 6,
		 "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECT");
    add_srs_wkt (p, 7,
		 "ION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standar");
    add_srs_wkt (p, 8,
		 "d_parallel_1\",31.88333333333333],PARAMETER[\"standard_p");
    add_srs_wkt (p, 9,
		 "arallel_2\",30.11666666666667],PARAMETER[\"latitude_of_o");
    add_srs_wkt (p, 10,
		 "rigin\",29.66666666666667],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 11,
		 ",-100.3333333333333],PARAMETER[\"false_easting\",700000]");
    add_srs_wkt (p, 12,
		 ",PARAMETER[\"false_northing\",3000000],AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 13, ",\"32139\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32140, "epsg", 32140,
		      "NAD83 / Texas South Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=30.28333333333333 +lat_2=28.38333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=27.83333333333333 +lon_0=-99 +x_0=600000 +y_0");
    add_proj4text (p, 2,
		   "=4000000 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Texas South Central\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 8,
		 "tandard_parallel_1\",30.28333333333333],PARAMETER[\"stan");
    add_srs_wkt (p, 9,
		 "dard_parallel_2\",28.38333333333333],PARAMETER[\"latitud");
    add_srs_wkt (p, 10,
		 "e_of_origin\",27.83333333333333],PARAMETER[\"central_mer");
    add_srs_wkt (p, 11,
		 "idian\",-99],PARAMETER[\"false_easting\",600000],PARAMET");
    add_srs_wkt (p, 12,
		 "ER[\"false_northing\",4000000],AUTHORITY[\"EPSG\",\"3214");
    add_srs_wkt (p, 13, "0\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32141, "epsg", 32141, "NAD83 / Texas South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=27.83333333333333 +lat_2=26.16666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=25.66666666666667 +lon_0=-98.5 +x_0=300000 +y");
    add_proj4text (p, 2,
		   "_0=5000000 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Texas South\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard");
    add_srs_wkt (p, 8,
		 "_parallel_1\",27.83333333333333],PARAMETER[\"standard_pa");
    add_srs_wkt (p, 9,
		 "rallel_2\",26.16666666666667],PARAMETER[\"latitude_of_or");
    add_srs_wkt (p, 10,
		 "igin\",25.66666666666667],PARAMETER[\"central_meridian\"");
    add_srs_wkt (p, 11,
		 ",-98.5],PARAMETER[\"false_easting\",300000],PARAMETER[\"");
    add_srs_wkt (p, 12,
		 "false_northing\",5000000],AUTHORITY[\"EPSG\",\"32141\"],");
    add_srs_wkt (p, 13, "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32142, "epsg", 32142, "NAD83 / Utah North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=41.78333333333333 +lat_2=40.71666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=40.33333333333334 +lon_0=-111.5 +x_0=500000 +");
    add_proj4text (p, 2,
		   "y_0=1000000 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Utah North\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_p");
    add_srs_wkt (p, 8,
		 "arallel_1\",41.78333333333333],PARAMETER[\"standard_para");
    add_srs_wkt (p, 9,
		 "llel_2\",40.71666666666667],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 10,
		 "in\",40.33333333333334],PARAMETER[\"central_meridian\",-");
    add_srs_wkt (p, 11,
		 "111.5],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 12,
		 "alse_northing\",1000000],AUTHORITY[\"EPSG\",\"32142\"],A");
    add_srs_wkt (p, 13, "XIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32143, "epsg", 32143,
		      "NAD83 / Utah Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.65 +lat_2=39.01666666666667 +lat_0=3");
    add_proj4text (p, 1,
		   "8.33333333333334 +lon_0=-111.5 +x_0=500000 +y_0=2000000 ");
    add_proj4text (p, 2, "+ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Utah Central\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard");
    add_srs_wkt (p, 8,
		 "_parallel_1\",40.65],PARAMETER[\"standard_parallel_2\",3");
    add_srs_wkt (p, 9,
		 "9.01666666666667],PARAMETER[\"latitude_of_origin\",38.33");
    add_srs_wkt (p, 10,
		 "333333333334],PARAMETER[\"central_meridian\",-111.5],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 12,
		 "ing\",2000000],AUTHORITY[\"EPSG\",\"32143\"],AXIS[\"X\",");
    add_srs_wkt (p, 13, "EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32144, "epsg", 32144, "NAD83 / Utah South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.35 +lat_2=37.21666666666667 +lat_0=3");
    add_proj4text (p, 1,
		   "6.66666666666666 +lon_0=-111.5 +x_0=500000 +y_0=3000000 ");
    add_proj4text (p, 2, "+ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Utah South\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standard_p");
    add_srs_wkt (p, 8,
		 "arallel_1\",38.35],PARAMETER[\"standard_parallel_2\",37.");
    add_srs_wkt (p, 9,
		 "21666666666667],PARAMETER[\"latitude_of_origin\",36.6666");
    add_srs_wkt (p, 10,
		 "6666666666],PARAMETER[\"central_meridian\",-111.5],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_easting\",500000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",3000000],AUTHORITY[\"EPSG\",\"32144\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 13, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32145, "epsg", 32145, "NAD83 / Vermont");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=42.5 +lon_0=-72.5 +k=0.999964286 +x_0");
    add_proj4text (p, 1,
		   "=500000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_de");
    add_proj4text (p, 2, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Vermont\",GEOGCS[\"NAD83\",DATUM[\"Nort");
    add_srs_wkt (p, 1,
		 "h_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137,29");
    add_srs_wkt (p, 2,
		 "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EP");
    add_srs_wkt (p, 3,
		 "SG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 4,
		 ",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORIT");
    add_srs_wkt (p, 5,
		 "Y[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],UNIT");
    add_srs_wkt (p, 6,
		 "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"");
    add_srs_wkt (p, 7,
		 "Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",");
    add_srs_wkt (p, 8,
		 "42.5],PARAMETER[\"central_meridian\",-72.5],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.999964286],PARAMETER[\"false_easting\",");
    add_srs_wkt (p, 10,
		 "500000],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 11, "\",\"32145\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32146, "epsg", 32146,
		      "NAD83 / Virginia North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=39.2 +lat_2=38.03333333333333 +lat_0=37");
    add_proj4text (p, 1,
		   ".66666666666666 +lon_0=-78.5 +x_0=3500000 +y_0=2000000 +");
    add_proj4text (p, 2, "ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Virginia North\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 8,
		 "rd_parallel_1\",39.2],PARAMETER[\"standard_parallel_2\",");
    add_srs_wkt (p, 9,
		 "38.03333333333333],PARAMETER[\"latitude_of_origin\",37.6");
    add_srs_wkt (p, 10,
		 "6666666666666],PARAMETER[\"central_meridian\",-78.5],PAR");
    add_srs_wkt (p, 11,
		 "AMETER[\"false_easting\",3500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 12,
		 "hing\",2000000],AUTHORITY[\"EPSG\",\"32146\"],AXIS[\"X\"");
    add_srs_wkt (p, 13, ",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32147, "epsg", 32147,
		      "NAD83 / Virginia South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=37.96666666666667 +lat_2=36.76666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=36.33333333333334 +lon_0=-78.5 +x_0=3500000 +");
    add_proj4text (p, 2,
		   "y_0=1000000 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Virginia South\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 8,
		 "rd_parallel_1\",37.96666666666667],PARAMETER[\"standard_");
    add_srs_wkt (p, 9,
		 "parallel_2\",36.76666666666667],PARAMETER[\"latitude_of_");
    add_srs_wkt (p, 10,
		 "origin\",36.33333333333334],PARAMETER[\"central_meridian");
    add_srs_wkt (p, 11,
		 "\",-78.5],PARAMETER[\"false_easting\",3500000],PARAMETER");
    add_srs_wkt (p, 12,
		 "[\"false_northing\",1000000],AUTHORITY[\"EPSG\",\"32147\"");
    add_srs_wkt (p, 13, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32148, "epsg", 32148,
		      "NAD83 / Washington North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=48.73333333333333 +lat_2=47.5 +lat_0=47");
    add_proj4text (p, 1,
		   " +lon_0=-120.8333333333333 +x_0=500000 +y_0=0 +ellps=GRS");
    add_proj4text (p, 2, "80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Washington North\",GEOGCS[\"NAD83\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6");
    add_srs_wkt (p, 2,
		 "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"426");
    add_srs_wkt (p, 6,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"sta");
    add_srs_wkt (p, 8,
		 "ndard_parallel_1\",48.73333333333333],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_2\",47.5],PARAMETER[\"latitude_of_origin\",4");
    add_srs_wkt (p, 10,
		 "7],PARAMETER[\"central_meridian\",-120.8333333333333],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 12,
		 "hing\",0],AUTHORITY[\"EPSG\",\"32148\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 13, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32149, "epsg", 32149,
		      "NAD83 / Washington South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=47.33333333333334 +lat_2=45.83333333333");
    add_proj4text (p, 1,
		   "334 +lat_0=45.33333333333334 +lon_0=-120.5 +x_0=500000 +");
    add_proj4text (p, 2, "y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Washington South\",GEOGCS[\"NAD83\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6");
    add_srs_wkt (p, 2,
		 "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHO");
    add_srs_wkt (p, 3,
		 "RITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORIT");
    add_srs_wkt (p, 4,
		 "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432");
    add_srs_wkt (p, 5,
		 "8,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"426");
    add_srs_wkt (p, 6,
		 "9\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PRO");
    add_srs_wkt (p, 7,
		 "JECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"sta");
    add_srs_wkt (p, 8,
		 "ndard_parallel_1\",47.33333333333334],PARAMETER[\"standa");
    add_srs_wkt (p, 9,
		 "rd_parallel_2\",45.83333333333334],PARAMETER[\"latitude_");
    add_srs_wkt (p, 10,
		 "of_origin\",45.33333333333334],PARAMETER[\"central_merid");
    add_srs_wkt (p, 11,
		 "ian\",-120.5],PARAMETER[\"false_easting\",500000],PARAME");
    add_srs_wkt (p, 12,
		 "TER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"32149\"],");
    add_srs_wkt (p, 13, "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32150, "epsg", 32150,
		      "NAD83 / West Virginia North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=40.25 +lat_2=39 +lat_0=38.5 +lon_0=-79.");
    add_proj4text (p, 1,
		   "5 +x_0=600000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / West Virginia North\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 8,
		 "tandard_parallel_1\",40.25],PARAMETER[\"standard_paralle");
    add_srs_wkt (p, 9,
		 "l_2\",39],PARAMETER[\"latitude_of_origin\",38.5],PARAMET");
    add_srs_wkt (p, 10,
		 "ER[\"central_meridian\",-79.5],PARAMETER[\"false_easting");
    add_srs_wkt (p, 11,
		 "\",600000],PARAMETER[\"false_northing\",0],AUTHORITY[\"E");
    add_srs_wkt (p, 12, "PSG\",\"32150\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32151, "epsg", 32151,
		      "NAD83 / West Virginia South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=38.88333333333333 +lat_2=37.48333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=37 +lon_0=-81 +x_0=600000 +y_0=0 +ellps=GRS80");
    add_proj4text (p, 2, " +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / West Virginia South\",GEOGCS[\"NAD83\",");
    add_srs_wkt (p, 1,
		 "DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 8,
		 "tandard_parallel_1\",38.88333333333333],PARAMETER[\"stan");
    add_srs_wkt (p, 9,
		 "dard_parallel_2\",37.48333333333333],PARAMETER[\"latitud");
    add_srs_wkt (p, 10,
		 "e_of_origin\",37],PARAMETER[\"central_meridian\",-81],PA");
    add_srs_wkt (p, 11,
		 "RAMETER[\"false_easting\",600000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 12,
		 "hing\",0],AUTHORITY[\"EPSG\",\"32151\"],AXIS[\"X\",EAST]");
    add_srs_wkt (p, 13, ",AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32152, "epsg", 32152,
		      "NAD83 / Wisconsin North");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=46.76666666666667 +lat_2=45.56666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=45.16666666666666 +lon_0=-90 +x_0=600000 +y_0");
    add_proj4text (p, 2, "=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Wisconsin North\",GEOGCS[\"NAD83\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",63");
    add_srs_wkt (p, 2,
		 "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"stan");
    add_srs_wkt (p, 8,
		 "dard_parallel_1\",46.76666666666667],PARAMETER[\"standar");
    add_srs_wkt (p, 9,
		 "d_parallel_2\",45.56666666666667],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 10,
		 "f_origin\",45.16666666666666],PARAMETER[\"central_meridi");
    add_srs_wkt (p, 11,
		 "an\",-90],PARAMETER[\"false_easting\",600000],PARAMETER[");
    add_srs_wkt (p, 12,
		 "\"false_northing\",0],AUTHORITY[\"EPSG\",\"32152\"],AXIS");
    add_srs_wkt (p, 13, "[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32153, "epsg", 32153,
		      "NAD83 / Wisconsin Central");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=45.5 +lat_2=44.25 +lat_0=43.83333333333");
    add_proj4text (p, 1,
		   "334 +lon_0=-90 +x_0=600000 +y_0=0 +ellps=GRS80 +datum=NA");
    add_proj4text (p, 2, "D83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Wisconsin Central\",GEOGCS[\"NAD83\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",");
    add_srs_wkt (p, 2,
		 "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTH");
    add_srs_wkt (p, 3,
		 "ORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORI");
    add_srs_wkt (p, 4,
		 "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943");
    add_srs_wkt (p, 5,
		 "28,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"42");
    add_srs_wkt (p, 6,
		 "69\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PR");
    add_srs_wkt (p, 7,
		 "OJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"st");
    add_srs_wkt (p, 8,
		 "andard_parallel_1\",45.5],PARAMETER[\"standard_parallel_");
    add_srs_wkt (p, 9,
		 "2\",44.25],PARAMETER[\"latitude_of_origin\",43.833333333");
    add_srs_wkt (p, 10,
		 "33334],PARAMETER[\"central_meridian\",-90],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_easting\",600000],PARAMETER[\"false_northing\",0],A");
    add_srs_wkt (p, 12,
		 "UTHORITY[\"EPSG\",\"32153\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 13, ",NORTH]]");
    p = add_epsg_def (first, last, 32154, "epsg", 32154,
		      "NAD83 / Wisconsin South");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=44.06666666666667 +lat_2=42.73333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=42 +lon_0=-90 +x_0=600000 +y_0=0 +ellps=GRS80");
    add_proj4text (p, 2, " +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Wisconsin South\",GEOGCS[\"NAD83\",DATU");
    add_srs_wkt (p, 1,
		 "M[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",63");
    add_srs_wkt (p, 2,
		 "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHOR");
    add_srs_wkt (p, 3,
		 "ITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY");
    add_srs_wkt (p, 4,
		 "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328");
    add_srs_wkt (p, 5,
		 ",AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269");
    add_srs_wkt (p, 6,
		 "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJ");
    add_srs_wkt (p, 7,
		 "ECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"stan");
    add_srs_wkt (p, 8,
		 "dard_parallel_1\",44.06666666666667],PARAMETER[\"standar");
    add_srs_wkt (p, 9,
		 "d_parallel_2\",42.73333333333333],PARAMETER[\"latitude_o");
    add_srs_wkt (p, 10,
		 "f_origin\",42],PARAMETER[\"central_meridian\",-90],PARAM");
    add_srs_wkt (p, 11,
		 "ETER[\"false_easting\",600000],PARAMETER[\"false_northin");
    add_srs_wkt (p, 12,
		 "g\",0],AUTHORITY[\"EPSG\",\"32154\"],AXIS[\"X\",EAST],AX");
    add_srs_wkt (p, 13, "IS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32155, "epsg", 32155,
		      "NAD83 / Wyoming East");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-105.1666666666667 +k=0.9");
    add_proj4text (p, 1,
		   "999375 +x_0=200000 +y_0=0 +ellps=GRS80 +datum=NAD83 +uni");
    add_proj4text (p, 2, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Wyoming East\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",40.5],PARAMETER[\"central_meridian\",-105.166666666");
    add_srs_wkt (p, 9,
		 "6667],PARAMETER[\"scale_factor\",0.9999375],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",200000],PARAMETER[\"false_northing\",0],");
    add_srs_wkt (p, 11,
		 "AUTHORITY[\"EPSG\",\"32155\"],AXIS[\"X\",EAST],AXIS[\"Y\"");
    add_srs_wkt (p, 12, ",NORTH]]");
    p = add_epsg_def (first, last, 32156, "epsg", 32156,
		      "NAD83 / Wyoming East Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-107.3333333333333 +k=0.9");
    add_proj4text (p, 1,
		   "999375 +x_0=400000 +y_0=100000 +ellps=GRS80 +datum=NAD83");
    add_proj4text (p, 2, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Wyoming East Central\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",40.5],PARAMETER[\"central_meridian\",-107.3");
    add_srs_wkt (p, 9,
		 "333333333333],PARAMETER[\"scale_factor\",0.9999375],PARA");
    add_srs_wkt (p, 10,
		 "METER[\"false_easting\",400000],PARAMETER[\"false_northi");
    add_srs_wkt (p, 11,
		 "ng\",100000],AUTHORITY[\"EPSG\",\"32156\"],AXIS[\"X\",EA");
    add_srs_wkt (p, 12, "ST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32157, "epsg", 32157,
		      "NAD83 / Wyoming West Central");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-108.75 +k=0.9999375 +x_0");
    add_proj4text (p, 1,
		   "=600000 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_de");
    add_proj4text (p, 2, "fs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Wyoming West Central\",GEOGCS[\"NAD83\"");
    add_srs_wkt (p, 1,
		 ",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980");
    add_srs_wkt (p, 2,
		 "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],A");
    add_srs_wkt (p, 3,
		 "UTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTH");
    add_srs_wkt (p, 4,
		 "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519");
    add_srs_wkt (p, 5,
		 "94328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 6,
		 "4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],");
    add_srs_wkt (p, 7,
		 "PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude");
    add_srs_wkt (p, 8,
		 "_of_origin\",40.5],PARAMETER[\"central_meridian\",-108.7");
    add_srs_wkt (p, 9,
		 "5],PARAMETER[\"scale_factor\",0.9999375],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_easting\",600000],PARAMETER[\"false_northing\",0],AUT");
    add_srs_wkt (p, 11,
		 "HORITY[\"EPSG\",\"32157\"],AXIS[\"X\",EAST],AXIS[\"Y\",N");
    add_srs_wkt (p, 12, "ORTH]]");
    p = add_epsg_def (first, last, 32158, "epsg", 32158,
		      "NAD83 / Wyoming West");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=40.5 +lon_0=-110.0833333333333 +k=0.9");
    add_proj4text (p, 1,
		   "999375 +x_0=800000 +y_0=100000 +ellps=GRS80 +datum=NAD83");
    add_proj4text (p, 2, " +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Wyoming West\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",40.5],PARAMETER[\"central_meridian\",-110.083333333");
    add_srs_wkt (p, 9,
		 "3333],PARAMETER[\"scale_factor\",0.9999375],PARAMETER[\"");
    add_srs_wkt (p, 10,
		 "false_easting\",800000],PARAMETER[\"false_northing\",100");
    add_srs_wkt (p, 11,
		 "000],AUTHORITY[\"EPSG\",\"32158\"],AXIS[\"X\",EAST],AXIS");
    add_srs_wkt (p, 12, "[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32161, "epsg", 32161,
		      "NAD83 / Puerto Rico & Virgin Is.");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=18.43333333333333 +lat_2=18.03333333333");
    add_proj4text (p, 1,
		   "333 +lat_0=17.83333333333333 +lon_0=-66.43333333333334 +");
    add_proj4text (p, 2,
		   "x_0=200000 +y_0=200000 +ellps=GRS80 +datum=NAD83 +units=");
    add_proj4text (p, 3, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Puerto Rico & Virgin Is.\",GEOGCS[\"NAD");
    add_srs_wkt (p, 1,
		 "83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS ");
    add_srs_wkt (p, 2,
		 "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"");
    add_srs_wkt (p, 3,
		 "]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,");
    add_srs_wkt (p, 4,
		 "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329");
    add_srs_wkt (p, 5,
		 "251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG");
    add_srs_wkt (p, 6,
		 "\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001");
    add_srs_wkt (p, 7,
		 "\"]],PROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMET");
    add_srs_wkt (p, 8,
		 "ER[\"standard_parallel_1\",18.43333333333333],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"standard_parallel_2\",18.03333333333333],PARAMETER[\"l");
    add_srs_wkt (p, 10,
		 "atitude_of_origin\",17.83333333333333],PARAMETER[\"centr");
    add_srs_wkt (p, 11,
		 "al_meridian\",-66.43333333333334],PARAMETER[\"false_east");
    add_srs_wkt (p, 12,
		 "ing\",200000],PARAMETER[\"false_northing\",200000],AUTHO");
    add_srs_wkt (p, 13,
		 "RITY[\"EPSG\",\"32161\"],AXIS[\"X\",EAST],AXIS[\"Y\",NOR");
    add_srs_wkt (p, 14, "TH]]");
    p = add_epsg_def (first, last, 32164, "epsg", 32164,
		      "NAD83 / BLM 14N (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-99 +k=0.9996 +x_0=500000.00");
    add_proj4text (p, 1,
		   "1016002 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=us-ft +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / BLM 14N (ftUS)\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"US survey foot\",0.3048006096012192,AUTHORITY[");
    add_srs_wkt (p, 7,
		 "\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercator\"],");
    add_srs_wkt (p, 8,
		 "PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_");
    add_srs_wkt (p, 9,
		 "meridian\",-99],PARAMETER[\"scale_factor\",0.9996],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_easting\",1640416.67],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32164\"],AXIS[\"X\",EAST");
    add_srs_wkt (p, 12, "],AXIS[\"Y\",NORTH]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_26 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 32165, "epsg", 32165,
		      "NAD83 / BLM 15N (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-93 +k=0.9996 +x_0=500000.00");
    add_proj4text (p, 1,
		   "1016002 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=us-ft +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / BLM 15N (ftUS)\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"US survey foot\",0.3048006096012192,AUTHORITY[");
    add_srs_wkt (p, 7,
		 "\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercator\"],");
    add_srs_wkt (p, 8,
		 "PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_");
    add_srs_wkt (p, 9,
		 "meridian\",-93],PARAMETER[\"scale_factor\",0.9996],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_easting\",1640416.67],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32165\"],AXIS[\"X\",EAST");
    add_srs_wkt (p, 12, "],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32166, "epsg", 32166,
		      "NAD83 / BLM 16N (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-87 +k=0.9996 +x_0=500000.00");
    add_proj4text (p, 1,
		   "1016002 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=us-ft +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / BLM 16N (ftUS)\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"US survey foot\",0.3048006096012192,AUTHORITY[");
    add_srs_wkt (p, 7,
		 "\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercator\"],");
    add_srs_wkt (p, 8,
		 "PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_");
    add_srs_wkt (p, 9,
		 "meridian\",-87],PARAMETER[\"scale_factor\",0.9996],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_easting\",1640416.67],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32166\"],AXIS[\"X\",EAST");
    add_srs_wkt (p, 12, "],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32167, "epsg", 32167,
		      "NAD83 / BLM 17N (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-81 +k=0.9996 +x_0=500000.00");
    add_proj4text (p, 1,
		   "1016002 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=us-ft +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / BLM 17N (ftUS)\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"US survey foot\",0.3048006096012192,AUTHORITY[");
    add_srs_wkt (p, 7,
		 "\"EPSG\",\"9003\"]],PROJECTION[\"Transverse_Mercator\"],");
    add_srs_wkt (p, 8,
		 "PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_");
    add_srs_wkt (p, 9,
		 "meridian\",-81],PARAMETER[\"scale_factor\",0.9996],PARAM");
    add_srs_wkt (p, 10,
		 "ETER[\"false_easting\",1640416.67],PARAMETER[\"false_nor");
    add_srs_wkt (p, 11,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32167\"],AXIS[\"X\",EAST");
    add_srs_wkt (p, 12, "],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32180, "epsg", 32180,
		      "NAD83 / SCoPQ zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-55.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / SCoPQ zone 2\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-55.5],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"scale_factor\",0.9999],PARAMETER[\"false_easting\",304");
    add_srs_wkt (p, 10,
		 "800],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"32180\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32181, "epsg", 32181, "NAD83 / MTM zone 1");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-53 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / MTM zone 1\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",0],PARAMETER[\"central_meridian\",-53],PARAMETER[\"sc");
    add_srs_wkt (p, 9,
		 "ale_factor\",0.9999],PARAMETER[\"false_easting\",304800]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"32");
    add_srs_wkt (p, 11, "181\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 32182, "epsg", 32182, "NAD83 / MTM zone 2");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-56 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / MTM zone 2\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",0],PARAMETER[\"central_meridian\",-56],PARAMETER[\"sc");
    add_srs_wkt (p, 9,
		 "ale_factor\",0.9999],PARAMETER[\"false_easting\",304800]");
    add_srs_wkt (p, 10,
		 ",PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"32");
    add_srs_wkt (p, 11, "182\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 32183, "epsg", 32183, "NAD83 / MTM zone 3");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-58.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / MTM zone 3\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",0],PARAMETER[\"central_meridian\",-58.5],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",30480");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "32183\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 32184, "epsg", 32184, "NAD83 / MTM zone 4");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-61.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / MTM zone 4\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",0],PARAMETER[\"central_meridian\",-61.5],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",30480");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "32184\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 32185, "epsg", 32185, "NAD83 / MTM zone 5");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-64.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / MTM zone 5\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",0],PARAMETER[\"central_meridian\",-64.5],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",30480");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "32185\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 32186, "epsg", 32186, "NAD83 / MTM zone 6");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-67.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / MTM zone 6\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",0],PARAMETER[\"central_meridian\",-67.5],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",30480");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "32186\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 32187, "epsg", 32187, "NAD83 / MTM zone 7");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-70.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / MTM zone 7\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",0],PARAMETER[\"central_meridian\",-70.5],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",30480");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "32187\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 32188, "epsg", 32188, "NAD83 / MTM zone 8");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-73.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / MTM zone 8\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",0],PARAMETER[\"central_meridian\",-73.5],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",30480");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "32188\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 32189, "epsg", 32189, "NAD83 / MTM zone 9");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-76.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / MTM zone 9\",GEOGCS[\"NAD83\",DATUM[\"N");
    add_srs_wkt (p, 1,
		 "orth_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137");
    add_srs_wkt (p, 2,
		 ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"");
    add_srs_wkt (p, 3,
		 "EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS");
    add_srs_wkt (p, 4,
		 "G\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHO");
    add_srs_wkt (p, 5,
		 "RITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],U");
    add_srs_wkt (p, 6,
		 "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION");
    add_srs_wkt (p, 7,
		 "[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin");
    add_srs_wkt (p, 8,
		 "\",0],PARAMETER[\"central_meridian\",-76.5],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",30480");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11, "32189\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 32190, "epsg", 32190, "NAD83 / MTM zone 10");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-79.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / MTM zone 10\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-79.5],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"scale_factor\",0.9999],PARAMETER[\"false_easting\",304");
    add_srs_wkt (p, 10,
		 "800],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11, "\"32190\"],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH]]");
    p = add_epsg_def (first, last, 32191, "epsg", 32191, "NAD83 / MTM zone 11");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-82.5 +k=0.9999 +x_0=304800 ");
    add_proj4text (p, 1, "+y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / MTM zone 11\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-82.5],PARAMETER[");
    add_srs_wkt (p, 9,
		 "\"scale_factor\",0.9999],PARAMETER[\"false_easting\",304");
    add_srs_wkt (p, 10,
		 "800],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",");
    add_srs_wkt (p, 11,
		 "\"32191\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORT");
    add_srs_wkt (p, 12, "H]]");
    p = add_epsg_def (first, last, 32192, "epsg", 32192, "NAD83 / MTM zone 12");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-81 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / MTM zone 12\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-81],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",30480");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "32192\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 32193, "epsg", 32193, "NAD83 / MTM zone 13");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-84 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / MTM zone 13\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-84],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",30480");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "32193\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 32194, "epsg", 32194, "NAD83 / MTM zone 14");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-87 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / MTM zone 14\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-87],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",30480");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "32194\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 32195, "epsg", 32195, "NAD83 / MTM zone 15");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-90 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / MTM zone 15\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-90],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",30480");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "32195\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 32196, "epsg", 32196, "NAD83 / MTM zone 16");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-93 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / MTM zone 16\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-93],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",30480");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "32196\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 32197, "epsg", 32197, "NAD83 / MTM zone 17");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-96 +k=0.9999 +x_0=304800 +y");
    add_proj4text (p, 1, "_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / MTM zone 17\",GEOGCS[\"NAD83\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637813");
    add_srs_wkt (p, 2,
		 "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[");
    add_srs_wkt (p, 3,
		 "\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E");
    add_srs_wkt (p, 4,
		 "PSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUT");
    add_srs_wkt (p, 5,
		 "HORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]]");
    add_srs_wkt (p, 6,
		 ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTI");
    add_srs_wkt (p, 7,
		 "ON[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_orig");
    add_srs_wkt (p, 8,
		 "in\",0],PARAMETER[\"central_meridian\",-96],PARAMETER[\"");
    add_srs_wkt (p, 9,
		 "scale_factor\",0.9999],PARAMETER[\"false_easting\",30480");
    add_srs_wkt (p, 10,
		 "0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 11,
		 "32197\"],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]");
    add_srs_wkt (p, 12, "]");
    p = add_epsg_def (first, last, 32198, "epsg", 32198,
		      "NAD83 / Quebec Lambert");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=60 +lat_2=46 +lat_0=44 +lon_0=-68.5 +x_");
    add_proj4text (p, 1,
		   "0=0 +y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Quebec Lambert\",GEOGCS[\"NAD83\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",637");
    add_srs_wkt (p, 2,
		 "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI");
    add_srs_wkt (p, 3,
		 "TY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[");
    add_srs_wkt (p, 4,
		 "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,");
    add_srs_wkt (p, 5,
		 "AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"");
    add_srs_wkt (p, 6,
		 "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJEC");
    add_srs_wkt (p, 7,
		 "TION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"standa");
    add_srs_wkt (p, 8,
		 "rd_parallel_1\",60],PARAMETER[\"standard_parallel_2\",46");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"latitude_of_origin\",44],PARAMETER[\"centr");
    add_srs_wkt (p, 10,
		 "al_meridian\",-68.5],PARAMETER[\"false_easting\",0],PARA");
    add_srs_wkt (p, 11,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"32198\"");
    add_srs_wkt (p, 12, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32199, "epsg", 32199,
		      "NAD83 / Louisiana Offshore");
    add_proj4text (p, 0,
		   "+proj=lcc +lat_1=27.83333333333333 +lat_2=26.16666666666");
    add_proj4text (p, 1,
		   "667 +lat_0=25.5 +lon_0=-91.33333333333333 +x_0=1000000 +");
    add_proj4text (p, 2, "y_0=0 +ellps=GRS80 +datum=NAD83 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"NAD83 / Louisiana Offshore\",GEOGCS[\"NAD83\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\"");
    add_srs_wkt (p, 2,
		 ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUT");
    add_srs_wkt (p, 3,
		 "HORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR");
    add_srs_wkt (p, 4,
		 "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994");
    add_srs_wkt (p, 5,
		 "328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4");
    add_srs_wkt (p, 6,
		 "269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],P");
    add_srs_wkt (p, 7,
		 "ROJECTION[\"Lambert_Conformal_Conic_2SP\"],PARAMETER[\"s");
    add_srs_wkt (p, 8,
		 "tandard_parallel_1\",27.83333333333333],PARAMETER[\"stan");
    add_srs_wkt (p, 9,
		 "dard_parallel_2\",26.16666666666667],PARAMETER[\"latitud");
    add_srs_wkt (p, 10,
		 "e_of_origin\",25.5],PARAMETER[\"central_meridian\",-91.3");
    add_srs_wkt (p, 11,
		 "3333333333333],PARAMETER[\"false_easting\",1000000],PARA");
    add_srs_wkt (p, 12,
		 "METER[\"false_northing\",0],AUTHORITY[\"EPSG\",\"32199\"");
    add_srs_wkt (p, 13, "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32201, "epsg", 32201,
		      "WGS 72 / UTM zone 1N");
    add_proj4text (p, 0, "+proj=utm +zone=1 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 1N\",GEOGCS[\"WGS 72\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-177],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"32201\"],AXIS[\"Easting");
    add_srs_wkt (p, 11, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32202, "epsg", 32202,
		      "WGS 72 / UTM zone 2N");
    add_proj4text (p, 0, "+proj=utm +zone=2 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 2N\",GEOGCS[\"WGS 72\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-171],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"32202\"],AXIS[\"Easting");
    add_srs_wkt (p, 11, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32203, "epsg", 32203,
		      "WGS 72 / UTM zone 3N");
    add_proj4text (p, 0, "+proj=utm +zone=3 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 3N\",GEOGCS[\"WGS 72\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-165],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"32203\"],AXIS[\"Easting");
    add_srs_wkt (p, 11, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32204, "epsg", 32204,
		      "WGS 72 / UTM zone 4N");
    add_proj4text (p, 0, "+proj=utm +zone=4 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 4N\",GEOGCS[\"WGS 72\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-159],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"32204\"],AXIS[\"Easting");
    add_srs_wkt (p, 11, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32205, "epsg", 32205,
		      "WGS 72 / UTM zone 5N");
    add_proj4text (p, 0, "+proj=utm +zone=5 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 5N\",GEOGCS[\"WGS 72\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-153],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"32205\"],AXIS[\"Easting");
    add_srs_wkt (p, 11, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32206, "epsg", 32206,
		      "WGS 72 / UTM zone 6N");
    add_proj4text (p, 0, "+proj=utm +zone=6 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 6N\",GEOGCS[\"WGS 72\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-147],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"32206\"],AXIS[\"Easting");
    add_srs_wkt (p, 11, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32207, "epsg", 32207,
		      "WGS 72 / UTM zone 7N");
    add_proj4text (p, 0, "+proj=utm +zone=7 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 7N\",GEOGCS[\"WGS 72\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-141],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"32207\"],AXIS[\"Easting");
    add_srs_wkt (p, 11, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32208, "epsg", 32208,
		      "WGS 72 / UTM zone 8N");
    add_proj4text (p, 0, "+proj=utm +zone=8 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 8N\",GEOGCS[\"WGS 72\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-135],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"32208\"],AXIS[\"Easting");
    add_srs_wkt (p, 11, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32209, "epsg", 32209,
		      "WGS 72 / UTM zone 9N");
    add_proj4text (p, 0, "+proj=utm +zone=9 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 9N\",GEOGCS[\"WGS 72\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-129],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"32209\"],AXIS[\"Easting");
    add_srs_wkt (p, 11, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32210, "epsg", 32210,
		      "WGS 72 / UTM zone 10N");
    add_proj4text (p, 0, "+proj=utm +zone=10 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 10N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-123],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"32210\"],AXIS[\"Easting");
    add_srs_wkt (p, 11, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32211, "epsg", 32211,
		      "WGS 72 / UTM zone 11N");
    add_proj4text (p, 0, "+proj=utm +zone=11 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 11N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-117],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"32211\"],AXIS[\"Easting");
    add_srs_wkt (p, 11, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32212, "epsg", 32212,
		      "WGS 72 / UTM zone 12N");
    add_proj4text (p, 0, "+proj=utm +zone=12 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 12N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-111],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"32212\"],AXIS[\"Easting");
    add_srs_wkt (p, 11, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32213, "epsg", 32213,
		      "WGS 72 / UTM zone 13N");
    add_proj4text (p, 0, "+proj=utm +zone=13 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 13N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-105],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",0],AUTHORITY[\"EPSG\",\"32213\"],AXIS[\"Easting");
    add_srs_wkt (p, 11, "\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32214, "epsg", 32214,
		      "WGS 72 / UTM zone 14N");
    add_proj4text (p, 0, "+proj=utm +zone=14 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 14N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-99],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32214\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32215, "epsg", 32215,
		      "WGS 72 / UTM zone 15N");
    add_proj4text (p, 0, "+proj=utm +zone=15 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 15N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-93],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32215\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32216, "epsg", 32216,
		      "WGS 72 / UTM zone 16N");
    add_proj4text (p, 0, "+proj=utm +zone=16 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 16N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-87],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32216\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32217, "epsg", 32217,
		      "WGS 72 / UTM zone 17N");
    add_proj4text (p, 0, "+proj=utm +zone=17 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 17N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-81],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32217\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32218, "epsg", 32218,
		      "WGS 72 / UTM zone 18N");
    add_proj4text (p, 0, "+proj=utm +zone=18 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 18N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-75],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32218\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32219, "epsg", 32219,
		      "WGS 72 / UTM zone 19N");
    add_proj4text (p, 0, "+proj=utm +zone=19 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 19N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-69],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32219\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32220, "epsg", 32220,
		      "WGS 72 / UTM zone 20N");
    add_proj4text (p, 0, "+proj=utm +zone=20 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 20N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-63],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32220\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32221, "epsg", 32221,
		      "WGS 72 / UTM zone 21N");
    add_proj4text (p, 0, "+proj=utm +zone=21 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 21N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-57],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32221\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32222, "epsg", 32222,
		      "WGS 72 / UTM zone 22N");
    add_proj4text (p, 0, "+proj=utm +zone=22 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 22N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-51],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32222\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32223, "epsg", 32223,
		      "WGS 72 / UTM zone 23N");
    add_proj4text (p, 0, "+proj=utm +zone=23 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 23N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-45],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32223\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32224, "epsg", 32224,
		      "WGS 72 / UTM zone 24N");
    add_proj4text (p, 0, "+proj=utm +zone=24 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 24N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-39],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32224\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32225, "epsg", 32225,
		      "WGS 72 / UTM zone 25N");
    add_proj4text (p, 0, "+proj=utm +zone=25 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 25N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-33],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32225\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32226, "epsg", 32226,
		      "WGS 72 / UTM zone 26N");
    add_proj4text (p, 0, "+proj=utm +zone=26 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 26N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-27],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32226\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32227, "epsg", 32227,
		      "WGS 72 / UTM zone 27N");
    add_proj4text (p, 0, "+proj=utm +zone=27 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 27N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-21],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32227\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32228, "epsg", 32228,
		      "WGS 72 / UTM zone 28N");
    add_proj4text (p, 0, "+proj=utm +zone=28 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 28N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-15],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32228\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32229, "epsg", 32229,
		      "WGS 72 / UTM zone 29N");
    add_proj4text (p, 0, "+proj=utm +zone=29 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 29N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-9],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",0],AUTHORITY[\"EPSG\",\"32229\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32230, "epsg", 32230,
		      "WGS 72 / UTM zone 30N");
    add_proj4text (p, 0, "+proj=utm +zone=30 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 30N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-3],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",0],AUTHORITY[\"EPSG\",\"32230\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32231, "epsg", 32231,
		      "WGS 72 / UTM zone 31N");
    add_proj4text (p, 0, "+proj=utm +zone=31 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 31N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",3],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 10,
		 "ing\",0],AUTHORITY[\"EPSG\",\"32231\"],AXIS[\"Easting\",");
    add_srs_wkt (p, 11, "EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32232, "epsg", 32232,
		      "WGS 72 / UTM zone 32N");
    add_proj4text (p, 0, "+proj=utm +zone=32 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 32N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",9],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 10,
		 "ing\",0],AUTHORITY[\"EPSG\",\"32232\"],AXIS[\"Easting\",");
    add_srs_wkt (p, 11, "EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32233, "epsg", 32233,
		      "WGS 72 / UTM zone 33N");
    add_proj4text (p, 0, "+proj=utm +zone=33 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 33N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",15],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",0],AUTHORITY[\"EPSG\",\"32233\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32234, "epsg", 32234,
		      "WGS 72 / UTM zone 34N");
    add_proj4text (p, 0, "+proj=utm +zone=34 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 34N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",21],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",0],AUTHORITY[\"EPSG\",\"32234\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32235, "epsg", 32235,
		      "WGS 72 / UTM zone 35N");
    add_proj4text (p, 0, "+proj=utm +zone=35 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 35N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",27],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",0],AUTHORITY[\"EPSG\",\"32235\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32236, "epsg", 32236,
		      "WGS 72 / UTM zone 36N");
    add_proj4text (p, 0, "+proj=utm +zone=36 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 36N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",33],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",0],AUTHORITY[\"EPSG\",\"32236\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32237, "epsg", 32237,
		      "WGS 72 / UTM zone 37N");
    add_proj4text (p, 0, "+proj=utm +zone=37 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 37N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",39],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",0],AUTHORITY[\"EPSG\",\"32237\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32238, "epsg", 32238,
		      "WGS 72 / UTM zone 38N");
    add_proj4text (p, 0, "+proj=utm +zone=38 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 38N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",45],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",0],AUTHORITY[\"EPSG\",\"32238\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32239, "epsg", 32239,
		      "WGS 72 / UTM zone 39N");
    add_proj4text (p, 0, "+proj=utm +zone=39 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 39N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",51],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",0],AUTHORITY[\"EPSG\",\"32239\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32240, "epsg", 32240,
		      "WGS 72 / UTM zone 40N");
    add_proj4text (p, 0, "+proj=utm +zone=40 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 40N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",57],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",0],AUTHORITY[\"EPSG\",\"32240\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32241, "epsg", 32241,
		      "WGS 72 / UTM zone 41N");
    add_proj4text (p, 0, "+proj=utm +zone=41 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 41N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",63],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",0],AUTHORITY[\"EPSG\",\"32241\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32242, "epsg", 32242,
		      "WGS 72 / UTM zone 42N");
    add_proj4text (p, 0, "+proj=utm +zone=42 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 42N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",69],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",0],AUTHORITY[\"EPSG\",\"32242\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32243, "epsg", 32243,
		      "WGS 72 / UTM zone 43N");
    add_proj4text (p, 0, "+proj=utm +zone=43 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 43N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",75],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",0],AUTHORITY[\"EPSG\",\"32243\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32244, "epsg", 32244,
		      "WGS 72 / UTM zone 44N");
    add_proj4text (p, 0, "+proj=utm +zone=44 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 44N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",81],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",0],AUTHORITY[\"EPSG\",\"32244\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32245, "epsg", 32245,
		      "WGS 72 / UTM zone 45N");
    add_proj4text (p, 0, "+proj=utm +zone=45 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 45N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",87],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",0],AUTHORITY[\"EPSG\",\"32245\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32246, "epsg", 32246,
		      "WGS 72 / UTM zone 46N");
    add_proj4text (p, 0, "+proj=utm +zone=46 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 46N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",93],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",0],AUTHORITY[\"EPSG\",\"32246\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32247, "epsg", 32247,
		      "WGS 72 / UTM zone 47N");
    add_proj4text (p, 0, "+proj=utm +zone=47 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 47N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",99],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",0],AUTHORITY[\"EPSG\",\"32247\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32248, "epsg", 32248,
		      "WGS 72 / UTM zone 48N");
    add_proj4text (p, 0, "+proj=utm +zone=48 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 48N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",105],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32248\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32249, "epsg", 32249,
		      "WGS 72 / UTM zone 49N");
    add_proj4text (p, 0, "+proj=utm +zone=49 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 49N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",111],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32249\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32250, "epsg", 32250,
		      "WGS 72 / UTM zone 50N");
    add_proj4text (p, 0, "+proj=utm +zone=50 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 50N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",117],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32250\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32251, "epsg", 32251,
		      "WGS 72 / UTM zone 51N");
    add_proj4text (p, 0, "+proj=utm +zone=51 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 51N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",123],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32251\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32252, "epsg", 32252,
		      "WGS 72 / UTM zone 52N");
    add_proj4text (p, 0, "+proj=utm +zone=52 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 52N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",129],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32252\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32253, "epsg", 32253,
		      "WGS 72 / UTM zone 53N");
    add_proj4text (p, 0, "+proj=utm +zone=53 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 53N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",135],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32253\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32254, "epsg", 32254,
		      "WGS 72 / UTM zone 54N");
    add_proj4text (p, 0, "+proj=utm +zone=54 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 54N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",141],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32254\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32255, "epsg", 32255,
		      "WGS 72 / UTM zone 55N");
    add_proj4text (p, 0, "+proj=utm +zone=55 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 55N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",147],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32255\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32256, "epsg", 32256,
		      "WGS 72 / UTM zone 56N");
    add_proj4text (p, 0, "+proj=utm +zone=56 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 56N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",153],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32256\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32257, "epsg", 32257,
		      "WGS 72 / UTM zone 57N");
    add_proj4text (p, 0, "+proj=utm +zone=57 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 57N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",159],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32257\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32258, "epsg", 32258,
		      "WGS 72 / UTM zone 58N");
    add_proj4text (p, 0, "+proj=utm +zone=58 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 58N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",165],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32258\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32259, "epsg", 32259,
		      "WGS 72 / UTM zone 59N");
    add_proj4text (p, 0, "+proj=utm +zone=59 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 59N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",171],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32259\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32260, "epsg", 32260,
		      "WGS 72 / UTM zone 60N");
    add_proj4text (p, 0, "+proj=utm +zone=60 +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 60N\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",177],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",0],AUTHORITY[\"EPSG\",\"32260\"],AXIS[\"Easting\"");
    add_srs_wkt (p, 11, ",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32301, "epsg", 32301,
		      "WGS 72 / UTM zone 1S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=1 +south +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 1S\",GEOGCS[\"WGS 72\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-177],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",10000000],AUTHORITY[\"EPSG\",\"32301\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32302, "epsg", 32302,
		      "WGS 72 / UTM zone 2S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=2 +south +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 2S\",GEOGCS[\"WGS 72\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-171],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",10000000],AUTHORITY[\"EPSG\",\"32302\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32303, "epsg", 32303,
		      "WGS 72 / UTM zone 3S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=3 +south +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 3S\",GEOGCS[\"WGS 72\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-165],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",10000000],AUTHORITY[\"EPSG\",\"32303\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32304, "epsg", 32304,
		      "WGS 72 / UTM zone 4S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=4 +south +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 4S\",GEOGCS[\"WGS 72\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-159],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",10000000],AUTHORITY[\"EPSG\",\"32304\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32305, "epsg", 32305,
		      "WGS 72 / UTM zone 5S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=5 +south +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 5S\",GEOGCS[\"WGS 72\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-153],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",10000000],AUTHORITY[\"EPSG\",\"32305\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32306, "epsg", 32306,
		      "WGS 72 / UTM zone 6S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=6 +south +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 6S\",GEOGCS[\"WGS 72\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-147],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",10000000],AUTHORITY[\"EPSG\",\"32306\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32307, "epsg", 32307,
		      "WGS 72 / UTM zone 7S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=7 +south +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 7S\",GEOGCS[\"WGS 72\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-141],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",10000000],AUTHORITY[\"EPSG\",\"32307\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32308, "epsg", 32308,
		      "WGS 72 / UTM zone 8S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=8 +south +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 8S\",GEOGCS[\"WGS 72\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-135],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",10000000],AUTHORITY[\"EPSG\",\"32308\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32309, "epsg", 32309,
		      "WGS 72 / UTM zone 9S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=9 +south +ellps=WGS72 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 9S\",GEOGCS[\"WGS 72\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORIT");
    add_srs_wkt (p, 2,
		 "Y[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRIM");
    add_srs_wkt (p, 3,
		 "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-129],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",10000000],AUTHORITY[\"EPSG\",\"32309\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32310, "epsg", 32310,
		      "WGS 72 / UTM zone 10S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=10 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 10S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-123],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",10000000],AUTHORITY[\"EPSG\",\"32310\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32311, "epsg", 32311,
		      "WGS 72 / UTM zone 11S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=11 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 11S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-117],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",10000000],AUTHORITY[\"EPSG\",\"32311\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32312, "epsg", 32312,
		      "WGS 72 / UTM zone 12S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=12 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 12S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-111],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",10000000],AUTHORITY[\"EPSG\",\"32312\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32313, "epsg", 32313,
		      "WGS 72 / UTM zone 13S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=13 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 13S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-105],PARAMETER[\"scale_factor\",0.9996],");
    add_srs_wkt (p, 9,
		 "PARAMETER[\"false_easting\",500000],PARAMETER[\"false_no");
    add_srs_wkt (p, 10,
		 "rthing\",10000000],AUTHORITY[\"EPSG\",\"32313\"],AXIS[\"");
    add_srs_wkt (p, 11, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32314, "epsg", 32314,
		      "WGS 72 / UTM zone 14S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=14 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 14S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-99],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32314\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32315, "epsg", 32315,
		      "WGS 72 / UTM zone 15S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=15 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 15S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-93],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32315\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32316, "epsg", 32316,
		      "WGS 72 / UTM zone 16S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=16 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 16S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-87],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32316\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32317, "epsg", 32317,
		      "WGS 72 / UTM zone 17S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=17 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 17S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-81],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32317\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32318, "epsg", 32318,
		      "WGS 72 / UTM zone 18S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 18S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-75],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32318\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32319, "epsg", 32319,
		      "WGS 72 / UTM zone 19S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 19S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-69],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32319\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32320, "epsg", 32320,
		      "WGS 72 / UTM zone 20S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 20S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-63],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32320\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32321, "epsg", 32321,
		      "WGS 72 / UTM zone 21S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 21S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-57],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32321\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32322, "epsg", 32322,
		      "WGS 72 / UTM zone 22S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=22 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 22S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-51],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32322\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32323, "epsg", 32323,
		      "WGS 72 / UTM zone 23S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=23 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 23S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-45],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32323\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32324, "epsg", 32324,
		      "WGS 72 / UTM zone 24S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=24 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 24S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-39],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32324\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32325, "epsg", 32325,
		      "WGS 72 / UTM zone 25S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=25 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 25S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-33],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32325\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32326, "epsg", 32326,
		      "WGS 72 / UTM zone 26S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=26 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 26S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-27],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32326\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32327, "epsg", 32327,
		      "WGS 72 / UTM zone 27S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=27 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 27S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-21],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32327\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32328, "epsg", 32328,
		      "WGS 72 / UTM zone 28S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=28 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 28S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-15],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32328\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32329, "epsg", 32329,
		      "WGS 72 / UTM zone 29S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=29 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 29S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-9],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"32329\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32330, "epsg", 32330,
		      "WGS 72 / UTM zone 30S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=30 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 30S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",-3],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"32330\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32331, "epsg", 32331,
		      "WGS 72 / UTM zone 31S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=31 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 31S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",3],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 10,
		 "ing\",10000000],AUTHORITY[\"EPSG\",\"32331\"],AXIS[\"Eas");
    add_srs_wkt (p, 11, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32332, "epsg", 32332,
		      "WGS 72 / UTM zone 32S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=32 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 32S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",9],PARAMETER[\"scale_factor\",0.9996],PAR");
    add_srs_wkt (p, 9,
		 "AMETER[\"false_easting\",500000],PARAMETER[\"false_north");
    add_srs_wkt (p, 10,
		 "ing\",10000000],AUTHORITY[\"EPSG\",\"32332\"],AXIS[\"Eas");
    add_srs_wkt (p, 11, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32333, "epsg", 32333,
		      "WGS 72 / UTM zone 33S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=33 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 33S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",15],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"32333\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32334, "epsg", 32334,
		      "WGS 72 / UTM zone 34S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=34 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 34S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",21],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"32334\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32335, "epsg", 32335,
		      "WGS 72 / UTM zone 35S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=35 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 35S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",27],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"32335\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32336, "epsg", 32336,
		      "WGS 72 / UTM zone 36S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=36 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 36S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",33],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"32336\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32337, "epsg", 32337,
		      "WGS 72 / UTM zone 37S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=37 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 37S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",39],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"32337\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32338, "epsg", 32338,
		      "WGS 72 / UTM zone 38S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=38 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 38S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",45],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"32338\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32339, "epsg", 32339,
		      "WGS 72 / UTM zone 39S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=39 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 39S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",51],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"32339\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32340, "epsg", 32340,
		      "WGS 72 / UTM zone 40S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=40 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 40S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",57],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"32340\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32341, "epsg", 32341,
		      "WGS 72 / UTM zone 41S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=41 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 41S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",63],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"32341\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32342, "epsg", 32342,
		      "WGS 72 / UTM zone 42S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=42 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 42S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",69],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"32342\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32343, "epsg", 32343,
		      "WGS 72 / UTM zone 43S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=43 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 43S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",75],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"32343\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32344, "epsg", 32344,
		      "WGS 72 / UTM zone 44S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=44 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 44S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",81],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"32344\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32345, "epsg", 32345,
		      "WGS 72 / UTM zone 45S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=45 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 45S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",87],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"32345\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_27 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 32346, "epsg", 32346,
		      "WGS 72 / UTM zone 46S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=46 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 46S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",93],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"32346\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32347, "epsg", 32347,
		      "WGS 72 / UTM zone 47S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=47 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 47S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",99],PARAMETER[\"scale_factor\",0.9996],PA");
    add_srs_wkt (p, 9,
		 "RAMETER[\"false_easting\",500000],PARAMETER[\"false_nort");
    add_srs_wkt (p, 10,
		 "hing\",10000000],AUTHORITY[\"EPSG\",\"32347\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32348, "epsg", 32348,
		      "WGS 72 / UTM zone 48S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=48 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 48S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",105],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32348\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32349, "epsg", 32349,
		      "WGS 72 / UTM zone 49S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=49 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 49S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",111],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32349\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32350, "epsg", 32350,
		      "WGS 72 / UTM zone 50S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=50 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 50S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",117],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32350\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32351, "epsg", 32351,
		      "WGS 72 / UTM zone 51S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=51 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 51S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",123],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32351\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32352, "epsg", 32352,
		      "WGS 72 / UTM zone 52S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=52 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 52S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",129],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32352\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32353, "epsg", 32353,
		      "WGS 72 / UTM zone 53S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=53 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 53S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",135],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32353\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32354, "epsg", 32354,
		      "WGS 72 / UTM zone 54S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=54 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 54S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",141],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32354\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32355, "epsg", 32355,
		      "WGS 72 / UTM zone 55S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=55 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 55S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",147],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32355\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32356, "epsg", 32356,
		      "WGS 72 / UTM zone 56S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=56 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 56S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",153],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32356\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32357, "epsg", 32357,
		      "WGS 72 / UTM zone 57S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=57 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 57S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",159],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32357\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32358, "epsg", 32358,
		      "WGS 72 / UTM zone 58S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 58S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",165],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32358\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32359, "epsg", 32359,
		      "WGS 72 / UTM zone 59S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=59 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 59S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",171],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32359\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32360, "epsg", 32360,
		      "WGS 72 / UTM zone 60S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=60 +south +ellps=WGS72 +units=m +no_defs");
    add_proj4text (p, 1, "");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72 / UTM zone 60S\",GEOGCS[\"WGS 72\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORI");
    add_srs_wkt (p, 2,
		 "TY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6322\"]],PRI");
    add_srs_wkt (p, 3,
		 "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"");
    add_srs_wkt (p, 4,
		 "degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"");
    add_srs_wkt (p, 5,
		 "]],AUTHORITY[\"EPSG\",\"4322\"]],UNIT[\"metre\",1,AUTHOR");
    add_srs_wkt (p, 6,
		 "ITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Mercator");
    add_srs_wkt (p, 7,
		 "\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"cent");
    add_srs_wkt (p, 8,
		 "ral_meridian\",177],PARAMETER[\"scale_factor\",0.9996],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"false_easting\",500000],PARAMETER[\"false_nor");
    add_srs_wkt (p, 10,
		 "thing\",10000000],AUTHORITY[\"EPSG\",\"32360\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32401, "epsg", 32401,
		      "WGS 72BE / UTM zone 1N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=1 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.81");
    add_proj4text (p, 1, "4,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 1N\",GEOGCS[\"WGS 72BE\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-177],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32401\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32402, "epsg", 32402,
		      "WGS 72BE / UTM zone 2N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=2 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.81");
    add_proj4text (p, 1, "4,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 2N\",GEOGCS[\"WGS 72BE\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-171],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32402\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32403, "epsg", 32403,
		      "WGS 72BE / UTM zone 3N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=3 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.81");
    add_proj4text (p, 1, "4,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 3N\",GEOGCS[\"WGS 72BE\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-165],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32403\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32404, "epsg", 32404,
		      "WGS 72BE / UTM zone 4N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=4 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.81");
    add_proj4text (p, 1, "4,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 4N\",GEOGCS[\"WGS 72BE\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-159],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32404\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32405, "epsg", 32405,
		      "WGS 72BE / UTM zone 5N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=5 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.81");
    add_proj4text (p, 1, "4,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 5N\",GEOGCS[\"WGS 72BE\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-153],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32405\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32406, "epsg", 32406,
		      "WGS 72BE / UTM zone 6N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=6 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.81");
    add_proj4text (p, 1, "4,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 6N\",GEOGCS[\"WGS 72BE\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-147],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32406\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32407, "epsg", 32407,
		      "WGS 72BE / UTM zone 7N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=7 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.81");
    add_proj4text (p, 1, "4,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 7N\",GEOGCS[\"WGS 72BE\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-141],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32407\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32408, "epsg", 32408,
		      "WGS 72BE / UTM zone 8N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=8 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.81");
    add_proj4text (p, 1, "4,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 8N\",GEOGCS[\"WGS 72BE\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-135],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32408\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32409, "epsg", 32409,
		      "WGS 72BE / UTM zone 9N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=9 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.81");
    add_proj4text (p, 1, "4,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 9N\",GEOGCS[\"WGS 72BE\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-129],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32409\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32410, "epsg", 32410,
		      "WGS 72BE / UTM zone 10N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=10 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 10N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-123],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32410\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32411, "epsg", 32411,
		      "WGS 72BE / UTM zone 11N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=11 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 11N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-117],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32411\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32412, "epsg", 32412,
		      "WGS 72BE / UTM zone 12N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=12 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 12N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-111],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32412\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32413, "epsg", 32413,
		      "WGS 72BE / UTM zone 13N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=13 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 13N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-105],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32413\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32414, "epsg", 32414,
		      "WGS 72BE / UTM zone 14N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=14 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 14N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-99],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32414\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32415, "epsg", 32415,
		      "WGS 72BE / UTM zone 15N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=15 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 15N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-93],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32415\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32416, "epsg", 32416,
		      "WGS 72BE / UTM zone 16N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=16 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 16N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-87],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32416\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32417, "epsg", 32417,
		      "WGS 72BE / UTM zone 17N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=17 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 17N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-81],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32417\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32418, "epsg", 32418,
		      "WGS 72BE / UTM zone 18N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 18N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-75],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32418\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32419, "epsg", 32419,
		      "WGS 72BE / UTM zone 19N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 19N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-69],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32419\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32420, "epsg", 32420,
		      "WGS 72BE / UTM zone 20N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 20N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-63],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32420\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32421, "epsg", 32421,
		      "WGS 72BE / UTM zone 21N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 21N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-57],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32421\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32422, "epsg", 32422,
		      "WGS 72BE / UTM zone 22N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=22 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 22N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-51],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32422\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32423, "epsg", 32423,
		      "WGS 72BE / UTM zone 23N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=23 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 23N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-45],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32423\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32424, "epsg", 32424,
		      "WGS 72BE / UTM zone 24N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=24 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 24N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-39],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32424\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32425, "epsg", 32425,
		      "WGS 72BE / UTM zone 25N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=25 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 25N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-33],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32425\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32426, "epsg", 32426,
		      "WGS 72BE / UTM zone 26N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=26 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 26N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-27],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32426\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32427, "epsg", 32427,
		      "WGS 72BE / UTM zone 27N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=27 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 27N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-21],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32427\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32428, "epsg", 32428,
		      "WGS 72BE / UTM zone 28N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=28 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 28N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-15],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32428\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32429, "epsg", 32429,
		      "WGS 72BE / UTM zone 29N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=29 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 29N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-9],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32429\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32430, "epsg", 32430,
		      "WGS 72BE / UTM zone 30N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=30 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 30N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-3],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32430\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32431, "epsg", 32431,
		      "WGS 72BE / UTM zone 31N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=31 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 31N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",3],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 10,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"32431\"],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32432, "epsg", 32432,
		      "WGS 72BE / UTM zone 32N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=32 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 32N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",9],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 10,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"32432\"],AXIS[\"E");
    add_srs_wkt (p, 12, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32433, "epsg", 32433,
		      "WGS 72BE / UTM zone 33N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=33 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 33N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",15],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32433\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32434, "epsg", 32434,
		      "WGS 72BE / UTM zone 34N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=34 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 34N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",21],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32434\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32435, "epsg", 32435,
		      "WGS 72BE / UTM zone 35N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=35 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 35N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",27],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32435\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32436, "epsg", 32436,
		      "WGS 72BE / UTM zone 36N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=36 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 36N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",33],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32436\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32437, "epsg", 32437,
		      "WGS 72BE / UTM zone 37N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=37 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 37N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",39],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32437\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32438, "epsg", 32438,
		      "WGS 72BE / UTM zone 38N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=38 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 38N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",45],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32438\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32439, "epsg", 32439,
		      "WGS 72BE / UTM zone 39N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=39 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 39N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",51],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32439\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32440, "epsg", 32440,
		      "WGS 72BE / UTM zone 40N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=40 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 40N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",57],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32440\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32441, "epsg", 32441,
		      "WGS 72BE / UTM zone 41N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=41 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 41N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",63],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32441\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32442, "epsg", 32442,
		      "WGS 72BE / UTM zone 42N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=42 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 42N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",69],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32442\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32443, "epsg", 32443,
		      "WGS 72BE / UTM zone 43N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=43 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 43N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",75],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32443\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32444, "epsg", 32444,
		      "WGS 72BE / UTM zone 44N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=44 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 44N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",81],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32444\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32445, "epsg", 32445,
		      "WGS 72BE / UTM zone 45N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=45 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 45N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",87],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32445\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32446, "epsg", 32446,
		      "WGS 72BE / UTM zone 46N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=46 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 46N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",93],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32446\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32447, "epsg", 32447,
		      "WGS 72BE / UTM zone 47N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=47 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 47N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",99],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",0],AUTHORITY[\"EPSG\",\"32447\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32448, "epsg", 32448,
		      "WGS 72BE / UTM zone 48N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=48 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 48N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",105],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32448\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32449, "epsg", 32449,
		      "WGS 72BE / UTM zone 49N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=49 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 49N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",111],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32449\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32450, "epsg", 32450,
		      "WGS 72BE / UTM zone 50N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=50 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 50N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",117],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32450\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32451, "epsg", 32451,
		      "WGS 72BE / UTM zone 51N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=51 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 51N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",123],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32451\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32452, "epsg", 32452,
		      "WGS 72BE / UTM zone 52N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=52 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 52N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",129],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32452\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32453, "epsg", 32453,
		      "WGS 72BE / UTM zone 53N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=53 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 53N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",135],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32453\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32454, "epsg", 32454,
		      "WGS 72BE / UTM zone 54N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=54 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 54N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",141],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32454\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32455, "epsg", 32455,
		      "WGS 72BE / UTM zone 55N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=55 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 55N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",147],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32455\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32456, "epsg", 32456,
		      "WGS 72BE / UTM zone 56N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=56 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 56N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",153],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32456\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32457, "epsg", 32457,
		      "WGS 72BE / UTM zone 57N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=57 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 57N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",159],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32457\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32458, "epsg", 32458,
		      "WGS 72BE / UTM zone 58N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 58N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",165],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32458\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32459, "epsg", 32459,
		      "WGS 72BE / UTM zone 59N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=59 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 59N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",171],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32459\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32460, "epsg", 32460,
		      "WGS 72BE / UTM zone 60N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=60 +ellps=WGS72 +towgs84=0,0,1.9,0,0,0.8");
    add_proj4text (p, 1, "14,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 60N\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",177],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",0],AUTHORITY[\"EPSG\",\"32460\"],AXIS[\"");
    add_srs_wkt (p, 12, "Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32501, "epsg", 32501,
		      "WGS 72BE / UTM zone 1S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=1 +south +ellps=WGS72 +towgs84=0,0,1.9,0");
    add_proj4text (p, 1, ",0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 1S\",GEOGCS[\"WGS 72BE\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-177],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32501\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32502, "epsg", 32502,
		      "WGS 72BE / UTM zone 2S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=2 +south +ellps=WGS72 +towgs84=0,0,1.9,0");
    add_proj4text (p, 1, ",0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 2S\",GEOGCS[\"WGS 72BE\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-171],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32502\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32503, "epsg", 32503,
		      "WGS 72BE / UTM zone 3S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=3 +south +ellps=WGS72 +towgs84=0,0,1.9,0");
    add_proj4text (p, 1, ",0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 3S\",GEOGCS[\"WGS 72BE\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-165],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32503\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32504, "epsg", 32504,
		      "WGS 72BE / UTM zone 4S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=4 +south +ellps=WGS72 +towgs84=0,0,1.9,0");
    add_proj4text (p, 1, ",0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 4S\",GEOGCS[\"WGS 72BE\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-159],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32504\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32505, "epsg", 32505,
		      "WGS 72BE / UTM zone 5S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=5 +south +ellps=WGS72 +towgs84=0,0,1.9,0");
    add_proj4text (p, 1, ",0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 5S\",GEOGCS[\"WGS 72BE\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-153],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32505\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32506, "epsg", 32506,
		      "WGS 72BE / UTM zone 6S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=6 +south +ellps=WGS72 +towgs84=0,0,1.9,0");
    add_proj4text (p, 1, ",0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 6S\",GEOGCS[\"WGS 72BE\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-147],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32506\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32507, "epsg", 32507,
		      "WGS 72BE / UTM zone 7S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=7 +south +ellps=WGS72 +towgs84=0,0,1.9,0");
    add_proj4text (p, 1, ",0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 7S\",GEOGCS[\"WGS 72BE\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-141],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32507\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32508, "epsg", 32508,
		      "WGS 72BE / UTM zone 8S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=8 +south +ellps=WGS72 +towgs84=0,0,1.9,0");
    add_proj4text (p, 1, ",0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 8S\",GEOGCS[\"WGS 72BE\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-135],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32508\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32509, "epsg", 32509,
		      "WGS 72BE / UTM zone 9S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=9 +south +ellps=WGS72 +towgs84=0,0,1.9,0");
    add_proj4text (p, 1, ",0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 9S\",GEOGCS[\"WGS 72BE\",DA");
    add_srs_wkt (p, 1,
		 "TUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-129],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32509\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32510, "epsg", 32510,
		      "WGS 72BE / UTM zone 10S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=10 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 10S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-123],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32510\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32511, "epsg", 32511,
		      "WGS 72BE / UTM zone 11S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=11 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 11S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-117],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32511\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32512, "epsg", 32512,
		      "WGS 72BE / UTM zone 12S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=12 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 12S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-111],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32512\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32513, "epsg", 32513,
		      "WGS 72BE / UTM zone 13S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=13 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 13S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-105],PARAMETER[\"scale_factor\"");
    add_srs_wkt (p, 10,
		 ",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32513\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32514, "epsg", 32514,
		      "WGS 72BE / UTM zone 14S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=14 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 14S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-99],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32514\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32515, "epsg", 32515,
		      "WGS 72BE / UTM zone 15S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=15 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 15S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-93],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32515\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32516, "epsg", 32516,
		      "WGS 72BE / UTM zone 16S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=16 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 16S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-87],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32516\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32517, "epsg", 32517,
		      "WGS 72BE / UTM zone 17S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=17 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 17S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-81],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32517\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32518, "epsg", 32518,
		      "WGS 72BE / UTM zone 18S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 18S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-75],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32518\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32519, "epsg", 32519,
		      "WGS 72BE / UTM zone 19S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 19S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-69],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32519\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32520, "epsg", 32520,
		      "WGS 72BE / UTM zone 20S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 20S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-63],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32520\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32521, "epsg", 32521,
		      "WGS 72BE / UTM zone 21S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 21S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-57],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32521\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32522, "epsg", 32522,
		      "WGS 72BE / UTM zone 22S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=22 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 22S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-51],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32522\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32523, "epsg", 32523,
		      "WGS 72BE / UTM zone 23S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=23 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 23S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-45],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32523\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32524, "epsg", 32524,
		      "WGS 72BE / UTM zone 24S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=24 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 24S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-39],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32524\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32525, "epsg", 32525,
		      "WGS 72BE / UTM zone 25S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=25 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 25S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-33],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32525\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32526, "epsg", 32526,
		      "WGS 72BE / UTM zone 26S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=26 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 26S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-27],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32526\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32527, "epsg", 32527,
		      "WGS 72BE / UTM zone 27S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=27 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 27S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-21],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32527\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32528, "epsg", 32528,
		      "WGS 72BE / UTM zone 28S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=28 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 28S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-15],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32528\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32529, "epsg", 32529,
		      "WGS 72BE / UTM zone 29S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=29 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 29S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-9],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",10000000],AUTHORITY[\"EPSG\",\"32529\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32530, "epsg", 32530,
		      "WGS 72BE / UTM zone 30S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=30 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 30S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",-3],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",10000000],AUTHORITY[\"EPSG\",\"32530\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32531, "epsg", 32531,
		      "WGS 72BE / UTM zone 31S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=31 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 31S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",3],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 10,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_northing\",10000000],AUTHORITY[\"EPSG\",\"32531\"],A");
    add_srs_wkt (p, 12, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32532, "epsg", 32532,
		      "WGS 72BE / UTM zone 32S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=32 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 32S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",9],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 10,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 11,
		 "lse_northing\",10000000],AUTHORITY[\"EPSG\",\"32532\"],A");
    add_srs_wkt (p, 12, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32533, "epsg", 32533,
		      "WGS 72BE / UTM zone 33S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=33 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 33S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",15],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",10000000],AUTHORITY[\"EPSG\",\"32533\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32534, "epsg", 32534,
		      "WGS 72BE / UTM zone 34S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=34 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 34S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",21],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",10000000],AUTHORITY[\"EPSG\",\"32534\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32535, "epsg", 32535,
		      "WGS 72BE / UTM zone 35S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=35 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 35S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",27],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",10000000],AUTHORITY[\"EPSG\",\"32535\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32536, "epsg", 32536,
		      "WGS 72BE / UTM zone 36S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=36 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 36S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",33],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",10000000],AUTHORITY[\"EPSG\",\"32536\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32537, "epsg", 32537,
		      "WGS 72BE / UTM zone 37S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=37 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 37S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",39],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",10000000],AUTHORITY[\"EPSG\",\"32537\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32538, "epsg", 32538,
		      "WGS 72BE / UTM zone 38S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=38 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 38S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",45],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",10000000],AUTHORITY[\"EPSG\",\"32538\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32539, "epsg", 32539,
		      "WGS 72BE / UTM zone 39S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=39 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 39S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",51],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",10000000],AUTHORITY[\"EPSG\",\"32539\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32540, "epsg", 32540,
		      "WGS 72BE / UTM zone 40S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=40 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 40S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",57],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",10000000],AUTHORITY[\"EPSG\",\"32540\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32541, "epsg", 32541,
		      "WGS 72BE / UTM zone 41S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=41 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 41S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",63],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",10000000],AUTHORITY[\"EPSG\",\"32541\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32542, "epsg", 32542,
		      "WGS 72BE / UTM zone 42S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=42 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 42S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",69],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",10000000],AUTHORITY[\"EPSG\",\"32542\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32543, "epsg", 32543,
		      "WGS 72BE / UTM zone 43S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=43 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 43S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",75],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",10000000],AUTHORITY[\"EPSG\",\"32543\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32544, "epsg", 32544,
		      "WGS 72BE / UTM zone 44S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=44 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 44S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",81],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",10000000],AUTHORITY[\"EPSG\",\"32544\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32545, "epsg", 32545,
		      "WGS 72BE / UTM zone 45S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=45 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 45S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",87],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",10000000],AUTHORITY[\"EPSG\",\"32545\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32546, "epsg", 32546,
		      "WGS 72BE / UTM zone 46S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=46 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 46S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",93],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",10000000],AUTHORITY[\"EPSG\",\"32546\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32547, "epsg", 32547,
		      "WGS 72BE / UTM zone 47S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=47 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 47S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",99],PARAMETER[\"scale_factor\",0");
    add_srs_wkt (p, 10,
		 ".9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"f");
    add_srs_wkt (p, 11,
		 "alse_northing\",10000000],AUTHORITY[\"EPSG\",\"32547\"],");
    add_srs_wkt (p, 12, "AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32548, "epsg", 32548,
		      "WGS 72BE / UTM zone 48S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=48 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 48S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",105],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32548\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32549, "epsg", 32549,
		      "WGS 72BE / UTM zone 49S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=49 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 49S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",111],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32549\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32550, "epsg", 32550,
		      "WGS 72BE / UTM zone 50S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=50 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 50S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",117],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32550\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32551, "epsg", 32551,
		      "WGS 72BE / UTM zone 51S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=51 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 51S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",123],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32551\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32552, "epsg", 32552,
		      "WGS 72BE / UTM zone 52S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=52 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 52S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",129],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32552\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32553, "epsg", 32553,
		      "WGS 72BE / UTM zone 53S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=53 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 53S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",135],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32553\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
#endif /* full EPSG initialization enabled/disabled */
    if (first == last)
	last = first;		/* suppressing stupid compiler warnings */
}

static void
initialize_epsg_28 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
    struct epsg_defs *p;
#if OMIT_EPSG == 0		/* full EPSG initialization enabled */
    p = add_epsg_def (first, last, 32554, "epsg", 32554,
		      "WGS 72BE / UTM zone 54S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=54 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 54S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",141],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32554\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32555, "epsg", 32555,
		      "WGS 72BE / UTM zone 55S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=55 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 55S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",147],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32555\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32556, "epsg", 32556,
		      "WGS 72BE / UTM zone 56S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=56 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 56S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",153],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32556\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32557, "epsg", 32557,
		      "WGS 72BE / UTM zone 57S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=57 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 57S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",159],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32557\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32558, "epsg", 32558,
		      "WGS 72BE / UTM zone 58S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 58S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",165],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32558\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32559, "epsg", 32559,
		      "WGS 72BE / UTM zone 59S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=59 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 59S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",171],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32559\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32560, "epsg", 32560,
		      "WGS 72BE / UTM zone 60S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=60 +south +ellps=WGS72 +towgs84=0,0,1.9,");
    add_proj4text (p, 1, "0,0,0.814,-0.38 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 72BE / UTM zone 60S\",GEOGCS[\"WGS 72BE\",D");
    add_srs_wkt (p, 1,
		 "ATUM[\"WGS_1972_Transit_Broadcast_Ephemeris\",SPHEROID[\"");
    add_srs_wkt (p, 2,
		 "WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TO");
    add_srs_wkt (p, 3,
		 "WGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324");
    add_srs_wkt (p, 4,
		 "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]");
    add_srs_wkt (p, 5,
		 "],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 6,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4324\"]],UNIT[\"metre\"");
    add_srs_wkt (p, 7,
		 ",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse");
    add_srs_wkt (p, 8,
		 "_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMET");
    add_srs_wkt (p, 9,
		 "ER[\"central_meridian\",177],PARAMETER[\"scale_factor\",");
    add_srs_wkt (p, 10,
		 "0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"");
    add_srs_wkt (p, 11,
		 "false_northing\",10000000],AUTHORITY[\"EPSG\",\"32560\"]");
    add_srs_wkt (p, 12, ",AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");

#endif /* unconditionally loading the WGS82/UTM zones */
    p = add_epsg_def (first, last, 32601, "epsg", 32601,
		      "WGS 84 / UTM zone 1N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=1 +ellps=WGS84 +datum=WGS84 +units=m +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 1N\",GEOGCS[\"WGS 84\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-177],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"32601\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32602, "epsg", 32602,
		      "WGS 84 / UTM zone 2N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=2 +ellps=WGS84 +datum=WGS84 +units=m +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 2N\",GEOGCS[\"WGS 84\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-171],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"32602\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32603, "epsg", 32603,
		      "WGS 84 / UTM zone 3N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=3 +ellps=WGS84 +datum=WGS84 +units=m +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 3N\",GEOGCS[\"WGS 84\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-165],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"32603\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32604, "epsg", 32604,
		      "WGS 84 / UTM zone 4N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=4 +ellps=WGS84 +datum=WGS84 +units=m +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 4N\",GEOGCS[\"WGS 84\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-159],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"32604\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32605, "epsg", 32605,
		      "WGS 84 / UTM zone 5N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=5 +ellps=WGS84 +datum=WGS84 +units=m +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 5N\",GEOGCS[\"WGS 84\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-153],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"32605\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32606, "epsg", 32606,
		      "WGS 84 / UTM zone 6N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=6 +ellps=WGS84 +datum=WGS84 +units=m +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 6N\",GEOGCS[\"WGS 84\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-147],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"32606\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32607, "epsg", 32607,
		      "WGS 84 / UTM zone 7N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=7 +ellps=WGS84 +datum=WGS84 +units=m +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 7N\",GEOGCS[\"WGS 84\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-141],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"32607\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32608, "epsg", 32608,
		      "WGS 84 / UTM zone 8N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=8 +ellps=WGS84 +datum=WGS84 +units=m +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 8N\",GEOGCS[\"WGS 84\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-135],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"32608\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32609, "epsg", 32609,
		      "WGS 84 / UTM zone 9N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=9 +ellps=WGS84 +datum=WGS84 +units=m +no");
    add_proj4text (p, 1, "_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 9N\",GEOGCS[\"WGS 84\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-129],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"32609\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32610, "epsg", 32610,
		      "WGS 84 / UTM zone 10N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=10 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 10N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-123],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"32610\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32611, "epsg", 32611,
		      "WGS 84 / UTM zone 11N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=11 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 11N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-117],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"32611\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32612, "epsg", 32612,
		      "WGS 84 / UTM zone 12N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=12 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 12N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-111],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"32612\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32613, "epsg", 32613,
		      "WGS 84 / UTM zone 13N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=13 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 13N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-105],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",0],AUTHORITY[\"EPSG\",\"32613\"],AXIS[\"E");
    add_srs_wkt (p, 11, "asting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32614, "epsg", 32614,
		      "WGS 84 / UTM zone 14N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=14 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 14N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-99],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32614\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32615, "epsg", 32615,
		      "WGS 84 / UTM zone 15N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=15 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 15N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-93],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32615\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32616, "epsg", 32616,
		      "WGS 84 / UTM zone 16N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=16 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 16N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-87],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32616\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32617, "epsg", 32617,
		      "WGS 84 / UTM zone 17N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=17 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 17N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-81],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32617\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32618, "epsg", 32618,
		      "WGS 84 / UTM zone 18N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 18N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-75],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32618\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32619, "epsg", 32619,
		      "WGS 84 / UTM zone 19N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 19N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-69],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32619\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32620, "epsg", 32620,
		      "WGS 84 / UTM zone 20N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 20N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-63],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32620\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32621, "epsg", 32621,
		      "WGS 84 / UTM zone 21N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 21N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-57],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32621\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32622, "epsg", 32622,
		      "WGS 84 / UTM zone 22N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=22 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 22N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-51],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32622\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32623, "epsg", 32623,
		      "WGS 84 / UTM zone 23N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=23 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 23N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-45],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32623\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32624, "epsg", 32624,
		      "WGS 84 / UTM zone 24N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=24 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 24N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-39],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32624\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32625, "epsg", 32625,
		      "WGS 84 / UTM zone 25N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=25 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 25N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-33],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32625\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32626, "epsg", 32626,
		      "WGS 84 / UTM zone 26N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=26 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 26N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-27],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32626\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32627, "epsg", 32627,
		      "WGS 84 / UTM zone 27N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=27 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 27N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-21],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32627\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32628, "epsg", 32628,
		      "WGS 84 / UTM zone 28N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=28 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 28N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-15],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32628\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32629, "epsg", 32629,
		      "WGS 84 / UTM zone 29N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=29 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 29N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-9],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"32629\"],AXIS[\"Eas");
    add_srs_wkt (p, 11, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32630, "epsg", 32630,
		      "WGS 84 / UTM zone 30N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=30 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 30N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-3],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"32630\"],AXIS[\"Eas");
    add_srs_wkt (p, 11, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32631, "epsg", 32631,
		      "WGS 84 / UTM zone 31N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=31 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 31N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",3],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 9,
		 "6],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_northing\",0],AUTHORITY[\"EPSG\",\"32631\"],AXIS[\"East");
    add_srs_wkt (p, 11, "ing\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32632, "epsg", 32632,
		      "WGS 84 / UTM zone 32N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=32 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 32N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",9],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 9,
		 "6],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_northing\",0],AUTHORITY[\"EPSG\",\"32632\"],AXIS[\"East");
    add_srs_wkt (p, 11, "ing\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32633, "epsg", 32633,
		      "WGS 84 / UTM zone 33N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=33 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 33N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",15],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"32633\"],AXIS[\"Eas");
    add_srs_wkt (p, 11, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32634, "epsg", 32634,
		      "WGS 84 / UTM zone 34N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=34 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 34N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",21],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"32634\"],AXIS[\"Eas");
    add_srs_wkt (p, 11, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32635, "epsg", 32635,
		      "WGS 84 / UTM zone 35N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=35 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 35N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",27],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"32635\"],AXIS[\"Eas");
    add_srs_wkt (p, 11, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32636, "epsg", 32636,
		      "WGS 84 / UTM zone 36N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=36 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 36N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",33],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"32636\"],AXIS[\"Eas");
    add_srs_wkt (p, 11, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32637, "epsg", 32637,
		      "WGS 84 / UTM zone 37N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=37 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 37N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",39],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"32637\"],AXIS[\"Eas");
    add_srs_wkt (p, 11, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32638, "epsg", 32638,
		      "WGS 84 / UTM zone 38N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=38 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 38N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",45],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"32638\"],AXIS[\"Eas");
    add_srs_wkt (p, 11, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32639, "epsg", 32639,
		      "WGS 84 / UTM zone 39N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=39 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 39N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",51],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"32639\"],AXIS[\"Eas");
    add_srs_wkt (p, 11, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32640, "epsg", 32640,
		      "WGS 84 / UTM zone 40N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=40 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 40N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",57],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"32640\"],AXIS[\"Eas");
    add_srs_wkt (p, 11, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32641, "epsg", 32641,
		      "WGS 84 / UTM zone 41N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=41 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 41N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",63],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"32641\"],AXIS[\"Eas");
    add_srs_wkt (p, 11, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32642, "epsg", 32642,
		      "WGS 84 / UTM zone 42N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=42 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 42N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",69],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"32642\"],AXIS[\"Eas");
    add_srs_wkt (p, 11, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32643, "epsg", 32643,
		      "WGS 84 / UTM zone 43N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=43 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 43N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",75],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"32643\"],AXIS[\"Eas");
    add_srs_wkt (p, 11, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32644, "epsg", 32644,
		      "WGS 84 / UTM zone 44N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=44 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 44N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",81],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"32644\"],AXIS[\"Eas");
    add_srs_wkt (p, 11, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32645, "epsg", 32645,
		      "WGS 84 / UTM zone 45N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=45 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 45N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",87],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"32645\"],AXIS[\"Eas");
    add_srs_wkt (p, 11, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32646, "epsg", 32646,
		      "WGS 84 / UTM zone 46N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=46 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 46N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",93],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"32646\"],AXIS[\"Eas");
    add_srs_wkt (p, 11, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32647, "epsg", 32647,
		      "WGS 84 / UTM zone 47N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=47 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 47N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",99],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",0],AUTHORITY[\"EPSG\",\"32647\"],AXIS[\"Eas");
    add_srs_wkt (p, 11, "ting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32648, "epsg", 32648,
		      "WGS 84 / UTM zone 48N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=48 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 48N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",105],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32648\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32649, "epsg", 32649,
		      "WGS 84 / UTM zone 49N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=49 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 49N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",111],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32649\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32650, "epsg", 32650,
		      "WGS 84 / UTM zone 50N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=50 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 50N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",117],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32650\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32651, "epsg", 32651,
		      "WGS 84 / UTM zone 51N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=51 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 51N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",123],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32651\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32652, "epsg", 32652,
		      "WGS 84 / UTM zone 52N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=52 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 52N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",129],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32652\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32653, "epsg", 32653,
		      "WGS 84 / UTM zone 53N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=53 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 53N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",135],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32653\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32654, "epsg", 32654,
		      "WGS 84 / UTM zone 54N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=54 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 54N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",141],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32654\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32655, "epsg", 32655,
		      "WGS 84 / UTM zone 55N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=55 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 55N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",147],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32655\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32656, "epsg", 32656,
		      "WGS 84 / UTM zone 56N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=56 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 56N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",153],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32656\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32657, "epsg", 32657,
		      "WGS 84 / UTM zone 57N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=57 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 57N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",159],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32657\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32658, "epsg", 32658,
		      "WGS 84 / UTM zone 58N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 58N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",165],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32658\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32659, "epsg", 32659,
		      "WGS 84 / UTM zone 59N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=59 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 59N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",171],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32659\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32660, "epsg", 32660,
		      "WGS 84 / UTM zone 60N");
    add_proj4text (p, 0,
		   "+proj=utm +zone=60 +ellps=WGS84 +datum=WGS84 +units=m +n");
    add_proj4text (p, 1, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 60N\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",177],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",0],AUTHORITY[\"EPSG\",\"32660\"],AXIS[\"Ea");
    add_srs_wkt (p, 11, "sting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32661, "epsg", 32661, "WGS 84 / UPS North");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=90 +lat_ts=90 +lon_0=0 +k=0.994 +x_0=");
    add_proj4text (p, 1,
		   "2000000 +y_0=2000000 +ellps=WGS84 +datum=WGS84 +units=m ");
    add_proj4text (p, 2, "+no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UPS North\",GEOGCS[\"WGS 84\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]]");
    add_srs_wkt (p, 3,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 4,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 5,
		 "122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,A");
    add_srs_wkt (p, 6,
		 "UTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_Stereogr");
    add_srs_wkt (p, 7,
		 "aphic\"],PARAMETER[\"latitude_of_origin\",90],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",0],PARAMETER[\"scale_factor\",0.994");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"false_easting\",2000000],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_northing\",2000000],AUTHORITY[\"EPSG\",\"32661\"],AXIS[");
    add_srs_wkt (p, 11, "\"Northing\",UNKNOWN],AXIS[\"Easting\",UNKNOWN]]");
    p = add_epsg_def (first, last, 32662, "epsg", 32662,
		      "WGS 84 / Plate Carree (deprecated)");
    add_proj4text (p, 0,
		   "+proj=eqc +lat_ts=0 +lat_0=0 +lon_0=0 +x_0=0 +y_0=0 +ell");
    add_proj4text (p, 1, "ps=WGS84 +datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / Plate Carree (deprecated)\",GEOGCS[\"W");
    add_srs_wkt (p, 1,
		 "GS 84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,2");
    add_srs_wkt (p, 2,
		 "98.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"E");
    add_srs_wkt (p, 3,
		 "PSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG");
    add_srs_wkt (p, 4,
		 "\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHOR");
    add_srs_wkt (p, 5,
		 "ITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UN");
    add_srs_wkt (p, 6,
		 "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[");
    add_srs_wkt (p, 7,
		 "\"Equirectangular\"],PARAMETER[\"latitude_of_origin\",0]");
    add_srs_wkt (p, 8,
		 ",PARAMETER[\"central_meridian\",0],PARAMETER[\"false_eas");
    add_srs_wkt (p, 9,
		 "ting\",0],PARAMETER[\"false_northing\",0],AUTHORITY[\"EP");
    add_srs_wkt (p, 10, "SG\",\"32662\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]");
    p = add_epsg_def (first, last, 32664, "epsg", 32664,
		      "WGS 84 / BLM 14N (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-99 +k=0.9996 +x_0=500000.00");
    add_proj4text (p, 1,
		   "1016002 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=us-ft +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / BLM 14N (ftUS)\",GEOGCS[\"WGS 84\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.25722356");
    add_srs_wkt (p, 2,
		 "3,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"632");
    add_srs_wkt (p, 3,
		 "6\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"US surv");
    add_srs_wkt (p, 6,
		 "ey foot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-99],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",1640416.67],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"32664\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORT");
    add_srs_wkt (p, 12, "H]]");
    p = add_epsg_def (first, last, 32665, "epsg", 32665,
		      "WGS 84 / BLM 15N (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-93 +k=0.9996 +x_0=500000.00");
    add_proj4text (p, 1,
		   "1016002 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=us-ft +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / BLM 15N (ftUS)\",GEOGCS[\"WGS 84\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.25722356");
    add_srs_wkt (p, 2,
		 "3,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"632");
    add_srs_wkt (p, 3,
		 "6\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"US surv");
    add_srs_wkt (p, 6,
		 "ey foot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-93],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",1640416.67],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"32665\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORT");
    add_srs_wkt (p, 12, "H]]");
    p = add_epsg_def (first, last, 32666, "epsg", 32666,
		      "WGS 84 / BLM 16N (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-87 +k=0.9996 +x_0=500000.00");
    add_proj4text (p, 1,
		   "1016002 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=us-ft +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / BLM 16N (ftUS)\",GEOGCS[\"WGS 84\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.25722356");
    add_srs_wkt (p, 2,
		 "3,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"632");
    add_srs_wkt (p, 3,
		 "6\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"US surv");
    add_srs_wkt (p, 6,
		 "ey foot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-87],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",1640416.67],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"32666\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORT");
    add_srs_wkt (p, 12, "H]]");
    p = add_epsg_def (first, last, 32667, "epsg", 32667,
		      "WGS 84 / BLM 17N (ftUS)");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=-81 +k=0.9996 +x_0=500000.00");
    add_proj4text (p, 1,
		   "1016002 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=us-ft +n");
    add_proj4text (p, 2, "o_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / BLM 17N (ftUS)\",GEOGCS[\"WGS 84\",DAT");
    add_srs_wkt (p, 1,
		 "UM[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.25722356");
    add_srs_wkt (p, 2,
		 "3,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"632");
    add_srs_wkt (p, 3,
		 "6\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"");
    add_srs_wkt (p, 4,
		 "]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\"");
    add_srs_wkt (p, 5,
		 ",\"9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"US surv");
    add_srs_wkt (p, 6,
		 "ey foot\",0.3048006096012192,AUTHORITY[\"EPSG\",\"9003\"");
    add_srs_wkt (p, 7,
		 "]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latit");
    add_srs_wkt (p, 8,
		 "ude_of_origin\",0],PARAMETER[\"central_meridian\",-81],P");
    add_srs_wkt (p, 9,
		 "ARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_east");
    add_srs_wkt (p, 10,
		 "ing\",1640416.67],PARAMETER[\"false_northing\",0],AUTHOR");
    add_srs_wkt (p, 11,
		 "ITY[\"EPSG\",\"32667\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORT");
    add_srs_wkt (p, 12, "H]]");
    p = add_epsg_def (first, last, 32701, "epsg", 32701,
		      "WGS 84 / UTM zone 1S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=1 +south +ellps=WGS84 +datum=WGS84 +unit");
    add_proj4text (p, 1, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 1S\",GEOGCS[\"WGS 84\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-177],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",10000000],AUTHORITY[\"EPSG\",\"32701\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32702, "epsg", 32702,
		      "WGS 84 / UTM zone 2S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=2 +south +ellps=WGS84 +datum=WGS84 +unit");
    add_proj4text (p, 1, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 2S\",GEOGCS[\"WGS 84\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-171],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",10000000],AUTHORITY[\"EPSG\",\"32702\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32703, "epsg", 32703,
		      "WGS 84 / UTM zone 3S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=3 +south +ellps=WGS84 +datum=WGS84 +unit");
    add_proj4text (p, 1, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 3S\",GEOGCS[\"WGS 84\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-165],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",10000000],AUTHORITY[\"EPSG\",\"32703\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32704, "epsg", 32704,
		      "WGS 84 / UTM zone 4S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=4 +south +ellps=WGS84 +datum=WGS84 +unit");
    add_proj4text (p, 1, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 4S\",GEOGCS[\"WGS 84\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-159],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",10000000],AUTHORITY[\"EPSG\",\"32704\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32705, "epsg", 32705,
		      "WGS 84 / UTM zone 5S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=5 +south +ellps=WGS84 +datum=WGS84 +unit");
    add_proj4text (p, 1, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 5S\",GEOGCS[\"WGS 84\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-153],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",10000000],AUTHORITY[\"EPSG\",\"32705\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32706, "epsg", 32706,
		      "WGS 84 / UTM zone 6S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=6 +south +ellps=WGS84 +datum=WGS84 +unit");
    add_proj4text (p, 1, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 6S\",GEOGCS[\"WGS 84\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-147],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",10000000],AUTHORITY[\"EPSG\",\"32706\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32707, "epsg", 32707,
		      "WGS 84 / UTM zone 7S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=7 +south +ellps=WGS84 +datum=WGS84 +unit");
    add_proj4text (p, 1, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 7S\",GEOGCS[\"WGS 84\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-141],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",10000000],AUTHORITY[\"EPSG\",\"32707\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32708, "epsg", 32708,
		      "WGS 84 / UTM zone 8S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=8 +south +ellps=WGS84 +datum=WGS84 +unit");
    add_proj4text (p, 1, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 8S\",GEOGCS[\"WGS 84\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-135],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",10000000],AUTHORITY[\"EPSG\",\"32708\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32709, "epsg", 32709,
		      "WGS 84 / UTM zone 9S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=9 +south +ellps=WGS84 +datum=WGS84 +unit");
    add_proj4text (p, 1, "s=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 9S\",GEOGCS[\"WGS 84\",DATUM[");
    add_srs_wkt (p, 1,
		 "\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,A");
    add_srs_wkt (p, 2,
		 "UTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-129],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",10000000],AUTHORITY[\"EPSG\",\"32709\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32710, "epsg", 32710,
		      "WGS 84 / UTM zone 10S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=10 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 10S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-123],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",10000000],AUTHORITY[\"EPSG\",\"32710\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32711, "epsg", 32711,
		      "WGS 84 / UTM zone 11S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=11 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 11S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-117],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",10000000],AUTHORITY[\"EPSG\",\"32711\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32712, "epsg", 32712,
		      "WGS 84 / UTM zone 12S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=12 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 12S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-111],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",10000000],AUTHORITY[\"EPSG\",\"32712\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32713, "epsg", 32713,
		      "WGS 84 / UTM zone 13S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=13 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 13S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-105],PARAMETER[\"scale_factor\",0.");
    add_srs_wkt (p, 9,
		 "9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fa");
    add_srs_wkt (p, 10,
		 "lse_northing\",10000000],AUTHORITY[\"EPSG\",\"32713\"],A");
    add_srs_wkt (p, 11, "XIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32714, "epsg", 32714,
		      "WGS 84 / UTM zone 14S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=14 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 14S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-99],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32714\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32715, "epsg", 32715,
		      "WGS 84 / UTM zone 15S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=15 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 15S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-93],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32715\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32716, "epsg", 32716,
		      "WGS 84 / UTM zone 16S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=16 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 16S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-87],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32716\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32717, "epsg", 32717,
		      "WGS 84 / UTM zone 17S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=17 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 17S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-81],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32717\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32718, "epsg", 32718,
		      "WGS 84 / UTM zone 18S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=18 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 18S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-75],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32718\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32719, "epsg", 32719,
		      "WGS 84 / UTM zone 19S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=19 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 19S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-69],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32719\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32720, "epsg", 32720,
		      "WGS 84 / UTM zone 20S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=20 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 20S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-63],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32720\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32721, "epsg", 32721,
		      "WGS 84 / UTM zone 21S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=21 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 21S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-57],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32721\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32722, "epsg", 32722,
		      "WGS 84 / UTM zone 22S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=22 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 22S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-51],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32722\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32723, "epsg", 32723,
		      "WGS 84 / UTM zone 23S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=23 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 23S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-45],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32723\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32724, "epsg", 32724,
		      "WGS 84 / UTM zone 24S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=24 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 24S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-39],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32724\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32725, "epsg", 32725,
		      "WGS 84 / UTM zone 25S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=25 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 25S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-33],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32725\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32726, "epsg", 32726,
		      "WGS 84 / UTM zone 26S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=26 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 26S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-27],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32726\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32727, "epsg", 32727,
		      "WGS 84 / UTM zone 27S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=27 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 27S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-21],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32727\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32728, "epsg", 32728,
		      "WGS 84 / UTM zone 28S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=28 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 28S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-15],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32728\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32729, "epsg", 32729,
		      "WGS 84 / UTM zone 29S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=29 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 29S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-9],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"32729\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32730, "epsg", 32730,
		      "WGS 84 / UTM zone 30S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=30 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 30S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",-3],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"32730\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32731, "epsg", 32731,
		      "WGS 84 / UTM zone 31S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=31 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 31S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",3],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 9,
		 "6],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_northing\",10000000],AUTHORITY[\"EPSG\",\"32731\"],AXIS");
    add_srs_wkt (p, 11, "[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32732, "epsg", 32732,
		      "WGS 84 / UTM zone 32S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=32 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 32S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",9],PARAMETER[\"scale_factor\",0.999");
    add_srs_wkt (p, 9,
		 "6],PARAMETER[\"false_easting\",500000],PARAMETER[\"false");
    add_srs_wkt (p, 10,
		 "_northing\",10000000],AUTHORITY[\"EPSG\",\"32732\"],AXIS");
    add_srs_wkt (p, 11, "[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32733, "epsg", 32733,
		      "WGS 84 / UTM zone 33S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=33 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 33S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",15],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"32733\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32734, "epsg", 32734,
		      "WGS 84 / UTM zone 34S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=34 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 34S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",21],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"32734\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32735, "epsg", 32735,
		      "WGS 84 / UTM zone 35S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=35 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 35S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",27],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"32735\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32736, "epsg", 32736,
		      "WGS 84 / UTM zone 36S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=36 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 36S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",33],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"32736\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32737, "epsg", 32737,
		      "WGS 84 / UTM zone 37S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=37 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 37S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",39],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"32737\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32738, "epsg", 32738,
		      "WGS 84 / UTM zone 38S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=38 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 38S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",45],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"32738\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32739, "epsg", 32739,
		      "WGS 84 / UTM zone 39S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=39 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 39S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",51],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"32739\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32740, "epsg", 32740,
		      "WGS 84 / UTM zone 40S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=40 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 40S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",57],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"32740\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32741, "epsg", 32741,
		      "WGS 84 / UTM zone 41S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=41 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 41S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",63],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"32741\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32742, "epsg", 32742,
		      "WGS 84 / UTM zone 42S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=42 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 42S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",69],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"32742\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32743, "epsg", 32743,
		      "WGS 84 / UTM zone 43S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=43 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 43S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",75],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"32743\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32744, "epsg", 32744,
		      "WGS 84 / UTM zone 44S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=44 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 44S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",81],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"32744\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32745, "epsg", 32745,
		      "WGS 84 / UTM zone 45S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=45 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 45S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",87],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"32745\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32746, "epsg", 32746,
		      "WGS 84 / UTM zone 46S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=46 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 46S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",93],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"32746\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32747, "epsg", 32747,
		      "WGS 84 / UTM zone 47S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=47 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 47S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",99],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "96],PARAMETER[\"false_easting\",500000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",10000000],AUTHORITY[\"EPSG\",\"32747\"],AXI");
    add_srs_wkt (p, 11, "S[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32748, "epsg", 32748,
		      "WGS 84 / UTM zone 48S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=48 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 48S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",105],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32748\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32749, "epsg", 32749,
		      "WGS 84 / UTM zone 49S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=49 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 49S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",111],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32749\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32750, "epsg", 32750,
		      "WGS 84 / UTM zone 50S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=50 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 50S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",117],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32750\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32751, "epsg", 32751,
		      "WGS 84 / UTM zone 51S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=51 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 51S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",123],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32751\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32752, "epsg", 32752,
		      "WGS 84 / UTM zone 52S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=52 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 52S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",129],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32752\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32753, "epsg", 32753,
		      "WGS 84 / UTM zone 53S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=53 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 53S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",135],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32753\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32754, "epsg", 32754,
		      "WGS 84 / UTM zone 54S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=54 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 54S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",141],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32754\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32755, "epsg", 32755,
		      "WGS 84 / UTM zone 55S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=55 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 55S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",147],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32755\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
}

static void
initialize_epsg_29 (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
    struct epsg_defs *p;
    p = add_epsg_def (first, last, 32756, "epsg", 32756,
		      "WGS 84 / UTM zone 56S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=56 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 56S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",153],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32756\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32757, "epsg", 32757,
		      "WGS 84 / UTM zone 57S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=57 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 57S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",159],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32757\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32758, "epsg", 32758,
		      "WGS 84 / UTM zone 58S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=58 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 58S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",165],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32758\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32759, "epsg", 32759,
		      "WGS 84 / UTM zone 59S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=59 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 59S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",171],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32759\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32760, "epsg", 32760,
		      "WGS 84 / UTM zone 60S");
    add_proj4text (p, 0,
		   "+proj=utm +zone=60 +south +ellps=WGS84 +datum=WGS84 +uni");
    add_proj4text (p, 1, "ts=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UTM zone 60S\",GEOGCS[\"WGS 84\",DATUM");
    add_srs_wkt (p, 1,
		 "[\"WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,");
    add_srs_wkt (p, 2,
		 "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"");
    add_srs_wkt (p, 3,
		 "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],");
    add_srs_wkt (p, 4,
		 "UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"");
    add_srs_wkt (p, 5,
		 "9122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,");
    add_srs_wkt (p, 6,
		 "AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Me");
    add_srs_wkt (p, 7,
		 "rcator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[");
    add_srs_wkt (p, 8,
		 "\"central_meridian\",177],PARAMETER[\"scale_factor\",0.9");
    add_srs_wkt (p, 9,
		 "996],PARAMETER[\"false_easting\",500000],PARAMETER[\"fal");
    add_srs_wkt (p, 10,
		 "se_northing\",10000000],AUTHORITY[\"EPSG\",\"32760\"],AX");
    add_srs_wkt (p, 11, "IS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
    p = add_epsg_def (first, last, 32761, "epsg", 32761, "WGS 84 / UPS South");
    add_proj4text (p, 0,
		   "+proj=stere +lat_0=-90 +lat_ts=-90 +lon_0=0 +k=0.994 +x_");
    add_proj4text (p, 1,
		   "0=2000000 +y_0=2000000 +ellps=WGS84 +datum=WGS84 +units=");
    add_proj4text (p, 2, "m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / UPS South\",GEOGCS[\"WGS 84\",DATUM[\"");
    add_srs_wkt (p, 1,
		 "WGS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUT");
    add_srs_wkt (p, 2,
		 "HORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]]");
    add_srs_wkt (p, 3,
		 ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN");
    add_srs_wkt (p, 4,
		 "IT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9");
    add_srs_wkt (p, 5,
		 "122\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,A");
    add_srs_wkt (p, 6,
		 "UTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Polar_Stereogr");
    add_srs_wkt (p, 7,
		 "aphic\"],PARAMETER[\"latitude_of_origin\",-90],PARAMETER");
    add_srs_wkt (p, 8,
		 "[\"central_meridian\",0],PARAMETER[\"scale_factor\",0.99");
    add_srs_wkt (p, 9,
		 "4],PARAMETER[\"false_easting\",2000000],PARAMETER[\"fals");
    add_srs_wkt (p, 10,
		 "e_northing\",2000000],AUTHORITY[\"EPSG\",\"32761\"],AXIS");
    add_srs_wkt (p, 11, "[\"Northing\",UNKNOWN],AXIS[\"Easting\",UNKNOWN]]");
    p = add_epsg_def (first, last, 32766, "epsg", 32766, "WGS 84 / TM 36 SE");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=36 +k=0.9996 +x_0=500000 +y_");
    add_proj4text (p, 1,
		   "0=10000000 +ellps=WGS84 +datum=WGS84 +units=m +no_defs");
    add_srs_wkt (p, 0,
		 "PROJCS[\"WGS 84 / TM 36 SE\",GEOGCS[\"WGS 84\",DATUM[\"W");
    add_srs_wkt (p, 1,
		 "GS_1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTH");
    add_srs_wkt (p, 2,
		 "ORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],");
    add_srs_wkt (p, 3,
		 "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI");
    add_srs_wkt (p, 4,
		 "T[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"91");
    add_srs_wkt (p, 5,
		 "22\"]],AUTHORITY[\"EPSG\",\"4326\"]],UNIT[\"metre\",1,AU");
    add_srs_wkt (p, 6,
		 "THORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Transverse_Merc");
    add_srs_wkt (p, 7,
		 "ator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"");
    add_srs_wkt (p, 8,
		 "central_meridian\",36],PARAMETER[\"scale_factor\",0.9996");
    add_srs_wkt (p, 9,
		 "],PARAMETER[\"false_easting\",500000],PARAMETER[\"false_");
    add_srs_wkt (p, 10,
		 "northing\",10000000],AUTHORITY[\"EPSG\",\"32766\"],AXIS[");
    add_srs_wkt (p, 11, "\"Easting\",EAST],AXIS[\"Northing\",NORTH]]");
#if OMIT_EPSG == 0		/* resuming conditional EPSG initialization */

    p = add_epsg_def (first, last, 40000, "gfoss.it", 40000,
		      "Italy mainland zone 1 GB Roma40");
    add_proj4text (p, 0,
		   "+proj=tmerc+lat_0=0 +lon_0=9  +k=0.9996 +x_0=1500000 +y_");
    add_proj4text (p, 1,
		   "0=0 +ellps=intl +units=m +towgs84=-104.1,-49.1,-9.9,0.97");
    add_proj4text (p, 2, "1,-2.917,0.714,-11.68 +no_defs");
    add_srs_wkt (p, 0, "");
    p = add_epsg_def (first, last, 40001, "gfoss.it", 40001,
		      "Italy mainland zone 2 GB Roma40");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=15 +k=0.9996 +x_0=2520000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=intl +units=m +towgs84=-104.1,-49.1,-9.9,0.9");
    add_proj4text (p, 2, "71,-2.917,0.714,-11.68 +no_defs");
    add_srs_wkt (p, 0, "");
    p = add_epsg_def (first, last, 40002, "gfoss.it", 40002,
		      "Italy Sardinia GB Roma40");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=9  +k=0.9996 +x_0=1500000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=intl +units=m +towgs84=-168.6,-34.0,38.6,-0.");
    add_proj4text (p, 2, "374,-0.679,-1.379,-9.48 +no_defs");
    add_srs_wkt (p, 0, "");
    p = add_epsg_def (first, last, 40003, "gfoss.it", 40003,
		      "Italy Sicily GB Roma40");
    add_proj4text (p, 0,
		   "+proj=tmerc +lat_0=0 +lon_0=9  +k=0.9996 +x_0=1500000 +y");
    add_proj4text (p, 1,
		   "_0=0 +ellps=intl +units=m +towgs84=-50.2,-50.4,84.8,-0.6");
    add_proj4text (p, 2, "90,-2.012,0.459,-28.08  +no_defs");
    add_srs_wkt (p, 0, "");
#endif /* full EPSG initialization enabled/disabled */
}

static void
initialize_epsg (struct epsg_defs **first, struct epsg_defs **last)
{
/* initializing the EPSG defs list */
    initialize_epsg_00 (first, last);
    initialize_epsg_01 (first, last);
    initialize_epsg_02 (first, last);
    initialize_epsg_03 (first, last);
    initialize_epsg_04 (first, last);
    initialize_epsg_05 (first, last);
    initialize_epsg_06 (first, last);
    initialize_epsg_07 (first, last);
    initialize_epsg_08 (first, last);
    initialize_epsg_09 (first, last);
    initialize_epsg_10 (first, last);
    initialize_epsg_11 (first, last);
    initialize_epsg_12 (first, last);
    initialize_epsg_13 (first, last);
    initialize_epsg_14 (first, last);
    initialize_epsg_15 (first, last);
    initialize_epsg_16 (first, last);
    initialize_epsg_17 (first, last);
    initialize_epsg_18 (first, last);
    initialize_epsg_19 (first, last);
    initialize_epsg_20 (first, last);
    initialize_epsg_21 (first, last);
    initialize_epsg_22 (first, last);
    initialize_epsg_23 (first, last);
    initialize_epsg_24 (first, last);
    initialize_epsg_25 (first, last);
    initialize_epsg_26 (first, last);
    initialize_epsg_27 (first, last);
    initialize_epsg_28 (first, last);
    initialize_epsg_29 (first, last);
}

static void
free_epsg (struct epsg_defs *first)
{
/* memory cleanup - destroying the EPSG list */
    struct epsg_defs *p = first;
    struct epsg_defs *pn;
    while (p)
      {
	  pn = p->next;
	  free_epsg_def (p);
	  p = pn;
      }
}

static int
populate_spatial_ref_sys (sqlite3 * handle)
{
/* populating the EPSG dataset into the SPATIAL_REF_SYS table */
    struct epsg_defs *first = NULL;
    struct epsg_defs *last = NULL;
    struct epsg_defs *p;
    char sql[1024];
    char *errMsg = NULL;
    int ret;
    sqlite3_stmt *stmt;

/* initializing the EPSG defs list */
    initialize_epsg (&first, &last);

/* starting a transaction */
    ret = sqlite3_exec (handle, "BEGIN", NULL, 0, &errMsg);
    if (ret != SQLITE_OK)
      {
	  fprintf (stderr, "%s\n", errMsg);
	  sqlite3_free (errMsg);
	  goto error;
      }

/* preparing the SQL parameterized statement */
    strcpy (sql, "INSERT INTO spatial_ref_sys ");
    strcat (sql,
	    "(srid, auth_name, auth_srid, ref_sys_name, proj4text, srs_wkt) ");
    strcat (sql, "VALUES (?, ?, ?, ?, ?, ?)");
    ret = sqlite3_prepare_v2 (handle, sql, strlen (sql), &stmt, NULL);
    if (ret != SQLITE_OK)
      {
	  fprintf (stderr, "%s\n", sqlite3_errmsg (handle));
	  goto error;
      }
    p = first;
    while (p)
      {
	  if (p->srid < 0 || p->auth_name == NULL)
	      break;
	  sqlite3_reset (stmt);
	  sqlite3_clear_bindings (stmt);
	  sqlite3_bind_int (stmt, 1, p->srid);
	  sqlite3_bind_text (stmt, 2, p->auth_name, strlen (p->auth_name),
			     SQLITE_STATIC);
	  sqlite3_bind_int (stmt, 3, p->auth_srid);
	  sqlite3_bind_text (stmt, 4, p->ref_sys_name, strlen (p->ref_sys_name),
			     SQLITE_STATIC);
	  sqlite3_bind_text (stmt, 5, p->proj4text, strlen (p->proj4text),
			     SQLITE_STATIC);
	  if (strlen (p->srs_wkt) == 0)
	      sqlite3_bind_null (stmt, 6);
	  else
	      sqlite3_bind_text (stmt, 6, p->srs_wkt, strlen (p->srs_wkt),
				 SQLITE_STATIC);
	  ret = sqlite3_step (stmt);
	  if (ret == SQLITE_DONE || ret == SQLITE_ROW)
	      ;
	  else
	    {
		fprintf (stderr, "%s\n", sqlite3_errmsg (handle));
		sqlite3_finalize (stmt);
		goto error;
	    }
	  p = p->next;
      }
    sqlite3_finalize (stmt);

/* confirming the transaction */
    ret = sqlite3_exec (handle, "COMMIT", NULL, 0, &errMsg);
    if (ret != SQLITE_OK)
      {
	  fprintf (stderr, "%s\n", errMsg);
	  sqlite3_free (errMsg);
	  goto error;
      }

/* freeing the EPSG defs list */
    free_epsg (first);

    return 1;
  error:
/* trying to perform a ROLLBACK anyway */
    ret = sqlite3_exec (handle, "ROLLBACK", NULL, 0, &errMsg);
    if (ret != SQLITE_OK)
      {
	  fprintf (stderr, "%s\n", errMsg);
	  sqlite3_free (errMsg);
      }

/* freeing the EPSG defs list */
    free_epsg (first);

    return 0;
}

static int
exists_spatial_ref_sys (sqlite3 * handle)
{
/* checking if the SPATIAL_REF_SYS table exists */
    int ret;
    int ok = 0;
    char sql[1024];
    char **results;
    int n_rows;
    int n_columns;
    char *err_msg = NULL;

    strcpy (sql,
	    "SELECT name FROM sqlite_master WHERE type = 'table' AND name LIKE 'spatial_ref_sys'");
    ret =
	sqlite3_get_table (handle, sql, &results, &n_rows, &n_columns,
			   &err_msg);
    if (ret != SQLITE_OK)
      {
/* some error occurred */
	  fprintf (stderr, "XX %s\n", err_msg);
	  sqlite3_free (err_msg);
	  return 0;
      }
    if (n_rows > 0)
	ok = 1;
    sqlite3_free_table (results);
    return ok;
}

static int
check_spatial_ref_sys (sqlite3 * handle)
{
/* checking if the SPATIAL_REF_SYS table has an appropriate layout */
    int ret;
    int i;
    const char *name;
    char sql[1024];
    char **results;
    int n_rows;
    int n_columns;
    char *err_msg = NULL;
    int rs_srid = 0;
    int auth_name = 0;
    int auth_srid = 0;
    int srtext = 0;
    int ref_sys_name = 0;
    int proj4text = 0;
    int srs_wkt = 0;

    strcpy (sql, "PRAGMA table_info(spatial_ref_sys)");
    ret =
	sqlite3_get_table (handle, sql, &results, &n_rows, &n_columns,
			   &err_msg);
    if (ret != SQLITE_OK)
      {
/* some error occurred */
	  fprintf (stderr, "%s\n", err_msg);
	  sqlite3_free (err_msg);
	  return 0;
      }
    if (n_rows > 0)
      {
	  for (i = 1; i <= n_rows; i++)
	    {
		name = results[(i * n_columns) + 1];
		if (strcasecmp (name, "srid") == 0)
		    rs_srid = 1;
		if (strcasecmp (name, "auth_name") == 0)
		    auth_name = 1;
		if (strcasecmp (name, "auth_srid") == 0)
		    auth_srid = 1;
		if (strcasecmp (name, "srtext") == 0)
		    srtext = 1;
		if (strcasecmp (name, "ref_sys_name") == 0)
		    ref_sys_name = 1;
		if (strcasecmp (name, "proj4text") == 0)
		    proj4text = 1;
		if (strcasecmp (name, "srs_wkt") == 0)
		    srs_wkt = 1;
	    }
      }
    sqlite3_free_table (results);
    if (rs_srid && auth_name && auth_srid && ref_sys_name && proj4text
	&& srs_wkt)
	ret = 1;
    else
	ret = 0;
    return ret;
}

static int
spatial_ref_sys_count (sqlite3 * handle)
{
/* checking if the SPATIAL_REF_SYS table is empty */
    int ret;
    int i;
    int count = 0;
    char sql[1024];
    char **results;
    int n_rows;
    int n_columns;
    char *err_msg = NULL;

    strcpy (sql, "SELECT Count(*) FROM spatial_ref_sys");
    ret =
	sqlite3_get_table (handle, sql, &results, &n_rows, &n_columns,
			   &err_msg);
    if (ret != SQLITE_OK)
      {
/* some error occurred */
	  fprintf (stderr, "%s\n", err_msg);
	  sqlite3_free (err_msg);
	  return 0;
      }
    if (n_rows > 0)
      {
	  for (i = 1; i <= n_rows; i++)
	    {
		count = atoi (results[(i * n_columns) + 0]);
	    }
      }
    sqlite3_free_table (results);
    return count;
}

SPATIALITE_DECLARE void
spatial_ref_sys_init (sqlite3 * handle, int verbose)
{
/* populating the EPSG dataset into the SPATIAL_REF_SYS table */
    if (!exists_spatial_ref_sys (handle))
      {
	  if (verbose)
	      fprintf (stderr, "the SPATIAL_REF_SYS table doesn't exists\n");
	  return;
      }
    if (!check_spatial_ref_sys (handle))
      {
	  if (verbose)
	      fprintf (stderr,
		       "the SPATIAL_REF_SYS table has an unsupported layout\n");
	  return;
      }
    if (spatial_ref_sys_count (handle))
      {
	  if (verbose)
	      fprintf (stderr,
		       "the SPATIAL_REF_SYS table already contains some row(s)\n");
	  return;
      }
    if (populate_spatial_ref_sys (handle))
      {
	  if (verbose)
	      fprintf (stderr,
		       "OK: the SPATIAL_REF_SYS table was succesfully populated\n");
	  return;
      }
}
